/*
 * PIDB.c: An interface for SML
 */

#include <stdio.h>
#include <sys/file.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <Sapid/Sapid.h>
#include <Sapid/PIDB.h>

#ifndef PATH_MAX
#define PATH_MAX 1023
#endif

void 	find_macro(char *, int, int);
int 	macro_exists(char *, int, int);
void 	agg_macro_init(char *, int, int);
char	get_a_macro(void);
int  	no_macro(void);
char	*name(void);
int	narg(void);
char	*call_fname(void);
int	call_start(void);
int	call_end(void);

DEFINEinfo	*get_defineInfo(void);
int		get_defineInfoN(void);
EXPANDinfo	*get_expandInfo(void);
int   		get_expandInfoN();
   
void	set_info(DEFINEinfo *, int, EXPANDinfo *, int);

void	conv_I_to_C(char *, int, char *, int);
long	get_C_offset(void);
char	*get_C_filename(void);

static char 	*macro_name;
static int 	exp_id;
static int   	macro_narg;
static int	region_start;
static int	region_end;

void 	find_macro(char *filename, int offset, int length)
{
    MacroExpandRegion	target;
    int			n;
    int			argn;

    target.type = EXPAND_FILE;
    target.start = offset;
    target.end = offset + length - 1;
    target.fname = filename;
#ifdef PIDB_USE_FILENAME_INDEX
    target.fname_idx = pidbConvFileId(filename);
#endif

    if ((exp_id = isInMacroExpand(&target, -1, -1)) < 0) {
	macro_name = NULL;
    } else {
#if 0
	macro_name = expandInfo[exp_id].mac->name;
	macro_narg = expandInfo[exp_id].mac->narg;
#else
	if ((argn = isInMacroExpandArg(exp_id, &target, &n)) < 0) {
	    macro_name = expandInfo[exp_id].mac->name;
	    macro_narg = expandInfo[exp_id].mac->narg;
	    region_start = expandInfo[exp_id].input.start;
	    region_end = expandInfo[exp_id].input.end;
	} else {
	    macro_name = "";
	    region_start = expandInfo[exp_id].arg[argn].start
	    	+ expandInfo[exp_id].arg[argn].trailing_space;
	    region_end = expandInfo[exp_id].arg[argn].end
	    	- expandInfo[exp_id].arg[argn].following_space;
	}
#endif
    }
}

int 	macro_exists(char *filename, int offset, int length)
{
    MacroExpandRegion	target;
    MacroCallRegion	result;
    int			exp_id;
    int			n;

    target.type = EXPAND_FILE;
    target.start = offset;
    target.end = offset + length - 1;
    target.fname = filename;
#ifdef PIDB_USE_FILENAME_INDEX
    target.fname_idx = pidbConvFileId(filename);
#endif

#if 1
    return isInMacroExpand(&target, -1, -1) >= 0;
#else
    return (((exp_id = isInMacroExpand(&target, -1, -1)) >= 0)
	    && (isInMacroExpandArg(exp_id, &target, &n) < 0));
#endif
}


static int		agg_ind;
static EXPANDinfo	*agg_p;
static int		agg_start;
static int		agg_end;

void 	agg_macro_init(char *filename, int offset, int length)
{
    char	*p;

    agg_ind = 0;
    agg_p = expandInfo;
    agg_start = offset;
    agg_end = offset + length - 1;
}

char	get_a_macro(void)
{
    for (; agg_ind < expandInfoN; ++agg_ind, ++agg_p) {
	if (agg_start <= agg_p->output.start && agg_p->output.end <= agg_end) {
	    macro_name = agg_p->mac->name;
	    macro_narg = agg_p->mac->narg;
	    ++agg_ind;
	    ++agg_p;
	    return;
	}
    }
    macro_name = NULL;
}

int  	no_macro(void)
{
    return macro_name == NULL;
}

char	*name(void)
{
    return macro_name;
}

int	narg(void)
{
    return macro_narg;
}

char	*call_fname(void)
{
    return expandInfo[exp_id].input.fname;
}

int	call_start(void)
{
#if 0
    return expandInfo[exp_id].input.start;
#else
    return region_start;
#endif
}

int	call_end(void)
{
#if 0
    return expandInfo[exp_id].input.end;
#else
    return region_end;
#endif
}


DEFINEinfo	*get_defineInfo(void)
{
    return defineInfo;
}

int		get_defineInfoN(void)
{
    return defineInfoN;
}

EXPANDinfo	*get_expandInfo(void)
{
    return expandInfo;
}

int   		get_expandInfoN()
{
    return expandInfoN;
}
   
void	set_info(DEFINEinfo *def, int defN, EXPANDinfo *exp, int expN)
{
    defineInfo = def;
    defineInfoN = defN;
    expandInfo = exp;
    expandInfoN = expN;
}

static	int	C_offset;
static	char	C_filename[PATH_MAX + 1];

void	conv_I_to_C(char *filename, int offset, char *ibuf, int ibuf_len)
{
    C_offset = cppCharConvItoC(filename, offset, ibuf, ibuf_len,
			       C_filename);

    if (C_offset < 0) {		/* No corresponding character */
	C_filename[0] = '\0';
    }
}

long	get_C_offset(void)
{
    return C_offset;
}

char	*get_C_filename(void)
{
    return C_filename;
}

void	spdSetReadSDBformat_binary(void)
{
    spdSetReadSDBformat(SAPID_BINARY_FORMAT_SDB);
}
