-- expr_type.sml: Make relations between expressions and types
-- Author: Yoshida Atsushi <yoshida@tutkie.tut.ac.jp>
-- Copyright (C) 1995, Yoshida Atsushi
-- $Id: mkExprType.sml,v 1.1.1.1 2001/05/14 06:09:23 yoshida Exp $
--
-- COPYRIGHT NOTICE: This code is provided WITHOUT ANY WARRANTY
-- and is subject to the terms of the SATHER LIBRARY GENERAL PUBLIC
-- LICENSE contained in the file: Sather/Doc/License of the
-- Sather distribution. The license is also available from ICSI,
-- 1947 Center St., Suite 600, Berkeley CA 94704, USA.
-------------------------------------------------------------------
class EXPR_TYPE is
   -- ٤Ƥμη׻SDB ؽ񤭽Ф

   db_file: STR is
      return SDB_DEF::sdb_i_model_path + "expr_type";
   end;  -- db_file

   shared cmd: STR := "mkExprType1";

   -- configuration

   --  Empty expression is void type?
   const empty_expr_is_void: BOOL := true;

   -- end of configuration

   main(args: ARRAY{STR}): INT is 
      protect
	 args := SML_INIT::analyze_args(args);
	 args := analyze_args(args);

	 if args.size /= 1 and args.size /= 2 then
	    print_usage;
	    return 1;
	 end; -- if

	 outf: FILE;
	 if args.size = 1 then
	    -- ä˽ꤵʤ ./SDB β expr_type ˽񤭹
	    outf := FILE::open_for_write(db_file);
	    if outf.error then
	       #ERR + "Can't open file: " + db_file + "\n";
	       return 2;
	    end; -- if
	 elsif args[1] = "-" then
	    -- ɸ
	    outf := FILE::stdout;
	 else
	    -- ꤵ줿
	    outf := FILE::open_for_write(args[1]);
	    if outf.error then
	       #ERR + "Can't open file: " + args[1] + "\n";
	       return 3;
	    end; -- if
	 end;

	 SML_INIT::all;

	 -- SDBν񼰤˹碌ϥȥ꡼ 
	 stream: $DB_FILE;
	 if SDB_DEF::sdb_format_is_text then
	    stream := #DB_TEXT_FILE(outf);
	 elsif SDB_DEF::sdb_format_is_binary then
	    stream := #DB_BIN_FILE(outf);
	 else
	    raise "Unknown SDB format is implemented.";
	 end; -- if 

	 loop
	    e ::= {e: EXPR}.elt!;
	    t ::= guess_type(e);
	    if ~void(t) then
	       stream.add(e.id, t.id);
	    end;  -- if
	 end; -- loop
	 stream.close;
	 SAPID_INFO::set("expr_type", stream.size);
	 SAPID_INFO::flush;
	 return 0;
      when $STR then
	 #ERR + "mkExprType: Error: " + exception.str + "\n";
	 return 1;
      else
	 #ERR + "mkExprType: Error: Unknown Internal Error\n";
	 return 1;
      end;
   end; -- main

   private print_usage is
      #ERR + "Usage: " + cmd
	    + " [-h] [-sdbd SDB_DIR] [-d SDB_DIR] [output_filename]\n";
   end; -- print_usage

   private print_warn(msg: STR) is
      #ERR +  cmd + ": Warning: " + msg + "\n";
   end; -- print_warn

   analyze_args(args: ARRAY{STR}): ARRAY{STR} is
      largs ::= #LIST{STR};
      i ::= 0; loop while!(args.has_ind(i));
	 a ::= args[i];
	 case a 
	 when "-d" then
	    i := i + 1;
	    if ~args.has_ind(i) then
	       raise "No SDB directory is specified.";
	    end;  -- if
	    SDB_DEF::sdb_file_path := args[i];
	 when "-h" then
	    print_usage;
	    UNIX::exit(1);
	 else
	    if a.is_prefix("-") then
	       print_usage;
	       UNIX::exit(1);
	    end;  -- if
	    largs := largs.append(a);
	 end;  -- case
	 i := i + 1;
      end;  -- loop
      return ARRAY{STR}::create_from(largs);
   end; -- analyze_args

   private shared type_tbl: FMAP{$SML_CPNT, A_TYPE};

   guess_type(expr: $EXPR): A_TYPE is
      t: A_TYPE;
      if type_tbl.test(expr) then
	 -- #OUT + "guess_type:(" + type_tbl.get(expr) +  ") " + expr + "\n";
	 return  type_tbl.get(expr);
      end; -- if
      typecase expr
      when EXPR then
	 case expr.sort
	 when "EMPTY" then
	    if empty_expr_is_void then
	       t := void_type;
	    else
	       t := void;
	    end;  -- if
	 when "BREAK", "CONTINUE", "GOTO" then
	    t := void;
	 when "RETURN" then
	    re ::= expr.exprs.first_elt;
	    if ~void(re) then
--	       t := guess_type(re);
	       dummy ::= guess_type(re);
	    end; -- if
	 when "COMMA" then
	    loop
	       e ::= expr.ordered_exprs.elt!;
	       t := guess_type(e);
	    end; -- loop
	 when "FUNCCALL" then
	    es ::= expr.ordered_exprs;
	    func_t ::= guess_type(es[0]);
	    if void(func_t) then
	       -- May be pointer of function. sdbin has never supported.
	       t := void;
	    else
	       t := pointed_type(func_t);
	       if es.size > 1 then
		  loop
		     e ::= es.elt!(1);
		     dummy ::= guess_type(e);
		  end; -- loop
	       end; -- if
	    end; -- if
	 when "INITLIST" then
	    t := void;		-- Which type?
--	 when "ARITHMETIC", "CONSTANT" then
	 when "ARITHMETIC" then
	    op ::= expr.op;
	    if void(op) then
	       -- may be pointer of function.
	       type_tbl := type_tbl.insert(expr, void);
	       return void;
	    end;
	    case op.sort
	    when "MULTIPLY", "DIVIDE", "MODULO", "ADD", "SUBTRACT",
		     "RIGHT_SHIFT", "LEFT_SHIFT" then
	       t0 ::= guess_type(expr.ordered_exprs[0]);
	       t1 ::= guess_type(expr.ordered_exprs[1]);
	       protect
		  t := unify(t0, t1);
	       when UNIFICATION_ERROR then
		  print_warn(exception.str);
		  t := void;
	       end;  -- protect
	    when "EQUAL", "NOT_EQUAL", "BINARY_AND", "BINARY_XOR",
		     "BINARY_OR", "LOGICAL_AND", "LOGICAL_OR",
		     "LESS_THAN", "LESS_EQUAL", "GREATER_THAN", "GREATER_EQUAL"
	    then
	       dummy ::= guess_type(expr.ordered_exprs[0]);
	       dummy := guess_type(expr.ordered_exprs[1]);
	       t := int_type;
	    when "NOT" then
	       dummy ::= guess_type(expr.exprs.first_elt);
	       t := int_type;
	    when "ASSIGN" then
	       dummy ::= guess_type(expr.right_expr);
	       t := guess_type(expr.left_expr);
	    when "ASSIGN_ADD", "ASSIGN_SUB", "ASSIGN_MUL", "ASSIGN_DIV",
		     "ASSIGN_MOD", "ASSIGN_AND", "ASSIGN_XOR", "ASSIGN_OR",
		     "ASSIGN_LEFT_SHIFT", "ASSIGN_RIHGT_SHIFT" then
	       t := guess_type(expr.left_expr);
	       dummy ::= guess_type(expr.right_expr);
	    when "SIZEOF" then
	       dummy ::= guess_type(expr.exprs.first_elt);
	       -- t := int_type;
	       t := unsigned_int_type;
	    when "SIZEOF_TYPE" then
	       t := unsigned_int_type;
	    when "CAST" then
	       t := op.a_type;
	    when "LEFT_ARRAY", "RIGHT_ARRAY" then
	       child_t ::= guess_type(expr.ordered_exprs[0]);
	       t := pointed_type(child_t);
	       dummy ::= guess_type(expr.ordered_exprs[1]);
	    when "ADDRESS" then
	       child_t ::= guess_type(expr.exprs.first_elt);
	       loop
		  pt ::= child_t.pointer_type.elt!;
		  if pt.sort = "POINTER" then
		     t := pt;
		     break!;
		  end; -- if
	       end; -- loop
	       if void(t) then
		  print_warn("No type object for addressed expr(id:" 
				     + expr.id + ") exists in SDB.");
	       end; -- if
	    when "POINTER" then
	       child_t ::= guess_type(expr.exprs.first_elt);
	       t := pointed_type(child_t);
	    when "QUESTION", "COLON" then  -- was "TRI"
	       cond ::= guess_type(expr.ordered_exprs[0]);
	       t0 ::= guess_type(expr.ordered_exprs[1]);
	       t1 ::= guess_type(expr.ordered_exprs[2]);
	       protect
		  t := unify(t0, t1);
	       when UNIFICATION_ERROR then
		  print_warn("Illegal typed tri expr(id:" + expr.id 
			     + ")(type:" + cond.id + " ? " 
			     + t0.id + " : " + t1.id +")");
		  t := void;
	       end;  -- protect
	    else
	       -- when "DOT"
	       t := guess_type(expr.exprs.first_elt);
	    end; -- case
	 else
	    print_warn("Unknown type of expression["
		       + expr.id + "] " + expr.sort);
	 end; -- case
      when IDENT_EXPR then
	 t := expr.ident.a_type;
      when CONST_EXPR then
	 cons ::= expr.cons;
	 case cons.a_type
	 when "INT" then
	    t := int_type(cons.str);
	 when "FLOAT" then
	    t := double_type(cons.str);
	 when "CHAR" then
	    --	    t := char_type(cons.str);
	    t := int_type(cons.str);
	 when "STRING" then
	    t := string_type(cons.str);
	 else
	    print_warn("Unknown type of constant [" + cons.id + "]");
	 end; -- case
      end; -- if
      -- #OUT + "guess_type:(" + t +  ") " + expr + "\n";
      type_tbl := type_tbl.insert(expr, t);
--      if void(t) then
--	 raise #GUESS_ERROR("No type object exists for expr: " + expr.id);
--      end;  -- if
      return t;
   end; -- guess_type

   unify(type0, type1: A_TYPE): A_TYPE is
      if void(type0) then
	 return type1;
      elsif void(type1) then
	 return type0
      end;
      t0 ::= type0;
      t1 ::= type1;
      t: A_TYPE;
      loop while!(t0.sort = "TYPE_DEF");
	 t0 := t0.ident.a_type;
      end; -- loop
      loop while!(t1.sort = "TYPE_DEF");
	 t1 := t1.ident.a_type;
      end; -- loop
      --      if type0.sort = "TYPE_DEF" then
      --	 t0 := type0.ident.a_type;
      --      end; -- if
      --      if type1.sort = "TYPE_DEF" then
      --	 t1 := type1.ident.a_type;
      --      end; -- if

      if t0 = t1 then
	 t := t0;
      elsif (t0.sort = "POINTER" or t0.sort = "ARRAY")
	 and (t1.primitive_type = "INT" or t1.primitive_type = "CHAR"
	      or t1.primitive_type = "NONE") then
	 t := t0;
      elsif (t1.sort = "POINTER" or t1.sort = "ARRAY")
	 and (t0.primitive_type = "INT" or t0.primitive_type = "CHAR"
	      or t1.primitive_type = "NONE") then
	 t := t1;
      elsif (t0.sort = "POINTER" or t0.sort = "ARRAY")
	 and t1.sort = "POINTER" then
	 t := int_type;
      elsif t0.sort /= t1.sort then
	 raise #UNIFICATION_ERROR(t0, t1);
      elsif t0.primitive_type = "DOUBLE" and t0.size_spec = "LONG" then
	 t := t0;
      elsif t1.primitive_type = "DOUBLE" and t1.size_spec = "LONG" then
	 t := t1;
      elsif t0.primitive_type = "DOUBLE" then
	 t := t0;
      elsif t1.primitive_type = "DOUBLE" then
	 t := t1;
      elsif t0.primitive_type = "FLOAT" then
	 t := t0;
      elsif t1.primitive_type = "FLOAT" then
	 t := t1;
      elsif (t0.primitive_type = "INT" or t0.primitive_type = "NONE")
	 and t0.size_spec = "LONG" and t0.sign = "UNSIGNED" then
	 t :=  t0;
      elsif (t1.primitive_type = "INT" or t0.primitive_type = "NONE")
	 and t1.size_spec = "LONG" and t1.sign = "UNSIGNED" then
	 t :=  t1;
      elsif (t0.primitive_type = "INT"  or t0.primitive_type = "NONE")
	 and t0.sign = "UNSIGNED" then
	 -- I believe UNSIGNED int + LONG int = UNSIGNED int.
	 -- If this is not true, I must check LONG before UNSIGNED.
	 t :=  t0;
      elsif (t1.primitive_type = "INT"  or t1.primitive_type = "NONE")
	 and t1.sign = "UNSIGNED" then
	 t :=  t1;
      elsif (t0.primitive_type = "INT" or t0.primitive_type = "NONE")
	 and t0.size_spec = "LONG" then
	 t :=  t0;
      elsif (t1.primitive_type = "INT"  or t1.primitive_type = "NONE")
	 and t1.size_spec = "LONG" then
	 t :=  t1;
      else
	 t := int_type;
      end; -- if
      return t;
   end; -- if
      

   -- ܷ
   --
   -- ϡSDB  signed/unsigned, long ʤɤξѰդƤ
   -- ٤ȤǤ뤬ޤ̤бʤΤǡmkExprType Ǥ
   -- б롣ͼΤ signed/unsigned, long ζ̤
   -- դʤ(ѥ˰¸뤿)
   
   int_type(rep: STR): A_TYPE is
      -- suffix
      case rep[rep.length - 1]
      when 'U', 'u' then
	 return unsigned_int_type;
      when 'L', 'l' then
	 return long_int_type;
      else end; -- case

      return int_type;
   end; -- int_type

   private shared saved_int_type: A_TYPE;
   private shared no_int_type: BOOL;
   int_type: A_TYPE is
      if no_int_type then return void; end;
      if ~void(saved_int_type) then return saved_int_type; end;
      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "INT" and t.size_spec = "NONE"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_int_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_int_type := true;
      print_warn("No \"int\" type exists in SDB.");
      return void;
   end; -- int_type

   private shared saved_unsigned_int_type: A_TYPE;
   private shared no_unsigned_int_type: BOOL;
   unsigned_int_type: A_TYPE is
      if no_unsigned_int_type then return void; end;
      if ~void(saved_unsigned_int_type) then
	 return saved_unsigned_int_type;
      end; -- if

      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "INT" and t.size_spec = "NONE"
		  and t.qualifier = "NONE" and t.sign = "UNSIGNED" then
	    saved_unsigned_int_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_unsigned_int_type := true;
      print_warn("No \"unsigned int\" type exists in SDB.");
      return void;
   end; -- unsigned_int_type

   private shared saved_long_int_type: A_TYPE;
   private shared no_long_int_type: BOOL;
   long_int_type: A_TYPE is
      if no_long_int_type then return void; end;
      if ~void(saved_long_int_type) then
	 return saved_long_int_type;
      end; -- if

      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "INT" and t.size_spec = "LONG"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_int_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_long_int_type := true;
      print_warn("No \"long int\" type exists in SDB.");
      return void;
   end; -- long_int_type

   char_type(rep: STR): A_TYPE is
      -- ĥ줿ʸ L'x' ϰäƤʤ
      return char_type;
   end; -- char_type
   
   private shared saved_char_type: A_TYPE;
   private shared no_char_type: BOOL;
   char_type: A_TYPE is
      if no_char_type then return void; end;
      if ~void(saved_char_type) then return saved_char_type; end;
      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "CHAR" and t.size_spec = "NONE"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_char_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_char_type := true;
      print_warn("No \"char\" type exists in SDB.");
      return void;
   end; -- if

   private shared saved_float_type: A_TYPE;
   private shared no_float_type: BOOL;
   float_type: A_TYPE is
      if no_float_type then return void; end;
      if ~void(saved_float_type) then return saved_float_type; end;
      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "FLOAT" and t.size_spec = "NONE"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_float_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_float_type := true;
      print_warn("No \"float\" type exists in SDB.");
      return void;
   end; -- if

   private shared saved_double_type: A_TYPE;
   private shared no_double_type: BOOL;
   double_type(rep: STR): A_TYPE is
      case rep[rep.length - 1]
      when 'F', 'f' then
	 return float_type;
      when 'L', 'l' then
	 return long_double_type;
      else end; -- case
      if no_double_type then return void; end;
      if ~void(saved_double_type) then return saved_double_type; end;
      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "DOUBLE" and t.size_spec = "NONE"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_double_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_double_type := true;
      print_warn("No \"double\" type exists in SDB.");
      return void;
   end; -- if

   private shared saved_long_double_type: A_TYPE;
   private shared no_long_double_type: BOOL;
   long_double_type: A_TYPE is
      if no_long_double_type then return void; end;
      if ~void(saved_long_double_type) then
	 return saved_long_double_type;
      end; -- if

      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "DOUBLE" and t.size_spec = "LONG"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_long_double_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_long_double_type := true;
      print_warn("No \"long double\" type exists in SDB.");
      return void;
   end; -- if

   private shared saved_string_type: A_TYPE;
   private shared no_string_type: BOOL;
   string_type(rep: STR): A_TYPE is
      -- wide-character ʸ L"..." ϰʤ
      
      if no_string_type then return void; end;
      if ~void(saved_string_type) then return saved_string_type; end;
      if ~void(char_type) then
	 pt ::= char_type.pointer_type;
	 if ~void(pt) then
	    loop
	       -- t ::= char_type.pointer_type.elt!;
	       t ::= pt.elt!;
	       if t.sort = "POINTER" then
		  saved_string_type := t;
		  return t;
	       end; -- if
	    end; -- loop
	 end; -- if
      end;  -- if
      no_string_type := true;
      print_warn("No \"char*\" type exists in SDB.");
      return void;
   end; -- if

   private shared saved_void_type: A_TYPE;
   private shared no_void_type: BOOL;
   void_type: A_TYPE is
      if no_void_type then return void; end;
      if ~void(saved_void_type) then return saved_void_type; end;
      loop
	 t ::= A_TYPE::all.elt!;
	 if t.primitive_type = "VOID" and t.size_spec = "NONE"
		  and t.qualifier = "NONE" and t.sign = "NONE" then
	    saved_void_type := t;
	    return t;
	 end; -- if
      end; -- loop
      no_void_type := true;
      print_warn("No \"void\" type exists in SDB.");
      return void;
   end; -- if

   private pointed_type(t: A_TYPE): A_TYPE is
      if void(t) then
	 return void;
      end;  -- if
      if t.sort = "TYPE_DEF" then
	 return t.ident.a_type.pointed_type;
      else
	 return t.pointed_type;
      end; -- if
   end; -- pointed_type

end; -- class EXPR_TYPE
-------------------------------------------------------------------
class UNIFICATION_ERROR < $STR is
   attr type0: A_TYPE;
   attr type1: A_TYPE;

   create(t0: A_TYPE, t1: A_TYPE): SAME is
      r ::= new;
      r.type0 := t0;
      r.type1 := t1;
      return r;
   end;  -- create

   str: STR is
      return "UNIFICATION_ERROR: " + type0 + "<" + type0.id + ">, " 
	    + type1 + "<" + type1.id + ">";
   end;  -- str
end; -- class UNIFICATION_ERROR
-------------------------------------------------------------------
class GUESS_ERROR is
   attr msg: STR;

   create: SAME is
      return new;
   end;  -- create

   create(m: STR): SAME is
      r ::= new;
      r.msg := m;
      return r;
   end;  -- create

   str: STR is
      return msg;
   end;  -- str
end; -- class GNUESS_ERROR
-------------------------------------------------------------------
abstract class $DB_FILE is
   add(id1, id2: SDB_ID);
   file(f: FILE);
   size: INT;
   close;
end; -- class $DB_FILE
-------------------------------------------------------------------
class DB_TEXT_FILE < $DB_FILE is
   const fs: CHAR := '\001';	-- Field Separator
   const ls: CHAR := '\n';	-- Line Separator

   attr file: FILE;

   attr size: INT;

   create: SAME is
      return new;
   end; -- create

   create(f: FILE): SAME is
      r ::= new;
      r.file := f;
      return r;
   end; -- create

   add(id1: SDB_ID, id2: SDB_ID) is
      file.plus("" + id1 + fs + id2 + fs + ls);
      size := size + 1;
   end; -- add

   close is
      file.close;
   end; -- close
end; -- class DB_TEXT_FILE
-------------------------------------------------------------------
class DB_BIN_FILE < $DB_FILE is
   attr file: FILE;

   attr size: INT;

   create: SAME is
      return new;
   end; -- create

   create(f: FILE): SAME is
      r ::= new;
      r.file := f;
      return r;
   end; -- create

   add(id1: SDB_ID, id2: SDB_ID) is
      int_id1 ::= #C_INT(id1.id);
      int_id2 ::= #C_INT(id2.id);
      fp ::= file.fp;
      SYS::inlined_C("#int_id1 = htonl(#int_id1)");
      SYS::inlined_C("#int_id2 = htonl(#int_id2)");
      SYS::inlined_C("fwrite(&#int_id1, sizeof(int), 1, #fp)");
      SYS::inlined_C("fwrite(&#int_id2, sizeof(int), 1, #fp)");
      size := size + 1;
   end; -- add

   close is
      file.close;
   end; -- close
end; -- class DB_BIN_FILE
-------------------------------------------------------------------
class BIN_FILE < $OSTREAM is
   include FILE;
   
   plus_int(i: INT) is
      ci ::= #C_INT(i);
      filep ::= fp;
      SYS::inlined_C("#ci = (int)ntohl((long)#ci)");
      SYS::inlined_C("fwrite(&#ci, sizeof(int), 1, #filep)");
   end; -- plus
end; -- class BIN_FILE
-------------------------------------------------------------------
