#!/usr/local/bin/wish4.1 -f

#
# Directory of current SDB
#
set SDB "SDB"

#
# Text base viewer
#

set shapid_cmd shapid
set shapid   [open |$shapid_cmd r+]

proc put_cmd {cmd} {
    global shapid

    puts $shapid $cmd
    flush $shapid
}

proc get_ans {} {
    global shapid

    set ans ""
    gets $shapid line
    while {$line != "done."} {
	set ans "$ans$line\n"
	gets $shapid line
    }
    return $ans
}

set mode c
trace variable mode w change_mode
proc change_mode {name element op} {
    global mode
    put_cmd "mode $mode"
    get_ans
}

frame .viewer
pack .viewer

#
# List of classes and relations
#
set title "Classes"
frame .viewer.classes
pack .viewer.classes -side left -expand y -fill y

label .viewer.classes.title -text $title
pack .viewer.classes.title

listbox .viewer.classes.ls -yscroll ".viewer.classes.scl set"
pack .viewer.classes.ls -side left -expand yes -fill both
#.viewer.classes.ls configure -selectbackground red2

regsub -all {SDB/|PIDB|[a-z]+_[a-z]+} [glob "$SDB/*"] "" list
eval ".viewer.classes.ls insert end $list"

#foreach i [exec /bsd43/bin/ls -1 $SDB | egrep -v (PIDB|_)] {
#    .viewer.classes.ls insert end $i
#}

bind .viewer.classes.ls <Double-Button-1> {show_obj_list}

# scroll bar
scrollbar .viewer.classes.scl -command ".viewer.classes.ls yview"
pack .viewer.classes.scl -side right -fill y

#
# List of objects
#
set title "Object IDs"
frame .viewer.objects
pack .viewer.objects -expand yes -fill y

label .viewer.objects.title -text $title
pack .viewer.objects.title

listbox .viewer.objects.obj -yscroll ".viewer.objects.yscl set"
pack .viewer.objects.obj  -side left   -expand yes -fill y

scrollbar .viewer.objects.yscl -command ".viewer.objects.obj yview"
pack .viewer.objects.yscl -side right -fill y

bind .viewer.objects.obj <Double-Button-1> {show_detail "from_list"}

proc show_obj_list {} {
    global SDB
    .viewer.classes.ls config -cursor watch
    update idletasks
    .viewer.objects.obj delete 0 end
    set ind [.viewer.classes.ls curselection]
    set name [.viewer.classes.ls get $ind]
#    .viewer.classes.ls select clear 0
#    .viewer.classes.ls select anchor $ind
    foreach i [exec cut -f2 ${SDB}/${name}] {
	.viewer.objects.obj insert end $i
    }
    .viewer.classes.ls config -cursor arrow
}

#
# Options
#
frame .viewer.options
pack .viewer.options -expand yes -fill y -side left -pady 30

set object_id ""

frame .viewer.options.id
pack .viewer.options.id -anchor w
label .viewer.options.id.label -text "ID" -width 3
entry .viewer.options.id.ent -textvariable object_id  -width 14 -border 5
pack .viewer.options.id.label .viewer.options.id.ent -side left -fill x

bind .viewer.options <Enter> {focus .viewer.options.id.ent}

frame .viewer.options.buttons
pack .viewer.options.buttons -side top
button .viewer.options.buttons.go -text "Search" \
	-command {show_detail "from_entry"} -width 7
button .viewer.options.buttons.cl -text "Clear" \
	-command {set object_id ""} -width 7
button .viewer.options.buttons.text -text "Text" \
	-command {show_text} -width 7
pack .viewer.options.buttons.go .viewer.options.buttons.cl  \
	.viewer.options.buttons.text -side top 

# mode
frame .viewer.options.mode
pack .viewer.options.mode -side top
radiobutton .viewer.options.mode.c -text "C-mode" -variable mode \
            -value c -anchor w
radiobutton .viewer.options.mode.i -text "I-mode" -variable mode \
            -value i -anchor w
pack .viewer.options.mode.c .viewer.options.mode.i

button .viewer.options.quit -text "QUIT" -command {destroy .} -width 7
pack .viewer.options.quit  -side bottom

#
# Details of an object
#

set title "Details"
frame .viewer.details
pack  .viewer.details -expand yes -fill y

label .viewer.details.title -text $title
pack .viewer.details.title

proc mk_details_txt {} {
    text .viewer.details.txt  -width 40  \
	    -yscroll ".viewer.details.yscl set"
    pack .viewer.details.txt -side left -expand yes -fill y
}

mk_details_txt

scrollbar .viewer.details.yscl -command ".viewer.details.txt yview"
pack .viewer.details.yscl -side right -fill y


proc show_detail {type} {
    global object_id
    .viewer config -cursor watch
    .viewer.details.txt config -cursor watch
    update idletasks

    destroy .viewer.details.txt
    mk_details_txt

    if {$type == "from_list"} {
	set ind [.viewer.objects.obj curselection]
	set name [.viewer.objects.obj get $ind]
    } else {
	if {$object_id == ""} {
	    set ind [.viewer.objects.obj curselection]
	    if {$ind == ""} {
		.viewer config -cursor arrow
		return
	    }
	    set name [.viewer.objects.obj get $ind]
	} else {
	    set name $object_id
	}
    }
    put_cmd "what $name"
    set data [get_ans]
#    set data [exec  $SDB_view $name]

#    .viewer.details.txt delete 1.0 end
    .viewer.details.txt insert 1.0 $data

# ID  show_detail_id  bind 
    set pos [.viewer.details.txt search -regexp {[iI][dD]:} 1.0 end]
    while {$pos != ""} {
	set pos "$pos + 4 chars"
	set eol "$pos wordend"
#	set eol [.viewer.details.txt search -regexp {$} $pos end]
	.viewer.details.txt tag add tag${pos} $pos $eol
	.viewer.details.txt tag bind tag${pos} <ButtonPress-1> \
	    "show_detail_id [.viewer.details.txt get $pos $eol]"
	.viewer.details.txt tag bind tag${pos} <Enter> \
	    ".viewer.details.txt tag configure \"tag${pos}\" -overstrike t"
	.viewer.details.txt tag bind tag${pos} <Leave> \
	    ".viewer.details.txt tag configure \"tag${pos}\" -overstrike f"
	.viewer.details.txt tag configure tag${pos} -underline t
	set pos [.viewer.details.txt search -regexp {[iI][dD]:} $eol end]
    }
    .viewer config -cursor arrow
}

proc show_detail_id {id} {
    global object_id
    set object_id $id
    show_detail {from_details_id}
}

set add_app_frame 0

proc show_text {} {
    global object_id
    global add_app_frame
    global mode
    if {$object_id == ""} {
	set ind [.viewer.objects.obj curselection]
	if {$ind == ""} {
	    return
	}
	set name [.viewer.objects.obj get $ind]
    } else {
	set name $object_id
    }

    put_cmd "print $name"
    set data [get_ans]
#    set data [exec  $SDB_print $arg  $name]

    if {$add_app_frame == 0} {
	frame .app
	pack .app
	set add_app_frame 1
	label .app.name -text "Text"
	pack .app.name  -side top -expand yes -fill y
	text .app.data -yscroll ".app.yscl set" -height 10
	pack .app.data -side left -expand yes -fill y

	scrollbar .app.yscl -command ".app.data yview"
	pack .app.yscl -side right -fill y
    }
    .app.data delete 1.0 end
    update idletasks
    .app.data insert 1.0 $data
}

pack .viewer.classes .viewer.objects .viewer.details -side left
