/* 
 * Program:             $RCSfile: AR4.h,v $  $Revision: 30.57 $
 * 
 * Purpose:
 * 
 * Author:              Y.Momiyama  1995/10/01
 *                      S.Yamamoto  1996/07/13
 * 
 * (C) Copyright:       Y.Momiyama and S.Yamamoto  1995 - 2003
 *                      This file is a product of the project Sapid.
 */

#if !defined(_SPD_AR4_H)
#define _SPD_AR4_H

/*
 * Configuration.
 */

#define USE_MACRO_spdGetName            /* Added by S.Yamamoto 1998/07/03. */

/*
 * Defined if either SAPID_FAST_db_get_object_id or
 * SAPID_FAST_db_get_relation_id is defined.
 */
#define SAPID_FAST_db_get_xxx_id

#define SAPID_FAST_spdGetClassIdByName  /* Added by M.Hakamata 1998/07/15. */

/******************************************************************************/

/*
 *                      SDB Access Routines.
 */

typedef struct {
    int         ivalue[SpdObjIValN];
    char        *svalue[SpdObjSValN];
} SpdDBObj;

typedef struct {
    int         ivalue[SpdRelIValN];
} SpdDBRel;

typedef struct {
    SpdDBObj    obj;
    SpdClassId  classId;
    int         *tmp;
} SpdDBObjCursor;

typedef struct {
    SpdDBRel    rel;
    SpdClassId  classId;
    int         *tmp;
} SpdDBRelCursor;

/* In "ar.c". */

typedef struct _SpdCursor {
    int         state;
    SpdId       objectId;
    int         attrId;
    SpdObjId    fileId;                 /* Added by H.Ohhashi 1998/05/09. */
    char        model;
    char        *relationName;
    SpdQueue    queue;
    SpdDBObjCursor      *op;
    SpdDBRelCursor      *pp;
    SpdDBRelCursor      *cp;
    struct _SpdCursor   *csr;
} SpdCursor;

#ifdef SAPID_AR4_KERNEL
#include "AR4Sub.h"
#endif

#include <Sapid/SMU.h>      /* For SpdXxxArray. */
#include <Sapid/SapidModel.h>

/******************************************************************************/

/* Defined in "ar4.c". */

SpdObjId spdGetABlockId(SpdObjId fileId, int offset);
SpdObjIdArray spdGetBlockIdArrayIncludingRegion(SpdRegion region);
SpdBoolean spdIsInHeaderFile(SpdObjId fileId, int offset);
SpdObjId spdGetPfileIdIncludingIoffset(SpdObjId fileId, int offset);
SpdObjIdArray spdGetBlockIdArrayOfMcall(SpdRegion region);
int spdConvOffsetItoC(SpdObjId i_fileId, int i_offset);
SpdRegion spdConvRegionItoCbyPIDB(SpdRegion regionI);
SpdRegion spdConvRegionItoC(SpdRegion regionI);
SpdRegion spdConvRegionCtoI(SpdRegion regionC);
SpdRegion spdConvOccToRegion(SpdOcc occ);
SpdOcc spdConvRegionToOcc(SpdRegion region);
char *spdGetTextByRegion(SpdRegion region);
SpdObjId spdGetCfileIdByPfileId(SpdObjId pfileId);

/* Defined in "ar.c". */

SpdString spdGetClassNameById(SpdClassId classId);
SpdClassId spdGetClassIdByName(SpdConstString className);

/* 
 * Modified by S.Yamamoto 1998/07/04.
 * SpdId spdGetAttrIdByName(SpdString className, SpdString attrName);
 */

#define spdGetAttrIdByName(className, attrName) \
            spdGetAttrIdByName3(SPD_NON_CID, (className), (attrName))

/* Defined in "getClassId.c". */

SpdClassId spdGetClassId(SpdObjId objectId);
SpdString spdGetClassName(SpdObjId objectId);

SpdBoolean spdIsValidClassOrRelationId(SpdClassId classId);
SpdBoolean spdIsValidClassId(SpdClassId classId);
SpdBoolean spdIsValidRelationId(SpdClassId relId);

/******************************************************************************/

#ifdef USE_MACRO_spdGetName

#define spdGetName(objId)       spdGetAttrValString((objId), "name")
#define spdGetLength(objId)     spdGetAttrValIntOfObj((objId), "length")
#define spdGetOffset(relId)     spdGetAttrValIntOfRel((relId), "offset")

#else /* not USE_MACRO_spdGetName */

SpdString spdGetName(SpdObjId objectId);
/* SpdConstString spdGetName(SpdObjId objectId); */
int spdGetLength(SpdObjId objectId);
int spdGetOffset(SpdRelId relationId);

#endif /* not USE_MACRO_spdGetName */

int spdGetRelativeOffset(SpdRelId relationId);
int spdGetAbsoluteOffset(SpdRelId relationId, SpdObjId fileId);
void spdFreeCursor(SpdCursor *csr);
SpdCursor *spdGetObjIdInit(SpdConstString className, SpdConstString name);
SpdObjId spdGetObjId(SpdCursor *csr);
SpdObjId spdGetAnObjId(SpdConstString className, SpdConstString name);
SpdCursor *spdGetRelInit(SpdObjId objectId, SpdConstString relationName, SpdConstString attrName);
SpdRel spdGetRel(SpdCursor *csr);
SpdRel spdGetARel(SpdObjId objectId, SpdConstString relationName, SpdConstString attrName);
SpdCursor *spdGetRelObjInit(SpdObjId objectId, SpdConstString relationName, SpdConstString attrName);
SpdObjId spdGetRelObj(SpdCursor *csr);
SpdObjId spdGetARelObj(SpdObjId objectId, SpdConstString relationName, SpdConstString attrName);
SpdCursor *spdGetRelIdInit(SpdObjId objectId1, SpdObjId objectId2, SpdConstString relationName);
SpdRelId spdGetRelId(SpdCursor *csr);
SpdRelId spdGetARelId(SpdObjId objectId1, SpdObjId objectId2, SpdConstString relationName);
SpdString spdGetObjText(SpdObjId objectId);
int spdGetLengthOfObj(SpdObjId objectId);

SpdObjId spdGetFileId(SpdRelId relationId);

/* Defined in "db.c". */

SpdObjId spdTarget(SpdString modelName, SpdString programName);
void spdReleaseTarget(void);
SpdDBId spdOpenSDB(void);
void spdCloseSDB(SpdDBId dbId);

SpdBoolean  spdIsValidId(SpdId id);

/* Defined in "getOcc.c". */

SpdCursor *spdGetOffsetOfObjInit(SpdObjId objectId);
SpdOcc spdGetOffsetOfObj(SpdCursor *csr);
SpdOcc spdGetAnOffsetOfObj(SpdObjId objectId);
SpdCursor *spdGetIncludedOccInit(SpdObjId objectId,
                                 SpdConstString objectType,
                                 SpdObjId fileId);
SpdOcc spdGetIncludedOcc(SpdCursor *csr);

/* Defined in "spdWriteSDB.c". */

int spdWriteClass(SpdClassId classId);
int spdWriteAllClass(SpdString modelName);

/* Defined in "getAttrVal.c". */

/*
 * Optimized functions.
 * Added by S.Yamamoto 1997/02/09.
 */

int spdGetAttrValInt(SpdId objectId, SpdConstString attrName);
SpdString spdGetAttrValString(SpdId objectId, SpdString attrName);
/* SpdConstString spdGetAttrValString(SpdId objectId, SpdString attrName); */

int spdGetAttrValIntOfObj(SpdObjId objectId, SpdConstString attrName);
int spdGetAttrValIntOfRel(SpdRelId relationId, SpdConstString attrName);

/******************************************************************************/

/*
 * Added by S.Yamamoto 1998/09/02. 
 */

#ifdef SAPID_AR4_KERNEL
SpdAttrNameId spdGetAttrNameId(SpdConstString cp);
#endif

SpdAttrId spdGetAttrIdByName3(SpdClassId classId, SpdConstString className,
                              SpdConstString attrName);

/******************************************************************************/

/*
 * Added by S.Yamamoto 1999/10/26. 
 */

SpdModelId spdGetModelId(SpdClassId classId);

#ifdef USE_spdIsClassId
SpdBoolean spdIsClassId(SpdClassId classId, SpdModelId modelId);
#endif

/* Added by S.Yamamoto 2001/10/06. */
SpdConstString spdGetModelNameById(SpdModelId modelId);

/******************************************************************************/

/*
 * Added by S.Yamamoto 2000/04/06. 
 */

SpdConstString spdGetAttrName(SpdClassId classId, SpdAttrId attrId);

int spdGetNumOfAttrs(SpdClassId classId);
int spdGetNumOfIntAttrs(SpdClassId classId);
int spdGetNumOfStrAttrs(SpdClassId classId);

SpdAttrId spdGetFirstAttrIdInt(SpdClassId classId);
SpdAttrId spdGetLastAttrIdInt(SpdClassId classId);
SpdAttrId spdGetFirstAttrIdStr(SpdClassId classId);
SpdAttrId spdGetLastAttrIdStr(SpdClassId classId);

/******************************************************************************/

/*
 * Added into "AR4_model.c" by S.Yamamoto 2000/04/08. 
 */

void spd_reinitialize_models(void);

SpdModelId spdUseModel(SpdDBModel *model);

SpdClassId spdGetFirstCid(SpdModelId modelId);
SpdClassId spdGetLastCid(SpdModelId modelId);
SpdClassId spdGetFirstRid(SpdModelId modelId);
SpdClassId spdGetLastRid(SpdModelId modelId);

SpdBoolean spdIsValidModelId(SpdModelId modelId);
SpdModelId spdGetModelIdByName(SpdConstString modelName);

#ifdef SAPID_AR4_KERNEL
void spdSetModelDir(SpdModelId modelId, SpdConstString pathname);
SpdConstString spdGetModelDir(SpdModelId modelId);
#endif

/******************************************************************************/

/*
 * Added into "AR4_model.c" by S.Yamamoto 2000/04/21. 
 */

int spdReadDB(SpdModelId modelId, SpdDBId dbId);
int spdCreateDB(SpdModelId modelId, SpdDBId dbId);

#endif /* _SPD_AR4_H */

extern SpdDBModel spdModels[SPD_MID_NUM];
