/* 
 * Program:             $RCSfile: AR4_model.c,v $  $Revision: 30.56 $
 * 
 * Purpose:
 * 
 * Author:              S.Yamamoto  1999/10/26
 * 
 * (C) Copyright:       S.Yamamoto  1999 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: AR4_model.c,v 30.56 2003/08/16 10:32:54 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DEBUG_020401 */

#if defined(lint)
#define DW_020401                   printf      /* Just for cheking. */
#elif defined(DEBUG_020401)
#define DW_020401                   spdMessage
#else
#define DW_020401                   (void)
#endif

SpdDBModel   spdModels[SPD_MID_NUM];

static void spdReadAModel(SpdModelId modelId, SpdDBId dbId);
static void spd_initialize_models(void);
static void spdInitPmodelAR(void);

/* 
 * Clear flags in model table "spdModels[]".
 */

static void spd_initialize_models(void)
{
    SpdModelId      mId;

    spdModels[SPD_MID_I_MODEL] = I_model;
    spdModels[SPD_MID_P_MODEL] = P_model;
    spdModels[SPD_MID_AREA_MODEL] = Area_model;
    spdModels[SPD_MID_B_MODEL] = B_model;
    spdModels[SPD_MID_D_MODEL] = D_model;
    spdModels[SPD_MID_SDA_MODEL] = SDA_model;
    spdModels[SPD_MID_STMT_MODEL] = Stmt_model;
    spdModels[SPD_MID_SDA_AREA_MODEL] = SDAarea_model;
    spdModels[SPD_MID_TYPE_VIEW] = TypeView_model;
    spdModels[SPD_MID_COMMENT_VIEW] = Comment_view;
    spdModels[SPD_MID_FLOW_VIEW] = Flow_view;
    spdModels[SPD_MID_FCD_VIEW] = FCD_view;
    spdModels[SPD_MID_J_MODEL] = J_model;

    for (mId = SPD_FIRST_MID; mId <=  SPD_LAST_MID; mId++) {
        spdModels[mId].used = FALSE;
    }
}

/* 
 * Clear flags in model table "spdModels[]".
 * ϿǥϡץνλޤǻѲǽǤ뤿ᡤusedϥ
 * ꥢʤ
 */

void spd_reinitialize_models(void)
{
    SpdModelId      mId;

    for (mId = SPD_FIRST_MID; mId <=  SPD_LAST_MID; mId++) {
        spdModels[mId].read = FALSE;
        spdModels[mId].created = FALSE;
    }
}

/* 
 * ǥѤǥϿ롥
 * ϿǥϡץνλޤǻѲǽǤ롥
 * 
 * !! ϿäAPIѰդƤʤ
 */

SpdModelId spdUseModel(SpdDBModel *model)
{
    SpdClassId          i;
    static SpdBoolean   is_first_call = TRUE;

    if (is_first_call) {
        spd_initialize_models();
        is_first_call = FALSE;
    }

    /* Check argument. */
    if (model == NULL) {
        spdErrorExit("spdUseModel(): Invalid model.\n");
    } else if (model->id != spdModels[model->id].id) {
        spdErrorExit("spdUseModel(): %d: Invalid model.\n", model->id);
    }

    if (model->id < SPD_FIRST_MID || SPD_LAST_MID < model->id) {
        spdErrorExit("spdUseModel(): %d: Invalid model.\n", model->id);
    }

    /* Check validity of the model. */
    if ((model->lastCid - model->firstCid + 1) != model->numOfClasses ||
        (model->lastRid - model->firstRid + 1) != model->numOfRelations ||
        (model->numOfClasses + model->numOfRelations) != model->numOfAll) {
        spdMessage("spdUseModel(): %d: Invalid model attribute.\n", model->id);
    }
    for (i = model->firstCid; i < model->lastCid; i++) {
        if (!spdIsValidClassId(i)) {
            spdErrorExit("spdUseModel(): %d, %d: Invalid class id.\n",
                         model->id, i);
        }
    }
    for (i = model->firstRid; i < model->lastRid; i++) {
        if (!spdIsValidRelationId(i)) {
            spdErrorExit("spdUseModel(): %d, %d: Invalid relationship id.\n",
                         model->id, i);
        }
    }

    if (!spdModels[model->id].used) {
        spdModels[model->id] = *model;
        spdModels[model->id].used = TRUE;
        spdModels[model->id].read = FALSE;
        spdModels[model->id].created = FALSE;
    }

    return (model->id);
}

/*
 * 饹°ǥμֹ롥
 */

SpdModelId spdGetModelId(SpdClassId classId)
{
    if (SPD_FIRST_CID <= classId && classId <= SPD_LAST_RID) {
        return (class_table[classId].modelId);
    } else {
        return (SPD_NON_MID);
    }
}

/*
 * ǥμֹ椫ǥ̾롥
 */

SpdConstString spdGetModelNameById(SpdModelId modelId)
{
    if (SPD_FIRST_MID <= modelId && modelId <= SPD_LAST_MID) {
        return (spdModels[modelId].name);
    } else {
        return (NULL);
    }
}

/******************************************************************************/

/*
 * "SpdBoolean spdReadDB(SpdModelId modelId, SpdDBId dbId)" is better.
 */

int spdReadDB(SpdModelId modelId, SpdDBId dbId)
{
    SpdModelId      mId;
    SpdBoolean      isFinished;

    if (modelId != SPD_NON_MID &&
        (modelId < SPD_FIRST_MID || SPD_LAST_MID < modelId)) {
        spdErrorExit("spdReadDB(): %d: Invalid model id.\n", modelId);
    }

    isFinished = TRUE;
    for (mId = SPD_FIRST_MID; mId <=  SPD_LAST_MID; mId++) {
        if (spdModels[mId].used && spdModels[mId].read == FALSE) {
            isFinished = FALSE;
        }
    }
    if (isFinished == TRUE) {
        /* All DB are read. */
        return (0);
    }

    for (mId = SPD_FIRST_MID; mId <= SPD_LAST_MID; mId++) {
        if (spdModels[mId].used) {
            set_initial_db_num(mId);
        }
    }
    for (mId = SPD_FIRST_MID; mId <= SPD_LAST_MID; mId++) {
        DW_020401("spdReadDB(): spdModels[%d].used %d.\n",
                  mId, spdModels[mId].used);
        if (spdModels[mId].used) {
            spdReadAModel(mId, dbId);
            spdModels[mId].read = TRUE;
        }
    }
    make_index();
    if (spdModels[SPD_MID_P_MODEL].read) {
        spdInitPmodelAR();
    }
    return (0);
}

int spdCreateDB(SpdModelId modelId, SpdDBId dbId)
{
    spdErrorExit("spdCreateDB(): %d: Not supported yet in AR4.\n", modelId);
    SPD_DUMMY_RETURN(0);
}

int spdCommitDB(SpdModelId modelId, SpdDBId dbId)
{
    spdErrorExit("spdCommitDB(): %d: Not supported yet in AR4.\n", modelId);
    SPD_DUMMY_RETURN(0);
}

/******************************************************************************/

/* 
 * Validity of argument "modelId" already checked.
 */

static void spdReadAModel(SpdModelId modelId, SpdDBId dbId)
{
    SpdClassId      i;
    SpdDBModel      model = spdModels[modelId];

    for (i = model.firstCid; i <= model.lastCid; i++) {
        read_obj_class(i, spdGetDbInfoNumOfObjs(i), spdReadAndInsertObj);
    }
    for (i = model.firstRid; i <= model.lastRid; i++) {
        read_rel_class(i, spdGetDbInfoNumOfObjs(i), spdReadAndInsertRel);
    }
}

/******************************************************************************/

SpdBoolean spdIsValidModelId(SpdModelId modelId)
{
    return (SPD_FIRST_MID <= modelId && modelId <=  SPD_LAST_MID);
}

/******************************************************************************/

SpdClassId spdGetFirstCid(SpdModelId modelId)
{
    if (!spdIsValidModelId(modelId)) {
        spdErrorExit("spdGetFirstCid(): %d: Invalid model id.\n", modelId);
    }
    return (spdModels[modelId].firstCid);
}

SpdClassId spdGetLastCid(SpdModelId modelId)
{
    if (!spdIsValidModelId(modelId)) {
        spdErrorExit("spdGetLastCid(): %d: Invalid model id.\n", modelId);
    }
    return (spdModels[modelId].lastCid);
}

SpdClassId spdGetFirstRid(SpdModelId modelId)
{
    if (!spdIsValidModelId(modelId)) {
        spdErrorExit("spdGetFirstRid(): %d: Invalid model id.\n", modelId);
    }
    return (spdModels[modelId].firstRid);
}

SpdClassId spdGetLastRid(SpdModelId modelId)
{
    if (!spdIsValidModelId(modelId)) {
        spdErrorExit("spdGetLastRid(): %d: Invalid model id.\n", modelId);
    }
    return (spdModels[modelId].lastRid);
}

/******************************************************************************/

/*
 * Added by S.Yamamoto 2000/07/16.
 */

#ifdef SAPID_AR4_KERNEL

/*
 * Set new pathname of the model directory (ex. "PSDB").
 * Added by N.Fukuyasu 2003/08/11.
 */

void spdSetModelDir(SpdModelId modelId, SpdConstString pathname)
{
    if (!spdIsValidModelId(modelId)) {
        spdErrorExit("spdSetModelDir(): %d: Invalid model id.\n", modelId);
    }
    if (spdModels[modelId].used) {
        spdModels[modelId].pathname = pathname;
    }
}

SpdConstString spdGetModelDir(SpdModelId modelId)
{
    if (!spdIsValidModelId(modelId)) {
        spdErrorExit("spdGetModelDir(): %d: Invalid model id.\n", modelId);
    }
    return (spdModels[modelId].used ? spdModels[modelId].pathname : NULL);
}

#endif /* SAPID_AR4_KERNEL */

/******************************************************************************/

static void spdInitPmodelAR(void)
{
    SpdObjIdArray   prog_ids;
    int             i;
    SpdObjId        lib_id, file_id;
    SpdCursor       *csr, *csr2;

    prog_ids = spdGetObjIdArray("program", NULL, NULL);
    for (i = 0; i < prog_ids.size; i++) {
        csr = spdGetRelObjInit(prog_ids.id[i], "prog_file", "any");
        while ((file_id = spdGetRelObj(csr)) != SPD_NON_ID) {
            spdCreateBlockTable(file_id);
        }
        spdFreeCursor(csr);
        /* Modified for class "library" by H.Ohhashi 1999/07/02. */
        csr = spdGetRelObjInit(prog_ids.id[i], "prog_lib", "any");
        while ((lib_id = spdGetRelObj(csr)) != SPD_NON_ID) {
            csr2 = spdGetRelObjInit(lib_id, "lib_file", "any");
            while ((file_id = spdGetRelObj(csr2)) != SPD_NON_ID) {
                spdCreateBlockTable(file_id);
            }
            spdFreeCursor(csr2);
        }
        spdFreeCursor(csr);
    }
    spdFreeObjIdArray(prog_ids);
}

/******************************************************************************/

/* ǥ̾ǥμֹ롥 */

SpdModelId spdGetModelIdByName(SpdConstString modelName)
{
    SpdModelId      mId;

    for (mId = SPD_FIRST_MID; mId <=  SPD_LAST_MID; mId++) {
        if (SPD_STREQ(spdModels[mId].name, modelName)) {
        return (spdModels[mId].id);
        }
    }
    return (SPD_NON_MID);
}
