/* 
 * Program:             $RCSfile: ar.c,v $  $Revision: 30.55 $
 * 
 * Purpose:             Access routines.
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: ar.c,v 30.55 2003/06/28 15:31:47 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdRel       NOT_FOUND_REL = {SPD_NON_ID, SPD_NON_ID};

/* Cursor management. *********************************************************/

SpdCursor *spdAllocCursor(void)
{
    /* SPD_EMPTY_QUEUE must be {0, 0}. */
    return (spdCalloc(1, sizeof(SpdCursor)));
}

void spdFreeCursor(SpdCursor *csr)
{
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return;
    }

    if (csr->relationName) {
        spdFree(csr->relationName);
    }
    spdQueueClean(&(csr->queue));
    if (csr->op) {
        db_get_object_name_free(csr->op);
    }
    if (csr->pp) {
        db_get_relation_p_id_free(csr->pp);
    }
    if (csr->cp) {
        db_get_relation_c_id_free(csr->cp);
    }
    if (csr->csr) {
        spdFreeCursor(csr->csr);
    }
    spdFree(csr);
}

/******************************************************************************/

SpdCursor *spdGetRelByAttrInit(SpdObjId objectId, SpdConstString relationName,
                               SpdConstString attrName)
{
    SpdCursor *csr;
    SpdClassId classId;
    
    csr = spdAllocCursor();
    if (SPD_STREQ(attrName, "p-id")) {
        csr->attrId = SPD_AID_REL_P_ID;
    } else if (SPD_STREQ(attrName, "c-id")) {
        csr->attrId = SPD_AID_REL_C_ID;
    } else {
        csr->attrId = spdGetAttrIdByName(relationName, attrName);
    }
    if (csr->attrId != SPD_AID_REL_P_ID &&
        csr->attrId != SPD_AID_REL_C_ID) {
        spdFreeCursor(csr);
        return (NULL);
    }
    classId = SPD_STREQ(relationName, "any") ?
        C_ANY : spdGetClassIdByName(relationName);
    if (csr->attrId == SPD_AID_REL_P_ID) {
        csr->pp = db_get_relation_p_id_init(classId, objectId);
        if (!csr->pp) {
            spdFreeCursor(csr);
            return (NULL);
        }
    } else {
        csr->cp = db_get_relation_c_id_init(classId, objectId);
        if (!csr->cp) {
            spdFreeCursor(csr);
            return (NULL);
        }
    }
    return (csr);
}

SpdRel spdGetRelByAttrRaw(SpdCursor *csr)
{
    SpdRel rel;
    SpdDBRel *tmp;
    
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return (NOT_FOUND_REL);
    }

    if (csr->attrId == SPD_AID_REL_P_ID) {
        tmp = db_get_relation_p_id(csr->pp);
    } else {
        tmp = db_get_relation_c_id(csr->cp);
    }
    if (tmp == NULL) {
        return (NOT_FOUND_REL);
    }
    rel.relationId = tmp->Rel_id;
    rel.objectId = 
        (csr->attrId == SPD_AID_REL_P_ID) ? tmp->Rel_c_id : tmp->Rel_p_id;
    return (rel);
}

SpdRel spdGetRelByAttr(SpdCursor *csr)
{
    SpdRel          rel;
    
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return (NOT_FOUND_REL);
    }

    rel = spdGetRelByAttrRaw(csr);
    return (rel);
}

SpdCursor *spdGetRelAnyInit(SpdObjId objectId, SpdConstString relationName)
{
    SpdCursor *csr;
    
    csr = spdAllocCursor();
    csr->state = 0;
    csr->objectId = objectId;
    csr->relationName = spdStrdup(relationName);
    csr->csr = spdGetRelByAttrInit(objectId, relationName, "p-id");
    if (!csr->csr) {
        spdFreeCursor(csr);
        return (NULL);
    }
    return (csr);
}

SpdRel spdGetRelAny(SpdCursor *csr)
{
    SpdRel          rel;

    if (csr == NULL) {
        return (NOT_FOUND_REL);
    }
    
    rel = spdGetRelByAttr(csr->csr);
    if (csr->state == 0 && rel.relationId == -1) {
        spdFreeCursor(csr->csr);
        csr->csr = spdGetRelByAttrInit(csr->objectId, csr->relationName,
                                       "c-id");
        csr->state = 1;
        rel = spdGetRelByAttr(csr->csr);
    }
    return (rel);
}

SpdCursor *spdGetRelInit(SpdObjId objectId,
                         SpdConstString relationName, SpdConstString attrName)
{
    SpdCursor *csr;
    
    if (SPD_STREQ(attrName, "any")) {
        csr = spdGetRelAnyInit(objectId, relationName);
    } else {
        csr = spdGetRelByAttrInit(objectId, relationName, attrName);
        if (csr)
            csr->state = -1;
    }
    return (csr);
}

SpdRel spdGetRel(SpdCursor *csr)
{
    if (csr == NULL) {
        return (NOT_FOUND_REL);
    }
    return (csr->state == -1 ? spdGetRelByAttr(csr) : spdGetRelAny(csr));
}

SpdRel spdGetARel(SpdObjId objectId, SpdConstString relationName,
                  SpdConstString attrName)
{
    SpdCursor *csr;
    SpdRel rel;
    
    if ((csr = spdGetRelInit(objectId, relationName, attrName)) == NULL) {
        return (NOT_FOUND_REL);
    }
    rel = spdGetRel(csr);
    spdFreeCursor(csr);
    return (rel);
}

SpdCursor *spdGetRelObjInit(SpdObjId objectId, SpdConstString relationName,
                            SpdConstString attrName)
{
    return (spdGetRelInit(objectId, relationName, attrName));
}

SpdObjId spdGetRelObj(SpdCursor *csr)
{
    /* Note, validity of csr is checked in spdGetRel(). */
    return (spdGetRel(csr).objectId);
}

SpdObjId spdGetARelObj(SpdObjId objectId, SpdConstString relationName,
                       SpdConstString attrName)
{
    return (spdGetARel(objectId, relationName, attrName).objectId);
}

SpdCursor *spdGetRelIdInit(SpdObjId objectId1, SpdObjId objectId2,
                           SpdConstString relationName)
{
    SpdCursor *csr;
    
    csr = spdAllocCursor();
    csr->objectId = objectId2;
    csr->relationName = spdStrdup(relationName);
    csr->csr = spdGetRelInit(objectId1, relationName, "any");
    if (!csr->csr) {
        spdFreeCursor(csr);
        return (NULL);
    }
    return (csr);
}

SpdRelId spdGetRelId(SpdCursor *csr)
{
    SpdRel  rel;
    
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return (SPD_NON_ID);
    }

    do {
        rel = spdGetRel(csr->csr);
    } while (rel.objectId != -1 && rel.objectId != csr->objectId);
    return (rel.relationId);
}

SpdRelId spdGetARelId(SpdObjId objectId1, SpdObjId objectId2,
                      SpdConstString relationName)
{
    SpdCursor *csr;
    SpdRelId id;
    
    if ((csr = spdGetRelIdInit(objectId1, objectId2, relationName)) == NULL) {
        return (SPD_NON_ID);
    }
    id = spdGetRelId(csr);
    spdFreeCursor(csr);
    return (id);
}

SpdRelId spdGetFileDeclRelId(SpdRelId relationId, SpdObjId fileId)
{
    SpdObjId objectId;
    SpdRel rel;
    SpdRelId relId;
    SpdCursor *csr;

    if (spdGetClassId(relationId) == SPD_RID_FILE_DECL) {
        return (relationId);
    }
    objectId = spdGetAttrValIntOfRel(relationId, "p-id");
    csr = spdGetRelInit(objectId, "any", "c-id");
    while ((rel = spdGetRel(csr)).relationId != SPD_NON_ID) {
        if (spdGetClassId(rel.relationId) == SPD_RID_FILE_DECL) {
            if (fileId == SPD_NON_ID ||
                fileId == spdGetAttrValIntOfRel(rel.relationId, "file_id")) {
                relId = rel.relationId;
                break;
            }
        } else if (SPD_IS_TYP_CMP(spdGetClassId(rel.relationId))) {
            relId = spdGetFileDeclRelId(rel.relationId, fileId);
            if (relId != SPD_NON_ID)
                break;
        }
    }
    spdFreeCursor(csr);

    return ((rel.relationId == SPD_NON_ID) ? SPD_NON_ID : relId);
}

SpdObjId spdGetFileId(SpdRelId relationId)
{
    SpdRelId    relId;

    relId = spdGetFileDeclRelId(relationId, SPD_NON_ID);
    return ((relId == SPD_NON_ID) ?
            SPD_NON_ID : spdGetAttrValIntOfRel(relId, "file_id"));
}

SpdString spdGetObjText(SpdObjId objectId)
{
    char        *name;
    SpdClassId  classId = spdGetClassId(objectId);

    if (classId == SPD_CID_CONSTANT) {
        return (spdGetAttrValString(objectId, "value"));
    } else if ((name = spdGetName(objectId)) != NULL) {
        return (name);
    } else if (SPD_IS_TYP_CMP(classId)) {
        SpdObjId    fileId;
        int         offset, length;
        SpdCursor   *csr;
        SpdOcc      occ;
        FILE        *fp;

        static char *buf = NULL;

        if (classId == SPD_CID_FILE) {
            fileId = objectId;
            offset = 0;
            length = spdGetFileLength(fileId);
        } else {
            csr = spdGetOffsetOfObjInit(objectId);
            while ((occ = spdGetOffsetOfObj(csr)).relationId != -1 &&
                   (occ.fileId == -1 || occ.offset == -1)) {
                /* Empty. */
            }
            spdFreeCursor(csr);
            if (occ.relationId == -1) {
                return (NULL);
            }
            fileId = occ.fileId;
            offset = occ.offset;
            length = spdGetLength(objectId);
        }
        buf = spdRealloc(buf, length + 1);
        fp = smuOpenSrcFile(fileId);
        SPD_FSEEK(fp, offset, SEEK_SET);
        spdFreadChar(buf, length, fp);
        spdFclose(fp);
        return (buf);
    } else {
        return (NULL);
    }
}

int spdGetLengthOfObj(SpdObjId objectId)
{
    char *buf;
    
    buf = spdGetObjText(objectId);
    return (buf == NULL ? -1 : (int)strlen(buf));
}
