/* 
 * Program:             $RCSfile: ar4.c,v $  $Revision: 30.40 $
 * 
 * Purpose:
 * 
 * Author:              Y.Momiyama  1995/10/01
 *                      S.Yamamoto  1996/07/13
 * 
 * (C) Copyright:       Y.Momiyama and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: ar4.c,v 30.40 2003/06/22 14:41:54 yamamoto Exp yamamoto $";
#endif

#include <limits.h>
#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5_KERNEL)
#include <Sapid/ImodelAR.h>
#include <Sapid/PmodelAR.h>
#endif

static SpdRegion    nonRegion = SPD_NON_REGION;
static Block_Table  nonTable = SPD_NON_BLOCK;

#if defined(SAPID_AR4_KERNEL)

static All_Block_Table  *all_block_table = NULL;
static int              all_block_table_n = 0;

static Block_Table Search(int offset, int start, int end, int n)
{
    int     center;

    while (TRUE) {
        center = start + (end - start) / 2;
        if (end <= start) {
            return (all_block_table[n].block_table[center]);
        }
        if (offset < all_block_table[n].block_table[center+1].offset) {
            end = center;
        } else {
            start = center + 1;
        }
    }
    /*NOTREACHED*/
}

/* "*.i"μֹoffsetblock_table롥 */

static Block_Table spdGetTableOfOffset(SpdObjId i_fileId, int i_offset)
{
    int     n;

    for (n = 0; n < all_block_table_n; n++) {
        if (all_block_table[n].fileId == i_fileId) {
            break;
        }
    }
    return ((n == all_block_table_n) ?
            nonTable :
            Search(i_offset, 0, all_block_table[n].block_table_n-1, n));
}

/* "*.c"offsetޤޤpblockμֹ롥 */

static SpdObjId spdGetBlockIncludingOffset(SpdObjId pfileId, int offset)
{
    SpdCursor *csr;
    SpdRel blk_rel;
    int block_offset, block_length;

    csr = spdGetRelInit(pfileId, "file_block", "p-id");
    while ((blk_rel = spdGetRel(csr)).objectId != SPD_NON_ID) {
        block_offset = spdGetOffset(blk_rel.relationId);
        block_length = spdGetLength(blk_rel.objectId);
        if (offset >= block_offset &&
            offset <= block_offset + block_length) {
            spdFreeCursor(csr);
            return (blk_rel.objectId);
        }
    }
    spdFreeCursor(csr);
    return (SPD_NON_ID);
}

/* pblockμֹ椫block_table롥 */

static Block_Table spdGetTableOfBlockId(SpdObjId block_id)
{
    int         i, n;
    SpdObjId    i_fileId;

    i_fileId = spdGetARelObj(block_id, "file_block", "c-id");
    i_fileId = spdGetCfileIdByPfileId(i_fileId);
    i_fileId = spdGetARelObj(i_fileId, "cfile_ifile", "p-id");
    for (n = 0; n < all_block_table_n; n++) {
        if (all_block_table[n].fileId == i_fileId) {
            break;
        }
    }
    if (n == all_block_table_n) {
      return (nonTable);
    }
    for (i = 0; i < all_block_table[n].block_table_n; i++) {
        if (block_id == all_block_table[n].block_table[i].id) {
            return (all_block_table[n].block_table[i]);
        }
    }
    return (nonTable);
}

/* "*.[chi]"fileIdoffsetBlock_Table롥 */

static Block_Table spdGetBlockTable(SpdObjId fileId, int offset)
{
    if (spdGetClassId(fileId) == SPD_CID_PFILE) {
      /* P-modelpfileξ硥 */
        return (spdGetTableOfBlockId(spdGetBlockIncludingOffset(fileId,
                                                                offset)));
    } else {
      /* I-model  file ξ硤pfile_id롥 */
      return (spdGetTableOfOffset(fileId, offset));
    }
}

void spdCreateBlockTable(SpdObjId ifileId)
{
    int block_table_size;
    int all_block_table_size;
    SpdRelArray iblocks;
    int i;

    all_block_table_size = sizeof(All_Block_Table);
    if (all_block_table == NULL) {
        all_block_table = spdMalloc(all_block_table_size);
    } else {
        all_block_table = spdRealloc(all_block_table,
                                     all_block_table_size
                                     * (all_block_table_n + 1));
    }

    all_block_table[all_block_table_n].fileId = ifileId;

    block_table_size = sizeof(Block_Table);
    iblocks = spdGetRelSortedArray(ifileId, "ifile_iblock", "p-id",
                                   NULL, spdCompareRelOffset);
    all_block_table[all_block_table_n].block_table =
        spdMalloc(block_table_size * iblocks.size);
    all_block_table[all_block_table_n].block_table_n = iblocks.size;

    for (i = 0; i < iblocks.size; i++) {
        all_block_table[all_block_table_n].block_table[i].id =
            spdGetARelObj(iblocks.rel[i].objectId, "pblock_iblock", "c-id");
        all_block_table[all_block_table_n].block_table[i].offset =
            spdGetAttrValInt(iblocks.rel[i].relationId, "offset");
        all_block_table[all_block_table_n].block_table[i].length =
            spdGetAttrValInt(iblocks.rel[i].objectId, "length");
    }

    spdFreeRelArray(iblocks);

    all_block_table_n++;
}    

void spdReleaseBlockTable(void)
{
    /* 
     * if (all_block_table != NULL) {
     *     spdFree(all_block_table);
     *     all_block_table = NULL;
     *     all_block_table_n = 0;
     * }
     */

    if (all_block_table != NULL) {
        int     n;

        for (n = 0; n < all_block_table_n; n++) {
            if (all_block_table[n].block_table != NULL) {
                spdFree(all_block_table[n].block_table);
            }
        }
        spdFree(all_block_table);
        all_block_table = NULL;
        all_block_table_n = 0;
    }
}

/******************************************************************************/

#elif defined(SAPID_AR5_KERNEL)

static const SpdRegionIndexNode *searchRegIndex(const SpdRegionIndex *regIndex, 
                                                int offset)
{
    int     start, end, center;

    start = 0;
    end = regIndex->size - 1;

    while (TRUE) {
        center = start + (end - start) / 2;
        if (end <= start) {
            return (&(regIndex->nodes[center]));
        }
        if (offset < regIndex->nodes[center+1].offset) {
            end = center;
        } else {
            start = center + 1;
        }
    }
    /*NOTREACHED*/
}

static Block_Table spdGetBlockTable(SpdObjId fileId, int offset)
{
    Block_Table tbl = nonTable;
    const SpdRegionIndex *regIndex;
    const SpdRegionIndexNode *regIndexNode;
    SpdObjId pblockId, iblockId;
    SpdRelId relId;

    regIndex = spdGetPmodelRegIndex(fileId);
    if (regIndex == NULL) {
        return nonTable;
    }

    regIndexNode = searchRegIndex(regIndex, offset);

    if (spdGetClassId(fileId) == SPD_CID_FILE) {
        /* file  ID ꤷ */
        tbl.offset = regIndexNode->offset;
        tbl.length = regIndexNode->length;
        iblockId = spdGetAttrIntVal(regIndexNode->relationId, 
                                    SAPIDDB_ATTR_C_ID);
        pblockId = spdGetAParentObjId(iblockId, SPD_RID_PBLOCK_IBLOCK);
        tbl.id = pblockId;
    } else if (spdGetClassId(fileId) == SPD_CID_PFILE) {
        /* pfile  ID ꤷ */
        pblockId = spdGetAttrIntVal(regIndexNode->relationId, 
                                    SAPIDDB_ATTR_C_ID);
        tbl.id = pblockId;
        tbl.length = spdGetAttrIntValByName(pblockId, "length");
        relId = spdGetAParentRelId(pblockId, SPD_RID_FILE_BLOCK);
        tbl.offset = spdGetAttrIntValByName(relId, "offset");
    }
    return tbl;
}

#endif /* defined(SAPID_AR5_KERNEL) */

/* fileIdǻꤵ줿"*.[chi]"offsetޤpblockμֹ롥 */

SpdObjId spdGetABlockId(SpdObjId fileId, int offset)
{
    return (spdGetBlockTable(fileId, offset).id);
}

/* "*.[chi]"occregionѴ롥 */

SpdRegion spdConvOccToRegion(SpdOcc occ)
{
    SpdRegion   reg;

    reg.fileId = occ.fileId;
    reg.offset = occ.offset;
    reg.length = spdGetLength(occ.objectId);
    return (reg);
}

static SpdBoolean isIncludedRegionInRelation(SpdRegion i_reg, SpdRel rel)
{
    int     offset;
    int     length;

    offset = spdGetAbsoluteOffset(rel.relationId, i_reg.fileId);
    length = spdGetLength(rel.objectId);
    if (offset == SPD_NON_OFFSET) {
        return (FALSE);
    }
    /* 
     * if (offset <= i_reg.offset &&
     *     i_reg.offset + i_reg.length <= offset + length) {
     *     return TRUE;
     * } else {
     *     return FALSE;
     * }
     */
    return (offset <= i_reg.offset &&
            i_reg.offset + i_reg.length <= offset + length);
}

/* ".i"regionoccѴ롥 */

SpdOcc spdConvRegionToOcc(SpdRegion region)
{
    SpdCursor *csr;
    SpdObjId objId;
    SpdRel rel;
    SpdRelId relId;
    SpdBoolean findFlag;
    SpdOcc occ;
    
    /* եID  region ƽ */
    objId = region.fileId;

    relId = SPD_NON_ID;
    findFlag = 1;

    while (findFlag) {
        /* Ҥɤ */
        findFlag = 0;
        csr = spdGetRelInit(objId, "any", "p-id");
        while ((rel = spdGetRel(csr)).objectId != SPD_NON_ID) {
            if (isIncludedRegionInRelation(region, rel)) {
                relId = rel.relationId;
                objId = rel.objectId;
                findFlag = 1;
            }
        }
        spdFreeCursor(csr);
    }

    occ.objectId = spdGetAttrValIntOfRel(relId, "c-id");
    occ.relationId = relId;
    occ.fileId = region.fileId;
    occ.offset = spdGetAbsoluteOffset(relId, occ.fileId);

    return occ;
}

/* pfileId顢򥤥󥯥롼ɤƤ
   (Ū˥󥯥롼ɤƤ뺬ܤ)"*.c"pfileId롥 */

SpdObjId spdGetCfileIdByPfileId(SpdObjId pfileId)
{
    SpdObjId pfile_id = pfileId;
    SpdObjId pblock_id;
    SpdObjId hfile_id;

    while ((hfile_id = spdGetARelObj(pfile_id, "hfile_file", "c-id")) !=
           SPD_NON_ID) {
        pblock_id = spdGetARelObj(hfile_id, "incld_hfile", "c-id");
        pfile_id = spdGetARelObj(pblock_id, "file_block", "c-id");
    }

    return pfile_id;
}

/* .c  pfile  ID  */

static SpdObjId spdGetPfileIdOfOffset(SpdObjId fileId, int offset)
{
    Block_Table block_table_one;
    
    block_table_one = spdGetBlockTable(fileId, offset);
    if (offset < block_table_one.offset ||
        offset > block_table_one.offset+block_table_one.length) {
        return SPD_NON_ID;
    }

    return spdGetARelObj(block_table_one.id, "file_block", "c-id");
}

/* .i  fileId, offset  .c Ǥ offset 
   ޥƽФʬPIDB ˤäƼ */

static SpdRegion spdConvRegionItoCbyPIDB_in(SpdRegion regionI)
{
    SpdRegion           regionC;
    /* SpdConstString      ifilename; */
    SpdString           ifilename;
    MacroExpandRegion   target;
    MacroCallRegion     result;
    int                 r;
    
    regionC.fileId = spdGetPfileIdOfOffset(regionI.fileId, regionI.offset);
    if (regionC.fileId == SPD_NON_ID) {
        return nonRegion;
    }
    
    /* PIDB  */
    /* !! Quick hack for suppressing warning about "cast". */
    ifilename = (SpdString)spdGetName(regionI.fileId);
    if (spdInitializePidbCache(regionI.fileId) != 0) {
        spdErrorExit("Can't initialize: \"%s\".", ifilename);
    }

    /* ޥ椫ɤå */
    target.type  = EXPAND_FILE;
    target.start = regionI.offset;
    target.end   = regionI.offset + regionI.length - 1;
    target.fname = ifilename;
#ifdef PIDB_USE_FILENAME_INDEX
    target.fname_idx = pidbConvFileId(ifilename);
#endif
    r = reverseParseAnOutputText(&target, &result);
    if (r < 0) {                            /* -1 or SAPID_CPP_NOT_SUPPORTED. */
        return nonRegion;
    }

    if (r == 0) {                           /* бʸ */
        regionC.offset = result.start;
        regionC.length = result.end - result.start + 1;
    } else {                                /* ޥʳ */
        int         ilen;
        SpdString   ibuf;
        char        cname[FILENAME_MAX + 1];
    
        ilen = spdReadFileInBuf(regionI.fileId, &ibuf);

        regionC.offset = cppCharNotInMacroItoC(ifilename, regionI.offset,
                                               ibuf, ilen, cname);
        regionC.length = cppCharNotInMacroItoC(ifilename,
                                               regionI.offset + regionI.length,
                                               ibuf, ilen, cname);

        if (regionC.offset < 0 || regionC.length < 0) {
            regionC = nonRegion;
        } else {
            regionC.length = regionC.length - regionC.offset;
        }

        spdFree(ibuf);
    }
    
    return regionC;
}

SpdRegion spdConvRegionItoCbyPIDB(SpdRegion regionI)
{
    SpdRegion regionC;
    int block_offset, block_length, block_sort;
    SpdRel rel;
    Block_Table block_table_one;
    SpdObjId blockId1, blockId2;

    /* P-model ɤ߹Ǥʤ硢region  *.c ξ硢Τޤ֤ */
    if (spdGetOption("mode") == SAPID_I_MODE ||
        spdGetClassId(regionI.fileId) == SPD_CID_PFILE) {
        return regionI;
    }

    block_table_one = spdGetBlockTable(regionI.fileId, regionI.offset);
    if (regionI.offset < block_table_one.offset ||
        regionI.offset > block_table_one.offset+block_table_one.length) {
        return nonRegion;
    }
    blockId1 = block_table_one.id;
    rel = spdGetARel(block_table_one.id, "file_block", "c-id");
    regionC.fileId = rel.objectId;
    block_offset = spdGetOffset(rel.relationId);
    block_sort = spdGetAttrValInt(block_table_one.id, "sort");
    switch (block_sort) {
      case PBLOCK_NORMAL:
        regionC.offset = block_offset+(regionI.offset-block_table_one.offset);
        break;
      case PBLOCK_MACROCALL:
        /* ޥƽФγϰ */
        regionC.offset = block_offset;
        break;
      default:
        /* бregionʤ */
        return nonRegion;
        break;
    }

    block_table_one = spdGetBlockTable(regionI.fileId,
                                       regionI.offset + regionI.length);
    if ((regionI.offset + regionI.length < block_table_one.offset) ||
        (regionI.offset + regionI.length
         > block_table_one.offset + block_table_one.length)) {
        return nonRegion;
    }
    blockId2 = block_table_one.id;
    rel = spdGetARel(block_table_one.id, "file_block", "c-id");
    /* ϤȽλ֤ǥե뤬ۤʤȤregionʤ */
    if (regionC.fileId != rel.objectId) {
        return nonRegion;
    }
    block_offset = spdGetOffset(rel.relationId);
    block_sort = spdGetAttrValInt(block_table_one.id, "sort");
    switch (block_sort) {
      case PBLOCK_NORMAL:
        regionC.length = block_offset+(regionI.offset + regionI.length
                                       - block_table_one.offset);
        break;
      case PBLOCK_MACROCALL:
        /* ޥƽФνλ */
        block_length = spdGetLength(block_table_one.id);
        regionC.length = block_offset + block_length;

        /* macroCall ǤƱ block regionǤ PIDB ȤäѴ */
        if (blockId1 == blockId2) {
            return spdConvRegionItoCbyPIDB_in(regionI);
        }
        break;
      default:
        return nonRegion;
        break;
    }

    regionC.length -= regionC.offset;
    return regionC;
}

/* region  *.c  region ֤
   P-model ɤ߹ǤʤϡΤޤ֤ */

SpdRegion spdConvRegionItoC(SpdRegion regionI)
{
    SpdRegion regionC;
    int block_offset, block_length, block_sort;
    SpdRel rel;
    Block_Table block_table_one;
    /* Deleted by S.Yamamoto 1998/05/23. */
    /* SpdObjId blockId1, blockId2; */

    /* P-model ɤ߹Ǥʤ硢region  *.c ξ硢Τޤ֤ */
    if (spdGetOption("mode") == SAPID_I_MODE ||
        spdGetClassId(regionI.fileId) == SPD_CID_PFILE) {
        return regionI;
    }

    block_table_one = spdGetBlockTable(regionI.fileId, regionI.offset);
    if (regionI.offset < block_table_one.offset ||
        regionI.offset > block_table_one.offset+block_table_one.length) {
        return nonRegion;
    }
    /* Deleted by S.Yamamoto 1998/05/23. */
    /* blockId1 = block_table_one.id; */
    rel = spdGetARel(block_table_one.id, "file_block", "c-id");
    regionC.fileId = rel.objectId;
    block_offset = spdGetOffset(rel.relationId);
    block_sort = spdGetAttrValInt(block_table_one.id, "sort");
    switch (block_sort) {
      case PBLOCK_NORMAL:
        regionC.offset = block_offset+(regionI.offset-block_table_one.offset);
        break;
      case PBLOCK_MACROCALL:
        /* ޥƽФγϰ */
        regionC.offset = block_offset;
        break;
      default:
        /* бregionʤ */
        return nonRegion;
        break;
    }

    block_table_one = spdGetBlockTable(regionI.fileId,
                                       regionI.offset + regionI.length);
    if ((regionI.offset + regionI.length < block_table_one.offset) ||
        (regionI.offset + regionI.length
         > block_table_one.offset + block_table_one.length)) {
        return nonRegion;
    }
    /* Deleted by S.Yamamoto 1998/05/23. */
    /* blockId2 = block_table_one.id; */
    rel = spdGetARel(block_table_one.id, "file_block", "c-id");
    /* ϤȽλ֤ǥե뤬ۤʤȤregionʤ */
    if (regionC.fileId != rel.objectId) {
        return nonRegion;
    }
    block_offset = spdGetOffset(rel.relationId);
    block_sort = spdGetAttrValInt(block_table_one.id, "sort");
    switch (block_sort) {
      case PBLOCK_NORMAL:
        regionC.length = block_offset+(regionI.offset + regionI.length
                                       - block_table_one.offset);
        break;
      case PBLOCK_MACROCALL:
        /* ޥƽФνλ */
        block_length = spdGetLength(block_table_one.id);
        regionC.length = block_offset + block_length;
        break;
      default:
        return nonRegion;
        break;
    }

    regionC.length -= regionC.offset;
    return regionC;
}

/* region  *.i  region ֤ */

SpdRegion spdConvRegionCtoI(SpdRegion regionC)
{
    SpdRegion regionI;
    Block_Table block_table_one;
    int block_sort, block_offset, block_length;
    SpdRel rel;
    SpdObjId pfile_id;

    /* region  *.i ξ硢Τޤ֤
       P-model ɤ߹ǤʤϡΤޤ֤ */
    if (spdGetOption("mode") == SAPID_I_MODE ||
        spdGetClassId(regionC.fileId) == SPD_CID_FILE) {
        return regionC;
    }

    block_table_one = spdGetBlockTable(regionC.fileId, regionC.offset);
    rel = spdGetARel(block_table_one.id, "file_block", "c-id");
    block_offset = spdGetOffset(rel.relationId);
    block_sort = spdGetAttrValInt(block_table_one.id, "sort");

    switch (block_sort) {
      case PBLOCK_NORMAL:
        regionI.offset = block_table_one.offset + (regionC.offset
                                                   - block_offset);
        break;
      case PBLOCK_MACROCALL:
        /* Ÿ줿ޥƬ */
        regionI.offset = block_table_one.offset;
        break;
      default:
        /* бregionʤ */
        return nonRegion;
        break;
    }

    block_table_one = spdGetBlockTable(regionC.fileId,
                                       regionC.offset + regionC.length);
    rel = spdGetARel(block_table_one.id, "file_block", "c-id");
    block_offset = spdGetOffset(rel.relationId);
    block_sort = spdGetAttrValInt(block_table_one.id, "sort");

    switch (block_sort) {
      case PBLOCK_NORMAL:
        regionI.length = block_table_one.offset
            + (regionC.offset + regionC.length - block_offset);
        break;
      case PBLOCK_MACROCALL:
        /* Ÿ줿ޥνλ */
        block_length = spdGetLength(block_table_one.id);
        regionI.length = block_offset + block_length;
        break;
      default:
        /* бregionʤ */
        return nonRegion;
        break;
    }
    regionI.length -= regionI.offset;

    pfile_id = spdGetCfileIdByPfileId(regionC.fileId);
    regionI.fileId = spdGetARelObj(pfile_id, "cfile_ifile", "p-id");

    return regionI;
}

/* region եƥȤΤΤ
   ޽ѤδؿǤΤǡ塹ˤϡδؿ
   Ѥʤ褦ѹ뤳Ȥ˾ޤ
   ϥ桼 */

char *spdGetTextByRegion(SpdRegion region)
{
    int size;
    char *buf = NULL;
    char *text;
    int i;

    if (region.fileId == SPD_NON_ID) return NULL;

    size = spdReadFileInBuf(region.fileId, &buf);

    if (buf == NULL)
        return NULL;

    if (region.offset < 0 || region.offset >= size ||
        region.length <= 0 || region.offset + region.length > size) {
        spdFree(buf);
        return NULL;
    }

    text = spdMalloc(region.length + 1);
    for (i = 0; i < region.length; i++) {
        text[i] = buf[region.offset + i];
    }
    text[i] = '\0';
    spdFree(buf);
    return text;
}

/* "*.i"offsetإåե˴ޤޤ뤫ݤ */

SpdBoolean spdIsInHeaderFile(SpdObjId fileId, int offset)
{
    SpdObjId blockId, pfileId, hfileId;

    blockId = spdGetABlockId(fileId, offset);
    pfileId = spdGetARelObj(blockId, "file_block", "c-id");
    hfileId = spdGetARelObj(pfileId, "hfile_file", "c-id");
    if (hfileId == SPD_NON_ID) {
        /* hfile ¸ߤʤΤǥإåեǤϤʤ */
        return FALSE;
    } else {
        return TRUE;
    }
}

/* *.i  offset ޤ pfileId  */

SpdObjId spdGetPfileIdIncludingIoffset(SpdObjId fileId, int offset)
{
    SpdObjId blockId, pfileId;

    blockId = spdGetABlockId(fileId, offset);
    pfileId = spdGetARelObj(blockId, "file_block", "c-id");

    return pfileId;
}

/* *.i  region (Ǥ)ޤޤ pblock Τ
   sort פΤ ID 򤹤٤֤
   sort  SPD_NON_ID ꤹƤ pblock ֤
   βɬ */

SpdObjIdArray spdGetBlockIdArrayOfRegion(SpdRegion region, int sort)
{
    SpdObjIdArray   blockArray = SPD_EMPTY_ARRAY;
    SpdObjId        start_blockId, end_blockId;
    SpdObjId        i;
    int             block_sort;

    if (region.fileId == SPD_NON_ID || region.offset == SPD_NON_OFFSET ||
        region.length <= 0) {
        return (blockArray);
    }

    start_blockId = spdGetABlockId(region.fileId, region.offset);
    end_blockId = spdGetABlockId(region.fileId, region.offset + region.length);
    if (start_blockId == SPD_NON_ID ||
        end_blockId == SPD_NON_ID ||
        start_blockId > end_blockId) {
        return (blockArray);
    }

    for (i = start_blockId; i < end_blockId; i++) {
        block_sort = spdGetAttrValInt(i, "sort");
        if (sort == SPD_NON_ID || block_sort == sort) {
            blockArray.id = spdRealloc(blockArray.id,
                                       sizeof(SpdObjId) * (blockArray.size + 1));
            blockArray.id[blockArray.size] = i;
            blockArray.size++;
        }
    }
    return (blockArray);
}

/* "*.i"regionޤpblockId֤βɬס */

SpdObjIdArray spdGetBlockIdArrayIncludingRegion(SpdRegion region)
{
    return spdGetBlockIdArrayOfRegion(region, SPD_NON_ID);
}

/* "*.i"region(Ǥ)ޤޤޥƽФpblockId֤
   βɬס */

SpdObjIdArray spdGetBlockIdArrayOfMcall(SpdRegion region)
{
    return spdGetBlockIdArrayOfRegion(region, PBLOCK_MACROCALL);
}
