/* 
 * Program:             $RCSfile: class_table.c,v $  $Revision: 30.37 $
 * 
 * Purpose:             SDB schema.
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: class_table.c,v 30.37 2003/04/07 10:53:17 yamamoto Exp yamamoto $";
#endif

#ifdef SAPID_AR5_KERNEL
#undef SAPID_AR5_KERNEL
#define SAPID_AR4_KERNEL
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* 
 * Class and Relations for C Language.
 */

Class_t class_table[] = {
    /* I-model classes. */
    {"program", /* 0 */
     SPD_CID_PROGRAM,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"file", /* 1 */
     SPD_CID_FILE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"declaration", /* 2 */
     SPD_CID_DECLARATION,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, -1},
     3,
     {"id", "length", "sort", "storage"},
     -1,
     {NULL}
    },
    {"declarator", /* 3 */
     SPD_CID_DECLARATOR,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, 1, 2, -1},
     2,
     {"id", "length", "sort"},
     -1,
     {NULL}
    },
    {"identifier", /* 4 */
     SPD_CID_IDENTIFIER,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_O_,
     {0, 1, 2, 3, SVAL(0), -1},
     3,
     {"id", "length", "sort", "scope"},
     0,
     {"name"}
    },
    {"optionaldecl", /* 5 */
     SPD_CID_OPTIONALDECL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, 1, 2, -1},
     2,
     {"id", "length", "sort"},
     -1,
     {NULL}
    },
    {"block", /* 6 */
     SPD_CID_BLOCK,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, 1, 2, -1},
     2,
     {"id", "length", "sort"},
     -1,
     {NULL}
    },
    {"expression", /* 7 */
     SPD_CID_EXPRESSION,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, -1},
     3,
     {"id", "length", "sort", "lrtype"},
     -1,
     {NULL}
    },
    {"label", /* 8 */
     SPD_CID_LABEL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_O_,
     {0, 1, 2, SVAL(0), -1},
     2,
     {"id", "length", "sort"},
     0,
     {"name"}
    },
    {"constant", /* 9 */
     SPD_CID_CONSTANT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_O_,
     {0, 1, 2, SVAL(0), -1},
     2,
     {"id", "length", "type"},
     0,
     {"value"}
    },
    {"type", /* 10 */
     SPD_CID_TYPE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_O_,
     {0, 1, 2, 3, 4, 5, SVAL(0), -1},
     5,
     {"id", "sort", "qualifier", "type", "size", "sign"},
     0,
     {"name"}
    },
    {"operator", /* 11 */
     SPD_CID_OPERATOR,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_O_,
     {0, 1, 2, -1},
     2,
     {"id", "length", "sort"},
     -1,
     {NULL}
    },
    {"library", /* 12 */
     SPD_CID_LIBRARY,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_I_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    /* I-model relationships. */
    {"blk_blk", /* 13 */
     SPD_RID_BLK_BLK,
     SPD_CID_BLOCK, SPD_CID_BLOCK,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "parent_block_id", "children_block_id", "offset"},
     -1,
     {NULL}
    },
    {"blk_decl", /* 14 */
     SPD_RID_BLK_DECL,
     SPD_CID_BLOCK, SPD_CID_DECLARATION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "parent_block_id", "children_declaration_id", "offset"},
     -1,
     {NULL}
    },
    {"blk_expr", /* 15 */
     SPD_RID_BLK_EXPR,
     SPD_CID_BLOCK, SPD_CID_EXPRESSION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "parent_block_id", "children_expression_id", "offset"},
     -1,
     {NULL}
    },
    {"blk_label", /* 16 */
     SPD_RID_BLK_LABEL,
     SPD_CID_BLOCK, SPD_CID_LABEL,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "block_id", "label_id", "offset"},
     -1,
     {NULL}
    },
    {"case_cexpr", /* 17 */
     SPD_RID_CASE_CEXPR,
     SPD_CID_LABEL, SPD_CID_EXPRESSION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "label_id", "expression_id", "offset"},
     -1,
     {NULL}
    },
    {"const_ref", /* 18 */
     SPD_RID_CONST_REF,
     SPD_CID_EXPRESSION, SPD_CID_CONSTANT,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "expression_id", "const_id", "offset"},
     -1,
     {NULL}
    },
    {"decl_decl", /* 19 */
     SPD_RID_DECL_DECL,
     SPD_CID_DECLARATION, SPD_CID_DECLARATOR,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "declaration_id", "declarator_id", "offset"},
     -1,
     {NULL}
    },
    {"decl_ident", /* 20 */
     SPD_RID_DECL_IDENT,
     SPD_CID_DECLARATOR, SPD_CID_IDENTIFIER,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "declarator_id", "identifier_id", "offset"},
     -1,
     {NULL}
    },
    {"decl_opt", /* 21 */
     SPD_RID_DECL_OPT,
     SPD_CID_DECLARATOR, SPD_CID_OPTIONALDECL,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "declarator_id", "optionaldecl_id", "offset"},
     -1,
     {NULL}
    },
    {"decl_type", /* 22 */
     SPD_RID_DECL_TYPE,
     SPD_CID_DECLARATION, SPD_CID_TYPE,
     SPD_MID_I_MODEL,
     C_TYP_R_,
     {1, 2, 3, -1},
     3,
     {"id", "declaration_id", "type_id", "offset"},
     -1,
     {NULL}
    },
    {"expr_expr", /* 23 */
     SPD_RID_EXPR_EXPR,
     SPD_CID_EXPRESSION, SPD_CID_EXPRESSION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "outer_expression_id", "inner_expression_id", "offset"},
     -1,
     {NULL}
    },
    {"expr_op", /* 24 */
     SPD_RID_EXPR_OP,
     SPD_CID_EXPRESSION, SPD_CID_OPERATOR,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "expression_id", "operator_id", "offset"},
     -1,
     {NULL}
    },
    {"expr_type", /* 25 */
     SPD_RID_EXPR_TYPE,
     SPD_CID_EXPRESSION, SPD_CID_TYPE,
     SPD_MID_I_MODEL,
     C_TYP_R_,
     {1, 2, -1},
     2,
     {"id", "expression_id", "type_id"},
     -1,
     {NULL}
    },
    {"file_decl", /* 26 */
     SPD_RID_FILE_DECL,
     SPD_CID_FILE, SPD_CID_DECLARATION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "file_id", "declaration_id", "offset"},
     -1,
     {NULL}
    },
    {"func_body", /* 27 */
     SPD_RID_FUNC_BODY,
     SPD_CID_OPTIONALDECL, SPD_CID_BLOCK,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "function_id", "block_id", "offset"},
     -1,
     {NULL}
    },
    {"func_parm", /* 28 */
     SPD_RID_FUNC_PARM,
     SPD_CID_OPTIONALDECL, SPD_CID_DECLARATION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "optionaldecl_id", "declaration_id", "offset"},
     -1,
     {NULL}
    },
    {"goto_label", /* 29 */
     SPD_RID_GOTO_LABEL,
     SPD_CID_EXPRESSION, SPD_CID_LABEL,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "expression_id", "label_id", "offset"},
     -1,
     {NULL}
    },
    {"ident_ref", /* 30 */
     SPD_RID_IDENT_REF,
     SPD_CID_EXPRESSION, SPD_CID_IDENTIFIER,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "expression_id", "identifier_id", "offset"},
     -1,
     {NULL}
    },
    {"ident_type", /* 31 */
     SPD_RID_IDENT_TYPE,
     SPD_CID_IDENTIFIER, SPD_CID_TYPE,
     SPD_MID_I_MODEL,
     C_TYP_R_,
     {1, 2, -1},
     2,
     {"id", "identifier_id", "type_id"},
     -1,
     {NULL}
    },
    {"init_expr", /* 32 */
     SPD_RID_INIT_EXPR,
     SPD_CID_OPTIONALDECL, SPD_CID_EXPRESSION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "optionaldecl_id", "expression_id", "offset"},
     -1,
     {NULL}
    },
    {"op_type", /* 33 */
     SPD_RID_OP_TYPE,
     SPD_CID_OPERATOR, SPD_CID_TYPE,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operator_id", "type_id"},
     -1,
     {NULL}
    },
    {"prog_file", /* 34 */
     SPD_RID_PROG_FILE,
     SPD_CID_PROGRAM, SPD_CID_FILE,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "program_id", "file_id"},
     -1,
     {NULL}
    },
    {"struct_decl", /* 35 */
     SPD_RID_STRUCT_DECL,
     SPD_CID_OPTIONALDECL, SPD_CID_DECLARATION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "optionaldecl_id", "declaration_id", "offset"},
     -1,
     {NULL}
    },
    {"type_cexpr", /* 36 */
     SPD_RID_TYPE_CEXPR,
     SPD_CID_TYPE, SPD_CID_EXPRESSION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "type_id", "expression_id", "offset"},
     -1,
     {NULL}
    },
    {"type_is", /* 37 */
     SPD_RID_TYPE_IS,
     SPD_CID_TYPE, SPD_CID_IDENTIFIER,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "type_id", "identifier_id"},
     -1,
     {NULL}
    },
    {"type_st", /* 38 */
     SPD_RID_TYPE_ST,
     SPD_CID_TYPE, SPD_CID_TYPE,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "type_pointer_id", "type_pointed_id"},
     -1,
     {NULL}
    },
    {"prot_parm", /* 39 */
     SPD_RID_PROT_PARM,
     SPD_CID_OPTIONALDECL, SPD_CID_DECLARATION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "optionaldecl_id", "declaration_id", "offset"},
     -1,
     {NULL}
    },
    {"funcptr_parm", /* 40 */
     SPD_RID_FUNCPTR_PARM,
     SPD_CID_OPTIONALDECL, SPD_CID_DECLARATION,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "optionaldecl_id", "declaration_id", "offset"},
     -1,
     {NULL}
    },
    {"decr_type", /* 41 */
     SPD_RID_DECR_TYPE,
     SPD_CID_DECLARATOR, SPD_CID_TYPE,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "declarator_id", "type_id"},
     -1,
     {NULL}
    },
    {"prog_lib", /* 42 */
     SPD_RID_PROG_LIB,
     SPD_CID_PROGRAM, SPD_CID_LIBRARY,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "program_id", "library_id"},
     -1,
     {NULL}
    },
    {"lib_file", /* 43 */
     SPD_RID_LIB_FILE,
     SPD_CID_LIBRARY, SPD_CID_FILE,
     SPD_MID_I_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "library_id", "file_id"},
     -1,
     {NULL}
    },
    /* P-model classes. */
    {"pfile", /* 44 */
     SPD_CID_PFILE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"pblock", /* 45 */
     SPD_CID_PBLOCK,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, 2, -1},
     2,
     {"id", "length", "sort"},
     -1,
     {NULL}
    },
    {"macroPattern", /* 46 */
     SPD_CID_MACROPATTERN,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "length"},
     -1,
     {NULL}
    },
    {"macroBody", /* 47 */
     SPD_CID_MACROBODY,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, SVAL(0), -1},
     1,
     {"id", "length"},
     0,
     {"name"}
    },
    {"macroName", /* 48 */
     SPD_CID_MACRONAME,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_O_,
     {0, 1, SVAL(0), -1},
     1,
     {"id", "length"},
     0,
     {"name"}
    },
    {"macroOptional", /* 49 */
     SPD_CID_MACROOPTIONAL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "length"},
     -1,
     {NULL}
    },
    {"macroDeclArg", /* 50 */
     SPD_CID_MACRODECLARG,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_O_,
     {0, 1, SVAL(0), -1},
     1,
     {"id", "length"},
     0,
     {"name"}
    },
    {"macroCall", /* 51 */
     SPD_CID_MACROCALL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "length"},
     -1,
     {NULL}
    },
    {"macroCallArg", /* 52 */
     SPD_CID_MACROCALLARG,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, SVAL(0), -1},
     1,
     {"id", "length"},
     0,
     {"name"}
    },
    {"hfile", /* 53 */
     SPD_CID_HFILE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, 2, SVAL(0), -1},
     2,
     {"id", "length", "sort"},
     0,
     {"name"}
    },
    {"iblock", /* 54 */
     SPD_CID_IBLOCK,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_P_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "length"},
     -1,
     {NULL}
    },
    /* P-model relationships. */
    {"file_block", /* 55 */
     SPD_RID_FILE_BLOCK,
     SPD_CID_PFILE, SPD_CID_PBLOCK,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "pfile_id", "pblock_id", "offset"},
     -1,
     {NULL}
    },
    {"call_mcall", /* 56 */
     SPD_RID_CALL_MCALL,
     SPD_CID_PBLOCK, SPD_CID_MACROCALL,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "pblock_id", "macroCall_id", "offset"},
     -1,
     {NULL}
    },
    {"mbody_mcall", /* 57 */
     SPD_RID_MBODY_MCALL,
     SPD_CID_MACROBODY, SPD_CID_MACROCALL,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "macroBody_id", "macroCall_id", "offset"},
     -1,
     {NULL}
    },
    {"mcall_mname", /* 58 */
     SPD_RID_MCALL_MNAME,
     SPD_CID_MACROCALL, SPD_CID_MACRONAME,
     SPD_MID_P_MODEL,
     C_TYP_R_,
     {1, 2, -1},
     2,
     {"id", "macroCall_id", "macroName_id"},
     -1,
     {NULL}
    },
    {"mcall_mcalla", /* 59 */
     SPD_RID_MCALL_MCALLA,
     SPD_CID_MACROCALL, SPD_CID_MACROCALLARG,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "macroCall_id", "macroCallArg_id", "offset"},
     -1,
     {NULL}
    },
    {"def_mpatt", /* 60 */
     SPD_RID_DEF_MPATT,
     SPD_CID_PBLOCK, SPD_CID_MACROPATTERN,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "pblock_id", "macroPattern_id", "offset"},
     -1,
     {NULL}
    },
    {"def_mbody", /* 61 */
     SPD_RID_DEF_MBODY,
     SPD_CID_PBLOCK, SPD_CID_MACROBODY,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "pblock_id", "macroBody_id", "offset"},
     -1,
     {NULL}
    },
    {"mpatt_mname", /* 62 */
     SPD_RID_MPATT_MNAME,
     SPD_CID_MACROPATTERN, SPD_CID_MACRONAME,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "macroPattern_id", "macroName_id", "offset"},
     -1,
     {NULL}
    },
    {"mpatt_mopt", /* 63 */
     SPD_RID_MPATT_MOPT,
     SPD_CID_MACROPATTERN, SPD_CID_MACROOPTIONAL,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "macroPattern_id", "macroOptional_id", "offset"},
     -1,
     {NULL}
    },
    {"mopt_mdecla", /* 64 */
     SPD_RID_MOPT_MDECLA,
     SPD_CID_MACROOPTIONAL, SPD_CID_MACRODECLARG,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "macroOptional_id", "macroDeclArg_id", "offset"},
     -1,
     {NULL}
    },
    {"mbody_mdecla", /* 65 */
     SPD_RID_MBODY_MDECLA,
     SPD_CID_MACROBODY, SPD_CID_MACRODECLARG,
     SPD_MID_P_MODEL,
     C_TYP_R_,
     {1, 2, 3, 4, 5, -1},
     5,
     {"id", "macroBody_id", "macroDeclArg_id", "order", "skip", "sharp"},
     -1,
     {NULL}
    },
    {"incld_hfile", /* 66 */
     SPD_RID_INCLD_HFILE,
     SPD_CID_PBLOCK, SPD_CID_HFILE,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "pblock_id", "hfile_id", "offset"},
     -1,
     {NULL}
    },
    {"hfile_file", /* 67 */
     SPD_RID_HFILE_FILE,
     SPD_CID_HFILE, SPD_CID_PFILE,
     SPD_MID_P_MODEL,
     C_TYP_R_,
     {1, 2, -1},
     2,
     {"id", "hfile_id", "pfile_id"},
     -1,
     {NULL}
    },
    {"mcalla_mcall", /* 68 */
     SPD_RID_MCALLA_MCALL,
     SPD_CID_MACROCALLARG, SPD_CID_MACROCALL,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "macroCallArg_id", "macroCall_id", "offset"},
     -1,
     {NULL}
    },
    {"cfile_ifile", /* 69 */
     SPD_RID_CFILE_IFILE,
     SPD_CID_PFILE, SPD_CID_FILE,
     SPD_MID_P_MODEL,
     C_TYP_R_,
     {1, 2, -1},
     2,
     {"id", "pfile_id", "file_id"},
     -1,
     {NULL}
    },
    {"ifile_iblock", /* 70 */
     SPD_RID_IFILE_IBLOCK,
     SPD_CID_FILE, SPD_CID_IBLOCK,
     SPD_MID_P_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "file_id", "iblock_id", "offset"},
     -1,
     {NULL}
    },
    {"pblock_iblock", /* 71 */
     SPD_RID_PBLOCK_IBLOCK,
     SPD_CID_PBLOCK, SPD_CID_IBLOCK,
     SPD_MID_P_MODEL,
     C_TYP_R_,
     {1, 2, -1},
     2,
     {"id", "pblock_id", "iblock_id"},
     -1,
     {NULL}
    },
    /* Area-model classes. */
    {"storage", /* 72 */
     SPD_CID_STORAGE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_AREA_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"area", /* 73 */
     SPD_CID_AREA,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_AREA_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, -1},
     4,
     {"id", "sort", "address", "size", "indirect"},
     -1,
     {NULL}
    },
    {"value", /* 74 */
     SPD_CID_VALUE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_AREA_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, -1},
     3,
     {"id", "sort", "type", "value"},
     -1,
     {NULL}
    },
    /* Area-model relationships. */
    {"storage_storage", /* 75 */
     SPD_RID_STORAGE_STORAGE,
     SPD_CID_STORAGE, SPD_CID_STORAGE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "p_storage_id", "c_storage_id"},
     -1,
     {NULL}
    },
    {"storage_area", /* 76 */
     SPD_RID_STORAGE_AREA,
     SPD_CID_STORAGE, SPD_CID_AREA,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "storage_id", "area_id"},
     -1,
     {NULL}
    },
    {"area_area", /* 77 */
     SPD_RID_AREA_AREA,
     SPD_CID_AREA, SPD_CID_AREA,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "parent_area_id", "child_area_id", "order"},
     -1,
     {NULL}
    },
    {"area_value", /* 78 */
     SPD_RID_AREA_VALUE,
     SPD_CID_AREA, SPD_CID_VALUE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "area_id", "value_id", "order"},
     -1,
     {NULL}
    },
    {"point_to", /* 79 */
     SPD_RID_POINT_TO,
     SPD_CID_VALUE, SPD_CID_AREA,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "pointer_id", "area_id"},
     -1,
     {NULL}
    },
    {"funcptr_to", /* 80 */
     SPD_RID_FUNCPTR_TO,
     SPD_CID_VALUE, SPD_CID_IDENTIFIER,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "pointer_id", "function_id"},
     -1,
     {NULL}
    },
    {"prog_storage", /* 81 */
     SPD_RID_PROG_STORAGE,
     SPD_CID_PROGRAM, SPD_CID_STORAGE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "program_id", "storage_id", "order"},
     -1,
     {NULL}
    },
    {"func_storage", /* 82 */
     SPD_RID_FUNC_STORAGE,
     SPD_CID_OPTIONALDECL, SPD_CID_STORAGE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "optdecl_id", "storage_id", "order"},
     -1,
     {NULL}
    },
    {"block_storage", /* 83 */
     SPD_RID_BLOCK_STORAGE,
     SPD_CID_BLOCK, SPD_CID_STORAGE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "block_id", "storage_id", "order"},
     -1,
     {NULL}
    },
    {"variable_area", /* 84 */
     SPD_RID_VARIABLE_AREA,
     SPD_CID_IDENTIFIER, SPD_CID_AREA,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "variable_id", "area_id", "order"},
     -1,
     {NULL}
    },
    {"member_area", /* 85 */
     SPD_RID_MEMBER_AREA,
     SPD_CID_MEMBER, SPD_CID_AREA,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "member_id", "area_id"},
     -1,
     {NULL}
    },
    {"string_area", /* 86 */
     SPD_RID_STRING_AREA,
     SPD_CID_CONSTANT, SPD_CID_AREA,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "string_id", "area_id"},
     -1,
     {NULL}
    },
    {"area_type", /* 87 */
     SPD_RID_AREA_TYPE,
     SPD_CID_AREA, SPD_CID_TYPE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "area_id", "type_id"},
     -1,
     {NULL}
    },
    {"value_type", /* 88 */
     SPD_RID_VALUE_TYPE,
     SPD_CID_VALUE, SPD_CID_TYPE,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "value_id", "type_id"},
     -1,
     {NULL}
    },
    {"value_enum", /* 89 */
     SPD_RID_VALUE_ENUM,
     SPD_CID_VALUE, SPD_CID_ENUMERATOR,
     SPD_MID_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "value_id", "enum_id"},
     -1,
     {NULL}
    },
    /* B-model classes. */
    {"execution", /* 90 */
     SPD_CID_EXECUTION,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_B_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"operation", /* 91 */
     SPD_CID_OPERATION,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_B_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    /* B-model relationships. */
    {"prog_exec", /* 92 */
     SPD_RID_PROG_EXEC,
     SPD_CID_PROGRAM, SPD_CID_EXECUTION,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "program_id", "execution_id", "order"},
     -1,
     {NULL}
    },
    {"exec_op", /* 93 */
     SPD_RID_EXEC_OP,
     SPD_CID_EXECUTION, SPD_CID_OPERATION,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "execution_id", "operation_id", "order"},
     -1,
     {NULL}
    },
    {"statement_op", /* 94 */
     SPD_RID_STATEMENT_OP,
     SPD_NON_CID, SPD_CID_OPERATION,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "statement_id", "operation_id", "order"},
     -1,
     {NULL}
    },
    {"alloc_storage", /* 95 */
     SPD_RID_ALLOC_STORAGE,
     SPD_CID_OPERATION, SPD_CID_STORAGE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "storage_id"},
     -1,
     {NULL}
    },
    {"free_storage", /* 96 */
     SPD_RID_FREE_STORAGE,
     SPD_CID_OPERATION, SPD_CID_STORAGE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "storage_id"},
     -1,
     {NULL}
    },
    {"define_value", /* 97 */
     SPD_RID_DEFINE_VALUE,
     SPD_CID_OPERATION, SPD_CID_VALUE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "value_id"},
     -1,
     {NULL}
    },
    {"refer_value", /* 98 */
     SPD_RID_REFER_VALUE,
     SPD_CID_OPERATION, SPD_CID_VALUE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "value_id"},
     -1,
     {NULL}
    },
    {"call_value", /* 99 */
     SPD_RID_CALL_VALUE,
     SPD_CID_OPERATION, SPD_CID_VALUE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "value_id"},
     -1,
     {NULL}
    },
    {"return_value", /* 100 */
     SPD_RID_RETURN_VALUE,
     SPD_CID_OPERATION, SPD_CID_VALUE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "value_id"},
     -1,
     {NULL}
    },
    {"branch_value", /* 101 */
     SPD_RID_BRANCH_VALUE,
     SPD_CID_OPERATION, SPD_CID_VALUE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "operation_id", "value_id"},
     -1,
     {NULL}
    },
    {"depend_on", /* 102 */
     SPD_RID_DEPEND_ON,
     SPD_CID_VALUE, SPD_CID_VALUE,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "value_id", "depend_value_id"},
     -1,
     {NULL}
    },
    {"call_func", /* 103 */
     SPD_RID_CALL_FUNC,
     SPD_CID_OPERATION, SPD_CID_IDENTIFIER,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "call_id", "func_id"},
     -1,
     {NULL}
    },
    {"call_return", /* 104 */
     SPD_RID_CALL_RETURN,
     SPD_CID_OPERATION, SPD_CID_OPERATION,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "call_id", "return_id"},
     -1,
     {NULL}
    },
    {"entry_exit", /* 105 */
     SPD_RID_ENTRY_EXIT,
     SPD_CID_OPERATION, SPD_CID_OPERATION,
     SPD_MID_B_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "entry_id", "exit_id"},
     -1,
     {NULL}
    },
    /* D-model classes. */
    {"software", /* 106 */
     SPD_CID_SOFTWARE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"document", /* 107 */
     SPD_CID_DOCUMENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"docfile", /* 108 */
     SPD_CID_DOCFILE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"name"}
    },
    {"docfrag", /* 109 */
     SPD_CID_DOCFRAG,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, 1, 2, SVAL(0), -1},
     2,
     {"id", "length", "sort"},
     0,
     {"name"}
    },
    {"textstring", /* 110 */
     SPD_CID_TEXTSTRING,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, 1, SVAL(0), -1},
     1,
     {"id", "length"},
     0,
     {"text"}
    },
    {"attribute", /* 111 */
     SPD_CID_ATTRIBUTE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, 1, SVAL(0), -1},
     1,
     {"id", "length"},
     0,
     {"name"}
    },
    {"docident", /* 112 */
     SPD_CID_DOCIDENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_D_MODEL,
     C_TYP_OC,
     {0, 1, 2, SVAL(0), -1},
     2,
     {"id", "length", "sort"},
     0,
     {"name"}
    },
    /* D-model relationships. */
    {"soft_doc", /* 113 */
     SPD_RID_SOFT_DOC,
     SPD_CID_SOFTWARE, SPD_CID_DOCUMENT,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "software_id", "document_id"},
     -1,
     {NULL}
    },
    {"soft_prog", /* 114 */
     SPD_RID_SOFT_PROG,
     SPD_CID_SOFTWARE, SPD_CID_PROGRAM,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "software_id", "program_id"},
     -1,
     {NULL}
    },
    {"doc_docfile", /* 115 */
     SPD_RID_DOC_DOCFILE,
     SPD_CID_DOCUMENT, SPD_CID_DOCFILE,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "document_id", "docfile_id"},
     -1,
     {NULL}
    },
    {"docfile_docfrag", /* 116 */
     SPD_RID_DOCFILE_DOCFRAG,
     SPD_CID_DOCFILE, SPD_CID_DOCFRAG,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "docfile_id", "docfrag_id", "offset"},
     -1,
     {NULL}
    },
    {"docelem_docfrag", /* 117 */
     SPD_RID_DOCELEM_DOCFRAG,
     SPD_CID_DOCFRAG, SPD_CID_DOCFRAG,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "docelem_id", "docfrag_id", "offset"},
     -1,
     {NULL}
    },
    {"docelem_attribute", /* 118 */
     SPD_RID_DOCELEM_ATTRIBUTE,
     SPD_CID_DOCFRAG, SPD_CID_ATTRIBUTE,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "docelem_id", "attribute_id", "offset"},
     -1,
     {NULL}
    },
    {"attribute_value", /* 119 */
     SPD_RID_ATTRIBUTE_VALUE,
     SPD_CID_ATTRIBUTE, SPD_CID_TEXTSTRING,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "attribute_id", "value_id", "offset"},
     -1,
     {NULL}
    },
    {"textfrag_textstring", /* 120 */
     SPD_RID_TEXTFRAG_TEXTSTRING,
     SPD_CID_DOCFRAG, SPD_CID_TEXTSTRING,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "textfrag_id", "textstring_id", "offset"},
     -1,
     {NULL}
    },
    {"defelem_docident", /* 121 */
     SPD_RID_DEFELEM_DOCIDENT,
     SPD_CID_DOCFRAG, SPD_CID_DOCIDENT,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "defelem_id", "docident_id", "offset"},
     -1,
     {NULL}
    },
    {"identelem_docident", /* 122 */
     SPD_RID_IDENTELEM_DOCIDENT,
     SPD_CID_DOCFRAG, SPD_CID_DOCIDENT,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "identelem_id", "docident_id", "offset"},
     -1,
     {NULL}
    },
    {"srcelem_srcfrag", /* 123 */
     SPD_RID_SRCELEM_SRCFRAG,
     SPD_CID_DOCFRAG, SPD_CID_DECLARATION,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "srcelem_id", "srcfrag_id", "offset"},
     -1,
     {NULL}
    },
    {"docident_identifier", /* 124 */
     SPD_RID_DOCIDENT_IDENTIFIER,
     SPD_CID_DOCIDENT, SPD_CID_IDENTIFIER,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "docident_id", "identifier_id", "offset"},
     -1,
     {NULL}
    },
    {"linkelem_srcfrag", /* 125 */
     SPD_RID_LINKELEM_SRCFRAG,
     SPD_CID_DOCFRAG, SPD_CID_DECLARATION,
     SPD_MID_D_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "linkelem_id", "srcfrag_id"},
     -1,
     {NULL}
    },
    /* SDA-model classes. */
    {"sdafunc", /* 126 */
     SPD_CID_SDAFUNC,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_SDA_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"varg", /* 127 */
     SPD_CID_VARG,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_SDA_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    /* SDA-model relationships. */
    {"next_stmt", /* 128 */
     SPD_RID_NEXT_STMT,
     SPD_CID_STATEMENT, SPD_CID_STATEMENT,
     SPD_MID_SDA_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "prev_stmt_id", "next_stmt_id", "sort"},
     -1,
     {NULL}
    },
    {"data_depend", /* 129 */
     SPD_RID_DATA_DEPEND,
     SPD_CID_EXPRESSION, SPD_CID_EXPRESSION,
     SPD_MID_SDA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "def_expr_id", "ref_expr_id"},
     -1,
     {NULL}
    },
    {"func_sfunc", /* 130 */
     SPD_RID_FUNC_SFUNC,
     SPD_CID_DECLARATION, SPD_CID_SDAFUNC,
     SPD_MID_SDA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "func_id_", "sfunc_id"},
     -1,
     {NULL}
    },
    {"sfunc_varg", /* 131 */
     SPD_RID_SFUNC_VARG,
     SPD_CID_SDAFUNC, SPD_CID_VARG,
     SPD_MID_SDA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "sfunc_id", "varg_id"},
     -1,
     {NULL}
    },
    {"sfunc_stmt", /* 132 */
     SPD_RID_SFUNC_STMT,
     SPD_CID_SDAFUNC, SPD_CID_STATEMENT,
     SPD_MID_SDA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "sfunc_id", "stmt_id"},
     -1,
     {NULL}
    },
    {"ident_varg", /* 133 */
     SPD_RID_IDENT_VARG,
     SPD_CID_IDENTIFIER, SPD_CID_VARG,
     SPD_MID_SDA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "ident_id", "varg_id"},
     -1,
     {NULL}
    },
    /* Stmt-model classes. */
    {"statement", /* 134 */
     SPD_CID_STATEMENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_STMT_MODEL,
     C_TYP_OC,
     {0, 1, 2, -1},
     2,
     {"id", "length", "sort"},
     -1,
     {NULL}
    },
    /* Stmt-model relationships. */
    {"for_expr", /* 135 */
     SPD_RID_FOR_EXPR,
     SPD_CID_STATEMENT, SPD_CID_EXPRESSION,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "state_id", "expr_id", "offset"},
     -1,
     {NULL}
    },
    {"cond_expr", /* 136 */
     SPD_RID_COND_EXPR,
     SPD_CID_STATEMENT, SPD_CID_EXPRESSION,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "state_id", "expr_id", "offset"},
     -1,
     {NULL}
    },
    {"iter_expr", /* 137 */
     SPD_RID_ITER_EXPR,
     SPD_CID_STATEMENT, SPD_CID_EXPRESSION,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "state_id", "expr_id", "offset"},
     -1,
     {NULL}
    },
    {"stmt_expr", /* 138 */
     SPD_RID_STMT_EXPR,
     SPD_CID_STATEMENT, SPD_CID_EXPRESSION,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "state_id", "expr_id", "offset"},
     -1,
     {NULL}
    },
    {"stmt_label", /* 139 */
     SPD_RID_STMT_LABEL,
     SPD_CID_STATEMENT, SPD_CID_LABEL,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "state_id", "label_id", "offset"},
     -1,
     {NULL}
    },
    {"return_expr", /* 140 */
     SPD_RID_RETURN_EXPR,
     SPD_CID_STATEMENT, SPD_CID_EXPRESSION,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "return_id", "expr_id", "offset"},
     -1,
     {NULL}
    },
    {"block_decl", /* 141 */
     SPD_RID_BLOCK_DECL,
     SPD_CID_STATEMENT, SPD_CID_DECLARATION,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "block_id", "decl_id", "offset"},
     -1,
     {NULL}
    },
    {"block_stmt", /* 142 */
     SPD_RID_BLOCK_STMT,
     SPD_CID_STATEMENT, SPD_CID_STATEMENT,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "block_id", "state_id", "offset"},
     -1,
     {NULL}
    },
    {"control_body", /* 143 */
     SPD_RID_CONTROL_BODY,
     SPD_CID_STATEMENT, SPD_CID_STATEMENT,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "ctrl_id", "body_id", "offset"},
     -1,
     {NULL}
    },
    {"else_body", /* 144 */
     SPD_RID_ELSE_BODY,
     SPD_CID_STATEMENT, SPD_CID_STATEMENT,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "ctrl_id", "body_id", "offset"},
     -1,
     {NULL}
    },
    {"stmt_iblock", /* 145 */
     SPD_RID_STMT_IBLOCK,
     SPD_CID_STATEMENT, SPD_CID_BLOCK,
     SPD_MID_STMT_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "block_id", "iblock_id"},
     -1,
     {NULL}
    },
    /* SDAarea-model classes. */
    {"sArea", /* 146 */
     SPD_CID_SAREA,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_SDA_AREA_MODEL,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    /* SDAarea-model relationships. */
    {"ident_area", /* 147 */
     SPD_RID_IDENT_SAREA,
     SPD_CID_IDENTIFIER, SPD_CID_SAREA,
     SPD_MID_SDA_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "ident_id", "area_id"},
     -1,
     {NULL}
    },
    {"consist_of", /* 148 */
     SPD_RID_CONSIST_OF,
     SPD_CID_SAREA, SPD_CID_SAREA,
     SPD_MID_SDA_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "p_area_id", "c_area_id"},
     -1,
     {NULL}
    },
    {"sarea_sarea", /* 149 */
     SPD_RID_SAREA_SAREA,
     SPD_CID_SAREA, SPD_CID_SAREA,
     SPD_MID_SDA_AREA_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "p_area_id", "c_area_id"},
     -1,
     {NULL}
    },
    /* Type-view classes. */
    {"member", /* 150 */
     SPD_CID_MEMBER,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_TYPE_VIEW,
     C_TYP_OC,
     {0, -1},
     0,
     {"id"},
     -1,
     {NULL}
    },
    {"enumerator", /* 151 */
     SPD_CID_ENUMERATOR,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_TYPE_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "value"},
     -1,
     {NULL}
    },
    /* Type-view relationships. */
    {"struct_member", /* 152 */
     SPD_RID_STRUCT_MEMBER,
     SPD_CID_TYPE, SPD_CID_MEMBER,
     SPD_MID_TYPE_VIEW,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "struct_id", "member_id", "order"},
     -1,
     {NULL}
    },
    {"member_type", /* 153 */
     SPD_RID_MEMBER_TYPE,
     SPD_CID_MEMBER, SPD_CID_TYPE,
     SPD_MID_TYPE_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "member_id", "type_id"},
     -1,
     {NULL}
    },
    {"member_name", /* 154 */
     SPD_RID_MEMBER_NAME,
     SPD_CID_MEMBER, SPD_CID_IDENTIFIER,
     SPD_MID_TYPE_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "member_id", "identifier_id"},
     -1,
     {NULL}
    },
    {"enum_enumerator", /* 155 */
     SPD_RID_ENUM_ENUMERATOR,
     SPD_CID_TYPE, SPD_CID_ENUMERATOR,
     SPD_MID_TYPE_VIEW,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "enum_id", "enumerator_id", "order"},
     -1,
     {NULL}
    },
    {"enumerator_name", /* 156 */
     SPD_RID_ENUMERATOR_NAME,
     SPD_CID_ENUMERATOR, SPD_CID_IDENTIFIER,
     SPD_MID_TYPE_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "enumerator_id", "identifier_id"},
     -1,
     {NULL}
    },
    {"array_num", /* 157 */
     SPD_RID_ARRAY_NUM,
     SPD_CID_TYPE, SPD_CID_CONSTANT,
     SPD_MID_TYPE_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "array_id", "constant_id"},
     -1,
     {NULL}
    },
    /* Comment-view classes. */
    {"pcomment", /* 158 */
     SPD_CID_PCOMMENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_COMMENT_VIEW,
     C_TYP_OC,
     {0, 1, 2, 3, 4, 5, 6, SVAL(0), -1},
     6,
     {"id", "length", "lint_sort", "offset", "line_num_begin", "line_num_end", "offset_from_linehead"},
     0,
     {"str"}
    },
    {"lcomment", /* 159 */
     SPD_CID_LCOMMENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_COMMENT_VIEW,
     C_TYP_OC,
     {0, SVAL(0), -1},
     0,
     {"id"},
     0,
     {"str"}
    },
    /* Comment-view relationships. */
    {"pfile_pcomment", /* 160 */
     SPD_RID_PFILE_PCOMMENT,
     SPD_CID_PFILE, SPD_CID_PCOMMENT,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "pfile_id", "pcomment_id", "offset"},
     -1,
     {NULL}
    },
    {"pcomment_lcomment", /* 161 */
     SPD_RID_PCOMMENT_LCOMMENT,
     SPD_CID_PCOMMENT, SPD_CID_LCOMMENT,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "pcomment_id", "lcomment_id"},
     -1,
     {NULL}
    },
    {"lcomment_pfile", /* 162 */
     SPD_RID_LCOMMENT_PFILE,
     SPD_CID_LCOMMENT, SPD_CID_PFILE,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "lcomment_id", "pfile_id"},
     -1,
     {NULL}
    },
    {"lcomment_pblock", /* 163 */
     SPD_RID_LCOMMENT_PBLOCK,
     SPD_CID_LCOMMENT, SPD_CID_PBLOCK,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "lcomment_id", "pblock_id"},
     -1,
     {NULL}
    },
    {"lcomment_declaration", /* 164 */
     SPD_RID_LCOMMENT_DECLARATION,
     SPD_CID_LCOMMENT, SPD_CID_DECLARATION,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "lcomment_id", "declaration_id"},
     -1,
     {NULL}
    },
    {"lcomment_label", /* 165 */
     SPD_RID_LCOMMENT_LABEL,
     SPD_CID_LCOMMENT, SPD_CID_LABEL,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "lcomment_id", "label_id"},
     -1,
     {NULL}
    },
    {"lcomment_expression", /* 166 */
     SPD_RID_LCOMMENT_EXPRESSION,
     SPD_CID_LCOMMENT, SPD_CID_EXPRESSION,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "lcomment_id", "expression_id"},
     -1,
     {NULL}
    },
    {"lcomment_block", /* 167 */
     SPD_RID_LCOMMENT_BLOCK,
     SPD_CID_LCOMMENT, SPD_CID_BLOCK,
     SPD_MID_COMMENT_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "lcomment_id", "block_id"},
     -1,
     {NULL}
    },
    /* Flow-view classes. */
    {"blkEntry", /* 168 */
     SPD_CID_BLK_ENTRY,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"blkGoInto", /* 169 */
     SPD_CID_BLK_GO_INTO,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"blkExit", /* 170 */
     SPD_CID_BLK_EXIT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"blkBreak", /* 171 */
     SPD_CID_BLK_BREAK,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"blkContinue", /* 172 */
     SPD_CID_BLK_CONTINUE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"blkGoAway", /* 173 */
     SPD_CID_BLK_GO_AWAY,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"blkReturn", /* 174 */
     SPD_CID_BLK_RETURN,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    {"switchBranch", /* 175 */
     SPD_CID_SWITCH_BRANCH,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FLOW_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    /* Flow-view relationships. */
    {"next_node", /* 176 */
     SPD_RID_NEXT_NODE,
     SPD_CID_BLK_ENTRY, SPD_CID_BLK_ENTRY,
     SPD_MID_FLOW_VIEW,
     C_TYP_RC,
     {1, 2, 3, -1},
     3,
     {"id", "curId", "nextId", "sort"},
     -1,
     {NULL}
    },
    {"blk_sdanode", /* 177 */
     SPD_RID_BLK_SDANODE,
     SPD_CID_BLK_ENTRY, SPD_CID_BLK_ENTRY,
     SPD_MID_FLOW_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "blkId", "sdanodeId"},
     -1,
     {NULL}
    },
    {"data_dependence", /* 178 */
     SPD_RID_DATA_DEPENDENCE,
     SPD_CID_BLK_ENTRY, SPD_CID_EXPRESSION,
     SPD_MID_FLOW_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "defObjId", "refExprId"},
     -1,
     {NULL}
    },
    {"def_used", /* 179 */
     SPD_RID_DEF_USED,
     SPD_CID_BLK_ENTRY, SPD_CID_EXPRESSION,
     SPD_MID_FLOW_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "defObjId", "usedExprId"},
     -1,
     {NULL}
    },
    /* FCD-view classes. */
    {"fcdg", /* 180 */
     SPD_CID_FCDG,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_FCD_VIEW,
     C_TYP_OC,
     {0, 1, -1},
     1,
     {"id", "sort"},
     -1,
     {NULL}
    },
    /* FCD-view relationships. */
    {"func_fcdg", /* 181 */
     SPD_RID_FUNC_FCDG,
     SPD_CID_DECLARATION, SPD_CID_FCDG,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "func_id", "fcdg_id"},
     -1,
     {NULL}
    },
    {"libdata_depend", /* 182 */
     SPD_RID_LIBDATA_DEPEND,
     SPD_CID_FCDG, SPD_CID_FCDG,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "source_id", "target_id"},
     -1,
     {NULL}
    },
    {"libctrl_depend", /* 183 */
     SPD_RID_LIBCTRL_DEPEND,
     SPD_CID_FCDG, SPD_CID_FCDG,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "source_id", "target_id"},
     -1,
     {NULL}
    },
    {"fcdg_expr", /* 184 */
     SPD_RID_FCDG_EXPR,
     SPD_CID_FCDG, SPD_CID_EXPRESSION,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "fcdg_id", "expr_id"},
     -1,
     {NULL}
    },
    {"fcdg_lib", /* 185 */
     SPD_RID_FCDG_LIB,
     SPD_CID_FCDG, SPD_CID_IDENTIFIER,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "fcdg_id", "lib_id"},
     -1,
     {NULL}
    },
    {"def_var", /* 186 */
     SPD_RID_DEF_VAR,
     SPD_CID_FCDG, SPD_CID_EXPRESSION,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "fcdg_id", "variable_id"},
     -1,
     {NULL}
    },
    {"cond_op", /* 187 */
     SPD_RID_COND_OP,
     SPD_CID_FCDG, SPD_CID_OPERATOR,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "cond_id", "op_id"},
     -1,
     {NULL}
    },
    {"cond_val", /* 188 */
     SPD_RID_COND_VAL,
     SPD_CID_FCDG, SPD_CID_CONSTANT,
     SPD_MID_FCD_VIEW,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "fcdg_id", "value_id"},
     -1,
     {NULL}
    },
    /* J-model classes. */
    {"File", /* 189 */
     SPD_CID_J_FILE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "encoding"},
     0,
     {"str"}
    },
    {"Class", /* 190 */
     SPD_CID_J_CLASS,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "sort"},
     0,
     {"str"}
    },
    {"Interface", /* 191 */
     SPD_CID_J_INTERFACE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "sort"},
     0,
     {"str"}
    },
    {"Variable", /* 192 */
     SPD_CID_J_VARIABLE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "sort"},
     0,
     {"str"}
    },
    {"Routine", /* 193 */
     SPD_CID_J_ROUTINE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "sort"},
     0,
     {"str"}
    },
    {"Type", /* 194 */
     SPD_CID_J_TYPE,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), SVAL(1), -1},
     4,
     {"id", "sort", "offset", "length", "strOffset"},
     1,
     {"pkgName", "str"}
    },
    {"Modifier", /* 195 */
     SPD_CID_J_MODIFIER,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, SVAL(0), -1},
     3,
     {"id", "offset", "length", "strOffset"},
     0,
     {"str"}
    },
    {"SBlock", /* 196 */
     SPD_CID_J_SBLOCK,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "sort"},
     0,
     {"str"}
    },
    {"Statement", /* 197 */
     SPD_CID_J_STATEMENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "length", "sort", "offset", "strOffset"},
     0,
     {"str"}
    },
    {"Expression", /* 198 */
     SPD_CID_J_EXPRESSION,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, 5, SVAL(0), -1},
     5,
     {"id", "length", "sort", "offset", "strOffset", "lrsort"},
     0,
     {"str"}
    },
    {"Literal", /* 199 */
     SPD_CID_J_LITERAL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "offset", "length", "strOffset", "sort"},
     0,
     {"str"}
    },
    {"Operator", /* 200 */
     SPD_CID_J_OPERATOR,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "length", "sort", "offset", "strOffset"},
     0,
     {"str"}
    },
    {"QName", /* 201 */
     SPD_CID_J_QNAME,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, SVAL(0), -1},
     3,
     {"id", "offset", "length", "strOffset"},
     0,
     {"str"}
    },
    {"Special", /* 202 */
     SPD_CID_J_SPECIAL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, SVAL(0), -1},
     3,
     {"id", "offset", "length", "strOffset"},
     0,
     {"str"}
    },
    {"Label", /* 203 */
     SPD_CID_J_LABEL,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, 4, SVAL(0), -1},
     4,
     {"id", "length", "sort", "offset", "strOffset"},
     0,
     {"str"}
    },
    {"Comment", /* 204 */
     SPD_CID_J_COMMENT,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, 1, 2, 3, SVAL(0), -1},
     3,
     {"id", "offset", "length", "strOffset"},
     0,
     {"str"}
    },
    {"Wildcard", /* 205 */
     SPD_CID_J_WILDCARD,
     SPD_NON_CID, SPD_NON_CID,
     SPD_MID_J_MODEL,
     C_TYP_OC,
     {0, -1},
     0,
     {"id"},
     -1,
     {NULL}
    },
    /* J-model relationships. */
    {"define_package", /* 206 */
     SPD_RID_J_DEFINE_PACKAGE,
     SPD_CID_J_FILE, SPD_CID_J_QNAME,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "file", "package"},
     -1,
     {NULL}
    },
    {"import_package", /* 207 */
     SPD_RID_J_IMPORT_PACKAGE,
     SPD_CID_J_FILE, SPD_CID_J_QNAME,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "file", "package"},
     -1,
     {NULL}
    },
    {"define_class", /* 208 */
     SPD_RID_J_DEFINE_CLASS,
     SPD_CID_J_FILE, SPD_CID_J_CLASS,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "file", "class"},
     -1,
     {NULL}
    },
    {"implement_interface", /* 209 */
     SPD_RID_J_IMPLEMENT_INTERFACE,
     SPD_CID_J_CLASS, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "interface"},
     -1,
     {NULL}
    },
    {"extend_class", /* 210 */
     SPD_RID_J_EXTEND_CLASS,
     SPD_CID_J_CLASS, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "subClass", "superClass"},
     -1,
     {NULL}
    },
    {"class_modifier", /* 211 */
     SPD_RID_J_CLASS_MODIFIER,
     SPD_CID_J_CLASS, SPD_CID_J_MODIFIER,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "modifier"},
     -1,
     {NULL}
    },
    {"class_variable", /* 212 */
     SPD_RID_J_CLASS_VARIABLE,
     SPD_CID_J_CLASS, SPD_CID_J_VARIABLE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "variable"},
     -1,
     {NULL}
    },
    {"class_routine", /* 213 */
     SPD_RID_J_CLASS_ROUTINE,
     SPD_CID_J_CLASS, SPD_CID_J_ROUTINE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "routine"},
     -1,
     {NULL}
    },
    {"class_sblock", /* 214 */
     SPD_RID_J_CLASS_SBLOCK,
     SPD_CID_J_CLASS, SPD_CID_J_SBLOCK,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "sblock"},
     -1,
     {NULL}
    },
    {"class_argument", /* 215 */
     SPD_RID_J_CLASS_ARGUMENT,
     SPD_CID_J_CLASS, SPD_CID_J_EXPRESSION,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "expression"},
     -1,
     {NULL}
    },
    {"class_class", /* 216 */
     SPD_RID_J_CLASS_CLASS,
     SPD_CID_J_CLASS, SPD_CID_J_CLASS,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "class"},
     -1,
     {NULL}
    },
    {"class_interface", /* 217 */
     SPD_RID_J_CLASS_INTERFACE,
     SPD_CID_J_CLASS, SPD_CID_J_INTERFACE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "class", "interface"},
     -1,
     {NULL}
    },
    {"define_interface", /* 218 */
     SPD_RID_J_DEFINE_INTERFACE,
     SPD_CID_J_FILE, SPD_CID_J_INTERFACE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "file", "interface"},
     -1,
     {NULL}
    },
    {"extend_interface", /* 219 */
     SPD_RID_J_EXTEND_INTERFACE,
     SPD_CID_J_INTERFACE, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "subInterface", "superInterface"},
     -1,
     {NULL}
    },
    {"interface_modifier", /* 220 */
     SPD_RID_J_INTERFACE_MODIFIER,
     SPD_CID_J_INTERFACE, SPD_CID_J_MODIFIER,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "interface", "modifier"},
     -1,
     {NULL}
    },
    {"interface_variable", /* 221 */
     SPD_RID_J_INTERFACE_VARIABLE,
     SPD_CID_J_INTERFACE, SPD_CID_J_VARIABLE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "interface", "variable"},
     -1,
     {NULL}
    },
    {"interface_routine", /* 222 */
     SPD_RID_J_INTERFACE_ROUTINE,
     SPD_CID_J_INTERFACE, SPD_CID_J_ROUTINE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "interface", "routine"},
     -1,
     {NULL}
    },
    {"interface_class", /* 223 */
     SPD_RID_J_INTERFACE_CLASS,
     SPD_CID_J_INTERFACE, SPD_CID_J_CLASS,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "interface", "class"},
     -1,
     {NULL}
    },
    {"interface_interface", /* 224 */
     SPD_RID_J_INTERFACE_INTERFACE,
     SPD_CID_J_INTERFACE, SPD_CID_J_INTERFACE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "interface", "interface"},
     -1,
     {NULL}
    },
    {"variable_modifier", /* 225 */
     SPD_RID_J_VARIABLE_MODIFIER,
     SPD_CID_J_VARIABLE, SPD_CID_J_MODIFIER,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "variable", "modifier"},
     -1,
     {NULL}
    },
    {"variable_type", /* 226 */
     SPD_RID_J_VARIABLE_TYPE,
     SPD_CID_J_VARIABLE, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "variable", "type"},
     -1,
     {NULL}
    },
    {"initialize_variable", /* 227 */
     SPD_RID_J_INITIALIZE_VARIABLE,
     SPD_CID_J_VARIABLE, SPD_CID_J_EXPRESSION,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "variable", "initializer"},
     -1,
     {NULL}
    },
    {"parameter_variable", /* 228 */
     SPD_RID_J_PARAMETER_VARIABLE,
     SPD_CID_J_ROUTINE, SPD_CID_J_VARIABLE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "routine", "variable"},
     -1,
     {NULL}
    },
    {"return_type", /* 229 */
     SPD_RID_J_RETURN_TYPE,
     SPD_CID_J_ROUTINE, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "routine", "returnType"},
     -1,
     {NULL}
    },
    {"throw_type", /* 230 */
     SPD_RID_J_THROW_TYPE,
     SPD_CID_J_ROUTINE, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "routine", "throwType"},
     -1,
     {NULL}
    },
    {"routine_body", /* 231 */
     SPD_RID_J_ROUTINE_BODY,
     SPD_CID_J_ROUTINE, SPD_CID_J_STATEMENT,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "routine", "body"},
     -1,
     {NULL}
    },
    {"routine_modifier", /* 232 */
     SPD_RID_J_ROUTINE_MODIFIER,
     SPD_CID_J_ROUTINE, SPD_CID_J_MODIFIER,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "routine", "modifier"},
     -1,
     {NULL}
    },
    {"type_class", /* 233 */
     SPD_RID_J_TYPE_CLASS,
     SPD_CID_J_TYPE, SPD_CID_J_CLASS,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "type", "class"},
     -1,
     {NULL}
    },
    {"type_interface", /* 234 */
     SPD_RID_J_TYPE_INTERFACE,
     SPD_CID_J_TYPE, SPD_CID_J_INTERFACE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "type", "interface"},
     -1,
     {NULL}
    },
    {"type_type", /* 235 */
     SPD_RID_J_TYPE_TYPE,
     SPD_CID_J_TYPE, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "parentType", "childType"},
     -1,
     {NULL}
    },
    {"type_qname", /* 236 */
     SPD_RID_J_TYPE_QNAME,
     SPD_CID_J_TYPE, SPD_CID_J_QNAME,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "type", "qname"},
     -1,
     {NULL}
    },
    {"qname_qname", /* 237 */
     SPD_RID_J_QNAME_QNAME,
     SPD_CID_J_QNAME, SPD_CID_J_QNAME,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "parentQName", "chidlQName"},
     -1,
     {NULL}
    },
    {"qname_dot", /* 238 */
     SPD_RID_J_QNAME_DOT,
     SPD_CID_J_QNAME, SPD_CID_J_OPERATOR,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "qname", "dot"},
     -1,
     {NULL}
    },
    {"sblock_body", /* 239 */
     SPD_RID_J_SBLOCK_BODY,
     SPD_CID_J_SBLOCK, SPD_CID_J_STATEMENT,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "sblock", "body"},
     -1,
     {NULL}
    },
    {"local_class", /* 240 */
     SPD_RID_J_LOCAL_CLASS,
     SPD_CID_J_STATEMENT, SPD_CID_J_CLASS,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "statement", "class"},
     -1,
     {NULL}
    },
    {"local_variable", /* 241 */
     SPD_RID_J_LOCAL_VARIABLE,
     SPD_CID_J_STATEMENT, SPD_CID_J_VARIABLE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "statement", "variable"},
     -1,
     {NULL}
    },
    {"statement_label", /* 242 */
     SPD_RID_J_STATEMENT_LABEL,
     SPD_CID_J_STATEMENT, SPD_CID_J_LABEL,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "statement", "label"},
     -1,
     {NULL}
    },
    {"statement_statement", /* 243 */
     SPD_RID_J_STATEMENT_STATEMENT,
     SPD_CID_J_STATEMENT, SPD_CID_J_STATEMENT,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "outerStatement", "innerStatement"},
     -1,
     {NULL}
    },
    {"statement_expression", /* 244 */
     SPD_RID_J_STATEMENT_EXPRESSION,
     SPD_CID_J_STATEMENT, SPD_CID_J_EXPRESSION,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "statement", "expression"},
     -1,
     {NULL}
    },
    {"refer_label", /* 245 */
     SPD_RID_J_REFER_LABEL,
     SPD_CID_J_LABEL, SPD_CID_J_LABEL,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "referer", "referee"},
     -1,
     {NULL}
    },
    {"case_value", /* 246 */
     SPD_RID_J_CASE_VALUE,
     SPD_CID_J_LABEL, SPD_CID_J_EXPRESSION,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "label", "expression"},
     -1,
     {NULL}
    },
    {"expression_operand", /* 247 */
     SPD_RID_J_EXPRESSION_OPERAND,
     SPD_CID_J_EXPRESSION, SPD_CID_J_EXPRESSION,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "operandExpression"},
     -1,
     {NULL}
    },
    {"expression_operator", /* 248 */
     SPD_RID_J_EXPRESSION_OPERATOR,
     SPD_CID_J_EXPRESSION, SPD_CID_J_OPERATOR,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "operator"},
     -1,
     {NULL}
    },
    {"expression_expression", /* 249 */
     SPD_RID_J_EXPRESSION_EXPRESSION,
     SPD_CID_J_EXPRESSION, SPD_CID_J_EXPRESSION,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "parentExpression", "childExpression"},
     -1,
     {NULL}
    },
    {"refer_variable", /* 250 */
     SPD_RID_J_REFER_VARIABLE,
     SPD_CID_J_EXPRESSION, SPD_CID_J_VARIABLE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "variable"},
     -1,
     {NULL}
    },
    {"refer_literal", /* 251 */
     SPD_RID_J_REFER_LITERAL,
     SPD_CID_J_EXPRESSION, SPD_CID_J_LITERAL,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "literal"},
     -1,
     {NULL}
    },
    {"refer_special", /* 252 */
     SPD_RID_J_REFER_SPECIAL,
     SPD_CID_J_EXPRESSION, SPD_CID_J_SPECIAL,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "special"},
     -1,
     {NULL}
    },
    {"refer_type", /* 253 */
     SPD_RID_J_REFER_TYPE,
     SPD_CID_J_EXPRESSION, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "type"},
     -1,
     {NULL}
    },
    {"refer_qname", /* 254 */
     SPD_RID_J_REFER_QNAME,
     SPD_CID_J_EXPRESSION, SPD_CID_J_QNAME,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "qname"},
     -1,
     {NULL}
    },
    {"has_type", /* 255 */
     SPD_RID_J_HAS_TYPE,
     SPD_CID_J_EXPRESSION, SPD_CID_J_TYPE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "type"},
     -1,
     {NULL}
    },
    {"invoke_routine", /* 256 */
     SPD_RID_J_INVOKE_ROUTINE,
     SPD_CID_J_EXPRESSION, SPD_CID_J_ROUTINE,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "expression", "routine"},
     -1,
     {NULL}
    },
    {"symbol_comment", /* 257 */
     SPD_RID_J_SYMBOL_COMMENT,
     SPD_CID_J_WILDCARD, SPD_CID_J_COMMENT,
     SPD_MID_J_MODEL,
     C_TYP_RC,
     {1, 2, -1},
     2,
     {"id", "symbol", "comment"},
     -1,
     {NULL}
    },
    /* Terminater. */
    {NULL,
     SPD_NON_CID,
     SPD_NON_CID, SPD_NON_CID,
     SPD_NON_MID,
     0,
     {-1},
     -1,
     {NULL},
     -1,
     {NULL}}
};
