/* 
 * Program:             $RCSfile: db.c,v $  $Revision: 30.47 $
 * 
 * Purpose:             Access routines ver. 4.
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: db.c,v 30.47 2003/07/08 05:46:40 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#ifdef SAPID_FAST_db_get_xxx_id
#include <Sapid/spdSet2.h>
#endif

/* #define DEBUG_OBJ_AND_LINK_NUM */
/* #define DEBUG_SET_CLASS_INFO_TABLE */

#if defined(lint)
#define DW_CLASS_INFO_TABLE         printf      /* Just for cheking. */
#elif defined(DEBUG_SET_CLASS_INFO_TABLE)
#define DW_CLASS_INFO_TABLE         spdMessage
#else
#define DW_CLASS_INFO_TABLE         (void)
#endif

typedef int (*fnp)(const void *, const void *);

/* Object arrays. */

SpdDBObj    *object_db;
int         object_db_n;
int         *object_name_key;

static SpdSet2  object_id_index;

/* Relation arrays. */

SpdDBRel    *relation_db;
int         relation_db_n;

static int  *relation_p_id_key;
static int  *relation_c_id_key;

static int initial_obj_db_num;
static int initial_link_db_num;
static int object_db_size;
static int relation_db_size;
 
static SpdBoolean   is_relative_offset;

static void initialize_db_memory(void);
static void release_db_memory(void);

static void make_obj_id_index(void);
static void make_obj_name_index(void);
static void make_rel_p_id_index(void);
static void make_rel_c_id_index(void);
static void set_class_info(void);

/* 
 * ꤵ줿ǥμΤӴϢǿ¤׻
 * initial_{obj,link}_db_num˲û롥
 */

void set_initial_db_num(SpdModelId modelId)
{
    SpdClassId  cid;

    for (cid = spdGetFirstCid(modelId); cid <= spdGetLastRid(modelId); cid++) {
        if (SPD_IS_CLS_ID(cid)) {
            initial_obj_db_num += spdGetDbInfoNumOfObjs(cid);
        } else {
            initial_link_db_num += spdGetDbInfoNumOfObjs(cid);
        }
    }

#ifdef DEBUG_OBJ_AND_LINK_NUM
    spdMessage("modelId %2d, initial_obj_db_num %d\n",
               modelId, initial_obj_db_num);
    spdMessage("modelId %2d, initial_link_db_num %d\n",
               modelId, initial_link_db_num);
#endif
}

void append_object(SpdDBObj obj)
{
    if (object_db_size == object_db_n) {
        if (object_db_size == 0) {
            object_db_size = (initial_obj_db_num == 0) ?
                SAPID_SDB_OBJ_NUM : initial_obj_db_num;
            object_db = spdMalloc(sizeof(SpdDBObj) * object_db_size);
        } else {
            object_db_size += SAPID_SDB_OBJ_INC;
            object_db = spdRealloc(object_db,
                                   sizeof(SpdDBObj) * object_db_size);
        }
    }
    memcpy(&(object_db[object_db_n]), &obj, sizeof(SpdDBObj));
    object_db_n++;
}

int append_relation(SpdDBRel rel)
{
    if (relation_db_size == relation_db_n) {
        if (relation_db_size == 0) {
            relation_db_size = (initial_link_db_num == 0) ?
                SAPID_SDB_LINK_NUM : initial_link_db_num;
            relation_db = spdMalloc(sizeof(SpdDBRel) * relation_db_size);
        } else {
            relation_db_size += SAPID_SDB_LINK_INC;
            relation_db = spdRealloc(relation_db,
                                     sizeof(SpdDBRel) * relation_db_size);
        }
    }
    memcpy(&(relation_db[relation_db_n]), &rel, sizeof(SpdDBRel));
    return relation_db_n++;
}

/******************************************************************************/

static int cmp_obj_name(const int *a, const int *b)
{
    SpdClassId  classId_a = spdGetClassId((SpdObjId)(object_db[*a].Obj_id));
    SpdClassId  classId_b = spdGetClassId((SpdObjId)(object_db[*b].Obj_id));
    int         retVal;

    if ((retVal = classId_a - classId_b) != 0) {
        return (retVal);
    }

    /* 
     * Modified by R.Miura 1998/08/14.
     * svalue[0]name°ǤȤ򳰤ޥObj_name
     * Τᡤsvalue[0]name°Ǥʤ饹ˤĤƤ0֤
     *
     * return (spdSafetyStrcmp(object_db[*a].Obj_name, object_db[*b].Obj_name));
     */
    if (class_table[classId_a].sattr[0] == NULL ||
        !SPD_STREQ(class_table[classId_a].sattr[0], "name")) {
        return (0);
    }
    return (spdSafetyStrcmp(object_db[*a].svalue[0], object_db[*b].svalue[0]));
}

static int cmp_rel_p_id(const int *a, const int *b)
{
    int         retVal;

    if ((retVal = relation_db[*a].Rel_p_id - relation_db[*b].Rel_p_id) != 0) {
        return (retVal);
    }

    /* 
     * idϡ
     *     classId * SPD_ID_SPACE + 0, 1,...
     * ȳդ뤿ᡤid˥ȤclassIdˤʤ롥
     */
    return (relation_db[*a].Rel_id - relation_db[*b].Rel_id);
}

static int cmp_rel_c_id(const int *a, const int *b)
{
    int         retVal;

    if ((retVal = relation_db[*a].Rel_c_id - relation_db[*b].Rel_c_id) != 0) {
        return (retVal);
    }

    /* ؿcmp_rel_p_id()ΥȤ򻲾Ȥ뤳ȡ */
    return (relation_db[*a].Rel_id - relation_db[*b].Rel_id);
}

/******************************************************************************/

void make_index(void)
{
    make_obj_id_index();
    make_obj_name_index();
    make_rel_p_id_index();
    make_rel_c_id_index();
    set_class_info();                       /* For direct access. */
}

static void make_obj_id_index(void)
{
    int     i;

    object_id_index = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_INT_VALUE);
#if 0
    /* ϸ̤⤢롣 */
    object_id_index->initialBktNum = object_db_n / 3;
    object_id_index->poolCapacityIncrement = object_db_n;
#endif
    for (i = 0; i < object_db_n; i++)
        spdAddElemOI(object_id_index, object_db[i].Obj_id, i);
}

static void make_obj_name_index(void)
{
    int     i;

    object_name_key = spdRealloc(object_name_key, object_db_n * sizeof(int));
    for (i = 0; i < object_db_n; i++) {
        object_name_key[i] = i;
    }
    SPD_QSORT(object_name_key, object_db_n, sizeof(int), cmp_obj_name);
}

static void make_rel_p_id_index(void)
{
    int     i;

    relation_p_id_key = spdRealloc(relation_p_id_key,
                                   relation_db_n * sizeof(int));
    for (i = 0; i < relation_db_n; i++) {
        relation_p_id_key[i] = i;
    }
    SPD_QSORT(relation_p_id_key, relation_db_n, sizeof(int), cmp_rel_p_id);
}

static void make_rel_c_id_index(void)
{
    int     i;

    relation_c_id_key = spdRealloc(relation_c_id_key,
                                   relation_db_n * sizeof(int));
    for (i = 0; i < relation_db_n; i++) {
        relation_c_id_key[i] = i;
    }
    SPD_QSORT(relation_c_id_key, relation_db_n, sizeof(int), cmp_rel_c_id);
}

static void set_class_info(void)
{
    int         id;
    SpdClassId  cId;

    for (cId = SPD_FIRST_CID; cId <= SPD_LAST_RID; cId++) {
        class_table[cId].min_id = SPD_NON_ID;
        class_table[cId].max_id = SPD_NON_ID;
    }

    for (id = 0; id < object_db_n; id++) {
        cId = spdGetClassId((SpdObjId)(object_db[id].Obj_id));
        if (class_table[cId].min_id == SPD_NON_ID ||
            object_db[id].Obj_id < class_table[cId].min_id) {
            class_table[cId].min_id = object_db[id].Obj_id;
            class_table[cId].min_index = id;
        }
        if (class_table[cId].max_id == SPD_NON_ID ||
            class_table[cId].max_id < object_db[id].Obj_id) {
            class_table[cId].max_id = object_db[id].Obj_id;
        }
    }
    for (id = 0; id < relation_db_n; id++) {
        cId = spdGetClassId((SpdRelId)(relation_db[id].Rel_id));
        if (class_table[cId].min_id == SPD_NON_ID ||
            relation_db[id].Rel_id < class_table[cId].min_id) {
            class_table[cId].min_id = relation_db[id].Rel_id;
            class_table[cId].min_index = id;
        }
        if (class_table[cId].max_id == SPD_NON_ID ||
            class_table[cId].max_id < relation_db[id].Rel_id) {
            class_table[cId].max_id = relation_db[id].Rel_id;
            /* class_table[cId].max_index = id; */
        }
    }

    for (cId = SPD_FIRST_CID; cId <= SPD_LAST_RID; cId++) {
        DW_CLASS_INFO_TABLE("set_class_info(): \"%s\", min_id %d, max_id %d, num %d, min_index %d.\n",
                            class_table[cId].name,
                            class_table[cId].min_id, class_table[cId].max_id,
                            spdGetDbInfoNumOfObjs(cId),
                            class_table[cId].min_index);
    }
}

/******************************************************************************/

static int cmp2_rel_p_id(const SpdDBRelCursor *a, const int *b)
{
    if (a->rel.Rel_p_id != relation_db[*b].Rel_p_id || a->classId == C_ANY) {
        return a->rel.Rel_p_id - relation_db[*b].Rel_p_id;
    }
    return (a->classId - spdGetClassId((SpdRelId)(relation_db[*b].Rel_id)));
}

static int cmp2_rel_c_id(const SpdDBRelCursor *a, const int *b)
{
    if (a->rel.Rel_c_id != relation_db[*b].Rel_c_id || a->classId == C_ANY) {
        return a->rel.Rel_c_id - relation_db[*b].Rel_c_id;
    }
    return (a->classId - spdGetClassId((SpdRelId)(relation_db[*b].Rel_id)));
}

/******************************************************************************/

SpdDBObj *db_get_object_id(SpdObjId objId)
{
    SpdElem     *elem;

    elem = spdSearchElemO(object_id_index, objId);
    return ((elem == NULL) ? NULL : 
            &(object_db[spdSetGetValueI(elem->value)]));
}

SpdDBRel *db_get_relation_id(SpdRelId relId)
{
    SpdClassId  cId;

    if ((cId = spdGetClassId(relId)) == SPD_NON_CID) {
        return (NULL);
    }
    return (&relation_db[class_table[cId].min_index +
                         relId - class_table[cId].min_id]);
}

SpdDBRel *db_get_relation_p_id(SpdDBRelCursor *dp)
{
    if (!dp->tmp) {
        dp->tmp = bxsearch(dp, relation_p_id_key, (size_t)relation_db_n, 
                           sizeof(int), (fnp)cmp2_rel_p_id);
    } else {
        if (dp->tmp - relation_p_id_key >= relation_db_n - 1 ||
            cmp2_rel_p_id(dp, (++(dp->tmp))) != 0) {
            dp->tmp = NULL;
        }
    }
    return ((dp->tmp == NULL) ? NULL : &(relation_db[*(dp->tmp)]));
}

SpdDBRelCursor *db_get_relation_p_id_init(SpdClassId classId, SpdRelId p_id)
{
    SpdDBRelCursor  *dp;
    
    dp = spdMalloc(sizeof(SpdDBRelCursor));
    dp->classId = classId;
    dp->rel.Rel_p_id = p_id;
    dp->tmp = NULL;
    return dp;
}

void db_get_relation_p_id_free(SpdDBRelCursor *dp)
{
    spdFree(dp);
}

/******************************************************************************/

SpdDBRel *db_get_relation_c_id(SpdDBRelCursor *dp)
{
    if (!dp->tmp) {
        dp->tmp = bxsearch(dp, relation_c_id_key, (size_t)relation_db_n, 
                           sizeof(int), (fnp)cmp2_rel_c_id);
    } else {
        if (dp->tmp - relation_c_id_key >= relation_db_n - 1 ||
            cmp2_rel_c_id(dp, (++(dp->tmp))) != 0) {
            dp->tmp = NULL;
        }
    }
    return ((dp->tmp == NULL) ? NULL : &(relation_db[*(dp->tmp)]));
}

SpdDBRelCursor *db_get_relation_c_id_init(SpdClassId classId, SpdRelId c_id)
{
    SpdDBRelCursor *dp;
    
    dp = spdMalloc(sizeof(SpdDBRelCursor));
    dp->classId = classId;
    dp->rel.Rel_c_id = c_id;
    dp->tmp = NULL;
    return (dp);
}

void db_get_relation_c_id_free(SpdDBRelCursor *dp)
{
    spdFree(dp);
}

SpdBoolean spdIsValidId(SpdId id)
{
    return (db_get_object_id(id) != NULL || db_get_relation_id(id) != NULL);
}

/* spdTarget(). ***************************************************************/

/* 
 * Default value for spdOption() in AR4.
 */

static SpdOption spdDefaultOption[] = {
    { "mode", SAPID_C_MODE },
    { "prototype", SPDSTR_PROTOTYPE_NON },
    { "typeQualifier", OFF },
    { NULL, SAPID_NON_OPTION }
};

/*
 * AR4νؿ
 * programNameꤵ줿ϡԤprogramId֤
 * programNameNULLξϡԤǤդprogramId֤
 */

SpdObjId spdTarget(SpdString modelName, SpdString programName)
{
    SpdDBId         dbId;
    
    if (SPD_STREQ(modelName, "I-model")) { /* Use model. */
        spdUseModel(&I_model);
    } else if (SPD_STREQ(modelName, "P-model")) {
        spdUseModel(&I_model);
        spdUseModel(&P_model);
    } else {
        spdErrorExit("spdTarget(): \"%s\": Invalid model.\n", modelName); 
    }
    dbId = spdOpenSDB();                    /* Open SDB. */
    spdReadDB(SPD_NON_MID, dbId);    /* Read all DB. */
    return (spdGetAnObjId("program", programName));
}

SpdDBId spdOpenSDB(void)
{
    SpdInfo             *info;
    SpdString           relative_offset;
    static SpdBoolean   is_first_call = TRUE;

    if (is_first_call) {
        atexit(spdReleaseTarget);
        initialize_db_memory();             /* Initialize heap memory. */
        is_first_call = FALSE;
    } else {
        spdReleaseTarget();                 /* Re-initialize heap memory. */
    }

    if (!smuIsValidSDB()) {                 /* Check validity of SDB. */
        spdErrorExit(SPD_MSG_INVALID_SDB, spdGetSDBdir());
    }

    spdOpenDbInfo();                  /* Read num-of-elements table. */

    /* Read format informations. */
    info = spdOpenSDBinfo();
    spdInsertOption(spdDefaultOption);
    if ((relative_offset = spdGetInfoString(info, "RELATIVE_OFFSET")) != NULL) {
        is_relative_offset = SPD_STREQ(relative_offset, "TRUE");
    }
    if (SPD_STREQ(spdGetInfoString(info, "BINARY_FORMAT_SDB"), "TRUE")) {
        spdSetReadSDBformat(SAPID_BINARY_FORMAT_SDB);
        spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
    } 
    spdCloseInfo(info);

    return (SPD_FIRST_DB_ID);
}

void spdCloseSDB(SpdDBId dbId)
{
    spdReleaseTarget();
}

void spdReleaseTarget(void)
{
    spd_reinitialize_models();
    release_db_memory();
    initialize_db_memory();

    /* Added by S.Yamamoto 2001/07/05. */
    spdReleasePidbCache();
    spdReleaseLineTbl();

    /* Added by N.Fukuyasu 2001/07/09. */
    spdReleaseBlockTable();
}

/*
 * Function names "spd_initialize_SDB()" and "spd_release_SDB()" are better.
 */

static void initialize_db_memory(void)
{
    initial_obj_db_num = 0;
    initial_link_db_num = 0;

    object_db = NULL;
    object_db_n = 0;
    object_db_size = 0;
    object_name_key = NULL;

    object_id_index = NULL;

    relation_db = NULL;
    relation_db_n = 0;
    relation_db_size = 0;
    relation_p_id_key = NULL;
    relation_c_id_key = NULL;

    /* Initialize num-of-elements table. */
    spdInitDbInfo();
}

static void release_db_memory(void)
{
    int         i;
    
    /* Release object_db[]. */
    for (i = 0; i < object_db_n; i++) {
        spdFreeStrsInObj(object_db[i]);
    }
    if (object_db != NULL) {
        spdFree(object_db);
        object_db = NULL;
    }

    /* Release relation_db[]. */
    if (relation_db != NULL) {
        spdFree(relation_db);
        relation_db = NULL;
    }

    if (object_id_index != NULL){
        spdFreeSet2(object_id_index);
    }

    if (object_name_key != NULL) {
        spdFree(object_name_key);
        object_name_key = NULL;
    }

    if (relation_p_id_key != NULL) {
        spdFree(relation_p_id_key);
        relation_p_id_key = NULL;
    }

    if (relation_c_id_key != NULL) {
        spdFree(relation_c_id_key);
        relation_c_id_key = NULL;
    }
}

/******************************************************************************/

/*
 * Functions spdGetRelativeOffset() and spdGetAbsoluteOffset() refer
 * variable "is_relative_offset".
 */

int spdGetRelativeOffset(SpdRelId relationId)
{
    int         offset;
    SpdRelId    fileDeclRelId;

    offset = spdGetAttrValIntOfRel(relationId, "offset");

    if (is_relative_offset == TRUE || offset == -1) {
        return (offset);
    }

    if (spdGetClassId(relationId) == SPD_RID_FILE_DECL) {
        return (offset);
    }

    if ((fileDeclRelId = spdGetFileDeclRelId(relationId, SPD_NON_ID)) ==
        SPD_NON_ID) {
        return (-1);
    }

    return (offset - spdGetAttrValIntOfRel(fileDeclRelId, "offset"));
}

int spdGetAbsoluteOffset(SpdRelId relationId, SpdObjId fileId)
{
    int         offset;
    SpdRelId    fileDeclRelId;

    offset = spdGetAttrValIntOfRel(relationId, "offset");
    if (is_relative_offset == FALSE || offset == -1) {
        return (offset);
    } else if (spdGetClassId(relationId) == SPD_RID_FILE_DECL) {
        return (offset);
    } else if ((fileDeclRelId = spdGetFileDeclRelId(relationId, fileId)) ==
               SPD_NON_ID) {
        return (-1);
    } else {
        return (offset + spdGetAttrValIntOfRel(fileDeclRelId, "offset"));
    }
}
