/* 
 * Program:             $RCSfile: getAttrVal.c,v $  $Revision: 30.11 $
 * 
 * Purpose:             Getting attribute value.
 * 
 * Author:              K.Hayashi  1995/02/20
 *                      S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: getAttrVal.c,v 30.11 2003/05/24 13:47:02 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/******************************************************************************/

#ifdef SAPID_SAFETY_MODE_0

#define INVALID_ATTR(f, l, a, c) \
        spdWarning("\"%s\", %d: Invalid attribute \"%s\" of class \"%s\".\n", \
                   (f), (l), (a), (c))

#else /* not SAPID_SAFETY_MODE_0 */

#define INVALID_ATTR(f, l, a, c)        /* Empty. */

#endif /* not SAPID_SAFETY_MODE_0 */

#ifdef SAPID_SAFETY_MODE_1

#define INVALID_OBJECT_ID(f, l, o) \
        spdWarning("\"%s\", %d: Invalid object id %d.\n", (f), (l), (o))
#define INVALID_RELATION_ID(f, l, r) \
        spdWarning("\"%s\", %d: Invalid relation id %d.\n", (f), (l), (r))

#else /* not SAPID_SAFETY_MODE_1 */

#define INVALID_OBJECT_ID(f, l, o)      /* Empty. */
#define INVALID_RELATION_ID(f, l, r)    /* Empty. */

#endif /* not SAPID_SAFETY_MODE_1 */

/******************************************************************************/

#ifdef SAPID_SAFETY_MODE_0

int _spdGetAttrValInt(char *d_file, int d_line,
                      int objectId, SpdConstString attrName)
{
    return (SPD_IS_CLS_ID(spdGetClassId(objectId)) ?
            _spdGetAttrValIntOfObj(d_file, d_line, objectId, attrName) :
            _spdGetAttrValIntOfRel(d_file, d_line, objectId, attrName));
}

#else /* not SAPID_SAFETY_MODE_0 */

int spdGetAttrValInt(SpdId objectId, SpdConstString attrName)
{
    return (SPD_IS_CLS_ID(spdGetClassId(objectId)) ?
            spdGetAttrValIntOfObj(objectId, attrName) :
            spdGetAttrValIntOfRel(objectId, attrName));
}

#endif /* not SAPID_SAFETY_MODE_0 */

#ifdef SAPID_SAFETY_MODE_0
int _spdGetAttrValIntOfObj(char *d_file, int d_line,
                           SpdObjId objectId, SpdString attrName)
#else
int spdGetAttrValIntOfObj(SpdObjId objectId, SpdConstString attrName)
#endif
{
    SpdDBObj    *obj;
    int         i;

    if ((obj = db_get_object_id(objectId)) == NULL) {
        INVALID_OBJECT_ID(d_file, d_line, objectId);
        return (-1);
    }

    i = spdGetAttrIdByName3(spdGetClassId(objectId), NULL, attrName);
    if (0 <= i && i < SpdObjIValN) {
        return (obj->ivalue[i]);
    } else {
        INVALID_ATTR(d_file, d_line, attrName, className);
        return (-1);
    }
}

#ifdef SAPID_SAFETY_MODE_0
int _spdGetAttrValIntOfRel(char *d_file, int d_line,
                           SpdRelId relationId, SpdString attrName)
#else
int spdGetAttrValIntOfRel(SpdRelId relationId, SpdConstString attrName)
#endif
{
    SpdDBRel    *rel;
    int         i;

    if ((rel = db_get_relation_id(relationId)) == NULL) {
        INVALID_RELATION_ID(d_file, d_line, relationId);
        return (-1);
    }

    i = spdGetAttrIdByName3(spdGetClassId(relationId), NULL, attrName);
    if (0 <= i && i < SpdRelIValN) {
        return (rel->ivalue[i]);
    } else {
        INVALID_ATTR(d_file, d_line, attrName, className);
        return (-1);
    }
}

/*
 * Note by S.Yamamoto 1998/08/26.
 * We assume that no relation has string attribute.
 */

#ifdef SAPID_SAFETY_MODE_0
SpdString _spdGetAttrValString(char *d_file, int d_line,
                               SpdId objectId, SpdString attrName)
#else
SpdString spdGetAttrValString(SpdId objectId, SpdString attrName)
#endif
{
    SpdDBObj    *obj;
    SpdClassId  classId = spdGetClassId(objectId);
    int         i;

    if (SPD_IS_REL_ID(classId) || (obj = db_get_object_id(objectId)) == NULL) {
        INVALID_OBJECT_ID(d_file, d_line, objectId);
        return (NULL);
    }

    i = spdGetAttrIdByName3(classId, NULL, attrName);
    if (SpdObjIValN <= i) {
        return (obj->svalue[i - SpdObjIValN]);
    } else {
        INVALID_ATTR(d_file, d_line, attrName, className);
        return (NULL);
    }
}

/******************************************************************************/

#ifndef USE_MACRO_spdGetName

SpdString spdGetName(SpdObjId objectId)
{
    return (spdGetAttrValString(objectId, "name"));
}

int spdGetLength(SpdObjId objectId)
{
    return (spdGetAttrValIntOfObj(objectId, "length"));
}

int spdGetOffset(SpdRelId relationId)
{
    return (spdGetAttrValIntOfRel(relationId, "offset"));
}

#endif /* not USE_MACRO_spdGetName */
