/* 
 * Program:             $RCSfile: getObjId.c,v $  $Revision: 30.9 $
 * 
 * Purpose:             
 * 
 * Author:              K.Hayashi  1995/02/20
 *                      S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: getObjId.c,v 30.9 2003/06/28 15:31:47 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

typedef int     (*fnp)(const void *, const void *);

static int  cmp2_obj_name(const SpdDBObjCursor *a, const int *b);
static int  cmp3_obj_name(const SpdDBObjCursor *a, const int *b);

static SpdDBObjCursor   *db_get_object_name_init(SpdClassId classId,
                                                 SpdConstString name);
static SpdDBObj         *db_get_object_name(SpdDBObjCursor *dp);

/* spdGetObjId(). *************************************************************/

SpdCursor *spdGetObjIdInit(SpdConstString className, SpdConstString name)
{
    SpdClassId  classId;
    SpdCursor   *csr;
    
    if ((classId = spdGetClassIdByName(className)) == -1) {
        return (NULL);
    }

    /* 
     * Added by R.Miura 1998/08/14.
     * "name"°ʤˤ⤫餺"name"ꤵƤν
     * svalue1ĤʤȤꤷƤ롥
     */
    if (name != NULL &&
        (class_table[classId].sattr[0] == NULL ||
        !SPD_STREQ(class_table[classId].sattr[0], "name"))) {
        return (NULL);
    }

    csr = spdAllocCursor();
    csr->op = db_get_object_name_init(classId, name);
    return (csr);
}

SpdObjId spdGetObjId(SpdCursor *csr)
{
    SpdDBObj *obj;
    
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return (SPD_NON_ID);
    }
    if ((obj = db_get_object_name(csr->op)) != NULL) {
        return (obj->Obj_id);
    }
    return (SPD_NON_ID);
}

SpdObjId spdGetAnObjId(SpdConstString className, SpdConstString name)
{
    SpdCursor   *csr;
    SpdObjId    id;
    
    if ((csr = spdGetObjIdInit(className, name)) == NULL) {
        return (SPD_NON_ID);
    }
    id = spdGetObjId(csr);
    spdFreeCursor(csr);
    return (id);
}

/* db_get_object_name(). ******************************************************/

/* 
 * static SpdDBObjCursor *db_get_object_name_init(int classId, char *name)
 */

static SpdDBObjCursor *db_get_object_name_init(SpdClassId classId,
                                               SpdConstString name)
{
    SpdDBObjCursor *dp;
    
    dp = spdMalloc(sizeof(SpdDBObjCursor));
    dp->classId = classId;
    dp->obj.svalue[0] = (name == NULL) ? NULL : spdStrdup(name);
    dp->tmp = NULL;
    return (dp);
}

static SpdDBObj *db_get_object_name(SpdDBObjCursor *dp)
{
    fnp     compar = (fnp)((dp->obj.svalue[0] == NULL) ?
                           cmp3_obj_name : cmp2_obj_name);

    if (dp->tmp == NULL) {
        dp->tmp = bxsearch(dp, object_name_key, (size_t)object_db_n,
                           sizeof(int), compar);
    } else {
        if (object_db_n - 1 <= dp->tmp - object_name_key ||
            compar(dp, (++(dp->tmp))) != 0) {
            dp->tmp = NULL;
        }
    }
    return (dp->tmp == NULL ? NULL : &(object_db[*(dp->tmp)]));
}

void db_get_object_name_free(SpdDBObjCursor *dp)
{
    spdFree(dp->obj.svalue[0]);
    spdFree(dp);
}

/******************************************************************************/

/* 
 * static int cmp2_obj_name(const SpdDBObj *a, const int *b)
 * {
 *     if ((*a).Obj_typ != object_db[*b].Obj_typ) {
 *         return (*a).Obj_typ - object_db[*b].Obj_typ;
 *     }
 *     return (spdSafetyStrcmp((*a).svalue[0], object_db[*b].svalue[0]));
 * }
 * 
 * static int cmp3_obj_name(const SpdDBObj *a, const int *b)
 * {
 *     return ((*a).Obj_typ - object_db[*b].Obj_typ);
 * }
 */

static int cmp2_obj_name(const SpdDBObjCursor *a, const int *b)
{
    SpdClassId classId_b;
    
    classId_b = spdGetClassId((SpdObjId)(object_db[*b].Obj_id));
    if (a->classId != classId_b) {
        return (a->classId - classId_b);
    }
    return (spdSafetyStrcmp(a->obj.svalue[0], object_db[*b].svalue[0]));
}

static int cmp3_obj_name(const SpdDBObjCursor *a, const int *b)
{
    /*
     * This function should be invoked in db_get_object_name().
     * Always ((*a).svalue[0] == NULL) is valid.
     */
    /* assert((*a).svalue[0] == NULL); */

    return (a->classId - spdGetClassId((SpdObjId)(object_db[*b].Obj_id)));
}
