/* 
 * Program:             $RCSfile: getOcc.c,v $  $Revision: 30.24 $
 * 
 * Purpose:             Occطδؿ("ar.c"ʬ)
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: getOcc.c,v 30.24 2003/06/28 15:31:48 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdOcc   spdNonOcc = SPD_NON_OCC;

static SpdObjId spdGetPFileId(SpdRelId relationId)
{
    SpdObjId    fileId = SPD_NON_ID;
    SpdObjId    objectId;
    SpdRel      rel;
    SpdCursor   *csr;
    
    objectId = spdGetAttrValIntOfRel(relationId, "p-id");
    /* ľ夬pfileξ硥 */
    if (spdGetClassId(objectId) == SPD_CID_PFILE) {
        return (spdGetAttrValIntOfRel(relationId, "pfile_id"));
    }

    /* ƤΥ֥ƤФơ */
    csr = spdGetRelInit(objectId, "any", "c-id");
    while ((rel = spdGetRel(csr)).relationId != SPD_NON_ID) {
        if (SPD_IS_TYP_CMP(spdGetClassId(rel.relationId))) {
            fileId = spdGetPFileId(rel.relationId);
            /* pfileȯ줿 */
            if (fileId != SPD_NON_ID) {
                break;
            }
        }
    }
    spdFreeCursor(csr);

    /* pfileȯƤʤȤSPD_NON_ID֤ */
    return (fileId);
}

/******************************************************************************/

SpdCursor *spdGetOffsetOfObjInit(SpdObjId objectId)
{
    SpdCursor   *csr;
    SpdRel      rel;
    
    csr = spdAllocCursor();
    csr->objectId = objectId;
    csr->csr = spdGetRelInit(objectId, "any", "c-id");
    while ((rel = spdGetRel(csr->csr)).relationId != SPD_NON_ID) {
        if (SPD_IS_TYP_CMP(spdGetClassId(rel.relationId))) {
            spdQueueEnqueue(&(csr->queue), rel.relationId);
        }
    }
    return (csr);
}

SpdOcc spdGetOffsetOfObj(SpdCursor *csr)
{
    SpdCursor   *relcsr;
    SpdRel      rel;
    SpdOcc      occ;
    SpdRelId    relationId;
    SpdModelId  modelId;
    SpdClassId  classId;
    SpdString   relName;
    
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return (spdNonOcc);
    }
    relationId = spdQueueDequeue(&(csr->queue));
    if (relationId == -1) {
        return (spdNonOcc);
    }
    occ.objectId = csr->objectId;           /* Not compatible with old AR. */
    occ.relationId = relationId;

    classId = spdGetClassId(relationId);
    relName = spdGetClassNameById(classId);

    relcsr = spdGetRelInit(occ.objectId, relName, "c-id");

    while ((rel = spdGetRel(relcsr)).relationId != SPD_NON_ID) {
        if (rel.relationId == relationId) {
            break;
        }
    }

    if (rel.relationId != SPD_NON_ID) {
        modelId = spdGetModelId(spdGetClassId(rel.objectId));
        if (modelId == SPD_MID_P_MODEL) {
            occ.fileId = spdGetPFileId(relationId);
            occ.offset = spdGetOffset(relationId);
        } else if (modelId == SPD_MID_I_MODEL) {
            occ.fileId = spdGetFileId(relationId);
            occ.offset = spdGetAbsoluteOffset(relationId, occ.fileId);
        } else {
            occ.fileId = SPD_NON_ID;
            occ.offset = SPD_NON_OFFSET;
        }
    }

    return ((occ.fileId == SPD_NON_ID) ? spdNonOcc : occ);
}

/*
 * Added by S.Yamamoto 2001/06/19.
 */

SpdOcc spdGetAnOffsetOfObj(SpdObjId objectId)
{
    SpdCursor   *csr;
    SpdOcc      occ;

    if ((csr = spdGetOffsetOfObjInit(objectId)) == NULL) {
        return (spdNonOcc);
    }
    occ = spdGetOffsetOfObj(csr);
    spdFreeCursor(csr);

    return (occ.objectId == SPD_NON_ID ? spdNonOcc : occ);
}

/******************************************************************************/

SpdCursor *spdGetIncludedOccInit(SpdObjId objectId, SpdConstString objectType,
                                 SpdObjId fileId)
{
    SpdCursor   *csr;
    SpdRel      rel;
    SpdObjId    o_id;
    SpdRelId    relId;
    SpdClassId  classId = spdGetClassIdByName(objectType);
    SpdBoolean  isAny = SPD_STREQ(objectType, "any");
    SpdSet      set = SAPID_EMPTY_SET;

    static SpdQueue  rqueue = SPD_EMPTY_QUEUE;

    csr = spdAllocCursor();
    spdQueueClean(&rqueue);

    csr->csr = spdGetRelInit(objectId, "any", "p-id");
    while ((rel = spdGetRel(csr->csr)).relationId != SPD_NON_ID) {
        if (SPD_IS_TYP_CMP(spdGetClassId(rel.relationId))) {
            spdQueueEnqueue(&rqueue, rel.relationId);
        }
    }
    spdAddElem(set, objectId, NULL);
    while ((relId = spdQueueDequeue(&rqueue)) != SPD_NON_ID) {
        o_id = spdGetAttrValIntOfRel(relId, "c-id");
        if (isAny || spdGetClassId(o_id) == classId) {
            spdQueueEnqueue(&(csr->queue), relId);
        }
        if (spdSetContains(set, o_id)) {
            continue;
        }
        spdAddElem(set, o_id, NULL);
        spdFreeCursor(csr->csr);
        csr->csr = spdGetRelInit(o_id, "any", "p-id");
        while ((rel = spdGetRel(csr->csr)).relationId != SPD_NON_ID) {
            if (SPD_IS_TYP_CMP(spdGetClassId(rel.relationId))) {
                spdQueueEnqueue(&rqueue, rel.relationId);
            }
        }
    }
    spdFreeSet(set);

    /* Added by H.Ohhashi 1998/05/09. */
    csr->fileId = fileId;

    return (csr);
}

SpdOcc spdGetIncludedOcc(SpdCursor *csr)
{
    SpdOcc      occ;
    SpdRelId    relationId;
    
    if (csr == NULL) {
        spdWarning("Invalid cursor.\n");
        return (spdNonOcc);
    }
    relationId = spdQueueDequeue(&(csr->queue));
    if (relationId == -1) {
        return (spdNonOcc);
    }
    occ.objectId = spdGetAttrValIntOfRel(relationId, "c-id");
    occ.relationId = relationId;

#ifdef SAPID_AR4
    /* P-model ΤȤ */
    if (occ.objectId >= SPD_OBJ_ID_OFFSET(SPD_CID_PFILE) &&
        occ.objectId < SPD_OBJ_ID_OFFSET(SPD_CID_HFILE) + SPD_ID_SPACE) {
        /* P-modelξfileIdΤɤ狼ʤ */
        /* by H.Ohhashi 1998/05/09  */
        occ.fileId = spdGetPFileId(relationId);
        occ.offset = spdGetOffset(relationId);
    } else {
        /* Modified by H.Ohhashi 1998/05/09 */
        /* occ.fileId = spdGetFileId(relationId); */
        if (csr->fileId == SPD_NON_ID) {
            occ.fileId = spdGetFileId(relationId);
        } else {
            occ.fileId = csr->fileId;
        }
        occ.offset = spdGetAbsoluteOffset(relationId, occ.fileId);
    }
#endif

    return occ;
}
