/* 
 * Program:             $RCSfile: read_sdb.c,v $  $Revision: 30.62 $
 * 
 * Purpose:             
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: read_sdb.c,v 30.62 2003/01/12 15:10:56 yamamoto Exp yamamoto $";
#endif

/* 
 * spdTarget(modelName, programName)
 *     spdReadDB(modelId, dbId)
 *         spdReadAModel(modelId, dbId)
 *             read_obj_class(classId, f)
 *                 (*f)(fp, classId)                        spdReadAndInsertObj.
 *                     spdReadObj(fp, classId)              Dependent on AR4.
 *                         spdReadIntAttr{Text,Binary}(fp)  Read a integer.
 *                         spdReadStrAttr{Text,Binary}(fp)  Read a string.
 *                     append_object()                      Dependent on AR4.
 *                     spdReadRel(fp, classId, relId)       Dependent on AR4.
 *                         spdReadIntAttr{Text,Binary}(fp)  Read a integer.
 *                         spdReadStrAttr{Text,Binary}(fp)  Read a string.
 *                     append_relation()                    Dependent on AR4.
 *             read_rel_class(classId, f)
 *                 ...
 *         read_Pmodel(void)                                Static.
 *             ...
 */

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define SPD_DUMP_READ */

void read_obj_class(SpdClassId classId, int num,
                    void (*fpReadAndInsert)(FILE *fp, SpdClassId classId))
{
    char        *className = spdGetClassNameById(classId);
    FILE        *fp;

    if (num == 0) {
        return;
    }

#ifdef SPD_DUMP_READ
    spdMessage("read_obj_class(): className \"%s\", num %d.\n", className, num);
#endif

    fp = spdOpenSDBfile(spdGetModelDir(spdGetModelId(classId)),
                        className, SAPID_O_READ);
    while (0 < num--) {
        (*fpReadAndInsert)(fp, classId);
    }
    if (feof(fp)) {
        spdErrorExit("read_obj_class(): %s: Too small SDB.\n", className);
    } else if (getc(fp) != EOF) {
        spdErrorExit("read_obj_class(): %s: Too large SDB.\n", className);
        spdMessage("    feof(): %d.\n", feof(fp));
        spdMessage("    ftell(): %d.\n", ftell(fp));
        spdMessage("    ferror(): %d.\n", ferror(fp));
    }
    spdFclose(fp);

#ifdef SAPID_READ_ROUTINE_MEMORY_PROFILE
    spdMessage("read_obj_class(): \"%s\".\n", className);
    spdPrintMemProf();
#endif
}

void read_rel_class(SpdClassId classId, int num,
                    void (*fpReadAndInsert)(FILE *fp, SpdClassId classId,
                                            SpdRelId id))
{
    char        *relationName = spdGetClassNameById(classId);
    SpdRelId    id = SPD_OBJ_ID_OFFSET(classId);
    FILE        *fp;

    if (num == 0) {
        return;
    }

#ifdef SPD_DUMP_READ
    spdMessage("read_rel_class(): relationName \"%s\", num %d.\n",
               relationName, num);
#endif

    fp = spdOpenSDBfile(spdGetModelDir(spdGetModelId(classId)),
                        relationName, SAPID_O_READ);
    while (0 < num--) {
        /* Ϣidɤ߹߻դ롥*/
        (*fpReadAndInsert)(fp, classId, id++);
    }
    if (feof(fp)) {
        spdErrorExit("read_rel_class(): %s: Too small SDB.\n", relationName);
    } else if (getc(fp) != EOF) {
        spdErrorExit("read_rel_class(): %s: Too large SDB.\n", relationName);
        spdMessage("    feof(): %d.\n", feof(fp));
        spdMessage("    ftell(): %d.\n", ftell(fp));
        spdMessage("    ferror(): %d.\n", ferror(fp));
    }
    spdFclose(fp);

#ifdef SAPID_READ_ROUTINE_MEMORY_PROFILE
    spdMessage("read_rel_class(): \"%s\".\n", relationName);
    spdPrintMemProf();
#endif
}

/******************************************************************************/

void spdReadAndInsertObj(FILE *fp, SpdClassId classId)
{
    append_object(spdReadObj(fp, classId));
}

void spdReadAndInsertRel(FILE *fp, SpdClassId classId, SpdRelId id)
{
    append_relation(spdReadRel(fp, classId, id));
}

/******************************************************************************/

SpdDBObj spdReadObj(FILE *fp, SpdClassId classId)
{
    int         i, argn;
    SpdDBObj    obj;

    SPD_CLEAR_DB_ITEM(&obj, sizeof(SpdDBObj));
    if (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB) {
        for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
            if (SPD_IS_SVAL(argn)) {        /* String. */
                obj.svalue[argn - IVAL_NUM] = spdReadStrAttrText(fp);
            } else {                        /* Integer. */
                obj.ivalue[argn] = spdReadIntAttrText(fp);
            }
        }
        /* Do strict format checking. */
        if (fgetc(fp) != '\n') {
            spdErrorExit("spdReadObj(): Invalid SDB format of obj %d.\n",
                         obj.Obj_id);
        }
    } else {                                /* Binary format. */
        for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
            if (SPD_IS_SVAL(argn)) {        /* String. */
                obj.svalue[argn - IVAL_NUM] = spdReadStrAttrBinary(fp);
            } else {                        /* Integer. */
                obj.ivalue[argn] = spdReadIntAttrBinary(fp);
            }
        }
    }

    if (spdGetClassId((SpdObjId)(obj.Obj_id)) != classId) {
        spdErrorExit("spdReadObj(): Invalid class id %d of object %d.\n",
                     spdGetClassId((SpdObjId)(obj.Obj_id)),
                     obj.Obj_id);
    }

    for (i = 0; i < SpdObjSValN; i++) {
        if (obj.svalue[i] != NULL) {
            obj.svalue[i] = spdStrdup(obj.svalue[i]);
        }
    }

    return (obj);
}

SpdDBRel spdReadRel(FILE *fp, SpdClassId classId, SpdRelId relId)
{
    int         i, argn;
    SpdDBRel    rel;
  
    SPD_CLEAR_DB_ITEM(&rel, sizeof(SpdDBRel));
    if (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB) {
        for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
            rel.ivalue[argn] = spdReadIntAttrText(fp);
        }
        /* Do strict format checking. */
        if (fgetc(fp) != '\n') {
            spdErrorExit("spdReadRel(): Invalid SDB format.\n");
        }
    } else {                                /* Binary format. */
        for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
            rel.ivalue[argn] = spdReadIntAttrBinary(fp);
        }
    }
    rel.Rel_id = relId;
    return rel;
}

/*
 * spdReadObj()ǳݤΰ롥
 */

void spdFreeStrsInObj(SpdDBObj obj)
{
    int     i;

    for (i = 0; i < SpdObjSValN; i++) {
        if (obj.svalue[i] != NULL) {
            spdFree(obj.svalue[i]);
        }
    }
}

/******************************************************************************/

#ifdef NOT_USED

void print_db_obj(SpdDBObj *db_obj)
{
    int         i, argn;
    SpdClassId  classId;

    classId = spdGetClassId(db_obj->Obj_id);
    printf("db_obj <");
    for (i = 0; 0 <= (argn = class_table[classId].seq[i]); i++) {
        if (SPD_IS_SVAL(argn)) {            /* String. */
            printf("\"%s\" ", db_obj->svalue[argn - IVAL_NUM]);
        } else {                            /* Integer. */
            printf("%d ", db_obj->ivalue[argn]);
        }
    }
    printf(">\n");
}

void print_db_rel(SpdDBRel *db_rel)
{
    int         i, argn;
    SpdClassId  classId;

    classId = spdGetClassId(db_rel->Rel_id);
    printf("db_rel <");
    for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
        printf("%d ", db_rel->ivalue[argn]);
    }
    printf(">\n");
}

#endif /* NOT_USED */
