/* 
 * Program:             $RCSfile: spdGlobalObj.c,v $  $Revision: 30.19 $
 * 
 * Purpose:             SDBˡprogram, library, fileʤɤ
 *                      global class갷δؿ
 * 
 * Author:              N.Fukuyasu  2001/01/19
 *                      S.Yamamoto  2001/01/19
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdGlobalObj.c,v 30.19 2003/06/28 15:31:48 yamamoto Exp yamamoto $";
#endif

/* 
 * spdSearchGlobalObj(SpdString className, SpdString name)
 * spdSearchGlobalObjEx(SpdString className, SpdString name)
 * SpdObjId spdCreateGlobalObj(SpdString className, SpdString name)
 * spdWriteGlobalClass(SpdString className)
 *     read_global_class(SpdClassId classId)
 *         read_obj_class(classId, obj_num, read_and_insert_global_obj);
 *             read_and_insert_global_obj(FILE *fp, SpdClassId classId)
 *                 append_global_object(spdReadObj(fp, classId));
 */

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/write_sdb_sub.h>

static SpdClassId   global_class_id = SPD_NON_CID;
static int          a_id, a_name;
static SpdObjId     next_global_id;     /* ID + 1ݻ롥 */

static SpdDBObj     *global_obj = NULL;
static int          global_obj_size = 0;
static int          global_obj_num = 0;

static void free_global_class(void);
static void append_global_object(SpdDBObj obj);
static void read_and_insert_global_obj(FILE *fp, SpdClassId classId);
static SpdClassId read_global_class(SpdClassId classId);
static void write_global_class(SpdClassId classId);

static int equalSimpifiedPath(SpdConstString path1, SpdConstString path2);

/*
 * classNameǻꤵ줿饹°롤̾nameΥ֥Ȥμ
 * ֹ֤¸ߤʤϡSPD_NON_ID֤
 */

SpdObjId spdSearchGlobalObj(SpdString className, SpdString name)
{
    SpdClassId  classId;
    int         i;

    if (className == NULL || name == NULL) {
        return (SPD_NON_ID);
    }

    classId = spdGetClassIdByName(className);
    if (read_global_class(classId) == SPD_NON_ID) {
        return (SPD_NON_ID);
    }

    for (i = 0; i < global_obj_num; i++) {
        if (SPD_STREQ(global_obj[i].svalue[a_name], name)) {
            return (global_obj[i].ivalue[a_id]);
        }
    }
    return (SPD_NON_ID);
}

/*
 * classNameǻꤵ줿饹°롤̾nameΥ֥Ȥμ
 * ֹ֤spdSimplifyPath()ȤäӤ롥
 * ¸ߤʤϡSPD_NON_ID֤
 */

SpdObjId spdSearchGlobalObjEx(SpdString className, SpdString name)
{
    SpdObjId    objId;
    int         i;

    if (className == NULL || name == NULL) {
        return (SPD_NON_ID);
    }

    /* Compare names simply. */
    if ((objId = spdSearchGlobalObj(className, name)) != SPD_NON_ID) {
        return (objId);
    }

    /* Compare names with using spdSimplifyPath(). */
    for (i = 0; i < global_obj_num; i++) {
        if (equalSimpifiedPath(global_obj[i].svalue[a_name], name)) {
            return (global_obj[i].ivalue[a_id]);
        }
    }
    return (SPD_NON_ID);
}

/*
 * className饹ˡ̾nameΥ֥Ȥɲä롥
 * ϡɲä֥Ȥμֹ֤
 * ԤϡSPD_NON_ID֤
 *
 * Note: "id""name"ʳ°Ǥ롥
 */

SpdObjId spdCreateGlobalObj(SpdString className, SpdString name)
{
    static SpdClassId   classId = SPD_NON_CID;
    SpdObjId            new_id;
    SpdDBObj            obj;

    if (className == NULL || name == NULL) {
        return (SPD_NON_ID);
    }

    classId = spdGetClassIdByName(className);
    if (read_global_class(classId) == SPD_NON_ID) {
        return (SPD_NON_ID);
    }

    SPD_CLEAR_DB_ITEM(&obj, sizeof(SpdDBObj));
    new_id = next_global_id++;
    obj.ivalue[a_id]   = new_id;
    obj.svalue[a_name] = spdStrdup(name);
    append_global_object(obj);
    return (new_id);
}

void spdWriteGlobalClass(SpdString className)
{
    SpdClassId  classId;

    if (className == NULL) {
        return;
    }
    classId = spdGetClassIdByName(className);
    write_global_class(classId);
}

/******************************************************************************/

static void free_global_class(void)
{
    int     i, j;

    for (i = 0; i < global_obj_num; i++) {
        for (j = 0; j < SpdObjSValN; j++) {
            spdFree(global_obj[i].svalue[j]);
        }
    }
    spdFree(global_obj);

    global_obj = NULL;
    global_obj_size = 0;
    global_obj_num = 0;
}

/* "<SAPID_TOP>/AR/AR4/db.c"append_object()ꡥ */

static void append_global_object(SpdDBObj obj)
{
    static SpdBoolean   is_first_call = TRUE;

    if (global_obj_size == global_obj_num) {
        global_obj_size += SAPID_SDB_OBJ_INC;
        global_obj = (SpdDBObj *)spdRealloc(global_obj,
                                            sizeof(SpdDBObj) * global_obj_size);
    }
    memcpy(&(global_obj[global_obj_num]), &obj, sizeof(SpdDBObj));
    global_obj_num++;

    if (next_global_id < obj.ivalue[a_id] + 1) {
        next_global_id = obj.ivalue[a_id] + 1;
    }
    if (is_first_call) {
        atexit(free_global_class);
        is_first_call = FALSE;
    }
}

/* "<SAPID_TOP>/AR/AR4/read_sdb.c"spdReadAndInsertObj()ꡥ */

static void read_and_insert_global_obj(FILE *fp, SpdClassId classId)
{
    append_global_object(spdReadObj(fp, classId));
}

/*
 * classIdǻꤵ줿饹Τߤɤ߹ࡥɤ߹饹μֹ֤
 * !! ͤSpdBooleanʬפ
 */

static SpdClassId read_global_class(SpdClassId classId)
{
    SpdInfo     *info;
    int         obj_num;
    SpdString   is_binary;

    if (classId != SPD_CID_PROGRAM && classId != SPD_CID_LIBRARY &&
        classId != SPD_CID_FILE) {
        return (SPD_NON_CID);
    }
    if (classId == global_class_id) {
        /* ꤵ줿饹ɤ߹Ǥ롥 */
        return (classId);
    } else if (global_obj != NULL) {
        /* ꤵ줿饹ʳɤ߹Ǥ롥 */
        free_global_class();
    }

    global_class_id = classId;
    /* next_global_idͤϡ饹ɤ߹ߤǡ
       append_global_object()ˤäѹǽ롥*/
    next_global_id = SPD_OBJ_ID_OFFSET(classId);

    /* a_id = spdGetAttrIdByName3(classId, NULL, "id"); */
    assert(spdGetAttrIdByName3(classId, NULL, "id") == SPD_AID_OBJ_ID);
    a_id = SPD_AID_OBJ_ID;
    a_name = spdGetAttrIdByName3(classId, NULL, "name") - SpdObjIValN;

    /* program, library, file°ΤI-model (2001/01/19) */
    spdUseModel(&I_model);

    info = spdOpenSDBinfo();
    is_binary = spdGetInfoString(info, "BINARY_FORMAT_SDB");
    if (is_binary != NULL && SPD_STREQ(is_binary, "TRUE")) {
        spdSetReadSDBformat(SAPID_BINARY_FORMAT_SDB);
    }
    obj_num = spdGetInfoNumOfObjs(info, spdGetClassNameById(classId));
    spdCloseInfo(info);

    read_obj_class(classId, obj_num, read_and_insert_global_obj);
    return (classId);
}

/* classIdǻꤵ줿饹񤭽Ф */

static void write_global_class(SpdClassId classId)
{
    FILE        *fp;
    char        *className;
    int         i;
    SpdObjId    id;
    SpdString   name;
    SpdInfo     *info;

    if (classId != SPD_CID_PROGRAM && classId != SPD_CID_LIBRARY &&
        classId != SPD_CID_FILE) {
        return;
    }
    if (classId != global_class_id) {
        return;
    }

    className = spdGetClassNameById(classId);
    fp = spdOpenSDBfile(I_MODEL_DIR, className, SAPID_O_WRITE);
    for (i = 0; i < global_obj_num; i++) {
        id = global_obj[i].ivalue[a_id];
        name = global_obj[i].svalue[a_name];
        switch (classId) {
          case SPD_CID_FILE:
            write_file_sub(fp, id, name);
            break;
          case SPD_CID_PROGRAM:
            write_program_sub(fp, id, name);
            break;
          case SPD_CID_LIBRARY:
            write_library_sub(fp, id, name);
            break;
          default:
            spdErrorExit("write_global_class(): %d: Invalid classId.\n",
                         classId);
            break;
        }
    }
    spdFclose(fp);

    info = spdOpenSDBinfo();
    spdSetInfoInt(info, className, global_obj_num);
    spdSetInfoString(info, "BINARY_FORMAT_SDB",
                     spdGetWriteSDBformat() == SAPID_BINARY_FORMAT_SDB ?
                     "TRUE" : "FALSE");
    spdCloseInfo(info);
}

static int equalSimpifiedPath(SpdConstString path1, SpdConstString path2)
{
    SpdString   tmp1, tmp2;
    int         retVal;

    if ((tmp1 = spdSimplifyPath(path1)) == NULL) {
        return (-1);
    }
    if ((tmp2 = spdSimplifyPath(path2)) == NULL) {
        spdFree(tmp1);
        return (-2);
    }
    retVal = SPD_STREQ(tmp1, tmp2);
    spdFree(tmp1);
    spdFree(tmp2);
    return (retVal);
}
