/* 
 * Program:             $RCSfile: spdWriteSDB.c,v $  $Revision: 30.14 $
 * 
 * Purpose:             Write object_db[] and relation_db[] out.
 * 
 * Author:              R.Miura  1997/10/20
 *                      S.Yamamoto  1997/10/20
 * 
 * (C) Copyright:       R.Miura and S.Yamamoto  1997 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdWriteSDB.c,v 30.14 2003/06/28 15:31:48 yamamoto Exp yamamoto $";
#endif

/* 
 * int spdWriteClass(SpdClassId classId);
 * int spdWriteAllClass(SpdString modelName);
 * 
 * History:
 * 1997/10/22, 󥿥եѹؿ̾ѹ
 * 1999/11/03, ѹspdOpenSDBfile(), spdWriteDB{Obj,Rel}()б
 */

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/write_sdb_sub.h>

/*
 * Open file.
 */

static FILE *openClassFile(SpdClassId classId)
{
    return (spdOpenSDBfile(spdGetModelDir(spdGetModelId(classId)),
                           spdGetClassNameById(classId), SAPID_O_WRITE));
}

/******************************************************************************/

/*
 * Write object class.
 */

static void writeOneObjClass(SpdClassId classId, FILE *fp, SpdInfo *info)
{
    SpdDBObj    *obj = object_db;
    int         i, counter = 0;

    for (i = 0; i < object_db_n; i++) {
        if (spdGetClassId((SpdObjId)(obj[i].Obj_id)) == classId) {
            spdWriteDBObj(fp, &obj[i]);
            counter++;
        }
    }
    spdSetInfoInt(info, spdGetClassNameById(classId), counter);
}    

/*
 * Write relation class.
 */

static void writeOneRelClass(SpdClassId classId, FILE *fp, SpdInfo *info)
{
    SpdDBRel    *rel = relation_db;
    int         i, counter = 0;

    for (i = 0; i < relation_db_n; i++) {
        if (spdGetClassId((SpdRelId)(rel[i].Rel_id)) == classId) {
            spdWriteDBRel(fp, &rel[i]);
            counter++;
        }
    }
    spdSetInfoInt(info, spdGetClassNameById(classId), counter);
}

/******************************************************************************/

/*
 * Write all class.
 */

static void writeAllClass(SpdModelId modelId)
{
    FILE **fp_obj, **fp_rel;
    int *obj_counter, *rel_counter;
    int i, j;
    SpdInfo *info;
    SpdClassId first_obj_class_id, last_obj_class_id;
    SpdClassId first_rel_class_id, last_rel_class_id;
    int obj_class_num, rel_class_num;
    SpdClassId classId;


    /* ƥǥͭͤꡥ */
    switch (modelId) {
      case SPD_MID_I_MODEL:
        first_obj_class_id = SPD_FIRST_CID_I_MODEL;
        last_obj_class_id  = SPD_LAST_CID_I_MODEL;
        first_rel_class_id = SPD_FIRST_RID_I_MODEL;
        last_rel_class_id  = SPD_LAST_RID_I_MODEL;
        break;
      case SPD_MID_P_MODEL:
        first_obj_class_id = SPD_FIRST_CID_P_MODEL;
        last_obj_class_id  = SPD_LAST_CID_P_MODEL;
        first_rel_class_id = SPD_FIRST_RID_P_MODEL;
        last_rel_class_id  = SPD_LAST_RID_P_MODEL;
        break;
      default:
        return;
    }

    /* fp_obj, fp_rel, obj_counter, rel_counter ν */
    obj_class_num = last_obj_class_id - first_obj_class_id + 1;
    rel_class_num = last_rel_class_id - first_rel_class_id + 1;
    fp_obj = (FILE **)spdMalloc(sizeof(FILE *) * obj_class_num);
    fp_rel = (FILE **)spdMalloc(sizeof(FILE *) * rel_class_num);
    obj_counter = (int *)spdMalloc(sizeof(int) * obj_class_num);
    rel_counter = (int *)spdMalloc(sizeof(int) * rel_class_num);
    for (classId = first_obj_class_id;
         classId <= last_obj_class_id; classId++) {
        i = classId - first_obj_class_id;
        fp_obj[i] = openClassFile(classId);
        obj_counter[i] = 0;
    }
    for (classId = first_rel_class_id;
         classId <= last_rel_class_id; classId++) {
        i = classId - first_rel_class_id;
        fp_rel[i] = openClassFile(classId);
        rel_counter[i] = 0;
    }

    /* Object class. */
    for (i = 0; i < object_db_n; i++) {
        classId = spdGetClassId((SpdObjId)(object_db[i].Obj_id));
        if (first_obj_class_id <= classId && classId <= last_obj_class_id) {
            j = classId - first_obj_class_id;
            spdWriteDBObj(fp_obj[j], &object_db[i]);
            obj_counter[j]++;
        }
    }

    /* Relation class. */
    for (i = 0; i < relation_db_n; i++) {
        classId = spdGetClassId((SpdRelId)(relation_db[i].Rel_id));
        if (first_rel_class_id <= classId && classId <= last_rel_class_id) {
            j = classId - first_rel_class_id;
            spdWriteDBRel(fp_rel[j], &relation_db[i]);
            rel_counter[j]++;
        }
    }

    for (classId = first_obj_class_id;
         classId <= last_obj_class_id; classId++) {
        i = classId - first_obj_class_id;
        spdFclose(fp_obj[i]);
    }
    for (classId = first_rel_class_id;
         classId <= last_rel_class_id; classId++) {
        i = classId - first_rel_class_id;
        spdFclose(fp_rel[i]);
    }
    spdFree(fp_obj);
    spdFree(fp_rel);

    /* info νϡ */
    info = spdOpenSDBinfo();
    for (classId = first_obj_class_id;
         classId <= last_obj_class_id; classId++) {
        i = classId - first_obj_class_id;
        spdSetInfoInt(info, spdGetClassNameById(classId), obj_counter[i]);
    }
    for (classId = first_rel_class_id;
         classId <= last_rel_class_id; classId++) {
        i = classId - first_rel_class_id;
        spdSetInfoInt(info, spdGetClassNameById(classId), rel_counter[i]);
    }
    spdCloseInfo(info);

    spdFree(obj_counter);
    spdFree(rel_counter);
}

/******************************************************************************/

/* 
 * int spdWriteClass(SpdClassId classId);
 * classIdǻꤵ줿饹spdGetSDBdir()SDB˽Ϥ롥
 * ˥ե뤬¸ߤϾ񤭤롥ｪλ0򡤰۾ｪλ
 * -1֤
 */

int spdWriteClass(SpdClassId classId)
{
    SpdModelId modelId = spdGetModelId(classId);
    FILE *fp;
    SpdInfo *info;

    /* Checking invalid classId. */
    if (modelId != SPD_MID_I_MODEL && modelId != SPD_MID_P_MODEL) {
        return -1;
    }

    fp = openClassFile(classId);
    info = spdOpenSDBinfo();

    if (SPD_IS_CLS_ID(classId)) {
        writeOneObjClass(classId, fp, info); /* Object class. */
    } else {
        writeOneRelClass(classId, fp, info); /* Relation class. */
    }

    spdFclose(fp);
    spdCloseInfo(info);

    return 0;
}

/* 
 * int spdWriteAllClass(SpdString modelName);
 * spdGetSDBdir()SDBmodelNameǻꤵ줿ǥΤ٤Ƥ
 * 饹Ϥ롥 ˥ե뤬¸ߤϾ񤭤롥ｪ
 * λ 0򡤰۾ｪλ-1֤ 
 */

int spdWriteAllClass(SpdString modelName)
{
    SpdModelId modelId;

    if (SPD_STREQ(modelName, "I-model")) {
        modelId = SPD_MID_I_MODEL;
    } else if (SPD_STREQ(modelName, "P-model")) {
        modelId = SPD_MID_P_MODEL;
    } else {
        return -1;
    }

    writeAllClass(modelId);

    return 0;
}
