/* 
 * Program:             $RCSfile: AR.c,v $  $Revision: 30.11 $
 * 
 * Purpose:             桼롼
 * 
 * Author:              T.Hidaka  1998/05/21
 *                      S.Yamamoto  1998/05/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: AR.c,v 30.11 2003/07/08 05:46:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

SpdIdSet spdGetObjIdSet(SpdClassId classId, SpdConstString name)
{
    SpdIdSet ext, res;
    SpdId id;
    SetCursor csr;

    if (name == NULL) {
        return spdGetClassExtent(classId);
    }

    ext = spdGetClassExtentP(classId); /* no need to free */
    res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
    for (csr = spdSetCursorCreate(ext); spdSetCursorNext(&csr);) {
        id = spdSetGetValueO(csr.curElem->value);
        if (SPD_STREQ(name, spdGetAttrStrValByName(id, "name"))) {
            spdAddElemOK(res, id);
        }
    }
    return res;
}

SpdId spddbGetAnObjId(SpdClassId classId, SpdConstString name)
{
    SpdIdSet ext;
    SpdId id, res = SPD_NON_ID;
    SetCursor csr;

    ext = spdGetClassExtentP(classId); /* no need to free */

    if (ext == NULL) {
        return SPD_NON_ID;
    }
    
    for (csr = spdSetCursorCreate(ext); spdSetCursorNext(&csr);) {
        id = spdSetGetValueO(csr.curElem->value);
        if (name == NULL ||
            SPD_STREQ(name, spdGetAttrStrValByName(id, "name"))) {
            res = id;
            break;
        }
    }
    return res;
}

SpdSet2 spdGetRelSet(SpdObjId objectId, SpdClassId relClassId,
                    SpdConstString attrName)
{
    SpdSet2 res;
    SpdIdSet ext;
    SpdBoolean parent, child;
    int attrNo;

    parent = child = FALSE;
    
    /* attrName ơҤΤ줫 */
    if (SPD_STREQ(attrName, "any")) {
        parent = child = TRUE;
    } else {
        attrNo = spdGetAttrNoByName(relClassId, attrName);
        switch (attrNo) {
        case SAPIDDB_ATTR_P_ID:
            child = TRUE;
            break;
        case SAPIDDB_ATTR_C_ID:
            parent = TRUE;
            break;
        }
    }

    /* ƤδϢ */
    if (parent) {
        res = spdGetParentRelSet(objectId, relClassId);
    } else {
        res = NULL;
    }

    /* ҤδϢ */
    if (child) {
        ext = spdGetChildRelSet(objectId, relClassId);
        if (res) {
            res = spdAddSet2(res, ext);
            spdFreeSet2(ext);
        } else {
            res = ext;
        }
    }

    /* ɬפʤ֤ */
    if (res == NULL) {
        res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
        res->valueFree = spdElemFreeSpdRel;
        res->valueCopy = spdElemCopySpdRel;
    }

    return res;
}

SpdSet2 spdGetRelObjSet(SpdObjId objectId, SpdClassId relClassId,
                       SpdConstString attrName)
{
    SpdIdSet res;
    SpdIdSet ext;
    SpdBoolean parent, child;
    int attrNo;

    parent = child = FALSE;
    
    /* attrName ơҤΤ줫 */
    if (SPD_STREQ(attrName, "any")) {
        parent = child = TRUE;
    } else {
        attrNo = spdGetAttrNoByName(relClassId, attrName);
        switch (attrNo) {
        case SAPIDDB_ATTR_P_ID:
            child = TRUE;
            break;
        case SAPIDDB_ATTR_C_ID:
            parent = TRUE;
            break;
        }
    }
    
    /* ƤδϢ */
    if (parent) {
        res = spdGetParentObjIdSet(objectId, relClassId);
    } else {
        res = NULL;
    }

    /* ҤδϢ */
    if (child) {
        ext = spdGetChildObjIdSet(objectId, relClassId);
        if (res) {
            res = spdAddSet2(res, ext);
            spdFreeSet2(ext);
        } else {
            res = ext;
        }
    }

    /* ɬפʤ֤ */
    if (res == NULL) {
        res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
    }

    return res;
}

SpdSet2 spdGetRelIdSet(SpdObjId objectId1, SpdObjId objectId2,
                      SpdClassId relClassId)
{
    SpdIdSet relIds1, relIds2;
    SpdIdSet parents1, parents2;
    SpdIdSet childs1, childs2;

    parents1 = spdGetParentRelIdSet(objectId1, relClassId);
    parents1->valueCopy = spdElemMoveSpdRel;
    childs1 = spdGetChildRelIdSet(objectId1, relClassId);
    childs1->valueCopy = spdElemMoveSpdRel;

    parents2 = spdGetParentRelIdSet(objectId2, relClassId);
    parents2->valueCopy = spdElemMoveSpdRel;
    childs2 = spdGetChildRelIdSet(objectId2, relClassId);
    childs2->valueCopy = spdElemMoveSpdRel;

    /* :
     * νϰŪʤΤǡ free Τ
     *  {parents{1,2}, childs{1,2}}->valueCopy := spdElemMoveSpdRel. 
     * Ȥ뤳ȤǡϢ(SpdRel)֥ȤѤƤ롣
     */

    /* objectId1 λҤɤǤ objectId2 οƤǤΤ */
    relIds1 = spdIntersectionSet2(childs1, parents2);

    /* objectId2 λҤɤǤ objectId1 οƤǤΤ */
    relIds2 = spdIntersectionSet2(childs2, parents1);

    /* ½֤ */
    relIds2 = spdAddSet2(relIds2, relIds1);
    
    spdFreeSet2(parents1);
    spdFreeSet2(childs1);
    spdFreeSet2(parents2);
    spdFreeSet2(childs2);
    spdFreeSet2(relIds1);

    return relIds2;
}

/*
   ֥ȤϢ򤿤ɤäƸĤΥ饹Υ֥Ȥ
   ĤȤ SpdRel ν֤
   */
static SpdSet2 spdSearchObj(SpdSet2 res, SpdObjId objectId, 
                           SpdClassId target, SpdBoolean dir)
{
    SpdSet2 rels;
    SpdRel *relp;
    SetCursor csr;

    assert(res != NULL);

    /* ҤɤδϢƼ */
    if (dir) {
        rels = spdGetParentRelSet(objectId, SPD_NON_ID);
    } else {
        rels = spdGetChildRelSet(objectId, SPD_NON_ID);
    }

    for (csr = spdSetCursorCreate(rels); spdSetCursorNext(&csr);) {
        relp = (SpdRel *)spdSetGetValueP(csr.curElem->value);
        if (spdGetClassId(relp->objectId) == target) {
            /* ̤Ͽ -- ϴϢID */
            spdAddElemOP(res, relp->relationId, relp);
            spdSetSetValueP(&csr.curElem->value, NULL);
            /* *relp  rels  res ˰ư */
        }
        /* Ƶ */
        res = spdSearchObj(res, relp->objectId, target, dir);
    }
    spdFreeSet2(rels);

    return res;
}

SpdSet2 spdSearchChildObj(SpdObjId objectId, SpdClassId target)
{
    SpdSet2 res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    res->valueFree = spdElemFreeSpdRel;
    res->valueCopy = spdElemCopySpdRel;
    return spdSearchObj(res, objectId, target, FALSE);
}

SpdSet2 spdSearchParentObj(SpdObjId objectId, SpdClassId target)
{
    SpdSet2 res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    res->valueFree = spdElemFreeSpdRel;
    res->valueCopy = spdElemCopySpdRel;
    return spdSearchObj(res, objectId, target, TRUE);
}
