/* 
 * Program:             $RCSfile: compat.c,v $  $Revision: 30.10 $
 * 
 * Purpose:             桼롼
 *                      AR3/AR4ߴؿ
 * 
 * Author:              T.Hidaka  1998/05/21
 *                      S.Yamamoto  1998/05/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: compat.c,v 30.10 2003/07/08 05:46:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

SpdString spdGetSDBdir(void);

int spdGetOffset(SpdRelId relationId)
{
    return spdGetAttrIntValByName(relationId, "offset");
}

SpdConstString spdGetClassName(SpdId objectId)
{
    return spdGetClassNameById(spdGetClassId(objectId));
}

SpdConstString spdGetName(SpdId objectId)
{
    return spdGetAttrStrValByName(objectId, "name");
}

int spdGetLength(SpdId objectId)
{
    return spdGetAttrIntValByName(objectId, "length");
}

int spdGetAttrValInt(SpdId objectId, SpdConstString attrName)
{
    return spdGetAttrIntValByName(objectId, attrName);
}

SpdConstString spdGetAttrValString(SpdId objectId, SpdConstString attrName)
{
    return spdGetAttrStrValByName(objectId, attrName);
}

int spdGetAttrValIntOfObj(SpdObjId objectId, SpdConstString attrName)
{
    return spdGetAttrIntValByName(objectId, attrName);
}

int spdGetAttrValIntOfRel(SpdRelId relationId, SpdConstString attrName)
{
    return spdGetAttrIntValByName(relationId, attrName);
}

/* SpdCursorѤؿ */

void spdFreeCursor(SpdCursor *csr)
{
    assert(csr->set != NULL);
    spdFreeSet2(csr->set);
    spdFree(csr);
}

SpdCursor *spddb_CreateCursor(SpdSet2 set)
{
    SpdCursor *csr;

    csr = spdMalloc(sizeof(SpdCursor));
    *csr = spdSetCursorCreate(set);
    return csr;
}

SpdElem *spddb_IterCursor(SpdCursor *csr)
{
    if (spdSetCursorNext(csr)) {
        return csr->curElem;
    } else {
        return NULL;
    }
}

SpdCursor *spdGetObjIdInit(SpdConstString className, SpdConstString name)
{
    return spddb_CreateCursor(spdGetObjIdSet(spdGetClassIdByName(className),
                                             name));
}

SpdObjId spdGetObjId(SpdCursor *csr)
{
    SpdElem *res;
    SpdObjId objId;

    res = spddb_IterCursor(csr);
    if (res) {
        objId = spdSetGetValueO(res->value);
    } else {
        objId = SPD_NON_ID;
    }

    return objId;
}

SpdObjId spdGetAnObjId(SpdConstString className, SpdConstString name)
{
    SpdCursor *csr;
    SpdObjId objId;
    
    csr = spdGetObjIdInit(className, name);
    objId = spdGetObjId(csr);
    spdFreeCursor(csr);

    return objId;
}

SpdCursor *spdGetRelInit(SpdObjId objectId, SpdConstString relationName,
                         SpdConstString attrName)
{
    return spddb_CreateCursor(spdGetRelSet(objectId,
                                           spdGetClassIdByName(relationName),
                                           attrName));
}

SpdRel spdGetRel(SpdCursor *csr)
{
    SpdElem *res;
    SpdRel rel;
    static SpdRel nonrel = {SPD_NON_ID, SPD_NON_ID};

    res = spddb_IterCursor(csr);
    if (res) {
        rel = *(SpdRel *)spdSetGetValueP(res->value);
    } else {
        rel = nonrel;
    }

    return rel;
}

SpdRel spdGetARel(SpdObjId objectId, SpdConstString relationName,
                  SpdConstString attrName)
{
    SpdCursor *csr;
    SpdRel rel;
    
    csr = spdGetRelInit(objectId, relationName, attrName);
    rel = spdGetRel(csr);
    spdFreeCursor(csr);

    return rel;
}

SpdCursor *spdGetRelObjInit(SpdObjId objectId, SpdConstString relationName,
                            SpdConstString attrName)
{
    return spddb_CreateCursor(spdGetRelObjSet(objectId,
                                              spdGetClassIdByName(relationName),
                                              attrName));
}

SpdRelId spdGetRelObj(SpdCursor *csr)
{
    SpdElem *res;
    SpdRelId relId;

    res = spddb_IterCursor(csr);
    if (res) {
        relId = spdSetGetValueO(res->value);
    } else {
        relId = SPD_NON_ID;
    }

    return relId;
}

SpdRelId spdGetARelObj(SpdObjId objectId, SpdConstString relationName,
                       SpdConstString attrName)
{
    SpdCursor *csr;
    SpdRelId relId;
    
    csr = spdGetRelObjInit(objectId, relationName, attrName);
    relId = spdGetRelObj(csr);
    spdFreeCursor(csr);

    return relId;
}

SpdCursor *spdGetRelIdInit(SpdObjId objectId1, SpdObjId objectId2,
                           SpdConstString relationName)
{
    return spddb_CreateCursor(spdGetRelIdSet(objectId1, objectId2,
                                             spdGetClassIdByName(relationName)));
}

SpdRelId spdGetRelId(SpdCursor *csr)
{
    SpdElem *res;
    SpdRelId relId;

    res = spddb_IterCursor(csr);
    if (res) {
        relId = spdSetGetValueO(res->value);
    } else {
        relId = SPD_NON_ID;
    }

    return relId;
}

SpdRelId spdGetARelId(SpdObjId objectId1, SpdObjId objectId2,
                      SpdConstString relationName)
{
    SpdCursor *csr;
    SpdRelId relId;
    
    csr = spdGetRelIdInit(objectId1, objectId2, relationName);
    relId = spdGetRelId(csr);
    spdFreeCursor(csr);

    return relId;
}

/* ǡ١ϴط */
int spdWriteClass(SpdClassId classId, SpdConstString dirName)
{
    SpdConstString oldSDBdir;

    oldSDBdir = spdStrdup(spdGetSDBdir());
    spdSetSDBdir(dirName);
    
    spdWriteAClass(classId, SPD_NON_ID);

    spdSetSDBdir(oldSDBdir);
    spdFree(oldSDBdir);
    
    return 0;
}

