/* 
 * Program:             $RCSfile: object.c,v $  $Revision: 30.46 $
 * 
 * Purpose:             ֥Ϣ
 * 
 * Author:              T.Hidaka  1998/05/21
 *                      S.Yamamoto  1998/05/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: object.c,v 30.46 2003/07/31 12:29:57 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdOption spdDefaultOption[] = {
    {"mode", SAPID_C_MODE},
    {"prototype", SPDSTR_PROTOTYPE_NON},
    {"typeQualifier", OFF},
    {NULL, SAPID_NON_OPTION}
};

/* 
 * ֥ȤΥǥå
 * ʣΥǡ١ɤ߹ȤѤ롥
 */

typedef struct {
    SpdSet2 objects;         /* oid򥭡Ȥ֥Ȥν硥 */
    SpdSet2 extents;         /* classId򥭡Ȥ
                               ֥IDνν硥 */
    SpdSet2 ridxes;          /* ϢΥǥåν硥 */
} SpdDBIndex;

/* ƴϢ饹ȤΥǥå */

typedef struct {
    SpdRelId rel;
    SpdObjId pid;
    SpdObjId cid;
} SpdRelCache;

typedef struct {
    int size;
    SpdRelCache *prels;
    SpdRelCache *crels;
} SpdRelIndex;

static SpdDBId defaultDBId = SPD_NON_ID;
static SpdDBIndex **indexes = NULL;
static int nindexes = 0;

static SpdDBId spddb_CreateDBIndex(void);
static void spddb_FreeDBIndex(SpdDBId dbId);
static SpdObjId mkNewObjId(SpdClassId classId);
static void spddb_FreeDBObject(SpdElemValue elem);
static void spddb_FreeSet(SpdElemValue elem);

/* Ǥη */
#define OBJID 0
#define RELID 1
#define REL   2

/* Ϣ򤿤ɤ */
#define PARENT 0
#define CHILD  1

/* Copied from "AR/AR4/ar.c" */

static void check_version(SpdInfo *info)
{
    if (spdGetInfoInt(info, "SapidSDBVersion", 0) != SapidSDBVersion) {
        spdErrorExit("Invalid SDB version %d (should be %d).\n",
                     spdGetInfoInt(info, "SapidSDBVersion", 0),
                     SapidSDBVersion);
    }
    if (spdGetInfoInt(info, "SapidPIDBVersion", 0) != SapidPIDBVersion) {
        spdErrorExit("Invalid PIDB version %d (should be %d).\n",
                     spdGetInfoInt(info, "SapidPIDBVersion", 0),
                     SapidPIDBVersion);
    }
}

/*
 * ֥ȤΥǥåؿ
 */

/* SpdDBId"SpdDBIndex *"Ѵ롥 */

static SpdDBIndex *spddb_GetDBIndex(SpdDBId dbId)
{
    if (dbId == SPD_NON_ID) {
        dbId = defaultDBId;
    }
    if (dbId == SPD_NON_ID || dbId >= nindexes) {
        spdErrorExit("Database hasn't been initialized.\n");
    }
    if (indexes[dbId] == NULL) {
        spdErrorExit("Database has already freed.\n");
    }
    return indexes[dbId];
}

static void spddb_DeleteDBIndex(SpdDBId dbId)
{
    if (dbId == SPD_NON_ID) {
        dbId = defaultDBId;
    }
    if (dbId == SPD_NON_ID || dbId >= nindexes) {
        spdErrorExit("Database hasn't been initialized.\n");
    }
    indexes[dbId] = NULL;
}

/* ϢΥǥå롥 */

static int spddb_compare_parent_id(const SpdRelCache *a, const SpdRelCache *b)
{
    SpdObjId    ap, bp;

    ap = a->pid;
    bp = b->pid;
    return ((ap - bp != 0) ? ap - bp : 
            spdGetClassId(a->rel) - spdGetClassId(b->rel));
}

static int spddb_compare_child_id(const SpdRelCache *a, const SpdRelCache *b)
{
    SpdObjId    ap, bp;

    ap = a->cid;
    bp = b->cid;
    return ((ap - bp != 0) ? ap - bp : 
            spdGetClassId(a->rel) - spdGetClassId(b->rel));
}

/* !! ǥå/μˡѹ */

static void spddb_FreeARelIndex(SpdElemValue elem)
{
    SpdRelIndex *ridx = spdSetGetValueP(elem);

    if (ridx == NULL) {
        return;
    }

    /* ǥåȤ롥 */
    if (ridx->prels) {
        spdFree(ridx->prels);
        ridx->prels = NULL;
    }
    if (ridx->crels) {
        spdFree(ridx->crels);
        ridx->crels = NULL;
    }
    spdFree(ridx);
}

static void spddb_FreeRelIndex(SpdDBIndex *idx, SpdClassId classId)
{
    if (idx->ridxes == NULL) {
        return;
    }

    if (classId == SPD_NON_ID) {
        /* ƤΥǥå */
        spdFreeSet2(idx->ridxes);
        idx->ridxes = NULL;
    } else {
        /* classIdǻꤵ줿饹Υǥå */
        spdDelElemO(idx->ridxes, classId);
    }
}

static SpdRelCache *spddb_allocIdArray(SpdSet2 set)
{
    SetCursor csr;
    SpdRelCache *res;
    SpdRelCache *p;

    p = res = (SpdRelCache *)spdMalloc(sizeof(SpdRelCache) * set->elemNum);
    for (csr = spdSetCursorCreate(set); spdSetCursorNext(&csr);) {
        p->rel = spdSetGetValueO(csr.curElem->value);
        p->pid = spdGetAttrIntVal(p->rel, SAPIDDB_ATTR_P_ID);
        p->cid = spdGetAttrIntVal(p->rel, SAPIDDB_ATTR_C_ID);
        p++;
    }
    return res;
}


/* !! ǥå/μˡѹ */

static SpdRelIndex *
spddb_MakeRelIndex(SpdDBIndex *idx, SpdClassId classId, int dir)
{
    SpdElem *elem;
    SpdSet2 ids;
    SpdRelIndex *ridx = NULL;

    /* classId ϢǤʤкʤ */
    if (classId == SPD_NON_ID || 
        !(spdGetClassSort(classId) & SAPIDDB_SORT_REL)) {
        return NULL;
    }
    
    /* ǥåν礬ʤн缫Τ */
    if (idx->ridxes == NULL) {
        idx->ridxes = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
        idx->ridxes->valueFree = spddb_FreeARelIndex;
    } else if ((elem = spdSearchElemO(idx->ridxes, classId)) != NULL) {
        /* Ǥ˥ǥåкʤ */
        ridx = spdSetGetValueP(elem->value);
        switch (dir) {
        case PARENT:
            if (ridx->crels != NULL) return ridx;
            break;
        case CHILD:
            if (ridx->prels != NULL) return ridx;
            break;
        }
    }
    
    /* ǥå */
    elem = spdSearchElemO(idx->extents, classId);
    if (elem == NULL) {
        return NULL;
    }
    ids = spdSetGetValueP(elem->value);
    assert(ids != NULL);

    if (ridx == NULL) {
        ridx = (SpdRelIndex *)spdMalloc(sizeof(SpdRelIndex));
        ridx->size = ids->elemNum;
        ridx->crels = NULL;
        ridx->prels = NULL;
        spdAddElemOP(idx->ridxes, classId, ridx); 
    }
    switch (dir) {
    case PARENT:
        ridx->crels = spddb_allocIdArray(ids);
        SPD_QSORT(ridx->crels, ridx->size, sizeof(SpdRelCache),
                    spddb_compare_child_id);
        break;
    case CHILD:
        ridx->prels = spddb_allocIdArray(ids);
        SPD_QSORT(ridx->prels, ridx->size, sizeof(SpdRelCache),
                    spddb_compare_parent_id);
        break;
    default:
        spdErrorExit("Unknown dir: %d\n", dir);
    }
    return ridx;
}

/* ֥ȤΥǥå롥 */

static SpdDBId spddb_CreateDBIndex(void)
{
    SpdDBIndex  *idx;
    SpdDBId     dbId;
    
    idx = spdMalloc(sizeof(SpdDBIndex));

    dbId = nindexes++;
    indexes = spdRealloc(indexes, sizeof(SpdDBIndex *) * nindexes);
    indexes[dbId] = idx;
    
    idx->objects = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    idx->objects->valueFree = spddb_FreeDBObject;
    idx->objects->valueCopy = NULL; /* shares SpdDBObjs. */

    idx->extents = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    idx->extents->valueFree = spddb_FreeSet;
    idx->extents->valueCopy = NULL;
    idx->ridxes = NULL;
    
    if (defaultDBId == SPD_NON_ID) {
      defaultDBId = dbId;
    }
    
    return dbId;
}

/*
 * ֥Ȥؿ
 */

/* IdObjectѴ롥 */

static SpdDBObj *spddb_GetDBObjByIdInDB(SpdDBIndex *idx, SpdId id)
{
    static SpdDBIndex *cacheIdx = NULL;
    static SpdId cacheId = SPD_NON_ID;
    static SpdDBObj *cacheDBObj = NULL;
    SpdElem *elem;
    
    if (idx == NULL || id == SPD_NON_ID) {
        cacheIdx = NULL;
        cacheId = SPD_NON_ID;
        cacheDBObj = NULL;
        return NULL;
    }

    if (cacheId != SPD_NON_ID && cacheId == id && cacheIdx == idx) {
        return cacheDBObj;
    }

    cacheIdx = idx;
    cacheId = id;
    elem = spdSearchElemO(idx->objects, id);
    cacheDBObj = elem != NULL ? spdSetGetValueP(elem->value) : NULL;
    return cacheDBObj;
}

SpdDBObj *spddb_GetDBObjById(SpdId id)
{
    return spddb_GetDBObjByIdInDB(spddb_GetDBIndex(defaultDBId), id);
}

SpdDBObj spddb_CreateDBObj(SpdClassId classId, SpdId objId)
{
    SpdDBObj    dbObj;
    int         nattr;
    
    /* SpdDBObj κ */
    dbObj.id = objId;
    dbObj.classId = classId;
    /* value γ */
    nattr = spddb_GetValueNum(classId);
    if (nattr) {
        dbObj.value = spdMalloc(sizeof(SpdDBValue) * nattr);
        spddb_InitDBObjValue(&dbObj);
    } else {
        dbObj.value = NULL;
    }

    return dbObj;
}

static void spddb_FreeDBObject(SpdElemValue elem)
{
    SpdDBObj *dbObj;

    dbObj = spdSetGetValueP(elem);
    spddb_FreeDBObjValue(dbObj);
    if (dbObj->value != NULL) {
        spdFree(dbObj->value);
    }
    spdFree(dbObj);
}

#if 0  /* ȤƤʤ? */
static void spddb_FreeDBValue(SpdDBIndex *idx, SpdId objId)
{
    SpdDBObj *dbObj;

    dbObj = spddb_GetDBObjByIdInDB(idx, objId);

    if (dbObj == NULL)
        return;

    if (dbObj->value) {
        spddb_FreeDBObjValue(dbObj);
        spdFree(dbObj->value);
    }
}
#endif

/* ǡ١Υǥå롥 */

static void spddb_FreeDBIndex(SpdDBId dbId)
{
    SpdDBIndex  *idx;

    idx = spddb_GetDBIndex(dbId);

    if (idx == NULL)                        /* ˳ */
        return;

    if (idx->objects) {
        spdFreeSet2(idx->objects);
    }

    if (idx->extents) {
        spdFreeSet2(idx->extents);
    }

    spddb_FreeRelIndex(idx, SPD_NON_ID);
    
    if (dbId == defaultDBId) {
      defaultDBId = SPD_NON_ID;
    }
    spdFree(idx);
    
    spddb_DeleteDBIndex(dbId);
}

static void spddb_FreeSet(SpdElemValue elem)
{
    spdFreeSet2(spdSetGetValueP(elem));
}

void spddb_WriteSDBInfo(SpdInfo *info, SpdDBId dbId)
{
    /* Empty. */
}

/* SapidDBνȽλ */

SpdDBId spdOpenSDB(void)
{
    SpdString           SDBdir;
    SpdInfo             *info;
    static SpdBoolean   is_first_call = TRUE;
    
    SDBdir = spdGetSDBdir();
    if (!spdIsDirectory(SDBdir)) {
        spdMkdirR(SDBdir);
    }

    /* Copied from spdTarget() in "AR4/ar.c". */
    info = spdOpenSDBinfo();
    check_version(info);
    spdInsertOption(spdDefaultOption);
    if (SPD_STREQ(spdGetInfoString(info, "BINARY_FORMAT_SDB"), "TRUE")) {
        spdSetReadSDBformat(SAPID_BINARY_FORMAT_SDB);
        spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
    } 
    spdCloseInfo(info);

    /* Added by S.Yamamoto 2001/11/23. */
    if (is_first_call) {
        spdInitDbInfo();
        is_first_call = FALSE;
    }
    spdOpenDbInfo();

    return spddb_CreateDBIndex();
}

void spdCloseSDB(SpdDBId dbId)
{
    spddb_FreeDBIndex(dbId);

    /* Added by S.Yamamoto 2001/07/05. */
    spdReleasePidbCache();
    spdReleaseLineTbl();

    /* Added by S.Yamamoto 2001/11/23. */
    spdInitDbInfo();
}

void spddb_FreeAllDBIndex(void)
{
    int i;
    
    for (i = 0; i < nindexes; i++) {
        spddb_FreeDBIndex(i);
    }
    spdFree(indexes);
    indexes = NULL;
    nindexes = 0;
}

/* ǥåϿ롥 */
/* *dbobj ʣʤᡢspdMalloc() ΰݤɬפ */
SpdId spddb_RegisterDBObj(SpdDBObj *dbobj, SpdDBId dbId)
{
    SpdIdSet    classextent;
    SpdElem     *elem;
    SpdDBIndex  *idx;

    if (dbobj->id == SPD_NON_ID) {
        dbobj->id = mkNewObjId(dbobj->classId);
    }        

    idx = spddb_GetDBIndex(dbId);

    /* Added by S.Yamamoto 2001/09/07. */
    if (SPD_ID_SPACE < (dbobj->id - SPD_OBJ_ID_OFFSET(dbobj->classId + 1))) {
        spdErrorExit("spddb_RegisterDBObj(): Class \"%s\", id %d, num %d: Too many objects.\n",
                     spdGetClassNameById(dbobj->classId), dbobj->id, 
                     dbobj->id - SPD_OBJ_ID_OFFSET(dbobj->classId) + 1);
    }

    /* ֥ȤνϿ롥 */
    if (spdAddElemOP(idx->objects, dbobj->id, dbobj) == NULL) {
        /* object idʣƤ롥 */
        spdErrorExit("spddb_RegisterDBObj(): %d: Duplicate object id.\n", 
                     dbobj->id);
    }
    
    /* 饹ȤνϿ롥 */
#if 0
    elem = spdSearchElemO(idx->extents, dbobj->classId);
    if (elem != NULL) {
        classextent = spdSetGetValueP(elem->value);
    } else if (dbobj->classId != SPD_NON_ID) {
        classextent = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
        spdAddElemOP(idx->extents, dbobj->classId, classextent);
    }
    if (classextent != NULL) {
        spdAddElemOK(classextent, dbobj->id);
    }
#else
    if (dbobj->classId != SPD_NON_ID) {
        elem = spdSearchElemO(idx->extents, dbobj->classId);
        if (elem != NULL) {
            classextent = spdSetGetValueP(elem->value);
        } else {
            classextent = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
            spdAddElemOP(idx->extents, dbobj->classId, classextent);
        }
        spdAddElemOK(classextent, dbobj->id);
    }
#endif

    if (spdGetClassSort(dbobj->classId) & SAPIDDB_SORT_REL) {
        /* ϢϿ줿餳ޤǤΥǥå̵롥 */
        spddb_FreeRelIndex(idx, dbobj->classId);
    }
    
    return dbobj->id;
}

/*
 * ǥå饪֥Ȥ롥DapidΤpublicˤ
 */

void spddb_UnRegisterDBObj(SpdId id, SpdDBId dbId)
{
    SpdDBIndex *idx;
    SpdClassId classId;
    SpdIdSet classextent;

    classId = spdGetClassId(id);
    
    idx = spddb_GetDBIndex(dbId);
    
    /* ֥Ȱ */
    spdDelElemO(idx->objects, id);

    /* 饹ȤΥ֥ID */
    classextent = spdSetGetValueP(spdSearchElemO(idx->extents, classId)->value);
    if (classextent) {
        spdDelElemO(classextent, id);
    }
    /* ϢΥ饹Υǥå̵ */
    spddb_FreeRelIndex(idx, classId);
}

/* ǥեȤΥǡ١ѹ */

SpdDBId spdSetDefaultDB(SpdDBId dbId)
{
    SpdDBId oldDBId;

    oldDBId = defaultDBId;
    defaultDBId = dbId;

    return oldDBId;
}

/* ǥեȤΥǡ١μ */

SpdDBId spdGetDefaultDB(void)
{
    return defaultDBId;
}

/* 
 * ֥Ȥؿ
 */

/* Classextent (֥)롥 */

SpdIdSet spdGetClassExtentP(SpdClassId classId)
{
    SpdDBIndex  *idx;
    SpdIdSet    pres;
    SpdElem     *elem;

    idx = spddb_GetDBIndex(defaultDBId);
    elem = spdSearchElemO(idx->extents, classId);
    if (elem == NULL) {
        return NULL;
    }
    pres = spdSetGetValueP(elem->value);
    return pres;  /* ʣƤʤȤ!! */
}

SpdIdSet spdGetClassExtent(SpdClassId classId)
{
    SpdIdSet res = spdGetClassExtentP(classId);
    return res != NULL ? spdDupSet2(res) :
        spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
}

/* ֥ȤȴϢμֹ֤ */

static SpdObjId mkNewObjId(SpdClassId classId)
{
    if (spdIsValidClassOrRelationId(classId)) {
        SpdObjId    tmpMaxId;

        tmpMaxId = spdGetDbInfoTmpMaxId(classId);
        tmpMaxId = (tmpMaxId == SPD_NON_ID) ? 
            SPD_OBJ_ID_OFFSET(classId) : tmpMaxId + 1;
        spdSetDbInfoTmpMaxId(classId, tmpMaxId);
        /* !! Do range check! */
        return (tmpMaxId);
    } else {
        spdErrorExit("mkNewObjId(): %d: Invalid class id.\n", classId);
        SPD_DUMMY_RETURN(SPD_NON_ID);
    }
}

/*
 * ֥Ȥ롥
 * Note: argument objId is not used.
 */

SpdId spdCreateObj(SpdClassId classId, SpdId objId)
{
    SpdDBObj    *obj;

    obj = (SpdDBObj *)spdMalloc(sizeof(SpdDBObj));
    *obj = spddb_CreateDBObj(classId, mkNewObjId(classId));
    if (obj->id == SPD_NON_ID) {
        spdErrorExit("spdCreateObj(): classId %d, objId %d.\n",
                     classId, obj->id);
    }
    return spddb_RegisterDBObj(obj, defaultDBId);
}

/* 󥯤롥 */

SpdRelId spdCreateRel(SpdClassId classId, SpdRelId relId,
                      SpdObjId parentId, SpdObjId childId)
{
    relId = spdCreateObj(classId, relId);
    spdSetAttrIntVal(relId, SAPIDDB_ATTR_P_ID, parentId);
    spdSetAttrIntVal(relId, SAPIDDB_ATTR_C_ID, childId);
    return relId;
}

/*
 * ֥Ȥ롥
 * id֥Ȥξϡ֥ȤդƤϢ롥
 */ 

void spdDeleteObj(SpdId id)
{
    if (spdGetClassSort(spdGetClassId(id)) & SAPIDDB_SORT_OBJ) {
        SpdSet2         rels;
        SpdRel          *rel;
        SetCursor       csr;

        rels = spdGetParentRelSet(id, SPD_NON_ID);
        for (csr = spdSetCursorCreate(rels); spdSetCursorNext(&csr);) {
            rel = spdSetGetValueP(csr.curElem->value);
            spdDeleteObj(rel->relationId);
        }
        spdFreeSet2(rels);
        
        rels = spdGetChildRelSet(id, SPD_NON_ID);
        for (csr = spdSetCursorCreate(rels); spdSetCursorNext(&csr);) {
            rel = spdSetGetValueP(csr.curElem->value);
            spdDeleteObj(rel->relationId);
        }
        spdFreeSet2(rels);
     }

    /* ǥå */
    spddb_UnRegisterDBObj(id, defaultDBId);
}

/*----------------------------------------------------------------------
  ֥Ȥ°ͤؿ
  ----------------------------------------------------------------------*/

/* ֥Ȥ°ͤμ */

static int spddb_GetAttrIntVal_in(SpdDBObj *obj, int attrNo)
{
    SpdDBType   tp;
    int         seq;
    
    if (obj == NULL || obj->value == NULL) {
        return SAPID_NON_VALUE;
    }

    tp = spdGetAttrType(obj->classId, attrNo);

    switch (tp) {
      case SPD_DBT_INT:
      case SPD_DBT_OID:
      case SPD_DBT_ENUM:
        seq = spddb_GetAttrSeq(obj->classId, attrNo);
        if (seq >= 0) {
            return spddb_GetAttrBitfieldVal(obj->classId, attrNo,
                                            obj->value[seq].ivalue);
        } else {
            switch (seq) {
              case SAPIDDB_ATTR_ID:
                return obj->id;
              case SAPIDDB_ATTR_CLASSID:
                return obj->classId;
              default:
                return SAPID_NON_VALUE;
            }
        }
        
      default:
        return SAPID_NON_VALUE;
    }
}

/* ֥Ȥ°륯饹 */

SpdClassId spdGetClassId(SpdId id)
{
    return spddb_GetAttrIntVal_in(spddb_GetDBObjById(id),
                                  SAPIDDB_ATTR_CLASSID);
}

int spdGetAttrIntVal(SpdId id, int attrNo)
{
    return spddb_GetAttrIntVal_in(spddb_GetDBObjById(id), attrNo);
}

int spdGetAttrIntValByName(SpdId id, SpdConstString attrName)
{
    SpdDBObj *obj = spddb_GetDBObjById(id);

    if (obj == NULL)
        return SAPID_NON_VALUE;
    
    return spddb_GetAttrIntVal_in(obj, spdGetAttrNoByName(obj->classId,
                                                          attrName));
}

static SpdConstString spddb_GetAttrStrVal_in(SpdDBObj *obj, int attrNo)
{
    SpdDBType   tp;
    
    if (obj == NULL || obj->value == NULL) {
        return NULL;
    }

    tp = spdGetAttrType(obj->classId, attrNo);
    switch (tp) {
      case SPD_DBT_STR:
        return obj->value[spddb_GetAttrSeq(obj->classId, attrNo)].svalue;
      case SPD_DBT_ENUM:
        return spdGetEnumStrVal(obj->classId, attrNo,
                                spddb_GetAttrIntVal_in(obj, attrNo));
      case SPD_DBT_OID:
      default:
        return NULL;
    }
}

SpdConstString spdGetAttrStrVal(SpdId id, int attrNo)
{
    return spddb_GetAttrStrVal_in(spddb_GetDBObjById(id), attrNo);
}

SpdConstString spdGetAttrStrValByName(SpdId id, SpdConstString attrName)
{
    SpdDBObj *obj = spddb_GetDBObjById(id);

    if (obj == NULL)
        return NULL;

    return spddb_GetAttrStrVal_in(obj, 
                                  spdGetAttrNoByName(obj->classId, attrName));
}

/* ֥Ȥ°ͤ */

static int spddb_SetAttrIntVal_in(SpdDBObj *obj, int attrNo, int value)
{
    SpdDBType   tp;
    int         seq;
    
    if (obj == NULL) {
        return SAPID_NON_VALUE;
    }

    tp = spdGetAttrType(obj->classId, attrNo);

    switch (tp) {
      case SPD_DBT_INT:
      case SPD_DBT_OID:
      case SPD_DBT_ENUM:
        seq = spddb_GetAttrSeq(obj->classId, attrNo);
        return obj->value[seq].ivalue =
            spddb_SetAttrBitfieldVal(obj->classId, attrNo,
                                     obj->value[seq].ivalue, value);
      default:
        return SAPID_NON_VALUE;
    }
}

int spdSetAttrIntVal(SpdId id, int attrNo, int value)
{
    return spddb_SetAttrIntVal_in(spddb_GetDBObjById(id), attrNo, value);
}

int spdSetAttrIntValByName(SpdId id, SpdConstString attrName, int value)
{
    SpdDBObj *obj = spddb_GetDBObjById(id);

    if (obj == NULL)
        return SAPID_NON_VALUE;
    
    return spddb_SetAttrIntVal_in(obj,
                                  spdGetAttrNoByName(obj->classId, attrName),
                                  value);
}

static SpdConstString spddb_SetAttrStrVal_in(SpdDBObj *obj, int attrNo,
                                             SpdConstString value)
{
    SpdDBType   tp;
    SpdString   *psvalue;
    
    if (obj == NULL || obj->value == NULL) {
        return NULL;
    }

    tp = spdGetAttrType(obj->classId, attrNo);
    switch (tp) {
      case SPD_DBT_STR:
        /* ʸ󷿤°ͤ񤭴 */
        psvalue = &(obj->value[spddb_GetAttrSeq(obj->classId, attrNo)].svalue);
        if (*psvalue)
            spdFree(*psvalue);
        return *psvalue = spdStrdup(value);
      case SPD_DBT_ENUM:
        /* 󷿤°ͤ񤭴 */
        obj->value[spddb_GetAttrSeq(obj->classId, attrNo)].ivalue =
            spdGetEnumIntVal(obj->classId, attrNo, value);
        return spdGetEnumStrVal(obj->classId, attrNo,
                                spddb_GetAttrIntVal_in(obj, attrNo));
      case SPD_DBT_OID:
      default:
        return NULL;
    }
}

SpdConstString spdSetAttrStrVal(SpdId id, int attrNo, SpdConstString value)
{
    return spddb_SetAttrStrVal_in(spddb_GetDBObjById(id), attrNo, value);
}

SpdConstString spdSetAttrStrValByName(SpdId id, SpdConstString attrName,
                                      SpdConstString value)
{
    SpdDBObj *obj = spddb_GetDBObjById(id);

    if (obj == NULL)
        return NULL;
    
    return spddb_SetAttrStrVal_in(obj,
                                  spdGetAttrNoByName(obj->classId, attrName),
                                  value);
}

/*----------------------------------------------------------------------
  ϢΥؿ
  ----------------------------------------------------------------------*/

/* ϢõѴؿ */

static int spddb_compare_srch_parent_id(const SpdObjId objId,
                                        const SpdRelCache *relc)
{
    return objId - relc->pid;
}

static int spddb_compare_srch_child_id(const SpdObjId objId,
                                       const SpdRelCache *relc)
{
    return objId - relc->cid;
}

static int spddb_rel_bsearch(SpdObjId id, SpdRelCache *rels, int n,
                             int (*compar)(const SpdObjId, const SpdRelCache *))
{
    int left, right, mid;

    left = 0;
    right = n - 1;

    while (left < right) {
        mid = (left + right) / 2;
        if (compar(id, rels + mid) > 0)
            left = mid + 1;
        else
            right = mid;
    }

    if (compar(id, rels + left) != 0) {
        return -1;
    } else {
        while (left > 0 && compar(id, rels + left - 1) == 0) {
            left--;
        }
    }

    return left;
}

/* Υ饹δϢ򤿤ɤäƽresɲ */

static SpdSet2 spddb_GetARelSet(SpdSet2 res, SpdObjId id, 
                                SpdClassId relClassId,
                                int type, int dir,
                                SpdConstString objAttr, int objValue,
                                SpdConstString relAttr, int relValue)
{
    SpdDBIndex *idx;
    SpdRelIndex *ridx;
    SpdRelCache *rels = NULL;
    int (*compare)(const SpdObjId, const SpdRelCache *) = NULL;
    int p;
    SpdObjId objId;
    SpdRelId relId;
    SpdRel *rel;

    /* relClassId  SPD_NON_ID or Ϣ饹Ǥʤϲ⤷ʤ */
    if (relClassId == SPD_NON_ID ||
        !(spdGetClassSort(relClassId) & SAPIDDB_SORT_REL)) {
        return res;
    }
    idx = spddb_GetDBIndex(defaultDBId);
    ridx = spddb_MakeRelIndex(idx, relClassId, dir);
    if (ridx == NULL) { /* relClassId idx->extents  ϿƤʤȤ */
        return res;                         /* ʥ饹ID? */
    }

    switch (dir) {
      case PARENT:
        rels = ridx->crels;
        compare = spddb_compare_srch_child_id;
        break;
      case CHILD:
        rels = ridx->prels;
        compare = spddb_compare_srch_parent_id;
        break;
    }

    if (rels == NULL || ridx->size <= 0) {
        return res;
    }

    p = spddb_rel_bsearch(id, rels, ridx->size, compare);
    
    /* ߤĤʤ */
    if (p < 0) {
        return res;
    }
    
    if (type == RELID) {
        for (; p < ridx->size && compare(id, rels + p) == 0; p++) {
            relId = (rels + p)->rel;
            spdAddElemOK(res, relId);
        }
    } else {
        for (; p < ridx->size && compare(id, rels + p) == 0; p++) {
            relId = (rels + p)->rel;
            if (dir == PARENT) {
                objId = (rels + p)->pid;
            } else {
                objId = (rels + p)->cid;
            }
            
            if ((objAttr == NULL || 
                 spdGetAttrIntValByName(objId, objAttr) == objValue) &&
                (relAttr == NULL || 
                 spdGetAttrIntValByName(relId, relAttr) == relValue)) {
                if (type == OBJID) {
                    spdAddElemOK(res, objId);
                } else {
                    rel = (SpdRel *)spdMalloc(sizeof(SpdRel));
                    rel->relationId = relId;
                    rel->objectId = objId;
                    spdAddElemOP(res, relId, rel);
                }
            }
        }
    }
    return res;
}

/* Ϣ򤿤ɤ */
/* relClassId == SPD_NON_ID ξƤΥ饹δϢ򤿤ɤ */

static SpdSet2 spddb_GetRelSet(SpdObjId id, SpdClassId relClassId, 
                              int type, int dir,
                              SpdConstString objAttr, int objValue,
                              SpdConstString relAttr, int relValue)
{
    SpdSet2 res = NULL;
    SpdSet2 classes;

    switch (type) {
    case OBJID:
    case RELID:
        res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
        break;
    case REL:
        res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE); 
        res->valueFree = spdElemFreeSpdRel;
        res->valueCopy = spdElemCopySpdRel;
        break;
    }

    if (relClassId == SPD_NON_ID) {
        SetCursor csr;
        SpdClassId classId;

        classes = spdGetAllClasses(SPD_NON_ID);
        for (csr = spdSetCursorCreate(classes); spdSetCursorNext(&csr);) {
            classId = spdSetGetValueO(csr.curElem->value);
            if (spdGetClassSort(classId) & SAPIDDB_SORT_REL) {
                res = spddb_GetARelSet(res, id, classId, type, dir,
                                       objAttr, objValue, relAttr, relValue);
            }
        }
        spdFreeSet2(classes);
    } else {
        res = spddb_GetARelSet(res, id, relClassId, type, dir,
                               objAttr, objValue, relAttr, relValue);
    }
    
    return res;
}

/* ϢĤɤ */

static SpdRel spddb_GetARel(SpdObjId id, SpdClassId relClassId, int dir,
                            SpdConstString objAttr, int objValue,
                            SpdConstString relAttr, int relValue)
{
    SpdSet2 classes;
    SpdSet2 res;
    SpdClassId classId;
    SpdRel rel;

    res = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    res->valueFree = spdElemFreeSpdRel;
    res->valueCopy = spdElemCopySpdRel;
    if (relClassId == SPD_NON_ID) {
        SetCursor csr;
        classes = spdGetAllClasses(SPD_NON_ID);
        for (csr = spdSetCursorCreate(classes); spdSetCursorNext(&csr);) {
            classId = spdSetGetValueO(csr.curElem->value);
            if (spdGetClassSort(classId) & SAPIDDB_SORT_REL) {
                res = spddb_GetARelSet(res, id, classId, REL, dir,
                                       objAttr, objValue, relAttr, relValue);
                if (res->elemNum) {
                    break;
                }
            }
            
        }
        spdFreeSet2(classes);
    } else {
        res = spddb_GetARelSet(res, id, relClassId, REL, dir,
                               objAttr, objValue, relAttr, relValue);
    }

    if (res->elemNum > 0) {
        SetCursor csr = spdSetCursorCreate(res);
        spdSetCursorNext(&csr);
        rel = *(SpdRel *)spdSetGetValueP(csr.curElem->value);
    } else {
        rel.objectId = SPD_NON_ID;
        rel.relationId = SPD_NON_ID;
    }

    spdFreeSet2(res);

    return rel;
}

/* ֥ȤλҤΥ֥Ȥ */

SpdIdSet spdGetChildObjIdSet(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetRelSet(id, relClassId, OBJID, CHILD,
                           NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdObjId spdGetAChildObjId(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE).objectId;
}

SpdIdSet spdGetChildObjIdSetWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                        SpdConstString objAttr, int objValue)
{
    return spddb_GetRelSet(id, relClassId, OBJID, CHILD,
                           objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdObjId spdGetAChildObjIdWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                      SpdConstString objAttr, int objValue)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         objAttr, objValue, NULL, SAPID_NON_VALUE).objectId;
}

SpdIdSet spdGetChildObjIdSetWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                        SpdConstString relAttr, int relValue)
{
    return spddb_GetRelSet(id, relClassId, OBJID, CHILD,
                           NULL, SAPID_NON_VALUE, relAttr, relValue);
}

SpdObjId spdGetAChildObjIdWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                      SpdConstString relAttr, int relValue)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         NULL, SAPID_NON_VALUE, relAttr, relValue).objectId;
}

/* ֥ȤοƤΥ֥Ȥ */

SpdIdSet spdGetParentObjIdSet(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetRelSet(id, relClassId, OBJID, PARENT,
                           NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdObjId spdGetAParentObjId(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE).objectId;
}

SpdIdSet spdGetParentObjIdSetWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                         SpdConstString objAttr, int objValue)
{
    return spddb_GetRelSet(id, relClassId, OBJID, PARENT,
                           objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdObjId spdGetAParentObjIdWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                       SpdConstString objAttr, int objValue)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         objAttr, objValue, NULL, SAPID_NON_VALUE).objectId;
}

SpdIdSet spdGetParentObjIdSetWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                         SpdConstString relAttr, int relValue)
{
    return spddb_GetRelSet(id, relClassId, OBJID, PARENT,
                           NULL, SAPID_NON_VALUE, relAttr, relValue);
}

SpdObjId spdGetAParentObjIdWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                       SpdConstString relAttr, int relValue)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         NULL, SAPID_NON_VALUE, relAttr, relValue).objectId;
}

/* ֥ȤλҤδϢ */

SpdSet2 spdGetChildRelSet(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetRelSet(id, relClassId, REL, CHILD,
                           NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdRel spdGetAChildRel(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdSet2 spdGetChildRelSetWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                    SpdConstString objAttr, int objValue)
{
    return spddb_GetRelSet(id, relClassId, REL, CHILD,
                           objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdRel spdGetAChildRelWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                  SpdConstString objAttr, int objValue)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdSet2 spdGetChildRelSetWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                    SpdConstString relAttr, int relValue)
{
    return spddb_GetRelSet(id, relClassId, REL, CHILD,
                           NULL, SAPID_NON_VALUE, relAttr, relValue);
}

SpdRel spdGetAChildRelWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                  SpdConstString relAttr, int relValue)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         NULL, SAPID_NON_VALUE, relAttr, relValue);
}

/* ֥ȤοƤδϢ */

SpdSet2 spdGetParentRelSet(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetRelSet(id, relClassId, REL, PARENT,
                           NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdRel spdGetAParentRel(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdSet2 spdGetParentRelSetWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                     SpdConstString objAttr, int objValue)
{
    return spddb_GetRelSet(id, relClassId, REL, PARENT,
                           objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdRel spdGetAParentRelWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                   SpdConstString objAttr, int objValue)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdSet2 spdGetParentRelSetWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                     SpdConstString relAttr, int relValue)
{
    return spddb_GetRelSet(id, relClassId, REL, PARENT,
                           NULL, SAPID_NON_VALUE, relAttr, relValue);
}

SpdRel spdGetAParentRelWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                   SpdConstString relAttr, int relValue)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         NULL, SAPID_NON_VALUE, relAttr, relValue);
}

/* ֥ȤλҤδϢID */

SpdIdSet spdGetChildRelIdSet(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetRelSet(id, relClassId, RELID, CHILD,
                           NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdRelId spdGetAChildRelId(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE).relationId;
}

SpdIdSet spdGetChildRelIdSetWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                        SpdConstString objAttr, int objValue)
{
    return spddb_GetRelSet(id, relClassId, RELID, CHILD,
                           objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdRelId spdGetAChildRelIdWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                      SpdConstString objAttr, int objValue)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         objAttr, objValue, NULL, SAPID_NON_VALUE).relationId;
}

SpdIdSet spdGetChildRelIdSetWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                        SpdConstString relAttr, int relValue)
{
    return spddb_GetRelSet(id, relClassId, RELID, CHILD,
                           NULL, SAPID_NON_VALUE, relAttr, relValue);
}

SpdRelId spdGetAChildRelIdWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                      SpdConstString relAttr, int relValue)
{
    return spddb_GetARel(id, relClassId, CHILD,
                         NULL, SAPID_NON_VALUE, relAttr, relValue).relationId;
}

/* ֥ȤοƤδϢID */

SpdIdSet spdGetParentRelIdSet(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetRelSet(id, relClassId, RELID, PARENT,
                           NULL, SAPID_NON_VALUE, NULL, SAPID_NON_VALUE);
}

SpdRelId spdGetAParentRelId(SpdObjId id, SpdClassId relClassId)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         NULL, SAPID_NON_VALUE, 
                         NULL, SAPID_NON_VALUE).relationId;
}

SpdIdSet spdGetParentRelIdSetWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                         SpdConstString objAttr, int objValue)
{
    return spddb_GetRelSet(id, relClassId, RELID, PARENT,
                           objAttr, objValue, NULL, SAPID_NON_VALUE);
}

SpdRelId spdGetAParentRelIdWithObjAttr(SpdObjId id, SpdClassId relClassId,
                                       SpdConstString objAttr, int objValue)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         objAttr, objValue, NULL, SAPID_NON_VALUE).relationId;
}

SpdIdSet spdGetParentRelIdSetWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                         SpdConstString relAttr, int relValue)
{
    return spddb_GetRelSet(id, relClassId, RELID, PARENT,
                           NULL, SAPID_NON_VALUE, relAttr, relValue);
}

SpdRelId spdGetAParentRelIdWithRelAttr(SpdObjId id, SpdClassId relClassId,
                                       SpdConstString relAttr, int relValue)
{
    return spddb_GetARel(id, relClassId, PARENT,
                         NULL, SAPID_NON_VALUE, relAttr, relValue).relationId;
}

/******************************************************************************/

#ifdef DAPID

/* 
 * DBˤϿʤ֥ȡ
 * ٥ɤ߽񤭤˻Ѥ롥CreateTempObj()顤ɬ
 * DeleteTempObj()뤳ȡspdCreateObj()spdDeleteObj()ȤäƤ⤤
 * ΤId𤵤ʤȤʤΤǡСإåɤˤʤä
 */

SpdDBObj spdCreateTempObj(SpdClassId classId, SpdObjId objId)
{
    SpdDBObj obj;

    obj = spddb_CreateDBObj(classId, objId);
    return obj;
}

void spdDeleteTempObj(SpdDBObj *obj)
{
    if (obj->value) {
        spddb_FreeDBObjValue(obj);
        spdFree(obj->value);
    }
}

int spdSetTempObjAttrIntValByName(SpdDBObj *obj,
                                  SpdConstString attrName,
                                  int value)
{
    if (obj == NULL) {
        return (SAPID_NON_VALUE);
    }
    return (spddb_SetAttrIntVal_in(obj,
                                   spdGetAttrNoByName(obj->classId, attrName),
                                   value));
}

SpdConstString spdSetTempObjAttrStrValByName(SpdDBObj *obj,
                                             SpdConstString attrName,
                                             SpdConstString value)
{
    if (obj == NULL) {
        return (NULL);
    }
    return (spddb_SetAttrStrVal_in(obj,
                                   spdGetAttrNoByName(obj->classId, attrName),
                                   value));
}

SpdConstString spdGetTempObjAttrStrValByName(SpdDBObj *obj,
                                             SpdConstString attrName)
{
    if (obj == NULL) {
        return (NULL);
    }
    return (spddb_GetAttrStrVal_in(obj,
                                   spdGetAttrNoByName(obj->classId, attrName)));
}

int spdGetTempObjAttrIntValByName(SpdDBObj *obj, SpdConstString attrName)
{
    if (obj == NULL) {
        return (SAPID_NON_VALUE);
    }
    return (spddb_GetAttrIntVal_in(obj, spdGetAttrNoByName(obj->classId,
                                                           attrName)));
}

SpdObjId spdReadTempObj(FILE *fp, SpdClassId classId, SpdDBObj *obj)
{
    return (spddb_ReadTuple(fp, classId, obj));
}

#endif /* DAPID */
