/* 
 * Program:             $RCSfile: read_db.c,v $  $Revision: 30.42 $
 * 
 * Purpose:             DBɤ߹ߴؿ
 * 
 * Author:              T.Hidaka  1998/05/21
 *                      S.Yamamoto  1998/05/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: read_db.c,v 30.42 2003/07/08 05:46:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define SPD_DUMP_READ */

/* Debug for AR5's exaustive memory use 2002/10/06. */
/* #define SPD_DUMP_READ_1 */

/******************************************************************************/

#ifdef DAPID

FILE *spdOpenDSDB_AR5(char *dname, int mode)
{
    return spdOpenDSDB(dname, mode);
}

#ifdef COMMENTED

FILE *spdOpenISDB_AR5(char *dname, int mode)
{
    /*
     * !! Ż: spdOpenISDB()SpdObjIdʤΤǡ
     * δؿɬפʤ齤Ʋ(ܡ2000/03/19)
     */
    return spdOpenISDB(dname, mode);
}

#endif /* COMMENTED */

#endif /* DAPID */

/******************************************************************************/

static FILE *openClassDB(SpdClassId classId)
{
    SpdConstString  path;
    SpdConstString  className;
    SpdModelId      modelId;

    modelId = spdGetModelId(classId);
    path = spdGetDBPathOfModel(modelId);
    className = spdGetClassNameById(classId);
    if (path == NULL || className == NULL ||
        !spdCheckSDBClassFile(path, className)) {
        return NULL;
    }

    return (spdOpenSDBfile(path, className, SAPID_O_READ));
}

static void closeClassDB(FILE *fp)
{
    spdFclose(fp);
}

int spdReadAClass(SpdClassId classId, SpdDBId dbId)
{
    SpdConstString  className;
    SpdIdSet        classes;
    FILE            *fp;
    SpdDBObj        *obj_p;
    int             obj_num;

#ifdef SPD_DUMP_READ_1
    static int      totalObjNum = 0;
#endif

    classes = spdGetAllClasses(SPD_NON_ID);
    if (classId == SPD_NON_ID || !spdContainsO(classes, classId)) {
        spdFreeSet2(classes);
        return (SAPID_ERR_INVALID_CLASSID);
    }
    spdFreeSet2(classes);

    className = spdGetClassNameById(classId);
    obj_num = spdGetDbInfoNumOfObjs(classId);

#ifdef SPD_DUMP_READ
    spdMessage("spdReadAClass(): classId %d, className \"%s\", ",
               classId, className);
    spdMessageCont("obj_num %d.\n", obj_num);
#endif

    if ((fp = openClassDB(classId)) == NULL) {
        return (SAPID_ERR_CANT_OPEN_SDB);
    }
    while (0 < obj_num--) {
        obj_p = (SpdDBObj *)spdMalloc(sizeof(SpdDBObj));
        *obj_p = spddb_CreateDBObj(classId, SPD_NON_ID);
        spddb_ReadTuple(fp, classId, obj_p);
        if (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB &&
            fgetc(fp) != '\n') {
            spdFree(obj_p);
            spdFclose(fp);
            return (SAPID_ERR_INVALID_SDB_FORMAT);
        }
        spddb_RegisterDBObj(obj_p, dbId);

        /* Added by S.Yamamoto 2001/11/23. */
        spdSetDbInfoMaxId(classId, obj_p->id);
        spdSetDbInfoTmpMaxId(classId, obj_p->id);

#ifdef SPD_DUMP_READ_1
        spdMessage("    className \"%s\", objNum %d, totalObjNum %d.\n",
                   className, obj_num, ++totalObjNum);
        spdPrintMemProf();
#endif
    }
    if (feof(fp)) {
        spdErrorExit("spdReadAClass(): %s: Too small SDB.\n", className);
    } else if (getc(fp) != EOF) {
        spdErrorExit("spdReadAClass(): %s: Too large SDB.\n", className);
        spdMessage("    feof(): %d.\n", feof(fp));
        spdMessage("    ftell(): %d.\n", ftell(fp));
        spdMessage("    ferror(): %d.\n", ferror(fp));
    }
    closeClassDB(fp);

    return 0;
}

static int spdReadAModel(SpdModelId modelId, SpdDBId dbId)
{
    SpdIdSet    models, classes;
    SpdClassId  classId;
    int         res;
    SetCursor   csr;

    models = spdGetAllModels();
    if (modelId == SPD_NON_ID || !spdContainsO(models, modelId)) {
        spdFreeSet2(models);
        return (SAPID_ERR_INVALID_MODELID);
    }
    spdFreeSet2(models);

    classes = spdGetAllClasses(modelId);

    /* ƤΥ֥Ȥɤ߹ࡥ */
    for (csr = spdSetCursorCreate(classes); spdSetCursorNext(&csr);) {
        classId = spdSetGetValueO(csr.curElem->value);
        if (spdGetClassSort(classId) & SAPIDDB_SORT_OBJ) {
            if ((res = spdReadAClass(classId, dbId)) < 0) {
                spdFreeSet2(classes);
                return res;
            }
        }
        
    }
    /* ƤδϢɤ߹ࡥ */
    for (csr = spdSetCursorCreate(classes); spdSetCursorNext(&csr);) {
        classId = spdSetGetValueO(csr.curElem->value);
        if (spdGetClassSort(classId) & SAPIDDB_SORT_REL) {
            spdReadAClass(classId, dbId);
        }
    }

    spdFreeSet2(classes);
    return 0;
}

int spdReadDB(SpdModelId modelId, SpdDBId dbId)
{
    SpdIdSet    models;
    int         res;
    SetCursor   csr;

    if (modelId == SPD_NON_MID) {
        models = spdGetAllModels();
        for (csr = spdSetCursorCreate(models); spdSetCursorNext(&csr);) {
            modelId = spdSetGetValueO(csr.curElem->value);
            if ((res = spdReadAModel(modelId, dbId)) < 0) {
                spdFreeSet2(models);
                return res;
            }
        }
        spdFreeSet2(models);
    } else {
        if ((res = spdReadAModel(modelId, dbId)) < 0) {
            return res;
        }
    }
    /* spdDumpDbInfo(); */
    return 0;
}

static int spdCreateAModel(SpdModelId modelId, SpdDBId dbId)
{
    /* !! Υեɬפ뤫⤷ʤ */
    return 0;
}

int spdCreateDB(SpdModelId modelId, SpdDBId dbId)
{
    SpdIdSet    models;
    int         res;
    SetCursor   csr;

    if (modelId == SPD_NON_ID) {
        models = spdGetAllModels();
        for (csr = spdSetCursorCreate(models); spdSetCursorNext(&csr);) {
            modelId = spdSetGetValueO(csr.curElem->value);
            if ((res = spdCreateAModel(modelId, dbId)) < 0) {
                spdFreeSet2(models);
                return res;
            }
        }
        spdFreeSet2(models);
    } else {
        if ((res = spdCreateAModel(modelId, dbId)) < 0) {
            return res;
        }
    }
    return 0;
}
