/* 
 * Program:             $RCSfile: mkSchema4.c,v $  $Revision: 40.7 $
 * 
 * Purpose:             Generate class table.
 * 
 * Author:              S.Yamamoto  1995/03/20
 * 
 * (C) Copyright:       S.Yamamoto  1995 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkSchema4.c,v 40.7 2003/06/28 15:31:53 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#ifdef VALIDATION
#include "classTable.c"
#else 
extern Class_t      class_table[];
#endif

#define OBSOLETE_SEQ

#define UNKNOWN_CLASS_ID    "*** Unknown class id. ***"

SpdConstString getTypMacroName(int id);

static SpdModelId getModelId(SpdClassId i);

int main(int argc, char *argv[])
{
    SpdId           i;
    int             j, k;
    SpdConstString  program_name;          /* This program name. */
    SpdConstString  className1, className2;
    SpdModelId      IModelId;
    
#ifndef OBSOLETE_SEQ
    int     ival_start;
#endif

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);

    /* Just dummy for spdGetModelNameById(). */
    IModelId = spdUseModel(&I_model);

#ifdef VALIDATION
    fprintf(stderr, "VALIDATION\n");
#endif  

    printf("Class_t class_table[] = {\n");

    for (i = SPD_FIRST_CID; i <= SPD_LAST_RID; i++) {
        if (i == spdGetFirstCid(spdGetModelId(i))) {
            printf("    /* %s classes. */\n",
                   spdGetModelNameById(getModelId(i)));
        } else if (i == spdGetFirstRid(spdGetModelId(i))) {
            printf("    /* %s relationships. */\n",
                   spdGetModelNameById(getModelId(i)));
        }

        /* "name". */
        printf("    {\"%s\", /* %d */\n", class_table[i].name, i);

        /* "class_id". */
        if (i != class_table[i].class_id) {
            spdErrorExit("Invalid class id %d (should be %d).\n",
                         class_table[i].class_id, i);
        }
        printf("     %s,\n", spdGetClassMacroName(i));

        /* "rel_class". */
        className1 = spdGetClassMacroName(class_table[i].rel_class1);
        className2 = spdGetClassMacroName(class_table[i].rel_class2);
        printf("     %s, %s,\n",
               ((className1 == NULL) ? UNKNOWN_CLASS_ID : className1),
               ((className2 == NULL) ? UNKNOWN_CLASS_ID : className2));

        /* "modelId". */
        printf("     %s,\n", spdGetModelMacroName(getModelId(i)));

        /* "typ". */
        printf("     %s,\n", getTypMacroName(class_table[i].typ));

#ifdef OBSOLETE_SEQ
        /* "seq[]". */
        printf("     {");
        for (j = 0; /* Empty. */; j++) {
            if (j != 0) {
                printf(", ");
            }
            if (class_table[i].seq[j] < IVAL_NUM) {
                printf("%d", class_table[i].seq[j]);
            } else {
                printf("SVAL(%d)", class_table[i].seq[j] - IVAL_NUM);
            }
            if (class_table[i].seq[j] == -1) {
                break;
            }
        }
        printf("},\n");
#else  /* not OBSOLETE_SEQ */
        printf("     {");
        /* Integer. */
        ival_start = SPD_IS_CLS_ID(i) ? 0 : 1;
        for (j = ival_start; j < SPD_MAX(SpdObjIValN, SpdRelIValN); j++) {
            if (class_table[i].iattr[j] == NULL) {
                break;
            }
            if (j != ival_start) {
                printf(", ");
            }
            printf("%d", j);
        }
        /* String. */
        for (k = 0; k < SpdObjSValN; k++) {
            if (class_table[i].sattr[k] == NULL) {
                break;
            }
            if (j != ival_start || k != 0) {
                printf(", ");
            }
            printf("SVAL(%d)", k);
        }            
        if (j != ival_start || k != 0) {
            printf(", ");
        }
        printf("-1");
        printf("},\n");
#endif /* not OBSOLETE_SEQ */

        /* "max_iattr". */
        for (k = SPD_MAX(SpdObjIValN, SpdRelIValN) - 1; 0 <= k; k--) {
            if (class_table[i].iattr[k] != NULL) {
                break;
            }
        }
        printf("     %d,\n", k);

        /* "iattr[SPD_MAX(SpdObjIValN, SpdRelIValN)]". */
        printf("     {");
        for (j = 0; j < SPD_MAX(SpdObjIValN, SpdRelIValN); j++) {
            if (class_table[i].iattr[j] != NULL) {
                if (j != 0) {
                    printf(", ");
                }
                printf("\"%s\"", class_table[i].iattr[j]);
            } else {
                break;
            }
        }
        if (j == 0) {
            printf("NULL");
        }
        printf("},\n");

        /* "max_sattr". */
        for (k = SpdObjSValN - 1; 0 <= k; k--) {
            if (class_table[i].sattr[k] != NULL) {
                break;
            }
        }
        printf("     %d,\n", k);

        /* "sattr[SpdObjSValN]". */
        printf("     {");
        for (j = 0; j < SpdObjSValN; j++) {
            if (class_table[i].sattr[j] != NULL) {
                if (j != 0) {
                    printf(", ");
                }
                printf("\"%s\"", class_table[i].sattr[j]);
            } else {
                break;
            }
        }
        if (j == 0) {
            printf("NULL");
        }
        printf("}\n");
        
        printf("    },\n");
    }

    printf("    /* Terminater. */\n");
    printf("    {NULL,\n");
    printf("     SPD_NON_CID,\n");
    printf("     SPD_NON_CID, SPD_NON_CID,\n");
    printf("     SPD_NON_MID,\n");
    printf("     0,\n");
    printf("     {-1},\n");
    printf("     -1,\n");
    printf("     {NULL},\n");
    printf("     -1,\n");
    printf("     {NULL}}\n");
    printf("};\n");

    exit (EXIT_SUCCESS);
}

/* 
 * λǤclass_table[]򻲾ȤǤʤȤդ뤳ȡ
 */

static SpdModelId getModelId(SpdClassId i)
{
    if (SPD_FIRST_CID_I_MODEL <= i && i <= SPD_LAST_RID_I_MODEL) {
        return (SPD_MID_I_MODEL);
    } else if (SPD_FIRST_CID_P_MODEL <= i && i <= SPD_LAST_RID_P_MODEL) {
        return (SPD_MID_P_MODEL);
    } else if (SPD_FIRST_CID_AREA_MODEL <= i && i <= SPD_LAST_RID_AREA_MODEL) {
        return (SPD_MID_AREA_MODEL);
    } else if (SPD_FIRST_CID_B_MODEL <= i && i <= SPD_LAST_RID_B_MODEL) {
        return (SPD_MID_B_MODEL);
    } else if (SPD_FIRST_CID_D_MODEL <= i && i <= SPD_LAST_RID_D_MODEL) {
        return (SPD_MID_D_MODEL);
    } else if (SPD_FIRST_CID_SDA_MODEL <= i && i <= SPD_LAST_RID_SDA_MODEL) {
        return (SPD_MID_SDA_MODEL);
    } else if (SPD_FIRST_CID_STMT_MODEL <= i && i <= SPD_LAST_RID_STMT_MODEL) {
        return (SPD_MID_STMT_MODEL);
    } else if (SPD_FIRST_CID_SA_MODEL <= i && i <= SPD_LAST_RID_SA_MODEL) {
        return (SPD_MID_SDA_AREA_MODEL);
    } else if (SPD_FIRST_CID_TYPE_VIEW <= i && i <= SPD_LAST_RID_TYPE_VIEW) {
        return (SPD_MID_TYPE_VIEW);
    } else if (SPD_FIRST_CID_COMMENT_VIEW <= i && i <= SPD_LAST_RID_COMMENT_VIEW) {
        return (SPD_MID_COMMENT_VIEW);
    } else if (SPD_FIRST_CID_FLOW_VIEW <= i && i <= SPD_LAST_RID_FLOW_VIEW) {
        return (SPD_MID_FLOW_VIEW);
    } else if (SPD_FIRST_CID_FCD_VIEW <= i && i <= SPD_LAST_RID_FCD_VIEW) {
        return (SPD_MID_FCD_VIEW);
    } else if (SPD_FIRST_CID_J_MODEL <= i && i <= SPD_LAST_RID_J_MODEL) {
        return (SPD_MID_J_MODEL);
    } else {
        return (SPD_NON_MID);
    }
}
