/* 
 * Program:             $RCSfile: mkSchema4From5.c,v $  $Revision: 40.6 $
 * 
 * Purpose:             Convert ver.4 class table to ver.5 one.
 * 
 * Author:              S.Yamamoto  2001/03/21
 * 
 * (C) Copyright:       S.Yamamoto  2001 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkSchema4From5.c,v 40.6 2003/06/28 15:31:54 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdConstString   program_name;              /* This program name. */

static char         *modelName = NULL;
static SpdBoolean   flag_opt_i = OFF;
static SpdBoolean   use_AR4_TYP = OFF;

static void parse_option(int argc, char *argv[]);
static void print_usage(SpdConstString command_name);
static void dump_model_class(SpdModelId modelId);
static void dump_model_rel(SpdModelId modelId);
static int cnvSortToTyp(int sort);

SpdConstString getTypMacroName(int id);
SpdConstString getSortMacroName(int id);

int spdGetTyp(SpdClassId classId);

int main(int argc, char *argv[])
{
    SpdId       i;
    SpdModelId  IModelId;
    SpdModelId  PModelId;
    SpdModelId  AreaModelId;
    SpdModelId  BModelId;
    SpdModelId  DModelId;
    SpdModelId  SDAModelId;
    SpdModelId  StmtModelId;
    SpdModelId  SareaModelId;
    SpdModelId  TypeViewId;
    SpdModelId  CommentViewId;
    SpdModelId  FlowViewId;
    SpdModelId  FCD_view_id;
    SpdModelId  JModelId;

    /* Just for debug. */
    setbuf(stdout, NULL); 

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parse_option(argc, argv);
    argc -= optind;
    argv += optind;

    IModelId = spdUseModel(&I_model);
    PModelId = spdUseModel(&P_model);
    AreaModelId = spdUseModel(&Area_model);
    BModelId = spdUseModel(&B_model);
    DModelId = spdUseModel(&D_model);
    SDAModelId = spdUseModel(&SDA_model);
    StmtModelId = spdUseModel(&Stmt_model);
    SareaModelId = spdUseModel(&SDAarea_model);
    TypeViewId = spdUseModel(&TypeView_model);
    CommentViewId = spdUseModel(&Comment_view);
    FlowViewId = spdUseModel(&Flow_view);
    FCD_view_id = spdUseModel(&FCD_view);
    JModelId = spdUseModel(&J_model);

    printf("Class_t class_table[] = {\n");
    if (modelName == NULL) {
        for (i = SPD_FIRST_MID; i <= SPD_LAST_MID; i++) {
            dump_model_class(i);
            dump_model_rel(i);
        }
    } else {
        for (i = SPD_FIRST_MID; i <= SPD_LAST_MID; i++) {
            if (SPD_STREQ(modelName, spdGetModelNameById(i))) {
                dump_model_class(i);
                dump_model_rel(i);
                break;
            }
        }
    }
    printf("    /* Terminater. */\n");
    printf("    {NULL,\n");
    printf("     SPD_NON_CID,\n");
    printf("     SPD_NON_CID, SPD_NON_CID,\n");
    printf("     SPD_NON_MID,\n");
    printf("     0,\n");
    printf("     {-1},\n");
    printf("     -1,\n");
    printf("     {NULL},\n");
    printf("     -1,\n");
    printf("     {NULL}}\n");
    printf("};\n");

    return (EXIT_SUCCESS);
}

static void dump_model_class(SpdModelId modelId)
{
    SpdObjId    i;

    printf("    /* %s classes. */\n", spdGetModelNameById(modelId));
    for (i = spdGetFirstCid(modelId); i <= spdGetLastCid(modelId); i++) {
        int         numOfIntAttr = 0, numOfStrAttr = 0;

        if (spdGetClassNameById(i) == NULL) {
            spdErrorExit("dump_model_class(): %d: Invalid class id.\n", i);
        }
        if (flag_opt_i) {
            printf("    {\"%s\", /* %d */\n", spdGetClassNameById(i), i);
        } else {
            printf("    {\"%s\",\n", spdGetClassNameById(i));
        }
        printf("     %s,\n", spdGetClassMacroName(i));
        printf("     SPD_NON_CID, SPD_NON_CID,\n");
        printf("     %s,\n", spdGetModelMacroName(spdGetModelId(i)));
        printf("     %s,\n",
               use_AR4_TYP && 
               (modelId == SPD_MID_I_MODEL || modelId == SPD_MID_P_MODEL) ?
               getTypMacroName(spdGetTyp(i)) :
               getTypMacroName(cnvSortToTyp(spdGetDBClassById(i)->sort)));
        {
            SpdDBClass  *dbClass;
            int         j;
            int         is_first = TRUE;

            dbClass = spdGetDBClassById(i);
            printf("     {");
            for (j = 0; j < SpdDBMaxFieldN; j++) {
                int         *aFields = dbClass->fields;
                SpdDBAttr   *aAttrs = dbClass->attrs;

                if (is_first) {
                    is_first = FALSE;
                } else {
                    printf(", ");
                }
                if (aFields[j] == SAPIDDB_ATTR_ID) {
                    printf("0");
                    numOfIntAttr++;
                } else if (aFields[j] == SAPIDDB_ATTR_NONE) {
                    printf("-1");
                    break;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_INT ||
                           aAttrs[aFields[j]].type == SPD_DBT_ENUM) {
                    printf("%d", aFields[j] + 1);
                    numOfIntAttr++;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_STR) {
                    printf("SVAL(%d)", numOfStrAttr);
                    numOfStrAttr++;
                }
            }
            printf("},\n");
        }

        /* "max_iattr". */
        printf( "     %d,\n", numOfIntAttr - 1);

        /* "iattr[SPD_MAX(SpdObjIValN, SpdRelIValN)]". */
        {
            SpdDBClass  *dbClass;
            int         j;

            dbClass = spdGetDBClassById(i);
            printf("     {");
            for (j = 0; j < SpdDBMaxFieldN; j++) {
                int         *aFields = dbClass->fields;
                SpdDBAttr   *aAttrs = dbClass->attrs;

                if (aFields[j] == SAPIDDB_ATTR_ID) {
                    printf("\"id\"");
                } else if (aFields[j] == SAPIDDB_ATTR_NONE) {
                    break;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_INT ||
                           aAttrs[aFields[j]].type == SPD_DBT_ENUM) {
                    if (j != 0) {
                        printf(", ");
                    }
                    printf("\"%s\"", aAttrs[aFields[j]].name);
                } else if (aAttrs[aFields[j]].type == SPD_DBT_STR) {
                    break;
                }
                
            }
            printf("},\n");
        }

        /* "max_sattr". */
        printf( "     %d,\n", numOfStrAttr - 1);

        /* "sattr[SpdObjSValN]". */
        {
            SpdDBClass  *dbClass;
            int         j;
            int         is_first = TRUE;

            dbClass = spdGetDBClassById(i);
            printf("     {");
            for (j = 0; j < SpdDBMaxFieldN; j++) {
                int         *aFields = dbClass->fields;
                SpdDBAttr   *aAttrs = dbClass->attrs;

                if (aFields[j] == SAPIDDB_ATTR_ID) {
                    continue;
                } else if (aFields[j] == SAPIDDB_ATTR_NONE) {
                    break;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_STR) {
                    if (is_first) {
                        is_first = FALSE;
                    } else {
                        printf(", ");
                    }
                    printf("\"%s\"", aAttrs[aFields[j]].name);
                }
                
            }
            if (is_first) {
                printf("NULL");
            }
            printf("}\n");
        }

        printf("    },\n");
    }
}

static void dump_model_rel(SpdModelId modelId)
{
    SpdRelId    i;

    printf("    /* %s relationships. */\n", spdGetModelNameById(modelId));
    for (i = spdGetFirstRid(modelId); i <= spdGetLastRid(modelId); i++) {
        int         numOfIntAttr = 0, numOfStrAttr = 0;
        SpdDBClass  *dbClass = spdGetDBClassById(i);

        if (spdGetClassNameById(i) == NULL) {
            spdErrorExit("dump_model_rel(): %d: Invalid rel id.\n", i);
        }
        if (flag_opt_i) {
            printf("    {\"%s\", /* %d */\n", spdGetClassNameById(i), i);
        } else {
            printf("    {\"%s\",\n", spdGetClassNameById(i));   
        }
        printf("     %s,\n", spdGetClassMacroName(i));
        printf("     %s, %s,\n",
               spdGetClassMacroName(dbClass->attrs[0].classId),
               spdGetClassMacroName(dbClass->attrs[1].classId));
        printf("     %s,\n", spdGetModelMacroName(spdGetModelId(i)));
        printf("     %s,\n",
               use_AR4_TYP && 
               (modelId == SPD_MID_I_MODEL || modelId == SPD_MID_P_MODEL) ?
               getTypMacroName(spdGetTyp(i)) :
               getTypMacroName(cnvSortToTyp(spdGetDBClassById(i)->sort)));
        {

            int         j;
            int         is_first = TRUE;

            printf("     {");
            for (j = 0; j < SpdDBMaxFieldN; j++) {
                int         *aFields = dbClass->fields;
                SpdDBAttr   *aAttrs = dbClass->attrs;

                if (is_first) {
                    is_first = FALSE;
                } else {
                    printf(", ");
                }
                if (aFields[j] == 0 || aFields[j] == 1) {
                    printf("%d", aFields[j] + 1);
                    numOfIntAttr++;
                    continue;
                } else if (aFields[j] == SAPIDDB_ATTR_NONE) {
                    printf("-1");
                    break;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_INT ||
                           aAttrs[aFields[j]].type == SPD_DBT_ENUM) {
                    printf("%d", aFields[j] + 1);
                    numOfIntAttr++;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_STR) {
                    printf("SVAL(%d)", numOfStrAttr);
                    numOfStrAttr++;
                }
            }
            printf("},\n");
        }

        /* "max_iattr". */
        printf( "     %d,\n", numOfIntAttr);

        /* "iattr[SPD_MAX(SpdObjIValN, SpdRelIValN)]". */
        {
            SpdDBClass  *dbClass;
            int         j;

            dbClass = spdGetDBClassById(i);
            printf("     {\"id\"");
            for (j = 0; j < SpdDBMaxFieldN; j++) {
                int         *aFields = dbClass->fields;
                SpdDBAttr   *aAttrs = dbClass->attrs;

                if (aFields[j] == SAPIDDB_ATTR_NONE) {
                    break;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_INT ||
                           aAttrs[aFields[j]].type == SPD_DBT_ENUM || 
                           aAttrs[aFields[j]].type == SPD_DBT_OID) {
                    printf(", \"%s\"", aAttrs[aFields[j]].name);
                } else if (aAttrs[aFields[j]].type == SPD_DBT_STR) {
                    break;
                }
                
            }
            printf("},\n");
        }

        /* "max_sattr". */
        printf( "     %d,\n", numOfStrAttr - 1);

        /* "sattr[SpdObjSValN]". */
        {
            SpdDBClass  *dbClass;
            int         j;
            int         is_first = TRUE;

            dbClass = spdGetDBClassById(i);
            printf("     {");
            for (j = 0; j < SpdDBMaxFieldN; j++) {
                int         *aFields = dbClass->fields;
                SpdDBAttr   *aAttrs = dbClass->attrs;

                if (aFields[j] == SAPIDDB_ATTR_ID) {
                    continue;
                } else if (aFields[j] == SAPIDDB_ATTR_NONE) {
                    break;
                } else if (aAttrs[aFields[j]].type == SPD_DBT_STR) {
                    if (is_first) {
                        is_first = FALSE;
                    } else {
                        printf(", ");
                    }
                    printf("\"%s\"", aAttrs[aFields[j]].name);
                }
                
            }
            if (is_first) {
                printf("NULL");
            }
            printf("}\n");
        }

        printf("    },\n");
    }
}

static void parse_option(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:im:t")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'i':
            flag_opt_i = ON;
            break;
          case 'm':                         /* Model name. */
            modelName = optarg;
            break;
          case 't':                         /* Use AR4's TYP for {IP}-model. */
            use_AR4_TYP = ON;
            break;
          default:
            spdError("'%c': Invalid option.\n", ch);
            print_usage(program_name);
            break;
        }
    }
}

static int cnvSortToTyp(int sort)
{
    if (sort & SAPIDDB_SORT_OBJ) {
        if (sort & SAPIDDB_SORT_OBJ_TEXT) {
            /* ƥɽ(offsetlength) */
            if (sort & SAPIDDB_SORT_OBJ_TREENODE) {
                /* 롼ȤϢéäȤ1̤ˤéʤ */
                return (C_TYP_OC);
            } else { 
                return (C_TYP_O_);
            }
        } else {
            /* ƥɽʤ(ǰμ) */
            if (sort & SAPIDDB_SORT_OBJ_TREENODE) {
                /* 롼ȤϢéäȤ1̤ˤéʤ */
                return (C_TYP_OC);
            } else { 
                return (C_TYP_O_);
            }
        }
    } else if (sort & SAPIDDB_SORT_REL) {
        if (sort & SAPIDDB_SORT_REL_COMPOSITE) {
            /* طǤ롥Ƥ1ĤǻҤʣ(ڹ¤)*/
            return (C_TYP_RC);
        } else if (sort & SAPIDDB_SORT_REL_REFERENCE) {
            /* ȴطǤ롥ƤʣФƻҤ1ġ */
            return (C_TYP_R_);
        } else if (sort & SAPIDDB_SORT_REL_ONE_TO_ONE) {
            /* бطǤ롥ƤȻҤ11Ǥ롥 */
            return (C_TYP_R_);
        } else {
            spdErrorExit("cnvSortToTyp(): %d, %s: Invalid sort.\n",
                         sort, getSortMacroName(sort));
            SPD_DUMMY_RETURN(C_TYP_OC);
        }
    } else {
        spdErrorExit("cnvSortToTyp(): %d, %s: Invalid sort.\n",
                     sort, getSortMacroName(sort));
        SPD_DUMMY_RETURN(C_TYP_OC);
    }
}

/*
 * Printing usage.
 */

static void print_usage(SpdConstString command_name)
{
    const char  *usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [options] [className...]\n"
        "DESCRIPTION\n"
        "    -d SDB-directory: Directory for SDB (\"SDB\").\n"
        "    -i: Print class and relationship id.\n"
        "    -m modelName: Specify model name.\n"
        "    -t: Use AR4's TYP.\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
