/* 
 * Program:             $RCSfile: mkWriteSDB.c,v $  $Revision: 30.23 $
 * 
 * Purpose:             Source code generator for output routines.
 * 
 * Author:              R.Miura  1997/10/24
 *                      S.Yamamoto  1997/10/24
 * 
 * (C) Copyright:       R.Miura and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkWriteSDB.c,v 30.23 2003/08/17 07:48:09 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#define CFILE               "write_sdb_sub.c"
#define HFILE               "write_sdb_sub.h"

#define UNKNOWN_CLASS_ID    "*** Unknown class id ***"

#define USAGE_FORMAT        "Usage: %s\n"

static const char *headerComment = 
    "/* \n"
    " * Program:             $RCSfile: mkWriteSDB.c,v $  $Revision: 30.23 $\n"
    " * \n"
    " * (C) Copyright:       R.Miura and S.Yamamoto  1997 - 1998\n"
    " *                      This file is a product of the project Sapid.\n"
    " * \n"
    " *                      Generated by: %s in Sapid-%d.%d.%d.\n"
    " */\n\n";

static const char *rcsidDefine =
    "#ifndef lint\n"
    "static char rcsid[] =\n"
    "    \"$Id: mkWriteSDB.c,v 30.23 2003/08/17 07:48:09 yamamoto Exp yamamoto $\";\n"
    "#endif\n\n";

static const char *includeStatement =
    "#include <stdio.h>\n"
    "#include <Sapid/Sapid.h>\n"
    "#include <Sapid/write_sdb_sub.h>\n\n";

/******************************************************************************/
 
/* 
 * FuncA롼
 * 
 * ƥ饹ӴϢФơ줾ؿwrite_xxxx_sub()롥
 * δؿ°ͤȤľܼ롥
 * 
 * :
 *
 * void write_program_sub(FILE *fp, int id, SpdString name)
 * {
 *         switch (spdGetWriteSDBformat()) {
 *             case SAPID_TEXT_FORMAT_SDB:
 *                 fprintf(fp, "%d%c%s%c\n",
 *                         id, SDB_ATTR_SEPARATOR,
 *                         name, SDB_ATTR_SEPARATOR);
 *                 break;
 *             case SAPID_BINARY_FORMAT_SDB:
 *                 spdWriteIntAttrBinary(fp, id);
 *                 fputs(name, fp);
 *                 fputc('\0', fp);
 *                 break;
 *             default:
 *                 spdErrorExit("Invalid write format.\n");
 *                 break;
 *         }
 * }
 */

/* 
 * ؿƬʬ(prototypeˤ)
 * void write_program_sub(FILE *fp, int id, SpdString name)
 */

static void generateFuncATop(FILE *fp, SpdClassId classId)
{
    int i, argn;

    fprintf(fp, "void write_%s_sub", class_table[classId].name);

    /* Parameter. */
    fprintf(fp, "(FILE *fp");
    for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
        if (SPD_IS_SVAL(argn)) {            /* String. */
            fprintf(fp, ", SpdString %s",
                    (class_table[classId].sattr)[argn - IVAL_NUM]);
        } else {                            /* Integer. */
            fprintf(fp, ", int %s",
                    (class_table[classId].iattr)[argn]);
        }
    }
    fprintf(fp, ")");
}

/*
 * ؿΤ롥
 */

static void generateFuncABody(FILE *fp, SpdClassId classId)
{
    int i, argn;

    fprintf(fp, "\n{\n");
    fprintf(fp, "\tswitch (spdGetWriteSDBformat()) {\n");

    /* 
     * Text format.
     */
    fprintf(fp, "\t  case SAPID_TEXT_FORMAT_SDB:\n");
    fprintf(fp, "\t\tfprintf(fp, ");

    fprintf(fp, "\"");
    for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
        if (SPD_IS_SVAL(argn)) {            /* String. */
            fprintf(fp, "%%s%%c");
        } else {                            /* Integer. */
            fprintf(fp, "%%d%%c");
        }
    }
    fprintf(fp, "\\n\"");

    for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
        if (SPD_IS_SVAL(argn)) {            /* String. */
            fprintf(fp, ",\n\t\t\t%s, SDB_ATTR_SEPARATOR",
                    (class_table[classId].sattr)[argn - IVAL_NUM]);
        } else {                            /* Integer. */
            fprintf(fp, ",\n\t\t\t%s, SDB_ATTR_SEPARATOR",
                    (class_table[classId].iattr)[argn]);
        }
    }
    fprintf(fp, ");\n");
    fprintf(fp, "\t\tbreak;\n");

    /* 
     * Binary format.
     */
    fprintf(fp, "\t  case SAPID_BINARY_FORMAT_SDB:\n");

    for (i = 0; (argn = class_table[classId].seq[i]) >= 0; i++) {
        if (SPD_IS_SVAL(argn)) {            /* String. */
            fprintf(fp, "\t\tfputs(%s, fp);\n",
                    (class_table[classId].sattr)[argn - IVAL_NUM]);
            fprintf(fp, "\t\tfputc('\\0', fp);\n");
        } else {                            /* Integer. */
            fprintf(fp, "\t\tspdWriteIntAttrBinary(fp, %s);\n",
                    (class_table[classId].iattr)[argn]);
        }
    }
    fprintf(fp, "\t\tbreak;\n");
    
    /* 
     * Invalid format.
     */
    fprintf(fp, "\t  default:\n");
    fprintf(fp, "\t\tspdErrorExit(\"Invalid write format.\\n\");\n");
    fprintf(fp, "\t\tbreak;\n");
    fprintf(fp, "\t}\n");
    fprintf(fp, "}\n\n");
}

/******************************************************************************/

/* 
 * FuncB롼
 * 
 * 饹ν񤭹ߴؿspdWriteDBObj(), Ϣν񤭹ߴؿspdWriteDBRel()
 * 롥FuncAȤϰۤʤꡤSpdDB{Obj,Rel}ι¤ΤȤ롥
 *
 * :
 * 
 * void spdWriteDBObj(FILE *fp, SpdDBObj *obj)
 * {
 *         switch (spdGetWriteSDBformat()) {
 *             case SAPID_TEXT_FORMAT_SDB:
 *                 switch (spdGetClassId((SpdObjId)(obj->Obj_id))) {
 *                     case SPD_CID_PROGRAM:
 *                         fprintf(fp, "%d%c%s%c\n",
 *                                 obj->ivalue[0], SDB_ATTR_SEPARATOR,
 *                                 obj->svalue[0], SDB_ATTR_SEPARATOR);
 *                         break;
 *                     case SPD_CID_FILE:
 *                         ...
 */

/* 
 * إå롥
 */

static void generateFuncBHeader(FILE *fp)
{
    fprintf(fp, "\n");
    fprintf(fp, "void spdWriteDBObj(FILE *fp, SpdDBObj *obj);\n");
    fprintf(fp, "void spdWriteDBRel(FILE *fp, SpdDBRel *rel);\n");
    fprintf(fp, "\n");
}

/* 
 * spdWriteDBObj()롥
 */

static void generateFuncBObj(FILE *fp)
{
    int             i, argn;
    SpdId           id;
    SpdConstString  macro;

    fprintf(fp, "void spdWriteDBObj(FILE *fp, SpdDBObj *obj)\n{\n");
    fprintf(fp, "\tswitch (spdGetWriteSDBformat()) {\n");

    /* 
     * Text format.
     */
    fprintf(fp, "\t  case SAPID_TEXT_FORMAT_SDB:\n");
    fprintf(fp, "\t\tswitch (spdGetClassId((SpdObjId)(obj->Obj_id))) {\n");

    for (id = SPD_FIRST_CID_IP_MODEL; id <= SPD_LAST_RID_IP_MODEL; id++) {
        if (!SPD_IS_CLS_ID(id)) {
            continue;
        }

        macro = spdGetClassMacroName(id);
        fprintf(fp, "\t\t  case %s:\n", (macro ? macro : UNKNOWN_CLASS_ID));
        fprintf(fp, "\t\t\tfprintf(fp, ");

        /* եޥåȤ롥 */
        fprintf(fp, "\"");
        for (i = 0; (argn = class_table[id].seq[i]) >= 0; i++) {
            if (SPD_IS_SVAL(argn)) {        /* String. */
                fprintf(fp, "%%s%%c");
            } else {                        /* Integer. */
                fprintf(fp, "%%d%%c");
            }
        }
        fprintf(fp, "\\n\"");

        /* 롥 */
        for (i = 0; (argn = class_table[id].seq[i]) >= 0; i++) {
            if (SPD_IS_SVAL(argn)) {        /* String. */
                fprintf(fp, ",\n\t\t\t\tobj->svalue[%d], SDB_ATTR_SEPARATOR",
                        argn - IVAL_NUM);
            } else {                        /* Integer. */
                fprintf(fp, ",\n\t\t\t\tobj->ivalue[%d], SDB_ATTR_SEPARATOR",
                        argn);
            }
        }
        fprintf(fp, ");\n");

        fprintf(fp, "\t\t\tbreak;\n");
    }

    /* Default. */
    fprintf(fp, "\t\t  default:\n");
    fprintf(fp, "\t\t\tspdErrorExit(\"Invalid class Id.\\n\");\n");
    fprintf(fp, "\t\t\tbreak;\n");
    fprintf(fp, "\t\t}\n");
    fprintf(fp, "\t\tbreak;\n");

    /* 
     * Binary format.
     */
    fprintf(fp, "\t  case SAPID_BINARY_FORMAT_SDB:\n");
    fprintf(fp, "\t\tswitch (spdGetClassId((SpdObjId)(obj->Obj_id))) {\n");

    for (id = SPD_FIRST_CID_IP_MODEL; id <= SPD_LAST_RID_IP_MODEL; id++) {
        if (!SPD_IS_CLS_ID(id)) {
            continue;
        }

        macro = spdGetClassMacroName(id);
        fprintf(fp, "\t\t  case %s:\n", (macro ? macro : UNKNOWN_CLASS_ID));

        for (i = 0; (argn = class_table[id].seq[i]) >= 0; i++) {
            if (SPD_IS_SVAL(argn)) {        /* String. */
                fprintf(fp, "\t\t\tfputs(obj->svalue[%d], fp);\n",
                        argn - IVAL_NUM);
                fprintf(fp, "\t\t\tfputc('\\0', fp);\n");
            } else {                        /* Integer. */
                fprintf(fp,
                        "\t\t\tspdWriteIntAttrBinary(fp, obj->ivalue[%d]);\n",
                        argn);
            }
        }

        fprintf(fp, "\t\t\tbreak;\n");
    }

    /* Default. */
    fprintf(fp, "\t\t  default:\n");
    fprintf(fp, "\t\t\tspdErrorExit(\"Invalid class Id.\\n\");\n");
    fprintf(fp, "\t\t\tbreak;\n");
    fprintf(fp, "\t\t}\n");
    fprintf(fp, "\t\tbreak;\n");

    /* 
     * Invalid format.
     */
    fprintf(fp, "\t  default:\n");
    fprintf(fp, "\t\tspdErrorExit(\"Invalid write format.\\n\");\n");
    fprintf(fp, "\t\tbreak;\n");
    fprintf(fp, "\t}\n");

    fprintf(fp, "}\n\n");
}

/* 
 * spdWriteDBRel()롥
 */

static void generateFuncBRel(FILE *fp)
{
    int             i, argn;
    SpdId           id;
    SpdConstString  macro;

    fprintf(fp, "void spdWriteDBRel(FILE *fp, SpdDBRel *rel)\n{\n");
    fprintf(fp, "\tswitch (spdGetWriteSDBformat()) {\n");

    /* 
     * Text format.
     */
    fprintf(fp, "\t  case SAPID_TEXT_FORMAT_SDB:\n");
    fprintf(fp, "\t\tswitch (spdGetClassId((SpdRelId)(rel->Rel_id))) {\n");

    for (id = SPD_FIRST_CID_IP_MODEL; id <= SPD_LAST_RID_IP_MODEL; id++) {
        if (!SPD_IS_REL_ID(id)) {
            continue;
        }

        macro = spdGetClassMacroName(id);
        fprintf(fp, "\t\t  case %s:\n", (macro ? macro : UNKNOWN_CLASS_ID));
        fprintf(fp, "\t\t\tfprintf(fp, ");

        /* եޥåȤ롥 */
        fprintf(fp, "\"");
        for (i = 0; class_table[id].seq[i] >= 0; i++) {
            fprintf(fp, "%%d%%c");
        }
        fprintf(fp, "\\n\"");

        /* 롥 */
        for (i = 0; (argn = class_table[id].seq[i]) >= 0; i++) {
            fprintf(fp, ",\n\t\t\t\trel->ivalue[%d], SDB_ATTR_SEPARATOR", 
                    argn);
        }
        fprintf(fp, ");\n");

        fprintf(fp, "\t\t\tbreak;\n");
    }

    /* Default. */
    fprintf(fp, "\t\t  default:\n");
    fprintf(fp, "\t\t\tspdErrorExit(\"Invalid class Id.\\n\");\n");
    fprintf(fp, "\t\t\tbreak;\n");
    fprintf(fp, "\t\t}\n");
    fprintf(fp, "\t\tbreak;\n");

    /* 
     * Binary format.
     */
    fprintf(fp, "\t  case SAPID_BINARY_FORMAT_SDB:\n");
    fprintf(fp, "\t\tswitch (spdGetClassId((SpdRelId)(rel->Rel_id))) {\n");

    for (id = SPD_FIRST_CID_IP_MODEL; id <= SPD_LAST_RID_IP_MODEL; id++) {
        if (!SPD_IS_REL_ID(id)) {
            continue;
        }

        macro = spdGetClassMacroName(id);
        fprintf(fp, "\t\t  case %s:\n", (macro ? macro : UNKNOWN_CLASS_ID));

        for (i = 0; (argn = class_table[id].seq[i]) >= 0; i++) {
            fprintf(fp, "\t\t\tspdWriteIntAttrBinary(fp, rel->ivalue[%d]);\n",
                    argn);
        }

        fprintf(fp, "\t\t\tbreak;\n");
    }

    /* Default. */
    fprintf(fp, "\t\t  default:\n");
    fprintf(fp, "\t\t\tspdErrorExit(\"Invalid class Id.\\n\");\n");
    fprintf(fp, "\t\t\tbreak;\n");
    fprintf(fp, "\t\t}\n");
    fprintf(fp, "\t\tbreak;\n");

    /* 
     * Invalid format.
     */
    fprintf(fp, "\t  default:\n");
    fprintf(fp, "\t\tspdErrorExit(\"Invalid write format.\\n\");\n");
    fprintf(fp, "\t\tbreak;\n");
    fprintf(fp, "\t}\n");

    fprintf(fp, "}\n\n");
}

/******************************************************************************/

/* 
 * Cե롥
 */

static void generateCFile(FILE *fp, SpdConstString program)
{
    SpdClassId  classId;

    fprintf(fp, headerComment,
            program, SapidVersion, SapidRevision, SapidPatchLevel);
    fprintf(fp, rcsidDefine);
    fprintf(fp, includeStatement);

    /* FuncA. */
    fprintf(fp, "/* FuncA. */\n\n");
    for (classId = SPD_FIRST_CID_IP_MODEL; 
     classId <= SPD_LAST_RID_IP_MODEL; classId ++) {
        generateFuncATop(fp, classId);
        generateFuncABody(fp, classId);
    }

    /* FuncB. */
    fprintf(fp, "/* FuncB. */\n\n");
    generateFuncBObj(fp);
    generateFuncBRel(fp);
}

/* 
 * إåե롥
 */

static void generateHFile(FILE *fp, SpdConstString program)
{
    SpdClassId  classId;

    fprintf(fp, headerComment,
            program, SapidVersion, SapidRevision, SapidPatchLevel);

    /* FuncA. */
    for (classId = SPD_FIRST_CID_IP_MODEL;
         classId <= SPD_LAST_RID_IP_MODEL; classId ++) {
        generateFuncATop(fp, classId);
        fprintf(fp, ";\n");
    }

    /* FuncB. */
    generateFuncBHeader(fp);
}    

/******************************************************************************/

int main(int argc, char *argv[])
{
    SpdConstString  program_name;               /* This program name. */
    FILE            *fp;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);

    if (argc != 1) {
        spdErrorExit(USAGE_FORMAT, program_name);
    }

    /* Generate C program file. */
    fp = spdFopen(CFILE, SPD_FOPEN_WT_MODE);
    generateCFile(fp, program_name);
    spdFclose(fp);

    /* Generate header file. */
    fp = spdFopen(HFILE, SPD_FOPEN_WT_MODE);
    generateHFile(fp, program_name);
    spdFclose(fp);

    return (EXIT_SUCCESS);
}
