/* 
 * Program:             $RCSfile: j2jx.java,v $  $Revision: 41.73 $
 * 
 * Purpose:             Generate elements list for JX-model3.
 *
 * Author:              H.Yoshida  2002/12/12
 *                      S.Yamamoto  2002/08/04
 *
 * (C) Copyright:       H.Yoshida and S.Yamamoto  2002 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: j2jx.java,v 41.73 2003/07/08 05:48:05 yamamoto Exp yamamoto $
 */

package JXModel3.j2jx;

import java.io.*;
import java.util.*;
import gnu.getopt.*;

import Japid.Freja.*;
import Japid.util.Stopwatch;

public class j2jx {
    private String      targetClassName;
    private String      outpath;
    private boolean     isIntegerObjId = false;
    
    // Output stream.
    private PrintStream _outetb;

    // Source program buffer.
    private byte[]      _src;
    
    // Range class.
    class Range {
        public int  sp, ep;
    
        public Range(int s, int e) {
            sp = s;
            ep = e;
        }
        public int len() {
            return ep - sp;
        }
    }

    // Counter for derived id. Incremented in makeIDAttrDerived().
    private static long derivedId = 0;

    // Main.
    public static void main(String args[]) {
        Stopwatch timer = new Stopwatch().start();
        
        if (args.length < 2) {
            FrjWorld.errorExit("java JXModel3.j2jx.j2jx <targetClassName> <elementTable>");
        }

        j2jx      ex = new j2jx();
        if (!ex.parse_options(args, args.length)) {
            FrjWorld.error("Invalid option or argument.");
        }
        FrejaEntry entry = new FrejaEntry(FrjWorld.J_MODEL_PATH,
                                          ex.targetClassName);
        if (!entry.exists()) {
            FrjWorld.errorExit("\"" + ex.targetClassName + "\": No such SDB.");
        }

        try {
            ex.init(ex.outpath, entry);

            ex.extractFile(FrjExtents.getObjects(FrjFile.getClassNameStatically()));
            
            for (Enumeration e = FrjExtents.getObjects(FrjClass.getClassNameStatically());
                 e.hasMoreElements();) ex.extractClass(e);
            for (Enumeration e = FrjExtents.getObjects(FrjInterface.getClassNameStatically());
                 e.hasMoreElements();) ex.extractInterface(e);
            for (Enumeration e = FrjExtents.getObjects(FrjSBlock.getClassNameStatically());
                 e.hasMoreElements();) ex.extractSBlock(e);
            for (Enumeration e = FrjExtents.getObjects(FrjRoutine.getClassNameStatically());
                 e.hasMoreElements();) ex.extractRoutine(e);
            for (Enumeration e = FrjExtents.getObjects(FrjVariable.getClassNameStatically());
                 e.hasMoreElements();) ex.extractVariable(e);
            for (Enumeration e = FrjExtents.getObjects(FrjStatement.getClassNameStatically());
                 e.hasMoreElements();) ex.extractStatement(e);
            for (Enumeration e = FrjExtents.getObjects(FrjLabel.getClassNameStatically());
                 e.hasMoreElements();) ex.extractLabel(e);
            for (Enumeration e = FrjExtents.getObjects(FrjType.getClassNameStatically());
                 e.hasMoreElements();) ex.extractType(e);
            for (Enumeration e = FrjExtents.getObjects(FrjExpression.getClassNameStatically());
                 e.hasMoreElements();) ex.extractExpression(e);
            for (Enumeration e = FrjExtents.getObjects(FrjQName.getClassNameStatically());
                 e.hasMoreElements();) ex.extractQName(e);
            for (Enumeration e = FrjExtents.getObjects(FrjLiteral.getClassNameStatically());
                 e.hasMoreElements();) ex.extractLiteral(e);

            for (Enumeration e = FrjExtents.getObjects(FrjComment.getClassNameStatically());
                 e.hasMoreElements();) ex.extractComment(e);
        } catch (FileNotFoundException e) {
            return;
        }
        timer.printTimeProf("j2jx");
        return;
    }

    // Parse options.
    private boolean parse_options(String args[], int argc) {
        int     ch;
        String  check;
        Getopt  g = new Getopt("j2jx", args, "d:iv");

        while ((ch = g.getopt()) != -1) {
            switch (ch) {
              case 'd':
                FrjWorld.spdSetSDBDir(g.getOptarg());
                break;
              case 'i':
                isIntegerObjId = true;
                break;
              default:
                System.err.println("'" + ch + "': Invalid option.");
                // print_usage(program_name);
                break;
            }
        }
        argc -= g.getOptind();
        if (argc != 2) {
            FrjWorld.errorExit("No arguments.");
        }
        targetClassName = new String(args[g.getOptind() + 0]);
        outpath = new String(args[g.getOptind() + 1]);
        return true;
    }

    // Initalization.
    private void init(String outpath, FrejaEntry entry)
        throws FileNotFoundException {
        _outetb = new PrintStream(new FileOutputStream(outpath));

        FrjWorld.load(entry);
        // String srcstr = entry.readText(0, entry.getFileSize());
        // _src = srcstr.getBytes();
        _src = entry.readBytes(0, entry.getFileSize());
    }

    // Output element spec.
    private void putElementSpec(String name, int offset, int length, int depth,
                                String attr) {
        String stroff = Integer.toString(offset);
        String strlen = Integer.toString(length);
        String strdep = Integer.toString(depth);
        String attrsp;
        if (0 < attr.length() && attr.charAt(0) != ' ') attrsp = " " + attr;
            else attrsp = attr;

        // !! Use FrjObject.INVALID_ID.
        if (offset == -1 || length == -1) {
             // FrjWorld.warning
        } else {
            _outetb.println(stroff + "\t" + strlen + "\t" + strdep + "\t" +
                            name + attrsp);
        }
    }

    // Search string offset to forward.
    private int searchOffsetFW(int start, int end, String target) {
        byte[] tb = target.getBytes();

        for (int pos = start; pos <= end - tb.length; pos++) {
            int q = 0;

            while (q < tb.length && _src[pos + q] == tb[q]) {
                q++;
            }
            if (q == tb.length) {
                return pos;
            }
        }
        return -1;
    }

    // Search string offset to backward.
    private int searchOffsetBW(int start, int end, String target) {
        byte[] tb = target.getBytes();

        for (int pos = start - tb.length; end <= pos; pos--) {
            int q = 0;
            while (q < tb.length && _src[pos + q] == tb[q]) {
                q++;
            }
            if (q == tb.length) {
                return pos;
            }
        }
        return -1;
    }

    // Get range of type family.
    private Range getTypeFamily(FrjType t) {
        int sp = t.getOffset();
        int ep = sp + t.getLength();

        while (true) {
            FrjType t2 = t.getChildType();
            if (t2 == null) {
                break;
            }

            int nsp = t2.getOffset();
            int nep = nsp + t2.getLength();
            if (nsp < sp) sp = nsp;
            if (ep < nep) ep = nep;
            t = t2;
        }

        FrjQName q = t.getQName();
        if (q != null) {
            if (q.getOffset() < sp) {
                sp = q.getOffset();
            }
        }

        return new Range(sp, ep);
    }

    // Make modifier's attribute.
    private String makeModAttr(Enumeration e) {
        String attr = "";

        while (e.hasMoreElements()) {
            FrjModifier c = (FrjModifier)e.nextElement();
            String      s = c.getStr();

            if (s.equals("abstract")) attr += " abstract=\"yes\"";
            if (s.equals("final")) attr += " final=\"yes\"";
            if (s.equals("transient")) attr += " transient=\"yes\"";
            if (s.equals("volatile")) attr += " volatile=\"yes\"";
            if (s.equals("static")) attr += " static=\"yes\"";
            if (s.equals("synchronized")) attr += " synchro=\"yes\"";
            if (s.equals("native")) attr += " native=\"yes\"";
            if (s.equals("strictfp")) attr += " strictfp=\"yes\"";

            if (s.equals("public")) attr += " access=\"Public\"";
            if (s.equals("protected")) attr += " access=\"Protected\"";
            if (s.equals("private")) attr += " access=\"Private\"";
        }
        return attr;
    }

    // Make operator's attribute.
    private String makeOpAttr(FrjOperator c) {
        String  attr = "";
        String  s = c.getStr();
        int     p = c.getSort();

        if (s.equals(".")) attr += " sort=\"DOT\"";
        else if (s.equals("=")) attr += " sort=\"Assign\"";
        else if (s.equals("+=")) attr += " sort=\"AssignADD\"";
        else if (s.equals("-=")) attr += " sort=\"AssignSUB\"";
        else if (s.equals("*=")) attr += " sort=\"AssignMUL\"";
        else if (s.equals("/=")) attr += " sort=\"AssignDIV\"";
        else if (s.equals("%=")) attr += " sort=\"AssignMOD\"";
        else if (s.equals("&=")) attr += " sort=\"AssignAND\"";
        else if (s.equals("|=")) attr += " sort=\"AssignOR\"";
        else if (s.equals("^=")) attr += " sort=\"AssignXOR\"";
        else if (s.equals("<<=")) attr += " sort=\"AssignSHL\"";
        else if (s.equals(">>=")) attr += " sort=\"AssignSHR\"";
        else if (s.equals(">>>=")) attr += " sort=\"AssignSHRZ\"";

        else if (p == FrjOperator.REFER_DIM) attr += " sort=\"ArrayAccess\"";
        else if (p == FrjOperator.NEW_OBJECT) attr += " sort=\"InstanceCreation\"";
        else if (p == FrjOperator.NEW_ARRAY) attr += " sort=\"ArrayCreation\"";
        else if (p == FrjOperator.CAST) attr += " sort=\"Cast\"";
        else if (p == FrjOperator.PAREN) attr += " sort=\"Paren\"";
        else if (p == FrjOperator.POSTFIX) {
            if (s.equals("++")) attr += " sort=\"PostINC\"";
            if (s.equals("--")) attr += " sort=\"PostDEC\"";
        } else if (p == FrjOperator.PREFIX) {
            if (s.equals("++")) attr += " sort=\"PreINC\"";
            if (s.equals("--")) attr += " sort=\"PreDEC\"";
            if (s.equals("!")) attr += " sort=\"LogicalNOT\"";
            if (s.equals("~")) attr += " sort=\"NOT\"";
            if (s.equals("+")) attr += " sort=\"Plus\"";
            if (s.equals("-")) attr += " sort=\"Minus\"";
        } else if (p == FrjOperator.BINARY) {
            if (s.equals("instanceof")) attr += " sort=\"InstanceOf\"";
            if (s.equals("&")) attr += " sort=\"AND\"";
            if (s.equals("^")) attr += " sort=\"XOR\"";
            if (s.equals("|")) attr += " sort=\"OR\"";
            if (s.equals("*")) attr += " sort=\"MUL\"";
            if (s.equals("/")) attr += " sort=\"DIV\"";
            if (s.equals("%")) attr += " sort=\"MOD\"";
            if (s.equals("+")) attr += " sort=\"ADD\"";
            if (s.equals("-")) attr += " sort=\"SUB\"";
            if (s.equals("<<")) attr += " sort=\"SHL\"";
            if (s.equals(">>")) attr += " sort=\"SHR\"";
            if (s.equals(">>>")) attr += " sort=\"SHRZ\"";
            if (s.equals("<")) attr += " sort=\"CondLT\"";
            if (s.equals(">")) attr += " sort=\"CondGT\"";
            if (s.equals("<=")) attr += " sort=\"CondLE\"";
            if (s.equals("==")) attr += " sort=\"CondEQ\"";
            if (s.equals("!=")) attr += " sort=\"CondNE\"";
            if (s.equals(">=")) attr += " sort=\"CondGE\"";
            if (s.equals("&&")) attr += " sort=\"LogicalAND\"";
            if (s.equals("||")) attr += " sort=\"LogicalOR\"";
        } else if (p == FrjOperator.TERNARY) {
            attr += " sort=\"Selection\"";
        }

        return attr;
    }

    // Make ID attribute.
    private String makeIDAttr(String pre, long id) {
        if (isIntegerObjId) {
            // !! JPD_BIT_WIDTH_SIGN + JPD_BIT_WIDTH_CRID + JPD_BIT_WIDTH_OID =
            //    32.
            id -= ((long)1 << 32);
        }
        return " " + pre + "=\"s" + Long.toString(id) + "\"";
    }
    
    // Make derived ID attribute.
    private String makeIDAttrDerived(String pre) {
        return " " + pre + "=\"d" + Long.toString(derivedId++) + "\"";
    }

    // 
    // Extraction code following.
    // 

    private void extractFile(Enumeration e) {
        FrjFile     c = (FrjFile)e.nextElement();
        String      attr = makeIDAttr("id", c.getID());

        putElementSpec("File", 0, c.getLength(), 10, attr);

        // Package.
        {
            FrjQName q = (FrjQName)c.getDefinedPackage();
            if (q != null) {
                int     off = searchOffsetBW(q.getOffset(), 0, "package");
                int     off2 = searchOffsetFW(q.getOffset(), _src.length, ";");
                int     len = off2 - off + 1;
                String  attr4 = makeIDAttrDerived("id");

                putElementSpec("Package", off, len, 10, attr4);
            }
        }

        // Import.
        for (e = c.getImportedPackages(); e.hasMoreElements();) {
            FrjQName q = (FrjQName)e.nextElement();

            int     off = searchOffsetBW(q.getOffset(), 0, "import");
            int     off2 = searchOffsetFW(q.getOffset(), _src.length, ";");
            int     len = off2 - off + 1;
            String  attr4 = makeIDAttrDerived("id");

            putElementSpec("Import", off, len, 10, attr4);
        }
    }

    private void extractClass(Enumeration e) {
        FrjClass c = (FrjClass)e.nextElement();

        // Class.
        String attr = makeIDAttr("id", c.getID());
        attr += makeModAttr(c.getModifiers());
        putElementSpec("Class", c.getOffset(), c.getLength(), 10, attr);
            
        // ident.
        if (0 < c.getStrOffset()) {
            String attr8 = makeIDAttr("defid", c.getID());
            putElementSpec("ident", c.getStrOffset(), c.getStrLength(),
                           0, attr8);
        }

        // Body.
        int off = searchOffsetFW(c.getOffset(),
                                 c.getOffset() + c.getLength(), "{");
        int len = c.getLength() + c.getOffset() - off;
        putElementSpec("Members", off, len, 8, "");

        // Search superclass.
        {
            FrjType     t = c.getSuperClassType();
            if (t != null) {
                Range   r = getTypeFamily(t);
                String  attr4 = makeIDAttrDerived("id");

                attr4 += makeIDAttr("typefirst", t.getID());
                putElementSpec("ExtdOpt", r.sp, r.len(), 4, attr4);
            }
        }

        // Search interfaces.
        for (Enumeration e2 = c.getInterfaceTypes();
             e2.hasMoreElements();) {
            FrjType     t = (FrjType)e2.nextElement();
            Range       r = getTypeFamily(t);
            String      attr4 = makeIDAttrDerived("id");

            attr4 += makeIDAttr("typefirst", t.getID());
            putElementSpec("ImplOpt", r.sp, r.len(), 4, attr4);
        }
    }

    private void extractInterface(Enumeration e) {
        FrjInterface c = (FrjInterface)e.nextElement();

        // Intf.
        String attr = makeIDAttr("id", c.getID());
        String mod = makeModAttr(c.getModifiers());
        putElementSpec("Intf", c.getOffset(), c.getLength(), 10, attr + mod);

        // ident.
        if (0 < c.getStrOffset()) {
            String attr8 = makeIDAttr("defid", c.getID());
            putElementSpec("ident", c.getStrOffset(), c.getStrLength(), 0, 
                           attr8);
        }

        // Body.
        int off = searchOffsetFW(c.getOffset(),
                                 c.getLength() + c.getOffset(), "{");
        int len = c.getLength() + c.getOffset() - off;
        putElementSpec("Members", off, len, 8, "");

        // Search interfaces.
        for (Enumeration e2 = c.getSuperInterfaceTypes();
             e2.hasMoreElements();) {
            FrjType     t = (FrjType)e2.nextElement();
            Range       r = getTypeFamily(t);
            String      attr4 = makeIDAttrDerived("id");

            attr4 += makeIDAttr("typefirst", t.getID());
            putElementSpec("ImplOpt", r.sp, r.len(), 4, attr4);
        }
    }

    private void extractRoutine(Enumeration e) {
        FrjRoutine  c = (FrjRoutine)e.nextElement();
        int         s = c.getSort();
        String      elem;

        // Method and Ctor.
        if (s == FrjRoutine.METHOD) elem = "Method";
        else if (s == FrjRoutine.CONSTRUCTOR) elem = "Ctor";
        else return; // Error.

        String attr = makeIDAttr("id", c.getID());
        if (s == FrjRoutine.METHOD) {
            attr += makeIDAttr("typefirst", c.getReturnType().getID());
        }
        attr += makeModAttr(c.getModifiers());
        putElementSpec(elem, c.getOffset(), c.getLength(), 10, attr);
        
        // ident.
        if (0 < c.getStrOffset() && s == FrjRoutine.METHOD) {
            String      attr8 = makeIDAttr("defid", c.getID());

            putElementSpec("ident", c.getStrOffset(), c.getStrLength(), 0, 
                           attr8);
        }
        if (s == FrjRoutine.CONSTRUCTOR) {
            String      attr8 = makeIDAttr("defid", c.getID());
            String      tname = c.getDefinedClassInterface().getStr();
            int         pos = searchOffsetFW(c.getOffset(),
                                     c.getOffset() + c.getLength(), tname);
            putElementSpec("ident", pos, tname.length(), 0, attr8);
        }

        // Search throws.
        for (Enumeration e2 = c.getThrowTypes(); e2.hasMoreElements();) {
            FrjType     t = (FrjType)e2.nextElement();
            Range       r = getTypeFamily(t);
            String      attr4 = makeIDAttrDerived("id");

            attr4 += makeIDAttr("typefirst", t.getID());
            putElementSpec("ThrwOpt", r.sp, r.len(), 4, attr4);
        }
    }

    private void extractVariable(Enumeration e) {
        FrjVariable c = (FrjVariable)e.nextElement();
        int         s = c.getSort();
        String      elem;

        // Field/LocalVar/Param.
        if (s == FrjVariable.CLASS_MEM) elem = "Field";
        else if (s == FrjVariable.INTERFACE_MEM) elem = "Field";
        else if (s == FrjVariable.LOCAL_VAR) elem = "Local";
        else if (s == FrjVariable.PARAMETER) elem = "Param";
        else return;

        String      attr = makeIDAttr("id", c.getID());
        attr += makeIDAttr("typefirst", c.getType().getID());
        
        // Search decl range.
        Range       r = new Range(c.getOffset(), c.getOffset() + c.getLength());
        FrjType     t = c.getType();
        Range       r2 = getTypeFamily(t);
        if (r2.sp < r.sp) r.sp = r2.sp;
        if (r.ep < r2.ep) r.ep = r2.ep;
    
        for (Enumeration e2 = c.getModifiers(); e2.hasMoreElements();) {
            FrjModifier m = (FrjModifier)e2.nextElement();
            if (m.getOffset() < r.sp) {
                r.sp = m.getOffset();
            }
        }
    
        FrjExpression x = c.getInitializer();
        if (x != null) {
            r.ep = x.getOffset() + x.getLength();
        }

        // Chain check.
        int comma = searchOffsetBW(c.getOffset(), r.sp, ",");
        if (r.sp < comma) {
            r.sp = c.getOffset();
            attr += " omitted=\"yes\"";
        }
        
        attr += makeModAttr(c.getModifiers());
        putElementSpec(elem, r.sp, r.len(), 10, attr);
        
        // Ident.
        String attr8 = makeIDAttr("defid", c.getID());
        putElementSpec("ident", c.getOffset(), c.getLength(), 0, attr8);
    }

    private void extractType(Enumeration e) {
        FrjType     c = (FrjType)e.nextElement();
        String      attr = makeIDAttr("id", c.getID());
        int         s = c.getSort();

        // Type.
        if (s == FrjType.ARRAY) attr += " sort=\"Array\"";
        if (s == FrjType.PRIMITIVE) attr += " sort=\"Primitive\"";
        if (s == FrjType.OBJECT) {
            // FrjClassInterface d = (FrjClassInterface)c.getReferringClassInterface();
            // if (d instanceof FrjClass) attr += " sort=\"Class\"";
            // if (d instanceof FrjInterface) attr += " sort=\"Interface\"";
            attr += " sort=\"Object\"";
            FrjQName q = (FrjQName)c.getQName();
            if (q != null && q.hasValidID()) {
                attr += makeIDAttr("qname", q.getID());
            }
        }

        if (s == FrjType.ARRAY) {
            FrjType     d = (FrjType)c.getChildType();

            if (d != null && d.hasValidID()) {
                attr += makeIDAttr("typenext", d.getID());
            }
        }

        putElementSpec("Type", c.getOffset(), c.getLength(), 2, attr);

        // ident.
        if (s == FrjType.OBJECT) {
            String              attr8 = "";
            FrjClassInterface   d = (FrjClassInterface)c.getReferringClassInterface();

            if (d != null && d.hasValidID()) {
                attr8 = makeIDAttr("defid", d.getID());
            }
            putElementSpec("ident", c.getOffset(), c.getLength(), 0, attr8);
        }
    }

    private void extractSBlock(Enumeration e) {
        FrjSBlock       c = (FrjSBlock)e.nextElement();
        String          attr = makeIDAttr("id", c.getID());

        putElementSpec("SInit", c.getOffset(), c.getLength(), 10, attr);
    }

    private void extractStatement(Enumeration e) {
        FrjStatement    c = (FrjStatement)e.nextElement();
        String          attr = makeIDAttr("id", c.getID());
        String          attr2 = " ";

        int s = c.getSort();
        if (s == FrjStatement.EMPTY) attr2 += "sort=\"EMPTY\"";
        if (s == FrjStatement.EXPRESSION) attr2 += "sort=\"EXPR\"";
        if (s == FrjStatement.BLOCK) attr2 += "sort=\"BLOCK\"";
        if (s == FrjStatement.DO) attr2 += "sort=\"DOWHILE\"";
        if (s == FrjStatement.WHILE) attr2 += "sort=\"WHILE\"";
        if (s == FrjStatement.FOR) attr2 += "sort=\"FOR\"";
        if (s == FrjStatement.IF) attr2 += "sort=\"IFELSE\"";
        if (s == FrjStatement.SWITCH) attr2 += "sort=\"SWITCH\"";
        if (s == FrjStatement.BREAK) attr2 += "sort=\"BREAK\"";
        if (s == FrjStatement.CONTINUE) attr2 += "sort=\"CONT\"";
        if (s == FrjStatement.RETURN) attr2 += "sort=\"RETURN\"";
        if (s == FrjStatement.THROW) attr2 += "sort=\"THROW\"";
        if (s == FrjStatement.SYNCHRONIZED) attr2 += "sort=\"SYNCHRO\"";
        if (s == FrjStatement.EXCEPTION) attr2 += "sort=\"TRY\"";
        if (s == FrjStatement.ASSERT) attr2 += "sort=\"ASSERT\"";

        if (attr2 == " ") {
            return; // Error.
        }

        putElementSpec("Stmt", c.getOffset(), c.getLength(), 8, attr + attr2);
    }

    private void extractExpression(Enumeration e) {
        FrjExpression   c = (FrjExpression)e.nextElement();
        String          attr = makeIDAttr("id", c.getID());
        String          attr2 = "";
        int             s = c.getSort();

        if (s == FrjExpression.EMPTY) attr2 += " sort=\"Empty\"";
        if (s == FrjExpression.REF_LITERAL) attr2 += " sort=\"Literal\"";
        if (s == FrjExpression.REF_TYPE) attr2 += " sort=\"TypeRef\"";
        if (s == FrjExpression.REF_VAR) attr2 += " sort=\"VarRef\"";
        if (s == FrjExpression.CONSTRUCTOR_CALL) attr2 += " sort=\"CtorCall\"";
        if (s == FrjExpression.METHOD_CALL) attr2 += " sort=\"MethodCall\"";
        if (s == FrjExpression.LISTED) attr2 += " sort=\"Listed\"";
        if (s == FrjExpression.OPERATION) {
            FrjOperator d = (FrjOperator)c.getOperator();
            attr2 += makeOpAttr(d);
        }
        if (s == FrjExpression.REF_SPECIAL) {
            byte    ch = _src[c.getOffset()], ch2 = _src[c.getOffset() + 1];

            if (ch == 's' && ch2 == 'u') {
                // Skip "super".
                return;
            } else if (ch == 'c' && ch2 == 'l') {
                // Added by S.Yamamoto 2003/02/26.
                attr2 += " sort=\"Class\"";
            } else if (ch == 't' && ch2 == 'h') {
                attr2 += " sort=\"This\"";
            } else {
                if (ch == 'n' && ch2 == 'u') {
                    putElementSpec("literal", c.getOffset(), c.getLength(),
                                   0, "sort=\"NULL\"");
                } else if ((ch == 't' && ch2 == 'r') ||
                           (ch == 'f' && ch2 == 'a')) {
                    putElementSpec("literal", c.getOffset(), c.getLength(),
                                   0, "sort=\"BOOL\"");
                } else {
                    FrjWorld.message("Notice: id " + c.getID() +
                                     ", offset " + c.getOffset() +
                                     ": Invalid REF_SPECIAL.");
                }
                attr2 += " sort=\"Literal\"";
            }
        }
        if (attr2.equals("")) {
            return;     // If REF_QNAME and REF_SPECIAL (super), not Expr.
        }

        // Cast type.
        if (s == FrjExpression.REF_TYPE) {
            FrjType     t = c.getType();

            if (t != null) {
                attr += makeIDAttr("typefirst", t.getID());
            }
        }

        // Variable.
        if (s == FrjExpression.REF_VAR) {
            // Create ident node.
            String      attr8 = "";
            FrjVariable d = (FrjVariable)c.getReferree();
            int         k = c.getLRSort();

            if (d != null && d.hasValidID()) {
                attr8 = makeIDAttr("defid", d.getID());
            }

            if (k == FrjExpression.R_VALUE || k == FrjExpression.LR_VALUE) {
                attr2 += " read=\"yes\"";
            }
            if (k == FrjExpression.L_VALUE || k == FrjExpression.LR_VALUE) {
                attr2 += " write=\"yes\"";
            }

            putElementSpec("ident", c.getOffset(), c.getLength(), 0, attr8);
        }

        // Routine.
        if (s == FrjExpression.METHOD_CALL) {
            // Create ident node.
            String      attr8 = "";
            FrjRoutine  d = (FrjRoutine)c.getRoutine();

            if (d != null && d.hasValidID()) {
                attr8 = makeIDAttr("defid", d.getID());
            }

            putElementSpec("ident", c.getStrOffset(), c.getStrLength(), 0,
                           attr8);
        }

        int     w = (s == FrjExpression.LISTED) ? 5 : 4;

        putElementSpec("Expr", c.getOffset(), c.getLength(), w, attr + attr2);
    }

    private void extractLiteral(Enumeration e) {
        FrjLiteral  c = (FrjLiteral)e.nextElement();
        String      attr = makeIDAttr("id", c.getID());
        int         s = c.getSort();

        if (s == FrjLiteral.CHAR) { attr += " sort=\"CHAR\""; }
        else if (s == FrjLiteral.STRING) { attr += " sort=\"STR\""; }
        else if (s == FrjLiteral.UNICODE) { attr += " sort=\"CHAR\""; }
        else if (s == FrjLiteral.INT) { attr += " sort=\"INT\""; }
        else if (s == FrjLiteral.DOUBLE) { attr += " sort=\"FLOAT\""; }
        else { FrjWorld.errorExit(s + ": Invalid literal sort."); }

        putElementSpec("literal", c.getOffset(), c.getLength(), 0, attr);
    }

    private void extractQName(Enumeration e) {
        FrjQName    c = (FrjQName)e.nextElement();
        String      attr = makeIDAttr("id", c.getID());

        putElementSpec("QName", c.getOffset(), c.getLength(), 2, attr);
        FrjOperator d = (FrjOperator) c.getDot();

        byte ch = _src[c.getOffset()];
        if (ch != '*' && !(d != null && d.hasValidID()) ) {
            putElementSpec("ident", c.getOffset(), c.getLength(), 0, "");
        }
    }

    private void extractLabel(Enumeration e) {
        FrjLabel    c = (FrjLabel)e.nextElement();
        int         s = c.getSort();
        int         sp = c.getOffset();

        if (s == FrjLabel.REFERRED) {
            String      attr2 = makeIDAttr("defid", c.getReferree().getID());

            putElementSpec("ident", sp, c.getLength(), 0, attr2);
        } else {
            String  attr = makeIDAttr("id", c.getID());
            int     ep = searchOffsetFW(c.getOffset(), _src.length, ":");

            if (s == FrjLabel.DEFINED) {
                String  attr2 = makeIDAttr("defid", c.getID());

                putElementSpec("ident", sp, c.getLength() - 1, 0, attr2);
            }
            putElementSpec("Label", sp, ep - sp + 1, 8, attr);
        }
    }

    private void extractComment(Enumeration e) {
        FrjComment  c = (FrjComment)e.nextElement();
        String      attr = makeIDAttr("id", c.getID());

        putElementSpec("comment", c.getOffset(), c.getLength(), 0, attr);
    }
}
