/* 
 * Program:             $RCSfile: AddGlobalLink.java,v $  $Revision: 50.4 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hotta  2003/09/02
 *                      S.Yamamoto  2003/09/02
 *
 * (C) Copyright:       Y.Hotta and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/*
 * $Id: AddGlobalLink.java,v 50.4 2003/11/26 23:25:19 yamamoto Exp yamamoto $
 */

package JXModel3.jx2html;

import java.io.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import sapid.util.*;
import gnu.getopt.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.*;

public class AddGlobalLink {
    String inxml;
    String pstxml;
    String outxml;
    Document pstdoc  = null;
    boolean flag_type = false;
    boolean flag_qname = false;
    boolean flag_expr = false;
    boolean flag_ident = false;
    boolean flag_ident2 = false;
    String defid;
    String linkfile;

    public AddGlobalLink(String[] args) throws Exception {
        checkOption(args, args.length);
        addLink();
    }

    public void checkOption(String args[], int argc) {
        int c;
        Getopt g = new Getopt("AddGlobalLink", args, "i:p:o:");
        while ((c = g.getopt()) != -1) {
            switch (c) {
            case 'i':
                inxml = g.getOptarg();
                break;
            case 'p':
                pstxml = g.getOptarg();
                break;
            case 'o':
                outxml = g.getOptarg();
                break;
            default:
                break;
            }
        }
    }

    public void addLink() throws Exception {
        Document doc = null;
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder        db  = dbf.newDocumentBuilder();
        doc     = db.parse(new File(inxml));
        pstdoc  = db.parse(new File(pstxml));
        String encode = SpdEncoding.getEncoding(inxml).toString();
        Element root = doc.getDocumentElement();
        walk(root);

        TransformerFactory transFactory = TransformerFactory.newInstance();
        Transformer transformer = transFactory.newTransformer();
        DOMSource source = new DOMSource(doc);
        StreamResult result = new StreamResult(new FileOutputStream(new File(outxml)));
        transformer.transform(source, result);
    }

    public void walk(Node n) {
        for (Node ch = n.getFirstChild(); ch != null; ch = ch.getNextSibling()) {
            if (ch.getNodeType() == Node.ELEMENT_NODE) {
                if (ch.getNodeName().equals("Type")){
                    NamedNodeMap typeattrlist = ch.getAttributes();
                    Node typeattr = typeattrlist.getNamedItem("id");
                    setLinkData(typeattr.getNodeValue());
                    flag_type = true;
                } else if (ch.getNodeName().equals("QName")) {
                    NamedNodeMap qnameattrlist = ch.getAttributes();
                    Node qnameattr = qnameattrlist.getNamedItem("id");
                    setLinkData(qnameattr.getNodeValue());
                    flag_qname = true;
                } else if (ch.getNodeName().equals("Expr")) {
                    NamedNodeMap exprattrlist = ch.getAttributes();
                    Node exprsort = exprattrlist.getNamedItem("sort");
                    if(exprsort.getNodeValue().equals("VarRef") ||
                       exprsort.getNodeValue().equals("MethodCall")) {
                        Node exprattr = exprattrlist.getNamedItem("id");
                        setLinkData(exprattr.getNodeValue());
                        flag_expr = true;
                    }
                } else if (ch.getNodeName().equals("ident")){
                    if ((flag_type || flag_qname || flag_expr) && !ch.hasAttributes()) {
                        flag_type = false;
                        flag_qname = false;
                        flag_expr = false;
                        if (flag_ident) {
                            Element elm = (Element) ch;
                            elm.setAttribute("file", linkfile);
                            elm.setAttribute("defid", defid);
                            flag_ident = false;
                        }
                    }
                } else if (ch.getNodeName().equals("kw")){
                    if (flag_type) {
                        flag_type = false;
                    }
                }
                walk(ch);
            }
        }
    }

    public void setLinkData(String id){
        Element pstroot = pstdoc.getDocumentElement();
        Node no =(Node)pstroot;
        NodeList classRefList = no.getChildNodes();
        if (classRefList.getLength() > 0) {
            for (int i = 0; i < classRefList.getLength(); i++) {
                Node n = classRefList.item(i);
                boolean flag_ref = false;
                if (n.getNodeName().equals("FieldRef")){
                    flag_ref = true;
                }else if (n.getNodeName().equals("ClassRef")){
                    flag_ref = true;
                }else if (n.getNodeName().equals("MethodRef")){
                    flag_ref = true;
                }
                if (n.getNodeType() == Node.ELEMENT_NODE && flag_ref) {
                    NamedNodeMap attrlist = n.getAttributes();
                    Node attr = attrlist.getNamedItem("id");
                    if (attr.getNodeValue().equals(id)){
                        NodeList nodeList = n.getChildNodes();
                        for (int j = 0; j < nodeList.getLength(); j++) {
                        Node refTgt = nodeList.item(j);
                            if(refTgt.getNodeType() == Node.ELEMENT_NODE &&
                               refTgt.getNodeName().equals("RefTgt")) {
                                NamedNodeMap refTgtattrlist = refTgt.getAttributes();
                                NodeList refTgtAList = refTgt.getChildNodes();
                                Node attrId = refTgtattrlist.getNamedItem("id");
                                defid = attrId.getNodeValue();
                                Node attrexists = refTgtattrlist.getNamedItem("exists");
                                if (attrexists.getNodeValue().equals("InPrj")) {
                                    flag_ident2 = true;
                                }
                                for (int k = 0; k < refTgtAList.getLength(); k++) {
                                    Node refTgtA = refTgtAList.item(k);
                                    if (refTgtA.getNodeType() == Node.ELEMENT_NODE) {
                                        NamedNodeMap refTgtAattrlist = refTgtA.getAttributes();
                                        Node attrHref = refTgtAattrlist.getNamedItem("href");
                                        linkfile = attrHref.getNodeValue();
                                        if (linkfile.equals("") || defid.equals("") ) {
                                            flag_ident = false;
                                        } else if(flag_ident2) {
                                            flag_ident = true;
                                            flag_ident2 = false;
                                        } else {
                                            flag_ident = false;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    public static void main(String[] args) throws Exception {
        AddGlobalLink agl = new AddGlobalLink(args);
    }
}
