/* 
 * Program:             $RCSfile: MkOutline.java,v $  $Revision: 50.23 $
 * 
 * Purpose:     
 *
 * Author:              Y.Sawada  2003/10/01
 *                      S.Yamamoto  2003/10/01
 *
 * (C) Copyright:       Y.Sawada and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: MkOutline.java,v 50.23 2003/11/26 23:25:19 yamamoto Exp yamamoto $
 */

package JXModel3.jx2html;

import java.io.*;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import sapid.util.*;
import gnu.getopt.*;

public class MkOutline {
    private   StringBuffer strBuf;
    private   String inputStr;
    protected boolean noHeader;
    protected String inFile;
    private   String midFile;
    protected String outFile;

    // ϥեΥ󥳡
    protected String encode;
    // xsltե뷲Υѥ
    protected String xsltPath;

    protected String topDir;

    public MkOutline()
    {
        strBuf = new StringBuffer();
        outFile = null;
        inputStr = null;
        noHeader = false;
        midFile = null;
        xsltPath = "xslt/";
        topDir = "../";
    }

    public MkOutline(String[] args)
    {
        this();
        parseOptions(args, args.length);
        setEncoding();
    }

    public MkOutline(String inputFile, String outputFile)
    {
        this();
        inFile = inputFile;
        outFile = outputFile;
        setEncoding();
    }

    public MkOutline(String inputFile, String outputFile, String xsltpath)
    {
        this(inputFile, outputFile);
        xsltPath = xsltpath;
    }

    public String getInFile()
    {
        return inFile;
    }
    public String getOutFile()
    {
        return outFile;
    }
    public String getEncoding()
    {
        return encode;
    }
    public String getXsltpath()
    {
        return xsltPath;
    }
    public String getStrBuf()
    {
        return new String(strBuf);
    }
    public String getTopDir()
    {
        return topDir;
    }

    public void setInFile(String str)
    {
        inFile = str;
    }
    public void setOutFile(String str)
    {
        outFile = str;
    }
    public void setInputString(String str)
    {
        inputStr = str;
    }
    public void setTopDir(String str)
    {
        topDir = str;
    }
    protected void setEncoding()
    {
        try {
            encode = SpdEncoding.getEncoding(inFile).toString();
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public StringBuffer append(String str)
    {
        return strBuf.append(str);
    }

    public StringBuffer append(StringBuffer sb)
    {
        return strBuf.append(sb);
    }

    public StringBuffer insert(int offset, String str)
    {
        return strBuf.insert(offset, str);
    }

    // xsl Transform XMLappend
    public void xslTransform(String xsltFile)
    {
        String[] paramName = { "topDir" };
        String[] paramValue = { topDir };

        xslTransform(xsltFile, paramName, paramValue);
    }

    public void xslTransform(String xsltFile, String[] paramName, String[] paramValue)
    {
        try {
            TransformerFactory factory = TransformerFactory.newInstance();

            Transformer transfm =
                factory.newTransformer(new StreamSource(xsltPath + xsltFile));

            transfm.setOutputProperty("encoding", encode);

            if (paramName.length != paramValue.length) {
                throw new IllegalArgumentException("Length of paramName(" +
                                                   paramName.length +
                                                   ")is different from paramValue(" +
                                                   paramValue.length + ")");
            }
            for (int i = 0; i < paramName.length; i++) {
                transfm.setParameter(paramName[i], paramValue[i]);
            }

            ByteArrayOutputStream os = new ByteArrayOutputStream();
            if (inputStr == null) {
                transfm.transform(new StreamSource(inFile),
                                  new StreamResult(os));
            } else {
                transfm.transform(new StreamSource(new StringReader(inputStr)),
                                  new StreamResult(os));
            }
            strBuf.append(os.toString());
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    // strbufƤ񤭽Ф
    public void flushBuffer()
    {
        flushBuffer(outFile);
    }
    public void flushBuffer(String outputFile)
    {
        if (outputFile != null) {
            try {
                PrintWriter pw = new PrintWriter
                    (new BufferedWriter(new FileWriter(outputFile)));
                pw.println(strBuf);
                pw.close();
            } catch(Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        } else {
            System.out.println(strBuf);
        }
        releaseBuffer();
    }

    public void releaseBuffer()
    {
        strBuf = new StringBuffer();
    }

    private String makeHeader(String title, String css, String encoding,
                              String depth, String[] jsfiles)
    {
        StringBuffer sb = new StringBuffer(
            "<?xml version=\"1.0\" encoding=\"" + encoding + "\" ?>\n" +
            "<!DOCTYPE html " +
            "PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" " +
            "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n" +
            "<html xmlns=\"http://www.w3.org/1999/xhtml\" " +
            "xml:lang=\"ja\" lang=\"ja\">\n" +
            "  <head>\n" +
            "    <title>" + title + "</title>\n" +
            "    <meta http-equiv=\"Content-Type\" " +
            "content=\"text/html; charset=" + encoding + "\" />\n" +
            "    <meta http-equiv=\"Content-Style-Type\" " +
            "content=\"text/css\" />\n" +
            "    <meta http-equiv=\"Content-Script-Type\" " +
            "content=\"text/javascript\" />\n" +
            "    <link href=\"" + depth + css +
            "\" type=\"text/css\" rel=\"stylesheet\" />\n");
        if (jsfiles != null) {
            for (int i = 0; i < jsfiles.length; i++) {
                sb.append("    <script src=\"" + depth + topDir +
                          "js/" + jsfiles[i] +
                          "\" type=\"text/javascript\"></script>\n");
            }
        }
        sb.append("  </head>\n");
        return new String(sb);
    }
    protected String makeHeader(String title, String css, String encoding,
                                String[] jsfiles)
    {
        return makeHeader(title, css, encoding, "", jsfiles);
    }
    protected String makeHeader(String title, String css, String encoding)
    {
        return makeHeader(title, css, encoding, "", null);
    }

    protected static String appendSlash(String str)
    {
        return str.endsWith("/") ? str : str + "/";
    }

    public static String packageDepth(String str)
    {
        return str.replace('.', '/').replaceAll("[^/]+", "..");
    }
    public boolean parseOptions(String args[], int argc)
    {
        int     ch;
        Getopt  g = new Getopt("MkOutline", args, "x:t:m:n");

        while ((ch = g.getopt()) != -1) {
            switch (ch) {
            case 'x':
                xsltPath = appendSlash(g.getOptarg());
                break;
            case 't':
                topDir = appendSlash(g.getOptarg());
                break;
            case 'm':
                midFile = g.getOptarg();
                break;
            case 'n':
                noHeader = true;
                break;
            default:
                // print_usage(program_name);
                break;
            }
        }
        argc -= g.getOptind();
        switch (argc) {
        case 1:
            inFile = new String(args[g.getOptind()]);
            break;
        case 2:
            inFile = new String(args[g.getOptind()]);
            outFile = new String(args[g.getOptind() + 1]);
            break;
        default:
            System.err.println("command [options] InputDir [OutputFile]\n" +
                               "OPTOION:\n" +
                               " -x XSLTdir\n" +
                               " -m\n" +
                               " -t TOPdir\n" +
                               " -n\n");
            System.exit(1);
        }
        return true;
    }

    public static void main(String[] args)
    {
        MkOutline mkol = new MkOutline(args);

        String inputFile = mkol.getInFile();

        mkol.append("<?xml version=\"1.0\" encoding=\"" +
                    mkol.getEncoding() + "\" ?>\n");
        mkol.append("<File name=\"" +
                    (new File(inputFile)).getName().
                    replaceFirst("\\.java\\.xml$", "") +
                    "\">\n");
        mkol.xslTransform("outline.xsl");
        mkol.append("</File>\n");

        String str = mkol.getStrBuf();
        String absPath = (new File(inputFile)).getAbsolutePath();

        if (mkol.midFile != null) {
            mkol.flushBuffer(mkol.midFile);
        } else {
            mkol.releaseBuffer();
        }

        // ΤޤϤ
        mkol.setInputString(str);

        mkol.xslTransform("treeView.xsl");

        String strBuf = mkol.getStrBuf();
        String packageStr  = strBuf.substring(0, strBuf.indexOf(">")).
            replaceFirst("<!--Package:(.*)--", "$1");

        String[] jsfiles = {"tree.js", "display.js"};

        if (!mkol.noHeader) {
            mkol.insert(0, mkol.
                        makeHeader("Outline",
                                   "Jx2Html.css",
                                   mkol.getEncoding(),
                                   packageStr.length() > 0
                                   ? appendSlash(packageDepth(packageStr))
                                   : "",
                                   jsfiles) +
                        "<body>\n" +
                        "<div>\n" +
                        inputFile.substring(0, inputFile.lastIndexOf('.')).
                        replaceAll(".+/", "") + " :");
        }

        if (!mkol.noHeader) {
            mkol.strBuf.append("</div>\n" +
                               "</body>\n" +
                               "</html>");
        }
        mkol.flushBuffer();
    }
}
