/* 
 * Program:             $RCSfile: Xref.java,v $  $Revision: 50.13 $
 * 
 * Purpose:     
 *
 * Author:              T.Takahashi  2003/10/30
 *                      S..Yamamoto  2003/10/30
 *
 * (C) Copyright:       T.Takahashi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Xref.java,v 50.13 2003/11/26 23:25:19 yamamoto Exp yamamoto $
 */

package JXModel3.jx2html;

import java.io.*;
import java.util.*;
import gnu.getopt.*;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;

class Xref extends MkOutline
{
    private String midFile;
    private String outlinePath;
    private Hashtable hashtbl;

    public Xref()
    {
        super();
        midFile = null;
        outlinePath = null;
        hashtbl = new Hashtable();
        setEncoding();
    }
    public Xref(String[] args)
    {
        this();
        parseOptions(args, args.length);
    }
    public String getMidFile()
    {
        return midFile;
    }

    // System  default property 
    protected void setEncoding()
    {
        String str = System.getProperty("file.encoding");

        if (str.equals("MS932")) {
            encode = "Shift_JIS";
        } else {
            encode = str;
        }
    }

    public void concatXMLFiles()
    {
        DirWalkIterator dw = new DirWalkIterator(inFile,
                                                 ".*\\.java\\.ol\\.xml$",
                                                 true);
        File tmpFile;
        StringBuffer sb = new StringBuffer();
        sb.append("<?xml version=\"1.0\" encoding=\"" + encode + "\" ?>\n" +
                  "<xsl:stylesheet version=\"1.0\" " +
                  "xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\">\n" +
                  " <xsl:output method=\"xml\" " +
                  "indent=\"yes\" " +
                  "encoding=\" +" + encode + "\"/>\n" +
                  " <xsl:template match=\"/\">\n" +
                  "  <FileList>\n");
        while (dw.hasNext()) {
            sb.append("   <xsl:copy-of select=\"" +
                      "document('" +
                      ((File)dw.next()).getAbsolutePath() +
                      "')\"/>\n");
        }
        sb.append("  </FileList>\n" +
                  " </xsl:template>\n" +
                  "</xsl:stylesheet>");
        xslStringTransform(new String(sb));
    }
    private void xslStringTransform(String xslString)
    {
        try {
            TransformerFactory factory = TransformerFactory.newInstance();

            Transformer transfm =
                factory.newTransformer(new StreamSource
                    (new StringReader(xslString)));

            transfm.setOutputProperty("encoding", encode);
            String inputStr =
                "<?xml version=\"1.0\" encoding=\"Shift_JIS\"?>\n" +
                "<root>\n</root>";

            ByteArrayOutputStream os = new ByteArrayOutputStream();

            transfm.transform(new StreamSource(new StringReader(inputStr)),
                              new StreamResult(os));
            append(os.toString());
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    public boolean parseOptions(String args[], int argc)
    {
        int     ch;
        Getopt  g = new Getopt("Xref", args, "x:m:t:n");

        while ((ch = g.getopt()) != -1) {
            switch (ch) {
            case 'x':
                xsltPath = appendSlash(g.getOptarg());
                break;
            case 'm':
                midFile = g.getOptarg();
                break;
            case 't':
                topDir = appendSlash(g.getOptarg());
                break;
            default:
                // print_usage(program_name);
                break;
            }
        }
        argc -= g.getOptind();
        switch (argc) {
        case 1:
            inFile = new String(args[g.getOptind()]);
            break;
        case 2:
            inFile = new String(args[g.getOptind()]);
            outFile = new String(args[g.getOptind() + 1]);
            break;
        default:
            System.err.println("command [options] InputDir [OutputFile]\n" +
                               "OPTOION:\n" +
                               " -x XSLTdir\n" +
                               " -m MidOutputFile\n" +
                               " -t TOPdir\n");
            System.exit(1);
        }
        return true;
    }

    public static void main(String args[])
    {
        Xref xref = new Xref(args);

        xref.concatXMLFiles();

        String str = xref.getStrBuf();

        if (xref.getMidFile() != null) {
            xref.flushBuffer(xref.getMidFile());
        } else {
            xref.releaseBuffer();
        }

        // ΤޤϤˤޤ魯
        xref.setInputString(str);

        String topDir = xref.getTopDir();
        String[] jsfiles = {"tree.js", "display.js", "fileset.js"};

        xref.append(xref.makeHeader("Xref",
                                    xref.topDir + "Jx2Html.css",
                                    //"Jx2Html.css",
                                    xref.getEncoding(),
                                    jsfiles));
        xref.xslTransform("xref.xsl");
        xref.append("</html>");
        xref.flushBuffer();
    }
}

class DirWalkIterator implements Iterator
{
    private LinkedList stack;
    private HashSet hashSet;
    private boolean canonical;
    private String regExp;

    DirWalkIterator(String fileName)
    {
        stack = new LinkedList();
        stack.addFirst(new File(fileName));
        hashSet = new HashSet();
        canonical = false;
        regExp = null;
    }
    DirWalkIterator(String fileName, String regExp)
    {
        this(fileName);
        this.regExp = regExp;
    }
    DirWalkIterator(String fileName, boolean canonical)
    {
        this(fileName);
        this.canonical = canonical;
    }
    DirWalkIterator(String fileName, String regExp, boolean canonical)
    {
        this(fileName);
        this.regExp = regExp;
        this.canonical = canonical;
    }
    private boolean matchRegExp(File file, String regExp)
    {
        if (regExp == null) {
            return true;
        } else {
            return file.getName().matches(regExp);
        }
    }
    private boolean canonicalp(File file)
    {
        if (canonical) {
            String canonicalPath = new String();
            try {
                canonicalPath = file.getCanonicalPath();
            } catch (Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
            return hashSet.add(canonicalPath);
        } else {
            return true;
        }
    }

    public boolean hasNext()
    {
        if (stack.size() == 0)
            return false;
        File tmpFile = (File)stack.getFirst();

        if (tmpFile.isDirectory()) {
            File[] fileList = tmpFile.listFiles();
            for (int i = fileList.length - 1; i >= 0; i--) {
                if (fileList[i].isDirectory() ||
                    matchRegExp(fileList[i], regExp) &&
                    canonicalp(fileList[i])) {
                    stack.add(1, fileList[i]);
                }
            }
            stack.removeFirst();
            return hasNext();
        }
        return true;
    }

    public Object next()
    {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        return stack.removeFirst();
    }

    public void remove()
    {
        if (hasNext()) {
            stack.removeFirst();
        }
    }
}
