/* 
 * Program:             $RCSfile: JXTester.java,v $  $Revision: 40.21 $
 * 
 * Purpose:             
 *
 * Author:              Y.Sawada  2003/02/19
 *                      S.Yamamoto  2003/02/19
 *
 * (C) Copyright:       Y.Sawada and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: JXTester.java,v 40.21 2003/07/08 05:48:06 yamamoto Exp yamamoto $
 */

package JXModel3.jxtester;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import Japid.util.Stopwatch;

class JXTester extends DefaultHandler {
    boolean         test;
    Stack           elements;
    Locator         locator;
    Locator         startloc;
    StringBuffer    sb;
    String          sort;
    String          newline = "unknown";
    static String   targetFile;
    
    Pattern pattern;
    Matcher matcher;

    char[] sp = {' ', '\t'};

    String[] nl = {"\n", "\r", "\r\n"};

    String[] kw = {
        "abstract", "default", "if", "private", "this",
        "boolean", "do", "implements", "protected", "throw",
        "break", "double", "import", "public", "throws",
        "byte", "else", "instanceof", "return", "transient",
        "case", "extends", "int", "short", "try",
        "catch", "final", "interface", "static", "void",
        "char", "finally", "long", "strictfp", "volatile",
        "class", "float", "native", "super", "while",
        "const", "for", "new", "switch",
        "continue", "goto", "package", "synchronized",
        
        "assert"
    };

    String[] op = {
        //Operators
        "=", ">", "<", "!", "~", "?", ":",
        "==", "<=", ">=", "!=", "&&", "||", "++", "--",
        "+", "-", "*", "/", "&", "|", "^", "%", "<<", ">>", ">>>",
        "+=", "-=", "*=", "/=", "&=", "|=", "^=", "%=", "<<=", ">>=", ">>>=",
        // Separators.
        "(", ")", "{", "}", "[", "]", ";", ",", ".",
        // !!
        "[]"
    };

    String[] literalBOOL = {"true", "false"};

    String literalNULL = "null";

    char[] literalINT = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        'a', 'b', 'c', 'd', 'e', 'f', 'A', 'B', 'C', 'D', 'E', 'F',
        'l', 'L', 'x', 'X'
    };

    char[] literalFLOAT = {
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        'd', 'D', 'e', 'E', 'f', 'F', '+', '-', '.'
    };

    char[] ident1 = {
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
        'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
        'u', 'v', 'w', 'x', 'y', 'z',
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
        'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
        'U', 'V', 'W', 'X', 'Y', 'Z',
        '_', '$'
    };
    char[] idnet2 = {
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
        'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
        'u', 'v', 'w', 'x', 'y', 'z',
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
        'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
        'U', 'V', 'W', 'X', 'Y', 'Z',
        '_', '$',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
    };

    public static void main(String[] args) throws Exception {
        Stopwatch timer = new Stopwatch().start();
        
        targetFile = new String(args[0]);
        InputSource is = new InputSource(new FileInputStream(args[0]));
        String uri = new File(args[0]).toURI().toString();
        is.setSystemId(uri);
        SAXParserFactory spf = SAXParserFactory.newInstance();
        spf.setValidating(true);
        SAXParser sp = spf.newSAXParser();
        sp.parse(is, new JXTester());
        
        timer.printTimeProf("JXTester");
        return;
    }

    public void startDocument() {
        test = true;
        elements = new Stack();
    }

    public void startElement(String namespaceURI, String localName,
                             String qName, Attributes atts) {
        startloc = new LocatorImpl(locator);
        elements.push(qName);
        sb = new StringBuffer();
        if (qName.equals("literal")) {
            sort = atts.getValue("sort");
        }
    }

    public void endElement(String namespaceURI, String localName,
                           String qName) {
        String str = sb.toString();
                
        if (qName.equals("ident")) {
            identCheck(str);
        } else if (qName.equals("literal")) {
            literalCheck(str);
        } else if (qName.equals("comment")) {
            commentCheck(str);
        } else if (qName.equals("kw")) {
            strCheck(str, kw, "isn't a keyword.");
        } else if (qName.equals("op")) {
            strCheck(str, op, "is neither an opearter nor a separator.");
        } else if (qName.equals("sp")) {
            charCheck(str, sp);
        } else if (qName.equals("nl")) {
            if (newline.equals("unknown")) {
                if (strCheck(str, nl, "must match newline.")) {
                    newline = str;
                }
            } else {
                strCheck(str, newline, "must match newline.");
            }
        }
        elements.pop();
    }

    public void characters(char[] ch, int start, int length) {
        sb.append(ch, start, length);
    }

    void charCheck(String str, char[] c) {
        boolean     flag = false;
        char[]      ca = str.toCharArray();

        for (int i = 0; i < ca.length; i++) {
            for (int j = 0; j < c.length; j++) {
                if (ca[i] == c[j]) {
                    flag = true;
                    break;
                } else {
                    flag = false;
                }
            }
            if (!flag) {
                contentsError("must not contents '" + String.valueOf(ca[i]) +
                              "'.", str);
                flag = true;
            }
        }
        if (ca.length == 0) {
            contentsError("is empty.", str);
        }
    }

    boolean strCheck(String str, String str2, String em) {
        if (!str.equals(str2)) {
            contentsError(em, str);
            return false;
        }
        return true;
    }

    boolean strCheck(String str, String[] list, String em) {
        boolean     flag = false;

        for (int i = 0; i < list.length; i++) {
            if (str.equals(list[i])) {
                flag = true;
                break;
            }
        }
        if (!flag) {
            contentsError(em, str);
            return false;
        }
        return true;
    }

    boolean strCheck2(String str, String str2, String em) {
        if (str.equals(str2)) {
            contentsError(em, str);
            return false;
        }
        return true;
    }

    boolean strCheck2(String str, String[] list, String em) {
        boolean     flag = false;
        
        for (int i = 0; i < list.length; i++) {
            if (str.equals(list[i])) {
                flag = true;
                break;
            }
        }
        if (flag) {
            contentsError(em, str);
            return false;
        }
        return true;
    }

    void nlCheck(String str) {
        for (int i = 0; i < nl.length; i++) {
            if (str.indexOf(nl[i]) != -1) {
                contentsError("must not contents \"newline\".", str);
            }
        }
    }

    void commentCheck(String str) {
        int     leng = str.length();
        String  c1 = new String(str.substring(0, 2));
                
        if (c1.equals("//")) {
            nlCheck(str);
        } else if (c1.equals("/*")) {
            String c2 = new String(str.substring(2, leng - 2));
            if (c2.indexOf("*/") != -1) {
                contentsError("must end with \"*/\".", str);
            }
            String c3 = new String(str.substring(leng - 2, leng));
            if (!c3.equals("*/")) {
                contentsError("must end with \"*/\".", str);
            }
        } else {
            contentsError("must start with \"/*\" or \"//\".", str);
        }
    }

    void literalCheck(String str) {
        if (sort.equals("INT")) {
            charCheck(str, literalINT);
        } else if (sort.equals("FLOAT")) {
            charCheck(str, literalFLOAT);
        } else if (sort.equals("STR")) {
            char c1 = str.charAt(0);
            char c2 = str.charAt(str.length() - 1);
            if (!(c1 == '"' && c2 == '"')) {
                contentsError("must match \"...\".", str);
            }
            nlCheck(str);
        } else if (sort.equals("CHAR")) {
            pattern = Pattern.compile("'.'|'\\\\[btnfr\"'\\\\]'|'\\\\[0-7]'|'\\\\[0-7][0-7]'|'\\\\[0-3][0-7][0-7]'|'\\\\u+[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]'");
            matcher = pattern.matcher(str);
            if(!matcher.matches()){
                contentsError("is invalid character.", str);
            }
        } else if (sort.equals("BOOL")) {
            strCheck(str, literalBOOL,
                     "must match BooleanLiteral (true or false).");
        } else if (sort.equals("NULL")) {
            strCheck(str, literalNULL, "must match NullLiteral (null).");
        }
    }
        
    void identCheck(String str) {
        boolean     flag;
        char        ic;
        char[]      ilist;
        
        for (int i = 0; i < str.length(); i++) {
            flag = false;
            ic = str.charAt(i);
            if (i == 0) {
                ilist = ident1;
            } else {
                ilist = idnet2;
            }
            for (int j = 0; j < ilist.length; j++) {
                if (ic == ilist[j]) {
                    flag = true;
                    break;
                }
            }
            if (i == 0 && !flag) {
                contentsError("must start with [a-zA-Z_$].", str);
            } else if (!flag) {
                contentsError("must match [a-zA-Z0-9_$].", str);
            }
        }
        strCheck2(str, kw, "must not match keyword.");
        strCheck2(str, literalBOOL,
                  "must not match BooleanLiteral (true or false).");
        strCheck2(str, literalNULL, "must not match NullLiteral (null).");
    }

    void contentsError(String em, String str) {
        test = false;

        System.err.print("Sapid: JXTester: ");
        System.err.print("\"" + targetFile + "\":");
        System.err.print(" JX-model error");
        System.err.print(" l." + startloc.getLineNumber());
        System.err.print(" c." + startloc.getColumnNumber());
        System.err.println(".");

        System.err.print("Sapid: JXTester: ");
        System.err.print("    Element \"" + elements.peek() + "\" ");
        System.err.print("(" + str + ") "); 
        System.err.println(em);
    }

    public void error(SAXParseException ex) throws SAXException {
        test = false;

        System.err.print("Sapid: JXTester: ");
        System.err.print("\"" + targetFile + "\":");
        System.err.print(" Validation error");
        System.err.print(" l." + Integer.toString(ex.getLineNumber()));
        System.err.print(" c." + Integer.toString(ex.getColumnNumber()));
        System.err.println(".");

        System.err.print("Sapid: JXTester: ");
        System.err.println("    " + ex.getMessage());
    }

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }
}
