/**
 * JX1.3 to pSPIE-XML
 * This is a JX-model 1.3 sample application.
 * 
 * Copyright (C) 2003 T.Ohno & H.Yoshida
 */

package JXModel3.pspie;

import JXModel3.xsdml.*;
import java.io.*;

public class MakePSpieXml {
    
    static FileOutputStream fos;
    static PrintWriter pw;
    static boolean check_source = true;
    static boolean check_nl = false;
    static String check_sfl = "";
    static boolean check_comment = false;
    static boolean check_head = false;
    static boolean check_body = false;
    static SpdElementList commentlist;
    static int linenum = 1;

    
    public static void main(String args[]) 
    {
    	if (args.length<2) {
    		System.err.println("usage: [in:JX1.3-xml] [out:pSPIE-xml]");
    		System.exit(0);
    	}
    	
		Repository r = new Repository();
		SpdDocument d = r.loadXSDML(args[0], false);
		SpdElement file = d.getFileElement();
		
		//PrintWriter
	    try {
		    fos = new FileOutputStream(args[1]);
		    pw = new PrintWriter(fos);
		} catch (FileNotFoundException e) {
			System.err.println("FileNotFound.");
			System.exit(0);
		}
		    
	    //initialize
	    pw.print("<?xml version='1.0'?>");
	    pw.print("<!DOCTYPE File SYSTEM \"pSPIE.dtd\">");
	    pw.print("<File>");
		    
	    SpdElementList filechildlist = file.getChildren();
	    filechildlist.invokeForEach(new ChildAllHandler());
		    
	    //treat last source
	    if (!check_source) {
			pw.print("</source>");
			check_source = true;
	    }
	
	    if (check_body) {
			pw.print("</Body>");
			check_body = false;
	    }    
		    
	    pw.print("</File>");
	    //FileWriter close
	    pw.close();
    }
    
	static class ChildAllHandler implements SpdElementHandler {
		public void action(SpdElement e)
		{
//			System.err.println("processing.. "+e.getName()+" in l."+e.getLine());

			//kw 
			if (e.getName().equals("kw")) {
			    check_nl = false;
				
			    //treat source
			    if (!check_source) {
				pw.print("</source>");
				check_source = true;
			    }
	
			    if (!check_body & !check_head) {
				linenum = e.getLine();
				pw.print("<Body line=\""+linenum+"\">");
				
				check_body = true;
			    }
			    		    
			    pw.print("<kw>");
			    pw.print(escapeString(e.getText()));
			    pw.print("</kw>");
			}
			//comment 
			else if (e.getName().equals("comment")) {
			    check_nl = false;
			    
			    //treat source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
	
			    check_comment = checkCommentNextElement(e);
			    
			    if (check_comment) {
					commentlist = new SpdElementList();
					commentlist.add(e);
					commentlist = getCommentNextList(e, commentlist);
			    }
			    else {
					if (check_body) {
					    pw.print("</Body>");
					    check_body = false;
					}
		
					linenum = e.getLine();
					pw.print("<comment line=\""+linenum+"\">");
					pw.print(escapeString(e.getText()));
					pw.print("</comment>");
			    }
			}
			//nl
			else if (e.getName().equals("nl")) {
			    if (!check_comment) {
					check_nl = true;
				    if (!checkPreNl(e) && !checkNextNl(e)) {
						//treat source
						if (!check_source) {
						    pw.print("</source>");
						    check_source = true;
						}
						if (!check_body & !check_head) {
							linenum = e.getLine(); 
						    pw.print("<Body line=\""+linenum+"\">");
						    check_body = true;
						}
						else {
							linenum = e.getLine()+1; 
							pw.print("\n<linenum line=\""+linenum+"\""+"/>");
						}
				    }
			    }
			}
			//sp
			else if (e.getName().equals("sp")) {
			    if(!check_comment) {
					if (check_nl == false) {
					    if (!check_body & !check_head) {
							linenum = e.getLine();
							pw.print("<Body line=\""+linenum+"\">");
							check_body = true;
					    }
					    
					    if (check_source) {
							pw.print("<source>");
							check_source = false;
					    }
					    pw.print(escapeString(e.getText()));
					}
					
					check_nl = false;
			    }
			}
			//ident literal 
			else if (e.getName().equals("ident")
				 |e.getName().equals("literal")) {

			    if (!check_body & !check_head) {
					linenum = e.getLine();
					pw.print("<Body line=\""+linenum+"\">");
					check_body = true;
			    }
	
			    check_nl = false;
			    
			    if (check_source) {
					pw.print("<source>");
					check_source = false;
			    }
			    pw.print(escapeString(e.getText()));
			}
			//op
			else if (e.getName().equals("op")) {
			    if (!check_body & !check_head) {
					linenum = e.getLine();
					pw.print("<Body line=\""+linenum+"\">");
					check_body = true;
			    }
	
			    check_nl = false;
			    
			    if (check_source) {
					pw.print("<source>");
					check_source = false;
			    }
			    
			    pw.print(escapeString(e.getText()));
	
			    if (!check_sfl.equals("")) {
					pw.print("</source>");
					pw.print("</"+check_sfl+">");
					check_body = false;
					check_source = true;
					check_sfl = "";
			    }
			}
	
			// Stmt Members
			else if (e.getName().equals("Stmt")
				 |e.getName().equals("Members")) {
			    //treat source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
			    
			    if (check_head) {
					pw.print("</Head>");
					check_head = false;
					check_body = false;
			    }
				
			    SpdElementList childlist = e.getChildren();
			    childlist.invokeForEach(new ChildAllHandler());
	
			}
			
			// Package Import Type QName Expr Label 
			else if (e.getName().equals("Package")
				 |e.getName().equals("Import")
				 |e.getName().equals("Type")
				 |e.getName().equals("QName")
				 |e.getName().equals("Expr")
				 |e.getName().equals("Label")
				 |e.getName().equals("SInit")
				 |e.getName().equals("Field")
				 |e.getName().equals("Local")) {
			    
			    SpdElementList childlist = e.getChildren();
			    childlist.invokeForEach(new ChildAllHandler());
			}
			
			//Class Intf Method Ctor 
			else if (e.getName().equals("Class")
				 |e.getName().equals("Intf")
				 |e.getName().equals("Method")
				 |e.getName().equals("Ctor")) {
			    //treat source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
	
			    if (check_body) {
					pw.print("</Body>");
					check_body = false;
			    }    
	
			    linenum = e.getLine();
			    pw.print("<"+e.getName()+" line=\""+linenum+"\">");
				
			    if (check_comment) {
					commentlist.invokeForEach(new CommentHandler());
					check_comment = false;
			    }
	
			    if (check_body) {
					pw.print("</Body>");
					check_body = false;
			    }    
			    
			    pw.print("<Head>");
			    check_head = true;
		    
			    SpdElementList childlist = e.getChildren();
			    removeOpKw(childlist);
			    childlist.invokeForEach(new ChildAllHandler());
	
			    //treat last source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
			    
			    if (check_body) {
					pw.print("</Body>");
					check_body = false;
			    }    
			    
			    pw.print("</"+e.getName()+">");
			}
			//Param ExtdOpt ThrowOpt 
			else {
			    //treat source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
	
			    if (check_body) {
					pw.print("</Body>");
					check_body = false;
			    }
			    
			    pw.print("<"+e.getName()+">");
			    
			    check_body = true;
			    SpdElementList childlist = e.getChildren();
			    childlist.invokeForEach(new ChildAllHandler());

			    //treat last source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
			    check_body = false;
			    pw.print("</"+e.getName()+">");
			}
		}
		public void removeOpKw(SpdElementList l)
		{
			for (int i=l.count()-1; i>=0; i--) {
				SpdElement e = l.getElementAt(i);
				String t = e.getText();
				if (t.equals("class")
					||t.equals("interface")
					||t.equals("(")
					||t.equals(")")
					||t.equals("extends")
					||t.equals("implements")
					||t.equals("throws"))
					 l.remove(e);
			}
		}
    }

    static class CommentHandler implements SpdElementHandler 
    {
		public void action(SpdElement e)
		{
			//comment 
			if (e.getName().equals("comment")) {
			    check_nl = false;
				    
			    //treat source
			    if (!check_source) {
					pw.print("</source>");
					check_source = true;
			    }
		
			    if (check_body) {
					pw.print("</Body>");
					check_body = false;
			    }
			    linenum = e.getLine();
			    pw.print("<"+e.getName()+" line=\""+
				     linenum+"\">");
			    pw.print(escapeString(e.getText()));
			    pw.print("</"+e.getName()+">");
			}
	    }	
    }

	//Markup change (<|>|&) to (&lt;|&gt;|&amp;)
	static String escapeString(String s) 
	{
		String res;
	
		res = s.replaceAll("<","&lt;");
		res = res.replaceAll(">","&gt;");
		res = res.replaceAll("&","&amp;");
	
		return res;
	}

	static boolean checkPreNl(SpdElement e) 
	{
		SpdElement ne = e.getPreceding();
		if (ne==null) return true;
		
		if (ne.getName().equals("sp")
			||ne.getName().equals("nl")) {
			return checkPreNl(ne);
		}				
		else if (ne.getName().equals("Class")
			||ne.getName().equals("Ctor")
			||ne.getName().equals("comment")
			||ne.getName().equals("Intf")
			||ne.getName().equals("Method")) {
			return true;
		}
		return false;
	}

	static boolean checkNextNl(SpdElement e) 
	{
		SpdElement ne = e.getFollowing();
		if (ne==null) return true;
		
		if (ne.getName().equals("sp")
			||ne.getName().equals("nl")) {
			return checkNextNl(ne);
		}				
		else if (ne.getName().equals("Class")
			||ne.getName().equals("Ctor")
			||ne.getName().equals("comment")
			||ne.getName().equals("Intf")
			||ne.getName().equals("Method")) {
			return true;
		}
		return false;
	}

    
	//check wheater or not the comment's next node  Class, Ctor, Intf, Method
	static boolean checkCommentNextElement(SpdElement e) 
	{
		boolean check = false;
		SpdElement ne = e.getFollowing();
		
		if (ne!= null) {
			if (ne.getName().equals("nl")
				||ne.getName().equals("sp")
				||ne.getName().equals("Members")) {
				check = checkCommentNextElement(ne);
			}
		
			else if (ne.getName().equals("Class")
				||ne.getName().equals("Ctor")
				||ne.getName().equals("Intf")
				||ne.getName().equals("Method")) {
				check = true;
			}
	
			else {
				check = false;
			}
		}
		return check;
	}

	// list of (comment ---> Class or Ctor or Intf or Method)
	static SpdElementList getCommentNextList (SpdElement e, SpdElementList elmlist) 
	{
		SpdElement ne = e.getFollowing();
	
		if (ne!= null) {
			if (ne.getName().equals("nl")||ne.getName().equals("sp")) {
				elmlist.add(ne);
				elmlist = getCommentNextList(ne, elmlist);
			}
		}
		
		return elmlist;
	}

}

