<?xml version="1.0"?>

<xsl:stylesheet version="1.0"
 xmlns:xsl="http://www.w3.org/1999/XSL/Transform">

<xsl:output method="html"/>

<xsl:template match="/">
<html>
  <head>
	<title>pSPIE</title>
	<link rel="stylesheet" type="text/css" href="pSPIE.css"/>
  </head>
  <body bgcolor="#FFFFFF">
    <h3>Publishing-oriented SPIE</h3>
    <hr/>
    <xsl:apply-templates select="File"/>
  </body>
</html>
</xsl:template>

<xsl:template match="File">
  <xsl:apply-templates select="Body|Class|Intf|comment"/>
</xsl:template>

<xsl:template match="Class">
  <table>
    <tr>
      <td class="cbody">
        <xsl:value-of select="@line"/>:
        <b>[CLASS] </b>
        <xsl:apply-templates select="Head"/>
        <xsl:apply-templates select="Class|Intf|Ctor|Method|SInit|Field|Body|comment"/>
      </td>
    </tr>
  </table>
</xsl:template> 

<xsl:template match="Intf">
  <table>
    <tr>
      <td class="ibody">
        <xsl:value-of select="@line"/>:
        <b>[INTERFACE] </b>
        <xsl:apply-templates select="Head"/>
        <xsl:apply-templates select="Class|Intf|Ctor|Method|SInit|Field|Body|comment"/>
      </td>
    </tr>
  </table>
</xsl:template>

<xsl:template match="Method">
  <table> 
    <tr>
      <td class="mbody">
        <xsl:value-of select="@line"/>:
        <b>[METHOD] </b>
        <xsl:apply-templates select="Head"/>
        <xsl:apply-templates select="Body|Local|Class|Intf|comment"/>
      </td>
    </tr>
  </table>
</xsl:template>

<xsl:template match="Ctor">
  <table>
    <tr>
      <td class="ctbody">
        <xsl:value-of select="@line"/>:
        <b>[CONSTRUCTOR] </b>
        <xsl:apply-templates select="Head"/>
        <xsl:apply-templates select="Body|Local|Class|Intf|comment"/>
      </td>
    </tr>
  </table>
</xsl:template>

<xsl:template match="Head">
  <xsl:apply-templates select="source|kw|linenum"/>
  <xsl:apply-templates select="ThrwOpt|ImplOpt|ExtdOpt|Param"/>
</xsl:template>    	


<xsl:template match="Param">
  <span class="param">
  <b>[PARAMETER] </b>
  <xsl:apply-templates select="source|comment|kw|linenum"/></span>
</xsl:template>    	

<xsl:template match="ImplOpt">
  <span class="option">
  <b>[IMPLEMENTS] </b>
  <xsl:apply-templates select="source|comment|linenum"/></span>
</xsl:template>    	

<xsl:template match="ExtdOpt">
  <span class="option">
  <b>[EXTENDS] </b>
  <xsl:apply-templates select="source|comment|linenum"/></span>
</xsl:template>    	

<xsl:template match="ThrowOpt">
  <span class="option">
  <b>[THROWS] </b>
  <xsl:apply-templates select="source|comment|linenum"/></span>
</xsl:template>    	

<xsl:template match="Body">
  <table>
    <tr>
      <td class="body">
        <xsl:value-of select="@line"/>:
        <xsl:apply-templates select="source|kw|linenum"/>
      </td>
    </tr>
  </table>
</xsl:template>    	

<xsl:template match="source">
  <xsl:value-of disable-output-escaping="yes" select="."/>
</xsl:template>    	

<xsl:template match="kw">
  <font color="red">
    <xsl:value-of select="."/>
  </font>
</xsl:template>    	

<xsl:template  match="comment">
  <table class="comment">
    <tr>
      <td>
        <xsl:value-of select="."/>
      </td>
    </tr>
  </table>
</xsl:template>    	

<xsl:template match="linenum">
  <br/><xsl:value-of select="@line"/>:
</xsl:template>    	

<xsl:template match="@line">
  <span class="lineno">
  <xsl:value-of select="."/>:
  </span>
</xsl:template>

</xsl:stylesheet>    
