/* 
 * Program:             $RCSfile: Compactor.java,v $  $Revision: 50.1 $
 * 
 * Purpose:             
 *
 * Author:              T.Takahashi  2003/06/01
 *                      S.Yamamoto  2003/06/01
 *
 * (C) Copyright:       T.Takahashi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Compactor.java,v 50.1 2003/08/10 04:25:10 yamamoto Exp yamamoto $
 */

package JXModel3.renumber;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;

public class Compactor extends JXModelOperator
{
    public int n;
    public String idPrefix;

    public Compactor()
    {
        super();
        n = 0;
        idPrefix = "s";
    }
    public void compactId()
    {
        Node node = (Node)document.getDocumentElement();
        compactId(node);
    }
    public void compactId(Node node)
    {
        NamedNodeMap attributes = node.getAttributes();
        Node id = attributes.getNamedItem("id");

        if (id != null) {
            String hashStr = id.getNodeValue();
            myPrintln(node.getNodeName() + " " + id.getNodeName() + " " + hashStr);
            recHt.put(id.getNodeValue(), idPrefix + Integer.toString(n));
            id.setNodeValue(idPrefix + Integer.toString(n++));
            myPrintln(id.getNodeName() + " " + id.getNodeValue());
            Vector vector = (Vector)suspHt.get(hashStr);
            if (vector != null) {
                Node refNode = (Node)vector.remove(0);
                while (true) {
                    myPrintln("  " + refNode.getNodeName() + " " + refNode.getNodeValue());
                    refNode.setNodeValue((String)recHt.get(hashStr));
                    myPrintln("  " + refNode.getNodeName() + " " + refNode.getNodeValue());
                    try{
                        refNode = (Node)vector.remove(0);
                    }catch (ArrayIndexOutOfBoundsException m) {
                        break;
                    }
                }
            }
        }
        switch (str2hash(node.getNodeName())) {
        case JX_ImplOpt:
        case JX_ExtdOpt:
        case JX_ThrwOpt:
        case JX_Method:
        case JX_Field:
        case JX_Local:
        case JX_Param:
        case JX_Expr:
            referenceId("typefirst", node);
            break;
        case JX_Type:
            referenceId("typenext", node);
            referenceId("qname", node);
            break;
        case JX_ident:
            referenceId("defid", node);
            break;
        }
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.TEXT_NODE)
                break;
            compactId(childNode);
        }
    }
    public void referenceId(String str, Node node)
    {
        NamedNodeMap attributes = node.getAttributes();
        Node id = attributes.getNamedItem(str);

        if (id != null) {
            myPrintln("  " + id.getNodeName() + " " + id.getNodeValue());
            String hashStr = (String)recHt.get(id.getNodeValue());
            if (hashStr != null)
                id.setNodeValue(hashStr);
            else {
                String refId = id.getNodeValue();
                Vector vector = (Vector)suspHt.get(refId);
                if (vector == null) {
                    vector = new Vector();
                    vector.add(id);
                    suspHt.put(id.getNodeValue(), vector);
                } else {
                    vector.add(id);
                }
            }
            myPrintln("  " + id.getNodeName() + " " + id.getNodeValue());
        }
    }
    public void setPrefix(String str)
    {
        idPrefix = str;
    }
    public void setN(int num)
    {
        n = num;
    }
    public static void main(String[] args)
        throws Exception
    {
        Compactor compactor;
        String filename = "";
        boolean debug = false;

        switch (args.length) {
        case 1:
            filename = args[0];
            break;
        case 2:
            if (args[0].equals("-d")) {
                filename = args[1];
                debug = true;
            } else
                Compactor.errMsg(0);
            break;
        default:
            Compactor.errMsg(0);
            break;
        }
        compactor = new Compactor();
        compactor.generateDOMTree(filename);
        compactor.setDebug(debug);
        compactor.compactId();
        compactor.outputXML();
    }
}
