/* 
 * Program:             $RCSfile: Diff.java,v $  $Revision: 50.3 $
 * 
 * Purpose:             An O(NP) sequence comparison algorithm.
 *
 * Author:              T.Takahashi  2003/06/01
 *                      S.Yamamoto  2003/06/01
 *
 * (C) Copyright:       T.Takahashi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Diff.java,v 50.3 2003/08/10 08:47:09 yamamoto Exp yamamoto $
 */

package JXModel3.renumber;

/**
 * An O(NP) sequence comparison algorithm
 * by S.Wu, E.W.Myers, U.Manber, and W.Miller.
 * 
 * C.f. "http://www.cs.aisona.edu/people/gene/PAPERS/np_diff.ps".
 */

class Diff
{
    Object[] obja;
    Object[] objb;
    int offset;
    int[] fp;
    String[] path;
    boolean debug;
    int check;

    public Diff(Object[] a, Object[] b)
    {
        if (a.length > b.length) {
            obja = b;
            objb = a;
            check = 1;
        } else {
            obja = a;
            objb = b;
            check = 0;
        }
        offset = obja.length + 1;
        path = new String[a.length + b.length + 3];
        fp = new int[a.length + b.length + 3];
        for (int i = 0; i < a.length + b.length + 3; i++) {
            path[i] = new String();
            fp[i] = -1;
        }
        debug = false;
    }
    public void setDebug(boolean bool)
    {
        debug = bool;
    }
    public void myPrintln(String s)
    {
        if (debug)
            System.out.println(s);
    }
    public void myPrintln()
    {
        if (debug)
            System.out.println();
    }
    public boolean compare(Object a, Object b)
    {
        myPrintln((String)a + " " + (String)b);
        return ((String)a).equals((String)b);
    }
    public int snake(int k, int y1, int y2) {
        int i = k + offset;
        int y;
        if (y1 + 1 > y2) {
            y = y1 + 1;
            path[i] = path[i-1];
            path[i] += String.valueOf(1 + check);
        } else {
            y = y2;
            path[i] = path[i+1];
            path[i] += String.valueOf(2 - check);
        }
        int x = y - k;
        myPrintln(" " + k + " " + x + " " + y);
        while (x < obja.length && y < objb.length &&
               compare(obja[x], objb[y])) {
            x++;
            y++;
            path[i] += String.valueOf(0);
        }
        return y;
    }
    public String makeMap()
    {
        int p = 0;
        int d = objb.length - obja.length;
        while (true) {
            for (int k = -p; k < d; k++)
                fp[k + offset] = snake(k, fp[k - 1 + offset],
                                       fp[k + 1 + offset]);
            for (int k = d + p; k > d; k--)
                fp[k + offset] = snake(k, fp[k - 1 + offset],
                                       fp[k + 1 + offset]);
            fp[d + offset] = snake(d, fp[d - 1 + offset],
                                   fp[d + 1 + offset]);
            if (fp[d + offset] == objb.length){
                myPrintln("reach");
                break;
            }
            p += 1;
        }
        return path[d + offset].substring(1);
    }
}
