/* 
 * Program:             $RCSfile: JXModelOperator.java,v $  $Revision: 50.1 $
 * 
 * Purpose:             
 *
 * Author:              T.Takahashi  2003/06/01
 *                      S.Yamamoto  2003/06/01
 *
 * (C) Copyright:       T.Takahashi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/*
 * $Id: JXModelOperator.java,v 50.1 2003/08/10 04:26:07 yamamoto Exp yamamoto $
 */

package JXModel3.renumber;

import java.io.*;
import java.util.Hashtable;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;

public class JXModelOperator
{
    public static final short JX_File      =  0;
    public static final short JX_Package   =  1;
    public static final short JX_Import    =  2;
    public static final short JX_Class     =  3;
    public static final short JX_Intf      =  4;
    public static final short JX_Ctor      =  5;
    public static final short JX_Method    =  6;
    public static final short JX_SInit     =  7;
    public static final short JX_Field     =  8;
    public static final short JX_Param     =  9;
    public static final short JX_Local     = 10;
    public static final short JX_ExtdOpt   = 11;
    public static final short JX_ImplOpt   = 12;
    public static final short JX_ThrwOpt   = 13;
    public static final short JX_Members   = 14;
    public static final short JX_QName     = 15;
    public static final short JX_Type      = 16;
    public static final short JX_Stmt      = 17;
    public static final short JX_Label     = 18;
    public static final short JX_Expr      = 19;
    public static final short JX_ident     = 20;
    public static final short JX_literal   = 21;
    public static final short JX_comment   = 22;
    public static final short JX_kw        = 23;
    public static final short JX_op        = 24;
    public static final short JX_sp        = 25;
    public static final short JX_nl        = 26;

    public static final short Stmt_EMPTY   = 101;
    public static final short Stmt_EXPR    = 102;
    public static final short Stmt_BLOCK   = 103;
    public static final short Stmt_DOWHILE = 104;
    public static final short Stmt_WHILE   = 105;
    public static final short Stmt_FOR     = 106;
    public static final short Stmt_IFELSE  = 107;
    public static final short Stmt_SWITCH  = 108;
    public static final short Stmt_BREAK   = 109;
    public static final short Stmt_CONT    = 110;
    public static final short Stmt_RETURN  = 111;
    public static final short Stmt_SYNC    = 112;
    public static final short Stmt_THROW   = 113;
    public static final short Stmt_TRY     = 114;
    public static final short Stmt_ASSERT  = 115;

    static Hashtable hashtable = new Hashtable();

    protected Hashtable recHt;
    protected Hashtable suspHt;
    protected Document document;
    protected String encoding;

    protected boolean debug;
    protected String outputFilename;
    protected String systemId;

    JXModelOperator()
    {
        recHt = new Hashtable();
        suspHt = new Hashtable();
        debug = false;
    }
    protected void generateDOMTree(String filename)
    {
        outputFilename = filename;
        try {
            document = DocumentBuilderFactory
                .newInstance().newDocumentBuilder()
                .parse(new File(filename));
            systemId = document.getDoctype().getSystemId();
            encoding = getEncoding(filename);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    protected String getEncoding(String filename)
        throws Exception
    {
        BufferedReader wi_buffer =
            new BufferedReader(new FileReader(filename));
        String tmp = wi_buffer.readLine();
        int start = tmp.indexOf("encoding=") + 10;//"encoding=".length
        if (start >= 10){
            int end = start;
            while (tmp.charAt(end) != '"') {
                end++;
            }
            return tmp.substring(start, end);
        } else {
            return null;
        }
    }
    protected void outputXML(Document doc)
    {
        try {
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            transformer.setOutputProperty("encoding", encoding);
            transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, systemId);
            DOMSource source = new DOMSource(doc);
            PrintWriter osw = new PrintWriter(System.out);
            StreamResult result = new StreamResult(osw);
            transformer.transform(source, result);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    protected void outputXML()
    {
        outputXML(document);
    }
    protected void outputXML(String filename)
    {
        try {
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            transformer.setOutputProperty("encoding", encoding);
            transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, systemId);
            DOMSource source = new DOMSource(document);
            File newXML = new File(filename);
            FileOutputStream os = new FileOutputStream(newXML);
            StreamResult result = new StreamResult(os);
            transformer.transform(source, result);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

    }
    // *.xml -> *_id.xml(OutputFilename)
    protected void outputXML(String filename, String id)
    {
        try {
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            transformer.setOutputProperty("encoding", encoding);
            transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, systemId);
            DOMSource source = new DOMSource(document);

            String[] splitName = filename.split("\\.");
            splitName[splitName.length - 2] += "_" + id;
            filename = splitName[0];
            for (int i = 1; i < splitName.length; i++)
                filename += "." + splitName[i];

            File newXML = new File(filename);
            FileOutputStream os = new FileOutputStream(newXML);
            StreamResult result = new StreamResult(os);
            transformer.transform(source, result);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

    }
    public short str2hash(String str)
    {
        return ((Short)hashtable.get(str)).shortValue();
    }
    protected void myPrintln(String str)
    {
        if (debug)
            System.err.println(str);
    }
    protected void myPrintln()
    {
        if (debug)
            System.err.println();
    }
    protected void setDebug(boolean bool)
    {
        debug = bool;
    }
    protected void setOutputFilename(String filename)
    {
        outputFilename = filename;
    }
    protected static void errMsg(int i)
    {
        switch(i) {
        case 0:
            System.err.println("Usage: command [-d] Filename");
        }
        System.exit(1);
    }
    static {
        hashtable.put("File"   , new Short(JX_File));
        hashtable.put("Package", new Short(JX_Package));
        hashtable.put("Import" , new Short(JX_Import));
        hashtable.put("Class"  , new Short(JX_Class));
        hashtable.put("Intf"   , new Short(JX_Intf));
        hashtable.put("Ctor"   , new Short(JX_Ctor));
        hashtable.put("Method" , new Short(JX_Method));
        hashtable.put("SInit"  , new Short(JX_SInit));
        hashtable.put("Field"  , new Short(JX_Field));
        hashtable.put("Param"  , new Short(JX_Param));
        hashtable.put("Local"  , new Short(JX_Local));
        hashtable.put("ExtdOpt", new Short(JX_ExtdOpt));
        hashtable.put("ImplOpt", new Short(JX_ImplOpt));
        hashtable.put("ThrwOpt", new Short(JX_ThrwOpt));
        hashtable.put("Members", new Short(JX_Members));
        hashtable.put("QName"  , new Short(JX_QName));
        hashtable.put("Type"   , new Short(JX_Type));
        hashtable.put("Stmt"   , new Short(JX_Stmt));
        hashtable.put("Label"  , new Short(JX_Label));
        hashtable.put("Expr"   , new Short(JX_Expr));
        hashtable.put("ident"  , new Short(JX_ident));
        hashtable.put("literal", new Short(JX_literal));
        hashtable.put("comment", new Short(JX_comment));
        hashtable.put("kw"     , new Short(JX_kw));
        hashtable.put("op"     , new Short(JX_op));
        hashtable.put("sp"     , new Short(JX_sp));
        hashtable.put("nl"     , new Short(JX_nl));

        hashtable.put("EMPTY"  , new Short(Stmt_EMPTY));
        hashtable.put("EXPR"   , new Short(Stmt_EXPR));
        hashtable.put("BLOCK"  , new Short(Stmt_BLOCK));
        hashtable.put("DOWHILE", new Short(Stmt_DOWHILE));
        hashtable.put("WHILE"  , new Short(Stmt_WHILE));
        hashtable.put("FOR"    , new Short(Stmt_FOR));
        hashtable.put("IFELSE" , new Short(Stmt_IFELSE));
        hashtable.put("SWITCH" , new Short(Stmt_SWITCH));
        hashtable.put("BREAK"  , new Short(Stmt_BREAK));
        hashtable.put("CONT"   , new Short(Stmt_CONT));
        hashtable.put("RETURN" , new Short(Stmt_RETURN));
        hashtable.put("SYNC"   , new Short(Stmt_SYNC));
        hashtable.put("THROW"  , new Short(Stmt_THROW));
        hashtable.put("TRY"    , new Short(Stmt_TRY));
        hashtable.put("ASSERT" , new Short(Stmt_ASSERT));
    }
}
