/* 
 * Program:             $RCSfile: Renumber.java,v $  $Revision: 50.5 $
 * 
 * Purpose:             
 *
 * Author:              T.Takahashi  2003/06/01
 *                      S.Yamamoto  2003/06/01
 *
 * (C) Copyright:       T.Takahashi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/*
 * $Id: Renumber.java,v 50.5 2003/08/10 05:12:25 yamamoto Exp yamamoto $
 */

package JXModel3.renumber;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;

public class Renumber extends JXModelOperator
{
    public int nestLevel;
    public Document newdocument;

    public Vector oldComment;    //String list
    public Vector oldCommentId;  //String list
    public Vector newComment;    //String list
    public Vector newCommentNode;//Node list

    private static final short Stmt_BOOL         = 200;
    private static final short Stmt_SEXPR        = 201;
    private static final short Stmt_ROOTBLOCK    = 202;
    private static final short Stmt_TRYBLOCK     = 203;
    private static final short Stmt_CATCHBLOCK   = 204;
    private static final short Stmt_CATCHLocal   = 205;
    private static final short Stmt_FINALLYBLOCK = 206;

    static {
        hashtable.put("BOOL",          new Short(Stmt_BOOL));
        hashtable.put("SEXPR",         new Short(Stmt_SEXPR));
        hashtable.put("ROOTBLOCK",     new Short(Stmt_ROOTBLOCK));
        hashtable.put("tryBLOCK",      new Short(Stmt_TRYBLOCK));
        hashtable.put("catchBLOCK",    new Short(Stmt_CATCHBLOCK));
        hashtable.put("catchLocal",    new Short(Stmt_CATCHLocal));
        hashtable.put("finallyBLOCK",  new Short(Stmt_FINALLYBLOCK));
    }

    Renumber()
    {
        super();
        oldComment = new Vector();
        oldCommentId = new Vector();
        newComment = new Vector();
        newCommentNode = new Vector();

        nestLevel = 0;
    }
    public void generateDOMTree(String oldFilename, String newFilename)
    {
        outputFilename = newFilename;
        try {
            document = DocumentBuilderFactory
                .newInstance().newDocumentBuilder()
                .parse(new File(oldFilename));

            newdocument = DocumentBuilderFactory
                .newInstance().newDocumentBuilder()
                .parse(new File(newFilename));
            systemId = newdocument.getDoctype().getSystemId();
            encoding = getEncoding(outputFilename);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    void replaceId()
    {
        Node node0 = (Node)document.getDocumentElement();
        Node node1 = (Node)newdocument.getDocumentElement();

        myPrintln("FileId " + getNodeId(node1) + " -> " + getNodeId(node0));
        setNodeId(node1, getNodeId(node0));
        nestLevel++;

        NodeList nodeList = node0.getChildNodes();
        Hashtable classTable = new Hashtable();
        Hashtable packageTable = new Hashtable();
        Hashtable importTable = new Hashtable();
        for(int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = (Node)nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                switch (str2hash(childNode.getNodeName())) {
                case JX_Class:
                case JX_Intf:
                    myPrintln(getIdent(childNode) + " => " + getNodeId(childNode));
                    classTable.put(getIdent(childNode), childNode);
                    break;
                case JX_Package:
                    myPrintln("Entry : package " +
                              flatQName(searchChildByName(childNode, "QName")) +
                              " => " + getNodeId(childNode));
                    packageTable.put(flatQName(searchChildByName(childNode, "QName")),
                                     childNode);
                    break;
                case JX_Import:
                    myPrintln("Entry : import  " +
                              flatQName(searchChildByName(childNode, "QName")) +
                              " => " + getNodeId(childNode));
                    importTable.put(flatQName(searchChildByName(childNode, "QName")),
                                    childNode);
                    break;
                case JX_comment:
                    oldComment.addElement(getFirstChildValue(childNode));
                    oldCommentId.addElement(getNodeId(childNode));
                    break;
                }
            }
        }
        nodeList = node1.getChildNodes();
        for(int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = (Node)nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                switch (str2hash(childNode.getNodeName())) {
                case JX_Class:
                case JX_Intf:
                    searchClass(childNode, classTable);
                    break;
                case JX_Package: {
                    Node tmpNode = (Node)packageTable.
                        get(flatQName(searchChildByName(childNode, "QName")));
                    if (tmpNode != null) {
                        myPrintln("Search: package" + " " +
                                  getNodeId(childNode) + " -> " + getNodeId(tmpNode));
                        replaceQNameId(tmpNode, childNode);
                    }
                    break;
                }
                case JX_Import: {
                    Node tmpNode = (Node)importTable.
                        get(flatQName(searchChildByName(childNode, "QName")));
                    if (tmpNode != null) {
                        myPrintln("Search: import" + " " +
                                  getNodeId(childNode) + " -> " + getNodeId(tmpNode));
                        replaceQNameId(tmpNode, childNode);
                    }
                    break;
                }
                case JX_comment:
                    newComment.addElement(getFirstChildValue(childNode));
                    newCommentNode.addElement(childNode);
                    break;
                }
            }
        }
        nestLevel--;
        replaceCommentId();
    }
    public void searchClass(Node node, Hashtable hashtable)
    {
        Node tmpNode = (Node)hashtable.get(getIdent(node));
        if (tmpNode != null) {
            myPrintln("Search: Class  " + getIdent(node) + " " +
                      getNodeId(node) + " -> " + getNodeId(tmpNode));
            setNodeId(node, getNodeId(tmpNode));
            NodeList childNodeList0 = tmpNode.getChildNodes();
            NodeList childNodeList1 = node.getChildNodes();
            for (int j = 0; j < childNodeList0.getLength(); j++) {
                Node childNode0 = (Node)childNodeList0.item(j);
                Node childNode1 = (Node)childNodeList1.item(j);
                if (childNode0.getNodeType() == Node.ELEMENT_NODE &&
                    childNode1.getNodeType() == Node.ELEMENT_NODE) {
                    String childNodeName0 = childNode0.getNodeName();
                    String childNodeName1 = childNode1.getNodeName();
                    if (childNodeName0.equals(childNodeName1)) {
                        switch (str2hash(childNodeName0)) {
                        case JX_Members:
                            compareMember(childNode0, childNode1);
                            break;
                        case JX_ExtdOpt:
                            replaceFieldId(childNode0, childNode1);
                            break;
                        case JX_ImplOpt:
                            replaceFieldId(childNode0, childNode1);
                            break;
                        case JX_ident:
                            referenceId("defid", childNode1);
                            break;
                        }
                    }
                }
            }
        } else {
            recursiveReferenceId(node);
        }
    }
    public Node searchChildByName(Node node, String s)
    {
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals(s))
                    return childNode;
            }
        }
        return null;
    }
    // id ֤ȤȤˡ֤ id Ͽ
    public void setNodeId(Node node, String str)
    {
        String nodeId = getNodeId(node);

        recHt.put(nodeId, str);
        node.getAttributes().getNamedItem("id").setNodeValue(str);

        Vector vector = (Vector)suspHt.get(nodeId);
        if (vector != null) {
            Node refNode = (Node)vector.remove(0);
            while (true) {
                refNode.setNodeValue((String)recHt.get(nodeId));
                try{
                    refNode = (Node)vector.remove(0);
                }catch (ArrayIndexOutOfBoundsException m) {
                    break;
                }
            }
        }
    }
    public Node getAttributeByName(Node node, String s)
    {
        return node.getAttributes().getNamedItem(s);
    }
    public String getFirstChildValue(Node node)
    {
        return node.getChildNodes().item(0).getNodeValue();
    }
    public String getNodeId(Node node)
    {
        return node.getAttributes().getNamedItem("id").getNodeValue();
    }
    public String getIdent(Node node)
    {
        Node childNode = searchChildByName(node, "ident");
        if (childNode == null)
            return null;
        else
            return childNode.getChildNodes().item(0).getNodeValue();
    }
    public void compareMember(Node node0, Node node1)
    {
        NodeList nodeList0 = node0.getChildNodes();
        NodeList nodeList1 = node1.getChildNodes();

        Hashtable htClass = new Hashtable();
        Hashtable htMethod = new Hashtable(); // Constractor Ͽ
        Hashtable htSInit = new Hashtable();
        Hashtable htField = new Hashtable();
        Vector vecSInit = new Vector();
        nestLevel++;

        for (int i = 0; i < nodeList0.getLength(); i++) {
            Node childNode = nodeList0.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                switch (str2hash(childNode.getNodeName())) {
                case JX_Class:
                case JX_Intf:
                    myPrintln("Entry : Class  " + getIdent(childNode) + " => " +
                              getNodeId(childNode));
                    htClass.put(getIdent(childNode), childNode);
                    break;
                case JX_Method:
                case JX_Ctor:
                    // СɤΤᡤ᥽å̾+ꥹȤ򥭡
                    String funcProto = getMethodPrototype(childNode);
                    myPrintln("Entry : " + childNode.getNodeName() + " " + funcProto);
                    htMethod.put(funcProto, childNode);
                    break;
                case JX_Field:
                    myPrintln("Entry : Field  " + getIdent(childNode));
                    htField.put(getIdent(childNode), childNode);
                    break;
                case JX_comment:
                    oldComment.addElement(getFirstChildValue(childNode));
                    oldCommentId.addElement(getNodeId(childNode));
                    break;
                case JX_SInit:
                    vecSInit.addElement(childNode);
                    break;
                }
            }
        }
        for (int i = 0; i < nodeList1.getLength(); i++) {
            Node childNode = nodeList1.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                switch (str2hash(childNode.getNodeName())) {
                case JX_Class:
                case JX_Intf:
                    searchClass(childNode, htClass);
                    break;
                case JX_Ctor: {
                    myPrintln(getMethodPrototype(childNode));
                    Node tmpNode = (Node)htMethod.get(getMethodPrototype(childNode));
                    if (tmpNode != null) {
                        myPrintln("Search: Ctor  " + getMethodPrototype(childNode) + " " +
                                  getNodeId(childNode) + " -> " + getNodeId(tmpNode));
                        replaceMethodId(tmpNode, childNode);
                    } else {
                        recursiveReferenceId((Node)childNode);
                    }
                    break;
                }
                case JX_Method: {
                    myPrintln(getMethodPrototype(childNode));
                    Node tmpNode = (Node)htMethod.get(getMethodPrototype(childNode));
                    if (tmpNode != null) {
                        myPrintln("Search: Method " + getMethodPrototype(childNode) + " " +
                                  getNodeId(childNode) + " -> " + getNodeId(tmpNode));
                        replaceMethodId(tmpNode, childNode);
                    } else {
                        recursiveReferenceId((Node)childNode);
                    }
                    break;
                }
                case JX_Field: {
                    Node gChildNode = searchChildByName(childNode, "ident");
                    Node tmpNode = (Node)htField.get(getFirstChildValue(gChildNode));
                    if (tmpNode != null) {
                        myPrintln("Search: Field  " + getIdent(childNode) + " " +
                                  getNodeId(childNode) + " -> " + getNodeId(tmpNode));
                        referenceId("defid", gChildNode);
                        replaceFieldId(tmpNode, childNode);
                    }
                    break;
                }
                case JX_comment:
                    newComment.addElement(getFirstChildValue(childNode));
                    newCommentNode.addElement(childNode);
                    break;
                case JX_SInit: {
                    Node tmpNode = (Node)vecSInit.remove(0);
                    if (tmpNode != null) {
                        setNodeId(childNode, getNodeId(tmpNode));
                        compareStmt(searchChildByName(tmpNode, "Stmt"),
                                    searchChildByName(childNode, "Stmt"));
                    } else {
                        recursiveReferenceId(childNode);
                    }
                    break;
                }
                }
            }
        }
        nestLevel--;
    }
    public Vector[] flatStmt(Node node, boolean bool, String kw)
    {
        NodeList nodeList = node.getChildNodes();

        Vector[] retValue;
        Vector vecExpr = new Vector();
        Vector vecNode = new Vector();

        if (node.getNodeType() == Node.ELEMENT_NODE) {
            switch (str2hash(node.getNodeName())) {
            case JX_Stmt:
                String sort = node.getAttributes().
                    getNamedItem("sort").getNodeValue();
                switch (str2hash(sort)) {
                case Stmt_EXPR:
                case Stmt_THROW:
                    myPrintln(sort + " " + flatExpr(searchChildByName(node, "Expr")));
                    vecExpr.addElement(sort + " " + flatExpr(searchChildByName(node, "Expr")));
                    vecNode.addElement(node);
                    break;
                case Stmt_BLOCK: {
                    myPrintln(kw + "BLOCK");
                    vecExpr.addElement(kw + "BLOCK");
                    vecNode.addElement(node);
                    NodeList childNodeList = node.getChildNodes();
                    for (int j = 0; j < childNodeList.getLength(); j++) {
                        Node childNode = childNodeList.item(j);
                        if (childNode.getNodeType() == Node.ELEMENT_NODE)
                            switch (str2hash(childNode.getNodeName())) {
                            case JX_Stmt:
                                retValue = flatStmt(childNode, bool, "");
                                vecExpr.addAll(retValue[0]);
                                vecNode.addAll(retValue[1]);
                                break;
                            case JX_Expr:
                                myPrintln("EXPR " + flatExpr(childNode));
                                vecExpr.addElement("EXPR " + flatExpr(childNode));
                                vecNode.addElement(childNode);
                                break;
                            case JX_Local:
                                myPrintln("Local " + getIdent(childNode));
                                vecExpr.addElement("Local " + getIdent(childNode));
                                vecNode.addElement(childNode);
                                break;
                            case JX_Label:
                                String labelName = getFirstChildValue(childNode.
                                                                      getChildNodes().item(0));
                                String labelNum = getLabelValue(childNode);
                                if (labelNum != null) {
                                    labelName += " " + labelNum;
                                }
                                myPrintln("Label " + labelName);
                                vecExpr.addElement("Label " + labelName);
                                vecNode.addElement(childNode);
                                break;
                            case JX_Class:
                            case JX_Intf:
                                String className = getIdent(childNode);
                                myPrintln(childNode.getNodeName() + " " + className);
                                vecExpr.addElement(childNode.getNodeName() + " " + className);
                                vecNode.addElement(childNode);
                                break;
                            case JX_comment:
                                if (bool) {
                                    oldComment.addElement(getFirstChildValue(childNode));
                                    oldCommentId.addElement(getNodeId(childNode));
                                } else {
                                    newComment.addElement(getFirstChildValue(childNode));
                                    newCommentNode.addElement(childNode);
                                }
                                break;
                            }
                    }
                    break;
                }
                case Stmt_DOWHILE:
                case Stmt_WHILE:
                case Stmt_IFELSE: {
                    vecExpr.addElement(sort);
                    vecNode.addElement(node);
                    myPrintln(sort);
                    NodeList childNodeList = node.getChildNodes();
                    for (int j = 0; j < childNodeList.getLength(); j++) {
                        Node childNode = childNodeList.item(j);
                        if (childNode.getNodeType() == Node.ELEMENT_NODE)
                            switch (str2hash(childNode.getNodeName())) {
                            case JX_Stmt:
                                retValue = flatStmt(childNode, bool, "");
                                vecExpr.addAll(retValue[0]);
                                vecNode.addAll(retValue[1]);
                                break;
                            case JX_Expr:
                                myPrintln("BOOL " +  flatExpr(childNode));
                                vecExpr.addElement("BOOL " + flatExpr(childNode));
                                vecNode.addElement(childNode);
                                break;
                            case JX_comment:
                                if (bool) {
                                    oldComment.addElement(getFirstChildValue(childNode));
                                    oldCommentId.addElement(getNodeId(childNode));
                                } else {
                                    newComment.addElement(getFirstChildValue(childNode));
                                    newCommentNode.addElement(childNode);
                                }
                                break;
                            }
                    }
                    break;
                }
                case Stmt_FOR:
                case Stmt_ASSERT: {
                    vecExpr.addElement(sort);
                    vecNode.addElement(node);
                    myPrintln(sort);
                    int l;
                    if (sort.equals("FOR")) {
                        l = 1;
                    } else {
                        l = 0;
                    }
                    NodeList childNodeList = node.getChildNodes();
                    for (int j = 0, k = 0; j < childNodeList.getLength(); j++) {
                        Node childNode = childNodeList.item(j);
                        if (childNode.getNodeType() == Node.ELEMENT_NODE)
                            switch (str2hash(childNode.getNodeName())) {
                            case JX_Stmt:
                                retValue = flatStmt(childNode, bool, "");
                                vecExpr.addAll(retValue[0]);
                                vecNode.addAll(retValue[1]);
                                break;
                            case JX_Expr:
                                if (k == l) {
                                    myPrintln("BOOL " +  flatExpr(childNode));
                                    vecExpr.addElement("BOOL " +
                                                       flatExpr(childNode));
                                    vecNode.addElement(childNode);
                                } else {
                                    myPrintln("EXPR " + flatExpr(childNode));
                                    vecExpr.addElement("EXPR " + flatExpr(childNode));
                                    vecNode.addElement(childNode);
                                }
                                break;
                            case JX_op:
                                if (getFirstChildValue(childNode).equals(";") ||
                                    getFirstChildValue(childNode).equals(":"))
                                    k++;
                                break;
                            case JX_Local:
                                myPrintln("Local " + getIdent(childNode));
                                vecExpr.addElement("Local " + getIdent(childNode));
                                vecNode.addElement(childNode);
                                break;
                            case JX_comment:
                                if (bool) {
                                    oldComment.addElement(getFirstChildValue(childNode));
                                    oldCommentId.addElement(getNodeId(childNode));
                                } else {
                                    newComment.addElement(getFirstChildValue(childNode));
                                    newCommentNode.addElement(childNode);
                                }
                                break;
                            }
                    }
                    break;
                }
                case Stmt_SWITCH:
                case Stmt_SYNC: {
                    myPrintln(sort);
                    vecExpr.addElement(sort);
                    vecNode.addElement(node);
                    NodeList childNodeList = node.getChildNodes();
                    for (int j = 0; j < childNodeList.getLength(); j++) {
                        Node childNode = childNodeList.item(j);
                        if (childNode.getNodeType() == Node.ELEMENT_NODE)
                            switch (str2hash(childNode.getNodeName())) {
                            case JX_Stmt:
                                retValue = flatStmt(childNode, bool, "");
                                vecExpr.addAll(retValue[0]);
                                vecNode.addAll(retValue[1]);
                                break;
                            case JX_Expr:
                                vecExpr.addElement("SEXPR " + flatExpr(childNode));
                                vecNode.addElement(childNode);
                                break;
                            case JX_comment:
                                if (bool) {
                                    oldComment.addElement(getFirstChildValue(childNode));
                                    oldCommentId.addElement(getNodeId(childNode));
                                } else {
                                    newComment.addElement(getFirstChildValue(childNode));
                                    newCommentNode.addElement(childNode);
                                }
                                break;
                            }
                    }
                    break;
                }
                case Stmt_BREAK:
                case Stmt_CONT: {
                    myPrintln(sort);
                    Node childNode = searchChildByName(node, "ident");
                    if (childNode != null)
                        vecExpr.addElement(sort + " " + getFirstChildValue(childNode));
                    else
                        vecExpr.addElement(sort);
                    vecNode.addElement(node);
                    break;
                }
                case Stmt_RETURN: {
                    Node childNode = searchChildByName(node, "Expr");
                    if (childNode != null) {
                        myPrintln("RETURN " + flatExpr(childNode));
                        vecExpr.addElement("RETURN " + flatExpr(childNode));
                    } else {
                        myPrintln("RETURN");
                        vecExpr.addElement("RETURN");
                    }
                    vecNode.addElement(node);
                    break;
                }
                case Stmt_TRY: {
                    myPrintln("TRY");
                    vecExpr.addElement("TRY");
                    vecNode.addElement(node);
                    NodeList childNodeList = node.getChildNodes();
                    String keyword = "";
                    for (int j = 0; j < childNodeList.getLength(); j++) {
                        Node childNode = childNodeList.item(j);
                        if (childNode.getNodeType() == Node.ELEMENT_NODE)
                            switch (str2hash(childNode.getNodeName())) {
                            case JX_Stmt:
                                retValue = flatStmt(childNode, bool, keyword);
                                vecExpr.addAll(retValue[0]);
                                vecNode.addAll(retValue[1]);
                                break;
                            case JX_comment:
                                if (bool) {
                                    oldComment.addElement(getFirstChildValue(childNode));
                                    oldCommentId.addElement(getNodeId(childNode));
                                } else {
                                    newComment.addElement(getFirstChildValue(childNode));
                                    newCommentNode.addElement(childNode);
                                }
                                break;
                            case JX_kw:
                                keyword = getFirstChildValue(childNode);
                                break;
                            case JX_Local:
                                myPrintln(kw + "Local " +
                                          getIdent(searchChildByName(childNode, "Type")) +
                                          " " + getIdent(childNode));
                                vecExpr.addElement(kw + "Local " +
                                                   getIdent(searchChildByName(childNode, "Type")) +
                                                   " " + getIdent(childNode));
                                vecNode.addElement(childNode);
                                break;
                            }
                    }
                    break;
                }
                case Stmt_EMPTY:
                    vecExpr.addElement("EMPTY");
                    vecNode.addElement(node);
                    break;
                }
            }
        }
        Vector[] vecArray = {vecExpr, vecNode};
        return vecArray;
    }
    public void compareStmt(Node node0, Node node1)
    {
        myPrintln("@@@@OldFile@@@@");
        Vector[] retValue0 = flatStmt(node0, true, "ROOT");
        myPrintln("@@@@NewFile@@@@");
        Vector[] retValue1 = flatStmt(node1, false, "ROOT");

        myPrintln("---------------");

        Diff diff = new Diff(retValue0[0].toArray(), retValue1[0].toArray());
        String pass = diff.makeMap();
        Node tmpNode0;
        Node tmpNode1;

        for (int i = 0, p0 = 0, p1 = 0; i < pass.length(); i++) {
            switch (pass.charAt(i)) {
            case '0':
                tmpNode0 = (Node)retValue0[1].get(p0);
                tmpNode1 = (Node)retValue1[1].get(p1);
                nestLevel++;
                myPrintln("@" + retValue0[0].get(p0) + " " + retValue1[0].get(p1));
                String strId[] = ((String)retValue0[0].get(p0)).split(" ", 2);
                switch (str2hash(strId[0])) {
                case Stmt_EXPR:
                case Stmt_BOOL:
                case Stmt_SEXPR:
                case Stmt_THROW:
                    replaceExprId(tmpNode0, tmpNode1);
                    break;
                case Stmt_DOWHILE:
                case Stmt_WHILE:
                case Stmt_IFELSE:
                case Stmt_FOR:
                case Stmt_SWITCH:
                case Stmt_TRY:
                case Stmt_BLOCK:
                case Stmt_ROOTBLOCK:
                case Stmt_TRYBLOCK:
                case Stmt_CATCHBLOCK:
                case Stmt_FINALLYBLOCK:
                case Stmt_SYNC:
                case Stmt_ASSERT:
                    setNodeId(tmpNode1, getNodeId(tmpNode0));
                    break;
                case Stmt_BREAK:
                case Stmt_CONT:
                    setNodeId(tmpNode1, getNodeId(tmpNode0));
                    recursiveReferenceId(tmpNode1); //label?
                    break;
                case Stmt_RETURN:
                    setNodeId(tmpNode1, getNodeId(tmpNode0));
                    if (((String)retValue0[0].get(p0)).length() > 6) {
                        replaceExprId(searchChildByName(tmpNode0, "Expr"),
                                      searchChildByName(tmpNode1, "Expr"));
                    }
                    break;
                case JX_Local:
                case Stmt_CATCHLocal: {
                    replaceFieldId(tmpNode0, tmpNode1);
                    // type
                    referenceId("typefirst", tmpNode1);
                    // ident
                    referenceId("defid", searchChildByName(tmpNode1, "ident"));
                    Node childNode0 = searchChildByName(tmpNode0, "Expr");
                    Node childNode1 = searchChildByName(tmpNode1, "Expr");
                    if (childNode0 != null && childNode1 != null &&
                        flatExpr(childNode0).equals(flatExpr(childNode1))) {
                        replaceExprId(childNode0, childNode1);
                    }
                    break;
                }
                case JX_Label: {
                    setNodeId(tmpNode1, getNodeId(tmpNode0));
                    Node childNode0 = searchChildByName(tmpNode0, "Expr");
                    if (childNode0 != null) {
                        Node childNode1 = searchChildByName(tmpNode1, "Expr");
                        replaceExprId(childNode0, childNode1);
                    }
                    break;
                }
                }
                nestLevel--;
                p0++;
                p1++;
                break;
            case '1':  // insert
                tmpNode1 = (Node)retValue1[1].get(p1);
                recursiveReferenceId(tmpNode1);
                p1++;
                break;
            case '2': // delete
                p0++;
                break;
            }
        }
    }
    public int skipString(Node node, NodeList nodeList, int i)
    {
        while (node.getNodeType() == Node.ELEMENT_NODE &&
               node.getNodeName().equals("kw") ||
               node.getNodeName().equals("nl") ||
               node.getNodeName().equals("sp") ||
               node.getNodeName().equals("op")) {
            node = nodeList.item(++i);
            if (node == null)
                return -1;
        }
        return i;
    }
    public void replaceFieldId(Node node0, Node node1)
    {
        myPrintln(getNodeId(node1) + " -> " + getNodeId(node0));
        setNodeId(node1, getNodeId(node0));
        referenceId("typefirst", node1);

        NodeList nodeList0 = node0.getChildNodes();
        NodeList nodeList1 = node1.getChildNodes();
        boolean identp = true;

        for (int i = 0, j = 0; i < nodeList0.getLength() && j < nodeList1.getLength(); i++, j++) {
            Node childNode0 = nodeList0.item(i);
            Node childNode1 = nodeList1.item(j);
            if (childNode0.getNodeType() == Node.ELEMENT_NODE) {
                i = skipString(childNode0, nodeList0, i);
                j = skipString(childNode1, nodeList1, j);
                if (i == -1 || j == -1) 
                    break;
                childNode0 = nodeList0.item(i);
                childNode1 = nodeList1.item(j);

                myPrintln("--" + childNode0.getNodeName());
                if (childNode0.getNodeName().equals("comment")) {
                    oldComment.addElement(getFirstChildValue(childNode0));
                    oldCommentId.addElement(getNodeId(childNode0));
                    childNode0 = nodeList0.item(++i);
                    if (childNode0 == null)
                        break;
                    i = skipString(childNode0, nodeList0, i);
                    if (i == -1)
                        break;
                    childNode0 = nodeList0.item(i);
                }
                if (childNode1.getNodeName().equals("comment")) {
                    newComment.addElement(getFirstChildValue(childNode1));
                    newCommentNode.addElement(childNode1);
                    childNode1 = nodeList1.item(++j);
                    if (childNode1 == null)
                        break;
                    j = skipString(childNode1, nodeList1, j);
                    if (j == -1)
                        break;
                    childNode1 = nodeList1.item(j);
                }
                if (childNode0.getNodeName().equals("Type")) {
                    if (identp) {
                        if (getFirstChildValue(childNode0.getChildNodes().item(0)).
                            equals(getFirstChildValue(childNode1.getChildNodes().item(0)))) {
                            myPrintln(getNodeId(childNode1) + " -> " + getNodeId(childNode0));
                            setNodeId(childNode1, getNodeId(childNode0));
                            // <Type> reference id
                            referenceId("typenext", childNode1);
                            referenceId("qname", childNode1);
                            referenceId("defid", childNode1.getChildNodes().item(0));
                            identp = false;
                        }
                    } else {// Array Operation
                        setNodeId(childNode1, getNodeId(childNode0));
                        referenceId("typenext", childNode1);
                    }
                } else if (childNode0.getNodeName().equals("ident")) {
                    referenceId("defid", childNode1);
                }
            }
        }
    }
    public void replaceMethodId(Node node0, Node node1)
    {
        replaceFieldId(node0, node1);
        NodeList nodeList0 = node0.getChildNodes();
        NodeList nodeList1 = node1.getChildNodes();
        referenceId("typefirst", node1);

        for (int i = 0, j = 0; i < nodeList0.getLength() && j < nodeList1.getLength(); i++, j++) {
            Node childNode0 = nodeList0.item(i);
            Node childNode1 = nodeList1.item(j);
            if (childNode0.getNodeType() == Node.ELEMENT_NODE) {
                i = skipString(childNode0, nodeList0, i);
                j = skipString(childNode1, nodeList1, j);
                if (i == -1 || j == -1)
                    break;
                childNode0 = nodeList0.item(i);
                childNode1 = nodeList1.item(j);

                // replaceFieldId ϿƤ
                if (childNode0.getNodeName().equals("comment")) {
                    childNode0 = nodeList0.item(++i);
                    if (childNode0 == null)
                        break;
                    i = skipString(childNode0, nodeList0, i);
                    if (i == -1)
                        break;
                    childNode0 = nodeList0.item(i);
                }
                if (childNode1.getNodeName().equals("comment")) {
                    childNode1 = nodeList1.item(++j);
                    if (childNode1 == null)
                        break;
                    j = skipString(childNode1, nodeList1, j);
                    if (j == -1)
                        break;
                    childNode1 = nodeList1.item(j);
                }
                if (childNode0.getNodeName().equals("Param")) {
                    replaceFieldId(childNode0, childNode1);
                    recursiveOldCommentAdd(childNode0);
                    recursiveNewCommentAdd(childNode1);
                } else if (childNode0.getNodeName().equals("Stmt")) {
                    compareStmt(childNode0, childNode1);
                } else if (childNode0.getNodeName().equals("ThrwOpt")) {
                    recursiveOldCommentAdd(childNode0);
                    if (childNode1.getNodeType() == Node.ELEMENT_NODE &&
                        childNode0.getNodeName().equals(childNode1.getNodeName())) {
                        recursiveNewCommentAdd(childNode1);
                        myPrintln(getNodeId(childNode1) + " -> " + getNodeId(childNode0));
                        setNodeId(childNode1, getNodeId(childNode0));
                        // <ThrwOpt> reference id
                        referenceId("typefirst", childNode1);
                        Node typeNode0 = searchChildByName(childNode0, "Type");
                        Node typeNode1 = searchChildByName(childNode1, "Type");
                        if (getFirstChildValue(typeNode0.getChildNodes().item(0)).
                            equals(getFirstChildValue(typeNode1.getChildNodes().item(0)))) {
                            myPrintln(getNodeId(typeNode1) + " -> " + getNodeId(typeNode0));
                            setNodeId(typeNode1, getNodeId(typeNode0));
                            // <Type> reference id
                            referenceId("typenext", typeNode1);
                            referenceId("qname", typeNode1);
                        }
                    }
                }
            }
        }
    }
    public void replaceQNameId(Node node0, Node node1)
    {
        myPrintln(getNodeId(node1) + " -> " + getNodeId(node0));
        setNodeId(node1, getNodeId(node0));
        NodeList nodeList0 = node0.getChildNodes();
        NodeList nodeList1 = node1.getChildNodes();
        for (int i = 0, j = 0; i < nodeList0.getLength() && j < nodeList1.getLength(); i++, j++) {
            Node childNode0 = nodeList0.item(i);
            Node childNode1 = nodeList1.item(j);
            if (childNode0.getNodeType() == Node.ELEMENT_NODE) {
                i = skipString(childNode0, nodeList0, i);
                j = skipString(childNode1, nodeList1, j);
                if (i == -1 || j == -1)
                    break;
                childNode0 = nodeList0.item(i);
                childNode1 = nodeList1.item(j);

                if (childNode0.getNodeName().equals("comment")) {
                    oldComment.addElement(getFirstChildValue(childNode0));
                    oldCommentId.addElement(getNodeId(childNode0));
                    childNode0 = nodeList0.item(++i);
                    if (childNode0 == null)
                        break;
                    i = skipString(childNode0, nodeList0, i);
                    if (i == -1)
                        break;
                    childNode0 = nodeList0.item(i);
                }
                if (childNode1.getNodeName().equals("comment")) {
                    newComment.addElement(getFirstChildValue(childNode1));
                    newCommentNode.addElement(childNode1);
                    childNode1 = nodeList1.item(++j);
                    if (childNode1 == null)
                        break;
                    j = skipString(childNode1, nodeList1, j);
                    if (j == -1)
                        break;
                    childNode1 = nodeList1.item(j);
                }
                if (childNode0.getNodeName().equals("QName")) {
                    nestLevel++;
                    replaceQNameId(childNode0, childNode1);
                    nestLevel--;
                } else if (childNode0.getNodeName().equals("literal")) {
                    myPrintln(getNodeId(childNode1) + " -> " + getNodeId(childNode0));
                    setNodeId(childNode1, getNodeId(childNode0));
                }
            }
        }
    }
    public void replaceExprId(Node node0, Node node1)
    {
        myPrintln(getNodeId(node1) + " -> " + getNodeId(node0));
        setNodeId(node1, getNodeId(node0));
        referenceId("typefirst", node1);

        NodeList nodeList0 = node0.getChildNodes();
        NodeList nodeList1 = node1.getChildNodes();
        for (int i = 0, j = 0; i < nodeList0.getLength() && j < nodeList1.getLength(); i++, j++) {
            Node childNode0 = nodeList0.item(i);
            Node childNode1 = nodeList1.item(j);
            if (childNode0.getNodeType() == Node.ELEMENT_NODE) {
                i = skipString(childNode0, nodeList0, i);
                j = skipString(childNode1, nodeList1, j);
                if (i == -1 || j == -1)
                    break;
                childNode0 = nodeList0.item(i);
                childNode1 = nodeList1.item(j);

                if (childNode0.getNodeName().equals("comment")) {
                    oldComment.addElement(getFirstChildValue(childNode0));
                    oldCommentId.addElement(getNodeId(childNode0));
                    childNode0 = nodeList0.item(++i);
                    if (childNode0 == null)
                        break;
                    i = skipString(childNode0, nodeList0, i);
                    if (i == -1)
                        break;
                    childNode0 = nodeList0.item(i);
                }
                if (childNode1.getNodeName().equals("comment")) {
                    newComment.addElement(getFirstChildValue(childNode1));
                    newCommentNode.addElement(childNode1);
                    childNode1 = nodeList1.item(++j);
                    if (childNode1 == null)
                        break;
                    j = skipString(childNode1, nodeList1, j);
                    if (j == -1)
                        break;
                    childNode1 = nodeList1.item(j);
                }
                if (childNode0.getNodeName().equals("Expr")) {
                    nestLevel++;
                    replaceExprId(childNode0, childNode1);
                    nestLevel--;
                } else if (childNode0.getNodeName().equals("literal")) {
                    String gChildValue = getFirstChildValue(childNode0);
                    if (!gChildValue.equals("null") &&
                        !gChildValue.equals("true") && !gChildValue.equals("false")) {
                        myPrintln(getNodeId(childNode1) + " -> " + getNodeId(childNode0));
                        setNodeId(childNode1, getNodeId(childNode0));
                    }
                } else if (childNode0.getNodeName().equals("Type")) {
                    recursiveOldCommentAdd(childNode0);
                    // InstanceCreation
                    if (getFirstChildValue(childNode0.getChildNodes().item(0)).
                        equals(getFirstChildValue(childNode1.getChildNodes().item(0)))) {
                        myPrintln(getNodeId(childNode1) + " -> " + getNodeId(childNode0));
                        setNodeId(childNode1, getNodeId(childNode0));
                        referenceId("typenext", childNode1);
                        referenceId("qname", childNode1);
                        referenceId("defid", childNode1.getChildNodes().item(0));
                        recursiveNewCommentAdd(childNode1);
                    }
                } else if (childNode0.getNodeName().equals("QName")) {
                    nestLevel++;
                    replaceQNameId(childNode0, childNode1);
                    nestLevel--;
                } else if (childNode1.getNodeName().equals("ident")) {
                    referenceId("defid", childNode1);
                }
            }
        }
    }
    /* (String[])Vector#toArray¹Ի顼ˤʤ뤿ᡥ*/
    public String[] vectorToString(Vector vec)
    {
        String[] str = new String[vec.size()];
        for (int i = 0; i < vec.size(); i++) {
            str[i] = (String)vec.get(i);
            myPrintln(str[i]);
        }
        myPrintln("");
        return str;
    }
    /* QNameʸ֤ */
    public String flatQName(Node node)
    {
        String str ="";
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("QName")) {
                    str += flatQName(childNode);
                } else if (childNode.getNodeName().equals("ident") ||
                           childNode.getNodeName().equals("op")) {
                    str += getFirstChildValue(childNode);
                }
            }
        }
        return str;
    }
    /* ʸ֤ */
    public String flatExpr(Node node)
    {
        String str = "";
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("Expr")) {
                    str += flatExpr(childNode);
                } else if (childNode.getNodeName().equals("ident") ||
                           childNode.getNodeName().equals("literal") ||
                           childNode.getNodeName().equals("op")) {
                    str += childNode.getChildNodes().item(0).getNodeValue();
                } else if (childNode.getNodeName().equals("QName")) {
                    str += flatQName(childNode);
                }
            }
        }
        return str;
    }
    /* ٥ʸ֤ */
    public String getLabelValue(Node node)
    {
        Node childNode = searchChildByName(node, "Expr");
        if (childNode == null) {
            return null;
        } else {
            return getFirstChildValue(childNode.getChildNodes().item(0));
        }
    }
    /* ؿΥץȥʸ֤ */
    public String getMethodPrototype(Node node)
    {
        String methodPrototype = getIdent(node) + "(";
        NodeList nodeList = node.getChildNodes();
        boolean a = true;
                                    
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("Param")) {
                    String strs = getDecl(childNode);
                    if (a) {
                        methodPrototype += strs;
                        a = false;
                    }else{
                        methodPrototype += "," + strs;
                    }
                }
            }
        }
        return methodPrototype + ")";
    }
    /* ʸ֤ */
    public String getDecl(Node node)
    {
        String str = "";
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("Type")) {
                    str += getFirstChildValue(childNode.getChildNodes().item(0));
                }else if (childNode.getNodeName().equals("ident")) {
                    str += " " + getFirstChildValue(childNode);
                }
            }
        }
        return str;
    }
    public void referenceId(String str, Node node)
    {
        Node id = getAttributeByName(node, str);
                                    
        if (id != null) {
            String hashStr = (String)recHt.get(id.getNodeValue());
            if (hashStr != null)
                id.setNodeValue(hashStr);
            else {
                String refId = id.getNodeValue();
                Vector vector = (Vector)suspHt.get(refId);
                if (vector == null) {
                    vector = new Vector();
                    vector.addElement(id);
                    suspHt.put(id.getNodeValue(), vector);
                } else {
                    vector.addElement(id);
                }
            }
        }
    }
    public void recursiveReferenceId(Node node)
    {
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("ident")) {
                    referenceId("defid", childNode);
                } else {
                    recursiveReferenceId(childNode);
                }
            }
        }
    }
    public void recursiveOldCommentAdd(Node node)
    {
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("comment")) {
                    oldComment.addElement(getFirstChildValue(childNode));
                    oldCommentId.addElement(getNodeId(childNode));
                } else {
                    recursiveOldCommentAdd(childNode);
                }
            }
        }
    }
    public void recursiveNewCommentAdd(Node node)
    {
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                if (childNode.getNodeName().equals("comment")) {
                    newComment.addElement(getFirstChildValue(childNode));
                    newCommentNode.addElement(childNode);
                } else {
                    recursiveNewCommentAdd(childNode);
                }
            }
        }
    }
    public void replaceCommentId()
    {
        myPrintln("--- comment id ---");
        Diff diff = new Diff(oldComment.toArray(), newComment.toArray());
        String pass = diff.makeMap();
        for (int i = 0, p0 = 0, p1 = 0; i < pass.length(); i++) {
            switch (pass.charAt(i)) {
            case '0':
                myPrintln(getFirstChildValue((Node)newCommentNode.get(p1)));
                myPrintln(getAttributeByName((Node)newCommentNode.get(p1), "id").
                          getNodeValue() + " -> " + (String)oldCommentId.get(p0));
                getAttributeByName((Node)newCommentNode.get(p1), "id").
                    setNodeValue((String)oldCommentId.get(p0));
                p0++;
                p1++;
                break;
            case '1':
                p1++;
                break;
            case '2':
                p0++;
                break;
            default:
                break;
            }
        }
    }
    // indent
    protected void myPrintln(String str)
    {
        if (debug) {
            for (int i = 0; i < nestLevel; i++)
                System.err.print("  ");
            System.err.println(str);
        }
    }
    public static void main(String[] args)
    {
        String filename0 = "";
        String filename1 = "";
        boolean debug = false;
        boolean out = false;
        int offset = 0;

        switch (args.length) {
        case 6:
            break;
        case 7:
            if (args[0].equals("-d"))
                debug = true;
            else if (args[0].equals("-m"))
                out = true;
            else
                Renumber.errMsg(0);
            offset = 1;
            break;
        case 8:
            debug = true;
            out = true;
            offset = 2;
            break;
        default:
            Renumber.errMsg(0);
            break;
        }
        try {
            Renumber renumber = new Renumber();
            Compactor compactor0 = new Compactor();
            Compactor compactor1 = new Compactor();

            compactor0.generateDOMTree(args[4 + offset]);
            compactor0.setDebug(debug);
            compactor0.setPrefix(args[1 + offset]);
            compactor0.compactId();
            if (out)
                compactor0.outputXML(args[4 + offset], "com");

            renumber.document = compactor0.document;


            compactor1.generateDOMTree(args[5 + offset]);
            compactor1.setDebug(debug);
            compactor1.setPrefix(args[3 + offset]);
            compactor1.compactId();
            if (out)
                compactor1.outputXML(args[5 + offset], "com");

            renumber.newdocument = compactor1.document;
            renumber.systemId = compactor1.systemId;
            renumber.encoding = compactor1.encoding;

            renumber.setDebug(debug);
            renumber.replaceId();
            renumber.outputXML(renumber.newdocument);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
}
