/* 
 * Program:             $RCSfile: xbd3.java,v $  $Revision: 40.7 $
 * 
 * Purpose:             XML builder.
 *
 * Author:              Y.Hiramatsu  2003/02/02
 *                      S.Yamamoto  2003/02/02
 *
 * (C) Copyright:       Y.Hiramatsu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: xbd3.java,v 40.7 2003/04/15 10:12:18 yamamoto Exp yamamoto $
 */

package JXModel3.xbd3;

import java.io.*;
import java.util.*;
import java.text.MessageFormat;
import java.text.ParseException;
import gnu.getopt.*;

interface Cmp {
    public boolean cmp(Object o1, Object o2);
}

interface sorting {
    void swap(Vector v, int i, int j);
    
    void sort(Vector v, int left, int right, Cmp cmp);
    void merge(Vector vec1, Vector vec2, Vector vec3, Cmp cmp);
}

public class xbd3 {
    class XMLNODE {
        String   name;
        String   attrs;
        long     length;  // (bitWidthOf(long) - BIT_WIDTH_WEIGHT) bit.
        short    weight;  // BIT_WIDTH_WEIGHT (define as 8) bit.    
    }
    
    class INSTYPE {
        final int T_END = 1;
        final int T_EMPTY = 2;
        final int T_START = 3;
    }
    
    class INSPTR implements Cmp {
        long      offset;
        int       type;
        XMLNODE   lpnode;
        
        // Desc: ֤դ륯饹
        public boolean cmp(Object o1, Object o2) {
            INSPTR r1 = (INSPTR)o1;
            INSPTR r2 = (INSPTR)o2;
            if (r1.offset == r2.offset) {
                if (r1.type == r2.type) {
                    long l1 = r1.lpnode.length;
                    long l2 = r2.lpnode.length;
                    short w1 = r1.lpnode.weight;
                    short w2 = r2.lpnode.weight;
                    
                    if (r1.type == T_START) {
                        if (l1 == l2) return (w1 > w2);
                        return (l1 > l2);
                    } else if (r1.type == T_END) {
                        if (l1 == l2) return (w1 < w2);
                        return (l1 < l2);
                    } else if (r1.type == T_EMPTY) {
                        return (w1 < w2);
                    }
                    // assert r1.type != r2.type;
                }
                return (r1.type < r2.type);
            }
            return (r1.offset < r2.offset);
        }
    }
    
    class Sorting implements sorting {
        public void swap(Vector v, int i, int j) {
            INSPTR tmp = (INSPTR)v.get(i);
            v.set(i, (INSPTR)v.get(j));
            v.set(j, tmp);
        }

        public void sort(Vector v, int left, int right, Cmp cmp) {
            if (left >= right)
                return;
            swap(v, left, (left + right) / 2);
            int last = left;
            for(int i = left + 1; i <= right; i++)
                if(cmp.cmp((INSPTR)v.get(i), (INSPTR)v.get(left))
                   == true)
                    swap(v, ++last, i);
            swap(v, left, last);
            sort(v, left, last - 1, cmp);
            sort(v, last + 1, right, cmp);
        }
        
        public void merge(Vector vec1, Vector vec2, Vector vec3, Cmp cmp) {
            int len1 = vec1.size();
            int len2 = vec2.size();
            int i1 = 0, i2 = 0;
            boolean active = true;
            while(active){
                if(cmp.cmp((INSPTR)vec1.get(i1),
                            (INSPTR)vec2.get(i2)) == true){
                    vec3.add((INSPTR)vec1.get(i1));
                    i1++;
                    if(i1 >= len1) {
                        for(int i = i2; i < len2; i++) {
                            vec3.add((INSPTR)vec2.get(i));
                        }
                        active = false;
                    }
                }
                else{
                    vec3.add((INSPTR)vec2.get(i2));
                    i2++;
                    if(i2 >= len2) {
                        for(int i = i1; i < len1; i++) {
                            vec3.add((INSPTR)vec1.get(i));
                        }
                        active = false;
                    }
                }
            }
        }        
    }
    
    // Output stream.
    private PrintStream _ps;

    // Source program buffer.
    private byte[] _src;
    static String     _textnode_name = "T";
    
    private static String _fileid;
    private static String program_name;
    private static boolean _opt_no_textnode = false;
    private static boolean _opt_entityref1 = false;
    private static boolean _opt_entityref2 = false;
    private static boolean _opt_linenumout = false;
    private static boolean _is_out_to_file = false;
    private static boolean _is_verbose = false;
    private static boolean _is_Verbose = false;
    
    private static String _src_path;
    private static String _meta_path;
    private static int      _linecount = 0;
    private static final int T_END = 1;
    private static final int T_EMPTY = 2;
    private static final int T_START = 3;
    private static final int EOF = -1;
    // ask the number of FILENAME_MAX
    private static final int FILENAME_MAX = 255;
    
    // Desc: element tableɤƳǼɲä롥

    private boolean load_metainfo_file(BufferedReader fp, PrintWriter fpout,
                                       ArrayList node_list, Vector ins_ptr_list)
    {
        int line = 0;

        if (_is_verbose) {
            System.out.println("Loading element table.");
        }

        while (true) {
            final int   BSIZE = FILENAME_MAX + 1;
            String      buf = new String();

            // Read a line.
            try {
                if((buf = fp.readLine()) == null) {
                    break;
                }
                line++;
                if(buf.lastIndexOf('#', 1) > 0) {
                    // Skip comment
                    continue;
                } else if (buf.lastIndexOf('!', 1) == 0) {
                    // XML string.
                    fpout.println(buf.substring(1));
                    continue;
                }
            } catch (Exception e) {
                System.out.println("Error" + e);
            }

            // Scan a line.
            String name = new String();
            String attrs = new String();
            String result[] = new String[5];
            int     offset = 0, length = 0, col = 0;
            short weight = 0;
            
            StringTokenizer st1 = new StringTokenizer(buf, "\t");
            
            while(st1.hasMoreTokens() && col != 4) {
                result[col] = new String(st1.nextToken());
                col++;
            }
            offset = Integer.parseInt(result[0]);
            length = Integer.parseInt(result[1]);
            Integer it = new Integer(result[2]);
            weight = it.shortValue();
            

            StringTokenizer st2 = new StringTokenizer(result[3]);

            int k = 0;
            while (st2.hasMoreTokens()) {
                if(k == 0) {
                    name = st2.nextToken();
                } else if (k == 1){
                    attrs = " " + st2.nextToken();
                } else {
                    attrs += st2.nextToken(",");
                }
                k++;
            }
            if (_is_Verbose) {
                System.out.println("line" + line + ":" + offset + "," + length + "," + weight + ",\"" + name + "\"");
            }
            if (col < 4 || length < 0) {
                System.err.println(offset + "," + length + "," + weight + ",\"" + name + "\"," + ": Invalid format in l." + line + ".");
                continue;
            }

            // Add node.
            XMLNODE     node = new XMLNODE();
            
            node.name = new String(name);
            node.length = length;
            node.weight = weight;
            node.attrs = new String(attrs);
            node_list.add(node);
            
            // Add insert position.
            
            INSPTR ipr = new INSPTR();
            INSPTR ipr2 = new INSPTR();
            int ni = node_list.lastIndexOf(node);
            ipr.lpnode = (XMLNODE)node_list.get(ni);
            ipr2.lpnode = (XMLNODE)node_list.get(ni);
            ipr.offset = offset;
            if (length > 0) {
                ipr.type = T_START;
                ins_ptr_list.add(ipr);    // Start tag.
                ipr2.offset = offset + length;
                ipr2.type = T_END;
                ins_ptr_list.add(ipr2);    // End tag.
            } else {
                ipr.type = T_EMPTY;
                ins_ptr_list.add(ipr);    // Empty tag.
            }
        }

        if (node_list.size() == 0) {
            System.err.println("No elements in element table.");
            return false;
        }
        
        if (_is_verbose) {
            System.out.println("Number of nodes in input file ="+ node_list.size() +".");
        }
        
        // Sort tag position.
        Sorting sg = new Sorting();
        sg.sort(ins_ptr_list, 0, ins_ptr_list.size() - 1, new INSPTR());

        // Check root node.
        Iterator ip = ins_ptr_list.iterator();
        INSPTR ins = (INSPTR)ip.next();
        
        if (ins.offset != 0) {
            System.err.println("The first tag position isn't zero." + ins.offset);
        }
        return true;
    }

    // Desc: ƥȥΡɤ򸡽Фɲä롥
    
    private void merge_text_node(ArrayList node_list, Vector ins_ptr_list) {
        if (_is_verbose) {
            System.out.println("Merging text node.");
        }
        // assert ins_ptr_list.size() > 1 ;

        Vector ins_ptr_list_old = new Vector();
        Vector ins_ptr_list_text = new Vector();
        ListIterator ip;
        ListIterator ip2;
        
        // create text node buffer
        ip = ins_ptr_list.listIterator(0);
        ip2 = ins_ptr_list.listIterator(1);

        while (ip2.hasNext()) {
            INSPTR ins = (INSPTR)ip.next();
            INSPTR ins2 = (INSPTR)ip2.next();

            final long  off1 = ins.offset;
            final long  off2 = ins2.offset;
            
            // assert off1 <= off2;
            if ((off1 < off2) && (ins.type != T_START ||
                                  ins2.type != T_END)) {
                // Add text node.
                XMLNODE     node = new XMLNODE();

                node.name = new String(_textnode_name);
                node.length = off2 - off1;
                node_list.add(node);

                // Add insert position.
                INSPTR  ipr = new INSPTR();
                INSPTR  ipr2 = new INSPTR();
                int ni = node_list.lastIndexOf(node);
                ipr.lpnode = (XMLNODE)node_list.get(ni);
                ipr2.lpnode = (XMLNODE)node_list.get(ni);
                ipr.offset = off1;
                ipr.type = T_START;
                ins_ptr_list_text.add(ipr);   // Start tag.
                ipr2.offset = off2;
                ipr2.type = T_END;
                ins_ptr_list_text.add(ipr2);   // End tag.
            }
        }

        if (_is_verbose) {
            System.out.println("Number of nodes after text node merging =" +
                               node_list.size() + ".");
        }
        
        // Merge text tag position.
        if (0 < ins_ptr_list_text.size()) {
            // Copy original.
            int s = ins_ptr_list.size();
            ins_ptr_list_old.ensureCapacity(s);
            ins_ptr_list_old = (Vector)ins_ptr_list.clone();
            
            // Expand buffer area.
            int sn = s + ins_ptr_list_text.size();
            ins_ptr_list.clear();
            ins_ptr_list.ensureCapacity(sn);

            // Merge original and text nodes.
            Sorting st = new Sorting();
            st.merge(ins_ptr_list_old, ins_ptr_list_text,
                     ins_ptr_list, new INSPTR());
        }
        return;
    }
    
    // Desc: XMLʸʸϤ롥
    
    private void output_xml_char(PrintWriter fpout, final int ch) {
        boolean crpred = false; // 'CR'
        if (!crpred && (ch == '\n' || ch == '\r')) {
            _linecount++;
            if (_opt_linenumout) {
                fpout.print("<!--" + _linecount + "-->");
            }
        }
        crpred = (ch == '\r');

        String buf = String.valueOf((char)ch);

        switch (ch) {               // üʸ
          case '<':
            buf = new String("&lt;");
        break;
          case '>':
            buf = new String("&gt;");
            break;
          case '&':
            buf = new String("&amp;"); 
        break;
          case 0x9:
            if (_opt_entityref1) buf = new String("&#x9;");
            break;
          case 0x20:
            if (_opt_entityref1) buf = new String("&#x20;");
            break;
          case '\n':
        if (_opt_entityref2) buf = new String("&#xA;");
            break;
          case '\r':
        if (_opt_entityref2) buf = new String("&#xD;");
            break;
        }
        fpout.print(buf);
    }
    
    // Desc: XMLʸ˥Ϥ롥

    private void output_xml_tag(PrintWriter fpout, final XMLNODE lpnode,
                                int type) {
        final String  tn = lpnode.name;
        final String  ta;
        if(lpnode.attrs != null)
            ta = lpnode.attrs;
        else
            ta = "";
        
        switch (type) {
          case T_START:
            fpout.print("<" + tn + ta + ">");
            break;
          case T_END:
            fpout.print("</" +tn + ">");
            break;
          case T_EMPTY:
            fpout.print("<" + tn + ta + "/>");
            break;
        }
    }

    // Desc: ѴΥᥤؿ

    private boolean process_main(BufferedReader fp, PrintWriter fpout,
                                 final Vector ins_ptr_list) {
        long    pos = 0;
        int     ch = 0;
        Iterator ip;
        INSPTR   insip;
        
        ip = ins_ptr_list.iterator();

        while(ip.hasNext()){
            insip = (INSPTR)ip.next();
            final long nextpos = insip.offset;

            // Copy characters until next inset position.
            for (int i = 0; i < nextpos - pos; i++) {
                try {
                    ch = fp.read();
                    if (ch == EOF) {
                        /* Error occurs. */ 
                        System.err.println("Detect the end of file before the last tag position.");
                        return false;
                    }
                }
                catch (Exception e){
                    System.out.println("e" + e);
                }
                output_xml_char(fpout, ch);
            }
            pos = nextpos;
        
            // Insert tag
            output_xml_tag(fpout, insip.lpnode, insip.type);
        }
        // It must not read next character.
        try {
            int cp;
            if ((cp = fp.read()) != EOF) {
                System.err.println("The last tag position isn't the end of file." + cp);
            }
        } catch(Exception e) {
        }
        if (_is_verbose) {
            System.out.println("Line counter =" +_linecount + ".");
        }

        return true;
    }

    // Desc: ȥݥȡ

    private boolean parse_param(String args[], int argc) {
        int     ch;
        String  check;
        Getopt  g = new Getopt("xbd3", args, "d:flnr:vV");

        while ((ch = g.getopt()) != -1) {
            switch (ch) {
              case 'd':
                // !! Temporaily skip.
                break;
              case 'f':
                _is_out_to_file = true;
                break;
              case 'l':
                _opt_linenumout = true;
                break;
              case 'n':
                _opt_no_textnode = true;
                break;
              case 'r':
                check = g.getOptarg();
                if (check.equals("1")) {
                    _opt_entityref1 = true;
                } else if (check.equals("2")) {
                    _opt_entityref2 = true;
                } else {
                    System.err.println("Invalid \"-r\" option.");
                    print_usage(program_name);
                }
                break;
              case 'v':
                _is_verbose = true;
                break;
              case 'V':
                _is_Verbose = true;
                break;
              default:
                System.err.println("'" + ch +"': Invalid option.");
                print_usage(program_name);
                break;
            }
        }
        argc -= g.getOptind();
        if (argc != 2) {
            System.err.println("No arguments.");
            print_usage(program_name);
        }
        _src_path = new String(args[0 + g.getOptind()]);
        _meta_path = new String(args[1 + g.getOptind()]);
        return true;
    }

    public static void main(String args[]) {
        boolean         retb;
        BufferedReader  fpsrc;
        BufferedReader  fpmeta;
        PrintWriter     fpout;
        xbd3            xbd = new xbd3();
        
        if (args.length < 2) {
            print_usage(new String("xbd3"));
            System.exit(1);
        }
        program_name = new String("xbd3");
        
        if (!xbd.parse_param(args, args.length)) {
            print_usage(program_name);
        }

        try {
            // Open source file.
            fpsrc = new BufferedReader(new FileReader(_src_path));
            // Open element table.
            fpmeta = new BufferedReader(new FileReader(_meta_path));
            
            // Open output file.
            if (_is_out_to_file) {
                String dst_path;
                dst_path = new String(_src_path + ".xml");
                fpout = new PrintWriter(new BufferedWriter(new FileWriter(dst_path)));
            } else {
                fpout = new PrintWriter(System.out, true);
            }
            
            // Ρɥꥹȡ
            ArrayList node_list = new ArrayList();
            
            // ֡
            Vector ins_ptr_list = new Vector();
            
            // Load element table.
            retb = xbd.load_metainfo_file(fpmeta, fpout, node_list, ins_ptr_list);
            if (!retb) {
                System.err.println("load_metainfo_file(): " + retb + ": Failed.");
                System.exit(1);
            }
            if (!_opt_no_textnode) {
                xbd.merge_text_node(node_list, ins_ptr_list);
            }
            
            // Build process.
            retb = xbd.process_main(fpsrc, fpout, ins_ptr_list);
            if (!retb) {
                System.err.println("process_main(): " + retb + ": Failed.");
                System.exit(1);
            }
            
            // Close files.
            fpsrc.close();
            fpmeta.close();
            fpout.close();
        } catch (Exception e) {
            System.err.println("open: \"" + program_name +
                               "\": Can't open file" + e);
            System.exit(1);
        }
        // return (EXIT_SUCCESS);
    }
    
    static void print_usage(String command_name) {
        System.err.println();
        System.err.println("NAME");
        System.err.println("    " + command_name);
        System.err.println("SYNOPSIS");
        System.err.println("    java JXModel.xbd3." + command_name + " [-fln][-r1][-r2] <targetJavafile> <elementTable>");
        System.err.println("DESCRIPTION");
        System.err.println("    ");
        System.err.println("NOTE");
        System.err.println("    elementTable format is:");
        System.err.println("    offset:%d length:%d weight:%d name:%d name:%s [attrs:%s]");
        // System.err.println("VERSION");
        // System.err.println("    " + SapidVersion + "." + SapidRevision +
        //                    "." + SapidPatchLevel);
        System.exit(1);
    }
}
