/*
 * Created on 2003/06/10
 * SpdElementList.java - Copyright (c) 2003 H.Yoshida
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS IS FREEWARE. NO WARRANTY.
 */

package JXModel3.xsdml;

import java.util.*;

/**
 * XSDML Element List Class
 * @author H.Yoshida
 */
public class SpdElementList {

	protected Vector m_elementvec;

	/**
	 * The constructor.
	 */
	public SpdElementList() {
		m_elementvec = new Vector(10);
	}

	/**
	 * Add element to this list.
	 * @param e
	 */
	public void add(SpdElement e) {
		m_elementvec.addElement(e);
	}

	/**
	 * Remove element.
	 * @param e
	 */
	public void remove(SpdElement e) {
		m_elementvec.removeElement(e);
	}

	/**
	 * Get element count.
	 */
	public int count() {
		return m_elementvec.size();
	}

	/**
	 * Is empty?
	 */
	public boolean isEmpty() {
		return (count()==0);
	}

	/**
	 * Invoke handler's method for each elements in the list.
	 * @param e
	 */
	public void invokeForEach(SpdElementHandler xseh) {
		for (int i=0; i<m_elementvec.size(); i++) {
			SpdElement xe = (SpdElement)m_elementvec.get(i);
			xseh.action(xe);
		}
	}
	
	/**
	 * Get first element.
	 */
	public SpdElement first() {
		if (isEmpty()) return null;
		return (SpdElement)m_elementvec.firstElement();
	}

	/**
	 * Get last element.
	 */
	public SpdElement last() {
		if (isEmpty()) return null;
		return (SpdElement)m_elementvec.lastElement();
	}
	
	/**
	 * Get an element.
	 * @param num = index
	 */
	public SpdElement getElementAt(int num) {
		if (num>=0 && num < count())
			return (SpdElement)m_elementvec.get(num);
		else return null;
	}
}

