/* 
 * Program:             $RCSfile: Creja.java,v $  $Revision: 41.7 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Creja.java,v 41.7 2003/02/15 11:56:26 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;

public class Creja {
    static boolean debug = false;
    private static Vector paths = new Vector();

    private static Enumeration getCrejaPaths() {
        return paths.elements();
    }
        
    /**
     * Searches FrejaEntry of className for package(packageName).
     * e.g. To get java.util.Vector, <P>
     * <PRE> Creaja.getFrtejaEntry("java.util", "Vector"); </PRE>
     */
    public static FrejaEntry getFrejaEntry(String packageName,
                                           String className) {
        return getFrejaEntry(packageName + "." + className);
    }

    /**
     * Search FrejaEntry of className for package
     *  (definedPackage, importedPacakges).
     * This method is the most general style of getFrejaEntry(...).
     *
     * <P>
     * example:<P>
     * <PRE>
     *   FrjFile myFile;
     *   ...
     *   FrjClass c = Creja.getFrejaEntry(myFile.getDefinedPacakge(),
     *                                    myFile.getImportedPacakges(),
     *                                    className);
     * </PRE>
     */
    public static FrejaEntry getFrejaEntry(FrjQName definedPackage,
                                           Enumeration importedPackages,
                                           String className) {
        if (className.indexOf(".") > 0) { // not >= 0
            // full quolified className (e.g. java.util.Vector)
            return getFrejaEntry(className);
        }
        /* else */

        FrejaEntry entry;
        if (definedPackage != null) {
            String packageName = definedPackage.getFullString();
            entry = getFrejaEntry(packageName, className);
            if (entry != null) {
                FrjWorld.debug(debug, "BINGO! "
                                      + packageName + " " + className
                                      + " is " + entry.getFrejaPath());
                return getCanonicalFrejaEntry(entry);
            }
        }
        else {
            // Search null package (i.e. SDB).
            entry = new FrejaEntry(FrjWorld.J_MODEL_PATH, className);
            if (entry.exists()) {
                return getCanonicalFrejaEntry(entry);
            } else {
                FrejaEntry.freeFrejaID(entry);
            }
        }

        if (importedPackages != null) {
            while (importedPackages.hasMoreElements()) {
                FrjQName qname =  (FrjQName) importedPackages.nextElement();
                String packageName = qname.getFullString();
                if (packageName.endsWith(".*")) {
                    int index = packageName.lastIndexOf(".*");
                    packageName = packageName.substring(0, index);
                    entry = getFrejaEntry(packageName, className);
                }
                else {  // import java.util.Vector; STYLE
                    int index = packageName.lastIndexOf(".") +1;
                    String pkgClassName = packageName.substring(index);
                    if (pkgClassName.equals(className)) {
                        entry = getFrejaEntry(packageName);
                    }
                    else {
                        continue;
                    }
                }
                if (entry != null) {
                    FrjWorld.debug(debug, "BINGO! "
                                          + packageName + " " + className
                                          + " is " + entry.getFrejaPath());
                    return getCanonicalFrejaEntry(entry);
                }
            }
        }
        // implicitly imported java.lang
        entry = getFrejaEntry0("java.lang." + className);
        if (entry != null) {
            return getCanonicalFrejaEntry(entry);
        }
        return null;

    }

    /**
     * Searches FrejaEntry of className for package(importedPacakges).
     */
    public static FrejaEntry getFrejaEntry(Enumeration importedPackages,
                                           String className) {
        return getFrejaEntry(null, importedPackages, className);
    }

    /**
     * Searches FrejaEntry of className regardless package.
     */
    public static FrejaEntry getFrejaEntry(String className) {
        FrejaEntry entry;
        if (className.indexOf(".") > 0) { // not >= 0
            // Search null package (i.e. SDB).
            entry = new FrejaEntry(FrjWorld.J_MODEL_PATH, className);
            if (entry.exists()) {
                return getCanonicalFrejaEntry(entry);
            }
            FrejaEntry.freeFrejaID(entry);                      
        }

        entry = getFrejaEntry0(className);
        if (entry != null) {
            return getCanonicalFrejaEntry(entry);
        }

        // implicitly imported java.lang
        entry = getFrejaEntry0("java.lang." + className);
        if (entry != null) {
            return getCanonicalFrejaEntry(entry);
        }

        return null;
    }

    private static FrejaEntry getFrejaEntry0(String className) {
        if (className.endsWith(".java")) {
            FrjWorld.warning(className + " has .java suffix.");
            int index = className.lastIndexOf(".java");
            className = className.substring(0, index);
        }

    crejaPathLoop:
        for (Enumeration dirNames = getCrejaPaths();
             dirNames.hasMoreElements();) {
            String dirName = (String) dirNames.nextElement();
            File dir = new File(dirName);
            if (!dir.isDirectory()) {
                continue crejaPathLoop;
            }
            StringTokenizer st = new StringTokenizer(className, ".");
        frejaPathLoop:
            while (st.hasMoreTokens()) {
                String pkg = st.nextToken();
                FrjWorld.debug(debug, "matching " + dirName
                                      + " " + pkg
                                      + " " + className);
                String[] dirFiles = dir.list();
                for (int i = 0; i < dirFiles.length; i++) {
                    if (pkg.equals(dirFiles[i])) {
                        if (!st.hasMoreTokens()) {
                            File frejaDir = new File(dirName
                                                     + File.separator
                                                     + dirFiles[i]);
                            if (isFrejaDir(frejaDir)) {
                                FrjWorld.debug(debug,
                                                      "!BINGO! " + dirName 
                                                      + " " + pkg
                                                      + " " + className);
                                int lastDot = className.lastIndexOf(".");
                                if (lastDot > 0) {
                                    return new FrejaEntry(dirName, pkg,
                                                          className.substring(0, lastDot));
                                } else {
                                    return new FrejaEntry(dirName, pkg);
                                }
                            } else {
                                FrjWorld.debug(debug,
                                                      "!MISMATCH! " + dirName 
                                                      + " " + pkg
                                                      + " " + className);
                            }
                            continue crejaPathLoop;
                        }
                        dirName +=  File.separator + dirFiles[i];
                        dir = new File(dirName);
                        if (!dir.isDirectory()) {
                            continue crejaPathLoop;
                        }
                        continue frejaPathLoop;
                    }
                }
                continue crejaPathLoop;
            }
        }
        return null;
    }

    private static FrejaEntry getCanonicalFrejaEntry(FrejaEntry entry) {
        if (entry == null) {
            return null;
        } else if (FrjExtents.isLoaded(entry)) {
            FrejaEntry centry =  FrjExtents.getLoadedFrejaEntry(entry);
            if (entry != centry) {
                FrejaEntry.freeFrejaID(entry);
            }
            return centry;
        } else {
            return entry;
        }
    }

    /**
     * Returns class names in SDB directory.
     */
    public static Enumeration getClasses() {
        return getClasses(new File(FrjWorld.J_MODEL_PATH));
    }

    /**
     * Returns class names defined in the package(packageName).
     * NOTE: Returned class names doesn't include package name.
     */
    public static Enumeration getClasses(String packageName) {
        return getClasses(getPackageDir(packageName));
    }

    private static Enumeration getClasses(File pkgDir) {
        Vector v = new Vector();
        if (pkgDir != null) {
            String[] dirs =  pkgDir.list();
            for (int i = 0; i < dirs.length; i++) {
                File dir = new File(pkgDir, dirs[i]);
                if (isFrejaDir(dir)) {
                    v.addElement(dirs[i]);
                }
            }
        }
        return v.elements();
    }

    /**
     * Returns sub packages defined in the package(packageName).
     */
    public static Enumeration getSubPackages(String packageName) {
        return getSubPackages(packageName, getPackageDir(packageName));
    }

    private static Enumeration getSubPackages(String prefix, File pkgDir) {
        Vector v = new Vector();
        if (pkgDir != null) {
            String[] dirs =  pkgDir.list();
            for (int i = 0; i < dirs.length; i++) {
                File dir = new File(pkgDir, dirs[i]);
                if (!isFrejaDir(dir)) {
                    v.addElement(prefix + "." + dirs[i]);
                }
            }
        }
        return v.elements();
    }

    private static File getPackageDir(String packageName) {
    pathLoop:
        for (Enumeration dirNames = getCrejaPaths();
             dirNames.hasMoreElements();) {
            String dirName = (String) dirNames.nextElement();
            File dir = new File(dirName);
            if (!dir.isDirectory()) {
                continue pathLoop;
            }
            StringTokenizer st = new StringTokenizer(packageName, ".");
            while (st.hasMoreTokens()) {
                String pkgIdName = st.nextToken(); // identifier
                String[] dirFiles = dir.list();
                for (int i = 0; i < dirFiles.length; i++) {
                    if (pkgIdName.equals(dirFiles[i])) {
                        dirName +=  File.separator + dirFiles[i];
                        dir = new File(dirName);
                        if (!dir.isDirectory()) {
                            continue pathLoop;
                        }
                        if (!st.hasMoreTokens()) {
                            return dir;
                        }
                    }
                }
            }
        }
        return null;
    }

    private static String homeDir = System.getProperty("user.home");
    private static final String prefFileName = ".crejapath";
    private static final String delimiter = " ";
    private static File prefFile = null;
                
    private static void readPref() {
        prefFile = new File(homeDir, prefFileName);
        if (prefFile.exists()) {        // exists specified file
            try {
                FileInputStream fin = new FileInputStream(prefFile);
                BufferedReader din
                    = new BufferedReader(new InputStreamReader(fin));
                String line;
                if ((line = din.readLine()) != null) {
                    StringTokenizer st
                        = new StringTokenizer(line, delimiter);
                    while (st.hasMoreTokens()) {
                        paths.addElement(st.nextToken());
                    }
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        } else {
            // Not set Creja, adding default path.
            paths.addElement(FrjWorld.J_MODEL_PATH);
        }
    }

    private static void savePref() {
        if (prefFile == null) {
            prefFile = new File(homeDir, prefFileName);
        }
        try {
            FileWriter fout = new FileWriter(prefFile);
            for (Enumeration e = paths.elements();
                 e.hasMoreElements();) {
                fout.write((String) e.nextElement());
                fout.write(delimiter);
            }
            fout.write("\n");
            fout.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private static void addPath(String path) {
        paths.addElement(path);
    }

    private static boolean isFrejaDir(File dir) {
        if (dir.exists() && dir.isDirectory()) {
            String dirName = dir.getPath();
            // existing some file in Freja ?
            File routineFile = new File(dirName + File.separator
                                        + FrjRoutine.getClassNameStatically());
            File class_routineFile = new File(dirName + File.separator
                                              + FrjRelations.CLASS_ROUTINE);
            if (routineFile.exists() && routineFile.isFile()
                && class_routineFile.exists() && class_routineFile.isFile()) {
                return true;
            }
        }
        return false;
    }


    static {
        readPref();
    }

    /**
     * Sets Creja path on command line.
     * Command line arguments are lists of path.
     * If no command line arguments are specified,
     *  shows current Creja path.
     *
     * <P>
     * example:
     * <PRE>
     *   % java Japid.Freja.Creja SDB /usr/local/Japid/Creja
     *   % java Japid.Freja.creja
     *   SDB
     *   /usr/home/hachisu/Research/Japid/Creja
     *   %
     * </PRE>
     */
    public static void main(String args[]) {
        if (args.length == 0) {
            for (Enumeration e = getCrejaPaths();
                 e.hasMoreElements();) {
                System.out.println((String) e.nextElement());
            }
        } else {
            paths.removeAllElements();
            for (int i = 0; i < args.length; i++) {
                addPath(args[i]);
            }
            savePref();
        }
    }
}
