/* 
 * Program:             $RCSfile: CrejaUtil.java,v $  $Revision: 41.6 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: CrejaUtil.java,v 41.6 2003/02/15 11:56:26 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;

public class CrejaUtil {
    static boolean debug = false;

    public static void main(String args[]) {
        try {
            for (int i = 0; i < args.length; i++) {
                if (args[i].equals("-w")
                    || args[i].equals("-which")) {
                    which(args[++i]);
                }
                else if (args[i].equals("-classes")) {
                    classes(args[++i]);
                }
                else if (args[i].equals("-subpkg")) {
                    subPackages(args[++i]);
                }
                else {
                    usage();
                    System.exit(1);
                }
            }
        }
        catch (Exception e) {
            System.err.println("Error: " + e.getMessage());
            usage();
            System.exit(1);
        }
                        
    }

    static void which(String className) {
        if (className == null) {
            usage();
            System.exit(1);
        }
        FrejaEntry entry = Creja.getFrejaEntry(className);
        if (entry == null || !entry.exists()) {
            FrjWorld.error("No such SDB " + className);
        }
        else {
            System.out.println(className + ": " + entry.getFrejaPath());
        }
    }

    static void classes(String pkgName) {
        if (pkgName == null) {
            usage();
            System.exit(1);
        }

        System.out.println(pkgName + ":");
        for (Enumeration e = Creja.getClasses(pkgName);
             e.hasMoreElements();) {
            String className = (String) e.nextElement();
            System.out.println(pkgName + "." + className);
        }
    }

    static void subPackages(String pkgName) {
        if (pkgName == null) {
            usage();
            System.exit(1);
        }

        System.out.println(pkgName + ":");
        for (Enumeration e = Creja.getSubPackages(pkgName);
             e.hasMoreElements();) {
            String subPkgName = (String) e.nextElement();
            System.out.println(subPkgName);
        }
    }

    static void usage() {
        String mes = "Usage: java Japid.Freja.Creja [options]\n"
            + "options:\n"
            + "-w/-which className:  shows Freja\n"
            + "-classes pkgName:     shows classes in pkgName\n"
            + "-subpkg pkgName:      shows sub packages in pkgName";
        System.err.println(mes);
    }
}
