/* 
 * Program:             $RCSfile: DBQuery.java,v $  $Revision: 40.15 $
 * 
 * Purpose:             The class read data from Freja.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: DBQuery.java,v 40.15 2003/02/15 11:56:36 yamamoto Exp yamamoto $
 */

package Japid.Freja.DBAccess;

import java.io.*;
import java.util.*;
import Japid.Freja.*;

/*
 * SDBη(ܺ٤"<SAPID_TOP>/doc/TechMemo/SDB-format.ja"򻲾Ȥ뤳):
 * 
 * TEXT mode:
 *     intǤθ'\1'
 *     stringǤθ'\1'
 *     tupleν'\n'
 * BINARY mode:
 *     intǤ4byteΥ(ڤ̵)
 *     stringǤθ'\0'
 *     tupleζڤ̵
 */

public class DBQuery {
    private String                  fileName;
    private BufferedReader          dbFileInText;
    private BufferedInputStream     dbFileInBinary;
    private BufferedWriter          dbFileOutText;
    private BufferedOutputStream    dbFileOutBinary;

    private FrejaEntry  entry;
    private int         readSDBformat;
    private int         writeSDBformat;

    private String className;

    // !! ɤ餫ʤΤξѰդΤϥʥ󥻥󥹤
    // K.Hoshino 2001/10/16.  
    private DBClass dbc;
    private DBRel dbr;

    private final static int integerSize = 4;

    public DBQuery(FrejaEntry entry, String className) {
        this.fileName = entry.getFrejaPath() + File.separator + className;
        this.entry = entry;
        this.className = className;
        if (entry.exists()) {
            readSDBformat = entry.jpdSDBformat.jpdGetReadSDBformat();
            writeSDBformat = entry.jpdSDBformat.jpdGetWriteSDBformat();
        }

        dbc = FrjWorld.dbTable.getDBClassByName(className);
        dbr = FrjWorld.dbTable.getDBRelByName(className);
    }

    /* see "lib/SpdUtil/spdIO.c" and "AR/AR5/scheme.c". */

    /**
     * Read data from Database and return tokenized strings.
     * if EOF return null.
     */
    public DBClassObject getAClassTuple() {
        DBClassObject obj = new DBClassObject(dbc);
        for (int i = 0; i < obj.getAttrNum(); i++) {
            try {
                if (obj.attrIsIntAt(i)) {
                    obj.setIntValue(jpdReadIntAttr());
                } else if (obj.attrIsStringAt(i)) {
                    obj.setStringValue(jpdReadStrAttr());
                } else {
                    FrjWorld.error("Invliad SDB table.");
                }
            } catch (DBException e) {
                // !! exception򤳤Τ褦˻ȤȤɤʤΤ⤷ʤ
                // K.Hoshino 2001.
                if (i == 0) {
                    return null;
                } else {
                    FrjWorld.error(fileName + ": Invalid SDB Format.");
                }
            }
        }
        
        /* Tupleλ */
        jpdReadTupleFinishCode();
        
        return obj;
    }

    public DBRelObject getARelTuple() {
        DBRelObject obj = new DBRelObject(dbr);
        for (int i = 0; i < obj.getAttrNum(); i++) {
            try {
                if (obj.attrIsIntAt(i)) {
                    obj.setIntValue(jpdReadIntAttr());
                } else if (obj.attrIsStringAt(i)) {
                    obj.setStringValue(jpdReadStrAttr());
                } else {
                    FrjWorld.error("Invliad SDB table.");
                }
            } catch (DBException e) {
                if (i == 0) {
                    return null;
                } else {
                    FrjWorld.error("Bad SDB Format in file " + fileName);
                }
            }
        }
        
        /* Tupleλ */
        jpdReadTupleFinishCode();

        return obj;
    }

    public boolean trylock4Reading() throws DBException {
        File dbFile = new File(fileName);
        try {
            if (dbFile.exists()) {      // exists specified file
                FileInputStream fin = new FileInputStream(dbFile);
                dbFileInText = null;
                dbFileInBinary = null;
                if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
                    dbFileInText = new BufferedReader(new InputStreamReader(fin));
                } else {
                    dbFileInBinary = new BufferedInputStream(fin);
                }
                return true;
            } else {
                throw new DBException("File " + fileName + " not found.");
            }
        }
        catch (DBException e) {
            throw new DBException(e.getMessage());
        }
        catch (Exception e) {
            throw new DBException("Unknown error occured when reading "
                                  + fileName + ".");
        }
    }
    
    public void unlock4Reading() {
        try {
            if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
                if (dbFileInText != null) {
                    dbFileInText.close();
                }
            } else {
                if (dbFileInBinary != null) {
                    dbFileInBinary.close();
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Writes data to SDB.
     */
    public void writeData(String line) {
        try {
            if (dbFileOutText != null) {
                dbFileOutText.write(line);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * write class data to Database.
     */
    public void writeAClassTuple(DBClassObject obj) {
        for (int i = 0; i < obj.getAttrNum(); i++) {
            if (obj.attrIsIntAt(i)) {
                jpdWriteIntAttr(obj.getIntValueAt(i));
            } else if (obj.attrIsStringAt(i)) {
                jpdWriteStrAttr(obj.getStringValueAt(i));
            } else {
                FrjWorld.error("Invliad SDB table.");
            }
        }
        
        /* Tupleλ */
        jpdWriteTupleFinishCode();
    }

    /**
     * write relation data to Database.
     */
    public void writeARelTuple(DBRelObject rel) {
        for (int i = 0; i < rel.getAttrNum(); i++) {
            if (rel.attrIsIntAt(i)) {
                jpdWriteIntAttr(rel.getIntValueAt(i));
            } else if (rel.attrIsStringAt(i)) {
                jpdWriteStrAttr(rel.getStringValueAt(i));
            } else {
                FrjWorld.error("Invliad SDB table.");
            }
        }
        
        /* Tupleλ */
        jpdWriteTupleFinishCode();
    }

    public boolean trylock4Writing() throws DBException {
        File dbFile = new File(fileName);
        try {
            if (dbFile.exists() && dbFile.canWrite()) {
                FileOutputStream fout = new FileOutputStream(dbFile);
                dbFileOutText = null;
                dbFileOutBinary = null;
                if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
                    dbFileOutText = new BufferedWriter(new OutputStreamWriter(fout));
                } else {
                    dbFileOutBinary = new BufferedOutputStream(fout);
                }
                return true;
            } else {
                throw new DBException("Cannot write to " + fileName + ".");
            }
        }
        catch (DBException e) {
            throw new DBException(e.getMessage());
        }
        catch (Exception e) {
            throw new DBException("Unknown error occured when reading "
                                  + fileName + ".");
        }
    }
    
    public void unlock4Writing() {
        try {
            if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
                if (dbFileOutText != null) {
                    dbFileOutText.flush();
                    dbFileOutText.close();
                }
            } else {
                if (dbFileOutBinary != null) {
                    dbFileOutBinary.flush();
                    dbFileOutBinary.close();
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    // -------------------------------------------------------------------------

    public void jpdReadTupleFinishCode() {
        if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
            int c;
            char ch;
            try {
                /* attrڤФ */
                c = dbFileInText.read();
                ch = (char)c;
                if (ch != JDBConsts.SDB_TEXT_TUPLE_SEPARATOR) {
                    FrjWorld.errorExit(fileName + ": spdReadTupleFinishCode().");
                }
            } catch (IOException e) {
                FrjWorld.errorExit(fileName + "Can't read data file.");
            }
        }
    }

    // -------------------------------------------------------------------------

    public int jpdReadIntAttr() throws DBException {
        if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
            return jpdReadIntAttrText();
        } else {
            return jpdReadIntAttrBinary();
        }
    }

    private int jpdReadIntAttrText() throws DBException {
        int c;
        char ch;

        String s = new String("");

        try {
            /* attrڤФ */
            c = dbFileInText.read();
            ch = (char) c;
            if (c == -1) {
                throw new DBException("data finish");
            }
            while (ch != JDBConsts.SDB_ATTR_SEPARATOR && c != -1) {
                s = s + ch;
                c = dbFileInText.read();
                ch = (char) c;
            }
            
            if (c == -1) {
                FrjWorld.errorExit(fileName + ": jpdReadIntAttrText().");
            }
        } catch (IOException e) {
            FrjWorld.errorExit(fileName + ": Can't read data file.");
        }

        return Integer.parseInt(s);
    }
    
    private int jpdReadIntAttrBinary() throws DBException {
        byte[] val = new byte[integerSize];

        int size = 4;

        try {
            size = dbFileInBinary.read(val, 0, integerSize);
            if (size != integerSize) {
                if (size <= 0) {
                    throw new DBException("data finish");
                } else {
                    FrjWorld.error(fileName + ": jpdReadIntAttrBinary().");
                    for (int i = 0; i < size; i++) {
                        System.err.print(Byte.toString(val[i]) + " ");
                    }
                    System.err.println();
                    System.exit(1);
                }
            }
        } catch (IOException e) {
            FrjWorld.errorExit(fileName + ": Can't read data file.");
        }

        return ((new java.math.BigInteger(val)).intValue());
    }

    // -------------------------------------------------------------------------

    public String jpdReadStrAttr() throws DBException {
        if (readSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
            return jpdReadStrAttrText();
        } else {
            return jpdReadStrAttrBinary();
        }
    }

    private String jpdReadStrAttrText() throws DBException {
        int c;
        char ch;

        String s = new String("");

        try {
            /* attrڤФ */
            c = dbFileInText.read();
            ch = (char) c;
            if (c == -1) {
                throw new DBException("data finish");
            }
            while (ch != JDBConsts.SDB_ATTR_SEPARATOR && c != -1) {
                s = s + ch;
                c = dbFileInText.read();
                ch = (char) c;
            }
            if (c == -1) {
                FrjWorld.errorExit(fileName + ": jpdReadStrAttrText().");
            }
        } catch (IOException e) {
            FrjWorld.errorExit(fileName + ": Can't read data file.");
        }

        return s;
    }

    private String jpdReadStrAttrBinary() throws DBException {
        Vector byteVector = new Vector();

        byte[] onebyte = new byte[1];

        try {
            int size = 1;

            size = dbFileInBinary.read(onebyte, 0, 1);
            if (size <= 0) {
                throw new DBException("data finish");
            }
            
            while (size > 0 && onebyte[0] != 0) {
                byteVector.add(new Byte(onebyte[0]));
                size = dbFileInBinary.read(onebyte, 0, 1);                
            }
        } catch (IOException e) {
            FrjWorld.errorExit(fileName + "Can't read data file.");
        }

        if (onebyte[0] != 0) {
            FrjWorld.errorExit(fileName + ": jpdReadStrAttrBinary().");
        }

        Byte[] byteArray = (Byte[]) byteVector.toArray(new Byte[0]);

        byte[] bytes = new byte[byteArray.length];
        for (int i = 0; i < byteArray.length; i++) {
            bytes[i] = byteArray[i].byteValue();
        }

        return (new String(bytes));
    }

    // -------------------------------------------------------------------------

    public void jpdWriteTupleFinishCode() {
        if (writeSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) {
            try {
                dbFileOutText.write((int) JDBConsts.SDB_TEXT_TUPLE_SEPARATOR);
            } catch (IOException e) {
                FrjWorld.error("Can't write to data file. : " + fileName);
                System.exit(1);
            }
        } else {
        }
    }

    // -------------------------------------------------------------------------

    public void jpdWriteIntAttr(int val) {
        if (writeSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) { 
            jpdWriteIntAttrText(val);
        } else {
            jpdWriteIntAttrBinary(val);
        }
    }

    private void jpdWriteIntAttrText(int val) {
        try {
            dbFileOutText.write(Integer.toString(val));
            dbFileOutText.write((int) JDBConsts.SDB_ATTR_SEPARATOR);
        } catch (IOException e) {
            FrjWorld.error("Can't write to data file. : " + fileName);
            System.exit(1);
        }
    }

    private void jpdWriteIntAttrBinary(int val) {
        try {
            dbFileOutBinary.write(intToBytes(val));
        } catch (IOException e) {
            FrjWorld.error("Can't write to data file. : " + fileName);
            System.exit(1);
        }
    }
    private static byte[] intToBytes(int val) {
        byte[] ret_bytes = new byte[integerSize];
        for (int i = 0; i < integerSize; i++) {
            int tmpval = val >> (integerSize - i - 1) * 8;
            ret_bytes[i] = (new Integer(tmpval)).byteValue();
        }
        return ret_bytes;
    }

    // -------------------------------------------------------------------------

    public void jpdWriteStrAttr(String str) {
        if (writeSDBformat == JpdSDBformat.JAPID_TEXT_FORMAT_SDB) { 
            jpdWriteStrAttrText(str);
        } else {
            jpdWriteStrAttrBinary(str);
        }
    }

    private void jpdWriteStrAttrText(String str) {
        try {
            dbFileOutText.write(str);
            dbFileOutText.write((int) JDBConsts.SDB_ATTR_SEPARATOR);
        } catch (IOException e) {
            FrjWorld.error("Can't write to data file. : " + fileName);
            System.exit(1);
        }
    }

    private void jpdWriteStrAttrBinary(String str) {
        try {
            dbFileOutBinary.write(str.getBytes());
            dbFileOutBinary.write((int) '\0');
        } catch (IOException e) {
            FrjWorld.error("Can't write to data file. : " + fileName);
            System.exit(1);
        }
    }
}
