/* 
 * Program:             $RCSfile: JpdInfo.java,v $  $Revision: 40.13 $
 * 
 * Purpose:             Handler for InfoFile.
 *
 * Author:              K.Hoshino  2001/10/02
 *                      S.Yamamoto  2001/10/02
 *
 * (C) Copyright:       K.Hoshino and S.Yamamoto  2001 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: JpdInfo.java,v 40.13 2003/02/15 11:56:37 yamamoto Exp yamamoto $
 */

package Japid.Freja.DBAccess;

import java.io.*;
import java.util.*;
import Japid.Freja.*;

/**
 * see lib/SpdUtil/spdInfo.c
 *
 * SDBBINARYASCIIۤʤ뤫⤷ʤ
 * (ʣΰ㤦ǤmillˤʣSDBǤ
 * Creja PathˤŪ˰Ȥꤹȡ
 * 줾SDB(Ĥޤmill)κۼɬפ롥)
 *
 * BINARYASCIIξInfoե
 * SDBγJava饹ɽǥ쥯ȥ겼֤٤
 *
 * K.Hoshino 2001/10/06
 */

public class JpdInfo {
    public File infoFile;

    /** Crejaǥ쥯ȥ˺Infoեdefault̾ */
    public static final String JAPID_INFO_FILE_NAME = ".info";
    /** Java饹եǥ쥯ȥ˺Infoեdefault̾ */
    public static final String CLASS_INFO_FILE_NAME = ".info";

    /** JpdInfoDataν */
    public Vector jpdInfoData;

    public JpdInfo(){
        infoFile = null;
        jpdInfoData = null;
    }

    public JpdInfo(String infoFileName){
        this();

        this.infoFile = new File(infoFileName);
        readAllInfo();
    }

    public JpdInfo(String infoFilePath, String infoFileName){
        this();

        this.infoFile = new File(infoFilePath, infoFileName);
        readAllInfo();
    }

    public JpdInfo(File infoFile){
        this();

        this.infoFile = infoFile;
        readAllInfo();
    }

    /**
     * ꤵ줿Java饹եǥ쥯ȥ(FrejaȸäƤΤ)
     * ˺줿Infoեopen.
     */
    public JpdInfo(FrejaEntry entry){
        this();

        String fileName = entry.getFrejaPath() + File.separator + CLASS_INFO_FILE_NAME;
        this.infoFile = new File(fileName);
        //this.infoFile = new File(entry.getFrejaPath(), CLASS_INFO_FILE_NAME);
        readAllInfo();
    }

    /** InfoDataΥꥢ */
    public void clearJpdInfoData() {
        jpdInfoData = new Vector();
    }

    private void readAllInfo() {
        BufferedReader infoFileIn = null;

        String oneLine = null;
        String attr;
        String val;
        int linenum = 1;

        /* open info file */
        try {
            FileInputStream fin = new FileInputStream(infoFile);
            infoFileIn = new BufferedReader(new InputStreamReader(fin));
        } catch (IOException e) {
            if (e instanceof FileNotFoundException) {
                FrjWorld.error("InfoFile " + infoFile + " not found.");
            } else {
                FrjWorld.error("IOException occured.");
                e.printStackTrace();
            }
        }

        clearJpdInfoData();

        try {
            while ((oneLine = infoFileIn.readLine()) != null) {
                int index = 0;
                char c = '\0';
                
                if (oneLine.length() != 0) {
                    /* Get attribute name. */
                    attr = new String("");
                    while (index < oneLine.length()) {
                        c = oneLine.charAt(index);
                        
                        if (c == '=') {
                            break;
                        } else {
                            attr = attr + c;
                            index++;
                        }
                    }
                    
                    if (c != '=') 
                        FrjWorld.error("Invalid Info format : line " + linenum + ".");
                    index++;
                    
                    /* Get value name. */
                    val = new String("");
                    while (index < oneLine.length()) {
                        c = oneLine.charAt(index);
                        
                        val = val + c;
                        index++;
                    }
                    
                    /* elementɲ */
                    jpdInfoData.add(new JpdInfoData(attr, val));
                    
                    linenum++;
                }
            }
        } catch (IOException e) {
            FrjWorld.error("IOException occered when reading " + infoFile + ".");
        }
        
        /* close info file */
        try {
            infoFileIn.close();
        } catch (IOException e) {
            FrjWorld.error("Can't close " + infoFile + ".");
        }
    }
    
    public void writeAllInfo() {
        BufferedWriter infoFileOut = null;

        /* open info file */
        try {
            FileOutputStream fout = new FileOutputStream(infoFile);
            infoFileOut = new BufferedWriter(new OutputStreamWriter(fout));
          } catch (IOException e) {
              if (e instanceof FileNotFoundException) {
                  FrjWorld.error("InfoFile " + infoFile + " not found.");
              } else {
                  FrjWorld.error("IOException occured.");
                  e.printStackTrace();
              }
        }

        try {
            for (int i = 0; i < jpdInfoData.size(); i++) {
                JpdInfoData jid = (JpdInfoData) jpdInfoData.elementAt(i);
                infoFileOut.write(jid.toString());
            }
        } catch (IOException e) {
            FrjWorld.error("IOException occered when reading " + infoFile + ".");
        }

        /* close info file */
        try {
            infoFileOut.flush();
            infoFileOut.close();
        } catch (IOException e) {
            FrjWorld.error("Can't close " + infoFile + ".");
        }
    }

    // ---------------------------------------------------------------------

    private JpdInfoData getJpdInfoData(String attr){
        JpdInfoData jid = null;
        for (int i = 0; i < jpdInfoData.size(); i++) {
            jid = (JpdInfoData) jpdInfoData.elementAt(i);
            if (jid.attrIs(attr)) break;
        }

        if (jid != null && jid.attrIs(attr)) return jid;
        else {
            FrjWorld.error("No such attribute \"" + attr + "\" in "
                                  + infoFile + ".");
            return null;
        }
    }

    /** attrǤͤInt֤ */
    public int jpdGetInfoInt(String attr) {
        if (attr == null) return -1;

        JpdInfoData jid = getJpdInfoData(attr);

        if (jid != null) return jid.getValInt();
        else return -1;
    }

    /** attrǤͤString֤ */
    public String jpdGetInfoString(String attr) {
        if (attr == null) return null;

        JpdInfoData jid = getJpdInfoData(attr);

        if (jid != null) return jid.getValString();
        else return null;
    }

    /** attrǤͤintset */
    public boolean jpdSetInfoInt(String attr, int val) {
        if (attr == null) return false;

        JpdInfoData jid = getJpdInfoData(attr);

        if (jid != null) {
            jid.setValInt(val);
        } else {
            /* elementοɲ */
            jpdInfoData.add(new JpdInfoData(attr, val));
        }

        return true;
    }

    /** attrǤͤStringset */
    public boolean jpdSetInfoString(String attr, String val) {
        if (attr == null) return false;

        JpdInfoData jid = getJpdInfoData(attr);

        if (jid != null) {
            jid.setValString(val);
        } else {
            /* elementοɲ */
            jpdInfoData.add(new JpdInfoData(attr, val));
        }

        return true;
    }
}

/**
 * Infoե
 * ͤStringݻƤ
 */
class JpdInfoData {

    public String attr;
    public String val;

    public JpdInfoData(String attr, int val) {
        this.attr = new String(attr);
        this.val = Integer.toString(val);
    }

    public JpdInfoData(String attr, String val) {
        this.attr = new String(attr);
        this.val = new String(val);
    }

    public boolean attrIs(String attr) {
        return attr.equals(this.attr);
    }

    public String getValString() {
        return val;
    }

    public int getValInt() {
        if (val.equals("TRUE")) return 1;
        else if (val.equals("FALSE")) return 0;
        else return Integer.parseInt(val);
    }

    public void setValString(String val) {
        this.val = new String(val);
    }

    public void setValInt(int val) {
        setValString(Integer.toString(val));
    }

    public String toString() {
        String returnStr = attr + "=" + val + "\n";
        return returnStr;
    }
}
