/* 
 * Program:             $RCSfile: ExternalAccessibleLoader.java,v $  $Revision: 41.4 $
 * 
 * Purpose:             The super class of all the DBClasses.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: ExternalAccessibleLoader.java,v 41.4 2003/02/12 03:39:30 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.util.*;
import Japid.Freja.*;

public class ExternalAccessibleLoader implements Loader {
    static Vector       classNames = new Vector();
    static Vector       otherClassNames = new Vector();
    static Vector       relNames = new Vector();
    static Vector       otherRelNames = new Vector();
    static Vector       partialLoadedEntries = new Vector();

    public void load(FrejaEntry entry) {
        DefaultLoader.load(getClassNames(), null, entry);
        DefaultLoader.load(new FrjRelation(), getRelationNames(),
                           null, entry);
        // Set related objects.
        for (Enumeration e = FrjExtents.getAllObjects(entry.getFrejaID());
             e.hasMoreElements();) {
            ((FrjObject) e.nextElement()).setRelatedObjects();
        }
        partialLoadedEntries.addElement(entry);
    }

    public static void loadOthers(FrejaEntry entry) {
        DefaultLoader.load(getOtherClassNames(), null, entry);
        DefaultLoader.load(new FrjRelation(), getOtherRelationNames(),
                           null, entry);
        // Set related objects.
        for (Enumeration e = FrjExtents.getAllObjects(entry.getFrejaID());
             e.hasMoreElements();) {
            ((FrjObject) e.nextElement()).resetRelatedObjects();
        }
        partialLoadedEntries.removeElement(entry);
    }

    public static boolean isPartialLoaded(FrejaEntry entry) {
        return partialLoadedEntries.contains(entry);
    }

    static {
        classNames.addElement("Japid.Freja.FrjFile");
        classNames.addElement("Japid.Freja.FrjClass");
        classNames.addElement("Japid.Freja.FrjInterface");
        classNames.addElement("Japid.Freja.FrjQName");
        classNames.addElement("Japid.Freja.FrjType");
        classNames.addElement("Japid.Freja.FrjVariable");
        classNames.addElement("Japid.Freja.FrjRoutine");
        classNames.addElement("Japid.Freja.FrjModifier");

        otherClassNames.addElement("Japid.Freja.FrjStatement");
        otherClassNames.addElement("Japid.Freja.FrjLabel");
        otherClassNames.addElement("Japid.Freja.FrjExpression");
        otherClassNames.addElement("Japid.Freja.FrjSpecial");
        otherClassNames.addElement("Japid.Freja.FrjOperator");
        otherClassNames.addElement("Japid.Freja.FrjLiteral");
        otherClassNames.addElement("Japid.Freja.FrjSBlock");
        otherClassNames.addElement("Japid.Freja.FrjComment");

        relNames.addElement(FrjRelations.DEFINE_PACKAGE);
        relNames.addElement(FrjRelations.IMPORT_PACKAGE);
        relNames.addElement(FrjRelations.DEFINE_CLASS);
        relNames.addElement(FrjRelations.IMPLEMENT_INTERFACE);
        relNames.addElement(FrjRelations.EXTEND_CLASS);
        relNames.addElement(FrjRelations.CLASS_MODIFIER);
        relNames.addElement(FrjRelations.CLASS_VARIABLE);
        relNames.addElement(FrjRelations.CLASS_ROUTINE);
        relNames.addElement(FrjRelations.CLASS_CLASS);          /* JDK 1.1.*/
        relNames.addElement(FrjRelations.CLASS_INTERFACE);      /* JDK 1.1.*/
        relNames.addElement(FrjRelations.DEFINE_INTERFACE);
        relNames.addElement(FrjRelations.EXTEND_INTERFACE);
        relNames.addElement(FrjRelations.INTERFACE_MODIFIER);
        relNames.addElement(FrjRelations.INTERFACE_VARIABLE);
        relNames.addElement(FrjRelations.INTERFACE_ROUTINE);
        relNames.addElement(FrjRelations.INTERFACE_CLASS);      /* JDK 1.1.*/
        relNames.addElement(FrjRelations.INTERFACE_INTERFACE);  /* JDK 1.1.*/
        relNames.addElement(FrjRelations.VARIABLE_MODIFIER);
        relNames.addElement(FrjRelations.VARIABLE_TYPE);
        relNames.addElement(FrjRelations.PARAMETER_VARIABLE);
        relNames.addElement(FrjRelations.RETURN_TYPE);
        relNames.addElement(FrjRelations.THROW_TYPE);
        relNames.addElement(FrjRelations.ROUTINE_MODIFIER);
        relNames.addElement(FrjRelations.TYPE_CLASS);
        relNames.addElement(FrjRelations.TYPE_INTERFACE);
        relNames.addElement(FrjRelations.TYPE_TYPE);
        relNames.addElement(FrjRelations.TYPE_QNAME);
        relNames.addElement(FrjRelations.QNAME_QNAME);

        otherRelNames.addElement(FrjRelations.CLASS_ARGUMENT);  /* JDK 1.1.*/
        otherRelNames.addElement(FrjRelations.CLASS_SBLOCK);
        otherRelNames.addElement(FrjRelations.INITIALIZE_VARIABLE);
        otherRelNames.addElement(FrjRelations.ROUTINE_BODY);
        otherRelNames.addElement(FrjRelations.QNAME_DOT);
        otherRelNames.addElement(FrjRelations.SBLOCK_BODY);
        otherRelNames.addElement(FrjRelations.LOCAL_CLASS);     /* JDK 1.1.*/
        otherRelNames.addElement(FrjRelations.LOCAL_VARIABLE);
        otherRelNames.addElement(FrjRelations.STATEMENT_LABEL);
        otherRelNames.addElement(FrjRelations.STATEMENT_STATEMENT);
        otherRelNames.addElement(FrjRelations.STATEMENT_EXPRESSION);
        otherRelNames.addElement(FrjRelations.REFER_LABEL);
        otherRelNames.addElement(FrjRelations.CASE_VALUE);
        otherRelNames.addElement(FrjRelations.EXPRESSION_OPERAND);
        otherRelNames.addElement(FrjRelations.EXPRESSION_OPERATOR);
        otherRelNames.addElement(FrjRelations.EXPRESSION_EXPRESSION);
        otherRelNames.addElement(FrjRelations.REFER_VARIABLE);
        otherRelNames.addElement(FrjRelations.REFER_LITERAL);
        otherRelNames.addElement(FrjRelations.REFER_SPECIAL);
        otherRelNames.addElement(FrjRelations.REFER_TYPE);
        otherRelNames.addElement(FrjRelations.REFER_QNAME);
        otherRelNames.addElement(FrjRelations.HAS_TYPE);
        otherRelNames.addElement(FrjRelations.INVOKE_ROUTINE);
        otherRelNames.addElement(FrjRelations.SYMBOL_COMMENT);
    }

    static public Enumeration getClassNames() {
        return classNames.elements();
    }
    static public Enumeration getRelationNames() {
        return relNames.elements();
    }
    static public Enumeration getOtherClassNames() {
        return otherClassNames.elements();
    }
    static public Enumeration getOtherRelationNames() {
        return otherRelNames.elements();
    }
}
