/* 
 * Program:             $RCSfile: F2SClass.java,v $  $Revision: 41.9 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: F2SClass.java,v 41.9 2003/02/15 11:56:39 yamamoto Exp yamamoto $
 */

package Japid.Freja.F2S;

import java.io.*;
import java.util.*;
import Japid.Freja.*;
import Japid.Freja.DBAccess.*;

public class F2SClass extends FrjClass implements F2SObject {
    // Constructors
    public F2SClass(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }

    public F2SClass() {
        super();
    }

    protected void create(FrejaEntry entry, DBClassObject obj) {
        new F2SClass(entry, obj);
    }

    public String toSource() {
        String source = "";
        sortRelatedObjects(new SortByOffset());

        if (sort != ANONYMOUS_CLASS) {
            source += getIndention();
            for (Enumeration e = getModifiers();
                 e.hasMoreElements();) {
                source += ((FrjModifier) e.nextElement()).getStr() + " ";
            }

            source += "class " + this.getStr() + " ";
            if (superClass != null) {
                source += "extends "
                    + ((F2SType) superClass).toSource() + " "; 
            }
                    
            Enumeration ei = getInterfaceTypes();
            if (ei.hasMoreElements()) {
                source += "implements ";
                String comma = "";
                while (ei.hasMoreElements()) {
                    source += comma + ((F2SType) ei.nextElement()).toSource();
                    comma = ", ";
                }
            }
        }

        source += "{\n";
        for (Enumeration e = getMemberClasses();
             e.hasMoreElements();) {
            F2SClass f2sClass = (F2SClass) e.nextElement();
            f2sClass.setIndentLevel(getIndentLevel() + 1);
            source += f2sClass.toSource() + "\n";
        }

        for (Enumeration e = getMemberInterfaces();
             e.hasMoreElements();) {
            F2SInterface f2sInterface = (F2SInterface) e.nextElement();
            f2sInterface.setIndentLevel(getIndentLevel() + 1);
            source += f2sInterface.toSource() + "\n";
        }

        for (Enumeration e = getMembers();
             e.hasMoreElements();) {
            source += getIndention();
            source += "    " + ((F2SVariable) e.nextElement()).toSource()
                + ";\n";
        }

        for (Enumeration e = getRoutines();
             e.hasMoreElements();) {
            F2SRoutine f2sRoutine = (F2SRoutine) e.nextElement();
            f2sRoutine.setIndentLevel(getIndentLevel() + 1);
            source += f2sRoutine.toSource() + "\n";
        }

        for (Enumeration e = getSBlocks();
             e.hasMoreElements();) {
            FrjSBlock sblock = (FrjSBlock) e.nextElement();
            if (sblock.getSort() == FrjSBlock.STATIC_INITIALIZER) {
                source += getIndention() + "    static ";
            }
            F2SStatement f2sStatement = (F2SStatement) sblock.getBody();
            f2sStatement.setIndentLevel(getIndentLevel() + 1);
            source += f2sStatement.toSource() + "\n";
        }

        source += getIndention() + "}";
        if(sort != ANONYMOUS_CLASS) {
            source += "\n";
        }
        return source;
    }

    private int indentLevel = 0;

    private int getIndentLevel() {
        return indentLevel;
    }

    void setIndentLevel(int indentLevel) {
        this.indentLevel = indentLevel;
    }

    private String getIndention() {
        String indent = "";
        int width = getIndentLevel() * INDENT_WIDTH;
        for(int i = 0; i < width; i++){
            indent += " ";
        }
        return indent;
    }
}
