/* 
 * Program:             $RCSfile: F2SExpression.java,v $  $Revision: 41.9 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: F2SExpression.java,v 41.9 2003/02/15 11:56:39 yamamoto Exp yamamoto $
 */

package Japid.Freja.F2S;

import java.io.*;
import java.util.*;
import Japid.Freja.*;
import Japid.Freja.DBAccess.*;

public class F2SExpression extends FrjExpression implements F2SObject {
    static int initializerLevel = 0;
    // distinguish listed expression (for init and array initializer)
    // array initializer needs '{' and '}'

    // Constructors
    public F2SExpression(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }

    public F2SExpression() {
        super();
    }

    protected void create(FrejaEntry entry, DBClassObject obj) {
        new F2SExpression(entry, obj);
    }

    public String toSource() {
        String source = "";
        sortRelatedObjects(new SortByOffset());
        if (sort == EMPTY) {
            return source;
        }
        else if (sort == REF_VAR) {
            source += getStr();
        }
        else if (sort == REF_ARRAY) {
            source += "REF ARRAY";
        }
        else if (sort == REF_LITERAL) {
            source += ((FrjLiteral) referree).getStr();
        }
        else if (sort == REF_TYPE) {
            source += ((F2SType) referree).toSource();
        }
        else if (sort == REF_SPECIAL) {
            source += ((FrjSpecial) referree).getStr();
        }
        else if (sort == METHOD_CALL) {
            source += getStr() + "(";

            String comma = "";
            for (Enumeration e = getChildExpressions();
                 e.hasMoreElements();) {
                F2SExpression f2sExpression = (F2SExpression) e.nextElement();
                f2sExpression.setIndentLevel(getIndentLevel() + 1);
                source += comma + f2sExpression.toSource();
                comma = ", ";
            }
            source += ")";
        }
        else if (sort == CONSTRUCTOR_CALL) {
            if (referree instanceof FrjSpecial) { // this or super
                source += referree.getStr();
            }
            else { // type
                source += ((F2SType) referree).toSource();
            }
            source += "(";
            String comma = "";
            for (Enumeration e = getChildExpressions();
                 e.hasMoreElements();) {
                source += comma + ((F2SExpression) e.nextElement()).toSource();
                comma = ", ";
            }
            source += ")";
        }
        else if (sort == OPERATION) {
            String firstSource = null, secondSource = null, thirdSource = null;
            if (firstOperand != null) {
                F2SExpression f2sExpression = (F2SExpression) get1stOperand();
                f2sExpression.setIndentLevel(getIndentLevel() + 1);
                firstSource = f2sExpression.toSource();
            }
            if (secondOperand != null) {
                secondSource = ((F2SExpression) get2ndOperand()).toSource();
            }
            if (thirdOperand != null) {
                thirdSource = ((F2SExpression) get3rdOperand()).toSource();
            }
            int opsort = operator.getSort();
            if (opsort == FrjOperator.PREFIX) {
                source += operator.getStr() + firstSource;
            }
            else if (opsort == FrjOperator.POSTFIX) {
                source += firstSource + operator.getStr();
            }
            else if (opsort == FrjOperator.NEW_ARRAY) { // new
                source += "new ";
                source += firstSource;  // type

                for (Enumeration e = getChildExpressions();
                     e.hasMoreElements();) {
                    source += "["
                        + ((F2SExpression) e.nextElement()).toSource()
                        + "]";
                }
            }
            else if (opsort == FrjOperator.NEW_OBJECT) { // new
                source += "new ";
                source += firstSource;  // constructor call
            }
            else if (opsort == FrjOperator.PAREN) {
                source += "(" + firstSource + ")";
            }
            else if (opsort == FrjOperator.BINARY) {
                source += firstSource;
                String opstr = operator.getStr();
                if (opstr.equals(".")) {
                    source += opstr;
                }
                else {
                    source += " " + opstr + " ";
                }
                source += secondSource;
            }
            else if (opsort == FrjOperator.REFER_DIM) { // array access
                source += firstSource;
                source += "[" + secondSource + "]";
            }
            else if (opsort == FrjOperator.CAST) { // cast
                source += "(" + firstSource +  ")";
                source += secondSource;
            }
            else if (opsort == FrjOperator.TERNARY) { // cast
                source += firstSource + "? ";
                source += secondSource + ": " + thirdSource;
            }
        }
        else if (sort == REF_QNAME) {
            if (referree != null) {
                source += ((F2SQName) referree).toSource();
            }
            else {
                source += "NULL QNAME";
            }
        }
        else if (sort == LISTED) {
            String comma = "";
            if (initializerLevel > 0) {
                source += " {";
            }
            for (Enumeration e = getChildExpressions();
                 e.hasMoreElements();) {
                source += comma
                    + ((F2SExpression) e.nextElement()).toSource();
                comma = ", ";
            }
            if (initializerLevel > 0) {
                source += "} ";
            }

        }
        else if (sort == DIM_EXPRESSION) {
            if (firstOperand != null) {
                source += ((F2SExpression) firstOperand).toSource();
            }
        }
        else if (sort == ANONYMOUS_CLASS) {
            F2SType type = (F2SType) referree;
            source += type.toSource();
            source += " (";
            String comma = "";
            for (Enumeration e = getChildExpressions();
                 e.hasMoreElements();) {
                source += comma + ((F2SExpression) e.nextElement()).toSource();
                comma = ", ";
            }
            source += ") ";
            F2SClass c = (F2SClass) getAnonymousClass();
            c.setIndentLevel(getIndentLevel() + 1);
            source += c.toSource();
        }
        else if (sort == ANONYMOUS_ARRAY) {
            source += "new ";
            source += ((F2SExpression) firstOperand).toSource();
            // REF_TYPE Expresssion
            initializerLevel++;
            source += ((F2SExpression) secondOperand).toSource();
            // Listed Expresssion
            initializerLevel--;
        }
        else if (sort == REF_ARRAY_LENGTH) {
            source += "length";
        }
        else {
            FrjWorld.error("No such Expression sort " + this);
        }
        return source;
    }

    private int indentLevel = 0;

    private int getIndentLevel() {
        return indentLevel;
    }

    void setIndentLevel(int indentLevel) {
        this.indentLevel = indentLevel;
    }

    private String getIndention() {
        String indent = "";
        int width = getIndentLevel() * INDENT_WIDTH;
        for(int i = 0; i < width; i++){
            indent += " ";
        }
        return indent;
    }
}
