/* 
 * Program:             $RCSfile: F2SStatement.java,v $  $Revision: 41.13 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: F2SStatement.java,v 41.13 2003/02/15 11:56:40 yamamoto Exp yamamoto $
 */

package Japid.Freja.F2S;

import java.io.*;
import java.util.*;
import Japid.Freja.*;
import Japid.Freja.DBAccess.*;

public class F2SStatement extends FrjStatement implements F2SObject {
    // Constructors
    public F2SStatement(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }

    public F2SStatement() {
        super();
    }

    protected void create(FrejaEntry entry, DBClassObject obj) {
        new F2SStatement(entry, obj);
    }

    public String toSource() {
        String source = "";
        sortRelatedObjects(new SortByOffset());

        for (Enumeration e = getLabels();
             e.hasMoreElements();) {
            FrjLabel label = (FrjLabel) e.nextElement();
            if (label.getSort() != FrjLabel.REFERRED) {
                F2SLabel f2sLabel = (F2SLabel) label;
                f2sLabel.setIndentLevel(getIndentLevel());
                source += f2sLabel.toSource();
            }
        }
                        
        if (sort == EMPTY) {
            source += getIndention() + ";\n";
        }
        else if (sort == EXPRESSION) {
            Enumeration e = getExpressions();
            source += getIndention();
            F2SExpression f2sExpression = (F2SExpression) e.nextElement();
            f2sExpression.setIndentLevel(getIndentLevel() + 1);
            source += f2sExpression.toSource();
            source += ";\n";
        }
        else if (sort == BLOCK) {
            source += "{\n";
            setIndentLevel(getIndentLevel() + 1);
            source += childrenToSource();
            setIndentLevel(getIndentLevel() - 1);
            source += getIndention() + "}\n";
        }
        else if (sort == DEF_VAR) {
            for (Enumeration e = getLocalVariables();
                 e.hasMoreElements();) {
                source += getIndention();
                source += ((F2SVariable) e.nextElement()).toSource() + ";\n";
            }
        }
        else if (sort == IF) {
            source += getIndention() + "if (";
            Enumeration e = getExpressions();
            source += ((F2SExpression) e.nextElement()).toSource();
            source += ") ";

            e = getChildStatements();

            // THEN part
            F2SStatement f2sStatementThen = (F2SStatement) e.nextElement();
            f2sStatementThen.setIndentLevel(getIndentLevel());
            source += f2sStatementThen.toSource();
            if (e.hasMoreElements()) { // ELSE part
                source += getIndention() + "else ";
                F2SStatement f2sStatementElse = (F2SStatement) e.nextElement();
                f2sStatementElse.setIndentLevel(getIndentLevel());
                source += f2sStatementElse.toSource();
            }
        }
        else if (sort == SWITCH) {
            source += getIndention() + "switch (";
            Enumeration e = getExpressions();
            source += ((F2SExpression) e.nextElement()).toSource();
            source += ") ";
            source += childrenToSource();
        }
        else if (sort == WHILE) {
            source += getIndention() + "while (";
            Enumeration e = getExpressions();
            source += ((F2SExpression) e.nextElement()).toSource();
            source += ") ";
            source += childrenToSource();
        }
        else if (sort == DO) {
            source += getIndention() + "do ";
            source += childrenToSource();
            source += getIndention() + "while (";
            Enumeration e = getExpressions();
            source += ((F2SExpression) e.nextElement()).toSource();
            source += ");\n";
        }
        else if (sort == FOR) {
            String semicolon = "";
            source += getIndention() + "for (";
            String comma = "";
            Enumeration vars = getLocalVariables();
            if (vars.hasMoreElements()) {
                source += F2SVariable.toSource(vars);
                semicolon = "; ";
            }
            for (Enumeration e = getExpressions();
                 e.hasMoreElements();) {
                source += semicolon
                    + ((F2SExpression) e.nextElement()).toSource();
                semicolon = "; ";
            }
            source += ") ";
            source += childrenToSource();
        }
        else if (sort == BREAK) {
            source += getIndention() + "break";
            for (Enumeration e = getLabels();
                 e.hasMoreElements();) {
                FrjLabel label = (FrjLabel) e.nextElement();
                if (label.getSort() == FrjLabel.REFERRED) {
                    source += " " + label.getStr();
                }
            }
            source += ";\n";
        }
        else if (sort == CONTINUE) {
            source += getIndention() + "continue";
            for (Enumeration e = getLabels();
                 e.hasMoreElements();) {
                FrjLabel label = (FrjLabel) e.nextElement();
                if (label.getSort() == FrjLabel.REFERRED) {
                    source += " " + label.getStr();
                }
            }
            source += ";\n";
        }
        else if (sort == RETURN) {
            source += getIndention() + "return";
            Enumeration e = getExpressions();
            if (e.hasMoreElements()) {
                source += " " + ((F2SExpression) e.nextElement()).toSource();
            }
            source += ";\n";
        }
        else if (sort == THROW) {
            source += getIndention() + "throw ";
            Enumeration e = getExpressions();
            source += ((F2SExpression) e.nextElement()).toSource();
            source += ";\n";
        }
        else if (sort == EXCEPTION) {
            source += childrenToSource();
        }
        else if (sort == TRY) {
            source += getIndention() + "try ";
            source += childrenToSource();
        }
        else if (sort == CATCH) {
            source += getIndention() + "catch (";
            Enumeration e = getLocalVariables();
            source += ((F2SVariable) e.nextElement()).toSource();
            source += ") ";
            source += childrenToSource();
        }
        else if (sort == FINALLY) {
            source += getIndention() + "finally ";
            source += childrenToSource();
        }
        else if (sort == SYNCHRONIZED) {
            source += getIndention() + "synchronized (";
            Enumeration e = getExpressions();
            source += ((F2SExpression) e.nextElement()).toSource();
            source += ") ";
            source += childrenToSource();
        }
        else if (sort == SWITCH_LABELED_EMPTY) {
            // do nothing
            // SWITCH_LABELED_EMPTY statement is a dummy statement
        }
        else if (sort == DEF_LOCAL_CLASS) {
            F2SClass f2sClass = (F2SClass) localClass;
            f2sClass.setIndentLevel(getIndentLevel() + 1);
            source += f2sClass.toSource();
        }
        else if (sort == ASSERT) {
            String semicolon = "";
            source += getIndention() + "assert ";
            Enumeration e = getExpressions();
            F2SExpression expr = (F2SExpression) e.nextElement();
            source += expr.toSource();
            if (e.hasMoreElements()) {
                expr =(F2SExpression) e.nextElement();
                source += ": " + expr.toSource();
            }
            source += ";\n";
        }

        else {
            FrjWorld.error("No such Statement sort " + this);
        }
        return source;
    }

    private String childrenToSource() {
        String source = "";
        for (Enumeration e = getChildStatements();
             e.hasMoreElements();) {
            F2SStatement f2sStatement = (F2SStatement) e.nextElement();
            f2sStatement.setIndentLevel(getIndentLevel());
            source += f2sStatement.toSource();
        }
        return source;
    }

    private int indentLevel = 0;

    private int getIndentLevel() {
        return indentLevel;
    }

    void setIndentLevel(int indentLevel) {
        this.indentLevel = indentLevel;
    }

    private String getIndention() {
        String indent = "";
        int width = getIndentLevel() * INDENT_WIDTH;
        for(int i = 0; i < width; i++){
            indent += " ";
        }
        return indent;
    }
}
