/* 
 * Program:             $RCSfile: FrjClass.java,v $  $Revision: 41.7 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjClass.java,v 41.7 2003/02/12 03:39:31 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjClass extends FrjClassInterface {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Class";

    /**
     * Table from a sort to string. Initialized by static block.
     */
    protected static String     sortNameTable[];

    // Attribute.
    public int                  sort;

    // Related objects.
    protected Vector            sblocks = null;
    protected FrjType           superClass = null;
    protected Vector            interfaces = null;
    protected Vector            arguments = null;
    protected FrjStatement      definingStatement = null;

    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;

    public final static int
        PACKAGE_MEMBER = 0,
        NESTED_TOPLEVEL = 1,
        MEMBER_CLASS = 2,
        LOCAL_CLASS = 3,
        ANONYMOUS_CLASS = 4;
    protected final static int
        SORT_FIRST_MEMBER = PACKAGE_MEMBER,
        SORT_LAST_MEMBER = ANONYMOUS_CLASS;

    // Constructors.
    public FrjClass(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }

    public FrjClass(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }

    public FrjClass() {
        super();
        sort = INVALID_ID;
    }

    // Class methods.
    /** Return the class name of this object. */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** Create object from SDB. */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjClass(entry, obj);
    }

    // Instance methods.
    /** Return the class name dynamically. */
    public String getClassName() {
        return CLASS_NAME;
    }

    /** Return the sort. */
    public int getSort() {
        return sort;
    }

    /** Check whether this object has a valid sort or not. */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }

    /** Return the sort as string. */
    public String getSortString() {
        // if (!hasValidSort()) {
        //     return "INVALID SORT";
        // }
        // return sortNameTable[sort];

        return (hasValidSort() ? sortNameTable[sort] : "INVALID SORT");
    }

    /** Return the name of attribute and its value as string. */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** Convert object to string. */
    public String toString() {
        // String res = super.toString();
        // res += ", sort: " + getSortString();
        // return res;

        return (super.toString() + ", sort: " + getSortString());
    }

    protected String toSdbline() {
        // String res = super.toSdbline();
        // res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        // return res;

        return (super.toSdbline() + getSort() + JDBConsts.SDB_ATTR_SEPARATOR);
    }

    // About relations.
    public void setRelatedObjects() {
        memberClasses = new Vector();
        memberInterfaces = new Vector();
        routines = new Vector();
        this.members = new Vector();
        modifiers = new Vector();
        interfaces = new Vector();
        arguments = new Vector();
        sblocks = new Vector();
        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        file = null;
        memberClasses = null;
        memberInterfaces = null;
        routines = null;
        this.members = null;
        modifiers = null;
        interfaces = null;
        arguments = null;
        sblocks = null;
        superClass = null;
        outerClass = null;
        outerInterface = null;
        definingStatement = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.DEFINE_CLASS)) {
            file = (FrjFile) obj;
        } else if (relName.equals(FrjRelations.EXTEND_CLASS)) {
            superClass = (FrjType) obj;
            superClassInterfaceTypes.addElement(obj);
        } else if (relName.equals(FrjRelations.IMPLEMENT_INTERFACE)) {
            interfaces.addElement((FrjType) obj);
            superClassInterfaceTypes.addElement(obj);
        } else if (relName.equals(FrjRelations.CLASS_ROUTINE)) {
            routines.addElement((FrjRoutine) obj);
        } else if (relName.equals(FrjRelations.CLASS_VARIABLE)) {
            members.addElement((FrjVariable) obj);
        } else if (relName.equals(FrjRelations.CLASS_MODIFIER)) {
            modifiers.addElement((FrjModifier) obj);
        } else if (relName.equals(FrjRelations.CLASS_SBLOCK)) {
            sblocks.addElement((FrjSBlock) obj);
        } else if (relName.equals(FrjRelations.CLASS_ARGUMENT)) {
            arguments.addElement((FrjExpression) obj);
        } else if (relName.equals(FrjRelations.CLASS_CLASS)) {
            if (rel.get1stObject().getID() == this.getID()) { // this is outer
                memberClasses.addElement((FrjClass) obj);
            }
            else {      // this is member 
                outerClass = (FrjClass) obj;
            }
        } else if (relName.equals(FrjRelations.CLASS_INTERFACE)) {
            memberInterfaces.addElement((FrjInterface) obj);
        } else if (relName.equals(FrjRelations.INTERFACE_CLASS)) {
            outerInterface = (FrjInterface) obj;
        } else if (relName.equals(FrjRelations.TYPE_CLASS)) {
            types.addElement(obj);
        } else if (relName.equals(FrjRelations.LOCAL_CLASS)) {
            definingStatement = (FrjStatement) obj;
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(memberClasses);
        sorter.sort();
        sorter.setVector(memberInterfaces);
        sorter.sort();
        sorter.setVector(routines);
        sorter.sort();
        sorter.setVector(members);
        sorter.sort();
        sorter.setVector(modifiers);
        sorter.sort();
        sorter.setVector(interfaces);
        sorter.sort();
        sorter.setVector(sblocks);
        sorter.sort();
        sorter.setVector(arguments);
        sorter.sort();
    }

    /** Traverses <I>class_sblock</I>. */
    public Enumeration getSBlocks() {
        return sblocks.elements();
    }
    /** Traverses <I>class_arguments</I>. */
    public Enumeration getArguments() {
        return arguments.elements();
    }

    /**
     * Traverses <I>extend_class</I>.
     * @return the super class (FrjType)
     */
    public FrjType getSuperClassType() {
        return superClass;
    }
    /**
     * Traverses <I>extend_class</I>.
     * @return sub classes<P>
     *  NOTE: This doesn't return all sub classes.
     */
    public Enumeration getSubClasses() {
        Vector subClasses = new Vector();
        for (Enumeration e = getTypes();
             e.hasMoreElements();) {
            Object o = e.nextElement();
            if (o instanceof FrjType) {
                FrjType type = (FrjType) o;
                FrjObject pt = type.getParentObject();
                if (pt instanceof FrjClass) {
                    subClasses.addElement(pt);
                }
            }
        }
        return subClasses.elements();
    }

    /** Traverses <I>implement_interface</I>. */
    public Enumeration getInterfaceTypes() {
        return interfaces.elements();
    }
    /** Traverses <I>class_type</I>. */
    public Enumeration getTypes() {
        if (types.size() == 0) {
            FrjType type = new FrjType(entry, this.str);
            FrjRelation rel = new FrjRelation(FrjRelations.TYPE_CLASS,
                                              this, type);
            resetRelatedObjects();
            type.setRelatedObjects();
        }
        return types.elements();
    }

    /** Return parent object if exists. */
    public FrjObject getParentObject() {
        if (file != null) {
            return file;
        } else if (outerClass != null) {
            return outerClass;
        } else if (outerInterface != null) {
            return outerInterface;
        } else if (definingStatement != null) {
            return definingStatement;
        } else {
            return null;
        }
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjClass) obj).sort = sort;
    }

    // Traverse methods.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        traverseObjects(getModifiers(), tm);

        tm.doMyself(this);

        if (superClass != null) {
            superClass.traverse(tm);
        }
        traverseObjects(getArguments(), tm);
        traverseObjects(getInterfaceTypes(), tm);
        traverseObjects(getMembers(), tm);
        traverseObjects(getRoutines(), tm);
        traverseObjects(getMemberClasses(), tm);
        traverseObjects(getMemberInterfaces(), tm);
        traverseObjects(getSBlocks(), tm);

        tm.doAfterTraverse(this);
    }

    static {
        sortNameTable = new String[SORT_LAST_MEMBER + 1];

        sortNameTable[PACKAGE_MEMBER] = "Pakage Member";
        sortNameTable[NESTED_TOPLEVEL] = "Nested Top Level";
        sortNameTable[MEMBER_CLASS] = "Member Class";
        sortNameTable[LOCAL_CLASS] = "Local Class";
        sortNameTable[ANONYMOUS_CLASS] = "Anonymous Class";
    }
}
