/* 
 * Program:             $RCSfile: FrjComment.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjComment.java,v 41.5 2003/02/15 11:56:27 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjComment extends FrjObject {
    // Class variables
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Comment";

    // Related objects
    protected FrjObject         symbol = null;

    // Constructors
    public FrjComment(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }

    public FrjComment() {
        super();
    }

    // Class methods
    /** return the class name of this object */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjComment(entry, obj);
    }

    // Instance method
    /** return the class name dynamically */
    public String getClassName() {
        return CLASS_NAME;
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        symbol = null;
    }
                
    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        }
        else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    protected void setRelatedComment(FrjObject obj) {
        symbol = obj;
    }

    public void sortRelatedObjects(VectorSorter sorter) {
    }

    /**
     * Returns whether this comment is javadoc comment.
     * NOTE: Checks only whether comment starts with "/**",
      * does not check the place of comment.
      * If there is following code,<P>
      *  <PRE> /** javadoc comment ? * / a = 1 + 2; </PRE>
       * isJavadocComment() returns true.
       *
       * @return: true if the comment starts with "/**"
        */
    public boolean isDocComment() {
        return getStr().startsWith("/**");
    }

    /** Return parent object if exists. */
    public FrjObject getParentObject() {
        return getSymbol();
    }

    public FrjObject getSymbol() {
        return symbol;
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        tm.doMyself(this);
        tm.doAfterTraverse(this);
    }

    /** Sets symbol to <I>obj</I>. */
    public void setSymbol(FrjObject obj) {
        FrjRelation rel
            = new FrjRelation(FrjRelations.SYMBOL_COMMENT, obj, this);
        try {
            this.setRelatedObject(obj, rel);
            obj.setRelatedObject(this, rel);
        }
        catch (NoSuchRelatedObjectException e) {
            FrjWorld.warning(e.getMessage());
        }
    }
}
