/* 
 * Program:             $RCSfile: FrjExpression.java,v $  $Revision: 41.16 $
 * 
 * Purpose:             The class corresponding to Expression.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjExpression.java,v 41.16 2003/02/15 11:56:27 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjExpression extends FrjObject {
    // Class variables.
    /** A class name of this class */
    public final static String  CLASS_NAME = "Expression";

    /** table from a sort to string */
    // initialized by static block
    protected static String     sortNameTable[];
    // initialized by static block
    protected static String     lrsortNameTable[];

    // Attributes.
    public int                          sort;
    public int                          lrsort;

    // Related objects
    protected FrjStatement      statement = null;
    protected FrjExpression     parentExpression = null;
    protected FrjLabel          caseLabel = null;
    protected FrjVariable       initialized = null;
    protected Vector            childExpressions = null;
    protected FrjExpression     firstOperand = null;
    protected FrjExpression     secondOperand = null;
    protected FrjExpression     thirdOperand = null;
    protected FrjOperator       operator = null;
    protected FrjObject         referree = null;        // this refers to referree
    protected FrjRoutine        routine = null;
    protected Vector            methods = null;
    protected FrjClass          argClass = null;
    // anonymous class of class_argument

    protected FrjType           myType = null;          // has_type

    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START,
        LRSORT_INDEX = SORT_INDEX +1;
    protected static int                _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int             EMPTY = _i++;
    public final static int             REF_VAR = _i++;
    public final static int             REF_ARRAY = _i++;
    public final static int             REF_LITERAL = _i++;
    public final static int             REF_TYPE = _i++;
    public final static int             REF_SPECIAL = _i++;
    public final static int             METHOD_CALL = _i++;
    public final static int             CONSTRUCTOR_CALL = _i++;
    public final static int             OPERATION = _i++;
    public final static int             REF_QNAME = _i++;
    public final static int             LISTED = _i++;
    public final static int             DIM_EXPRESSION = _i++;
    public final static int             ANONYMOUS_CLASS = _i++;
    public final static int             ANONYMOUS_ARRAY = _i++;
    public final static int             REF_ARRAY_LENGTH = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // sentinel

    protected static int _j = 0;
    protected final static int  LRSORT_FIRST_MEMBER = _j;       // sentinel
    public final static int             NONE = _j++;
    public final static int             L_VALUE = _j++;
    public final static int             R_VALUE = _j++;
    public final static int             LR_VALUE = _j++;
    public final static int             OBJECT_OPERATION = _j++;
    protected final static int  LRSORT_LAST_MEMBER = --_j;      // sentinel
        
    public static boolean debug = false;

    // Constructors
    public FrjExpression(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
        // this.lrsort = Integer.parseInt(attrs[LRSORT_INDEX]);
    }
    public FrjExpression(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }
    public FrjExpression() {
        super();
        sort = INVALID_ID;
    }

    // Class methods.
    /** return the class name of this object statically */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjExpression(entry, obj);
    }

    // Instance methods.
    /** return the class name of this object */
    public String getClassName() {
        return CLASS_NAME;
    }

    // about relation
    /** return the sort */
    public int getSort() {
        return sort;
    }
    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** return the lrsort */
    public int getLRSort() {
        return lrsort;
    }
    /** check this object has a valid lrsort */
    public boolean hasValidLRSort() {
        return (LRSORT_FIRST_MEMBER <= lrsort
                && lrsort <= LRSORT_LAST_MEMBER);
    }
    /** return the lrsort as string */
    public String getLRSortString() {
        if (!hasValidLRSort()) {        /* error */
            return "INVALID LRSORT";
        }
        return lrsortNameTable[lrsort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        attrs[LRSORT_INDEX] = new FrjAttrsAsString("lrsort",
                                                   getLRSortString());
        return attrs;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        res += ", lrsort: " + getLRSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        res += getLRSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // about relation
    public void setRelatedObjects() {
        childExpressions = new Vector();
        methods = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        statement = null;
        parentExpression = null;
        caseLabel = null;
        initialized = null;
        childExpressions = null;
        firstOperand = null;
        secondOperand = null;
        thirdOperand = null;
        operator = null;
        referree = null;        // this refers to referree
        routine = null;
        methods = null;
        argClass = null;
        myType = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.STATEMENT_EXPRESSION)) {
            statement = (FrjStatement) obj;
        } else if (relName.equals(FrjRelations.INITIALIZE_VARIABLE)) {
            initialized = (FrjVariable) obj;
        } else if (relName.equals(FrjRelations.CASE_VALUE)) {
            caseLabel = (FrjLabel) obj;
        } else if (relName.equals(FrjRelations.EXPRESSION_OPERAND)) {
            if (rel.get1stObject().getID() == this.getID()) { // obj is a child
                setOperand((FrjExpression) obj);
            } else {      // obj is a parent
                parentExpression = (FrjExpression) obj;
            }
        } else if (relName.equals(FrjRelations.EXPRESSION_EXPRESSION)) {
            if (rel.get1stObject().getID() == this.getID()) { // obj is a child
                childExpressions.addElement(obj);
            } else {      // obj is a parent
                parentExpression = (FrjExpression) obj;
            }
        } else if (relName.equals(FrjRelations.EXPRESSION_OPERATOR)) {
            operator = (FrjOperator) obj;
        } else if (relName.equals(FrjRelations.REFER_VARIABLE)
                 ||     relName.equals(FrjRelations.REFER_LITERAL)
                 ||     relName.equals(FrjRelations.REFER_SPECIAL)
                 ||     relName.equals(FrjRelations.REFER_TYPE)
                 ||     relName.equals(FrjRelations.REFER_QNAME)) {
            if (referree != null) {
                String msg = "Expression already referes to other object.\n";
                msg += "  this: " + this + "\n";
                msg += "  old: " + referree + "\n";
                msg += "  new: " + obj;
                throw new NoSuchRelatedObjectException(msg);                            
            } else {
                referree = obj;
            }
        } else if (relName.equals(FrjRelations.HAS_TYPE)) {
            if (myType != null) {
                String msg = "Expression already referes to other object.\n";
                msg += "  this: " + this + "\n";
                msg += "  old: " + myType + "\n";
                msg += "  new: " + obj;
                throw new NoSuchRelatedObjectException(msg);                            
            } else {
                myType = (FrjType) obj;
            }
        } else if (relName.equals(FrjRelations.INVOKE_ROUTINE)) {
            if (routine != null) {
                String msg = "Expression already referes to other object.\n";
                msg += "  this: " + this + "\n";
                msg += "  old: " + routine + "\n";
                msg += "  new: " + obj;
                throw new NoSuchRelatedObjectException(msg);                            
            } else {
                routine = (FrjRoutine) obj;
            }
        } else if (relName.equals(FrjRelations.CLASS_ARGUMENT)) {
            argClass = (FrjClass) obj;
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(childExpressions);
        sorter.sort();
    }

    /** Traverses <I>statement_expression</I>. */
    public FrjStatement getStatement() {
        return statement;
    }
    /** Traverses <I>expression_expression</I>. */
    public FrjExpression getParentExpression() {
        return parentExpression;
    }
    /** Traverses <I>variable_initializer</I>. */
    public FrjVariable getInitializedVariable() {
        return initialized;
    }
    /** Traverses <I>case_value</I>. */
    public FrjLabel getCaseLabel() {
        return caseLabel;
    }
    /** Traverses <I>expression_expression</I>. */
    public Enumeration getChildExpressions() {
        return childExpressions.elements();
    }
    /** Traverses <I>expression_operands</I>. */
    public Enumeration getOperands() {
        Vector v = new Vector();
        if (firstOperand != null) {
            v.addElement(firstOperand);
        }
        if (secondOperand != null) {
            v.addElement(secondOperand);
        }
        if (thirdOperand != null) {
            v.addElement(thirdOperand);
        }
        return v.elements();
    }
    /** Returns the first operand. */
    public FrjExpression get1stOperand() {
        return firstOperand;
    }
    /** Returns the second operand. */
    public FrjExpression get2ndOperand() {
        return secondOperand;
    }
    /** Returns the third operand. */
    public FrjExpression get3rdOperand() {
        return thirdOperand;
    }
    /** Traverses <I>expression_operator</I>. */
    public FrjOperator getOperator() {
        return operator;
    }
    /**
     * Traverses <I>refer_variable</I> or <I>refer_literal</I>
     * or <I>refer_special</I> or <I>refer_type</I> or <I>refer_qname</I>.
     */
    public FrjObject getReferree() {
        return referree;
    }

    /**
     * Traverses <I>invoke_routine</I> lexically.
     * Gotten routine is not always invoked, because of method overriding.
     * If you get methods may be invoked, use getMethods().
     */
    public FrjRoutine getRoutine() {
        if (routine == null) {
            setInvokeRoutine();
        }
        return routine;
    }

    protected void addMethod(FrjRoutine m) {
        if (!methods.contains(m)) {
            methods.addElement(m);
        }
    }

    /**
     * Traverses <I>invoke_routine</I> semantically.
     * Returns methods may be invoked.
     * You must call InvocationAnalyzer#analyze() before calling this method.
     */
    public Enumeration getMethods() {
        return methods.elements();
    }

    private void setOperand(FrjExpression operand) {
        if (firstOperand == null) {
            firstOperand = operand;
        } else if (secondOperand == null) {
            if (!firstOperand.hasValidOffset()) {
                                /* error */
            } else if (firstOperand.getOffset() < operand.getOffset()) {
                secondOperand = operand;
            } else {      // swap
                secondOperand = firstOperand;
                firstOperand = operand;
            }
        } else {
            if (!firstOperand.hasValidOffset()
                || !secondOperand.hasValidOffset()) {
                                /* error */
            } else if (secondOperand.getOffset() < operand.getOffset()) {
                thirdOperand = operand;
            } else {
                thirdOperand = secondOperand;
                if (firstOperand.getOffset() < operand.getOffset()) {
                    // swap second 
                    secondOperand = operand;
                } else {  // swap first
                    secondOperand = firstOperand;
                    firstOperand = operand;
                }
            }
        }
    }

    /**
     * Traverses <I>ref_type</I> and <I>type_class</I>.
     */
    public FrjClass getAnonymousClass() {
        if (sort != ANONYMOUS_CLASS) {
            return null;
        }
        /* else */
        FrjType type = (FrjType) referree;
        if (type == null) {
            FrjWorld.warning("FrjExpression is ANONYMOUS_CLASS,"
                                    + "but it has no class.");
            return null;
        }
        /* else */
        return (FrjClass) type.getReferringClassInterface();
    }

    /**
     * Traverses <I>class_argument</I>
     */
    public FrjClass getArgClass() {
        return argClass;
    }


    /**
     * Traverses <I>has_type</I>.
     * NOTE: It doens't travers <I>refer_type</I>.
     * And, you must call FrjExpression#setType() before calling this method.
     */
    public FrjType getType() {
        return myType;
    }

    /**
     * Set type of expression(<I>has_type</I>).
     * This method assumes that type of child expressions is already set.
     * if expression is method call, this also call setInvokeMethod().
     */
    public void setType() {
        if (myType != null) {   // already set
            return;
        }
                
        if (sort == REF_VAR) {
            resolveReferringVariable();
            if (referree != null) {
                                // type of refering variable
                myType = ((FrjVariable) referree).getType();
            }
        } else if (sort == REF_LITERAL) {
            int lsort = ((FrjLiteral) referree).getSort();
            if (lsort == FrjLiteral.CHAR || lsort == FrjLiteral.UNICODE) {
                myType = new FrjType(entry, FrjType.CHAR);
            } else if (lsort == FrjLiteral.STRING) {
                myType = new FrjType(entry, "String", "java.lang");
            } else if (lsort == FrjLiteral.INT) {
                myType = new FrjType(entry, FrjType.INT);
            } else if (lsort == FrjLiteral.DOUBLE) {
                myType = new FrjType(entry, FrjType.DOUBLE);
            } else {
                FrjWorld.error("LITERAL is crazy");
            }
        } else if (sort == REF_TYPE) {
            // instanceof, cast, new
            myType = (FrjType) referree;
        } else if (sort == REF_SPECIAL) {
            setSpecialType();
        } else if (sort == METHOD_CALL) {
            this.setInvokeMethod();
            if (routine == null) {
                myType = null;
            } else {
                myType = routine.getReturnType();
            }
        } else if (sort == CONSTRUCTOR_CALL) {
            this.setInvokeConstructor();
        } else if (sort == OPERATION) {
            setOperationType();
        } else if (sort == REF_QNAME) {
            setQNameType();
            return;
        } else if (sort == LISTED) {
            // child expressions' type must be already set
            myType = null;
        } else if (sort == DIM_EXPRESSION) {
            myType = null;
        } else if (sort == REF_ARRAY_LENGTH) {
            // this must be set at sort == REF_VAR
            // myType = new FrjType(entry, FrjType.INT);
        }

        if (myType != null) {
            if (myType.getFrejaEntry() != entry) {
                                // other Freja's FrjType
                FrjType newType = new FrjType(entry, myType.getStr(),
                                              myType.getPackageName());
                FrjType nt = newType;
                for (FrjType t = myType.getChildType();
                     t != null; t = t.getChildType()) {
                    FrjType ct = new FrjType(entry, t.getStr(),
                                             t.getPackageName());
                    FrjRelation r
                        = new FrjRelation(FrjRelations.TYPE_TYPE, nt, ct);
                    nt = ct;
                }
                myType = newType;
            }
            FrjRelation rel
                = new FrjRelation(FrjRelations.HAS_TYPE, this, myType);
            myType.resetRelatedObjects();
        }
    }


    /**
     * Sets myType to type if myType equals null.
     * This method is invoked by setSpecialType().
     */
    private void setType(FrjType type) {
        if (myType != null) {
            this.myType = type;
            FrjRelation rel
                = new FrjRelation(FrjRelations.HAS_TYPE, this, myType);
            myType.resetRelatedObjects();
        }
    }           

    private void setOperationType() {
        if (myType != null || this.sort != OPERATION) {
            return;
        }
        if (thirdOperand != null) {     // ?:
            myType = secondOperand.getType();
        } else if (secondOperand != null) {
            FrjType type1 = firstOperand.getType();
            FrjType type2 = secondOperand.getType();
            if (operator.isDot()) {
                // secondOperand().getStr() may return null.
                if (type1 == null) {
                    myType = null;
                } else if ("length".equals(secondOperand.getStr()) &&
                           type1.getSort() == FrjType.ARRAY) {
                    // "array.length".
                    myType = new FrjType(entry, FrjType.INT);
                    secondOperand.myType = this.myType;
                    secondOperand.sort = REF_ARRAY_LENGTH;
                } else {
                    myType = type2;
                }
            } else if (type1 == null) {
                myType = null;
            } else if ((operator.getSort()) == FrjOperator.REFER_DIM) {
                myType = type1.getChildType();
            } else if (operator.isAssignment()) {
                                // assignment expression
                myType = type1;
            } else if (operator.isConditional()) {
                                // conditional expression
                myType = new FrjType(entry, FrjType.BOOLEAN);
            } else if (operator.isBitwise()) {
                if (type1.getStr() == FrjType.BOOLEAN) {
                    // logical operator
                    myType = type1;
                } else if (type2.getStr() == FrjType.BOOLEAN) {
                    // logical operator
                    myType = type2;
                } else { // bitwise operator
                    myType = type1;
                }
            } else if ((operator.getSort()) == FrjOperator.CAST) {
                myType = type1;
            } else {
                try {
                    myType = type1.convert(type2);
                }
                catch (TypesNotConvertableException e) {
                    FrjWorld.warning((e.getMessage()));
                    myType = null;
                }
                catch (Exception e) {
                    e.printStackTrace();
                    myType = null;
                }
            }
        } else if (firstOperand != null) {
            myType = firstOperand.getType();
        } else {
            FrjWorld.error("No operand is set in " + this);
        }
    }
                
    private void setQNameType() {
        /*
         * QName is
         * 1) single-qname class (e.g. FrjWorld)
         * 2) dot-trailed class (e.g. Japid.Freja.FrjWorld)
         * 3) dot-trailed class and variable
         *      (e.g. System.out, java.lang.System.out)
         * 4) super class's variable (e.g. str, FrjObject#str)
         */
        if (myType != null || this.sort != REF_QNAME) {
            return;
        }

        if (!FrjWorld.isAutoLoading()) {
            return;
        }

        FrjFile file = 
            (FrjFile)getAncestorObject(FrjFile.getClassNameStatically());
        if (file == null) {
            FrjWorld.vwarning(this + ": Expression has no parent.\n");
            return;
        }

        FrjQName qname= (FrjQName)referree;
        Vector varQNames = new Vector();

        /* Search for entry. */
        while (qname != null) {
            String fullQNameString = qname.getFullString();
            FrejaEntry entry = Creja.getFrejaEntry(file.getDefinedPackage(),
                                                   file.getImportedPackages(),
                                                   fullQNameString);
            if (entry == null) {
                VectorSorter sorter = new SortByOffset(qname.getChildQNames());
                sorter.sort();
                Enumeration e = sorter.getEnumeration();
                if (e.hasMoreElements()) {
                    qname = (FrjQName) e.nextElement(); // car
                    varQNames.insertElementAt((FrjQName) e.nextElement(), 0);
                    // cdr, add first
                } else {
                    qname = null;       // break
                }
                continue;
            }

            /* else */
            // Style 1) or 2) or 3),  QName is a class or interface
            if (FrjExtents.isLoaded(entry)) {
                entry = FrjExtents.getLoadedFrejaEntry(entry);
            } else {
                FrjWorld.load(entry);
            }
            FrjWorld.debug(debug,
                                  "QName referring class " + fullQNameString);

            // convert qnames referring type to expression
            //  e.g. Package.Class.mem1.mem2.mem3
            //       ^^^^^^^^^^^^^
            // System.out.println();
            // Expr(OP)a +-- Expr(REF_QNAME)b --ref_qname-- QName(System.out)c
            //           +-- Expr(INVOKE_MTD)d
            //           +-- Operator(.)e
            // ===>
            // Expr(OP)a +-- Expr(OP)g +-- Expr(REF_TYPE)b
            //           |             |       +--ref_type-- Type(System)c'
            //           |             +-- Expr(REF_VAR)f
            //           |                     +--ref_var-- Variable(out)c'
            //           +-- Expr(INVOKE_MTD)d
            //           +-- Operator(.)e

            FrjQName pkgQName = null;
            FrjQName typeQName = null;
            String pname = null;
            FrjOperator pkgTypeDot = qname.getDot();
            if (pkgTypeDot == null) { // simple Class
                typeQName = qname;
            } else {  // including package
                VectorSorter sorter = new SortByOffset(qname.getChildQNames());
                sorter.sort();
                Enumeration e = sorter.getEnumeration();
                if (e.hasMoreElements()) {
                    pkgQName = (FrjQName) e.nextElement();
                    typeQName = (FrjQName) e.nextElement();
                    pname = pkgQName.getFullString();
                }
            }

            FrjType newType = new FrjType(this.getFrejaEntry(),
                                          typeQName.getStr(),
                                          pname);
            // copy attribute
            newType.offset = typeQName.getOffset();
            newType.length = typeQName.getLength();
            if (pkgTypeDot != null) {
                new FrjRelation(FrjRelations.TYPE_QNAME,
                                newType, pkgQName);
                pkgQName.resetRelatedObjects();
            }

            // dummy relation, for using newType.getReferringClassInterface()
            new FrjRelation(FrjRelations.HAS_TYPE, this, newType);
            newType.setRelatedObjects();

            // convert qnames referring vars to expression
            //  e.g. Package.Class.mem1.mem2.mem3
            //                     ^^^^^^^^^^^^^^
            // QName +-- QName +-- QName(mem1)
            //       |         +-- QName(mem2)
            //       |         +-- Operator(.)
            //       +-- QName(mem3)
            //       +-- Operator(.)
            // ===>
            // Expr(OP) +-- Expr(OP) +-- Expr(REF_VAR, mem1)
            //          |            +-- Expr(REF_VAR, mem2)
            //          |            +-- Operator(.)
            //          +-- Expr(REF_VAR, mem3)
            //          +-- Operator(.)

            FrjOperator typeVarDot = null;
            FrjExpression varExpr = null;
            if (varQNames.size() > 0) { // Style 3)
                FrjWorld.debug(debug,
                                      "QName has variable " + fullQNameString);

                Enumeration e = varQNames.elements();
                FrjQName varQName = (FrjQName) e.nextElement(); // mem1
                varExpr = new FrjExpression(this.getFrejaEntry());
                                // copy attribute
                varExpr.str = varQName.getStr();
                varExpr.offset = varQName.getOffset();
                varExpr.length = varQName.getLength();
                varExpr.sort = REF_VAR;
                                // set varExpr's type
                FrjClassInterface ci = newType.getReferringClassInterface();
                FrjVariable var = ci.search4Member(varExpr.getStr());
                if (var != null) {      
                    FrjType varType = var.getType();
                    if (varType.getFrejaEntry() != varExpr.entry) {
                        varType = new FrjType(varExpr.entry, varType.getStr(),
                                              varType.getPackageName());
                    }
                    new FrjRelation(FrjRelations.HAS_TYPE, varExpr, varType);
                    varExpr.resetRelatedObjects();
                } else {
                    FrjWorld.warning("No such variable "
                                            + varExpr.getStr()
                                            + " defined in " +
                                            getClassInterfaceType(ci));
                }

                typeVarDot = varQName.getParentQName().getDot();
                typeVarDot.removeRelatedObjects();
                varQName.remove();
                                
                while (e.hasMoreElements()) { // order: mem2, mem3
                    FrjQName cdrQName = (FrjQName) e.nextElement();
                    FrjExpression carExpr = varExpr;
                    FrjExpression cdrExpr
                        = new FrjExpression(this.getFrejaEntry());
                    // copy attribute
                    cdrExpr.str = cdrQName.getStr();
                    cdrExpr.offset = cdrQName.getOffset();
                    cdrExpr.length = cdrQName.getLength();
                    cdrExpr.sort = REF_VAR;
                    // set varExpr's type
                    ci = carExpr.getType().getReferringClassInterface();
                    var = ci.search4Member(cdrExpr.getStr());
                    FrjType cdrType = null;
                    if (var != null) {
                        cdrType = var.getType();
                        if (cdrType.getFrejaEntry() != cdrExpr.entry) {
                            cdrType = new FrjType(cdrExpr.entry,
                                                  cdrType.getStr());
                        }
                        new FrjRelation(FrjRelations.HAS_TYPE,
                                        cdrExpr, cdrType);
                    } else {
                        FrjWorld.warning("No such variable "
                                                + cdrExpr.getStr()
                                                + " defined in "
                                                + getClassInterfaceType(ci));
                    }

                    varExpr = new FrjExpression(this.getFrejaEntry());
                    varExpr.sort = OPERATION;
                    new FrjRelation(FrjRelations.EXPRESSION_OPERAND,
                                    varExpr, carExpr);
                    new FrjRelation(FrjRelations.EXPRESSION_OPERAND,
                                    varExpr, cdrExpr);
                    new FrjRelation(FrjRelations.EXPRESSION_OPERATOR,
                                    varExpr, typeVarDot);
                    new FrjRelation(FrjRelations.HAS_TYPE,
                                    varExpr, cdrType);
                    carExpr.resetRelatedObjects();
                    cdrExpr.setRelatedObjects();
                    typeVarDot = cdrQName.getParentQName().getDot();
                    typeVarDot.removeRelatedObjects();
                    cdrQName.remove();
                }
                varExpr.resetRelatedObjects();
            }

            try {
                this.removeRelatedObject(newType); // remove dummy relation
                newType.removeRelatedObject(this); // remove dummy relation
            }
            catch (NoSuchRelatedObjectException e) {
                e.printStackTrace();
            }

            // remove qname
            if (qname != referree) { // includeing variable, Style 3)
                referree.remove();
            }
            qname.remove();
            if (pkgTypeDot != null) {
                typeQName.remove();
                pkgTypeDot.remove();
            }

            FrjExpression newExpr
                = new FrjExpression(this.getFrejaEntry());
            this.copyTo(newExpr);
            FrjExtents.replaceObject(newExpr);

            if (varExpr == null) {
                newExpr.sort = REF_TYPE;
                new FrjRelation(FrjRelations.REFER_TYPE, newExpr, newType);
            } else {
                FrjExpression typeExpr
                    = new FrjExpression(this.getFrejaEntry());
                                // copy attribute
                typeExpr.offset = newType.getOffset();
                typeExpr.length = newType.getLength();
                typeExpr.sort = REF_TYPE;
                new FrjRelation(FrjRelations.HAS_TYPE, typeExpr, newType);
                new FrjRelation(FrjRelations.REFER_TYPE, typeExpr, newType);

                newExpr.sort = OPERATION;
                new FrjRelation(FrjRelations.EXPRESSION_OPERAND,
                                newExpr, typeExpr);
                new FrjRelation(FrjRelations.EXPRESSION_OPERAND,
                                newExpr, varExpr);
                new FrjRelation(FrjRelations.EXPRESSION_OPERATOR,
                                newExpr, typeVarDot);
            }

            newExpr.resetRelatedObjects();
            for (Enumeration e = newExpr.getRelatedObjects();
                 e.hasMoreElements();) {
                FrjObject obj = (FrjObject) e.nextElement();
                if (obj == null) {
                    // removed object by converting qname to expression
                    //                                  FrjWorld.warning("Vector Enumerator returns null");
                } else {
                    obj.resetRelatedObjects();
                }
            }
            newType.resetRelatedObjects();
            return;
        }

        // Style 4)
        qname = (FrjQName) referree;
        String fullQNameString = qname.getFullString();
        if (fullQNameString.indexOf(".") < 0) {
            FrjClassInterface ci
                = (FrjClassInterface) this.getDefinerClassInterface();
            FrjVariable var = ci.search4Member(qname.getStr());

            if (var != null) {
                myType = var.getType();
                Japid.Freja.FrjWorld.debug(debug, "QName " + fullQNameString
                                  + " is super class "
                                  + var.getDefinerClassInterface().getStr() 
                                  + "'s member and type is "
                                  + myType.getStr() + ".");
            } else {
                FrjWorld.vwarning("Undefined QName \"" + fullQNameString + "\".");
            }
        } else {
            FrjWorld.vwarning("Undefined QName \"" + fullQNameString + "\".");
        }
    }

    private void setSpecialType() {
        FrjClassInterface ci
            = (FrjClassInterface) this.getDefinerClassInterface();
        FrjSpecial sp = (FrjSpecial) referree;

        if (sp.isThis()) {      // this or ClassName.this
            FrjExpression parent = getParentExpression();
            if (parent != null && parent.getSort() == OPERATION &&
                parent.getOperator().isDot()) {
                FrjExpression first = parent.get1stOperand();
                if (first == this) { // this.var
                    myType = getClassInterfaceType(ci);
                } else {
                    // ClassName.this
                    first.setType();
                    myType = parent.get1stOperand().getType();
                }
            } else { // simply this
                myType = getClassInterfaceType(ci);
            }
        } else if (sp.isSuper()) {
            Enumeration e = ci.getSuperClassInterfaceTypes();
            if (e.hasMoreElements()) {
                myType = (FrjType) e.nextElement();
                                // must be the super type or interface
            } else {
                if (!getDefinerClassInterface().getStr()
                    .equals("Object")) {
                    myType = new FrjType(entry, "Object");
                } else {
                    FrjWorld.warning("No extended class");
                    myType = null;
                }
            }
        } else if (sp.isTrue() || sp.isFalse()) {
            myType = new FrjType(entry, FrjType.BOOLEAN);
        } else if (sp.isNull()) {
            myType = new FrjType(entry, "null");// null objective class
        } else if (sp.isClass()) {        // Class Literal
            FrjExpression parent = getParentExpression();
            if (parent != null
                && parent.getSort() == OPERATION
                && parent.getOperator().isDot()) {
                parent.get1stOperand().setType();
                                // Class Literal is object of java.lang.Class
                myType = new FrjType(this.getFrejaEntry(),
                                     "Class");
                parent.setType(myType);
            } else {
                FrjWorld.warning("illegal class literal: " + this);
            }
        } else {
            FrjWorld.error("No such special symbol "
                                  + referree.getStr() + ".");
        }
    }

    protected FrjType getClassInterfaceType(FrjClassInterface ci) {
        for (Enumeration e = ci.getTypes();
             e.hasMoreElements();) {
            FrjType t = (FrjType) e.nextElement();
            if (t.getFrejaEntry() == getFrejaEntry()) {
                return t;
            }
        }
        return null;
    }

    private void resolveReferringVariable() {
        if (!(this.getSort() == REF_VAR && referree == null)) {
            return;
        }

        FrjExpression parent = this.getParentExpression();
        FrjClassInterface memberClass = null;
        if (parent != null && parent.getSort() == OPERATION &&
            parent.getOperator().isDot()) {
            FrjExpression parent1stOperand =  parent.get1stOperand();
            if (parent1stOperand != this) {
                parent1stOperand.setType();
                FrjType t = parent1stOperand.getType();
                if (t == null) {
                    memberClass = null;
                } else {
                    memberClass = t.getReferringClassInterface();
                    if (memberClass == null
                        && t.getSort() == FrjType.ARRAY
                        && this.getStr().equals("length")) {
                        // this is length of array
                        // this type can be set by setTypes()
                        return;
                    }
                }
            } else {
                memberClass = this.getDefinerClassInterface();
            }
        } else {
            memberClass =  this.getDefinerClassInterface();
        }

        while (memberClass != null) {
            referree = memberClass.search4Member(getStr());
            if (referree != null) {
                new FrjRelation(FrjRelations.REFER_VARIABLE, this, referree);
                referree.resetRelatedObjects();
                return;

            }
            FrjObject parentObj = memberClass.getParentObject();
            if (parentObj instanceof FrjClass) {
                // parentObj is outer class
                memberClass = (FrjClass) parentObj;
            } else {
                memberClass = null;
            }
        }

        FrjWorld.vwarning("FrjExpression#resolveReferringVariable(): " +
                          this + ": Can't get type of class.\n");
    }

    private FrjType searchEntry4Type(FrejaEntry entry, String className) {
        // QName is class ?
        for (Enumeration e = FrjExtents.getObjects(entry.getFrejaID(),
                                                   FrjClass.getClassNameStatically());
             e.hasMoreElements();) {
            FrjClass c = (FrjClass) e.nextElement();
            if (c.getStr().equals(className)) {
                FrjWorld.debug(debug, "QName " + className +
                               " is class " + c.getStr());
                return getClassInterfaceType(c);
            }
        }
        // QName is interface?
        for (Enumeration
                 e = FrjExtents.getObjects(entry.getFrejaID(),
                                           FrjInterface.getClassNameStatically());
             e.hasMoreElements();) {
            FrjInterface i = (FrjInterface) e.nextElement();
            if (i.getStr().equals(className)) {
                FrjWorld.debug(debug, "QName " + className +
                               " is interface " + i.getStr());
                return getClassInterfaceType(i);
            }
        }
        FrjWorld.warning("QName " + className
                                + " must be class, but unknown.");
        return null;
    }

    public static void setTypes() {
        setTypes(FrjExtents.getObjects(FrjExpression.CLASS_NAME),
                 FrjExtents.getObjects(FrjFile.CLASS_NAME));
    }

    public static void setTypes(FrejaEntry entry) {
        setTypes(FrjExtents.getObjects(entry, FrjExpression.CLASS_NAME),
                 FrjExtents.getObjects(entry, FrjFile.CLASS_NAME));
    }

    protected static void setTypes(Enumeration exprs, Enumeration files) {
        // translate REF_QNAME to valid expression
        while (exprs.hasMoreElements()) {
            FrjExpression expr = (FrjExpression) exprs.nextElement();
            if (expr.getSort() == REF_QNAME) {
                expr.setType();
            }
        }
        while (files.hasMoreElements()) {
            ((FrjObject) files.nextElement()).traverse(new TypeSetter());
        }
    }


    public void setReferree() {
        if (sort == REF_VAR && referree == null) {
            this.resolveReferringVariable();
        }
    }

    public static void setReferrees() {
        setReferrees(FrjExtents.getObjects(CLASS_NAME));
    }

    public static void setReferrees(FrejaEntry entry) {
        setReferrees(FrjExtents.getObjects(entry, CLASS_NAME));
    }

    protected static void setReferrees(Enumeration e) {
        while (e.hasMoreElements()) {
            ((FrjExpression) e.nextElement()).setReferree();
        }
    }

    public static void setInvokeRoutines() {
        setInvokeRoutines(FrjExtents.getObjects(CLASS_NAME));
    }

    public static void setInvokeRoutines(FrejaEntry entry) {
        setInvokeRoutines(FrjExtents.getObjects(entry, CLASS_NAME));
    }

    protected static void setInvokeRoutines(Enumeration e) {
        while (e.hasMoreElements()) {
            FrjExpression expr = (FrjExpression) e.nextElement();
            if (expr.getSort() == METHOD_CALL) {
                expr.setInvokeMethod();
            } else if (expr.getSort() == CONSTRUCTOR_CALL) {
                expr.setInvokeConstructor();
            }
        }
    }

    public void setInvokeRoutine() {
        if (this.getSort() == METHOD_CALL) {
            this.setInvokeMethod();
        } else if (this.getSort() == CONSTRUCTOR_CALL) {
            this.setInvokeConstructor();
        }
    }

    public static void setInvokeMethods() {
        setInvokeMethods(FrjExtents.getObjects(CLASS_NAME));
    }

    public static void setInvokeMethods(FrejaEntry entry) {
        setInvokeMethods(FrjExtents.getObjects(entry, CLASS_NAME));
    }

    protected static void setInvokeMethods(Enumeration e) {
        while (e.hasMoreElements()) {
            FrjExpression expr = (FrjExpression) e.nextElement();
            if (expr.getSort() == METHOD_CALL) {
                expr.setInvokeMethod();
            }
        }
    }
    public void setInvokeMethod() {
        if (this.getSort() != FrjExpression.METHOD_CALL || routine != null) {
            return;
        }

        // set arguments' type
        for (Enumeration args = getChildExpressions();
             args.hasMoreElements();) {
            FrjExpression arg = (FrjExpression) args.nextElement();
            arg.setType();
        }
        //              this.traverse(new TypeSetter());

        FrjTypeList tl = this.argument2TypeList();
                
        FrjClassInterface methodClass = null;
        FrjExpression parent = getParentExpression();
        if (parent != null && parent.getSort() == OPERATION &&
            parent.getOperator().isDot()) {
            // obj.method(...) or method(...).XXX STYLE 
            FrjExpression parent1stOperand =  parent.get1stOperand();
            if (parent1stOperand != this) {
                parent1stOperand.setType();
                FrjType t = parent1stOperand.getType();
                if (t == null) {
                    methodClass = null;
                } else {
                    methodClass = t.getReferringClassInterface();
                }
            } else {
                methodClass =  this.getDefinerClassInterface();
            }
        } else {
            methodClass =  this.getDefinerClassInterface();
        }

        if (methodClass == null) {
            //                  System.out.println("   %%% can't get class");
            FrjWorld.vwarning("FrjExpression#setInvokeMethod(): "
                              + "Can't get type of class.\n"
                              + "  method: " + this);
            return;
        }

        FrjRoutine r = (FrjRoutine) methodClass
            .search4Method(this.getStr(), this.argument2TypeList());
        if (r != null) {
            new FrjRelation(FrjRelations.INVOKE_ROUTINE, this, r);
            resetRelatedObjects();
            r.resetRelatedObjects();
        } else {
            FrjWorld.vwarning("FrjExpression#setInvokeMethod(): "
                              + "Can't specify invoked method.\n"
                              + "  method: " + this);

        }
    }

    public void setInvokeConstructor() {
        if (this.getSort() != FrjExpression.CONSTRUCTOR_CALL ||
            routine != null) {
            return;
        }

        if (myType == null) {
            // referree is FrjType (refer_type)
            if (referree instanceof FrjSpecial) {
                FrjSpecial sp = (FrjSpecial) referree;
                FrjClassInterface ci =
                    (FrjClassInterface)this.getDefinerClassInterface();
                if (sp.isThis()) {              // this() style.
                    myType = getClassInterfaceType(ci);
                } else if (sp.isSuper()) {      // super() style.
                    Enumeration e = ci.getSuperClassInterfaceTypes();
                    if (e.hasMoreElements()) {
                        myType = (FrjType) e.nextElement();
                    } else {
                        if (!getDefinerClassInterface().getStr().equals("Object")) {
                            myType = new FrjType(entry, "Object");
                        } else {
                            FrjWorld.warning("No extended class");
                            myType = null;
                        }
                    }
                } else {
                    FrjWorld.error("No such special symbol "
                                          + sp.getStr() + ".");
                }
            } else {
                myType = (FrjType) referree;
            }
        }
        if (myType != null) {
            FrjClassInterface ci = myType.getReferringClassInterface();
            if (ci != null) {
                // Note: constuctor call's str is null.
                FrjRoutine r = (FrjRoutine)
                    ci.search4Constructor(this.argument2TypeList());
                if (r != null) {
                    new FrjRelation(FrjRelations.INVOKE_ROUTINE, this, r);
                    FrjType myType0 = myType;
                    resetRelatedObjects();
                    r.resetRelatedObjects();
                    myType = myType0;
                    return;
                }
            }
            // Can't find constructor definition
            FrjWorld.vwarning("FrjExpression#setType(): "
                              + "Can't specify constructor.\n"
                              + "  constructor: " + this);
        }
    }

    public FrjObject getParentObject() {
        if (statement != null) {
            return statement;
        }
        if (parentExpression != null) {
            return parentExpression;
        } else if (initialized != null) {
            return initialized;
        } else {
            return caseLabel;
        }
    }

    public FrjTypeList argument2TypeList() {
        if (!(getSort() == METHOD_CALL || getSort() == CONSTRUCTOR_CALL)) {
            return null;
        }

        VectorSorter    sorter = new SortByOffset(getChildExpressions());
        sorter.sort();
        FrjTypeList argTypeList = new FrjTypeList();
        for (Enumeration e = sorter.getVector().elements();
             e.hasMoreElements();) {
            FrjExpression arg = (FrjExpression) e.nextElement();
            arg.setType();
            FrjType t = (FrjType) arg.getType();
            if (t == null) {
                if (FrjWorld.isVerbose()) {
                    FrjWorld.error("FrjExpression#argumnet2TypeList:"
                                          + "Can't set type \n"
                                          + "  method: " + this + "\n"
                                          + "  arg:    " + arg);
                }
                return null;
            } else {
                argTypeList.addType(t);
            }
        }
        return argTypeList;
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjExpression) obj).sort = sort;
        ((FrjExpression) obj).lrsort = lrsort;
    }

    // Traverse method.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);

        tm.doMyself(this);

        traverseObjects(getChildExpressions(), tm);
        if (operator != null) {
            traverseObjects(getOperands(), tm);
            operator.traverse(tm);
        }
        if (referree != null && !(referree instanceof FrjVariable)) {
            referree.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }

    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];
                
        sortNameTable[EMPTY] = "Empty";
        sortNameTable[REF_VAR] = "Refer Variable";
        sortNameTable[REF_ARRAY] = "Refer Array";
        sortNameTable[REF_LITERAL] = "Refer Literal";
        sortNameTable[REF_TYPE] = "Refer Type";
        sortNameTable[REF_SPECIAL] = "Refer Special";
        sortNameTable[METHOD_CALL] = "Method Call";
        sortNameTable[CONSTRUCTOR_CALL] = "Constructor Call";
        sortNameTable[OPERATION] = "Operation";
        sortNameTable[REF_QNAME] = "Refer QName";
        sortNameTable[LISTED] = "Listed Expression";
        sortNameTable[DIM_EXPRESSION] = "Dimension Expression";
        sortNameTable[ANONYMOUS_CLASS] = "Anonymous Class";
        sortNameTable[ANONYMOUS_ARRAY] = "Anonymous Array";
        sortNameTable[REF_ARRAY_LENGTH] = "Refer Array Length";

        lrsortNameTable = new String[LRSORT_LAST_MEMBER +1];
        lrsortNameTable[NONE] = "NONE";
        lrsortNameTable[L_VALUE] = "Left Value";
        lrsortNameTable[R_VALUE] = "Right Value";
        lrsortNameTable[LR_VALUE] = "Left and Right Value";
        lrsortNameTable[OBJECT_OPERATION] = "Object Operation";
    }
}
