/* 
 * Program:             $RCSfile: FrjExtents.java,v $  $Revision: 41.9 $
 * 
 * Purpose:             The super class of all the DBClasses
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjExtents.java,v 41.9 2003/04/06 07:42:39 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;

public class FrjExtents {
    // Class variables
    /** dictionary from an id to the object that has the id */
    protected static Dictionary oid2ObjectDict = new Hashtable(256);

    /**
          * Dictionary from an object to the id.
          * This is used in replaceObject().
          */
    protected static Dictionary object2OIDDict = new Hashtable(256);

    /** dictionary from a class name to a vector of objects of the class */
    protected static Dictionary className2VectorDict = new Hashtable(32, 0.9f);

    /** dictionary from an id to the object that has the id */
    protected static Dictionary rid2RelationDict = new Hashtable(256);

    /** dictionary from a class name to a vector of objects of the class */
    protected static Vector     relNames = new Vector();

    /** hashtable of loaded FrejaEntry */
    protected static Hashtable  loadedEntries = new Hashtable();

    // Class methods
    /** return all class names defined in Freja */
    public static Enumeration getClassNames() {
        return className2VectorDict.keys();
    }

    /** add an object to extents */
    static void addObject(FrjObject obj) {
        try {
            Long dictKey = new Long(obj.getID());
            if (oid2ObjectDict.get(dictKey) == null) {
                getClassVector(obj.getClassName()).addElement(obj);
                oid2ObjectDict.put(dictKey, obj);
                object2OIDDict.put(obj, dictKey);
            }
            else {
                FrjWorld.warning("An object is already added." +
                                        "  " + obj);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void replaceObject(FrjObject obj) {
        try {
            Vector v = getClassVector(obj.getClassName());
            Long idKey = (Long) object2OIDDict.get(obj);
            if (idKey != null) {        // obj is already added
                object2OIDDict.remove(obj);
                oid2ObjectDict.remove(idKey);
                v.removeElement(obj);
            }
            FrjObject old = getObjectByID(obj.getID());
            v.setElementAt(obj, v.indexOf(old));
            oid2ObjectDict.put(new Long(obj.getID()), obj);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void removeObject(FrjObject obj) {
        try {
            Vector v = getClassVector(obj.getClassName());
            v.removeElement(obj);
            oid2ObjectDict.remove(new Long(obj.getID()));
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }


    /** return all objects */
    public static Enumeration getAllObjects() {
        return oid2ObjectDict.elements();
    }
    /** return all objects in Freja having frejaID */
    public static Enumeration getAllObjects(int frejaID) {
        return selectObjects(frejaID, getAllObjects());
    }
    /** return all objects in FrejaEntry */
    public static Enumeration getAllObjects(FrejaEntry entry) {
        return getAllObjects(entry.getFrejaID());
    }
                
    /** return all objects of a className */
    public static Enumeration getObjects(String className) {
        try {
            return getClassVector(className).elements();
        }
        catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    /** return all objects of a className in Freja having frejaID */
    public static Enumeration getObjects(int frejaID, String className) {
        return selectObjects(frejaID, getObjects(className));
    }

    /** return all objects of a className in FrejaEntry */
    public static Enumeration getObjects(FrejaEntry entry, String className) {
        return getObjects(entry.getFrejaID(), className);
    }

    /** return the object having an id */
    public static FrjObject getObjectByID(long id) {
        return (FrjObject) oid2ObjectDict.get(new Long(id));
    }

    private static Vector getClassVector(String className)
        throws NoSuchFrjClassOrRelationException {
        Vector v = (Vector) className2VectorDict.get(className);
        if (v == null) {
            throw new NoSuchFrjClassOrRelationException
                ("Class " + className + " is not defined.");
        }
        return v;
    }

    public static int getObjectNums() {
        return oid2ObjectDict.size();
    }

    private static Enumeration selectObjects(int frejaID, Enumeration e) {
        Vector v = new Vector();
        while (e.hasMoreElements()) {
            FrjObject obj = (FrjObject) e.nextElement();
            if (FrjID.getFrejaID(obj.getID()) == frejaID) {
                v.addElement(obj);
            }
        }
        return v.elements();
    }
        

    // Methods about relations
    /** Adds the relation to extents. */
    static void addRelation(FrjRelation rel) {
        try {
            //                  getRelationVector(rel.getRelationName()).addElement(rel);
            rid2RelationDict.put(new Long(rel.getID()), rel);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
    /** Removes the relation from extents. */
    static void removeRelation(FrjRelation rel) {
        try {
            //                  getRelationVector(rel.getRelationName()).addElement(rel);
            rid2RelationDict.remove(new Long(rel.getID()));
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /** return all relation names defined in Freja */
    public static Enumeration getRelationNames() {
        return relNames.elements();
    }

    /** return all relations */
    public static Enumeration getAllRelations() {
        return rid2RelationDict.elements();
    }

    /** return all relations in Freja having frejaID */
    public static Enumeration getAllRelations(int frejaID) {
        return selectRelations(frejaID, getAllRelations());
    }

    /** return all relations in FrejaEntry */
    public static Enumeration getAllRelations(FrejaEntry entry) {
        return getAllRelations(entry.getFrejaID());
    }

    /** return all relations of a relName */
    public static Enumeration getRelations(String relName) {
        Vector relations = new Vector();
        FrjRelation     rel;
        for (Enumeration e = getAllRelations();
             e.hasMoreElements();) {
            rel = (FrjRelation) e.nextElement();
            if (rel.getRelationName().equals(relName)) {
                relations.addElement(rel);
            }
        }
        return relations.elements();
    }

    /** return all relations of a relName in Freja having frejaID */
    public static Enumeration getRelations(int frejaID, String relName) {
        return selectRelations(frejaID, getRelations(relName));
    }

    /** return all relations of a relName in FrejaEntry */
    public static Enumeration getRelations(FrejaEntry entry, String relName) {
        return getRelations(entry.getFrejaID(), relName);
    }

    /** return the object having an id */
    public static FrjRelation getRelationByID(int id) {
        return (FrjRelation) rid2RelationDict.get(new Long(id));
    }

    public static int getRelationNums() {
        return rid2RelationDict.size();
    }

    private static Enumeration selectRelations(int frejaID, Enumeration e) {
        Vector v = new Vector();
        while (e.hasMoreElements()) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            if (FrjID.getFrejaID(rel.getID()) == frejaID) {
                v.addElement(rel);
            }
        }
        return v.elements();
    }

    /** add loaded FrejaEntry */
    public static void addFrejaEntry(FrejaEntry entry) {
        if (entry != null
            && !loadedEntries.containsKey(entry.getCanonicalFrejaCode())) {
            loadedEntries.put(entry.getCanonicalFrejaCode(), entry);
        }
    }

    /** check whether entry is already loaded */
    public static boolean isLoaded(FrejaEntry entry) {
        return loadedEntries.containsKey(entry.getCanonicalFrejaCode());
    }

    /** return FrejaEntry already loaded */
    public static FrejaEntry getLoadedFrejaEntry(FrejaEntry entry) {
        return (FrejaEntry) loadedEntries.get(entry.getCanonicalFrejaCode());
    }

    /** return all FrejaEntrys */
    public static Enumeration getAllFrejaEntrys() {
        return loadedEntries.elements();
    }

    /** clear all objects and relations */
    public static void clear() {
        oid2ObjectDict = new Hashtable(256);
        object2OIDDict = new Hashtable(64);
        className2VectorDict = new Hashtable(32, 0.9f);
        rid2RelationDict = new Hashtable(256);
        loadedEntries = new Hashtable();
        initClassName2VectorDict();
        FrejaEntry.freeAllFrejaID();
    }


    static private void initClassName2VectorDict() {
        // Classes
        className2VectorDict.put("File", new Vector());
        className2VectorDict.put("Class", new Vector());
        className2VectorDict.put("Interface", new Vector());
        className2VectorDict.put("Routine", new Vector());
        className2VectorDict.put("Variable", new Vector());
        className2VectorDict.put("Type", new Vector());
        className2VectorDict.put("Modifier", new Vector());
        className2VectorDict.put("SBlock", new Vector());
        className2VectorDict.put("QName", new Vector());
        className2VectorDict.put("Statement", new Vector());
        className2VectorDict.put("Label", new Vector());
        className2VectorDict.put("Expression", new Vector());
        className2VectorDict.put("Special", new Vector());
        className2VectorDict.put("Operator", new Vector());
        className2VectorDict.put("Literal", new Vector());
        className2VectorDict.put("Comment", new Vector());

    }
        
    static private void initRelNames() {
        // Relations
        // These codes are converted by DBTable.c
        relNames.addElement(FrjRelations.DEFINE_PACKAGE);
        relNames.addElement(FrjRelations.IMPORT_PACKAGE);
        relNames.addElement(FrjRelations.DEFINE_CLASS);
        relNames.addElement(FrjRelations.IMPLEMENT_INTERFACE);
        relNames.addElement(FrjRelations.EXTEND_CLASS);
        relNames.addElement(FrjRelations.CLASS_MODIFIER);
        relNames.addElement(FrjRelations.CLASS_VARIABLE);
        relNames.addElement(FrjRelations.CLASS_ROUTINE);
        relNames.addElement(FrjRelations.CLASS_SBLOCK);
        relNames.addElement(FrjRelations.CLASS_ARGUMENT);                       /* JDK 1.1. */
        relNames.addElement(FrjRelations.CLASS_CLASS);                          /* JDK 1.1. */
        relNames.addElement(FrjRelations.CLASS_INTERFACE);                      /* JDK 1.1. */
        relNames.addElement(FrjRelations.DEFINE_INTERFACE);
        relNames.addElement(FrjRelations.EXTEND_INTERFACE);
        relNames.addElement(FrjRelations.INTERFACE_MODIFIER);
        relNames.addElement(FrjRelations.INTERFACE_VARIABLE);
        relNames.addElement(FrjRelations.INTERFACE_ROUTINE);
        relNames.addElement(FrjRelations.INTERFACE_CLASS);                      /* JDK 1.1. */
        relNames.addElement(FrjRelations.INTERFACE_INTERFACE);          /* JDK 1.1. */
        relNames.addElement(FrjRelations.VARIABLE_MODIFIER);
        relNames.addElement(FrjRelations.VARIABLE_TYPE);
        relNames.addElement(FrjRelations.INITIALIZE_VARIABLE);
        relNames.addElement(FrjRelations.PARAMETER_VARIABLE);
        relNames.addElement(FrjRelations.RETURN_TYPE);
        relNames.addElement(FrjRelations.THROW_TYPE);
        relNames.addElement(FrjRelations.ROUTINE_BODY);
        relNames.addElement(FrjRelations.ROUTINE_MODIFIER);
        relNames.addElement(FrjRelations.TYPE_CLASS);
        relNames.addElement(FrjRelations.TYPE_INTERFACE);
        relNames.addElement(FrjRelations.TYPE_TYPE);
        relNames.addElement(FrjRelations.TYPE_QNAME);
        relNames.addElement(FrjRelations.QNAME_QNAME);
        relNames.addElement(FrjRelations.QNAME_DOT);
        relNames.addElement(FrjRelations.SBLOCK_BODY);
        relNames.addElement(FrjRelations.LOCAL_CLASS);                          /* JDK 1.1. */
        relNames.addElement(FrjRelations.LOCAL_VARIABLE);
        relNames.addElement(FrjRelations.STATEMENT_LABEL);
        relNames.addElement(FrjRelations.STATEMENT_STATEMENT);
        relNames.addElement(FrjRelations.STATEMENT_EXPRESSION);
        relNames.addElement(FrjRelations.REFER_LABEL);
        relNames.addElement(FrjRelations.CASE_VALUE);
        relNames.addElement(FrjRelations.EXPRESSION_OPERAND);
        relNames.addElement(FrjRelations.EXPRESSION_OPERATOR);
        relNames.addElement(FrjRelations.EXPRESSION_EXPRESSION);
        relNames.addElement(FrjRelations.REFER_VARIABLE);
        relNames.addElement(FrjRelations.REFER_LITERAL);
        relNames.addElement(FrjRelations.REFER_SPECIAL);
        relNames.addElement(FrjRelations.REFER_TYPE);
        relNames.addElement(FrjRelations.REFER_QNAME);
        relNames.addElement(FrjRelations.HAS_TYPE);
        relNames.addElement(FrjRelations.INVOKE_ROUTINE);
        relNames.addElement(FrjRelations.SYMBOL_COMMENT);
    }

    // static initializer
    static {
        initClassName2VectorDict();
        initRelNames();
    }

}
