/* 
 * Program:             $RCSfile: FrjID.java,v $  $Revision: 41.19 $
 * 
 * Purpose:             A wrapper class of SDB's object id.
 *
 * Author:              H.Kuwabara  2001/10/02
 *                      S.Yamamoto  2001/10/02
 *
 * (C) Copyright:       H.Kuwabara and S.Yamamoto  2001 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjID.java,v 41.19 2003/02/15 11:56:28 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import Japid.Freja.DBAccess.*;

public class FrjID {
    /*
     * |-----|---------|--------|-----|--------|---------|
     *  1     15        16       1     9        22
     *  sign  not used  FrejaId  sign  ClassId  ObjectId
     *                  FID      SIGN  CRID     OID
     */

    private static final int JPD_BIT_WIDTH_FID = 16;
    private static final int JPD_BIT_WIDTH_SDBID = 32;
    private static final int JPD_BIT_WIDTH_SIGN = 1;
    private static final int JPD_BIT_WIDTH_CRID = 9;
    private static final int JPD_BIT_WIDTH_OID = 22;

    public static final int MAX_FREJA = 1 << JPD_BIT_WIDTH_FID;
    
    private static final long MASK_GET_FID = 
        (long)(MAX_FREJA - 1) << JPD_BIT_WIDTH_SDBID;
    private static final long MASK_GET_CRID =
        ((1 << JPD_BIT_WIDTH_CRID) - 1) << JPD_BIT_WIDTH_OID;

    private static final long newIDSpace = 0;
    private static long newOid = 0;
    private static long newRid = 0;

    /**
     * Generates object id from FrejaEntry and id gotten from SDB.
     */
    public static final long generateObjectID(FrejaEntry entry, int sdbID) {
        int classID = getClassRelID(sdbID);

        if ((sdbID & MASK_GET_FID) != 0) {
            FrjWorld.errorExit(entry.getFrejaPath() + ": Too many objects.");
        }
        if (entry.getMaxCRID(classID) < sdbID) {
            entry.setMaxCRID(classID, sdbID);
        }
        return ((long)entry.getFrejaID() << JPD_BIT_WIDTH_SDBID) | sdbID;
    }

    /**
     * Generates new object id from FrejaEntry.
     */
    public static final long generateObjectID(FrejaEntry entry,
                                              String className) {
        int classID = SymSort.getSymSortValByName(className);

        if (classID == SymSort.JPD_INVALID_ID) {
            FrjWorld.errorExit(className + ": Can't find class.");
        } else if ((entry.incrementMaxCRID(classID) & MASK_GET_FID) != 0) {
            FrjWorld.errorExit(entry.getFrejaPath() +
                               ": Too many created objects.");
        }

        return ((long)entry.getFrejaID() << JPD_BIT_WIDTH_SDBID) | 
               (classID << JPD_BIT_WIDTH_OID) | entry.getMaxCRID(classID);
    }

    /**
     * Generates new relation id from FrejaEntry.
     */
    public static final long generateRelationID(FrejaEntry entry,
                                                String relName) {
        int relID = FrjWorld.dbTable.getDBClassRelIDByName(relName);

        if (relID == SymSort.JPD_INVALID_ID) {
            FrjWorld.errorExit("Can't find relation : " + relName);
        } else if ((entry.incrementMaxCRID(relID) & MASK_GET_FID) != 0) {
            FrjWorld.errorExit(entry.getFrejaPath() +
                               ": Too many created relations.");
        }

        return ((long)entry.getFrejaID() << JPD_BIT_WIDTH_SDBID) | 
               (relID << JPD_BIT_WIDTH_OID) | entry.getMaxCRID(relID);
    }

    /** Generates new object id. */
    public static final long generateObjectID() {
        return (newIDSpace << JPD_BIT_WIDTH_SDBID) + newOid++;
    }

    /** Generates new relation id. */
    public static final long generateRelationID() {
        return (newIDSpace << JPD_BIT_WIDTH_SDBID) + newRid++;
    }

    /** Returns data id in SDB. */
    public static final int getFrejaDataID(long id) {
        return (int)(id & ~MASK_GET_FID);
    }

    /** Returns full object id. */
    public static final long getID(int frejaID, int dataID) {
        return ((long)frejaID << JPD_BIT_WIDTH_SDBID) + dataID;
    }

    /** Returns Freja id. */
    public static final int getFrejaID(long oid) {
        return (int)((oid & MASK_GET_FID) >> JPD_BIT_WIDTH_SDBID);
    }

    /** Returns ClassRel id in SDB. */
    public static final int getClassRelID(long oid) {
        return (int)((oid & MASK_GET_CRID) >> JPD_BIT_WIDTH_OID);
    }
}
