/* 
 * Program:             $RCSfile: FrjIdentifier.java,v $  $Revision: 41.4 $
 * 
 * Purpose:             The class corresponding to Expression
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjIdentifier.java,v 41.4 2003/02/12 03:39:33 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

abstract public class FrjIdentifier extends FrjObject {
    // Class variables.

    protected Vector            modifiers = null;

    // Consturctors.
    public FrjIdentifier(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }
    public FrjIdentifier(FrejaEntry entry) {
        super(entry);
    }
    public FrjIdentifier() {
        super();
    }

    /**
     * Traverses <I>class_modifier</I> or <I>interface_modifier</I>
     * or <I>variable_modifier</I> or <I>routine_modifier<I>.
     */
    public Enumeration getModifiers() {
        return modifiers.elements();
    }

    public boolean isPublic() {
        return checkModifier("public");
    }
    public boolean isProtected() {
        return checkModifier("protected");
    }
    public boolean isPrivate() {
        return checkModifier("private");
    }
    public boolean isDefaultAccess() {
        return (!isPublic() && !isProtected() && !isPrivate());
    }
    public boolean isStatic() {
        return checkModifier("static");
    }
    public boolean isFinal() {
        return checkModifier("final");
    }
    public boolean isAbstract() {
        return checkModifier("abstract");
    }
    public boolean isNative() {
        return checkModifier("native");
    }
    public boolean isSynchronized() {
        return checkModifier("synchronized");
    }

    /* Reserved:
     * public boolean isTransient() {
     *     return checkModifier("transient");
     * }     
     * public boolean isVolatile() {
     *     return checkModifier("volatile");
     * }     
     */

    private boolean checkModifier(String mod) {
        for (Enumeration e = modifiers.elements();
             e.hasMoreElements();) {
            FrjModifier m = (FrjModifier)e.nextElement();
            if (m.getStr().equals(mod)) {
                return true;
            }
        }
        return false;
    }
}
