/* 
 * Program:             $RCSfile: FrjInterface.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjInterface.java,v 41.5 2003/02/15 11:56:28 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjInterface extends FrjClassInterface {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Interface";

    /** table from a sort to string */
    protected static String     sortNameTable[];        // initialized by static block

    // Attribute.
    public int                          sort;

    // Related objects.
    protected Vector            superInterfaces = null;

    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int                _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int             PACKAGE_MEMBER = _i++;
    public final static int             NESTED_TOPLEVEL = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // sentinel

    // Constructors.
    public FrjInterface(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }

    public FrjInterface(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }

    public FrjInterface() {
        super();
        sort = INVALID_ID;
    }

    // Class methods.
    /** return the class name of this object */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjInterface(entry, obj);
    }

    // Instance method.
    /** return the class name dynamically */
    public String getClassName() {
        return CLASS_NAME;
    }

    /** return the sort */
    public int getSort() {
        return sort;
    }

    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // About relations.
    public void setRelatedObjects() {
        memberClasses = new Vector();
        memberInterfaces = new Vector();
        routines = new Vector();
        members = new Vector();
        modifiers = new Vector();
        superInterfaces = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        file = null;
        memberClasses = null;
        memberInterfaces = null;
        routines = null;
        members = null;
        modifiers = null;
        superInterfaces = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.INTERFACE_CLASS)) {
            memberClasses.addElement((FrjClass) obj);
        } else if (relName.equals(FrjRelations.INTERFACE_INTERFACE)) {
            if (rel.get1stObject().getID() == this.getID()) { // this is outer
                memberInterfaces.addElement((FrjInterface) obj);
            } else {      // this is member 
                outerInterface = (FrjInterface) obj;
            }
        } else if (relName.equals(FrjRelations.CLASS_INTERFACE)) {
            outerClass = (FrjClass) obj;
        } else if (relName.equals(FrjRelations.DEFINE_INTERFACE)) {
            file = (FrjFile) obj;
        } else if (relName.equals(FrjRelations.EXTEND_INTERFACE)) {
            superInterfaces.addElement((FrjType) obj);
            superClassInterfaceTypes.addElement(obj);
        } else if (relName.equals(FrjRelations.INTERFACE_ROUTINE)) {
            routines.addElement((FrjRoutine) obj);
        } else if (relName.equals(FrjRelations.INTERFACE_VARIABLE)) {
            members.addElement((FrjVariable) obj);
        } else if (relName.equals(FrjRelations.INTERFACE_MODIFIER)) {
            modifiers.addElement((FrjModifier) obj);
        } else if (relName.equals(FrjRelations.TYPE_INTERFACE)) {
            types.addElement(obj);
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(memberClasses);
        sorter.sort();
        sorter.setVector(memberInterfaces);
        sorter.sort();
        sorter.setVector(routines);
        sorter.sort();
        sorter.setVector(members);
        sorter.sort();
        sorter.setVector(modifiers);
        sorter.sort();
        sorter.setVector(superInterfaces);
        sorter.sort();
    }

    /**
     * Traverses <I>extend_interface</I>.
     * @return super interfaces (FrjType)
     */
    public Enumeration getSuperInterfaceTypes() {
        return superInterfaces.elements();
    }

    /**
     * Traverses <I>extend_interface</I>.
     * @return sub interfaces (FrjType)<P>
     *  NOTE: This doesn't return all sub interfaces.
     */
    public Enumeration getSubInterfaces() {
        Vector subInterfaces = new Vector();
        for (Enumeration e = getTypes();
             e.hasMoreElements();) {
            Object o = e.nextElement();
            if (o instanceof FrjType) {
                FrjType type = (FrjType) o;
                FrjObject pt = type.getParentObject();
                if (pt instanceof FrjInterface) {
                    subInterfaces.addElement(pt);
                }
            }
        }
        return subInterfaces.elements();
    }

    /**
     * Traverses <I>implement_interface</I>.
     * @return classes implemented this (FrjType) <P>
     *  NOTE: This doesn't return all classes implemented this.
     */
    public Enumeration getImplementedClasses() {
        Vector implementedClasses = new Vector();
        for (Enumeration e = getTypes();
             e.hasMoreElements();) {
            Object o = e.nextElement();
            if (o instanceof FrjType) {
                FrjType type = (FrjType) o;
                FrjObject pt = type.getParentObject();
                if (pt instanceof FrjClass) {
                    implementedClasses.addElement(pt);
                }
            }
        }
        return implementedClasses.elements();
    }

    /** Traverses <I>interface_type</I>. */
    public Enumeration getTypes() {
        if (types.size() == 0) {
            FrjType type = new FrjType(entry, this.str);
            FrjRelation rel = new FrjRelation(FrjRelations.TYPE_INTERFACE,
                                              this, type);
            resetRelatedObjects();
            type.setRelatedObjects();
        }
        return types.elements();
    }


    /** return parent object if exists */
    public FrjObject getParentObject() {
        if (file != null) {
            return file;
        } else if (outerClass != null) {
            return outerClass;
        } else if (outerInterface != null) {
            return outerInterface;
        } else {
            return null;
        }
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjInterface) obj).sort = sort;
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        traverseObjects(getModifiers(), tm);

        tm.doMyself(this);

        traverseObjects(getSuperInterfaceTypes(), tm);
        traverseObjects(getMembers(), tm);      
        traverseObjects(getRoutines(), tm);
        traverseObjects(getMemberClasses(), tm);
        traverseObjects(getMemberInterfaces(), tm);

        tm.doAfterTraverse(this);
    }

    /* static initializer */
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];

        sortNameTable[PACKAGE_MEMBER] = "Pakage Member";
        sortNameTable[NESTED_TOPLEVEL] = "Nested Top Level";
    }

}
