/* 
 * Program:             $RCSfile: FrjLiteral.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to Expression.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjLiteral.java,v 41.5 2003/02/12 03:39:34 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjLiteral extends FrjObject {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Literal";

    /** Table from a sort to string. */
    protected static String sortNameTable[];    // Initialized by static block.

    // Attribute.
    public int                  sort;

    // Related objects.
    protected FrjExpression     expression = null;

    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int             CHAR = _i++;
    public final static int             STRING = _i++;
    public final static int             UNICODE = _i++;
    public final static int             INT = _i++;
    public final static int             DOUBLE = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // sentinel

    // Constructors.
    public FrjLiteral(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }
    public FrjLiteral() {
        super();
        sort = INVALID_ID;
    }
    public FrjLiteral(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }


    // Class methods.
    /** Return the class name of this object statically. */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }
                        
    /** Create object from SDB. */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjLiteral(entry, obj);
    }

    // Instance methods.
    /** Return the class name of this object. */
    public String getClassName() {
        return CLASS_NAME;
    }

    // About attribute.
    /** Return the sort. */
    public int getSort() {
        return sort;
    }
    /** Check this object has a valid sort. */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** Return the sort as string. */
    public String getSortString() {
        if (!hasValidSort()) {  /* Error. */
            return null;
        }
        return sortNameTable[sort];
    }

    /** Return the name of attribute and its value as string. */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }


    /** Convert object to string. */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // About relation.
    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        expression = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.REFER_LITERAL)) {
            expression = (FrjExpression) obj;
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
    }

    /** Traverses <I>refer_literal</I>. */
    public FrjExpression getExpression() {
        return expression;
    }

    public FrjObject getParentObject() {
        return expression;
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjLiteral) obj).sort = sort;
    }

    // Traverse method.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        tm.doMyself(this);
        tm.doAfterTraverse(this);
    }
                        
    // Static initializer.
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];
                
        sortNameTable[CHAR] = "Char";
        sortNameTable[STRING] = "String";
        sortNameTable[UNICODE] = "Unicode";
        sortNameTable[INT] = "Int";
        sortNameTable[DOUBLE] = "Double";
    }
}
