/* 
 * Program:             $RCSfile: FrjModifier.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjModifier.java,v 41.5 2003/02/12 03:39:34 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjModifier extends FrjObject {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Modifier";

    // Related objects.
    /**
     * Modifiereds are Identifiers.
     * Modifiered is plural because consider following example.
     * <PRE> public int a, b, c; </PRE>
     * "public" modifies Variables "a", "b" and "c".
     */
    protected Vector    modifiereds = null;

    // Constructors.
    public FrjModifier(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }

    public FrjModifier(FrejaEntry entry) {
        super(entry);
    }

    public FrjModifier() {
        super();
    }

    // Class methods.
    /** Return the class name of this object. */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** Create object from SDB. */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjModifier(entry, obj);
    }

    // Instance method.
    /** Return the class name dynamically. */
    public String getClassName() {
        return CLASS_NAME;
    }

    // About relation.
    public void setRelatedObjects() {
        modifiereds = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        modifiereds = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {
        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            modifiereds.addElement(obj);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(modifiereds);
        sorter.sort();
    }

    /**
     * Traverses <I>class_modifier</I> or <I>interface_modifier</I>
     * or <I>variable_modifier</I> or <I>routine_modifier<I>.
     * @return modifiereds (FrjIdentifier) <P>
     * A modifiered is identifier
     * and is plural because consider following example.<P>
     * <PRE> public int a, b, c; </PRE><P>
     * "public" modifies Variables "a", "b" and "c".
     */
    public Enumeration getModifiereds() {
        return modifiereds.elements();
    }

    /** Return parent object if exists. */
    public FrjObject getParentObject() {
        if (modifiereds == null) {
            return null;
        } else {
            return (FrjObject) getModifiereds().nextElement();
        }
    }

    // Traverse method.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        tm.doMyself(this);
        tm.doAfterTraverse(this);
    }
}
