/* 
 * Program:             $RCSfile: FrjQName.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjQName.java,v 41.5 2003/02/12 03:39:35 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjQName extends FrjObject {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "QName";

    // Related objects.
    protected FrjFile           file = null;
    protected FrjObject         referrer = null;
    protected FrjQName          parentQName = null;
    protected Vector            childQNames = null;
    protected FrjOperator       dot = null;
        
    // Constructors.
    public FrjQName(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }

    public FrjQName(FrejaEntry entry) {
        super(entry);
    }

    public FrjQName() {
        super();
    }

    // Class methods.
    /** return the class name of this object */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** Create object from SDB. */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjQName(entry, obj);
    }

    // Instance method.
    /** Return the class name dynamically. */
    public String getClassName() {
        return CLASS_NAME;
    }
        
    /** Returns full quarified name. */
    public String getFullString() {
        String result = "";
        VectorSorter sorter = new SortByOffset(getChildQNames());
        sorter.sort();
        Enumeration e = sorter.getVector().elements();
        if (e.hasMoreElements()) {
            String dot = "";
            while (e.hasMoreElements()) {
                result += dot + ((FrjQName) e.nextElement()).getFullString();
                dot = ".";
            }
        } else {
            result = getStr();
        }
        return result;
    }

    // About relations.
    public void setRelatedObjects() {
        childQNames = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        file = null;
        referrer = null;
        parentQName = null;
        childQNames = null;
        dot = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.DEFINE_PACKAGE)
            || relName.equals(FrjRelations.IMPORT_PACKAGE)) {
            file = (FrjFile) obj;
        } else if (relName.equals(FrjRelations.QNAME_QNAME)) {
            if (rel.get1stObject().getID() == this.getID()) {
                // Argument obj is a child.
                childQNames.addElement(obj);
            } else {
                // Argument obj is a parent.
                parentQName = (FrjQName) obj;
            }
        } else if (relName.equals(FrjRelations.QNAME_DOT)) {
            dot = (FrjOperator) obj;
        } else if (relName.equals(FrjRelations.TYPE_QNAME) ||
                   relName.equals(FrjRelations.REFER_QNAME)) {
            if (referrer != null) {
                String msg = "QName is already referred to other object.\n";
                msg += "  this: " + this + "\n";
                msg += "  old: " + referrer + "\n";
                msg += "  new: " + obj;
                throw new NoSuchRelatedObjectException(msg);                            
            } else {
                referrer = obj;
            }
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(childQNames);
        sorter.sort();
    }

    /** Traverses <I>define_package</I> or <I>import_package</I>. */
    public FrjFile getFile() {
        return file;
    }
    /** Traverses <I>refer_qname</I> or <I>type_qname</I>. */
    public FrjObject getReferrer() {
        return referrer;
    }
    /**
     * Traverses <I>qname_qname</I>
     * @return parent FrjQName
     */
    public FrjQName getParentQName() {
        return parentQName;
    }
    /**
     * Traverses <I>qname_qname</I>
     * @return child FrjQNames
     */
    public Enumeration getChildQNames() {
        return childQNames.elements();
    }
    /** Traverses <I>qname_dot</I>. */
    public FrjOperator getDot() {
        return dot;
    }

    /** return parent object if exists */
    public FrjObject getParentObject() {
        if (parentQName != null) {
            return parentQName;
        } else if (file != null) {
            return file;
        } else {
            return referrer;
        }
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);

        tm.doMyself(this);

        if (dot != null) {
            traverseObjects(getChildQNames(), tm);
            dot.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }
}
