/* 
 * Program:             $RCSfile: FrjRelation.java,v $  $Revision: 41.9 $
 * 
 * Purpose:             The classes about relations.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjRelation.java,v 41.9 2003/02/15 11:56:29 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjRelation implements Loadable {
    // Class Variables
    protected String            relName;

    // Attributes
    public long                 id;
    public long                 oid1;
    public long                 oid2;
    /** reference counter of related objects */
    private int                         refCount = 0;

    // Constructors
    public FrjRelation() {
        this.relName = null;
    }

    public FrjRelation(String relName) {
        this.relName = relName;
    }

    public FrjRelation(FrejaEntry entry, String relName, DBRelObject obj) {
        this.relName = relName;

        /* RelationClassξȰäƤ줾java饹ѰդƤ餺
           FrjObjectȤͤۤʤ롣Ȥ߷פɬפȤʤ */
        /* relation  id ɤ߹߻˺ */
        id = FrjID.generateRelationID(entry, relName);
        oid1 = FrjID.generateObjectID(entry, obj.getIntValueAt(0));
        oid2 = FrjID.generateObjectID(entry, obj.getIntValueAt(1));

        if (oid1 < 0) {
            FrjWorld.error("Object " + oid1
                                  + " is not instatiated.");
            return;
        }
        if (oid2 < 0) {
            FrjWorld.error("Object " + oid2
                                  + " is not instatiated.");
            return;
        }

        FrjExtents.addRelation(this);
        refCount = 2;
        add();
    }

    public FrjRelation(String relName, FrjObject obj1, FrjObject obj2) {
        this.relName = relName;
        FrejaEntry entry1 = obj1.getFrejaEntry();
        if (entry1 == obj2.getFrejaEntry()) {
            // obj1 and obj2 in the same source
            id = FrjID.generateRelationID(entry1, relName);
        }
        else {
            id = FrjID.generateRelationID();
        }
        if (obj1 == null || obj2 == null) {
            return;
        }
        oid1 = obj1.getID();
        oid2 = obj2.getID();

        refCount = 2;
        FrjExtents.addRelation(this);
        add();
    }

    // Class methods
    /** return the class name of this object */
    public String getRelationName() {
        return relName; 
    }

    /** return the class name of this object */
    public long getID() {
        return id; 
    }
    public FrjObject get1stObject() {
        return FrjExtents.getObjectByID(oid1);
    }
    public FrjObject get2ndObject() {
        return FrjExtents.getObjectByID(oid2);
    }

    void unlinkRelatedObject() {
        if (--refCount == 0) {
            FrjExtents.removeRelation(this);
        }
    }

    /**
     * Loads objects from SDB.
     */
    public void load(FrejaEntry entry, String relName) {
        DBRelObject obj = null;

        DBQuery dbQuery = new DBQuery(entry, relName);
        try {
            if (!dbQuery.trylock4Reading()) {
                return;
            }
            while ((obj = dbQuery.getARelTuple()) != null) {
                new FrjRelation(entry, relName, obj);
            }
            dbQuery.unlock4Reading();
        }
        catch (DBException e) {
            FrjWorld.warning(e.getMessage());
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Saves objects to SDB.
     */
    public void save(FrejaEntry entry, String relName) {
        DBQuery dbQuery = new DBQuery(entry, relName);
        try {
            if (!dbQuery.trylock4Writing()) {
                return;
            }
                        
            VectorSorter sorter
                = new SortByRID(FrjExtents.getRelations(entry.getFrejaID(),
                                                        relName));
            sorter.sort();
            for (Enumeration e = sorter.getEnumeration();
                 e.hasMoreElements();) {
                FrjRelation rel = (FrjRelation) e.nextElement();
                dbQuery.writeARelTuple(DBRelObject.getDBRelObject(rel));
            }
            dbQuery.unlock4Writing();
        }
        catch (DBException e) {
            FrjWorld.warning(e.getMessage());
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    protected String toSdbline() {
        String res = "";
        res += FrjID.getFrejaDataID(id) + JDBConsts.SDB_ATTR_SEPARATOR;
        res += FrjID.getFrejaDataID(oid1) + JDBConsts.SDB_ATTR_SEPARATOR;
        res += FrjID.getFrejaDataID(oid2) + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    /**
     * add this relation to objects
     */ 
    protected void add() {
        FrjObject obj1 = FrjExtents.getObjectByID(oid1);
        FrjObject obj2 = FrjExtents.getObjectByID(oid2);
        if (obj1 != null && obj2 != null) {
            obj1.relations.addElement(this);
            obj2.relations.addElement(this);
        }
    }

    // Instance methods
    public FrjObject getRelatedObject(FrjObject obj) {
        if (obj.getID() == oid1) {
            return FrjExtents.getObjectByID(oid2);
        }
        else if (obj.getID() == oid2) {
            return FrjExtents.getObjectByID(oid1);
        }
        else {
            return null;
        }
    }

    public String toString() {
        return "[" + relName + "] " + "ID: " + id
            + ", OID1: " + oid1 + ", OID2: " + oid2;
    }
}
