/* 
 * Program:             $RCSfile: FrjRelations.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The classes about relations.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjRelations.java,v 41.5 2003/02/12 03:39:36 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjRelations extends Vector {
    /** Return all related objects contained this relations. */
    Enumeration getRelatedObjects(FrjObject obj) {
        Vector  v = new Vector();
        for (Enumeration e = this.elements();
             e.hasMoreElements();) {
            v.addElement(((FrjRelation) e.nextElement())
                         .getRelatedObject(obj));
        }
        return v.elements();
    }

    /** Return all the 2nd related objects contained this relations. */
    Enumeration get2ndRelatedObjects(FrjObject obj) {
        Vector  v = new Vector();
        for (Enumeration e = this.elements();
             e.hasMoreElements();) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            if (rel.get1stObject().getID() == obj.getID()) {
                v.addElement(rel.getRelatedObject(obj));
            }
        }
        return v.elements();
    }

    /** Return all the 1st related objects contained this relations. */
    Enumeration get1stRelatedObjects(FrjObject obj) {
        Vector  v = new Vector();
        for (Enumeration e = this.elements();
             e.hasMoreElements();) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            if (rel.get2ndObject().getID() == obj.getID()) {
                v.addElement(rel.getRelatedObject(obj));
            }
        }
        return v.elements();
    }

    // Constants.
    // These constants and table are converted by "DBTable.c".
    public static String
        DEFINE_PACKAGE = "define_package",
        IMPORT_PACKAGE = "import_package",
        DEFINE_CLASS = "define_class",
        IMPLEMENT_INTERFACE = "implement_interface",
        EXTEND_CLASS = "extend_class",
        CLASS_MODIFIER = "class_modifier",
        CLASS_VARIABLE = "class_variable",
        CLASS_ROUTINE = "class_routine",
        CLASS_SBLOCK = "class_sblock",
        CLASS_ARGUMENT = "class_argument",              /* JDK 1.1. */
        CLASS_CLASS = "class_class",                    /* JDK 1.1. */
        CLASS_INTERFACE = "class_interface",            /* JDK 1.1. */
        DEFINE_INTERFACE = "define_interface",
        EXTEND_INTERFACE = "extend_interface",
        INTERFACE_MODIFIER = "interface_modifier",
        INTERFACE_VARIABLE = "interface_variable",
        INTERFACE_ROUTINE = "interface_routine",
        INTERFACE_CLASS = "interface_class",            /* JDK 1.1. */
        INTERFACE_INTERFACE = "interface_interface",    /* JDK 1.1. */
        VARIABLE_MODIFIER = "variable_modifier",
        VARIABLE_TYPE = "variable_type",
        INITIALIZE_VARIABLE = "initialize_variable",
        PARAMETER_VARIABLE = "parameter_variable",
        RETURN_TYPE = "return_type",
        THROW_TYPE = "throw_type",
        ROUTINE_BODY = "routine_body",
        ROUTINE_MODIFIER = "routine_modifier",
        TYPE_CLASS = "type_class",
        TYPE_INTERFACE = "type_interface",
        TYPE_TYPE = "type_type",
        TYPE_QNAME = "type_qname",
        QNAME_QNAME = "qname_qname",
        QNAME_DOT = "qname_dot",
        SBLOCK_BODY = "sblock_body",
        LOCAL_CLASS = "local_class",                    /* JDK 1.1. */
        LOCAL_VARIABLE = "local_variable",
        STATEMENT_LABEL = "statement_label",
        STATEMENT_STATEMENT = "statement_statement",
        STATEMENT_EXPRESSION = "statement_expression",
        REFER_LABEL = "refer_label",
        CASE_VALUE = "case_value",
        EXPRESSION_OPERAND = "expression_operand",
        EXPRESSION_OPERATOR = "expression_operator",
        EXPRESSION_EXPRESSION = "expression_expression",
        REFER_VARIABLE = "refer_variable",
        REFER_LITERAL = "refer_literal",
        REFER_SPECIAL = "refer_special",
        REFER_TYPE = "refer_type",
        REFER_QNAME = "refer_qname",
        HAS_TYPE = "has_type",
        INVOKE_ROUTINE = "invoke_routine",
        SYMBOL_COMMENT = "symbol_comment";
}
