/* 
 * Program:             $RCSfile: FrjSBlock.java,v $  $Revision: 41.4 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjSBlock.java,v 41.4 2003/02/15 11:56:30 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjSBlock extends FrjObject {
    // Class variables
    /** A class name of this class. */
    public final static String  CLASS_NAME = "SBlock";

    /** table from a sort to string */
    protected static String     sortNameTable[];        // initialized by static block

    // Attribute
    public int                          sort;

    // Related object
    protected FrjClass          definedClass = null;
    protected FrjStatement      body = null;

    // Constatnts
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int                _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int             STATIC_INITIALIZER = _i++;
    public final static int             INSTANCE_INITIALIZER = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // sentinel

    // Constructors
    public FrjSBlock(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        //this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }

    public FrjSBlock(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }

    public FrjSBlock() {
        super();
        sort = INVALID_ID;
    }

    // Class methods
    /** return the class name of this object */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjSBlock(entry, obj);
    }

    // Instance method
    /** return the class name dynamically */
    public String getClassName() {
        return CLASS_NAME;
    }

    /** return the sort */
    public int getSort() {
        return sort;
    }

    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }
        
    // Methods about relation
    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        definedClass = null;
        body = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.CLASS_SBLOCK)) {
            definedClass = (FrjClass) obj;
        }
        else if (relName.equals(FrjRelations.SBLOCK_BODY)) {
            body = (FrjStatement) obj;
        }
        else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        }
        else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
    }

    /** Traverses <I>class_sblock</I>. */
    public FrjClass getDefinedClass() {
        return definedClass;
    }
    /** Traverses <I>sblock_body</I>. */
    public FrjStatement getBody() {
        return body;
    }
    public FrjObject getParentObject() {
        return definedClass;
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjSBlock) obj).sort = sort;
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        tm.doMyself(this);

        if (body != null) {
            body.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }

    /* static initializer */
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];

        sortNameTable[STATIC_INITIALIZER] = "Static Initializer";
        sortNameTable[INSTANCE_INITIALIZER] = "Instance Initializer";
    }

}
