/* 
 * Program:             $RCSfile: FrjTypeList.java,v $  $Revision: 41.6 $
 * 
 * Purpose:             The class corresponding to Expression.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjTypeList.java,v 41.6 2003/02/12 03:39:38 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.util.*;

/**
  * A class represents list of FrjType.
  * This is used on routine's parameter or argument list and so on.
  */
public class FrjTypeList {
    private Vector list;

    public FrjTypeList() {
        list = new Vector();
    }

    public void addType(FrjType type) {
        if (type == null) {
            FrjWorld.error("You add null to FrjTypeList.");
            return;
        }
        list.addElement(type);
    }

    public Enumeration elements() {
        return list.elements();
    }

    public int size() {
        return list.size();
    }

    public boolean equals(FrjTypeList tl) {
        if (this.size() != tl.size()) {
            return false;
        }

        Enumeration e1 = this.elements();
        Enumeration e2 = tl.elements();
        while (e1.hasMoreElements()) {  // e1 and e2 are the same size
            FrjType t1 = (FrjType) e1.nextElement();
            FrjType t2 = (FrjType) e2.nextElement();
                
            if (!t1.equals(t2)) {
                return false;
            }
        }       
        return true;
    }

    public boolean isCompatible(FrjTypeList tl) {
        if (this.size() != tl.size()) {
            return false;
        }

        Enumeration e1 = this.elements();
        Enumeration e2 = tl.elements();
        while (e1.hasMoreElements()) {  // e1 and e2 are the same size
            FrjType t1 = (FrjType) e1.nextElement();
            FrjType t2 = (FrjType) e2.nextElement();
                
            if (!t1.isCompatible(t2)) {
                return false;
            }
        }       
        return true;
    }

    public String toString() {  // for Debug
        String result = "";
        for (Enumeration e = elements();
             e.hasMoreElements();) {
            result += ", ";
            FrjType t = (FrjType) e.nextElement();
            result += t.getStr();
            for (FrjType t2 = t.getChildType();
                 t2 != null; t2 = t2.getChildType()) {
                result += t2.getStr();
            }
        }

        // if (result.length() > 2) {      // 2 is the lenght of ", "
        //     return result.substring(2);
        // } else {
        //     return result;
        // }

        // 2 means the lenght of ", ".
        return (2 < result.length()) ? result.substring(2) : result;
    }
}       
