/* 
 * Program:             $RCSfile: FrjWorld.java,v $  $Revision: 41.20 $
 * 
 * Purpose:             The main class of Freja.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjWorld.java,v 41.20 2003/09/07 05:33:41 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjWorld {
    private static String           SDB_PATH = "SDB";
    private final static String     J_MODEL_DIR = "J-MODEL";
    public static String            J_MODEL_PATH = SDB_PATH + File.separator +
        J_MODEL_DIR;

    private static Loader           loader = new DefaultLoader();
    private static boolean          autoLoadingFlag = false;
    private static boolean          verbose = false;

    public static final int         ENC_NONE = 0,
                                    ENC_FILE = 1,
                                    ENC_SYSTEM = 2;
    private static final int        ENC_FIRST_MEMBER = ENC_NONE,
                                    ENC_LAST_MEMBER = ENC_SYSTEM;
    private static int              encStyle = ENC_SYSTEM;
    private static String           systemEncoding = "ASCII";

    /**
     * If this flag is true,
     * FID part (which is in ID, see FrjID.java) will be set to 0.
     */
    private static boolean FIDTOZERO = false;

    public static void setFidToZero(boolean fidToZero) {
        FIDTOZERO = fidToZero;
    }

    public static boolean isFidToZero() {
        return FIDTOZERO;
    }

    /** Set SDB path. */
    public static void spdSetSDBDir(String dir) {
        SDB_PATH = dir;
        J_MODEL_PATH = SDB_PATH + File.separator + J_MODEL_DIR;
    }

    /** Database Class/Rel table. */
    public static DBTable dbTable = null;

    /** Sets loader. */
    public static void setLoader(Loader ldr) {
        loader = ldr;
    }

    /** Returns loader. */
    public static Loader getLoader() {
        return loader;
    }

    /** Loads objects from <I>entry</I>. */
    public static void load(FrejaEntry entry) {
        if (!FrjExtents.isLoaded(entry)) {
            if (FrjWorld.isVerbose()) {
                FrjWorld.message("loading " + entry.getFrejaPath());
            }
            loader.load(entry);
            FrjExtents.addFrejaEntry(entry);
        }
    }

    public static void setVerbose(boolean v) {
        verbose = v;
    }
    public static boolean isVerbose() {
        return verbose;
    }

    public static void setEncodingStyle(int es) {
        if (ENC_FIRST_MEMBER <= es && es <= ENC_LAST_MEMBER) {
            encStyle = es;
        } else {
            FrjWorld.warning(es + ": Invalid encoding style.");
        }
    }
    public static int getEncodingStyle() {
        return encStyle;
    }
    public static void setSystemEncoding(String enc) {
        systemEncoding = enc;
    }
    public static String getSystemEncoding() {
        return systemEncoding;
    }

    /** Saves object of <I>entry</I>. */
    public static void save(FrejaEntry entry) {
        // save objects
        for (Enumeration e = FrjExtents.getClassNames();
             e.hasMoreElements();) {
            String name = (String) e.nextElement();
            try {
                Class c = Class.forName("Japid.Freja.Frj" + name);
                FrjObject saver = (FrjObject) c.newInstance();
                saver.save(entry, name);
            }
            catch (ClassNotFoundException ext) {
                FrjWorld.warning("Java class " + "Japid.Freja.Frj"
                                        + name + " is not defined");
            }
            catch (Exception ext) {
                ext.printStackTrace();
            }
        }
        // save relations
        FrjRelation saver = new FrjRelation();
        for (Enumeration e = FrjExtents.getRelationNames();
             e.hasMoreElements();) {
            String relName = (String) e.nextElement();
            saver.save(entry, relName);
        }
        // save infofile
        entry.saveJpdInfo();
    }

    /**
     * If flag is true, uses Creja.<br>
     * NOTE: This method is defined for historical compatibility.
     * Using setAutoLoading() method is recommended.
     */
    public static void setUsingCreja(boolean flag) {
        autoLoadingFlag = flag;
    }
    /**
     * @return whether uses Creja
     * NOTE: This method is defined for historical compatibility.
     * Using isAutoLoading() method is recommended.
     */
    public static boolean usesCreja() {
        return autoLoadingFlag;
    }

    /**
     * If flag is true, auto loading other classes.<br>
     */
    public static void setAutoLoading(boolean flag) {
        autoLoadingFlag = flag;
    }

    /**
     * @return whether auto loading
     */
    public static boolean isAutoLoading() {
        return autoLoadingFlag;
    }

    /**
     * Prints debug message to stderr.
     */
    public static void debug(boolean debug, String s) {
        if (debug) {
            System.out.flush();
            System.err.flush();
            System.err.println("Debug: " + s);
            System.err.flush();
        }
    }

    /**
     * Prints error message to stderr.
     */
    public static void error(String s) {
        System.out.flush();
        System.err.flush();
        System.err.println("Error: " + s);
        System.err.flush();
    }

    /**
     * Prints error message to stderr, and then exit.
     */
    public static void errorExit(String s) {
        System.out.flush();
        System.err.flush();
        System.err.println("Error: " + s);
        System.err.flush();
        System.exit(1);
    }

    /**
     * Prints message to stderr.
     */
    public static void message(String s) {
        System.out.flush();
        System.err.flush();
        System.err.println("Japid: " + s);
        System.err.flush();
    }

    /**
     * Prints warning message to stderr.
     */
    public static void warning(String s) {
        System.out.flush();
        System.err.flush();
        System.err.println("Warning: " + s);
        System.err.flush();
    }

    /**
     * Prints verbose warning message to stderr.
     */
    public static void vwarning(String s) {
        if (verbose) {
            warning(s);
        }
    }

    static {
        dbTable = new DBTable();
    }
}
