/* 
 * Program:             $RCSfile: SetExprType.java,v $  $Revision: 41.13 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: SetExprType.java,v 41.13 2003/02/15 11:56:33 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.util.*;

public class SetExprType {
    public static void main(String args[]) {
        if (args.length == 0) {
            usage();
            return;
        }
        FrjWorld.setAutoLoading(true);

        Vector classes = new Vector();
        boolean verify = false;
        boolean verbose = false;
        boolean recursive = false;
        boolean clear = false;
        int i = 0;
        for (/* Empty. */; i < args.length; i++) {
            if (args[i].equals("-c")) {
                clear = true;
            } else if (args[i].equals("-nc")) {
                FrjWorld.setAutoLoading(false);
            } else if (args[i].equals("-p")) {
                String pkgName = args[++i];
                for (Enumeration e = Creja.getClasses(pkgName);
                     e.hasMoreElements(); /* Empty. */) {
                    String className = (String) e.nextElement();

                    classes.addElement(pkgName + "." + className);
                }
            } else if (args[i].equals("-rp")) {
                Vector pkgs = new Vector();

                pkgs.addElement(args[++i]);
                for (Enumeration e = pkgs.elements();
                     e.hasMoreElements();) {
                    String pkgName = (String) e.nextElement();
                    for (Enumeration e2 = Creja.getClasses(pkgName);
                         e2.hasMoreElements(); /* Empty. */) {
                        String className = (String) e2.nextElement();
                        classes.addElement(pkgName + "." + className);
                    }
                    for (Enumeration e2 = Creja.getSubPackages(pkgName);
                         e2.hasMoreElements(); /* Empty. */) {
                        pkgs.addElement(e2.nextElement());
                    }
                }
            } else if (args[i].equals("-v")) {
                verbose = true;
                FrjWorld.setVerbose(true);
            } else if (args[i].equals("-Dcreja")) {
                Creja.debug = true;
            } else if (args[i].equals("-De")) {
                FrjExpression.debug = true;
            } else if (args[i].equals("-Dfrejaentry")) {
                FrejaEntry.debug = true;
            } else if (args[i].equals("-V")) {
                verify = true;
            } else if (args[i].startsWith("-")) {
                usage();
                System.exit(1);
            } else {
                break;
            }
        }

        for (/* Empty. */; i < args.length; i++) {
            if (args[i].endsWith(".java")) {
                int index = args[i].lastIndexOf(".java");
                classes.addElement(args[i].substring(0, index));
            } else {
                classes.addElement(args[i]);
            }
        }
                
        System.err.println("Set expression's type of " +
                           classes.size() + " class(es).");

        /* SDBǡɤ߹ */
        Loader defaultLoader = new DefaultLoader();
        ExternalAccessibleLoader externalLoader
            = new ExternalAccessibleLoader();
        for (Enumeration e = classes.elements();
             e.hasMoreElements(); /* Empty. */) {
            String className = (String) e.nextElement();
            FrjWorld.setLoader(defaultLoader);
            FrejaEntry entry = Creja.getFrejaEntry(className);
            if (entry == null) {
                FrjWorld.error("No such SDB " + className);
                continue;
            }
            FrjWorld.message("SetExprType: \"" + entry.getFrejaPath() + "\".");
            if (verbose) {
                FrjWorld.message("Reading \"" + entry.getFrejaPath() + "\".");
            }
            if (externalLoader.isPartialLoaded(entry)) {
                if (verbose) {
                    FrjWorld.message("Loading \"" + entry.getFrejaPath() + "\".");
                }
                externalLoader.loadOthers(entry);

            } else {
                FrjWorld.load(entry);
            }

            if (verify) {
                for (Enumeration e2 = FrjExtents.getObjects(FrjExpression.getClassNameStatically());
                     e2.hasMoreElements(); /* Empty. */) {
                    FrjExpression expr = (FrjExpression)e2.nextElement();

                    if (expr.getSort() == FrjExpression.EMPTY ||
                        expr.getSort() == FrjExpression.LISTED) {
                        continue;
                    }
                    if (expr.getType() == null) {
                        System.out.println("No type: " + expr);
                    }
                }
                continue;
            }
            FrjWorld.setLoader(externalLoader);
            if (verbose) {
                FrjWorld.message("Set expression's type.");
            }
            /* (Expression)ηꤹ롥 */
            FrjExpression.setTypes(entry);
            FrjType.setReferringClassInterfaces(entry);
            FrjExpression.setInvokeMethods(entry);

            if (verbose) {
                FrjWorld.message("Saving \"" + entry.getFrejaPath() + "\".");
            }
            FrjWorld.save(entry);
            // FrjWorld.message("Finish saving.");
            if (verbose) {
                Runtime runtime = Runtime.getRuntime();
                System.err.println("Object: " + FrjExtents.getObjectNums()
                                   + " Relation: " + FrjExtents.getRelationNums());
                System.err.println("Total: " + runtime.totalMemory()
                                   + " Free: " + runtime.freeMemory());
            }
            if (clear) {
                FrjExtents.clear();
            }
        }
    }

    static void usage() {
        System.err.println("Usage: [-c][-nc][-p packageName][-rp pckageName][-v]" +
                           "[-D...][-V] filenames");
        System.err.println("   -nc: Not use Creja.");
        System.err.println("   -v: Verbose mode.");
        System.err.println("   -V: Verify SDB data.");
        System.err.println("   -De: Print debug data for Expression.");
        System.err.println("   -Dcreja: Print debug data for Creja.");
        System.err.println("   -Dfrejaentry: Print debug data for FrejaEntry.");
    }
}
