/* 
 * Program:             $RCSfile: GetClassSlice.java,v $  $Revision: 40.5 $
 * 
 * Purpose:             A test class of "Japid.Freja".
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: GetClassSlice.java,v 40.5 2003/02/15 11:56:44 yamamoto Exp yamamoto $
 */

package Japid.Tools;

import java.util.*;
import Japid.Freja.*;

public class GetClassSlice {
    static Vector exceptPackages = new Vector();

    public static void main(String args[]) {
        try {
            internalMain(args);
        }
        finally {
            Runtime runtime = Runtime.getRuntime();
            System.err.println("Object: " + FrjExtents.getObjectNums()
                               + " Relation: " + FrjExtents.getRelationNums());
            System.err.println("Total: " + runtime.totalMemory()
                               + " Free: " + runtime.freeMemory());
        }
    }


    private static void internalMain(String args[]) {
        if (args.length == 0) {
            return;
        }
        FrjWorld.setUsingCreja(true);

        boolean verbose = false;
        int i = 0;
        for (; i < args.length; i++) {
            if (args[i].equals("-e")) {
                exceptPackages.addElement(args[++i]);
            }
            else if (args[i].equals("-v")
                     || args[i].equals("-verbose")) {
                FrjWorld.setVerbose(true);
                verbose = true;
            }
            else if (args[i].startsWith("-")) {
                usage();
                System.exit(1);
            }
            else {
                break;
            }
        }
        /* SDBǡɤ߹ */
                
        Loader loader =  new DefaultLoader();
        FrjWorld.setLoader(loader);
        for (; i < args.length; i++) {
            String className = args[i];
            FrejaEntry entry = Creja.getFrejaEntry(className);
            if (entry == null) {
                FrjWorld.error("No such SDB " + className);
                continue;
            }
            FrjWorld.load(entry);
            FrjFile f = (FrjFile) FrjExtents.getObjects("File").nextElement();
            /*
              for (Enumeration e = FrjExtents.getObjects("Type");
              e.hasMoreElements();) {
              FrjType type = (FrjType) e.nextElement();
              if (verbose) {
              System.err.println(type.getStr());
              }
              type.getReferringClassInterface();
              }
            */
            for (Enumeration e = f.getClasses();
                 e.hasMoreElements();) {
                FrjClass c = (FrjClass) e.nextElement();
                if (className.endsWith(c.getStr())) {
                    System.out.println(c.getFullClassName());
                    c.traverse(new TraverseType());
                }
            }
            for (Enumeration e = f.getInterfaces();
                 e.hasMoreElements();) {
                FrjInterface itf = (FrjInterface) e.nextElement();
                if (className.endsWith(itf.getStr())) {
                    System.out.println(itf.getFullClassName());
                    itf.traverse(new TraverseType());
                }
            }
        }
    }

    static boolean excepts(String fullClassName) {
        for (Enumeration e = exceptPackages.elements();
             e.hasMoreElements();) {
            String pkg = (String) e.nextElement();
            if (fullClassName.startsWith(pkg)) {
                return true;
            }
        }
        return false;
    }

    static void usage() {
        System.err.println("options: [-e package] [-v] filenames");
        System.err.println(" -e package:  don't traverse package.");
        System.err.println(" -v/-verbose: verbose mode.");
    }

}

class TraverseType implements TraverseMethod {
    static Dictionary types = new Hashtable();
    private FrjClassInterface   curClass;
    static int line = 0;
    // types->key:    class name
    //      ->value:  dict->key:    printed class
    //                    ->value:  any
    private int depth;

    TraverseType() {
        depth = 0;
    }
    TraverseType(int d) {
        depth = d;
    }

    public void doBeforeTraverse(FrjObject obj) {
    }

    public void doMyself(FrjObject obj) {
        if (obj instanceof FrjType) {
            FrjType type = (FrjType) obj;
            if (type.getSort() == FrjType.OBJECT) {
                Dictionary printed = (Dictionary) types.get(type.getStr());
                if (printed == null) {
                    Dictionary d = new Hashtable();
                    d.put(curClass, curClass);
                    types.put(type.getStr(), d);

                    print(type, false);

                    FrjClassInterface rc = type.getReferringClassInterface();
                    if (rc != null && !rc.getStr().equals("Object")) {
                        if (!GetClassSlice.excepts(rc.getFullClassName())) {
                            rc.traverse(new TraverseType(depth +1));
                        }
                    }
                }
                else if (printed.get(curClass) == null) {
                    printed.put(curClass, curClass);
                    print(type, true);
                }
            }
        }
        else if (obj instanceof FrjClassInterface) {
            curClass = (FrjClassInterface) obj;
        }
    }

    public void doAfterTraverse(FrjObject obj) {
    }

    private void print(FrjType type, boolean suppressed) {
        putSpace(depth *2);
        boolean isSuper = false;
        for (Enumeration e
                 = curClass.getSuperClassInterfaceTypes();
             e.hasMoreElements();) {
            FrjType t = (FrjType) e.nextElement();
            if (t.getStr().equals(type.getStr())) {
                isSuper = true;
                System.out.print(" +");
                break;
            }
        }
        if (!isSuper) {
            putSpace(2);
        }
        String  fullClassName;
        FrjClassInterface ci = type.getReferringClassInterface();
        fullClassName = (ci == null) ? type.getStr()
            : ci.getFullClassName();
        System.out.println(fullClassName
                           + ((suppressed)? " [suppressed]": ""));
    }

    void putSpace(int n) {
        for (int i = 0; i < n; i++) {
            System.out.print(" ");
        }
    }
}
