/* 
 * Program:             $RCSfile: Syab2.java,v $  $Revision: 40.11 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Syab2.java,v 40.11 2003/02/15 11:56:43 yamamoto Exp yamamoto $
 */

package Japid.Tools;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import Japid.Freja.*;

class Syab2 extends Frame {
    FrejaEntry                          jfile;
    String                                      className;
    static int                          frejaID = 1;
    // private ObjectsChoice                    objectsChoice;
    private SpecifiedObjectsListBox             specifiedObjectsListBox;
    private RelationsListBox            relationsListBox;
    private ObjectAttrsText                     specifiedObjectAttrsText;
    private ObjectAttrsText                     relatedObjectAttrsText;
    private SourceCodeText                      sourceCodeText;
    // private Button                                   quitButton;
    // private Button                                   backButton;
    private Button                                      whereButton;
    private CheckboxMenuItem            sortOffsetMenuItem;
    private CheckboxMenuItem            sortIDMenuItem;
    // Stack                                            objectStack;
    Font        appFont = new Font("Helvitica", Font.PLAIN, 14);
    Font        textFont = new Font("Helvitica", Font.PLAIN, 14);

    final VectorSorter sortByOffset = new SortByOffset();
    final VectorSorter sortByID = new SortByID();
    VectorSorter sorter = sortByOffset;

    private final String        M_FILE = "File", M_RELOAD = "Reload",
        M_QUIT = "Quit", M_LIST = "List",
        M_OPTION = "Option", M_SORT_OFFSET = "Sort By Offset",
        M_SORT_ID = "Sort By ID";


    Syab2(String className, String loaderString, int frejaID) {
        super("Syab --- Simple, Yet Another Browser");
        this.className = className;
        if (frejaID >= FrjID.MAX_FREJA) {
            this.frejaID = frejaID % FrjID.MAX_FREJA;
        } else {
            this.frejaID = frejaID;
        }

        if (loaderString != null) {
            try { 
                Class c = Class.forName(loaderString);
                FrjWorld.setLoader((Loader) c.newInstance());
            }
            catch (Exception e) {
                System.err.println("No Such Loader Class: " + loaderString);
                e.printStackTrace();
                System.exit(1);
            }
        }
        load();
        createWidget();
        this.pack();
    }

    Syab2(String className) {
        this(className, null, frejaID);
    }

    void load() {
        jfile = Creja.getFrejaEntry(className);
        if (jfile == null) {
            FrjWorld.error("No such SDB " + className);
            System.exit(1);
        }

        FrjWorld.load(jfile);

        System.out.println("loaded.");
        // objectStack = new Stack();
    }

    void reload() {
        FrjExtents.clear();

        specifiedObjectsListBox.removeAll();
        relationsListBox.removeAll();
        specifiedObjectAttrsText.clearText();
        relatedObjectAttrsText.clearText();

        load();
        sourceCodeText.reload();
    }

    void createWidget() {
        setFont(appFont);

        ListBoxItemListener lbIListener = new ListBoxItemListener();
        ListBoxActionListener lbAListener = new ListBoxActionListener();
        ButtonActionListener bAListener = new ButtonActionListener();
        CheckboxMenuItemItemListener cmiIListener = new CheckboxMenuItemItemListener();
        ChoiceItemListener cIListener = new ChoiceItemListener();
        MenuItemActionListener miAListener = new MenuItemActionListener();

        MenuBar menubar = new MenuBar();
        Menu fileMenu = new Menu(M_FILE);
        MenuItem m_reload = new MenuItem(M_RELOAD);
        m_reload.addActionListener(miAListener);
        fileMenu.add(m_reload);
        MenuItem m_quit = new MenuItem(M_QUIT);
        m_quit.addActionListener(miAListener);
        fileMenu.add(m_quit);
        menubar.add(fileMenu);

        Menu listMenu = new Menu(M_LIST);
        VectorSorter s = new SortStrings(FrjExtents.getClassNames());
        s.sort();
        for (Enumeration e = s.getEnumeration(); e.hasMoreElements();) {
            MenuItem m_classname = new MenuItem((String) e.nextElement());
            m_classname.addActionListener(miAListener);
            listMenu.add(m_classname);
        }
        menubar.add(listMenu);
                
        Menu optionMenu = new Menu(M_OPTION);
        sortOffsetMenuItem = new CheckboxMenuItem(M_SORT_OFFSET);
        sortOffsetMenuItem.addItemListener(cmiIListener);
        sortOffsetMenuItem.setState(true);
        sortIDMenuItem = new CheckboxMenuItem(M_SORT_ID);
        sortIDMenuItem.addItemListener(cmiIListener);
        sortIDMenuItem.setState(false);
        optionMenu.add(sortOffsetMenuItem);
        optionMenu.add(sortIDMenuItem);
        menubar.add(optionMenu);
        setMenuBar(menubar);
                
        GridBagLayout gbl = new GridBagLayout();

        setLayout(gbl);

        //              quitButton = new Button("QUIT");
        //              quitButton.addActionListener(bAListener);
        //              backButton = new Button("BACK");
        //              backButton.addActionListener(bAListener);
        whereButton = new Button("where");
        whereButton.addActionListener(bAListener);
        //              objectsChoice = new ObjectsChoice(this);
        //              objectsChoice.addItemListener(cIListener);
        specifiedObjectsListBox
            = new SpecifiedObjectsListBox(this, 10);
        specifiedObjectsListBox.addItemListener(lbIListener);
        specifiedObjectsListBox.addActionListener(lbAListener);
        relationsListBox
            = new RelationsListBox(this, 10);
        relationsListBox.addItemListener(lbIListener);
        relationsListBox.addActionListener(lbAListener);
        sourceCodeText = new SourceCodeText(this, 20, 100);
        specifiedObjectAttrsText = new ObjectAttrsText(this, 10, 10);
        relatedObjectAttrsText = new ObjectAttrsText(this, 10, 10);
                
        GridBagConstraints gbcText = new GridBagConstraints();
        gbcText.weightx = 0;
        gbcText.weighty = 0;

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.insets = new Insets(2, 10, 15, 2);
        gbc.weightx = 0;
        gbc.weighty = 0;
        //
        // +------------------------------+
        // | +----+  +----+ +----+ +----+ |
        // | |    |  |    | |    | |    | |
        // | +----+  +----+ +----+ +----+ |
        // | src cd                  ===  |
        // | +--------------------------+ |
        // | |                          | |
        // | +--------------------------+ |
        // +------------------------------+

        int y = 0, x = 0;
        add(new Label(" "), gbl, gbcText, x++, y, 1, 1); // dummy
        //              add(objectsChoice, gbl, gbc, x++, y, 1, 1);
        //              add(backButton, gbl, gbc, center -1, y, 1, 1);
        //              add(quitButton, gbl, gbc, center +3, y, 1, 1);


        y++; x = 0;
        add(new Label("Specified Objects"), gbl, gbcText, x++, y, 1, 1);
        add(new Label("Attributes"), gbl, gbcText, x++, y, 1, 1);
        add(new Label("Relations"),  gbl, gbcText, x++, y, 1, 1);
        add(new Label("Attributes"), gbl, gbcText, x++, y, 1, 1);

        gbc.weightx = 0.8;
        gbc.weighty = 0.8;
        y++; x = 0;

        gbc.fill = GridBagConstraints.BOTH;
        add(specifiedObjectsListBox, gbl, gbc, x++, y, 1, 1);
        //              gbc.fill = GridBagConstraints.HORIZONTAL;
        add(specifiedObjectAttrsText, gbl, gbc, x++, y, 1, 1);

        gbc.fill = GridBagConstraints.BOTH;
        add(relationsListBox, gbl, gbc, x++, y, 1, 1);
        //              gbc.fill = GridBagConstraints.HORIZONTAL;
        add(relatedObjectAttrsText, gbl, gbc, x++, y, 1, 1);

        y++; x = 0;
        gbc.fill = GridBagConstraints.NONE;
        //              add(whereButton, gbl, gbc, center +3, y, 1, 1);
        gbcText.anchor = GridBagConstraints.SOUTHWEST;
        add(new Label("  Source Code"), gbl, gbcText, 0, y, 1, 1);
        gbcText.anchor = GridBagConstraints.CENTER;
        add(whereButton, gbl, gbcText, 3, y, 1, 1);

        y++;
        gbc.fill = GridBagConstraints.BOTH;
        add(sourceCodeText, gbl, gbc, 0, y, 4, 1);

        addWindowListener(new MyWindowListener());
    }

    void specifiedObjectClicked(FrjObject selected) {
        specifiedObjectAttrsText.setObject(selected);
        relationsListBox.setRelations(selected);
        relatedObjectAttrsText.clearText();
        sourceCodeText.setText(selected);
    }

    void specifiedObjectDoubleClicked(FrjObject selected) {
        specifiedObjectAttrsText.setObject(selected);
        relationsListBox.setRelations(selected);
        relatedObjectAttrsText.clearText();
        sourceCodeText.setText(selected);
    }

    void relationClicked(FrjObject selected) {
        relatedObjectAttrsText.setObject(selected);
        sourceCodeText.setText(selected);
    }
    void relationDoubleClicked(FrjObject selected) {
        specifiedObjectsListBox.setObject(selected);
        specifiedObjectDoubleClicked(selected);
        relatedObjectAttrsText.clearText();
        sourceCodeText.setText(selected);
    }

    /****************** Listeners ***********************/

    protected class ListBoxItemListener implements java.awt.event.ItemListener {
        public void itemStateChanged(java.awt.event.ItemEvent e) {
            Object e_source = e.getSource();

            if (e_source == specifiedObjectsListBox) {
                FrjObject selected
                    =  specifiedObjectsListBox.getSelectedObject();
                if (selected == null) {
                    System.err.println("No objects are selcted.");
                } else {
                    specifiedObjectClicked(selected);
                }
            } else if (e_source == relationsListBox) {
                FrjObject selected
                    =  relationsListBox.getRelatedObject();
                if (selected == null) {
                    System.err.println("No relaions are selcted.");
                } else {
                    relationClicked(selected);
                }
            }
        }
    }

    protected class ListBoxActionListener implements java.awt.event.ActionListener {
        public void actionPerformed(java.awt.event.ActionEvent e) {
            Object e_source = e.getSource();

            if (e_source == specifiedObjectsListBox) {
                FrjObject selected
                    =  specifiedObjectsListBox.getSelectedObject();
                if (selected == null) {
                    System.err.println("No objects are selcted.");
                } else {
                    specifiedObjectDoubleClicked(selected);
                }
            } else if (e_source == relationsListBox) {
                FrjObject selected
                    =  relationsListBox.getRelatedObject();
                relationDoubleClicked(selected);
            }
        }
    }

    protected class ButtonActionListener implements java.awt.event.ActionListener {
        public void actionPerformed(java.awt.event.ActionEvent e) {
            Object e_source = e.getSource();

            if (e_source == whereButton) {
                int     start = sourceCodeText.getSelectionStart();
                int     end = sourceCodeText.getSelectionEnd();
                specifiedObjectsListBox.setObjects(start, end);
                specifiedObjectAttrsText.clearText();
                relationsListBox.deleteAllRelations();
                relatedObjectAttrsText.clearText();
            }
        }
    }

    protected class CheckboxMenuItemItemListener implements java.awt.event.ItemListener {
        public void itemStateChanged(java.awt.event.ItemEvent e) {
            Object e_source = e.getSource();

            String m_label = ((CheckboxMenuItem) e_source).getLabel();
            if (m_label.equals(M_SORT_OFFSET)) {
                sorter = sortByOffset;
                sortIDMenuItem.setState(false);
            } else if (m_label.equals(M_SORT_ID)) {
                sorter = sortByID;
                sortOffsetMenuItem.setState(false);
            } else {
                                // error
            }
        }
    }

    protected class ChoiceItemListener implements java.awt.event.ItemListener {
        public void itemStateChanged(java.awt.event.ItemEvent e) {
            Object e_source = e.getSource();
        }
    }

    protected class MenuItemActionListener implements java.awt.event.ActionListener {
        public void actionPerformed(java.awt.event.ActionEvent e) {
            Object e_source = e.getSource();

            String m_label = ((MenuItem) e_source).getLabel();
            if (m_label.equals(M_RELOAD)) {
                reload();
            } else if (m_label.equals(M_QUIT)) {
                dispose();
                System.exit(0);
            } else {      // List Menu
                specifiedObjectsListBox.setObjects(m_label);
                specifiedObjectAttrsText.clearText();
                relationsListBox.deleteAllRelations();
                relatedObjectAttrsText.clearText();
            }
        }
    }

    /****************** Listeners end ***********************/

    private void add(Component c, GridBagLayout gbl, GridBagConstraints gbc,
                     int x, int y, int w, int h) {
        gbc.gridx = x;
        gbc.gridy = y;
        gbc.gridwidth = w;
        gbc.gridheight = h;
        gbl.setConstraints(c, gbc);
        add(c);
    }

    static long changeID(long id) {
        return FrjID.getID(frejaID, FrjID.getFrejaDataID(id));
    }

    public static void main(String args[]) {
        Syab2   s;
        String className, loaderString = null;
        int frejaID = 1;

        if (args.length == 0) {
            usage();
            System.exit(1);
        }

        int i = 0;
        for (; i < args.length; i++) {
            if (args[i].equals("-loader")) {
                loaderString = args[++i];
            } else if (args[i].startsWith("-frejaID")) {
                frejaID = Integer.parseInt(args[++i]);
            } else if (args[i].startsWith("-")) {
                usage();
                System.exit(1);
            } else {
                break;
            }
        }
        if (i == args.length) {
            usage();
            System.exit(1);
        }
        className = args[i];
                        
        s = new Syab2(className, loaderString, frejaID);
        s.show();
    }

    static void usage() {
        System.err.println("Usage: java Japid.Tools.Syab2 [-loader loader] " +
                           " [-frejaID num] ClassName");
    }

}

class ObjectsChoice extends Choice {
    Syab2       syab;
    ObjectsChoice(Syab2 syab) {
        super();
        for (Enumeration e = FrjExtents.getClassNames();
             e.hasMoreElements();) {
            super.addItem((String) e.nextElement());
        }
        this.syab = syab;
    }
}

abstract class ObjectsListBox extends java.awt.List {
    Vector                      objects;
    Syab2                       syab;

    ObjectsListBox(Syab2 syab, int size) {
        super(size, false);
        objects = new Vector();
        setFont(syab.textFont);
        this.syab = syab;
    }

    void addObject(FrjObject obj) {
        if (obj != null) {
            objects.addElement(obj);
            super.add("[" + obj.getClassName() + "] "
                      + Syab2.changeID(obj.getID()));
        }
    }

    void deleteAllObjects() {
        objects = new Vector();
        super.removeAll();
    }

    FrjObject getSelectedObject() {
        return (FrjObject) objects.elementAt(super.getSelectedIndex());
    }

    abstract void setObject(FrjObject obj);
}

class RelationsListBox extends java.awt.List {
    Vector                      relations;
    Syab2                       syab;
    FrjObject           specifiedObject;

    RelationsListBox(Syab2 syab, int size) {
        super(size, false);
        relations = new Vector();
        setFont(syab.textFont);
        this.syab = syab;
    }

    void setRelations(FrjObject obj) {
        specifiedObject = obj;
        deleteAllRelations();

        FrjObject parent = obj.getParentObject();
        FrjRelation parentRel = null;

        for (Enumeration e = obj.getRelations();
             e.hasMoreElements();) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            if (rel.getRelatedObject(obj).equals(parent)) {
                parentRel = rel;
                addParentRelation(parentRel);
                break;
            }
        }

        for (Enumeration e = obj.getRelations();
             e.hasMoreElements();) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            if (rel == parentRel) {
                continue;
            }
            addRelation(rel);
        }
    }

    void addParentRelation(FrjRelation rel) {
        if (rel != null) {
            relations.addElement(rel);
            super.add("[[" + rel.getRelationName() + "]] "
                      + Syab2.changeID(rel.getID()));
        }
    }
    void addRelation(FrjRelation rel) {
        if (rel != null) {
            relations.addElement(rel);
            super.add("[" + rel.getRelationName() + "] "
                      + Syab2.changeID(rel.getID()));
        }
    }

    void deleteAllRelations() {
        relations = new Vector();
        super.removeAll();
    }

    FrjObject getRelatedObject() {
        FrjRelation rel
            =  (FrjRelation) relations.elementAt(super.getSelectedIndex());
        return rel.getRelatedObject(specifiedObject);
    }
}

class SpecifiedObjectsListBox extends ObjectsListBox {
    SpecifiedObjectsListBox(Syab2 syab, int size) {
        super(syab, size);
    }

    void setObject(FrjObject obj) {
        deleteAllObjects();
        addObject(obj);
    }

    void setObjects(String className) {
        deleteAllObjects();
        syab.sorter.setEnumeration(FrjExtents.getObjects(className));
        syab.sorter.sort();
        for (Enumeration e = syab.sorter.getVector().elements();
             e.hasMoreElements();) {
            addObject((FrjObject) e.nextElement());
        }
    }

    void setObjects(int start, int end) {
        deleteAllObjects();
        for (Enumeration e = FrjExtents.getAllObjects();
             e.hasMoreElements();) {
            FrjObject obj = (FrjObject) e.nextElement();
            if (obj.getOffset() <= start
                && end <= obj.getOffset() + obj.getLength()) {
                addObject(obj);
            }
        }
    }

}

class ObjectAttrsText extends TextArea {
    Syab2       syab;

    ObjectAttrsText(Syab2 syab, int rows, int cols) {
        super(rows, cols);
        this.syab = syab;
        setFont(syab.textFont);
    }

    void setObject(FrjObject obj) {
        FrjAttrsAsString        attrs[];
        attrs = obj.getAttrsAsString();

        String text = "";
        for (int i = 0; attrs[i] != null; i++) {
            text += attrs[i].name + ": " + attrs[i].value + "\n";
        }
                
        clearText();
        setText(text);
    }

    void clearText() {
        setText("");
    }
}

class SourceCodeText extends TextArea {
    Syab2       syab;

    SourceCodeText(Syab2 syab, int rows, int cols) {
        super(rows, cols);
        this.syab = syab;
        setFont(syab.textFont);
        load();
    }

    void load() {
        String text = syab.jfile.readText(0, syab.jfile.getFileSize());
        setText(text);
    }
                
    void reload() {
        clearText();
        load();
    }

    void setText(FrjObject obj) {
        select(obj.getOffset(), obj.getOffset() + obj.getLength());
    }
        
    void setColoredText(FrjObject obj) {
        String beforePattern, pattern, afterPattern;
        int     offset = obj.getOffset();
        int     length = obj.getLength();

        beforePattern = syab.jfile.readText(0, offset);
        pattern = syab.jfile.readText(offset, length);
        afterPattern = syab.jfile.readText(offset+length, syab.jfile.getFileSize() - (offset+length));
        clearText();
                
        setForeground(Color.black);
        append(beforePattern);
        setForeground(Color.red);
        append(pattern);
        setForeground(Color.black);
        append(afterPattern);

        // select(offset, offset+length);

    }
        
    void clearText() {
        setText("");
    }
}

class MyWindowListener extends WindowAdapter {
    public void windowClosing(WindowEvent e) {
        System.exit(0);
    }
}
