/* 
 * Program:             $RCSfile: Frame.java,v $  $Revision: 40.4 $
 * 
 * Purpose:             
 *
 * Author:              H.Kuwabara  2003/06/02
 *                      S.Yamamoto  2003/06/02
 *
 * (C) Copyright:       H.Kuwabara and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Frame.java,v 40.4 2003/07/08 05:47:56 yamamoto Exp yamamoto $
 */

package Japid.Tools.Syab3;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.GridLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;

import Japid.Tools.Syab3.action.*;

public class Frame extends JFrame {
    private Main main;
    private ViewModelMediator mediator;

    private Frame() {
    }

    public Frame(Main main) {
        super(Resource.getInstance().getString("frame.title"));
        this.main = main;

        setDefaultCloseOperation(DISPOSE_ON_CLOSE);
        addWindowListener(new WindowHandler());
    }

    public void init(ViewModelMediator mediator, 
                     ActionManager actionManager) {
        this.mediator = mediator;

        setJMenuBar(new MenuBar(mediator, actionManager));

        JList listObjects = new JList(mediator.getObjectsListModel());
        JList listRelations = new JList(mediator.getRelationsListModel());
        ObjectAttributesArea specifiedOAA = new ObjectAttributesArea();
        ObjectAttributesArea relatedOAA = new ObjectAttributesArea();
        SourceCodeArea sourceCodeArea = new SourceCodeArea(main.getModel());

        listObjects.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        listRelations.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        ObjectSelect objectSelect = new ObjectSelect(mediator, listObjects);
        listObjects.addListSelectionListener(objectSelect);
        listObjects.addMouseListener(objectSelect);

        RelationSelect relationSelect = new RelationSelect(mediator, 
                                                           listRelations);
        listRelations.addListSelectionListener(relationSelect);
        listRelations.addMouseListener(relationSelect);

        mediator.registObjectAttributes(specifiedOAA);
        mediator.registRelationAttributes(relatedOAA);
        mediator.registSourceCodeArea(sourceCodeArea);

        Resource res = Resource.getInstance();

        JButton buttonWhere = new JButton(res.getString("button.where"));
        KeyStroke stroke = KeyStroke.getKeyStroke(res.getString("button.where.mnemonic"));
        if (stroke != null) {
            buttonWhere.setMnemonic(stroke.getKeyCode());
        }
        buttonWhere.addActionListener(new Where(mediator));

        Container container = getContentPane();
        JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        splitPane.setContinuousLayout(false);
        splitPane.setOneTouchExpandable(true);
        container.add(splitPane);

        JPanel panelM = new JPanel();
        panelM.setLayout(new GridLayout(1, 4));

        JPanel panelS = new JPanel(new BorderLayout());
        panelS.add(new JLabel(res.getString("list.objects"), SwingConstants.CENTER), 
                  BorderLayout.NORTH);
        panelS.add(new JScrollPane(listObjects), BorderLayout.CENTER);
        panelM.add(panelS);

        panelS = new JPanel(new BorderLayout());
        panelS.add(new JLabel(res.getString("attrs.object"), SwingConstants.CENTER), 
                  BorderLayout.NORTH);
        panelS.add(new JScrollPane(specifiedOAA), BorderLayout.CENTER);
        panelM.add(panelS);

        panelS = new JPanel(new BorderLayout());
        panelS.add(new JLabel(res.getString("list.relations"), SwingConstants.CENTER), 
                  BorderLayout.NORTH);
        panelS.add(new JScrollPane(listRelations), BorderLayout.CENTER);
        panelM.add(panelS);

        panelS = new JPanel(new BorderLayout());
        panelS.add(new JLabel(res.getString("attrs.relation"), SwingConstants.CENTER), 
                  BorderLayout.NORTH);
        panelS.add(new JScrollPane(relatedOAA), BorderLayout.CENTER);
        panelM.add(panelS);

        splitPane.setTopComponent(panelM);

        panelM = new JPanel();
        panelM.setLayout(new BoxLayout(panelM, BoxLayout.Y_AXIS));

        panelS = new JPanel();
        panelS.setLayout(new BoxLayout(panelS, BoxLayout.X_AXIS));
        panelS.add(new JLabel(res.getString("source")));
        panelS.add(Box.createHorizontalGlue());
        panelS.add(buttonWhere);
        panelM.add(panelS);
        panelM.add(new JScrollPane(sourceCodeArea));

        splitPane.setBottomComponent(panelM);

        pack();

        sourceCodeArea.requestFocus();
    }

    public void setOpenClassName(String className) {
        setTitle(Resource.getInstance().getString("frame.title") + " - " + 
                                                  className);
    }

    private class WindowHandler extends WindowAdapter {
        public void windowClosing(WindowEvent e) {
            main.terminate();
        }
    }
}
