/* 
 * Program:             $RCSfile: Main.java,v $  $Revision: 40.3 $
 * 
 * Purpose:             
 *
 * Author:              H.Kuwabara  2003/06/02
 *                      S.Yamamoto  2003/06/02
 *
 * (C) Copyright:       H.Kuwabara and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: Main.java,v 40.3 2003/07/08 05:47:57 yamamoto Exp yamamoto $
 */

package Japid.Tools.Syab3;

import Japid.Freja.*;
import Japid.Tools.Syab3.action.*;

public final class Main {
    private DataModel model;

    private Main() {
    }

    public Main(String className, String loaderName, int frejaID) {
        model = new DataModel(className, loaderName, frejaID);
        Frame frame = new Frame(this);
        frame.setOpenClassName(className);

        MessageDialog.setParentComponent(frame);

        ViewModelMediator mediator = new ViewModelMediator();
        mediator.registObjectsListModel(new SpecifiedObjectsListModel());
        mediator.registRelationsListModel(new RelationsListModel());

        ActionManager actionManager = new ActionManager();
        Resource res = Resource.getInstance();
        actionManager.put(Load.class.getName(), 
                          new Load(model, frame, mediator));
        actionManager.put(Reload.class.getName(), new Reload(model, mediator));
        actionManager.put(Exit.class.getName(), new Exit(this));
        actionManager.put(ClassSelect.class.getName(), 
                          new ClassSelect(mediator));

        frame.init(mediator, actionManager);
        frame.setSize(800, 600);
        frame.setVisible(true);

        model.reload();
        mediator.reload();
    }

    public DataModel getModel() {
        return model;
    }

    public void terminate() {
        System.exit(0);
    }

    public static void main(String[] args) {
        String loaderName = null;
        String className;
        int frejaID = 1;

        if (args.length < 1) {
            usage();
            System.exit(1);
        }

        int i = 0;
        for ( ; i < args.length; i++) {
            if (args[i].equals("-loader")) {
                loaderName = args[++i];
            } else if (args[i].startsWith("-frejaID")) {
                frejaID = Integer.parseInt(args[++i]);
            } else if (args[i].startsWith("-")) {
                usage();
            } else {
                break;
            }
        }

        if (i == args.length) {
            usage();
        }
        className = args[i];

        Main main = new Main(className, loaderName, frejaID);
    }

    private static void usage() {
        FrjWorld.errorExit(Resource.getInstance().getString("Usage"));
    }
}
