/* 
 * Program:             $RCSfile: MenuBar.java,v $  $Revision: 40.3 $
 * 
 * Purpose:             
 *
 * Author:              H.Kuwabara  2003/06/02
 *                      S.Yamamoto  2003/06/02
 *
 * (C) Copyright:       H.Kuwabara and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: MenuBar.java,v 40.3 2003/07/08 05:47:57 yamamoto Exp yamamoto $
 */

package Japid.Tools.Syab3;

import java.awt.event.ActionListener;
import java.util.Iterator;

import javax.swing.AbstractButton;
import javax.swing.ButtonGroup;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.KeyStroke;

import Japid.Freja.*;
import Japid.Tools.Syab3.action.*;

public class MenuBar extends JMenuBar {
    private MenuBar() {
    }

    public MenuBar(ViewModelMediator mediator, ActionManager manager) {
        Resource res = Resource.getInstance();

        JMenu menu = new JMenu(res.getString("menu.file"));
        setButtonMnemonic(menu, res.getString("menu.file.mnemonic"));
        menu.add(makeMenuItem(res.getString("menu.file.load"), 
                              res.getString("menu.file.load.mnemonic"), 
                              res.getString("menu.file.load.accel"), 
                              manager.get(Load.class.getName())));
        menu.add(makeMenuItem(res.getString("menu.file.reload"), 
                              res.getString("menu.file.reload.mnemonic"), 
                              res.getString("menu.file.reload.accel"), 
                              manager.get(Reload.class.getName())));
        menu.addSeparator();
        menu.add(makeMenuItem(res.getString("menu.file.exit"), 
                              res.getString("menu.file.exit.mnemonic"), 
                              res.getString("menu.file.exit.accel"), 
                              manager.get(Exit.class.getName())));
        add(menu);

        menu = new JMenu(res.getString("menu.list"));
        setButtonMnemonic(menu, res.getString("menu.list.mnemonic"));
        ActionListener select = manager.get(ClassSelect.class.getName());
        VectorSorter sorter = new SortStrings(FrjExtents.getClassNames());
        sorter.sort();
        for (Iterator i = sorter.getVector().iterator(); i.hasNext(); ) {
            String className = (String)i.next();
            JMenuItem item = new JMenuItem(className);
            item.setActionCommand(className);
            item.addActionListener(select);
            menu.add(item);
        }
        add(menu);

        menu = new JMenu(res.getString("menu.option"));
        setButtonMnemonic(menu, res.getString("menu.option.mnemonic"));
        ButtonGroup buttonGroup = new ButtonGroup();
        JRadioButtonMenuItem item = makeRadioButtonMenuItem(res.getString("menu.option.offset"), 
                                                            res.getString("menu.option.offset.mnemonic"), 
                                                            new OffsetSorterSelect(mediator));
        item.setSelected(true);
        buttonGroup.add(item);
        menu.add(item);
        item = makeRadioButtonMenuItem(res.getString("menu.option.id"), 
                                       res.getString("menu.option.id.mnemonic"), 
                                       new IDSorterSelect(mediator));
        menu.add(item);
        buttonGroup.add(item);
        add(menu);
    }

    /**
     *  Creates a menu item.
     *  @param text a text of the menu item
     *  @param action the action of the menu item
     *  @return the menu item
     */
    protected JMenuItem makeMenuItem(String text, 
                                     ActionListener action) {
        JMenuItem item = new JMenuItem(text);
        item.addActionListener(action);

        return item;
    }

    /**
     *  Creates a menu item.
     *  @param text a text of the menu item
     *  @param mnemonic the keyboard mnemonic for the menu item
     *  @param action the action of the menu item
     *  @return the menu item
     */
    protected JMenuItem makeMenuItem(String text, 
                                     String mnemonic, 
                                     ActionListener action) {
        JMenuItem item = makeMenuItem(text, action);
        setButtonMnemonic(item, mnemonic);

        return item;
    }

    /**
     *  Creates a menu item.
     *  @param text a text of the menu item
     *  @param mnemonic the keyboard mnemonic for the menu item
     *  @param accelerator the key combination for direct access
     *  @param action the action of the menu item
     *  @return the menu item
     */
    protected JMenuItem makeMenuItem(String text, 
                                     String mnemonic, 
                                     String accelerator, 
                                     ActionListener action) {
        JMenuItem item = makeMenuItem(text, mnemonic, action);
        item.setAccelerator(KeyStroke.getKeyStroke(accelerator));

        return item;
    }

    /**
     *  Creates a radio button menu item.
     *  @param text a text of the menu item
     *  @param action the action of the menu item
     */
    protected JRadioButtonMenuItem makeRadioButtonMenuItem(String text, 
                                                           ActionListener action) {
        JRadioButtonMenuItem item = new JRadioButtonMenuItem(text);
        item.addActionListener(action);

        return item;
    }

    /**
     *  Creates a menu item.
     *  @param text a text of the menu item
     *  @param mnemonic the keyboard mnemonic for the menu item
     *  @param action the action of the menu item
     */
    protected JRadioButtonMenuItem makeRadioButtonMenuItem(String text, 
                                                           String mnemonic, 
                                                           ActionListener action) {
        JRadioButtonMenuItem item = makeRadioButtonMenuItem(text, action);
        setButtonMnemonic(item, mnemonic);

        return item;
    }

    /**
     *  Sets the keyboard mnemonic on the specified button.
     *  @param button the specified button
     *  @param mnemonic the letter which represents the mnemonic
     */
    protected void setButtonMnemonic(AbstractButton button, 
                                     String mnemonic) {
        KeyStroke stroke = KeyStroke.getKeyStroke(mnemonic);
        if (stroke != null) {
            button.setMnemonic(stroke.getKeyCode());
        }
    }
}
