/* 
 * Program:             $RCSfile: SpecifiedObjectsListModel.java,v $  $Revision: 40.3 $
 * 
 * Purpose:             
 *
 * Author:              H.Kuwabara  2003/06/02
 *                      S.Yamamoto  2003/06/02
 *
 * (C) Copyright:       H.Kuwabara and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: SpecifiedObjectsListModel.java,v 40.3 2003/07/08 05:47:58 yamamoto Exp yamamoto $
 */

package Japid.Tools.Syab3;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;

import javax.swing.DefaultListModel;
import javax.swing.JList;

import Japid.Freja.*;

public class SpecifiedObjectsListModel extends DefaultListModel {
    private ArrayList listObjects;

    private VectorSorter sorter;
    private static final VectorSorter sorterByOffset = new SortByOffset();
    private static final VectorSorter sorterByID = new SortByID();

    public SpecifiedObjectsListModel() {
        listObjects = new ArrayList();
        useSorterByOffset();
    }

    public void addObject(FrjObject object) {
        if (object != null) {
            listObjects.add(object);
            addElement("[" + object.getClassName() + "] " + 
                       DataModel.convertToFullID(object.getID()));
        }
    }

    public void removeAllObjects() {
        listObjects.clear();
        removeAllElements();
    }

    public FrjObject getObject(int index) {
        if (index >= 0 && index < listObjects.size()) {
            return (FrjObject)listObjects.get(index);
        } else {
            return null;
        }
    }

    public void setObject(FrjObject object) {
        removeAllObjects();
        addObject(object);
    }

    public void setObjects(String className) {
        removeAllObjects();
        sorter.setEnumeration(FrjExtents.getObjects(className));
        sorter.sort();
        for (Iterator i = sorter.getVector().iterator(); i.hasNext(); ) {
            addObject((FrjObject)i.next());
        }
    }

    public void setObjects(int start, int end) {
        removeAllObjects();
        for (Enumeration e = FrjExtents.getAllObjects(); e.hasMoreElements(); ) {
            FrjObject object = (FrjObject)e.nextElement();
            if (object.getOffset() <= start && 
                end <= object.getOffset() + object.getLength()) {
                addObject(object);
            }
        }
    }

    public void useSorterByOffset() {
        sorter = sorterByOffset;
    }

    public void useSorterByID() {
        sorter = sorterByID;
    }
}
