/* 
 * Program:             $RCSfile: XRef.java,v $  $Revision: 40.6 $
 * 
 * Purpose:             
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: XRef.java,v 40.6 2003/02/15 11:56:44 yamamoto Exp yamamoto $
 */

package Japid.Tools;

import java.awt.*;
import java.util.*;
import Japid.Freja.*;
import Japid.Freja.F2S.*;

class XRef extends Frame {
    //  private PackageListBox                  packageListBox;
    private TextField                           classNameText;
    private ClassListBox                        classListBox;
    private TypeListBox                         superClassListBox;
    private TypeListBox                         interfaceListBox;
    private MemberListBox                       memberListBox;
    private TypeListBox                         returnListBox;
    private TypeListBox                         parameterListBox;
    private TypeListBox                         throwListBox;
    private ExprListBox                         exprListBox;
    private SourceCodeArea                      sourceCodeArea;
    private TextField                           messageField;
    private Button                                      analyzeButton;
    private CheckboxMenuItem            publicMenuItem;
    private CheckboxMenuItem            packageMenuItem;
    private CheckboxMenuItem            protectedMenuItem;
    private CheckboxMenuItem            privateMenuItem;

    Font                appFont = new Font("Helvitica", Font.PLAIN, 14);
    Font                textFont = new Font("Helvitica", Font.PLAIN, 14);
    String              packageName;
    boolean             showPublic = true, showPackage = true,
        showProtected = true, showPrivate = true;

    private final String        M_FILE = "File", M_QUIT = "Quit",
        M_ACCESS = "Access",
        M_SHOW_PUBLIC = "Show public",
        M_SHOW_PACKAGE = "Show package",
        M_SHOW_PROTECTED = "Show protected",
        M_SHOW_PRIVATE = "Show private";

    XRef(String pkgName) {
        super("Xref");
        packageName = pkgName;
        createWidget();
        //              packageListBox.addObject(pkgName);
        VectorSorter sorter = new SortStrings(Creja.getClasses(pkgName));
        sorter.sort();
        for (Enumeration e = sorter.getEnumeration();
             e.hasMoreElements();) {
            String className = (String) e.nextElement();
            classListBox.addObject(className);
        }
        FrjWorld.setAutoLoading(true);
        FrjWorld.setLoader(new F2SExternalAccessibleLoader());
    }
        
    void createWidget() {
        setFont(appFont);

        MenuBar menubar = new MenuBar();
        Menu fileMenu = new Menu(M_FILE);
        fileMenu.add(new MenuItem(M_QUIT));
        menubar.add(fileMenu);

        Menu accessMenu = new Menu(M_ACCESS);
        publicMenuItem = new CheckboxMenuItem(M_SHOW_PUBLIC);
        publicMenuItem.setState(showPublic);
        packageMenuItem = new CheckboxMenuItem(M_SHOW_PACKAGE);
        packageMenuItem.setState(showPackage);
        protectedMenuItem = new CheckboxMenuItem(M_SHOW_PROTECTED);
        protectedMenuItem.setState(showProtected);
        privateMenuItem =  new CheckboxMenuItem(M_SHOW_PRIVATE);
        privateMenuItem.setState(showPrivate);
        accessMenu.add(publicMenuItem);
        accessMenu.add(packageMenuItem);
        accessMenu.add(protectedMenuItem);
        accessMenu.add(privateMenuItem);
        menubar.add(accessMenu);

        setMenuBar(menubar);
                
        GridBagLayout gbl = new GridBagLayout();

        setLayout(gbl);

        analyzeButton = new Button("Analyze");

        //              packageListBox = new PackageListBox(this, 8);
        classNameText = new TextField(60);
        classNameText.setEditable(false);
        classNameText.setFont(appFont);
        classListBox = new ClassListBox(this, 8);
        superClassListBox = new TypeListBox(this, 1);
        interfaceListBox = new TypeListBox(this, 4);
        memberListBox = new MemberListBox(this, 8);
        returnListBox = new TypeListBox(this, 1);
        parameterListBox = new TypeListBox(this, 4);
        exprListBox = new ExprListBox(this, 6);
        throwListBox = new TypeListBox(this, 2);
        sourceCodeArea = new SourceCodeArea(this, 20, 5);
        messageField = new TextField(60);
        messageField.setEditable(false);
        messageField.setFont(appFont);
                
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.insets = new Insets(2, 2, 2, 2);
        gbc.weightx = 0.8;
        gbc.weighty = 0.8;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.anchor = GridBagConstraints.WEST;

        //
        //    0     1     2     4     5
        //   +-----------------------+
        //       |    class name         |
        //   +-cls-+-----+-mem-+-----+
        //   |     +-spc-+     +-ret-+
        //   |     |     |     |     |
        //   |     +-int-+     +-prm-+
        //   |     |     |     |     |
        //   |     |     |     +-thr-+
        //   |     |     |     |     |
        //   +-----+-----+-----+-----+
        //   |     source            |
        //   +-----------------------+

        int y = 0;
        int classListX = 0, classAttrX = 1, memberListX = 2, memberAttrX = 4,
            exprListX = 5;

        Panel cp = new Panel();
        cp.add(new Label("Class:"));
        cp.add(classNameText);
        add(cp, gbl, gbc, 0, y, 3, 1);

        y++;
        add(new Label(" classes "), gbl, gbc, classListX, y, 1, 1);
        add(new Label(" members and methods"), gbl, gbc, memberListX, y, 1, 1);

        y++;
        add(classListBox, gbl, gbc, classListX, y, 1, 6);
        add(new Label("super"), gbl, gbc, classAttrX, y, 1, 1);
        gbc.weighty = 0.1;
        add(memberListBox, gbl, gbc, memberListX, y, 1, 5);
        gbc.weighty = 0.8;
        add(new Label("type"), gbl, gbc, memberAttrX, y, 1, 1);
        Panel p = new Panel();
        p.add(new Label("Dependency"));
        p.add(analyzeButton);
        add(p, gbl, gbc, exprListX, y, 1, 1);

        y++;
        add(superClassListBox, gbl, gbc, classAttrX, y, 1, 1);
        add(returnListBox, gbl, gbc, memberAttrX, y, 1, 1);
        add(exprListBox, gbl, gbc, exprListX, y, 1, 5);

        y++;
        add(new Label("interface"), gbl, gbc, classAttrX, y, 1, 1);
        add(new Label("parameters"), gbl, gbc, memberAttrX, y, 1, 1);

        y++;
        add(interfaceListBox, gbl, gbc, classAttrX, y, 1, 3);
        add(parameterListBox, gbl, gbc, memberAttrX, y, 1, 1);

        y++;
        add(new Label("throws"), gbl, gbc, memberAttrX, y, 1, 1);

        y++;
        add(throwListBox, gbl, gbc, memberAttrX, y, 1, 1);

        y++;
        // dummy
        add(new Label(" source code    "), gbl, gbc, classListX, y, 1, 1);
        add(new Label("                "), gbl, gbc, classAttrX, y, 1, 1);
        add(new Label("                                         "),
            gbl, gbc, memberListX, y, 1, 1);
        add(new Label(" "), gbl, gbc, 3, y, 1, 1);
        add(new Label("                "), gbl, gbc, memberAttrX, y, 1, 1);

        y++;
        add(sourceCodeArea, gbl, gbc, 0, y, 6, 1);

        y++;
        Panel mp = new Panel();
        mp.add(new Label("message:"));
        mp.add(messageField);
        add(mp, gbl, gbc, 0, y, 5, 1);
                
        pack();
    }

    private void add(Component c, GridBagLayout gbl, GridBagConstraints gbc,
                     int x, int y, int w, int h) {
        gbc.gridx = x;
        gbc.gridy = y;
        gbc.gridwidth = w;
        gbc.gridheight = h;
        gbl.setConstraints(c, gbc);
        add(c);
    }

    /****************** Listeners ***********************/

    protected class ListBoxItemListener implements java.awt.event.ItemListener {
        public void itemStateChanged(java.awt.event.ItemEvent e) {
            Object e_source = e.getSource();

            message("");
                
            if (e_source == classListBox) {
                FrjClass selected
                    =  (FrjClass) classListBox.getSelectedObject();
                if (selected != null) {
                    setClassListBox(selected);
                }
            } else if (e_source == memberListBox) {
                FrjObject selected
                    =  (FrjObject) memberListBox.getSelectedObject();
                if (selected != null) {
                    setMemberListBox(selected);
                }
            } else if (e_source == exprListBox) {
                FrjObject selected
                    =  (FrjObject) exprListBox.getSelectedObject();
                if (selected != null) {
                    setExprListBox((FrjExpression) selected);
                }
            } else {
                                // error
            }
        }
    }

    protected class ListBoxActionListener implements java.awt.event.ActionListener {
        public void actionPerformed(java.awt.event.ActionEvent e) {
            Object e_source = e.getSource();

            if (e_source instanceof TypeListBox) {
                FrjType t = ((TypeListBox) (e_source)).getSelectedType();
                FrjClass c = (FrjClass) t.getReferringClassInterface();
                if (c != null) {
                    setClassListBox(c);
                    classListBox.selectObject(c);
                }
            } else {
                                // error
            }
        }
    }

    protected class MenuItemActionListener implements java.awt.event.ActionListener {
        public void actionPerformed(java.awt.event.ActionEvent e) {
            Object e_source = e.getSource();

            message("");

            if (e_source.equals(M_QUIT)) {
                dispose();
                System.exit(0);
            } else if (e_source.equals(M_SHOW_PUBLIC)) {
                showPublic = publicMenuItem.getState();
                memberListBox
                    .setClass((FrjClass)
                              classListBox.getSelectedObject());
            } else if (e_source.equals(M_SHOW_PACKAGE)) {
                showPackage = packageMenuItem.getState();
                memberListBox
                    .setClass((FrjClass) classListBox.getSelectedObject());
            } else if (e_source.equals(M_SHOW_PROTECTED)) {
                showProtected = protectedMenuItem.getState();
                memberListBox
                    .setClass((FrjClass) classListBox.getSelectedObject());
            } else if (e_source.equals(M_SHOW_PRIVATE)) {
                showPrivate = privateMenuItem.getState();
                memberListBox
                    .setClass((FrjClass) classListBox.getSelectedObject());
            } else {
                                // error
            }
        }
    }

    protected class ButtonActionListener implements java.awt.event.ActionListener {
        public void actionPerformed(java.awt.event.ActionEvent e) {
            Object e_source = e.getSource();

            message("");

            if (e_source == analyzeButton) {
                FrjObject obj = memberListBox.getSelectedObject();
                if (obj instanceof FrjVariable) {
                    Enumeration e2
                        = FrjExtents.getObjects(obj.getFrejaEntry().getFrejaID(),
                                                "Expression");
                    if (!e2.hasMoreElements()) {
                        // loadƤexpression¿ʬ¸ߤ
                        loadOthers(obj.getFrejaEntry());
                    }
                    exprListBox.setVariable((FrjVariable) obj);
                } else {
                    // error
                }
            }
        }
    }

    /****************** Listeners end ***********************/

    void setClassListBox(FrjClass fclass) {
        FrjFile file = fclass.getFile();
        FrjQName pkg = file.getDefinedPackage();
        String fullClassName;
        if (pkg != null) {
            fullClassName = pkg.getFullString() + "." + fclass.getStr();
        } else {
            fullClassName = fclass.getStr();
        }
        classNameText.setText(fullClassName);
        superClassListBox.deleteAllObjects();
        superClassListBox.addObject(fclass.getSuperClassType());
        interfaceListBox.deleteAllObjects();
        interfaceListBox.addObjects(fclass.getInterfaceTypes());
        memberListBox.setClass(fclass);
        exprListBox.deleteAllObjects();
        sourceCodeArea.clearText();
        sourceCodeArea.setText(fclass);
    }

    void setMemberListBox(FrjObject obj) {
        if (obj == null) {
            return;
        }
        sourceCodeArea.selectText(obj);
        returnListBox.deleteAllObjects();
        parameterListBox.deleteAllObjects();
        throwListBox.deleteAllObjects();
        if (obj instanceof FrjVariable) {
            FrjVariable var = (FrjVariable) obj;
            returnListBox.addObject(var.getType());
        } else if (obj instanceof FrjRoutine) {
            FrjRoutine routine = (FrjRoutine) obj;
            returnListBox.addObject(routine.getReturnType());
            parameterListBox.addObjects(routine.getParameters());
            throwListBox.addObjects(routine.getThrowTypes());
        }
        exprListBox.deleteAllObjects();
    }

    void setExprListBox(FrjExpression expr) {
        sourceCodeArea.selectText(expr);
    }

    FrejaEntry load(String className) {
        message("loading " + className + "...");
        FrejaEntry entry;
        synchronized (messageField) {
            entry = Creja.getFrejaEntry(packageName, className);
        }
        FrjWorld.load(entry);
        VectorSorter sorter = new SortByOffset();
        for (Enumeration e = FrjExtents.getAllObjects(entry.getFrejaID());
             e.hasMoreElements();) {
            ((FrjObject) e.nextElement()).sortRelatedObjects(sorter);
        }
        message("loaded.");
        return entry;
    }

    void loadOthers(FrejaEntry entry) {
        message("analyzing ");
        synchronized (messageField) {
            F2SExternalAccessibleLoader.loadOthers(entry);
        }
        VectorSorter sorter = new SortByOffset();
        for (Enumeration e = FrjExtents
                 .getObjects(entry.getFrejaID(), "Variable");
             e.hasMoreElements();) {
            ((FrjObject) e.nextElement()).sortRelatedObjects(sorter);
        }

        message("done.");
    }
        
    void message(String mes) {
        messageField.setText(mes);
        messageField.getToolkit().sync();
    }

    public static void main(String args[]) {
        XRef    xref;
                
        if (args.length < 1) {
            System.err.println("Usage: Japid.Tools.XRef packageName");
            System.exit(1);
        }
        xref = new XRef(args[0]);
        xref.show();
    }
}
                        
abstract class ListBox extends java.awt.List {
    Vector                      objects;
    XRef                        xref;
    Dimension           d;

    ListBox(XRef xref, int size) {
        super(size, false);
        objects = new Vector();
        setFont(xref.textFont);
        this.xref = xref;
        d = getPreferredSize();
    }

    void addObject(FrjObject obj) {
        if (obj != null) {
            objects.addElement(obj);
            super.add(toString(obj));
        }
    }

    void addObject(String name) {
        objects.addElement(name);
        super.add(name);
    }

    void selectObject(FrjObject obj) {
        int     index = objects.indexOf(obj);
        super.select(index);
    }

    void deleteAllObjects() {
        objects = new Vector();
        super.removeAll();
    }

    FrjObject getSelectedObject() {
        return (FrjObject) objects.elementAt(super.getSelectedIndex());
    }
                
    abstract String toString(FrjObject obj);

    /** ListΥͤˤ */
    /**
       public void addNotify() {
       super.addNotify();
       resize(d.width, d.height);
       }
    **/
}

class PackageListBox extends ListBox {
    PackageListBox(XRef xref, int size) {
        super(xref, size);
    }

    String toString(FrjObject obj) {
        return ((F2SObject) obj).toSource();
    }
}

class ClassListBox extends ListBox {
    ClassListBox(XRef xref, int size) {
        super(xref, size);
    }

    void setFile(FrjFile file) {
        if (file == null) {
            return;
        }
        for (Enumeration e = file.getClasses();
             e.hasMoreElements();) {
            this.addObject((FrjObject) e.nextElement());
        }
    }

    FrjObject getSelectedObject() {
        Object obj = objects.elementAt(super.getSelectedIndex());
        if (obj instanceof String) {
            String className = (String) obj;
            xref.load(className);
            for (Enumeration e = FrjExtents.getObjects("Class");
                 e.hasMoreElements();) {
                FrjClass c = (FrjClass) e.nextElement();
                if (c.getStr().equals(className)) {
                    objects.setElementAt(c, objects.indexOf(className));
                    return c;
                }
            }
            return null;
        } else {
            return (FrjObject) obj;
        }
    }

    String toString(FrjObject obj) {
        return obj.getStr();
    }
}

class MemberListBox extends ListBox {
    MemberListBox(XRef xref, int size) {
        super(xref, size);
    }

    void setClass(FrjClass fclass) {
        if (fclass == null) {
            return;
        }
        deleteAllObjects();
        for (Enumeration e = fclass.getMembers();
             e.hasMoreElements();) {
            this.addVariable((FrjVariable) e.nextElement());
        }
        for (Enumeration e = fclass.getRoutines();
             e.hasMoreElements();) {
            this.addRoutine((FrjRoutine) e.nextElement());
        }
    }

    void addRoutine(FrjRoutine routine) {
        /** for debug
            System.out.println(routine.getStr());
            System.out.println("  public: " + xref.showPublic + " " +
            routine.isPublic());
            System.out.println("  pacakge: " + xref.showPackage + " " +
            routine.isDefaultAccess());
            System.out.println("  protected: " + xref.showProtected + " " +
            routine.isProtected());
            System.out.println("  private: " + xref.showPrivate + " " +
            routine.isPrivate());
        **/
        if (xref.showPublic && routine.isPublic()) {
            addObject(routine);
        } else if (xref.showPackage && routine.isDefaultAccess()) {
            addObject(routine);
        } else if (xref.showProtected && routine.isProtected()) {
            addObject(routine);
        } else if (xref.showProtected && routine.isPrivate()) {
            addObject(routine);
        }
    }
                
    void addVariable(FrjVariable var) {
        /** for debug
            System.out.println(var.getStr());
            System.out.println("  public: " + xref.showPublic + " " +
            var.isPublic());
            System.out.println("  pacakge: " + xref.showPackage + " " +
            var.isDefaultAccess());
            System.out.println("  protected: " + xref.showProtected + " " +
            var.isProtected());
            System.out.println("  private: " + xref.showPrivate + " " +
            var.isPrivate());
        **/
        if (xref.showPublic && var.isPublic()) {
            addObject(var);
        } else if (xref.showPackage && var.isDefaultAccess()) {
            addObject(var);
        } else if (xref.showProtected && var.isProtected()) {
            addObject(var);
        } else if (xref.showProtected && var.isPrivate()) {
            addObject(var);
        }
    }
                
    String toString(FrjObject obj) {
        if (obj instanceof FrjVariable) {
            return variableString((FrjVariable) obj);
        } else if (obj instanceof FrjRoutine) {
            return routineString((FrjRoutine) obj);
        } else {
            return "";
        }
    }

    static String variableString(FrjVariable var) {
        String res = "";
        res += ((F2SType) var.getType()).toSource();
        res += " " + var.getStr() + ";";
        return res;
    }

    static String routineString(FrjRoutine r) {
        String res = "";
        F2SType rt = (F2SType) r.getReturnType();
        if (rt != null) {
            res += rt.toSource() + " ";
        }
        res += r.getStr() + "(";
        String comma = "";
        for (Enumeration e1 = r.getParameters();
             e1.hasMoreElements();) {
            res += comma + ((F2SVariable) e1.nextElement()).toSource();
            comma = ", ";
        }
        res += ");";
        return res;
    }

}

class TypeListBox extends ListBox {
    TypeListBox(XRef xref, int size) {
        super(xref, size);
    }
    void addObjects(Enumeration e) {
        while (e.hasMoreElements()) {
            addObject((FrjObject) e.nextElement());
        }
    }

    FrjType getSelectedType() {
        FrjObject obj = super.getSelectedObject();
        if (obj instanceof FrjType) {
            return (FrjType) obj;
        } else if (obj instanceof FrjVariable) {
            return ((FrjVariable) obj).getType();
        }
        return null;
    }

    String toString(FrjObject obj) {
        if (obj instanceof F2SType) {
            return ((F2SType) obj).toSource();
        } else if (obj instanceof FrjVariable) {
            return MemberListBox.variableString((FrjVariable) obj);
        } else {
            return "";
        }
    }
}

class ExprListBox extends ListBox {
    ExprListBox(XRef xref, int size) {
        super(xref, size);
    }

    void setVariable(FrjVariable var) {
        for (Enumeration e = var.getReferrers();
             e.hasMoreElements();) {
            addObject((FrjObject) e.nextElement());
        }
    }

    String toString(FrjObject obj) {
        if (obj instanceof FrjExpression) {
            FrjExpression expr = (FrjExpression) obj;
            FrjRoutine r = (FrjRoutine) obj.getAncestorObject("Routine");
            String lrStr;
            if (expr.getLRSort() == FrjExpression.L_VALUE
                || expr.getLRSort() == FrjExpression.LR_VALUE) {
                lrStr = "L";
            } else if (expr.getLRSort() == FrjExpression.R_VALUE) {
                lrStr = "R";
            } else if (expr.getLRSort() == FrjExpression.OBJECT_OPERATION) {
                lrStr = "O";
            } else {
                lrStr = "U";
            }
            return lrStr + ": " + r.getStr();
        } else {
            return "";
        }
    }
}

class SourceCodeArea extends TextArea {
    private     int textOffset;
    XRef        xref;

    SourceCodeArea(XRef xref, int rows, int cols) {
        super(rows, cols);
        this.xref = xref;
        setFont(xref.textFont);
        setEditable(false);
        setBackground(Color.white);
    }

    void setText(FrjObject obj) {
        textOffset = obj.getOffset();
        setText(obj.getText());
    }

    void selectText(FrjObject obj) {
        select(obj.getOffset() - textOffset,
               obj.getOffset() + obj.getLength() - textOffset);
    }
        
    void clearText() {
        setText("");
    }
}
