/*
 *     GenClassHTML.java  Oct 21, 2003
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package Japid.genxrefxml;

import org.jtool.platform.pgstruct.*;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.util.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class GenClassHTML {

    public static void main(String[] args) {
        if (args.length > 0) {
            GenClassHTML gen = new GenClassHTML();
            gen.view(args[0]);
        } else {
            System.err.println("Usage: java GenClassHTML xml_filename");
        }
    }

    public void view(String fileName) {
        String content = write(fileName);

        System.out.println("---- ");
        System.out.print(content);
        System.out.println("----");
    }

    private StringBuffer buf;

    public String write(String fileName) {
        try {
            String content = ReadFile.read(fileName);
            Document doc = DOMTreeCreator.create(content);
            return write(doc);

        } catch (IOException e) { }
        return null;
    }

    public String write(Document doc) {
        QueryJavaClass jc = new QueryJavaClass(doc.getDocumentElement());
        buf = new StringBuffer();

        if (jc.isInterface()) {
            return writeInterface(jc);
        } else {
            return writeClass(jc);
        }
    }

    private String writeClass(QueryJavaClass jc) {
        writeInfo(jc, "CLASS: ", "");

        writeSuperClass(jc, "  ");
        writeSubClass(jc, "  ");
        writeSuperInterface(jc, "  ");

        writeConstructor(jc, "  ");
        writeInnerClass(jc, "  ");

        writeReference(jc, "  ");

        writeField(jc, "  ");
        writeMethod(jc, "  ");

        return buf.toString();
    }

    public String writeInterface(QueryJavaClass jc) {
        writeInfo(jc, "INTERFACE: ", "");

        writeSuperInterface(jc, "  ");
        writeSubInterface(jc, "  ");
        writeDescendantClass(jc, "  ");

        writeReference(jc, "  ");

        writeField(jc, "  ");
        writeMethod(jc, "  ");

        return buf.toString();
    }

    private void writeInfo(QueryJavaClass jc, String title, String indent) {
        buf.append(indent + title + jc.getName() + "\n");
        buf.append(indent + "  ID = " + jc.getId() + "\n");
        buf.append(indent + "  FQN = " + jc.getFQN() + "\n");
        buf.append(indent + "  MODIFIER = " + jc.getModifier() + "\n");

        buf.append(indent + "  A = " + jc.getAnchorText() +  "\n");
        buf.append(indent + "  File = " + jc.getFileName() +  "\n");
        buf.append(indent + "  Line = " + jc.getLine() +  "\n");
        buf.append("\n");
    }

    private void writeElement(QueryJavaElement je, String title, String indent) {
        buf.append(indent + title + "\n");
        buf.append(indent + "  ID = " + je.getId() + "\n");

        buf.append(indent + "  A = " + je.getAnchorText() +  "\n");
        buf.append(indent + "  File = " + je.getFileName() + "\n");
        buf.append(indent + "  Line = " + je.getLine() + "\n");
        buf.append("\n");
    }

    private void writeSuperClass(QueryJavaClass jc, String indent) {
        Iterator it = jc.getSuperClassNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaClass c = new QueryJavaClass(elem);
            writeElement(c, "SUPER CLASS: " + c.getName(), indent);

            writeSuperClass(c, indent + "  ");
        }
    }

    private void writeSubClass(QueryJavaClass jc, String indent) {
        Iterator it = jc.getSubClassNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaClass c = new QueryJavaClass(elem);
            writeElement(c, "SUB CLASS: " + c.getName(), indent);

            writeSubClass(c, indent + "  ");
        }
    }

    private void writeSuperInterface(QueryJavaClass jc, String indent) {
        Iterator it = jc.getSuperInterfaceNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaClass c = new QueryJavaClass(elem);
            writeElement(c, "SUPER INTERFACE: " + c.getName(), indent);

            writeSuperInterface(c, indent + "  ");
        }
    }

    private void writeSubInterface(QueryJavaClass jc, String indent) {
        Iterator it = jc.getSubInterfaceNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaClass c = new QueryJavaClass(elem);
            writeElement(c, "SUB INTERFACE: " + c.getName(), indent);

            writeSubInterface(c, indent + "  ");
        }
    }

    private void writeDescendantClass(QueryJavaClass jc, String indent) {
        Iterator it = jc.getDescendantClassNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaClass c = new QueryJavaClass(elem);
            writeElement(c, "DESCENDANT CLASS: " + c.getName(), indent);

            writeDescendantClass(c, indent + "  ");
        }
    }

    private void writeConstructor(QueryJavaClass jc, String indent) {
        Iterator it = jc.getConstructorNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaMethod jm = new QueryJavaMethod(elem);
            writeElement(jm, "CONSTRUCTOR: " + jm.getSignature(), indent);
        }
    }

    private void writeInnerClass(QueryJavaClass jc, String indent) {
        Iterator it = jc.getInnerClassNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaClass c = new QueryJavaClass(elem);
            writeElement(c, "INNER CLASS: " + c.getName(), indent);
        }
    }

    private void writeField(QueryJavaClass jc, String indent) {
        Iterator it = jc.getFieldNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaField jf = new QueryJavaField(elem);
            writeElement(jf, "FIELD: " + jf.getName(), indent);
        }
    }

    private void writeMethod(QueryJavaClass jc, String indent) {
        Iterator it = jc.getMethodNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaMethod jm = new QueryJavaMethod(elem);
            writeElement(jm, "METHOD: " + jm.getSignature(), indent);
        }
    }

    private void writeReference(QueryJavaClass jc, String indent) {
        Iterator it = jc.getReferringNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            QueryJavaElement src = new QueryJavaElement(elem);
            writeElement(src, "REFER FROM: ", indent);
        }
    }
}
