/*
 *     GenFieldHTML.java  Oct 21, 2003
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package Japid.genxrefxml;

import org.jtool.platform.pgstruct.*;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.util.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class GenFieldHTML {

    public static void main(String[] args) {
        if (args.length > 0) {
            GenFieldHTML gen = new GenFieldHTML();
            gen.view(args[0]);
        } else {
            System.err.println("Usage: java GenFieldHTML xml_filename");
        }
    }

    public void view(String fileName) {
        String content = write(fileName);

        System.out.println("---- ");
        System.out.print(content);
        System.out.println("----");
    }

    private StringBuffer buf;

    public String write(String fileName) {
        try {
            String content = ReadFile.read(fileName);
            Document doc = DOMTreeCreator.create(content);
            return write(doc);

        } catch (IOException e) { }
        return null;
    }

    public String write(Document doc) {
        QueryJavaClass jc = new QueryJavaClass(doc.getDocumentElement());
        buf = new StringBuffer();

        Iterator it = jc.getFieldNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaField jf = new QueryJavaField(elem);
            writeField(jf);
        }

        return buf.toString();
    }

    private void writeField(QueryJavaField jf) {
        writeInfo(jf, "");

        writeReference(jf, "  ");
    }

    private void writeInfo(QueryJavaField jf, String indent) {
        buf.append(indent + "FIELD: " + jf.getName() + "\n");
        buf.append(indent + "  ID = " + jf.getId() + "\n");
        buf.append(indent + "  TYPE = " + jf.getFQN() + "\n");
        buf.append(indent + "  MODIFIER = " + jf.getModifier() + "\n");

        buf.append(indent + "  A = " + jf.getAnchorText() +  "\n");
        buf.append(indent + "  File = " + jf.getFileName() +  "\n");
        buf.append(indent + "  Line = " + jf.getLine() +  "\n");
        buf.append("\n");
    }

    private void writeElement(QueryJavaElement je, String title, String indent) {
        buf.append(indent + title + "\n");
        buf.append(indent + "  ID = " + je.getId() + "\n");

        buf.append(indent + "  A = " + je.getAnchorText() +  "\n");
        buf.append(indent + "  File = " + je.getFileName() + "\n");
        buf.append(indent + "  Line = " + je.getLine() + "\n");
    }

    private void writeReference(QueryJavaField jf, String indent) {
        Iterator it = jf.getReferringNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            QueryJavaElement je = new QueryJavaElement(elem);
            writeElement(je, "REFER FROM:", indent + "  ");

            writeReadOrWriteAttribute(elem, indent + "  ");
            buf.append("\n");
        }
    }
        
    private void writeReadOrWriteAttribute(Element elem, String indent) {
        if (elem.getAttribute(Constants.ReadAttr).equals(Constants.YES)) {
            buf.append(indent + "  READ\n");
        }
        
        if (elem.getAttribute(Constants.WriteAttr).equals(Constants.YES)) {
            buf.append(indent + "  WRITE\n");
        }
    }
}
