/*
 *     GenMethodHTML.java  Oct 21, 2003
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package Japid.genxrefxml;

import org.jtool.platform.pgstruct.*;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.util.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class GenMethodHTML {

    public static void main(String[] args) {
        if (args.length > 0) {
            GenMethodHTML gen = new GenMethodHTML();
            gen.view(args[0]);
        } else {
            System.err.println("Usage: java GenMethodHTML xml_filename");
        }
    }

    public void view(String fileName) {
        String content = write(fileName);

        System.out.println("---- ");
        System.out.print(content);
        System.out.println("----");
    }

    private StringBuffer buf;

    public String write(String fileName) {
        try {
            String content = ReadFile.read(fileName);
            Document doc = DOMTreeCreator.create(content);
            return write(doc);

        } catch (IOException e) { }
        return null;
    }

    public String write(Document doc) {
        QueryJavaClass jc = new QueryJavaClass(doc.getDocumentElement());
        buf = new StringBuffer();

        Iterator it = jc.getMethodNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaMethod jm = new QueryJavaMethod(elem);
            writeMethod(jm);
        }

        return buf.toString();
    }

    private void writeMethod(QueryJavaMethod jm) {
        writeInfo(jm, "");

        writeElement(jm.getThrowsNodes(), "THROWS: ", "  ");
        
        writeElement(jm.getOverridingNodes(), "OVERRIDING: ", "  ");
        writeElement(jm.getOverloadingNodes(), "OVERLOADING: ", "  ");
        writeElement(jm.getImplementingNodes(), "IMPLEMENTING: ", "  ");

        writeElement(jm.getOverriddenNodes(), "OVERRIDDEN: ", "  ");
        writeElement(jm.getOverloadedNodes(), "OVERLOADED: ", "  ");
        writeElement(jm.getImplementedNodes(), "IMPLEMENTED: ", "  ");

        writeElement(jm.getReferringNodes(), "REFER FROM:", "  ");
    }

    private void writeInfo(QueryJavaMethod jm, String indent) {
        buf.append(indent + "METHOD: " + jm.getSignature() + "\n");
        buf.append(indent + "  ID = " + jm.getId() + "\n");
        buf.append(indent + "  RETURN TYPE = " + jm.getFQN() + "\n");
        buf.append(indent + "  MODIFIER = " + jm.getModifier() + "\n");

        buf.append(indent + "  A = " + jm.getAnchorText() +  "\n");
        buf.append(indent + "  File = " + jm.getFileName() +  "\n");
        buf.append(indent + "  Line = " + jm.getLine() +  "\n");
        buf.append("\n");
    }

    private void writeElement(QueryJavaElement je, String title, String indent) {
        buf.append(indent + title + "\n");
        buf.append(indent + "  ID = " + je.getId() + "\n");

        buf.append(indent + "  A = " + je.getAnchorText() +  "\n");
        buf.append(indent + "  File = " + je.getFileName() + "\n");
        buf.append(indent + "  Line = " + je.getLine() + "\n");
        buf.append("\n");
    }

    private void writeElement(ArrayList nodes, String title, String indent) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            QueryJavaElement je = new QueryJavaElement(elem);
            writeElement(je, title, indent);
        }
    }
}
