/*
 *     GenSummaryHTML.java  Oct 18, 2003
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package Japid.genxrefxml;

import org.jtool.platform.pgstruct.*;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.util.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class GenSummaryHTML {

    public static void main(String[] args) {
        if (args.length > 0) {
                GenSummaryHTML gen = new GenSummaryHTML();
                gen.view(args[0]);
        } else {
            System.err.println("Usage: java GenSummaryHTML xml_filename");
        }
    }

    public void view(String fileName) {
        String content = write(fileName);

        System.out.println("---- ");
        System.out.print(content);
        System.out.println("----");
    }

    private StringBuffer buf;

    public String write(String fileName) {
        try {
            String content = ReadFile.read(fileName);
            Document doc = DOMTreeCreator.create(content);
            return write(doc);

        } catch (IOException e) { }
        return null;
    }

    public String write(Document doc) {
        QueryProject jp = new QueryProject(doc.getDocumentElement());
        buf = new StringBuffer();
        
        writePackage(jp);
        writeClass(jp.getClassNodes(), "CLASS");
        writeClass(jp.getInterfaceNodes(), "INTERFACE");
        writeMember(jp.getFieldNodes(), "FIELD");
        writeMember(jp.getMethodNodes(), "METHOD");

        return buf.toString();
    }

    private void writePackage(QueryProject jp) {
        buf.append("PACKAGE {\n");
        Iterator it = jp.getPackageNames().iterator();
        while (it.hasNext()) {
            String name = (String)it.next();
            buf.append("  " + name + "\n");
        }
        buf.append("} PACKAGE\n");
    }

    private void writeClass(ArrayList nodes, String title) {
        buf.append("\n");
        buf.append(title + " {\n");

        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            QueryJavaClass jc = new QueryJavaClass(elem);
            buf.append("  " + jc.getName());
            buf.append(" [ " + jc.getFileName() +  " ]\n");
        }
        buf.append("} " + title + "\n");
    }

    private void writeMember(ArrayList nodes, String title) {
        buf.append("\n");
        buf.append(title + " {\n");

        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            QueryJavaElement je = new QueryJavaElement(elem);
            QueryJavaClass jc = new QueryJavaClass((Element)elem.getParentNode());

            buf.append("  " + je.getText());
            buf.append(" [ " + jc.getFileName() +  " ]\n");
        }
        buf.append("} " + title + "\n");
    }

    private String getPath(String path) {
        return path.substring(0, path.lastIndexOf(".")) + ".html";
    }
}
