/* 
 * Program:             $RCSfile: DirWalkIterator.java,v $  $Revision: 50.2 $
 * 
 * Purpose:             
 *
 * Author:              Y.Sawada  2003/10/01
 *                      S.Yamamoto  2003/10/01
 *
 * (C) Copyright:       Y.Sawada and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: DirWalkIterator.java,v 50.2 2003/11/17 13:26:29 yamamoto Exp yamamoto $
 */

package jspie3;

import java.io.*;
import java.util.*;

public class DirWalkIterator implements Iterator
{
    private LinkedList stack;
    private HashSet hashSet;
    private boolean canonical;
    private String regExp;

    public DirWalkIterator(String fileName)
    {
        stack = new LinkedList();
        stack.addFirst(new File(fileName));
        hashSet = new HashSet();
        canonical = false;
        regExp = null;
    }
    public DirWalkIterator(String fileName, String regExp)
    {
        this(fileName);
        this.regExp = regExp;
    }
    public DirWalkIterator(String fileName, boolean canonical)
    {
        this(fileName);
        this.canonical = canonical;
    }
    public DirWalkIterator(String fileName, String regExp, boolean canonical)
    {
        this(fileName);
        this.regExp = regExp;
        this.canonical = canonical;
    }
    private boolean matchRegExp(File file, String regExp)
    {
        if (regExp == null) {
            return true;
        } else {
            return file.getName().matches(regExp);
        }
    }
    private boolean canonicalp(File file)
    {
        if (canonical) {
            String canonicalPath = new String();
            try {
                canonicalPath = file.getCanonicalPath();
            } catch (Exception e) {
                e.printStackTrace();
            }
            return hashSet.add(canonicalPath);
        } else {
            return true;
        }
    }

    public boolean hasNext()
    {
        if (stack.size() == 0)
            return false;
        File tmpFile = (File)stack.getFirst();

        if (tmpFile.isDirectory()) {
            File[] fileList = tmpFile.listFiles();
            for (int i = fileList.length - 1; i >= 0; i--) {
                if (fileList[i].isDirectory() ||
                    matchRegExp(fileList[i], regExp) &&
                    canonicalp(fileList[i])) {
                    stack.add(1, fileList[i]);
                }
            }
            stack.removeFirst();
            return hasNext();
        }
        return true;
    }

    public Object next()
    {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        return stack.removeFirst();
    }

    public void remove()
    {
        if (hasNext()) {
            stack.removeFirst();
        }
    }
}
