/* 
 * Program:             $RCSfile: MemoTreeVisitor.java,v $  $Revision: 50.3 $
 * 
 * Purpose:             
 *
 * Author:              Y.Sawada  2003/10/01
 *                      S.Yamamoto  2003/10/01
 *
 * (C) Copyright:       Y.Sawada and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: MemoTreeVisitor.java,v 50.3 2003/11/18 14:53:11 yamamoto Exp yamamoto $
 */

package jspie3;

import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

public class MemoTreeVisitor {
    protected Document linkDoc;
    protected StringBuffer result;
    protected int depth;
    protected String rootFileName;
    
    public MemoTreeVisitor(String path){
	try {
            FileReader fl = new FileReader(path + "/memo-link.xml");
	    BufferedReader br = new BufferedReader(fl);
            InputSource src = new InputSource(br);
	    DocumentBuilderFactory dbfactory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder builder = dbfactory.newDocumentBuilder();
	    linkDoc = builder.parse(src);
	    result = new StringBuffer();
	    depth = 0;
	    rootFileName = new String();
	} catch (Exception e) {
	    //	    e.printStackTrace();
	}
    }

    public void printResult(){
	System.out.println(result.toString());
    }

    public String getResult(){
	return result.toString();
    }

    public void destroy(){
	result = new StringBuffer();
    }

    public void memoTreeVisit(){
	Node root = linkDoc.getDocumentElement();
	NodeList nl = root.getChildNodes();
	for(int i = 0 ; i < nl.getLength(); i++){
	    Node node = nl.item(i);
	    if(node.getNodeName().equals("link") && !isChild(node)){
		rootFileName = node.getChildNodes().item(1).getFirstChild().getNodeValue();
		travel(node,((Element)node).getAttribute("id"));
	    }
	}
    }
    
    public void memoTreeVisit(String targetFileName){
	Node root = linkDoc.getDocumentElement();
	NodeList nl = root.getChildNodes();
	for(int i = 0 ; i < nl.getLength(); i++){
	    Node node = nl.item(i);
	    if(node.getNodeName().equals("link") && !isChild(node)){
		rootFileName = node.getChildNodes().item(1).getFirstChild().getNodeValue();
		if(rootFileName != null && rootFileName.equals(targetFileName)){
		    travel(node,((Element)node).getAttribute("id"));
		}
	    }
	}
    }
    
    protected void travel(Node node, String linkIds){
	visit(node,linkIds);
	depth ++;
	String linkId = ((Element)node).getAttribute("id");
	NodeList nl = node.getChildNodes();
	Element memo = (Element)nl.item(3);
	String memoId = memo.getAttribute("id");
	ArrayList children = getChildren(memoId,linkIds);
	for(int i = 0; i < children.size(); i++){
	    String child = (String)children.get(i);
	    String str = linkIds + "_" + child;
	    Node childNode = getLinkNode(child);
	    travel(childNode,str);
	}
	depth --;
	visitAfter(node,linkIds);
    }
    
    protected void visit(Node node, String linkIds){
	for(int i = 0;i < depth; i++){
	    result.append("-");
	}
	result.append("linkIds : " + linkIds + "\n");
    }

    protected void visitAfter(Node node, String linkIds){
    }

    public ArrayList getChildren(String memoId, String linkIds) {
	ArrayList children = new ArrayList();
	Node root = linkDoc.getDocumentElement();
	NodeList nl = root.getChildNodes();
	for(int i = 0 ; i < nl.getLength(); i++){
	    Node node = nl.item(i);
	    if(node.getNodeName().equals("link")){
		String linkId = ((Element)node).getAttribute("id");
		NodeList nl2 = node.getChildNodes();
		Element pos = (Element)nl2.item(1);
		Element memo = (Element)nl2.item(3);
		String posId = pos.getAttribute("id");
		if(posId !=null && (posId.equals(memoId) || posId.equals(linkIds))){
		    children.add(linkId);
		}
	    }
	}	
	return children;
    }

    protected boolean isChild(Node node){
	NodeList nl = node.getChildNodes();
	Element pos = (Element)nl.item(1);
	String posId = pos.getAttribute("id");
	if(posId.matches("^m[1-9][0-9]*$") ||
	   posId.matches("^l[1-9][0-9]*(_l[1-9][0-9]*)*$")){
	    return true;
	}
	return false;
    }
    
    protected String getMemoId(String targetLinkId){
	Node root = linkDoc.getDocumentElement();
	NodeList nl = root.getChildNodes();
	for(int i = 0 ; i < nl.getLength(); i++){
	    Node node = nl.item(i);
	    if(node.getNodeName().equals("link")){
		String linkId = ((Element)node).getAttribute("id");
		if(linkId !=null && linkId.equals(targetLinkId)){
		    NodeList nl2 = node.getChildNodes();
		    Element memo = (Element)nl2.item(3);
		    String posId = memo.getAttribute("id");
		    return posId;
		}
	    }
	}
	return null;
    }

    protected Node getLinkNode(String targetLinkId){
	Node root = linkDoc.getDocumentElement();
	NodeList nl = root.getChildNodes();
	for(int i = 0 ; i < nl.getLength(); i++){
	    Node node = nl.item(i);
	    if(node.getNodeName().equals("link")){
		String linkId = ((Element)node).getAttribute("id");
		if(linkId !=null && (linkId.equals(targetLinkId))){
		   return node;
		}
	    }
	}
	return null;
    }
}
