/* 
 * Program:             $RCSfile: Stopwatch.java,v $  $Revision: 40.7 $
 * 
 * Purpose:             A simple timer for measuring performance.
 *
 * Author:              S.Wilson  2000/07/18
 *                      J.Kesselman  2000/07/18
 * Modified:            S.Yamamoto  2003/07/18
 *
 * (C) Copyright:       S.Wilson and J.Kesselman  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

/* $Id: Stopwatch.java,v 40.7 2003/07/25 13:21:20 yamamoto Exp yamamoto $ */

package Japid.util;

import java.text.*;

/**
  * A class to help benchmark code snippits.
  * It simulates a real stop watch.
  */
public class Stopwatch {

    private long startTime = -1;
    private long stopTime = -1;
    private boolean running = false;

    public Stopwatch start() {
        startTime = System.currentTimeMillis();
        running = true;
        return this;
    }
    public Stopwatch stop() {
        stopTime = System.currentTimeMillis();
        running = false;
        return this;
    }

    /**
     * Returns elapsed time in milliseconds. 
     * If the watch has never been started then return zero.
     */
    public long getElapsedTime() {
        if (startTime == -1) {
            return 0;
        }
        long endTime = 0;
        if (running) {
            return System.currentTimeMillis() - startTime;
        } else {
            return stopTime - startTime; 
        }     
    }
  
    public Stopwatch reset() {
        startTime = -1;
        stopTime = -1;
        running = false;
        return this;
    }

    public void printTimeProf(String command) {
        System.err.print("Sapid: " + command + ": ");
        printTimeProf();
    }

    public void printTimeProf() {
        long     diff = getElapsedTime();
        long     seconds = diff / 1000;
        long     milliseconds = diff % 1000;
        String   buf;

        if (3600 < seconds) {
            // "%d:%02d:%02d.%02d".
            buf = seconds / 3600 + ":" +
                (seconds % 3600) / 60 + ":" +
                seconds % 60 + "." +
                milliseconds / 10;
        } else {
            // "%d:%02d.%02d".
            buf = seconds / 60 + ":" +
                seconds % 60 + "." +
                milliseconds / 10;
        }

        DecimalFormat df1 = new DecimalFormat("0.00");
        StringBuffer sb1 = new StringBuffer();
        FieldPosition fp1 = new FieldPosition(NumberFormat.INTEGER_FIELD);
        
        // "Time profile: \"%.2fu %.2fs %s\".\n".
        System.err.println("Time profile: \"" +
                           df1.format((float)diff / 1000, sb1, fp1) + "u " +
                           "0.00s " +
                           buf + "\".");
    }
}
