/* 
 * Program:             $RCSfile: AreaModel.c,v $  $Revision: 30.26 $
 * 
 * Purpose:             AreaModel饹
 * 
 * Author:              T.Hidaka  1998/05/21
 *                      S.Yamamoto  1998/05/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: AreaModel.c,v 30.26 2003/04/07 16:02:47 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName storage_sort[] = {
    { SPD_VAL_NAME3(AREA_STORAGE_GLOBAL, "global") },
    { SPD_VAL_NAME3(AREA_STORAGE_FUNC, "func") },
    { SPD_VAL_NAME3(AREA_STORAGE_LOCAL, "local") },
    { SPD_VAL_NAME3(AREA_STORAGE_HEAP, "heap") },
    { 0, NULL, NULL }
};

static SpdValName area_sort[] = {
    { SPD_VAL_NAME3(AREA_BASIC, "basic") },
    { SPD_VAL_NAME3(AREA_STRUCT, "struct") },
    { SPD_VAL_NAME3(AREA_UNION, "union") },
    { SPD_VAL_NAME3(AREA_ARRAY, "array") },
    { 0, NULL, NULL }
};

static SpdValName area_indirect[] = {
    { SPD_VAL_NAME3(AREA_INDIRECT, "indirect") },
    { SPD_VAL_NAME3(AREA_DIRECT, "direct") },
    { 0, NULL, NULL }
};

static SpdValName value_sort[] = {
    { SPD_VAL_NAME3(VALUE_NONVALUE, "non value") },
    { SPD_VAL_NAME3(VALUE_BASIC, "basic") },
    { SPD_VAL_NAME3(VALUE_POINTER, "pointer") },
    { SPD_VAL_NAME3(VALUE_ENUM, "enum") },
    { 0, NULL, NULL }
};

static SpdValName value_type[] = {
    { SPD_VAL_NAME3(VALUE_NONVALUE, "non value") },
    { SPD_VAL_NAME3(VALUE_INTEGER, "integer") },
    { SPD_VAL_NAME3(VALUE_UINTEGER, "unsigned integer") },
    { SPD_VAL_NAME3(VALUE_POINTER, "pointer") },
    { SPD_VAL_NAME3(VALUE_FUNCPOINTER, "func pointer") },
    { SPD_VAL_NAME3(VALUE_REALPOINTER, "real pointer") },
    { 0, NULL, NULL }
};

static SpdDBClass Area_model_classes[] = {
    /* Classes. */
    {
        "storage",
        SPD_CID_STORAGE,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "sort", SPD_DBT_ENUM, storage_sort },
            /* { "level", SPD_DBT_INT }, */
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "area",
        SPD_CID_AREA,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, SAPIDDB_ATTR_NONE },
        {
            { "sort", SPD_DBT_ENUM, area_sort },
            { "address", SPD_DBT_INT },
            { "size", SPD_DBT_INT },
            { "indirect", SPD_DBT_ENUM, area_indirect },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "value",
        SPD_CID_VALUE,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "sort", SPD_DBT_ENUM, value_sort },
            { "type", SPD_DBT_ENUM, value_type },
            { "value", SPD_DBT_INT },
            /* !! binaryƳɬפ롥 */
            { NULL, SPD_DBT_NON },
        }
    },
    /* Relationships. */
    {
        "storage_area",
        SPD_RID_STORAGE_AREA,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE },
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "storage_storage",
        SPD_RID_STORAGE_STORAGE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"p_storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE },
            {"c_storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "area_area",
        SPD_RID_AREA_AREA,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"parent_area_id", SPD_DBT_OID, NULL, SPD_CID_AREA }, 
            {"child_area_id", SPD_DBT_OID, NULL, SPD_CID_AREA },
            {"order", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "area_value",
        SPD_RID_AREA_VALUE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA }, 
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE },
            {"order", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "point_to",
        SPD_RID_POINT_TO,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"pointer_id", SPD_DBT_OID, NULL, SPD_CID_VALUE },
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "funcptr_to",
        SPD_RID_FUNCPTR_TO,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"pointer_id", SPD_DBT_OID, NULL, SPD_CID_VALUE },
            {"function_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "prog_storage",
        SPD_RID_PROG_STORAGE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"program_id", SPD_DBT_OID, NULL, SPD_CID_PROGRAM }, 
            {"storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE },
            {"order", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "func_storage",
        SPD_RID_FUNC_STORAGE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"optdecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            {"storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE },
            {"order", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "block_storage",
        SPD_RID_BLOCK_STORAGE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK }, 
            {"storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE },
            {"order", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "variable_area",
        SPD_RID_VARIABLE_AREA,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"variable_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA },
            {"order", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "member_area",
        SPD_RID_MEMBER_AREA,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"member_id", SPD_DBT_OID, NULL, SPD_CID_MEMBER }, 
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "string_area",
        SPD_RID_STRING_AREA,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"string_id", SPD_DBT_OID, NULL, SPD_CID_CONSTANT },
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "area_type",
        SPD_RID_AREA_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"area_id", SPD_DBT_OID, NULL, SPD_CID_AREA },
            {"type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "value_type",
        SPD_RID_VALUE_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE },
            {"type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "value_enum",
        SPD_RID_VALUE_ENUM,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE },
            {"enum_id", SPD_DBT_OID, NULL, SPD_CID_ENUMERATOR }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        NULL, SPD_NON_ID, 0, {SAPIDDB_ATTR_NONE}
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel Area_model = {
    SPD_MID_AREA_MODEL,
    "Area-model",
    SPD_AREA_MODEL_VERSION,
    SPD_AREA_MODEL_DIR,

    SPD_FIRST_CID_AREA_MODEL,
    SPD_LAST_CID_AREA_MODEL,
    SPD_CID_NUM_AREA_MODEL,

    SPD_FIRST_RID_AREA_MODEL,
    SPD_LAST_RID_AREA_MODEL,
    SPD_RID_NUM_AREA_MODEL,

    SPD_CRID_NUM_AREA_MODEL,

#if defined(SAPID_AR5)

    Area_model_classes,
    NULL

#endif /* defined(SAPID_AR5) */
};
