/* 
 * Program:             $RCSfile: Bmodel.c,v $  $Revision: 30.17 $
 * 
 * Purpose:             Class definition of B-model.
 * 
 * Author:              T.Hidaka and S.Yamamoto  1998/03/15
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: Bmodel.c,v 30.17 2003/04/07 16:02:48 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName execution_sort[] = {
    { SPD_VAL_NAME3(OPERATION_ALLOC, "alloc") },
    { SPD_VAL_NAME3(OPERATION_FREE, "free") },
    { SPD_VAL_NAME3(OPERATION_DEFINE, "define") },
    { SPD_VAL_NAME3(OPERATION_REFER, "refer") },
    { SPD_VAL_NAME3(OPERATION_CALL, "call") },
    { SPD_VAL_NAME3(OPERATION_RETURN, "return") },
    { SPD_VAL_NAME3(OPERATION_BRANCH, "branch") },
    { SPD_VAL_NAME3(OPERATION_ENTRY, "entry") },
    { SPD_VAL_NAME3(OPERATION_EXIT, "exit") },
    { 0, NULL, NULL }
};

static SpdDBClass B_model_classes[] = {
    /* Classes. */
    {
        "execution",
        SPD_CID_EXECUTION,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "operation",
        SPD_CID_OPERATION,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "sort", SPD_DBT_ENUM, execution_sort },
            { NULL, SPD_DBT_NON },
        }
    },

    /* Relationships. */
    {
        "prog_exec",
        SPD_RID_PROG_EXEC,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"program_id", SPD_DBT_OID, NULL, SPD_CID_PROGRAM },
            {"execution_id", SPD_DBT_OID, NULL, SPD_CID_EXECUTION }, 
            {"order", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "exec_op",
        SPD_RID_EXEC_OP,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"execution_id", SPD_DBT_OID, NULL, SPD_CID_EXECUTION }, 
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"order", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "statement_op",
        SPD_RID_STATEMENT_OP,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"statement_id", SPD_DBT_OID, NULL, SPD_NON_ID},
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"order", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "alloc_storage",
        SPD_RID_ALLOC_STORAGE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "free_storage",
        SPD_RID_FREE_STORAGE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"storage_id", SPD_DBT_OID, NULL, SPD_CID_STORAGE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "define_value",
        SPD_RID_DEFINE_VALUE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "refer_value",
        SPD_RID_REFER_VALUE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "call_value",
        SPD_RID_CALL_VALUE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "return_value",
        SPD_RID_RETURN_VALUE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "branch_value",
        SPD_RID_BRANCH_VALUE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"operation_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION },
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "depend_on",
        SPD_RID_DEPEND_ON,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            {"depend_value_id", SPD_DBT_OID, NULL, SPD_CID_VALUE }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "call_func",
        SPD_RID_CALL_FUNC,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"call_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION }, 
            {"func_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "call_return",
        SPD_RID_CALL_RETURN,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"call_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION }, 
            {"return_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "entry_exit",
        SPD_RID_ENTRY_EXIT,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"entry_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION }, 
            {"exit_id", SPD_DBT_OID, NULL, SPD_CID_OPERATION }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        NULL, SPD_NON_ID, 0, {SAPIDDB_ATTR_NONE}
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel B_model = {
    SPD_MID_B_MODEL,
    "B-model",
    SPD_B_MODEL_VERSION,
    SPD_B_MODEL_DIR,

    SPD_FIRST_CID_B_MODEL,
    SPD_LAST_CID_B_MODEL,
    SPD_CID_NUM_B_MODEL,

    SPD_FIRST_RID_B_MODEL,
    SPD_LAST_RID_B_MODEL,
    SPD_RID_NUM_B_MODEL,

    SPD_CRID_NUM_B_MODEL,

#if defined(SAPID_AR5)

    B_model_classes,
    NULL,

#endif /* defined(SAPID_AR5) */
};
