/* 
 * Program:             $RCSfile: Dmodel.c,v $  $Revision: 30.19 $
 * 
 * Purpose:             D-model饹
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: Dmodel.c,v 30.19 2003/07/08 06:19:50 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName docfrag_sort[] = {
    { SPD_VAL_NAME3(DOCFRAG_TEXTFRAG, "textfrag") },
    { SPD_VAL_NAME3(DOCFRAG_BASICELEM, "basicelem") },
    { SPD_VAL_NAME3(DOCFRAG_DEFELEM, "defelem") },
    { SPD_VAL_NAME3(DOCFRAG_LINKELEM, "linkelem") },
    { SPD_VAL_NAME3(DOCFRAG_IDENTELEM, "identelem") },
    { SPD_VAL_NAME3(DOCFRAG_SRCELEM, "srcelem") },
    { SPD_VAL_NAME3(DOCFRAG_MACROELEM, "macroelem") },
    { 0, NULL, NULL }
};

static SpdValName docident_sort[] = {
    { SPD_VAL_NAME3(DID_VARIABLE, "variable") },
    { SPD_VAL_NAME3(DID_FUNCTION, "function") },
    { SPD_VAL_NAME3(DID_MEMBER, "member") },
    { SPD_VAL_NAME3(DID_ENUMERATOR, "enumerator") },
    { SPD_VAL_NAME3(DID_TYPENAME, "type") },
    { SPD_VAL_NAME3(DID_TAG, "tag") },
    { SPD_VAL_NAME3(DID_MACRONAME, "macro") },
    { SPD_VAL_NAME3(DID_FILE, "file") },
    { SPD_VAL_NAME3(DID_UNKNOWN, "unkown") },
    { 0, NULL, NULL }
};

static SpdDBClass D_model_classes[] = {
    /*---------- ֥ ----------*/
    {
        "software",
        SPD_CID_SOFTWARE,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "document",
        SPD_CID_DOCUMENT,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docfile",
        SPD_CID_DOCFILE,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docfrag",
        SPD_CID_DOCFRAG,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, docfrag_sort },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "textstring",
        SPD_CID_TEXTSTRING,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "text", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "attribute",
        SPD_CID_ATTRIBUTE,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docident",
        SPD_CID_DOCIDENT,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, docident_sort },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    /*---------- Ϣ ----------*/
    {
        "soft_doc",
        SPD_RID_SOFT_DOC,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "software_id", SPD_DBT_OID, NULL, SPD_CID_SOFTWARE }, 
            { "document_id", SPD_DBT_OID, NULL, SPD_CID_DOCUMENT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "soft_prog",
        SPD_RID_SOFT_PROG,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "software_id", SPD_DBT_OID, NULL, SPD_CID_SOFTWARE }, 
            { "program_id", SPD_DBT_OID, NULL, SPD_CID_PROGRAM },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "doc_docfile",
        SPD_RID_DOC_DOCFILE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "document_id", SPD_DBT_OID, NULL, SPD_CID_DOCUMENT }, 
            { "docfile_id", SPD_DBT_OID, NULL, SPD_CID_DOCFILE },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docfile_docfrag",
        SPD_RID_DOCFILE_DOCFRAG,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "docfile_id", SPD_DBT_OID, NULL, SPD_CID_DOCFILE }, 
            { "docfrag_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docelem_docfrag",
        SPD_RID_DOCELEM_DOCFRAG,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "docelem_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 
            { "docfrag_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docelem_attribute",
        SPD_RID_DOCELEM_ATTRIBUTE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "docelem_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 
            { "attribute_id", SPD_DBT_OID, NULL, SPD_CID_ATTRIBUTE },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "attribute_value",
        SPD_RID_ATTRIBUTE_VALUE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "attribute_id", SPD_DBT_OID, NULL, SPD_CID_ATTRIBUTE }, 
            { "value_id", SPD_DBT_OID, NULL, SPD_CID_TEXTSTRING },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "textfrag_textstring",
        SPD_RID_TEXTFRAG_TEXTSTRING,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "textfrag_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 
            { "textstring_id", SPD_DBT_OID, NULL, SPD_CID_TEXTSTRING },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "defelem_docident",
        SPD_RID_DEFELEM_DOCIDENT,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "defelem_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 
            { "docident_id", SPD_DBT_OID, NULL, SPD_CID_DOCIDENT },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "identelem_docident",
        SPD_RID_IDENTELEM_DOCIDENT,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "identelem_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 
            { "docident_id", SPD_DBT_OID, NULL, SPD_CID_DOCIDENT },
            { "offset", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "srcelem_srcfrag",
        SPD_RID_SRCELEM_SRCFRAG,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "srcelem_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 

            { "srcfrag_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION },
            /* !! ߤϡsrcfrag_iddeclaration˷Ǥ */

            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "docident_identifier",
        SPD_RID_DOCIDENT_IDENTIFIER,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "docident_id", SPD_DBT_OID, NULL, SPD_CID_DOCIDENT }, 
            { "identifier_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "linkelem_srcfrag",
        SPD_RID_LINKELEM_SRCFRAG,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "linkelem_id", SPD_DBT_OID, NULL, SPD_CID_DOCFRAG }, 
            { "srcfrag_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION }, 
            /*
             * !! ߤϡsrcfrag_iddeclaration˷Ǥ
             * ImodelΤ٤ƤΥ饹פɽ褦ˤ򵭽Ҥˤ
             * ɤ餤Τ?
             */
            { NULL, SPD_DBT_NON },
        }
    },
    {
        NULL, SPD_NON_ID, 0, { SAPIDDB_ATTR_NONE }
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel D_model = {
    SPD_MID_D_MODEL,
    "D-model",
    SPD_D_MODEL_VERSION,
    D_MODEL_DIR,

    SPD_FIRST_CID_D_MODEL,
    SPD_LAST_CID_D_MODEL,
    SPD_CID_NUM_D_MODEL,

    SPD_FIRST_RID_D_MODEL,
    SPD_LAST_RID_D_MODEL,
    SPD_RID_NUM_D_MODEL,

    SPD_CRID_NUM_D_MODEL,

#if defined(SAPID_AR5)

    D_model_classes,
    NULL,
#endif /* defined(SAPID_AR5) */
};
