/* 
 * Program:             $RCSfile: FCD-view.c,v $  $Revision: 40.4 $
 * 
 * Purpose:             Class definition of FCD (funciton call dependency) view.
 *
 * Author:              N.Atsumi  2001/11/29
 *                      S.Yamamoto  2001/11/29 
 *
 * (C) Copyright:       N.Atsumi and S.Yamamoto  2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: FCD-view.c,v 40.4 2003/04/07 16:02:58 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName fcdg_sort[] = {
    { SPD_VAL_NAME3(FCDG_LIBDEF, "libdef") },
    { SPD_VAL_NAME3(FCDG_LIBREF, "libref") },
    { SPD_VAL_NAME3(FCDG_LIBDEFREF, "libdefref") },
    { SPD_VAL_NAME3(FCDG_CONDITION, "condition") },
    { 0, NULL, NULL }
};

static SpdDBClass FCD_view_classes[] = {
    /* Classes. */
    {
        "fcdg",
        SPD_CID_FCDG,
        SAPIDDB_SORT_OBJ,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE},
        {
            {"sort", SPD_DBT_ENUM, fcdg_sort}, 
            {NULL, SPD_DBT_NON},
        }
    },
    /* Relationships. */
    {
        "func_fcdg",
        SPD_RID_FUNC_FCDG,
        SAPIDDB_SORT_REL_COMPOSITE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"func_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION}, 
            {"fcdg_id", SPD_DBT_OID, NULL, SPD_CID_FCDG},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "libdata_depend",
        SPD_RID_LIBDATA_DEPEND,
        SAPIDDB_SORT_REL_COMPOSITE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"source_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"target_id", SPD_DBT_OID, NULL, SPD_CID_FCDG},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "libctrl_depend",
        SPD_RID_LIBCTRL_DEPEND,
        SAPIDDB_SORT_REL_COMPOSITE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"source_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"target_id", SPD_DBT_OID, NULL, SPD_CID_FCDG},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "fcdg_expr",
        SPD_RID_FCDG_EXPR,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"fcdg_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"expr_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "fcdg_lib",
        SPD_RID_FCDG_LIB,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"fcdg_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"lib_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "def_var",
        SPD_RID_DEF_VAR,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"fcdg_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"variable_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "cond_op",
        SPD_RID_COND_OP,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"cond_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"op_id", SPD_DBT_OID, NULL, SPD_CID_OPERATOR},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "cond_val",
        SPD_RID_COND_VAL,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"fcdg_id", SPD_DBT_OID, NULL, SPD_CID_FCDG}, 
            {"value_id", SPD_DBT_OID, NULL, SPD_CID_CONSTANT},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        NULL, SPD_NON_ID, 0, {SAPIDDB_ATTR_NONE}
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel FCD_view = {
    SPD_MID_FCD_VIEW,
    "FCD-view",
    SPD_FCD_VIEW_VERSION,
    SPD_FCD_VIEW_DIR,

    SPD_FIRST_CID_FCD_VIEW,
    SPD_LAST_CID_FCD_VIEW,
    SPD_CID_NUM_FCD_VIEW,

    SPD_FIRST_RID_FCD_VIEW,
    SPD_LAST_RID_FCD_VIEW,
    SPD_RID_NUM_FCD_VIEW,

    SPD_CRID_NUM_FCD_VIEW,

#if defined(SAPID_AR5)

    FCD_view_classes,
    NULL,

#endif /* defined(SAPID_AR5) */
};
