/* 
 * Program:             $RCSfile: FlowView.c,v $  $Revision: 40.4 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/11/17
 *                      S.Yamamoto  2001/11/17
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: FlowView.c,v 40.4 2003/04/07 16:02:57 yamamoto Exp yamamoto $";
#endif

/* 
 * Class definition of FlowView.
 */

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName blk_sort[] = {
    { SPD_VAL_NAME3(SPD_FV_B_FUNC, "func") },
    { SPD_VAL_NAME3(SPD_FV_B_COMP, "comp") },
    { SPD_VAL_NAME3(SPD_FV_B_IF, "if") },
    { SPD_VAL_NAME3(SPD_FV_B_SWITCH, "switch") },
    { SPD_VAL_NAME3(SPD_FV_B_FOR, "for") },
    { SPD_VAL_NAME3(SPD_FV_B_WHILE, "while") },
    { SPD_VAL_NAME3(SPD_FV_B_DO, "do") },
    { 0, NULL, NULL }
};

static SpdValName edge_sort[] = {
    { SPD_VAL_NAME3(SPD_FV_E_NORMAL, "normal") },
    { SPD_VAL_NAME3(SPD_FV_E_BRANCH_TRUE, "true") },
    { SPD_VAL_NAME3(SPD_FV_E_BRANCH_FALSE, "false") },
    { SPD_VAL_NAME3(SPD_FV_E_LOOPBACK, "loopback") },
    { SPD_VAL_NAME3(SPD_FV_E_SWITCH_CASE, "case") },
    { SPD_VAL_NAME3(SPD_FV_E_SWITCH_DEFAULT, "default") },
    { SPD_VAL_NAME3(SPD_FV_E_SWITCH_NOMATCH, "nomatch") },
    { 0, NULL, NULL }
};

static SpdDBClass Flow_view_classes[] = {
    /* Classes. */
    {
        "blkEntry",
        SPD_CID_BLK_ENTRY,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blkGoInto",
        SPD_CID_BLK_GO_INTO,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blkExit",
        SPD_CID_BLK_EXIT,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blkBreak",
        SPD_CID_BLK_BREAK,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blkContinue",
        SPD_CID_BLK_CONTINUE,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blkGoAway",
        SPD_CID_BLK_GO_AWAY,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blkReturn",
        SPD_CID_BLK_RETURN,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, blk_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "switchBranch",
        SPD_CID_SWITCH_BRANCH,
        SAPIDDB_SORT_OBJ_TREENODE,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_INT },   /* dummy */
            { NULL, SPD_DBT_NON },
        }
    },

    /* Relationships. */
    {
        "next_node",
        SPD_RID_NEXT_NODE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"curId", SPD_DBT_OID, NULL, SPD_CID_SDA5_ANY_NODE},
            {"nextId", SPD_DBT_OID, NULL, SPD_CID_SDA5_ANY_NODE },
            {"sort", SPD_DBT_ENUM, edge_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blk_sdanode",
        SPD_RID_BLK_SDANODE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"blkId", SPD_DBT_OID, NULL, SPD_CID_SDA5_ANY_NODE },
            {"sdanodeId", SPD_DBT_OID, NULL, SPD_CID_SDA5_ANY_NODE },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "data_dependence",
        SPD_RID_DATA_DEPENDENCE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"defObjId", SPD_DBT_OID, NULL, SPD_CID_SDA5_ANY_NODE },
            {"refExprId", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "def_used",
        SPD_RID_DEF_USED,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"defObjId", SPD_DBT_OID, NULL, SPD_CID_SDA5_ANY_NODE },
            {"usedExprId", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        NULL, SPD_NON_ID, 0, {SAPIDDB_ATTR_NONE}
    }

};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel Flow_view = {
    SPD_MID_FLOW_VIEW,
    "Flow-view",
    SPD_FLOW_VIEW_VERSION,
    SPD_FLOW_VIEW_DIR,

    SPD_FIRST_CID_FLOW_VIEW,
    SPD_LAST_CID_FLOW_VIEW,
    SPD_CID_NUM_FLOW_VIEW,

    SPD_FIRST_RID_FLOW_VIEW,
    SPD_LAST_RID_FLOW_VIEW,
    SPD_RID_NUM_FLOW_VIEW,

    SPD_CRID_NUM_FLOW_VIEW,

#if defined(SAPID_AR5)

    Flow_view_classes,
    NULL,

#endif /* defined(SAPID_AR5) */
};
