/* 
 * Program:             $RCSfile: Imodel.c,v $  $Revision: 30.24 $
 * 
 * Purpose:             Class definition of I-model.
 * 
 * Author:              T.Hidaka  1998/05/21
 *                      S.Yamamoto  1998/05/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: Imodel.c,v 30.24 2003/07/30 16:26:47 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName declaration_sort[] = {
    { SPD_VAL_NAME3(DECL_NORMAL, "normal") },
    { SPD_VAL_NAME3(DECL_FUNCDEF, "funcdef") },
    { SPD_VAL_NAME3(DECL_MEMBER, "member") },
    { SPD_VAL_NAME3(DECL_ENUMERATOR, "enumerator") },
    { SPD_VAL_NAME3(DECL_PARAMETER, "parameter") },
    { SPD_VAL_NAME3(DECL_TYPE, "type") },
    { 0, NULL, NULL }
};

static SpdValName declaration_storage[] = {
    { SPD_VAL_NAME3(STORAGE_AUTO, "auto") },
    { SPD_VAL_NAME3(STORAGE_EXTERN, "extern") },
    { SPD_VAL_NAME3(STORAGE_STATIC, "static") },
    { SPD_VAL_NAME3(STORAGE_REGISTER, "register") },
    { SPD_VAL_NAME3(STORAGE_TYPEDEF, "typedef") },
    { 0, NULL, NULL }
};

static SpdValName declarator_sort[] = {
    { SPD_VAL_NAME3(DECL_VARIABLE, "variable") },
    { SPD_VAL_NAME3(DECL_MEMBER, "member") },
    { SPD_VAL_NAME3(DECL_FUNCDEF, "funcdef") },
    { SPD_VAL_NAME3(DECL_PROTOTYPE, "prototype") },
    { SPD_VAL_NAME3(DECL_TAG, "tag") },
    { SPD_VAL_NAME3(DECL_TYPE, "type") },
    { SPD_VAL_NAME3(DECL_ENUMERATOR, "enumerator") },
    { 0, NULL, NULL }
};

static SpdValName identifier_sort[] = {
    { SPD_VAL_NAME3(ID_VARIABLE, "variable") },
    { SPD_VAL_NAME3(ID_FUNCTION, "function") },
    { SPD_VAL_NAME3(ID_MEMBER, "member") },
    { SPD_VAL_NAME3(ID_ENUMERATOR, "enumerator") },
    { SPD_VAL_NAME3(ID_TYPENAME, "type") },
    { SPD_VAL_NAME3(ID_TAG, "tag") },
    { 0, NULL, NULL }
};

static SpdValName optionaldecl_sort[] = {
    { SPD_VAL_NAME3(OPT_PROTOTYPE, "prototype") },
    { SPD_VAL_NAME3(OPT_FUNCDEF, "funcdef") },
    { SPD_VAL_NAME3(OPT_INITIALIZER, "initializer") },
    { SPD_VAL_NAME3(OPT_STRUCT, "struct") },
    { SPD_VAL_NAME3(OPT_FUNCPTR, "funcptr") },
    { 0, NULL, NULL }
};

static SpdValName block_sort[] = {
    { SPD_VAL_NAME3(BLOCK_BASIC, "basic block") },
    { SPD_VAL_NAME3(BLOCK_IF, "if") },
    { SPD_VAL_NAME3(BLOCK_SWITCH, "switch") },
    { SPD_VAL_NAME3(BLOCK_FOR, "for") },
    { SPD_VAL_NAME3(BLOCK_WHILE, "while") },
    { SPD_VAL_NAME3(BLOCK_DO, "do") },
    { SPD_VAL_NAME3(BLOCK_COMP, "composite block") },
    { SPD_VAL_NAME3(BLOCK_BRANCH, "branch block") },
    { SPD_VAL_NAME3(BLOCK_HIERACHICAL, "hierachical block") },
    { 0, NULL, NULL }
};

static SpdValName expression_sort[] = {
    { SPD_VAL_NAME3(EXPR_EMPTY, "empty") },
    { SPD_VAL_NAME3(EXPR_BREAK, "break") },
    { SPD_VAL_NAME3(EXPR_CONTINUE, "continue") },
    { SPD_VAL_NAME3(EXPR_RETURN, "return") },
    { SPD_VAL_NAME3(EXPR_GOTO, "goto") },
    { SPD_VAL_NAME3(EXPR_SORT_ARITH, "arithmetic") },
    { SPD_VAL_NAME3(EXPR_SORT_LOGICAL, "logical") },
    { SPD_VAL_NAME3(EXPR_SORT_ARRAY, "array") },
    { SPD_VAL_NAME3(EXPR_SORT_MEMBER, "member") },
    { SPD_VAL_NAME3(EXPR_SORT_POINTER, "pointer") },
    { SPD_VAL_NAME3(EXPR_SORT_ADDRESS, "address") },
    { SPD_VAL_NAME3(EXPR_SORT_TRI, "?:") },
    { SPD_VAL_NAME3(EXPR_SORT_SIZEOF, "sizeof") },
    { SPD_VAL_NAME3(EXPR_SORT_CAST, "cast") },
    { SPD_VAL_NAME3(EXPR_SORT_COMMA, "comma") },
    { SPD_VAL_NAME3(EXPR_SORT_FUNCCALL, "function call") },
    { SPD_VAL_NAME3(EXPR_SORT_INITLIST, "initializer") },
    { SPD_VAL_NAME3(EXPR_SORT_BUILTIN, "builtin function") },
    { SPD_VAL_NAME3(EXPR_SORT_BIT, "bit") },
    { SPD_VAL_NAME3(EXPR_SORT_SHIFT, "shift") },
    { SPD_VAL_NAME3(EXPR_SORT_COMPARE, "compare") },
    { SPD_VAL_NAME3(EXPR_SORT_POINTER_MEMBER, "pointer member") },
    { 0, NULL, NULL }
};

static SpdValName expression_lrtype[] = {
    { SPD_VAL_NAME3(EXPR_LR_UNKNOWN, "unknown") },
    { SPD_VAL_NAME3(EXPR_L_VALUE, "left value") },
    { SPD_VAL_NAME3(EXPR_R_VALUE, "right value") },
    { SPD_VAL_NAME3(EXPR_PL_VALUE, "left value with pointer") },
    { SPD_VAL_NAME3(EXPR_PR_VALUE, "right value with pointer") },
    { SPD_VAL_NAME3(EXPR_P_REF, "reference") },
    { SPD_VAL_NAME3(EXPR_LR_VALUE, "left and right value") },
    { SPD_VAL_NAME3(EXPR_PLR_VALUE, "left and right value with pointer") },
    { 0, NULL, NULL }
};

static SpdValName label_sort[] = {
    { SPD_VAL_NAME3(LABEL_NAMED, "label") },
    { SPD_VAL_NAME3(LABEL_CASE, "case") },
    { SPD_VAL_NAME3(LABEL_DEFAULT, "default") },
    { 0, NULL, NULL }
};

static SpdValName constant_type[] = {
    { SPD_VAL_NAME3(CONST_INT, "int") },
    { SPD_VAL_NAME3(CONST_FLOAT, "float") },
    { SPD_VAL_NAME3(CONST_CHAR, "character") },
    { SPD_VAL_NAME3(CONST_STRING, "string") },
    { SPD_VAL_NAME3(CONST_ENUM, "enum") },
    { SPD_VAL_NAME3(CONST_WCHAR, "wide character") },
    { 0, NULL, NULL }
};

static SpdValName type_sort[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, "unknown") },
    { SPD_VAL_NAME3(TYPE_TYPE_DEF, "typedef") },
    { SPD_VAL_NAME3(TYPE_STRUCT_TAG, "struct tag") },
    { SPD_VAL_NAME3(TYPE_DECLARATION, "standard") },
    { SPD_VAL_NAME3(TYPE_POINTER, "pointer") },
    { SPD_VAL_NAME3(TYPE_ARRAY, "array") },
    { SPD_VAL_NAME3(TYPE_FUNC, "function") },
    { SPD_VAL_NAME3(TYPE_ENUM_SPECIFIER, "enumerate") },
    { SPD_VAL_NAME3(TYPE_UNION_TAG, "union tag") },
    { SPD_VAL_NAME3(TYPE_BUILTIN, "builtin") },
    { 0, NULL, NULL }
};

static SpdValName type_qualifier[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, "unknown") },
    { SPD_VAL_NAME3(TYPE_CONST, "const") },
    { SPD_VAL_NAME3(TYPE_VOLATILE, "volatile") },
    { 0, NULL, NULL }
};

static SpdValName type_type[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, "unknown") },
    { SPD_VAL_NAME3(TYPE_VOID, "void") },
    { SPD_VAL_NAME3(TYPE_INT, "int") },
    { SPD_VAL_NAME3(TYPE_CHAR, "char") },
    { SPD_VAL_NAME3(TYPE_FLOAT, "float") },
    { SPD_VAL_NAME3(TYPE_DOUBLE, "double") },
    { SPD_VAL_NAME3(TYPE_DOTS, "...") },
    { SPD_VAL_NAME3(TYPE_BUILTIN_TYPE, "builtin type") },
    { 0, NULL, NULL }
};

static SpdValName type_size[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, "unknown") },
    { SPD_VAL_NAME3(TYPE_SHORT, "short") },
    { SPD_VAL_NAME3(TYPE_LONG, "long") },
    { SPD_VAL_NAME3(TYPE_LONG_LONG, "long long") },
    { 0, NULL, NULL }
};

static SpdValName type_sign[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, "unknown") },
    { SPD_VAL_NAME3(TYPE_SIGNED, "signed") },
    { SPD_VAL_NAME3(TYPE_UNSIGNED, "unsigned") },
    { 0, NULL, NULL }
};

static SpdValName operator_sort[] = {
    { SPD_VAL_NAME3(OP_L_PAR, "(") },
    { SPD_VAL_NAME3(OP_R_PAR, ")") },
    { SPD_VAL_NAME3(OP_L_ARRAY, "[") },
    { SPD_VAL_NAME3(OP_R_ARRAY, "]") },
    { SPD_VAL_NAME3(OP_ARROW, "->") },
    { SPD_VAL_NAME3(OP_DOT, ".") },
    { SPD_VAL_NAME3(OP_NOT, "!") },
    { SPD_VAL_NAME3(OP_B_NOT, "~") },
    { SPD_VAL_NAME3(OP_INC, "++") },
    { SPD_VAL_NAME3(OP_DEC, "--") },
    { SPD_VAL_NAME3(OP_PLUS, "+") },
    { SPD_VAL_NAME3(OP_MINUS, "-") },
    { SPD_VAL_NAME3(OP_POINTER, "*") },
    { SPD_VAL_NAME3(OP_ADDRESS, "&") },
    { SPD_VAL_NAME3(OP_MUL, "*") },
    { SPD_VAL_NAME3(OP_DIV, "/") },
    { SPD_VAL_NAME3(OP_MOD, "%") },
    { SPD_VAL_NAME3(OP_ADD, "+") },
    { SPD_VAL_NAME3(OP_SUB, "-") },
    { SPD_VAL_NAME3(OP_SHL, "<<") },
    { SPD_VAL_NAME3(OP_SHR, ">>") },
    { SPD_VAL_NAME3(OP_LT, "<") },
    { SPD_VAL_NAME3(OP_LE, "<=") },
    { SPD_VAL_NAME3(OP_GT, ">") },
    { SPD_VAL_NAME3(OP_GE, ">=") },
    { SPD_VAL_NAME3(OP_EQ, "==") },
    { SPD_VAL_NAME3(OP_NE, "!=") },
    { SPD_VAL_NAME3(OP_B_AND, "&") },
    { SPD_VAL_NAME3(OP_B_XOR, "^") },
    { SPD_VAL_NAME3(OP_B_OR, "|") },
    { SPD_VAL_NAME3(OP_L_AND, "&&") },
    { SPD_VAL_NAME3(OP_L_OR, "||") },
    { SPD_VAL_NAME3(OP_ASSIGN, "=") },
    { SPD_VAL_NAME3(OP_A_ADD, "+=") },
    { SPD_VAL_NAME3(OP_A_SUB, "-=") },
    { SPD_VAL_NAME3(OP_A_MUL, "*=") },
    { SPD_VAL_NAME3(OP_A_DIV, "/=") },
    { SPD_VAL_NAME3(OP_A_MOD, "%=") },
    { SPD_VAL_NAME3(OP_A_AND, "&=") },
    { SPD_VAL_NAME3(OP_A_XOR, "^=") },
    { SPD_VAL_NAME3(OP_A_OR, "|=") },
    { SPD_VAL_NAME3(OP_A_SHL, "<<=") },
    { SPD_VAL_NAME3(OP_A_SHR, ">>=") },
    { SPD_VAL_NAME3(OP_COMMA, ",") },
    { SPD_VAL_NAME3(OP_SIZEOF, "sizeof") },
    { SPD_VAL_NAME3(OP_SIZEOF_TYPE, "sizeof(type)") },
    { SPD_VAL_NAME3(OP_CAST, "(type)") },
    { SPD_VAL_NAME3(OP_QUESTION, "?") },
    { SPD_VAL_NAME3(OP_COLON, ":") },
    { SPD_VAL_NAME3(OP_BUILTIN, "__builtin") },
    { SPD_VAL_NAME3(OP_BUILTIN_TYPE, "__builtin(type)") },
    { 0, NULL, NULL }
};

static SpdDBClass I_model_classes[] = {
    /* Classes. */
    {
        "program",
        SPD_CID_PROGRAM,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "file",
        SPD_CID_FILE,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "declaration",
        SPD_CID_DECLARATION,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, declaration_sort },
            { "storage", SPD_DBT_ENUM, declaration_storage },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "declarator",
        SPD_CID_DECLARATOR,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, declarator_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "identifier",
        SPD_CID_IDENTIFIER,
        SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, identifier_sort },
            { "scope", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "optionaldecl",
        SPD_CID_OPTIONALDECL,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, optionaldecl_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "block",
        SPD_CID_BLOCK,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, block_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "expression",
        SPD_CID_EXPRESSION,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, expression_sort },
            { "lrtype", SPD_DBT_ENUM, expression_lrtype },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "label",
        SPD_CID_LABEL,
        SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, label_sort }, 
            { "name", SPD_DBT_STR }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "constant",
        SPD_CID_CONSTANT,
        SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "type", SPD_DBT_ENUM, constant_type }, 
            { "value", SPD_DBT_STR }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "type",
        SPD_CID_TYPE,
        SAPIDDB_SORT_OBJ,
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, 5, SAPIDDB_ATTR_NONE },
        {
            { "sort", SPD_DBT_ENUM, type_sort }, 
            { "qualifier", SPD_DBT_ENUM, type_qualifier }, 
            { "type", SPD_DBT_ENUM, type_type }, 
            { "size", SPD_DBT_ENUM, type_size }, 
            { "sign", SPD_DBT_ENUM, type_sign }, 
            { "name", SPD_DBT_STR }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "operator",
        SPD_CID_OPERATOR,
        SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, operator_sort }, 
            { NULL, SPD_DBT_NON },
        }
    },
    /* Added by H.Ohhashi 1999/07/01. */
    {
        "library",
        SPD_CID_LIBRARY,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    /* Relationships. */
    {
        "blk_blk",
        SPD_RID_BLK_BLK,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "parent_block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK }, 
            { "children_block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blk_decl",
        SPD_RID_BLK_DECL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "parent_block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK }, 
            { "children_declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blk_expr",
        SPD_RID_BLK_EXPR,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "parent_block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK }, 
            { "children_expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "blk_label",
        SPD_RID_BLK_LABEL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK }, 
            { "label_id", SPD_DBT_OID, NULL, SPD_CID_LABEL },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "case_cexpr",
        SPD_RID_CASE_CEXPR,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "label_id", SPD_DBT_OID, NULL, SPD_CID_LABEL }, 
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "const_ref",
        SPD_RID_CONST_REF,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "const_id", SPD_DBT_OID, NULL, SPD_CID_CONSTANT },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "decl_decl",
        SPD_RID_DECL_DECL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION }, 
            { "declarator_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATOR },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "decl_ident",
        SPD_RID_DECL_IDENT,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "declarator_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATOR }, 
            { "identifier_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "decl_opt",
        SPD_RID_DECL_OPT,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "declarator_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATOR }, 
            { "optionaldecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "decl_type",
        SPD_RID_DECL_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION }, 
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE },
            { "offset", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "expr_expr",
        SPD_RID_EXPR_EXPR,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "outer_expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "inner_expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "expr_op",
        SPD_RID_EXPR_OP,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "operator_id", SPD_DBT_OID, NULL, SPD_CID_OPERATOR }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "expr_type",
        SPD_RID_EXPR_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "file_decl",
        SPD_RID_FILE_DECL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "file_id", SPD_DBT_OID, NULL, SPD_CID_FILE }, 
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "func_body",
        SPD_RID_FUNC_BODY,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "function_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            { "block_id", SPD_DBT_OID, NULL, SPD_CID_BLOCK },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "func_parm",
        SPD_RID_FUNC_PARM,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "optionaldecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "goto_label",
        SPD_RID_GOTO_LABEL,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "label_id", SPD_DBT_OID, NULL, SPD_CID_LABEL }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "ident_ref",
        SPD_RID_IDENT_REF,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "identifier_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "ident_type",
        SPD_RID_IDENT_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "identifier_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "init_expr",
        SPD_RID_INIT_EXPR,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "optionaldecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "op_type",
        SPD_RID_OP_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "operator_id", SPD_DBT_OID, NULL, SPD_CID_OPERATOR }, 
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "prog_file",
        SPD_RID_PROG_FILE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "program_id", SPD_DBT_OID, NULL, SPD_CID_PROGRAM }, 
            { "file_id", SPD_DBT_OID, NULL, SPD_CID_FILE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "struct_decl",
        SPD_RID_STRUCT_DECL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "optionaldecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "type_cexpr",
        SPD_RID_TYPE_CEXPR,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { "expression_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "type_is",
        SPD_RID_TYPE_IS,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { "identifier_id", SPD_DBT_OID, NULL, SPD_CID_IDENTIFIER }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "type_st",
        SPD_RID_TYPE_ST,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "type_pointer_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { "type_pointed_id", SPD_DBT_OID, NULL, SPD_CID_TYPE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "prot_parm",
        SPD_RID_PROT_PARM,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "optionaldecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "funcptr_parm",
        SPD_RID_FUNCPTR_PARM,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "optionaldecl_id", SPD_DBT_OID, NULL, SPD_CID_OPTIONALDECL }, 
            { "declaration_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "decr_type",
        SPD_RID_DECR_TYPE,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "declarator_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATOR }, 
            { "type_id", SPD_DBT_OID, NULL, SPD_CID_TYPE },
            { NULL, SPD_DBT_NON },
        }
    },
    /* Added by H.Ohhashi 1999/07/01. */
    {
        "prog_lib",
        SPD_RID_PROG_LIB,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "program_id", SPD_DBT_OID, NULL, SPD_CID_PROGRAM }, 
            { "library_id", SPD_DBT_OID, NULL, SPD_CID_LIBRARY }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "lib_file",
        SPD_RID_LIB_FILE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "library_id", SPD_DBT_OID, NULL, SPD_CID_LIBRARY }, 
            { "file_id", SPD_DBT_OID, NULL, SPD_CID_FILE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        NULL, SPD_NON_ID, 0, { SAPIDDB_ATTR_NONE }
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel I_model = {
    SPD_MID_I_MODEL,
    "I-model",
    SPD_I_MODEL_VERSION,
    I_MODEL_DIR,

    SPD_FIRST_CID_I_MODEL,
    SPD_LAST_CID_I_MODEL,
    SPD_CID_NUM_I_MODEL,

    SPD_FIRST_RID_I_MODEL,
    SPD_LAST_RID_I_MODEL,
    SPD_RID_NUM_I_MODEL,

    SPD_CRID_NUM_I_MODEL,

#if defined(SAPID_AR5)

    I_model_classes,
    NULL

#endif /* defined(SAPID_AR5) */
};
