/* 
 * Program:             $RCSfile: Jmodel.c,v $  $Revision: 40.11 $
 * 
 * Purpose:             Class definition of J-model.
 * 
 * Author:              S.Yamamoto  2002/07/15
 * 
 * (C) Copyright:       S.Yamamoto  2002 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: Jmodel.c,v 40.11 2003/08/10 14:13:59 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SapidOffsetOf.h>
#include "JB.h"

#if defined(SAPID_AR5)

/* static SpdDBClass J_model_classes[] = { */
SpdDBClass J_model_classes[] = {
    /* Classes. */
    {
        "File",
        SPD_CID_J_FILE, /* 189 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "encoding", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.file), SPD_OFFSET_OF(JpdFile, encoding) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Class",
        SPD_CID_J_CLASS, /* 190 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.jclass),  SPD_OFFSET_OF(JpdClass, sort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Interface",
        SPD_CID_J_INTERFACE, /* 191 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.interface), SPD_OFFSET_OF(JpdInterface, sort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Variable",
        SPD_CID_J_VARIABLE, /* 192 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.variable), SPD_OFFSET_OF(JpdVariable, sort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Routine",
        SPD_CID_J_ROUTINE, /* 193 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET } ,
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.routine), SPD_OFFSET_OF(JpdRoutine, sort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Type",
        SPD_CID_J_TYPE, /* 194 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, 5, SAPIDDB_ATTR_NONE },
        {
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.type),  SPD_OFFSET_OF(JpdType, sort) },
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "pkgName", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(u.type), SPD_OFFSET_OF(JpdType, pkgName) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Modifier",
        SPD_CID_J_MODIFIER, /* 195 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "SBlock",
        SPD_CID_J_SBLOCK, /* 196 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.sblock), SPD_OFFSET_OF(JpdSBlock, sort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Statement",
        SPD_CID_J_STATEMENT, /* 197 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.statement), SPD_OFFSET_OF(JpdStatement, sort) },
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Expression",
        SPD_CID_J_EXPRESSION, /* 198 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, 5, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.statement), SPD_OFFSET_OF(JpdStatement, sort) },
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "lrsort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.expression), SPD_OFFSET_OF(JpdExpression, lrsort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Literal",
        SPD_CID_J_LITERAL, /* 199 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.literal), SPD_OFFSET_OF(JpdLiteral, sort) },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Operator",
        SPD_CID_J_OPERATOR, /* 200 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.joperator), SPD_OFFSET_OF(JpdOperator, sort) },
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "QName",
        SPD_CID_J_QNAME, /* 201 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Special",
        SPD_CID_J_SPECIAL, /* 202 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Label",
        SPD_CID_J_LABEL, /* 203 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "sort", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(u.label), SPD_OFFSET_OF(JpdLabel, sort) },
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Comment",
        SPD_CID_J_COMMENT, /* 204 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, 0, 1, 2, 3, SAPIDDB_ATTR_NONE },
        {
            { "offset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(offset), SDB_NON_OFFSET },
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(length), SDB_NON_OFFSET },
            { "strOffset", SPD_DBT_INT, NULL, SPD_NON_ID, 0, SymDataOffset(strOffset), SDB_NON_OFFSET },
            { "str", SPD_DBT_STR, NULL, SPD_NON_ID, 0, SymDataOffset(str), SDB_NON_OFFSET },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "Wildcard",
        SPD_CID_J_WILDCARD, /* 205 */
        SAPIDDB_SORT_OBJ_TREENODE, /* !! */
        { SAPIDDB_ATTR_ID, SAPIDDB_ATTR_NONE },
        {
            { NULL, SPD_DBT_NON }
        },
    },
    /* Relationships. */
    {
        "define_package",
        SPD_RID_J_DEFINE_PACKAGE, /* 206 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "file", SPD_DBT_OID, NULL, SPD_CID_J_FILE },
            { "package", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "import_package",
        SPD_RID_J_IMPORT_PACKAGE, /* 207 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "file", SPD_DBT_OID, NULL, SPD_CID_J_FILE },
            { "package", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "define_class",
        SPD_RID_J_DEFINE_CLASS, /* 208 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "file", SPD_DBT_OID, NULL, SPD_CID_J_FILE },
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "implement_interface",
        SPD_RID_J_IMPLEMENT_INTERFACE, /* 209 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "extend_class",
        SPD_RID_J_EXTEND_CLASS, /* 210 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "subClass", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "superClass", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_modifier",
        SPD_RID_J_CLASS_MODIFIER, /* 211 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "modifier", SPD_DBT_OID, NULL, SPD_CID_J_MODIFIER },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_variable",
        SPD_RID_J_CLASS_VARIABLE, /* 212 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_routine",
        SPD_RID_J_CLASS_ROUTINE, /* 213 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_sblock",
        SPD_RID_J_CLASS_SBLOCK, /* 214 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "sblock", SPD_DBT_OID, NULL, SPD_CID_J_SBLOCK },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_argument",
        SPD_RID_J_CLASS_ARGUMENT, /* 215 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_class",
        SPD_RID_J_CLASS_CLASS, /* 216 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "class_interface",
        SPD_RID_J_CLASS_INTERFACE, /* 217 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "define_interface",
        SPD_RID_J_DEFINE_INTERFACE, /* 218 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "file", SPD_DBT_OID, NULL, SPD_CID_J_FILE },
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "extend_interface",
        SPD_RID_J_EXTEND_INTERFACE, /* 219 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "subInterface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { "superInterface", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "interface_modifier",
        SPD_RID_J_INTERFACE_MODIFIER, /* 220 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { "modifier", SPD_DBT_OID, NULL, SPD_CID_J_MODIFIER },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "interface_variable",
        SPD_RID_J_INTERFACE_VARIABLE, /* 221 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "interface_routine",
        SPD_RID_J_INTERFACE_ROUTINE, /* 222 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "interface_class",
        SPD_RID_J_INTERFACE_CLASS, /* 223 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "interface_interface",
        SPD_RID_J_INTERFACE_INTERFACE, /* 224 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "variable_modifier",
        SPD_RID_J_VARIABLE_MODIFIER, /* 225 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { "modifier", SPD_DBT_OID, NULL, SPD_CID_J_MODIFIER },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "variable_type",
        SPD_RID_J_VARIABLE_TYPE, /* 226 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { "type", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "initialize_variable",
        SPD_RID_J_INITIALIZE_VARIABLE, /* 227 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { "initializer", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "parameter_variable",
        SPD_RID_J_PARAMETER_VARIABLE, /* 228 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "return_type",
        SPD_RID_J_RETURN_TYPE, /* 229 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { "returnType", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "throw_type",
        SPD_RID_J_THROW_TYPE, /* 230 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { "throwType", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "routine_body",
        SPD_RID_J_ROUTINE_BODY, /* 231 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { "body", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "routine_modifier",
        SPD_RID_J_ROUTINE_MODIFIER, /* 232 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { "modifier", SPD_DBT_OID, NULL, SPD_CID_J_MODIFIER },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "type_class",
        SPD_RID_J_TYPE_CLASS, /* 233 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "type", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "type_interface",
        SPD_RID_J_TYPE_INTERFACE, /* 234 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "type", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { "interface", SPD_DBT_OID, NULL, SPD_CID_J_INTERFACE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "type_type",
        SPD_RID_J_TYPE_TYPE, /* 235 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "parentType", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { "childType", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "type_qname",
        SPD_RID_J_TYPE_QNAME, /* 236 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "type", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { "qname", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "qname_qname",
        SPD_RID_J_QNAME_QNAME, /* 237 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "parentQName", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { "chidlQName", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "qname_dot",
        SPD_RID_J_QNAME_DOT, /* 238 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "qname", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { "dot", SPD_DBT_OID, NULL, SPD_CID_J_OPERATOR },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "sblock_body",
        SPD_RID_J_SBLOCK_BODY, /* 239 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "sblock", SPD_DBT_OID, NULL, SPD_CID_J_SBLOCK },
            { "body", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "local_class",
        SPD_RID_J_LOCAL_CLASS, /* 240 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "statement", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { "class", SPD_DBT_OID, NULL, SPD_CID_J_CLASS },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "local_variable",
        SPD_RID_J_LOCAL_VARIABLE, /* 241 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "statement", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "statement_label",
        SPD_RID_J_STATEMENT_LABEL, /* 242 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "statement", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { "label", SPD_DBT_OID, NULL, SPD_CID_J_LABEL },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "statement_statement",
        SPD_RID_J_STATEMENT_STATEMENT, /* 243 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "outerStatement", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { "innerStatement", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "statement_expression",
        SPD_RID_J_STATEMENT_EXPRESSION, /* 244 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "statement", SPD_DBT_OID, NULL, SPD_CID_J_STATEMENT },
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "refer_label",
        SPD_RID_J_REFER_LABEL, /* 245 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "referer", SPD_DBT_OID, NULL, SPD_CID_J_LABEL },
            { "referee", SPD_DBT_OID, NULL, SPD_CID_J_LABEL },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "case_value",
        SPD_RID_J_CASE_VALUE, /* 246 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "label", SPD_DBT_OID, NULL, SPD_CID_J_LABEL },
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "expression_operand",
        SPD_RID_J_EXPRESSION_OPERAND, /* 247 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "operandExpression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "expression_operator",
        SPD_RID_J_EXPRESSION_OPERATOR, /* 248 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "operator", SPD_DBT_OID, NULL, SPD_CID_J_OPERATOR },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "expression_expression",
        SPD_RID_J_EXPRESSION_EXPRESSION, /* 249 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "parentExpression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "childExpression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "refer_variable",
        SPD_RID_J_REFER_VARIABLE, /* 250 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "variable", SPD_DBT_OID, NULL, SPD_CID_J_VARIABLE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "refer_literal",
        SPD_RID_J_REFER_LITERAL, /* 251 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "literal", SPD_DBT_OID, NULL, SPD_CID_J_LITERAL },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "refer_special",
        SPD_RID_J_REFER_SPECIAL, /* 252 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "special", SPD_DBT_OID, NULL, SPD_CID_J_SPECIAL },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "refer_type",
        SPD_RID_J_REFER_TYPE, /* 253 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "type", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "refer_qname",
        SPD_RID_J_REFER_QNAME, /* 254 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "qname", SPD_DBT_OID, NULL, SPD_CID_J_QNAME },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "has_type",
        SPD_RID_J_HAS_TYPE, /* 255 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "type", SPD_DBT_OID, NULL, SPD_CID_J_TYPE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "invoke_routine",
        SPD_RID_J_INVOKE_ROUTINE, /* 256 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "expression", SPD_DBT_OID, NULL, SPD_CID_J_EXPRESSION },
            { "routine", SPD_DBT_OID, NULL, SPD_CID_J_ROUTINE },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        "symbol_comment",
        SPD_RID_J_SYMBOL_COMMENT, /* 257 */
        SAPIDDB_SORT_REL_COMPOSITE, /* !! */
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "symbol", SPD_DBT_OID, NULL, SPD_CID_J_WILDCARD },
            { "comment", SPD_DBT_OID, NULL, SPD_CID_J_COMMENT },
            { NULL, SPD_DBT_NON }
        },
    },
    {
        NULL, SPD_NON_ID, 0, { SAPIDDB_ATTR_NONE },
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel J_model = {
    SPD_MID_J_MODEL,
    "J-model",
    SPD_J_MODEL_VERSION,
    J_MODEL_DIR,

    SPD_FIRST_CID_J_MODEL,
    SPD_LAST_CID_J_MODEL,
    SPD_CID_NUM_J_MODEL,

    SPD_FIRST_RID_J_MODEL,
    SPD_LAST_RID_J_MODEL,
    SPD_RID_NUM_J_MODEL,

    SPD_CRID_NUM_J_MODEL,

#if defined(SAPID_AR5)

    J_model_classes,
    NULL

#endif /* defined(SAPID_AR5) */
};
