/* 
 * Program:             $RCSfile: Pmodel.c,v $  $Revision: 30.19 $
 * 
 * Purpose:             Class definition of P-model.
 * 
 * Author:              T.Hidaka  1998/10/13
 * Modified:            S.Yamamoto  2000/05/01
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: Pmodel.c,v 30.19 2003/04/07 16:02:45 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName pblock_sort[] = {
    { SPD_VAL_NAME3(PBLOCK_COMMENT, "comment") },
    { SPD_VAL_NAME3(PBLOCK_MACROCALL, "macrocall") },
    { SPD_VAL_NAME3(PBLOCK_NORMAL, "normal") },
    { SPD_VAL_NAME3(PBLOCK_DIRECTIVE, "directive") },

    { SPD_VAL_NAME3(PBLOCK_DEFINE_FILE, "#define") },
    { SPD_VAL_NAME3(PBLOCK_DEFINE_OPTION, "define(cmdline)") },
    { SPD_VAL_NAME3(PBLOCK_DEFINE_INIT, "define(internal)") },
    { SPD_VAL_NAME3(PBLOCK_UNDEF, "#undef") },
    { SPD_VAL_NAME3(PBLOCK_INCLUDE, "#include") },
    { SPD_VAL_NAME3(PBLOCK_IF, "#if") },
    { SPD_VAL_NAME3(PBLOCK_ELSE, "#else") },
    { SPD_VAL_NAME3(PBLOCK_ELIF, "#elif") },
    { SPD_VAL_NAME3(PBLOCK_ENDIF, "#endif") },
    { SPD_VAL_NAME3(PBLOCK_IFDEF, "#ifdef") },
    { SPD_VAL_NAME3(PBLOCK_IFNDEF, "#ifndef") },
    { SPD_VAL_NAME3(PBLOCK_PRAGMA, "#pragma") },
    { SPD_VAL_NAME3(PBLOCK_LINE, "#line") },
    { SPD_VAL_NAME3(PBLOCK_ERROR, "#error") },
    { 0, NULL, NULL }
};

static SpdValName hfile_sort[] = {
    { SPD_VAL_NAME3(HFILE_ANGLE, "<...>") },
    { SPD_VAL_NAME3(HFILE_DQUOT, "\"...\"") },
    { 0, NULL, NULL }
};

static SpdDBClass P_model_classes[] = {
    /* Classes. */
    {
        "pfile",
        SPD_CID_PFILE,
        SAPIDDB_SORT_OBJ_TREENODE,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "pblock",
        SPD_CID_PBLOCK,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, pblock_sort },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroPattern",
        SPD_CID_MACROPATTERN,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroBody",
        SPD_CID_MACROBODY,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroName",
        SPD_CID_MACRONAME,
        SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroOptional",
        SPD_CID_MACROOPTIONAL,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroDeclArg",
        SPD_CID_MACRODECLARG,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "name", SPD_DBT_STR },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroCall",
        SPD_CID_MACROCALL,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "macroCallArg",
        SPD_CID_MACROCALLARG,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "name", SPD_DBT_STR }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "hfile",
        SPD_CID_HFILE,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { "sort", SPD_DBT_ENUM, hfile_sort }, 
            { "name", SPD_DBT_STR }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "iblock",
        SPD_CID_IBLOCK,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        { SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            { "length", SPD_DBT_INT, NULL, SPD_NON_ID, 0 },
            { NULL, SPD_DBT_NON },
        }
    },
    /* Relationships. */
    {
        "file_block",
        SPD_RID_FILE_BLOCK,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "pfile_id", SPD_DBT_OID, NULL, SPD_CID_PFILE }, 
            { "pblock_id", SPD_DBT_OID, NULL, SPD_CID_PBLOCK },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "call_mcall",
        SPD_RID_CALL_MCALL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "pblock_id", SPD_DBT_OID, NULL, SPD_CID_PBLOCK },
            { "macroCall_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALL },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mbody_mcall",
        SPD_RID_MBODY_MCALL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "macroBody_id", SPD_DBT_OID, NULL, SPD_CID_MACROBODY },
            { "macroCall_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALL },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mcall_mname",
        SPD_RID_MCALL_MNAME,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "macroCall_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALL },
            { "macroName_id", SPD_DBT_OID, NULL, SPD_CID_MACRONAME },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mcall_mcalla",
        SPD_RID_MCALL_MCALLA,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "macroCall_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALL },
            { "macroCallArg_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALLARG },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "def_mpatt",
        SPD_RID_DEF_MPATT,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "pblock_id", SPD_DBT_OID, NULL, SPD_CID_PBLOCK },
            { "macroPattern_id", SPD_DBT_OID, NULL, SPD_CID_MACROPATTERN },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "def_mbody",
        SPD_RID_DEF_MBODY,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "pblock_id", SPD_DBT_OID, NULL, SPD_CID_PBLOCK },
            { "macroBody_id", SPD_DBT_OID, NULL, SPD_CID_MACROBODY },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mpatt_mname",
        SPD_RID_MPATT_MNAME,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "macroPattern_id", SPD_DBT_OID, NULL, SPD_CID_MACROPATTERN },
            { "macroName_id", SPD_DBT_OID, NULL, SPD_CID_MACRONAME },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mpatt_mopt",
        SPD_RID_MPATT_MOPT,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "macroPattern_id", SPD_DBT_OID, NULL, SPD_CID_MACROPATTERN },
            { "macroOptional_id", SPD_DBT_OID, NULL, SPD_CID_MACROOPTIONAL },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mopt_mdecla",
        SPD_RID_MOPT_MDECLA,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "macroOptional_id", SPD_DBT_OID, NULL, SPD_CID_MACROOPTIONAL },
            { "macroDeclArg_id", SPD_DBT_OID, NULL, SPD_CID_MACRODECLARG },
            { "offset", SPD_DBT_INT },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mbody_mdecla",
        SPD_RID_MBODY_MDECLA,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, 3, 4, SAPIDDB_ATTR_NONE },
        {
            { "macroBody_id", SPD_DBT_OID, NULL, SPD_CID_MACROBODY },
            { "macroDeclArg_id", SPD_DBT_OID, NULL, SPD_CID_MACRODECLARG },
            { "order", SPD_DBT_INT }, 
            { "skip", SPD_DBT_INT }, 
            { "sharp", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "incld_hfile",
        SPD_RID_INCLD_HFILE,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "pblock_id", SPD_DBT_OID, NULL, SPD_CID_PBLOCK },
            { "hfile_id", SPD_DBT_OID, NULL, SPD_CID_HFILE },
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "hfile_file",
        SPD_RID_HFILE_FILE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "hfile_id", SPD_DBT_OID, NULL, SPD_CID_HFILE }, 
            { "pfile_id", SPD_DBT_OID, NULL, SPD_CID_PFILE }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "mcalla_mcall",
        SPD_RID_MCALLA_MCALL,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "macroCallArg_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALLARG }, 
            { "macroCall_id", SPD_DBT_OID, NULL, SPD_CID_MACROCALL }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "cfile_ifile",
        SPD_RID_CFILE_IFILE,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "pfile_id", SPD_DBT_OID, NULL, SPD_CID_PFILE }, 
            { "file_id", SPD_DBT_OID, NULL, SPD_CID_FILE },
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "ifile_iblock",
        SPD_RID_IFILE_IBLOCK,
        SAPIDDB_SORT_REL_COMPOSITE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            { "file_id", SPD_DBT_OID, NULL, SPD_CID_FILE }, 
            { "iblock_id", SPD_DBT_OID, NULL, SPD_CID_IBLOCK }, 
            { "offset", SPD_DBT_INT }, 
            { NULL, SPD_DBT_NON },
        }
    },
    {
        "pblock_iblock",
        SPD_RID_PBLOCK_IBLOCK,
        SAPIDDB_SORT_REL_ONE_TO_ONE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            { "pblock_id", SPD_DBT_OID, NULL, SPD_CID_PBLOCK }, 
            { "iblock_id", SPD_DBT_OID, NULL, SPD_CID_IBLOCK }, 
            { NULL, SPD_DBT_NON },
        }
    },

    {
        NULL, SPD_NON_ID, 0, { SAPIDDB_ATTR_NONE }
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel P_model = {
    SPD_MID_P_MODEL,
    "P-model",
    SPD_P_MODEL_VERSION,
    P_MODEL_DIR,

    SPD_FIRST_CID_P_MODEL,
    SPD_LAST_CID_P_MODEL,
    SPD_CID_NUM_P_MODEL,

    SPD_FIRST_RID_P_MODEL,
    SPD_LAST_RID_P_MODEL,
    SPD_RID_NUM_P_MODEL,

    SPD_CRID_NUM_P_MODEL,

#if defined(SAPID_AR5)

    P_model_classes,
    NULL

#endif /* defined(SAPID_AR5) */
};
