/* 
 * Program:             $RCSfile: PmodelAR.c,v $  $Revision: 30.13 $
 * 
 * Purpose:             P-model access functions.
 * 
 * Author:              T.Hidaka  1998/11/13
 * Modified:            S.Yamamoto  2000/05/01
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: PmodelAR.c,v 30.13 2003/07/08 07:46:19 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelARcompat.h>
#include <Sapid/PmodelAR.h>

static SpdSet allindexes = NULL;

static void freeIndexes(SpdSet *indexes)
{
    int             i;
    SpdRegionIndex  *regIndex;

    for (i = 0; i < (*indexes)->elemNum; i++) {
        regIndex = (SpdRegionIndex *)spdGetAElem(*indexes, NULL);
        spddb_freeRegionIndex(regIndex);
    }
    spdFreeSet(*indexes);
}

static void freeAllIndexes(void)
{
    int     i;
    SpdSet  *indexes;
    
    for (i = 0; i < allindexes->elemNum; i++) {
        indexes = (SpdSet *)spdGetAElem(allindexes, NULL);
        freeIndexes(indexes);
    }
    spdFreeSet(allindexes);
    allindexes = NULL;
}

static SpdSet *getIndexesOfDB(SpdDBId dbId)
{
    SpdSet              *pindexes;
    SpdSet              indexes;
    static SpdBoolean   is_first_call = TRUE;

    if (allindexes == NULL) {
        allindexes = spdCreateSet(sizeof(SpdSet));
        if (is_first_call) {
            atexit(freeAllIndexes);
            is_first_call = FALSE;
        }
    }
    if ((pindexes = spdSearchElem(allindexes, dbId)) == NULL) {
        indexes = spdCreateSet(sizeof(SpdRegionIndex));
        spdAddElem(allindexes, dbId, (Contents)&indexes);
        pindexes = spdSearchElem(allindexes, dbId);
    }
    return pindexes;
}

void spdFreePmodelRegIndex(SpdObjId fileId)
{
    SpdDBId         dbId;
    SpdSet          *indexes;
    SpdRegionIndex  *pregIndex;
    
    dbId = spdGetDefaultDB();
    indexes = getIndexesOfDB(dbId);
    if (indexes == NULL) {
        return;
    }
    pregIndex = spdSearchElem(*indexes, fileId);
    spddb_freeRegionIndex(pregIndex);
    spdDelElem(*indexes, fileId);
}

SpdRegionIndex *spdGetPmodelRegIndex(SpdObjId fileId)
{
    SpdDBId dbId;
    SpdSet *indexes;
    SpdRegionIndex regIndex;
    SpdRegionIndex *pregIndex;
    SpdClassId relClassId;
    
    dbId = spdGetDefaultDB();
    indexes = getIndexesOfDB(dbId);
    if (indexes == NULL) {
        return NULL;
    }

    pregIndex = spdSearchElem(*indexes, fileId);
    if (pregIndex == NULL) {
        switch (spdGetClassId(fileId)) {
          case SPD_CID_FILE:
            relClassId = SPD_RID_IFILE_IBLOCK;
            break;
          case SPD_CID_PFILE:
            relClassId = SPD_RID_FILE_BLOCK;
            break;
          default:
            relClassId = SPD_NON_ID;
            break;
        }
        if (relClassId != SPD_NON_ID) {
            regIndex = spddb_createRegionIndex(fileId, relClassId);
            spdAddElem(*indexes, fileId, (Contents)&regIndex);
            pregIndex = spdSearchElem(*indexes, fileId);
        }
    }

    return pregIndex;
}
