/* 
 * Program:             $RCSfile: SDAmodel.c,v $  $Revision: 31.14 $
 * 
 * Purpose:
 *
 * Author:              T.Suzuki  1998/10/22
 *                      S.Yamamoto  1997/10/22
 *
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: SDAmodel.c,v 31.14 2003/04/07 16:02:51 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName varg_sort[] = {
    { SPD_VAL_NAME3(VARG_IN, "normal") },
    { SPD_VAL_NAME3(VARG_OUT, "control") },
    { 0, NULL, NULL }
};

static SpdValName next_stmt_sort[] = {
    { SPD_VAL_NAME3(NEXTSTMT_NORMAL, "normal") },
    /* { SPD_VAL_NAME3(NEXTSTMT_START, "start") }, */
    { SPD_VAL_NAME3(NEXTSTMT_CONTROL, "control") },
    { SPD_VAL_NAME3(NEXTSTMT_LOOPBACK, "loopback") },
    { SPD_VAL_NAME3(NEXTSTMT_DECL, "decl") },
    { SPD_VAL_NAME3(NEXTSTMT_END, "end") },
    { 0, NULL, NULL }
};

static SpdDBClass SDA_model_classes[] = {
    /* Classes. */
    {
        "sdafunc",
        SPD_CID_SDAFUNC,
        SAPIDDB_SORT_OBJ,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_INT },   /* dummy */
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "varg",
        SPD_CID_VARG,
        SAPIDDB_SORT_OBJ,
        {SAPIDDB_ATTR_ID, 0, SAPIDDB_ATTR_NONE },
        {
            {"sort", SPD_DBT_ENUM, varg_sort },
            { NULL, SPD_DBT_NON },
        }
    },

    /* Relationships. */
    {
        "next_stmt",
        SPD_RID_NEXT_STMT,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, 2, SAPIDDB_ATTR_NONE },
        {
            {"prev_stmt_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"next_stmt_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT},
            {"sort", SPD_DBT_ENUM, next_stmt_sort },
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "data_depend",
        SPD_RID_DATA_DEPEND,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"def_expr_id", SPD_DBT_OID, NULL,SPD_CID_EXPRESSION},
            {"ref_expr_id", SPD_DBT_OID, NULL,SPD_CID_EXPRESSION},
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "func_sfunc",
        SPD_RID_FUNC_SFUNC,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"func_id_", SPD_DBT_OID, NULL, SPD_CID_DECLARATION},
            {"sfunc_id", SPD_DBT_OID, NULL, SPD_CID_SDAFUNC},
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "sfunc_varg",
        SPD_RID_SFUNC_VARG,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"sfunc_id", SPD_DBT_OID, NULL,SPD_CID_SDAFUNC},
            {"varg_id",  SPD_DBT_OID, NULL,SPD_CID_VARG},
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "sfunc_stmt",
        SPD_RID_SFUNC_STMT,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"sfunc_id", SPD_DBT_OID, NULL,SPD_CID_SDAFUNC},
            {"stmt_id",  SPD_DBT_OID, NULL,SPD_CID_STATEMENT},
            { NULL, SPD_DBT_NON },
        }
    },

    {
        "ident_varg",
        SPD_RID_IDENT_VARG,
        SAPIDDB_SORT_REL_REFERENCE,
        { 0, 1, SAPIDDB_ATTR_NONE },
        {
            {"ident_id", SPD_DBT_OID, NULL,SPD_CID_IDENTIFIER},
            {"varg_id",  SPD_DBT_OID, NULL,SPD_CID_VARG},
            { NULL, SPD_DBT_NON },
        }
    },

    {
        NULL, SPD_NON_ID, 0, {SAPIDDB_ATTR_NONE}
    }

};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel SDA_model = {
    SPD_MID_SDA_MODEL,
    "SDA-model",
    SPD_SDA_MODEL_VERSION,
    SDA4_DIR,

    SPD_FIRST_CID_SDA_MODEL,
    SPD_LAST_CID_SDA_MODEL,
    SPD_CID_NUM_SDA_MODEL,

    SPD_FIRST_RID_SDA_MODEL,
    SPD_LAST_RID_SDA_MODEL,
    SPD_RID_NUM_SDA_MODEL,

    SPD_CRID_NUM_SDA_MODEL,

#if defined(SAPID_AR5)

    SDA_model_classes,
    NULL,

#endif /* defined(SAPID_AR5) */
};
