/* 
 * Program:             $RCSfile: StmtModel.c,v $  $Revision: 31.17 $
 * 
 * Purpose:
 *
 * Author:              T.Suzuki  1998/10/22
 *                      S.Yamamoto  1998/10/22
 *
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: StmtModel.c,v 31.17 2003/04/07 16:02:53 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SAPID_AR5)

static SpdValName statement_sort[] = {
    { SPD_VAL_NAME3(STATE_BLOCK, "block") },
    { SPD_VAL_NAME3(STATE_IF, "if") },
    { SPD_VAL_NAME3(STATE_FOR, "for") },
    { SPD_VAL_NAME3(STATE_DO, "do") },
    { SPD_VAL_NAME3(STATE_WHILE, "while") },
    { SPD_VAL_NAME3(STATE_SWITCH, "switch") },
    { SPD_VAL_NAME3(STATE_NORMAL, "normal") },
    { SPD_VAL_NAME3(STATE_BREAK, "break") },
    { SPD_VAL_NAME3(STATE_CONTINUE, "continue") },
    { SPD_VAL_NAME3(STATE_GOTO, "goto") },
    { SPD_VAL_NAME3(STATE_RETURN, "return") },
    { 0, NULL, NULL }
};

static SpdDBClass Stmt_model_classes[] = {
    /* Classes. */
    {
        "statement",
        SPD_CID_STATEMENT,
        SAPIDDB_SORT_OBJ_TREENODE | SAPIDDB_SORT_OBJ_TEXT,
        {SAPIDDB_ATTR_ID, 0, 1, SAPIDDB_ATTR_NONE},
        {
            {"length", SPD_DBT_INT, NULL, SPD_NON_ID, 0},
            {"sort", SPD_DBT_ENUM, statement_sort}, 
            {NULL, SPD_DBT_NON},
        }
    },

    /* Relationships. */
    {
        "for_expr",
        SPD_RID_FOR_EXPR,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"state_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"expr_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "cond_expr",
        SPD_RID_COND_EXPR,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"state_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"expr_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "iter_expr",
        SPD_RID_ITER_EXPR,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"state_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"expr_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "stmt_expr",
        SPD_RID_STMT_EXPR,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"state_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"expr_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "stmt_label",
        SPD_RID_STMT_LABEL,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"state_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"label_id", SPD_DBT_OID, NULL, SPD_CID_LABEL},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "return_expr",
        SPD_RID_RETURN_EXPR,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"return_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"expr_id", SPD_DBT_OID, NULL, SPD_CID_EXPRESSION},
            {"offset", SPD_DBT_INT},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "block_stmt",
        SPD_RID_BLOCK_STMT,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"block_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"state_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "block_decl",
        SPD_RID_BLOCK_DECL,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"block_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"decl_id", SPD_DBT_OID, NULL, SPD_CID_DECLARATION},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON}
        }
    },
    {
        "control_body",
        SPD_RID_CONTROL_BODY,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"ctrl_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"body_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "else_body",
        SPD_RID_ELSE_BODY,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, 2, SAPIDDB_ATTR_NONE},
        {
            {"ctrl_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT}, 
            {"body_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT},
            {"offset", SPD_DBT_INT}, 
            {NULL, SPD_DBT_NON},
        }
    },
    {
        "stmt_iblock",
        SPD_RID_STMT_IBLOCK,
        SAPIDDB_SORT_REL_REFERENCE,
        {0, 1, SAPIDDB_ATTR_NONE},
        {
            {"block_id", SPD_DBT_OID, NULL, SPD_CID_STATEMENT},
            {"iblock_id", SPD_DBT_OID, NULL,SPD_CID_BLOCK},
            {NULL, SPD_DBT_NON},
        }
    },
    {
        NULL, SPD_NON_ID, 0, {SAPIDDB_ATTR_NONE}
    }
};

#endif /* defined(SAPID_AR5) */

/******************************************************************************/

SpdDBModel Stmt_model = {
    SPD_MID_STMT_MODEL,
    "Stmt-model",
    SPD_STMT_MODEL_VERSION,
    SPD_STMT_MODEL_DIR,

    SPD_FIRST_CID_STMT_MODEL,
    SPD_LAST_CID_STMT_MODEL,
    SPD_CID_NUM_STMT_MODEL,

    SPD_FIRST_RID_STMT_MODEL,
    SPD_LAST_RID_STMT_MODEL,
    SPD_RID_NUM_STMT_MODEL,

    SPD_CRID_NUM_STMT_MODEL,

#if defined(SAPID_AR5)

    Stmt_model_classes,
    NULL,

#endif /* defined(SAPID_AR5) */
};
