/* 
 * Program:             $RCSfile: blkConv.c,v $  $Revision: 30.31 $
 * 
 * Purpose:             Converter ordinary SDB to statement SDB.
 *
 * Author:              T.Suzuki  1998/10/22
 *                      S.Yamamoto  1997/10/22
 *
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: blkConv.c,v 30.31 2003/07/08 05:48:26 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelAR.h>
#include <Sapid/ImodelARcompat.h>

#define createRel2(nrel, nobj, rel) \
        createRel(nrel, nobj, (rel).objectId, (rel).relationId)

#define SPD_DELETE_OBJ(id)  spdDeleteObjDw(id)

static SpdBoolean   del_obj_flag = FALSE;
static SpdBoolean   preserve_I_model = FALSE;
static SpdBoolean   not_write_db = FALSE;
static SpdBoolean   verbose_flag = FALSE;

static SpdObjId in_blk_op(SpdObjId blk, SpdObjId p_new_blk, SpdSet2 *labels);
static void blk_comp(SpdObjId comp, SpdObjId newblk);

static void spdDeleteObjDw(SpdId id);

/*
 * ֥ȤޤȤƺ뤿ΥХåե
 */

static SpdSet2 deleteObjs = NULL;

static void addDelObj(SpdId id)
{
    if (deleteObjs == NULL) {
        deleteObjs = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE);
    }
    spdAddElemOK(deleteObjs, id);
}

static void delObjs(void)
{
    SetCursor csr;

    if (deleteObjs != NULL) {
        for (csr = spdSetCursorCreate(deleteObjs); spdSetCursorNext(&csr);) {
            SPD_DELETE_OBJ(spdSetGetValueO(csr.curElem->value));
        }
    }
}

static void spdDeleteObjDw(SpdId id)
{
    spdMessage("spdDeleteObjDw(): \"%s\", %d.\n",
               spdGetClassNameById(spdGetClassId(id)), id);
    spdDeleteObj(id);
}

/******************************************************************************/

static SpdRelId createRel(SpdClassId rel_cls, SpdObjId pid, SpdObjId cid,
                          SpdObjId old_rel)
{
    SpdRelId    rel = spdCreateRel(rel_cls, SPD_NON_ID, pid, cid);
    
    if (old_rel != SPD_NON_ID) {
        int     offs = spdGetOffset(old_rel);

        spdSetAttrIntValByName(rel, "offset", offs);
        addDelObj(old_rel);
    }
    if (verbose_flag) {
        spdMessage("createRel(): %d %d %d %d.\n", pid, cid, old_rel, rel_cls);
    }
    return (rel);
}

static SpdRelId createStmtObj(SpdObjId pid, int sort)
{
    SpdObjId    new_obj = spdCreateObj(SPD_CID_STATEMENT, SPD_NON_ID);
    int         length = spdGetAttrIntValByName(pid, "length");

    spdSetAttrIntValByName(new_obj, "sort",   sort);
    spdSetAttrIntValByName(new_obj, "length", length);
    return (new_obj);
}

/*
 * ʸȤξＰȤδϢ롥
 */

void static create_cond_expr(SpdObjId blk, SpdObjId c_stmt, int newsort)
{
    /* blk_exprcondition_exprȤƴϢ롥 */
    if (newsort != STATE_FOR) {
        /* Condition expression. */
        SpdRel rel = spdGetAChildRel(blk, SPD_RID_BLK_EXPR);
        createRel2(SPD_RID_COND_EXPR, c_stmt, rel);
    } else {
        /* 'for' has 3 conditional expressions. */
        SpdSet2  rels = spdGetChildRelSet(blk, SPD_RID_BLK_EXPR);
        SpdRel  r1, r2, r3, tmp;
        SetCursor csr;
        
        if (rels->elemNum != 3) {
            spdErrorExit("%d: Should have just 3 items.\n", rels->elemNum);
        }
        csr = spdSetCursorCreate(rels);
        spdSetCursorNext(&csr);
        r3 = *(SpdRel *)spdSetGetValueP(csr.curElem->value);
        spdSetCursorNext(&csr);
        r2 = *(SpdRel *)spdSetGetValueP(csr.curElem->value);
        spdSetCursorNext(&csr);
        r1 = *(SpdRel *)spdSetGetValueP(csr.curElem->value);

        /* 3 Item Bubble Sort */
        if (spdGetOffset(r2.relationId) > spdGetOffset(r3.relationId)) {
            tmp = r2; r2 = r3; r3 = tmp;
        }
        if (spdGetOffset(r1.relationId) > spdGetOffset(r2.relationId)){
            tmp = r1; r1 = r2; r2 = tmp;
        }
        if (spdGetOffset(r2.relationId) > spdGetOffset(r3.relationId)) {
            tmp = r2; r2 = r3; r3 =tmp;
        }

        createRel2(SPD_RID_FOR_EXPR, c_stmt, r1);
        createRel2(SPD_RID_COND_EXPR, c_stmt, r2);
        createRel2(SPD_RID_ITER_EXPR, c_stmt, r3);
        spdFreeSet2(rels);
    }
}

/*
 * ＰΤȤδϢ, Τβ
 *
 * blk_blk: {: ʸ, : ֥å} δϢ
 * c_stmt:  ʸ֥
 * relid:   c_stmtȿΤȤδ֤Ϣ
 */

static void create_cond_body(SpdRel blk_blk, SpdObjId c_stmt,
                             SpdClassId rel_cls)
{
    SpdObjId    blk = blk_blk.objectId;
    int         blkSort = spdGetAttrIntValByName(blk, "sort");

    if (blkSort == BLOCK_COMP) {
        /* statement(sort=block)֥Ȥ */
        SpdObjId b_stmt = createStmtObj(blk, STATE_BLOCK);
        spdCreateRel(SPD_RID_STMT_IBLOCK, SPD_NON_ID, b_stmt, blk);
        /* Ϣ롥 */
        createRel(rel_cls, c_stmt, b_stmt, blk_blk.relationId);
        /* compoundʤ顤֥å */
        blk_comp(blk, b_stmt);              /* blockȤν */
    } else {
        /* ľ inner-blockξ */
        SpdObjId newid = in_blk_op(blk, SPD_NON_ID, NULL); /* c_stmt */
        if (newid != SPD_NON_ID) {
            createRel(rel_cls, c_stmt, newid, blk_blk.relationId);
        }
    }
}

/*
 * Control Statement
 */

static void cond_op(SpdObjId blk, SpdObjId c_stmt, int newsort)
{
    SpdSet2   rels;
    SpdRel   r1, r2 = {SPD_NON_ID, SPD_NON_ID}, tmp;
    SetCursor csr;

    /* creat relation between ctrl-statement and it's-condition */
    create_cond_expr(blk, c_stmt, newsort);
    
    /* Create control-body, else-body relation */
    rels = spdGetChildRelSet(blk, SPD_RID_BLK_BLK);
    csr = spdSetCursorCreate(rels);

    /* Analyze Inner Body */
    spdSetCursorNext(&csr);
    r1 = *(SpdRel *)spdSetGetValueP(csr.curElem->value);
    if (rels->elemNum > 1) {
        /* when having 2 child block (if block), sort by offset */
        spdSetCursorNext(&csr);
        r2 = *(SpdRel *)spdSetGetValueP(csr.curElem->value);
        if (spdGetOffset(r1.relationId) > spdGetOffset(r2.relationId)){
            tmp = r1; r1 = r2; r2 = tmp;
        }
        create_cond_body(r1, c_stmt, SPD_RID_CONTROL_BODY);
        create_cond_body(r2, c_stmt, SPD_RID_ELSE_BODY);
    } else {
        /* Only One Condition Block*/
        create_cond_body(r1, c_stmt, SPD_RID_CONTROL_BODY);
    }
    spdFreeSet2(rels);
}

/* sort: one of STATE_RETURN/STATE_BREAK/STATE_CONTINUE/STATE_GOTO. */

SpdObjId
mkJumpObj(SpdObjId expr)
{
    SpdObjId    newid;
    SpdRelId    new_rel;
    int         stat_sort, expr_sort;
    SpdRel      crel = {SPD_NON_ID, SPD_NON_ID};

    expr_sort = spdGetAttrIntValByName(expr, "sort");

    if (expr_sort == EXPR_BREAK) {
        stat_sort = STATE_BREAK;
    } else if (expr_sort == EXPR_CONTINUE) {
        stat_sort = STATE_CONTINUE;
    } else if (expr_sort == EXPR_GOTO) {
        stat_sort = STATE_GOTO;
        crel = spdGetAChildRel(expr, SPD_RID_GOTO_LABEL);
        new_rel = SPD_RID_GOTO_LABEL;
    } else if (expr_sort == EXPR_RETURN) {
        stat_sort = STATE_RETURN;
        crel = spdGetAChildRel(expr, SPD_RID_EXPR_EXPR);
        new_rel = SPD_RID_RETURN_EXPR;
    } else {
        spdErrorExit("Expression %d has invalid sort %d.\n", expr, expr_sort);
        return 1;
    }

    newid = createStmtObj(expr, stat_sort);
    
    /* Ϣĥ (goto label, return_expr)*/
    if (crel.objectId != SPD_NON_ID) createRel2(new_rel, newid, crel);

    /* б expression롥 */
    /* SPD_DELETE_OBJ(expr); */
    
    return newid;
}

/*
 * Top Expression --> Statement
 */

SpdObjId top_expr_op(SpdRel expr_rel)
{
    SpdObjId    newid;
    SpdObjId    expr = expr_rel.objectId;
    int         sort = spdGetAttrIntValByName(expr, "sort");

    if (sort == EXPR_BREAK || sort == EXPR_CONTINUE ||
        sort == EXPR_GOTO  || sort == EXPR_RETURN) {
        newid = mkJumpObj(expr);
    } else {
        newid = createStmtObj(expr, STATE_NORMAL);
        createRel2(SPD_RID_STMT_EXPR, newid, expr_rel);
    }
    return newid;
}

static SpdSet2
add_labels(SpdSet2 *labels, SpdObjId nobj)
{
    SetCursor csr;

    if (labels == NULL || *labels == NULL) {
        return NULL;
    }
    for (csr = spdSetCursorCreate(*labels); spdSetCursorNext(&csr);) {
        SpdRel      *rp = spdSetGetValueP(csr.curElem->value);
        SpdRelId    rel = createRel2(SPD_RID_STMT_LABEL, nobj, *rp);
        
        spdSetAttrIntValByName(rel, "offset", spdGetOffset(rp->relationId));
    }
    spdFreeSet2(*labels);
    *labels = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    return *labels;
}

/*
 * ֥åν
 *  blk:       I-model block
 *  p_new_blk: ҤȤľstmt(sort=block)
 *
 *  :  blockФ뿷stmt-obj
 */

static SpdObjId in_blk_op(SpdObjId blk, SpdObjId p_new_blk, SpdSet2 *labels)
{
    int     nobj;
    int     newsort;
    int     sort = spdGetAttrIntValByName(blk, "sort");
    
    switch(sort) {
      case BLOCK_BASIC:
        {
            int         m_offset = -1;
            SpdSet2     rels = spdGetChildRelSet(blk, SPD_RID_BLK_EXPR);
            SpdObjId    m_stmt = SPD_NON_ID; 
            SetCursor   csr;
        
            for (csr = spdSetCursorCreate(rels); spdSetCursorNext(&csr);) {
                SpdRel  *relp = spdSetGetValueP(csr.curElem->value);
            
                nobj = top_expr_op(*relp);
                if (nobj != SPD_NON_ID) {
                    if (p_new_blk != SPD_NON_ID) {
                        createRel(SPD_RID_BLOCK_STMT,
                                  p_new_blk, nobj, relp->relationId);
                    }
                    if (m_stmt == SPD_NON_ID ||
                        m_offset > spdGetOffset(relp->relationId)) {
                        /* ǽstmt˥٥ɲä뤿ν */
                        /* (basicʳ comp_blkǹԤʤ) */
                        m_stmt = nobj;
                        m_offset = spdGetOffset(relp->relationId);
                    }
                }
            }
        
            if (m_offset != -1) {
                add_labels(labels, m_stmt);
            }
            spdFreeSet2(rels);
        }
        return (p_new_blk == SPD_NON_ID) ? nobj : SPD_NON_ID ;
        break;
      case BLOCK_COMP:
        /* state (sort: block)֥Ȥ롥 */
        nobj = createStmtObj(blk, STATE_BLOCK);
        spdCreateRel(SPD_RID_STMT_IBLOCK, SPD_NON_ID, nobj, blk);
        blk_comp(blk, nobj);
        return nobj;
        break;
      case BLOCK_IF:
      case BLOCK_SWITCH:
      case BLOCK_FOR:
      case BLOCK_WHILE:
      case BLOCK_DO:
        /* ֥(STATEMENT_IF/SWITCH/..)롥 */
        if (sort == BLOCK_IF)          newsort = STATE_IF;
        else if (sort == BLOCK_FOR)    newsort = STATE_FOR;
        else if (sort == BLOCK_DO)     newsort = STATE_DO;
        else if (sort == BLOCK_WHILE)  newsort = STATE_WHILE;
        else if (sort == BLOCK_SWITCH) newsort = STATE_SWITCH;

        nobj = createStmtObj(blk, newsort);
        cond_op(blk, nobj, newsort);
        return nobj;
        break;
      default:
        spdErrorExit("Block %d has invalid sort %d.\n", blk, sort);
        break;
    }
    return SPD_NON_ID;
}

static int compar_relOffset(const SpdQueueElemType *a,
                            const SpdQueueElemType *b)
{
    SpdRelId    rid_a = (*(SpdRel**)a)->relationId;
    SpdRelId    rid_b = (*(SpdRel**)b)->relationId;
    
    return (spdGetOffset(rid_a) - spdGetOffset(rid_b));
}

/* 
 * static int compar_offset(const SpdQueueElemType *a,
 *                          const SpdQueueElemType *b)
 * {
 *     SpdRelId    rid_a = spdGetAParentRelId(*(SpdObjId*)a, SPD_NON_ID);
 *     SpdRelId    rid_b = spdGetAParentRelId(*(SpdObjId*)b, SPD_NON_ID);
 *     return (spdGetOffset(rid_a) - spdGetOffset(rid_b));
 * }
 */

/*
 * Compound Block Analysis
 * comp:   I-model Compound Block
 * newblk: it's new statement(sort=blk)
 *
 * : ʤ
 */

static void blk_comp(SpdObjId comp, SpdObjId newblk)
{
    int         i;
    SpdSet2     ids = spdGetChildRelSet(comp, SPD_NON_ID);
    SpdQueue    q = spdQueueInit();
    SpdSet2     labels = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    SetCursor   csr;
    labels->valueFree = spdElemFreeSpdRel;
    labels->valueCopy = spdElemCopySpdRel;

    for (csr = spdSetCursorCreate(ids); spdSetCursorNext(&csr);) {
        _spdQueuePushFront(&q,
                           (_spd_queue_dummy.p = spdSetGetValueP(csr.curElem->value),
                            _spd_queue_dummy));
    }

    spdQueueSort(&q, compar_relOffset);

    for (i = 0; i < q.elem; i++) {
        SpdRel      nrel = *(SpdRel*)spdQueuePeek(&q, i);
        SpdObjId    nobj;
        
        if (verbose_flag) {
            spdMessage("blk_comp(): rel %d offs %d.\n",
                       nrel.objectId, spdGetOffset(nrel.relationId));
        }
        
        switch (spdGetClassId(nrel.objectId)) {
          case SPD_CID_BLOCK:
            nobj = in_blk_op(nrel.objectId, newblk, &labels);
            if (nobj != SPD_NON_ID) {
                createRel(SPD_RID_BLOCK_STMT, newblk, nobj, nrel.relationId);
                add_labels(&labels, nobj);
            }
            break;
          case SPD_CID_EXPRESSION:    /* Not reach. */
            spdWarning("Unexpected expression %d.\n", nrel.objectId);
            break;
          case SPD_CID_DECLARATION:
            createRel2(SPD_RID_BLOCK_DECL, newblk, nrel);
            break;
          case SPD_CID_LABEL:
            /* Not making blk_label here. state->label */
            {
                SpdRel *nrel_p = (SpdRel *)spdMalloc(sizeof(SpdRel));
                *nrel_p = nrel;
                spdAddElemOP(labels, nrel.objectId, nrel_p);
            }
            break;
          case SPD_CID_STATEMENT:
            break;
          default:
            spdErrorExit("blk_comp(): Invalid class \"%s\" %d.\n",
                         spdGetClassNameById(spdGetClassId(nrel.objectId)),
                         nrel.objectId);
            break;
        }
    }
    
    spdQueueClean(&q);
    spdFreeSet2(ids);
    spdFreeSet2(labels);
}

void conv(void)
{
    SpdSet2  decls = spdGetClassExtentP(SPD_CID_DECLARATION); /* no need to free */
    SetCursor csr;
    
    for (csr = spdSetCursorCreate(decls); spdSetCursorNext(&csr);) {
        SpdIdSet    decrs;
        SetCursor   csr_d;
        SpdObjId    decl = spdSetGetValueO(csr.curElem->value);
    
        if (spdGetAttrIntValByName(decl, "sort") != DECL_FUNCDEF) {
            continue;
        }
        decrs = spdGetChildObjIdSet(decl, SPD_RID_DECL_DECL);
    
        for (csr_d = spdSetCursorCreate(decrs); spdSetCursorNext(&csr_d);) {
            SpdObjId    decr = spdSetGetValueO(csr_d.curElem->value);
            SpdIdSet    opts = spdGetChildObjIdSet(decr, SPD_RID_DECL_OPT);
            SetCursor   csr_o;
        
            for (csr_o = spdSetCursorCreate(opts); spdSetCursorNext(&csr_o);) {
                SpdObjId    opt = spdSetGetValueO(csr_o.curElem->value);
                SpdRel      rel = spdGetAChildRel(opt, SPD_RID_FUNC_BODY);
        
                if (rel.objectId != SPD_NON_ID) {
                    SpdObjId newid = createStmtObj(rel.objectId, STATE_BLOCK);
                    spdCreateRel(SPD_RID_STMT_IBLOCK,
                                 SPD_NON_ID, newid, rel.objectId);
                    blk_comp(rel.objectId, newid);
                    createRel(SPD_RID_FUNC_BODY, decl, newid, rel.relationId);
                }
            }
            spdFreeSet2(opts);
        }
        spdFreeSet2(decrs);
    }

    if (del_obj_flag) {
        delObjs();
    }
}

/*
 * func_bodyI-MODELȶͭˤʤäƤΤǡ˺롥
 */

void clean_relation(void)
{
    SpdSet2  func_body = spdGetClassExtentP(SPD_RID_FUNC_BODY); /* no need to free */
    SetCursor csr;

    for (csr = spdSetCursorCreate(func_body); spdSetCursorNext(&csr);) {
        SpdRelId    rid = spdSetGetValueO(csr.curElem->value);
        SpdObjId    pid = spdGetAttrIntVal(rid, SAPIDDB_ATTR_P_ID);
        
        if (spdGetClassId(pid) == SPD_CID_DECLARATION) {
            SPD_DELETE_OBJ(rid);
        }
    }
}

void opt_analyze(int argc, char *argv[])
{
    char c;

    while ((c = getopt(argc, argv, "d:inrv")) != EOF) {
        switch (c) {
          case 'd':             /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'i':             /* Preserve I-model. */
            preserve_I_model = TRUE;
            break;
          case 'n':             /* Not write db (not used yet). */
            not_write_db = TRUE;
            break;
          case 'r':             /* Remove objects. */
            del_obj_flag = TRUE;
            break;
          case 'v':
            verbose_flag = TRUE;
            break;
          default:
            spdErrorExit("'%c': Invalid option.\n", c);
            break;
        }
    }
}

int main(int argc, char *argv[])
{
    int     md1, md2, md3;
    int     db;

    spdInitialize(spdBasename(argv[0]));
    opt_analyze(argc, argv);

    md1 = spdUseModel(&I_model);
    md2 = spdUseModel(&Stmt_model);
    md3 = spdUseModel(&TypeView_model);
    db = spdOpenSDB();
    spdReadDB(md1, db);
    spdReadDB(md3, db);
    spdCreateDB(md2, db);

    clean_relation();
    conv();

    if (!preserve_I_model) {
        spdCommitDB(md1, db);
    }
    spdCommitDB(md2, db);

    return EXIT_SUCCESS;
}
