/* 
 * Program:             $RCSfile: mkNext.c,v $  $Revision: 30.32 $
 * 
 * Purpose:             Make Control-flow.
 *
 * Author:              T.Suzuki  1998/10/22
 *                      S.Yamamoto  1997/10/22
 *
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkNext.c,v 30.32 2003/07/08 05:48:26 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelARcompat.h>
#include <Sapid/StmtModelAR.h>

/* Global Flags */
static int not_write_db = 0;   /* 1:not write db */
static int verbose_flag = 0;   /* 1:verbose-output-mode */
static int block_trv_mode = 1; /* 1:block mkNextǤɤ */

static int
compar_relOffset(const SpdQueueElemType *a, const SpdQueueElemType *b)
{
    SpdRelId rid_a = (*(SpdRel**)a)->relationId;
    SpdRelId rid_b = (*(SpdRel**)b)->relationId;
    return spdGetOffset(rid_a) - spdGetOffset(rid_b);
}

/*
 * Ϳ줿block stmt offset Ȥ(queue)֤
 * ͤϡSpdRel({objectId, relationId})ؤΥݥ
 */

static SpdQueue
getSortedStmts_common(SpdObjId blk, SpdClassId get_rel)
{
    /* Ҷ(stmt)򤹤٤ */
    SpdSet2 stmts = spdGetChildRelSet(blk, SPD_RID_BLOCK_STMT);
    SpdSet2 decls = spdGetChildRelSet(blk, SPD_RID_BLOCK_DECL);
    /* offset   */
    SpdQueue q  = spdQueueInit();
    SpdQueue q2 = spdQueueInit();
    SetCursor csr;

    for (csr = spdSetCursorCreate(stmts); spdSetCursorNext(&csr);) {
        _spd_queue_dummy.p = spdSetGetValueP(csr.curElem->value);
        _spdQueuePushFront(&q, _spd_queue_dummy);
        spdSetSetValueP(&csr.curElem->value, NULL);
    }
    for (csr = spdSetCursorCreate(decls); spdSetCursorNext(&csr);) {
        _spd_queue_dummy.p = spdSetGetValueP(csr.curElem->value);
        _spdQueuePushFront(&q, _spd_queue_dummy);
        spdSetSetValueP(&csr.curElem->value, NULL);
    }

    spdQueueSort(&q, compar_relOffset);

    while(q.elem>0)
        spdQueuePushFront(&q2, ((SpdRel*)_spdQueuePopBack(&q).p)->objectId);

    spdQueueClean(&q);
    spdFreeSet2(stmts);
    spdFreeSet2(decls);
    return q2;
}

static SpdQueue
getSortedStmts(SpdObjId blk)
{
    return getSortedStmts_common(blk, SPD_RID_BLOCK_STMT);
}

#if 0

/* ɻȤʤƤɤ */
static SpdQueue
getSortedDecls(SpdObjId blk, SpdClassId get_rel)
{
    return getSortedStmts_common(blk, SPD_RID_BLOCK_DECL);
}

#endif

/*
 * `'ʸ
 * Type A
 *  * ľ夬 blockʳstmt -->  stmt
 *  * ľ夬 blockstmt --> blockΡǽ blockʳstmt
 *  * ľ夬 ʤ --> Υ֥åμstmt, blk_stmt
 *                     if, switch ΤȤʸ
 *                     ifʳ loop  ʬ
 * Type B
 *  * ľ夬 blockޤblockʳstmt -->  stmt
 *  * ľ夬ʤ --> Υ֥åμ(Ʊ)
 */

static SpdObjId
getNextStmt(SpdObjId stmt, int *loopback_flag)
{
    int idx;
    SpdObjId nid;
    SpdQueue q;

    /* (block) block */
    SpdObjId pid = spdGetAParentObjId(stmt, SPD_RID_BLOCK_STMT);
    if (pid == SPD_NON_ID) /* declξ? */
        pid = spdGetAParentObjId(stmt, SPD_RID_BLOCK_DECL);

    /* cntrl-blkξ -- Ϥ block -- */
    if (pid == SPD_NON_ID) {

        pid = spdGetAParentObjId(stmt, SPD_RID_ELSE_BODY);
        if (pid != SPD_NON_ID) /* ifξ硤Ƥμʸ */
            return getNextStmt(pid, loopback_flag);
        else {
            pid = spdGetAParentObjId(stmt, SPD_RID_CONTROL_BODY);
            if (pid != SPD_NON_ID) {
                int sort = spdGetAttrIntValByName(pid, "sort");
                /* Ƥ loopξϿƼȤ stmt֤ */
                if ( sort != STATE_IF && sort != STATE_SWITCH) {
                    *loopback_flag = 1;
                    return pid;
                }
            
                /* Ƥ loopǤʤ(if, swtich)ΤȤϤμʸ */
                return getNextStmt(pid, loopback_flag);
            }
        }
        return SPD_NON_ID; /* ξϴؿ */
    }

    if (block_trv_mode &&
        spdGetClassId(stmt) == SPD_CID_STATEMENT &&
        spdGetAttrIntValByName(stmt, "sort") == STATE_BLOCK) {
        /* stmtstmtoffsetǥ */
        q = getSortedStmts(stmt);
        idx = -1;
    } else {
        /* pidƱ٥stmtoffsetǥ */
        q = getSortedStmts(pid);
        /* Ϥ stmt idx */
        idx = spdQueueSearchIdx(&q, stmt);
        if (idx == -1) {
            /* stmtĤʤ */
            spdQueueClean(&q);
            return SPD_NON_ID;
        }
    }

    /* ľʸ */
    nid = spdQueuePeek(&q, idx+1);

    while (1) {
        /* declξ */
        if (spdGetClassId(nid) == SPD_CID_DECLARATION)
            break;
    
        if (!block_trv_mode) {
            /* block-stmt mknextǤɤʤ褦ˤ */
            if (spdGetAttrIntValByName(nid, "sort") == STATE_BLOCK) {
                /* sub-blockξ */
                /* κǽʸ */
                SpdQueue q = getSortedStmts(nid);
                nid = spdQueuePeek(&q, 0);
                spdQueueClean(&q);
            }
        }
        
        /* ʸνξ -- ҤȤľblockμʸ */
        if (nid == SPD_NON_ID) {
            /* block-stmt ¤ꡤƵ */
            int tmp_block_trv = block_trv_mode; 

            block_trv_mode = 0;/* offˤʤ */
            nid = getNextStmt(pid, loopback_flag);
            block_trv_mode = tmp_block_trv;
        }
        break;
    }

    spdQueueClean(&q);
    
    return nid;
}

/*
 * (block)ǽμ¹ʸ
 * ʤ (¿ʬξ)Υ֥åμʸõ
 */

static SpdObjId
getFirstStmt(SpdObjId stmt, int *loopback_flag)
{
    SpdObjId tmpstmt;

    /* ϤΥå */
    if (spdGetAttrIntValByName(stmt, "sort") != STATE_BLOCK) {
        spdWarning("spdGetFirstStmt():required sort=BLOCK object.\n");
        return SPD_NON_ID;
    }

    tmpstmt = stmt;
    while (tmpstmt != SPD_NON_ID) {
        SpdQueue top = getSortedStmts(tmpstmt);
        tmpstmt = spdQueuePeek(&top, 0);
        spdQueueClean(&top);

        if (spdGetAttrIntValByName(tmpstmt, "sort") != STATE_BLOCK)
            return tmpstmt;
    }

    return getNextStmt(stmt, loopback_flag);
}

static SpdObjId
getUpperBlk(SpdObjId stmt, int in_sort)
{
    while ((stmt = sdaGetParentCtrlStmt(stmt)) != SPD_NON_ID) {
        int     sort = spdGetAttrIntValByName(stmt, "sort");

        if (in_sort == STATE_CONTINUE) {
            if (sort == STATE_FOR || sort == STATE_WHILE || sort == STATE_DO) {
                break;
            }
            continue;
        } else if (in_sort == STATE_BREAK) {
            if (sort == STATE_FOR || sort == STATE_WHILE ||
                sort == STATE_DO || sort == STATE_SWITCH) {
                break;
            }
        }
    }
    return stmt;
}

SpdObjId fin_body = SPD_NON_ID;

static int
add_next_rel(SpdObjId pid, SpdObjId cid, int nextSort)
{
    int rid;
    if (pid == SPD_NON_ID) return 0; /* ؿ ? */

    if (cid == SPD_NON_ID) {   /* ¿ʬؿȴ */
        cid = fin_body;
    }

    if (cid == fin_body) {
        /* ؿȴϢsortNEXTSTMT_ENDˤ */
        nextSort = NEXTSTMT_END;
    }

    if (verbose_flag) {
        printf("%d -> %d(%d)\n", pid, cid, nextSort);
    }

    rid = spdCreateRel(SPD_RID_NEXT_STMT, SPD_NON_ID, pid, cid);
    spdSetAttrIntValByName(rid, "sort", nextSort);
    return 1;
}

/* switchʸstmt(block)˴ޤޤ label */

static void
get_labels(SpdObjId stmt, SpdSet2 lbls)
{
    SpdIdSet objs = spdGetChildRelSet(stmt, SPD_NON_ID);
    SetCursor csr;

    for (csr = spdSetCursorCreate(objs); spdSetCursorNext(&csr);) {
        SpdRel *relp = spdSetGetValueP(csr.curElem->value);
        
        switch(spdGetClassId(relp->relationId)) {
        case SPD_RID_CONTROL_BODY:
        case SPD_RID_ELSE_BODY:
        case SPD_RID_BLOCK_STMT:
            if (spdGetAttrIntValByName(relp->objectId, "sort") != STATE_SWITCH)
                get_labels(relp->objectId, lbls);
            break;

        case SPD_RID_STMT_LABEL:
            if (spdGetAttrIntValByName(relp->objectId, "sort") != LABEL_NAMED)
                spdAddElemOK(lbls, relp->objectId);
            break;
        }
    }
    spdFreeSet2(objs);
    return;
}

static SpdObjId
mk_control_body(SpdObjId cur_stmt, SpdRelId rel)
{
    SpdObjId nxt_stmt = spdGetAChildObjId(cur_stmt, rel);

    if (spdGetAttrIntValByName(nxt_stmt, "sort") == STATE_BLOCK) {
        SpdQueue nxt_stmts = getSortedStmts(nxt_stmt);
        nxt_stmt = spdQueuePeek(&nxt_stmts, 0);
        spdQueueClean(&nxt_stmts);
    }
    add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_CONTROL);

    return nxt_stmt;
}

/*
 * next-relation 
 */

static void
mk_nextRel(SpdObjId decl, SpdObjId body)
{
    int loopback_flag = 0;
    SpdSet2 checkd_stmt /* ̤äʸϿ*/
        = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_KEY_IS_VALUE); 
    SpdQueue q = spdQueueInit();

    /* ǽΥ֥å stmt */
    SpdQueue top = getSortedStmts(body);
    if (top.elem > 0)
        spdQueueEnqueue(&q, spdQueuePeek(&top, 0));
    spdQueueClean(&top);

    /* ϴؿФơsdaΥ֥Ȥ ǽʸؤδϢ */
    {
        SpdObjId    strt_body = SPD_NON_ID;

        strt_body = fin_body = spdCreateObj(SPD_CID_SDAFUNC, SPD_NON_ID);
        spdCreateRel(SPD_RID_FUNC_SFUNC, SPD_NON_ID, decl, strt_body);

        if (q.elem > 0 && strt_body != SPD_NON_ID) {
            SpdObjId frst_stmt = spdQueuePeek(&q, 0);
            if (verbose_flag) {
                printf("%d -> %d(FIRST)\n", strt_body, frst_stmt);
            }
            spdCreateRel(SPD_RID_SFUNC_STMT,
                         SPD_NON_ID, strt_body, frst_stmt);
        }
        /* add_next_rel(strt_body, spdQueuePeek(&q, 0), NEXTSTMT_START); */
    }

    while (q.elem > 0) {

        int sort;
        SpdObjId cur_stmt = spdQueueDequeue(&q); /* Υ롼פǤβо */
        SpdObjId nxt_stmt = SPD_NON_ID;
        SpdQueue nxt_stmts= spdQueueInit();

        if (cur_stmt == SPD_NON_ID) continue;

        /* ˲ϺѤ */
        if (spdContainsO(checkd_stmt, cur_stmt))
            continue;

        spdAddElemOK(checkd_stmt, cur_stmt);

        sort = spdGetAttrIntValByName(cur_stmt, "sort");
        
        if (spdGetClassId(cur_stmt) == SPD_CID_DECLARATION) {
            /* ξ*/
            loopback_flag = 0;
            nxt_stmt = getNextStmt(cur_stmt, &loopback_flag);

            /* Ϣĥ (cur --> nxtĥ) */
            add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_DECL);
            spdQueueEnqueue(&q, nxt_stmt);
            continue;
        }

        switch (sort) {
          case STATE_BLOCK:
            if (!block_trv_mode) {
                spdErrorExit("%d: Invalid block statement.\n", cur_stmt);
            }
            nxt_stmt = getNextStmt(cur_stmt, &loopback_flag);

            /* Ϣĥ (cur --> nxtĥ) */
            if (loopback_flag == 0) {
                add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_NORMAL);
            } else {
                /* ¿ʬelseˤ̵Ϥ */
                add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_LOOPBACK);
            }
            
            spdQueueEnqueue(&q, nxt_stmt);
            break;
            /* top = getSortedStmts(cur_stmt); */
            /* spdQueueEnqueue(&q, spdQueuePeek(&top, 0)); */
            /* spdQueueClean(&top); */
            /* break; */
        
          case STATE_IF:
            nxt_stmt = mk_control_body(cur_stmt, SPD_RID_CONTROL_BODY);
            spdQueueEnqueue(&q, nxt_stmt);
            nxt_stmt = spdGetAChildObjId(cur_stmt, SPD_RID_ELSE_BODY);
            if (nxt_stmt != SPD_NON_ID) {
                if (spdGetAttrIntValByName(nxt_stmt, "sort") == STATE_BLOCK) {
                    nxt_stmts= getSortedStmts(nxt_stmt);
                    nxt_stmt = spdQueuePeek(&nxt_stmts, 0);
                    spdQueueEnqueue(&q, nxt_stmt);
                    spdQueueClean(&nxt_stmts);
                }
                add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_CONTROL);
            } else {
                goto next_stmt;
            }
            break;

          case STATE_SWITCH:
            {
                /* stmt  labelȤäơsortcaseΤΤ  */
                /* label --> labeld-stmt get, δ֤˴Ϣĥ */
                int default_case_flag = 0;
                SetCursor csr;

                SpdIdSet case_lbls = spdCreateSet2(SPDSET_OBJID_KEY,
                                                   SPDSET_KEY_IS_VALUE);
                /* = spdSearchChildObj(cur_stmt, SPD_CID_LABEL);*/
                get_labels(cur_stmt, case_lbls);
        
                for (csr = spdSetCursorCreate(case_lbls);
                     spdSetCursorNext(&csr);) {
                    SpdObjId lbl_stmt;
                    SpdObjId lbl = spdSetGetValueO(csr.curElem->value);
                    int lsort = spdGetAttrIntValByName(lbl, "sort");

                    if (lsort == LABEL_NAMED) continue;

                    lbl_stmt=spdGetAParentObjId(lbl, SPD_RID_STMT_LABEL);

                    /* Note: labelnext_stmtιǡ */
                    add_next_rel(cur_stmt, lbl, NEXTSTMT_CONTROL);
                    /* ݤʤΤǡblockΤȤʬ(¿ʬ) */
                    if (spdGetAttrIntValByName(lbl_stmt, "sort") == STATE_BLOCK) {
                        lbl_stmt = getFirstStmt(lbl_stmt, &loopback_flag);
                        add_next_rel(cur_stmt, lbl_stmt, NEXTSTMT_NORMAL);
                    } else {
                        add_next_rel(lbl, lbl_stmt, NEXTSTMT_NORMAL);
                    }

                    spdQueueEnqueue(&q, lbl_stmt);

                    if (lsort == LABEL_DEFAULT)  default_case_flag = 1;
                }
            
                spdFreeSet2(case_lbls);
                if (default_case_flag == 0)
                    goto next_stmt;         /* defaultʤȤ */
            }
            break;

          case STATE_FOR:
          case STATE_DO:
          case STATE_WHILE:
            /* block */
            nxt_stmt = mk_control_body(cur_stmt, SPD_RID_CONTROL_BODY);

            /* 
             * nxt_stmt = spdGetAChildObjId(cur_stmt, SPD_RID_CONTROL_BODY);
             * if (spdGetAttrIntValByName(nxt_stmt, "sort") == STATE_BLOCK) {
             *     nxt_stmts= getSortedStmts(nxt_stmt);
             *     nxt_stmt = spdQueuePeek(&nxt_stmts, 0);
             *     spdQueueClean(&nxt_stmts);
             * }
             * add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_CONTROL);
             */

            spdQueueEnqueue(&q, nxt_stmt);
            /* ʸ */

          case STATE_NORMAL:
          next_stmt:
            loopback_flag = 0;
            nxt_stmt = getNextStmt(cur_stmt, &loopback_flag);

            /* Ϣĥ (cur --> nxtĥ) */
            if (loopback_flag == 0)
                add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_NORMAL);
            else
                add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_LOOPBACK);
            spdQueueEnqueue(&q, nxt_stmt);
            break;

          case STATE_GOTO:
            nxt_stmt = spdGetAChildObjId(cur_stmt, SPD_RID_GOTO_LABEL);
            nxt_stmt = spdGetAParentObjId(nxt_stmt, SPD_RID_STMT_LABEL);
            add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_NORMAL);
            spdQueueEnqueue(&q, nxt_stmt);
            break;

          case STATE_RETURN:
            /* fin_bodyȤδ֤˴Ϣĥ */
            add_next_rel(cur_stmt, fin_body, NEXTSTMT_NORMAL);
            break;

          case STATE_CONTINUE:
            nxt_stmt = getUpperBlk(cur_stmt, STATE_CONTINUE);
            add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_LOOPBACK);
            spdQueueEnqueue(&q, nxt_stmt);
            break;

          case STATE_BREAK:
            nxt_stmt = getUpperBlk(cur_stmt, STATE_BREAK);
            nxt_stmt = getNextStmt(nxt_stmt, &loopback_flag);
            add_next_rel(cur_stmt, nxt_stmt, NEXTSTMT_NORMAL);
            spdQueueEnqueue(&q, nxt_stmt);
            break;

          default:
            spdErrorExit("Statemet %d has invalid sort %d.\n", cur_stmt, sort);
            break;
        }
    }

    spdQueueClean(&q);
    spdFreeSet2(checkd_stmt);
    return;
}

void
mkNextRel(void)
{
    SpdIdSet files = spdGetClassExtentP(SPD_CID_FILE); /* no need to free */
    SetCursor csr;

    for (csr = spdSetCursorCreate(files); spdSetCursorNext(&csr);) {
        SetCursor csr_d;
        SpdIdSet decls = spdGetChildObjIdSet(
            spdSetGetValueO(csr.curElem->value), SPD_RID_FILE_DECL);

        for (csr_d = spdSetCursorCreate(decls); spdSetCursorNext(&csr);) {
            /* for each func */
            SpdObjId decl = spdSetGetValueO(csr.curElem->value);
            SpdObjId body = spdGetAChildObjId(decl, SPD_RID_FUNC_BODY);
        
            if (body != SPD_NON_ID) mk_nextRel(decl, body);
        }
        
        spdFreeSet2(decls);
    }
}

void
opt_analyze(int argc, char *argv[])
{
    char c;

    while ((c = getopt(argc, argv, "d:nbv")) != EOF) {
        switch (c) {
          case 'd':             /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'b':
            block_trv_mode ^= 1;
            break;
          case 'n': 
            not_write_db ^= 1;
            break;
          case 'v':
            verbose_flag ^= 1;
            break;
          default:
            spdErrorExit("'%c': Invalid option.\n", c);
            break;
        }
    }
}

int
main(int argc, char *argv[])
{
    int     md1, md2, md3, md4;
    int     db;

    spdInitialize(spdBasename(argv[0]));
    opt_analyze(argc, argv);

    md1 = spdUseModel(&I_model);
    md2 = spdUseModel(&Stmt_model);
    md3 = spdUseModel(&SDA_model);
    md4 = spdUseModel(&TypeView_model);
    db = spdOpenSDB();

    spdReadDB(md1, db);
    spdReadDB(md2, db);
    spdReadDB(md4, db);
    spdCreateDB(md3, db);

    mkNextRel();

    if (!not_write_db) {
        spdCommitDB(md3, db);
    }
    
    return (EXIT_SUCCESS);
}

/******************************************************************************/

/* 
 * void static mk_next_decl(SpdObjId pid, SpdObjId frst_stmt)
 * {
 *     int         i;
 *     SpObjId     cid;
 *     SpdQueue    decls = spdGetSortedDecls(pid);
 * 
 *     for (i=0; i<decls->elem; i++) {
 *         cid = *(SpdObjId *)spdQueuePeek(&decls, i);
 *         if (verbose_flag) {
 *             printf("%d -> %d(decl)\n", pid, cid);
 *         }
 *         rid = spdCreateRel(SAPID_REL_D_NEXT_DECL, SPD_NON_ID, pid, cid);
 *         pid = cid;
 *     }
 *     rid = spdCreateRel(SAPID_REL_D_NEXT_DECl, SPD_NON_ID, pid, cid);
 *     return (1);
 * }
 */
