/* 
 * Program:             $RCSfile: statinfo.c,v $  $Revision: 30.11 $
 * 
 * Purpose:
 *
 * Author:              T.Suzuki  1998/10/22
 *                      S.Yamamoto  1997/10/22
 *
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: statinfo.c,v 30.11 2003/07/08 05:48:26 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "statinfo.h"

extern int verbose_flag;

/* ---------------------------------------- */
/*          ¹Ծ֤εϿݻ              */
/* ---------------------------------------- */

SdaStatInfo*
info_init(void)
{
    
    SdaStatInfo *n_inf = spdMalloc(sizeof(SdaStatInfo));
    n_inf->lpq = spdMalloc(sizeof(SpdQueue));
    n_inf->lpqc= spdMalloc(sizeof(SpdQueue));

    n_inf->stmt.objectId   = SPD_NON_ID;
    n_inf->stmt.relationId = SPD_NON_ID;
    n_inf->rdmp = spdCreateSet2(SPDSET_OBJID_KEY, SPDSET_POINTER_VALUE);
    *n_inf->lpq  = spdQueueInit();              /* 10 */
    *n_inf->lpqc = spdQueueInit();              /* 10 */

    return n_inf;
}

void
info_free(SdaStatInfo *inf)
{
    SetCursor csr;

    spdQueueClean(inf->lpq);
    spdQueueClean(inf->lpqc);
    spdFree(inf->lpq);
    spdFree(inf->lpqc);

    for (csr = spdSetCursorCreate(inf->rdmp); spdSetCursorNext(&csr);) {
        SpdQueue *q = spdSetGetValueP(csr.curElem->value);
        spdQueueClean(q);
        spdFree(q);
    }
    spdFreeSet2(inf->rdmp);
    inf->lpq = NULL;
    inf->lpqc= NULL;
    inf->rdmp= NULL;

    spdFree(inf);
}

void
info_loop_add(SdaStatInfo *inf, SpdObjId lp_stmt)
{
    spdQueueEnqueue(inf->lpq, lp_stmt);
    spdQueueEnqueue(inf->lpqc, 0);
}

SpdObjId
info_loop_top(const SdaStatInfo *inf)
{
    return spdQueueGetBack(inf->lpq);
}

int
info_loopcnt_get(const SdaStatInfo *inf, SpdObjId lp_stmt)
{
    int idx;
    if (lp_stmt == SPD_NON_ID) {
        lp_stmt = info_loop_top(inf);
    }

    idx = spdQueueSearchIdx(inf->lpq, lp_stmt);
    return spdQueueGetBack(inf->lpqc);
}

int
info_loopcnt_inc(SdaStatInfo *inf)
{
    int cnt = spdQueueDequeue(inf->lpqc);
    ++cnt;
    spdQueueEnqueue(inf->lpqc, cnt);
    return cnt;
}

SpdObjId
info_loop_pop(SdaStatInfo *inf, SpdObjId blk)
{
    if (blk != SPD_NON_ID && spdQueueGetBack(inf->lpq) != blk) {
        spdWarning("Mismatch Queue?(%d:%d)\n", spdQueueGetBack(inf->lpq), blk);
    }

    (void)spdQueueDequeue(inf->lpqc);
    return spdQueueDequeue(inf->lpq);
}

/*
 * ref-def Ϣɲ
 */
/* ñ def */

void
info_add_r_d(SdaStatInfo *inf, SpdObjId ref, SpdObjId def)
{
    int idx;
    SpdQueue *q = spdSetGetValueP(spdSearchElemO(inf->rdmp, ref)->value);

    if (q == NULL) {
        q = (SpdQueue *)spdMalloc(sizeof(SpdQueue));
        *q = spdQueueInit();
        spdAddElemOP(inf->rdmp, ref, q);
    }

    idx = spdQueueSearchIdx(q, def);

    if (idx < 0)
        spdQueueEnqueue(q, def);
}

/*  def */

void
info_add_r_ds(SdaStatInfo *inf, SpdObjId ref, SpdIdSet defs)
{
    SetCursor csr;

    SpdQueue *q = spdSetGetValueP(spdSearchElemO(inf->rdmp, ref)->value);

    if (q == NULL) {
        q = (SpdQueue *)spdMalloc(sizeof(SpdQueue));
        *q = spdQueueInit();
        spdAddElemOP(inf->rdmp, ref, q);
    }

    for (csr = spdSetCursorCreate(defs); spdSetCursorNext(&csr);) {
        SpdObjId def = spdSetGetValueO(csr.curElem->value);

        int idx = spdQueueSearchIdx(q, def);
        if (idx < 0) {
            spdQueueEnqueue(q, def);
        }
    }
    return;
}

SpdQueue*
info_get_def(const SdaStatInfo *inf, SpdObjId stmt)
{
    return (SpdQueue*)spdSetGetValueP(spdSearchElemO(inf->rdmp, stmt)->value);
}

void
info_del_def(SdaStatInfo *inf, SpdObjId def)
{
    SpdQueue *q = spdSetGetValueP(spdSearchElemO(inf->rdmp, def)->value);
    if (q != NULL) spdQueueClean(q);
    spdDelElemO(inf->rdmp, def);
}

SdaStatInfo*
info_dup(const SdaStatInfo *inf)
{
    int i, j;
    SdaStatInfo *n_inf = info_init();
    SetCursor csr;

    n_inf->stmt = inf->stmt;

    /* rd_mp  ԡ [ref - def] */
    for (csr = spdSetCursorCreate(inf->rdmp); spdSetCursorNext(&csr);) {
        SpdObjId ref = spdSetGetValueO(csr.curElem->key);
        SpdQueue *q = spdSetGetValueP(csr.curElem->value);

        for (j = 0; j < q->elem; j++) {
            info_add_r_d(n_inf, ref, spdQueuePeek(q, j));
        }
    }

    /* lp_q  ԡ -- [stmt -> counter] */
    for (i = 0; i < inf->lpq->elem; i++) {
        SpdObjId id = spdQueuePeek(inf->lpq, i);
        int     cnt = spdQueuePeek(inf->lpqc, i);
    
        spdQueuePushBack(n_inf->lpq, id);
        spdQueuePushBack(n_inf->lpqc, cnt);
    }

    return n_inf;
}

/*
 * Ƥɽ
 */

void
info_view_rdmp(SdaStatInfo *info)
{
    int j;
    SetCursor csr;

    if (verbose_flag) {
        printf("-rdmp-\n");
    }

    for (csr = spdSetCursorCreate(info->rdmp); spdSetCursorNext(&csr);) {
        SpdObjId area = spdSetGetValueO(csr.curElem->key);
        SpdQueue *q = spdSetGetValueP(csr.curElem->value);

        if (verbose_flag) {
            printf("%d --> ", area);
            for (j = 0; j < q->elem; j++) {
                printf("%d ", spdQueuePeek(q, j));
            }
            printf("\n");
        }
    }

    if (verbose_flag) {
        printf("------\n");
    }
}

void
info_view(SdaStatInfo *info)
{
    printf("----------\n");
    printf("stmt:{%d, %d}\n", info->stmt.objectId, info->stmt.relationId);

    info_view_rdmp(info);
    printf("----------\n");
}

int
info_reduce_rdmp(SdaStatInfo *inf1, SdaStatInfo *inf2)
{
    int j;
    SetCursor csr;

    for (csr = spdSetCursorCreate(inf1->rdmp); spdSetCursorNext(&csr);) {
        SpdObjId def;
        SpdQueue *d_exprs1, *d_exprs2;
    
        def = spdSetGetValueO(csr.curElem->key);
        d_exprs1 = spdSetGetValueP(csr.curElem->value); /* ref->def */
        d_exprs2 = spdSetGetValueP(spdSearchElemO(inf2->rdmp, def)->value);
        if (d_exprs2 == NULL)  continue;    /* not include */

        for (j = 0; j < d_exprs1->elem; j++) {
            SpdObjId expr = spdQueuePeek(d_exprs1, j);
            int      idx  = spdQueueSearchIdx(d_exprs2, expr);
            if (idx < 0) continue;

            /*  */
            spdQueueDelete(d_exprs2, idx, 1);
        }
        if (d_exprs2->elem == 0) {
            spdQueueClean(d_exprs2);
            spdDelElemO(inf2->rdmp, def);
        }
    }
    return TRUE;
}
