/* 
 * Program:             $RCSfile: waitq.c,v $  $Revision: 30.10 $
 * 
 * Purpose:
 *
 * Author:              T.Suzuki  1998/10/22
 *                      S.Yamamoto  1997/10/22
 *
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: waitq.c,v 30.10 2003/06/08 02:43:08 yamamoto Exp yamamoto $";
#endif

/*
 *
 * StatementԤ/¹Դλݻ⥸塼
 * 
 */

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/StmtModelAR.h>
#include "statinfo.h"

/*
 *  waitq : 
 *    regist_stmt: Map<stmt, queue_of_stat>
 *    
 *  queue_of_stat: Queue<SdaStatInfo*>
 *  
 */

/*
 *  execq :
 *    regist_stmt: map<stmt, queue_of_stat>
 *  
 *  queue_of_stat: Queue<SdaStatInfo*>
 */

typedef struct Waitq_t {
    SpdSet   regist_stmt;                   /* stmt --> queue of stat_a */
} Waitq;

static Waitq waitq, execq;

void info_view(SdaStatInfo *info);

/*
 * Waitqν
 */

void _waitq_init(Waitq *q)
{
    q->regist_stmt = spdCreateSet(sizeof(SpdQueue));
}

/*
 * Waitqβ
 */

void _waitq_free(Waitq *q)
{
    int i, j;
    SpdQueue *stat_q;
    SdaStatInfo *inf;

    for (i=0; i<q->regist_stmt->elemNum; i++) {
        stat_q = (SpdQueue*)spdGetAElem(q->regist_stmt, NULL);
    
        for (j=0; j<stat_q->elem; j++) {
            inf = _spdQueuePeek(stat_q, j).p;

            info_free(inf);
        }

        spdQueueClean(stat_q);
    }
    
    spdFreeSet(q->regist_stmt);
}

/* ߤϡ뤫ʤΤ */

int _waitq_size(Waitq *q)
{
    int     i;

    for (i = 0; i < q->regist_stmt->elemNum; i++) {
        SpdQueue    *stat_q = (SpdQueue*)spdGetAElem(q->regist_stmt, NULL);

        if (0 < stat_q->elem) {
            return (1);
        }
    }
    return (0);
}

/*
 * Ǥɲ
 */

void _waitq_add_info(Waitq *q, SdaStatInfo *inf)
{
    SpdObjId stmt = inf->stmt.objectId;
    SpdQueue *stat_q = spdSearchElem(q->regist_stmt, stmt);

    if (stat_q == NULL) {
        /*  */
        SpdQueue dmy = spdQueueInit();
        stat_q = (SpdQueue*)spdAddElem(q->regist_stmt, stmt, &dmy);
        spdQueueClean(&dmy);
    }

    _spdQueuePushBack(stat_q, (_spd_queue_dummy.p = inf, _spd_queue_dummy));
}

/*
 * ƱstmtСʣǤ
 */

void _waitq_reduce_rdmp(Waitq *wq, SdaStatInfo *inf)
{
    int i;
    SpdQueue*   stat_q;

    /* ԤΤ ƱstmtΤΤ */
    stat_q =
        (SpdQueue*)spdSearchElem(wq->regist_stmt, inf->stmt.objectId);

    if (stat_q == NULL) return;

    /* Ԥ˴ޤޤʬ */
    for (i=0; i<stat_q->elem; i++) {
        SdaStatInfo *c_inf = _spdQueuePeek(stat_q, i).p;
        info_reduce_rdmp(c_inf, inf);
    }
}

/*
 * ֤ɲ: ƱstmtСʣ򤱤Ԥʤ
 * stmt == SPD_NON_ID ΤȤΥ֥Ȥ(ԡ)Ǽ롣
 */

void _waitq_add_stmt(Waitq *wq, SdaStatInfo *inf, SpdRel stmt)
{
    SdaStatInfo *cpy_inf;

    if (stmt.objectId != SPD_NON_ID) {
        cpy_inf = info_dup(inf);
        cpy_inf->stmt = stmt;
    }
    else
        cpy_inf = inf;

    _waitq_reduce_rdmp(wq, cpy_inf);

    _waitq_add_info(wq, cpy_inf);
}

/*
 * infoloop_cntοͥõ֤
 */

SdaStatInfo *_get_max_level(Waitq *wq)
{
    int i, j, idx;
    int max_lvl = -1;
    SdaStatInfo *max_if = NULL;
    SpdQueue    *max_q  = NULL;
    
    for (i=0; i<wq->regist_stmt->elemNum; i++) {
        SpdQueue *stat_q =
            (SpdQueue*)spdGetAElem(wq->regist_stmt, NULL);
    
        for (j=0; j<stat_q->elem; j++) {
            SdaStatInfo *n_if = _spdQueuePeek(stat_q, j).p;
        
            if (n_if->lpq->elem > max_lvl) {
                idx = j;
                max_if  = n_if;
                max_q   = stat_q;
                max_lvl = n_if->lpqc->elem;
            }
        }
    }

    /* queueϺ */
    if (max_q != NULL) {
        spdQueueDelete(max_q, idx, 1);
    }

    return max_if;
}

SdaStatInfo *_get_nxt_info(Waitq *wq)
{
    int i, j, idx;
    int objectId = SPD_NON_ID;
    SdaStatInfo *max_if = NULL;
    SpdQueue    *max_q  = NULL;
    
    for (i=0; i<wq->regist_stmt->elemNum; i++) {
        SpdQueue *stat_q =
            (SpdQueue*)spdGetAElem(wq->regist_stmt, NULL);
    
        for (j=0; j<stat_q->elem; j++) {
            SdaStatInfo *n_if = _spdQueuePeek(stat_q, j).p;
        
            if (n_if->stmt.objectId != SPD_NON_ID &&
                (objectId == SPD_NON_ID || n_if->stmt.objectId < objectId)){
                idx = j;
                max_if  = n_if;
                max_q   = stat_q;
                objectId = n_if->stmt.objectId;
            }
        }
    }

    /* queueϺ */
    if (max_q != NULL) {
        spdQueueDelete(max_q, idx, 1);
    }

    return max_if;
}

void _waitq_view(Waitq *wq)
{
    int i, j;
    
    printf("==== waitq\n");
    for (i = 0; i < wq->regist_stmt->elemNum; i++) {
        SpdQueue    *stat_q = (SpdQueue*)spdGetAElem(wq->regist_stmt, NULL);

        for (j = 0; j < stat_q->elem; j++) {
            SdaStatInfo     *n_if = _spdQueuePeek(stat_q, j).p;

            printf("%d --> %x\n", n_if->stmt.objectId, n_if);
            info_view(n_if);
        }
    }
    printf("=====\n");
}

SpdQueue *_get_stat(Waitq *wq, SpdRel stmt)
{
    return (SpdQueue*)spdSearchElem(wq->regist_stmt, stmt.objectId);
}

/* ---------------------------------------- */

void waitq_init(void)
{
    _waitq_init(&waitq);
    _waitq_init(&execq);
}

void waitq_free(void)
{
    _waitq_free(&waitq);
    _waitq_free(&execq);
}

/*
 * (ʬʤɤˤ)ǲս¸
 */

void waitq_add_stmt(SdaStatInfo *inf, SpdRel stmt)
{
    /* äƤơƱǤСʤ */
    SpdQueue* stat_q;
    SpdRel    c_stmt;

    c_stmt = (stmt.objectId != SPD_NON_ID) ? stmt : inf->stmt;

    stat_q = _get_stat(&waitq, c_stmt);

    if (stat_q != NULL) {
        SpdRel  tmp_id = inf->stmt;

        inf->stmt = c_stmt;
        _waitq_reduce_rdmp(&waitq, inf);
        if (inf->rdmp->elemNum == 0)
            return;
        inf->stmt = tmp_id;
    }

    _waitq_add_stmt(&waitq, inf, stmt);
}

/*
 * ¹Ԥ֤Ͽ
 */

void execq_add_stmt(SdaStatInfo *inf, SpdRel rel)
{
    _waitq_add_stmt(&execq, inf, rel);
}

/*
 * check_queue
 */

SdaStatInfo *check_already_exec(SdaStatInfo *inf)
{
    SpdQueue *stat_q;

    stat_q = (SpdQueue*)spdSearchElem(execq.regist_stmt,
                                      inf->stmt.objectId);
    if (stat_q == NULL) {
        /* ޤԤ٤¹ԤƤʤ */
        return inf;
    }

    _waitq_reduce_rdmp(&waitq, inf);
    _waitq_reduce_rdmp(&execq, inf);

    if (inf->rdmp->elemNum == 0) {

        /* Ǥˡ¹ԤƤꡢ˶rdmapξ */
        return NULL;
    }

    return inf;
}

/*
 * Ԥ椫鼡˼¹Ԥʸ
 */

SdaStatInfo *getNextFromWQ(void)
{
    SdaStatInfo *nxt_info = NULL;

    /* ¹ԾֵϿɲ */

    /*    _waitq_view(&waitq);*/

    while(1) {
        SpdQueue    *stat_q;
        SdaStatInfo *tmp_info;
    
        /*  nxt_info = _get_max_level(&waitq);*/
        nxt_info = _get_nxt_info(&waitq);

        if (nxt_info == NULL) 
            break;
    
        tmp_info = check_already_exec(nxt_info);

        stat_q = (SpdQueue*)spdSearchElem(waitq.regist_stmt,
                                          nxt_info->stmt.objectId);
        if (stat_q->elem == 0)
            tmp_info = nxt_info;        /* ޤ̤¹ԤΤᡢ֤ʤ */

        if (tmp_info != NULL) {
            nxt_info = tmp_info;        /* ¹ԾΥɲ */
            execq_add_stmt(nxt_info, nxt_info->stmt);
            break;
        }

        info_free(nxt_info);
    }

    /* SdaStatInfo֤stmt˼˼¹Ԥ٤Ԥޤ */
    return nxt_info;
}

int waitq_size(void)
{
    return _waitq_size(&waitq);
}
