/* 
 * Program:             $RCSfile: SpdXMLDetag.java,v $
 * 
 * Purpose:             
 * 
 * Author:              K.Niimi  2003/10/11
 *                      S.Yamamoto  2003/10/11
 * 
 * (C) Copyright:       K.Niimi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: SpdXMLDetag.java,v 50.3 2003/10/14 15:00:04 yamamoto Exp yamamoto $
 */

package sapid.util;

import java.io.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 * To detag XML.
 *
 * @author Ken Niimi.
 */

public class SpdXMLDetag
{
    /** file name */
    private String fileName = null;

    /** StringBuffer */
    private StringBuffer buffer = null;

    /** validate flag */
    private boolean isValidate = false;

    /** document root element */
    private Element root = null;

    /** Line Separator */
    SpdLineSeparator lineSep = null;

    /**
     * Constructor
     * @param fileName
     * @param isValidate
     */
    public SpdXMLDetag(String fileName,boolean isValidate){
        this.fileName = fileName;
        this.isValidate = isValidate;
        buffer = new StringBuffer();
        root = makeRootElement();

        try{
            lineSep = SpdLineSeparator.getLineSeparator(fileName);
        }
        catch(FileNotFoundException e){
            SpdMsg.errorExit("Cannot find file [" + fileName + "]",e);
        }
        
        readXML(root);        
    }

    /** make root element */
    private Element makeRootElement(){
        Element elem = null;
        try{
            //make DocumentBuilderFactory
            DocumentBuilderFactory dbfactory = DocumentBuilderFactory.newInstance();
            dbfactory.setValidating(isValidate);
            //make DocumentBuilder
            DocumentBuilder builder = dbfactory.newDocumentBuilder();
            //parse
            Document doc = builder.parse(fileName);
            //get root Element
            elem = doc.getDocumentElement();
        }
        catch(Exception e){
            SpdMsg.errorExit("Document Element make error",e);
        }
        return elem;
    }

    /** detag XML */
    public String deTag(){
        return buffer.toString();
    }
    
    /** reading xml text  */
    private void readXML(Element elem){
        NodeList nlist = elem.getChildNodes();
        for(int i=0;i<nlist.getLength();i++){
            if(nlist.item(i).getNodeType()==org.w3c.dom.Node.TEXT_NODE){
                //replace \n to LineSeparator
                buffer.append(nlist.item(i).toString().replaceAll("\n",lineSep.toString()));
            } else {
                Element elem2 = (Element)nlist.item(i);
                readXML(elem2);
            }
        }
    }
}
