/*  
 * Program:             $RCSfile: JSXTExpression.java,v $  $Revision: 50.6 $
 * 
 * Purpose:             Expression (NonTerminalNode).
 *                      child: 
 * 
 * Author:              S.Takeshi  2003/08/03
 *                      S.Yamamoto  2003/08/03
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: JSXTExpression.java,v 50.6 2003/08/12 06:36:42 yamamoto Exp yamamoto $
 */

package wapid.jsxmodel;

import wapid.parser.js.*;

public class JSXTExpression extends JSXTMetaNode {
    public String node_name = "Expr";

    /** save expression sort */
    protected int expressionSort;

    public static final int NOT_SORTED = -1;

    public static final int MUL = 0;
    public static final int DIV = 1;
    public static final int MOD = 2;
    public static final int ADD = 3;
    public static final int SUB = 4;
    public static final int SHL = 5;
    public static final int SHR = 6;
    public static final int SHRZ = 7;
    public static final int NOT = 8;
    public static final int AND = 9;
    public static final int XOR = 10;
    public static final int OR = 11;
    public static final int ASSIGN = 12;
    public static final int ASSIGN_MUL = 13;
    public static final int ASSIGN_DIV = 14;
    public static final int ASSIGN_MOD = 15;
    public static final int ASSIGN_ADD = 16;
    public static final int ASSIGN_SUB = 17;
    public static final int ASSIGN_SHL = 18;
    public static final int ASSIGN_SHR = 19;
    public static final int ASSIGN_SHRZ = 20;
    //  public static final int ASSIGN_NOT = 21;
    public static final int ASSIGN_AND = 22;
    public static final int ASSIGN_XOR = 23;
    public static final int ASSIGN_OR = 24;
    public static final int DOT = 25;
    public static final int INSTANCE_OF = 26;
    public static final int PRE_INC = 27;
    public static final int PRE_DEC = 28;
    public static final int POST_INC = 29;
    public static final int POST_DEC = 30;
    public static final int PLUS = 31;
    public static final int MINUS = 32;
    public static final int LOGICAL_NOT = 33;
    public static final int LOGICAL_AND = 34;
    //  public static final int LOGICAL_XOR = 35;
    public static final int LOGICAL_OR = 36;
    public static final int COND_LT = 37;
    public static final int COND_GT = 38;
    public static final int COND_LE = 39;
    public static final int COND_EQ = 40;
    public static final int COND_NE = 41;
    public static final int COND_GE = 42;
    public static final int SELECTION = 43;
    //  public static final int CONSTRUCTOR_CALL = 44;
    //  public static final int SUPER_CONSTRUCTOR_CALL = 45;
    public static final int FUN_CALL = 46;
    public static final int LISTED = 47;
    public static final int ARRAY_ACCESS = 48;
    public static final int ALLOCATION = 49;
    public static final int ARRAY_CREATION = 50;
    //  public static final int CAST = 51;
    public static final int PAREN = 52;
    public static final int VAR_REF = 53;
    public static final int TYPE_REF = 54;
    public static final int LITERAL = 55;
    public static final int THIS = 56;
    //  public static final int CLASS = 57;
    public static final int EMPTY = 58;
    public static final int ARGUMENT = 59;
    public static final int VOID = 60;
    public static final int TYPEOF = 61;
    public static final int DELETE = 62;
    public static final int COND_SEQ = 63;
    public static final int COND_NSEQ = 64;

    public static final int COMPOSIT = 65;

    /** constructor */
    public JSXTExpression(
        int depth,
        Token firstToken,
        Token lastToken,
        int sort) {
        super(depth, firstToken, lastToken);
        setNodeName(node_name);
        this.expressionSort = sort;
        super.node_type = EXPR;
    }
    public JSXTExpression(
        int depth,
        Token firstToken,
        Token lastToken,
        boolean list) {
        super(depth, firstToken, lastToken);
        setNodeName(node_name);
        super.node_type = EXPRLIST;
    }

    /** int to Stirng : XML attribute */
    public String toAttribute() {
        StringBuffer sb = new StringBuffer(" sort=\"");
        switch (expressionSort) {
            case MUL : // MUL = 0;
                sb.append("Mul");
                break;
            case DIV : // DIV = 1;
                sb.append("Div");
                break;
            case MOD : // MOD = 2;
                sb.append("Mod");
                break;
            case ADD : // ADD = 3;
                sb.append("Add");
                break;
            case SUB : // SUB = 4;
                sb.append("Sub");
                break;
            case SHL : // SHL = 5;
                sb.append("LShift");
                break;
            case SHR : // SHR = 6;
                sb.append("RShift");
                break;
            case SHRZ : // SHRZ = 7;
                sb.append("RZeroShift");
                break;
            case NOT : // NOT = 8;
                sb.append("Not");
                break;
            case AND : // AND = 9;
                sb.append("And");
                break;
            case XOR : // XOR = 10;
                sb.append("Xor");
                break;
            case OR : // OR = 11;
                sb.append("Or");
                break;
            case ASSIGN : // ASSIGN = 12;
                sb.append("Assign");
                break;
            case ASSIGN_MUL : // ASSIGN_MUL = 13;
                sb.append("AssignMul");
                break;
            case ASSIGN_DIV : // ASSIGN_DIV = 14;
                sb.append("AssignDiv");
                break;
            case ASSIGN_MOD : // ASSIGN_MOD = 15;
                sb.append("AssignMod");
                break;
            case ASSIGN_ADD : // ASSIGN_ADD = 16;
                sb.append("AssignAdd");
                break;
            case ASSIGN_SUB : // ASSIGN_SUB = 17;
                sb.append("AssignSub");
                break;
            case ASSIGN_SHL : // ASSIGN_SHL = 18;
                sb.append("AssignShl");
                break;
            case ASSIGN_SHR : // ASSIGN_SHR = 19;
                sb.append("AssignShr");
                break;
            case ASSIGN_SHRZ : // ASSIGN_SHRZ = 20;
                sb.append("AssignShrZ");
                break;
                //          public static final int ASSIGN_NOT = 21;
            case ASSIGN_AND : // ASSIGN_AND = 22;
                sb.append("AssignAnd");
                break;
            case ASSIGN_XOR : // ASSIGN_XOR = 23;
                sb.append("AssignXor");
                break;
            case ASSIGN_OR : // ASSIGN_OR = 24;
                sb.append("AssignOr");
                break;
                //          public static final int DOT = 25;
                //          public static final int INSTANCE_OF = 26;
            case PRE_INC : // PRE_INC = 27;
                sb.append("PreInc");
                break;
            case PRE_DEC : // PRE_DEC = 28;
                sb.append("PreDec");
                break;
            case POST_INC : // POST_INC = 29;
                sb.append("PostInc");
                break;
            case POST_DEC : // POST_DEC = 30;
                sb.append("PostDec");
                break;
            case PLUS : // PLUS = 31;
                sb.append("Plus");
                break;
            case MINUS : // MINUS = 32;
                sb.append("Minus");
                break;
            case LOGICAL_NOT : // LOGICAL_NOT = 33
                sb.append("LogicalNot");
                break;
            case LOGICAL_AND : // LOGICAL_AND = 34
                sb.append("LogicalAnd");
                break;
                //          public static final int LOGICAL_XOR = 35;
            case LOGICAL_OR : // LOGICAL_OR = 36
                sb.append("LogicalOr");
                break;
            case COND_LT : // COND_LT = 37;
                sb.append("LessThan");
                break;
            case COND_GT : // COND_GT = 38;
                sb.append("GreaterThan");
                break;
            case COND_LE : // COND_LE = 39;
                sb.append("LessEqual");
                break;
            case COND_EQ : // COND_EQ = 40;
                sb.append("Equal");
                break;
            case COND_NE : // COND_NE = 41;
                sb.append("NotEqual");
                break;
            case COND_GE : // COND_GE = 42;
                sb.append("GreaterEqual");
                break;
            case SELECTION : // SELECTION = 43;
                sb.append("Select");
                break;
                //          public static final int CONSTRUCTOR_CALL = 44;
                //          public static final int SUPER_CONSTRUCTOR_CALL = 45;
            case FUN_CALL : // FUN_CALL = 46;
                sb.append("FunCall");
                break;
                //          public static final int LISTED = 47;
            case ARRAY_ACCESS : // ARRAY_ACCESS = 48;
                sb.append("ArrayAccess");
                break;
            case ALLOCATION : // ALLOCATION = 49;
                sb.append("Allocation");
                break;
                //          public static final int ARRAY_CREATION = 50;
                //          public static final int CAST = 51;
            case PAREN : // PAREN = 52;
                sb.append("Paren");
                break;
            case VAR_REF : // VAR_REF = 53;
                sb.append("VarRef");
                break;
                //          public static final int TYPE_REF = 54;
            case LITERAL : // LITERAL = 55;
                sb.append("Literal");
                break;
            case THIS : // THIS = 56;
                sb.append("This");
                break;
                //          public static final int CLASS = 57;
            case EMPTY : // EMPTY = 58;
                sb.append("Empty");
                break;
            case ARGUMENT : //  ARGUMENT = 59;
                sb.append("Argument");
                break;
            case VOID ://   VOID = 60;
                sb.append("Void");
                break;
            case TYPEOF : //    TYPEOF = 61;
                sb.append("TypeOf");
                break;
            case DELETE: // DELETE = 62;
                sb.append("Delete");
                break;
            case COND_SEQ : //  COND_SEQ = 63;
                sb.append("StrictEqual");
                break;
            case COND_NSEQ : // COND_SEQ = 63;
                sb.append("StrictNotEqual");
                break;

            default :
                sb.append("NOT_SORTED");
                break;
        }
        sb.append("\"");
        return sb.toString();
    }

}
