/*  
 * Program:             $RCSfile: JSXTGenerator.java,v $  $Revision: 50.8 $
 * 
 * Purpose:             JavaScriptXML tree generator.
 *                      This program visits AST and generate JSXT.
 * 
 * Author:              S.Takeshi  2003/08/03
 *                      S.Yamamoto  2003/08/03
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: JSXTGenerator.java,v 50.8 2003/09/09 07:36:48 yamamoto Exp yamamoto $
 */

package wapid.jsxmodel;

import java.util.*;
import wapid.parser.js.*;

public class JSXTGenerator implements EcmaScriptVisitor {
    // JSXT
    public Vector nodes = new Vector();
    public Vector specials = new Vector();
    // ID factory
    private IDFactory idf = new IDFactory();
    // depth of node
    private int depth = 0;

    /**
     * regist Node
     * @param node
     */
    public void registNode(JSXTMetaNode node) {
        if (!node.isTerminalNode || node.isIdent) {
            node.setID(idf.createID());
        }
        nodes.add(node);
        setNodeRelation(node);
    }

    /**
     * regist SpecialToken
     * @param n
     */
    public void registSpecialToken(SimpleNode n) {
        Vector tmp = n.checkSpecialToken();
        JSXTMetaNode mn = null;
        for (int i = 0; i < tmp.size(); i++) {
            mn = (JSXTMetaNode) tmp.elementAt(i);
            if (!duplicationCheck(mn)) {
                mn.depth += depth;
                nodes.add(mn);
                // for dupulicated check
                specials.add(mn);
                setNodeRelation(mn);
            }
        }
    }

    /**
     * Check duplication of node
     * @param n
     * @return  true: duplicated with some node
     */
    public boolean duplicationCheck(JSXTMetaNode n) {
        boolean result = false;
        for (int i = 0; i < specials.size(); i++) {
            JSXTMetaNode t = (JSXTMetaNode) specials.elementAt(i);
            if (t.getPosition().equals(n.getPosition())) {
                result = true;
            }
        }
        return result;
    }

    /**
     * set Relation : parent <---> children
     * 
     * @param n
     */
    public void setNodeRelation(JSXTMetaNode n) {
        int d = n.depth, num = nodes.size();
        JSXTMetaNode parent = n;
        if (depth > 1) {
            while ((d - 1) != parent.depth) {
                //System.out.println(n.depth + " " + parent.depth);
                parent = (JSXTMetaNode) (nodes.elementAt(--num));
            }
            parent.addChild(n);
        }
    }

    /////////////////////////////////////
    //  visit AST  --begin
    /////////////////////////////////////

    /**  */
    public Object visit(SimpleNode node, Object data) {
        data = node.childrenAccept(this, data);
        return data;
    }

    /* *******************************
     *   Non Terminal Node
     *  ***************************** */

    /** Root node. (Program) */
    public Object visit(ASTProgram node, Object data) {
        ++depth;
        JSXTProgram n =
            new JSXTProgram(depth, node.getFirstToken(), node.getLastToken());
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    
    /** EOF */
    public Object visit(ASTEndOfFile node, Object data) {
        ++depth;
        registSpecialToken(node);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }

    // // UNUSE
    // public Object visit(ASTStatementList node, Object data) {
    //     // ++indent;
    //     data = node.childrenAccept(this, data);
    //     // --indent;
    //     return data;
    // }

    public Object visit(ASTFunctionDeclaration node, Object data) {
        ++depth;
        JSXTFunctionDeclaration n =
            new JSXTFunctionDeclaration(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTFunctionDeclaration.NAMED);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTAnonymousFunction node, Object data) {
        ++depth;
        JSXTFunctionDeclaration n =
            new JSXTFunctionDeclaration(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTFunctionDeclaration.ANONYMOUS);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTFormalParameterList node, Object data) {
        data = node.childrenAccept(this, data);
        return data;
    }
    public Object visit(ASTParameter node, Object data) {
    ++depth;
    JSXTFormalParameter n =
        new JSXTFormalParameter(
            depth,
            node.getFirstToken(),
            node.getLastToken());
    registSpecialToken(node);
    registNode(n);
    data = node.childrenAccept(this, data);
    --depth;
    return data;
}
    public Object visit(ASTStatement node, Object data) {
        data = node.childrenAccept(this, data);
        return data;
    }

    /**
     * Statement Virtual Parent Node
     */

    // // UNUSE
    // public Object visit(ASTBlockStatement node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    // //UNUSE
    // public Object visit(ASTVariableDeclaration node, Object data) {
    //     JSXTVariableDeclaration n =
    //         new JSXTVariableDeclaration(
    //             depth,
    //             node.getFirstToken(),
    //             node.getLastToken());
    //     registNode(n);
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    public Object visit(ASTVariableStatement node, Object data) {
        ++depth;
        JSXTVariableDeclaration n =
            new JSXTVariableDeclaration(
                depth,
                node.getFirstToken(),
                node.getLastToken());
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTIfStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.IF);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTDoWhileStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.DOWHILE);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTWhileStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.WHILE);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTForStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.FOR);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTForInStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.FOR);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTForVarStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.FOR);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTForVarInStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.FOR);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTContinueStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.CONTINUE);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTBreakStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.BREAK);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTReturnStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.RETURN);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTWithStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.WITH);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTSwitchStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.SWITCH);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTLabelledStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.LABELLED);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTThrowStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.THROW);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTTryStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.TRY);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTEmptyStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.EMPTY);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTExpressionStatement node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.EXPRESSION);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }

    /**
     * Block ( implied in statement as "statementblock")
     */
    public Object visit(ASTBlock node, Object data) {
        ++depth;
        JSXTStatement n =
            new JSXTStatement(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTStatement.BLOCK);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }

    // // UNUSE
    // public Object visit(ASTTryBlock node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    // // UNUSE
    // public Object visit(ASTCatchBlock node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    // // UNUSE
    // public Object visit(ASTFinallyBlock node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    /**
     * Expression
     */
    public Object visit(ASTAllocationExpression node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.ALLOCATION);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTPostfixExpression node, Object data) {
        ++depth;
        int postOpType = -1;
        boolean isUnary = false;
        SimpleNode sn = (SimpleNode) (node.jjtGetChild(1));
        postOpType = checkExprSort(sn, isUnary);
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                postOpType);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTUnaryExpression node, Object data) {
        ++depth;
        int unaryOpType = -1;
        boolean isUnary = true;
        SimpleNode sn = (SimpleNode) (node.jjtGetChild(0));
        unaryOpType = checkExprSort(sn, isUnary);
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                unaryOpType);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTBinaryExpressionSequence node, Object data) {
        ++depth;
        int binaryOpType = -1;
        boolean isUnary = false;
        SimpleNode sn = (SimpleNode) (node.jjtGetChild(1));
        binaryOpType = checkExprSort(sn, isUnary);
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                binaryOpType);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTOrExpressionSequence node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.LOGICAL_OR);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTAndExpressionSequence node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.LOGICAL_AND);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTConditionalExpression node, Object data) {
        ++depth;
        int conditionType = -1;
        boolean isUnary = false;
        SimpleNode sn = (SimpleNode) (node.jjtGetChild(1));
        conditionType = checkExprSort(sn, isUnary);
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                conditionType);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTAssignmentExpression node, Object data) {
        ++depth;
        int assignType = -1;
        boolean isUnary = false;
        SimpleNode sn = (SimpleNode) (node.jjtGetChild(1));
        assignType = checkExprSort(sn, isUnary);
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                assignType);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }

    // public Object visit(ASTPrimaryExpression node, Object data) {
    //     ++depth;
    //     JSXTExpression n =
    //         new JSXTExpression(
    //             depth,
    //             node.getFirstToken(),
    //             node.getLastToken(),
    //             0);
    //     registSpecialToken(node);
    //     registNode(n);
    //     data = node.childrenAccept(this, data);
    //     --depth;
    //     return data;
    // }

    public Object visit(ASTLiteralExpression node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.LITERAL);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTEmptyExpression node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.EMPTY);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }

    // // UNUSE
    // public Object visit(ASTExpression node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    /** special Literal */
    public Object visit(ASTArrayLiteral node, Object data) {
        data = node.childrenAccept(this, data);
        return data;
    }
    public Object visit(ASTObjectLiteral node, Object data) {
        data = node.childrenAccept(this, data);
        return data;
    }

    // // UNUSE
    // public Object visit(ASTExpressionList node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }
    // // UNUSE
    // public Object visit(ASTArray node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }
    // // UNUSE
    // public Object visit(ASTCallExpression node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }
    // // UNUSE
    // public Object visit(ASTLeftHandSideExpression node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }
    // // UNUSE
    // public Object visit(ASTMemberExpression node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    /**
     * additive Information
     */
    /** this expression */
    public Object visit(ASTThisReference node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.THIS);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTInitializer node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.ASSIGN);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    // UNUSE
    public Object visit(ASTCompositeReference node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                true);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    // TODO var_ref? member_access?
    public Object visit(ASTPropertyIdentifierReference node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.VAR_REF);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTPropertyValueReference node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.ARRAY_ACCESS);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    public Object visit(ASTArgument node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.ARGUMENT);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    // // UNUSE
    // public Object visit(ASTQNameList node, Object data) {
    //     ++depth;
    //    // node = registQName(node, data);
    //     JSXTQName n =
    //         new JSXTQName(
    //             depth,
    //             node.getFirstToken(),
    //             node.getLastToken());
    //     registSpecialToken(node);
    //     registNode(n);
    //     data = node.childrenAccept(this, data);
    //     --depth;
    //     return data;
    // }

    // // UNUSE
    // public Object visit(ASTQName node, Object data) {
    //     ++depth;
    //     JSXTQName n =
    //         new JSXTQName(
    //             depth,
    //             node.getFirstToken(),
    //             node.getLastToken());
    //     registSpecialToken(node);
    //     registNode(n);
    //     data = node.childrenAccept(this, data);
    //     --depth;
    //     return data;
    // }
    // 

    /** Argument :  */
    public Object visit(ASTArgumentList node, Object data) {
        ++depth;
        JSXTExpression n =
            new JSXTExpression(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                JSXTExpression.FUN_CALL);
        registSpecialToken(node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    
    // // UNUSE : "Param" is used instead.
    // public Object visit(ASTFunctionCallParameters node, Object data) {
    //     data = node.childrenAccept(this, data);
    //     return data;
    // }

    /*
     * Terminal Element
     */
     
    /** literal */
    public Object visit(ASTLiteral node, Object data) {
        ++depth;
        String literal_sort = node.getType();
        JSXTLiteral n =
            new JSXTLiteral(
                depth,
                node.getFirstToken(),
                node.getLastToken(),
                literal_sort);
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    
    /** identifier */
    public Object visit(ASTIdentifier node, Object data) {
        ++depth;
        JSXTIdentifier n =
            new JSXTIdentifier(
                depth,
                node.getFirstToken(),
                node.getLastToken());
        n.isIdent = true;
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    
    /** operetor */
    public Object visit(ASTOperator node, Object data) {
        ++depth;
        JSXTOperator n =
            new JSXTOperator(depth, node.getFirstToken(), node.getLastToken());
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    
    /** keyword */
    public Object visit(ASTKeyWord node, Object data) {
        ++depth;
        JSXTKeyWord n =
            new JSXTKeyWord(depth, node.getFirstToken(), node.getLastToken());
        registSpecialToken((SimpleNode) node);
        registNode(n);
        data = node.childrenAccept(this, data);
        --depth;
        return data;
    }
    
    /////////////////////////////////////
    //  visit AST  --end
    /////////////////////////////////////

    /**
     * Check Expression Sort 
     * @param sn
     * @return sort (JSXTExpression constant)
     */
    private int checkExprSort(SimpleNode sn, boolean isUnary) {
        int sort = -1;
        if (!isUnary) {
            switch (sn.getFirstToken().kind) {
                // Assignment expression
                case EcmaScriptConstants.ASSIGN : // =
                    sort = JSXTExpression.ASSIGN;
                    break;
                case EcmaScriptConstants.PLUSASSIGN : // +=
                    sort = JSXTExpression.ASSIGN_ADD;
                    break;
                case EcmaScriptConstants.MINUSASSIGN : // -=
                    sort = JSXTExpression.ASSIGN_SUB;
                    break;
                case EcmaScriptConstants.STARASSIGN : // *=
                    sort = JSXTExpression.ASSIGN_MUL;
                    break;
                case EcmaScriptConstants.SLASHASSIGN : // /=
                    sort = JSXTExpression.ASSIGN_DIV;
                    break;
                case EcmaScriptConstants.ANDASSIGN : // &=
                    sort = JSXTExpression.ASSIGN_AND;
                    break;
                case EcmaScriptConstants.ORASSIGN : // |=
                    sort = JSXTExpression.ASSIGN_OR;
                    break;
                case EcmaScriptConstants.XORASSIGN : // ^=
                    sort = JSXTExpression.ASSIGN_XOR;
                    break;
                case EcmaScriptConstants.REMASSIGN : // %=
                    sort = JSXTExpression.ASSIGN_MOD;
                    break;
                case EcmaScriptConstants.LSHIFTASSIGN : // <<=
                    sort = JSXTExpression.ASSIGN_SHL;
                    break;
                case EcmaScriptConstants.RSIGNEDSHIFTASSIGN : // >>=
                    sort = JSXTExpression.ASSIGN_SHR;
                    break;
                case EcmaScriptConstants.RUNSIGNEDSHIFTASSIGN : // >>>=
                    sort = JSXTExpression.ASSIGN_SHRZ;
                    break;

                    // Binary expression
                case EcmaScriptConstants.STAR : // *
                    sort = JSXTExpression.MUL;
                    break;
                case EcmaScriptConstants.SLASH : // /
                    sort = JSXTExpression.DIV;
                    break;
                case EcmaScriptConstants.REM : // %
                    sort = JSXTExpression.MOD;
                    break;
                case EcmaScriptConstants.PLUS : // +
                    sort = JSXTExpression.ADD;
                    break;
                case EcmaScriptConstants.MINUS : // -
                    sort = JSXTExpression.SUB;
                    break;
                case EcmaScriptConstants.LSHIFT : // <<
                    sort = JSXTExpression.SHL;
                    break;
                case EcmaScriptConstants.RSIGNEDSHIFT : // >>
                    sort = JSXTExpression.SHR;
                    break;
                case EcmaScriptConstants.RUNSIGNEDSHIFT : // >>>
                    sort = JSXTExpression.SHRZ;
                    break;
                case EcmaScriptConstants.LT : // <
                    sort = JSXTExpression.COND_LT;
                    break;
                case EcmaScriptConstants.GT : // >
                    sort = JSXTExpression.COND_GT;
                    break;
                case EcmaScriptConstants.LE : // <=
                    sort = JSXTExpression.COND_LE;
                    break;
                case EcmaScriptConstants.GE : // >=
                    sort = JSXTExpression.COND_GE;
                    break;
                case EcmaScriptConstants.EQ : // ==
                    sort = JSXTExpression.COND_EQ;
                    break;
                case EcmaScriptConstants.NE : // !=
                    sort = JSXTExpression.COND_NE;
                    break;
                case EcmaScriptConstants.SEQ : // ===
                    sort = JSXTExpression.COND_SEQ;
                    break;
                case EcmaScriptConstants.NSEQ : // !==
                    sort = JSXTExpression.COND_NSEQ;
                    break;
                case EcmaScriptConstants.BIT_AND : // &
                    sort = JSXTExpression.AND;
                    break;
                case EcmaScriptConstants.XOR : // ^
                    sort = JSXTExpression.XOR;
                    break;
                case EcmaScriptConstants.BIT_OR : // |
                    sort = JSXTExpression.OR;
                    break;

                    // Postfix expression
                case EcmaScriptConstants.INCR : // ++
                    sort = JSXTExpression.POST_INC;
                    break;
                case EcmaScriptConstants.DECR : // --
                    sort = JSXTExpression.POST_DEC;
                    break;

                    // Selection ( a ? b : c )
                case EcmaScriptConstants.HOOK : // --
                    sort = JSXTExpression.SELECTION;
                    break;

                default :
                    sort = JSXTExpression.PAREN;
                    break;
            }
        } else {
            // Unary expression
            switch (sn.getFirstToken().kind) {
                case EcmaScriptConstants.VOID : // void
                    sort = JSXTExpression.VOID;
                    break;
                case EcmaScriptConstants.TYPEOF : // typeof
                    sort = JSXTExpression.TYPEOF;
                    break;
                case EcmaScriptConstants.DELETE : // delete
                    sort = JSXTExpression.DELETE;
                    break;
                case EcmaScriptConstants.INCR : // ++
                    sort = JSXTExpression.PRE_INC;
                    break;
                case EcmaScriptConstants.DECR : // --
                    sort = JSXTExpression.PRE_DEC;
                    break;
                case EcmaScriptConstants.PLUS : // +
                    sort = JSXTExpression.PLUS;
                    break;
                case EcmaScriptConstants.MINUS : // -
                    sort = JSXTExpression.MINUS;
                    break;
                case EcmaScriptConstants.TILDE : // ~
                    sort = JSXTExpression.NOT;
                    break;
                case EcmaScriptConstants.BANG : // !
                    sort = JSXTExpression.LOGICAL_NOT;
                    break;

                default :
                    sort = JSXTExpression.NOT_SORTED;
                    break;
            }
        }
        return sort;
    }
}
