/*  
 * Program:             $RCSfile: JSXTMetaNode.java,v $  $Revision: 50.6 $
 * 
 * Purpose:             JSXT meta node.
 *                      Each JSXT node extends this class.
 * 
 * Author:              S.Takeshi  2003/08/03
 *                      S.Yamamoto  2003/08/03
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: JSXTMetaNode.java,v 50.6 2003/08/12 06:20:34 yamamoto Exp yamamoto $
 */

package wapid.jsxmodel;

import java.util.Vector;
import wapid.parser.js.*;

public abstract class JSXTMetaNode 
{
    /**
     * Node Type
     */
    public final int NONE = -1;
    
    public final int EXPRLIST = 0; // special
    
    public final int EXPR   =   1;
    public final int PARAM  =   2;
    public final int FUNDEC =   3;
    public final int PROGRAM=   4;
    public final int STMT   =   5;
    public final int VARDEC =   6;

    public final int COM    =   7;
    public final int EOL    =   8;
    public final int IDENT  =   9;
    public final int KW     =   10;
    public final int LITERAL=   11;
    public final int OP     =   12;
    public final int SP     =   13;
    
    /** 
     * Token Information
     */
    // Sort of node
    public int node_type = NONE;
    // Name of this node
    public String node_name;
    // Node ID
    protected int id; 
    // depth of this Node
    protected int depth;
    // first Token of this Node
    protected Token firstToken;
    // last Token of this Node
    protected Token lastToken;
    // parent Node of this one
    protected JSXTMetaNode parent;
    // children Nodes of this one
    protected Vector children;

    // TODO use node_type
    // Node type : Terminal or NonTerminal
    protected boolean isTerminalNode = false;
    // Node type : identifier
    protected boolean isIdent = false;

    
    /**
     * JSXT Node
     * 
     * @param firstToken
     * @param lastToken
     */
    public JSXTMetaNode(int depth, Token firstToken,Token lastToken){
        this.depth = depth;
        this.firstToken = firstToken;
        this.lastToken = lastToken;
        parent = null;
        children = new Vector();
        setNodeName(node_name);
    }
    /**
     * End tag object
     * 
     * @param node_name
     */
    public JSXTMetaNode(String node_name){
        setNodeName(node_name);
        this.depth = 0;
    }
    
    /** to XML attribute (String) */
    public String toAttribute(){
        return "";
    }
    
    /** Escape Char */
    public String checkStr(String str){
        str = str.replaceAll("&","&amp;");
        str = str.replaceAll("<","&lt;");
        str = str.replaceAll(">","&gt;");
        str = str.replaceAll("\"","&quot;");
        str = str.replaceAll("\'","&apos;");    
        return str;
    }
    
    /** to String */
    public String toString(){
        return "";      
    }
    
    /** Set ID */
    public void setID(int id){
        this.id = id;
    }
    
    /** Get ID*/
    public String getID(){
        return " id=\""+id+"\"";
    }
    
    /** Set node name */
    public void setNodeName(String name){
        this.node_name = name;
    }
    /** Set as parent node */
    public void setParent(JSXTMetaNode parent){
        this.parent = parent;
    }
    /** Get parent node */
    public JSXTMetaNode getParent(){
        return parent;
    }
    /** Set children */
    public void setChildren(Vector children){
        this.children = children;
    }
    /** Get child node (index) */
    public JSXTMetaNode getChild(int index){
        return (JSXTMetaNode)(children.elementAt(index));
    }
    /** Add child node */
    public void addChild(JSXTMetaNode obj){
        obj.setParent(this);
        children.addElement((Object)obj);
    }
    /** get number of children */
    public int getNumChildren(){
        return children.size();
    }
    /** get position of first token */ 
    public String getPosition(){
        String pos = "pos=\"";
        pos += firstToken.beginLine + " " + firstToken.beginColumn+ "\"";
        return pos;
    }
}
