/*  
 * Program:             $RCSfile: XMLSourceGenerator.java,v $  $Revision: 50.8 $
 * 
 * Purpose:             JSXT to XML source.
 * 
 * Author:              S.Takeshi  2003/08/03
 *                      S.Yamamoto  2003/08/03
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: XMLSourceGenerator.java,v 50.8 2003/09/15 05:14:22 yamamoto Exp yamamoto $
 */

package wapid.jsxmodel;

import java.util.*;

public class XMLSourceGenerator {
	private String line_buffer = "";
	private String file_encoding = "";
	private String dtd_file_name = "JSX-model.dtd";
	public Vector sourceText = new Vector();

	/**
	 * constructor
	 */
	public XMLSourceGenerator(String file_encoding) {
		this.file_encoding = file_encoding;
		this.sourceText.add("");
	}

	/**
	 * generate XML source
	 * @param v
	 * @return
	 */
	public Vector generateSource(Vector v) {
		sourceText.add(
			"<?xml version=\"1.0\" encoding=\"" + file_encoding + "\" ?>\n");
		sourceText.add("<!DOCTYPE Program SYSTEM \"" + dtd_file_name + "\">\n");
		JSXTMetaNode n = (JSXTMetaNode) v.elementAt(0);
		nodeDump(n);
		return sourceText;
	}

	/**
	 * Node Dumper ( XML tag )
	 * @param tag
	 * @return
	 */
	public void nodeDump(JSXTMetaNode n) {
		String textStr = "";
		if (n.node_type == n.EOL) {
			JSXTEndOfLine tmpEOL = (JSXTEndOfLine) n;
			textStr = tmpEOL.firstToken.image;
			sourceText.add(
				"<"
					+ tmpEOL.node_name
					+ " "
					+ tmpEOL.toAttribute()
					+ tmpEOL.getLineNum()
					+ " "
					+ tmpEOL.getCharOffset()
					+ " offset=\""
					+ (tmpEOL.firstToken.beginColumn
						+ getByteOffset(line_buffer))
					+ "\">"
					+ checkStr(textStr)
					+ "</"
					+ tmpEOL.node_name
					+ ">");
			line_buffer = "";
		} else if (n.isTerminalNode && !n.isIdent) {
			if (n.node_type == n.COM) {
				JSXTComment tmpCom = (JSXTComment) n;
				if (tmpCom.commentSort == JSXTComment.SINGLE) {
					String tmpStr = n.firstToken.image;
					// remove EndOfLine ("\n", "\r", "\n\r") from Single line Comment
					if (tmpStr
						.substring(tmpStr.length() - 2, tmpStr.length())
						.equals("\r\n")) {
						textStr =
							tmpStr.substring(
								0,
								n.firstToken.image.length() - 2);
					} else {
						textStr =
							tmpStr.substring(
								0,
								n.firstToken.image.length() - 1);
					}
				} else {
					textStr = n.firstToken.image;
				}
			} else {
				textStr = n.firstToken.image;
			}
			sourceText.add(
				"<"
					+ n.node_name
					+ " "
					+ n.toAttribute()
					+ n.getPosition()
					+ ">"
					+ checkStr(textStr)
					+ "</"
					+ n.node_name
					+ ">");
		} else if (n.isTerminalNode && n.isIdent) {
			textStr = n.firstToken.image;
			sourceText.add(
				"<"
					+ n.node_name
					+ n.getID()
					+ " "
					+ n.toAttribute()
					+ n.getPosition()
					+ ">"
					+ checkStr(textStr)
					+ "</"
					+ n.node_name
					+ ">");
		} else if (n.node_type == n.EXPRLIST) {
			// for composite reference
			makeLeftRecursive(n, n.getNumChildren());
		} else {
			sourceText.add(
				"<" + n.node_name + n.getID() + n.toAttribute() + ">");
			// visit child node
			for (int i = 0; i < n.getNumChildren(); i++) {
				nodeDump(n.getChild(i));
			}
			// end tag
			sourceText.add("</" + n.node_name + ">");
		}

		// memory string of 1-line
		if (!textStr.equals("\n")) {
			line_buffer += textStr;
		}
	}

	/**
	 * @param line_buffer
	 * @return
	 */
	private int getByteOffset(String line_buffer) {
		int byteOffset_diff = 0;
		byte[] byteArray = line_buffer.getBytes();

		for (int i = 0; i < line_buffer.length(); i++) {
			String tmp = String.valueOf(line_buffer.charAt(i));
			if (tmp.getBytes().length != 1) {
				byteOffset_diff++;
			}
		}
		return byteOffset_diff;
	}

	/**
	 * make left recursive tree
	 * @param n
	 * @param num
	 */
	private void makeLeftRecursive(JSXTMetaNode n, int num) {
		JSXTExpression cur = (JSXTExpression) n.getChild(--num);
		sourceText.add(
			"<" + cur.node_name + cur.getID() + cur.toAttribute() + ">");
		if (num > 0) {
			makeLeftRecursive(n, num);
		}
		//*** visit child node
		for (int i = 0; i < cur.getNumChildren(); i++) {
			nodeDump(cur.getChild(i));
		}
		// EndTag
		sourceText.add("</" + n.node_name + ">");
	}

	/**
	 * Indent
	 * @param depth
	 * @return
	 */
	private String indentString(int depth) {
		StringBuffer sb = new StringBuffer();
		for (int i = 1; i < depth; ++i) {
			sb.append(" ");
		}
		return sb.toString();
	}

	/**
	 * XML Escape Char
	 * @param str
	 * @return
	 */
	public String checkStr(String str) {
		str = str.replaceAll("&", "&amp;");
		str = str.replaceAll("<", "&lt;");
		str = str.replaceAll(">", "&gt;");
		str = str.replaceAll("\"", "&quot;");
		str = str.replaceAll("\'", "&apos;");
		return str;
	}
}
