/*  
 * Program:             $RCSfile: JavaScriptXSDMLAnalyzer.java,v $  $Revision: 50.2 $
 * 
 * Purpose:             XSDML analyzer for JavaScript.
 * 	                Making a simple symbol table.
 * 
 * Author:              S.Takeshi  2003/09/06
 *                      S.Yamamoto  2003/09/06
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: JavaScriptXSDMLAnalyzer.java,v 50.2 2003/09/07 07:28:56 yamamoto Exp yamamoto $
 */

/**
 * XSDML analyzer for JavaScript.
 *     Making a simple symbol table.
 * 
 * @author shiki
 */

package wapid.jsxmodel.xref;

import wapid.jsxmodel.xsapid.*;

public class JavaScriptXSDMLAnalyzer {

	// 
	static final SpdElementFilter JSX_FUNDEC =
		new SpdElementFilterByName("FunDec");
	static final SpdElementFilter JSX_VARDEC =
		new SpdElementFilterByName("VarDec");
	static final SpdElementFilter JSX_STMT = new SpdElementFilterByName("Stmt");

	//
	static final SpdElementFilter JSX_VARREF =
		new SpdElementFilterBySort("VarRef");
	static final SpdElementFilter JSX_FUNCALL =
		new SpdElementFilterBySort("FunCall");

	// terminal node
	static final SpdElementFilter JSX_SP = new SpdElementFilterByName("sp");
	static final SpdElementFilter JSX_KW = new SpdElementFilterByName("kw");
	static final SpdElementFilter JSX_IDENT =
		new SpdElementFilterByName("ident");
	static final SpdElementFilter JSX_NL = new SpdElementFilterByName("nl");
	static final SpdElementFilter JSX_LITERAL =
		new SpdElementFilterByName("literal");
	static final SpdElementFilter JSX_COMMENT =
		new SpdElementFilterByName("comment");
	static final SpdElementFilter JSX_OP = new SpdElementFilterByName("op");

	private SpdElement e0;

	// Symbol Table	
	public SymbolTable st = new SymbolTable();

	public JavaScriptXSDMLAnalyzer(String XSDML) {
		Repository r = new Repository();
		SpdDocument d = r.loadXSDML(XSDML, false);
		this.e0 = d.getFileElement();
		st.sourceText = e0.getText();
	}

	/**
	 * XSDML Analyzer (for JSX-model)
	 *
	 */
	public SymbolTable JSXAnalyzer() {

		/* 
		 * get global variable (declared explicitly) 
		 */
		SpdElementList elg = e0.searchChild(JSX_VARDEC);
		boolean isGrobal = true;
		VariableHandler vh = new VariableHandler(isGrobal);
		st.STGrobalVar = vh.getVariables(elg);

		/*
		 *  get function and it's local variable 
		 * 
		 * TODO nested function
		 */
		SpdElementList elf = e0.searchChild(JSX_FUNDEC);
		FunctionHandler fh = new FunctionHandler();
		st.STFunction = fh.getFunctions(elf);

		/* 
		 * get FunctionRef  
		 * 	<FunDec>-- 
		 * 
		 * TODO nested function
		 */
		SpdElementList elrf = e0.searchChild(JSX_FUNDEC);
		ReferenceHandler rh = new ReferenceHandler();
		for (int i = 0; i < elrf.size(); i++) {
			st = rh.getReferenceIdentOfFunction(st, elrf.getElementAt(i));
		}

		/*
		 * get TopLevelRef 
		 * 	<Program><Stmt>-- 
		 * 	or 
		 * 	<Program><VarDec>-- 
		 * 
		 * TODO analyze TopLevelStatement
		 */
		SpdElementList elrv = e0.searchChild(JSX_VARDEC);
		SpdElementList elrs = e0.searchChild(JSX_STMT);
		for (int i = 0; i < elrv.size(); i++) {
			st = rh.getReferenceIdentOfTopLevelStmt(st, elrv.getElementAt(i));
		}
		for (int i = 0; i < elrs.size(); i++) {
			st = rh.getReferenceIdentOfTopLevelStmt(st, elrs.getElementAt(i));
		}

		/*
		 * get Token
		 * 
		 * change SpdElement -- searchDecendant()
		 */
		SpdElementFilterList filterList = new SpdElementFilterList();
		filterList.addElement(JSX_COMMENT);
		filterList.addElement(JSX_IDENT);
		filterList.addElement(JSX_KW);
		filterList.addElement(JSX_LITERAL);
		filterList.addElement(JSX_NL);
		filterList.addElement(JSX_OP);
		filterList.addElement(JSX_SP);

		SpdElementList elTokenList = e0.searchDecendantMulti(filterList);
		st.tokenList = elTokenList;

		return st;
	}
}
