/*  
 * Program:             $RCSfile: ReferenceHandler.java,v $  $Revision: 50.2 $
 * 
 * Purpose:             Reference handler.
 * 
 * Author:              S.Takeshi  2003/09/06
 *                      S.Yamamoto  2003/09/06
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: ReferenceHandler.java,v 50.2 2003/09/07 07:28:57 yamamoto Exp yamamoto $
 */

/**
 * Reference handler.
 * 
 * @author
 */

package wapid.jsxmodel.xref;

import java.util.*;
import wapid.jsxmodel.xsapid.*;

public class ReferenceHandler {

	final SpdElementFilter JSX_VARREF = new SpdElementFilterBySort("VarRef");
	final SpdElementFilter JSX_FUNCALL = new SpdElementFilterBySort("FunCall");
	final SpdElementFilter JSX_IDENT = new SpdElementFilterByName("ident");
	final SpdElementFilter JSX_EXPR = new SpdElementFilterByName("Expr");

	/**
	 * get FunctionRef  
	 * 	<FunDec>-- 
	 * 
	 * @param st
	 * @param e
	 * @return
	 * TODO nested function
	 */
	public SymbolTable getReferenceIdentOfFunction(
		SymbolTable st,
		SpdElement e) {

		/* search target function's SymbolTableFunction */
		SymbolTableFunction targetFun = null;
		for (int i = 0; i < st.STFunction.size(); i++) {
			SymbolTableFunction tmpStf =
				(SymbolTableFunction) (st.STFunction.elementAt(i));
			if (tmpStf.funID == Integer.parseInt(e.getID())) {
				targetFun = tmpStf;
				break;
			}
		}

		/* variable check */
		SpdElementList sel = e.searchDecendant(JSX_VARREF);
		// ident list
		SpdElementList checkList = new SpdElementList();
		for (int i = 0; i < sel.size(); i++) {
			SpdElement work = sel.getElementAt(i).getIndexedChild(JSX_IDENT, 1);
			checkList.addElement(work);
		}

		// check ident
		for (int i = 0; i < checkList.size(); i++) {
			SpdElement checkEl = checkList.getElementAt(i);
			String checkName = checkEl.getText();
			boolean defined = false;
			/* search local variable */
			int j = 0;
			Vector targetList = targetFun.STLocalVar;
			SymbolTableElement targetEl = null;
			while (!defined && j < targetList.size()) {
				targetEl = (SymbolTableElement) (targetList.elementAt(j));
				if (targetEl.symName.equals(checkName)) {
					defined = true;
					SymbolTableReference str =
						new SymbolTableReference(
							checkName,
							checkEl.getPositionString(),
							Integer.parseInt(checkEl.getID()),
							targetEl.symID);
//					targetFun.STRefLocalVar.add(str);
					targetEl.usedList.add(str);
				}
				j++;
			}

			/* search grobal variable reference */
			if (!defined) {
				int k = 0;
				targetList = st.STGrobalVar;
				while (!defined && k < targetList.size()) {
					targetEl = (SymbolTableElement) (targetList.elementAt(k));
					if (targetEl.symName.equals(checkName)) {
						defined = true;
						SymbolTableReference str =
							new SymbolTableReference(
								checkName,
								checkEl.getPositionString(),
								Integer.parseInt(checkEl.getID()),
								targetEl.symID);
						targetFun.STRefGrobalVar.add(str);
						targetEl.usedList.add(str);
					}
					k++;
				}
			}

			// undefined varref
			if (!defined) {
				//				System.out.println("[U] "+checkName);
			}

		}

		/* check function call */
		sel = e.searchDecendant(JSX_FUNCALL);
		checkList = new SpdElementList();
		// make ident list (function name)
		for (int i = 0; i < sel.size(); i++) {
			SpdElement work = sel.getElementAt(i).getIndexedChild(JSX_EXPR, 1);
			work = work.getIndexedChild(JSX_IDENT, 1);
			checkList.addElement(work);
		}
		// check ident
		for (int i = 0; i < checkList.size(); i++) {
			SpdElement checkEl = checkList.getElementAt(i);
			String checkName = checkEl.getText();
			boolean defined = false;

			int j = 0;
			Vector targetList = st.STFunction;
			SymbolTableFunction targetEl = null;
			while (!defined && j < targetList.size()) {
				targetEl = (SymbolTableFunction) (targetList.elementAt(j));
				if (targetEl.funName.equals(checkName)) {
					defined = true;
					SymbolTableReference str =
						new SymbolTableReference(
							checkName,
							checkEl.getPositionString(),
							Integer.parseInt(checkEl.getID()),
							targetEl.funID);
					targetFun.STRefFunCall.add(str);
					targetEl.usedList.add(str);
				}
				j++;
			}
		}

		return st;
	}

	/**
	 * get FunctionRef  
	 * 	<Program><Stmt>-- 
	 * 	<Program><VarDec>-- 
	 * 
	 * @param st
	 * @param e
	 * @return
	 */
	public SymbolTable getReferenceIdentOfTopLevelStmt(
		SymbolTable st,
		SpdElement e) {

		/* check grobal variable reference */
		SpdElementList sel = e.searchDecendant(JSX_VARREF);
		// ident list
		SpdElementList checkList = new SpdElementList();

		for (int i = 0; i < sel.size(); i++) {
			SpdElement work = sel.getElementAt(i).getIndexedChild(JSX_IDENT, 1);
			checkList.addElement(work);
		}

		// check ident
		for (int i = 0; i < checkList.size(); i++) {
			SpdElement checkEl = checkList.getElementAt(i);
			String checkName = checkEl.getText();
			boolean defined = false;

			/* search grobal variable reference */
			int j = 0;
			Vector targetList = st.STGrobalVar;
			while (!defined && j < targetList.size()) {
				SymbolTableElement targetEl =
					(SymbolTableElement) (targetList.elementAt(j));
				if (targetEl.symName.equals(checkName)) {
					defined = true;
					SymbolTableReference str =
						new SymbolTableReference(
							checkName,
							checkEl.getPositionString(),
							Integer.parseInt(checkEl.getID()),
							targetEl.symID);
					st.RefVar.add(str);
					targetEl.usedList.add(str);
				}
				j++;
			}

			// undefined varref
			if (!defined) {
				// test
				//				System.out.println("[U] "+checkName);
			}

		}

		/* check function call */
		sel = e.searchDecendant(JSX_FUNCALL);
		checkList = new SpdElementList();
		// make ident list (function name)
		for (int i = 0; i < sel.size(); i++) {
			SpdElement work = sel.getElementAt(i).getIndexedChild(JSX_EXPR, 1);
			work = work.getIndexedChild(JSX_IDENT, 1);
			checkList.addElement(work);
		}
		// check ident
		for (int i = 0; i < checkList.size(); i++) {
			SpdElement checkEl = checkList.getElementAt(i);
			String checkName = checkEl.getText();
			boolean defined = false;
			/* search local variable */
			int j = 0;
			Vector targetList = st.STFunction;
			SymbolTableFunction targetEl = null;
			while (!defined && j < targetList.size()) {
				targetEl = (SymbolTableFunction) (targetList.elementAt(j));
				if (targetEl.funName.equals(checkName)) {
					defined = true;
					SymbolTableReference str =
						new SymbolTableReference(
							checkName,
							checkEl.getPositionString(),
							Integer.parseInt(checkEl.getID()),
							targetEl.funID);
					st.RefFun.add(str);
				}
				j++;
			}
		}

		return st;
	}
}
