/*  
 * Program:             $RCSfile: XHTMLSourceGenerator.java,v $  $Revision: 50.4 $
 * 
 * Purpose:             XHTMLSourceGenerator.
 * 
 * Author:              S.Takeshi  2003/09/06
 *                      S.Yamamoto  2003/09/06
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: XHTMLSourceGenerator.java,v 50.4 2003/09/27 06:45:17 yamamoto Exp yamamoto $
 */

/**
 *  XHTMLSourceGenerator
 */

package wapid.jsxmodel.xref;

import java.io.*;
import java.util.*;

import wapid.jsxmodel.xsapid.*;
import sapid.util.*;

public class XHTMLSourceGenerator {

	private final int marginLineNum = 50; // bottom margin for scroll
	private String output_dir_name;
	private String file_encoding;

	public XHTMLSourceGenerator(String output_dir_ame, String file_encoding) {
		this.output_dir_name = output_dir_ame;
		this.file_encoding = file_encoding;
	}

	public void generateXHTMLSource(SymbolTable st) {

		try {
			File file = new File(output_dir_name);

			String tmpText = "";

			/* check directory */
			if (file.exists()) {
				if (file.isDirectory()) {
				} else {
					System.exit(1);
				}
			} else {
				SpdMsg.verbose(
					"   --> Create Directory "
						+ "[ "
						+ output_dir_name
						+ " ]");
				file.mkdirs();

			}

			/* make CSS */
			CSSGenerator cssg = new CSSGenerator();
			cssg.makeCSS(output_dir_name + "/basic.css");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/basic.css ]");

			/* make Source Program file */
			tmpText = makeSourceHTMLWithLink(st);
			this.makeHTML(tmpText, "Source", output_dir_name + "/source.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/source.html ]");

			/* make index.html */
			tmpText = this.makeIndexHTML();
			this.makeHTML(
				tmpText,
				output_dir_name,
				output_dir_name + "/index.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/index.html ]");

			/* make contents.html */
			tmpText = this.makeContentsHTML(st);
			this.makeHTML(
				tmpText,
				"content",
				output_dir_name + "/contents.html");

			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/contents.html ]");

			/* create list of grobal var : list_grobalVar.html */
			Vector lg = new Vector();
			for (int i = 0; i < st.STGrobalVar.size(); i++) {
				lg.add(st.STGrobalVar.elementAt(i));
			}

			tmpText = this.makeListOfGrovalVarHTML(lg);
			this.makeHTML(
				tmpText,
				"list_grobalVar",
				output_dir_name + "/list_grobalVar.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/list_grobalVar.html ]");

			// create spec of grobal var : spec_grobalVar.html
			//test
			tmpText = this.makeSpecOfGrovalVarHTML(lg);
			this.makeHTML(
				tmpText,
				"spec_grobalVar",
				output_dir_name + "/spec_grobalVar.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/spec_grobalVar.html ]");

			/* create list of local var : list_localVar.html */
			Vector ll = new Vector();
			for (int i = 0; i < st.STFunction.size(); i++) {
				SymbolTableFunction tmpf =
					(SymbolTableFunction) st.STFunction.elementAt(i);
				for (int j = 0; j < tmpf.STLocalVar.size(); j++) {
					ll.add(tmpf.STLocalVar.elementAt(j));
				}
			}
			tmpText = this.makeListOfLocalVarHTML(ll);
			this.makeHTML(
				tmpText,
				"list_localVar",
				output_dir_name + "/list_localVar.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/list_localVar.html ]");

			// create spec of local var : spec_localVar.html
			tmpText = this.makeSpecOfLocalVarHTML(ll);
			this.makeHTML(
				tmpText,
				"spec_localVar",
				output_dir_name + "/spec_localVar.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/spec_localVar.html ]");

			/* create list of function : list_function.html */
			Vector lf = new Vector();
			for (int i = 0; i < st.STFunction.size(); i++) {
				lf.add(st.STFunction.elementAt(i));
			}
			tmpText = this.makeListOfFunctionHTML(lf);
			this.makeHTML(
				tmpText,
				"list_function",
				output_dir_name + "/list_function.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/list_function.html ]");

			// create spec of function : spec_function.html
			// test
			tmpText = this.makeSpecOfFunctionHTML(lf);
			this.makeHTML(
				tmpText,
				"spec_function",
				output_dir_name + "/spec_function.html");
			SpdMsg.verbose(
				"   --> Create File [ "
					+ output_dir_name
					+ "/spec_function.html ]");
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}

	}

	/**
	 * make HTML File
	 * @param text
	 * @param title
	 * @param outputFileName
	 */
	private void makeHTML(String text, String title, String outputFileName) {
		try {
			int bottomMarginLineNum = 25;
			FileOutputStream fos = new FileOutputStream(outputFileName);
			OutputStreamWriter osw = new OutputStreamWriter(fos, file_encoding);

			// head section
			osw.write(
				"<?xml version=\"1.0\" encoding=\"" + file_encoding + "\"?>\n");
			osw.write(
				"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n");
			osw.write(
				"<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"ja\" lang=\"ja\">\n");
			osw.write("<head>");
			osw.write(
				"<link rel=\"stylesheet\" href=\"basic.css\" type=\"text/css\">");
			osw.write("<title>" + title + "</title>");
			osw.write("</head>\n");

			// body section
			osw.write(text);

			// foot section
			osw.write("</html>\n");
			osw.flush();
			osw.close();
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * make spec_localVar.html
	 * @param Vector
	 */
	private String makeSpecOfLocalVarHTML(Vector list) {

		String text = "";

		text += "<body><p><strong>[ Local Variable ]</strong></p>\n";

		for (int i = 0; i < list.size(); i++) {
			SymbolTableElement el = (SymbolTableElement) list.elementAt(i);
			text += "<a name=\"" + el.symID + "\"></a>";
			text += "<table border=\"1\" width=\"100%\">\n";
			text += "<tr><td><strong>Name</strong><br />"
				+ el.symName
				+ "</td>\n";
			text += "<td><strong>Defined Position</strong><br />"
				+ "<a href=\"./source.html#"
				+ el.symID
				+ "\">"
				+ "line : "
				+ el.getLineNum()
				+ " column : "
				+ el.getColumnNum()
				+ "</a>"
				+ "</td></tr>";
			text += "<tr><td colspan=\"2\"><strong>Used</strong><br />";

			for (int j = 0; j < el.usedList.size(); j++) {
				SymbolTableReference tmp =
					(SymbolTableReference) el.usedList.elementAt(j);
				text += "( at <a href=\"./source.html#"
					+ tmp.identID
					+ "\">line : "
					+ tmp.getLineNum()
					+ " column : "
					+ tmp.getColumnNum()
					+ " </a>)</br>";
			}

			text += "</td></tr>";
			text += "</table><hr/>\n";
		}

		// bottom margin
		for (int i = 0; i < marginLineNum; i++) {
			text += "<br />";
		}
		text += "</body>";

		return text;
	}

	/**
	 * make spec_grobalVar.html
	 * @param string
	 */
	private String makeSpecOfGrovalVarHTML(Vector list) {

		String text = "";

		text += "<body><p><strong>[ Grobal Variable ]</strong></p>\n";

		for (int i = 0; i < list.size(); i++) {
			SymbolTableElement el = (SymbolTableElement) list.elementAt(i);
			text += "<a name=\"" + el.symID + "\"></a>";
			text += "<table border=\"1\" width=\"100%\">\n";
			text += "<tr><td><strong>Name</strong><br />"
				+ el.symName
				+ "</td>\n";
			text += "<td><strong>Defined Position</strong><br />"
				+ "<a href=\"./source.html#"
				+ el.symID
				+ "\">"
				+ "line : "
				+ el.getLineNum()
				+ " column : "
				+ el.getColumnNum()
				+ "</a>"
				+ "</td></tr>";
			text += "<tr><td colspan=\"2\"><strong>Used</strong><br />";

			for (int j = 0; j < el.usedList.size(); j++) {
				SymbolTableReference tmp =
					(SymbolTableReference) el.usedList.elementAt(j);
				text += "( at <a href=\"./source.html#"
					+ tmp.identID
					+ "\">line : "
					+ tmp.getLineNum()
					+ " column : "
					+ tmp.getColumnNum()
					+ " </a>)</br>";
			}

			text += "</td></tr>";
			text += "</table><hr/>\n";
		}

		// bottom margin
		for (int i = 0; i < marginLineNum; i++) {
			text += "<br />";
		}
		text += "</body>";

		return text;
	}

	/**
	 * make spec_function.html
	 * @param string
	 */
	private String makeSpecOfFunctionHTML(Vector list) {

		String text = "";

		text += "<body><p><strong>[ Function ]</strong></p>\n";

		for (int i = 0; i < list.size(); i++) {
			SymbolTableFunction el = (SymbolTableFunction) list.elementAt(i);
			text += "<a name=\"" + el.funID + "\"></a>";
			text += "<table border=\"1\" width=\"100%\">\n";
			text += "<tr><td><strong>Name</strong><br />"
				+ el.funName
				+ "</td>\n";
			text += "<td><strong>Defined Position</strong><br />"
				+ "<a href=\"./source.html#"
				+ el.funNameIdentID
				+ "\">"
				+ "line : "
				+ el.getLineNum()
				+ " column : "
				+ el.getColumnNum()
				+ "</a>"
				+ "</td></tr>";

			/* check parameter and local variable of function */
			String paramVar = "", localVar = "";
			for (int j = 0; j < el.STLocalVar.size(); j++) {
				SymbolTableElement tmp =
					(SymbolTableElement) el.STLocalVar.elementAt(j);
				if (tmp.symType == SymbolTableConstants.PARAM) {
					paramVar += tmp.symName
						+ " (line : "
						+ tmp.getLineNum()
						+ " column : "
						+ tmp.getColumnNum()
						+ ")<br />\n";
				} else if (tmp.symType == SymbolTableConstants.LOCALVARIABLE) {
					localVar += tmp.symName
						+ " (line : "
						+ tmp.getLineNum()
						+ " column : "
						+ tmp.getColumnNum()
						+ ")<br />\n";
				}
			}

			/* this function has following parameter */
			text += "<tr><td valign=\"top\"><strong>Parameter</strong><br />";
			text += paramVar;
			text += "</td>";

			/* this function has following local variable */
			text += "<td valign=\"top\"><strong>Local Variable</strong><br />";
			text += localVar;
			text += "</td></tr>";

			/* this function uses following grobal variable */
			text
				+= "<tr><td colspan=\"2\"><strong>Use of Grobal Variable</strong><br />";
			for (int j = 0; j < el.STRefGrobalVar.size(); j++) {
				SymbolTableReference tmp =
					(SymbolTableReference) el.STRefGrobalVar.elementAt(j);
				//				text += "<a href=\"spec_grobalVar.html#"
				//					+ tmp.defID
				//					+ "\" target=\"righttop\">"
				//					+ tmp.identName
				//					+ "</a>"
				//					+ " (line : "
				//					+ tmp.getLineNum()
				//					+ " column : "
				//					+ tmp.getColumnNum()
				//					+ ")<br />\n";
				text += tmp.identName
					+ " ( at <a href=\"./source.html#"
					+ tmp.identID
					+ "\">line : "
					+ tmp.getLineNum()
					+ " column : "
					+ tmp.getColumnNum()
					+ " </a>)</br>";
			}
			text += "</td></tr>";

			// this function calls following function
			text += "<tr><td colspan=\"2\"><strong>Call</strong><br />";

			for (int j = 0; j < el.STRefFunCall.size(); j++) {
				SymbolTableReference tmp =
					(SymbolTableReference) el.STRefFunCall.elementAt(j);
				text += tmp.identName
					+ "( at "
					+ "<a href=\"./source.html#"
					+ tmp.identID
					+ "\">"
					+ "line : "
					+ tmp.getLineNum()
					+ " column : "
					+ tmp.getColumnNum()
					+ "</a> )</br>";
			}

			text += "</td></tr>";

			// this function is called by following function
			text += "<tr><td colspan=\"2\"><strong>Called</strong><br />";

			for (int j = 0; j < el.usedList.size(); j++) {
				SymbolTableReference tmp =
					(SymbolTableReference) el.usedList.elementAt(j);
				text +=
				//					" by <a href=\"./spec_function.html#"
				//					+ el.funID
				//					+ "\">"
				//					+ el.funName
				//					+ "</a>"

				"( at <a href=\"./source.html#"
					+ tmp.identID
					+ "\">"
					+ "line : "
					+ tmp.getLineNum()
					+ " column : "
					+ tmp.getColumnNum()
					+ "</a> )<br />";
			}

			text += "</td></tr>";
			text += "</table><hr/>\n";
		}

		// bottom margin
		for (int i = 0; i < marginLineNum; i++) {
			text += "<br />";
		}
		text += "</body>";

		return text;
	}

	/**
	 * make list_localVar.html
	 * @param string
	 */
	private String makeListOfLocalVarHTML(Vector list) {

		String text = "";

		text += "<body><p><strong>[ Local Variable ]</strong></p><br />\n";

		for (int i = 0; i < list.size(); i++) {
			SymbolTableElement el = (SymbolTableElement) list.elementAt(i);
			text += "<a href=\"spec_localVar.html#"
				+ el.symID
				+ "\" target=\"righttop\">"
				+ el.symName
				+ "</a><br />\n";
		}
		text += "</body>";

		return text;
	}

	/**
	 * make list_grobalVar.html
	 * @param string
	 */
	private String makeListOfGrovalVarHTML(Vector gList) {

		String text = "";

		text += "<body><p><strong>[ Grobal Variable ]</strong></p><br />\n";

		for (int i = 0; i < gList.size(); i++) {
			SymbolTableElement g = (SymbolTableElement) gList.elementAt(i);
			text += "<a href=\"spec_grobalVar.html#"
				+ g.symID
				+ "\" target=\"righttop\">"
				+ g.symName
				+ "</a><br />\n";
		}
		text += "</body>";

		return text;
	}

	/**
	 * make list_function.html
	 * @param string
	 */
	private String makeListOfFunctionHTML(Vector gList) {

		String text = "";

		text += "<body><p><strong>[ Function ]</strong></p><br />\n";

		for (int i = 0; i < gList.size(); i++) {
			SymbolTableFunction g = (SymbolTableFunction) gList.elementAt(i);
			text += "<a href=\"spec_function.html#"
				+ g.funID
				+ "\" target=\"righttop\">"
				+ g.funName
				+ "</a><br />\n";
		}

		text += "</body>";

		return text;
	}

	/**
	 * make index.html
	 * 
	 */
	private String makeIndexHTML() {

		String text = "";

		text += " <frameset cols=\"30%,*\">\n";
		text += "  <frameset rows=\"30%,*\">\n";
		text += "   <frame src=\"contents.html\" name=\"lefttop\" />\n";
		text += "   <frame src=\"list_function.html\" name=\"leftbottom\" />\n";
		text += "  </frameset>\n";
		text += "  <frame src=\"source.html\" name=\"righttop\" />\n";
		text += " </frameset>\n";

		return text;
	}

	/**
	 * make contents.html
	 * @param outputDirName
	 */
	private String makeContentsHTML(SymbolTable st) {

		String text = "";

		text += " <body>";
		text += " <p><strong>Contents</strong></p>";
		text += "  <a href=\"source.html\" target=\"righttop\">"
			+ st.sourceName
			+ "</a><br />";
		text += "<hr/>\n";
		text
			+= "  <a href=\"list_function.html\" target=\"leftbottom\">Function ("
			+ st.STFunction.size()
			+ ")</a><br />\n";
		text
			+= "  <a href=\"list_grobalVar.html\" target=\"leftbottom\">GrobalVariable ("
			+ st.STGrobalVar.size()
			+ ")</a><br />\n";
		int numOfLocalVar = 0;
		for (int i = 0; i < st.STFunction.size(); i++) {
			SymbolTableFunction stf =
				(SymbolTableFunction) st.STFunction.elementAt(i);
			numOfLocalVar += stf.STLocalVar.size();
		}
		text
			+= "  <a href=\"list_localVar.html\" target=\"leftbottom\">LocalVariable ("
			+ numOfLocalVar
			+ ")</a><br />\n";
		text += " </body>";

		return text;

	}

	/**
	 * make Source Program with more information
	 * 
	 * @param st
	 */
	private String makeSourceHTMLWithLink(SymbolTable st) {

		int countLineNum = 1;
		final int lineNumSpace = 4;
		String text = "";

		text += "<body><pre>   1: ";

		for (int i = 0; i < st.tokenList.size(); i++) {
			SpdElement token = (SpdElement) st.tokenList.getElementAt(i);

			if (token.getName().equals("nl")) { // add linenumber
				int figureOfCurrentLine =
					String.valueOf(countLineNum + 1).length();
				text += token.getText();
				for (int j = 0; j < lineNumSpace - figureOfCurrentLine; j++) {
					text += " ";
				}
				text += ++countLineNum + ": ";
			} else if (
				token.getName().equals("comment")
					&& token.getSort().equals("Multi")) { // addlinenumber
				String[] strArray = token.getText().split("\n");
				for (int j = 0; j < strArray.length; j++) {
					text += checkStr(strArray[j]);
					if (j < strArray.length - 1) {
						text += "\n";
						int figureOfCurrentLine =
							String.valueOf(countLineNum + 1).length();
						for (int k = 0;
							k < lineNumSpace - figureOfCurrentLine;
							k++) {
							text += " ";
						}
						text += ++countLineNum + ": ";
					}
				}
			} else if (token.getName().equals("ident")) {
				boolean isAlreadyLinked = false;
				SearchSymbolTable sst = new SearchSymbolTable(st);
				int linkID = -1;
				// search function
				if (!isAlreadyLinked) {
					linkID =
						sst.searchFunction(Integer.parseInt(token.getID()));
					if (linkID > 0) {
						text += "<a name=\""
							+ token.getID()
							+ "\">"
							+ "<a href=\"./spec_function.html#"
							+ linkID
							+ "\">"
							+ token.getText()
							+ "</a>"
							+ "</a>";
						isAlreadyLinked = true;
					}
				}
				// search grobalVar
				if (!isAlreadyLinked) {
					linkID =
						sst.searchGrobalVar(Integer.parseInt(token.getID()));
					if (linkID > 0) {
						text += "<a name=\""
							+ token.getID()
							+ "\">"
							+ "<a href=\"./spec_grobalVar.html#"
							+ linkID
							+ "\">"
							+ token.getText()
							+ "</a>"
							+ "</a>";
						isAlreadyLinked = true;
					}
				}
				// search LocalVar
				if (!isAlreadyLinked) {
					linkID =
						sst.searchLocalVar(Integer.parseInt(token.getID()));
					if (linkID > 0) {
						text += "<a name=\""
							+ token.getID()
							+ "\">"
							+ "<a href=\"./spec_localVar.html#"
							+ linkID
							+ "\">"
							+ token.getText()
							+ "</a>"
							+ "</a>";
						isAlreadyLinked = true;
					}
				}

				if (!isAlreadyLinked) {
					text += "<a name=\""
						+ token.getID()
						+ "\">"
						+ checkStr(token.getText())
						+ "</a>";
				}
			} else {
				text += checkStr(token.getText());
			}
		}

		text += "</pre></body>";

		// bottom margin
		for (int i = 0; i < marginLineNum; i++) {
			text += "<br />";
		}

		return text;
	}

	/**
	 * Transform HTML Escape Char
	 * @param str
	 * @return
	 */
	private String checkStr(String str) {
		str = str.replaceAll("&", "&amp;");
		str = str.replaceAll("<", "&lt;");
		str = str.replaceAll(">", "&gt;");
		str = str.replaceAll("\"", "&quot;");
		str = str.replaceAll("\'", "&apos;");
		return str;
	}
}
