/*  
 * Program:             $RCSfile: XRef.java,v $  $Revision: 50.6 $
 * 
 * Purpose:             Cross referencer.
 * 
 * Author:              S.Takeshi  2003/09/06
 *                      S.Yamamoto  2003/09/06
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: XRef.java,v 50.6 2003/10/12 05:38:04 yamamoto Exp yamamoto $
 */

/**
 * Cross referencer.
 * 
 * @author T.Shikimachi
 */

package wapid.jsxmodel.xref;

import java.io.*;

import gnu.getopt.*;
import sapid.util.*;

public class XRef {
	private static String file_encoding = null; // Default encoding.
	private static String output_dir_name = ""; // Default output Directory.
	private static String inputfilename = null;
	private static String default_output_dir_name = "XREF-JS";

	/**
	 * Entry point of Cross Referencer
	 * @param args
	 */
	public static void main(String args[]) {
		
		XRef xr = new XRef();
		SymbolTable st = new SymbolTable();
		
		// Initialize Sapid Message Utility
		//SpdMsg.setProjectName("Sapid");
		SpdMsg.setProgramName("JSX: XRef");
		
		// check options
		if (args.length < 1) { // less parameter
			SpdMsg.error("Invalid command");
			printHelp();
			return;
		}
		xr.check_options(args, args.length);
		
		// check file_encoding
		if(file_encoding == null){
            try{
                file_encoding = SpdEncoding.getEncoding(inputfilename).toString();
            }
            catch(FileNotFoundException e){
                SpdMsg.errorExit("Cannot find file[" + inputfilename + "].",e);
            }
		}
		
		// Tester.
		JavaScriptXSDMLAnalyzer jsxa = new JavaScriptXSDMLAnalyzer(inputfilename);
		st = jsxa.JSXAnalyzer();
		SpdMsg.verbose("Analyzing XML file");

		// TODO get sourcefile name
		st.sourceName = inputfilename.replaceAll(".xml","");
		/* generate XHTML sources */
		XHTMLSourceGenerator sg = new XHTMLSourceGenerator(output_dir_name,file_encoding);
		sg.generateXHTMLSource(st);
		SpdMsg.verbose("Finished!");

	}
	
	/** 
	 * Check Options. 
	 * @param args
	 * @param argc
	 */
	private void check_options(String args[], int argc) {
		int ch;
		boolean out_dir_is_decided = false;
		Getopt g = new Getopt("Sapid: JSX", args, "e:D:vh");

		while ((ch = g.getopt()) != -1) {
			switch (ch) {
				case 'e' : // encoding for outputfile
					file_encoding = g.getOptarg();
					break;
				case 'D' : // output Directory
					output_dir_name = g.getOptarg();
					out_dir_is_decided = true;
					break;
				case 'v' : // verbose mode
					SpdMsg.setVerboseMode(true);
					break;
				case 'h' :
					printHelp();
					System.exit(0);
				default :
					printHelp();
					System.exit(1);
					break;
			}
		}
		// get inputfilename and outputdirname
		inputfilename = new String(args[g.getOptind() + 0]);
		if (!out_dir_is_decided) {
			String[] tmp = inputfilename.split("/");
			if (tmp.length == 1) { // if inputfile in current dir
				output_dir_name = default_output_dir_name;
			} else {
				for (int i = 0; i < tmp.length - 1; i++) {
					if(i==tmp.length-2){
						output_dir_name += tmp[i]+"/"+default_output_dir_name;
					}else{
						output_dir_name += tmp[i] + "/";
					}
				}
			}
		}
	}
	
	/**
	 * Print Help Message
	 */
	public static void printHelp(){
		SpdMsg.println("NAME");
		SpdMsg.println("   XRef - Cross Referencer for JSXModel");
		SpdMsg.println("SYNOPSIS");
		SpdMsg.println("   java wapid.jsxmodel.xref.XRef [OPTION] SOURCE");
		SpdMsg.println("DESCRIPTION");
		SpdMsg.println("   -h                : Show this help.");
		SpdMsg.println("   -D                : Specify Output Directory. (Default : \"./XREF-JS\")");
		SpdMsg.println("   -e                : Specify Output File Encoding. (Default : same of source file)");
		SpdMsg.println("   -v                : Print verbose message.");
	}
}
