/*
 * Created on 2003/06/07
 * XSDMLElement.java - Copyright (c) 2003 H.Yoshida
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS IS FREEWARE. NO WARRANTY.
 */

package wapid.jsxmodel.xsapid;

import org.w3c.dom.*;

/**
 * XSDML Element Class
 * @author H.Yoshida
 */
public class SpdElement {
	/**
	 * The attached XML element object.
	 * @see http://java.sun.com/j2se/1.4/ja/docs/ja/api/org/w3c/dom/Element.html
	 */
	public Element domelm = null;

	/**
	 * The constructor.
	 * Attach a XML element.
	 */
	public SpdElement(Element elm) {
		domelm = elm;
	}

	/**
	 * Search for child elements.
	 * @return xsapid.XSDMLElementList
	 */
	public SpdElementList searchChild(SpdElementFilter f) {
		SpdElementList el = new SpdElementList();

		NodeList nl = domelm.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element) n);
				if (f != null) {
					if (f.isAcceptable(e))
						el.addElement(e);
				} else
					el.addElement(e);
			}
		}

		return el;
	}

	/**
	 * Search for decendant elements.
	 * @return xsapid.XSDMLElementList
	 */
	public SpdElementList searchDecendant(SpdElementFilter f) {
		SpdElementList el = new SpdElementList();

		searchDecendantR(el, f); // recursive
		return el;
	}
	private SpdElementList searchDecendantR(
		SpdElementList el,
		SpdElementFilter f) {

		NodeList nl = domelm.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element) n);
				if (f != null) {
					if (f.isAcceptable(e))
						el.addElement(e);
				} else
					el.addElement(e);
				e.searchDecendantR(el, f); // recursive
			}
		}

		return el;
	}

	/**
	 * Search for decendant elements for some filters.
	 * @return xsapid.XSDMLElementList
	 */
	public SpdElementList searchDecendantMulti(SpdElementFilterList fl) {
		SpdElementList el = new SpdElementList();

		searchDecendantRMulti(el, fl); // recursive
		return el;
	}
	private SpdElementList searchDecendantRMulti(
		SpdElementList el,
		SpdElementFilterList fl) {

		NodeList nl = domelm.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element) n);
				// check for each filter
				for (int j = 0; j < fl.size(); j++) {
					SpdElementFilter f = fl.getElementAt(j);

					if (f != null) {
						if (f.isAcceptable(e))
							el.addElement(e);
					} else
						el.addElement(e);
				}
				e.searchDecendantRMulti(el, fl); // recursive
			}
		}

		return el;
	}

	public SpdElement get1stChild(SpdElementFilter f) {
		return searchChild(f).getElementAt(0);
	}
	public SpdElement get2ndChild(SpdElementFilter f) {
		return searchChild(f).getElementAt(1);
	}
	public SpdElement get3rdChild(SpdElementFilter f) {
		return searchChild(f).getElementAt(2);
	}

	public SpdElement getIndexedChild(SpdElementFilter f, int index) {
		return searchChild(f).getElementAt(index - 1);
	}

	/**
	 * Get element name.
	 * @return
	 */
	public String getName() {
		return domelm.getTagName();
	}

	/**
	 * Get 'id' attribute value.
	 * @return
	 */
	public String getID() {
		return domelm.getAttribute("id");
	}

	/**
	 * Get 'sort' attribute value.
	 * @return
	 */
	public String getSort() {
		return domelm.getAttribute("sort");
	}

	/**
	 * Get column counter.
	 * @return
	 */
	public int getColumnNum() {
		String[] posa = getPositionString().split(" ");
		int c = Integer.parseInt(posa[0]);
		return c;
	}

	/**
	 * Get line counter.
	 * @return
	 */
	public int getLineNum() {
		String[] posa = getPositionString().split(" ");
		int l = Integer.parseInt(posa[1]);
		return l;
	}

	/**
	 * Get position string.
	 * @return
	 */
	public String getPositionString() {
		String val = domelm.getAttribute("pos");
		if (!val.equals(""))
			return val;

		NodeList nl = domelm.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType() != Node.ELEMENT_NODE)
				continue;
			Element e = (Element) n;
			String ret = new SpdElement(e).getPositionString();
			if (!ret.equals(""))
				return ret;
		}
		return "";
	}

	/**
	 * Get text representation.
	 * @return
	 */
	public String getText() {
		String ret = "";
		NodeList nl = domelm.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element) n;
				ret += new SpdElement(e).getText();
			}
			if (n.getNodeType() == Node.TEXT_NODE) {
				ret += n.getNodeValue();
			}
		}
		return ret;
	}
}
