/*  
 * Program:             $RCSfile: AbstractSimpleNode.java,v $  $Revision: 50.6 $
 * 
 * Purpose:             Extention of SimpleNode.
 * 
 * Author:              S.Takeshi  2003/08/03
 *                      S.Yamamoto  2003/08/03
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: AbstractSimpleNode.java,v 50.6 2003/08/12 06:09:36 yamamoto Exp yamamoto $
 */

package wapid.parser.js;

import java.util.*;
import wapid.jsxmodel.JSXTComment;
import wapid.jsxmodel.JSXTEndOfLine;
import wapid.jsxmodel.JSXTSpace;

abstract public class AbstractSimpleNode {
    /** for position */
    protected int line = 0;
    protected int column = 0;

    /** Token of Node (first) */
    protected Token firstToken;

    /** Token of Node (last) */
    protected Token lastToken;

    /**
     * set line number
     * @param line
     */
    public void setLineNumber(int line) {
        this.line = line;
    }
    /**
     * get line number
     * @return line
     */
    public int getLineNumber() {
        return line;
    }

    /** 
     * set column number 
     * @param column
     */
    public void setColumnNumber(int column) {
        this.column = column;
    }
    /**
     * get column number
     * @return column
     */
    public int getColumnNumber() {
        return column;
    }

    /**
     * set firstToken
     * @param firstToken
     */
    public void setFirstToken(Token firstToken) {
        this.firstToken = firstToken;
    }
    /**
     * get firstToken
     * @return firstToken
     */
    public Token getFirstToken() {
        return firstToken;
    }

    /**
     * set lastToken
     * @param lastToken
     */
    public void setLastToken(Token lastToken) {
        this.lastToken = lastToken;
    }
    /**
     * get lastToken
     * @return lastToken
     */
    public Token getLastToken() {
        return lastToken;
    }

    /**
     * get special token
     */
    public Token getSpecialToken() {
        return firstToken.specialToken;
    }

    /**
     * Node have CommentToken or not.
     */
    public Vector checkSpecialToken() {
        final int   OUT_COMMENT = 0;
        final int   IN_COMMENT = 1;
        Vector      v = new Vector();
	
        if (firstToken.specialToken != null) {
            Token tok = firstToken.specialToken;
            // trace specialtoken
            while (tok.specialToken != null) {
                tok = tok.specialToken;
            }
            while (tok != null) {
                if (tok.kind == EcmaScriptConstants.SINGLE_LINE_COMMENT) {
                    Token fc = tok, tmp = tok, lc = tok;
                    while (tok.kind == EcmaScriptConstants.IN_SINGLE_LINE_COMMENT) {
                        lc = tok;
                        tok = tok.next;
                    }
                    JSXTComment sc = new JSXTComment(OUT_COMMENT, fc, lc,
						     JSXTComment.SINGLE);
                    v.addElement(sc);
                } else if (
                    tok.kind == EcmaScriptConstants.MULTI_LINE_COMMENT) {
                    Token fc = tok;
                    while (tok.kind == EcmaScriptConstants.IN_MULTI_LINE_COMMENT) {
                        tok = tok.next;
                    }
                    JSXTComment sc = new JSXTComment(OUT_COMMENT, fc, tok,
						     JSXTComment.MULTI);
                    v.add(sc);
                } else if (tok.kind == EcmaScriptConstants.EOL) {
                    JSXTEndOfLine eol = new JSXTEndOfLine(OUT_COMMENT, tok, tok);
                    v.add(eol);
                } else if (tok.kind == EcmaScriptConstants.SPACE) {
                    JSXTSpace sp = new JSXTSpace(OUT_COMMENT, tok, tok);
                    v.add(sp);
                }
                tok = tok.next;
            }
        }
        return v;
    }
}
