/*  
 * Program:             $RCSfile: MakeJSX.java,v $  $Revision: 50.13 $
 * 
 * Purpose:             MakeJSX.
 * 
 * Author:              S.Takeshi  2003/08/03
 *                      S.Yamamoto  2003/08/03
 * 
 * (C) Copyright:       S.Takeshi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: MakeJSX.java,v 50.13 2003/10/12 05:37:41 yamamoto Exp yamamoto $
 */

package wapid.parser.js;

import java.io.*;
import java.util.*;
import gnu.getopt.*;
import wapid.jsxmodel.*;
import sapid.util.*;

public class MakeJSX {

	private static String file_encoding = null; // Default encoding.
	private static String output_dir_name = ""; // Default output Directory.
	private static String inputfilename = null;

	/* Parser entry point. */
	public static void main(String args[]) {

		MakeJSX mj = new MakeJSX();

		// Initialize Sapid Message Utility
		//SpdMsg.setProjectName("Sapid");
		SpdMsg.setProgramName("JSX");

		/* check options */
		if (args.length < 1) { // less parameter
			SpdMsg.error("Invalid command");
			printHelp();
			return;
		}
		mj.check_options(args, args.length);
		
		/* preserve file encoding */
		if(file_encoding == null){
            try{
                file_encoding = SpdEncoding.getEncoding(inputfilename).toString();
            }
            catch(FileNotFoundException e){
                SpdMsg.errorExit("Cannot find file[" + inputfilename + "].",e);
            }
            
		}

		try {
			EcmaScript parser = null;
			FileInputStream fis = new FileInputStream(inputfilename);
			InputStreamReader isr = new InputStreamReader(fis, "JISAutoDetect");

			SpdMsg.verbose(
				"Parser: Reading from file " + "[ " + inputfilename + " ]");

			parser = new EcmaScript(isr);

			// Parsing and make AST.
			SimpleNode n = parser.Program();
			SpdMsg.verbose("Parser: Parsed correctly");

			// Close FileInputStream.
			fis.close();

			// Generate JSXT from AST.
			JSXTGenerator v = new JSXTGenerator();
			n.jjtAccept(v, null);
			SpdMsg.verbose("TreeGenerator: Created JSXT");

			// Create XML source.
			XMLSourceGenerator xsg = new XMLSourceGenerator(file_encoding);

			/* make output directory (If it isn't exist.) */
			if (output_dir_name != null) {
				File output_dir = new File(output_dir_name);
				if (!output_dir.exists() || !output_dir.isDirectory()) {
					output_dir.mkdirs();
					SpdMsg.verbose(
						"  --> Create Directory "
							+ "[ "
							+ output_dir_name
							+ " ]");
				}
			}

			// Output xml file.
			FileOutputStream fos =
				new FileOutputStream(
					mj.checkOutputFileName(output_dir_name, inputfilename)
						+ ".xml");
			OutputStreamWriter osw = new OutputStreamWriter(fos, file_encoding);
			Vector source = xsg.generateSource(v.nodes);
			for (int i = 0; i < source.size(); i++) {
				osw.write((String) source.elementAt(i));
				osw.flush();
			}
			fos.close();

			SpdMsg.verbose("XMLGenerator: Created XML repository ...");
			SpdMsg.verbose("  --> Output Dir  [ " + output_dir_name + " ]");
			SpdMsg.verbose(
				"  --> Output File [ "
					+ mj.getOnlyFileName(inputfilename)
					+ ".xml ]");

			// copy "JSX-model.dtd" to output directory.
			mj.copyDTD(output_dir_name);

			SpdMsg.verbose("Finished!");

		} catch (FileNotFoundException e) {
			SpdMsg.errorExit(
				"Parser: File " + inputfilename + " not found.",
				e);
		} catch (UnsupportedEncodingException e) {
			SpdMsg.errorExit("Parser: Unsupported Encode.", e);
		} catch (IOException e) {
			SpdMsg.errorExit("", e);
		} catch (ParseException e) {
			System.err.println(e.getMessage());
			SpdMsg.errorExit("Parser: Encountered parse errors.", e);
		} catch (Exception e) {
			SpdMsg.errorExit("Parser: ", e);
		}
	}

	/**
	 * copy DTD file
	 * @param output_dir_name
	 */
	private void copyDTD(String output_dir_name) {
		InputStream inStream =
			getClass().getResourceAsStream(
				"/wapid/parser/js/resources/JSX-model.dtd");
		FileOutputStream foutstream = null;
		try {
			foutstream =
				new FileOutputStream(
					output_dir_name
						+ System.getProperty("file.separator")
						+ "JSX-model.dtd");
			for (;;) {
				byte[] byteRead = new byte[1024];
				int iRead = inStream.read(byteRead);
				if (iRead == (-1))
					break;
				foutstream.write(byteRead, 0, iRead);
			}
			inStream.close();
			foutstream.close();
		} catch (FileNotFoundException e) {
			SpdMsg.errorExit("DTD File not found. ", e);
		} catch (Exception e) {
			SpdMsg.errorExit("", e);
		}
	}

	/**
	 * check outputfile name
	 * @param inputfilename
	 * @return 
	 */
	private String checkOutputFileName(
		String output_dir_name,
		String inputfilename) {

		if (output_dir_name == null) {
			return inputfilename;
		} else {
			return output_dir_name
				+ System.getProperty("file.separator")
				+ getOnlyFileName(inputfilename);
		}
	}

	/**
	 * @param inputfilename
	 * @return
	 */
	private String getOnlyFileName(String inputfilename) {
		String[] filename;
		filename = inputfilename.split("/");
		if (filename.length > 1) {
			return filename[filename.length - 1];
		} else {
			return inputfilename;
		}
	}

	/** 
	 * Check Options. 
	 * @param args
	 * @param argc
	 */
	private void check_options(String args[], int argc) {
		int ch;
		boolean out_dir_is_decided = false;
		Getopt g = new Getopt("Sapid: JSX", args, "e:D:vh");

		while ((ch = g.getopt()) != -1) {
			switch (ch) {
				case 'e' : // encoding for outputfile
					file_encoding = g.getOptarg();
					break;
				case 'D' : // output Directory
					output_dir_name = g.getOptarg();
					out_dir_is_decided = true;
					break;
				case 'v' :
					SpdMsg.setVerboseMode(true);
					break;
				case 'h' :
					printHelp();
					System.exit(0);
				default :
					printHelp();
					System.exit(1);
					break;
			}
		}
		// get inputfilename and outputdirname
		inputfilename = new String(args[g.getOptind() + 0]);
		if (!out_dir_is_decided) {
			String[] tmp = inputfilename.split("/");
			if (tmp.length == 1) {
				// if inputfile in current directory
				//  set output directory to current directory.
				output_dir_name = "./";
			} else {
				for (int i = 0; i < tmp.length - 1; i++) {
					output_dir_name += tmp[i] + "/";
				}
			}
		}
	}

	/**
	 * Print Help Message
	 */
	public static void printHelp() {
		SpdMsg.println("NAME");
		SpdMsg.println("   MakeJSXmodel - JSXModel maker");
		SpdMsg.println("SYNOPSIS");
		SpdMsg.println("   java wapid.parser.js.MakeJSX [OPTION] JavaScriptSource");
		SpdMsg.println("DESCRIPTION");
		SpdMsg.println("   -h                : Show this help.");
		SpdMsg.println("   -D                : Specify Output Directory. (Default : \"./\" current)");
		SpdMsg.println("   -e                : Specify Output File Encoding.");
		SpdMsg.println("   -v                : Print verbose message.");
	}
}
