/* 
 * Program:             $RCSfile: WXModel.java,v $
 * 
 * Purpose:             
 * 
 * Author:              K.Niimi  2003/10/11
 *                      S.Yamamoto  2003/10/11
 * 
 * (C) Copyright:       K.Niimi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: WXModel.java,v 50.3 2003/10/14 14:56:42 yamamoto Exp yamamoto $
 */

package wapid.parser.wxmodel;

import java.io.*;
import sapid.util.*;

/**
 * WXModel ɽ饹
 *
 * @author Ken Niimi.
 */

public class WXModel
{
    /** Ϥե̾ݻ */
    private String fileName = null;

    /** SDBѤΥ饹 */
    private WXDBAccessor acc = null;

    /** ΥեΥ󥳡ǥ */
    private String encode = null;

    /** ΥեΥ饤󥻥ѥ졼 */
    private String lineSeparator = null;
    
    /** ɤ߹ХȤ礭 */
    private static final int READ_BYTE_LEN = 4096;
    
    /** 󥹥ȥ饯 */
    public WXModel(String fileName){
        this.fileName = fileName;
        acc = new WXDBAccessor();
        try{
            encode = SpdEncoding.getEncoding(fileName).toString();
            lineSeparator = SpdLineSeparator.getLineSeparator(fileName).toString();
        }
        catch(java.io.FileNotFoundException e){
            SpdMsg.errorExit("Cannot find file[" + fileName + "].",e);
        }catch(Exception e){
            SpdMsg.errorExit("Error.",e);
        }
        
        SpdMsg.verbose("Encoding : " + encode);
    }

    /** ϥǥ쥯ȥ */
    public void setSDBName(String dirName){
        acc.setSDBName(dirName);
    }
    
    /** ϥ󥳡ǥ󥰤λ */
    public void setEncoding(String encode){
        this.encode = encode;
    }
    
    /** ϤԤSDBؽϤ */
    public void analyze(){
        //ѡκ
        WXModelParser parser = null;
        try{
            FileInputStream fis = new FileInputStream(fileName);
            InputStreamReader isr = new InputStreamReader(fis, "JISAutoDetect");
            parser = new WXModelParser(isr);
        }
        catch(Exception e){
            SpdMsg.errorExit("WXModel parser not made",e);
        }
        
        //WXModel 
        try{
            SimpleNode root = parser.makeASTTree();
            SpdMsg.verbose("AST made");
            MkWXModelVisitor v = new MkWXModelVisitor(encode,lineSeparator);
            root.jjtAccept(v,null);
            SpdMsg.verbose("WXModel made");
            acc.writeSDB(fileName,v.toString(),encode);
            SpdMsg.verbose("Output => [" + acc.getSDBFileName(fileName) + "]");
        }
        catch(Exception e){
            SpdMsg.errorExit("Parsing Error",e);
        }
    }

    /** ꥸʥ */
    public boolean isSameToOriginal(){
        String sdbFileName = acc.getSDBFileName(fileName);
        SpdMsg.verbose("Detagging [" + sdbFileName + "] ...");
        //ValidateFlag == true
        SpdXMLDetag detagger = new SpdXMLDetag(sdbFileName,true);
        String dbStr = detagger.deTag();
        //System.out.print(dbStr);
        byte[] dbBytes = null;
        try{
            dbBytes = dbStr.getBytes(encode);
        }
        catch(java.io.UnsupportedEncodingException e){
            SpdMsg.errorExit("UnsupportedEncodingException in compareOriginal",e);
        }

        SpdMsg.verbose("Compare to Original");
        FileInputStream srcFis = null;
        try{
            srcFis = new FileInputStream(fileName);

            byte[] orgBytes = new byte[READ_BYTE_LEN];
            int read_byte_len = 0;
            int db_byte_pos = 0;
            while( (read_byte_len = srcFis.read(orgBytes))!= -1 ){
                for(int i=0;i<read_byte_len;i++){
                    //System.out.println("" + dbBytes[db_byte_pos] +":" + orgBytes[i]);
                    if(dbBytes[db_byte_pos]!=orgBytes[i]){
                        //SpdMsg.verbose("Differ from Original");
                        //System.out.println(orgBytes);
                        return false;
                    }
                    db_byte_pos++;
                }
            }
        }
        catch(java.io.IOException e){
            SpdMsg.errorExit("Error in compareOriginal",e);
        }

        //SpdMsg.verbose("Same to Original");
        return true;
    }
}
