/* 
 * Program:             $RCSfile: PIDB_cleanUp.c,v $  $Revision: 30.1 $
 * 
 * Purpose:
 * 
 * Author:              
 *
 * Modified:            S.Yamamoto  1997/10/04
 * 
 * Note:                Modified preprocessor in GCC ver. 1.4.2.
 *
 * (C) Copyright:       S.Yamamoto  1997
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid_PIDB_cleanUp_c[] =
    "$Id: PIDB_cleanUp.c,v 30.1 1998/09/24 11:26:53 yamamoto Exp yamamoto $";
#endif

/* #define DEBUG_SAPID_DELETE_MACRO */

#ifdef USE_SAPID_DELETE_MACRO

#define SAVED_DEFN_NUM      2048

static DEFINITION   *saved_defn[SAVED_DEFN_NUM];
static int          saved_defn_n = 0;

static void         pidbReleaseHashTable(void);
static void         pidbDeleteMacro(HASHNODE *hp);
static void         pidbInsertSavedDefn(DEFINITION *defn);

/* Function delete_macro() is copied from "cccp.c" in GCC. */

/* Note that the DEFINITION of a macro is removed from the hash table
   but its storage is not freed.  This would be a storage leak
   except that it is not reasonable to keep undefining and redefining
   large numbers of macros many times.
   In any case, this is necessary, because a macro can be #undef'd
   in the middle of reading the arguments to a call to it.
   If #undef freed the DEFINITION, that would crash.  */

delete_macro (hp)
     HASHNODE *hp;
{
    if (hp->prev != NULL)
        hp->prev->next = hp->next;
    if (hp->next != NULL)
        hp->next->prev = hp->prev;

    /* make sure that the bucket chain header that
       the deleted guy was on points to the right thing afterwards. */
    if (hp == *hp->bucket_hdr)
        *hp->bucket_hdr = hp->next;

    if (hp->type == T_MACRO) {
        pidbInsertSavedDefn(hp->value.defn);
    }

#if 0
    if (hp->type == T_MACRO) {
        DEFINITION *d = hp->value.defn;
        struct reflist *ap, *nextap;

        for (ap = d->pattern; ap != NULL; ap = nextap) {
            nextap = ap->next;
            spdFree(ap);
        }
        spdFree(d);
    }
#endif
    spdFree(hp);
}

static void pidbInsertSavedDefn(DEFINITION *defn)
{
    if (saved_defn_n < SAVED_DEFN_NUM) {
        saved_defn[saved_defn_n++] = defn;
    } else {
        spdWarning("pidbInsertSavedDefn(): Too many saved macros.\n" );
    }
}

void pidbDeleteMacro(HASHNODE *hp)
{

    if (hp->prev != NULL)
        hp->prev->next = hp->next;
    if (hp->next != NULL)
        hp->next->prev = hp->prev;

    /* make sure that the bucket chain header that
       the deleted guy was on points to the right thing afterwards. */
    if (hp == *hp->bucket_hdr)
        *hp->bucket_hdr = hp->next;

    if (hp->type == T_MACRO) {
        DEFINITION      *d = hp->value.defn;
        struct reflist  *ap, *nextap;

        for (ap = d->pattern; ap != NULL; ap = nextap) {
            nextap = ap->next;
            spdFree(ap);
        }
        spdFree(d);
    }

    spdFree(hp);
}

static void pidbReleaseHashTable(void)
{
    int     i;

    /* 
     * spdMessage("pidbReleaseHashTable(): saved_defn_n %d.\n",
     *            saved_defn_n);
     */

    for (i = 0; i < saved_defn_n; i++) {
        DEFINITION      *d = saved_defn[i];
        struct reflist  *ap, *nextap;

        for (ap = d->pattern; ap != NULL; ap = nextap) {
            nextap = ap->next;
            spdSafetyFree(ap);
        }
        spdSafetyFree(d);
    }

#ifdef COMMENTED

    int         bucket;
    HASHNODE    *hp, *next_hp;

    for (bucket = 0; bucket < HASHSIZE; bucket++) {
        for (hp = hashtab[bucket]; hp != NULL; hp= next_hp) {
            next_hp = hp->next;
            
#ifdef DEBUG_SAPID_DELETE_MACRO
            if (hp->type == T_MACRO) {
                DEFINITION  *defn = hp->value.defn;

                spdMessage("pidbReleaseHashTable(): \"%s\".\n",
                           defn->expansion);
            }
#endif

            pidbDeleteMacro(hp);
        }
    }

#endif /* COMMENTED */
}

#endif /* USE_SAPID_DELETE_MACRO */
