/* 
 * Program:             $RCSfile: DebugSyntaxTree.c,v $  $Revision: 40.11 $
 * 
 * Purpose:             Functions for debugging.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: DebugSyntaxTree.c,v 40.11 2002/12/28 06:49:20 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "JB.h"
#include "FuncProto.h"

#ifdef YYDEBUG

void putFile(SymData *file, char *msg)
{
    SymData *tmp;

    if (file == NULL || file->symSort != SPD_CID_J_FILE) {
        return;
    }

    jpdDebugMessage(TRUE, "%s FILE %s:\n", msg, file->str); 

    for (tmp = file->u.file->classOrInterface; tmp != NULL;) {
        if (tmp->symSort == SPD_CID_J_CLASS) {
            putClass(tmp, "");
            tmp = tmp->u.jclass->nextClassOrInterface;
        } else if (tmp->symSort == SPD_CID_J_INTERFACE) {
            putInterface(tmp, "");
            tmp = tmp->u.interface->nextClassOrInterface;
        } else {
            spdErrorExit("\"%s\" (%n) is nether class nor interface.\n",
                         tmp->str, tmp->symSort);
        }
    }
    
    if (file->u.file->definedPackage != NULL) {
        putQName(file->u.file->definedPackage, "Defined Package");
    }

    for (tmp = file->u.file->importedPackage;
         tmp != NULL; tmp = tmp->u.qname->nextQName) {
        putQName(tmp, "Imported Package");
    }
}

void putModifiers(SymData *symData, char *msg)
{

    if (symData == NULL || symData->symSort != SPD_CID_J_MODIFIER) {
        return;
    }
    jpdDebugMessage(TRUE, "%s Modifier: ", msg);
    while (symData->u.modifier->next != NULL) {
        jpdDebugMessage(TRUE, "%s ", symData->str);
        symData = symData->u.modifier->next;
    }
    jpdDebugMessage(TRUE, "%s \n", symData->str);   
}

void putType(SymData *type, char *msg)
{
    if (type == NULL || type->symSort != SPD_CID_J_TYPE) {
        jpdDebugMessage(TRUE, "%s Type: NO TYPE DEFINED\n", msg);
        return;
    }
    jpdDebugMessage(TRUE, "%s Type: ", msg);
    while (type->u.type->next != NULL) {
        jpdDebugMessage(TRUE, "%s ", type->str);
        type = type->u.type->next;
    }
    jpdDebugMessage(TRUE, "%s \n", type->str);
    putQName(type->u.type->qname, "Type's QName");

    if (type->u.type->nextType != NULL) {
        putType(type->u.type->nextType, msg);
    }
}

void putVariableDecl(SymData *vars, char *msg)
{
    SymData *var;

    if (vars == NULL) {
        return;
    }
    jpdDebugMessage(TRUE, "%s Vairable Decl:\n", msg);
    for (var = vars; var != NULL;
         var = var->u.variable->next) {
        putVariable(var, " ");
        if (var->u.variable->type != NULL) {
            putType(var->u.variable->type, "  ");
        }
        if (var->u.variable->modifier != NULL) {
            putModifiers(var->u.variable->modifier, "  ");
        }
        if (var->u.variable->initializer != NULL) {
            putExpression(var->u.variable->initializer, "INITIALIZED ");
        }
    }
}

void putVariable(SymData *var, char *msg)
{
    char *sortStr;

    if (var == NULL || var->symSort != SPD_CID_J_VARIABLE) {
        return;
    }
    switch(var->u.expression->sort) {
      case CLASS_MEM:
        sortStr = "Class Member";
        break;
      case INTERFACE_MEM:
        sortStr = "Interface Member";
        break;
      case PARAMETER:
        sortStr =  "Parameter";
        break;
      case LOCAL_VAR:
        sortStr = "Local Variable";
        break;
    }
    jpdDebugMessage(TRUE, "%s Var Name: %s, Sort: %s\n",
                    msg, var->str, sortStr);
}

void putLiteral(SymData *symData, char *msg)
{
    char literalStr[1024];
    
    if (symData == NULL || symData->symSort != SPD_CID_J_LITERAL) {
        jpdDebugMessage(TRUE, "%sLiteral: NO LITERAL DEFINED\n", msg);
        return;
    }
    switch (symData->u.literal->sort) {
      case L_CHAR:
        sprintf(literalStr, "Char value: %c", symData->u.literal->value.c);
        break;
      case L_STRING:
        sprintf(literalStr, "String value: %s",
                symData->u.literal->value.s);
        break;
      case L_UNICODE:
        sprintf(literalStr, "unicode value: %c",
                symData->u.literal->value.c);
        break;
      case L_INT:
        sprintf(literalStr, "int value: %d", symData->u.literal->value.i);
        break;
      case L_DOUBLE:
        sprintf(literalStr, "float value: %f",
                symData->u.literal->value.f);
        break;
      default:
        spdErrorExit("Unknown sort: %d.\n", symData->u.literal->sort);
        return;
    }
    jpdDebugMessage(TRUE, "%s Literal: %s\n", msg, literalStr);
}

void putExpression(SymData *expr, char *msg)
{
    SymData     *tmp;

    if (expr == NULL || expr->symSort != SPD_CID_J_EXPRESSION) {
        return;
    }

    jpdDebugMessage(TRUE, "%s Expression: ", msg);

    switch (expr->u.expression->sort) {
      case E_EMPTY:
        jpdDebugMessage(TRUE, "Empty Expression");
        break;
      case E_REF_VAR:
        jpdDebugMessage(TRUE, "Ref Var %s: ", expr->str);
        putVariable(expr->u.expression->first, "defined ");
        break;
      case E_REF_ARRAY:
        jpdDebugMessage(TRUE, "Ref Array\n");
        break;
      case E_REF_LITERAL:
        putLiteral(expr->u.expression->first, "Ref Literal ");
        break;
      case E_REF_TYPE:
        putType(expr->u.expression->first, "Ref Type ");
        break;
      case E_REF_SPECIAL:
        jpdDebugMessage(TRUE, "Ref Special: %s\n",
                        expr->u.expression->first->str);

        break;
      case E_METHOD_CALL:
        jpdDebugMessage(TRUE, "Invoke Method: %s\n",
                        expr->str);
        for (tmp = expr->u.expression->next;
             tmp != NULL; tmp = tmp->u.expression->next) {
            putExpression(tmp, "  Argument");
        }
        break;
      case E_OPERATION:
        jpdDebugMessage(TRUE, "Operation: %s\n",
                        expr->u.expression->joperator->str);
        if (expr->u.expression->first != NULL) {
            putExpression(expr->u.expression->first, "1st: ");
        }
        if (expr->u.expression->second != NULL) {
            putExpression(expr->u.expression->second, "2nd: ");
        }
        if (expr->u.expression->third != NULL) {
            putExpression(expr->u.expression->third, "3rd: ");
        }
        fprintf(stderr, "\n");
        break;
      case E_REF_QNAME:
        putQName(expr->u.expression->first, "Ref QName ");
        break;
      default:
        spdErrorExit("Unknown expression: %d (l.%d, c.%d).\n",
                     expr->u.expression->sort, yylineno, c_count);
    }
}

void putStatement(SymData *stmt, char *msg)
{
    char        str[256], str2[128];
    SymData     *tmp;
    int         i;

    if (stmt == NULL || stmt->symSort != SPD_CID_J_STATEMENT) {
        return;
    }
    
    sprintf(str, "%s Statement: ", msg);
    
    switch (stmt->u.statement->sort) {
      case S_EMPTY:
        jpdDebugMessage(TRUE, "%s EMPTY\n", str);
        break;
      case S_EXPRESSION:
        jpdDebugMessage(TRUE, "%s Expression\n", str);
        putExpression(getStatementExpression(stmt), "");
        break;
      case S_DEF_VAR:
        jpdDebugMessage(TRUE, "%s DEF_VAR\n", str);
        putVariableDecl(getStatementVariableDecl(stmt), "  ");
        break;
      case S_BLOCK:
        jpdDebugMessage(TRUE, "%s BLOCK\n", str);
        for (tmp = stmt->u.statement->next, i = 1;
             tmp != NULL; tmp = tmp->u.statement->next, i++) {
            sprintf(str2, "BLOCK %d: ", i);
            putStatement(tmp, str2);
        }
        break;
      case S_IF:
        jpdDebugMessage(TRUE, "%s IF\n", str);
        putExpression(getStatementCondition(stmt), "CONDITION ");
        putStatement(getStatementIfThen(stmt), "THEN ");
        putStatement(getStatementIfElse(stmt), "ELSE ");
        break;
      case S_SWITCH:
        jpdDebugMessage(TRUE, "%s SWITCH\n", str);
        putExpression(getStatementCondition(stmt), "SWITCH EXPR ");
        putStatement(getStatementIfThen(stmt), "ACTION ");
        break;
      case S_WHILE:
        jpdDebugMessage(TRUE, "%s WHILE\n", str);
        putExpression(getStatementCondition(stmt), "CONDITION ");
        putStatement(getStatementAction(stmt), "ACTION ");
        break;
      case S_DO:
        jpdDebugMessage(TRUE, "%s DO\n", str);
        putExpression(getStatementCondition(stmt), "CONDITION ");
        putStatement(getStatementAction(stmt), "ACTION ");
        break;
      case S_FOR:
        jpdDebugMessage(TRUE, "%s FOR\n", str);
        if (getStatementForInit(stmt) != NULL) {
            if (getStatementForInit(stmt)->symSort == SPD_CID_J_STATEMENT) {
                /* ѿ */
                putStatement(getStatementForInit(stmt), "INIT ");
            } else {                          /*  */
                for (tmp = getStatementForInit(stmt);
                     tmp != NULL; tmp = tmp->u.expression->next) {
                    putExpression(tmp, "INIT ");
                }
            }
        } else {
            spdErrorExit("FOR INIT is null.\n");
        }
        putExpression(getStatementCondition(stmt), "CONDITION ");
        for (tmp = getStatementForIncr(stmt);
             tmp != NULL; tmp = tmp->u.expression->next) {
            putExpression(tmp, "POST ");
        }
        putStatement(getStatementAction(stmt), "ACTION ");
        break;
      case S_BREAK:
        jpdDebugMessage(TRUE, "%s BREAK\n", str);       
        break;
      case S_CONTINUE:
        jpdDebugMessage(TRUE, "%s CONTINUE\n", str);
        break;
      case S_RETURN:
        jpdDebugMessage(TRUE, "%s RETURN ", str);
        putExpression(getStatementExpression(stmt), "Return Value ");
        break;
      case S_THROW:
        jpdDebugMessage(TRUE, "%s THROW ", str);
        putExpression(getStatementExpression(stmt), "Throw Object ");
        break;
      case S_EXCEPTION:
        jpdDebugMessage(TRUE, "%s EXCEPTION ", str);
        putStatement(getStatementExceptionTry(stmt), "TRY ");
        putStatement(getStatementExceptionCatches(stmt), "CATCHES ");
        putStatement(getStatementExceptionFinally(stmt), "FINALLY ");
        break;
      case S_TRY:
        jpdDebugMessage(TRUE, "%s TRY ", str);
        putStatement(getStatementTryBlock(stmt), "Try Block");
        break;
      case S_CATCH:
        jpdDebugMessage(TRUE, "%s CATCH ", str);
        for (tmp = stmt; tmp != NULL; tmp = tmp->u.statement->next) {
            putVariableDecl(getStatementCatchArgument(tmp), "Catch Arg ");
            putStatement(getStatementCatchBlock(tmp), "Catch Block ");
        }
        break;
      case S_FINALLY:
        jpdDebugMessage(TRUE, "%s FINALLY ", str);
        putStatement(getStatementFinallyBlock(stmt), "Finally Block");
        break;
      case S_SYNCHRONIZED:
        jpdDebugMessage(TRUE, "%s SYNCHRONIZED ", str);
        putExpression(getStatementSynchronizedExpression(stmt),
                      "Synchronized Expression");
        putStatement(getStatementSynchronizedStatement(stmt),
                     "Synchronized Statement");
        break;
      default:
        spdErrorExit("Unknown statement: %d (l.%d, c.%d).\n",
                     stmt->u.statement->sort, yylineno, c_count);
    }
    if (hasStatementLabel(stmt)) {
        putLabel(getStatementLabel(stmt), "LABEL");
    }
}

void putLabel(SymData *label, char *msg)
{
    switch (label->u.label->sort) {
      case DEF_LABEL:
        jpdDebugMessage(TRUE, "%s DEF LABEL %s\n", msg, label->str);
        break;
      case REF_LABEL:
        jpdDebugMessage(TRUE, "%s REF LABEL %s\n", msg, label->str);
        break;
      case CASE_LABEL:
        jpdDebugMessage(TRUE, "%s CASE LABEL ", msg);
        putExpression(label->u.label->caseValue, "CASE VALUE");
        break;
      case DEFAULT_LABEL:
        jpdDebugMessage(TRUE, "%s DEFAULT LABEL ", msg);
        break;
      default:
        spdErrorExit("Unknown label: %d (l.%d, c.%d).\n",
                     label->u.label->sort, yylineno, c_count);
    }
}

void putQName(SymData *qname, char *msg)
{
    SymData *first;
    char    str[1024];

    if (qname == NULL || qname->symSort != SPD_CID_J_QNAME) {
        return;
    }

    /* sprintf(str, "%s QName: ", msg); */
    if (qname->u.qname->dot == NULL) {
        jpdDebugMessage(TRUE, "%s", qname->str);
    } else {
        putQName(qname->u.qname->first, NULL);
        jpdDebugMessage(TRUE, ".", qname->str);
        putQName(qname->u.qname->second, NULL);
    }
}

void putRoutine(SymData *routine, char *msg)
{
    if (routine == NULL || routine->symSort != SPD_CID_J_ROUTINE) {
        return;
    }
    
    if (routine->u.routine->sort == METHOD) {
        jpdDebugMessage(TRUE, "%s METHOD: %s\n", msg, routine->str);
    } else {
        jpdDebugMessage(TRUE, "%s CONSTRUCTOR: %s\n", msg, routine->str);
    }

    putModifiers(routine->u.routine->modifier, "");
    putType(routine->u.routine->returnType, "  Return Type");
    putType(routine->u.routine->throwType, "  Throw Type");
    putVariableDecl(routine->u.routine->parameter, "  Parameters");
    /*  putStatement(routine->u.routine->body, "Body");*/
}

void putClass(SymData *jclass, char *msg)
{
    SymData     *tmp;

    if (jclass == NULL || jclass->symSort != SPD_CID_J_CLASS) {
        return;
    }
    
    jpdDebugMessage(TRUE, "%s CLASS: %s\n", msg, jclass->str);
    putModifiers(jclass->u.jclass->modifier, "");
    putType(jclass->u.jclass->super, "Super class: ");
    for (tmp = jclass->u.jclass->interface;
         tmp != NULL; tmp = tmp->u.type->nextType) {
        putType(tmp, "Implements interface: ");
    }

    for (tmp = jclass->u.jclass->member; tmp != NULL;) {
        switch (tmp->symSort) {             /* Фϵսˤʤ */
          case SPD_CID_J_VARIABLE:          /* ѿ */
            putVariable(tmp, "Class Member ");
            tmp = tmp->u.variable->nextMember;
            break;
          case SPD_CID_J_ROUTINE:           /* ᥽åɡ󥹥ȥ饯 */
            putRoutine(tmp, "Routine ");
            tmp = tmp->u.routine->nextMember;
            break;
          case SPD_CID_J_SBLOCK:            /* Ū */
            jpdDebugMessage(TRUE, "Static Block \n");
            tmp = tmp->u.sblock->nextMember;
            break;
          default:
            spdErrorExit("%d: Invalid class member.\n", tmp->symSort);
        }
        
    }
}

void putInterface(SymData *interface, char *msg)
{
    SymData     *tmp;

    if (interface == NULL || interface->symSort != SPD_CID_J_INTERFACE) {
        return;
    }
    
    jpdDebugMessage(TRUE, "%s INTERFACE: %s\n", msg, interface->str);
    putModifiers(interface->u.interface->modifier, "");

    for (tmp = interface->u.interface->super;
         tmp != NULL; tmp = tmp->u.type->nextType) {
        putType(tmp, "Super : ");
    }

    for (tmp = interface->u.interface->member; tmp != NULL;) {
        switch (tmp->symSort) {             /* Фϵսˤʤ */
          case SPD_CID_J_VARIABLE:          /* ѿ */
            putVariable(tmp, "Interface Member ");
            tmp = tmp->u.variable->nextMember;
            break;
          case SPD_CID_J_ROUTINE:           /* ᥽åɡ󥹥ȥ饯 */
            putRoutine(tmp, "Routine ");
            tmp = tmp->u.routine->nextMember;
            break;
          case SPD_CID_J_SBLOCK:            /* Ū */
            jpdDebugMessage(TRUE, "Static Block \n");
            tmp = tmp->u.sblock->nextMember;
            break;
          default:
            spdErrorExit("%d: Invalid interface member.\n", tmp->symSort);
        }
    }
}

#endif /* YYDEBUG */
