/* 
 * Program:             $RCSfile: FuncProto.h,v $  $Revision: 40.9 $
 * 
 * Purpose:
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* java-lex.l */
int         yylex(void);

/* java-yacc.y */
int         yyparse(void);

/* SymTab.c */
int         addVar2MemberSymTab(SymData *symData);
SymData     *searchForMemberSymTab(char *name, VarSymTab *mst);
int         addVar2LocalSymTab(SymData *symData);
SymData     *searchForLocalSymTab(char *name, VarSymTab *lst);
int         addLabel2LabelSymTab(SymData *label, LabelSymTab *lst);
SymData     *searchForLabelSymTab(char *name, LabelSymTab *lst);
VarSymTab   *getVarSymTab(SymTabSort sort);
void        freeVarSymTab(VarSymTab *st);
void        initFirstSymTab(void);
VarSymTab   *pushMemberSymTab(void);
void        popMemberSymTab(void);
VarSymTab   *pushLocalSymTab(void);
void        popLocalSymTab(void);
LabelSymTab *getLabelSymTab(void);
void        freeLabelSymTab(LabelSymTab *lst);
void        printSymTabStack(void);

void        initLabelSymTab(void);
LabelSymTab *pushLabelSymTab(void);
void        popLabelSymTab(void);

/* SyntaxTree.c */
SymData     *setClass(SymData *jclass);
SymData     *setAnonClass(SymData *newObj, SymData *body);
SymData     *setNewObjectPrefix(SymData *newOp, SymData *typeName,
                                SymData *lp, SymData *args, SymData *rp);
SymData     *setInterface(SymData *interface);
SymData     *setRoutine(SymData *method, RoutineSort sort);
SymData     *setModifier(SymData *symData);
SymData     *setType(SymData *symData, TypeSort sort);
SymData     *setVariable(SymData *symData);
SymData     *setMemberVariable(SymData *vars, SymData *type, SymData *mod);
SymData     *setLocalVariable(SymData *vars, SymData *type, SymData *mod,
                              VarSort sort);
SymData     *setQName(SymData *symData);
SymData     *setStatement(SymData *stmt, StatementSort stmtSort);
SymData     *setExpression(SymData *expr, ExpressionSort exprSort,
                           SymData *anOperator,
                           SymData *first, SymData *second, SymData *third);
SymData     *qname2PrimaryNoNewArray(SymData *qname0);
SymData     *setConstructorCall(SymData *newObj);
SymData     *setClassLiteralExpression(SymData *type,
                                       SymData *dot,
                                       SymData *jclass);
SymData     *setSpecialName(SymData *symData);
SymData     *setOperator(SymData *symData, OperatorSort sort);
SymData     *setLabel(SymData *label, LabelSort sort);
SymData     *setCaseLabel(SymData *label, SymData *expr);
SymData     *setDefaultLabel(SymData *label);
SymData     *setSBlock(SymData *sblock, SBlockSort sort, SymData *block);
SymData     *qname2Expression(SymData *qname, LRSort lrsort);
SymData     *qname2Type(SymData *qname0);
void        setLRSortR(SymData *expr, LRSort lr);
void        lockLRSortR(SymData *expr);

SymDataStack    *setSymDataStack(SymData *symData);
SymDataStack    *setQNameStack(SymDataStack *stack, SymData *qname);
SymDataStack    *getSymDataStack(void);

SymData     *popSymDataStack(SymDataStack *stack);
SymData     *pushSymDataStack(SymDataStack *stack, SymData *elem);
int         isEmptySymDataStack(SymDataStack *stack);
SymData     *getSymData(void);
void        freeSymData(SymData *symData);
SymData     *setOffsetAndLength(SymData *symData, SymData *first, SymData *last);

/* DebugSyntaxTree.c */
void        putFile(SymData *file, char *msg);
void        putModifiers(SymData *symData, char *msg);
void        putType(SymData *type, char *msg);
void        putVariableDecl(SymData *vars, char *msg);
void        putVariable(SymData *var, char *msg);
void        putLiteral(SymData *symData, char *msg);
void        putExpression(SymData *expr, char *msg);
void        putStatement(SymData *stmt, char *msg);
void        putLabel(SymData *label, char *msg);
void        putQName(SymData *qname, char *msg);
void        putRoutine(SymData *routine, char *msg);
void        putClass(SymData *jclass, char *msg);
void        putInterface(SymData *interface, char *msg);

/* WriteDB.c */
void        writeDBClass(SymData *symData);
void        writeDBRel(char *relName, SymData *obj1, SymData *obj2);
void        writeFile(SymData *file);
void        writeClass(SymData *jclass);
void        writeInterface(SymData *interface);
void        writeVariable(SymData *vars);
void        writeRoutine(SymData *routine);
void        writeType(SymData *type);
void        writeSBlock(SymData *symData);
void        writeStatement(SymData *stmt);
void        writeExpression(SymData *expr);
void        writeLiteral(SymData *symData);
void        writeOperator(SymData *anOperator);
void        writeQName(SymData *qname);
void        writeSpecial(SymData *special);
void        writeLabel(SymData *label);
void        writeDBRelAboutVariable(char *relName, SymData *obj, SymData *vars);
void        writeDBRelAboutTypeList(char *relName, SymData *obj, SymData *type);
void        initDB(void);
void        mkSymlink(char *classname);

/* Util.c */
void        _jpdDebugMessage(int debugFlag, const char *format, ...);
