/* 
 * Program:             $RCSfile: JB.h,v $  $Revision: 40.68 $
 * 
 * Purpose:             Declarations of types and so on.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef _SPD_JB_H
#define _SPD_JB_H

extern SpdConstString   frejaDir;

/*
 * Basic values.
 */

#define YYSTYPE             SymData *

#define JAPID_PKG_INFO      ".pkg.info"

#define GET_DBCLASS(index)  (J_model_classes[(index)])
#define GET_DBREL(index)    (J_model_classes[(index) + SPD_CID_NUM_J_MODEL])

/*
 * XML values.
 */

#define JDBIN_ETB           "jdbin.etb"
#define J2JX_ETB            "j2jx.etb"
#define XML_HEADER_NAME     "header.etb"
#define JX_MODEL_DTD_FILE   "JX-model3.dtd"

#define JX_MODEL_XML_HEADER \
    "!<?xml version=\"1.0\" encoding=\"%s\"?>\n" \
    "!<!DOCTYPE File SYSTEM \"" JX_MODEL_DTD_FILE "\">\n"

/**
 * Type definitions.
 */
typedef struct SymData          SymData;
typedef struct JpdFile          JpdFile;
typedef struct JpdClass         JpdClass;
typedef struct JpdInterface     JpdInterface;
typedef struct JpdVariable      JpdVariable;
typedef struct JpdRoutine       JpdRoutine;
typedef struct JpdType          JpdType;
typedef struct JpdModifier      JpdModifier;
typedef struct JpdSBlock        JpdSBlock;
typedef struct JpdStatement     JpdStatement;
typedef struct JpdExpression    JpdExpression;
typedef struct JpdLiteral       JpdLiteral;
typedef struct JpdOperator      JpdOperator;
typedef struct JpdQName         JpdQName;
typedef struct JpdLabel         JpdLabel;
/* JDK 1.1. */
/* AnonClass뤿ΰŪʥס */
typedef struct JpdNewObjectPrefix JpdNewObjectPrefix;

typedef struct SymDataStack     SymDataStack;

typedef struct varSymTabElem    VarSymTabElem;
typedef struct varSymTab        VarSymTab;
typedef struct labelSymTabElem  LabelSymTabElem;
typedef struct labelSymTab      LabelSymTab;

typedef struct DBClassAttr      DBClassAttr;
typedef struct DBClassTable     DBClassTable;
typedef struct DBRelAttr        DBRelAttr;
typedef struct DBRelTable       DBRelTable;

/**
 * Definition of eumerations.
 */

#define JDBIN_DEFAULT_POOL_SIZE     4096

/* !! */
#define JPD_INVALID_ID  (-1)

/* typedef SpdClassId      SymSort; */

typedef enum varsort {
    CLASS_MEM = 0, INTERFACE_MEM, PARAMETER, LOCAL_VAR
} VarSort;

/* JDK 1.1. */
typedef enum ClassSort {
    PACKAGE_MEMBER_CLASS = 0, NESTED_TOP_LEVEL_CLASS, MEMBER_CLASS,
    LOCAL_CLASS, ANON_CLASS
} ClassSort;

/* JDK 1.1. */
typedef enum InterfaceSort {
    PACKAGE_MEMBER_INTERFACE = 0, NESTED_TOP_LEVEL_INTERFACE
} InterfaceSort;

typedef enum RoutineSort {
    METHOD = 0, CONSTRUCTOR
} RoutineSort;

typedef enum typesort {
    PRIMITIVE_TYPE = 0, OBJECT_TYPE, ARRAY_TYPE
} TypeSort;

typedef enum SBlockSort {
    STATIC_INITIALIZER = 0, INSTANCE_INITIALIZER
} SBlockSort;

/* "S_" means Statement. */

typedef enum StatementSort {
    S_EMPTY = 0, S_EXPRESSION,              /* S_LABELED, */
    S_BLOCK, S_DEF_VAR,
    S_IF, S_SWITCH, S_WHILE, S_DO, S_FOR, S_BREAK, S_CONTINUE,
    S_RETURN, S_THROW, S_EXCEPTION, S_TRY, S_CATCH, S_FINALLY, S_SYNCHRONIZED,
    /* JDK 1.1. */
    S_SWITCH_LABELED_EMPTY, S_LOCAL_CLASS,
    /* JDK 1.4 */
    S_ASSERT
} StatementSort;

/* "E_" means Expression. */

typedef enum ExpressionSort {
    E_EMPTY = 0,                            /* for3Ĥμ̤뤿˻Ȥ */
    E_REF_VAR, E_REF_ARRAY, E_REF_LITERAL, E_REF_TYPE,
    E_REF_SPECIAL, E_METHOD_CALL, E_CONSTRUCTOR_CALL,
    E_OPERATION, E_REF_QNAME, E_LISTED, E_DIM_EXPR,
    /* JDK 1.1. */
    E_ANON_CLASS, E_ANON_ARRAY,
    E_REF_ARRAY_LENGTH                      /* used by the access library */
} ExpressionSort;

typedef enum LRSort {
    NONE = 0, L_VALUE, R_VALUE, LR_VALUE, OBJECT_OPERATION
} LRSort;

/*
 * "O_" means Operator.
 * Note: Macro "O_BINARY" conflicts with one defined in some systems.
 */

typedef enum OperatorSort {
    JPD_O_PREFIX = 0,
    JPD_O_POSTFIX,
    JPD_O_BINARY,
    JPD_O_TERNARY,
    JPD_O_PAREN, 
    JPD_O_CAST,
    JPD_O_NEW_ARRAY,
    JPD_O_NEW_OBJECT,
    JPD_O_ALLOC_DIM,
    JPD_O_REF_DIM
} OperatorSort;

/* "L_" means Literal. */

typedef enum literalSort {
    L_CHAR = 0, L_STRING, L_UNICODE, L_INT, L_DOUBLE
} LiteralSort;

typedef enum labelsort {
    DEF_LABEL = 0, REF_LABEL, CASE_LABEL, DEFAULT_LABEL
} LabelSort;

typedef enum commentSort {
    C_BLOCK = 0, C_INLINE, C_LINE
} CommentSort;

/**
 * JB饹
 */

struct JpdFile {
    SymData     *classOrInterface;
    SymData     *definedPackage;
    SymData     *importedPackage;
    SpdEncode   encoding;
};

struct JpdClass {
    ClassSort       sort;
    SymData         *modifier;
    SymData         *super;                 /* Super class. */
    SymData         *interface;             /* Implimented interfaces. */
    SymData         *member;
    /* 饹Υѿ᥽åɡ󥹥ȥ饯 */
    /* ŪҤΥꥹ                           */
    VarSymTab       *memberSymTab;          /* Class's symbol table. */
    SymData         *nextClassOrInterface;
    /* եƤ뼡Υ饹ؤ */
    /* JDK 1.1. */
    SymData         *nextMember;
    /* Х饹μΥѿ뤤ϥ᥽åɤʸؤ */
    SymData         *arguments;
    /* Anon-Classΰ */
};

struct JpdInterface {
    InterfaceSort   sort;
    SymData         *modifier;
    SymData         *super;                 /* extended interfaces */
    SymData         *member;
    /* 饹Υѿ᥽åɡ󥹥ȥ饯   */
    /* ŪҤΥꥹ                             */
    VarSymTab       *memberSymTab;          /* interface's Symbol Table */
    SymData         *nextClassOrInterface;
    /* եƤ뼡Υ饹ؤ          */
    /* JDK 1.1. */
    SymData     *nextMember;
    /* 饹μΥѿ뤤ϥ᥽åɤʸؤ     */
};

struct JpdVariable {
    VarSort     sort;                       /* ѿμ */
    SymData     *modifier;
    SymData     *type;
    SymData     *initializer;
    SymData     *next;
    /* int x, y; ʤɤΤ褦˰ĤʸʣѿȤ */
    /*  ѿ̾ꥹȤˤ뤿˻Ȥ                      */
    /*  ޤΤȤintɽJpdTypex, y鶦ͭ     */
    SymData     *nextMember;
    /* 饹μΥѿ뤤ϥ᥽åɤʸؤ     */
};

struct JpdRoutine {
    RoutineSort sort;
    SymData     *parameter;
    SymData     *modifier;
    SymData     *returnType;
    SymData     *throwType;
    SymData     *body;
    SymData     *nextMember;
    /* 饹μΥѿ뤤ϥ᥽åɤʸؤ     */
};

struct JpdType {
    TypeSort    sort;                       /* μ */
    char        *pkgName;                   /* ѥå̾. millǤNULLꤹ */
    SymData     *qname;                     /* A.B.Type  A.Bʬ */
    SymData     *next;
    /* nextȤäꥹȤˤ */
    /* e.g. double[][] ::= [] -next-> [] -next-> double */
    SymData     *nextType;
    /* nextTypeϥ᥽åΤȤthrowsʣΥפ         */
    /* ꥹȤˤȤ˻Ѥ졤줬ؤϴǤ   */
    /* JDK 1.1. */
    SymData     *jclass;
    /* type_class δϢ(AnonClassǻȤ) */
};

struct JpdModifier {
    SymData     *next;
    /* ʣνҤꤵƤȤϥꥹȤˤ */
    /* e.g. public static ::= public -next-> static   */
};

struct JpdSBlock {
    SBlockSort  sort;
    SymData     *block;                     /* block */

    /* 饹μΥѿ뤤ϥ᥽åɤʸؤ */
    SymData     *nextMember;
};

struct JpdStatement {
    StatementSort   sort;                   /* ʸμ*/
    SymData         *expression1, *expression2, *expression3;
    /* ؤδϢʸμ(expr1), whileʸξ(expr1),   */
    /*  forʸν(expr1), (expr2), ü(expr3)  */
    /*  ifʸξ(expr1)ʤɤǻȤ롥                   */
    SymData         *statement1, *statement2;
    /* ʸؤδϢwhile, forʸΥ(statement1)   */
    /*  ifʸthen(statement1), else(statement2)        */
    /* expression#, statement# ϥޥ̤ƥ   */
    SymData         *label;
    SymData         *next;
    /* ֥åμstatementꥹȤˤ */
    SymData         *block;
    /* ֥åκǽstatemntؤ */
    /*
     * e.g. { stmt1; {stmt2;} stmt3}
     * OUTBLOCK: -block-> stmt3
     * +-next-> INBLOCK: -block-> stmt2
     * +-next-> stmt2
     */
};

struct JpdExpression {
    ExpressionSort  sort;                   /* μ */
    LRSort          lrsort;                 /* ͡ */
    SpdBoolean      two_pass_resolve;
       /* writeExpression()ѿλõݤjdbinѿ */
    SymData         *joperator;             /* 黻 */
    SymData         *first;                 /* 쥪ڥ or ѿƥλ */
    /* ƤФ᥽åɤؤλ */
    SymData         *second;                /* 󥪥ڥ, ᥽åɤΰؤ */
    SymData         *third;                 /* 軰ڥ */
    SymData         *next;                  /* ᥽åɤΰʤɼꥹȤˤȤ */
    /* Ȥ롥NULL terminator. */
};

/**
 * ƥ()ɽJpd饹
 * JpdLiteralͤ"java-yacc.y"ǤϤʤ"java-lex.l"ꤵ롥
 */

struct JpdLiteral {
    LiteralSort     sort;                   /* ƥμ */
    union literal {
        int     i;
        double  f;
        char    c;
        char    *s;
    } value;
};

struct JpdOperator {
    OperatorSort    sort;
};

/**
 * Qualified Name is  IDENTIFIER ( . IDENTIFIER )*
 *  e.g. java.util.Vector
 * It is expressed by reverse ordered list
 *  e.g. java.util.Vector ::= Vecotr -> util -> java
 * It is managed in QNameSymTab
 */

struct JpdQName {                           /* Qualified Name */
    SymData     *first;
    SymData     *second;
    SymData     *dot;
    SymData     *nextQName;
};

/* JpdSpecial */

struct JpdLabel {
    LabelSort   sort;                       /* def, ref, case, default */
    SymData     *caseValue;                 /* case٥(expression) */
    SymData     *defined;
    /* sort == REF_LABEL ΤȤ٥ƤʬؤΥ */
    SymData     *next;
    /* case 1: case2: ʤɤΥ٥뤬ʣƤʸǥꥹȤˤ */
};

/* JDK 1.1AnonClass뤿Τ */
struct JpdNewObjectPrefix {
    SymData     *newOp;
    SymData     *typeName;
    SymData     *args;
    SymData     *lp;
    SymData     *rp;
    VarSymTab   *memberSymTab;              /* class's Symbol Table */
};

/**
 * Note: "YYSTYPE = SymData *"
 * 
 * yaccΥåˤ"SymData *"Ѥޤ졤ʸڤSymDataȤäƹۤ롥
 * ʸڤƥΡɤ˸ͭΥǡ¸뤿ˡ
 * SymData"union u"JpdXxxؤΥݥ󥿤ġ
 * 㤨СJpdVariableϹʸڤVariableΡɤΤǤϤʤΡɤ
 * SymDataVariableΡɤ˸ͭξJpdVariableǼ롥
 *  
 * SymDatagetSymData()ؿˤäidꤵ롥
 */

struct SymData {
    int     id;

    /* Initialized to JPD_INVALID_ID and set on writing DB. */
    
    SpdClassId  symSort;                    /* Sort of SymData. */
    CommentSort sort2;                      /* Sort of comments. */ 

    int         offset;                     /* Offset of symbol. */
    
    /* Length of charactors that consist of symbol. */
    int         length;
    
    int         strOffset;                  /* Offset of string. */
    
    /* String representation of symbol. */
    char        *str;

    SymData     *comment;                   /* Comment for this symbol. */
    
    /*
     * In order to avoid conflicting with C++ keywords "class" and "operator".
     * "j" means Java.
     */
    union {
        JpdFile         *file;
        JpdClass        *jclass;
        JpdInterface    *interface;
        JpdVariable     *variable;
        JpdRoutine      *routine;
        JpdType         *type;
        JpdModifier     *modifier;
        JpdSBlock       *sblock;
        JpdStatement    *statement;
        JpdExpression   *expression;
        JpdLiteral      *literal;
        JpdOperator     *joperator;
        JpdQName        *qname;
        JpdLabel        *label;             /* JpdSpecial has no union value. */
        JpdNewObjectPrefix  *newObject;
    } u;
};

#define MAX_STACK_SIZE 128

struct SymDataStack {
    SymData     *stack[MAX_STACK_SIZE];
    SymData     **cur;
};

/**
 * Symbol table definition.
 */

typedef enum symTabSort {SENTINEL_TAB, MEMBER_TAB, LOCAL_TAB} SymTabSort;

struct varSymTab {
    SymTabSort      sort;
    VarSymTabElem   *table;
    VarSymTabElem   *cur;                   /* current position of table */
    VarSymTab       *next;
    VarSymTab       *prev;
    VarSymTab       *prevMember;
};

struct varSymTabElem {
    SymData         *symData;               /* pointer of the Symbol's SymData */
};

struct labelSymTab {
    LabelSymTabElem     *table;
    LabelSymTabElem     *cur;               /* current position of table */
};

struct labelSymTabElem {
    SymData *symData;                       /* pointer of the Label's SymData */
};

/* ¤ΥС˴ؤ롤¤ΤƬΥեåȡ */
#define SDB_NON_OFFSET  (-1)

/* 
 * Bit pattern of J-model object id:
 * 
 * @see Japid.Freja.FrjID.java
 */
#define JPD_MAX_OID                     ((1 << SPD_BIT_WIDTH_ID) - 1)

struct DBClassAttr {
    SpdDBType   type;
    char        *name;                      /* name of this attribute */
    int         offset;                     /* offset from SymData */
    int         unionOffset;                /* offset from SymData.u, */
    /* -1 means this attr is not member of uion */
};

struct DBClassTable {
    char        *name;
    SpdClassId  id;
    DBClassAttr attrs[SPD_ISVAL_NUM];
};

struct DBRelAttr {
    SpdDBType   type;
    char        *name;
    SpdClassId  classId;
};

struct DBRelTable {
    char        *name;
    DBRelAttr   attrs[SPD_ISVAL_NUM];
};

#ifdef YYSTYPE
#undef YYSTYPE
#endif /* YYSTYPE */
#define  YYSTYPE SymData*

struct JpdDebug {
    int         jclass, interface, type, variable, routine, modifier,
                statement, expression, qname, literal, symTab;

    SpdBoolean  etb;  /* Added by S.Yamamoto 2003/02/12. */
};

/**
 * Macros.
 */

#define symSort2TableIndex(_symSort)    ((_symSort) - SPD_FIRST_CID_J_MODEL)

/* JDK 1.1. */
#define setClassSort(_c, _s)            ((_c)->u.jclass->sort = (_s))
#define setInterfaceSort(_i, _s)        ((_i)->u.interface->sort = (_s))
#define setClassArguments(_c, _a)       ((_c)->u.jclass->arguments = (_a))
#define getClassArguments(_c)           ((_c)->u.jclass->arguments)

/* 
 * ʲʸ
 * 
 *     a[i] = b[i] + c[++j];
 * 
 * ˤ뼰κ(L)(R)
 * 
 *     a[i]      i: R,           a: L,       a[i]: L
 *     b[i]      i: R,           b: R,       b[i]: R
 *     c[++j]    j: LR,      c: R,       c[++j]: R
 * 
 * Τ褦ꤵޤ
 * 
 * a[i]b[i]ϤμϤǤϺͤͤȽǤ"="
 * դˤ뤫դˤ뤫狼äʳȽꤵޤޤź
 * Ϻդˤ뤫դˤ뤫˴ؤ餺a[i]iϺ͡c[++j]jϺ
 * ͤıͤǤjava-yacc.yǤϰʲΤ褦˽Ƥޤ
 * 
 * (1) źʤɤϡޤǲϤѤʳǺͤ͡ꤹ
 * 
 * (2) ʳμϡ"="ޤʬޤǲϤʳǡդ롥
 * (a[i], a, i)ФL_VALUE򡤱դ뼰(b[i], b, i, c[++j],
 * c, j)ФR_VALUE򡤴ؿsetLRSortR()ˤꤷޤ
 * 
 * (2)ΤȤˡźͤꤵƤޤȺΤǡ
 * (1)ʳź lockLRSort ŬѤͤιǤʤ褦
 * ƤޤŪˤlockLRSort()lrsortͤ10­ƤΤǡ
 * writeExpression()SDB˽񤯤Ȥunlockͤ򸵤ᤷƤޤ
 */

#define setLRSort(_e, _lr)  ((_e)->u.expression->lrsort = (_lr))
#define lockLRSort(_e)      ((isLockedLRSort(_e)) ? \
                             ((_e)->u.expression->lrsort): \
                             ((_e)->u.expression->lrsort += 10))
#define isLockedLRSort(_e)  ((_e)->u.expression->lrsort >= 10)
#define unlockLRSort(_e)    ((isLockedLRSort(_e)) ? \
                             ((_e)->u.expression->lrsort -= 10): \
                             (_e)->u.expression->lrsort)

#define setTwoPassResolve(_e, _r) \
    ((_e)->u.expression->two_pass_resolve = (_r))
#define isTwoPassResolve(_e) \
    ((_e)->u.expression->two_pass_resolve == TRUE)

#define setMethodArguments(_m, _a) \
    ((_m)->u.expression->second = (_a))
#define getMethodArguments(_m) \
    ((_m)->u.expression->second)

#define setDimExprs(_e, _d) ((_e)->u.expression->third = (_d))
#define getDimExprs(_e)     ((_e)->u.expression->third)

#define setStatementExpression(_stmt, _expr) \
    ((_stmt)->u.statement->expression1 = (_expr))
#define getStatementExpression(_stmt) \
    ((_stmt)->u.statement->expression1)

#define setStatementCondition(_stmt, _cond)\
    ((_stmt)->u.statement->expression1 = (_cond))
#define getStatementCondition(_stmt) \
    ((_stmt)->u.statement->expression1)

#define setStatementForInit(_stmt, _expr) \
    ((_stmt)->u.statement->expression2 = (_expr))
#define getStatementForInit(_stmt) \
    ((_stmt)->u.statement->expression2)

#define setStatementForIncr(_stmt, _expr) \
    ((_stmt)->u.statement->expression3 = (_expr))
#define getStatementForIncr(_stmt) \
    ((_stmt)->u.statement->expression3)

#define setStatementAction(_stmt, _expr) \
    ((_stmt)->u.statement->statement1 = (_expr))
#define getStatementAction(_stmt) \
    ((_stmt)->u.statement->statement1)

#define setStatementIfThen(_stmt, _expr) \
    ((_stmt)->u.statement->statement1 = (_expr))
#define getStatementIfThen(_stmt) \
    ((_stmt)->u.statement->statement1)

#define setStatementIfElse(_stmt, _expr) \
    ((_stmt)->u.statement->statement2 = (_expr))
#define getStatementIfElse(_stmt) \
    ((_stmt)->u.statement->statement2) 

#define setStatementVariableDecl(_stmt, _vars) \
    ((_stmt)->u.statement->expression1 = (_vars))
#define getStatementVariableDecl(_stmt) \
    ((_stmt)->u.statement->expression1)

/* JDK 1.1. */
#define setStatementClass(_stmt, _cls) \
    ((_stmt)->u.statement->expression1 = (_cls))
#define getStatementClass(_stmt) \
    ((_stmt)->u.statement->expression1)

#define setStatementLabel(_s, _l) \
{ \
    if (hasStatementLabel(_s)) { \
        SymData *last; \
        for (last = (_l); last->u.label->next != NULL; \
             last = last->u.label->next) \
            ; \
        last->u.label->next = (_s)->u.statement->label; \
        (_s)->u.statement->label = (_l); \
    } else { \
        (_s)->u.statement->label = (_l); \
    } \
}

#define getStatementLabel(_stmt) \
    ((_stmt)->u.statement->label)

#define hasStatementLabel(_stmt) \
    ((_stmt)->u.statement->label != (NULL))

#define setStatementExceptionTry(_stmt, _try) \
    ((_stmt)->u.statement->expression1 = (_try))
#define getStatementExceptionTry(_stmt) \
    ((_stmt)->u.statement->expression1)

#define setStatementExceptionCatches(_stmt, _catches) \
    ((_stmt)->u.statement->expression2 = (_catches))
#define getStatementExceptionCatches(_stmt) \
    ((_stmt)->u.statement->expression2)

#define setStatementExceptionFinally(_stmt, _finally) \
    ((_stmt)->u.statement->expression3 = (_finally))
#define getStatementExceptionFinally(_stmt) \
    ((_stmt)->u.statement->expression3)

#define setStatementCatchArgument(_stmt, _arg) \
    ((_stmt)->u.statement->expression1 = (_arg))
#define getStatementCatchArgument(_stmt) \
    ((_stmt)->u.statement->expression1)

#define setStatementTryBlock(_stmt, _block) \
    ((_stmt)->u.statement->statement1 = (_block))
#define getStatementTryBlock(_stmt) \
    ((_stmt)->u.statement->statement1)

#define setStatementCatchBlock(_stmt, _block) \
    ((_stmt)->u.statement->statement1 = (_block))
#define getStatementCatchBlock(_stmt) \
    ((_stmt)->u.statement->statement1)

#define setStatementFinallyBlock(_stmt, _block) \
    ((_stmt)->u.statement->statement1 = (_block))
#define getStatementFinallyBlock(_stmt) \
    ((_stmt)->u.statement->statement1)

#define setStatementSynchronizedExpression(_stmt, _expr) \
    ((_stmt)->u.statement->expression1 = (_expr))
#define getStatementSynchronizedExpression(_stmt) \
    ((_stmt)->u.statement->expression1)

#define setStatementSynchronizedStatement(_stmt, _synchronized) \
    ((_stmt)->u.statement->statement1 = (_synchronized))
#define getStatementSynchronizedStatement(_stmt) \
    ((_stmt)->u.statement->statement1)

/* JDK 1.4 */
#define setStatementAssert(_stmt, _expr1, _expr2) \
{ \
    (_stmt)->u.statement->expression1 = (_expr1); \
    (_stmt)->u.statement->expression2 = (_expr2); \
}
#define getStatementAssertExpression1(_stmt) \
    ((_stmt)->u.statement->expression1)
#define getStatementAssertExpression2(_stmt) \
    ((_stmt)->u.statement->expression2)

#define getAnonClassType(_expr) \
    ((_expr)->u.expression->first)

#define setComment(_a, _b) \
{ \
    SymData *a, *b = ((_b)->symSort == SPD_CID_J_COMMENT) ? \
                     (_b) : (_b)->comment; \
    for (a = (_a); a->comment != NULL; a = a->comment) \
        ; \
    a->comment = b; \
}

#ifdef YYDEBUG

#define jpdDebugMessage     _jpdDebugMessage
#define jpdDebugFunc2Call(flag, f, a1, a2)      {if ((flag)) (f)((a1), (a2));}

#else /* not YYDEBUG */

#define jpdDebugMessage     (void)
#define jpdDebugFunc2Call(flag, f, a1, a2) 

#endif /* not YYDEBUG */

/**
 * External declarations.
 */

extern char     *yyfilename;
extern int      yylineno;
extern int      c_count;
extern SymData  *comment;
extern FILE     *yyin;
extern FILE     *jdbinEtb;
extern FILE     *j2jxEtb;

extern int          yydebug;
extern SpdBoolean   verbose;
extern SpdBoolean   linkclass;
extern SpdBoolean   withAssert;     /* JDK 1.4 assert. */
extern SpdBoolean   outputJ2jxEtb;

extern struct JpdDebug jpdDebug;

extern LabelSymTab  *globalLabelSymTab;

extern DBClassTable dbClassTable[];
extern DBRelTable   dbRelTable[];

#include "j2jx.h"

#endif /* not _SPD_JB_H */
