/* 
 * Program:             $RCSfile: SymTab.c,v $  $Revision: 40.16 $
 * 
 * Purpose:
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: SymTab.c,v 40.16 2003/07/22 11:30:20 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "JB.h"
#include "y.tab.h"
#include "FuncProto.h"

#define MAX_SYMBOL_NUM      128

VarSymTab   *firstTab = NULL;
VarSymTab   *topMemberSymTab = NULL;
VarSymTab   *topLocalSymTab = NULL;
VarSymTab   *topSymTab = NULL;

#define MAX_LABEL_TAB   4

/* 
 * labelSymTabs[0]static blockѤΥ٥뵭ɽ
 * labelSymTabs[1]ʹߤϥ᥽åɡ󥹥ȥ饯ѡ
 */

LabelSymTab     labelSymTabs[MAX_LABEL_TAB];
LabelSymTab *topLabelSymTab = labelSymTabs;

/* 
 * ɽ
 * 
 * class C1 {
 *  int v1;
 *  void m1() {
 *      int l1;
 *      if (...) {
 *          int l2;
 *          class C2 {
 *              int v2;
 *              void m2() {
 *                  int l3;
 *              }
 *          }
 *      }
 *  }
 * }
 * 
 *          firstTab                             topMember  topLocal,top
 *                                                          
 * ɽ: sentinel -- C1(M) -- m1(L) -- if_blk(L) -- C2(M) -- m2(L)
 * ѿ:                v1       l1       l2           v2       l3
 * 
 */
                                        
int addVar2MemberSymTab(SymData *symData)
{
    if (topMemberSymTab->cur - topMemberSymTab->table >= MAX_SYMBOL_NUM - 2) {
        return -1;
    }
    topMemberSymTab->cur++;
    topMemberSymTab->cur->symData = symData;

    jpdDebugMessage(jpdDebug.symTab, "add %s to MemberSymTab %p\n",
                    symData->str, topMemberSymTab);
    return 0;
}

SymData *searchForMemberSymTab(char *name, VarSymTab *mst)
{
    VarSymTabElem       *p;

    if (mst == NULL) {
        mst = topMemberSymTab;
    }
    if (mst->sort != MEMBER_TAB) {
        spdWarning("BUG? in serachForMemberSymTab:"
                   "topMemberSymTab is not MEMBER_TAB: %d.\n", mst->sort);
        return NULL;
    }
    for (p = mst->cur; p->symData != NULL; p--) {
        jpdDebugMessage(jpdDebug.symTab,
                        "search %s for %s of SymTab %p(%d)\n",
                        name, p->symData->str, mst, mst->sort);
        if (SPD_STREQ(name, p->symData->str)) {
            jpdDebugMessage(jpdDebug.symTab,
                            "MATCH %s for %s of SymTab %p(%d)\n",
                            name, p->symData->str, mst, mst->sort);
            return p->symData;
        }
    }
    /* Search previous member symbol table */
    if (mst->prevMember != NULL) {
        return searchForMemberSymTab(name, mst->prevMember);
    }
    else {
        return NULL;
    }
}

int addVar2LocalSymTab(SymData *symData)
{
    if (topLocalSymTab->cur - topLocalSymTab->table >= MAX_SYMBOL_NUM - 2) {
        return -1;
    }
    topLocalSymTab->cur++;
    topLocalSymTab->cur->symData = symData;

    jpdDebugMessage(jpdDebug.symTab, "add %s to localSymTab %p\n",
                    symData->str, topLocalSymTab);

    return 0;
}

SymData *searchForLocalSymTab(char *name, VarSymTab *lst)
{
    VarSymTabElem       *p;

    if (lst == NULL) {
        lst = topLocalSymTab;
    }
    if (lst->sort != LOCAL_TAB) {
        /* maybe initializer of member variable */
        jpdDebugMessage(jpdDebug.symTab,
                        "in searchForLocalSymTab"
                        "topLocalSymTab is not LOCAL_TAB: %d\n",
                        lst->sort);
        return NULL;
    }

    for (p = lst->cur; p->symData != NULL; p--) {
        jpdDebugMessage(jpdDebug.symTab,
                        "search %s for %s of SymTab %p(%d)\n",
                        name, p->symData->str, lst, lst->sort);

        if (SPD_STREQ(name, p->symData->str)) {
            jpdDebugMessage(jpdDebug.symTab,
                            "MATCH %s for %s of SymTab %p(%d)\n",
                            name, p->symData->str, lst, lst->sort);
            return p->symData;

        }
    }

    /* Search previous local symbol table */
    for (lst = lst->prev; lst->sort == MEMBER_TAB; lst = lst->prev)
        ;
    if (lst->sort == SENTINEL_TAB) {
        return NULL;
    }
    else {
        return searchForLocalSymTab(name, lst);
    }
}

int addLabel2LabelSymTab(SymData *label, LabelSymTab *lst)
{
    if (lst == NULL) {
        lst = topLabelSymTab;
    }
    if (lst->cur - lst->table >= MAX_SYMBOL_NUM -2) {
        return -1;
    }
    lst->cur++;
    lst->cur->symData = label;
    return 0;
}

SymData *searchForLabelSymTab(char *name, LabelSymTab *lst)
{
    LabelSymTabElem     *p;

    if (lst == NULL) {
        if (topLabelSymTab == NULL) {
            return NULL;
        }
        else {
            lst = topLabelSymTab;
        }
    }
    for (p = lst->cur; p->symData != NULL; p--) {
        if (SPD_STREQ(name, p->symData->str)) {
            return p->symData;
        }
    }
    return NULL;
}

VarSymTab *getVarSymTab(SymTabSort sort)
{
    VarSymTab       *st;

    st = spdMalloc(sizeof(VarSymTab));
    st->sort = sort;
    st->table = spdMalloc(sizeof(VarSymTabElem) * MAX_SYMBOL_NUM);
    st->table[0].symData = NULL;            /* sentinel */
    st->cur = st->table;
    return st;
}

void freeVarSymTab(VarSymTab *st)
{
    if (st == NULL) {
        return;
    }
    spdFree(st->table);
    spdFree(st);
}

void initFirstSymTab(void)
{
    if (firstTab != NULL) {
        VarSymTab *last, *p;

        for (last = firstTab; last->next != NULL; last = last->next)
            ;
        for (p = last; p != firstTab; p = p->prev) {
            freeVarSymTab(p);
        }
    }
    else {
        firstTab = getVarSymTab(SENTINEL_TAB);
    }
    firstTab = getVarSymTab(SENTINEL_TAB);
    topMemberSymTab = firstTab;
    topLocalSymTab = firstTab;
    topSymTab = firstTab;
}

VarSymTab *pushMemberSymTab(void)
{
    VarSymTab *st = getVarSymTab(MEMBER_TAB);
    VarSymTab *mst;

    topSymTab->next = st;
    st->prev = topSymTab;
    topSymTab = st;
    topMemberSymTab = st;
    topSymTab->next = NULL;

    /* Search previous member symbol table */
    for (mst = st->prev; mst->sort == LOCAL_TAB; mst = mst->prev)
        ;
    st->prevMember = (mst->sort == MEMBER_TAB) ? mst: NULL;

    jpdDebugMessage(jpdDebug.symTab,
                    "push MemberSymTab: %p %d: ", st, st->sort);
    printSymTabStack();

    return st;
}

void popMemberSymTab(void)
{
    VarSymTab *tmp = topSymTab;

    if (topSymTab->sort != MEMBER_TAB) {
        spdErrorExit("topSymTab (%d) should be MEMBER_TAB.\n", topSymTab->sort);
    }
    topSymTab = topSymTab->prev;
    topSymTab->next = NULL;

    for (topMemberSymTab = topSymTab;
         topMemberSymTab->sort == LOCAL_TAB;
         topMemberSymTab = topMemberSymTab->prev) {
        /* Empty. */
    }

    for (topLocalSymTab = topSymTab;
         topLocalSymTab->sort == MEMBER_TAB;
         topLocalSymTab = topLocalSymTab->prev) {
        /* Empty. */
    }

    jpdDebugMessage(jpdDebug.symTab,
                    "pop MemberSymTab: %p %d: ", tmp, tmp->sort);
    printSymTabStack();

    /* writeClass()λ˥ѿɽ򻲾ȤΤfreeʤ */
    /* freeVarSymTab(tmp);*/
}



VarSymTab *pushLocalSymTab(void)
{
    VarSymTab *st = getVarSymTab(LOCAL_TAB);
    topSymTab->next = st;
    st->prev = topSymTab;
    topSymTab = st;
    topLocalSymTab = st;
    topSymTab->next = NULL;

    jpdDebugMessage(jpdDebug.symTab, "push LocalSymTab: %p %d: ", st, st->sort);
    printSymTabStack();

    return st;
}

void popLocalSymTab(void)
{
    VarSymTab *tmp;
    tmp = topSymTab;

    if (topSymTab->sort != LOCAL_TAB) {
        spdErrorExit("topSymTab (%d) should be LOCAL_TAB.\n", topSymTab->sort);
    }
    topSymTab = topSymTab->prev;
    topSymTab->next = NULL;

    topLocalSymTab = topSymTab;

    jpdDebugMessage(jpdDebug.symTab, "pop LocalSymTab: %p %d: ",
                    tmp, tmp->sort);
    printSymTabStack();

    freeVarSymTab(tmp);
}


void initLabelSymTab(void)
{
    /* set up labelSymTabs[0] which is the table for static block */
    labelSymTabs[0].table = spdMalloc(sizeof(LabelSymTabElem) * MAX_SYMBOL_NUM);
    labelSymTabs[0].table[0].symData = NULL; /* sentinel */
    labelSymTabs[0].cur = labelSymTabs[0].table;
}

LabelSymTab *pushLabelSymTab(void)
{
    /* Set up a table for a method or a constructor. */
    if (topLabelSymTab - labelSymTabs >= MAX_LABEL_TAB) {
        spdErrorExit("Too many label table elements.\n");
    }

    topLabelSymTab++;
    topLabelSymTab->table = spdMalloc(sizeof(LabelSymTabElem) * MAX_SYMBOL_NUM);
    topLabelSymTab->table[0].symData = NULL; /* sentinel */
    topLabelSymTab->cur = topLabelSymTab->table;
    return topLabelSymTab;
}

void popLabelSymTab(void)
{
    /* free a table for a method or a constructor */
    spdFree(topLabelSymTab->table);
    if (topLabelSymTab == labelSymTabs) {
        initLabelSymTab();
    }
    else {
        topLabelSymTab--;
    }
}

/* type is DEFAULT, SUPER, this(THIS) */

void printSymTabStack(void)
{
    VarSymTab *p = firstTab;
    for (; p != NULL; p = p->next) {
        jpdDebugMessage(jpdDebug.symTab, "%p(%d) -- ", p, p->sort);
    }
    jpdDebugMessage(jpdDebug.symTab, "\n");
    jpdDebugMessage(jpdDebug.symTab, "  topSymTab: %p(%d)\n",
                    topSymTab, topSymTab->sort);
    jpdDebugMessage(jpdDebug.symTab, "  topMemberSymTab: %p(%d)\n",
                    topMemberSymTab, topMemberSymTab->sort);
    jpdDebugMessage(jpdDebug.symTab, "  topLocalSymTab: %p(%d)\n",
                    topLocalSymTab, topLocalSymTab->sort);

}
