/* 
 * Program:             $RCSfile: WriteDB.c,v $  $Revision: 40.101 $
 * 
 * Purpose:
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: WriteDB.c,v 40.101 2003/07/25 13:14:51 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <sys/types.h>      /* For stat(). */
#include <sys/stat.h>       /* For stat(). */
#include <Sapid/Sapid.h>
#include "JB.h"
#include "FuncProto.h"

static int  dbClassID[SPD_CID_NUM_J_MODEL];     /* Number of instances. */

/* Output file names of classes and relations. */
static FILE *dbClassFile[SPD_CID_NUM_J_MODEL];  
static FILE *dbRelFile[SPD_RID_NUM_J_MODEL];

/* Ϥ󥹥󥹿Υ󥿡 */
static int  dbClassObjCounter[SPD_CID_NUM_J_MODEL];
static int  dbRelObjCounter[SPD_RID_NUM_J_MODEL];

static void init_obj_counter(void);
static void write_japid_version(void);
static int getRelIndex(char *relName);
static int getRelIndexOld(char *relName);

/* ѿ˻Ȥн褹뤿δؿ */
static void pushMemSymTabStack(VarSymTab *tab);
static VarSymTab *popMemSymTabStack(void);
static SymData *searchForMemSymTabStack(char *name);

/* !! Dirty hack. */
extern SpdDBClass J_model_classes[];

void setObjectID(SymData *symData)
{
    symData->id = SPD_OBJ_ID_OFFSET(symData->symSort) +
        dbClassID[symSort2TableIndex(symData->symSort)]++;
}

void writeDBClass(SymData *symData)
{
    char                *symP, *unionP;
    SpdDBClass          *table;
    SpdDBAttr           *attr;
    int                 i;
    FILE                *dbfile;

    symP = (char *)symData;
    table = &(GET_DBCLASS(symSort2TableIndex(symData->symSort)));
    if (table->id != symData->symSort) {
        spdBugExit("%d != %d: Invalid DBClassTable.\n",
                   table->id, symData->symSort);
    }

    if (symData->id == JPD_INVALID_ID) {
        setObjectID(symData);
    } else {
        /* symDataϴDB˽񤫤Ƥ롥 */
        return;
    }

    dbfile = dbClassFile[symSort2TableIndex(symData->symSort)];
    spddb_CreateTuple(dbfile);
    spdWriteIntAttr(dbfile, (symData->id));
    for (i = 0; i < SPD_ISVAL_NUM; i++) {
        attr = &(table->attrs[i]);
        if (attr->type == SPD_DBT_NON) {
            break;
        } else if (attr->type == SPD_DBT_INT) {
            if (attr->unionOffset == SDB_NON_OFFSET) {
                /* See "lib/SpdUtil/spdIO.c". */
                spdWriteIntAttr(dbfile, *((int *)(symP + attr->offset)));
            } else {
                unionP = *((char **)(symP + attr->offset));
                spdWriteIntAttr(dbfile, *((int *)(unionP + attr->unionOffset)));
            }
        } else if (attr->type == SPD_DBT_STR) {
            if (attr->unionOffset == SDB_NON_OFFSET) {
                spdWriteStrAttr(dbfile, *((char **)(symP + attr->offset)));
            } else {
                unionP = *((char **)(symP + attr->offset));
                spdWriteStrAttr(dbfile, *((char **)(unionP + attr->unionOffset)));
            }
        } else {
            spdBugExit("Invalid SDB_TYPE %d.\n", attr->type);
        }
            
    }
    spddb_CommitTuple(dbfile);
    
    /* Write comment. */
    if (symData->symSort != SPD_CID_J_COMMENT) {
        SymData *com;
        for (com = symData->comment; com != NULL; com = com->comment) {
            writeDBClass(com);
            writeDBRel("symbol_comment", symData, com);

            extractComment(com);
        }
    }

    /* Ϥ󥹥󥹿򥫥Ȥ롥 */
    dbClassObjCounter[symSort2TableIndex(symData->symSort)]++;
}

void writeDBRel(char *relName, SymData *obj1, SymData *obj2)
{
    int         relIndex;
    FILE        *dbfile;

    if (obj1 == NULL || obj2 == NULL) {
        return;
    }

    if ((relIndex = getRelIndex(relName)) < 0) {
        spdBugExit("No such relation %s.\n", relName);
    }

    dbfile = dbRelFile[relIndex];
    spddb_CreateTuple(dbfile);
    if (GET_DBREL(relIndex).attrs[0].classId != SPD_CID_J_WILDCARD &&
        (obj1->symSort != GET_DBREL(relIndex).attrs[0].classId ||
         obj2->symSort != GET_DBREL(relIndex).attrs[1].classId)) {
        spdBugExit("Relation %s must be %s(%d)-%s(%d), "
                   "but this relation is %s(%d)-%s(%d)\n",
                   relName,
                   GET_DBCLASS(symSort2TableIndex(GET_DBREL(relIndex).attrs[0].classId)).name,
                   GET_DBREL(relIndex).attrs[0].classId,
                   GET_DBCLASS(symSort2TableIndex(GET_DBREL(relIndex).attrs[1].classId)).name,
                   GET_DBREL(relIndex).attrs[1].classId,
                   GET_DBCLASS(symSort2TableIndex(obj1->symSort)).name,
                   obj1->symSort,
                   GET_DBCLASS(symSort2TableIndex(obj2->symSort)).name,
                   obj2->symSort);      
    }

    if (obj1->id == JPD_INVALID_ID) {
        writeDBClass(obj1);
    }
    spdWriteIntAttr(dbfile, obj1->id);
    if (obj2->id == JPD_INVALID_ID) {
        writeDBClass(obj2);
    }
    spdWriteIntAttr(dbfile, obj2->id);
    spddb_CommitTuple(dbfile);

    /* Ϥ󥹥󥹿򥫥Ȥ롥 */
    dbRelObjCounter[relIndex]++;
}

/**
 * Function writeXxx() writes object of JpdXxx to SDB using writeDBClass.
 */

void writeFile(SymData *file)
{
    SymData *tmp;

    if (file == NULL || file->symSort != SPD_CID_J_FILE) {
        return;
    }

    writeDBClass(file);
    setSavedJpdFile(file);

    for (tmp = file->u.file->classOrInterface; tmp != NULL;) {
        if (tmp->symSort == SPD_CID_J_CLASS) {
            writeClass(tmp);
            writeDBRel("define_class", file, tmp);
            tmp = tmp->u.jclass->nextClassOrInterface;
        } else if (tmp->symSort == SPD_CID_J_INTERFACE) {
            writeInterface(tmp);
            writeDBRel("define_interface", file, tmp);
            tmp = tmp->u.interface->nextClassOrInterface;
        } else {
            spdBugExit("\"%s\" (%n): Neither class nor interface.\n",
                       tmp->str, tmp->symSort);
        }
    }

    if (file->u.file->definedPackage != NULL) {
        writeQName(file->u.file->definedPackage);
        writeDBRel("define_package",
                   file, file->u.file->definedPackage);
    }

    for (tmp = file->u.file->importedPackage;
         tmp != NULL; tmp = tmp->u.qname->nextQName) {
        writeQName(tmp);
        writeDBRel("import_package", file, tmp);
    }

    /* Write "SDB/J-MODEL/<targetClassName>/.info". */
    write_japid_version();

    extractFile(file);
}

void writeClass(SymData *jclass)
{
    SymData     *arg, *tmp;

    if (jclass == NULL || jclass->symSort != SPD_CID_J_CLASS) {
        return;
    }
    
    writeDBClass(jclass);

    pushMemSymTabStack(jclass->u.jclass->memberSymTab);
    if (linkclass && jclass->u.jclass->sort == PACKAGE_MEMBER_CLASS) {
        mkSymlink(jclass->str);
    }

    for (tmp = jclass->u.jclass->modifier;
         tmp != NULL; tmp = tmp->u.modifier->next) {
        writeDBRel("class_modifier", jclass, tmp);
    }

    writeType(jclass->u.jclass->super);
    writeDBRel("extend_class", jclass, jclass->u.jclass->super);

    writeType(jclass->u.jclass->interface);
    writeDBRelAboutTypeList("implement_interface",
                            jclass, jclass->u.jclass->interface);

    for (arg = jclass->u.jclass->arguments;
         arg != NULL; arg = arg->u.expression->next) {
        writeExpression(arg);
        writeDBRel("class_argument", jclass, arg);
    }

    for (tmp = jclass->u.jclass->member; tmp != NULL;) {
        switch (tmp->symSort) {             /* Фϵսˤʤ롥 */
          case SPD_CID_J_VARIABLE:
            writeVariable(tmp);
            writeDBRelAboutVariable("class_variable", jclass, tmp);
            tmp = tmp->u.variable->nextMember;
            break;
          case SPD_CID_J_ROUTINE:
            writeRoutine(tmp);
            writeDBRel("class_routine", jclass, tmp);
            tmp = tmp->u.routine->nextMember;
            break;
          case SPD_CID_J_SBLOCK:
            writeSBlock(tmp);
            writeDBRel("class_sblock", jclass, tmp);
            tmp = tmp->u.sblock->nextMember;
            break;
          case SPD_CID_J_CLASS:
            writeClass(tmp);
            writeDBRel("class_class", jclass, tmp);
            tmp = tmp->u.jclass->nextMember;
            break;
          case SPD_CID_J_INTERFACE:
            writeInterface(tmp);
            writeDBRel("class_interface", jclass, tmp);
            tmp = tmp->u.interface->nextMember;
            break;
          default:
            spdBugExit("%d: Invalid class member.\n", tmp->symSort);
            break;
        }
    }

    extractClass(jclass);
    popMemSymTabStack();
}

void writeInterface(SymData *interface)
{
    SymData     *tmp;

    if (interface == NULL || interface->symSort != SPD_CID_J_INTERFACE) {
        return;
    }
    
    writeDBClass(interface);

    if (linkclass &&
        (interface->u.interface->sort == PACKAGE_MEMBER_INTERFACE)) {
        mkSymlink(interface->str);
    }

    for (tmp = interface->u.interface->modifier;
         tmp != NULL; tmp = tmp->u.modifier->next) {
        writeDBRel("interface_modifier", interface, tmp);
    }

    writeType(interface->u.interface->super);
    writeDBRelAboutTypeList("extend_interface",
                            interface, interface->u.interface->super);

    for (tmp = interface->u.interface->member; tmp != NULL;) {
        switch (tmp->symSort) {             /* Фϵսˤʤ */
          case SPD_CID_J_VARIABLE:          /* ѿ */
            writeVariable(tmp);
            writeDBRelAboutVariable("interface_variable", interface, tmp);
            tmp = tmp->u.variable->nextMember;
            break;
          case SPD_CID_J_ROUTINE:           /* ᥽åɡ󥹥ȥ饯 */
            writeRoutine(tmp);
            writeDBRel("interface_routine", interface, tmp);
            tmp = tmp->u.routine->nextMember;
            break;
          case SPD_CID_J_CLASS:             /* 饹 */
            writeClass(tmp);
            writeDBRel("interface_class", interface, tmp);
            tmp = tmp->u.jclass->nextMember;
            break;
          case SPD_CID_J_INTERFACE:         /* 󥿥ե */
            writeInterface(tmp);
            writeDBRel("interface_interface", interface, tmp);
            tmp = tmp->u.interface->nextMember;
            break;
          default:
            spdBugExit("%d can't be interface member\n", tmp->symSort);
            break;
        }
    }

    extractInterface(interface);
}

void writeVariable(SymData *vars)
{
    SymData *var, *tmp;

    for (var = vars; var != NULL; var = var->u.variable->next) {
        writeDBClass(var);
        if (var->u.variable->type != NULL) {
            writeType(var->u.variable->type);
            writeDBRel("variable_type", var, var->u.variable->type);
        }
        for (tmp = var->u.variable->modifier;
             tmp != NULL; tmp = tmp->u.modifier->next) {
            writeDBRel("variable_modifier", var, tmp);
        }
        if (var->u.variable->initializer != NULL) {
            writeExpression(var->u.variable->initializer);
            writeDBRel("initialize_variable",
                       var, var->u.variable->initializer);
        }

        extractVariable(var);
    }
}

void writeRoutine(SymData *routine)
{
    SymData     *tmp;

    if (routine == NULL || routine->symSort != SPD_CID_J_ROUTINE) {
        return;
    }
    
    writeDBClass(routine);
    
    for (tmp = routine->u.routine->modifier;
         tmp != NULL; tmp = tmp->u.modifier->next) {
        writeDBRel("routine_modifier", routine, tmp);
    }
    writeType(routine->u.routine->returnType);
    writeDBRel("return_type", routine, routine->u.routine->returnType);
    writeType(routine->u.routine->throwType);
    writeDBRelAboutTypeList("throw_type",
                            routine, routine->u.routine->throwType);
    writeVariable(routine->u.routine->parameter);
    writeDBRelAboutVariable("parameter_variable",
                            routine, routine->u.routine->parameter);
    writeStatement(routine->u.routine->body);
    writeDBRel("routine_body", routine, routine->u.routine->body);

    extractRoutine(routine);
}

void writeType(SymData *type)
{
    if (type == NULL || type->symSort != SPD_CID_J_TYPE) {
        return;
    }

    writeDBClass(type);
    if (type->u.type->next != NULL) {
        writeType(type->u.type->next);
        writeDBRel("type_type", type, type->u.type->next);
    }

    writeQName(type->u.type->qname);
    writeDBRel("type_qname", type, type->u.type->qname);

    if (type->u.type->nextType != NULL) {
        writeType(type->u.type->nextType);
    }

    if (type->u.type->jclass != NULL) {
        writeClass(type->u.type->jclass);
        writeDBRel("type_class", type, type->u.type->jclass);
    }
    
    extractType(type);
}

void writeSBlock(SymData *symData)
{
    if (symData == NULL || symData->symSort != SPD_CID_J_SBLOCK) {
        return;
    }

    writeDBClass(symData);
    writeStatement(symData->u.sblock->block);
    writeDBRel("sblock_body", symData, symData->u.sblock->block);

    extractSBlock(symData);
}

void writeStatement(SymData *stmt)
{
    SymData     *tmp;

    if (stmt == NULL || stmt->symSort != SPD_CID_J_STATEMENT) {
        return;
    }

    writeDBClass(stmt);
    switch (stmt->u.statement->sort) {
      case S_EMPTY:
        break;
      case S_EXPRESSION:
        writeExpression(getStatementExpression(stmt));
        writeDBRel("statement_expression", stmt, getStatementExpression(stmt));
        break;
      case S_DEF_VAR:
        writeVariable(getStatementVariableDecl(stmt));
        writeDBRelAboutVariable("local_variable",
                                stmt, getStatementVariableDecl(stmt));
        break;
      case S_BLOCK:
        /* S_BLOCK is just BLOCK, not static block. */
        for (tmp = stmt->u.statement->block;
             tmp != NULL; tmp = tmp->u.statement->next) {
            writeStatement(tmp);
            writeDBRel("statement_statement", stmt, tmp);
        }
        break;
      case S_IF:
        writeExpression(getStatementCondition(stmt));
        writeDBRel("statement_expression", stmt, getStatementCondition(stmt));
        writeStatement(getStatementIfThen(stmt));
        writeDBRel("statement_statement", stmt, getStatementIfThen(stmt));
        writeStatement(getStatementIfElse(stmt));
        writeDBRel("statement_statement", stmt, getStatementIfElse(stmt));
        break;
      case S_SWITCH:
        writeExpression(getStatementCondition(stmt));
        writeDBRel("statement_expression", stmt, getStatementCondition(stmt));
        writeStatement(getStatementAction(stmt));
        writeDBRel("statement_statement", stmt, getStatementAction(stmt));
        break;
      case S_WHILE:
        writeExpression(getStatementCondition(stmt));
        writeDBRel("statement_expression", stmt, getStatementCondition(stmt));
        writeStatement(getStatementAction(stmt));
        writeDBRel("statement_statement", stmt, getStatementAction(stmt));
        break;
      case S_DO:
        writeExpression(getStatementCondition(stmt));
        writeDBRel("statement_expression", stmt, getStatementCondition(stmt));
        writeStatement(getStatementAction(stmt));
        writeDBRel("statement_statement", stmt, getStatementAction(stmt));
        break;
      case S_FOR:
        /* Assertion. */
        if (getStatementForInit(stmt) == NULL) {
            spdBugExit("FOR INIT is null");
        }
        if (getStatementCondition(stmt) == NULL) {
            spdBugExit("FOR EXPR is null");
        }
        if (getStatementForIncr(stmt) == NULL) {
            spdBugExit("FOR INCR is null");
        }

        if (getStatementForInit(stmt)->symSort == SPD_CID_J_STATEMENT) {
            /* ѿ */ /*  */
            SymData *varStmt = getStatementForInit(stmt);
            writeVariable(getStatementVariableDecl(varStmt));
            writeDBRelAboutVariable("local_variable",
                                    stmt, getStatementVariableDecl(varStmt));
        } else {                            /*  */
            for (tmp = getStatementForInit(stmt);
                 tmp != NULL; tmp = tmp->u.expression->next) {
                writeExpression(tmp);
                writeDBRel("statement_expression", stmt, tmp);
            }
        }

        writeExpression(getStatementCondition(stmt));
        writeDBRel("statement_expression", stmt, getStatementCondition(stmt));

        for (tmp = getStatementForIncr(stmt);
             tmp != NULL; tmp = tmp->u.expression->next) {
            writeExpression(tmp);
            writeDBRel("statement_expression", stmt, tmp);
        }

        writeStatement(getStatementAction(stmt));
        writeDBRel("statement_statement", stmt, getStatementAction(stmt));
        break;
      case S_BREAK:
      case S_CONTINUE:
        break;
      case S_RETURN:
        writeExpression(getStatementExpression(stmt));
        writeDBRel("statement_expression", stmt, getStatementExpression(stmt));
        break;
      case S_THROW:
        writeExpression(getStatementExpression(stmt));
        writeDBRel("statement_expression", stmt, getStatementExpression(stmt));
        break;
      case S_EXCEPTION:
        writeStatement(getStatementExceptionTry(stmt));
        writeDBRel("statement_statement",
                   stmt, getStatementExceptionTry(stmt));

        for (tmp = getStatementExceptionCatches(stmt);
             tmp != NULL; tmp = tmp->u.statement->next) {
            writeStatement(tmp);            /* tmp is a catch block */
            writeDBRel("statement_statement", stmt, tmp);
        }
        writeStatement(getStatementExceptionFinally(stmt));
        writeDBRel("statement_statement",
                   stmt, getStatementExceptionFinally(stmt));
        break;
      case S_TRY:
        writeStatement(getStatementTryBlock(stmt));
        writeDBRel("statement_statement",
                   stmt, getStatementTryBlock(stmt));
        break;

      case S_CATCH:
        writeVariable(getStatementCatchArgument(stmt));
        writeDBRel("local_variable",
                   stmt, getStatementCatchArgument(stmt));
        writeStatement(getStatementCatchBlock(stmt));
        writeDBRel("statement_statement",
                   stmt, getStatementCatchBlock(stmt));
        break;
      case S_FINALLY:
        writeStatement(getStatementFinallyBlock(stmt));
        writeDBRel("statement_statement",
                   stmt, getStatementFinallyBlock(stmt));
        break;
      case S_SYNCHRONIZED:
        writeExpression(getStatementSynchronizedExpression(stmt));
        writeDBRel("statement_expression",
                   stmt, getStatementSynchronizedExpression(stmt));
        writeStatement(getStatementSynchronizedStatement(stmt));
        writeDBRel("statement_statement",
                   stmt, getStatementSynchronizedStatement(stmt));
        break;
      case S_SWITCH_LABELED_EMPTY:
        break;
      case S_LOCAL_CLASS:
        writeClass(getStatementClass(stmt));
        writeDBRel("local_class",
                   stmt, getStatementClass(stmt));
        break;
      case S_ASSERT:
        writeExpression(getStatementAssertExpression1(stmt));
        writeDBRel("statement_expression",
                   stmt, getStatementAssertExpression1(stmt));
        if (getStatementAssertExpression2(stmt) != NULL) {
            writeExpression(getStatementAssertExpression2(stmt));
            writeDBRel("statement_expression",
                       stmt, getStatementAssertExpression2(stmt));
        }
        break;
      default:
        spdBugExit("Unknown statement: %d (l.%d, c.%d).\n",
                   stmt->u.statement->sort, yylineno, c_count);
        break;
    }

    if (hasStatementLabel(stmt)) {
        SymData *label;
        for (label = getStatementLabel(stmt); label != NULL;
             label = label->u.label->next) {
            writeLabel(label);
            writeDBRel("statement_label", stmt, label);
        }
    }

    extractStatement(stmt);
}

void writeExpression(SymData *expr)
{
    SymData     *tmp;

    if (expr == NULL || expr->symSort != SPD_CID_J_EXPRESSION) {
        return;
    }

    unlockLRSort(expr);
    if (expr->u.expression->lrsort == JPD_INVALID_ID) {
        spdWarning("Expr: lrsort is not set %d (%d - %d).\n",
                   expr->u.expression->sort, expr->offset, expr->length);
    }
    writeDBClass(expr);

    switch (expr->u.expression->sort) {
      case E_EMPTY:
        break;
      case E_REF_VAR:
        if (expr->u.expression->first != NULL) {
            writeDBRel("refer_variable", expr, expr->u.expression->first);
        }
        else {
            if (isTwoPassResolve(expr)) {
                expr->u.expression->first = searchForMemSymTabStack(expr->str);
                if (expr->u.expression->first != NULL) {
                    writeDBRel("refer_variable",
                               expr, expr->u.expression->first);
                }
            }
        }
        break;
      case E_REF_ARRAY:
        break;
      case E_REF_LITERAL:
        writeLiteral(expr->u.expression->first);
        writeDBRel("refer_literal", expr, expr->u.expression->first);
        break;
      case E_REF_TYPE:
        writeType(expr->u.expression->first);
        writeDBRel("refer_type", expr, expr->u.expression->first);
        break;
      case E_REF_SPECIAL:
        writeSpecial(expr->u.expression->first);
        writeDBRel("refer_special", expr, expr->u.expression->first);
        break;
      case E_METHOD_CALL:
        for (tmp = getMethodArguments(expr);
             tmp != NULL; tmp = tmp->u.expression->next) {
            writeExpression(tmp);
            writeDBRel("expression_expression", expr, tmp);
        }
        break;
      case E_CONSTRUCTOR_CALL:
        if (expr->u.expression->first->symSort == SPD_CID_J_SPECIAL) {
            /* super(...) */
            writeSpecial(expr->u.expression->first);
            writeDBRel("refer_special", expr, expr->u.expression->first);
        } else if (expr->u.expression->first->symSort == SPD_CID_J_TYPE) {
            /* new Type(...) */
            writeType(expr->u.expression->first);
            writeDBRel("refer_type", expr, expr->u.expression->first);
        } else {
            spdErrorExit("Unknown constructor call to \"%s\" (%d).\n",
                         expr->u.expression->first->str,
                         expr->u.expression->first->u.expression->sort);
        }
            
        for (tmp = getMethodArguments(expr);
             tmp != NULL; tmp = tmp->u.expression->next) {
            writeExpression(tmp);
            writeDBRel("expression_expression", expr, tmp);
        }
        break;
      case E_OPERATION:
        writeOperator(expr->u.expression->joperator);
        writeDBRel("expression_operator", expr, expr->u.expression->joperator);
        switch (expr->u.expression->joperator->u.joperator->sort) {
          case JPD_O_PREFIX:
          case JPD_O_POSTFIX:
          case JPD_O_PAREN:
            writeExpression(expr->u.expression->first);
            writeDBRel("expression_operand", expr, expr->u.expression->first);
            break;
          case JPD_O_REF_DIM:
          case JPD_O_BINARY:
            writeExpression(expr->u.expression->first);
            writeDBRel("expression_operand", expr, expr->u.expression->first);
            writeExpression(expr->u.expression->second);
            writeDBRel("expression_operand", expr, expr->u.expression->second);
            break;
          case JPD_O_TERNARY:
            writeExpression(expr->u.expression->first);
            writeDBRel("expression_operand", expr, expr->u.expression->first);
            writeExpression(expr->u.expression->second);
            writeDBRel("expression_operand", expr, expr->u.expression->second);
            writeExpression(expr->u.expression->third);
            writeDBRel("expression_operand", expr, expr->u.expression->third);
            break;
          case JPD_O_CAST:
            writeExpression(expr->u.expression->first);
            writeDBRel("expression_operand", expr, expr->u.expression->first);
            writeExpression(expr->u.expression->second);
            writeDBRel("expression_operand", expr, expr->u.expression->second);
            break;
          case JPD_O_NEW_ARRAY:             /** type[size][], etc */
            writeExpression(expr->u.expression->first); /* type */
            writeDBRel("expression_operand", expr, expr->u.expression->first);

            /* [size][] */
            for (tmp = getDimExprs(expr);
                 tmp != NULL; tmp = tmp->u.expression->next) {
                writeExpression(tmp);
                writeDBRel("expression_expression", expr, tmp);
            }
            break;
          case JPD_O_NEW_OBJECT:
            writeExpression(expr->u.expression->first); /* constructor call */
            writeDBRel("expression_operand", expr, expr->u.expression->first);
            break;
          default:
            spdBugExit("Unknown operator: '%s'(%d)\n",
                       expr->u.expression->joperator->str,
                       expr->u.expression->joperator->u.joperator->sort);
        }
        break;
      case E_REF_QNAME:
        writeQName(expr->u.expression->first);
        writeDBRel("refer_qname", expr, expr->u.expression->first);
        break;
      case E_LISTED:
        for (tmp = expr->u.expression->first;
             tmp != NULL; tmp = tmp->u.expression->next) {
            writeExpression(tmp);
            writeDBRel("expression_expression", expr, tmp);
        }
        break;
      case E_DIM_EXPR:                      /* e.g. [], [10] of Type[10][] */
        writeOperator(expr->u.expression->joperator);
        writeDBRel("expression_operator", expr, expr->u.expression->joperator);
        if (expr->u.expression->first != NULL) {
            writeExpression(expr->u.expression->first);
            writeDBRel("expression_operand",
                       expr, expr->u.expression->first);
        }
        break;
      case E_ANON_CLASS:
        writeType(getAnonClassType(expr));
        writeDBRel("refer_type", expr, getAnonClassType(expr));
        {
            SymData     *jclass = (getAnonClassType(expr))->u.type->jclass;
            SymData     *arg;

            for (arg = jclass->u.jclass->arguments;
                 arg != NULL; arg = arg->u.expression->next) {
                writeDBRel("expression_expression", expr, arg);
            }
        }
        
        break;
      case E_ANON_ARRAY:
        writeOperator(expr->u.expression->joperator);
        writeDBRel("expression_operator",
                   expr, expr->u.expression->joperator);
        writeExpression(expr->u.expression->first);
        writeDBRel("expression_operand",
                   expr, expr->u.expression->first);
        writeExpression(expr->u.expression->second);
        writeDBRel("expression_operand",
                   expr, expr->u.expression->second);
        break;
      default:
        spdBugExit("Unknown expression: %d (offset %d, length %d).\n",
                   expr->u.expression->sort, expr->offset, expr->length);
        break;
    }
    
    extractExpression(expr);
}

void writeLiteral(SymData *symData)
{
    if (symData == NULL || symData->symSort != SPD_CID_J_LITERAL) {
        return;
    }

    writeDBClass(symData);

    extractLiteral(symData);
}

void writeOperator(SymData *anOperator)
{
    if (anOperator == NULL || anOperator->symSort != SPD_CID_J_OPERATOR) {
        return;
    }

    writeDBClass(anOperator);
}

void writeQName(SymData *qname)
{
    if (qname == NULL || qname->symSort != SPD_CID_J_QNAME) {
        return;
    }

    writeDBClass(qname);
    if (qname->u.qname->dot != NULL) {
        writeQName(qname->u.qname->first);
        writeDBRel("qname_qname", qname, qname->u.qname->first);
        writeQName(qname->u.qname->second);
        writeDBRel("qname_qname", qname, qname->u.qname->second);
        writeOperator(qname->u.qname->dot);
        writeDBRel("qname_dot", qname, qname->u.qname->dot);
    }

    extractQName(qname);
}

void writeSpecial(SymData *special)
{
    if (special == NULL || special->symSort != SPD_CID_J_SPECIAL) {
        return;
    }

    writeDBClass(special);
}

void writeLabel(SymData *label)
{
    if (label == NULL || label->symSort != SPD_CID_J_LABEL) {
        return;
    }

    writeDBClass(label);
    switch (label->u.label->sort) {
      case DEF_LABEL:
        break;
      case REF_LABEL:
        writeDBRel("refer_label", label, label->u.label->defined);
        break;
      case CASE_LABEL:
        writeExpression(label->u.label->caseValue);
        writeDBRel("case_value", label, label->u.label->caseValue);
        break;
      case DEFAULT_LABEL:
        break;
      default:
        spdBugExit("Unknown label: %d (l.%d, c.%d).\n",
                   label->u.label->sort, yylineno, c_count);
    }

    extractLabel(label);
}

void writeDBRelAboutVariable(char *relName, SymData *obj, SymData *vars)
{
    SymData     *var;

    for (var = vars; var != NULL; var = var->u.variable->next) {
        writeDBRel(relName, obj, var);
    }
}

void writeDBRelAboutTypeList(char *relName, SymData *obj, SymData *type)
{
    for (/* Nothing. */; type != NULL; type = type->u.type->nextType) {
        writeDBRel(relName, obj, type);
    }
}

void initDB(void)
{
    int                 i;
    static SpdBoolean   is_first_call = TRUE;

    /* ֥ȥ󥿤롥 */
    init_obj_counter();

    for (i = 0; i < SPD_CID_NUM_J_MODEL; i++) {
        if (GET_DBCLASS(i).id != SPD_FIRST_CID_J_MODEL + i) {
            spdBugExit("%d: Invalid class id.\n", i);
        }
    }

    if (is_first_call) {
        is_first_call = FALSE;
    } else {
        /* ̤ΥեϤ塥 */
        for (i = 0; i < SPD_CID_NUM_J_MODEL; i++) {
            spdFclose(dbClassFile[i]);
        }
        for (i = 0; i < SPD_RID_NUM_J_MODEL; i++) {
            spdFclose(dbRelFile[i]);
        }
    }

    for (i = 0; i < SPD_CID_NUM_J_MODEL; i++) {
        dbClassID[i] = 1;
        dbClassFile[i] = spdOpenFreja(frejaDir, GET_DBCLASS(i).name,
                                      SAPID_O_WRITE);
    }
    for (i = 0; i < SPD_RID_NUM_J_MODEL; i++) {
        dbRelFile[i] = spdOpenFreja(frejaDir, GET_DBREL(i).name, 
                                    SAPID_O_WRITE);
    }
}

/* ֥ȥ󥿤ν */

static void init_obj_counter(void)
{
    int     i;
    
    for (i = 0; i < SPD_CID_NUM_J_MODEL; i++) {
        dbClassObjCounter[i] = 0;
    }
    for (i = 0; i < SPD_RID_NUM_J_MODEL; i++) {
        dbRelObjCounter[i] = 0;
    }
}

/* 
 * Write "SDB/J-MODEL/<targetClassName>/.info".
 */

static void write_japid_version(void)
{
    SpdInfo         *info;
    int             i;
    static char     fileName[FILENAME_MAX + 1];

    sprintf(fileName, "%s/%s", J_MODEL_TMP_DIR, frejaDir);
    info = spdOpenJSDBinfo(fileName);

    spdSetInfoInt(info, "SapidVersion", SapidVersion);
    spdSetInfoInt(info, "SapidRevision", SapidRevision);
    spdSetInfoInt(info, "SapidPatchLevel", SapidPatchLevel);
    spdSetInfoInt(info, "SapidSDBVersion", SapidSDBVersion);
    spdSetInfoString(info, "BINARY_FORMAT_SDB",
                     spdGetWriteSDBformat() == SAPID_BINARY_FORMAT_SDB ?
                     "TRUE" : "FALSE");
    spdSetInfoString(info, "MILL", "TRUE");
    spdSetInfoInt(info, "SPD_FIRST_CID_J_MODEL", SPD_FIRST_CID_J_MODEL);
    spdSetInfoInt(info, "SPD_CID_NUM_J_MODEL", SPD_CID_NUM_J_MODEL);
    spdSetInfoInt(info, "SPD_RID_NUM_J_MODEL", SPD_RID_NUM_J_MODEL);

    /* ֥ȿ񤭽Ф */
    for (i = 0; i < SPD_CID_NUM_J_MODEL; i++) {
        if (dbClassObjCounter[i] > JPD_MAX_OID) {
            spdWarning("%s, %s, %d: Too many objects.\n", 
                       frejaDir, GET_DBCLASS(i).name, dbClassObjCounter[i]);
        }
        spdSetInfoInt(info, GET_DBCLASS(i).name, dbClassObjCounter[i]);
    }
    for (i = 0; i < SPD_RID_NUM_J_MODEL; i++) {
        if (dbRelObjCounter[i] > JPD_MAX_OID) {
            spdWarning("%s, %s, %d: Too many relations.\n", 
                       frejaDir, GET_DBREL(i).name, dbRelObjCounter[i]);
        }
        spdSetInfoInt(info, GET_DBREL(i).name, dbRelObjCounter[i]);
    }

    spdCloseInfo(info);
}

static int getRelIndex(char *relName)
{
    SpdClassId      rId = spdGetClassIdByName(relName);

    if (SPD_FIRST_RID_J_MODEL <= rId && rId <= SPD_LAST_RID_J_MODEL) {
        assert ((rId - SPD_FIRST_RID_J_MODEL) == getRelIndexOld(relName));
        return (rId - SPD_FIRST_RID_J_MODEL);
    } else {
        assert (-1 == getRelIndexOld(relName));
        return (-1);
    }
}

static int getRelIndexOld(char *relName)
{
    int     i;
    
    for (i = 0; i < SPD_RID_NUM_J_MODEL; i++) {
        if (SPD_STREQ(relName, GET_DBREL(i).name)) {
            return (i);
        }
    }
    return (-1);
}

/* 
 * For example in "<SAPID_TOP>/test/java-zip":
 *     spdSymlink("ZipPack", "SDB/J-MODEL/ZipPackDummy");
 */

#if !defined(WIN32)

void mkSymlink(char *className)
{
    char    toDir[FILENAME_MAX + 1];

    if (!SPD_STREQ(frejaDir, className)) {
        sprintf(toDir, "%s/" J_MODEL_TMP_DIR "/%s", spdGetSDBdir(), className);

#ifdef OBSOLETE_030127
        {
            struct stat     sbuf;

            if ((lstat(toDir, &sbuf) == 0) && (sbuf.st_mode & S_IFLNK)) {
                /*
                 * toDirϴ¸ߤsymbolic linkǤ롥
                 * ŤSDBʤǡƤӲϤǽ⤤
                 */
                spdUnlink(toDir);
            }
        }
#endif /* OBSOLETE_030127 */

        /* ۤȤɤξ硤toDir¸ߤʤǰΤunlink()롥 */
        unlink(toDir);

        spdSymlink(frejaDir, toDir);
    }
}

#else /* defined(WIN32) */

void mkSymlink(char *className)
{
    spdWarning("mkSymlink(): Not supported yet.\n");
}

#endif /* defined(WIN32) */

/******************************************************************************/

/*
 * ѿ˻Ȥн褹뤿ν
 * 
 * (1) writeClass()ǥ饹Υѿεɽ򥹥åѤǤ
 *     饹ΤǥåɬפȤʤ롥
 * (2) writeExpression()ǻΤʤѿФƤȤϡ
 *     åεɽѿõ
 */

#define MAX_CLASS_NEST_LEVEL    8
#define STACK_EMPTY_INDEX       (-1)

static VarSymTab    *memSymTabStack[MAX_CLASS_NEST_LEVEL];
static int          memSymTabStackTopIndex = STACK_EMPTY_INDEX;

static void pushMemSymTabStack(VarSymTab *tab)
{
    if (memSymTabStackTopIndex == MAX_CLASS_NEST_LEVEL-1) {
        spdErrorExit("Member symbol table stack overflow\n");
    }
    memSymTabStack[++memSymTabStackTopIndex] = tab;
}

static VarSymTab *popMemSymTabStack(void)
{
    if (memSymTabStackTopIndex == STACK_EMPTY_INDEX) {
        spdErrorExit("BUG: Member symbol table stack underflow\n");
    }
    return memSymTabStack[memSymTabStackTopIndex--];
}

/* åѤǤѿεɽѿnameSymDataõ */

static SymData *searchForMemSymTabStack(char *name)
{
    SymData *var = NULL;

    if (memSymTabStackTopIndex != STACK_EMPTY_INDEX) {
        var = searchForMemberSymTab(name,
                                    memSymTabStack[memSymTabStackTopIndex]);
    }
    return var;
}
