/* 
 * Program:             $RCSfile: j2jx.c,v $  $Revision: 40.63 $
 * 
 * Purpose:             Generate elements list for JX-model3.
 *
 * Author:              S.Yamamoto  2003/02/07
 *
 * (C) Copyright:       S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: j2jx.c,v 40.63 2003/07/19 09:44:44 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "JB.h"

/* #define CAST_TYPEFIRST*/
/* #define RESOLVE_METHOD */
/* #define RESOLVE_OBJECT_TYPE */

#define RESOLVE_METHOD_DW(b, a)         debugWrite((b), (a))
/* #define CAST_TYPEFIRST_DW(b, a)         debugWrite((b), (a)) */
#define CAST_TYPEFIRST_DW(b, a)
#define RESOLVE_OBJECT_TYPE_DW(b, a)    debugWrite((b), (a))

static SymData          *savedJpdFile = NULL;
static SpdString        fileBuf = NULL;
static int              fileLen = -1;

/* Counter for derived id. Incremented in makeIDAttrDerived(). */
static SpdObjId         derivedId = 0;

static int getAttrInt(SymData *symData, SpdConstString attrName);

#ifdef RESOLVE_METHOD
static SymData *resolveMethodSub(SymData *jclass, SpdConstString methodName);
#endif

/* !! Dirty hack. */
extern SpdDBClass J_model_classes[];

static char *debugWrite(char *buf, char *attr)
{
    return (jpdDebug.etb ? strcat(buf, attr) : NULL);
}

static int getId(SymData *data)
{
    return (getAttrInt(data, "id"));
}

static int getOffset(SymData *data)
{
    return (getAttrInt(data, "offset"));
}

static int getLength(SymData *data)
{
    return (getAttrInt(data, "length"));
}

static SpdBoolean hasValidID(SymData *data)
{
    return (data->id != JPD_INVALID_ID);
}

static Range mkRange(int s, int e)
{
    Range      ret;

    ret.sp = s;
    ret.ep = e;
    return (ret);
}

/* Get range of type family. */

static Range getTypeFamily(SymData *type)
{
    int sp = getOffset(type);
    int ep = sp + getLength(type);

    while (TRUE) {
        int nsp;
        int nep;
        SymData *t2 = type->u.type->next;

        if (t2 == NULL) {
            break;
        }

        nsp = getOffset(t2);
        nep = nsp + getLength(t2);
        if (nsp < sp) sp = nsp;
        if (ep < nep) ep = nep;
        type = t2;
    }

    {
        SymData     *q = type->u.type->qname;
        
        if (q != NULL) {
            if (getOffset(q) < sp) {
                sp = getOffset(q);
            }
        }
    }

    return (mkRange(sp, ep));
}

static int getAttrInt(SymData *symData, SpdConstString attrName)
{
    char                *symP, *unionP;
    SpdDBClass          *table;
    SpdDBAttr           *attr;
    int                 i;

    symP = (char *) symData;
    table = &(GET_DBCLASS(symSort2TableIndex(symData->symSort)));
    if (table->id != symData->symSort) {
        spdBugExit("%d != %d: Invalid DBClassTable.\n",
                   table->id, symData->symSort);
    }

    for (i = 0; i < SPD_ISVAL_NUM; i++) {
        attr = &(table->attrs[i]);
        if (attr->type == SPD_DBT_NON) {
            break;
        } else if (SPD_STREQ(attr->name, attrName)) {
            if (attr->type == SPD_DBT_INT) {
                if (attr->unionOffset == SDB_NON_OFFSET) {
                    /* See "lib/SpdUtil/spdIO.c". */
                    return (*((int *)(symP + attr->offset)));
                } else {
                    unionP = *((char **)(symP + attr->offset));

                    return (*((int *)(unionP + attr->unionOffset)));
                }
            } else {
                spdErrorExit("getAttrInt(): 1: \"%s\": Invalid attrName.\n",
                             attrName);
            }
        }
    }
    spdErrorExit("getAttrInt(): 2: \"%s\": Invalid attrName.\n", attrName);
    SPD_DUMMY_RETURN(0);
} 

static SpdString getAttrStr(SymData *symData, SpdConstString attrName)
{
    char                *symP, *unionP;
    SpdDBClass          *table;
    SpdDBAttr           *attr;
    int                 i;

    symP = (char *) symData;
    table = &(GET_DBCLASS(symSort2TableIndex(symData->symSort)));
    if (table->id != symData->symSort) {
        spdBugExit("%d != %d: Invalid DBClassTable.\n",
                   table->id, symData->symSort);
    }

    for (i = 0; i < SPD_ISVAL_NUM; i++) {
        attr = &(table->attrs[i]);
        if (attr->type == SPD_DBT_NON) {
            break;
        } else if (SPD_STREQ(attr->name, attrName)) {
            if (attr->type == SPD_DBT_STR) {
                if (attr->unionOffset == SDB_NON_OFFSET) {
                    return (*((char **)(symP + attr->offset)));
                } else {
                    unionP = *((char **)(symP + attr->offset));

                    return (*((char **)(unionP + attr->unionOffset)));
                }
            } else {
                spdErrorExit("getAttrInt(): 1: \"%s\": Invalid attrName.\n",
                             attrName);
            }
        }
    }
    spdErrorExit("getAttrInt(): 2: \"%s\": Invalid attrName.\n", attrName);
    SPD_DUMMY_RETURN(NULL);
}

#ifdef RESOLVE_METHOD

static SymData *resolveMethod(SymData *expr)
{
    SymData     *tmp, *ret = NULL;
    SpdString   methodName = getAttrStr(expr, "str");

    /* spdMessage("resolveMethod(): \"%s\".\n", methodName); */
    for (tmp = savedJpdFile->u.file->classOrInterface; tmp != NULL;) {
        /* spdMessage("++++\"%s\".\n", getAttrStr(tmp, "str")); */
        if (tmp->symSort == SPD_CID_J_CLASS) {
            /* spdMessage("aaaa\"%s\".\n", getAttrStr(tmp, "str")); */
            ret = resolveMethodSub(tmp, methodName);
            tmp = tmp->u.jclass->nextClassOrInterface;
        } else if (tmp->symSort == SPD_CID_J_INTERFACE) {
            /* spdMessage("bbbb\"%s\".\n", getAttrStr(tmp, "str")); */
            ret = resolveMethodSub(tmp, methodName);
            tmp = tmp->u.interface->nextClassOrInterface;
        } else {
            spdBugExit("\"%s\" (%n): Neither class nor interface.\n",
                       tmp->str, tmp->symSort);
        }
        if (ret != NULL) {
            return (ret);
        }
    }
    return (NULL);
}

static SymData *resolveMethodSub(SymData *jclass, SpdConstString methodName)
{
    SymData     *tmp;

    for (tmp = jclass->u.jclass->member; tmp != NULL;) {
        switch (tmp->symSort) {             /* Фϵսˤʤ롥 */
          case SPD_CID_J_VARIABLE:
            tmp = tmp->u.variable->nextMember;
            break;
          case SPD_CID_J_ROUTINE:
            /* spdMessage("    \"%s\".\n", getAttrStr(tmp, "str")); */
            /* !! Check number of arguments. */
            if (SPD_STREQ(methodName, getAttrStr(tmp, "str"))) {
                /* spdMessage("    Found \"%s\", %d.\n",           */
                /*            getAttrStr(tmp, "str"), getId(tmp)); */
                return (tmp);
            }
            tmp = tmp->u.routine->nextMember;
            break;
          case SPD_CID_J_SBLOCK:
            tmp = tmp->u.sblock->nextMember;
            break;
          case SPD_CID_J_CLASS:
            tmp = tmp->u.jclass->nextMember;
            break;
          case SPD_CID_J_INTERFACE:
            tmp = tmp->u.interface->nextMember;
            break;
          default:
            spdBugExit("%d: Invalid class member.\n", tmp->symSort);
            break;
        }
    }

  return (NULL);
}

#endif /* RESOLVE_METHOD */

void initExtractFile(void)
{
    if (fileBuf == NULL) {
        fileLen = spdReadJavaInBufByName(yyfilename, &fileBuf);
    }
}

void cleanUpExtractFile(void)
{
    if (fileBuf != NULL) {
        spdFree(fileBuf);
        fileBuf = NULL;
    }
    fileLen = -1;
}

/* Search string offset to forward. */

static int searchOffsetFW(int start, int end, SpdString target)
{
    int         pos;
    int         len = strlen(target);

    for (pos = start; pos <= end - len; pos++) {
        int     q = 0;

        while (q < len && fileBuf[pos + q] == target[q]) {
            q++;
        }
        if (q == len) {
            return pos;
        }
    }
    return (-1);
}

/* Search string offset to backward. */

static int searchOffsetBW(int start, int end, SpdString target)
{
    int         pos;
    int         len = strlen(target);

    for (pos = start - len; end <= pos; pos--) {
        int q = 0;
        while (q < len && fileBuf[pos + q] == target[q]) {
            q++;
        }
        if (q == len) {
            return pos;
        }
    }
    return (-1);
}

/* Output element spec. */

static void putElementSpec(SpdString name, int offset, int length,
                           int depth, SpdString attr)
{
    static char attrsp[SPD_XML_ATTR_LEN + 1];

    if (0 < strlen(attr) && attr[0] != ' ') {
        strcpy(attrsp, " ");
        strcat(attrsp, attr);
    } else {
        strcpy(attrsp, attr);
    }

    /* !! Use FrjObject.INVALID_ID. */
    if (offset == -1 || length == -1) {
        spdErrorExit("putElementSpec(): Invalid argument.");
    } else {
        fprintf(j2jxEtb, "%d\t%d\t%d\t%s%s\n",
                offset, length, depth, name, attrsp);
    }
}

/* Make ID attribute. */

static SpdString makeIDAttr(SpdString pre, long id)
{
    static char attr[SPD_XML_ATTR_LEN + 1];

    sprintf(attr, " %s=\"s%ld\"", pre, id);
    return (attr);
}

/* Make derived ID attribute. */

static SpdString makeIDAttrDerived(SpdString pre)
{
    static char attr[SPD_XML_ATTR_LEN + 1];

    sprintf(attr, " %s=\"d%ld\"", pre, derivedId++);
    return (attr);
}

/* Make modifier's attribute. */

static SpdString makeModAttr(SymData *symData)
{
    SymData     *tmp;
    static char attr[SPD_XML_ATTR_LEN + 1];

    attr[0] = '\0';
    
    if (symData->symSort == SPD_CID_J_CLASS) {
        tmp = symData->u.jclass->modifier;
    } else if (symData->symSort == SPD_CID_J_INTERFACE) {
        tmp = symData->u.interface->modifier;
    } else if (symData->symSort == SPD_CID_J_ROUTINE) {
        tmp = symData->u.routine->modifier;
    } else if (symData->symSort == SPD_CID_J_VARIABLE) {
        tmp = symData->u.variable->modifier;
    } else {
        spdErrorExit("makeModAttr(): %d: Invalid sort.\n", symData->symSort);
    } 

    for (/* Nothing */; tmp != NULL; tmp = tmp->u.modifier->next) {
        char    *cp = tmp->str;

        if (SPD_STREQ(cp, "abstract")) strcat(attr, " abstract=\"yes\"");
        if (SPD_STREQ(cp, "final")) strcat(attr, " final=\"yes\"");
        if (SPD_STREQ(cp, "transient")) strcat(attr, " transient=\"yes\"");
        if (SPD_STREQ(cp, "volatile")) strcat(attr, " volatile=\"yes\"");
        if (SPD_STREQ(cp, "static")) strcat(attr, " static=\"yes\"");
        if (SPD_STREQ(cp, "synchronized")) strcat(attr, " synchro=\"yes\"");
        if (SPD_STREQ(cp, "native")) strcat(attr, " native=\"yes\"");
        if (SPD_STREQ(cp, "strictfp")) strcat(attr, " strictfp=\"yes\"");
    
        if (SPD_STREQ(cp, "public")) strcat(attr, " access=\"Public\"");
        if (SPD_STREQ(cp, "protected")) strcat(attr, " access=\"Protected\"");
        if (SPD_STREQ(cp, "private")) strcat(attr, " access=\"Private\"");
    }
    return (attr);
}

/* Make operator's attribute. */

static SpdString makeOpAttr(SymData *symData)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    SpdString   s = getAttrStr(symData->u.expression->joperator, "str");
    int         p = getAttrInt(symData->u.expression->joperator, "sort");

    /* spdMessage("makeOpAttr(): \"%s\".\n", s); */
    attr[0] = '\0';

    if (SPD_STREQ(s, ".")) { strcpy(attr, " sort=\"DOT\""); }
    else if (SPD_STREQ(s, "=")) { strcpy(attr, " sort=\"Assign\""); }
    else if (SPD_STREQ(s, "+=")) { strcpy(attr, " sort=\"AssignADD\""); }
    else if (SPD_STREQ(s, "-=")) { strcpy(attr, " sort=\"AssignSUB\""); }
    else if (SPD_STREQ(s, "*=")) { strcpy(attr, " sort=\"AssignMUL\""); }
    else if (SPD_STREQ(s, "/=")) { strcpy(attr, " sort=\"AssignDIV\""); }
    else if (SPD_STREQ(s, "%=")) { strcpy(attr, " sort=\"AssignMOD\""); }
    else if (SPD_STREQ(s, "&=")) { strcpy(attr, " sort=\"AssignAND\""); }
    else if (SPD_STREQ(s, "|=")) { strcpy(attr, " sort=\"AssignOR\""); }
    else if (SPD_STREQ(s, "^=")) { strcpy(attr, " sort=\"AssignXOR\""); }
    else if (SPD_STREQ(s, "<<=")) { strcpy(attr, " sort=\"AssignSHL\""); }
    else if (SPD_STREQ(s, ">>=")) { strcpy(attr, " sort=\"AssignSHR\""); }
    else if (SPD_STREQ(s, ">>>=")) { strcpy(attr, " sort=\"AssignSHRZ\""); }

    else if (p == JPD_O_REF_DIM) { strcpy(attr, " sort=\"ArrayAccess\""); }
    else if (p == JPD_O_NEW_OBJECT) { strcpy(attr, " sort=\"InstanceCreation\""); }
    else if (p == JPD_O_NEW_ARRAY) { strcpy(attr, " sort=\"ArrayCreation\""); }
    else if (p == JPD_O_CAST) { strcpy(attr, " sort=\"Cast\""); }
    else if (p == JPD_O_PAREN) { strcpy(attr, " sort=\"Paren\""); }
    else if (p == JPD_O_POSTFIX) {
        if (SPD_STREQ(s, "++")) { strcpy(attr, " sort=\"PostINC\""); }
        if (SPD_STREQ(s, "--")) { strcpy(attr, " sort=\"PostDEC\""); }
    } else if (p == JPD_O_PREFIX) {
        if (SPD_STREQ(s, "++")) { strcpy(attr, " sort=\"PreINC\""); }
        if (SPD_STREQ(s, "--")) { strcpy(attr, " sort=\"PreDEC\""); }
        if (SPD_STREQ(s, "!")) { strcpy(attr, " sort=\"LogicalNOT\""); }
        if (SPD_STREQ(s, "~")) { strcpy(attr, " sort=\"NOT\""); }
        if (SPD_STREQ(s, "+")) { strcpy(attr, " sort=\"Plus\""); }
        if (SPD_STREQ(s, "-")) { strcpy(attr, " sort=\"Minus\""); }
    } else if (p == JPD_O_BINARY) {
        if (SPD_STREQ(s, "instanceof")) { strcpy(attr, " sort=\"InstanceOf\""); }
        if (SPD_STREQ(s, "&")) { strcpy(attr, " sort=\"AND\""); }
        if (SPD_STREQ(s, "^")) { strcpy(attr, " sort=\"XOR\""); }
        if (SPD_STREQ(s, "|")) { strcpy(attr, " sort=\"OR\""); }
        if (SPD_STREQ(s, "*")) { strcpy(attr, " sort=\"MUL\""); }
        if (SPD_STREQ(s, "/")) { strcpy(attr, " sort=\"DIV\""); }
        if (SPD_STREQ(s, "%")) { strcpy(attr, " sort=\"MOD\""); }
        if (SPD_STREQ(s, "+")) { strcpy(attr, " sort=\"ADD\""); }
        if (SPD_STREQ(s, "-")) { strcpy(attr, " sort=\"SUB\""); }
        if (SPD_STREQ(s, "<<")) { strcpy(attr, " sort=\"SHL\""); }
        if (SPD_STREQ(s, ">>")) { strcpy(attr, " sort=\"SHR\""); }
        if (SPD_STREQ(s, ">>>")) { strcpy(attr, " sort=\"SHRZ\""); }
        if (SPD_STREQ(s, "<")) { strcpy(attr, " sort=\"CondLT\""); }
        if (SPD_STREQ(s, ">")) { strcpy(attr, " sort=\"CondGT\""); }
        if (SPD_STREQ(s, "<=")) { strcpy(attr, " sort=\"CondLE\""); }
        if (SPD_STREQ(s, "==")) { strcpy(attr, " sort=\"CondEQ\""); }
        if (SPD_STREQ(s, "!=")) { strcpy(attr, " sort=\"CondNE\""); }
        if (SPD_STREQ(s, ">=")) { strcpy(attr, " sort=\"CondGE\""); }
        if (SPD_STREQ(s, "&&")) { strcpy(attr, " sort=\"LogicalAND\""); }
        if (SPD_STREQ(s, "||")) { strcpy(attr, " sort=\"LogicalOR\""); }
    } else if (p == JPD_O_TERNARY) {
        strcpy(attr, " sort=\"Selection\"");
    }

    return (attr);
}

/******************************************************************************/

void setSavedJpdFile(SymData *file)
{
    savedJpdFile = file;
}

void extractFile(SymData *file)
{
    SpdString   attr;
    SymData     *tmp;

    if (!outputJ2jxEtb) {
        return;
    }

    initExtractFile();

    attr = makeIDAttr("id", getId(file));
    putElementSpec("File", 0, getLength(file), 10, attr);

    /* Package. */
    {
        SymData   *q = file->u.file->definedPackage;
        
        if (q != NULL) {
            int     off = searchOffsetBW(getOffset(q), 0, "package");
            int     off2 = searchOffsetFW(getOffset(q), fileLen, ";");
            int     len = off2 - off + 1;
  
            putElementSpec("Package", off, len, 10, makeIDAttrDerived("id"));
        }
    }

    /* Import. */
    for (tmp = file->u.file->importedPackage;
         tmp != NULL; tmp = tmp->u.qname->nextQName) {
        int         off = searchOffsetBW(getOffset(tmp), 0, "import");
        int         off2 = searchOffsetFW(getOffset(tmp), fileLen, ";");
        int         len = off2 - off + 1;
        SpdString   attr4 = makeIDAttrDerived("id");

        putElementSpec("Import", off, len, 10, attr4);
    }
}

void extractClass(SymData *jclass)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    int         offset, length;

    if (!outputJ2jxEtb) {
        return;
    }

    offset = getOffset(jclass);
    length = getLength(jclass);

    /* Class. */
    strcpy(attr, makeIDAttr("id", getId(jclass)));
    strcat(attr, makeModAttr(jclass));
    putElementSpec("Class", offset, length, 10, attr);

    /* ident. */
    if (0 < getAttrInt(jclass, "strOffset")) {
        putElementSpec("ident", getAttrInt(jclass, "strOffset"),
                       strlen(getAttrStr(jclass, "str")), 0,
                       makeIDAttr("defid", getId(jclass)));
    }

    /* Body. */
    {
        int off = searchOffsetFW(offset, offset + length, "{");
        int len = offset + length - off;

        putElementSpec("Members", off, len, 8, "");
    }

    /* Search superclass. */
    {
        SymData     *t = jclass->u.jclass->super;
        
        if (t != NULL) {
            Range   r = getTypeFamily(t);
            static char attr4[SPD_XML_ATTR_LEN + 1];
            
            strcpy(attr4, makeIDAttrDerived("id"));
            strcat(attr4, makeIDAttr("typefirst", getId(t)));
            putElementSpec("ExtdOpt", r.sp, r.ep - r.sp, 4, attr4);
        }
    }

    /* Search interfaces. */
    {
        SymData     *t;
        
        for (t = jclass->u.jclass->interface; t != NULL;
             t = t->u.type->nextType) {
            Range       r = getTypeFamily(t);
            static char attr4[SPD_XML_ATTR_LEN + 1];

            strcpy(attr4, makeIDAttrDerived("id"));
            strcat(attr4, makeIDAttr("typefirst", getId(t)));
            putElementSpec("ImplOpt", r.sp, r.ep - r.sp, 4, attr4);
        }
    }
}

void extractInterface(SymData *interface)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    int         offset, length;

    if (!outputJ2jxEtb) {
        return;
    }

    offset = getOffset(interface);
    length = getLength(interface);

    /* Intf. */
    strcpy(attr, makeIDAttr("id", getId(interface)));
    strcat(attr, makeModAttr(interface));
    putElementSpec("Intf", offset, length, 10, attr);

    /* ident. */
    if (0 < getAttrInt(interface, "strOffset")) {
        putElementSpec("ident", getAttrInt(interface, "strOffset"),
                       strlen(getAttrStr(interface, "str")), 0, 
                       makeIDAttr("defid", getId(interface)));
    }

    /* Body. */
    {
        int off = searchOffsetFW(offset, offset  + length, "{");
        int len = offset + length - off;
        
        putElementSpec("Members", off, len, 8, "");
    }

    /* Search interfaces. */
    {
        SymData     *t;

        for (t = interface->u.interface->super; t != NULL;
             t = t->u.type->nextType) {
            Range       r = getTypeFamily(t);
            static char attr4[SPD_XML_ATTR_LEN + 1];

            strcpy(attr4, makeIDAttrDerived("id"));
            strcat(attr4, makeIDAttr("typefirst", getId(t)));
            putElementSpec("ImplOpt", r.sp, r.ep - r.sp, 4, attr4);
        }
    }
}

void extractRoutine(SymData *routine)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    int         s;
    SpdString   elem;
    int         offset, length;

    if (!outputJ2jxEtb) {
        return;
    }

    s = getAttrInt(routine, "sort");
    offset = getOffset(routine);
    length = getLength(routine);

    /* Method and Ctor. */
    if (s == METHOD) {
        elem = "Method";
    } else if (s == CONSTRUCTOR) {
        elem = "Ctor";
    } else {
        spdErrorExit("extractRoutine(): %d: Invalid sort.\n");
        return; /* Error. */
    }

    strcpy(attr, makeIDAttr("id", getId(routine)));
    if (s == METHOD) {
        strcat(attr, makeIDAttr("typefirst",
                                getId(routine->u.routine->returnType)));
    }
    strcat(attr, makeModAttr(routine));
    putElementSpec(elem, offset, length, 10, attr);

    /* ident. */
    if (0 < getAttrInt(routine, "strOffset") && s == METHOD) {
        SpdString   attr8 = makeIDAttr("defid", getId(routine));

        putElementSpec("ident", getAttrInt(routine, "strOffset"),
                       strlen(getAttrStr(routine, "str")), 0, attr8);
    }
    if (s == CONSTRUCTOR) {
        SpdString   attr8 = makeIDAttr("defid", getId(routine));
        SpdString   tname = getAttrStr(routine, "str");
        int         pos = searchOffsetFW(getAttrInt(routine, "strOffset"),
                                         offset + length, tname);

        putElementSpec("ident", pos, strlen(tname), 0, attr8);
    }

    /* Search throws. */
    {
        SymData     *t;
        for (t = routine->u.routine->throwType; t != NULL;
             t = t->u.type->nextType) {
          Range       r = getTypeFamily(t);
          static char attr4[SPD_XML_ATTR_LEN + 1];

          strcpy(attr4, makeIDAttrDerived("id"));
          strcat(attr4, makeIDAttr("typefirst", getId(t)));
          putElementSpec("ThrwOpt", r.sp, r.ep - r.sp, 4, attr4);
        }
    }
}

void extractVariable(SymData *var)
{
    int         s;
    SpdString   elem;
    int         offset, length;
    SymData     *tmp;
    static char attr[SPD_XML_ATTR_LEN + 1];

    if (!outputJ2jxEtb) {
        return;
    }

    s = getAttrInt(var, "sort");
    offset = getOffset(var);
    length = getLength(var);

    /* Field/LocalVar/Param. */
    elem = NULL;
    if (s == CLASS_MEM) {
        elem = "Field";
    } else if (s == INTERFACE_MEM) {
        elem = "Field";
    } else if (s == LOCAL_VAR) {
        elem = "Local";
    } else if (s == PARAMETER) {
        elem = "Param";
    } else {
        spdErrorExit("extractVariable(): %d: Invalid sort.\n");
        return;
    }

    strcpy(attr, makeIDAttr("id", getId(var)));;
    strcat(attr, makeIDAttr("typefirst", getId(var->u.variable->type)));

    {
        /* Search decl range. */
        Range       r = mkRange(offset, offset + length);
        SymData     *t = var->u.variable->type;
        Range       r2 = getTypeFamily(t);
        
        if (r2.sp < r.sp) r.sp = r2.sp;
        if (r.ep < r2.ep) r.ep = r2.ep;

        for (tmp = var->u.variable->modifier;
             tmp != NULL; tmp = tmp->u.modifier->next) {
            if (getOffset(tmp) < r.sp) {
                r.sp = getOffset(tmp);
            }
        }

        tmp = var->u.variable->initializer;
        if (tmp != NULL) {
            r.ep = getOffset(tmp) + getLength(tmp);
        }
    
        /* Chain check. */
        {
            int comma = searchOffsetBW(getOffset(var), r.sp, ",");
            if (r.sp < comma) {
                r.sp = getOffset(var);
                strcat(attr, " omitted=\"yes\"");
            }
        }
        
        strcat(attr, makeModAttr(var));
        putElementSpec(elem, r.sp, r.ep - r.sp, 10, attr);

        /* Ident. */
        putElementSpec("ident",
                       getOffset(var), getLength(var), 0,
                       makeIDAttr("defid", getId(var)));
    }
}

void extractType(SymData *type)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    int         s;

    if (!outputJ2jxEtb) {
        return;
    }

    strcpy(attr, makeIDAttr("id", getId(type)));
    s = getAttrInt(type, "sort");

    /* Type. */
    if (s == ARRAY_TYPE) {
        strcat(attr, " sort=\"Array\"");
    } if (s == PRIMITIVE_TYPE) {
        strcat(attr, " sort=\"Primitive\"");
    } if (s == OBJECT_TYPE) {
        SymData     *q = type->u.type->qname;

        strcat(attr, " sort=\"Object\"");
        if (q != NULL && hasValidID(q)) {
            strcat(attr, makeIDAttr("qname", getId(q)));
        }
    }

    if (s == ARRAY_TYPE) {
        SymData     *d = type->u.type->next;

        if (d != NULL && hasValidID(d)) {
            strcat(attr, makeIDAttr("typenext", getId(d)));
        }
    }

    putElementSpec("Type", getOffset(type), getLength(type), 2, attr);

    /* !! ident. */
    if (s == OBJECT_TYPE) {
        static char attr8[SPD_XML_ATTR_LEN + 1];
#ifdef RESOLVE_OBJECT_TYPE
        SymData     *d;
#endif
        
        attr8[0] = '\0';

#ifdef RESOLVE_OBJECT_TYPE
        /* !! FrjClassInterface   d = (FrjClassInterface)
                                      c.getReferringClassInterface(); */
        
        d = type->u.type->jclass;
        if (d != NULL && hasValidID(d)) {
            strcpy(attr8, makeIDAttr("defid", getId(d)));
        }
#endif

        RESOLVE_OBJECT_TYPE_DW(attr8, " dw=\"OBJECT TYPE\"");
        putElementSpec("ident", getOffset(type), getLength(type), 0, attr8);
    }
}

void extractSBlock(SymData *sblock)
{
    static char attr[SPD_XML_ATTR_LEN + 1];

    if (!outputJ2jxEtb) {
        return;
    }
    
    strcpy(attr, makeIDAttr("id", getId(sblock)));
    putElementSpec("SInit", getOffset(sblock), getLength(sblock), 10, attr);
}

void extractStatement(SymData *stmt)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    int         s;

    if (!outputJ2jxEtb) {
        return;
    }

    strcpy(attr, makeIDAttr("id", getId(stmt)));
    s = getAttrInt(stmt, "sort");

    if (s == S_EMPTY) { strcat(attr, " sort=\"EMPTY\""); }
    else if (s == S_EXPRESSION) { strcat(attr, " sort=\"EXPR\""); }
    else if (s == S_BLOCK) { strcat(attr, " sort=\"BLOCK\""); }
    else if (s == S_DO) { strcat(attr, " sort=\"DOWHILE\""); }
    else if (s == S_WHILE) { strcat(attr, " sort=\"WHILE\""); }
    else if (s == S_FOR) { strcat(attr, " sort=\"FOR\""); }
    else if (s == S_IF) { strcat(attr, " sort=\"IFELSE\""); }
    else if (s == S_SWITCH) { strcat(attr, " sort=\"SWITCH\""); }
    else if (s == S_BREAK) { strcat(attr, " sort=\"BREAK\""); }
    else if (s == S_CONTINUE) { strcat(attr, " sort=\"CONT\""); }
    else if (s == S_RETURN) { strcat(attr, " sort=\"RETURN\""); }
    else if (s == S_THROW) { strcat(attr, " sort=\"THROW\""); }
    else if (s == S_SYNCHRONIZED) { strcat(attr, " sort=\"SYNCHRO\""); }
    else if (s == S_EXCEPTION) { strcat(attr, " sort=\"TRY\""); }
    else if (s == S_ASSERT) { strcat(attr, " sort=\"ASSERT\""); }
    else {
        /* !! S_DEF_VAR򰷤äƤʤΤǥ顼ȹͤʤǤ? */
        /* spdErrorExit("extractStatement(): %d.\n", s); */
        return; /* Error. */
    }

    putElementSpec("Stmt", getOffset(stmt), getLength(stmt), 8, attr);
}

void extractExpression(SymData *expr)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    static char attr2[SPD_XML_ATTR_LEN + 1];
    int         s;
    int         offset, length;

    if (!outputJ2jxEtb) {
        return;
    }

    strcpy(attr, makeIDAttr("id", getId(expr)));
    /* strcpy(attr2, ""); */
    attr2[0] = '\0';
    s = getAttrInt(expr, "sort");
    offset = getOffset(expr);
    length = getLength(expr);

    if (s == E_EMPTY) { strcpy(attr2, " sort=\"Empty\""); }
    else if (s == E_REF_LITERAL) { strcpy(attr2, " sort=\"Literal\""); }
    else if (s == E_REF_TYPE) { strcpy(attr2, " sort=\"TypeRef\""); }
    else if (s == E_REF_VAR) { strcpy(attr2, " sort=\"VarRef\""); }
    else if (s == E_CONSTRUCTOR_CALL) { strcpy(attr2, " sort=\"CtorCall\""); }
    else if (s == E_METHOD_CALL) { strcpy(attr2, " sort=\"MethodCall\""); }
    else if (s == E_LISTED) { strcpy(attr2, " sort=\"Listed\""); }
    else if (s == E_OPERATION) { strcat(attr2, makeOpAttr(expr)); }
    else if (s == E_REF_SPECIAL) {
        if (SPD_STREQ_N(&(fileBuf[offset]), "super", 5)) {
            /* Skip "super". */
            return;
        } else if (SPD_STREQ_N(&(fileBuf[offset]), "class", 5)) {
            /* Added by S.Yamamoto 2003/02/26. */
            strcpy(attr2, " sort=\"Class\"");
        } else if (SPD_STREQ_N(&(fileBuf[offset]), "this", 4)) {
            strcpy(attr2, " sort=\"This\"");
        } else {
            if (SPD_STREQ_N(&(fileBuf[offset]), "null", 4)) {
                putElementSpec("literal", offset, length, 0, "sort=\"NULL\"");
            } else if (SPD_STREQ_N(&(fileBuf[offset]), "true", 4) ||
                       SPD_STREQ_N(&(fileBuf[offset]), "false", 5)) {
                putElementSpec("literal", offset, length, 0, "sort=\"BOOL\"");
            } else {
                spdMessage("Notice: id %d, offset %d: Invalid REF_SPECIAL.\n",
                           getId(expr), offset);
            }
            strcpy(attr2, " sort=\"Literal\"");
        }
    }

    if (SPD_STREQ(attr2, "")) {
        return;     /* If REF_QNAME and REF_SPECIAL (super), not Expr. */
    }

    /* !! Cast type. */
    if (s == E_REF_TYPE) {

#ifdef CAST_TYPEFIRST
        SymData     *t = expr->u.expression->first;

        if (t != NULL) {
            strcat(attr, makeIDAttr("typefirst", getId(t)));
        }
#endif

        CAST_TYPEFIRST_DW(attr, " dw=\"CAST TYPEFIRST\"");
    }

    /* Variable. */
    if (s == E_REF_VAR) {
        /* Create ident node. */
        static char attr8[SPD_XML_ATTR_LEN + 1];
        SymData     *d = expr->u.expression->first;
        int         k = expr->u.expression->lrsort;
    
        attr8[0] = '\0';
        if (d != NULL && hasValidID(d)) {
            strcpy(attr8, makeIDAttr("defid", getId(d)));
        }
    
        if (k == R_VALUE || k == LR_VALUE) {
            strcat(attr2, " read=\"yes\"");
        }
        if (k == L_VALUE || k == LR_VALUE) {
            strcat(attr2, " write=\"yes\"");
        }
    
        putElementSpec("ident", offset, length, 0, attr8);
    }

    /* !! Routine. */
    if (s == E_METHOD_CALL) {
        /* Create ident node. */
        static char attr8[SPD_XML_ATTR_LEN + 1];
        SymData     *d = expr->u.expression->first;

        attr8[0] = '\0';

        /* 
         * spdMessage("extractExpression(): \"%20s\".\n",
         *            getAttrStr(expr, "str"));
         */
        assert(d == NULL);
        if (d != NULL && hasValidID(d)) {
            strcpy(attr8, makeIDAttr("defid", getId(d)));
        }

#ifdef RESOLVE_METHOD
        {
            SymData     *d = resolveMethod(expr);

            /* 
             * spdMessage("        : %d, %x.\n",
             *            getAttrInt(expr, "strOffset"), d);
             * if (d != NULL) { spdMessage("            : %d.\n", getId(d)); }
             */
            if (d != NULL && hasValidID(d)) {
                strcpy(attr8, makeIDAttr("defid", getId(d)));
            }
        }
#endif

        RESOLVE_METHOD_DW(attr8, " dw=\"METHOD CALL\"");
        putElementSpec("ident", getAttrInt(expr, "strOffset"),
                       strlen(getAttrStr(expr, "str")), 0, attr8);
    }

    strcat(attr, attr2);
    putElementSpec("Expr", getOffset(expr),
                   getLength(expr), (s == E_LISTED) ? 5 : 4, attr);
}

void extractLiteral(SymData *lit)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    int         s;

    if (!outputJ2jxEtb) {
        return;
    }

    strcpy(attr, makeIDAttr("id", getId(lit)));
    s = getAttrInt(lit, "sort");

    if (s == L_CHAR) { strcat(attr, " sort=\"CHAR\""); }
    else if (s == L_STRING) { strcat(attr, " sort=\"STR\""); }
    else if (s == L_UNICODE) { strcat(attr, " sort=\"CHAR\""); }
    else if (s == L_INT) { strcat(attr, " sort=\"INT\""); }
    else if (s == L_DOUBLE) { strcat(attr, " sort=\"FLOAT\""); }
    else { spdErrorExit("%d: Invalid literal sort.\n", s); }

    putElementSpec("literal", getOffset(lit), getLength(lit), 0, attr);
}

void extractQName(SymData *qname)
{
    static char attr[SPD_XML_ATTR_LEN + 1];

    if (!outputJ2jxEtb) {
        return;
    }

    strcpy(attr, makeIDAttr("id", getId(qname)));
    putElementSpec("QName", getOffset(qname), getLength(qname), 2, attr);
  
    {
        SymData     *d = qname->u.qname->dot;
        char        ch = fileBuf[getOffset(qname)];
      
        if (ch != '*' && !(d != NULL && hasValidID(d)) ) {
            putElementSpec("ident", getOffset(qname), getLength(qname), 0, "");
        }
    }
}

void extractLabel(SymData *label)
{
    static char attr[SPD_XML_ATTR_LEN + 1];
    static char attr2[SPD_XML_ATTR_LEN + 1];
    int         s;
    int         sp;

    if (!outputJ2jxEtb) {
        return;
    }

    s = getAttrInt(label, "sort");
    sp = getOffset(label);

    if (s == REF_LABEL) {
        strcpy(attr2, makeIDAttr("defid", getId(label->u.label->defined)));
        putElementSpec("ident", sp, getLength(label), 0, attr2);
    } else {
        int     ep = searchOffsetFW(getOffset(label), fileLen, ":");

        strcpy(attr, makeIDAttr("id", getId(label)));
        if (s == DEF_LABEL) {
            strcpy(attr2, makeIDAttr("defid", getId(label)));
            putElementSpec("ident", sp, getLength(label) - 1, 0, attr2);
        }
        putElementSpec("Label", sp, ep - sp + 1, 8, attr);
    }
}

void extractComment(SymData *com)
{
    if (!outputJ2jxEtb) {
        return;
    }
  
    putElementSpec("comment", getOffset(com), getLength(com), 0,
                   makeIDAttr("id", getId(com)));
}
