%{

/* 
 * Program:             $RCSfile: java-lex.l,v $  $Revision: 40.49 $
 * 
 * Purpose:
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: java-lex.l,v 40.49 2003/05/24 15:54:13 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "JB.h"
#include "y.tab.h"
#include "FuncProto.h"

/*
 * Important note: all actions should end with "count_char();" or
 * "return (TOKEN(XXX));" in order to keep value of variable "c_count" valid.
 */

/*
 * Debug flags.
 * TOKEN_DEBUG: For lex debug.
 */

/* #define TOKEN_DEBUG */

#ifdef TOKEN_DEBUG
#define TOKEN(x)        ((yydebug ? \
                          spdMessage("yylex(): l.%d, c.%d: \"%s\" -> %s.\n", \
                                 yylineno, c_count,  yytext, #x) : 0), \
                         count_char(), (x))
#else
#define TOKEN(x)        (count_char(), (x))
#endif

/*
 * Redefine YYLMAX.
 * Note: yytext is defined as yytext[YYLMAX].
 */

#undef YYLMAX
#define YYLMAX              (SPD_LINE_LEN + 1)

/*
 * For compatibility lex and flex.
 * Note: in flex the undocumented lex scanner internal variable yylineno is
 * not supported unless -l or "%option yylineno" is used. And we assume flex
 * is invoked without -l flag. Variable yylineno is not part of
 * the POSIX specification.
 */

#ifdef USE_FLEX
#define HAVE_NOT_YYLINENO
#endif

#ifdef HAVE_NOT_YYLINENO
int                 yylineno = 1;   /* Line number of ".java" file. */
#endif

#ifdef HAVE_NOT_YYLINENO
#define INC_YYLINENO()      (yylineno++)
#else 
#define INC_YYLINENO()      /* Empty. */
#endif

#define MAKE_WORD_NAME(x)   spdStrPoolFastCopy(str_pool, (x))

#if !defined(yywrap)
#define yywrap()            1       /* Treat EOF as an EOF. */
#endif

#define SET_YYLVAL_KEYWORD()    (set_yylval(), \
                                 output_etb(c_count, yyleng, 0, "kw", ""))
#define SET_YYLVAL_OPERATOR()   (set_yylval(), \
                                 output_etb(c_count, yyleng, 0, "op", ""))

extern YYSTYPE          yylval;
extern SpdStringPool    *str_pool;

int         yydebug = 0;
int         yydebug_line;

int         c_count = 0;            /* Offset in file. */

static char str_buf[SPD_STRING_LEN + 1];

static void set_yylval(void);
static void append_yytext(void);
static void count_char(void);
static void output_etb(int offset, int length, int weight, 
                       SpdConstString name, SpdConstString attr);
static void output_newline_tag(void);

%}

%start  COMM COMM_CXX STR

%%

<INITIAL>[ \t]+ {
    output_etb(c_count, yyleng, 0, "sp", "");
    count_char();
}
<INITIAL>"\r\n"|"\n"|"\r" {
    output_newline_tag();
    count_char();
    INC_YYLINENO();
    if (yydebug_line && yydebug_line <= yylineno) {
        yydebug = 1;
        yydebug_line = 0;
    }
}
<INITIAL>"\014"|"\032" {
    /* Skip '^L' (0xc) and '^Z' (0x1a). */
    output_etb(c_count, yyleng, 0, "sp", "");
    count_char();
}

<INITIAL>"//" {
    set_yylval();
    yylval->length = 0;
    spdFree(yylval->str);
    append_yytext();
    yylval->symSort = SPD_CID_J_COMMENT;
    yylval->sort2 = C_LINE;
    BEGIN(COMM_CXX);
    count_char();
}
<COMM_CXX>"\r\n"|"\n"|"\r" {
    yylval->str = MAKE_WORD_NAME(str_buf);
    comment = yylval;
    BEGIN(INITIAL);
    output_newline_tag();
    count_char();
    INC_YYLINENO();
}
<COMM_CXX>. {
    append_yytext();
    count_char();
}

<INITIAL>"/*" {
    set_yylval();
    yylval->length = 0;
    spdFree(yylval->str);
    append_yytext();
    yylval->symSort = SPD_CID_J_COMMENT;
    yylval->sort2 = C_INLINE;
    BEGIN(COMM);
    count_char();
}
<COMM>"*/" {
    append_yytext();
    yylval->str = MAKE_WORD_NAME(str_buf);
    comment = yylval;
    BEGIN(INITIAL);
    count_char();
}
<COMM>"\r\n"|"\n"|"\r" {
    /* !! output_newline_tag(); */
    append_yytext();
    yylval->sort2 = C_BLOCK;
    count_char();
    INC_YYLINENO();
}
<COMM>. {
    append_yytext();
    count_char();
}

<INITIAL>break          {SET_YYLVAL_KEYWORD(); return (TOKEN(BREAK));}
<INITIAL>class          {SET_YYLVAL_KEYWORD(); return (TOKEN(CLASS));}
<INITIAL>catch          {SET_YYLVAL_KEYWORD(); return (TOKEN(CATCH));}
<INITIAL>case           {SET_YYLVAL_KEYWORD(); return (TOKEN(CASE));}
<INITIAL>continue       {SET_YYLVAL_KEYWORD(); return (TOKEN(CONTINUE));}
<INITIAL>default        {SET_YYLVAL_KEYWORD(); return (TOKEN(DEFAULT));}
<INITIAL>do             {SET_YYLVAL_KEYWORD(); return (TOKEN(DO));}
<INITIAL>else           {SET_YYLVAL_KEYWORD(); return (TOKEN(ELSE));}
<INITIAL>extends        {SET_YYLVAL_KEYWORD(); return (TOKEN(EXTENDS));}
<INITIAL>finally        {SET_YYLVAL_KEYWORD(); return (TOKEN(FINALLY));}
<INITIAL>for            {SET_YYLVAL_KEYWORD(); return (TOKEN(FOR));}
<INITIAL>if             {SET_YYLVAL_KEYWORD(); return (TOKEN(IF));}
<INITIAL>implements     {SET_YYLVAL_KEYWORD(); return (TOKEN(IMPLEMENTS));}
<INITIAL>import         {SET_YYLVAL_KEYWORD(); return (TOKEN(IMPORT));}
<INITIAL>instanceof     {SET_YYLVAL_KEYWORD(); return (TOKEN(INSTANCEOF));}
<INITIAL>interface      {SET_YYLVAL_KEYWORD(); return (TOKEN(INTERFACE));}
<INITIAL>new            {SET_YYLVAL_KEYWORD(); return (TOKEN(NEW));}
<INITIAL>package        {SET_YYLVAL_KEYWORD(); return (TOKEN(PACKAGE));}
<INITIAL>return         {SET_YYLVAL_KEYWORD(); return (TOKEN(RETURN));}
<INITIAL>switch         {SET_YYLVAL_KEYWORD(); return (TOKEN(SWITCH));}
<INITIAL>throw          {SET_YYLVAL_KEYWORD(); return (TOKEN(THROW));}
<INITIAL>throws         {SET_YYLVAL_KEYWORD(); return (TOKEN(THROWS));}
<INITIAL>try            {SET_YYLVAL_KEYWORD(); return (TOKEN(TRY));}
<INITIAL>while          {SET_YYLVAL_KEYWORD(); return (TOKEN(WHILE));}

<INITIAL>abstract       {SET_YYLVAL_KEYWORD(); return (TOKEN(ABSTRACT));}
<INITIAL>final          {SET_YYLVAL_KEYWORD(); return (TOKEN(FINAL));}
<INITIAL>public         {SET_YYLVAL_KEYWORD(); return (TOKEN(PUBLIC));}
<INITIAL>protected      {SET_YYLVAL_KEYWORD(); return (TOKEN(PROTECTED));}
<INITIAL>private        {SET_YYLVAL_KEYWORD(); return (TOKEN(PRIVATE));}
<INITIAL>static         {SET_YYLVAL_KEYWORD(); return (TOKEN(STATIC));}
<INITIAL>transient      {SET_YYLVAL_KEYWORD(); return (TOKEN(TRANSIENT));}
<INITIAL>volatile       {SET_YYLVAL_KEYWORD(); return (TOKEN(VOLATILE));}
<INITIAL>native         {SET_YYLVAL_KEYWORD(); return (TOKEN(NATIVE));}
<INITIAL>synchronized   {SET_YYLVAL_KEYWORD(); return (TOKEN(SYNCHRONIZED));}
<INITIAL>strictfp       {SET_YYLVAL_KEYWORD(); return (TOKEN(STRICTFP));}
<INITIAL>super          {SET_YYLVAL_KEYWORD(); return (TOKEN(SUPER));}
<INITIAL>this           {SET_YYLVAL_KEYWORD(); return (TOKEN(THIS));}

<INITIAL>null           {set_yylval(); return (TOKEN(JNULL));}
<INITIAL>true           {set_yylval(); return (TOKEN(JTRUE));}
<INITIAL>false          {set_yylval(); return (TOKEN(JFALSE));}

<INITIAL>assert {
    if (withAssert) {
        /* JDK 1.4 assert. */
        SET_YYLVAL_KEYWORD();
        return (TOKEN(ASSERT));
    } else {
        set_yylval();
        TOKEN(IDENTIFIER);
    }
}

<INITIAL>void|byte|char|int|short|long|float|double|boolean {
    SET_YYLVAL_KEYWORD();
    return (TOKEN(PRIMITIVE_TYPE_SPECIFIER));
}

<INITIAL>\'.\'|\'\\.\'|\'\\[0-7]+\'|\'\\x[0-9a-zA-Z]+\' {
    /* Character. */
    set_yylval();
    yylval->symSort = SPD_CID_J_LITERAL;
    yylval->u.literal = spdMalloc(sizeof(JpdLiteral));
    yylval->u.literal->sort = L_CHAR;
    yylval->u.literal->value.c = *yytext;
    return (TOKEN(CHARACTER_LITERAL));
}

<INITIAL>\'\\u[0-9a-fA-F]{4}\' {
    /* Unicode character. */
    set_yylval();
    yylval->symSort = SPD_CID_J_LITERAL;
    yylval->u.literal = spdMalloc(sizeof(JpdLiteral));
    yylval->u.literal->sort = L_UNICODE;
    yylval->u.literal->value.c = *yytext;
    return (TOKEN(UNICODE_LITERAL));
}

<INITIAL>\" {
    set_yylval();       /* Set offset of yylval. */
    yylval->length = 0;
    spdFree(yylval->str);
    append_yytext();
    yylval->symSort = SPD_CID_J_LITERAL;
    yylval->u.literal = spdMalloc(sizeof(JpdLiteral));
    yylval->u.literal->sort = L_STRING;
    BEGIN(STR);
    count_char();
}
<STR>\" {
    append_yytext();
    yylval->str = yylval->u.literal->value.s = MAKE_WORD_NAME(str_buf);
    BEGIN(INITIAL);
    return (TOKEN(STRING_LITERAL));
}
<STR>"\\\r\n"|"\\\n"|"\\\r"|"\r\n"|"\n"|"\r" {
    spdErrorExit("yylex(): l.%d, c.%d: Invalid newline in string literal.\n",
                 yylineno, c_count);
}
<STR>\\.|. {
    append_yytext();
    count_char();
}

<INITIAL>([0-9]+|("0x"|"0X")[0-9a-fA-F]+)[ulUL]* {
    set_yylval();
    yylval->symSort = SPD_CID_J_LITERAL;
    yylval->u.literal = spdMalloc(sizeof(JpdLiteral));
    yylval->u.literal->sort = L_INT;
    yylval->u.literal->value.i = atoi(yytext);
    return (TOKEN(INTEGER_LITERAL));
}

<INITIAL>(([0-9]+"."|[0-9]+"."?([eE][-+]?[0-9]+)|[0-9]*"."[0-9]+([eE][-+]?[0-9]+)?)[fdFD]*|[0-9]+[fdFD]) {
    set_yylval();
    yylval->symSort = SPD_CID_J_LITERAL;
    yylval->u.literal = spdMalloc(sizeof(JpdLiteral));
    yylval->u.literal->sort = L_DOUBLE;
    yylval->u.literal->value.f = atof(yytext);
    return (TOKEN(FLOATING_LITERAL));
}

<INITIAL>[a-zA-Z_$][0-9a-zA-Z_$]* {
    set_yylval();
    return (TOKEN(IDENTIFIER));
}

<INITIAL>"&"    {SET_YYLVAL_OPERATOR(); return (TOKEN(AND));}
<INITIAL>"&&"   {SET_YYLVAL_OPERATOR(); return (TOKEN(AND_AND));}
<INITIAL>"&="   {SET_YYLVAL_OPERATOR(); return (TOKEN(AND_EQUAL));}
<INITIAL>"="    {SET_YYLVAL_OPERATOR(); return (TOKEN(EQUAL));}
<INITIAL>"=="   {SET_YYLVAL_OPERATOR(); return (TOKEN(EQUAL_EQUAL));}
<INITIAL>"!"    {SET_YYLVAL_OPERATOR(); return (TOKEN(EXCLAMATION));}
<INITIAL>"!="   {SET_YYLVAL_OPERATOR(); return (TOKEN(EXCLAMATION_EQUAL));}
<INITIAL>"^"    {SET_YYLVAL_OPERATOR(); return (TOKEN(HAT));}
<INITIAL>"^="   {SET_YYLVAL_OPERATOR(); return (TOKEN(HAT_EQUAL));}
<INITIAL>"<"    {SET_YYLVAL_OPERATOR(); return (TOKEN(LESS));}
<INITIAL>"<="   {SET_YYLVAL_OPERATOR(); return (TOKEN(LESS_EQUAL));}
<INITIAL>"<<"   {SET_YYLVAL_OPERATOR(); return (TOKEN(LESS_LESS));}
<INITIAL>"<<="  {SET_YYLVAL_OPERATOR(); return (TOKEN(LESS_LESS_EQUAL));}
<INITIAL>"-"    {SET_YYLVAL_OPERATOR(); return (TOKEN(MINUS));}
<INITIAL>"-="   {SET_YYLVAL_OPERATOR(); return (TOKEN(MINUS_EQUAL));}
<INITIAL>"--"   {SET_YYLVAL_OPERATOR(); return (TOKEN(MINUS_MINUS));}
<INITIAL>">"    {SET_YYLVAL_OPERATOR(); return (TOKEN(MORE));}
<INITIAL>">="   {SET_YYLVAL_OPERATOR(); return (TOKEN(MORE_EQUAL));}
<INITIAL>">>"   {SET_YYLVAL_OPERATOR(); return (TOKEN(MORE_MORE));}
<INITIAL>">>="  {SET_YYLVAL_OPERATOR(); return (TOKEN(MORE_MORE_EQUAL));}
<INITIAL>">>>"  {SET_YYLVAL_OPERATOR(); return (TOKEN(MORE_MORE_MORE));}
<INITIAL>">>>=" {SET_YYLVAL_OPERATOR(); return (TOKEN(MORE_MORE_MORE_EQUAL));}
<INITIAL>"|"    {SET_YYLVAL_OPERATOR(); return (TOKEN(OR));}
<INITIAL>"|="   {SET_YYLVAL_OPERATOR(); return (TOKEN(OR_EQUAL));}
<INITIAL>"||"   {SET_YYLVAL_OPERATOR(); return (TOKEN(OR_OR));}
<INITIAL>"%"    {SET_YYLVAL_OPERATOR(); return (TOKEN(PERCENT));}
<INITIAL>"%="   {SET_YYLVAL_OPERATOR(); return (TOKEN(PERCENT_EQUAL));}
<INITIAL>"+"    {SET_YYLVAL_OPERATOR(); return (TOKEN(PLUS));}
<INITIAL>"+="   {SET_YYLVAL_OPERATOR(); return (TOKEN(PLUS_EQUAL));}
<INITIAL>"++"   {SET_YYLVAL_OPERATOR(); return (TOKEN(PLUS_PLUS));}
<INITIAL>"/"    {SET_YYLVAL_OPERATOR(); return (TOKEN(SLASH));}
<INITIAL>"/="   {SET_YYLVAL_OPERATOR(); return (TOKEN(SLASH_EQUAL));}
<INITIAL>"*"    {SET_YYLVAL_OPERATOR(); return (TOKEN(STAR));}
<INITIAL>"*="   {SET_YYLVAL_OPERATOR(); return (TOKEN(STAR_EQUAL));}
<INITIAL>"~"    {SET_YYLVAL_OPERATOR(); return (TOKEN(TILDA));}
<INITIAL>","    {SET_YYLVAL_OPERATOR(); return (TOKEN(COMMA));}
<INITIAL>"."    {SET_YYLVAL_OPERATOR(); return (TOKEN(DOT));}
<INITIAL>"?"    {SET_YYLVAL_OPERATOR(); return (TOKEN(QUESTION));}

<INITIAL>[\[][ \t]*[\]]     {SET_YYLVAL_OPERATOR(); return (TOKEN(OP_DIM));}

<INITIAL>"("    {SET_YYLVAL_OPERATOR(); return (TOKEN('('));}
<INITIAL>")"    {SET_YYLVAL_OPERATOR(); return (TOKEN(')'));}
<INITIAL>"{"    {SET_YYLVAL_OPERATOR(); return (TOKEN('{'));}
<INITIAL>"}"    {SET_YYLVAL_OPERATOR(); return (TOKEN('}'));}
<INITIAL>"["    {SET_YYLVAL_OPERATOR(); return (TOKEN('['));}
<INITIAL>"]"    {SET_YYLVAL_OPERATOR(); return (TOKEN(']'));}
<INITIAL>";"    {SET_YYLVAL_OPERATOR(); return (TOKEN(';'));}
<INITIAL>":"    {SET_YYLVAL_OPERATOR(); return (TOKEN(COLON));}

. {
    spdWarning("yylex(): l.%d, c.%d: Invalid character '%c' (0x%x) skipped.\n",
               yylineno, c_count, yytext[0], yytext[0]);
    count_char();
}

%%

static void set_yylval(void)
{
    yylval = getSymData();
    yylval->length = yyleng;
    yylval->offset = c_count;
    yylval->strOffset = c_count;
    yylval->str = spdStrdup(yytext);
    if (comment != NULL) {
        yylval->comment = comment;
        comment = NULL;
    }
}

static void append_yytext(void)
{
    if (SPD_STRING_LEN < (yylval->length + yyleng)) {
        spdErrorExit("append_yytext(): \"%s\": Too long constant string.\n",
                     yytext);
    }
    strcpy(str_buf + yylval->length, yytext);
    yylval->length += yyleng;
}

static void count_char(void)
{
    c_count += yyleng;
}

static void output_etb(int offset, int length, int weight, 
                       SpdConstString name, SpdConstString attr)
{
    fprintf(jdbinEtb, "%d\t%d\t%d\t%s %s\n",
            offset, length, weight, name, attr);
}

static void output_newline_tag(void)
{
    static char attr[32];

    sprintf(attr, "line=\"%d\" offset=\"%d\"", yylineno, c_count);
    output_etb(c_count, yyleng, 0, "nl", attr);
}
