%{

/* 
 * Program:             $RCSfile: java-yacc.y,v $  $Revision: 40.100 $
 * 
 * Purpose:
 *
 * Grammar author:      D.Bronnikov  1996
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       D.Bronnikov, Y.Hachisu and S.Yamamoto  1996 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: java-yacc.y,v 40.100 2003/11/24 04:53:39 yamamoto Exp yamamoto $";
#endif

/* 
 * Copyright (C) 1996 Dmitri Bronnikov, All rights reserved.
 *
 * THIS GRAMMAR IS PROVIDED "AS IS" WITHOUT  ANY  EXPRESS  OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES  OF  MERCHANTABILITY  AND  FITNESS  FOR  A  PARTICULAR
 * PURPOSE, OR NON-INFRINGMENT.
 *
 * This grammar can be used by anyone without any restrictions. 
 *
 * Dmitri Bronnikov
 * 21485 Green Hill Rd
 * Farmington Hills, MI, 48335 
 * Bronnikov@aol.com
 *
 * -----------------------------------------------------------------
 *
 * This grammar is the first (dirty) version that managed to properly
 * parse all the Java demos available to me.
 *
 * It needs a lot of brushing, reordering, and may be buggy too.
 * Please let me know about any bugs/ambiguities found. 
 *
 * There are some tabs, nice picture should appear if an editor
 * is set to use tabsize = 8.
 *
 * -----------------------------------------------------------------
 *
 * Certain modifications have been applied to resolve conflicts, as
 * per the following list:
 *
 * All modifiers were made syntactically allowable to resolve
 * reduce/reduce conflicts.
 *
 * All expressions have been allowed in expression statement (this
 * modification is a parsing style change rather than conflict
 * resolution)
 *
 * All expressions allowed in type expression specifying cast type
 * in the cast expression.
 *
 * New token OP_DIM introduced for [] to allow the parser to reduce
 * primary expression when [ is encountered after a qualified name.
 * LEX definition for this token is [\[]{white_space}*[\]].
 *
 * -----------------------------------------------------------------
 *
 * Conflicts
 *
 * This grammar is translated by YACC with four 'good' shift/reduce
 * conflicts.
 *
 * The first one of the is ubiquitous if-then-else.
 *
 * The other three deal with '.' operator, which in Java may be
 * used both in qualified type names in declarations and as field
 * access operator in expressions.
 *
 * Please notice, that (unlike C/C++) lexical analizer should not
 * differentiate type name from variable name. One of the reasons 
 * is performance - some packages/classes may be remote. The other 
 * one is that Java existing compilers allow for sinonimous type
 * and variable name to be used in the same scope, depending on
 * context.
 */

#include <stdio.h>
#include <signal.h>
#include <Sapid/Sapid.h>
#include "JB.h"
#include "FuncProto.h"

#define MILL_OPTION_STRINGS             "abd:hjn:pvAD:ES"

/* !! Use SpdBoolean. */
static SpdBoolean   fullpath = TRUE;
SpdBoolean          linkclass = TRUE;
SpdBoolean          verbose = FALSE;

SpdBoolean          withAssert = FALSE;     /* JDK 1.4 assert. */
SpdBoolean          outputJ2jxEtb = FALSE;
static SpdBoolean   onlyParsing = FALSE;
static SpdBoolean   opt_S = FALSE;          /* Supress trivial warnings. */

/* ifndef YYDEBUGΤȤϺjpdDebugMessage()Τ˳褫Ƥ */
struct JpdDebug jpdDebug;

SpdConstString  frejaDir;

/* ETB files. */
FILE            *jdbinEtb;
FILE            *j2jxEtb = NULL;

char            *yyfilename;
SymData         *yylval;
SymData         *comment = NULL;
SpdStringPool   *str_pool = NULL;

static SpdConstString   program_name;       /* This program name. */
static SymData          *jpdFile;

extern char     *yytext;

static void parse_option(int argc, char *argv[]);
static void write_japid_version_top(void);
static void usage(void);
static void segvError(int x);
static void yyerror(char *msg);

static void write_pkg_info(SymData *qname);
static void write_pkg_info_sub(SymData *qname, FILE *fp);
static void write_XML_header(void);
static void do_analysis(void);
static SymData *setFile(SymData *file);
 
%}

%token ABSTRACT
%token BOOLEAN BREAK BYTE
%token CASE CAST CATCH CHAR CLASS CONST CONTINUE 
%token DEFAULT DO DOUBLE
%token ELSE EXTENDS
%token FINAL FINALLY FLOAT FOR
%token GENERIC GOTO
%token IF IMPLEMENTS IMPORT INNER INSTANCEOF INT INTERFACE
%token LONG
%token NATIVE NEW 
%token JNULL            /* Just NULL conflicts with one in C. */
%token OPERATOR OUTER
%token PACKAGE PRIVATE PROTECTED PUBLIC
%token REST RETURN
%token SHORT STATIC SUPER SWITCH SYNCHRONIZED
%token STRICTFP         /* Added for JDK 1.2. */
%token THIS THROW THROWS TRANSIENT TRY
%token VAR VOID VOLATILE
%token WHILE

%token JTRUE JFALSE     /* Just TRUE may conflict with one in C. */

%token ASSERT           /* JDK 1.4 assert. */

%token          AND
%token          AND_AND
%token          AND_EQUAL
%token          DOT
%token          EQUAL
%token          EQUAL_EQUAL
%token          EXCLAMATION
%token          EXCLAMATION_EQUAL
%token          HAT
%token          HAT_EQUAL
%token          LESS
%token          LESS_EQUAL
%token          LESS_LESS
%token          LESS_LESS_EQUAL
%token          MINUS
%token          MINUS_EQUAL
%token          MINUS_MINUS
%token          MORE
%token          MORE_EQUAL
%token          MORE_MORE
%token          MORE_MORE_MORE
%token          MORE_MORE_EQUAL
%token          MORE_MORE_MORE_EQUAL
%token          OR
%token          OR_EQUAL
%token          OR_OR
%token          PERCENT
%token          PERCENT_EQUAL
%token          PLUS
%token          PLUS_EQUAL
%token          PLUS_PLUS
%token          SLASH
%token          SLASH_EQUAL
%token          STAR
%token          STAR_EQUAL
%token          TILDA
%token          SEMICOLON
%token          COLON
%token          COMMA
%token          QUESTION
%token          OP_DIM

%token IDENTIFIER PRIMITIVE_TYPE_SPECIFIER MODIFIER
%token CHARACTER_LITERAL STRING_LITERAL UNICODE_LITERAL
%token INTEGER_LITERAL FLOATING_LITERAL

%nonassoc   EMPTY_MODIFIER
%nonassoc   '('
%nonassoc   FINAL       /* Important for final local variable. */

%nonassoc   LOWER_THAN_DOT
%nonassoc   DOT
%nonassoc   UPPER_THAN_DOT

%nonassoc   LOWER_THAN_ELSE
%nonassoc   ELSE

%start CompilationUnit

%%

TypeName /* ================================================================ */
    : PrimitiveType 
    {
        $$ = $1;
    }
    | QualifiedName 
    {
        $$ = qname2Type($1);
    }
    | ArrayType
    {
        $$ = $1;
    }
    ;

ArrayType /* =============================================================== */
    : TypeName OP_DIM 
    {
        setType($2, ARRAY_TYPE);
        $2->u.type->next = $1;  /* typeϵսΥꥹȤˤʤ */
        $$ = $2;
    }
    ;

PrimitiveType /* =========================================================== */
    : PRIMITIVE_TYPE_SPECIFIER 
    {
        setType($1, PRIMITIVE_TYPE);
        $$ = $1;
    }
    ;

QualifiedName /* =========================================================== */
    : IDENTIFIER 
    {
        $$ = setQName($1);
    }
    | QualifiedName DOT IDENTIFIER 
    {
        SymData *qname = getSymData();
        setQName(qname);
        setQName($3);
        setOperator($2, JPD_O_BINARY);
        qname->u.qname->first = $1;
        qname->u.qname->second = $3;
        qname->u.qname->dot = $2;
        $$ = qname;
        setOffsetAndLength($$, $1, $3);
    }
    ;

CompilationUnit /* ========================================================= */
    : PackageStatement ImportStatements TypeDeclarations 
    {
        jpdFile->u.file->definedPackage = $1;
        jpdFile->u.file->importedPackage = $2;
        jpdFile->u.file->classOrInterface = $3;
        $$ = jpdFile;
        if (comment != NULL) {
            $$->comment = comment;
            /* setComment($$, comment); */
        }
        jpdDebugFunc2Call(jpdDebug.jclass, putFile, $$, "");
    }
    ;

PackageStatement /* ======================================================== */
    : PACKAGE QualifiedName ';'
    {
        $$ = $2;
        
        setComment($$, $1);
        freeSymData($1);
        setComment($$, $3);
        freeSymData($3);

        write_pkg_info($2);
    }
    | /* Empty. */
    {
        $$ = NULL;

        write_pkg_info(NULL);
    }
    ;

ImportStatements /* ======================================================== */
    : ImportStatements ImportStatement
    {
        $2->u.qname->nextQName = $1;    /* սΥꥹ */
        $$ = $2;
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

ImportStatement /* ========================================================= */
    : IMPORT QualifiedName ';'
    {
        $$ = $2;

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $3);
        freeSymData($3);
    }
    | IMPORT QualifiedName DOT STAR ';'
    {
        SymData *qname = getSymData();

        setQName($4);               /* "*" QNameȤߤʤ */
        setQName(qname);
        setOperator($3, JPD_O_BINARY);
        qname->u.qname->first = $2;
        qname->u.qname->second = $4;
        qname->u.qname->dot = $3;
        $$ = qname;
        setOffsetAndLength($$, $2, $4);     

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $5);
        freeSymData($5);
    }
    ;

TypeDeclarations /* ======================================================== */
    :  TypeDeclarations TypeDeclaration
    {
        SymData *pushed = $2;

        if ($2->symSort == SPD_CID_J_CLASS) {
            $2->u.jclass->nextClassOrInterface = $1;
        } else if ($2->symSort == SPD_CID_J_INTERFACE) {
            $2->u.interface->nextClassOrInterface = $1;
        } else if ($2->symSort == SPD_CID_J_WILDCARD) {
            /* jdk 1.1ˤä"class ... {...};"ʤɤб뤿ᡥ */
            freeSymData($2);
            pushed = $1;
        }
        /* TypeDeclϵսΥꥹȤˤʤ롥 */
        $$ = pushed;
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

TypeDeclaration /* ========================================================= */
    : ClassDeclaration
    {
        $$ = $1;
        setClassSort($1, PACKAGE_MEMBER_CLASS);
    }
    | InterfaceDeclaration
    {
        $$ = $1;
        setInterfaceSort($1, PACKAGE_MEMBER_INTERFACE);
    }
    /* 
     * ʲruleʤȻפjdk 1.1.6      
     * "java/text/CompactByteArray.java"Ǥϡ饹θ";"Τ
     * ɲä
     */
    | ';'
    {
        $$ = $1;
        $$->symSort = SPD_CID_J_WILDCARD;
        if (!opt_S) {
            spdWarning("\"%s\" l.%d: Invalid type \";\".\n",
                       yyfilename, yylineno);
        }
    }
    ;

/* class˵ɽƤ뤿ClassDeclΥ롼ʬ䤷 */

ClassDeclaration  /* ======================================================= */
    : ClassPrefix ClassBody
    {
        $1->u.jclass->member = $2->u.routine->nextMember; /* $2Ȥϥߡ */
        popMemberSymTab();

        $$ = $1;
        setOffsetAndLength($$, $1, $2);

        setComment($$, $2);
        freeSymData($2);

        jpdDebugFunc2Call(jpdDebug.jclass, putClass, $$, "Define ");
    }
    ;

ClassPrefix /* ============================================================= */
    : Modifiers CLASS IDENTIFIER Super Interfaces
    {
        SymData *lastMod, *lastSym;

        setClass($3);
        $3->u.jclass->modifier = $1;
        $3->u.jclass->super = $4;
        $3->u.jclass->interface = $5;
        $3->u.jclass->memberSymTab = pushMemberSymTab();
        initLabelSymTab();

        $$ = $3;

        if ($5 != NULL) {
            lastSym = $5;
        } else if ($4 != NULL) {
            lastSym = $4;
        } else {
            lastSym = $3;
        }

        if ($1 == NULL) {
            setOffsetAndLength($$, $2, lastSym);
        } else {
            /* ֺǽ˽иModifierõ */
            for (lastMod = $1; lastMod->u.modifier->next != NULL;
                 lastMod = lastMod->u.modifier->next) {
                /* Empty. */
            }
            setOffsetAndLength($$, lastMod, lastSym);
        }

        setComment($$, $2);
        freeSymData($2);
    }
    ;

/* 
 * Modifiers rule is not applied Empty rule because StaticInitializer
 * rule and others (e.g. VariableDeclaration) in FieldDeclarations are conflict.
 */

Modifiers   /* ============================================================= */
    : Modifiers Modifier
    {
        $2->u.modifier->next = $1;  /* ModifiersϵսΥꥹȤˤʤ롥 */
        $$ = $2;
    }
    | /* Empty. */  %prec EMPTY_MODIFIER
    {   
        $$ = NULL;
    }
    ;

Modifier /* ================================================================ */
    : ABSTRACT 
    {
        $$ = setModifier($1);
    }
    | FINAL
    {
        $$ = setModifier($1);
    }
    | PUBLIC 
    {
        $$ = setModifier($1);
    }
    | PROTECTED 
    {
        $$ = setModifier($1);
    }
    | PRIVATE 
    {
        $$ = setModifier($1);
    }
    | STATIC 
    {
        $$ = setModifier($1);
    }
    | TRANSIENT 
    {
        $$ = setModifier($1);
    }
    | VOLATILE 
    {
        $$ = setModifier($1);
    }
    | NATIVE 
    {
        $$ = setModifier($1);
    }
    | SYNCHRONIZED 
    {
        $$ = setModifier($1);
    }
    | STRICTFP
    {
        /* Added for JDK 1.2. */
        $$ = setModifier($1);
    }
    ;

Super /* = QualifiedName is type name. ===================================== */
    : EXTENDS QualifiedName
    {
        $$ = qname2Type($2);

        setComment($$, $1);
        freeSymData($1);
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

Interfaces /* ============================================================== */
    : IMPLEMENTS InterfaceNameList
    {
        $$ = $2;

        setComment($$, $1);
        freeSymData($1);
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

InterfaceNameList /* ======================================================= */
    : TypeName      /* QualifiedNameäΤľ 1996/06/19. */
    {
        $$ = $1;
        jpdDebugFunc2Call(jpdDebug.type, putType, $1, "Interfaces Type");
    }
    | InterfaceNameList COMMA TypeName
    {
        $3->u.type->nextType = $1;      /* interfaceϵսˤʤ롥 */
        $$ = $3;

        setComment($$, $2);
        freeSymData($2);
    }
    ;

ClassBody /* =============================================================== */
    : '{' FieldDeclarations '}'
    {
        /* ClassDeclarationClassBodyoffset, length׻뤿 */
        /* ߡ롥ºݤRoutine"$1->u.routine->nextMember" */
        setRoutine($1, JPD_INVALID_ID);
        $1->u.routine->nextMember = $2;
        $$ = $1;
        setOffsetAndLength($$, $1, $3);

        setComment($$, $3);
        freeSymData($3);
    }
    ;

FieldDeclarations /* ======================================================= */
    : FieldDeclarations FieldDeclaration
    {
        SymData *pushed = $2;
        switch ($2->symSort) {              /* Фϵսˤʤ */
          case SPD_CID_J_VARIABLE:          /* ѿ */
            $2->u.variable->nextMember = $1;
            break;
          case SPD_CID_J_ROUTINE:           /* ᥽åɡ󥹥ȥ饯 */
            $2->u.routine->nextMember = $1;
            break;
          case SPD_CID_J_SBLOCK:            /* Ū */
            $2->u.sblock->nextMember = $1;
            break;
            /* JDK 1.1 */
          case SPD_CID_J_CLASS:             /* Х饹 */
            $2->u.jclass->nextMember = $1;
            break;
          case SPD_CID_J_INTERFACE:         /* Х饹 */
            $2->u.interface->nextMember = $1;
            break;
          case SPD_CID_J_WILDCARD:
            /* jdk 1.1äˤ void method {...};; */
            /* ʤɤб뤿Τ                       */
            freeSymData($2);
            pushed = $1;
            break;
          default:
            spdErrorExit("%s (l.%d, c.%d): %d: Can't be class member.\n",
                         yyfilename, yylineno, c_count, $2->symSort);
        }
        $$ = pushed;
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

FieldDeclaration /* ======================================================== */
    : FieldVariableDeclaration
    {
        $$ = $1;
    } 
    | MethodDeclaration
    {
        $$ = $1;
    } 
    | ConstructorDeclaration
    {
        $$ = $1;
    } 
    | StaticInitializer
    {
        $$ = $1;
    }
    /* Following rules are JDK 1.1. */
    | ClassDeclaration
    {
        SymData *mod;
        int topLevel = 0;
        for (mod = $1->u.jclass->modifier;
             mod != NULL; mod = mod->u.modifier->next) {
            if (SPD_STREQ(mod->str, "static")) {
                topLevel = 1;
            }
        }
        if (topLevel) {
            setClassSort($1, NESTED_TOP_LEVEL_CLASS);
        } else {
            setClassSort($1, MEMBER_CLASS);
        }
        $$ = $1;
    } 
    | InterfaceDeclaration
    {
        setInterfaceSort($1, NESTED_TOP_LEVEL_INTERFACE);
        $$ = $1;
    }
    | Block 
    {
        SymData *sblock;

        sblock = setSBlock(NULL, INSTANCE_INITIALIZER, $1);
        $$ = sblock;
        setOffsetAndLength($$, $1, $1);
        jpdDebugFunc2Call(jpdDebug.statement, putStatement,
                          sblock->u.sblock->block, "Static Block: ");
    }
    /* 
     * ʲruleʤȻפjdk 1.1.6
     * "java/net/Socket.java"Ǥϡ᥽åθ";"Τɲä
     */
    | ';'
    {
        $$ = $1;
        $$->symSort = SPD_CID_J_WILDCARD;
        if (!opt_S) {
            spdWarning("\"%s\" l.%d: Invalid member \";\".\n",
                       yyfilename, yylineno);
        }
    }
    ;

FieldVariableDeclaration /* = ѿʸ =========================== */
    : Modifiers TypeName VariableDeclarators ';'
    {
        $$ = setMemberVariable($3, $2, $1);

        setComment($$, $4);
        freeSymData($4);

        jpdDebugFunc2Call(jpdDebug.variable, putVariableDecl, $$, "MEMBER ");
        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Member Var Type");
    }
    ;

VariableDeclarators /* = ʣѿ ================================== */
    : VariableDeclarator 
    {
        $$ = $1;
    }
    | VariableDeclarators COMMA VariableDeclarator 
    {
        $3->u.variable->next = $1;  /* ѿ̾ϵսΥꥹȤˤʤ */
        $$ = $3;

        setComment($$, $2);
        freeSymData($2);
    }
    ;

VariableDeclarator /* = ѿν ===================================== */
    : DeclaratorName 
    {
        $$ = $1;
    }
    | DeclaratorName EQUAL VariableInitializer 
    {
        $1->u.variable->initializer = $3;
        $$ = $1;

        setComment($$, $2);
        freeSymData($2);
    }
    ;

DeclaratorName /* = ѿ̾ؿ̾ ======================================= */
    : IDENTIFIER 
    {
        /* λǤѿȻפäƽ */
        setVariable($1);
        $$ = $1;
    }
    | DeclaratorName OP_DIM 
    {
        /* ѿ(ؿ̾ǤϤꤨʤ) */
        setType($2, ARRAY_TYPE);
        $2->u.type->next = $1->u.variable->type;
        $1->u.variable->type = $2;
        $$ = $1;
    }
    ;

VariableInitializer /* ===================================================== */
    : Expression
    {
        $$ = $1;
    }
    | '{' ArrayInitializers '}'
    {
        /** Expr(E_LISTED) --first--> Expr --next--> ... --next--> */
        $$ = setExpression(NULL, E_LISTED, NULL, $2, NULL, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, $1, $3);

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $3);
        freeSymData($3);
    }
    ;

ArrayInitializers /* ======================================================= */
    : VariableInitializer
    {
        $$ = $1;
    }
    | ArrayInitializers COMMA VariableInitializer
    {
        $3->u.expression->next = $1;    /* initializerϵսΥꥹȤˤʤ */
        $$ = $3;

        setComment($$, $2);
        freeSymData($2);
    }
    | ArrayInitializers COMMA
    {
        $$ = $1;

        setComment($$, $2);
        freeSymData($2);
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

MethodDeclaration /* ======================================================= */
    : Modifiers TypeName MethodDeclarator Throws MethodBody  
    {
        SymData *lastMod;

        $3->u.routine->modifier = $1;
        if ($3->u.routine->returnType != NULL) {
            SymData *type = $3->u.routine->returnType;
            /* search last type */
            for (;type->u.type->next != NULL; type = type->u.type->next)
                ;
            type->u.type->next = $2;
        } else {
            $3->u.routine->returnType = $2;
        }
        
        $3->u.routine->throwType = $4;
        $3->u.routine->body = $5;

        popLocalSymTab();   /* parameterѤεɽ */
        popLabelSymTab();

        $$ = $3;

        if ($1 == NULL) {
            SymData *type, *firstType;
            type = firstType = $$->u.routine->returnType;
            for (; type != NULL; type = type->u.type->next) {
                if (type->offset < firstType->offset) {
                    firstType = type;
                }
            }
            setOffsetAndLength($$, firstType, $5);
        } else {
            /* ֺǽ˽иModifierõ */
            for (lastMod = $1; lastMod->u.modifier->next != NULL;
                 lastMod = lastMod->u.modifier->next) {
                /* Empty. */
            }
            setOffsetAndLength($$, lastMod, $5);
        }
    
        jpdDebugFunc2Call(jpdDebug.routine, putRoutine, $$, "Define ");
        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Routine Return Type");
        jpdDebugMessage(jpdDebug.routine, "\n\n");
    }
    ;

/* ᥽åɤεɽƤ뤿˥롼ʬ */

MethodDeclarator /* ======================================================== */
    /* : DeclaratorName '(' ParameterList ')' */
    : MethodDeclaratorPrefix ParameterList ')'
    {
        /* $1(DeclaratorName)VariableȤƽƤ */
        spdFree($1->u.variable);
        setRoutine($1, METHOD);
        $1->u.routine->parameter = $2;
        $$ = $1;

        pushLabelSymTab();

        setComment($$, $3);
        freeSymData($3);
    }
    | MethodDeclarator OP_DIM
    {
        /* type method()[]  */
        /*  cf. java.io.ByteArrayOutputStream#toByteArray() */
        setType($2, ARRAY_TYPE);
        if ($1->u.routine->returnType != NULL) {
            $2->u.type->next = $1->u.routine->returnType;
        }

        $1->u.routine->returnType = $2;
        $$ = $1;

        pushLabelSymTab();
    }
    ;

MethodDeclaratorPrefix /* ================================================== */
    : DeclaratorName '('
    {
        $$ = $1;
        pushLocalSymTab();

        setComment($$, $2);
        freeSymData($2);
    }
    ;

ParameterList /* =========================================================== */
    : Parameter
    {
        $$ = $1;
    }
    | ParameterList COMMA Parameter
    {
        $3->u.variable->next = $1;  /* parameterϵսˤʤ */
        $$ = $3;

        setComment($$, $2);
        freeSymData($2);
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

Parameter /* = ؿΰ ================================================ */
    : Modifiers TypeName DeclaratorName 
    {
        $$ = setLocalVariable($3, $2, $1, PARAMETER);
        jpdDebugFunc2Call(jpdDebug.variable, putVariableDecl,
                          $$, "PARAMETER ");
        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Parameter Type");
    }
    ;

Throws /* ================================================================== */
    : THROWS TypeNameList
    {
        $$ = $2;

        setComment($$, $1);
        freeSymData($1);
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;

TypeNameList /* ============================================================ */
    : TypeName
    {
        $$ = $1;
        jpdDebugFunc2Call(jpdDebug.type, putType, $1, "TypeNameList");
    }
    | TypeNameList COMMA TypeName
    {
        $3->u.type->nextType = $1;  /* TypeNameϵսˤʤ */
        $$ = $3;

        setComment($$, $2);
        freeSymData($2);
    }
    ;

MethodBody /* ============================================================== */
    : Block
    {
        $$ = $1;
    }
    | ';'
    {
        /* !! ad hocʽǽϹͤɬפꡥ */
        $$ = setStatement(NULL, S_EMPTY);
        setOffsetAndLength($$, $1, $1);
    }
    ;

ConstructorDeclaration /* ================================================== */
    : Modifiers ConstructorDeclarator Throws ConstructorBody 
    {
        SymData     *lastMod;

        $2->u.routine->modifier = $1;
        $2->u.routine->throwType = $3;
        $2->u.routine->body = $4;

        popLocalSymTab();   /* parameterѤεɽ */
        popLabelSymTab();

        $$ = $2;
        if ($1 == NULL) {
            setOffsetAndLength($$, $2, $4);
        } else {
            /* ֺǽ˽иModifierõ */
            for (lastMod = $1; lastMod->u.modifier->next != NULL;
                 lastMod = lastMod->u.modifier->next) {
                /* Empty. */
            }
            setOffsetAndLength($$, lastMod, $4);
        }

        jpdDebugFunc2Call(jpdDebug.routine, putRoutine, $$, "Define ");
        jpdDebugMessage(jpdDebug.routine, "\n\n");
    }
    ;

/* ѥ᡼εɽƤ뤿롼ʬ */

ConstructorDeclarator /* =================================================== */
    /* : TypeName '(' ParameterList ')'*/
    : ConstructorDeclaratorPrefix ParameterList ')'
    {
        /* $1(DeclaratorName)TypeNameȤƽƤ                 */
        /* TypeNameDOTǷ礵ƤʤOBJECT_TYPEˤʤäƤϤ   */
        spdFree($1->u.type);
        setRoutine($1, CONSTRUCTOR);
        $1->u.routine->parameter = $2;
        $$ = $1;

        pushLabelSymTab();

        setComment($$, $3);
        freeSymData($3);

        jpdDebugFunc2Call(jpdDebug.type, putType, $1, "Constructor Type");
    }           
    ;

ConstructorDeclaratorPrefix /* ============================================= */
    : TypeName '('
    {
        $$ = $1;
        pushLocalSymTab();

        setComment($$, $2);
        freeSymData($2);
    }
    ;

ConstructorBody /* ========================================================= */
    : Block
    {
        $$ = $1;
    }
    ;

StaticInitializer /* ======================================================= */
    : Modifiers Block           /* Modifiers must be STATIC */
    {
        SymData *sblock;

        if ($1 == NULL || !SPD_STREQ($1->str, "static")
            || $1->u.modifier->next != NULL) {
            yyerror("Static initializer must be \"static { ... }\"");
        }

        sblock = setSBlock(NULL, STATIC_INITIALIZER, $2);

        popLabelSymTab();

        $$ = sblock;
        setOffsetAndLength($$, $1, $2);     

        setComment($$, $1);
        freeSymData($1);

        jpdDebugFunc2Call(jpdDebug.statement, putStatement,
                          sblock->u.sblock->block, "Static Block: ");
    }
    ;

/* interface˵ɽƤ뤿InterfaceDeclΥ롼ʬ䤷 */

InterfaceDeclaration /* ==================================================== */
    : InterfacePrefix InterfaceBody
    {
        $1->u.interface->member = $2->u.routine->nextMember;
            /* $1Ȥϥߡ */;
        popMemberSymTab();
        
        $$ = $1;
        setOffsetAndLength($$, $1, $2);
        
        setComment($$, $2);
        freeSymData($2);

        jpdDebugFunc2Call(jpdDebug.interface, putInterface, $$, "Define ");
    }
    ;

InterfacePrefix /* ========================================================= */
    :  Modifiers INTERFACE IDENTIFIER ExtendsInterfaces
    {
        SymData *lastMod, *lastSym;

        setInterface($3);
        $3->u.interface->modifier = $1;
        $3->u.interface->super = $4;
        $3->u.interface->memberSymTab = pushMemberSymTab();
        
        $$ = $3;

        if ($4 != NULL) {
            lastSym = $4;
        } else {
            lastSym = $3;
        }

        if ($1 == NULL) {
            setOffsetAndLength($$, $2, lastSym);
        } else {
            /* ֺǽ˽иModifierõ */
            for (lastMod = $1; lastMod->u.modifier->next != NULL;
                 lastMod = lastMod->u.modifier->next)
                /* Empty. */;
            setOffsetAndLength($$, lastMod, lastSym);
        }
        
        setComment($$, $2);
        freeSymData($2);
    }
    ;

ExtendsInterfaces /* ======================================================= */
    : EXTENDS InterfaceNameList
    {
        $$ = $2;

        setComment($$, $1);
        freeSymData($1);
    }
    | /* Empty. */
    {
        $$ = NULL;
    }
    ;
    
InterfaceBody /* =========================================================== */
    : '{' FieldDeclarations '}'
    {
        /* InterfaceDeclarationInterfaceBodyoffset,length׻ */
        /* Υߡ롥ºݤRoutine $1->u.routine->nextMember */
        setRoutine($1, JPD_INVALID_ID);
        $1->u.routine->nextMember = $2;
        $$ = $1;
        setOffsetAndLength($$, $1, $3);

        setComment($$, $3);
        freeSymData($3);
    }
    ;

Block /* =================================================================== */
    : LeftBrace BlockStatements RightBrace
    {
        SymData *block = setStatement(NULL, S_BLOCK);
        block->u.statement->block = $2;
        $$ = block;
        setOffsetAndLength($$, $1, $3);

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $3);
        freeSymData($3);
    }
    | '{' '}'
    {
        SymData *block = setStatement(NULL, S_BLOCK);
        /* block->u.statement->next = NULL in setStatement */
        $$ = block;
        setOffsetAndLength($$, $1, $2);

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $2);
        freeSymData($2);
    }
    ;

/* ѿεɽ뤿ε§ */

LeftBrace /* =============================================================== */
    : '{'
    {
        pushLocalSymTab();
    }
    ;

RightBrace /* ============================================================== */
    : '}'
    {
        popLocalSymTab();
    }
    ;

BlockStatements /* ========================================================= */
    : BlockStatement
    {
        $$ = $1;
    }
    | BlockStatements BlockStatement
    {
        $2->u.statement->next = $1;     /* statementϵսˤʤ */
        $$ = $2;
    }
    ;

BlockStatement /* ========================================================== */
    : BlockStatementWithoutSwitchLabel
    {
        $$ = $1;
    }
    | SwitchLabeledStatement
    {
        $$ = $1;
    }
    /* following rule is JDK 1.1 */
    | ClassDeclaration
    {
        SymData *stmt;

        setClassSort($1, LOCAL_CLASS);

        stmt = setStatement(NULL, S_LOCAL_CLASS);
        setStatementClass(stmt, $1);
        $$ = stmt;
        setOffsetAndLength($$, $1, $1);
    }
    /* Υ롼finalѿȤξͤ򤱤뤿ɬ */
    | FINAL ClassDeclaration
    {
        SymData *stmt;

        setModifier($1);
        setClassSort($1, LOCAL_CLASS);
        $1->u.modifier->next = $2->u.jclass->modifier;
        $2->u.jclass->modifier = $1;

        stmt = setStatement(NULL, S_LOCAL_CLASS);
        setStatementClass(stmt, $2);
        $$ = stmt;
        setOffsetAndLength($$, $1, $2);

    }
    ;

BlockStatementWithoutSwitchLabel /* ======================================== */
    : LocalVariableDeclarationStatement
    {
        $$ = $1;

        jpdDebugFunc2Call(jpdDebug.statement, putStatement, $1, "");
    }
    | StatementWithoutSwitchLabel
    {
        $$ = $1;

        jpdDebugFunc2Call(jpdDebug.statement, putStatement, $1, "");
    }
    ;

LocalVariableDeclarationStatement /* = ѿʸ ============== */
    : TypeName VariableDeclarators ';'
    {
        SymData *vars, *stmt;
        
        vars = setLocalVariable($2, $1, NULL, LOCAL_VAR);
        stmt = setStatement(NULL, S_DEF_VAR);
        setStatementVariableDecl(stmt, vars);
        $$ = stmt;
        setOffsetAndLength($$, $1, $3);

        setComment($$, $3);
        freeSymData($3);

        jpdDebugFunc2Call(jpdDebug.variable, putVariableDecl, vars, "LOCAL ");
        jpdDebugFunc2Call(jpdDebug.type, putType, $1, "Local Var Type");
    }
    /* following rule is JDK 1.1 */
    | FINAL TypeName VariableDeclarators ';'
    {
        SymData *vars, *stmt;

        setModifier($1);        
        vars = setLocalVariable($3, $2, $1, LOCAL_VAR);
        stmt = setStatement(NULL, S_DEF_VAR);
        setStatementVariableDecl(stmt, vars);
        $$ = stmt;
        setOffsetAndLength($$, $1, $4);

        setComment($$, $4);
        freeSymData($4);

        jpdDebugFunc2Call(jpdDebug.variable, putVariableDecl, vars, "LOCAL ");
        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Local Var Type");

    }
    ;

StatementWithoutSwitchLabel /* ============================================= */
    : EmptyStatement
    {
        $$ = $1;
    }
    | ExpressionStatement
    {
        $$ = $1;
    }
    | SelectionStatement
    {
        $$ = $1;
    }
    | IterationStatement
    {
        $$ = $1;
    }
    | JumpStatement
    {
        $$ = $1;
    }
    | GuardingStatement
    {
        $$ = $1;
    }
    | LabeledStatement
    {
        $$ = $1;
    }
    | Block
    {
        $$ = $1;
    }
    | AssertStatement
    {
        /* JDK 1.4 assert. */
    }
    ;

AssertStatement /* ========================================================= */
   : ASSERT Expression ';'
   {
        $$ = setStatement($1, S_ASSERT);
        setStatementAssert($$, $2, NULL);
        setOffsetAndLength($$, $1, $3);

        setComment($$, $3);
        freeSymData($3);
   }
   | ASSERT Expression COLON Expression ';'
   {
        $$ = setStatement($1, S_ASSERT);
        setStatementAssert($$, $2, $4);
        setOffsetAndLength($$, $1, $5);

        setComment($$, $3);
        freeSymData($3);
        setComment($$, $5);
        freeSymData($5);
   }
   ; 

EmptyStatement /* ========================================================== */
    : ';'
    {
        $$ = setStatement(NULL, S_EMPTY);
        setOffsetAndLength($$, $1, $1);

        setComment($$, $1);
        freeSymData($1);
    }
    ;

SwitchBlock /* ============================================================= */
    : Block
    {
        $$ = $1;
    }
    | LeftBrace BlockStatements  SwitchLabels RightBrace
    {
        SymData *emptyStmt, *block;

        emptyStmt = setStatement(NULL, S_SWITCH_LABELED_EMPTY);
        setStatementLabel(emptyStmt, $3);

        emptyStmt->u.statement->next = $2;
        block = setStatement(NULL, S_BLOCK);
        block->u.statement->block = emptyStmt;
        $$ = block;

        setOffsetAndLength($$, $1, $4);

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $4);
        freeSymData($4);

    }
    | LeftBrace SwitchLabels RightBrace
    {
        SymData *emptyStmt, *block;

        emptyStmt = setStatement(NULL, S_SWITCH_LABELED_EMPTY);
        setStatementLabel(emptyStmt, $2);

        emptyStmt->u.statement->next = NULL;
        block = setStatement(NULL, S_BLOCK);
        block->u.statement->block = emptyStmt;
        $$ = block;

        setOffsetAndLength($$, $1, $3);

        setComment($$, $1);
        freeSymData($1);
        setComment($$, $3);
        freeSymData($3);
    }
    ;

LabeledStatement /* ======================================================== */
    : Label BlockStatement
    {
        setStatementLabel($2, $1);
        $$ = $2;
    }
    ;

SwitchLabeledStatement /* ================================================== */
    : SwitchLabels BlockStatementWithoutSwitchLabel
    {
        setStatementLabel($2, $1);
        $$ = $2;
    }
    ;

SwitchLabels /* ============================================================ */
    : SwitchLabel
    {
        $$ = $1;
    }
    | SwitchLabels SwitchLabel
    {
        SymData *last;

        for (last = $1; last->u.label->next != NULL;
             last = last->u.label->next)
            ;
        last->u.label->next = $2;
        $$ = $1;
    }
    ;

SwitchLabel /* ============================================================= */
    : CASE ConstantExpression COLON
    {
        SymData     *label;

        label = setCaseLabel($1, $2);
        setOffsetAndLength(label, $1, $2);
        $$ = label;

        /* !! ȤΥȡˤĤʤΤǡޤ⤷ʤ */
        setComment($$, $3);
 
        freeSymData($3);
    }
    | DEFAULT COLON
    {
        SymData *label;

        label = setDefaultLabel($1);
        setOffsetAndLength(label, $1, $1);
        $$ = label;

        /* !! ȤΥȡˤĤʤΤǡޤ⤷ʤ */
        setComment($$, $2);
        freeSymData($2);
    }
    ;

Label /* =================================================================== */
    : IDENTIFIER COLON
    {
        $$ = setLabel($1, DEF_LABEL);
        setOffsetAndLength($$, $1, $2);

        setComment($$, $2);
        freeSymData($2);
    }
    ;

ExpressionStatement /* ===================================================== */
    : Expression ';'
    {
        SymData *stmt = setStatement(NULL, S_EXPRESSION);
        setStatementExpression(stmt, $1);
        setOffsetAndLength(stmt, $1, $2);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
    }
    ;

Expression /* ============================================================== */
    : AssignmentExpression 
    {
        jpdDebugFunc2Call(jpdDebug.expression, putExpression, $1, "");
        $$ = $1;
    }
    ;

SelectionStatement /* ====================================================== */
    : IF '(' Expression ')' BlockStatement %prec LOWER_THAN_ELSE
                            /* may be Statement */
    {
        SymData *stmt = setStatement($1, S_IF);
        setStatementCondition(stmt, $3);
        setStatementIfThen(stmt, $5);
        /* setStatementIfElse(stmt, NULL) in setStatement */
        setOffsetAndLength(stmt, $1, $5);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);
    }
    | IF '(' Expression ')' BlockStatement ELSE BlockStatement /* may be Statement */
    {
        SymData *stmt = setStatement($1, S_IF);
        setStatementCondition(stmt, $3);
        setStatementIfThen(stmt, $5);
        setStatementIfElse(stmt, $7);
        setOffsetAndLength(stmt, $1, $7);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);
        setComment($$, $6);
        freeSymData($6);
    }
    | SWITCH '(' Expression ')' SwitchBlock
    {
        SymData *stmt = setStatement($1, S_SWITCH);
        setStatementCondition(stmt, $3);
        setStatementAction(stmt, $5);
        setOffsetAndLength(stmt, $1, $5);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);
    }
    ;

IterationStatement /* ====================================================== */
    : WHILE '(' Expression ')' BlockStatement /* may be Statement */
    {
        SymData *stmt = setStatement($1, S_WHILE);
        setStatementCondition(stmt, $3);
        setStatementAction(stmt, $5);
        setOffsetAndLength(stmt, $1, $5);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);
    }
    | DO BlockStatement WHILE '(' Expression ')' ';' /* may be Statement */
    {
        SymData *stmt = setStatement($1, S_DO);
        setStatementCondition(stmt, $5);
        setStatementAction(stmt, $2);
        setOffsetAndLength(stmt, $1, $7);
        $$ = stmt;

        setComment($$, $3);
        freeSymData($3);
        setComment($$, $4);
        freeSymData($4);
        setComment($$, $6);
        freeSymData($6);
        setComment($$, $7);
        freeSymData($7);
    }
    /*
     * ForʸޤΥ롼Ϥäȱ
     * Empty§Ȥ褦ˤQName->JustNameQName->TypeName
     * ȤƲᤵƤޤʸϤǤʤΤǡΤޤ޻Ĥ
     */
    | FOR '(' ForInit ForExpr ForIncr ')' BlockStatement /* may be Statement */
    {
        SymData *stmt = setStatement($1, S_FOR);
        setStatementForInit(stmt, $3);
        setStatementCondition(stmt, $4);
        setStatementForIncr(stmt, $5);
        setStatementAction(stmt, $7);
        setOffsetAndLength(stmt, $1, $7);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $6);
        freeSymData($6);
    }
    | FOR '(' ForInit ForExpr         ')' BlockStatement /* may be Statement */
    {
        SymData *emptyExpr, *stmt;

        emptyExpr = setExpression(NULL, E_EMPTY, NULL, NULL, NULL, NULL);
        setLRSort(emptyExpr, NONE);
        setOffsetAndLength(emptyExpr, $5, $5);
        emptyExpr->length = 0;

        stmt = setStatement($1, S_FOR);
        setStatementForInit(stmt, $3);
        setStatementCondition(stmt, $4);
        setStatementForIncr(stmt, emptyExpr);
        setStatementAction(stmt, $6);
        setOffsetAndLength(stmt, $1, $6);
        $$ = stmt;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $5);
        freeSymData($5);
    }
    ;

ForInit /* ================================================================= */
    :  Expressions ';'
    {
        /** Expr(E_LISTED) --first--> Expr --next--> ... --next--> */
        SymData *p;
        $$ = setExpression(NULL, E_LISTED, NULL, $1, NULL, NULL);
        setLRSort($$, NONE);
        /* search for first expression */
        for (p = $1; p->u.expression->next != NULL; p = p->u.expression->next)
            ;
        setOffsetAndLength($$, p, $1);

        setComment($$, $2);
        freeSymData($2);
    }
    | LocalVariableDeclarationStatement
    {
        $$ = $1;
    }
    | ';'
    {
        $$ = setExpression(NULL, E_EMPTY, NULL, NULL, NULL, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, $1, $1);
        $$->length = 0;

        setComment($$, $1);
        freeSymData($1);
    }
    ;

ForExpr /* ================================================================= */
    :  Expression ';'
    {
        $$ = $1;

        setComment($$, $2);
        freeSymData($2);
    }
    | ';'
    {
        $$ = setExpression(NULL, E_EMPTY, NULL, NULL, NULL, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, $1, $1);
        $$->length = 0;

        setComment($$, $1);
        freeSymData($1);
    }
    ;

ForIncr /* ================================================================= */
    : Expressions
    {
        SymData *p;
        $$ = setExpression(NULL, E_LISTED, NULL, $1, NULL, NULL);
        setLRSort($$, NONE);
        /* search for first expression */
        for (p = $1; p->u.expression->next != NULL; p = p->u.expression->next)
            ;
        setOffsetAndLength($$, p, $1);
    }
    ;

Expressions /* ============================================================= */
    : Expression
    {
        $$ = $1;
    }
    | Expressions COMMA Expression
    {
        $3->u.expression->next = $1;    /* expressionϵսˤʤ */
        $$ = $3;
    }
    ;

JumpStatement /* =========================================================== */
    : BREAK IDENTIFIER ';'
    {
        SymData *label;
        label = setLabel($2, REF_LABEL);

        setStatement($1, S_BREAK);
        setStatementLabel($1, label);
        setOffsetAndLength($1, $1, $3);
        $$ = $1;

        setComment($$, $3);
        freeSymData($3);
    }
    | BREAK ';'
    {
        setOffsetAndLength($1, $1, $2);
        $$ = setStatement($1, S_BREAK);
        /* setStatementJumpLabel($1, NULL) in setStatement */

        setComment($$, $2);
        freeSymData($2);
    }
    | CONTINUE IDENTIFIER ';'
    {
        SymData *label;
        label = setLabel($2, REF_LABEL);

        setStatement($1, S_CONTINUE);
        setStatementLabel($1, label);
        setOffsetAndLength($1, $1, $3);
        $$ = $1;

        setComment($$, $3);
        freeSymData($3);
    }
    | CONTINUE ';'
    {
        setOffsetAndLength($1, $1, $2);
        $$ = setStatement($1, S_CONTINUE);
        /* setStatementJumpLabel($1, NULL) in setStatement */

        setComment($$, $2);
        freeSymData($2);
    }
    | RETURN Expression ';'
    {
        setStatement($1, S_RETURN);
        setStatementExpression($1,  $2);
        setOffsetAndLength($1, $1, $3);
        $$ = $1;

        setComment($$, $3);
        freeSymData($3);
    }
    | RETURN ';'
    {
        setStatement($1, S_RETURN);
        /* $$->u.statement->expression1 = NULL in setStatement */
        setOffsetAndLength($1, $1, $2);
        $$ = $1;

        setComment($$, $2);
        freeSymData($2);
    }
    | THROW Expression ';'
    {
        setStatement($1, S_THROW);
        setStatementExpression($1, $2);
        setOffsetAndLength($1, $1, $3);
        $$ = $1;

        setComment($$, $3);
        freeSymData($3);
    }
    ;

GuardingStatement /* ======================================================= */
    /* : SYNCHRONIZED '(' Expression ')' BlockStatement /* Maybe Statement. */
    : Modifiers '(' Expression ')' BlockStatement /*%prec UPPER_THAN_MODIFIER*/
    {
        if (!SPD_STREQ($1->str, "synchronized")) {
            spdErrorExit("%s: Illegal Synchronized Statement, %s\n",
                         yyfilename, $1->str);
        }
        setStatement($1, S_SYNCHRONIZED);
        setStatementSynchronizedExpression($1, $3);
        setStatementSynchronizedStatement($1, $5);
        setOffsetAndLength($1, $1, $5);
        $$ = $1;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);
    }
    /* 
     * TRY
     * S_EXCEPTION
     *  ->expression1 = Try Statement
     *  ->expression2 = Catches
     *  ->expression3 = Finally Statement
     * S_TRY
     *  ->statement1 = Try Block
     * S_CATCHES
     *  ->expression1 = VariableDecl(argument)
     *  ->statement1  = Block
     *  ->next        = Catches
     * S_FINALLY
     *  ->statement1 = Finally Block
     * Ȥ
     */
    | TRY Block Finally
    {
        SymData *exception = setStatement(NULL, S_EXCEPTION);
        setStatement($1, S_TRY);
        setStatementTryBlock($1, $2);
        setOffsetAndLength($1, $1, $2);
        setStatementExceptionTry(exception, $1);
        setStatementExceptionFinally(exception, $3);
        setOffsetAndLength(exception, $1, $3);
        $$ = exception;
    }
    | TRY Block Catches
    {
        SymData *exception = setStatement(NULL, S_EXCEPTION);
        setStatement($1, S_TRY);
        setStatementTryBlock($1, $2);
        setOffsetAndLength($1, $1, $2);
        setStatementExceptionTry(exception, $1);
        setStatementExceptionCatches(exception, $3);
        setOffsetAndLength(exception, $1, $3);
        $$ = exception;
    }
    | TRY Block Catches Finally
    {
        SymData *exception = setStatement(NULL, S_EXCEPTION);
        setStatement($1, S_TRY);
        setStatementTryBlock($1, $2);
        setOffsetAndLength($1, $1, $2);
        setStatementExceptionTry(exception, $1);
        setStatementExceptionCatches(exception, $3);
        setStatementExceptionFinally(exception, $4);
        setOffsetAndLength(exception, $1, $4);
        $$ = exception;
    }
    ;

Catches /* ================================================================= */
    : Catch
    {
        $$ = $1;
    }
    | Catches Catch
    {
        $2->u.statement->next = $1;     /* Catchesϵսˤʤ롥 */
        $$ = $2;
    }
    ;

Catch /* =================================================================== */
    : CATCH '(' CatchArgument ')' Block
    {
        setStatement($1, S_CATCH);
        setStatementCatchArgument($1, $3);
        setStatementCatchBlock($1, $5);
        setOffsetAndLength($1, $1, $5);
        $$ = $1;

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);
    }
    ;

CatchArgument /* =========================================================== */
    : TypeName IDENTIFIER
    {
        setVariable($2);
        $$ = setLocalVariable($2, $1, NULL, LOCAL_VAR);

        jpdDebugFunc2Call(jpdDebug.type, putType, $1, "Catch Type");
    }
    ;

Finally /* ================================================================= */
    : FINALLY Block
    {
        setStatement($1, S_FINALLY);
        setStatementFinallyBlock($1, $2);
        setOffsetAndLength($1, $1, $2);
        $$ = $1;
    }
    ;

PrimaryExpression /* ======================================================= */
     : PrimaryNoNewArray
     {
         $$ = $1;
     }
     | ArrayCreationExpression
     {
         $$ = $1;
     }
     ;

PrimaryNoNewArray /* = ѿƥλȡ᥽åɵư ================ */
    /* : JustName */
    : QualifiedName %prec LOWER_THAN_DOT
    {
        $$ = qname2PrimaryNoNewArray($1);
    }
    | SpecialName
    {
        if (!SPD_STREQ($1->str, "this") &&
            !SPD_STREQ($1->str, "super") &&
            !SPD_STREQ($1->str, "null") &&
            !SPD_STREQ($1->str, "true") &&
            !SPD_STREQ($1->str, "false")) {
            spdErrorExit("\"%s\": Invalid special name in file \"%s\".\n",
                         $1->str, yyfilename);
        }
        $$ = setExpression(NULL, E_REF_SPECIAL, NULL, $1, NULL, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $1);
    }
    | NotJustNameNoNewArray
    {
        $$ = $1;
    }
    ;

NotJustNameNoNewArray /* =================================================== */
    : ClassInstanceCreationExpression 
    {
        $$ = $1;
    }
    | ComplexPrimary 
    {
        $$ = $1;
    }
    ;

SpecialName /* ============================================================= */
    : THIS 
    {
        setSpecialName($1);
        $$ = $1;
    }
    | SUPER 
    {
        setSpecialName($1);
        $$ = $1;
    }
    | JNULL 
    {
        setSpecialName($1);
        $$ = $1;
    }
    | JTRUE 
    {
        setSpecialName($1);
        $$ = $1;
    }
    | JFALSE 
    {
        setSpecialName($1);
        $$ = $1;
    }
    ;

ComplexPrimary /* ========================================================== */
    : Literal 
    {
        SymData *expr
            = setExpression(NULL, E_REF_LITERAL, NULL, $1, NULL, NULL);

        setLRSort(expr, R_VALUE);
        setOffsetAndLength(expr, $1, $1);
        $$ = expr;
    }
    | '(' Expression ')'
    {
        setOperator($1, JPD_O_PAREN);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);

        setComment($$, $3);
        freeSymData($3);
    }
    | ArrayAccess
    {
        $$ = $1;
    }
    | FieldAccess 
    {
        $$ = $1;
    }
    | MethodCall 
    {
        $$ = $1;
    }
    /* For JDK 1.1. */
    | QualifiedName DOT THIS
    {
        /* QualifiedName is ClassName. */
        SymData *refType, *refSpecial;

        /* Set type, operator and special. */
        setType($1, OBJECT_TYPE);
        setOperator($2, JPD_O_BINARY);
        setSpecialName($3);

        /* set REF_TYPE expr */
        refType = setExpression(NULL, E_REF_TYPE, NULL, $1, NULL, NULL);
        setLRSort(refType, R_VALUE);
        setOffsetAndLength(refType, $1, $1);

        /* set REF_SPECIAL expr */

        refSpecial = setExpression(NULL, E_REF_SPECIAL, NULL,
                                   $3, NULL, NULL);
        setLRSort(refSpecial, NONE);
        setOffsetAndLength(refSpecial, $3, $3);


        $$ = setExpression(NULL, E_OPERATION, $2, refType, refSpecial, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, refType, refSpecial);
    }
    | QualifiedName DOT SUPER
    {
        /* QualifiedName is ClassName. */
        SymData *refType, *refSpecial;

        /* Set type, operator and special. */
        setType($1, OBJECT_TYPE);
        setOperator($2, JPD_O_BINARY);
        setSpecialName($3);

        /* set REF_TYPE expr */
        refType = setExpression(NULL, E_REF_TYPE, NULL, $1, NULL, NULL);
        setLRSort(refType, R_VALUE);
        setOffsetAndLength(refType, $1, $1);

        /* set REF_SPECIAL expr */
        refSpecial = setExpression(NULL, E_REF_SPECIAL, NULL,
                                   $3, NULL, NULL);
        setLRSort(refSpecial, NONE);
        setOffsetAndLength(refSpecial, $3, $3);


        $$ = setExpression(NULL, E_OPERATION, $2, refType, refSpecial, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, refType, refSpecial);
    }
    /* TypeName DOT CLASSconflictΤTypeNameŸ롥 */
    | PrimitiveType DOT CLASS
    {
        $$ = setClassLiteralExpression($1, $2, $3);
    }
    | QualifiedName DOT CLASS
    {
        SymData *type  = qname2Type($1);
        $$ = setClassLiteralExpression(type, $2, $3);
    }
    | ArrayType DOT CLASS
    {
        $$ = setClassLiteralExpression($1, $2, $3);
    }
    ;

ArrayAccess /* ============================================================= */
    /* : JustName '[' Expression ']' */
    : QualifiedName '[' Expression ']'
    {
        SymData *var;
        SpdBoolean resolveLater = FALSE;

        lockLRSortR($3);
        setOperator($2, JPD_O_REF_DIM);
        /* JustName is QName or SpecialName: ᥽å̾ */
        if ($1->symSort == SPD_CID_J_SPECIAL) {   /* SpecialName */
            spdErrorExit("%s[] is ArrayAccess?\n", $1->str);
        } else if ($1->u.qname->dot == NULL) {
            /* DOTǷ礵ƤʤQualifiedName */
            var = searchForLocalSymTab($1->str, NULL);
            if (var == NULL) {
                var = searchForMemberSymTab($1->str, NULL);
                if (var == NULL) {
                    resolveLater = TRUE;
                    if (verbose) {
                        spdWarning("Unknown identifier (l.%d, c.%d):"
                                   " \"%s\".\n", yylineno, c_count, $1->str);
                    }
                }
            }
            spdFree($1->u.qname);

            setExpression($1, E_REF_VAR, NULL, var, NULL, NULL);
            setLRSort($1, R_VALUE);
            $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
            setLRSort($$, R_VALUE);
            setTwoPassResolve($$, resolveLater);
            setOffsetAndLength($$, $1, $4);

            jpdDebugMessage(jpdDebug.variable, "(l.%d, c.%d): ",
                            yylineno, c_count);
            jpdDebugFunc2Call(jpdDebug.variable, putVariable, var, "Refer");
        } else {  /* DOTǷ礵줿QualifiedName */
            SymData         *qname, *expr;

            for (qname = $1; qname->u.qname->first != NULL;
                 qname = qname->u.qname->first)
                ;
            
            var = searchForLocalSymTab(qname->str, NULL);
            if (var == NULL) {
                var = searchForMemberSymTab(qname->str, NULL);
            }

            if (var != NULL) {
                /* qnameѿ(֥), eg. obj.mem1.mem2 */
                SymData *p;
                expr = qname2Expression($1, R_VALUE);

                /* ѿ */             
                for (p = expr; p != NULL && p->symSort == SPD_CID_J_EXPRESSION;
                     p = p->u.expression->first) {
                    if (p->u.expression->sort == E_REF_VAR &&
                        SPD_STREQ(p->str, var->str)) {
                        p->u.expression->first = var;
                        break;
                    }
                }
            } else {
                /*
                 * Ǥa.b.ccѿȻפä
                 * Expr(OP) = Expr(REF_QNAME, a.b) OP(.) Expr(REF_VAR, c)
                 * Ȥ򤷤Ƥ롥
                 */
                SymData *first, *second;

                first = setExpression(NULL, E_REF_QNAME, NULL,
                                      $1->u.qname->first, NULL, NULL);
                setLRSort(first, R_VALUE);
                setOffsetAndLength(first, $1->u.qname->first,
                                $1->u.qname->first);
                spdFree($1->u.qname->second->u.qname);
                second = setExpression($1->u.qname->second, E_REF_VAR, NULL,
                                       NULL, NULL, NULL);
                setLRSort(second, R_VALUE);
                setOffsetAndLength(second, $1->u.qname->second,
                                $1->u.qname->second);
                expr = setExpression(NULL, E_OPERATION, $1->u.qname->dot,
                                   first, second, NULL);
                setLRSort(expr, R_VALUE);
                setOffsetAndLength(expr, first, second);
                spdFree($1->u.qname);              
            }
            $$ = setExpression(NULL, E_OPERATION, $2, expr, $3, NULL);
            setLRSort($$, R_VALUE);
            setOffsetAndLength($$, expr, $4);
            
        }
        setComment($$, $4);
        freeSymData($4);

        jpdDebugFunc2Call(jpdDebug.expression,
                              putExpression, $$, "ArrayAccess ");
    }
    | ComplexPrimary '[' Expression ']'
    {
        setOperator($2, JPD_O_REF_DIM);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        lockLRSortR($3);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $4);

        setComment($$, $4);
        freeSymData($4);
    }
    ;

FieldAccess /* === л ============================================== */
    : PrimaryExpression DOT IDENTIFIER 
    {
        /* PrimaryEPostfixEΥ롼 */
        /* MethodAccessȤζ̤ɬפ餷 */
        
        SymData *var;
        SpdBoolean resolveLater = FALSE;

        /*
         * PrimaryE is SpecialName or ArrayAccess or MethodCall
         * or ( Exression ).
         */
        if ($1->u.expression->sort == E_REF_SPECIAL) {
            if (SPD_STREQ($1->u.expression->first->str, "this")) {
                /* ѿؤΥ */
                var = searchForMemberSymTab($3->str, NULL);
                if (var == NULL) {
                    resolveLater = TRUE;
                    if (verbose) {
                        spdWarning("%s: Member variable \"%s\" isn't defined"
                                   "(l.%d, c.%d).\n",
                                   yyfilename, $3->str, yylineno, c_count);
                    }
                }

                setExpression($3, E_REF_VAR, NULL, var, NULL, NULL);
                setLRSort($3, R_VALUE);
                setTwoPassResolve($3, resolveLater);
                setOperator($2, JPD_O_BINARY);
                $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
                setLRSort($$, R_VALUE);
                setOffsetAndLength($$, $1, $3);

                
                jpdDebugMessage(jpdDebug.variable, "(l.%d, c.%d): ",
                                yylineno, c_count);
                jpdDebugFunc2Call(jpdDebug.variable,
                                  putVariable, var, "Refer THIS");
            } else if (SPD_STREQ($1->u.expression->first->str, "super")) { 
                setExpression($3, E_REF_VAR, NULL, NULL, NULL, NULL);
                setLRSort($3, R_VALUE);
                setOperator($2, JPD_O_BINARY);
                $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
                setLRSort($$, R_VALUE);
                setOffsetAndLength($$, $1, $3);

                jpdDebugMessage(jpdDebug.variable, "(l.%d, c.%d): ",
                                yylineno, c_count);
                jpdDebugFunc2Call(jpdDebug.variable,
                                  putVariable, var, "Refer Super");
            }
        } else {
            setOperator($2, JPD_O_BINARY);
            setExpression($3, E_REF_VAR, NULL, NULL, NULL, NULL);
            setLRSort($3, R_VALUE);
            $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
            setOffsetAndLength($$, $1, $3);
        }
    }
    | PostfixExpression DOT IDENTIFIER
    {
         setOperator($2, JPD_O_BINARY);
         setExpression($3, E_REF_VAR, NULL, NULL, NULL, NULL);
         setLRSort($3, R_VALUE);
         $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
         setLRSort($$, R_VALUE);
         setOffsetAndLength($$, $1, $3);
    }
    ;   

MethodCall /* ============================================================== */
    /* : JustName '(' ArgumentList ')' */
    : QualifiedName '(' ArgumentList ')'
    {
        SymData *var, *method;

        if ($1->u.qname->dot == NULL) {
            /* DOTǷ礵ƤʤQualifiedName i.e. ᥽å̾ */
            spdFree($1->u.qname);
            method = setExpression($1, E_METHOD_CALL,
                                   NULL, NULL, NULL, NULL);
            setLRSort($1, OBJECT_OPERATION);
            setMethodArguments(method, $3);
            setOffsetAndLength(method, $1, $4);
            $$ = method;
        } else {
            /* DOTǷ礵줿QualifiedName, ǸID᥽å̾ */
            /* ǽIDѿ̾QName */
            SymData         *expr, *qname, *dot, *p;

            method = $1->u.qname->second;
            qname = $1->u.qname->first;
            dot = $1->u.qname->dot;

            setComment($$, $1);
            freeSymData($1);
            method = setExpression(method, E_METHOD_CALL, NULL,
                                   NULL, NULL, NULL);
            setLRSort(method, OBJECT_OPERATION);
            setMethodArguments(method, $3); /* argument */
            setOffsetAndLength(method, method, $4);

            for (p = qname; p->u.qname->first != NULL; p = p->u.qname->first) {
                /* Empty. */ 
            }

            var = searchForLocalSymTab(p->str, NULL);
            if (var == NULL) {
                var = searchForMemberSymTab(p->str, NULL);
            }

            if (var != NULL) {
                /* elemѿ(֥), eg. obj.mem1.mem2 */
                expr = qname2Expression(qname, OBJECT_OPERATION);
                $$ = setExpression(NULL, E_OPERATION, dot,
                                   expr, method, NULL);
                setLRSort($$, OBJECT_OPERATION);
                for (p = expr; p->u.expression->first != NULL;
                     p = p->u.expression->first) {
                    /* Empty. */ 
                }
                /* obj ȤƤѿ˴ϢϤ */
                p = setExpression(p, E_REF_VAR, NULL,
                                  var, NULL, NULL);
                setLRSort(p, OBJECT_OPERATION);
                setOffsetAndLength($$, expr, method);
            } else {
                /* qnameϥѥå뤤ϥ饹 */
                expr = setExpression(NULL, E_REF_QNAME, NULL,
                                     qname, NULL ,NULL);
                setLRSort(expr, OBJECT_OPERATION);
                setOffsetAndLength(expr, qname, qname);
                /* QNameλȼ */
                $$ = setExpression(NULL, E_OPERATION, dot,
                                   expr, method, NULL);
                setLRSort($$, OBJECT_OPERATION);
                setOffsetAndLength($$, expr, method);
            }
        }

        lockLRSortR($$);

        setComment($$, $2);
        freeSymData($2);
        setComment($$, $4);
        freeSymData($4);

        jpdDebugFunc2Call(jpdDebug.expression,
                              putExpression, $$, "MethodCall ");
    }
    | PrimaryExpression DOT IDENTIFIER '(' ArgumentList ')'
    {
        SymData *method;

        method = setExpression($3, E_METHOD_CALL, NULL, NULL, NULL, NULL);
        setLRSort(method, OBJECT_OPERATION);
        setOffsetAndLength(method, method, $6);
        setMethodArguments(method, $5);
        setOperator($2, JPD_O_BINARY);
        setLRSortR($1, OBJECT_OPERATION);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, method, NULL);
        setLRSort($$, OBJECT_OPERATION);
        lockLRSortR($$);
        setOffsetAndLength($$, $1, method);

        setComment($$, $4);
        freeSymData($4);
        setComment($$, $6);
        freeSymData($6);
    }
    | SpecialName '(' ArgumentList ')'
    {
        SymData     *constructor;

        if (!SPD_STREQ($1->str, "this") && !SPD_STREQ($1->str, "super")) { 
            spdErrorExit("\"%s\": Invalid MethodCall in file \"%s\".\n",
                         $1->str, yyfilename);
        }
        constructor = setExpression(NULL, E_CONSTRUCTOR_CALL, NULL,
                                    $1, NULL, NULL);
        setLRSort(constructor, NONE);
        setMethodArguments(constructor, $3);
        setOffsetAndLength(constructor, $1, $4);
        $$ = constructor;
    }
    /* Following rule is JDK 1.1. */
    /* !! ʲΥɤϿͤƤʤ׺ƹ͡ */
    | PrimaryExpression DOT SUPER '(' ArgumentList ')' /* Constructor call */
    {
        SymData *constructor = setExpression(NULL, E_CONSTRUCTOR_CALL, NULL,
                                             $3, NULL, NULL);
        setLRSort(constructor, NONE);
        setMethodArguments(constructor, $5);
        setOffsetAndLength(constructor, $3, $6);

        setOperator($2, JPD_O_BINARY);
        setLRSortR($1, OBJECT_OPERATION);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, constructor, NULL);
        setLRSort($$, OBJECT_OPERATION);
        lockLRSortR($$);
        setOffsetAndLength($$, $1, constructor);

        setComment(constructor, $4);
        setComment(constructor, $6);
        freeSymData($4);
        freeSymData($6);
    }
    ;

ArgumentList /* ============================================================ */
    : Argument 
    {
        $$ = $1;
    }
    | ArgumentList COMMA Argument 
    {
        $3->u.expression->next = $1;    /* ϵսˤʤ */
        $$ = $3;

        setComment($$, $2);
        freeSymData($2);
    }
    | /* Empty. */ {
        $$ = NULL;
    }
    ;

Argument /* ================================================================ */
    : Expression 
    {
        $$ = $1;
    }
    ;

ClassInstanceCreationExpression /* ========================================= */
    /*  NewObjectPrefix: NEW TypeName '(' ArgumentList ')' */
    : NewObjectPrefix
    {
        $$ = setConstructorCall($1);
        popMemberSymTab();
    }
    /* following rules are JDK 1.1 */
    | NewObjectPrefix ClassBody
    {
        /*
         * Expr(OP) +-expr_opr-- Operator(new)
         *          +-expr_opd-- Expr(ANON_CLASS)
         *                         | ref_type
         *                       Type(OBJECTIVE)
         *                         | type_class
         *                       Class(ClassName) +-class_argument-- Expr(a1)
         *                                        +-class_argument-- Expr(a2)
         *                                        +- ...-- Body
         */

        $$ = setAnonClass($1, $2);
        popMemberSymTab();
    }
    | PrimaryExpression DOT NewObjectPrefix 
    {
        SymData *constCall = setConstructorCall($3);
        
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, constCall, NULL);
        setOffsetAndLength($$, $1, constCall);
        popMemberSymTab();
    }
    /* ʤQNamePrimaryExpr˴ޤޤΤǤε§פʤΤ */
    /* conflictδطǰʲε§ä                               */
    | QualifiedName DOT NewObjectPrefix 
    {
        SymData *expr = qname2PrimaryNoNewArray($1);
        SymData *constCall = setConstructorCall($3);
        
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, expr, constCall, NULL);
        setOffsetAndLength($$, expr, constCall);
        popMemberSymTab();
    }

    /* | PrimaryExpression NEW TypeName '(' ArgumentList ')' ClassBody */
    | PrimaryExpression DOT NewObjectPrefix  ClassBody
    {
        popMemberSymTab();
        // !!
        spdWarning("Not implemented yet: Instantiate inner class rule.\n");
    }
    ;

/*
 * Anonymous ClassΥѿεɽ뤿ε§
 * ŪJpdNewObjectPrefixˤƤ
 */

NewObjectPrefix /* ========================================================= */
    : NEW TypeName '(' ArgumentList ')'
    {
        $$ = setNewObjectPrefix($1, $2, $3, $4, $5);
    }
    ;

ArrayCreationExpression /* ================================================= */
    : NEW TypeName DimExprs Dims
    {
        SymData         *refType, *anOperator;
        SymData         *lastDim;
        
        anOperator = setOperator($1, JPD_O_NEW_ARRAY);
        refType = setExpression(NULL, E_REF_TYPE, NULL, $2, NULL, NULL);
        setLRSort(refType, R_VALUE);
        setOffsetAndLength(refType, $2, $2);

        /* e.g. new int[2][4][][];
         *   źϽиεսΥꥹ
         *   DimExprs($3)  [4]=>[2] Ȥ Expression(E_DIM_EXPR)Υꥹ
         *   Dims($4)      []=>[] Ȥ Expression(E_DIM_EXPR)Υꥹ
         *    []=>[]=>[4]=>[2] ȤꥹȤˤ
         */
        
        /* Search for the tail of list Dims. */
        for (lastDim = $4; lastDim->u.expression->next != NULL;
             lastDim = lastDim->u.expression->next) {
            /* Empty. */
        }
        lastDim->u.expression->next = $3;
        $$ = setExpression(NULL, E_OPERATION, anOperator,
                           refType, NULL, NULL);
        setLRSort($$, NONE);
        setDimExprs($$, $4);
        setOffsetAndLength($$, anOperator, $4);
        
        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Alloc ArrayType");
        jpdDebugFunc2Call(jpdDebug.expression, putExpression, $$,
                          "Allocated Array");
    }
    | NEW TypeName DimExprs
    {
        SymData         *refType, *anOperator;
        
        anOperator = setOperator($1, JPD_O_NEW_ARRAY);
        refType = setExpression(NULL, E_REF_TYPE, NULL, $2, NULL, NULL);
        setLRSort(refType, NONE);
        setOffsetAndLength(refType, $2, $2);
        $$ = setExpression(NULL, E_OPERATION, anOperator,
                           refType, NULL, NULL);
        setLRSort($$, R_VALUE);
        setDimExprs($$, $3);
        setOffsetAndLength($$, anOperator, $3);

        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Alloc ArrayType");
        jpdDebugFunc2Call(jpdDebug.expression, putExpression, $$,
                          "Allocated Array");
    }
    /*
     * Following rule is JDK 1.1.
     *     | NEW TypeName Dims LeftBrace ArrayInitializers RightBrace
     * '{' '}' ArrayInitializersΤconflictɤᡥ
     */
    | NEW ArrayType '{' ArrayInitializers '}'
    {
        SymData         *refType, *anOperator, *arrayInit;
        
        anOperator = setOperator($1, JPD_O_NEW_ARRAY);
        refType = setExpression(NULL, E_REF_TYPE, NULL, $2, NULL, NULL);
        setLRSort(refType, NONE);
        setOffsetAndLength(refType, $2, $2);

        /** Expr(E_LISTED) --first--> Expr --next--> ... --next--> */
        arrayInit = setExpression(NULL, E_LISTED, NULL, $4, NULL, NULL);
        setLRSort(arrayInit, NONE);
        setOffsetAndLength(arrayInit, $3, $5);

        setComment(arrayInit, $3);
        freeSymData($3);
        setComment(arrayInit, $5);
        freeSymData($5);

        $$ = setExpression(NULL, E_ANON_ARRAY, anOperator,
                           refType, arrayInit, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, anOperator, arrayInit);
    }
    ;

DimExprs /* ================================================================ */
    : DimExpr
    {
        $$ = $1;
    }
    | DimExprs DimExpr
    {
        $2->u.expression->next = $1; /* սΥꥹȤˤƤ */
        $$ = $2;
    }
    ;

DimExpr /* ================================================================= */
    : '[' Expression ']'
    {
        setOperator($1, JPD_O_ALLOC_DIM);
        $$ = setExpression(NULL, E_DIM_EXPR, $1, $2, NULL, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, $1, $3);

        setComment($$, $3);
        freeSymData($3);
    }
    ;

Dims /* ==================================================================== */
    : OP_DIM
    {
        setOperator($1, JPD_O_ALLOC_DIM);
        $$ = setExpression(NULL, E_DIM_EXPR, $1, NULL, NULL, NULL);
        setLRSort($$, NONE);
        setOffsetAndLength($$, $1, $1);
    }
    | Dims OP_DIM
    {
        SymData *dim;
        
        setOperator($2, JPD_O_ALLOC_DIM);
        dim = setExpression(NULL, E_DIM_EXPR, $2, NULL, NULL, NULL);
        setLRSort(dim, NONE);
        setOffsetAndLength(dim, $2, $2);
        
        dim->u.expression->next = $1; /* սΥꥹȤˤƤ */
        $$ = dim;
    }
    ;

PostfixExpression /* ======================================================= */
    : PrimaryExpression 
    {
        $$ = $1;
    }   
    | PostfixExpression PLUS_PLUS 
    {
        setOperator($2, JPD_O_POSTFIX);
        $$ =setExpression(NULL, E_OPERATION, $2, $1, NULL, NULL);
        setLRSortR($1, LR_VALUE);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | PostfixExpression MINUS_MINUS 
    {
        setOperator($2, JPD_O_POSTFIX);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, NULL, NULL);
        setLRSortR($1, LR_VALUE);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    ;

UnaryExpression /* ========================================================= */
    : PLUS_PLUS UnaryExpression 
    {
        setOperator($1, JPD_O_PREFIX);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSortR($2, LR_VALUE);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | MINUS_MINUS UnaryExpression 
    {
        setOperator($1, JPD_O_PREFIX);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSortR($2, LR_VALUE);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | PLUS UnaryExpression 
    {
        setOperator($1, JPD_O_PREFIX);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | MINUS UnaryExpression 
    {
        setOperator($1, JPD_O_PREFIX);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | UnaryExpressionNotPlusMinus 
    {
        $$ = $1;
    }   
    ;

UnaryExpressionNotPlusMinus /* ============================================= */
    : PostfixExpression 
    {
        $$ = $1;
    }   
    | TILDA UnaryExpression 
    {
        setOperator($1, JPD_O_PREFIX);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | EXCLAMATION UnaryExpression 
    {
        setOperator($1, JPD_O_PREFIX);
        $$ = setExpression(NULL, E_OPERATION, $1, $2, NULL, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $2);
    }
    | CastExpression 
    {
        $$ = $1;
    }
    ;

CastExpression /* ========================================================== */
    /* distinguish (p)-q and (p)--q from parened expression */
    : '(' PrimitiveType ')' UnaryExpression
    {
        SymData *refType;

        setType($2, PRIMITIVE_TYPE);
        refType = setExpression(NULL, E_REF_TYPE, NULL, $2, NULL, NULL);
        setLRSort(refType, NONE);
        setOffsetAndLength(refType, $2, $2);
        setOperator($1, JPD_O_CAST);
        $$ = setExpression(NULL, E_OPERATION, $1, $4, refType, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $4);

        setComment($$, $3);
        freeSymData($3);

        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Cast Type");
    }
    | '(' ArrayType ')' UnaryExpression
    {
        SymData *refType;

        refType = setExpression(NULL, E_REF_TYPE, NULL, $2, NULL, NULL);
        setLRSort(refType, NONE);
        setOffsetAndLength(refType, $2, $2);
        setOperator($1, JPD_O_CAST);
        $$ = setExpression(NULL, E_OPERATION, $1, $4, refType, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $4);

        setComment($$, $3);
        freeSymData($3);

        jpdDebugFunc2Call(jpdDebug.type, putType, $2, "Cast Type");
    }
    | '(' Expression ')' UnaryExpressionNotPlusMinus
    {
        SymData *refType;

        /* Expression($2) must be TypeExpression such as ObjectTypeName */
        if ($2->u.expression->sort == E_REF_VAR) {
            /* DOTǷ礵ƤʤQName */
            SymData *type;

            type = $2;
            spdFree($2->u.expression);

            setType(type, OBJECT_TYPE);
            refType = setExpression(NULL, E_REF_TYPE, NULL, type, NULL, NULL);
            setLRSort(refType, NONE);
            setOffsetAndLength(refType, type, type);
            setOperator($1, JPD_O_CAST);
            $$ = setExpression(NULL, E_OPERATION, $1, $4, refType, NULL);
            setLRSort($$, R_VALUE);
            setOffsetAndLength($$, $1, $4);

            jpdDebugFunc2Call(jpdDebug.type, putType, type, "Cast Type");
        }

        /* else if ($2->u.expression->sort == E_REF_QNAME) { */
        /* Commented 1997/12/23. */

        else if ($2->u.expression->sort == E_OPERATION) {
            /* (X.Y.Z)
             * Expr(OP)=Expr(REF_QNAME, X.Y) OP(.) Expr(REF_VAR, Z)
             * ȽƤ
             */
            SymData *qname, *type;
            
            /* Expr(OP) +-- Expr(REF_QNAME) -- QName(X.Y)
             *          +-- Operator(DOT)
             *          +-- Expr(REF_VAR, Z)
             */
            qname = ($2->u.expression->first)->u.expression->first;
            type = $2->u.expression->second;

            /* free Expr(REF_QNAME) */
            freeSymData($2->u.expression->first);
            /* free Operator(DOT) */
            freeSymData($2->u.expression->joperator);
            /* free Expr(REF_VAR, Z)'s union expression  */
            spdFree(type->u.expression);

            setType(type, OBJECT_TYPE);
            type->u.type->qname = qname;
            refType = setExpression(NULL, E_REF_TYPE, NULL, type, NULL, NULL);
            setLRSort(refType, NONE);
            setOffsetAndLength(refType, type, type);
            setOperator($1, JPD_O_CAST);
            $$ = setExpression(NULL, E_OPERATION, $1, $4, refType, NULL);
            setLRSort($$, NONE);
            setOffsetAndLength($$, $1, $4);
            setComment($$, $2);
            
            /* free Expr(OP) */
            freeSymData($2);

            jpdDebugFunc2Call(jpdDebug.type, putType, type, "Cast Type");
        } else {
            spdErrorExit("\"%s\" l.%d: Invalid type expression.\n",
                         yyfilename, yylineno);
        }

        setComment($$, $3);
        freeSymData($3);
    }
    ;

MultiplicativeExpression /* ================================================ */
    : UnaryExpression 
    {
        $$ = $1;
    }   
    | MultiplicativeExpression MultiplicativeOperator UnaryExpression 
    {
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

MultiplicativeOperator /* ================================================== */
    : STAR 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | SLASH 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | PERCENT 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    ;

AdditiveExpression /* ====================================================== */
    : MultiplicativeExpression 
    {
        $$ = $1;
    }
    | AdditiveExpression PLUS MultiplicativeExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    | AdditiveExpression MINUS MultiplicativeExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

ShiftExpression /* ========================================================= */
    : AdditiveExpression 
    {
        $$ = $1;
    }   
    | ShiftExpression ShiftOperator AdditiveExpression 
    {
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

ShiftOperator /* =========================================================== */
    : MORE_MORE 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | LESS_LESS 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | MORE_MORE_MORE 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    ;

RelationalExpression /* ==================================================== */
    : ShiftExpression 
    {
        $$ = $1;
    }   
    | RelationalExpression RelationalOperator ShiftExpression 
    {
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    | RelationalExpression INSTANCEOF TypeName 
    {
        SymData *refType;
        refType = setExpression(NULL, E_REF_TYPE, NULL, $3, NULL, NULL);
        setLRSort(refType, NONE);
        setOffsetAndLength(refType, $3, $3);
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, refType, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, refType);

        jpdDebugFunc2Call(jpdDebug.type, putType, $3, "Instanceof Type");
    }
    ;

RelationalOperator /* ====================================================== */
    : MORE 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | LESS 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | MORE_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | LESS_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    ;

EqualityExpression /* ====================================================== */
    : RelationalExpression 
    {
        $$ = $1;
    }   
    | EqualityExpression EqualityOperator RelationalExpression 
    {
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

EqualityOperator /* ======================================================== */
    : EQUAL_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | EXCLAMATION_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    ;
  
AndExpression /* =========================================================== */
    : EqualityExpression 
    {
        $$ = $1;
    }   
    | AndExpression AND EqualityExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

ExclusiveOrExpression /* =================================================== */
    : AndExpression 
    {
        $$ = $1;
    }   
    | ExclusiveOrExpression HAT AndExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

InclusiveOrExpression /* =================================================== */
    : ExclusiveOrExpression 
    {
        $$ = $1;
    }   
    | InclusiveOrExpression OR ExclusiveOrExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

ConditionalAndExpression /* ================================================ */
    : InclusiveOrExpression 
    {
        $$ = $1;
    }   
    | ConditionalAndExpression AND_AND InclusiveOrExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

ConditionalOrExpression /* ================================================= */
    : ConditionalAndExpression 
    {
        $$ = $1;
    }   
    | ConditionalOrExpression OR_OR ConditionalAndExpression 
    {
        setOperator($2, JPD_O_BINARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

ConditionalExpression /* =================================================== */
    : ConditionalOrExpression 
    {
        $$ = $1;
    }   
    | ConditionalOrExpression QUESTION Expression COLON ConditionalExpression 
    {
        setOperator($2, JPD_O_TERNARY);
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, $5);
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $5);

        setComment($$, $4);
        freeSymData($4);
    }
    ;

AssignmentExpression /* ==================================================== */
    : ConditionalExpression 
    {
        $$ = $1;
    }   
    | UnaryExpression AssignmentOperator AssignmentExpression 
    {
        $$ = setExpression(NULL, E_OPERATION, $2, $1, $3, NULL);
        if (SPD_STREQ($2->str, "=")) {      /* Simple equal (e.g. =). */
            setLRSortR($1, L_VALUE);
        } else {                            /* Complex equal (e.g. +=, *=). */
            setLRSortR($1, LR_VALUE);
        }
        setLRSort($$, R_VALUE);
        setOffsetAndLength($$, $1, $3);
    }
    ;

AssignmentOperator /* ====================================================== */
    : EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | STAR_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | SLASH_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | PERCENT_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | PLUS_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | MINUS_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | MORE_MORE_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | LESS_LESS_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | MORE_MORE_MORE_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | AND_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | HAT_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    | OR_EQUAL 
    {
        $$ = setOperator($1, JPD_O_BINARY);
    }
    ;

ConstantExpression
    : Expression
    {
        /* ConstantExpressionľExpressionȤФʤ */
        $$ = $1;
    }   
    ;

Literal /* = SymData of XXX_LITERAL is set in yylex at "java.l". =========== */
    : CHARACTER_LITERAL
    {
        jpdDebugFunc2Call(jpdDebug.literal, putLiteral, $1, "");
        $$ = $1;
    }
    | STRING_LITERAL 
    {
        jpdDebugFunc2Call(jpdDebug.literal, putLiteral, $1, "");
        $$ = $1;
    }
    | UNICODE_LITERAL 
    {
        jpdDebugFunc2Call(jpdDebug.literal, putLiteral, $1, "");
        $$ = $1;
    }
    | INTEGER_LITERAL 
    {
        jpdDebugFunc2Call(jpdDebug.literal, putLiteral, $1, "");
        $$ = $1;
    }
    | FLOATING_LITERAL 
    {
        jpdDebugFunc2Call(jpdDebug.literal, putLiteral, $1, "");
        $$ = $1;
    }
    ;

%%

int main(int argc, char *argv[])
{
    int             i;
    SpdModelId      JmodelId;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    JmodelId = spdUseModel(&J_model);

    signal(SIGSEGV, segvError);
    str_pool = spdStrPoolInit(JDBIN_DEFAULT_POOL_SIZE);
    /* Clear all flags with 0. */
    memset(&jpdDebug, 0, sizeof(struct JpdDebug));

    parse_option(argc, argv);
    argc -= optind;
    argv += optind;
    if (argc < 1) {
        spdErrorExit("No input file.\n");
    }

    for (i = 0; i < argc; i++) {
        yylineno = 1;
        c_count = 0;
        yyfilename = argv[i];
        initFirstSymTab();

        /* Set jpdFile. */
        jpdFile = getSymData();
        setFile(jpdFile);

        /* Set frejaDir. */
        frejaDir = spdBasename(spdStrdup(yyfilename));
        frejaDir = spdCnvFileExtension(frejaDir, ".java", "");

        /* Do analysis. */
        do_analysis();

        /* Output XML header file. */
        write_XML_header();

        spdFree(frejaDir);
    }

    spdStrPoolFree(str_pool);
    write_japid_version_top();
    return (EXIT_SUCCESS);
}

static void parse_option(int argc, char *argv[])
{
    int     j, ch;

    while ((ch = getopt(argc, argv, MILL_OPTION_STRINGS)) != -1) {
        switch (ch) {
          case 'a':                         /* JDK 1.4 assert on. */
            withAssert = TRUE;
            break;
          case 'b':                         /* Binary mode. */
            spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
            break;
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'h':
            usage();
            exit(EXIT_FAILURE);
            break;
          case 'j':                         /* Output j2jx's ETB. */
            outputJ2jxEtb = TRUE;
            break;
          case 'n':
            for (j = 0; j < strlen(optarg); j++) { 
                switch (optarg[j]) {
                  case 'f':
                    fullpath = FALSE;
                    break;
                  case 'l':
                    linkclass = FALSE;
                    break;
                  case 'a':                         /* JDK 1.4 assert off. */
                    withAssert = FALSE;
                    break;
                  default:
                    usage();
                    exit(EXIT_FAILURE);
                    break;
                }
            }
            break;
          case 'p':
            onlyParsing = TRUE;
            break;
          case 'v':
            verbose = TRUE;
#ifndef YYDEBUG
            yydebug = 1;
#endif
            break;
          case 'D':

#if YYDEBUG 
            for (j = 0; j < strlen(optarg); j++) { 
                switch (optarg[j]) {
                  case 'c':
                    jpdDebug.jclass = 1;
                    break;
                  case 'e':
                    jpdDebug.expression = 1;
                    break;
                  case 'i':
                    jpdDebug.interface = 1;
                    break;
                  case 'l':
                    jpdDebug.literal = 1;
                    break;
                  case 'm':
                    jpdDebug.modifier = 1;
                    break;
                  case 'q':
                    jpdDebug.qname = 1;
                    break;
                  case 'r':
                    jpdDebug.routine = 1;
                    break;
                  case 's':
                    jpdDebug.statement = 1;
                    break;
                  case 't':
                    jpdDebug.type = 1;
                    break;
                  case 'v':
                    jpdDebug.variable = 1;
                    break;
                  case 'y':
                    yydebug = 1;
                    break;
                  case 'A':
                    memset(&jpdDebug, 1, sizeof(struct JpdDebug));
                    /* !! ʤѿ1٤ */
                    break;
                  case 'S':
                    jpdDebug.symTab = 1;
                    break;
                  default:
                    usage();
                    exit(EXIT_FAILURE);
                    break;
                }
            }
#else  /* not YYDEBUG */
            spdWarning("Invalid \"-D\" option. "
                       "Please compile with YYDEBUG macro.\n");    
#endif /* not YYDEBUG */

            break;
          case 'E':
            jpdDebug.etb = 1;
            break;
          case 'S':
            opt_S = TRUE;
            break;
          default:
            usage();
            exit(EXIT_FAILURE);
            break;
        }
    }
}

/* 
 * Write "SDB/.info".
 */

static void write_japid_version_top(void)
{
    SpdInfo     *info;

    info = spdOpenSDBinfo();
    spdSetInfoInt(info, "SapidVersion", SapidVersion);
    spdSetInfoInt(info, "SapidRevision", SapidRevision);
    spdSetInfoInt(info, "SapidPatchLevel", SapidPatchLevel);
    spdSetInfoString(info, "MILL", "TRUE");
    spdCloseInfo(info);
}

static void usage(void)
{
    spdMessage("Usage: %s [-abhpvS][-d dir][-n{afl}][-D{ceilmqrstvAS}] files\n",
               program_name);
}

static void segvError(int x)
{
    spdErrorExit("Signal SEGV is caugt: \"%s\" l.%d c.%d: yytext \"%s\": \"%s\".\n",
                 yyfilename, yylineno, c_count, yytext, yylval->str);
}

static void yyerror(char *msg)
{
    spdError(((*(msg + strlen(msg) - 1) == '\n') ?
              "yyerror(): \"%s\" l.%d: yytext \"%s\": %s" :
              "yyerror(): \"%s\" l.%d: yytext \"%s\": %s\n"),
             yyfilename, yylineno, yytext, msg);
    exit(EXIT_FAILURE);
}

static void write_pkg_info(SymData *qname)
{
    FILE  *fp;

    fp = spdOpenFreja(frejaDir, JAPID_PKG_INFO, SAPID_O_WRITE);
    write_pkg_info_sub(qname, fp);
    fputs("\n", fp);
    spdFclose(fp);
}

static void write_pkg_info_sub(SymData *qname, FILE *fp)
{
    if (qname == NULL) {
        return;
    } if (qname->u.qname->dot == NULL) {
        fputs(qname->str, fp);
    } else {
        write_pkg_info_sub(qname->u.qname->first, fp);
        fputs(".", fp);
        write_pkg_info_sub(qname->u.qname->second, fp);
    }
}

/* Output XML header file. */

static void write_XML_header(void)
{
    FILE        *headerOut;

    headerOut = spdOpenFreja(frejaDir, XML_HEADER_NAME, SAPID_O_WRITE);
    fprintf(headerOut,
            JX_MODEL_XML_HEADER, spdEncodeToStr(jpdFile->u.file->encoding));
    spdFclose(headerOut);
}

static void do_analysis(void)
{     
    static char     dirName[FILENAME_MAX + 1];

    /* ǥ쥯ȥ"SDB/J-MODEL/<className>"롥 */
    sprintf(dirName, "%s/" J_MODEL_TMP_DIR "/%s", spdGetSDBdir(), frejaDir);
    spdMkdirR(dirName);

    yyin = spdFopen(yyfilename, SPD_FOPEN_RB_MODE);
    jpdFile->u.file->encoding = spdDetectEncoding(yyin);
    jdbinEtb = spdOpenFreja(frejaDir, JDBIN_ETB, SAPID_O_WRITE);
    if (outputJ2jxEtb) {
        initExtractFile();
        j2jxEtb = spdOpenFreja(frejaDir, J2JX_ETB, SAPID_O_WRITE);
    }
    yyparse();
    if (onlyParsing == FALSE) {
        initDB();
        writeFile(jpdFile);
    }
    if (outputJ2jxEtb) {
        spdFclose(j2jxEtb);
        cleanUpExtractFile();
    }
    spdFclose(jdbinEtb);
    spdFclose(yyin);
}

/* Fileꤹ롥 */

static SymData *setFile(SymData *file)
{
    static char     cwd[FILENAME_MAX + 1];

    file->offset = 0;
    file->length = spdGetFileLengthByName(yyfilename);

    if (fullpath && yyfilename[0] != SPD_PATH_SEP && yyfilename[1] != ':') {
        /* !! Dirty hack for Windows drive name. */
        getcwd(cwd, FILENAME_MAX);
        strcat(cwd, "/");
        strcat(cwd, yyfilename);
        file->str = spdStrdup(cwd);
    } else {
        file->str = spdStrdup(yyfilename);
    }

#if defined(__CYGWIN__)
    {
        char    *strtmp;

        strtmp = spdMalloc(FILENAME_MAX + 1);
        strtmp[0] = '\0';
        cygwin_conv_to_win32_path(file->str, strtmp);
        file->str = strtmp;
    }
#endif

    file->symSort = SPD_CID_J_FILE;
    file->u.file = spdMalloc(sizeof(JpdFile));
    file->u.file->classOrInterface = NULL;
    file->u.file->definedPackage = NULL;
    file->u.file->importedPackage = NULL;
    return (file);
}
