/* 
 * Program:             $RCSfile: mkFCDView.c,v $  $Revision: 40.11 $
 * 
 * Purpose:             
 *
 * Author:              N.Atsumi  2001/11/29
 *                      S.Yamamoto  2001/11/29 
 *
 * (C) Copyright:       N.Atsumi and S.Yamamoto  2001 - 2002 
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
"$Id: mkFCDView.c,v 40.11 2003/07/08 05:46:59 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelARcompat.h>
#include <Sapid/PmodelAR.h>

#define DEFAULT_ARRAY 16

static int  verbose_flag = 0;
static int  debug_flag = 0;

static SpdConstString prog_name; /* program name */

typedef struct _def_node {
    SpdObjId lib_id;                        /* library function Identifier */
    SpdObjId expr;                          /* library function expression */
    SpdObjId ret_var;                       /* return value expression */
    SpdObjId dep_src;
    SpdObjId node;                          /* define node id */
} DefNode;

typedef struct _cond_node {
    SpdObjId cond_elem;                     /* condition expression */
    SpdObjId cond;
    SpdObjId op;                            /* operator */
    SpdObjId value;                         /* compare value */
    SpdObjId node;                          /* condition node id */
} CondNode;

typedef struct _ctrl_lib {
    SpdObjId library;
    int size;
} CtrlLib;

typedef struct _ref_node {
    SpdObjId lib_id;                        /* library function Identifier */
    SpdObjId expr;                          /* library function expression */
    SpdObjId node;                          /* call node id */
} RefNode;

typedef struct _side_expr {
    SpdObjId right;
    SpdObjId left;
} SideExpr;

/*
 * library function  Identifier Id ֤
 */
static SpdObjId get_lib_id(SpdObjId expr)
{
    SpdObjId    id, decl;
    SpdCursor   *csr;

    id = spdGetARelObj(expr, "ident_ref", "p-id");
    csr = spdGetRelObjInit(id, "decl_ident", "c-id");
    while ((decl = spdGetRelObj(csr)) != SPD_NON_ID) {
        if (smuIsSystemObject(decl)) {
            spdFreeCursor(csr);
            return id;
        }
    }
    spdFreeCursor(csr);
    return SPD_NON_ID;
}

/*
 * expr  cast  expr ֤
 */
static SpdObjId rmCast(SpdObjId expr)
{
    SpdObjId    op;
    int         sort;

    op = spdGetARelObj(expr, "expr_op", "p-id");
    sort = spdGetAttrValIntOfObj(op, "sort");
    if ((sort == OP_L_PAR) || (sort == OP_R_PAR) || (sort == OP_CAST)) {
        expr = spdGetARelObj(expr, "expr_expr", "p-id");
        return rmCast(expr);
    }
    return expr;
}

/*
 * def_node ֤
 */
static DefNode get_def_node(SpdObjId expr, SpdObjId var)
{
    SpdRelArray     expr_rel;
    SpdObjId        id;
    DefNode         node;

    node.expr = SPD_NON_ID;
    expr = rmCast(expr);
    var = rmCast(var);
    expr_rel = spdGetRelSortedArray(expr, "expr_expr", "p-id", NULL,
                                    spdCompareRelOffset);
    if (expr_rel.size) {
        if ((id = get_lib_id(expr_rel.rel[0].objectId)) != SPD_NON_ID) {
            node.expr = expr;
            node.ret_var = var;
            node.lib_id = id;
            node.node = spdCreateObj(SPD_CID_FCDG, SPD_NON_ID);
            if (debug_flag)
                printf("[def_node]\tcreate obj %d\n", node.node);
        }
    }
    spdFreeRelArray(expr_rel);
    return node;
}

/*
 * declaration Ф def_node õ
 * 
 */
static void analyze_decl(SpdObjId block, SpdSet def)
{
    SpdCursor   *blk_decl_csr;
    SpdObjId    decl;

    blk_decl_csr = spdGetRelObjInit(block, "blk_decl", "p-id");
    while ((decl = spdGetRelObj(blk_decl_csr)) != SPD_NON_ID) {
        SpdCursor *decl_decl_csr;
        decl_decl_csr = spdGetRelObjInit(decl, "decl_decl", "p-id");
        while ((decl = spdGetRelObj(decl_decl_csr)) != SPD_NON_ID) {
            SpdObjId opt, ident;
            opt = spdGetARelObj(decl, "decl_opt", "p-id");
            if (opt != SPD_NON_ID) {
                SpdObjId init;
                init = spdGetARelObj(opt, "init_expr", "p-id");
                if (init != SPD_NON_ID) {
                    DefNode node;
                    ident = spdGetARelObj(decl, "decl_ident", "p-id");
                    node = get_def_node(init, ident);
                    node.dep_src = decl;
                    if (node.expr != SPD_NON_ID) {
                        spdAddElem(def, node.expr, &node);
                    }
                }
            }
        }
        spdFreeCursor(decl_decl_csr);
    }
    spdFreeCursor(blk_decl_csr);
}

/*
 * 黻Ҥξüexpression ֤
 */
static SideExpr get_side_expr(SpdObjId expr)
{
    SpdRelArray     array;
    SideExpr        side;

    array = spdGetRelSortedArray(expr, "expr_expr", "p-id", NULL,
                                 spdCompareRelOffset);
    if (array.size != 2) {
        side.right = SPD_NON_ID;
        side.left = SPD_NON_ID;
    } else {
        side.left = rmCast(array.rel[0].objectId);
        side.right = rmCast(array.rel[1].objectId);
    }
    spdFreeRelArray(array);
    return side;
}

/*
 * funccall expression Id ˻
 * def_node 
 */
static DefNode get_def_node_out(SpdObjId expr)
{
    SpdObjId    op;
    int         sort;
    DefNode     node;

    node.expr = SPD_NON_ID;
    expr = rmCast(expr);
    op = spdGetARelObj(expr, "expr_op", "p-id");
    sort = spdGetAttrValIntOfObj(op, "sort");
    if (sort == OP_ASSIGN) {
        SideExpr side;
        side = get_side_expr(expr);
        node = get_def_node(side.right, side.left);
        node.dep_src = expr;
    }
    return node;
}

/*
 * expr Id  constant Id ֤
 * ɬȤϸ¤ʤġޥξϡ
 */
static SpdObjId get_const(SpdObjId expr)
{
    SpdObjId    child, const_id;

    const_id = spdGetARelObj(expr, "const_ref", "p-id");
    if (const_id == SPD_NON_ID) {
        child = spdGetARelObj(expr, "expr_expr", "p-id");
        if (child != SPD_NON_ID)
            return get_const(child);
    } else {
        return const_id;
    }
    return SPD_NON_ID;
}

/*
 * cond_elem ＰΤ objectId ֤
 */
static SpdObjId get_condition(SpdObjId elem)
{
    SpdObjId    out;

    out = spdGetARelObj(elem, "expr_expr", "c-id");
    if (out != SPD_NON_ID)
        return get_condition(out);
    else return elem;
}

/*
 * Ｐ def_node õ
 */
static void analyze_cond(SpdObjId cond, SpdSet cond_set, SpdSet def_set)
{
    SpdObjId    op;
    int         sort;

    op = spdGetARelObj(cond, "expr_op", "p-id");
    sort = spdGetAttrValIntOfObj(op, "sort");
    if ((sort == OP_L_AND) || (sort == OP_L_OR) || (sort == OP_L_PAR) ||
        (sort == OP_R_PAR) || (sort == OP_CAST)) {
        SpdCursor *csr;
        SpdObjId elem;
        /* Ｐʬ */
        csr = spdGetRelObjInit(cond, "expr_expr", "p-id");
        while ((elem = spdGetRelObj(csr)) != SPD_NON_ID) {
            analyze_cond(elem, cond_set, def_set);
        }
        spdFreeCursor(csr);
    } else {
        /* ＰơФ */
        CondNode cond_node;
        DefNode def_node;
        SpdObjId const_id;
        SideExpr side_expr;
        def_node.expr = SPD_NON_ID;
        if ((sort == OP_EQ) || (sort == OP_NE) || (sort == OP_LT) ||
            (sort == OP_LE) || (sort == OP_GT) || (sort == OP_GE)) {
            side_expr = get_side_expr(cond);
            def_node = get_def_node_out(side_expr.right);
            if (def_node.expr != SPD_NON_ID) {
                const_id = get_const(side_expr.left);
            } else {
                def_node = get_def_node_out(side_expr.left);
                if (def_node.expr != SPD_NON_ID) {
                    const_id = get_const(side_expr.right);
                }
            }
        } else if (sort == OP_NOT) {
            const_id = SPD_NON_ID;
        }
        if (def_node.expr != SPD_NON_ID) {
            spdAddElem(def_set, def_node.expr, &def_node);
            cond_node.cond_elem = cond;
            cond_node.cond = get_condition(cond);
            cond_node.value = const_id;
            cond_node.op = op;
            cond_node.node = spdCreateObj(SPD_CID_FCDG, SPD_NON_ID);
            spdCreateRel(SPD_RID_LIBDATA_DEPEND, SPD_NON_ID,
                         def_node.node, cond_node.node);
            if (debug_flag) {
                printf("[def_node]\tcreate obj %d\n", cond_node.node);
                printf("[data_depend]\tcreate rel %d-%d\n", def_node.node,
                       cond_node.node);
            }
            spdAddElem(cond_set, cond_node.cond_elem, &cond_node);
        }
    } 
}

/*
 * basic block  def_node ߤĤ
 */
static void get_def_basic(SpdObjId expr, SpdSet def_set)
{
    SpdObjId    op;
    SideExpr    side;
    DefNode     node;
    int         sort;

    op = spdGetARelObj(expr, "expr_op", "p-id");
    sort = spdGetAttrValIntOfObj(op, "sort");
    if (sort == OP_ASSIGN) {
        side = get_side_expr(expr);
        node = get_def_node(side.right, side.left);
        node.dep_src = expr;
        if (node.expr != SPD_NON_ID) {
            spdAddElem(def_set, node.expr, &node);
        }
    }
}

/*
 * block Id ˻
 * cond_node, def_node դ
 */
static void analyze_block(SpdObjId block, SpdSet cond, SpdSet def)
{
    int sort;

    if (block == SPD_NON_ID) {
        return;
    }
    sort = spdGetAttrValIntOfObj(block, "sort");
    switch (sort) {
        SpdCursor *csr;
        SpdObjId expr;
        SpdRelArray expr_array;
        SpdObjId cond_expr;
      case BLOCK_BASIC:
        csr = spdGetRelObjInit(block, "blk_expr", "p-id");
        while ((expr = spdGetRelObj(csr)) != SPD_NON_ID) {
            get_def_basic(expr, def);
        }
        spdFreeCursor(csr);
        break;
      case BLOCK_IF:
        cond_expr = spdGetARelObj(block, "blk_expr", "p-id");
        analyze_cond(cond_expr, cond, def);
        block = spdGetARelObj(block, "blk_blk", "p-id");
        analyze_block(block, cond, def);
        break;
      case BLOCK_SWITCH:
        cond_expr = spdGetARelObj(block, "blk_expr", "p-id");
        analyze_cond(cond_expr, cond, def);
        block = spdGetARelObj(block, "blk_blk", "p-id");
        analyze_block(block, cond, def);
        break;
      case BLOCK_WHILE:
        cond_expr = spdGetARelObj(block, "blk_expr", "p-id");
        analyze_cond(cond_expr, cond, def);
        block = spdGetARelObj(block, "blk_blk", "p-id");
        analyze_block(block, cond, def);
        break;
      case BLOCK_DO:
        cond_expr = spdGetARelObj(block, "blk_expr", "p-id");
        analyze_cond(cond_expr, cond, def);
        block = spdGetARelObj(block, "blk_blk", "p-id");
        analyze_block(block, cond, def);
        break;
      case BLOCK_FOR:
        expr_array = spdGetRelArray(block, "blk_expr", "p-id", NULL);
        analyze_cond(expr_array.rel[1].objectId, cond, def);
        spdFreeRelArray(expr_array);
        block = spdGetARelObj(block, "blk_blk", "p-id");
        analyze_block(block, cond, def);
        break;
      case BLOCK_COMP:
        analyze_decl(block, def);
        csr = spdGetRelObjInit(block, "blk_blk", "p-id");
        while ((block = spdGetRelObj(csr)) != SPD_NON_ID) {
            analyze_block(block, cond, def);
        }
        spdFreeCursor(csr);
        break;
      default:
        break;
    }
}

/*
 * cast 黻ξ硤¦ expr ֤
 */
static SpdObjId addCast(SpdObjId expr)
{
    SpdObjId    out, op;
    int         sort;

    out = spdGetARelObj(expr, "expr_expr", "c-id");
    op = spdGetARelObj(out, "expr_op", "p-id");
    sort = spdGetAttrValIntOfObj(op, "sort");
    if ((sort == OP_L_PAR) || (sort == OP_R_PAR) || (sort == OP_CAST)) {
        return addCast(out);
    } else {
        return expr;
    }
}

/*
 * key  def_node ν¸ߤ뤫
 */
static DefNode check_def_node(SpdObjId key, SpdSet set)
{
    DefNode     *tmp, node;

    tmp = (DefNode *)spdSearchElem(set, key);
    if (tmp == NULL) {
        node.expr = SPD_NON_ID;
    } else {
        node = *tmp;
    }
    return node;
}

/*
 * def_node, cond_node, ref_node ν˻
 * ǡ¸ϤԤ
 */
static void analyze_depend(const SpdSet def_set, SpdSet cond_set,
                           SpdSet ref_set)
{
    DefNode     node;
    int         i;

    for (i = 0; i < def_set->elemNum; i++) {
        node = *(DefNode *)spdGetAElem(def_set, NULL);
        if (node.ret_var != SPD_NON_ID) {
            SpdCursor *csr;
            SpdObjId dep, dep_org;
            SpdRelArray expr_rel;
            /* def_node ¸طˤǤ */
            csr = spdGetRelObjInit(node.dep_src, "data_dependence", "p-id");
            while ((dep = spdGetRelObj(csr)) != SPD_NON_ID) {
                /* argument of library function? */
                SpdObjId out, op;
                int sort;
                dep_org = dep;
                dep = addCast(dep);
                /* ⤷dep ʤ out ؿƽ */
                out = spdGetARelObj(dep, "expr_expr", "c-id");
                if (out == SPD_NON_ID) {
                    continue;
                }
                expr_rel = spdGetRelSortedArray(out, "expr_expr", "p-id",
                                                NULL, spdCompareRelOffset);
                if (expr_rel.size) {
                    SpdObjId id;
                    /* expr_rel[0] ϴؿƽФδؿ̾ʬؤ expr */
                    id = get_lib_id(expr_rel.rel[0].objectId);
                    if (id != SPD_NON_ID) {
                        DefNode target_node;
                        /* dep is an argument of library function */
                        target_node = check_def_node(out, def_set);
                        if (target_node.node == node.node) continue;
                        if (target_node.expr != SPD_NON_ID) {
                            /* expr_rel --data depend--> target_node */
                            spdCreateRel(SPD_RID_LIBDATA_DEPEND, SPD_NON_ID,
                                         node.node, target_node.node);
                            if (debug_flag)
                                printf("[data_depend]\tcreate rel %d-%d\n",
                                       node.node, target_node.node);
                        } else {
                            RefNode ref_node;
                            ref_node.expr = out;
                            ref_node.lib_id = id;
                            /* node --data depend--> id */
                            ref_node.node = spdCreateObj(SPD_CID_FCDG,
                                                         SPD_NON_ID);
                            if (debug_flag)
                                printf("[ref_node]\tcreate obj %d\n",
                                       ref_node.node);
                            spdCreateRel(SPD_RID_LIBDATA_DEPEND, SPD_NON_ID,
                                         node.node, ref_node.node);
                            if (debug_flag)
                                printf("[data_depend]\tcreate rel %d-%d\n",
                                       node.node, ref_node.node);
                            spdAddElem(ref_set, ref_node.expr, &ref_node);
                        }
                    }
                }
                spdFreeRelArray(expr_rel);
                /* condition? */
                /* operator is OP_EQ, OP_NE, OP_GT, OP_GE, OP_LE, OP_LT */
                op = spdGetARelObj(out, "expr_op", "p-id");
                sort = spdGetAttrValIntOfObj(op, "sort");
                if ((sort == OP_EQ) || (sort == OP_NE) || (sort == OP_GT) ||
                    (sort == OP_GE) || (sort == OP_LT) || (sort == OP_LE)) {
                    /* condition */
                    CondNode cond_node;
                    SideExpr side;
                    side = get_side_expr(out);
                    cond_node.op = op;
                    cond_node.cond_elem = out;
                    cond_node.cond = get_condition(out);
                    cond_node.node = spdCreateObj(SPD_CID_FCDG, SPD_NON_ID);
                    if (debug_flag)
                        printf("[cond_node]\tcreate obj %d\n", cond_node.node);
                    if (side.right == dep_org) {
                        cond_node.value = get_const(side.left);
                    } else if (side.left == dep_org) {
                        cond_node.value = get_const(side.right);
                    } else {
                        cond_node.value = SPD_NON_ID;
                        spdWarning("Can't specify depend element[%d].\n", dep);
                    }
                    spdCreateRel(SPD_RID_LIBDATA_DEPEND, SPD_NON_ID,
                                 node.node, cond_node.node);
                    if (debug_flag)
                        printf("[data_depend]\tcreate rel %d-%d\n",
                               node.node, cond_node.node);
                    spdAddElem(cond_set, cond_node.cond_elem, &cond_node);
                }
            }
            spdFreeCursor(csr);
        }
    }
}

/*
 * src, ref_node ν˻
 * src ˴ޤޤ node  set_include ˳Ǽ
 */
static void check_include_ref(SpdObjId src, SpdSet set,
                              SpdSet set_include)
{
    int     i, src_len;
    SpdOcc  src_occ;

    src_occ = spdGetAnOffsetOfObj(src);
    src_len = spdGetLength(src);
    for (i = 0; i < set->elemNum; i++) {
        RefNode node;
        SpdOcc node_occ;
        int node_len;
        node = *(RefNode *)spdGetAElem(set, NULL);
        node_occ = spdGetAnOffsetOfObj(node.expr);
        node_len = spdGetLength(node.expr);
        if ((src_occ.offset <= node_occ.offset) &&
            ((src_occ.offset + src_len) >= (node_occ.offset + node_len))) {
            spdAddElem(set_include, node.node, &node.node);
        }
    }
}

/*
 * src, def_node ν˻
 * src ˴ޤޤ node  set_include ˳Ǽ
 */
static void check_include_def(SpdObjId src, const SpdSet set,
                              SpdSet set_include)
{
    int     i, src_len;
    SpdOcc  src_occ;

    src_occ = spdGetAnOffsetOfObj(src);
    src_len = spdGetLength(src);
    for (i = 0; i < set->elemNum; i++) {
        DefNode node;
        SpdOcc node_occ;
        int node_len;
        node = *(DefNode *)spdGetAElem(set, NULL);
        node_occ = spdGetAnOffsetOfObj(node.expr);
        node_len = spdGetLength(node.expr);
        if ((src_occ.offset <= node_occ.offset) &&
            ((src_occ.offset + src_len) >= (node_occ.offset + node_len))) {
            spdAddElem(set_include, node.node, &node.node);
        }
    }
}

/*
 * src, cond_node ν˻
 * src ˴ޤޤ node  set_include ˳Ǽ
 */
static void check_include_cond(SpdObjId src, SpdSet set,
                               SpdSet set_include)
{
    int     i, src_len;
    SpdOcc  src_occ;

    src_occ = spdGetAnOffsetOfObj(src);
    src_len = spdGetLength(src);
    for (i = 0; i < set->elemNum; i++) {
        CondNode node;
        SpdOcc node_occ;
        int node_len;
        node = *(CondNode *)spdGetAElem(set, NULL);
        node_occ = spdGetAnOffsetOfObj(node.cond_elem);
        node_len = spdGetLength(node.cond_elem);
        if ((src_occ.offset <= node_occ.offset) &&
            ((src_occ.offset + src_len) >= (node_occ.offset + node_len))) {
            spdAddElem(set_include, node.node, &node.node);
        }
    }
}

/*
 * def_node, cond_node, ref_node ν˻
 * 
 */
static void create_ctrl_rel(SpdSet cond_set, SpdSet def_set, SpdSet ref_set)
{
    CondNode    cond;
    int         i;

    for (i = 0; i < cond_set->elemNum; i++) {
        SpdObjId cond_blk, block;
        SpdCursor *blk_csr;
        cond = *(CondNode *)spdGetAElem(cond_set, NULL);
        cond_blk = spdGetARelObj(cond.cond, "blk_expr", "c-id");
        blk_csr = spdGetRelObjInit(cond_blk, "blk_blk", "p-id");
        while ((block = spdGetRelObj(blk_csr)) != SPD_NON_ID) {
            SpdSet ctrl_set;
            int j;
            ctrl_set = spdCreateSet(sizeof(SpdObjId));
            /* block ˴ޤޤ cond_node Ĵ٤ */
            check_include_cond(block, cond_set, ctrl_set);
            /* block ˴ޤޤ def_node Ĵ٤ */
            check_include_def(block, def_set, ctrl_set);
            /* block ˴ޤޤ ref_node Ĵ٤ */
            check_include_ref(block, ref_set, ctrl_set);
            /* Ϣ libctrl-depend  */
            for (j = 0; j < ctrl_set->elemNum; j++) {
                SpdObjId target;
                target = *(SpdObjId *)spdGetAElem(ctrl_set, NULL);
                spdCreateRel(SPD_RID_LIBCTRL_DEPEND, SPD_NON_ID, cond.node,
                             target);
                if (debug_flag)
                    printf("[ctrl_depend]\tcreate rel %d-%d\n",
                           cond.node, target);
            }
            spdFreeSet(ctrl_set);
        }
        spdFreeCursor(blk_csr);
    }
}

/*
 * cond_node ν˻
 * cond_node ˴طϢĥ
 */
static void create_cond_rel(SpdSet set)
{
    int     i;
    
    for (i = 0; i < set->elemNum; i++) {
        CondNode node;
        node = *(CondNode *)spdGetAElem(set, NULL);
        spdSetAttrIntValByName(node.node, "sort", FCDG_CONDITION);
        spdCreateRel(SPD_RID_FCDG_EXPR, SPD_NON_ID, node.node, node.cond_elem);
        spdCreateRel(SPD_RID_COND_OP, SPD_NON_ID, node.node, node.op);
        spdCreateRel(SPD_RID_COND_VAL, SPD_NON_ID, node.node, node.value);
        if (debug_flag)
            printf("set attribute(sort) \"condition\" %d\n"
                   "[fcdg_expr]\tcreate rel %d-%d\n"
                   "[cond_op]\t create rel %d-%d\n"
                   "[cond_val]\t create rel %d-%d\n",
                   node.node,
                   node.node, node.cond_elem,
                   node.node, node.op,
                   node.node, node.value);
    }
}

/*
 * ref_node νˤ
 * ref_node ˴طϢĥ
 */
static void create_ref_rel(SpdSet set)
{
    int     i;

    for (i = 0; i < set->elemNum; i++) {
        RefNode node;
        node = *(RefNode *)spdGetAElem(set, NULL);
        spdSetAttrIntValByName(node.node, "sort", FCDG_LIBREF);
        spdCreateRel(SPD_RID_FCDG_EXPR, SPD_NON_ID, node.node, node.expr);
        spdCreateRel(SPD_RID_FCDG_LIB, SPD_NON_ID, node.node, node.lib_id);
        if (debug_flag)
            printf("set attribute(sort) \"libref\" %d\n"
                   "[fcdg_expr]\tcreate rel %d-%d\n"
                   "[fcdg_lib]\t create rel %d-%d\n",
                   node.node,
                   node.node, node.expr,
                   node.node, node.lib_id);
    }
}

/*
 * def_node ν礪Ӵؿdeclaration Idˤ
 * def_node ˴طϢĥ
 */
static void create_def_rel(SpdSet set, SpdObjId func)
{
    int     i;
    for (i = 0; i < set->elemNum; i++) {
        DefNode     node;
        SpdObjId    dep, ctrl;
        
        node = *(DefNode *)spdGetAElem(set, NULL);
        spdCreateRel(SPD_RID_FCDG_EXPR, SPD_NON_ID, node.node, node.expr);
        spdCreateRel(SPD_RID_FCDG_LIB, SPD_NON_ID, node.node, node.lib_id);
        spdCreateRel(SPD_RID_DEF_VAR, SPD_NON_ID, node.node, node.ret_var);
        if (debug_flag)
            printf("[fcdg_expr]\tcreate rel %d-%d\n"
                   "[fcdg_lib]\tcreate rel %d-%d\n"
                   "[def_var]\tcreate rel %d-%d\n",
                   node.node, node.expr,
                   node.node, node.lib_id,
                   node.node, node.ret_var);
        dep = spdGetARelObj(node.node, "libdata_depend", "c-id");
        ctrl = spdGetARelObj(node.node, "libctrl_depend", "c-id");
        if ((dep == SPD_NON_ID) && (ctrl == SPD_NON_ID)) {
            spdSetAttrIntValByName(node.node, "sort", FCDG_LIBDEFREF);
            spdCreateRel(SPD_RID_FUNC_FCDG, SPD_NON_ID, func, node.node);
            if (debug_flag)
                printf("set attribute(sort) \"libdefref\" %d\n"
                       "[func_fcdg]\tcreate rel %d-%d\n",
                       node.node,
                       func, node.node);
        } else {
            spdSetAttrIntValByName(node.node, "sort", FCDG_LIBDEF);
            if (debug_flag)
                printf("set attribute(sort) \"libdef\" %d\n",
                       node.node);
        }
    }
}

/*
 * declaration of function ˻
 * FCDG
 */
static void analyze_func(SpdObjId func)
{
    SpdObjId    declr, opt, body;
    SpdSet      cond_set, def_set, ref_set;

    if (verbose_flag) {
        spdMessage("Making FCD for function[%d].\n", func);
    }
    declr = spdGetARelObj(func, "decl_decl", "p-id");
    opt = spdGetARelObj(declr, "decl_opt", "p-id");
    body = spdGetARelObj(opt, "func_body", "p-id");

    /* cond_node, ref_node, def_node Ǽ SpdSet Ѱ */
    cond_set = spdCreateSet(sizeof(CondNode));
    def_set = spdCreateSet(sizeof(DefNode));
    ref_set = spdCreateSet(sizeof(RefNode));

    /* block ΥơȥȤ(def_set, cond_set ) */
    analyze_block(body, cond_set, def_set);
    /* def_set ǡ¸ */
    analyze_depend(def_set, cond_set, ref_set);
    /* cond_set ¸ */
    create_ctrl_rel(cond_set, def_set, ref_set);
    /* cond_node, def_node, ref_node γƴϢ */
    create_cond_rel(cond_set);
    create_def_rel(def_set, func);
    create_ref_rel(ref_set);
    spdFreeSet(def_set);
    spdFreeSet(cond_set);
    spdFreeSet(ref_set);
}

/*
 * program Id Ƥ file Id 
 * file Ƥδؿ(analyze_func)
 */
static void analyze_prog(SpdObjId prog)
{
    SpdCursor   *file_csr;
    SpdObjId    file;

    file_csr = spdGetRelObjInit(prog, "prog_file", "p-id");
    while ((file = spdGetRelObj(file_csr)) != SPD_NON_ID) {
        SpdCursor   *decl_csr;
        SpdObjId    decl;
        SpdSet      func_set;
        int         i;
        
        func_set = spdCreateSet(sizeof(SpdObjId));
        decl_csr = spdGetRelObjInit(file, "file_decl", "p-id");
        while ((decl = spdGetRelObj(decl_csr)) != SPD_NON_ID) {
            int     sort;

            sort = spdGetAttrIntValByName(decl, "sort");
            if (sort == DECL_FUNCDEF) {
                spdAddElem(func_set, decl, &decl);
            }
        }
        spdFreeCursor(decl_csr);
        for (i = 0; i < func_set->elemNum; i++) {
            SpdObjId    func;

            func = *(SpdObjId *)spdGetAElem(func_set, NULL);
            analyze_func(func);
        }
        spdFreeSet(func_set);
    }
    spdFreeCursor(file_csr);
}

void printUsage(char *argv[])
{
    printf("NAME\n"
           "\t%s\n"
           "SYNOPSIS\n"
           "\t%s [options]\n"
           "DESCRIPTION\n"
           "\t-d SDB directory\t: Directory for SDB (\"SDB\").\n"
           "\t-h\t\t\t: Show this help.\n"
           "\t-v\t\t\t: Print verbose message.\n"
           "\t-D\t\t\t: Debug mode.\n"
           "VERSION\n"
           "\t%d.%d.%d\n",
           argv[0], argv[0], SapidVersion, SapidRevision, SapidPatchLevel);
    exit(EXIT_SUCCESS);
}

void opt_analyze(int argc, char *argv[])
{
    int     c;

    while ((c = getopt(argc, argv, "d:hvD")) != EOF) {
        switch (c) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'h':
            printUsage(argv);
            break;
          case 'v':                         /* Print verbose messages. */
            verbose_flag ^= 1;
            break;
          case 'D':
            debug_flag ^= 1;                /* Debug mode. */
            break;
          default:
            printUsage(argv);
            spdErrorExit("'%s': Invalid option.\n", argv[optind-1]);
            break;
        }
    }
}

int main(int argc, char *argv[])
{
    SpdModelId fl_id, i_id, p_id, fcd_id;
    SpdDBId db_id;
    SpdCursor *prog_csr;
    SpdObjId prog;

    spdInitMemProf();
    prog_name = spdBasename(argv[0]);
    spdInitialize(prog_name);
    opt_analyze(argc, argv);

    i_id = spdUseModel(&I_model);
    p_id = spdUseModel(&P_model);
    fl_id = spdUseModel(&Flow_view);
    fcd_id = spdUseModel(&FCD_view);
    db_id = spdOpenSDB();
    spdReadDB(i_id, db_id);
    spdReadDB(p_id, db_id);
    spdReadDB(fl_id, db_id);
    spdReadDB(fcd_id, db_id);

    prog_csr = spdGetObjIdInit("program", NULL);
    while ((prog = spdGetObjId(prog_csr)) != SPD_NON_ID) {
        analyze_prog(prog);
    }

    spdFreeCursor(prog_csr);
    spdMessage("Writing FCD database.\n");
    spdCommitDB(fcd_id, db_id);
    spdCloseSDB(db_id);

    return EXIT_SUCCESS;
}
