/* 
 * Program:             $RCSfile: calcReach.c,v $  $Revision: 40.9 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/12/11
 *                      S.Yamamoto  2001/12/11
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: calcReach.c,v 40.9 2003/06/04 16:59:29 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkDD.h"

static preObjListNode *createPOLNode(void);
static SpdBoolean searchPOLNode(SpdObjId objId);
static void addPOLNode(SpdObjId objId);
static void freePOList(void);
static SpdObjId getPreDeclId(SpdObjId declId);
static SpdObjId getLastDeclId(SpdObjId blkEntryId);
static void makePOList(SpdObjId objId);

extern int debugPrintFlag;
extern int debugPrintFlag2;

static preObjListNode *preObjListHead = NULL; /* ääꤹꥹȡ*/
static preObjListNode *preObjListTail = NULL;

/* preObjListؿ */

static preObjListNode *createPOLNode(void)
{
    return (preObjListNode *)spdMalloc(sizeof(preObjListNode));
}

static SpdBoolean searchPOLNode(SpdObjId objId)
{
    preObjListNode *p;

    p = preObjListHead;
    while (p != NULL) {
        if (p->objId == objId) return TRUE;
        p = p->next;
    }
    return FALSE;
}

static void addPOLNode(SpdObjId objId)
{
    preObjListNode *p;
  
    if (searchPOLNode(objId) == TRUE) {
        return;
    }

    p = createPOLNode();
    p->objId = objId;
    p->next = NULL;
    if (preObjListHead == NULL) {
        preObjListHead = p;
        preObjListTail = p;
    } else {
        preObjListTail->next = p;
        preObjListTail = p;
    }
 
    /* if (debugPrintFlag2) fprintf(stderr, "addPOLNode: %d\n", objId); */
}

static void freePOList(void)
{
    preObjListNode *p, *next;

    p = preObjListHead;
    while (p != NULL) {
        /* 
         * if (debugPrintFlag2) {
         *     fprintf(stderr, "freePOList: %d\n", p->objId);
         * }
         */
        next = p->next;
        spdFree(p);
        p = next;
    }
    preObjListHead = NULL;
    preObjListTail = NULL;
}

/******************************************************************************/

static SpdObjId getPreDeclId(SpdObjId declId)
{
    int i;
    SpdRelArray declBuf;
    SpdObjId blkId, preDeclId;

    blkId = spdGetARelObj(declId, "blk_decl", "children_declaration_id");
    if (blkId == SPD_NON_ID) {
        /* ؿΰdeclarationȡ롥 */
        return SPD_NON_ID;
    }
    declBuf = spdGetRelSortedArray(blkId, "blk_decl", "parent_block_id",
                                   NULL, spdCompareRelOffset);
    for (i = 0; i < declBuf.size; i++) {
        if (declBuf.rel[i].objectId == declId) {
            if (i == 0) {
                preDeclId = SPD_NON_ID;
            } else {
                preDeclId = declBuf.rel[i - 1].objectId;
            }
            break;
        }
    }
    spdFreeRelArray(declBuf);
    return preDeclId;
}

static SpdObjId getLastDeclId(SpdObjId blkEntryId)
{
    SpdRelArray declBuf;
    SpdObjId blkId, declId;

    blkId = spdGetARelObj(blkEntryId, "blk_sdanode", "sdanodeId");
    declBuf = spdGetRelSortedArray(blkId, "blk_decl", "parent_block_id",
                                   NULL, spdCompareRelOffset);
    if (declBuf.size == 0) {
        declId = SPD_NON_ID;
    } else {
        declId = declBuf.rel[declBuf.size - 1].objectId;
    }
    spdFreeRelArray(declBuf);
    return declId;
}

/* preObjList롥*/

static void makePOList(SpdObjId objId)
{
    SpdCursor *cursor;
    SpdObjId preObjId, declId;

    if (spdGetClassId(objId) == SPD_CID_DECLARATION) {
        declId = getPreDeclId(objId);
        if (declId != SPD_NON_ID) {
            addPOLNode(declId);
        } else {
            SpdObjId blkId, blkEntryId;
            blkId = spdGetARelObj(objId, "blk_decl", "children_declaration_id");
            if (blkId != SPD_NON_ID) {
                blkEntryId = getFVNode(blkId, SPD_CID_BLK_ENTRY);
                makePOList(blkEntryId);
            }
        }
        return;
    }

    cursor = spdGetRelObjInit(objId, "next_node", "nextId");
    while ((preObjId = spdGetRelObj(cursor)) != SPD_NON_ID) {
        if (spdGetClassId(preObjId) == SPD_CID_EXPRESSION) {
            addPOLNode(preObjId);
        } else if (spdGetClassId(preObjId) == SPD_CID_BLK_ENTRY) {
            /* ֥åμưѿĴ٤ */
            declId = getLastDeclId(preObjId);
            if (declId != SPD_NON_ID) {
                addPOLNode(declId);
            } else {
                makePOList(preObjId);
            }
        } else {
            makePOList(preObjId);
        }
    }
    spdFreeCursor(cursor);
}

/******************************************************************************/

/* ǥХåѴؿ */

void printReach(void)
{
    int i;
    topLevelListNode *curNode;
    subExprListNode *curSubNode;

    curNode = getTLLHead();
    while (curNode != NULL) {
        fprintf(stderr, "tLLNode %d   %d\n", curNode->objId, curNode->reach[0]);
        curSubNode = getSLHead();
        i = 0;
        while (curSubNode != NULL) {
            if (chkBitVector(curNode->reach, i) == TRUE) {
                fprintf(stderr, "   reach(%d): %d\n", i, curSubNode->objId);
            }
            curSubNode = curSubNode->next;
            i++;
        }
        curNode = curNode->next;
    }
}

void calcReach(void)
{
    topLevelListNode *curNode;
    bitVector *newReach, *newReachSum;
    SpdBoolean changed;

    newReach = (bitVector *)createBitVector();
    newReachSum = (bitVector *)createBitVector();

    do {
        changed = FALSE;

        /* ȥåץ٥뼰ΥΡɤФƽԤʤ*/
        curNode = getTLLHead();
        while (curNode != NULL) {
            preObjListNode *curPOLNode;
            /*printf("calcReach():curNode->objId %d\n", curNode->objId);*/

            makePOList(curNode->objId);     /* preObjListϡ롥 */
            resetBitVector(newReachSum);
            curPOLNode = preObjListHead;
            while (curPOLNode != NULL) {
                topLevelListNode *tLLNode;

                /* Υȥåץ٥뼰˼Ф*/
                tLLNode = (topLevelListNode *)searchTLLNode(curPOLNode->objId);
                if (tLLNode == NULL) {
                    /* դ餿ɤäˤ뼰
                     * (ȥåץ٥뼰ꥹȤϿƤʤ)̵롥
                     */
                    if (debugPrintFlag2) {
                        fprintf(stderr,
                                "TLLNode not found. curPOLNode->objId %d\n",
                                curPOLNode->objId);
                    }
                    curPOLNode = curPOLNode->next;
                    continue;
                }
                copyBitVector(newReach, tLLNode->kill);
                notBitVector(newReach);
                andBitVector(newReach, tLLNode->reach);
                orBitVector(newReach, tLLNode->def);
                orBitVector(newReachSum, newReach);
                curPOLNode = curPOLNode->next;
            }
            freePOList();                   /* preObjListϡ */

            if (isEqualBitVector(newReachSum, curNode->reach) == FALSE) {
                changed = TRUE;
                copyBitVector(curNode->reach, newReachSum);
            }
            curNode = curNode->next;
        }
    } while (changed == TRUE);

    spdFree(newReach);
    spdFree(newReachSum);
}
