/* 
 * Program:             $RCSfile: mkCD.c,v $  $Revision: 42.4 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/12/11
 *                      S.Yamamoto  2001/12/11
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkCD.c,v 42.4 2003/05/29 04:19:36 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelAR.h>
#include "mkCD.h"

static void printBlkSort(int blk_sort);
static void printEdgeSort(int edge_sort);

static void makeBlkNode(SpdObjId blk_id, int blk_sort, 
                SpdObjId *blkEntry_id, SpdObjId  *blkExit_id);
static void pushBlkStack(SpdObjId blk_id, int blk_sort,
                 SpdObjId blkEntry_id, SpdObjId blkExit_id);
static void popBlkStack(void);

static void makeBlkBreakRel(int cur_blkStack_n);
static void makeBlkContinueRel(int  cur_blkStack_n);
static void makeBlkReturnRel(int cur_blkStack_n);
static void makeBlkGoAwayRel(SpdObjId expr_id, int cur_blkStack_n);
static void makeNextRel(SpdObjId next_id, int edge_sort);

static void basicblkTop(SpdObjId blk_id);
static SpdObjId compblkTop(SpdObjId blk_id);
static void blkloop(SpdObjId blk_id);
static SpdObjId ifblkTop(SpdObjId blk_id);
static SpdObjId switchblkTop(SpdObjId blk_id);
static SpdObjId forblkTop(SpdObjId blk_id);
static SpdObjId whileblkTop(SpdObjId blk_id);
static SpdObjId doblkTop(SpdObjId blk_id);
static SpdObjId blkSwitch(SpdObjId blk_id);

static void funcTop(SpdObjId funcDecl_id);
static void file2func(SpdObjId file_id);
static void prog2file(SpdObjId prog_id);
static void opt_analyze(int argc, char *argv[]);


/*  Ϣnext_node ΥåΥȤꤹ뤿Υޥ */
#define NEXT_EDGE_IS_TRUE_EDGE          0
#define NEXT_EDGE_IS_FALSE_EDGE         1
#define NEXT_EDGE_IS_FREE               2

/* COMP֥åSWITCH֥åΤΤǤ뤳Ȥ򼨤ޥ */
#define NEXT_COMP_IS_SWITCH_COMP        0
#define NEXT_COMP_IS_NOT_SWITCH_COMP    1

/* ֥åΥåΥ */
#define BLK_STACK_MAX                   200

static SpdConstString program_name;                  /* This program name. */

static int      debugPrintFlag  = 0;  /* 0ȥǥХåѤɽФ */
static int      debugPrintFlag2 = 0;  /* ޤɬפʤʾФ */

static int      branchCount  = 0;     /* ifʸֹ̤桥 */

/* ΥåΥȤλꡥ */
static int      nextEdgeSort   = NEXT_EDGE_IS_FREE;

/* SWITCH֥åCOMPȽѡ */
static int      nextCompBelong = NEXT_COMP_IS_NOT_SWITCH_COMP;

static SpdObjId lastId;  /* ҤȤΥΡɤidɽѿ */

/* ֥åΥå */
/* [0]ϴؿȽиΥ٥롥 */
static SpdFvStackNode blkStack[BLK_STACK_MAX];

static int      blkStack_n = -1;

/******************** ǥХå˻Ȥؿ *********************************/

static void printBlkSort(int blk_sort)
{
    static char *blk_name[] = {
        "SPD_FV_B_FUNC", "SPD_FV_B_COMP", "SPD_FV_B_IF", "SPD_FV_B_SWITCH",
        "SPD_FV_B_FOR", "SPD_FV_B_WHILE", "SPD_FV_B_DO"
    };

    switch (blk_sort) {
      case SPD_FV_B_FUNC:
      case SPD_FV_B_COMP:
      case SPD_FV_B_IF:
      case SPD_FV_B_SWITCH:
      case SPD_FV_B_FOR:
      case SPD_FV_B_WHILE:
      case SPD_FV_B_DO:
        fprintf(stderr, "%s", blk_name[blk_sort]);
        break;
      default:
        spdErrorExit("printBlkSort()\n");
    }
}

static void printEdgeSort(int edge_sort)
{
    static char *edge_name[] = {
        "SPD_FV_E_NORMAL", "SPD_FV_E_BRANCH_TRUE",
        "SPD_FV_E_BRANCH_FALSE", "SPD_FV_E_LOOPBACK",
        "SPD_FV_E_SWITCH_CASE", "SPD_FV_E_SWITCH_DEFAULT",
        "SPD_FV_E_SWITCH_NOMATCH"
    };

    switch (edge_sort) {
      case SPD_FV_E_NORMAL:
      case SPD_FV_E_BRANCH_TRUE:
      case SPD_FV_E_BRANCH_FALSE:
      case SPD_FV_E_LOOPBACK:
      case SPD_FV_E_SWITCH_CASE:
      case SPD_FV_E_SWITCH_DEFAULT:
      case SPD_FV_E_SWITCH_NOMATCH:
        fprintf(stderr, "%s", edge_name[edge_sort]);
        break;
      default:
        spdErrorExit("printEdgeSort()\n");
    }
}

/******************** ƥ֥åѴؿ *************************************/

/*
 * blk_idΥ֥åβΡ(blkEntry)ȡ۽иΡ(blkExit)
 * äơid֤
 */

static void makeBlkNode(SpdObjId blk_id, int blk_sort,
                        SpdObjId *blkEntry_id, SpdObjId *blkExit_id)
{
    SpdObjId new_blkEntry_id, new_blkExit_id;
    SpdRelId new_rel_id;

    /*
     * Ϳ줿 blk_id  blkEntry blkExit Ρɤꡤ
     * ΥΡɤblk_id ȤδϢϤ롥
     */    
    new_blkEntry_id = spdCreateObj(SPD_CID_BLK_ENTRY, SPD_NON_ID);
    spdSetAttrIntValByName(new_blkEntry_id, "sort", blk_sort);

    if (debugPrintFlag) {
        fprintf(stderr, "@ Make object \"blkEntry\" \t(id %d)\n", new_blkEntry_id);
        fprintf(stderr, "  sort : \t"); 
        printBlkSort(blk_sort);
        fprintf(stderr, "\n");
    }
    *blkEntry_id = new_blkEntry_id;

    new_rel_id = spdCreateRel(SPD_RID_BLK_SDANODE, SPD_NON_ID,
                              blk_id, new_blkEntry_id);

    if (debugPrintFlag) {
        fprintf(stderr, "# Make relation \"blk_sdanode\" \t(id %d)\n", new_rel_id);
        fprintf(stderr, "  \"blkId\" \t(id %d)\n", blk_id);
        fprintf(stderr, "  \"sdanodeId\" \t(id %d)\n\n", new_blkEntry_id);
    }

    new_blkExit_id = spdCreateObj(SPD_CID_BLK_EXIT, SPD_NON_ID);
    spdSetAttrIntValByName(new_blkExit_id, "sort", blk_sort);
 
    if (debugPrintFlag) {
        fprintf(stderr, "@ Make object \"blkExit\" \t(id %d)\n", new_blkExit_id);
        fprintf(stderr, "  sort : \t");
        printBlkSort(blk_sort);
        fprintf(stderr, "\n");
    }
    *blkExit_id = new_blkExit_id;

    new_rel_id = spdCreateRel(SPD_RID_BLK_SDANODE, SPD_NON_ID,
                              blk_id, new_blkExit_id);

    if (debugPrintFlag) {
        fprintf(stderr, "# Make relation \"blk_sdanode\" \t(id %d)\n",
                new_rel_id);
        fprintf(stderr, "  \"blkId\" \t(id %d)\n", blk_id);
        fprintf(stderr, "  \"sdanodeId\" \t(id %d)\n\n", new_blkExit_id);
    }
}

/*
 * åѤࡥ
 */

static void pushBlkStack(SpdObjId blk_id, int blk_sort,
                         SpdObjId blkEntry_id, SpdObjId blkExit_id)
{
    /* ν */
    if (blkStack_n == BLK_STACK_MAX - 1) {
        spdErrorExit("Stack Over Flow\n");
    }
    
    blkStack_n++;
    blkStack[blkStack_n].blk_sort = blk_sort;
    blkStack[blkStack_n].blk_id = blk_id;
    blkStack[blkStack_n].blkEntry_id = blkEntry_id;
    blkStack[blkStack_n].blkExit_id = blkExit_id;
    blkStack[blkStack_n].blkBreak_flag = FALSE;
    blkStack[blkStack_n].blkBreak_id = SPD_NON_ID;
    blkStack[blkStack_n].blkContinue_flag = FALSE;
    blkStack[blkStack_n].blkContinue_id = SPD_NON_ID;
    blkStack[blkStack_n].blkReturn_flag = FALSE;
    blkStack[blkStack_n].blkReturn_id = SPD_NON_ID; 
  
    if (debugPrintFlag2) {
        fprintf(stderr, "Push into blkStack[%d]\n\n", blkStack_n);
    }
}

/*
 * å
 */

static void popBlkStack(void)
{
    blkStack[blkStack_n].blk_sort = SPD_NON_ID;
    blkStack[blkStack_n].blk_id = SPD_NON_ID;
    blkStack[blkStack_n].blkEntry_id = SPD_NON_ID;
    blkStack[blkStack_n].blkExit_id = SPD_NON_ID;
    blkStack[blkStack_n].blkBreak_flag = FALSE;
    blkStack[blkStack_n].blkBreak_id = SPD_NON_ID;
    blkStack[blkStack_n].blkContinue_flag = FALSE;
    blkStack[blkStack_n].blkContinue_id = SPD_NON_ID;
    blkStack[blkStack_n].blkReturn_flag = FALSE;
    blkStack[blkStack_n].blkReturn_id = SPD_NON_ID;
    blkStack_n--;
    
    if (debugPrintFlag2) {
        fprintf(stderr, "Pop from blkStack[%d]\n\n", blkStack_n + 1);
    }
}

/**************** Ϣ ****************************************/

/*
 * æѽиĤʤ롥
 */

static void makeBlkBreakRel(int cur_blkStack_n)
{
    int         cur_blk_sort;
    int         cur_blkBreak_flag;
    SpdObjId    new_blkBreak_id;
    SpdRelId    new_rel_id;
    SpdObjId    cur_blk_id;
    SpdObjId    cur_blkExit_id;
    SpdObjId    cur_blkBreak_id;

    /* ȤΥ֥åΡbreakѤνи˴ϢϤ */
    /* gotoξϡΥ֥åޤǥå򤵤Τܤä
     * ǤgotoѤδϢϤ롥
     */
    cur_blk_id = blkStack[cur_blkStack_n].blk_id;
    cur_blk_sort = blkStack[cur_blkStack_n].blk_sort;
    cur_blkBreak_flag = blkStack[cur_blkStack_n].blkBreak_flag;
 
    if (cur_blkBreak_flag == TRUE ) {
        /* ե饰ΩäƤν */
        /* COMPIFΤɤ餫Ǥ롥 */
        
        /* breakȡǤ˺Ƥ blkBreak Ĥʤ */
        cur_blkBreak_id = blkStack[cur_blkStack_n].blkBreak_id;
        makeNextRel(cur_blkBreak_id, SPD_FV_E_NORMAL);
    } else if (cur_blkBreak_flag == FALSE) {
        switch (cur_blk_sort) {
          case SPD_FV_B_COMP:
          case SPD_FV_B_IF:
            new_blkBreak_id = spdCreateObj(SPD_CID_BLK_BREAK, SPD_NON_ID);
            spdSetAttrIntValByName(new_blkBreak_id, "sort", cur_blk_sort);

            if (debugPrintFlag) {
                fprintf(stderr, "@ Make object \"blkBreak\" \t(id %d)\n",
                        new_blkBreak_id);
                fprintf(stderr, "  sort : \t");
                printBlkSort(cur_blk_sort);
                fprintf(stderr, "\n");
            }
            
            new_rel_id = spdCreateRel(SPD_RID_BLK_SDANODE, SPD_NON_ID,
                                      cur_blk_id, new_blkBreak_id);

            if (debugPrintFlag) {
                fprintf(stderr, "# Make relation \"blk_sdanode\" \t(id %d)\n",
                        new_rel_id);
                fprintf(stderr, "  \"blkId\" \t(id %d)\n", cur_blk_id);
                fprintf(stderr, "  \"sdanodeId\" \t(id %d)\n\n", new_blkBreak_id);
            }

            /* breakʸbreak_outlastIdˤʤäƤ롥 */
            makeNextRel(new_blkBreak_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkBreak_flag = TRUE;
            blkStack[cur_blkStack_n].blkBreak_id = new_blkBreak_id; /* иidϿ */
            makeBlkBreakRel(cur_blkStack_n - 1); /* ƵƤӽФ */
            break;
          case SPD_FV_B_SWITCH:
          case SPD_FV_B_FOR:
          case SPD_FV_B_WHILE:
          case SPD_FV_B_DO:
            /* ֥åνиˤĤʤ */
            cur_blkExit_id = blkStack[cur_blkStack_n].blkExit_id;
            makeNextRel(cur_blkExit_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkBreak_flag = TRUE;
            break;
          case SPD_FV_B_FUNC:
            fprintf(stderr, "error: makeBlkBreakRel()\n");
            spdErrorExit("break not within loop or switch\n");
          default:
            spdErrorExit("makeBlkBreakRel()\n");
        }    
    }
}

static void makeBlkContinueRel(int cur_blkStack_n)
{
    int         cur_blk_sort;
    int         cur_blkContinue_flag;
    SpdObjId    new_blkContinue_id;
    SpdRelId    new_rel_id;
    SpdObjId    continued_id;               /* continueǹԤĤΥΡɤid */
    SpdObjId    cur_blk_id;
    SpdObjId    cur_blkContinue_id;
    
    /* ȤΥ֥åΡsortѤνи˴ϢϤ롥 */
    cur_blk_id = blkStack[cur_blkStack_n].blk_id;
    cur_blk_sort = blkStack[cur_blkStack_n].blk_sort;
    cur_blkContinue_flag = blkStack[cur_blkStack_n].blkContinue_flag;
    
    if (cur_blkContinue_flag == TRUE ) {
        /* ե饰ΩäƤν */
        /* COMPIFʤ(SWITCHľbreakʤ) */
    
        /* continueȡǤ˺ƤblkContinueĤʤ */
        cur_blkContinue_id = blkStack[cur_blkStack_n].blkContinue_id;
        makeNextRel(cur_blkContinue_id, SPD_FV_E_NORMAL);

    } else if (cur_blkContinue_flag == FALSE) {        
        switch (cur_blk_sort) {
          case SPD_FV_B_COMP:
          case SPD_FV_B_IF:
          case SPD_FV_B_SWITCH:
            new_blkContinue_id = spdCreateObj(SPD_CID_BLK_CONTINUE,
                                              SPD_NON_ID);
            spdSetAttrIntValByName(new_blkContinue_id, "sort", cur_blk_sort);

            if (debugPrintFlag) {
                fprintf(stderr, "@ Make object \"blkContinue\" \t(id %d)\n",
                        new_blkContinue_id);
                fprintf(stderr, "  sort : \t");
                printBlkSort(cur_blk_sort);
                fprintf(stderr, "\n");
            }

            new_rel_id = spdCreateRel(SPD_RID_BLK_SDANODE, SPD_NON_ID,
                                      cur_blk_id, new_blkContinue_id);

            if (debugPrintFlag) {
                fprintf(stderr, "# Make relation \"blk_sdanode\" \t(id %d)\n",
                        new_rel_id);
                fprintf(stderr, "  \"blkId\" \t(id %d)\n", cur_blk_id);
                fprintf(stderr, "  \"sdanodeId\" \t(id %d)\n\n", new_blkContinue_id);
            }

            /* continueʸlastIdˤʤäƤ롥 */
            makeNextRel(new_blkContinue_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkContinue_flag = TRUE;
            blkStack[cur_blkStack_n].blkContinue_id = new_blkContinue_id; /*иidϿ*/
            makeBlkContinueRel(cur_blkStack_n - 1); /* ƵƤӽФ */
            break;
          case SPD_FV_B_FOR:
            /* forϸˤĤʤ */
            {
                SpdRelArray expr_buf;
     
                expr_buf = spdGetRelSortedArray(blkStack[cur_blkStack_n].blk_id,
                                                "blk_expr", "parent_block_id",
                                                NULL, spdCompareRelOffset);
                continued_id = expr_buf.rel[2].objectId;
                spdFreeRelArray(expr_buf);
            }
            makeNextRel(continued_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkContinue_flag = TRUE;
            break;
          case SPD_FV_B_WHILE:
          case SPD_FV_B_DO:
            /* whiledoϾＰˤĤʤ */
            continued_id = spdGetARelObj(blkStack[cur_blkStack_n].blk_id,
                                         "blk_expr", "parent_block_id");
            makeNextRel(continued_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkContinue_flag = TRUE;
            break;
          case SPD_FV_B_FUNC:
            fprintf(stderr, "error: makeBlkContinueRel()\n");
            spdErrorExit("continue not within loop\n");
          default:
            spdErrorExit("makeBlkContinueRel()\n");
        }    
    }
}

static void makeBlkReturnRel(int cur_blkStack_n)
{
    int         cur_blk_sort;
    int         cur_blkReturn_flag;
    SpdObjId    new_blkReturn_id;
    SpdRelId    new_rel_id;
    SpdObjId    cur_blk_id;
    SpdObjId    cur_blkReturn_id;
    SpdObjId    funcExit_id;

    cur_blk_id = blkStack[cur_blkStack_n].blk_id;
    cur_blk_sort = blkStack[cur_blkStack_n].blk_sort;
    cur_blkReturn_flag = blkStack[cur_blkStack_n].blkReturn_flag;

    if (cur_blkReturn_flag == TRUE) {
        cur_blkReturn_id = blkStack[cur_blkStack_n].blkReturn_id;
        makeNextRel(cur_blkReturn_id, SPD_FV_E_NORMAL);
    } else if (cur_blkReturn_flag == FALSE) {
        switch (cur_blk_sort) {
          case SPD_FV_B_COMP:
          case SPD_FV_B_IF:
          case SPD_FV_B_SWITCH:
          case SPD_FV_B_FOR:
          case SPD_FV_B_WHILE:
          case SPD_FV_B_DO:
        
            new_blkReturn_id = spdCreateObj(SPD_CID_BLK_RETURN, SPD_NON_ID);
            spdSetAttrIntValByName(new_blkReturn_id, "sort", cur_blk_sort);

            if (debugPrintFlag) {
                fprintf(stderr, "@ Make object \"blkReturn\" \t(id %d)\n",
                        new_blkReturn_id);
                fprintf(stderr, "  sort : \t");
                printBlkSort(cur_blk_sort);
                fprintf(stderr, "\n");
            }

            new_rel_id = spdCreateRel(SPD_RID_BLK_SDANODE, SPD_NON_ID,
                                      cur_blk_id, new_blkReturn_id);

            if (debugPrintFlag) {
                fprintf(stderr, "# Make relation \"blk_sdanode\" \t(id %d)\n",
                        new_rel_id);
                fprintf(stderr, "  \"blkId\" \t(id %d)\n", cur_blk_id);
                fprintf(stderr, "  \"sdanodeId\" \t(id %d)\n\n", new_blkReturn_id);
            }

            /* returnʸ blkReturn lastIdˤʤäƤ롥 */
            makeNextRel(new_blkReturn_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkReturn_flag = TRUE;
            blkStack[cur_blkStack_n].blkReturn_id = new_blkReturn_id; /*иidϿ*/
            makeBlkReturnRel(cur_blkStack_n - 1); /* ƵƤӽФ */
            break;
          case SPD_FV_B_FUNC:
            /* ؿνиˤĤʤ */
            funcExit_id = blkStack[cur_blkStack_n].blkExit_id;
            makeNextRel(funcExit_id, SPD_FV_E_NORMAL);
            blkStack[cur_blkStack_n].blkReturn_flag = TRUE;
            break;
          default:
            spdErrorExit("makeBlkReturnRel()\n");
        }        
    }
}

static void makeBlkGoAwayRel(SpdObjId expr_id, int cur_blkStack_n)
{
    int         cur_blk_sort;
    int         i;
    SpdObjId    new_blkGoAway_id;
    SpdRelId    new_rel_id;
    SpdRelArray label_buf;
    SpdObjId    cur_blk_id;
    SpdObjId    cur_label_id;
    SpdObjId    goto_label_id;

    cur_blk_id = blkStack[cur_blkStack_n].blk_id;
    cur_blk_sort = blkStack[cur_blkStack_n].blk_sort;
    
    switch (cur_blk_sort) {
      case SPD_FV_B_COMP:
        /* ŪΥ٥ĤĴ٤롥 */
        /* goto_label. */
        
        goto_label_id = spdGetARelObj(expr_id, "goto_label", "expression_id");

        label_buf = spdGetRelSortedArray(cur_blk_id, "blk_label", "block_id",
                                         NULL, spdCompareRelOffset);
        for (i = 0; i < label_buf.size; i++) {
            cur_label_id = label_buf.rel[i].objectId;
            
            if (cur_label_id == goto_label_id) {
                makeNextRel(goto_label_id, SPD_FV_E_NORMAL);
                spdFreeRelArray(label_buf);
                return;                     /* ȴޤ */
            }
        }
        
        spdFreeRelArray(label_buf);
        /* ŪΥ٥COMPǤʤä鲼ν³ */
      case SPD_FV_B_IF:
      case SPD_FV_B_SWITCH:
      case SPD_FV_B_FOR:
      case SPD_FV_B_WHILE:
      case SPD_FV_B_DO:
        new_blkGoAway_id = spdCreateObj(SPD_CID_BLK_GO_AWAY, SPD_NON_ID);
        spdSetAttrIntValByName(new_blkGoAway_id, "sort", cur_blk_sort);

        if (debugPrintFlag) {
            fprintf(stderr, "@ Make object \"blkGoAway\" \t(id %d)\n",
                    new_blkGoAway_id);
            fprintf(stderr, "  sort : \t");
            printBlkSort(cur_blk_sort);
            fprintf(stderr, "\n");
        }

        new_rel_id = spdCreateRel(SPD_RID_BLK_SDANODE, SPD_NON_ID,
                                  cur_blk_id, new_blkGoAway_id);

        if (debugPrintFlag) {
            fprintf(stderr, "# Make relation \"blk_sdanode\" \t(id %d)\n",
                    new_rel_id);
            fprintf(stderr, "  \"blkId\" \t(id %d)\n", cur_blk_id);
            fprintf(stderr, "  \"sdanodeId\" \t(id %d)\n\n", new_blkGoAway_id);
        }
 
        /* gotoʸblkGoAwaylastIdˤʤäƤ롥 */
        makeNextRel(new_blkGoAway_id, SPD_FV_E_NORMAL);
        makeBlkGoAwayRel(expr_id, cur_blkStack_n - 1); /* ƵƤӽФ */
        break;
      case SPD_FV_B_FUNC:
        /* goto֥åǻȤƤˤȯǤʤä */
        fprintf(stderr, "error: makeBlkGoAwayRel()\n");
        spdErrorExit("SDA don't allow goto enter a block.\n"); 
        break;
      default:
        spdErrorExit("error: makeBlkGoAwayRel()\n");
    }
}

/*
 * lastIdexpr_idΥ֥ȤĤʤ
 */

static void makeNextRel(SpdObjId next_id, int edge_sort)
{
    SpdRelId new_rel_id = 0;
    
    if (lastId == SPD_NON_ID) {
        lastId = next_id;
        return;
    }
    
    /* ѿȤä */
    if (nextEdgeSort == NEXT_EDGE_IS_TRUE_EDGE) {
        nextEdgeSort = NEXT_EDGE_IS_FREE;   /* ե饰򥯥ꥢ */
        edge_sort = SPD_FV_E_BRANCH_TRUE;
    } else if (nextEdgeSort == NEXT_EDGE_IS_FALSE_EDGE) {
        nextEdgeSort = NEXT_EDGE_IS_FREE;   /* ե饰򥯥ꥢ */
        edge_sort = SPD_FV_E_BRANCH_FALSE;
    }

    /* ϢϤ롥 */
    /* !! ΥХ: ϢĤAR5spdGetObjText()ưʤʤ롥*/
    
    new_rel_id = spdCreateRel(SPD_RID_NEXT_NODE, SPD_NON_ID,
                              lastId, next_id);
    spdSetAttrIntValByName(new_rel_id, "sort", edge_sort);
    
    if (debugPrintFlag) {
        fprintf(stderr, "# Make relation \"next_node\" \t(id %d)\n",
                new_rel_id);
        fprintf(stderr, "  %d (%s) -> ", lastId, spdGetObjText(lastId));
        fprintf(stderr, "%d (%s)\n", next_id, spdGetObjText(next_id));

        fprintf(stderr, "  \"sort\"\t: ");
        printEdgeSort(edge_sort);
        fprintf(stderr, "\n\n");
    }

    lastId = next_id;
}

/**************** Ƽ֥åΤν ****************************/

/*
 * sortBLOCK_BASICΥ֥åexprϢǤĤʤ롥
 * ޤifʸ(BLOCK_IFΥ֥å)ξＰΤˤȤ롥
 * whileʸξＰˤȤ롥
 */

static void basicblkTop(SpdObjId blk_id)
{
    SpdRelArray expr_buf;
    SpdObjId cur_expr_id;
    int i;

    expr_buf = spdGetRelSortedArray(blk_id, "blk_expr", "parent_block_id",
                                    NULL, spdCompareRelOffset);
    
    if (expr_buf.size > 0){
        for (i = 0; i < expr_buf.size; i++) {
            cur_expr_id = expr_buf.rel[i].objectId;
            makeNextRel(cur_expr_id, SPD_FV_E_NORMAL);
        }
    }
    spdFreeRelArray(expr_buf); 

    /*
     * BASIC֥åκǸexprå
     * ʸä顤ѴؿƤӡlastId롥
     */
    switch (spdGetAttrValInt(lastId, "sort")) {
      case EXPR_BREAK:
        makeBlkBreakRel(blkStack_n);
        lastId = SPD_NON_ID;              /* Ϣڤܡ */
        break;
      case EXPR_CONTINUE:
        makeBlkContinueRel(blkStack_n);
        lastId = SPD_NON_ID;
        break;
      case EXPR_RETURN:
        makeBlkReturnRel(blkStack_n);
        lastId = SPD_NON_ID;
        break;
      case EXPR_GOTO:
        makeBlkGoAwayRel(cur_expr_id, blkStack_n);
        /* gotoexprlabel롥 */
        lastId = SPD_NON_ID;
        break;
    }
}

/*
 * compound֥åѡ
 */

static SpdObjId compblkTop(SpdObjId blk_id)
{
    SpdObjId blkEntry_id, blkExit_id;
  
    makeBlkNode(blk_id, SPD_FV_B_COMP, &blkEntry_id, &blkExit_id);
    pushBlkStack(blk_id, SPD_FV_B_COMP, blkEntry_id, blkExit_id);

    makeNextRel(blkEntry_id, SPD_FV_E_NORMAL);
    
    /* fprintf(stderr, "#####COMP IN##### %d\n", blkEntry_id);*/
    blkloop(blk_id);
    /* fprintf(stderr, "####COMP OUT##### %d\n", blkExit_id);*/

    makeNextRel(blkExit_id, SPD_FV_E_NORMAL);

    popBlkStack();

    /* id֤ doʸΥ롼ץХåΤ뤿ν*/
    return blkEntry_id;
}

/*
 * compblkTop()ƤФ롥 ϤCOMP֥å
 * COMP֥åλҥ֥åidblkSwitch()Ƥ֡
 */

static void blkloop(SpdObjId blk_id)
{
    int cur_blk_offset, cur_label_offset;
    int cur_label_sort;
    int i, j;
    SpdRelArray blk_buf, label_buf;
    SpdObjId cur_blk_id, cur_label_id;
    SpdObjId new_switchBranch_id = SPD_NON_ID;
    /* COMP֥åSWITCHΤΤɤ򼨤ե饰 */
    SpdBoolean thisCompIsSwitchComp = FALSE;
    SpdBoolean thisCompHasDefault = FALSE;
    
#ifdef COMMENT_OUT
    SpdRelArray decl_buf;
#endif

#ifdef COMMENT_OUT                          /* ϥեդ鳰 */
    /*
     *  
     * ֥åƬdeclaration򥪥եåȤǥȤƼ
     */

    decl_buf = spdGetRelSortedArray(blk_id, "blk_decl", "parent_block_id",
                                    NULL, spdCompareRelOffset);
    for(i = 0; i < decl_buf.size; i++){
        /* To Do  եդˤĤʤ */
        makeNextRel(decl_buf.rel[i].objectId, SPD_FV_E_NORMAL);

        /* declarationʸɽɽ */
        if(debugPrintFlag2)
            fprintf(stderr, "declaration : %s\n", 
                    spdGetObjText(decl_buf.rel[i].objectId));
        
    }
    spdFreeRelArray(decl_buf);
#endif /* COMMENT_OUT */

    /*
     * switchʸCOMP֥åǤȤե饰ΩäƤ
     * switchBranch롥
     */
    if (nextCompBelong == NEXT_COMP_IS_SWITCH_COMP) {
        thisCompIsSwitchComp = TRUE;
        nextCompBelong = NEXT_COMP_IS_NOT_SWITCH_COMP; /* Ƥ */
        
        /* switch branch 롥 */
        new_switchBranch_id = spdCreateObj(SPD_CID_SWITCH_BRANCH,
                                           SPD_NON_ID);
        if (debugPrintFlag)
            fprintf(stderr, "@ Make object \"switchBranch\" \t(id %d)\n\n",
                    new_switchBranch_id);

        makeNextRel(new_switchBranch_id, SPD_FV_E_NORMAL);

        /* switch branch θϤĤʤʤ() */
        lastId = SPD_NON_ID;
    }
   
    /*
     *  
     * blk_expr δϢCOMP֥åΤΤ¸ߤʤ
     * (BASIC֥åȡif while for do֥åξＰǽФ)
     */

    /* "parent_block_id"ǤϤʤ"any"ˤ̵¥롼פˤʤΤա */
    blk_buf = spdGetRelSortedArray(blk_id, "blk_blk", "parent_block_id",
                                   NULL, spdCompareRelOffset);
    /* label */
    label_buf = spdGetRelSortedArray(blk_id, "blk_label", "block_id",
                                     NULL, spdCompareRelOffset);

    i = 0;
    j = 0;
    while (i < blk_buf.size || j < label_buf.size) {
        if (i < blk_buf.size) {
            cur_blk_id     = blk_buf.rel[i].objectId;
            cur_blk_offset = spdGetOffset(blk_buf.rel[i].relationId);
        } else {
            cur_blk_id = SPD_NON_ID;
            /* cur_blk_offset ϻȤΤǽƤϤʤ*/
        }
        
        if (j < label_buf.size) {
            cur_label_id = label_buf.rel[j].objectId;
            cur_label_offset = spdGetOffset(label_buf.rel[j].relationId);
            cur_label_sort = spdGetAttrValInt(cur_label_id, "sort");
        } else {
            cur_label_id = SPD_NON_ID;
            cur_label_offset = -1;
        }

        /* ֥å¤ȡ٥ǥեåȤξΤ롥 */
        if ((cur_label_offset > cur_blk_offset || cur_label_offset == -1) &&
            i < blk_buf.size) {
            
            blkSwitch(cur_blk_id);          /* ֥åѤν */
            i++;
        } else {
            /* ΥΡɤ٥ؤδϢϤ롥 */
            makeNextRel(cur_label_id, SPD_FV_E_NORMAL);
            
            if (thisCompIsSwitchComp == TRUE) {
                SpdObjId temp_lastId = lastId;
                
                /* switchBranch ٥ؤδϢϤ롥 */
                lastId = new_switchBranch_id;
                switch (cur_label_sort) {
                  case LABEL_CASE:
                    makeNextRel(cur_label_id, SPD_FV_E_SWITCH_CASE);
                    break;
                  case LABEL_DEFAULT:
                    thisCompHasDefault = TRUE;
                    makeNextRel(cur_label_id, SPD_FV_E_SWITCH_DEFAULT);
                    break;
                  default:
                    /* GOTOѤΥ٥SWITCHˤϴطʤΤǤʤˤ⤷ʤ */
                    break;
                }
                lastId = temp_lastId;
            }
            j++;
        }   
    }
    
    if (thisCompIsSwitchComp == TRUE && thisCompHasDefault == FALSE) {
        SpdObjId temp_lastId = lastId;
        
        /* NOMATCHδϢϤ롥 */
        lastId = new_switchBranch_id;
        makeNextRel(blkStack[blkStack_n].blkExit_id, SPD_FV_E_SWITCH_NOMATCH);
        
        lastId = temp_lastId;
    }
    
    spdFreeRelArray(blk_buf);
    spdFreeRelArray(label_buf);
}

/*
 * if֥åλҥ֥åblkSwitch()ؿϤ
 * ¾ˤ⿧ifѤν򤷤Ƥ롥
 */

static SpdObjId ifblkTop(SpdObjId blk_id)
{
    SpdObjId cond_expr_id;                  /* Ｐid thenθɬס */
    SpdObjId cur_blk_id;
    SpdObjId blkEntry_id, blkExit_id;
    SpdRelArray blk_buf;                    /* if֥åλҥ֥å(then else)ΥХåե */
    int curBranchCount;

    if (debugPrintFlag2) {
        curBranchCount = branchCount++;
        fprintf(stderr, "\nif block No. %d\n", curBranchCount);
    }

    makeBlkNode(blk_id, SPD_FV_B_IF, &blkEntry_id, &blkExit_id);
    pushBlkStack(blk_id, SPD_FV_B_IF, blkEntry_id, blkExit_id);

    makeNextRel(blkEntry_id, SPD_FV_E_NORMAL);
    
    /* Ｐ */
    basicblkTop(blk_id);
    cond_expr_id = lastId;

    /*  */
    blk_buf = spdGetRelSortedArray(blk_id, "blk_blk", "parent_block_id",
                                   NULL, spdCompareRelOffset);

    /* then */
    nextEdgeSort = NEXT_EDGE_IS_TRUE_EDGE;
    if (blk_buf.size >= 1){
        cur_blk_id = blk_buf.rel[0].objectId;

        if (debugPrintFlag2) {
            fprintf(stderr, "\"then\" of if block No. %d\n\n", curBranchCount);
        }
        
        blkSwitch(cur_blk_id);
   
        if (debugPrintFlag2) {
            fprintf(stderr, "End of \"then\" of if block No. %d\n\n",
                    curBranchCount);
        }

        /* thenκǸΥΡɤidendΡɤˤĤʤ */
        makeNextRel(blkExit_id, SPD_FV_E_NORMAL);
    }

    /* else */
    nextEdgeSort = NEXT_EDGE_IS_FALSE_EDGE;
    lastId = cond_expr_id;                  /*  */
    if(blk_buf.size == 2){
        cur_blk_id = blk_buf.rel[1].objectId;
        
        if (debugPrintFlag2) {
            fprintf(stderr, "\"else\" of if block No. %d\n\n", curBranchCount);
        }

        blkSwitch(cur_blk_id);
    
        if (debugPrintFlag2) {
            fprintf(stderr, "End of \"else\" of if block No. %d\n\n",
                    curBranchCount);
        }
        
        /* elseκǸΥΡɤidendΡɤˤĤʤ */
        makeNextRel(blkExit_id, SPD_FV_E_NORMAL);
    } else {
        /* else̵ν */
        /* cond_expr_id endΡɤˤĤʤ */
        makeNextRel(blkExit_id, SPD_FV_E_BRANCH_FALSE);
    }
    
    /* lastId ϲendΡɤˤʤäƤ롥 */
    popBlkStack();
    spdFreeRelArray(blk_buf);
    
    return blkEntry_id;
}

/* 
 * switchʸѡ
 */

static SpdObjId switchblkTop(SpdObjId blk_id)
{
    SpdObjId c_blk_id;
    SpdObjId blkEntry_id, blkExit_id;

    makeBlkNode(blk_id, SPD_FV_B_SWITCH, &blkEntry_id, &blkExit_id);
    pushBlkStack(blk_id, SPD_FV_B_SWITCH, blkEntry_id, blkExit_id);
  
    makeNextRel(blkEntry_id, SPD_FV_E_NORMAL);

    /*  */
    basicblkTop(blk_id);
  
    /*  */
    c_blk_id = spdGetARelObj(blk_id, "blk_blk", "parent_block_id");

    switch (spdGetAttrValInt(c_blk_id, "sort")) {
      case BLOCK_COMP:
        nextCompBelong = NEXT_COMP_IS_SWITCH_COMP;
        blkSwitch(c_blk_id);
        break;
      default:
        spdWarning("switchblkTop() : unreachable code\n");
        break;
    }
    
    makeNextRel(blkExit_id, SPD_FV_E_NORMAL);
  
    popBlkStack();

    return blkEntry_id;
}

/* 
 * forʸѡ
 */

static SpdObjId forblkTop(SpdObjId blk_id)
{
    SpdRelArray expr_buf;                   /* μΥХåե */
    SpdObjId c_blk_id;                      /* while֥åλҥ֥å */
    SpdObjId ini_expr, branch_expr, foot_expr;
    SpdObjId blkEntry_id, blkExit_id;

    if (debugPrintFlag2)
        fprintf(stderr, "\nfor loop start\n");

    /* иΡɤ롥 */
    makeBlkNode(blk_id, SPD_FV_B_FOR, &blkEntry_id, &blkExit_id);
    pushBlkStack(blk_id, SPD_FV_B_FOR, blkEntry_id, blkExit_id);

    /* Ĥʤ */
    makeNextRel(blkEntry_id, SPD_FV_E_NORMAL);

    /* ʤɤν */
    expr_buf = spdGetRelSortedArray(blk_id, "blk_expr", "parent_block_id",
                                    NULL, spdCompareRelOffset);

    if(expr_buf.size == 3){
        ini_expr = expr_buf.rel[0].objectId;
        branch_expr = expr_buf.rel[1].objectId;
        foot_expr = expr_buf.rel[2].objectId;
    } else {
        spdErrorExit("forblkTop()\n");
    }
    spdFreeRelArray(expr_buf);

    if (debugPrintFlag2) {
        fprintf(stderr, " ini %d (%s)\n", ini_expr, spdGetObjText(ini_expr));
        fprintf(stderr, " branch %d (%s)\n", branch_expr, spdGetObjText(branch_expr));
        fprintf(stderr, " foot %d (%s)\n", foot_expr, spdGetObjText(foot_expr));
    }

    /* ϢϤ롥 */
    makeNextRel(ini_expr, SPD_FV_E_NORMAL);
    makeNextRel(branch_expr, SPD_FV_E_NORMAL);
    nextEdgeSort = NEXT_EDGE_IS_TRUE_EDGE;  /* ѿȤä */

    /*  */
    c_blk_id = spdGetARelObj(blk_id, "blk_blk", "parent_block_id");

    blkSwitch(c_blk_id);

    /*  */
    makeNextRel(foot_expr, SPD_FV_E_NORMAL);

    /* 롼ץХå */
    makeNextRel(branch_expr, SPD_FV_E_LOOPBACK);

    /* 롼æ */
    /* branch_expr blkExit_idˤĤʤ */
    nextEdgeSort = NEXT_EDGE_IS_FALSE_EDGE; /* ѿȤä */
    lastId = branch_expr;
    /* иĤʤ */
    makeNextRel(blkExit_id, SPD_FV_E_NORMAL);

    popBlkStack();

    if (debugPrintFlag2)
        fprintf(stderr, "End of for\n\n");

    return blkEntry_id;
}

/*
 * while֥åѡ
 */

static SpdObjId whileblkTop(SpdObjId blk_id)
{
    SpdObjId c_blk_id;                      /* while֥åλҥ֥å */
    SpdObjId cond_expr_id;                  /* ＰidΤȡ */
    SpdObjId blkEntry_id, blkExit_id;

    if (debugPrintFlag2)
        fprintf(stderr, "\nwhile loop start\n");

    makeBlkNode(blk_id, SPD_FV_B_WHILE, &blkEntry_id, &blkExit_id);
    pushBlkStack(blk_id, SPD_FV_B_WHILE, blkEntry_id, blkExit_id);
    
    makeNextRel(blkEntry_id, SPD_FV_E_NORMAL);
    
    /* Ｐ */
    basicblkTop(blk_id);
    cond_expr_id = lastId;

    nextEdgeSort = NEXT_EDGE_IS_TRUE_EDGE;  /* ѿȤä */
    
    /*  */
    c_blk_id = spdGetARelObj(blk_id, "blk_blk", "parent_block_id");

    blkSwitch(c_blk_id);

    /* 롼ץХåδϢϤ롥 */
    makeNextRel(cond_expr_id, SPD_FV_E_LOOPBACK);

    /* 郎ʤäȤδϢ(롼æС) */
    /* cond_expr_id 鲾endΡɤؤδϢϤ롥 */
    nextEdgeSort = NEXT_EDGE_IS_FALSE_EDGE; /* ѿȤä */
    makeNextRel(blkExit_id, SPD_FV_E_NORMAL);
    
    popBlkStack();

    if (debugPrintFlag2)
        fprintf(stderr, "End of while\n\n");

    return blkEntry_id;
}

/*
 * do while֥å
 */

static SpdObjId doblkTop(SpdObjId blk_id)
{
    SpdObjId cond_expr_id;
    SpdObjId c_blk_id;                      /* do while֥åλҥ֥å */
    SpdObjId blkEntry_id, blkExit_id;
    SpdObjId loopBackTarget_id;             /* 롼ץХå补 */

    if (debugPrintFlag2)
        fprintf(stderr, "\ndo while loop start\n");

    makeBlkNode(blk_id, SPD_FV_B_DO, &blkEntry_id, &blkExit_id);
    pushBlkStack(blk_id, SPD_FV_B_DO, blkEntry_id, blkExit_id);

    makeNextRel(blkEntry_id, SPD_FV_E_NORMAL);

    /*  */
    c_blk_id = spdGetARelObj(blk_id, "blk_blk", "parent_block_id");

    if (spdGetAttrValInt(c_blk_id, "sort") == BLOCK_BASIC) {
        blkSwitch(c_blk_id);
        loopBackTarget_id = lastId;
    } else {
        loopBackTarget_id = blkSwitch(c_blk_id);
    }

    /* Ｐ */
    basicblkTop(blk_id);
    cond_expr_id = lastId;

    /* doʸΥ롼ץХåTRUEǼ */
    nextEdgeSort = NEXT_EDGE_IS_TRUE_EDGE;
    makeNextRel(loopBackTarget_id, SPD_FV_E_LOOPBACK);

    /* 郎ʤäȤδϢ(롼æ) */
    /* lastId(Ｐ)鲾endΡɤؤδϢϤ롥 */
    nextEdgeSort = NEXT_EDGE_IS_FALSE_EDGE;
    lastId = cond_expr_id;
    makeNextRel(blkExit_id, SPD_FV_E_NORMAL);

    popBlkStack();

    if (debugPrintFlag2)
        fprintf(stderr, "End of do while\n\n");

    return blkEntry_id;
}

/*
 * ֥åμऴȤ˽ؿǤ錄 
 */

static SpdObjId blkSwitch(SpdObjId blk_id)
{
    SpdObjId ret_id = SPD_NON_ID;

    /* ֤ͤret_id˼äƤΤdo֥åǻȤᡥ */
    /* blkEntryidϤƤ롥*/
    switch (spdGetAttrValInt(blk_id, "sort")) {
      case BLOCK_HIERACHICAL :              /*  */
        spdErrorExit("blkSwitch() : hierachial\n");
        break;
      case BLOCK_BRANCH :                   /*  */
        spdErrorExit("blkSwitch() : branch\n");
        break;
      case BLOCK_BASIC :
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : Basic Block\n");
        basicblkTop(blk_id);
        break;
      case BLOCK_COMP :
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : Compound Block\n");
        ret_id = compblkTop(blk_id);
        break;
      case BLOCK_IF:
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : If Block\n");
        ret_id = ifblkTop(blk_id);
        break;
      case BLOCK_SWITCH :
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : Switch Block\n");
        ret_id = switchblkTop(blk_id);
        break;
      case BLOCK_FOR :
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : For Block\n");
        ret_id = forblkTop(blk_id);
        break;
      case BLOCK_WHILE :
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : While Block\n");
        ret_id = whileblkTop(blk_id);
        break;
      case BLOCK_DO :
        if (debugPrintFlag2)
            fprintf(stderr, "blkSwitch() : Do Block\n");
        ret_id = doblkTop(blk_id);
        break;
      default:
        break;
    }
    return ret_id;
}

/**************** ؿΥ֥åޤǤν *************************/

/*
 * ؿƤ"declaration"idؿΥ֥åid롥
 * "block"id˴ؿblkSwitch()Ƥ֡
 */

static void funcTop(SpdObjId funcDecl_id)
{
    SpdObjId decr_id, opt_id, blk_id;
    SpdObjId func_entry_id, func_exit_id;

    /* ؿȽиβۥΡɤꥹåѤࡥ */
    /* ۥΡɤfuncDecl_id֤δϢblk_sdanodeϤ롥 */
    makeBlkNode(funcDecl_id, SPD_FV_B_FUNC, &func_entry_id, &func_exit_id);
    pushBlkStack(funcDecl_id, SPD_FV_B_FUNC, func_entry_id, func_exit_id);

    /* ؿȤ˽(blkEntry饹) */
    lastId = func_entry_id;

    decr_id = spdGetARelObj(funcDecl_id, "decl_decl", "declaration_id");
    opt_id = spdGetARelObj(decr_id, "decl_opt", "declarator_id");
    blk_id = spdGetARelObj(opt_id, "func_body", "function_id");
  
    blkSwitch(blk_id);

    /* ؿνиȺǸδϢϤ롥 */
    makeNextRel(func_exit_id, SPD_FV_E_NORMAL);
    popBlkStack();
}

/*
 * "file"idꡤΥեȴϢΤؿ"declaration"
 * id롥"declaration"id˴ؿfuncTop()Ƥ֡
 */

static void file2func(SpdObjId file_id)
{
    SpdCursor *decl_csr;
    SpdObjId  decl_id;

    decl_csr = spdGetRelObjInit(file_id, "file_decl", "file_id");
    while ((decl_id = spdGetRelObj(decl_csr)) != SPD_NON_ID) {
        if (spdGetAttrValInt(decl_id, "sort") == DECL_FUNCDEF) {
            if (debugPrintFlag) {
                SpdObjId funcDecr_id, funcName_id;

                /* ؿ̾ɽ */
                funcDecr_id = spdGetARelObj(decl_id, "decl_decl",
                                            "declaration_id");
                funcName_id = spdGetARelObj(funcDecr_id, "decl_ident",
                                            "declarator_id");
                fprintf(stderr, "function name : %s\n",
                        spdGetName(funcName_id));
            }
            funcTop(decl_id);
        }
    }
    spdFreeCursor(decl_csr);
}

/* 
 * "program"idꡤΥץȴϢΤ"file"id롥
 * ơ"file"id˴ؿfile2func()Ƥ֡
 */

static void prog2file(SpdObjId prog_id)
{
    SpdCursor *file_csr;
    SpdObjId  file_id;
    
    /* "prog_file"δϢ򤿤ɤ롥 */
    file_csr = spdGetRelObjInit(prog_id, "prog_file", "program_id");
    while ((file_id = spdGetRelObj(file_csr)) != SPD_NON_ID) {
        if (debugPrintFlag){
            fprintf(stderr, "file name : %s\n", spdGetName(file_id));
        }
        file2func(file_id);
    }
    spdFreeCursor(file_csr);
}

static void opt_analyze(int argc, char *argv[])
{
    char c;

    while ((c = getopt(argc, argv, "d:Dv")) != EOF) {
        switch (c) {
          case 'd':
            /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'D':
            /* debug: äϢɽ롥 */
            debugPrintFlag = 1;
            break;
          case 'v':
            /* verbose: åξɽ롥 */
            debugPrintFlag2 = 1;
            break;
          default:
            spdErrorExit("'%c': Invalid option.\n", c);
            break;
        }
    }
}

int main(int argc, char *argv[])
{   
    SpdModelId  i_id, fv_id;
    SpdDBId     sdb_id;
    
    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    opt_analyze(argc, argv);

    i_id = spdUseModel(&I_model);
    fv_id = spdUseModel(&Flow_view);
    sdb_id = spdOpenSDB();
    spdReadDB(i_id, sdb_id);
    spdCreateDB(fv_id, sdb_id);
    prog2file(spdGetAnObjId("program", NULL));
    spdMessage("Writing FlowView database.\n");
    spdCommitDB(fv_id, sdb_id);
    spdCloseSDB(sdb_id);
    return (EXIT_SUCCESS);
}
