/* 
 * Program:             $RCSfile: mkList.c,v $  $Revision: 40.14 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/12/11
 *                      S.Yamamoto  2001/12/11
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
"$Id: mkList.c,v 40.14 2003/06/04 16:59:28 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkDD.h"

static topLevelListNode *createTLLNode(void);
static void addTLLNode(SpdObjId objId);
static void addArgs(SpdObjId funcDeclId);
static void addDecls(SpdObjId objId);
static SpdObjId getBranchNext(SpdObjId exprId, SpdBoolean bool);
static SpdObjId evalNormal(SpdObjId objId);
static SpdObjId evalBranch(SpdObjId objId);
static SpdObjId evalSwitch(SpdObjId objId);
static void evalBranchSub(SpdObjId objId);

extern int debugPrintFlag;
extern int debugPrintFlag2;

/* ΥꥹȤϥץ˰Ĥ¸ߤΤǡ
 * Head  Tail ӰѿǤĤ褦ˤƤ롥
 */
static topLevelListNode *topLevelListHead = NULL;
static topLevelListNode *topLevelListTail = NULL;
static int listSize = 0;

/************** ꥹ *****************/

topLevelListNode *getTLLHead(void)
{
    return topLevelListHead;
}

static topLevelListNode *createTLLNode(void)
{
    return (topLevelListNode *)spdMalloc(sizeof(topLevelListNode));
}

topLevelListNode *searchTLLNode(SpdObjId objId)
{
    topLevelListNode *p;

    p = topLevelListHead;
    while(p != NULL) {
        if (p->objId == objId) return p;
        p = p->next;
    }
    return NULL;                            /* ꥹȤϿƤʤ */
}

static void addTLLNode(SpdObjId objId)
{
    topLevelListNode *p;

    if (searchTLLNode(objId) != NULL) {
        if (debugPrintFlag2) {
            fprintf(stderr, "addTLLNode(): %d is already in List.\n", objId);
        }
        return;                             /* ϿƤ롥 */
    }

    p = createTLLNode();
    p->objId = objId;
    p->def = NULL;

    /* λǤsubExprListΥʬʤΤǡ */
    /* def, kill, reach 뤳ȤϤǤʤ     */
    p->kill = NULL;
    p->reach = NULL;

    p->next = NULL;

    if (topLevelListHead == NULL) {
        topLevelListHead = p;
        topLevelListTail = p;
    } else {
        topLevelListTail->next = p;
        topLevelListTail = p;
    }
    listSize++;
    if (debugPrintFlag2) fprintf(stderr, "addTLLNode(): add %d.\n", objId);
}

/* ̤Ѵؿ */

int getTLLSize(void)
{
    return listSize;
}

/* ȥåץ٥뼰ꥹȤ */

void freeTLList(void)
{
    topLevelListNode *p, *next;

    p = topLevelListHead;
    while (p != NULL) {
        if (debugPrintFlag2) fprintf(stderr, "freeTLList: %d.\n", p->objId);
        next = p->next;
        spdFree(p->def);
        spdFree(p->kill);
        spdFree(p->reach);
        spdFree(p);
        p = next;
    }
    topLevelListHead = NULL;
    topLevelListTail = NULL;
    listSize = 0;
}

/***********  **************/

/* ꥹȤ˴ؿΰdeclarationɲä롥 */

static void addArgs(SpdObjId funcDeclId)
{
    int i;
    SpdObjId decrId, optId;
    SpdRelArray declBuf;

    decrId = spdGetARelObj(funcDeclId, "decl_decl", "declaration_id");
    optId = spdGetARelObj(decrId, "decl_opt", "declarator_id");
    declBuf = spdGetRelSortedArray(optId, "func_parm", "optionaldecl_id",
                                   NULL, spdCompareRelOffset);
    for (i = 0; i < declBuf.size; i++) {
        addTLLNode(declBuf.rel[i].objectId);
    }
    spdFreeRelArray(declBuf);
}

/* ꥹȤ˶ɽѿdeclarationɲä롥 */

static void addDecls(SpdObjId objId)
{
    /*
     * objIdCOMP֥åblkEntryidä硤
     * declarationꥹȤˤĤʤ
     */
    if (spdGetClassId(objId) == SPD_CID_BLK_ENTRY &&
        spdGetAttrValInt(objId, "sort") == SPD_FV_B_COMP) {
        int i;
        SpdObjId blkId;
        SpdRelArray declBuf;
    
        blkId = spdGetARelObj(objId, "blk_sdanode", "sdanodeId");
        declBuf = spdGetRelSortedArray(blkId, "blk_decl", "parent_block_id",
                                       NULL, spdCompareRelOffset);
        for (i = 0; i < declBuf.size; i++) {
            addTLLNode(declBuf.rel[i].objectId);
        }
        spdFreeRelArray(declBuf);
    }
}

static SpdObjId getBranchNext(SpdObjId exprId, SpdBoolean bool)
{
    SpdCursor *nextNode_cursor;
    SpdRel rel;
    SpdObjId nextNodeId = SPD_NON_ID;

    nextNode_cursor = spdGetRelInit(exprId, "next_node", "curId");
    do {
        int edgeSort;

        rel = spdGetRel(nextNode_cursor);
        edgeSort = spdGetAttrValInt(rel.relationId, "sort");
        if (edgeSort == SPD_FV_E_BRANCH_TRUE && bool == TRUE) {
            nextNodeId = rel.objectId;
            break;
        } else if (edgeSort == SPD_FV_E_BRANCH_FALSE && bool == FALSE) {
            nextNodeId =  rel.objectId;
            break;
        }
    } while (rel.objectId != SPD_NON_ID);
    spdFreeCursor(nextNode_cursor);

    return nextNodeId;
}

/*
 * ʬ̵Ȥ롥
 *  blkEntryidꤷƤ롥
 */

static SpdObjId evalNormal(SpdObjId objId)
{
    SpdObjId curId;
    SpdId classId;

    /* ܥ֥å˰Ĥexpression */
    if (spdGetClassId(objId) == SPD_CID_EXPRESSION) {
        int sort;
        addTLLNode(objId);
        sort = spdGetAttrValInt(objId, "sort");
        if (sort == EXPR_BREAK) {
            return SPD_NON_ID;
        } else if (sort == EXPR_CONTINUE) {
            return SPD_NON_ID;
        } else if (sort == EXPR_RETURN) {
            return SPD_NON_ID;
        } else if (sort == EXPR_GOTO) {
            /* ̤ɬפΤ٥Ͽ롥 */
            return SPD_NON_ID;
        }
        return SPD_NON_ID;
    }

    addDecls(objId);
  
    curId = spdGetARelObj(objId, "next_node", "curId");
    for(;;) {
        classId = spdGetClassId(curId);
        if (classId == SPD_CID_EXPRESSION) {
            int sort;
      
            addTLLNode(curId);
            sort = spdGetAttrValInt(curId, "sort");
            if (sort == EXPR_BREAK) {
                return SPD_NON_ID;
            } else if (sort == EXPR_CONTINUE) {
                return SPD_NON_ID;
            } else if (sort == EXPR_RETURN) {
                return SPD_NON_ID;
            } else if (sort == EXPR_GOTO) {
                /* ̤ɬפΤ٥Ͽ롥 */
                return SPD_NON_ID;
            }      
      
        } else if (classId == SPD_CID_LABEL) {
            /* ٥̤äȤϿ롥  */
        } else if (classId == SPD_CID_BLK_ENTRY) {
            /* declarationΥꥹȤ */
            /* ֥åμ򸫤evalNormalޤevalBranchƤ֡ */
      
            switch (spdGetAttrValInt(curId, "sort")) {
              case SPD_FV_B_COMP:
                curId = evalNormal(curId);
                if (curId == SPD_NON_ID) {
                    return SPD_NON_ID;    /* returnȵ */
                }
                break;
              case SPD_FV_B_IF:
              case SPD_FV_B_FOR:
              case SPD_FV_B_WHILE:
              case SPD_FV_B_DO:
                curId = evalBranch(curId);
                break;
              case SPD_FV_B_SWITCH:
                curId = evalSwitch(curId);
            }
      
        } else if (classId == SPD_CID_BLK_EXIT) {
            break;
        }
  
        curId = spdGetARelObj(curId, "next_node", "curId");
    }

    return curId;                           /* BLK_EXITid֤ */
}

static SpdObjId evalBranch(SpdObjId objId)
{
    int sort;
    SpdObjId curId, branchExprId, blkId;

    if (spdGetClassId(objId) != SPD_CID_BLK_ENTRY) {
        spdErrorExit("evalBranch().\n");
    }
    curId = objId;
    sort = spdGetAttrValInt(objId ,"sort");
    switch (sort) {
      case SPD_FV_B_FOR:
        curId = spdGetARelObj(curId, "next_node", "curId");
        addTLLNode(curId);                  /*  */
        /* ³ */
      case SPD_FV_B_IF:
      case SPD_FV_B_WHILE:
        curId = spdGetARelObj(curId, "next_node", "curId");
        addTLLNode(curId);                  /* Ｐ */
        break;
      case SPD_FV_B_DO:
        break;
      default:
        spdErrorExit("evalBranch().\n");
    }

    if (sort == SPD_FV_B_FOR || sort == SPD_FV_B_IF || sort == SPD_FV_B_WHILE) {
        branchExprId = curId;
        curId = getBranchNext(branchExprId, TRUE);
        /* evalNormal(curId);*/
        /* if(A) do { xxx } while(B) Τ褦ʥбĤꡥ  */
        evalBranchSub(curId);
    } else if (sort == SPD_FV_B_DO) {
        curId = spdGetARelObj(curId, "next_node", "curId");
        /* evalNormal(curId); */
        /* if(A) do { xxx } while(B) Τ褦ʥбĤꡥ  */
        evalBranchSub(curId);
        
        /* Ｐ */
        blkId = spdGetARelObj(objId, "blk_sdanode", "sdanodeId");
        addTLLNode(spdGetARelObj(blkId, "blk_expr", "parent_block_id"));
    }

    if (sort == SPD_FV_B_IF) {
        curId = getBranchNext(branchExprId, FALSE);
        /*
         * else if б 
         * elseʤб
         */
        if (spdGetClassId(curId) == SPD_CID_BLK_ENTRY &&
            spdGetAttrValInt(curId, "sort") == SPD_FV_B_IF) {
            evalBranch(curId);
        } else if (spdGetClassId(curId) != SPD_CID_BLK_EXIT) {
            /* evalNormal(curId); */
            /* if(A) do { xxx } while(B) Τ褦ʥбĤꡥ  */
            evalBranchSub(curId);
        }
    }

    if (sort == SPD_FV_B_FOR) {
        /*  */
        SpdCursor *preObjCursor;
        SpdRel preRel;

        preObjCursor = spdGetRelInit(branchExprId, "next_node", "nextId");
        for (;;) {
            preRel = spdGetRel(preObjCursor);
            if (preRel.objectId == SPD_NON_ID) break;
            if (spdGetAttrValInt(preRel.relationId, "sort") ==
                SPD_FV_E_LOOPBACK)
                break;
        }
        spdFreeCursor(preObjCursor);
        addTLLNode(preRel.objectId);
    }

    blkId = spdGetARelObj(objId, "blk_sdanode", "sdanodeId");
    return getFVNode(blkId, SPD_CID_BLK_EXIT); /* BLK_EXITid֤ */
}

static SpdObjId evalSwitch(SpdObjId objId)
{
    int sort;
    SpdObjId curId, blkId;
    SpdCursor *nextCursor;
    SpdRel rel;

    if (spdGetClassId(objId) != SPD_CID_BLK_ENTRY) {
        spdErrorExit("evalSwitch().\n");
    }
    curId = objId;
    sort = spdGetAttrValInt(objId ,"sort");
    if (sort != SPD_FV_B_SWITCH) {
        spdErrorExit("evalSwitch().\n");
    }

    curId = spdGetARelObj(curId, "next_node", "curId");
    addTLLNode(curId);                      /*  */

    curId = spdGetARelObj(curId, "next_node", "curId");
    /* switch(a)b++; Ȥξ */
    if (spdGetClassId(curId) == SPD_CID_BLK_EXIT) {
        return curId;
    }
  
    addDecls(curId);
    curId = spdGetARelObj(curId, "next_node", "curId"); /* switchBranch */
    nextCursor = spdGetRelInit(curId, "next_node", "curId");
    do {
        rel = spdGetRel(nextCursor);
   
        if (spdGetAttrValInt(rel.relationId, "sort") == SPD_FV_E_SWITCH_CASE ||
            spdGetAttrValInt(rel.relationId, "sort") == SPD_FV_E_SWITCH_DEFAULT) {
            evalNormal(rel.objectId);
        }
    } while (rel.objectId != SPD_NON_ID);
    spdFreeCursor(nextCursor);

    blkId = spdGetARelObj(objId, "blk_sdanode", "sdanodeId");
    return getFVNode(blkId, SPD_CID_BLK_EXIT); /* BLK_EXITid֤ */
}

static void evalBranchSub(SpdObjId curId)
{
    /* "if(A) do { xxx } while(B)"Τ褦ʥб  */
    if (spdGetClassId(curId) == SPD_CID_EXPRESSION) {
        evalNormal(curId);
        return;
    }
    
    /* ֥åμ򸫤evalNormalޤevalBranchƤ֡ */
    switch (spdGetAttrValInt(curId, "sort")) {
      case SPD_FV_B_COMP:
        evalNormal(curId);
        break;
      case SPD_FV_B_IF:
      case SPD_FV_B_FOR:
      case SPD_FV_B_WHILE:
      case SPD_FV_B_DO:
        evalBranch(curId);
        break;
      case SPD_FV_B_SWITCH:
        evalSwitch(curId);
        break;
      default:
        break;
    }
}

/* ȥåץ٥ExpressionΥꥹȤ롥 */

void mkTLList(SpdObjId declId)
{
    SpdObjId funcEntryId;

    addArgs(declId);
  
    funcEntryId = getFVNode(declId, SPD_CID_BLK_ENTRY);
    evalNormal(funcEntryId);
}
