/* 
 * Program:             $RCSfile: mkSubList.c,v $  $Revision: 40.10 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/12/11
 *                      S.Yamamoto  2001/12/11
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkSubList.c,v 40.10 2003/06/04 16:59:28 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkDD.h"

static subExprListNode *createSLNode(void);
static void addSLNode(SpdObjId objId, SpdObjId identId, SpdBoolean withPtrFlag,
              SpdBoolean defUnkownAreaFlag,
              SpdBoolean killFlag, SpdObjId topLevelObjId);
static SpdObjId isTypePointerExpr(SpdObjId exprId);
static SpdObjId getExprInfoSub(SpdObjId exprId);
static void analExpr(SpdObjId exprId, SpdObjId topObjId, SpdBoolean killFlag);
static void analTopLevelObj(SpdObjId objId);

static SpdObjId skipPar(SpdObjId exprId);
static SpdObjId getDefExprId(SpdObjId exprId);

extern int debugPrintFlag;
extern int debugPrintFlag2;

/* ΥꥹȤϥץ˰Ĥ¸ߤΤǡ
 * Head  Tail ѿǤĤ褦ˤƤ롥
 */
static subExprListNode *subExprListHead = NULL;
static subExprListNode *subExprListTail = NULL;
static int listSize = 0;

/************** ꥹ *****************/

subExprListNode *getSLHead(void)
{
    return subExprListHead;
}

static subExprListNode *createSLNode(void)
{
    return (subExprListNode *)spdMalloc(sizeof(subExprListNode));
}

subExprListNode *searchSLNode(SpdObjId objId)
{
    subExprListNode *p;

    p = subExprListHead;
    while(p != NULL) {
        if (p->objId == objId) return p;
        p = p->next;
    }
    return NULL;
}

int searchSLNodeNumber(SpdObjId objId)
{
    int i = 0;
    subExprListNode *p;

    p = subExprListHead;
    while (p != NULL) {
        if (p->objId == objId) return i;
        p = p->next;
        i++;
    }
    return -1;
}

static void addSLNode(SpdObjId objId, SpdObjId identId, SpdBoolean withPtrFlag,
                      SpdBoolean defUnkownAreaFlag, SpdBoolean killFlag,
                      SpdObjId topLevelObjId)
{
    subExprListNode *p;

    if (searchSLNode(objId) != NULL) {
        fprintf(stderr, "addSLNode: already in List.\n");
        return;                             /* ϿƤ롥 */
    }

    p = createSLNode();
    p->objId = objId;
    p->identId = identId;
    p->withPtrFlag = withPtrFlag;
    p->defUnkownAreaFlag = defUnkownAreaFlag;
    p->killFlag = killFlag;
    p->topLevelObjId = topLevelObjId;
    p->next = NULL;

    if (subExprListHead == NULL) {
        subExprListHead = p;
        subExprListTail = p;
    } else {
        subExprListTail->next = p;
        subExprListTail = p;
    }
    listSize++;
    if (debugPrintFlag2) {
        fprintf(stderr, "addSLNode(): add %d\n", objId);
        fprintf(stderr, "  identId %d  withPtrFlag %d",
                identId, withPtrFlag ? 1 : 0);
        fprintf(stderr, "  defUnknownAreaFlag %d\n", defUnkownAreaFlag);
        fprintf(stderr, "  killFlag %d  topLevelObjId %d\n",
                killFlag ? 1 : 0, topLevelObjId);
    }
}

int getSLSize(void)
{
    return listSize;
}

void freeSList(void)
{
    subExprListNode *p, *next;

    p = subExprListHead;
    while (p != NULL) {
        if (debugPrintFlag2) fprintf(stderr, "freeSList: %d\n", p->objId);
        next = p->next;
        spdFree(p);
        p = next;
    }
    subExprListHead = NULL;
    subExprListTail = NULL;
    listSize = 0;
}

/***********  **************/

/* exprId η pointer ɤĴ٤롥 */
/* ΤȤźݥ󥿤ȤߤʤƤ롥*/

static SpdObjId isTypePointerExpr(SpdObjId exprId)
{
    SpdObjId retType;

    retType = spdGetARelObj(exprId, "expr_type", "expression_id");
    if (spdGetAttrValInt(retType, "sort") == TYPE_POINTER ||
        spdGetAttrValInt(retType, "sort") == TYPE_ARRAY) {
        return TRUE;
    } else {
        return FALSE;
    }
}

/* åФ*/

static SpdObjId skipPar(SpdObjId exprId)
{
    SpdRelArray exprBuf;
    SpdObjId cExprId, opId;

    cExprId = exprId;
    while (cExprId != SPD_NON_ID) {
        int opSort;
        opId = spdGetARelObj(cExprId, "expr_op", "expression_id");
        opSort = spdGetAttrValInt(opId, "sort");
        if (opSort == OP_L_PAR || opSort == OP_R_PAR) {
            exprBuf = spdGetRelSortedArray(cExprId, "expr_expr",
                                           "outer_expression_id",
                                           NULL, spdCompareRelOffset);
            cExprId = exprBuf.rel[0].objectId;
            spdFreeRelArray(exprBuf);
        } else {
            break;
        }
    }
    return cExprId;
}

/* ݥ󥿤ȤäǡɤΥݥ󥿤򸵤˥ɥ쥹Ƥ뤫 */
/* ǽξϡSPD_NON_ID ޤ -2 ֤
 * (Ƶ꤯񤱤ʤƻʤι¤)
 */
static SpdObjId getExprInfoSub(SpdObjId exprId)
{
    int i;
    SpdRelArray exprBuf;
    SpdObjId identId;

    if (spdGetClassId(exprId) != SPD_CID_EXPRESSION) {
        spdWarning("getDefIdent(): exprId %d, classId %d: Not expression.\n",
                   exprId, spdGetClassId(exprId));
        return SPD_NON_ID;                /* ȤꤢSPD_NON_ID֤ */
    }

    exprBuf = spdGetRelSortedArray(exprId, "expr_expr", "outer_expression_id",
                                   NULL, spdCompareRelOffset);
    for (i = 0; i < exprBuf.size; i++) {
        SpdObjId cExprId, opId;

        cExprId = skipPar(exprBuf.rel[i].objectId);
        opId = spdGetARelObj(cExprId, "expr_op", "expression_id");
        identId = spdGetARelObj(cExprId, "ident_ref", "expression_id");

        /* ݥ󥿷μɤ */
        if (isTypePointerExpr(cExprId) == TRUE) {
            /* ݥ󥿻Ȥˤäơɥ쥹ȤƤ硥 */
            switch (spdGetAttrValInt(opId, "sort")) {
              case OP_POINTER:
              case OP_L_ARRAY:
              case OP_R_ARRAY:
              case OP_ARROW:
              case OP_DOT:
                spdFreeRelArray(exprBuf);
                return SPD_NON_ID;
                break;
            }
            /* ؿƽФˤäơɥ쥹ȤƤ硥 */
            if (spdGetAttrValInt(cExprId, "sort") == EXPR_SORT_FUNCCALL) {
                spdFreeRelArray(exprBuf);
                return SPD_NON_ID;
            }
            if (identId != SPD_NON_ID) {
                spdFreeRelArray(exprBuf);
                return identId;
            }
        }
        if ((identId = getExprInfoSub(cExprId)) != -2) return identId;
    }
    spdFreeRelArray(exprBuf);
    /*return SPD_NON_ID;*/
    /* ʤˤ⳺Τդʤä硥 */
    return -2;
} 

/* Dot Фϡ*killFlag FALSE ˤʤ롥
 * ¤ΤΥФ̤ǤƤʤǤ롥
 */

SpdObjId skipParAndDot(SpdObjId exprId, SpdBoolean *killFlag)
{
    SpdRelArray exprBuf;
    SpdObjId cExprId, opId;

    /* skip. */
    cExprId = exprId;
    while (cExprId != SPD_NON_ID) {
        int opSort;
        opId = spdGetARelObj(cExprId, "expr_op", "expression_id");
        opSort = spdGetAttrValInt(opId, "sort");
        if (opSort == OP_L_PAR || opSort == OP_R_PAR || opSort == OP_DOT) {
            exprBuf = spdGetRelSortedArray(cExprId, "expr_expr",
                                           "outer_expression_id",
                                           NULL, spdCompareRelOffset);
            cExprId = exprBuf.rel[0].objectId;
            spdFreeRelArray(exprBuf);
            if (opSort == OP_DOT && killFlag != NULL) {
                /* ¤ΤϥС¿뤫顥 */
                *killFlag = FALSE;
            }
        } else {
            break;
        }
    }
    return cExprId;
}

/* 
 *  κռͿ졤ξ֤("a = 10""a"ʬ)
 *
 *  identId()ɤIdentifierؤΤ֤
 *  withPtrFlag"*p"Τ褦˥ݥ󥿤𤷤Ƥ򼨤ե饰
 * (withPtrFlagǰ)
 *  
 *  defUnknownAreaFlag"**p"Τ褦˲ϤǤʤΰ򤵤򼨤ե饰
 *  Υե饰ϸ夫դäΤǡޤͭ˻ѤƤʤ⤷ʤ
 * 
 * ͤȤKillFlag֤
 */

SpdBoolean getExprInfo(SpdObjId exprId, SpdObjId *identId,
                       SpdBoolean *withPtrFlag,
                       SpdBoolean *defUnkownAreaFlag)
{
    SpdObjId cExprId, opId;
    SpdBoolean killFlag = TRUE;

    if (spdGetClassId(exprId) != SPD_CID_EXPRESSION) {
        spdWarning("getExprInfo(): exprId %d, classId %d: Not expression.\n",
                   exprId, spdGetClassId(exprId));
        /* ȤꤢkillFlag (TRUE)֤ */
        return killFlag;
    }

    cExprId = skipParAndDot(exprId, &killFlag);
  
    opId = spdGetARelObj(cExprId, "expr_op", "expression_id");
    if (opId == SPD_NON_ID) {
        /* ݥ󥿻̵硥 */
        *identId = spdGetARelObj(cExprId, "ident_ref", "expression_id");
        *withPtrFlag = FALSE;
        *defUnkownAreaFlag = FALSE;
        /* *killFlagʤFALSEǤ٤TRUEˤƤϤޤ顥 */
        return killFlag;                    /* killFlag */
    } else {
        /* ݥ󥿻Ȥξ硥 */
        switch (spdGetAttrValInt(opId, "sort")) {
          case OP_POINTER:
          case OP_L_ARRAY:
          case OP_R_ARRAY:
          case OP_ARROW:
            *withPtrFlag = TRUE;
            *identId = getExprInfoSub(cExprId);
            if (*identId == SPD_NON_ID) {
                /* identId ޤʤä硤defUnkownAreaFlag ΩƤ롥 */
                *defUnkownAreaFlag = TRUE;
            } else if (*identId == -2) {
                *defUnkownAreaFlag = TRUE;
                *identId = SPD_NON_ID;    /* -2SPD_NON_ID롥 */
            } else {
                /* identId ޤä硥 */
                *defUnkownAreaFlag = FALSE;
            }
      
            return FALSE;                   /* killFlag */
            break;
          default:
            spdWarning("No Match.");
            fprintf(stderr, "No Match: Op is %d.\n", opId);
            /* Ȥꤢ֤ͤϥåȤ롥 */
            return killFlag;
        }
    }
}

/* id顤ռid֤ */

static SpdObjId getDefExprId(SpdObjId exprId)
{
    SpdRelArray exprBuf;
    SpdObjId cExprId;
  
    /* left value expression */
    exprBuf = spdGetRelSortedArray(exprId, "expr_expr",
                                   "outer_expression_id",
                                   NULL, spdCompareRelOffset);
    cExprId = exprBuf.rel[0].objectId;
    spdFreeRelArray(exprBuf);

    return cExprId;
}

/* 
 * ͤʬ򤵤
 *  killFlag ϡǽanalTopLevelObj()Ϥ줿ͤȡ
 * "&&", "||"ʤɤʬˤäƤΤ(¿ʬ)ޤΤǤ롥
 * ʬȤkillFlagȤϴĤʤΤա
 */

static void analExpr(SpdObjId exprId, SpdObjId topObjId, SpdBoolean killFlag)
{
    int i;
    SpdRelArray exprBuf;
    SpdObjId opId, identId, cExprId;
    SpdBoolean withPtrFlag, defUnkownAreaFlag, secondKillFlag, tempKillFlag;
    /* secondKillFlagϡ&& || ʬǣܰʹߤʬΤkillFlag */

    /* exprIdͿ줿Ĵ٤롥 */
    opId = spdGetARelObj(exprId, "expr_op", "expression_id");

    /* 黻Ҥʤξ硥 */
    if (opId == SPD_NON_ID) {
        int     sort = spdGetAttrValInt(exprId, "sort");

        /* returnȴؿϡͤΤμʬĤΤ
         * ³ڹ¤򤿤ɤɬפ롥
         * ʳμʤ顤üȤߤʤ
         */
        if (sort != EXPR_RETURN && sort != EXPR_SORT_FUNCCALL) {
            return;
        }
    
        /* 黻Ҥļξ硥 */
    } else {
        secondKillFlag = killFlag;          /*  */

        switch (spdGetAttrValInt(opId, "sort")) {
            /* ͤ뼰 */
          case OP_INC:
          case OP_DEC:
          case OP_ASSIGN:
          case OP_A_ADD:
          case OP_A_SUB:
          case OP_A_MUL:
          case OP_A_DIV:
          case OP_A_MOD:
          case OP_A_AND:
          case OP_A_XOR:
          case OP_A_OR:
          case OP_A_SHL:
          case OP_A_SHR:
            /* killFlag 1λΤ tempKillFlag ͤȤ*/
            /* ĤޤꡤƤkillʤʤҤkillʤ */
            cExprId = getDefExprId(exprId);
            tempKillFlag = getExprInfo(cExprId, &identId,
                                       &withPtrFlag, &defUnkownAreaFlag);
            addSLNode(exprId, identId, withPtrFlag, defUnkownAreaFlag,
                      killFlag ? tempKillFlag : killFlag, topObjId);
            break;
          case OP_L_AND:
          case OP_L_OR:
          case OP_QUESTION:
          case OP_COLON:
            secondKillFlag = FALSE;
            break;
          default:
            break;
        }
    }

    exprBuf = spdGetRelSortedArray(exprId, "expr_expr", "outer_expression_id",
                                   NULL, spdCompareRelOffset);
    for (i = 0; i < exprBuf.size; i++) {
        analExpr(exprBuf.rel[i].objectId, topObjId,
                 i ? secondKillFlag : killFlag);
    }
    spdFreeRelArray(exprBuf);
}

/* 
 * ȥåץ٥뼰ʬexprʬ򤹤ؿ
 * expression  declarationid롥
 */

static void analTopLevelObj(SpdObjId objId)
{
    /* expression. */
    if (spdGetClassId(objId) == SPD_CID_EXPRESSION) {
        analExpr(objId, objId, TRUE);

        /* declaration. */
    } else if (spdGetClassId(objId) == SPD_CID_DECLARATION) {
        int i;
        SpdRelArray decrBuf;
        SpdBoolean killFlag = TRUE;

        /*
         * static, extern ѿưŪ˽ʤΤǡ
         * killFlag = 0 ˤ롥
         */
        if (spdGetAttrValInt(objId, "storage") == STORAGE_EXTERN ||
            spdGetAttrValInt(objId, "storage") == STORAGE_STATIC) {
            killFlag = FALSE;
        }
    
        decrBuf = spdGetRelSortedArray(objId, "decl_decl", "declaration_id",
                                       NULL, spdCompareRelOffset);
        for (i = 0; i < decrBuf.size; i++) {
            SpdObjId decrId, optId, identId;

            decrId = decrBuf.rel[i].objectId;
            optId = spdGetARelObj(decrId, "decl_opt", "declarator_id");
            if (optId  == SPD_NON_ID) {
                /* ʤʪ̵ȤɽʤȤʤ
                 * ⤷ʤ϶̤Ƥʤ
                 */
                identId = spdGetARelObj(decrId ,"decl_ident", "declarator_id");
                addSLNode(decrId, identId, FALSE, FALSE, killFlag, objId);
            } else {
                /* դdeclarator򥵥֥ꥹȤϿ */
                identId = spdGetARelObj(decrId ,"decl_ident", "declarator_id");
                addSLNode(decrId, identId, FALSE, FALSE, killFlag, objId);

                /* Ĵ٤롥 */
                /* staticνϡkillFlagFALSEˤʤȤʤ(?) */
                analExpr(spdGetARelObj(optId, "init_expr", "optionaldecl_id"),
                         objId, TRUE        /* killFlag (?)*/);
            }
        }
        spdFreeRelArray(decrBuf);
    }
}

/* ͤʬΥꥹȤ롥 */

void mkSList(void)
{
    topLevelListNode *curNode;

    curNode = getTLLHead();
    while (curNode != NULL) {
        /* fprintf(stderr, "mkSList():curNode->objId %d\n", curNode->objId); */
        analTopLevelObj(curNode->objId);
        curNode = curNode->next;
    }
}
