/* 
 * Program:             $RCSfile: mkPfileFile.c,v $  $Revision: 31.9 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohhashi  1999/11/27
 *                      S.Yamamoto  1999/11/27
 * 
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  1999 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkPfileFile.c,v 31.9 2003/06/28 15:32:08 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static void printUsage(SpdConstString prog_name);
static SpdString init_program(SpdString prog_path);
static void analyzeOption(int argc, char *argv[]);
static int getSDBinfo(SpdString className);
static void writePfileFile(void);

/*
 * Initialization.
 */

static SpdString init_program(SpdString prog_path)
{
    static SpdString program_name;

    if ((program_name = strrchr(prog_path, '/')) != NULL) {
        program_name++;
    } else {
        program_name = prog_path;
    }
    spdInitialize(program_name);

    return program_name;
}

/*
 * Analyze option.
 */

static void analyzeOption(int argc, char *argv[])
{
    int c;
    SpdString prog_name;

    prog_name = init_program(argv[0]);

    while ((c = getopt(argc, argv, "d:")) != EOF) {
        switch (c) {
          case 'd':
            spdSetSDBdir(optarg);
            break;
          default:
            printUsage(prog_name);
            break;
        }
    }
}

/*
 * Set read SDB format & return pfile objects number.
 */

static int getSDBinfo(SpdString className)
{
    int     obj_num;
    SpdInfo *info = spdOpenSDBinfo();

    /* Count pfile objects. */
    obj_num = spdGetInfoInt(info, className, 0);

    /* Set read SDB format . */
    if (SPD_STREQ(spdGetInfoString(info, "BINARY_FORMAT_SDB"), "TRUE")) {
        spdSetReadSDBformat(SAPID_BINARY_FORMAT_SDB);
    } else {
        spdSetReadSDBformat(SAPID_TEXT_FORMAT_SDB);
    }
    spdCloseInfo(info);

    return (obj_num);
}

/*
 * Write "<SDB>/SRC/pfile_file".
 */

static void writePfileFile(void)
{
    FILE        *fp_pfile, *fp_pfile_file;
    int         obj_num;
    SpdDBObj    pfile_obj;
    SpdString   class_name = spdGetClassNameById(SPD_CID_PFILE);
    SpdObjId    file_id;
    int         attr_id, attr_name;

    obj_num = getSDBinfo(class_name);

    fp_pfile = spdOpenSDBfile(P_MODEL_DIR, class_name, SAPID_O_READ);
    fp_pfile_file = spdOpenSDBfile(SAPID_SRC_DIR, SRC_PFILE_FILE,
                                   SAPID_O_WRITE);

    /* Read pfile obj, search fileId by name and write table. */
    attr_id = spdGetAttrIdByName3(SPD_CID_PFILE, class_name, "id");
    attr_name = spdGetAttrIdByName3(SPD_CID_PFILE, class_name, "name") -
        SpdObjIValN;
    while (obj_num--) {
        pfile_obj = spdReadObj(fp_pfile, SPD_CID_PFILE);
        file_id = spdSearchGlobalObj("file", pfile_obj.svalue[attr_name]);
    
        if (file_id == SPD_NON_ID) {
            spdWarning("Can't search file Id for pfileId(%d).\n",
                       pfile_obj.ivalue[attr_id]);
        }
        fprintf(fp_pfile_file, "%d %d\n", pfile_obj.ivalue[attr_id], file_id);

        spdFreeStrsInObj(pfile_obj);
    }

    spdFclose(fp_pfile);
    spdFclose(fp_pfile_file);
}

/*
 * Main.
 */

int main(int argc, char *argv[])
{
    analyzeOption(argc, argv);
    writePfileFile();
    return (EXIT_SUCCESS);
}

/*
 * Print usage message.
 */

static void printUsage(SpdConstString command_name)
{
    SpdConstString usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [options] [programName]\n"
        "DESCRIPTION\n"
        "    -d SDBdirectory.\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
