/* 
 * Program:             $RCSfile: block.c,v $  $Revision: 30.23 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: block.c,v 30.23 2003/07/20 11:46:07 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

#define MATCH_OK       0
#define MATCH_FAIL     1
#define MATCH_SKIP     2
#define MATCH_ABORT    3
#define MATCH_PADDING  4
#define MATCH_CONTINUE 5

/* ȡ1ļ롥length ˤ \ + Ԥޤ */

static void getIntToken(FileBuffer *fb, MyCursor *paramcsr, PPRegion *reg)
{
    int i, c;
    skipSpace(fb, paramcsr);

    reg->offset = paramcsr->current.offset;

    for (i = 0; i < MAX_TOKEN_LEN && isdigit(c = peekChar(fb, paramcsr));
         i++) {
        reg->str[i] = c;
        nextChar(fb, paramcsr);
    }
    reg->str[i] = '\0';

    reg->length = paramcsr->current.offset - reg->offset;
}

/* ʸȡ1ļ롥length ˤ \ + Ԥޤ */

static void getStringToken(FileBuffer *fb, MyCursor *paramcsr, PPRegion *reg)
{
    int i, c;
    skipSpace(fb, paramcsr);

    reg->offset = paramcsr->current.offset;

    if ((c =peekChar(fb, paramcsr)) == '"') {
        reg->str[0] = c;
        nextChar(fb, paramcsr);
    }

    /* !! ѥޥνͤƤʤΤǡ
      Ȥꤢ \" ϹθƤʤ */
    for (i = 1; i < MAX_TOKEN_LEN && (c = peekChar(fb, paramcsr)) != '"';
         i++) {
        reg->str[i] = c;
        nextChar(fb, paramcsr);
    }
    reg->str[i] = '"';
    nextChar(fb, paramcsr);
    reg->str[i+1] = '\0';

    reg->length = paramcsr->current.offset - reg->offset;
}

/* __FILE__ ʤɤν */

static SpdBoolean parsePredefinedMacro(CFileBuffer *cFileBuf, 
                                       IFileBuffer *iFileBuf)
{
    static struct predef_table_t {
        char *name;
        enum { predef_string, predef_int} type;
    } predefs[] = {
        { "__LINE__", predef_int },
        { "__FILE__", predef_string },
        { "__DATE__", predef_string },
        { "__TIME__", predef_string },
        { "__STDC__", predef_int },
        { NULL, predef_int }
    };

    int i;
    PPRegion creg, ireg;

    creg.offset = cFileBuf->csr.current.offset;
    ireg.offset = iFileBuf->csr.current.offset;
    getToken(&(cFileBuf->fb), &(cFileBuf->csr), &creg);
    
    for (i = 0; predefs[i].name != NULL; i++) {
        if (SPD_STREQ(predefs[i].name, creg.str)) 
            break;
    }

    if (predefs[i].name) {
        if (predefs[i].type == predef_int) {
            getIntToken(&(iFileBuf->fb), &(iFileBuf->csr), &ireg);
        } else {
            getStringToken(&(iFileBuf->fb), &(iFileBuf->csr), &ireg);
        }

        /* !! pblock٤? */

        return TRUE;
    }

    return FALSE;
}

/* 1ʬΥޥå󥰤Ԥ */

static int matchLine(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf, int mlres,
                     SpdBoolean flag)
{
    int res;
    int clines;

    /* ͤͤƱ */
    res = mlres;

    if (verbose > 3) {
        spdMessage("matchLine: from: %s(%d):%d -> %d, %s(%d):%d\n",
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                   cFileBuf->csr.current.offset, cFileBuf->nextBlock.offset,
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno,
                   iFileBuf->csr.current.offset);
    }

    clines = 0;
    if (isSkipedLine(cFileBuf, iFileBuf)) {
        /* .i ե뤬ԤǤ .c ե뤬ԤǤϤʤȤ
           ŸʤäԤȤߤʤ */
        res = MATCH_SKIP;

        if (verbose > 3) {
            spdMessage("matchLine : There's not expand text.\n");
        }

        if (cFileBuf->csr.next.offset <= cFileBuf->nextBlock.offset) {
            /* ιԤޤ block ³Ƥ */
            /* .c Ǥ 1ԤɤФʪԿ */
            if (verbose > 3) {
                spdMessage("matchLine : skipline %s(%d), %s(%d).\n",
                           cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                           iFileBuf->fb.filename, iFileBuf->csr.current.lineno);
            }
            clines = skipLogicalLine(&(cFileBuf->fb), &(cFileBuf->csr), 1);
            skipILine(cFileBuf, iFileBuf, clines);
        } else {
            /* block ԤǽäƤν (ŪˤꤨʤϤ) */
            if (verbose > 3) {
                spdMessage("matchLine : skipToOffset from %s(%d) to (%d).\n",
                           cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                           cFileBuf->nextBlock.offset);
            }
            skipToOffset(&(cFileBuf->fb), &(cFileBuf->csr),
                         cFileBuf->nextBlock.offset);
        }
    } else {
        /* Ÿʸ */
        /* .c եԤ .i եԤ */

        while (cFileBuf->csr.current.offset < cFileBuf->csr.next.offset &&
               iFileBuf->csr.current.offset < iFileBuf->csr.next.offset &&
               cFileBuf->csr.current.offset < cFileBuf->nextBlock.offset) {
            if (peekChar(&(cFileBuf->fb), &(cFileBuf->csr)) != 
                peekChar(&(iFileBuf->fb), &(iFileBuf->csr))) {

                /* ޥå󥰤˼Ԥ硤__FILE__ ʤɤ¸ߤʤ
                   Ĵ٤뤿˰ö롥*/
                if (peekChar(&(cFileBuf->fb), &(cFileBuf->csr)) == '_') {
                    return MATCH_ABORT;
                }

                if (verbose > 5) {
                    spdMessage("%s(%d)[%d] '%c' and %s(%d)[%d] '%c'\n",
                               cFileBuf->fb.filename,
                               cFileBuf->csr.current.lineno,
                               cFileBuf->csr.current.offset,
                               cFileBuf->fb.buf[cFileBuf->csr.current.offset], 
                               iFileBuf->fb.filename,
                               iFileBuf->csr.current.lineno,
                               iFileBuf->csr.current.offset,
                               iFileBuf->fb.buf[iFileBuf->csr.current.offset]);
                }

                res = MATCH_FAIL;
                break;
            }
            
            nextChar(&(cFileBuf->fb), &(cFileBuf->csr));
            nextChar(&(iFileBuf->fb), &(iFileBuf->csr));
            /* \ + Ԥå */
            if (flag && isContinueLine(&(cFileBuf->fb), &(cFileBuf->csr))) {
                if (verbose > 3) {
                    spdMessage("matchLine : There's continue line at %s(%d).\n",
                               cFileBuf->fb.filename, 
                               cFileBuf->csr.current.lineno);
                }
                iFileBuf->skip.fpadding = TRUE;
                res = MATCH_CONTINUE;
                break;
            }
        }

        if (cFileBuf->csr.current.offset == cFileBuf->csr.next.offset) {
            /* ޤã */
            /* .c Ǥ 1ԤɤФʪԿ */
            if (verbose > 3) {
                spdMessage("matchLine : There's expand text (continued). %s(%d), %s(%d).\n",
                           cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                           iFileBuf->fb.filename, iFileBuf->csr.current.lineno);
            }
            clines = skipLogicalLine(&(cFileBuf->fb), &(cFileBuf->csr), 1);
            skipILine(cFileBuf, iFileBuf, clines);

            if (iFileBuf->skip.fpadding) {
                /* block ڤ */
                if (verbose > 3) {
                    spdMessage("matchLine : Abort for padding lines.\n");
                }
                iFileBuf->skip.fpadding = FALSE;
        
                res = MATCH_PADDING;
            }
        } else {
            /* Ԥǽä */
            if (verbose > 3) {
                spdMessage("matchLine : There's expand text.\n");
                spdMessage("%s:: current: %d next: %d   %s:: current:%d next:%d\n",
                           cFileBuf->fb.filename, cFileBuf->csr.current.offset,
                           cFileBuf->csr.next.offset,
                           iFileBuf->fb.filename, iFileBuf->csr.current.offset,
                           iFileBuf->csr.next.offset);
            }
        }
    }
    if (verbose > 3) {
        spdMessage("matchLine: to: %s(%d):%d -> %d, %s(%d):%d\n",
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                   cFileBuf->csr.current.offset, cFileBuf->nextBlock.offset,
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno,
                   iFileBuf->csr.current.offset);
    }
    
    return res;
}

/* normal / comment block Υޥå󥰤Ȥ */

static SpdObjId matchNormalBlock(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf,
                                 int sort)
{
    SpdObjId pblockId, iblockId;
    SpdRelId relId;
    int ioffset, ilength;
    int skipflag;                           /* ԤŸƤ뤫ɤФƤ뤫Υե饰 */
    /* 0:   1: Ÿ  2: Ÿʤ */
    SpdBoolean errflag;
    int mlres;

    if (verbose > 1) {
        spdMessage("matchNormalBlock (%s): %s(%d):%d+%d, %s(%d):%d\n", 
                   (sort == PBLOCK_NORMAL ? "NORMAL" : 
                    (sort == PBLOCK_COMMENT ? "COMMENT" : "??")),
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                   cFileBuf->csr.current.offset,
                   cFileBuf->currentBlock.length,
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno,
                   iFileBuf->csr.current.offset);
    }

    skipflag = 0;
    do {
        mlres = MATCH_OK;
        ioffset = iFileBuf->csr.current.offset;
        ilength = -1;

        /* MATCH_CONTINUE ΤȤ offset Ĵ */
        cFileBuf->currentBlock.offset = cFileBuf->csr.current.offset;

        while (cFileBuf->csr.current.offset < cFileBuf->nextBlock.offset &&
               mlres != MATCH_CONTINUE) {
            errflag = TRUE;
            /* 1ʬΥޥå󥰤롥 */

            switch (mlres = matchLine(cFileBuf, iFileBuf, mlres, 
                                      (sort != PBLOCK_COMMENT))) {
              case MATCH_CONTINUE:          /* \ + Ԥǰ */
                /* !!  length Ǥ \ + Ԥޤ */
                cFileBuf->currentBlock.length = 
                    cFileBuf->csr.current.offset - cFileBuf->currentBlock.offset;
                /* \ + ʬ᤹ */
                cFileBuf->currentBlock.length -= 
                    countContinueLine(&(cFileBuf->fb), &(cFileBuf->csr)) * 2;
                
                /* FALLTHROUGH */
              case MATCH_OK:                /* ޥå */
                /* ֤ξ skipflag Ÿ */
                if (skipflag == 0)
                    skipflag = 1;
                /* skipflag ŸǤϤʤХ顼 */
                if (skipflag != 1)
                    errflag = FALSE;
                break;
              case MATCH_SKIP:              /* Ÿ夬 */
                /* ֤ξ skipflag Ÿʤ */
                if (skipflag == 0)
                    skipflag = 2;
                break;
              case MATCH_FAIL:              /* ޥå󥰼 */
                errflag = FALSE;
                break;
              case MATCH_ABORT:             /*  */
                /* cFileBuf->currentBlock, cFileBuf->nextBlock Ĵ */
                cFileBuf->currentBlock.length = 
                    cFileBuf->csr.current.offset - cFileBuf->currentBlock.offset;
                /* parsePredefinedMacro  iFileBuf ѹ뤿ᡤ
                   ilength׻¸ */
                ilength = iFileBuf->csr.current.offset - ioffset;
                if (!parsePredefinedMacro(cFileBuf, iFileBuf)) {
                    spdErrorExit("%s(%d) and %s(%d): Can't match block.\n",
                                 cFileBuf->fb.filename,
                                 cFileBuf->csr.current.lineno,
                                 iFileBuf->fb.filename,
                                 iFileBuf->csr.current.lineno);
                }
                cFileBuf->nextBlock.sort = cFileBuf->currentBlock.sort;
                cFileBuf->nextBlock.offset = cFileBuf->csr.current.offset;
                break;
              case MATCH_PADDING:           /* ǥޥƽФˤäƹĴ */
                /* cFileBuf->currentBlock, cFileBuf->nextBlock Ĵ */
                cFileBuf->currentBlock.length = 
                    cFileBuf->csr.current.offset - cFileBuf->currentBlock.offset;
                cFileBuf->nextBlock.sort = cFileBuf->currentBlock.sort;
                cFileBuf->nextBlock.offset = cFileBuf->csr.current.offset;
                break;
            }
            /* ޥå󥰤˼Ԥˤ⤳Ǥ exit ʤ褦ѹ
               (matchDirectiveк) */
            if (!errflag) {
                return SPD_NON_ID;
            }
        }

        pblockId = createPblock(cFileBuf->fb.fileId, sort, 
                                cFileBuf->currentBlock.offset,
                                cFileBuf->currentBlock.length);
        
        if (skipflag != 2) {
            iblockId = spdCreatePmodelObj(SPD_CID_IBLOCK);
            if (ilength < 0)
                ilength = iFileBuf->csr.current.offset - ioffset;
            spdSetAttrIntValByName(iblockId, "length", ilength);
            
            relId = spdCreatePmodelRel(SPD_RID_IFILE_IBLOCK,
                                       iFileBuf->fb.fileId, iblockId);
            spdSetAttrIntValByName(relId, "offset", ioffset);
            relId = spdCreatePmodelRel(SPD_RID_PBLOCK_IBLOCK, 
                                       pblockId, iblockId);
        }

    } while (mlres == MATCH_CONTINUE);

    return pblockId;
}

static SpdBoolean skipToNextComment(FileBuffer *fb, MyCursor *csr)
{
    MyCursor csr2;

    while (csr->current.offset < csr->next.offset) {
        csr2 = *csr;
        if (nextChar(fb, &csr2) == '/') {
            if (peekChar(fb, &csr2) == '*') {
                if (verbose > 20) {
                    spdMessage("skipToNextComment: found comment at %s:%d\n", 
                               fb->filename, csr2.current.offset);
                }
                
                return TRUE;
            }
        }
        *csr = csr2;
    }
    if (verbose > 20) {
        spdMessage("skipToNextComment: can't found comment\n");
    }

    return FALSE;
}

/* normal block õ롥
   ΤȤ˼blocksortʬΤǤ⥻åȤ */

static int findEndOfNormalBlock(CFileBuffer *cFileBuf, SpdBoolean isSkiped)
{
    int offset, lineno;
    int fcomment, fstring, fchar;
    int c;
    MyCursor csr, csr2;

    csr = cFileBuf->csr;

    offset = csr.current.offset;
    lineno = csr.current.lineno;

    printCursor("findEndOfNormalBlock: from:", &(cFileBuf->fb), cFileBuf->csr);

    while (!isEndOfFile(&(cFileBuf->fb), &csr)) {
        if (isBeginningOfLine(&(cFileBuf->fb), &csr)) {
            skipSpace(&(cFileBuf->fb), &csr);
            /* ιԤƬ # ʤǥ쥯ƥֹ */
            if (peekChar(&(cFileBuf->fb), &csr) == '#') {
                cFileBuf->nextBlock.sort = PBLOCK_DIRECTIVE;
                if (verbose) {
                    spdMessage("findEndOfNormalBlock: find directive\n");
                }
                break;
            }
        }

        /* оݹԤǡmacrocall ˥Ȥγϥȡ
           뤫ɤõ */
        fcomment = fstring = fchar = FALSE;

        while (csr.current.offset < csr.next.offset &&
               !(!isSkiped && 
                 csr.current.offset == cFileBuf->nextMcall.coffset)) {
            csr2 = csr;

            c = nextChar(&(cFileBuf->fb), &csr2);

            if (c == '\\') {
                nextChar(&(cFileBuf->fb), &csr2);
            } else if (c == '\'' && fstring == FALSE) {
                if (fchar)
                    fchar = FALSE;
                else
                    fchar = TRUE;
            } else if (c == '"' && fchar == FALSE) {
                if (fstring)
                    fstring = FALSE;
                else
                    fstring = TRUE;
            } else if (c == '/' && fstring == FALSE && fchar == FALSE) {
                if (peekChar(&(cFileBuf->fb), &csr2) == '*') {
                    fcomment = TRUE;
                    break;
                }
            }
            
            csr = csr2;

            if (verbose > 20) {
                spdMessage("findEndOfNormalBlock: at %s[%d], nextline: [%d] "
                           "nextMacro: [%d]\n", 
                           cFileBuf->fb.filename, csr.current.offset, 
                           csr.next.offset, cFileBuf->nextMcall.coffset);
            }
        }
        
        if (verbose > 3) {
            spdMessage("findEndOfNormalBlock: check at %d char `%c' "
                       "nextline: %d nextmacro: %d\n", 
                       csr.current.offset, peekChar(&(cFileBuf->fb), &csr),
                       csr.next.offset,
                       cFileBuf->nextMcall.coffset);
        }
    
        if (!isSkiped && csr.current.offset == cFileBuf->nextMcall.coffset) {
            /* macrocall Ĥä */
            cFileBuf->nextBlock.sort = PBLOCK_MACROCALL;
            if (verbose) {
                spdMessage("findEndOfNormalBlock: find Macrocall at %d\n",
                           cFileBuf->nextMcall.coffset);
            }
            break;
        } else if (fcomment) {
            /* ȤĤä */
            cFileBuf->nextBlock.sort = PBLOCK_COMMENT;
            if (verbose) {
                spdMessage("findEndOfNormalBlock: find comment\n");
            }
            break;
        }

        /* ɤߤĤ餺 1Ԥνã */
        if (csr.current.offset == csr.next.offset) {
            if (verbose > 3) {
                spdMessage("findEndOfNormalBlock: next line.\n");
            }
            skipLogicalLine(&(cFileBuf->fb), &csr, 1);
        }
    }

    if (verbose > 10) {
        spdMessage("findEndOfNormalBlock: %s(%d):%d+%d\n%.*s\n", 
                   cFileBuf->fb.filename, lineno, offset, 
                   csr.current.offset - offset, 
                   csr.current.offset - offset, cFileBuf->fb.buf + offset);
    }

    if (isEndOfFile(&(cFileBuf->fb), &csr)) {
        cFileBuf->nextBlock.sort = PBLOCK_EOF;
        csr.current.offset = cFileBuf->fb.size;
    }

    return csr.current.offset;
}

/* comment block õ */

static int findEndOfCommentBlock(const CFileBuffer *cFileBuf)
{
    MyCursor csr;

    printCursor("findEndOfCommentBlock: from:", &(cFileBuf->fb), cFileBuf->csr);

    csr = cFileBuf->csr;
    skipSpace(&(cFileBuf->fb), &csr);

    if (skipComment(&(cFileBuf->fb), &csr) == FALSE) {
        spdErrorExit("%s(%d): Can't find comment.\n", 
                     cFileBuf->fb.filename, csr.current.lineno);
    }

    return csr.current.offset;
}

static int getBlockSort(const CFileBuffer *cFileBuf, SpdBoolean isSkiped)
{
    SpdBoolean fbeginLine;
    MyCursor csr;
    int c;

    csr = cFileBuf->csr;
    printCursor("getBlockSort: from:", &(cFileBuf->fb), cFileBuf->csr);

    if (isEndOfFile(&(cFileBuf->fb), &csr)) {
        return PBLOCK_EOF;
    }

    fbeginLine = isBeginningOfLine(&(cFileBuf->fb), &csr);

    c = peekChar(&(cFileBuf->fb), &csr);

    if (!isSkiped && cFileBuf->nextMcall.coffset == csr.current.offset) {
        if (verbose)
            spdMessage("getBlockSort : macrocall\n");
        return PBLOCK_MACROCALL;
    }

    if (c == '/') {
        nextChar(&(cFileBuf->fb), &csr);
        if (nextChar(&(cFileBuf->fb), &csr) == '*') {
            if (verbose)
                spdMessage("getBlockSort : comment\n");
            return PBLOCK_COMMENT;
        }
    }

    if (fbeginLine && c == '#') {
        if (verbose)
            spdMessage("getBlockSort : directive\n");
        return PBLOCK_DIRECTIVE;
    }

    if (verbose)
        spdMessage("getBlockSort : normal\n");

    return PBLOCK_NORMAL;
}

/* 
 * cFileBuf->currentBlock  cFileBuf->nextBlock򥳥ԡ
 * cFileBuf->nextBlock ˤ cFileBuf->currentBlock μ pblock  sort 
 * cFileBuf->currentBlock.length 
 * 
 * cFileBuf->nextBlock.length ꤵʤ
 */

static void getNextBlockInfo(CFileBuffer *cFileBuf, SpdBoolean isSkiped)
{
    printCursor("getNextBlockInfo: from:", &(cFileBuf->fb), cFileBuf->csr);
    
    if (cFileBuf->nextBlock.sort == PBLOCK_UNKNOWN) {
        /*  pblock  sort ʾ */
        cFileBuf->nextBlock.sort = getBlockSort(cFileBuf, isSkiped);
    }

    /* cFileBuf->currentBlock  cFileBuf->nextBlock򥳥ԡ */
    cFileBuf->currentBlock = cFileBuf->nextBlock;

    /* nextBlock  sort  offset  */
    switch (cFileBuf->currentBlock.sort) {
      case PBLOCK_NORMAL:
        cFileBuf->nextBlock.offset = findEndOfNormalBlock(cFileBuf, isSkiped);
        /* cFileBuf->nextBlock.sort  findEndOfNormalBlock ꤵ */
        break;
      case PBLOCK_COMMENT:
        cFileBuf->nextBlock.offset = findEndOfCommentBlock(cFileBuf);
        cFileBuf->nextBlock.sort = PBLOCK_UNKNOWN;
        break;
      case PBLOCK_MACROCALL:
        cFileBuf->nextBlock.offset = 
            cFileBuf->csr.current.offset + cFileBuf->nextMcall.clength;
        cFileBuf->nextBlock.sort = PBLOCK_UNKNOWN;
        break;
      case PBLOCK_DIRECTIVE:
        cFileBuf->nextBlock.offset = cFileBuf->csr.next.offset;
        cFileBuf->nextBlock.sort = PBLOCK_UNKNOWN;
        break;
    }

    /* currentBlock  length  */
    if (verbose > 3) {
        spdMessage("getNextBlockInfo: current:%d[%d], next:%d[%d]\n", 
                   cFileBuf->currentBlock.offset, 
                   cFileBuf->currentBlock.sort,
                   cFileBuf->nextBlock.offset, 
                   cFileBuf->nextBlock.sort);
    }
    cFileBuf->currentBlock.length = 
        cFileBuf->nextBlock.offset - cFileBuf->currentBlock.offset;
}

/* block Υޥå */

void matchBlock(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf)
{
    SpdBoolean isSkiped;

    /* .i θ֤߰ɤФ٤ # ԤλϤ򥹥åפƤ
       򳫻Ϥ롤ȤѤĤȤɬ
       (%%isSkipedLine ϴؿ̾Ѥ) */
    isSkiped = isSkipedLine(cFileBuf, iFileBuf);
    if (verbose > 20) {
        spdMessage("matchBlock: at %s(%d), %d\n", 
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno, 
                   isSkiped);
    }

    getNextBlockInfo(cFileBuf, isSkiped);
    switch (cFileBuf->currentBlock.sort) {
      case PBLOCK_COMMENT:
      case PBLOCK_NORMAL:
        if (matchNormalBlock(cFileBuf, iFileBuf, cFileBuf->currentBlock.sort)
            == SPD_NON_ID) {
            if (verbose > 3) {
                printLine("matchNormalBlock:cfile\n", &(cFileBuf->fb),
                          &(cFileBuf->csr));
                printLine("matchNormalBlock:ifile\n", &(iFileBuf->fb),
                          &(iFileBuf->csr));
            }
            spdErrorExit("%s(%d) and %s(%d): Blocks don't match.\n",
                         cFileBuf->fb.filename,
                         cFileBuf->csr.current.lineno,
                         iFileBuf->fb.filename,
                         iFileBuf->csr.current.lineno);
        }
        break;
      case PBLOCK_MACROCALL:
        matchMacroCall(cFileBuf, iFileBuf);
        getNextMacroCall(cFileBuf, iFileBuf);
        break;
      case PBLOCK_DIRECTIVE:
        matchDirective(cFileBuf, iFileBuf);
        break;
      case PBLOCK_EOF:
        break;
    }
}

/* .c Υϰư.i ΤɤФ */
int matchCommentInDirective(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf)
{
    CFileBuffer orig_cfb;
    int orig_ilineno;

    orig_cfb = *cFileBuf;
    orig_ilineno = iFileBuf->csr.top.lineno;

    while (peekChar(&(iFileBuf->fb), &(iFileBuf->csr)) == '/') {
        if (skipToNextComment(&(cFileBuf->fb), &(cFileBuf->csr))) {
            cFileBuf->nextBlock.offset = findEndOfCommentBlock(cFileBuf);
            cFileBuf->currentBlock.offset = cFileBuf->csr.current.offset;
            cFileBuf->currentBlock.length = 
                cFileBuf->nextBlock.offset - cFileBuf->currentBlock.offset;
        
            if (matchNormalBlock(cFileBuf, iFileBuf, PBLOCK_COMMENT) 
                == SPD_NON_ID)
                break;
        } else {
            break;
        }
    }
    
    *cFileBuf = orig_cfb;

    return iFileBuf->csr.top.lineno - orig_ilineno;
}
