/* 
 * Program:             $RCSfile: define.c,v $  $Revision: 30.13 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: define.c,v 30.13 2003/07/20 11:46:08 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

static int isTokenChar(int c)
{
    return (isalnum(c) || c == '_');
}

/* ȡ1ļ롣length ˤ \ + Ԥޤ */

void getToken(FileBuffer *fb, MyCursor *paramcsr, PPRegion *reg)
{
    int         i, c;

    skipSpaceOrComment(fb, paramcsr);

    reg->offset = paramcsr->current.offset;

    for (i = 0; i < MAX_TOKEN_LEN && isTokenChar(c = peekChar(fb, paramcsr));
         i++) {
        reg->str[i] = c;
        nextChar(fb, paramcsr);
    }
    reg->str[i] = '\0';

    reg->length = paramcsr->current.offset - reg->offset;
}

static void createMacroDefine(SpdObjId pblockId, DEFINEinfo *pdef,
                              SpdBoolean fdefine, 
                              CFileBuffer *cFileBuf, MyCursor *paramcsr)
{
    PPRegion macroPattern, macroName, macroOptional, macroBody;
    PPRegion *macroDeclArg = NULL;
    ArgReference *pref;
    SpdRelId relId;
    int i;

    /* macroPattern  (length ϸ) */
    macroPattern.objId = spdCreatePmodelObj(SPD_CID_MACROPATTERN);

    if (fdefine) {
        macroPattern.offset = paramcsr->current.offset;
    } else {
        macroPattern.offset = -1;
        macroPattern.length = 0;
    }
    relId = spdCreatePmodelRel(SPD_RID_DEF_MPATT,
                               pblockId, macroPattern.objId);
    spdSetAttrIntValByName(relId, "offset", macroPattern.offset);

    /* macroName  */
    if (fdefine) {
        getToken(&(cFileBuf->fb), paramcsr, &macroName);
        if (!SPD_STREQ(macroName.str, pdef->name)) {
            spdErrorExit("%s(%d): PIDB has defineInfo of macro `%s'"
                         " but there's macro `%s'\n",
                         cFileBuf->fb.filename, paramcsr->top.lineno,
                         pdef->name, macroName.str);
        }
    } else {
        macroName.length = strlen(pdef->name);
        macroName.offset = -1;
    }
    
    macroName.objId = spdCreatePmodelObj(SPD_CID_MACRONAME);
    spdSetAttrStrValByName(macroName.objId, "name", pdef->name);
    spdSetAttrIntValByName(macroName.objId, "length", macroName.length);

    relId = spdCreatePmodelRel(SPD_RID_MPATT_MNAME,
                               macroPattern.objId, macroName.objId);
    spdSetAttrIntValByName(relId, "offset", macroName.offset);

    if (verbose)
        spdMessage("createMacroDefine: %s\n", pdef->name);

    if (pdef->narg >= 0) {
        /* ޥ¸ߤ */
        /* macroOptional  (length ϸ) */

        /* !! ޥɥ饤󥪥ץ˰Ĥޥ뤳Ȥ
           sdb4 ץȤбƤʤ */
        macroOptional.objId = spdCreatePmodelObj(SPD_CID_MACROOPTIONAL);
        if (fdefine) {
            macroOptional.offset = paramcsr->current.offset;
        } else {
            macroOptional.offset = -1;
            macroOptional.length = 0;
        }

        relId = spdCreatePmodelRel(SPD_RID_MPATT_MOPT,
                                   macroPattern.objId, macroOptional.objId);
        spdSetAttrIntValByName(relId, "offset", macroOptional.offset);

        if (fdefine && nextChar(&(cFileBuf->fb), paramcsr) != '(') {
            spdErrorExit("In analysing #define %s at %s[%d], "
                         "PIDB has %d argument info "
                         "but can't find `(' in source at [%d].\n",
                         macroName.str, 
                         cFileBuf->fb.filename, macroName.offset,
                         pdef->narg, paramcsr->current.offset);
        }

        macroDeclArg = spdMalloc(sizeof(PPRegion) * pdef->narg);
        for (i = 0; i < pdef->narg; i++) {
            if (fdefine) {
                getToken(&(cFileBuf->fb), paramcsr, &(macroDeclArg[i]));
                if (!SPD_STREQ(pdef->arg[i], 
                                 macroDeclArg[i].str)) {
                    if (verbose) {
                        spdWarning("In analysing #define %s at %s[%d], "
                                   "PIDB has argument `%s' but `%s' "
                                   "in source at [%d].\n",
                                   macroName.str, 
                                   cFileBuf->fb.filename, macroName.offset,
                                   pdef->arg[i], macroDeclArg[i].str,
                                   paramcsr->current.offset);
                    }
                }
            } else {
                macroDeclArg[i].offset = -1;
                macroDeclArg[i].length = strlen(pdef->arg[i]);
            }
            /* macroDeclArg  */
            macroDeclArg[i].objId = 
                spdCreatePmodelObj(SPD_CID_MACRODECLARG);
            spdSetAttrStrValByName(macroDeclArg[i].objId, 
                                   "name", pdef->arg[i]);
            spdSetAttrIntValByName(macroDeclArg[i].objId,
                                   "length", macroDeclArg[i].length);
        
            relId = spdCreatePmodelRel(SPD_RID_MOPT_MDECLA,
                                       macroOptional.objId, 
                                       macroDeclArg[i].objId);
            spdSetAttrIntValByName(relId, "offset", macroDeclArg[i].offset);

            if (fdefine) {
                /* , ɤФ */
                skipSpaceOrComment(&(cFileBuf->fb), paramcsr);
                if (peekChar(&(cFileBuf->fb), paramcsr) == ',') {
                    nextChar(&(cFileBuf->fb), paramcsr);
                }
            }
        }

        if (fdefine && nextChar(&(cFileBuf->fb), paramcsr) != ')') {
            spdErrorExit("In analysing #define %s at %s(%d), "
                         "PIDB has %d argument info "
                         "but can't find `)' in source at [%d].\n",
                         macroName.str, 
                         cFileBuf->fb.filename, macroName.offset,
                         pdef->narg, paramcsr->current.offset);
        }

        /* macroOptional  length  */
        if (fdefine)
            macroOptional.length =
                paramcsr->current.offset - macroOptional.offset;
        spdSetAttrIntValByName(macroOptional.objId, 
                               "length", macroOptional.length);
    }
    
    /* macroPattern  length  */
    if (fdefine)
        macroPattern.length = paramcsr->current.offset - macroPattern.offset;
    spdSetAttrIntValByName(macroPattern.objId, "length", macroPattern.length);

    /* macroBody  */
    macroBody.objId = spdCreatePmodelObj(SPD_CID_MACROBODY);
#if defined(USE_MACROBODY_HASH)
    addMacroBodyHash(macroName.objId, macroBody.objId);
#endif
    if (fdefine) {
        macroBody.offset = paramcsr->current.offset;
        macroBody.length = paramcsr->next.offset - macroBody.offset;
    } else {
        macroBody.offset = -1;
        macroBody.length = strlen(pdef->expansion);
    }
    spdSetAttrStrValByName(macroBody.objId, "name", 
                           pdef->expansion);
    relId = spdCreatePmodelRel(SPD_RID_DEF_MBODY,
                               pblockId, macroBody.objId);
    spdSetAttrIntValByName(relId, "offset", macroBody.offset);

    spdSetAttrIntValByName(macroBody.objId, "length", macroBody.length);

    /* mbody_mdecla  */
    for (i = 0; i < pdef->nref; i++) {
        pref = &(pdef->ref[i]);
        relId = spdCreatePmodelRel(SPD_RID_MBODY_MDECLA,
                                   macroBody.objId,
                                   macroDeclArg[pref->argn].objId);
        spdSetAttrIntValByName(relId, "order", i);
        spdSetAttrIntValByName(relId, "skip", pref->nchar);
        spdSetAttrIntValByName(relId, "sharp", pref->stringify);
    }

    /* PIDB-macroId -> macroName Υơ֥ */
    addMacroIndex(pdef->ID, macroName.objId);

    if (pdef->narg >= 0) {
        spdFree(macroDeclArg);
    }
}

/* #define ν */

void parseDefineDirective(CFileBuffer *cFileBuf, 
                          MyCursor *paramcsr, int ioffset, SpdObjId pblockId)
{
    DEFINEinfo *pdef;
    int i;

    /* PIDB  macroName 򸡺 */
    pdef = NULL;
    for (i = 0; i < defineInfoN; i++) {
        if (defineInfo[i].pos.type == DEFINE_FILE &&
            SPD_STREQ(defineInfo[i].pos.fname, 
                        cFileBuf->lineInfo.filename) &&
            defineInfo[i].pos.start == cFileBuf->csr.current.offset &&
            defineInfo[i].output == ioffset) {
            pdef = &(defineInfo[i]);
            break;
        }
    }
    
    if (pdef == NULL) {
        /* ɤФ줿 #define ΤϤ */
        if (verbose)
            spdMessage("#define at %s(%d) is skiped.\n", 
                       cFileBuf->fb.filename, cFileBuf->csr.current.lineno);
        return;
    }

    createMacroDefine(pblockId, pdef, TRUE, cFileBuf, paramcsr);
}

/* -D ˤä줿ޥν */

void parsePreDefinedOption(void)
{
    SpdObjId pblockId;
    DEFINEinfo *pdef;
    int i;
    
    for (i = 0; i < defineInfoN; i++) {
        pdef = &(defineInfo[i]);

        if (pdef->pos.type != DEFINE_FILE) {
            /* pblock  */
            pblockId = spdCreatePmodelObj(SPD_CID_PBLOCK);
            if (pdef->pos.type == DEFINE_INITIALIZATION) 
                spdSetAttrIntValByName(pblockId, "sort", PBLOCK_DEFINE_INIT);
            else
                spdSetAttrIntValByName(pblockId, "sort", PBLOCK_DEFINE_OPTION);
            spdSetAttrIntValByName(pblockId, "length", -1);

            createMacroDefine(pblockId, pdef, FALSE, NULL, NULL);
        }
    }
}
