/* 
 * Program:             $RCSfile: directive.c,v $  $Revision: 30.13 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: directive.c,v 30.13 2003/07/20 11:46:08 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

#define MAX_DIRECTIVE_LEN  9    /* for length of "include" + alpha */
/* !! FreeBSD  egcs Ǥ? include_next directive ¸ߤƤ... */

static int getDirectiveSort(const char *name)
{
    static struct {
        char *name;
        int sort;
    } directiveTable[] = {
        {"define", PBLOCK_DEFINE_FILE},
        {"elif", PBLOCK_ELIF},
        {"else", PBLOCK_ELSE},
        {"endif", PBLOCK_ENDIF},
        {"error", PBLOCK_ERROR},
        {"if", PBLOCK_IF},
        {"ifdef", PBLOCK_IFDEF},
        {"ifndef", PBLOCK_IFNDEF},
        {"include", PBLOCK_INCLUDE},
        {"line", PBLOCK_LINE},
        {"pragma", PBLOCK_PRAGMA},
        {"undef", PBLOCK_UNDEF},
        {NULL, -1}
    };

    int i;

    for (i = 0; directiveTable[i].name != NULL; i++) {
        if (SPD_STREQ(directiveTable[i].name, name)) {
            if (verbose) {
                spdMessage("getDirectiveSort : %s, %d\n", name, 
                           directiveTable[i].sort);
            }
            return directiveTable[i].sort;
        }
    }

    if (verbose) {
        spdMessage("getDirectiveSort : unknown directive `%s'\n", name);
    }
    return -1;
}

static int parseDirective(CFileBuffer *cFileBuf, MyCursor *paramcsr)
{
    MyCursor csr;
    int i, c;
    char directivename[MAX_DIRECTIVE_LEN + 1];
    int sort;

    csr = cFileBuf->csr;                    /* λǤƬ#ؤƤϤ */

    if (nextChar(&(cFileBuf->fb), &csr) != '#') {
        spdErrorExit("%s(%d) isn't directive line.\n", cFileBuf->fb.filename,
                     csr.current.lineno);
    } 

    skipSpaceOrComment(&(cFileBuf->fb), &csr);

    /* ǥ쥯ƥ̾ڤФ */
    for (i = 0; isalpha(c = nextChar(&(cFileBuf->fb), &csr)) || c == '_'; ) {
        if (i < MAX_DIRECTIVE_LEN) {
            directivename[i] = c;
            i++;
        }
    } 
    directivename[i] = '\0';
    sort = getDirectiveSort(directivename);

    skipSpaceOrComment(&(cFileBuf->fb), &csr);

    *paramcsr = csr;

    return sort;
}

/******************************************************************************/

static void pushIfexprStack(CFileBuffer *cFileBuf, SpdBoolean ifexpr)
{
    IFStack *new;

    new = spdMalloc(sizeof(IFStack));
    new->next = cFileBuf->ifstack.next;
    new->ifexpr = ifexpr;
    new->expanded = ifexpr;
    cFileBuf->ifstack.next = new;
}

static SpdBoolean popIfexprStack(CFileBuffer *cFileBuf)
{
    IFStack *old;
    SpdBoolean res;

    old = cFileBuf->ifstack.next;

    if (old == NULL) {
        spdWarning("#if stack is something wrong...\n");

        return FALSE;
    }

    cFileBuf->ifstack.next = old->next;
    res = old->ifexpr;

    spdSafetyFree(old);

    return res;
}

static void replaceIfexprStack(CFileBuffer *cFileBuf, SpdBoolean ifexpr)
{
    IFStack *old;

    old = cFileBuf->ifstack.next;

    if (old == NULL) {
        spdWarning("#if stack is something wrong...\n");
        return;
    }
    
    old->ifexpr = ifexpr;
    old->expanded = old->expanded || ifexpr;
}

static SpdBoolean fetchIfexprStack(CFileBuffer *cFileBuf)
{
    IFStack *p;

    p = cFileBuf->ifstack.next;

    if (p == NULL) {
        /* stack ˤʤˤʤȤ, #if γʤΤ TRUE */
        return TRUE;
    }

    return p->ifexpr;
}

static SpdBoolean isexpandedIfexprStack(CFileBuffer *cFileBuf)
{
    IFStack *p;

    p = cFileBuf->ifstack.next;

    if (p == NULL) {
        /* stack ˤʤˤʤȤ, ƤӽФʤϤ */
        return TRUE;
    }

    return p->expanded;
}

static SpdBoolean fetchOuterIfexprStack(CFileBuffer *cFileBuf)
{
    IFStack *p, *q;

    p = cFileBuf->ifstack.next;

    if (p == NULL) {
        /* stack ˤʤˤʤȤ, #if γʤΤ TRUE */
        return TRUE;
    }

    q = p->next;

    if (q == NULL) {
        /* stack ˤʤˤʤȤ, #if γʤΤ TRUE */
        return TRUE;
    }

    return q->ifexpr;
}

/*
 *  cFileBuf, iFileBuf "#if" directiveؤƤˤοͤ롥
 */

static SpdBoolean getIfexprInfoValue(CFileBuffer *cFileBuf,
                                     IFileBuffer *iFileBuf,
                                     SpdBoolean isexpanded)
{
    SpdBoolean res = FALSE;
    
    if (iFileBuf->ifexprId < ifexprInfoN && isexpanded) {
        if (verbose) {
            spdMessage("CFileBuf:: %d  ifexpr:: #if %s(%d) is %d\n", 
                       cFileBuf->csr.current.lineno,
                       ifexprInfo[iFileBuf->ifexprId].cfilename,
                       ifexprInfo[iFileBuf->ifexprId].clineno,
                       ifexprInfo[iFileBuf->ifexprId].ifexpr);
        }

        res = ifexprInfo[iFileBuf->ifexprId].ifexpr;

        iFileBuf->ifexprId++;
    }

    return res;
}

/******************************************************************************/

/* ץץåŸȥ֥Ⱥ */

SpdObjId matchDirective(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf)
{
    SpdObjId    pblockId;
    int         sort;
    MyCursor    paramcsr;
    int         clines, ilines;
    int         skip = 0;
    SpdBoolean  ifexpr;

    if (verbose) {
        spdMessage("matchDirective: %s(%d):%d, %s(%d):%d\n",
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                   cFileBuf->csr.current.offset,
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno,
                   iFileBuf->csr.current.offset);
    }

    sort = parseDirective(cFileBuf, &paramcsr);
    
    pblockId = createPblock(cFileBuf->fb.fileId, sort, 
                            cFileBuf->csr.current.offset, 
                            cFileBuf->csr.next.offset  -
                            cFileBuf->csr.current.offset);

    switch (sort) {
      case PBLOCK_DEFINE_FILE:
        if (fetchIfexprStack(cFileBuf) == TRUE) {
            parseDefineDirective(cFileBuf, &paramcsr, iFileBuf->csr.current.offset, pblockId);
        }
        break;
      case PBLOCK_INCLUDE:
        if (fetchIfexprStack(cFileBuf) == TRUE) {
            skip = parseIncludeDirective(cFileBuf, iFileBuf, &paramcsr,
                                         pblockId);
        }
        break;
      case PBLOCK_LINE:
        if (fetchIfexprStack(cFileBuf) == TRUE) {
            parseLineDirective(cFileBuf, iFileBuf, &paramcsr);
        }
        break;
      case PBLOCK_IF:
      case PBLOCK_IFDEF:
      case PBLOCK_IFNDEF:
        ifexpr = getIfexprInfoValue(cFileBuf, iFileBuf, 
                                    fetchIfexprStack(cFileBuf));
        pushIfexprStack(cFileBuf, ifexpr);
        break;
      case PBLOCK_ELIF:
        ifexpr = getIfexprInfoValue(cFileBuf, iFileBuf, 
                                    fetchOuterIfexprStack(cFileBuf) && 
                                    !isexpandedIfexprStack(cFileBuf));
        replaceIfexprStack(cFileBuf, ifexpr);
        break;
      case PBLOCK_ELSE:
        replaceIfexprStack(cFileBuf, 
                           fetchOuterIfexprStack(cFileBuf) && 
                           !isexpandedIfexprStack(cFileBuf));
        break;
      case PBLOCK_ENDIF:
        popIfexprStack(cFileBuf);
        break;
      case PBLOCK_UNDEF:
      case PBLOCK_PRAGMA:
      case PBLOCK_ERROR:
        /* Empty. */
        break;
    }

    /* ץץåΥȤΥޥå */
    ilines = matchCommentInDirective(cFileBuf, iFileBuf);

    /* .c 1ԥå(ǥ쥯ƥּΤȥ) */
    clines = skipLogicalLine(&(cFileBuf->fb), &(cFileBuf->csr), 1);

    /* .c ǤʪԿʬ .i ζԤ򥹥å */
    skipILine(cFileBuf, iFileBuf, clines - skip - ilines);

    return pblockId;
}
