/* 
 * Program:             $RCSfile: file.c,v $  $Revision: 30.21 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: file.c,v 30.21 2003/06/22 14:42:13 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

static void freeIFStack(IFStack *p);

/* FileBuffer  */

static CFileBuffer *readCFile(SpdObjId pfileId)
{
    CFileBuffer *cFileBuf;
    BlockInfo initialBlockInfo = { PBLOCK_UNKNOWN, 0, 0 };

    cFileBuf = spdMalloc(sizeof(CFileBuffer));
    if (readFileBuffer(&(cFileBuf->fb), pfileId) < 0) {
        spdFree(cFileBuf);
        return NULL;
    }

    if (verbose) {
        spdMessage("readCFile \"%s\".\n", spdGetName(pfileId));
    }

    cFileBuf->csr = initCursor(&(cFileBuf->fb));
    cFileBuf->currentBlock = cFileBuf->nextBlock = initialBlockInfo;
    cFileBuf->lineInfo.filename = spdStrdup(cFileBuf->fb.filename);
    cFileBuf->lineInfo.loffset = 0;
    cFileBuf->ifstack.next = NULL;
    return cFileBuf;
}

static IFileBuffer *readIFile(SpdObjId fileId)
{
    IFileBuffer *iFileBuf;

    iFileBuf = spdMalloc(sizeof(IFileBuffer));
    if (readFileBuffer(&(iFileBuf->fb), fileId) < 0) {
        spdFree(iFileBuf);
        return NULL;
    }
    iFileBuf->csr = initCursor(&(iFileBuf->fb));
    iFileBuf->skip.lineno = -1;
    iFileBuf->skip.filename = NULL;
    iFileBuf->skip.cloffset = 0;
    iFileBuf->skip.fpadding = FALSE;
    iFileBuf->mExpId = -1;
    iFileBuf->ifexprId = 0;
    return iFileBuf;
}

/* cfilename Ф Pmodel ǡ١κ */
/* ͤб pfile ֥Ȥ ID */

SpdObjId makePmodelOfCFile(IFileBuffer *iFileBuf)
{
    char *cfilename;
    CFileBuffer *cFileBuf;
    SpdObjId pfileId = SPD_NON_ID;

    iFileBuf->skip.lineno = -1;
    spdFree(iFileBuf->skip.filename);
    iFileBuf->skip.filename = NULL;
    iFileBuf->skip.cloffset = 0;
    iFileBuf->skip.fpadding = FALSE;

    if (verbose) {
        spdMessage("\"%s\": %d: Analyzing \".c\" file.\n",
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno);
    }

    if (checkEnterSharpDirective(iFileBuf, &cfilename) == FALSE) {
        spdErrorExit("\"%s\": %d: Can't find file.\n",
                     iFileBuf->fb.filename, iFileBuf->csr.current.lineno);
    }

    /* pfile ֥  */
    pfileId = spdCreatePmodelObj(SPD_CID_PFILE);
    spdSetAttrStrValByName(pfileId, "name", cfilename);
    spdFree(cfilename);
    
    /* .c(or .h) եɤ߹ */
    cFileBuf = readCFile(pfileId);
    if (cFileBuf == NULL) {
        return SPD_NON_ID;
    }
    getNextMacroCall(cFileBuf, iFileBuf);

    while (!isEndOfFile(&(cFileBuf->fb), &(cFileBuf->csr))) {
        matchBlock(cFileBuf, iFileBuf);
    }

    /* Ǹ夬 # Ԥǽ礬Τǡ.i å
       ɬפʤɤФ */
    isSkipedLine(cFileBuf, iFileBuf);

    /* 곫*/
    freeFileBuffer(&(cFileBuf->fb));
    freeIFStack(cFileBuf->ifstack.next);
    spdFree(cFileBuf->lineInfo.filename);
    spdFree(cFileBuf);

    return pfileId;
}

/* iFileId Ф Pmodel ǡ١κ */

void makePmodelOfIFile(SpdObjId ifileId)
{
    IFileBuffer *iFileBuf;
    SpdString   cfilename;
    SpdObjId    pfileId = SPD_NON_ID;
    SpdObjId    cfileId;

    cfileId = spdConvFileIdItoC(ifileId);
    /* !! Quick hack for suppressing warning about "cast". */
    cfilename = (SpdString)spdGetName(cfileId);
    
    /* PIDBɤ߹ࡥ */
    if (spdInitializePidbCache(ifileId) != 0) {
        spdErrorExit("Can't read PIDB: \"%s\".", cfilename);
    }

    createMacroIndex();

    parsePreDefinedOption();

    /* ".i"եɤ߹ࡥ */
    iFileBuf = readIFile(ifileId);
    if (iFileBuf == NULL) {
        return;
    }

    pfileId = makePmodelOfCFile(iFileBuf);

    spdCreatePmodelRel(SPD_RID_CFILE_IFILE, pfileId, ifileId);

    freeFileBuffer(&(iFileBuf->fb));
    spdFree(iFileBuf->skip.filename);

    freeMacroIndex();
    spdFree(iFileBuf);
}

/******************************************************************************/

static void freeIFStack(IFStack *p)
{
    while (p != NULL) {
        IFStack     *tmp;       

        tmp = p->next;
        spdFree(p);
        p = tmp;
    }
}
