/* 
 * Program:             $RCSfile: filebuf.c,v $  $Revision: 30.27 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: filebuf.c,v 30.27 2003/07/20 11:46:07 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

static SpdBoolean isSpaceOrTab(int c);

/* 
 * cFileμԤƬõ\+ԤǷ³Ƥȡ
 * ʸ󡿥ȤԤޤäƤԤȤư
 */

static int searchNextLogicalLine(const FileBuffer *fb, FOffset *current)
{
    int         fStr, fComment, fSlash, fAsterisk, fBackslash, fChar;
    FOffset     oldfo;

    oldfo = *current;
    fBackslash = fStr = fComment = fSlash = fAsterisk = fChar = FALSE;
    while (current->offset < fb->size) {
        if (fb->buf[current->offset] == '\n')
            current->lineno++;

        if (fBackslash == FALSE) {
            if (fComment == FALSE && fStr ==FALSE && fChar == FALSE) {
                if (fb->buf[current->offset] == '\n') {
                    break;
                }
                
                if (fb->buf[current->offset] == '"') {
                    fStr = TRUE;
                }

                if (fb->buf[current->offset] == '\'') {
                    fChar = TRUE;
                }
                
                if (fb->buf[current->offset] == '*' && fSlash == TRUE) {
                    fComment = TRUE;
                }
                
                if (fb->buf[current->offset] == '/') {
                    fSlash = TRUE;
                } else {
                    fSlash = FALSE;
                }
            } else {
                if (fStr) {
                    if (fb->buf[current->offset] == '"') {
                        fStr = FALSE;
                    }
                }
                
                if (fChar) {
                    if (fb->buf[current->offset] == '\'') {
                        fChar = FALSE;
                    }
                }
            }
        }

        if (fComment) {
            if (fAsterisk && fb->buf[current->offset] == '/') {
                fComment = FALSE;
            }

            if (fb->buf[current->offset] == '*') {
                fAsterisk = TRUE;
            } else {
                fAsterisk = FALSE;
            }
        }

        if (fb->buf[current->offset] == '\\' &&
            !spdInputEncodeIsWChar(fb->buf[current->offset - 1], 
                                   fb->buf[current->offset])
            && fBackslash == FALSE) {
            fBackslash = TRUE;
        } else {
            fBackslash = FALSE;
        }

        current->offset++;
    }
    if (current->offset < fb->size) {
        /* ԤɤФ */
        current->offset++;
    }

    if (verbose > 10) {
        spdMessage("searchNextLogicalLine: %s(%d)[%d] -> (%d)[%d]\n",
                   fb->filename, oldfo.lineno, oldfo.offset, 
                   current->lineno, current->offset);
    }

    return current->lineno - oldfo.lineno;
}

/* FileBuffer Ԥ csr->current  n ɤФ */
/* csr->current ꤵƤ뤳ȤȤ롥 */
/* csr->top, csr->next Ŭڤѹ롥 */
/* ɤФʪԿ֤ */

int skipLogicalLine(const FileBuffer *fb, MyCursor *csr, int n)
{
    int i;
    int nline;

    if (csr->current.offset == csr->next.offset) {
        csr->current.offset--;
        if (peekChar(fb, csr) == '\n') {
            csr->current.lineno--;
        }
    }
    for (i = 0, nline = 0; i < n && csr->current.offset < fb->size; i++) {
        nline += searchNextLogicalLine(fb, &(csr->current));
    }
    csr->next = csr->top = csr->current;
    searchNextLogicalLine(fb, &(csr->next));

    return nline;
}

/* FileBuffer ɤ߹ߤäɤå */

int isEndOfFile(const FileBuffer *fb, const MyCursor *csr)
{
    if (fb->size <= csr->current.offset) {
        return TRUE;
    } else {
        return FALSE;
    }
}

/* 
 * FileBuffer˥եɤ߹ࡥ
 * ޤޥmkPfileFileƤФƤʤΤǡϢ"pfile_file"ϻȤʤ
 * ΤᡤspdConvIdPfileToFile()Ȥʤ
 */

int readFileBuffer(FileBuffer *fb, SpdObjId fileId)
{
    FILE            *fp;
    struct stat     st;

    fb->fileId = fileId;
    fb->filename = spdStrdup(spdGetName(fileId));

    fp = smuOpenSrcFile(smuGetFileIdByName(fb->filename));

    /* Get file size. */
    spdFstat(fileno(fp), &st);
    fb->size = (int)(st.st_size);

    fb->buf = spdMalloc(fb->size + 1);
    fb->size = spdFreadChar(fb->buf, fb->size, fp);
    spdFclose(fp);

    /* եåȴط */

    return (fb->size);
}

void freeFileBuffer(FileBuffer *fb)
{
    if (fb->buf) {
        spdFree(fb->buf);
    }
    if (fb->filename) {
        spdFree(fb->filename);
    }
}

MyCursor initCursor(const FileBuffer *fb)
{
    MyCursor csr;

    csr.top.offset = 0;
    csr.top.lineno = 1;
    csr.next = csr.current = csr.top;
    searchNextLogicalLine(fb, &(csr.next));
    
    return csr;
}

/* FileBuffer  1ʸɤ */

int peekChar(const FileBuffer *fb, const MyCursor *csr)
{
    return fb->buf[csr->current.offset];
}

/* \+Ԥ¸ߤ뤫ɤå롥 */

SpdBoolean isContinueLine(const FileBuffer *fb, const MyCursor *csr)
{
    return (csr->current.offset - 2 >= csr->top.offset &&
            fb->buf[csr->current.offset - 2] == '\\' &&
            fb->buf[csr->current.offset - 1] == '\n' &&
            !spdInputEncodeIsWChar(fb->buf[csr->current.offset - 1], 
                                   fb->buf[csr->current.offset - 2]));
}

/* ɤФ\+Ԥ */

int countContinueLine(const FileBuffer *fb, const MyCursor *pcsr)
{
    MyCursor csr;
    int i;

    csr = *pcsr;

    for (i = 0; isContinueLine(fb, &csr); i++) {
        csr.current.offset -= 2;
    }

    return i;
}

/* \+ԤɤФ */

static void skipContinueLine(const FileBuffer *fb, MyCursor *csr)
{
    while (csr->current.offset < csr->next.offset &&
           fb->buf[csr->current.offset] == '\\' &&
           fb->buf[csr->current.offset + 1] == '\n' &&
           !spdInputEncodeIsWChar(fb->buf[csr->current.offset - 1],
                                  fb->buf[csr->current.offset])) {
        csr->current.offset += 2;
        csr->current.lineno++;
    }
}

/* FileBuffer θ֤߰1ʸɤǡʸ pli ư
   \+ԤɤФ */
int nextChar(const FileBuffer *fb, MyCursor *csr)
{
    int res;

    if (verbose > 30)
        spdMessage("nextChar: from %s(%d):%d next (%d):%d\n",
                   fb->filename, 
                   csr->current.lineno, csr->current.offset,
                   csr->next.lineno, csr->next.offset);

    /*  csr  \\ + ԤؤƤɤФ */
    if (csr->current.offset < csr->next.offset) {
        skipContinueLine(fb, csr);
    }

    res = peekChar(fb, csr);

    /* csr ǤʤСcsr ư */
    if (csr->current.offset < csr->next.offset) {
        csr->current.offset++;
    
        skipContinueLine(fb, csr);

        if (res == '\n') {
            csr->current.lineno++;
        }
    }

    if (verbose > 30)
        spdMessage("nextChar: to %s(%d):%d next (%d):%d\n",
                   fb->filename, 
                   csr->current.lineno, csr->current.offset,
                   csr->next.lineno, csr->next.offset);

    return res;
}

/* FileBufferΥڡɤФԤޤǤɤĤ齪λ롥 */
/* ڡμʸ֤ */

int skipSpace(const FileBuffer *fb, MyCursor *csr)
{
    int c = '\0';

    while (csr->current.offset < csr->next.offset &&
           isSpaceOrTab(c = peekChar(fb, csr)))
        nextChar(fb, csr);

    return c;
}

/* 
 * static SpdBoolean isSpaceOrTab(int c)
 * {
 *     if (c == ' ' || c == '\t')
 *         return TRUE;
 *     else
 *         return FALSE;
 * }
 */

static SpdBoolean isSpaceOrTab(int c)
{
    return (c == ' ' || c == '\t');
}

SpdBoolean skipComment(const FileBuffer *fb, MyCursor *csr)
{
    int c, c1;
    int lineno, offset;
    MyCursor tmp;

    lineno = csr->current.lineno;
    offset = csr->current.offset;

    /* ᤷ */
    tmp = *csr;

    /* Ȥγϥȡ򥹥åפ */
    if (!(nextChar(fb, csr) == '/' && 
          nextChar(fb, csr) == '*')) {
        printLine("skipComment : Can't find comment.\n", fb, csr);
        
        *csr = tmp;

        return FALSE;
    }
    
    c = 0;
    while ((c1 = nextChar(fb, csr)) && !(c == '*' && c1 == '/')) {
        if (csr->current.offset == csr->next.offset) {
            skipLogicalLine(fb, csr, 1);
        }
        c = c1;
    }

    if (verbose > 10) {
        spdMessage("skipComment: %s(%d)%d+%d\n%.*s\n",
                   fb->filename, lineno, offset, 
                   csr->current.offset - offset + 1, 
                   csr->current.offset - offset + 1, fb->buf + offset);
    }

    return TRUE;
}

int skipSpaceOrComment(const FileBuffer *fb, MyCursor *csr)
{
    do {
        skipSpace(fb, csr);
    } while (skipComment(fb, csr));

    return skipSpace(fb, csr);
}

/* offset ޤ csr ʤ롥ɤФԿ֤ */

int skipToOffset(const FileBuffer *fb, MyCursor *csr, int offset)
{
    int oldlineno;
    
    if (verbose > 10) {
        spdMessage("skipToOffset: from %s(%d):%d to %d\n",
                   fb->filename, csr->current.lineno,
                   csr->current.offset, offset);
    }

    oldlineno = csr->current.lineno;

    while (csr->next.offset <= offset) {
        skipLogicalLine(fb, csr, 1);
    }

    while (csr->current.offset < offset) {
        nextChar(fb, csr);
    }

    return csr->current.lineno - oldlineno;
}

/* lineno ޤ csr ʤ롥ɤФԿ֤ */

/* 
 * int skipToLineno(const FileBuffer *fb, MyCursor *csr, int lineno)
 * {
 *     int oldlineno;
 *     
 *     oldlineno = csr->current.lineno;
 * 
 *     while (csr->current.lineno < lineno) {
 *         skipLogicalLine(fb, csr, 1);
 *     }
 * 
 *     return csr->current.lineno - oldlineno;
 * }
 */

SpdBoolean isBeginningOfLine(const FileBuffer *fb, const MyCursor *csr)
{
    MyCursor c;
    
    c = *csr;
    c.current = c.top;
    while (c.current.offset < csr->current.offset) {
        if (!isSpaceOrTab(nextChar(fb, &c)))
            return FALSE;
    }
    return TRUE;
}

SpdBoolean isNullLine(const MyCursor *csr)
{
    return (csr->next.offset - csr->top.offset) == 1;
}
