/* 
 * Program:             $RCSfile: line.c,v $  $Revision: 30.14 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: line.c,v 30.14 2003/07/20 11:46:07 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

static SpdBoolean checkSkipSharpDirective(IFileBuffer *iFileBuf,
                                          char **cfilename, int *line);

static int scanIntValue(const FileBuffer *fb, MyCursor *csr, int *pValue)
{
    int i, c;
    static char buf[80];                    /* ʸɽĹ */

    if (!isdigit(c = peekChar(fb, csr)) || pValue == NULL) {
        return -1;
    }

    i = 0;

    while (isdigit(c = nextChar(fb,csr)))
        buf[i++] = c;

    buf[i] = '\0';

    *pValue = atoi(buf);

    return 0;
}

/* # lineno "filename" flag ιԤ򥹥󤹤롥
   ͤ flag flag ¸ߤʤ0, Ԥη
   -1 ֤
   filename, lineno  pFilename, pLineno ˤ줾֤
   *pFilename ϸƽ¦ǳɬפ롥*/

#define FLAG_NONE  0
#define FLAG_ENTER 1
#define FLAG_LEAVE 2

static int parseSharpDirective(IFileBuffer *iFileBuf,
                               char **pFilename, int *pLineno)
{
    int			flag;
    MyCursor		csr;
    const FileBuffer    *fb;
    int			i, c;
    static char		tmp[FILENAME_MAX + 1];  

    if (iFileBuf == NULL || pLineno == NULL || pFilename == NULL) {
        return -1;
    }

    fb = &(iFileBuf->fb);
    csr = iFileBuf->csr;

    if (verbose > 10) {
        spdMessage("Parsing # line at %s(%d)..\n", 
                   fb->filename, csr.top.lineno);
    }

    /* Ĥʤäͤ򥻥å */
    *pFilename = NULL;
    *pLineno = -1;

    /* Ƭζ򥹥å */
    skipSpace(fb, &csr);

    /* # ¸ߤå */
    if (nextChar(fb, &csr) != '#') {
        return -1;
    }

    /* 򥹥å */
    skipSpace(fb, &csr);

    /* lineno ɤ߹ */
    if (scanIntValue(fb, &csr, pLineno)) {
        printLine("parseSharpDirective : Can't find lineno in .i line \n", 
                  fb, &csr);
        printCursor("parseSharpDirective : ", fb, csr);

        return -1;
    }

    /* 򥹥å */
    skipSpace(fb, &csr);

    /* ե̾ɤ߹ */
    if (nextChar(fb, &csr) != '"') {
        printLine("parseSharpDirective : Can't find filename in .i line \n", 
                  fb, &csr);
        printCursor("parseSharpDirective : ", fb, csr);
        return -1;
    }

    /* *pFilename = spdMalloc(1); */
    i = 0;
    while ((c = nextChar(fb, &csr)) != '"') {
        /* *pFilename = spdRealloc(*pFilename, i + 2); */
        /* (*pFilename)[i] = c; */
        /* i++; */
        tmp[i++] = c;
    }
    /* (*pFilename)[i] = '\0'; */
    tmp[i] = '\0';
    *pFilename = spdStrdup(tmp);

    /* 򥹥å */
    skipSpace(fb, &csr);

    /* ե饰ɤ߹ */
    /* Ԥäե饰̵ */

    if (peekChar(fb, &csr) == '\n') {
        flag = FLAG_NONE;
    } else {
        scanIntValue(fb, &csr, &flag);
    }

    if (verbose) {
        spdMessage("parseSharpLine: %s(%d): Found # line to %s(%d) flag %d\n", 
                   iFileBuf->fb.filename, iFileBuf->csr.top.lineno,
                   *pFilename, *pLineno, flag);
    }

    return flag;
}

/* 
 * ֹĴ򼨤"#"ԤϤ롥
 * cfilenameˤϡ"#"Ԥ˴ޤޤƤ".c"Υե̾, lineˤ
 * ֹꤷ֤
 */

static SpdBoolean checkSkipSharpDirective(IFileBuffer *iFileBuf,
                                          char **cfilename, int *line)
{
    int flag;
    SpdBoolean res = FALSE;
    int l;
    char *filename;

    /* ⤷֤ʤмιԤذư */
    if (iFileBuf->csr.current.offset == iFileBuf->csr.next.offset) {
        skipLogicalLine(&(iFileBuf->fb), &(iFileBuf->csr), 1);
    }

    flag = parseSharpDirective(iFileBuf, &filename, &l);
    
    if (flag == FLAG_NONE) {
        *line = l;
        *cfilename = filename;
        res = TRUE; 

        /* cloffset  */
        iFileBuf->skip.cloffset = (*line - 1) - iFileBuf->csr.top.lineno;
    }

    return res;
}

static int skipILine_in(IFileBuffer *iFileBuf, const char *cfilename,
                        int clineno, int nlines)
{
    int     res = 0;

    /* nlinesϻȤʤ */
    if (verbose > 5) {
        spdMessage("skipILine: %s(%d):%d: cloffset:%d    to:%s(%d) skipto:%d\n",
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno,
                   iFileBuf->csr.current.offset, iFileBuf->skip.cloffset,
                   cfilename, clineno,
                   iFileBuf->skip.lineno);
    }

    if (clineno < 0) {
        /* .c бԤ̵(# Ԥν) */
        /* ñ1ɤФ */
        return skipLogicalLine(&(iFileBuf->fb), &(iFileBuf->csr), 1);
    }

    if (clineno < iFileBuf->skip.lineno || 
        (iFileBuf->skip.lineno > 0 && iFileBuf->skip.lineno + 1 < clineno) ||
        (iFileBuf->skip.filename && 
         !spdSafetyStreq(iFileBuf->skip.filename, cfilename))) {
        /* # Ԥιֹ礭Ϥʤˤ⤷ʤ */
        return 0;
    }

    /* #if Ÿʤäˡ롼ˤåɬ */
    if (iFileBuf->skip.lineno < 0) {
        if (checkSkipSharpDirective(iFileBuf, &(iFileBuf->skip.filename),
                                    &(iFileBuf->skip.lineno))) {
            if (verbose > 10) {
                spdMessage("skipILine: found # line to %s(%d), current: %d\n", 
                           iFileBuf->skip.filename, iFileBuf->skip.lineno,
                           clineno);
            }
        } else {
            if (verbose > 10) {
                spdMessage("skipILine: normal skip.\n");
            }
        }
    }

    if (clineno < iFileBuf->skip.lineno || 
        (iFileBuf->skip.filename && 
         !spdSafetyStreq(iFileBuf->skip.filename, cfilename))) {
        /* # Ԥιֹ礭Ϥʤˤ⤷ʤ */
        return 0;
    }

    if (verbose > 10) {
        spdMessage("skipILine: enter loop. %s(%d) to %d\n", cfilename,
                   iFileBuf->csr.top.lineno + iFileBuf->skip.cloffset,
                   clineno);
                   
    }
    while ((iFileBuf->skip.lineno >= 0 && iFileBuf->skip.lineno < clineno &&
            clineno <= iFileBuf->skip.lineno + 1) ||
           (iFileBuf->skip.lineno < 0 && 
            iFileBuf->csr.top.lineno + iFileBuf->skip.cloffset < clineno)) {
        if (iFileBuf->skip.lineno >= 0) {
            /* # ԤˤäĴξ */
        
            if (iFileBuf->skip.lineno <= clineno &&
                spdSafetyStreq(iFileBuf->skip.filename, cfilename)) {
                if (verbose > 10) {
                    spdMessage("skipILine: # line matching is end. "
                               "skipto %d and %s(%d)\n", 
                               iFileBuf->skip.lineno, cfilename, clineno);
                }
                iFileBuf->skip.lineno = -1;
                spdFree(iFileBuf->skip.filename);
                iFileBuf->skip.filename = NULL;
            } else {
                if (verbose > 10) {
                    spdMessage("skipILine: # line matching now.\n");
                }
                continue;
            }
        
        }

        /* ̾Υޥå󥰤ξ */
        res += skipLogicalLine(&(iFileBuf->fb), &(iFileBuf->csr), 1);

        if (iFileBuf->skip.lineno < 0) {
            if (checkSkipSharpDirective(iFileBuf, &(iFileBuf->skip.filename),
                                        &(iFileBuf->skip.lineno))) {
                if (verbose > 10) {
                    spdMessage("skipILine: found # line to %s(%d),"
                               " current: %d\n", 
                               iFileBuf->skip.filename, iFileBuf->skip.lineno,
                               clineno);
                }
                break;
            } else {
                if (verbose > 10) {
                    spdMessage("skipILine: normal skip.\n");
                }
            }
        }
    }
    if (verbose > 10) {
        spdMessage("skipILine: exit loop.\n");
    }

    return res;
}

int skipILine(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf, int nlines)
{
    return skipILine_in(iFileBuf, cFileBuf->lineInfo.filename, 
                        cFileBuf->csr.top.lineno + cFileBuf->lineInfo.loffset,
                        nlines);
}

static void finishSkipSharpLine(IFileBuffer *iFileBuf)
{
    iFileBuf->skip.lineno = -1;
    spdFree(iFileBuf->skip.filename);
    iFileBuf->skip.filename = NULL;
    skipLogicalLine(&(iFileBuf->fb), &(iFileBuf->csr), 1);
}

SpdBoolean isSkipedLine(const CFileBuffer *cFileBuf, IFileBuffer *iFileBuf)
{
    /* # ԤˤäɤФƤʤ TRUE */
    if (iFileBuf->skip.lineno >
        cFileBuf->csr.current.lineno + cFileBuf->lineInfo.loffset) {
        return TRUE;
    }

    /* ɤФä iFileBuf->skip  */
    if (iFileBuf->skip.lineno >= 0 && 
        spdSafetyStreq(cFileBuf->lineInfo.filename,
                       iFileBuf->skip.filename)) {
        finishSkipSharpLine(iFileBuf);
    }

    return isNullLine(&(iFileBuf->csr)) && !isNullLine(&(cFileBuf->csr));
}

SpdBoolean checkEnterSharpDirective(IFileBuffer *iFileBuf,
                                    char **cfilename)
{
    int line, flag;
    SpdBoolean res = FALSE;

    /* ⤷ƬǤʤмιԤذư */
    if (iFileBuf->csr.top.offset != iFileBuf->csr.current.offset) {
        skipLogicalLine(&(iFileBuf->fb), &(iFileBuf->csr), 1);
    }

    flag = parseSharpDirective(iFileBuf, cfilename, &line);
    
    if (flag < 0) {
        res = FALSE;
    } else {
        if (flag != FLAG_LEAVE && line == 1) {
            res = TRUE;

            /* cloffset  */
            iFileBuf->skip.cloffset = (line - 1) - iFileBuf->csr.top.lineno;
        }
    }

    skipILine_in(iFileBuf, *cfilename, -1, 1);

    return res;
}

/* եν(include )򼨤 # Ԥ */
/* cfilename ȥե̾פʤХ顼 */

SpdBoolean checkLeaveSharpDirective(IFileBuffer *iFileBuf,
                                    const char *cfilename, int *line)
{
    int flag;
    char *filename = NULL;
    SpdBoolean res = FALSE;

    /* ⤷ƬǤʤмιԤذư */
    if (iFileBuf->csr.top.offset != iFileBuf->csr.current.offset) {
        skipLogicalLine(&(iFileBuf->fb), &(iFileBuf->csr), 1);
    }

    flag = parseSharpDirective(iFileBuf, &filename, line);
    
    if (flag == FLAG_LEAVE) {
        if (spdSafetyStreq(cfilename, filename)) {
            res = TRUE;

            /* cloffset  */
            iFileBuf->skip.cloffset = (*line - 1) - iFileBuf->csr.top.lineno;
        }
    }

    if (filename)
        spdFree(filename);

    skipILine_in(iFileBuf, cfilename, -1, 1);

    return res;
}    

void parseLineDirective(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf,
                        MyCursor *paramcsr)
{
    int     i, c;
    int     clineno;
    /* char *filename = NULL; */
    static char tmp[FILENAME_MAX + 1];  

    /* lineno ɤ߹ */
    if (scanIntValue(&(cFileBuf->fb), paramcsr, &clineno)) {
        printLine("parseLineDirective : Can't find lineno in .c line \n", 
                  &(cFileBuf->fb), paramcsr);
        printCursor("parseLineDirective : ", &(cFileBuf->fb), *paramcsr);

        return;
    }

    /* 򥹥å */
    skipSpace(&(cFileBuf->fb), paramcsr);

    /* ե̾ɤ߹ */
    if (nextChar(&(cFileBuf->fb), paramcsr) == '"') {
        /* filename = spdMalloc(1); */
        i = 0;
        while ((c = nextChar(&(cFileBuf->fb), paramcsr)) != '"') {
            /* filename = spdRealloc(filename, i + 2); */
            /* filename[i] = c; */
            /* i++; */
            tmp[i++] = c;
        }
        /* filename[i] = '\0'; */
        tmp[i] = '\0';
    
        /* if (!SPD_STREQ(filename, cFileBuf->lineInfo.filename)) { */
        if (!SPD_STREQ(tmp, cFileBuf->lineInfo.filename)) {
            spdFree(cFileBuf->lineInfo.filename);
            /* cFileBuf->lineInfo.filename = spdStrdup(filename); */
            cFileBuf->lineInfo.filename = spdStrdup(tmp);
        }
        cFileBuf->lineInfo.loffset = clineno - cFileBuf->csr.top.lineno - 1;

        if (verbose) {
            spdMessage("parseLineDirective: Set filename %s offset %d.\n",
                       cFileBuf->lineInfo.filename,
                       cFileBuf->lineInfo.loffset);
        }
        getNextMacroCall(cFileBuf, iFileBuf);

        /* spdFree(filename); */
    }

    if (iFileBuf->skip.lineno >= 0) {
        /* .i  # ԤɤФʤ顢Ĵ */
        if (iFileBuf->skip.lineno ==
            cFileBuf->csr.current.lineno + cFileBuf->lineInfo.loffset &&
            spdSafetyStreq(cFileBuf->lineInfo.filename,
                           iFileBuf->skip.filename)) {
            /* ɤФä iFileBuf->skip  */
            if (verbose) {
                spdMessage("#line at %s(%d) have extracted "
                           "# line at %s(%d).\n",
                           cFileBuf->fb.filename, 
                           cFileBuf->csr.current.lineno,
                           iFileBuf->fb.filename, 
                           iFileBuf->csr.current.lineno);
                
            }
            finishSkipSharpLine(iFileBuf);

            return;
        }
    }
    /* ǤʤСWarning ФɤФ */
    if (iFileBuf->skip.lineno < 0) {
        spdWarning("#line at %s(%d) doesn't have extracted "
                   "# line at %s(%d).\n",
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno);
    } else if (iFileBuf->skip.lineno !=
               cFileBuf->csr.current.lineno + cFileBuf->lineInfo.loffset) {
        spdWarning("#line at %s(%d) and # line at %s(%d) has "
                   "different lineno %d and %d.\n",
                   cFileBuf->fb.filename, cFileBuf->csr.current.lineno,
                   iFileBuf->fb.filename, iFileBuf->csr.current.lineno,
                   cFileBuf->csr.current.lineno + cFileBuf->lineInfo.loffset,
                   iFileBuf->skip.lineno);
    }
               
    /* !! Ȥꤢ # ԤޤɤФ?? */
}
