/* 
 * Program:             $RCSfile: macrocall.c,v $  $Revision: 30.14 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: macrocall.c,v 30.14 2003/07/08 11:52:00 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <limits.h>
#include <Sapid/Sapid.h>
#include "mkPmodel.h"

#define COMPATIBLE_mkPIDB4a

static SpdObjId parseMacroCall(const CFileBuffer *cFileBuf, int offset,
                               int expId);

void getNextMacroCall(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf)
{
    /* PIDB 򸡺 coffset, clength, ioffset, ilength  */
    int offset;
    int i;
    EXPANDinfo *pexp;

    offset = iFileBuf->csr.current.offset;
    if (verbose > 1) {
        spdMessage("getNextMacroCall : search Macro at %s:%d from ID %d\n", 
                   iFileBuf->fb.filename, offset, iFileBuf->mExpId);
    }
    cFileBuf->nextMcall.coffset = INT_MAX;
    cFileBuf->nextMcall.clength = 0;

    for (i = iFileBuf->mExpId + 1; i < expandInfoN; i++) {
        pexp = &(expandInfo[i]);

        if (pexp->input.type == CALL_FILE && 
            pexp->output.start >= offset) {

            if (verbose > 1) {
                spdMessage("getNextMacroCall : find Macro %d at %s:%d\n", 
                           i,
                           iFileBuf->fb.filename, pexp->output.start);
                spdMessage("expInfo : %s  cFileBuf : %s\n",
                           pexp->input.fname, cFileBuf->lineInfo.filename);
            }

            if (SPD_STREQ(pexp->input.fname, cFileBuf->lineInfo.filename)) {
                cFileBuf->nextMcall.id = i;
                cFileBuf->nextMcall.coffset = pexp->input.start;
                cFileBuf->nextMcall.clength = 
                    pexp->input.end - pexp->input.start + 1;
                
                cFileBuf->nextMcall.ioffset = pexp->output.start;
                cFileBuf->nextMcall.ilength = 
                    pexp->output.end - pexp->output.start + 1;
            }

            break;
        }
    }

    /* macrocall Ĥʤä */
    if (cFileBuf->nextMcall.coffset == INT_MAX) {
        if (verbose > 1) {
            spdMessage("getNextMacroCall : not found Macro after %s(%d)\n", 
                       iFileBuf->fb.filename, offset);
        }
    }
}

static int calcOffset(const EXPANDinfo *pexp)
{
    int offset;

    if (verbose > 3)
        spdMessage("calcOffset: %d\n", pexp->ID);

    switch (pexp->input.type) {
      case CALL_FILE:
        if (verbose > 3)
            spdMessage("calcOffset: CALL_FILE %d\n", pexp->input.start);
        return pexp->input.start;
        break;
      case CALL_MACROEXPANSION:
        if (verbose > 3)
            spdMessage("calcOffset: CALL_MACROEXPANSION %d\n",
                       pexp->input.start);
        return -1;
        break;
      case CALL_MACROARG:
        if (pexp->input.macexp->input.type == CALL_FILE) {
            return pexp->input.macexp->arg[pexp->input.argn].start;
        } else {
            offset = calcOffset(pexp->input.macexp);
            if (offset < 0) {
                return offset;
            } else {
                return (offset + pexp->input.start +
                        pexp->input.macexp->arg[pexp->input.argn].start);
            }
        }
      default:
        return -1;
    } 
}

static void getMacroParam(const CFileBuffer *cFileBuf, PPRegion *reg, 
                          EXPANDinfo *pexp, int n)
{
    MacroArgInfo *parg;
    int i;

    parg = &(pexp->arg[n]);
    if (verbose > 1) {
        spdMessage("getMacroParam: expID:%d argn:%d\n", pexp->ID, n);
        spdMessage("getMacroParam: start:%d end:%d\n", parg->start, parg->end);
    }

    reg->offset = calcOffset(pexp);
    if (reg->offset >= 0) {
        if (verbose > 1)
            spdMessage("getMacroParam: offset %d + %d\n", reg->offset,
                       (parg->start - pexp->input.start));
        reg->offset += (parg->start - pexp->input.start);
        reg->length = parg->end - parg->start + 1;

        if (reg->offset + reg->length > cFileBuf->fb.size) {
            spdWarning("getMacroParam: Macro parameter %d of expID %d has "
                       "illegal offset %d + length %d.\n", n, pexp->ID,
                       reg->offset, reg->length);

            reg->length = 0;
        } else {
            strncpy(reg->str, cFileBuf->fb.buf + reg->offset, reg->length);
            /* \n 򥹥ڡ֤ */
            for (i = 0; i < reg->length; i++) {
                if (reg->str[i] == '\n') {
                    reg->str[i] = ' ';
                }
            }
        }
    } else {
        reg->offset = -1;
        reg->length = 0;
    }
    
    reg->str[reg->length] = '\0';

    if (verbose > 1) {
        spdMessage("getMacroParam: offset %d + %d chars in .c\n",
                   reg->offset, reg->length);
    }
}

/******************************************************************************/

#if defined(USE_MACROBODY_HASH)

static SpdSet macroBodyHash;

void initMacroBodyHash(void)
{
    macroBodyHash = spdCreateSet(sizeof(int));
}

void addMacroBodyHash(SpdObjId macroNameId, SpdObjId macroBodyId)
{
    spdAddElem(macroBodyHash, macroNameId, &macroBodyId);
}

void freeMacroBodyHash(void)
{
    spdFreeSet(macroBodyHash);
}

#endif /* defined(USE_MACROBODY_HASH) */

/******************************************************************************/

static SpdObjId getMacroBodyId(SpdObjId macroNameId)
{
#if defined(USE_MACROBODY_HASH)

    SpdObjId    *pmacroBodyId;

    pmacroBodyId = spdSearchElem(macroBodyHash, macroNameId);
    return (pmacroBodyId ? *pmacroBodyId : SPD_NON_ID);

#else

    SpdObjId    macroPatternId, pblockId;
    
    macroPatternId = spdGetAParentObjId(macroNameId, SPD_RID_MPATT_MNAME);
    pblockId = spdGetAParentObjId(macroPatternId, SPD_RID_DEF_MPATT);
    return spdGetAChildObjId(pblockId, SPD_RID_DEF_MBODY);

#endif
}

static SpdObjId parseMacroCall(const CFileBuffer *cFileBuf, int offset,
                               int expId)
{
    PPRegion macroCall, macroParam;
    EXPANDinfo *pexp, *pargexp;
    DEFINEinfo *pdef;
    SpdObjId macroNameId, macroCallId, macroBodyId;
    SpdRelId relId;
    int i, j;
    int off;

    if (expId < 0 || expId > expandInfoN) {
        spdMessage("Illegal expandInfo Id %d\n", expId);
        return SPD_NON_ID;
    }

    pexp = &(expandInfo[expId]);
    pdef = pexp->mac;

    if (verbose) {
        spdMessage("parseMacroCall: offset:%d  PIDB:expID:%d start:%d[%d]\n",
                   offset, expId + 1, pexp->input.start, pexp->input.type);
    }

    macroCall.length = pexp->input.end - pexp->input.start + 1;
    macroCall.offset = calcOffset(pexp);

    macroNameId = searchMacroIndex(pdef->ID);
    if (macroNameId == SPD_NON_ID) {
        spdErrorExit("Can't find index of macroID:%d\n", pdef->ID);
    }

    macroCall.objId = spdCreatePmodelObj(SPD_CID_MACROCALL);
    spdSetAttrIntValByName(macroCall.objId, "length", macroCall.length);
    relId = spdCreatePmodelRel(SPD_RID_MCALL_MNAME, 
                               macroCall.objId, macroNameId);
    spdSetAttrIntValByName(relId, "offset", macroCall.offset);

    if (verbose) {
        spdMessage("parseMacroCall: macroCall: %d name: %s offset: %d length: %d\n", 
                   macroCall.objId, 
                   spdGetAttrStrValByName(macroNameId, "name"),
                   macroCall.offset, macroCall.length);
    }

    if (pdef->narg >= 0) {
        for (i = 0; i < pdef->narg; i++) {
            getMacroParam(cFileBuf, &macroParam, pexp, i);
            macroParam.objId = spdCreatePmodelObj(SPD_CID_MACROCALLARG);
            spdSetAttrStrValByName(macroParam.objId, "name", macroParam.str);
            spdSetAttrIntValByName(macroParam.objId, 
                                   "length", macroParam.length);
            relId = spdCreatePmodelRel(SPD_RID_MCALL_MCALLA,
                                       macroCall.objId, macroParam.objId);
            spdSetAttrIntValByName(relId, "offset", macroParam.offset);

            for (j = 0; j < expandInfoN; j++) {
                pargexp = &(expandInfo[j]);
        
                if (pargexp->input.type == CALL_MACROARG && 
                    /* !! macexpID ź */
                    pargexp->input.macexpID + 1 == pexp->ID &&
                    pargexp->input.argn == i) {
                    /* ޥ°˥ޥƽФޤ硤
                       mcalla_mcall  */
                    off = calcOffset(pargexp);
                    macroCallId = parseMacroCall(cFileBuf, off, j);
                    relId = spdCreatePmodelRel(SPD_RID_MCALLA_MCALL,
                                               macroParam.objId,
                                               macroCallId);
                    spdSetAttrIntValByName(relId, "offset", off);
                }
            }
        }
    }

    /* ޥŸ̤˥ޥƽФޤ硤mbody_mcall  */
    /* !! ϡP-model  macroBody ʤΤŸʤΤ
      ̤ǤƤʤȤˤΡǥνʷ */
    macroBodyId = SPD_NON_ID;
    for (i = 0; i < expandInfoN; i++) {
        pargexp = &(expandInfo[i]);
    
        if (pargexp->input.type == CALL_MACROEXPANSION && 
            /* !! macexpID ź */
            pargexp->input.macexpID + 1 == pexp->ID) {
        
            if (macroBodyId == SPD_NON_ID) {
                /* macroBodyId ɬפˤʤޤǸʤ */
                macroBodyId = getMacroBodyId(macroNameId);
            }
#if defined(COMPATIBLE_mkPIDB4a)
            relId = spdGetAParentRelId(macroBodyId, SPD_RID_DEF_MBODY);
            off = spdGetOffset(relId);
#else
            off = calcOffset(pargexp);
#endif
            macroCallId = parseMacroCall(cFileBuf, off, i);
            relId = spdCreatePmodelRel(SPD_RID_MBODY_MCALL,
                                       macroBodyId, macroCallId);
            spdSetAttrIntValByName(relId, "offset", off);
        }
    }

    return macroCall.objId;
}

SpdObjId matchMacroCall(CFileBuffer *cFileBuf, IFileBuffer *iFileBuf)
{
    SpdObjId pblockId, macroCallId, iblockId;
    SpdRelId relId;
    int clines, ilines;

    if (verbose > 3) {
        spdMessage("matchMacroCall: skip %d chars in .c and %d chars in .i\n",
                   cFileBuf->nextMcall.clength, cFileBuf->nextMcall.ilength);
        spdMessage("matchMacroCall: from %d at .c and %d at .i\n",
                   cFileBuf->csr.current.offset, iFileBuf->csr.current.offset);
    }

    pblockId = createPblock(cFileBuf->fb.fileId, PBLOCK_MACROCALL, 
                            cFileBuf->nextMcall.coffset, 
                            cFileBuf->nextMcall.clength);

    macroCallId = parseMacroCall(cFileBuf, cFileBuf->csr.current.offset,
                                 cFileBuf->nextMcall.id);

    relId = spdCreatePmodelRel(SPD_RID_CALL_MCALL, pblockId, macroCallId);
    spdSetAttrIntValByName(relId, "offset", cFileBuf->nextMcall.coffset);
    
    iblockId = spdCreatePmodelObj(SPD_CID_IBLOCK);
    spdSetAttrIntValByName(iblockId, "length", 
                           cFileBuf->nextMcall.ilength);
    relId = spdCreatePmodelRel(SPD_RID_IFILE_IBLOCK,
                               iFileBuf->fb.fileId, iblockId);
    spdSetAttrIntValByName(relId, "offset", cFileBuf->nextMcall.ioffset);
    relId = spdCreatePmodelRel(SPD_RID_PBLOCK_IBLOCK, 
                               pblockId, iblockId);

    /* PIDB 򿮤ƥå */
    clines = skipToOffset(&(cFileBuf->fb), &(cFileBuf->csr), 
                          cFileBuf->nextMcall.coffset + cFileBuf->nextMcall.clength);

    if (iFileBuf->csr.current.offset != cFileBuf->nextMcall.ioffset) {
        if (verbose)
            spdMessage("matchMacroCall: skiped macrocall at %s(%d).\n",
                       iFileBuf->fb.filename, iFileBuf->csr.current.lineno);
        ilines = 0;
    } else {
        ilines = skipToOffset(&(iFileBuf->fb), &(iFileBuf->csr), 
                              cFileBuf->nextMcall.ioffset + cFileBuf->nextMcall.ilength);
        if (clines != ilines) {
            /* Ÿ̤ιԿ㤦硤.i μιԤ # ԤԤ */
            iFileBuf->skip.fpadding = TRUE;
        }
    }


    if (verbose > 3) {
        spdMessage("matchMacroCall: to %d at .c and %d at .i\n",
                   cFileBuf->csr.current.offset, iFileBuf->csr.current.offset);
    }

    iFileBuf->mExpId = cFileBuf->nextMcall.id;

    return macroCallId;
}
