/* 
 * Program:             $RCSfile: mkPmodel.c,v $  $Revision: 30.21 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1998/12/01
 *                      S.Yamamoto  1999/01/23
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1998 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkPmodel.c,v 30.21 2003/08/16 10:33:14 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelAR.h>
#include <Sapid/ImodelARcompat.h>
#include "mkPmodel.h"

static void makePSDB(SpdModelId PmodelId, SpdObjId fileId)
{
    static char psdb_path[FILENAME_MAX + 1];
    SpdInfo *info;

    makePmodelOfIFile(fileId);

    /*
     * "PSDB/<fileId>" is set as the PSDB directory and
     * "PSDB/<fileId>.info" is set as the Info file.
     * These structures should be shared with PIDB and ISDB.
     */

    sprintf(psdb_path, "%s/%s", PSDB_DIR, spdObjIdStr(fileId));
    spdSetDBPathOfModel(PmodelId, psdb_path);
    spdSetSDBinfoFile(PSDB_DIR, spdObjIdStr(fileId));

    info = spdOpenSDBinfo();
    spdSetInfoString(info, "BINARY_FORMAT_SDB",
                     (spdGetWriteSDBformat() == SAPID_BINARY_FORMAT_SDB ?
                      "TRUE" : "FALSE"));
    spdCloseInfo(info);
}

int verbose = 0;

int main(int argc, char **argv)
{
    SpdModelId  ImodelId, PmodelId;
    SpdDBId     dbId;
    const char  *target_program_name = NULL;
    char        *target_file_name = NULL;
    SpdObjId    progId, fileId;
    SpdIdSet    files;
    SpdInfo     *info;
    int         c;
    SetCursor   csr;

    spdInitialize(spdBasename(argv[0]));

    while ((c = getopt(argc, argv, "d:e:f:t:vV:")) != EOF) {
        switch (c) {
          case 'd':
            spdSetSDBdir(optarg);
            break;
          case 'e':
            spdSetInputEncode(optarg);
            break;
          case 'f':
            target_file_name = spdStrdup(optarg);
            break;
          case 't':
            /* target_program_name = spdStrdup(optarg); */
            target_program_name = optarg;
            break;
          case 'v':
            verbose ++;
            break;
          case 'V':
            verbose = atoi(optarg);
            break;
          case 'h':
          default:
            spdErrorExit("Usage: %s [-v][-d dir][-e encoding][-f i-fileName][-t targetProgramName][-V level]\n"
                         "    -d dir     : Directory for SDB (\"SDB\").\n"
                         "    -e encoding: Specify input encoding (\"C\").\n"
                         "    -f file    : Analyze only one file.\n"
                         "    -t taget   : Specify target name.\n"
                         "    -v         : Verbose mode.\n"
                         "    -V         : Set verbose level.\n"
                         , argv[0]);
            break;
        }
    }

    info = spdOpenSDBinfo();
    if (SPD_STREQ(spdGetInfoString(info, "BINARY_FORMAT_SDB"), "TRUE")) {
        spdSetReadSDBformat(SAPID_BINARY_FORMAT_SDB);
        spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
    }
    spdCloseInfo(info);

    ImodelId = spdUseModel(&I_model);
    PmodelId = spdUseModel(&P_model);
    
    dbId = spdOpenSDB();
    spdReadDB(ImodelId, dbId);
    spdCreateDB(PmodelId, dbId);

#if defined(USE_MACROBODY_HASH)
    initMacroBodyHash();
#endif

    if (target_file_name) {
        fileId = spddbGetAnObjId(SPD_CID_FILE, target_file_name);
        if (fileId == SPD_NON_ID) {
            spdErrorExit("\"%s\" is not found.\n", target_file_name);
        }
        spdMessage("File name \"%s\".\n", target_file_name);
        spdFree(target_file_name);
        makePSDB(PmodelId, fileId);
    } else {
        progId = spddbGetAnObjId(SPD_CID_PROGRAM, target_program_name);
        if (target_program_name == NULL) {
            /* target_program_name = spdStrdup(spdGetName(progId)); */
            target_program_name = spdGetName(progId);
            spdMessage("Program name \"%s\".\n", target_program_name);
        }
        files = spdGetChildObjIdSet(progId, SPD_RID_PROG_FILE);
        for (csr = spdSetCursorCreate(files); spdSetCursorNext(&csr);) {
            fileId = spdSetGetValueO(csr.curElem->value);
            makePmodelOfIFile(fileId);
        }
        spdFreeSet2(files);
    }

#if defined(USE_MACROBODY_HASH)
    freeMacroBodyHash();
#endif

    spdMessage("Writing P-model database.\n");
    spdCommitDB(PmodelId, dbId);

    /* spdFree(target_program_name); */
    spdCloseSDB(dbId);

    return (EXIT_SUCCESS);
}
