/* 
 * Program:             $RCSfile: builtin.c,v $  $Revision: 30.12 $
 * 
 * Purpose:
 * 
 * Author:              H.Ariga  1996/01/27
 *                      S.Yamamoto  1996/01/27
 * 
 * (C) Copyright:       H.Ariga and S.Yamamoto  1996 - 2003
 *                      This file is a product of the project Sapid.
 */

#include <stdio.h> 
#include <Sapid/Sapid.h>
#include "main.h"
#include "y.tab.h"
#include "prototype.h"

/*
 * Note, file name "SDB_conf.c", function name "load_SDB_conf()" are better.
 */

/* #define DEBUG_SDB_CONF */

#if defined(lint)
#define DW_SDB_CONF         printf      /* Just for cheking. */
#elif defined(DEBUG_SDB_CONF)
#define DW_SDB_CONF         spdMessage
#else
#define DW_SDB_CONF         (void)
#endif

#define SDB_CONF_FILE       "SDB.conf"

#define MAX_BUILTIN_FUNC                8
#define MAX_SKIPPED_TOKEN               32
#define MAX_SKIPPED_TOKEN_WITH_ARG      32
#define MAX_BUILTIN_CONST               8
#define MAX_BUILTIN_TYPE                8

#define MODE_UNKNOWN                    0
#define MODE_BUILTIN_FUNC               1
#define MODE_SKIPPED_TOKEN              2
#define MODE_SKIPPED_TOKEN_WITH_ARG     3
#define MODE_BUILTIN_CONST              4
#define MODE_BUILTIN_TYPE               5

static int      builtin_func_num = 0;
static char     builtin_funcs[MAX_BUILTIN_FUNC][SPD_ID_LEN + 1];

static int      skipped_token_num = 0;
static char     skipped_tokens[MAX_SKIPPED_TOKEN][SPD_ID_LEN + 1];

static int      skipped_token_with_arg_num = 0;
static char     skipped_tokens_with_arg[MAX_SKIPPED_TOKEN_WITH_ARG][SPD_ID_LEN + 1];

static int      builtin_const_num = 0;
static char     builtin_consts[MAX_BUILTIN_CONST][SPD_ID_LEN + 1];

static int      builtin_type_num = 0;
static char     builtin_types[MAX_BUILTIN_TYPE][SPD_ID_LEN + 1];

static int      mode = MODE_UNKNOWN;

void load_builtin(void)
{
    char    filename[FILENAME_MAX + 1];
    char    line_buff[SPD_LINE_LEN + 1];
    char    *cp, *env; 
    FILE    *fp;

    filename[0] = '\0'; 
    if ((env = getenv("SAPID_SDB_CONFDIR")) == NULL) {
        strcpy(filename, ".");
    } else {
        strcpy(filename, env);
    }
    strcat(filename, "/" SDB_CONF_FILE);

    if ((fp = spdSimpleFopen(filename, SPD_FOPEN_RT_MODE)) == NULL){
        spdWarning("\"%s\": Can't find file.\n", filename);
        return;
    }
    while (fgets(line_buff, SPD_LINE_LEN + 1, fp) != NULL) {
        /* Trim the last '\n'. */
        if ((cp = strrchr(line_buff, '\n')) != NULL) {
            *cp = '\0';
        } 

        if (line_buff[0] == '#' || line_buff[0] == '\0') {
            /* Skip a comment line and an empty line. */
            continue;
        }
        switch (line_buff[1]) {
        case 'b':
            if (SPD_STREQ(line_buff, "[builtin_type_parameter]")) {
                mode = MODE_BUILTIN_FUNC;
                continue;
            } else if (SPD_STREQ(line_buff, "[builtin_constant]")) {
                mode = MODE_BUILTIN_CONST;
                continue;
            } else if (SPD_STREQ(line_buff, "[builtin_type]")) {
                mode = MODE_BUILTIN_TYPE;
                continue;
            }
            break;
        case 's':
            if (SPD_STREQ(line_buff, "[skipped_token]")) {
                mode = MODE_SKIPPED_TOKEN;
                continue;
            } else if (SPD_STREQ(line_buff, "[skipped_token_with_arg]")) {
                mode = MODE_SKIPPED_TOKEN_WITH_ARG;
                continue;
            }
            break;
        }

        if (SPD_ID_LEN < strlen(line_buff)) {
            spdWarning("Too long line: \"%s\".\n", line_buff);
            continue;
        } else {
            SpdBoolean too_many = FALSE;
            switch (mode) {
            case MODE_BUILTIN_FUNC:
                if (builtin_func_num < MAX_BUILTIN_FUNC) {
                    DW_SDB_CONF("load_builtin(): builtin_func: %d: \"%s\"\n",
                                builtin_func_num, line_buff);
                    strcpy(builtin_funcs[builtin_func_num++], line_buff);
                } else {
                    too_many = TRUE;
                }
                break;
            case MODE_SKIPPED_TOKEN:
                if (skipped_token_num <= MAX_SKIPPED_TOKEN) {
                    DW_SDB_CONF("load_builtin(): skipped_token: %d: \"%s\"\n",
                                skipped_token_num, line_buff);
                    strcpy(skipped_tokens[skipped_token_num++], line_buff);
                } else {
                    too_many = TRUE;
                }
                break;
            case MODE_SKIPPED_TOKEN_WITH_ARG:
                if (skipped_token_with_arg_num < MAX_SKIPPED_TOKEN_WITH_ARG) {
                    DW_SDB_CONF("load_builtin(): skipped_token_with_arg: %d: \"%s\"\n",
                                skipped_token_with_arg_num, line_buff);
                    strcpy(skipped_tokens_with_arg[skipped_token_with_arg_num++],
                           line_buff);
                } else {
                    too_many = TRUE;
                }
                break;
            case MODE_BUILTIN_CONST:
                if (builtin_const_num <= MAX_BUILTIN_CONST) {
                    DW_SDB_CONF("load_builtin(): builtin_const: %d: \"%s\"\n",
                                builtin_const_num, line_buff);
                    strcpy(builtin_consts[builtin_const_num++], line_buff);
                }
                break;
            case MODE_BUILTIN_TYPE:
                if (builtin_type_num <= MAX_BUILTIN_TYPE) {
                    DW_SDB_CONF("load_builtin(): builtin_type: %d: \"%s\"\n",
                                builtin_type_num, line_buff);
                    strcpy(builtin_types[builtin_type_num++], line_buff);
                } else {
                    too_many = TRUE;
                }
                break;
            }
            if (too_many) {
                spdWarning("Too many entries: \"%s\".\n", line_buff);
                break;
            }
        }
    }
    spdFclose(fp);
}

int is_builtin_function(char *str)
{
    int     i;

    for (i = 0; i < builtin_func_num; i++) {
        if (SPD_STREQ(str, builtin_funcs[i])) {
            return (TRUE);
        }
    }
    return (FALSE);
}

int is_skipped_token(char *str)
{
    int     i;

    for (i = 0; i < skipped_token_num; i++) {
        if (SPD_STREQ(str, skipped_tokens[i])) {
            return (TRUE);
        }
    }
    return (FALSE);
}

int is_skipped_token_with_arg(char *str)
{
    int     i;

    for (i = 0; i < skipped_token_num; i++) {
        if (SPD_STREQ(str, skipped_tokens_with_arg[i])) {
            return (TRUE);
        }
    }
    return (FALSE);
}

int is_builtin_constant(char *str, char *type)
{
    int     i;

    for (i = 0; i < builtin_const_num; i++) {
        if (SPD_STREQ(str, builtin_consts[i] + 1)) {
            *type = builtin_consts[i][0];
            return (TRUE);
        }
    }
    return (FALSE);
}

int is_builtin_type(char *str)
{
    int     i;

    for (i = 0; i < builtin_type_num; i++) {
        if (SPD_STREQ(str, builtin_types[i])) {
            return (TRUE);
        }
    }
    return (FALSE);
}
