%{

/* 
 * Program:             $RCSfile: lex.l,v $  $Revision: 30.71 $
 * 
 * Purpose:
 * 
 * Author:              K.Mabuchi  1991/09/30
 *                      S.Yamamoto  1992/04/01
 * Modified:            H.Ariga  1993/07/01
 *                      A.Yoshida  1999/06/03
 * 
 * (C) Copyright:       K.Mabuchi, H.Ariga and S.Yamamoto  1991 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: lex.l,v 30.71 2003/07/30 16:27:31 yamamoto Exp yamamoto $";
#endif

#define CHECK_ONLY_FIRST_CHAR
/* reserved words can be identified by only the first charactor
   in each rule. */

#define IN_LEX              1

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "main.h"
#include "y.tab.h"
#include "prototype.h"

/*
 * Important note: all actions should end with "count_char();" or
 * "return (TOKEN(XXX));" in order to keep value of variable "c_count" valid.
 */

/*
 * Debug flags.
 * TOKEN_DEBUG: For lex debug.
 */

/* #define TOKEN_DEBUG */

#ifdef TOKEN_DEBUG
#define TOKEN(x)        (spdMessage("yylex(): l.%d, c.%d: \"%s\" -> %s.\n", \
                                    yylineno, c_count, yytext, #x), \
                         spdMessage("       : specifier_flag = %d.\n", \
                                    specifier_flag), \
                         count_char(), (x))
#define TYPE_TOKEN(x)   (TOKEN(x), specifier_flag = 1, (x))
#else
#define TOKEN(x)        (count_char(), (x))
#define TYPE_TOKEN(x)   (TOKEN(x), specifier_flag = 1, (x))
#endif

/*
 * Redefine YYLMAX.
 * Note: yytext is defined as yytext[YYLMAX].
 */

#undef YYLMAX
#define YYLMAX              (SPD_LINE_LEN + 1)

/*
 * For compatibility lex and flex.
 * Note: in flex the undocumented lex scanner internal variable yylineno is
 * not supported unless -l or "%option yylineno" is used. And we assume flex
 * is invoked without -l flag. Variable yylineno is not part of
 * the POSIX specification.
 */

#ifdef USE_FLEX
#define HAVE_NOT_YYLINENO
#endif

#ifdef HAVE_NOT_YYLINENO
int                 yylineno;       /* Line number of ".i" file. */
#endif

/* Source file is specified by "#line" directive. */
int                 source_lineno;  /* Line number of ".c" or ".h" file. */

#ifdef HAVE_NOT_YYLINENO
#define INC_YYLINENO()      (yylineno++, source_lineno++)
#else 
#define INC_YYLINENO()      (source_lineno++)
#endif

#define MAKE_WORD_NAME(x)   spdStrPoolFastCopy(str_pool, (x))

#if !defined(yywrap)
#define yywrap()            1       /* Treat EOF as an EOF. */
#endif

extern YYSTYPE          yylval;

extern int              storage;
extern ScopeListP       scope_list;
extern int              specifier_flag;
extern CSP              ptr_const;
extern SpdStringPool    *str_pool;
extern SpdBoolean       is_D_model;

int                 c_count;        /* Offset in file. */
int                 paren_level;    /* Level of '(' and ')'. */ 
FOP                 tmp_fo;
TypeListP           guessed_type;

static char         str_buf[SPD_STRING_LEN + 1];

/* STWA (Skipped Token With Arg): Added by A.Yoshida 1999/09/09. */
static int          stwa_level;
static int          stwa_buf_index;
static char         stwa_buf[SPD_LINE_LEN + 1];

/* SBIE (Skipped Block In Expression): Added by S.Yamamoto 200/02/28. */
static int          sbie_level;
static int          sbie_buf_index;
static char         sbie_buf[SPD_LINE_LEN + 1];

/* For yacc debug. */
extern int          yydebug;        /* Whether -v option specified or not. */
int                 yydebug_line;   /* Starting line of debug mode. */

static WP       make_yylval(void);
static void     append_yytext(void);
static void     count_char(void);

static IDOP     insert_guessed_typedefname(char *name);

%}

%start  STR COMM COMM_CXX CPPDECL CPPSTR CPPEND STWA SBIE

ESCAPED_CHAR    (\\.|\\[0-7]+|\\x[0-9a-zA-Z]+)

SJIS1           [\201-\237\340-\357]
SJIS2           [\100-\176\200-\374]

%%

<INITIAL>[ \t]+ {
    count_char();
}
<INITIAL>"\n" {
    c_count++;
    INC_YYLINENO();
    if (yydebug_line && yydebug_line <= yylineno) {
        yydebug = 1;
        yydebug_line = 0;
    }
}
<INITIAL>"\014"|"\032" {
    /* Skip '^L' (0xc) and '^Z' (0x1a). */
    c_count++;
}

<INITIAL>^"#" {
    BEGIN(CPPDECL);
    c_count++;
}
<CPPDECL>[ \t] {
    c_count++;
}
<CPPDECL>\" {
    BEGIN(CPPSTR);
    c_count++;
}
<CPPDECL>[0-9]+ {
    source_lineno = atoi(yytext) - 1;
    count_char();
}
<CPPDECL>[^0-9 \t\"\n].* {
    count_char();
}
<CPPDECL>"\n" {
    BEGIN(INITIAL);
    count_char();
    INC_YYLINENO();
}

<CPPSTR>\" {
    BEGIN(CPPEND);
    c_count++;
}
<CPPSTR>([^\"]|{SJIS1}|{SJIS2})* {
    if ((tmp_fo = find_file(yytext)) == NULL) {
        tmp_fo = insert_file(yytext);
    }
    scope_list = insert_scope_list(scope_list, tmp_fo->scope);
    count_char();
}

<CPPEND>[^\n]*"\n" {
    BEGIN(INITIAL);
    count_char();
    INC_YYLINENO();
}

<INITIAL>"/*" {
    BEGIN(COMM);
    c_count += 2;
}
<COMM>"*/" {
    BEGIN(INITIAL);
    c_count += 2;
}
<COMM>"\n" {
    c_count++;
    INC_YYLINENO();
}
<COMM>[ \t]              c_count++;
<COMM>.                  c_count++;

<INITIAL>"//" {
    BEGIN(COMM_CXX);
    c_count += 2;
}
<COMM_CXX>"\n" {
    BEGIN(INITIAL);
    c_count++;
    INC_YYLINENO();
}
<COMM_CXX>[ \t]              c_count++;
<COMM_CXX>.                  c_count++;

<STWA>"(" {
    stwa_level++;
    stwa_buf[stwa_buf_index++] = yytext[0];
    c_count++;
}
<STWA>")" {
    stwa_level--;
    stwa_buf[stwa_buf_index++] = yytext[0];
    if (stwa_level == 0) {
        stwa_buf[stwa_buf_index] = '\0';
        /* spdMessage("yylex(): l.%d, c.%d: Skiped: \"%s\"\n", */
        /*            yylineno, c_count, stwa_buf);            */
        BEGIN(INITIAL);
    }
    c_count++;
}
<STWA>"\n" {
    stwa_buf[stwa_buf_index++] = yytext[0];
    c_count++;
    INC_YYLINENO();
}
<STWA>. {
    stwa_buf[stwa_buf_index++] = yytext[0];
    c_count++;
}

<INITIAL>break           {yylval = make_yylval(); return (TOKEN(BREAK));}
<INITIAL>case            {yylval = make_yylval(); return (TOKEN(CASE));}
<INITIAL>continue        {yylval = make_yylval(); return (TOKEN(CONTINUE));}
<INITIAL>default         {yylval = make_yylval(); return (TOKEN(DEFAULT));}
<INITIAL>do              {yylval = make_yylval(); return (TOKEN(DO));}
<INITIAL>else            {yylval = make_yylval(); return (TOKEN(ELSE));}
<INITIAL>enum            {yylval = make_yylval(); return (TOKEN(ENUM));}
<INITIAL>for             {yylval = make_yylval(); return (TOKEN(FOR));}
<INITIAL>goto            {yylval = make_yylval(); return (TOKEN(GOTO));}
<INITIAL>if              {yylval = make_yylval(); return (TOKEN(IF));}
<INITIAL>return          {yylval = make_yylval(); return (TOKEN(RETURN));}
<INITIAL>sizeof          {yylval = make_yylval(); return (TOKEN(SIZEOF));}
<INITIAL>struct|union {
    yylval = make_yylval();
#ifdef CHECK_ONLY_FIRST_CHAR
    switch (yytext[0]) {
    case 's':  /* struct */
        yylval->sort = TYPE_STRUCT_TAG;
        break;
    case 'u':  /* union */
        yylval->sort = TYPE_UNION_TAG;
        break;
    }
#else
    if (SPD_STREQ("struct", yytext)) {
        yylval->sort = TYPE_STRUCT_TAG;
    } else if (SPD_STREQ("union", yytext)) {
        yylval->sort = TYPE_UNION_TAG;
    }
#endif
#ifdef DEBUG_STRUCT_UNION
    fprintf(stderr, "lex.l: struct or union: sort = %d %s\n",
            yylval->sort, yytext);
#endif
    return (TOKEN(STRUCT_OR_UNION));
}
<INITIAL>switch          {yylval = make_yylval(); return (TOKEN(SWITCH));}
<INITIAL>while           {yylval = make_yylval(); return (TOKEN(WHILE));}
<INITIAL>const|volatile {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
#ifdef CHECK_ONLY_FIRST_CHAR
    switch (yytext[0]) {
    case 'c':  /* const */
        yylval->sort = TYPE_CONST;
        break;
    case 'v':  /* volatile */
        yylval->sort = TYPE_VOLATILE;
        break;
    }
#else
    if (SPD_STREQ("const", yytext)) {
        yylval->sort = TYPE_CONST;
    } if (SPD_STREQ("volatile", yytext)) {
        yylval->sort = TYPE_VOLATILE;
    }
#endif
    return (TOKEN(TYPE_QUALIFIER));
}
<INITIAL>typedef {
    storage = STORAGE_TYPEDEF;
    yylval = make_yylval();
    return (TOKEN(STORAGE_CLASS_SPECIFIER));
}
<INITIAL>auto|register|static|extern {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
#ifdef CHECK_ONLY_FIRST_CHAR
    switch (yytext[0]) {
    case 'a':  /* auto */
        yylval->sort = VAR_AUTO;
        storage = STORAGE_AUTO;
        break;
    case 'r':  /* register */
        yylval->sort = VAR_REGISTER;
        storage = STORAGE_REGISTER;
        break;
    case 's':  /* static */
        yylval->sort = VAR_STATIC;
        storage = STORAGE_STATIC;
        break;
    case 'e':  /* extern */
        yylval->sort = VAR_EXTERN;
        storage = STORAGE_EXTERN;
        break;
    }
#else
    if (SPD_STREQ("auto", yytext)) {
        yylval->sort = VAR_AUTO;
        storage = STORAGE_AUTO;
    } else if (SPD_STREQ("register", yytext)) {
        yylval->sort = VAR_REGISTER;
        storage = STORAGE_REGISTER;
    } else if (SPD_STREQ("static", yytext)) {
        yylval->sort = VAR_STATIC;
        storage = STORAGE_STATIC;
    } else if (SPD_STREQ("extern", yytext)) {
        yylval->sort = VAR_EXTERN;
        storage = STORAGE_EXTERN;
    }
#endif
    return (TOKEN(STORAGE_CLASS_SPECIFIER));
}
<INITIAL>void|char|int|float|double {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
#ifdef CHECK_ONLY_FIRST_CHAR
    switch (yytext[0]) {
        /* no need to check every characters. */
    case 'v': /* void */
        yylval->sort = TYPE_VOID;
        break;
    case 'i': /* int */
        yylval->sort = TYPE_INT;
        break;
    case 'c': /* char */
        yylval->sort = TYPE_CHAR;
        break;
    case 'f': /* float */
        yylval->sort = TYPE_FLOAT;
        break;
    case 'd': /* double */
        yylval->sort = TYPE_DOUBLE;
        break;
    }
#else
    if (SPD_STREQ("void", yytext)) {
        yylval->sort = TYPE_VOID;
    } else if (SPD_STREQ("int", yytext)) {
        yylval->sort = TYPE_INT;
    } else if (SPD_STREQ("char", yytext)) {
        yylval->sort = TYPE_CHAR;
    } else if (SPD_STREQ("float", yytext)) {
        yylval->sort = TYPE_FLOAT;
    } else if (SPD_STREQ("double", yytext)) {
        yylval->sort = TYPE_DOUBLE;
    }
#endif
    return (TYPE_TOKEN(TYPE_SPECIFIER));
}
<INITIAL>short|long {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
#ifdef CHECK_ONLY_FIRST_CHAR
    switch (yytext[0]) {
    case 's':  /* short */
        yylval->sort = TYPE_SHORT;
        break;
    case 'l':  /* long */
        yylval->sort = TYPE_LONG;
        break;
    }
#else
    if (SPD_STREQ("short", yytext)) {
        yylval->sort = TYPE_SHORT;
    } else if (SPD_STREQ("long", yytext)) {
        yylval->sort = TYPE_LONG;
    }
#endif
    return (TOKEN(TYPE_SPECIFIER_SIZE));
}
<INITIAL>signed|unsigned {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
#ifdef CHECK_ONLY_FIRST_CHAR
    switch (yytext[0]) {
    case 's':  /* signed */
        yylval->sort = TYPE_SIGNED;
        break;
    case 'u':  /* unsigned */
        yylval->sort = TYPE_UNSIGNED;
        break;
    }
#else
    if (SPD_STREQ("signed", yytext)) {
        yylval->sort = TYPE_SIGNED;
    } else if (SPD_STREQ("unsigned", yytext)) {
        yylval->sort = TYPE_UNSIGNED;
    }
#endif
    return (TOKEN(TYPE_SPECIFIER_SIGN));
}

<INITIAL>\'(.|{ESCAPED_CHAR})\' {
    /* This rule should precede the rule of multibyte character. */
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
    return (TOKEN(CHARACTER_CONSTANT));
}
<INITIAL>L?\'(([^\\]|{ESCAPED_CHAR})(.|{ESCAPED_CHAR}))\' {
    if (yytext[0] != 'L') {
        spdWarning("No prefix 'L' for multibyte character %s at l.%d.\n",
                   yytext, yylineno);
    }
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
    return (TOKEN(W_CHARACTER_CONSTANT));
}

<INITIAL>\" {
    yylval = make_yylval();
    yylval->length = 0;
    append_yytext();
    BEGIN(STR);
    count_char();
}
<STR>\" {
    append_yytext();
    yylval->name = MAKE_WORD_NAME(str_buf);
    BEGIN(INITIAL);
    return (TOKEN(STRING_CONSTANT));
}
<STR>\\\n|\n {
    append_yytext();
    count_char();
    INC_YYLINENO();
}
<STR>{SJIS1}{SJIS2} {
    if (!spdInputEncodeIs(SPD_EC_SJIS)) {
        REJECT;
    } else {
        append_yytext();
        count_char();
    }
}
<STR>\\.|. {
    append_yytext();
    count_char();
}

<INITIAL>([0-9]+|("0x"|"0X")[0-9a-fA-F]+)[ulUL]* {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
    return (TOKEN(INTEGER_CONSTANT));
}
<INITIAL>([0-9]+"."|[0-9]+"."?([eE][-+]?[0-9]+)|[0-9]*"."[0-9]+([eE][-+]?[0-9]+)?)[flFL]* {
    yylval = make_yylval();
    yylval->name = MAKE_WORD_NAME(yytext);
    return (TOKEN(FLOATING_CONSTANT));
}

<INITIAL>[a-zA-Z_][0-9a-zA-Z_]* {
    /* Modified by A.Yoshida 1999/06/03. */
    /* Removed ENUMERATION_CONST: Variables, enumerators and typedef-ed type
       share a same name space. Original code had separated name spaces for 
       each sort of objects.  This separation ocurred the wrong analysis
       that local variable name in current scope is a enumerator declared
       out of the scope. The new code returns IDENTIFIER even if the token
       is a enumerator. Looking up name spaces is to be done in "yacc.y".
    */
    /* Important note: To decide the token is a label name or not, knowledge 
       of context is required, i.e. "the token is followed by a colon or not."
       Without this knowledge, it is impossible to decide the token is
       IDENTIFIER or TYPEDEF_NAME. So, "yacc.y" should be treat TYPEDEF_NAME
       as a label name in the context of label statement. This problem had
       been fixed already.
    */
    /* Modified by A.Yoshida 1999/06/10. */
    /* All tokens except tokens same with typedef-ed names or builtin 
       functions' names are treated as IDENTIFIER. In the contexts of
       tag, enumerator and label name, TYPEDEF_NAME may be retured.
       Because tag, enumerator and label name have unique name spaces, 
       TYPEDEF_NAME is illegal in syntax. But, by allowing TYPEDEF_NAME
       in these context and treating it same as IDENTIFER in yacc.y, no
       syntax error nor ambiguity in parsing exists.
     */

#ifdef DEBUG_950320
    if (yydebug) {
        spdMessge("S flag %d\n", specifier_flag);
    }
#endif

    if (is_skipped_token(yytext)) {
        spdMessage("yylex(): \"%s\": Token skipped at l.%d.\n",
                   yytext, yylineno);
        count_char();
        /* Nothing returned. */
    } else if (is_skipped_token_with_arg(yytext)) {
        stwa_level = 0;
        strcpy(stwa_buf, yytext);
        stwa_buf_index = yyleng;
        /* Skip a expression which enclused with '(' and ')'. */
        BEGIN(STWA);
        count_char();
    } else {
        char    type;

        yylval = make_yylval();
        yylval->name = MAKE_WORD_NAME(yytext);
        if (is_builtin_function(yytext)) {
            return (TOKEN(BUILTIN_TYPE_PARAMETER));
        } else if (is_builtin_constant(yytext, &type)) {
            spdWarning("\"%s\": Builtin constant used at l.%d.\n",
                       yytext, yylineno);
            switch (type) {
              case 's':
                return TOKEN(STRING_CONSTANT);
                break;                      /* NOTREACHED */
              case 'i':
                return TOKEN(INTEGER_CONSTANT);
                break;                      /* NOTREACHED */
              default:
                spdWarning("Unknown builtin constant type: %c.\n", type);
                yyerror("Unknown builtin constant type.\n");
            }
        } else if (!specifier_flag && 
                   ((yylval->ident = find_type2(yylval->name, ID_TYPENAME))
                    != NULL)) {
            /* The purpose of specifier_flag is to distinguish
               member names in struct/union declarations and variable names 
               in declarators from typedef-ed names. Even if a member
               (or variable) name is same with typedef-ed names declared 
               before, the member name should be treat as identifier. */
            return (TYPE_TOKEN(TYPEDEF_NAME));
        } else if (!specifier_flag && is_builtin_type(yylval->name)) {
            return (TYPE_TOKEN(BUILTIN_TYPE_NAME));
        } else if (is_D_model && !specifier_flag) {
            TSP     tmp_type;
            
            if (yydebug) {
                spdMessage("\"%s\": Guessed typedef name.\n", yylval->name);
            }
            if ((tmp_type = find_guessed_typedef_type(yylval->name)) != NULL) {
                yylval->ident = tmp_type->ident;
            } else {
                yylval->ident = insert_guessed_typedefname(yylval->name);
            }
            return (TYPE_TOKEN(GUESSED_TYPEDEF_NAME));
        } else {
            return (TOKEN(IDENTIFIER));
        }
    }
}

<INITIAL>"->"   {yylval = make_yylval(); return (TOKEN(ARROW));}
<INITIAL>"."    {yylval = make_yylval(); return (TOKEN(DOT));}
<INITIAL>"&"    {yylval = make_yylval(); return (TOKEN(AND));}
<INITIAL>"&&"   {yylval = make_yylval(); return (TOKEN(AND_AND));}
<INITIAL>"&="   {yylval = make_yylval(); return (TOKEN(AND_EQUAL));}
<INITIAL>"..."  {yylval = make_yylval(); return (TOKEN(DOTS));}
<INITIAL>"="    {yylval = make_yylval(); return (TOKEN(EQUAL));}
<INITIAL>"=="   {yylval = make_yylval(); return (TOKEN(EQUAL_EQUAL));}
<INITIAL>"!"    {yylval = make_yylval(); return (TOKEN(EXCLAMATION));}
<INITIAL>"!="   {yylval = make_yylval(); return (TOKEN(EXCLAMATION_EQUAL));}
<INITIAL>"^"    {yylval = make_yylval(); return (TOKEN(HAT));}
<INITIAL>"^="   {yylval = make_yylval(); return (TOKEN(HAT_EQUAL));}
<INITIAL>"<"    {yylval = make_yylval(); return (TOKEN(LESS));}
<INITIAL>"<="   {yylval = make_yylval(); return (TOKEN(LESS_EQUAL));}
<INITIAL>"<<"   {yylval = make_yylval(); return (TOKEN(LESS_LESS));}
<INITIAL>"<<="  {yylval = make_yylval(); return (TOKEN(LESS_LESS_EQUAL));}
<INITIAL>"-"    {yylval = make_yylval(); return (TOKEN(MINUS));}
<INITIAL>"-="   {yylval = make_yylval(); return (TOKEN(MINUS_EQUAL));}
<INITIAL>"--"   {yylval = make_yylval(); return (TOKEN(MINUS_MINUS));}
<INITIAL>">"    {yylval = make_yylval(); return (TOKEN(MORE));}
<INITIAL>">="   {yylval = make_yylval(); return (TOKEN(MORE_EQUAL));}
<INITIAL>">>"   {yylval = make_yylval(); return (TOKEN(MORE_MORE));}
<INITIAL>">>="  {yylval = make_yylval(); return (TOKEN(MORE_MORE_EQUAL));}
<INITIAL>"|"    {yylval = make_yylval(); return (TOKEN(OR));}
<INITIAL>"|="   {yylval = make_yylval(); return (TOKEN(OR_EQUAL));}
<INITIAL>"||"   {yylval = make_yylval(); return (TOKEN(OR_OR));}
<INITIAL>"%"    {yylval = make_yylval(); return (TOKEN(PERCENT));}
<INITIAL>"%="   {yylval = make_yylval(); return (TOKEN(PERCENT_EQUAL));}
<INITIAL>"+"    {yylval = make_yylval(); return (TOKEN(PLUS));}
<INITIAL>"+="   {yylval = make_yylval(); return (TOKEN(PLUS_EQUAL));}
<INITIAL>"++"   {yylval = make_yylval(); return (TOKEN(PLUS_PLUS));}
<INITIAL>"/"    {yylval = make_yylval(); return (TOKEN(SLASH));}
<INITIAL>"/="   {yylval = make_yylval(); return (TOKEN(SLASH_EQUAL));}
<INITIAL>"*"    {yylval = make_yylval(); return (TOKEN(STAR));}
<INITIAL>"*="   {yylval = make_yylval(); return (TOKEN(STAR_EQUAL));}
<INITIAL>"~"    {yylval = make_yylval(); return (TOKEN(TILDA));}
<INITIAL>","    {yylval = make_yylval(); return (TOKEN(COMMA));}

<INITIAL>":"    {yylval = make_yylval(); return (TOKEN(yytext[0]));}
<INITIAL>";"    {yylval = make_yylval(); return (TOKEN(yytext[0]));}
<INITIAL>"?"    {yylval = make_yylval(); return (TOKEN(yytext[0]));}
<INITIAL>"["    {yylval = make_yylval(); return (TOKEN(yytext[0]));}
<INITIAL>"]"    {yylval = make_yylval(); return (TOKEN(yytext[0]));}

<INITIAL>"(" {
    paren_level++;
    yylval = make_yylval();
    return (TOKEN(yytext[0]));
}
<INITIAL>")" {
    paren_level--;
    yylval = make_yylval();
    return (TOKEN(yytext[0]));
}

<INITIAL>"{" {
    if (paren_level == 0) {
        yylval = make_yylval();
        specifier_flag = 0; 
        return (TOKEN(yytext[0]));
    } else {
        sbie_level = 1;
        strcpy(sbie_buf, yytext);
        sbie_buf_index = yyleng;
        /* Skip a block which enclused with '{' and '}'. */
        BEGIN(SBIE);
        c_count++;
    }
}
<INITIAL>"}" {
    yylval = make_yylval();
    return (TOKEN(yytext[0]));
}

<SBIE>"{" {
    sbie_level++;
    sbie_buf[sbie_buf_index++] = yytext[0];
    c_count++;
}
<SBIE>"}" {
    sbie_level--;
    sbie_buf[sbie_buf_index++] = yytext[0];
    if (sbie_level == 0) {
        sbie_buf[sbie_buf_index] = '\0';
        /* spdMessage("yylex(): l.%d, c.%d: Skiped: \"%s\"\n", */
        /*            yylineno, c_count, sbie_buf);            */
        yylval = make_yylval();
        yylval->name = MAKE_WORD_NAME(SPD_SKIPPED_BLOCK_IN_EXPRESSION);
        BEGIN(INITIAL);
        /* Return duumy token. */ 
        return TOKEN(INTEGER_CONSTANT);
    } else {
        c_count++;
    }
}
<SBIE>"\n" {
    sbie_buf[sbie_buf_index++] = yytext[0];
    c_count++;
    INC_YYLINENO();
}
<SBIE>. {
    sbie_buf[sbie_buf_index++] = yytext[0];
    c_count++;
}

. {
    spdWarning("yylex(): l.%d, c.%d: Invalid character '%c' (0x%x) skipped.\n",
               yylineno, c_count, yytext[0], yytext[0]);
    count_char();
}

%%

static WP make_yylval(void)
{
    WP      wp;
  
    wp = spdMalloc(sizeof(struct Word_data));
    wp->sort = TYPE_UNKNOWN;
    wp->point = c_count;
    wp->head_point = c_count;
    wp->length = yyleng;
    wp->lr_value = EXPR_LR_UNKNOWN;
    wp->name = NULL;
    wp->type = NULL;
    wp->expr = NULL;
    wp->dataptr = NULL;
    wp->type_spec2 = NULL;
    wp->type_spec2_last = NULL;
    wp->type_list2 = NULL;
    wp->ident = NULL;
    wp->ident_list = NULL;
    return (wp);
}

/******************************************************************************/

static IDOP insert_guessed_typedefname(char *name)
{
    TSP     ptr_ts;
    IDOP    tmp_ident;

    ptr_ts = insert_type(name, NULL);
    ptr_ts->sort = TYPE_TYPE_DEF;
    ptr_ts->pointer = NULL;                 /* ptr_type */
    guessed_type = insert_type_list(guessed_type, ptr_ts, NULL);
    /* 
     * tmp_fo->scope->type_t = insert_type_list(tmp_fo->scope->type_t,
     *                                          ptr_ts, NULL);
     */
    tmp_ident = insert_identifier(name, ID_TYPENAME);
    tmp_ident->data.type = 0;               /* Type isn't specified. */
    tmp_ident->length = strlen(name);       /* !! ǤΤ */
    tmp_ident->type = NULL;                 /* ptr_type; */

    ptr_ts->ident = tmp_ident;           
    /* ident_decl = insert_identifier_list(NULL, tmp_ident, decr); */

    /* typedef줿̻Ҥ򸽺ߤΥפηơ֥Ͽ롥 */
    /* add_typedef_table(tmp_ident); */

    return (tmp_ident);
}

static void append_yytext(void)
{
    if (SPD_STRING_LEN < (yylval->length + yyleng)) {
        spdErrorExit("append_yytext(): \"%s\": Too long constant string.\n",
                     yytext);
    }
    strcpy(str_buf + yylval->length, yytext);
    yylval->length += yyleng;
}

static void count_char(void)
{
    c_count += yyleng;
}
