/* 
 * Program:             $RCSfile: main.h,v $  $Revision: 30.18 $
 * 
 * Purpose:             SDB analyzer.
 * 
 * Author:              K.Mabuchi  1991/09/30
 *                      S.Yamamoto  1992/04/01
 * Modified:            H.Ariga  1993/07/01
 * 
 * (C) Copyright:       K.Mabuchi, H.Ariga and S.Yamamoto  1991 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef _SDB_MAIN_H
#define _SDB_MAIN_H

#include "ISDB.h"

/*
 * Debug configuration.
 * YYDEBUG: For yacc debug.
 */

/* 
 * #define MEMBER_DEBUG 1
 * #define TYPE_DEBUG 1
 * #define POINTER_DEBUG 1
 * #define BLOCK_DEBUG 1
 * #define TYPE_REF_DEBUG 1
 * #define DECL_DEBUG 1
 * #define PARM_DEBUG 1
 * #define FILE_DEBUG 1
 * #define EXPR_DEBUG 1
 * 
 * #define VAR_DEBUG 1
 * #define CONST_DEBUG 1
 * #define FUNCTION_DEBUG 1
 * #define ENUM_DEBUG 1
 * #define DEBUG_OP_TYPE 1
 * #define DEBUG_STRUCT_UNION 1
 */

#define YYDEBUG         1

/*
 * Basic values.
 */

#define YYSTYPE         WP
#define NON             0
#define NO_NAME         NULL

/*
 * Configurations.
 */

#define SDBIN_DEFAULT_POOL_SIZE     4096
#define SPEC_NAME_COPY(s, n)        ((s) = spdStrPoolCopy(str_pool, (n)))

extern SpdStringPool *str_pool;

#ifdef USE_FLEX   
void    yyrestart(FILE *input_file);
#endif

/*
 * Miscellaneous macro.
 */

#define SET_NEW_SOURCE(s, w_ptr, f) \
    (((w_ptr) != NULL) ? app_source_LC((s), (w_ptr)->head_point, (f)) : NULL)

/*
 * Structure pointer definitions.
 */

typedef struct PO           *POP;
typedef struct FO           *FOP;
typedef struct DNO          *DNOP;
typedef struct DRO          *DROP;
typedef struct OD           *ODP;
typedef struct BlockObject  *BOP;
typedef struct ExprObject   *EOP;
typedef struct LabelObject  *LOP;
typedef struct OperatorObject   *OOP;

typedef struct TS2          *TS2P;
typedef struct TS           *TSP;

typedef struct TS2_List     *TS2_ListP;

typedef struct FO_List      *FO_ListP;
typedef struct DO_List      *DO_ListP;  /* Type name "DNO_ListP" is better. */

typedef struct DS_List      *DS_ListP;
typedef struct DS_List      *DataListP;

typedef struct MEMS_List    *MEMS_ListP;

typedef struct SrcLC        *SrcLCP;

/*
 * Structure definitions.
 */

typedef struct Word_data {
    int     sort;
    int     point;
    int     length;
    int     head_point;
    int     lr_value;
    TSP     type;

    TS2P      type_spec2;               /* word ɽ type κǽ */
    TS2P      type_spec2_last;          /* word ɽ type κǸ */
    TS2_ListP type_list2;               /* type Υꥹ */
    struct _ident_object *ident;        /* type_is, ident_type  */
    struct _ident_list *ident_list;     /* ident_type  */

    char    *name;
    EOP     expr;
    void    *dataptr;
} WD, *WP;

/* Line and column. */

typedef struct SrcLC {
    int         point;
    FOP         file;
    SrcLCP      next;
} SrcLC;

typedef SrcLC               Src_LC;         /* Don't use any more. */
typedef SrcLC               *Src_LC_P;      /* Don't use any more. */

typedef struct _Point {
    int     head_point;
    int     tail_point;
    int     flag;
    void    *ptr;
} Point;

/* Scope. */

struct Var_Table {
    struct DS *var_s; 
    struct Var_Table *next;
};

struct Enum_Table {
    struct CS  *const_s;
    struct Enum_Table *next;
};

typedef struct _Label_Table {
    struct LabelObject *label_s;
    struct _Label_Table *next;
} LabelTable,*LabelTablePtr;

typedef struct _Tag_Table {
    struct _ident_object *ident;        /* tag ɽ̻ */
    SpdBoolean member;                  /* member Ѥߤɤ */
    struct _Tag_Table *next;
} Tag_Table, *Tag_TableP;

/* Scope type for struct level member type. */

#define SCOPE_GLOBAL    1
#define SCOPE_FILE      2
#define SCOPE_FUNCTION  4
#define SCOPE_BLOCK     8
#define SCOPE_LOCAL     (SCOPE_FUNCTION | SCOPE_BLOCK)

typedef struct _Level {
    struct Var_Table *var_t;
    struct Enum_Table *const_t;
    struct TS_List *type_t;
    struct _Label_Table *label_t;

    struct _ident_list *typedef_t; /* typedef 줿μ̻ҤΥơ֥ */
    Tag_TableP tag_t;              /* tag μ̻ҤΥơ֥ */

    int    type;
    struct _Level *next;
} Level, *LevelPtr;

/* 
 * Type spec 2: ֥ȤΥǡ¤
 */

typedef struct TS2 {
    int         id;

    /* attributes */
    char        *name;
    int         sort;
    int         qualifier;      /* const, volatile */
    int         type;           /* void, char, int, float, double, ... */
    int         length;         /* short, long */
    int         sign;           /* signed, unsigned */

    /* relations */
    TS2P        type_st_prev;           /* type_st (int <-- *) */
    TS2_ListP   type_st_next;           /* type_st (int --> *) */
    struct _ident_object *ident;        /* type_is */
    struct ExprObjectList *array;       /* type_cexpr */

    SrcLCP      source;

    /*
     * TSP λͤǤϡ"const unsined long" ϰĤΥ֥Ȥˤʤ褦
     * Ƥ롥
     * TS2P Ǥϡ"const" "unsined" "long" ̡Υ֥ȤȤ
     * Τ TSP ˹碌뤿ˡ夫򤷤Ƥ롥
     * פȤʤä֥Ȥ򼨤 deleted flag Ѱդ
     * type_spec2.c ꤵ write_sdb.c ѤƤ롥
     */
    SpdBoolean  deleted;        /* ޡ */
    TS2P        u_type;         /* ̾󤻤줿 */

    TS2P        next;
    TS2P        prev;   /* Is this necessary? */
} TypeSpec2;

typedef struct TS2_List {
    TS2P type_spec;
    SrcLCP source;
    TS2_ListP next;
    TS2_ListP prev;
} TypeList2;

/* 
 * Type spec.
 * ֥ȤΥǡ¤
 */

typedef struct TS {
    int         id;                     /* ֹ桥 */
    char        *name;                  /* ̻̾ */
    int         sort;
        /* μࡥ
           TYPE_{DECLARATION,POINTER,ARRAY,FUNC,...}. */
    int         class;                          
        /* Storage class (auto, register, static or extern).
           Use macros STORAGE_XXX defined in "declaration.h". */
    int         qualifier;              /* (const or volatile). */
    int         type;                   /* (void, char, int, float or
                                           double). */
    int         length;                 /* Ĺ(short or long). */
    int         sign;                   /* ̵ͭ(signed or unsigned). */
    TSP         pointer;
    struct ExprObjectList *array;
    struct MEMS_List *struc;
    struct CS_List *enum_const;
    DNOP        decl;
    struct _ident_object *ident;
    SrcLCP      source;
    TSP         next;
    TSP         prev;

    /*
     * make_type_spec() ǳݤ줿ΰ free 뤿 ts_list2 ǻѡ
     * ts_list Ͽ줿 ts_next = ts_prev = NULL;
     */
    TSP         ts_next;
    TSP         ts_prev;
} TypeSpec;

/* ֥ȤΥꥹ */

typedef struct TS_List {
    int id;
    TSP type_s;
    SrcLCP source;
    struct TS_List *next;
    struct TS_List *prev;
} *TypeListP;

/* 
 * Constant spec.
 */

typedef struct CS {
    int id;
    char *name;
    int type;                   /* Υ */
    char *value;
    int length;                 /* Ĺ */
    struct CS_List *const_s;    /* δϢ */
    DNOP decl;
    SrcLCP source;
    struct CS *next;
    struct CS *prev;
} ConstSpec, *CSP;

typedef struct CS_List {
    struct CS *const_s;
    SrcLCP source;
    struct CS_List *next;
    struct CS_List *prev;
} *ConstListP;

/* 
 * Data (variable) spec.
 * ѿ֥ȤΥǡ¤
 */

typedef struct DS {
    int id;
    char *name;
    int sort;
    int class;
    TSP type;
    DNOP decl;
    SrcLCP source;
    struct DS *next;
    struct DS *prev;
} DataSpec, *DSP;

typedef struct DS_List {
    int         id;
    struct DS *data_s;
    SrcLCP      source;
    DS_ListP    next;
    DS_ListP    prev;
} DS_List;

/* 
 * ؿ֥ȤΥǡ¤
 */

typedef struct FS {
    int id;
    char *name;
    int length;
    TSP type;
    struct FS *next;
    struct FS *prev;
} FunctionSpec, *FSP;

/******************************************************************************/

typedef struct FS_List {
    struct FS *function_s;
    SrcLCP source;
    struct FS_List *next;
    struct FS_List *prev;
} *FunctionListP;

/*
 * Identifier object (IDO).
 */

typedef struct _ident_object {
    int id;
    int length;
    int sort;
    char *name;
    int ISDB_storage;

    union {
        struct DS *var;
        struct FS *func;
        struct MEMS *mem;
        struct CS *cons;
        TSP type;
    } data;
    TSP type;

    TS2P type2;

    struct _ident_object *next;
    struct _ident_object *prev;
} IDO, *IDOP;

typedef struct _ident_list {
    struct _ident_object *ident;
    SrcLCP source;
    struct _ident_list *next;
    struct _ident_list *prev;
} IDO_List, *IDO_ListP;

/* 
 * Scope.
 */

typedef struct SCOPE {
    struct Var_Table *var_t;
    struct Enum_Table *const_t;
    struct TS_List *type_t;
} Scope, *SCP;

typedef struct SCOPE_List {
    SCP         scope;
    struct SCOPE_List *next;
} *ScopeListP;

/******************************************************************************/

/* ֥ȺΤΥݥ󥿤
   ֥ȤΥꥹȤؤΥݥ󥿤Υå */

struct Stack_Queue {
    TSP in;
    TSP out;
    struct Stack_Queue *next;
    struct Stack_Queue *prev;
};

/* ֥ȺΤΥå */

struct Stack_Type {     
    TSP sort;
    TSP list;
    int point;
    struct Stack_Queue *queue;
    struct Stack_Type *next;
    struct Stack_Type *prev;
};

/******************************************************************************/

/*
 * Program object (PO).
 *
 * Note: Used only in D-model.
 */

typedef struct PO {
    int         id;
    char        *name;
    FO_ListP    file;
    POP         next;
} PO;

/*
 * File object (FO).
 */

typedef struct FO {
    int         id;
    char        *name;
    FOP         next;
    DO_ListP    decl;
    FO_ListP    fo_list;
    SCP         scope;
} FO;

typedef struct FO_List {
    FOP         file_obj;
    FO_ListP    next;
} FO_List;

/*
 * Declaration object (DNO).
 */

typedef struct DNO {
    int         id;
    int         length;
    int         storage;
    int         sort;
    DS_ListP    var;
    struct TS_List *type;
    struct TS_List *type_ref;
    struct CS_List *const_l;
    struct FS_List *func_l;
    struct MEMS_List *mem_l;
    struct BlockObjectList *block_l;
    struct _declarator_list *dt_l;

    TS2_ListP   type2;  /* decl_type */

    DNOP        next;
    DNOP        prev;
} DNO;

typedef struct DO_List {
    DNOP        decl;
    SrcLCP      source;
    DO_ListP    next;
    DO_ListP    prev;
} DO_List;

/* func_parmΤΥå */

struct Stack_Parm {
    DO_ListP    decl_l;
    struct Stack_Parm *next;
};

/* direct_declΤΥå */

struct Stack_Func {
    int         direct_decl;
    struct Stack_Func *next;
};

/* storageΤΥå */

struct Stack_Storage {
    int         storage;
    struct Stack_Storage *next;
};

/*
 * Declarator object (DRO).
 */

typedef struct DRO {
    int         id;
    int         length;
    int         sort;
    TSP         type;                       /* decr_type (1997/05/31). */

    TS2P        type2;  /* decr_type */

    struct _ident_list *id_l;               /* decl_ident. */
    struct _optional_decl_list *opt_l;      /* decl_opt. */
    DROP        next;
    DROP        prev;
} DRO;

typedef struct _declarator_list {
    DROP        dt;
    SrcLCP      source;
    struct _declarator_list *next;
    struct _declarator_list *prev;
} DT_List, *DT_ListP;

/* decl_declΤΥå */

struct Stack_Decl {
    DT_ListP decl_decl;
    struct Stack_Decl *next;
};

/*
 * OptionalDecl object (OD).
 */

typedef struct OD {
    int         id;
    int         length;
    int         sort;
    DO_ListP    decl_l;
    struct BlockObjectList *block_l;
    struct ExprObjectList  *initexpr_l;
    struct TS_List *prototype;
    ODP         next;
    ODP         prev;
} OD;

typedef struct _optional_decl_list {
    ODP         opt;
    SrcLCP      source;
    struct _optional_decl_list *next;
    struct _optional_decl_list *prev;
} OD_List, *OD_ListP;

/*
 * Block object (BO).
 */

typedef struct BlockObject {
    int         id;
    int         sort;
    int         length;
    DO_ListP    decl_list;
    struct ExprObjectList *expr_list;
    struct BlockObjectList *blk_list;
    struct LabelObjectList *label_list;
    struct BlockObject *next;
    struct BlockObject *prev;
} BO;

typedef struct BlockObjectList {
    struct BlockObject *block_s;
    SrcLCP      source;
    struct BlockObjectList *next;
    struct BlockObjectList *prev;
} *BO_ListP;

/*
 * Expression object (EO).
 */

typedef struct ExprObject {
    int         id;
    int         length;
    int         sort;
    int         lrtype;
    IDO_ListP   id_l;
    ConstListP  csl;
    MEMS_ListP  memsl;
    DataListP   dsl;
    FunctionListP fsl;
    TypeListP   type;
    struct LabelObjectList *label_list;
    struct ExprObjectList *expr_list;
    struct OperatorObjectList *op_list;
    EOP         next;
    EOP         prev;
} EO;

typedef struct ExprObjectList {
    EOP         expr;
    SrcLCP      source;
    struct ExprObjectList *next;
    struct ExprObjectList *prev;
} *EO_ListP;

/*
 * Label object (LO).
 */

typedef struct LabelObject {
    int         id;
    char        *name;
    int         sort;
    int         length;
    struct ExprObjectList *expr_l;
    struct LabelObject *next;
    struct LabelObject *prev;
} LO;

typedef struct LabelObjectList {
    struct LabelObject *label_s;
    SrcLCP      source;
    struct LabelObjectList *next;
    struct LabelObjectList *prev;
} *LO_ListP;

/*
 * Operator object (OO).
 */

typedef struct OperatorObject {
    int         id;
    int         length;
    int         sort;
    TSP         type;

    TS2P        type2;

    struct OperatorObject *next;
    struct OperatorObject *prev;
} OO;

typedef struct OperatorObjectList {
    struct OperatorObject *operator_s;
    SrcLCP      source;
    struct OperatorObjectList *next;
    struct OperatorObjectList *prev;
} *OO_ListP;

/* ¤ΥС */

typedef struct MEMS {
    int         id;
    char        *name;
    TSP         type;
    SrcLCP      source;
    struct MEMS *next;
    struct MEMS *prev;
} MemberSpec, *MEMSP;

typedef struct MEMS_List {
    struct MEMS *member_s;
    SrcLCP      source;
    struct MEMS_List *next;
    struct MEMS_List *prev;
} MEMS_List;

typedef struct UNDEF_List {
    char        *name;
    SrcLCP      source;
    struct UNDEF_List *next;
    struct UNDEF_List *prev;
} UL,*ULP;

/******************************************************************************/

/* 
 * Extern declaration.
 */

/* ƥ֥ȤΥꥹȤƬΥݥ󥿡 */

extern TSP      ts_list;
extern TS2P     ts2_list;
extern CSP      cs_list;
extern DSP      ds_list;
extern FOP      fo_list;

extern POP      po_list;        /* Note: Used only in D-model. */
extern FSP      fs_list;
extern DNOP     do_list;        /* Variable name "dno_list" is better. */
extern DROP     dt_list;        /* Variable name "dro_list" is better. */
extern ODP      od_list;
extern BOP      bo_list;
extern EOP      eo_list;
extern LOP      lo_list;
extern OOP      oo_list;

extern MEMSP    mems_list;
extern IDOP     id_list;

extern ScopeListP   scope_list;
extern ULP          undef_list;

/*
 * make_type_spec() ǳݤ줿ΰ free 뤿Υꥹȡ
 * ts_list Ͽ줿 ts_next = ts_prev = NULL;
 */

extern TSP ts_list2;

/******************************************************************************/

/* Added by S.Yamamoto 1998/05/07. */

#define INC_ID(x)       ((x)++)

/* Added by A.Yoshida 1999/08/02. */
/* 
 * This flag is for revising type objects of struct/union who do not have 
 * pointers to members though the members are defined.  This revising
 * happens when both const struct and non-const struct are used.  In this
 * case, the identifier whose type is the struct may link with an incorrect 
 * type object. This problem will be solved by implementing "type spec2"
 * completely.
 */

#ifndef REVISE_STRUCT_WITHOUT_MEMBER
#define REVISE_STRUCT_WITHOUT_MEMBER
#endif

#endif /* _SDB_MAIN_H */
