/* 
 * Program:             $RCSfile: scope.c,v $  $Revision: 30.13 $
 * 
 * Purpose:
 * 
 * Author:              K.Mabuchi  1991/09/30
 *                      S.Yamamoto  1992/04/01
 * Modified:            H.Ariga  1995/10/05
 *                      A.Yoshida  1999/06/03
 * 
 * (C) Copyright:       K.Mabuchi, H.Ariga and S.Yamamoto  1991 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: scope.c,v 30.13 2003/06/28 15:32:15 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "main.h"
#include "prototype.h"

static LevelPtr     level = NULL;

static SpdBoolean lookup_pe_ident_in_level(LevelPtr lev, char *name,
                                           struct Var_Table **v_tab,
                                           struct Enum_Table **e_tab);

void insert_level(int type)
{
    LevelPtr new;
#ifdef LEVEL_DEBUG
    void disp_lev();
    extern int yylineno;
#endif
    
    new = (LevelPtr)spdCalloc(sizeof(Level),1);
    new->var_t = NULL;
    new->const_t = NULL;
    new->type_t = NULL;
    new->label_t = NULL;
    new->typedef_t = NULL;
    if ( type == SCOPE_LOCAL ){
        if ( level->type == SCOPE_FILE )
            type = SCOPE_FUNCTION;
        else
            type = SCOPE_BLOCK;
    }
    new->type = type;
    new->next = level;
    level = new;
#ifdef LEVEL_DEBUG
    printf("after insert_level %d",yylineno);
    disp_lev();
#endif
}

void escape_level(void)
{
    LevelPtr    tmp;

#ifdef LEVEL_DEBUG
    void        disp_lev();
    extern int  yylineno;

    printf("before escape_level %d",yylineno);
    disp_lev();
#endif

    if (level) {
        free_Var_Table(level->var_t);
        free_Enum_Table(level->const_t);
        free_TS_ListP(level->type_t);
        free_Label_Table(level->label_t);
        free_IDO_ListP(level->typedef_t);
        free_Tag_TableP(level->tag_t);

        tmp = level;
        level = level->next;
        spdFree(tmp);
    }
}

LevelPtr get_current_level(void)
{
    return level;
}

/* ѿꥹȤϿ */

struct Var_Table *insert_var_table(struct Var_Table *ptr, DSP ptr_vs)
{
    struct Var_Table *new;
    
#ifdef DEBUG_REDEF
    printf("variable %s added to a table\n",ptr_vs->name);
#endif  
    new = (struct Var_Table *)spdCalloc(sizeof(struct Var_Table),1);
    new->var_s = ptr_vs;
    
    new->next = ptr;
    return new;
}

/* enum ꥹȤϿ */

struct Enum_Table *insert_enum_table(struct Enum_Table *ptr, CSP ptr_cs)
{
    struct Enum_Table *new;
    
    new = (struct Enum_Table *)spdCalloc(sizeof(struct Enum_Table),1);
    new->const_s = ptr_cs;
    
    new->next = ptr;
    return new;
}

/* label ơ֥ꥹȤϿ */

LabelTablePtr insert_label_table(LabelTablePtr ptr, LOP ptr_ls)
{
    LabelTablePtr new;
    
    new = (LabelTablePtr)spdCalloc(sizeof(LabelTable),1);
    new->label_s = ptr_ls;
    new->next = ptr;
    return new;
}

/*
 * tag ơ֥˼̻ҤϿ
 */
Tag_TableP insert_tag_table(Tag_TableP tag_t, IDOP ident, SpdBoolean member)
{
    Tag_TableP tmp_tag;

    tmp_tag = (Tag_TableP)spdCalloc(1, sizeof(Tag_Table));
    tmp_tag->ident = ident;
    tmp_tag->member = member;
    tmp_tag->next = tag_t;
    return tmp_tag;
}

void free_Tag_TableP(Tag_TableP tag_t)
{
    Tag_TableP tmp_tag;
    while (tag_t != NULL) {
        tmp_tag = tag_t;
        tag_t = tag_t->next;
        spdFree(tmp_tag);
    }
}

LevelPtr find_level_has_type(int type)
{
    /* פμब type ǤΤõ
     * ֤ פΥݥ
     *        դʤ NULL
     */   
              
    LevelPtr tmp;

    for (tmp = level ; tmp && !(tmp->type & type)  ; tmp = tmp-> next);
    return tmp;
}

int find_identifier_in_same_scope(char *name,int type)
{
    int r = 0;
    LevelPtr lev;
    struct Var_Table *var_t;
    struct Enum_Table *const_t;
    /* struct TS_List *type_t; */

    lev = find_level_has_type(type);
    if (lev != NULL && lookup_pe_ident_in_level(lev, name, &var_t, &const_t)) {
        r = 1;
    }
    return r; 
}

void add_var_table_list(struct DS_List *ds_l)
{

#ifdef DEBUG_REDEF
    void disp_lev();
#endif  

    for (/* Empty. */; ds_l != NULL; ds_l = ds_l->next){
        level->var_t = insert_var_table(level->var_t, ds_l->data_s);
    }

#ifdef DEBUG_REDEF
    printf("scope\n");
    disp_lev();
#endif  
}

void add_var_table(DSP ptr_data)
{
    level->var_t = insert_var_table(level->var_t, ptr_data);
}

void insert_var_pos(DSP ptr_vs, WP ptr)
{
    extern FOP current_file;
    ptr_vs->source = app_source_LC(ptr_vs->source,ptr->point,current_file);
}

void add_const_table(CSP ptr_data)
{
    level->const_t = insert_enum_table(level->const_t, ptr_data);
}

/* ؿפ˥٥ɲ */

void add_label_table(LOP ptr_ls)
{
    LevelPtr lev;
    if ( ptr_ls->sort == LABEL_NAMED ){
        lev = find_level_has_type(SCOPE_FUNCTION);
        if ( lev ){
            lev->label_t = insert_label_table(lev->label_t, ptr_ls);
#ifdef DEBUG_FIND_LABEL         
            printf("find_label: added %s\n",ptr_ls->name);
#endif                 
        }
        else {
            spdWarning("Cannot find function scope");
        }
    }
}

/*
 * ߤΥפ typedef ơ֥˷̻Ҥɲ
 */
void add_typedef_table(IDOP ident)
{
    level->typedef_t = insert_identifier_list(level->typedef_t, ident, NULL);
}

/*
 * ߤΥפ tag ơ֥˷̻Ҥɲ
 * member Ƥ뤫ɤ
 */
void add_tag_table(IDOP ident, SpdBoolean member)
{
    level->tag_t = insert_tag_table(level->tag_t, ident, member);
}

IDOP get_tag_ident(Tag_TableP tag_t)
{
    return (tag_t == NULL ? NULL : tag_t->ident);
}

void set_tag_table_member_def(Tag_TableP tag_t)
{
    if (tag_t == NULL) return;
    tag_t->member = TRUE;
}

SpdBoolean tag_has_member_def(Tag_TableP tag_t)
{
    return (tag_t == NULL ? FALSE : tag_t->member);
}

DSP find_var_pos(char *name)
{
    LevelPtr lev;
    struct Var_Table *tab;
    
    for (lev = level; lev != NULL ; lev = lev->next)
        for (tab = lev->var_t; tab != NULL; tab = tab->next)
            if (SPD_STREQ(name, tab->var_s->name)) return tab->var_s;
    return NULL;
}

CSP find_enum_pos(char *name)
{
    LevelPtr lev;
    struct Enum_Table *tab;
    
    if (name)
        for (lev = level; lev != NULL ; lev = lev->next)
            for (tab = lev->const_t; tab != NULL; tab = tab->next)
                if (SPD_STREQ(name, tab->const_s->name)) return tab->const_s;
    return NULL;
}

LOP find_label(char *name)
{
    LevelPtr lev;
    LabelTablePtr ltp;
    /* LOP lab; */
    lev = find_level_has_type(SCOPE_FUNCTION);
    if (lev) {
        for (ltp = lev->label_t; ltp; ltp = ltp->next) {
            if (ltp->label_s->name == NULL) {
                continue;
            }

#ifdef DEBUG_FIND_LABEL         
            printf("find_label: predef %s cmp with %s\n",
                   ltp->label_s->name,name);
#endif                 

            if (SPD_STREQ(ltp->label_s->name, name)) {
                break;
            }
        }
    }
    return ( ltp ? ltp->label_s : NULL );
}

/*
 * ߤΥפ name  sort ημ̻Ҥ򸫤Ĥ롥
 * ĤʤХפéäƱȤ򷫤֤
 */
IDOP find_type2(char *name, int sort)
{
    LevelPtr lev;
    IDOP ident;

    /* Added by A.Yoshida 1999/06/03. */
    /* For check existence of the same name identifier of another sort. */
    struct Var_Table *v_tab;
    struct Enum_Table *e_tab;

    /*
     * lex.l  typedef ɤĴ٤礫
     * yacc.y ǤǤϿƤ빽¤()ΥǤ롥
     * name  NULL ǤϤʤ
     */
    /* if (name == NULL) return NULL; */

    for (lev = level; lev != NULL; lev = lev->next) {
        ident = find_identifier_list_by_name(lev->typedef_t, name, sort);
        if (ident != NULL) {
            return ident;
        }
        /* Added by A.Yoshida 1999/06/03. */
        /* If the same name identifier is declared as variable or enumerator,
           the identifier of the name is not typedef-ed identifier. */
        if (lookup_pe_ident_in_level(lev, name, &v_tab, &e_tab)) {
            return NULL;
        }
    }
    return NULL;
}

Tag_TableP find_tag_in_scope(char *name, LevelPtr lev)
{
    Tag_TableP tmp_tag;
    IDOP ident;

    /* if (name == NULL) return NULL; */
    if (lev == NULL) return NULL;
    for (tmp_tag = lev->tag_t; tmp_tag != NULL; tmp_tag = tmp_tag->next) {
        ident = tmp_tag->ident;
        if (ident != NULL && ident->name != NULL
            && SPD_STREQ(name, ident->name)) return tmp_tag;
    }
    return NULL;
}

Tag_TableP find_tag_in_current(char *name)
{
    return find_tag_in_scope(name, level);
}

Tag_TableP find_tag_recursive(char *name)
{
    Tag_TableP  tmp_tag;
    LevelPtr    lev;

    for (lev = level; lev != NULL; lev = lev->next) {
        if ((tmp_tag = find_tag_in_scope(name, lev)) != NULL) return tmp_tag;
    }
    return NULL;
}

void disp_lev(void)
{
    LevelPtr            lev;
    struct Var_Table    *tab;
    IDO_ListP           type_table;
    
    fprintf(stderr, "- scope.c ----------------------------------------\n");
    for (lev = level; lev != NULL ; lev = lev->next) {
        fprintf(stderr, " {}level %d:",lev->type);
        for (tab = lev->var_t; tab != NULL; tab = tab->next)
            fprintf(stderr, " %s",tab->var_s->name);
        for (type_table = lev->typedef_t; type_table != NULL;
             type_table = type_table->next)
            fprintf(stderr, " %s(type)", type_table->ident->name);
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "--------------------------------------------------\n");
}

/*
 * lookup_pe_ident():
 * lookups the variable and enuemrator named 'name' in name table from
 * inner level to outer level.  'pe' means Primary Expression. Variables
 * and enumerators consist of primary expressions.
 *
 * Created by A.Yoshida 1999/06/03.
 */

SpdBoolean lookup_pe_ident(char *name, DSP *var_p, CSP *const_p)
{
    LevelPtr lev;
    struct Var_Table *v_tab;
    struct Enum_Table *e_tab;
    
    *var_p = NULL;
    *const_p = NULL;

    for (lev = level; lev != NULL ; lev = lev->next) { /* for each scope */
        if (lookup_pe_ident_in_level(lev, name, &v_tab, &e_tab)) {
            if (v_tab != NULL) {
                *var_p = v_tab->var_s;
            } else /* if (e_tab != NULL) */ {
                *const_p = e_tab->const_s;
            }
            return TRUE;
        }
    }
    return FALSE;
}

/*
 * lookup_ident_in_level(): lookups identifier in a level.
 * (auxiliary function)
 *
 * Created by A.Yoshida 1999/06/04.
 */
static SpdBoolean lookup_pe_ident_in_level(LevelPtr lev, char *name,
                                           struct Var_Table **v_tab,
                                           struct Enum_Table **e_tab)
{
    
    struct Var_Table *vtp;
    struct Enum_Table *etp;

    *v_tab = NULL;
    *e_tab = NULL;
    for (vtp = lev->var_t; vtp != NULL; vtp = vtp->next) {
        if (SPD_STREQ(name, vtp->var_s->name)) {
            *v_tab = vtp;
            return TRUE;
        }
    }
    for (etp = lev->const_t; etp != NULL; etp = etp->next) {
        if (SPD_STREQ(name, etp->const_s->name)) {
            *e_tab = etp;
            return TRUE;
        }
    }
    return FALSE;
}
