/* 
 * Program:             $RCSfile: link4b_read.c,v $  $Revision: 40.9 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu  2001/09/20
 *                      S.Yamamoto  2001/09/20
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2001 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: link4b_read.c,v 40.9 2003/07/20 11:46:16 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(SDBLINK4B)
#include "sdblink4b.h"
#elif defined(SPD_MERGE_SDB)
#include "sdblink4c.h"
#endif

/* #define DEBUG_SPDSET_SDBLINK4B */

extern SpdDBObj         *object_db;
extern SpdDBRel         *relation_db;
extern int              object_db_n;
extern int              relation_db_n;
extern SpdSet           class_object_set[];
extern RelationArray    relation_array[];

static int  *id_offset = NULL;
static int  id_offset_size = 0;

#if defined(DEBUG_SPDSET_SDBLINK4B)
static Contents traceSpdAddElem(SpdSet set, int key, Contents contents);
static Contents traceSpdGetAElem(SpdSet set, int *key);
#endif

/*
 * class_index() returns the unique number i (0 <= i < SDBLINK_CLASS_NUM)
 * for object class.
 */

int class_index(SpdClassId class_id)
{
    return (spdGetModelId(class_id) == SPD_MID_P_MODEL ?
            class_id - SPD_RID_NUM_I_MODEL : class_id);
}

/*
 * rel_index() returns the unique number i (0 <= i < SDBLINK_REL_NUM)
 * for relation class.
 */

int rel_index(SpdClassId class_id)
{
    return (spdGetModelId(class_id) == SPD_MID_P_MODEL ?
            class_id - SPD_CID_NUM_I_MODEL - SPD_CID_NUM_P_MODEL :
            class_id - SPD_CID_NUM_I_MODEL);
}

int offset_index(int file_no, SpdClassId class_id)
{
    return (file_no * SDBLINK_CLASS_NUM + class_index(class_id));
}

/**********************************************************************/

/*
 * id_offset[file_no][class_id]ϡΥեˤơ饹μֹ椬
 * class_idǤ륪֥Ȥɤ߹Ȥˡ֥Ȥμֹ
 * ɲä٤offsetɽ
 */

void init_id_offset(int file_num)
{
    id_offset_size = (file_num + 1) * SDBLINK_CLASS_NUM;
    id_offset = spdCalloc(id_offset_size, sizeof(int));
}

void free_id_offset(void)
{
    if (id_offset != NULL) {
        spdFree(id_offset);
        id_offset = NULL;
        id_offset_size = 0;
    }
}

int get_id_offset(int file_no, SpdClassId class_id)
{
    int i = offset_index(file_no, class_id);

    if (i < 0 || id_offset_size <= i) {
        spdErrorExit("get_id_offset(): invalid index %d.\n", i);
    }
    return id_offset[offset_index(file_no, class_id)];
}

void set_id_offset(int file_no, SpdClassId class_id, SpdObjId object_id)
{
    int i = offset_index(file_no, class_id);
    int new_id_offset = object_id - SPD_OBJ_ID_OFFSET(class_id) + 1;

    if (i < 0 || id_offset_size <= i) {
        spdErrorExit("set_id_offset(): invalid index %d.\n", i);
    }
    if (id_offset[i] < new_id_offset) {
        id_offset[i] = new_id_offset;
    }
}

/**********************************************************************/

/*
 * ֥ȤϿ롥
 * AR4Ʊͤobject_db[]Ͽclass_object_set[]ˤϿ롥
 */

void add_new_object(SpdClassId classId, SpdDBObj db_obj)
{
    MyObject  newObj;
  
    /* 
     * Do array boundary check. Commented for speed-up.
     * 
     * if (object_db_n  >= object_db_alloc_n) {
     *     spdErrorExit("Over the allocated memory of object_db.\n");
     * }
     */

    memcpy(&(object_db[object_db_n]), &db_obj, sizeof(SpdDBObj));

    newObj.db_key = object_db_n;
    newObj.p_link.elem_num = 0;
    newObj.c_link.elem_num = 0;
    spdAddElem(class_object_set[class_index(classId)],
               object_db[newObj.db_key].Obj_id, &newObj);

#if defined(DEBUG_SPDSET_SDBLINK4B)
    traceSpdAddElem(class_object_set[class_index(classId)],
                    object_db[newObj.db_key].Obj_id, &newObj);
#endif

    object_db_n++;
}

/*
 * 졼Ͽ롥
 * AR4Ʊͤrelation_db[]Ͽrelation_array[]ˤϿ롥
 */

void add_new_relation(SpdClassId classId, SpdDBRel db_rel)
{
    RelationArray   *array;

    /* 
     * Do array boundary check. Commented for speed-up.
     *     
     * if (relation_db_n  >= relation_db_alloc_n) {
     *     spdErrorExit("Over the allocated memory of relation_db.\n");
     * }
     */

    memcpy(&(relation_db[relation_db_n]), &db_rel, sizeof(SpdDBRel));

    array = &relation_array[rel_index(classId)];
    /* 
     * Do array boundary check. Commented for speed-up.
     *
     * if (array->elem_num >= obj_and_rel_num_table[classId]) {
     *     spdErrorExit("Over the allocated memory of relation_array.\n");
     * }
     */
    array->elem[array->elem_num].db_key = relation_db_n;
    array->elem[array->elem_num].flag = FALSE;
    array->elem_num++;

    relation_db_n++;
}

/******************************************************************************/

SpdId get_equal_db_obj(SpdDBObj *db_obj, SpdClassId class_id)
{
    int		i, dummy;
    MyObject    *obj;

    for (i = 0; i < class_object_set[class_index(class_id)]->elemNum; i++) {

#if defined(DEBUG_SPDSET_SDBLINK4B)
        dummy = -999;           /* Just for debugging. */
        traceSpdGetAElem(class_object_set[class_index(class_id)], &dummy);
#endif

        obj = spdGetAElem(class_object_set[class_index(class_id)], &dummy);

#if defined(DEBUG_SPDSET_SDBLINK4B)
        traceSpdGetAElem(class_object_set[class_index(class_id)], &dummy);
#endif

        if (is_identical_db_obj(&object_db[obj->db_key], db_obj, class_id)) {
            return (object_db[obj->db_key].Obj_id); /* Found. */
        }
    }
    return (SPD_NON_ID);                  /* Not found. */
}

#if COMMENTED

void read_a_class(FILE *fp, int file_no, SpdClassId class_id, SpdInfo *info,
                  int (*identify)(SpdClassId, SpdObjId))
{
    SpdDBObj db_obj;

    /*
     * Initialize the offset table for class_id of next file by the
     * maximum object id of the class.
     */
    set_id_offset(file_no + 1, class_id,
                  get_id_offset(file_no, class_id) +
                  SPD_OBJ_ID_OFFSET(class_id) - 1);

    obj_num = spdGetInfoInt(info, spdGetClassNameById(class_id), 0);

    while (obj_num--) {
        db_obj = spdReadObj(fp, class_id);
        db_obj.Obj_id += get_id_offset(file_no, class_id);
        set_id_offset(file_no + 1, class_id, db_obj.Obj_id);
        if (identify(class_id, db_obj)) {
            /* ̾ */
            /* file class λۤʤ? */
            spdFreeStrsInObj(db_obj);
        } else {
            /* ɲ */
            add_new_object(class_id, db_obj);
        }
    }
}

#endif /* COMMENTED */

/******************************************************************************/

#if defined(DEBUG_SPDSET_SDBLINK4B)

static Contents traceSpdAddElem(SpdSet set, int key, Contents contents)
{
    spdMessage("traceSpdAddElem(): key %d.\n", key);
}

static Contents traceSpdGetAElem(SpdSet set, int *key)
{
    spdMessage("traceSpdGetAElem(): *key %d.\n", *key);
}

#endif /* defined(DEBUG_SPDSET_SDBLINK4B) */
