/*
 * Program:             $RCSfile: sdblink4b.h,v $  $Revision: 30.22 $
 *
 * Purpose:
 *
 * Author:              S.Kawai  1995/10/01
 *                      S.Yamamoto  1996/06/02
 *
 * (C) Copyright:       S.Kawai and S.Yamamoto  1995 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef _SDBLINK_H
#define _SDBLINK_H

#define COND_CLASS      1
#define COND_REL        2
#define COND_PROGRAM    3
#define COND_OPERATOR   4
#define COND_SKIP       5
#define COND_IDENT      6

#define SEARCH_FIRST    0
#define SEARCH_NEXT     1
#define P_ID            0
#define C_ID            1

#ifdef SPD_MERGE_SDB

#define SDB_MERGE_CMD   "spdMergeSDB"

#define SAPID_I_P_CLASS_AND_REL_NUM \
    (SPD_CRID_NUM_I_MODEL + SPD_CRID_NUM_P_MODEL)
#define SAPID_I_P_CLASS_NUM \
    (SPD_CID_NUM_I_MODEL + SPD_CID_NUM_P_MODEL)
#define SAPID_I_P_REL_NUM \
    (SPD_RID_NUM_I_MODEL + SPD_RID_NUM_P_MODEL)

#endif /* SPD_MERGE_SDB */

#if defined(SPD_MERGE_SDB)
#define SDBLINK_CLASS_NUM           SAPID_I_P_CLASS_NUM
#define SDBLINK_REL_NUM             SAPID_I_P_REL_NUM
#define SDBLINK_CLASS_AND_REL_NUM   SAPID_I_P_CLASS_AND_REL_NUM
#else
#define SDBLINK_CLASS_NUM           SPD_CID_NUM_I_MODEL
#define SDBLINK_REL_NUM             SPD_RID_NUM_I_MODEL
#define SDBLINK_CLASS_AND_REL_NUM   SPD_CRID_NUM_I_MODEL
#endif

/******************************************************************************/

/* 󥯡 */

typedef struct {
    struct _object      *obj;
    struct _relation    *rel;
} Link;

typedef struct {
    int         elem_num;
    Link        *elem;
} LinkArray;

/* Ρ */

typedef struct _object {
    int         db_key;
    LinkArray   p_link;
    LinkArray   c_link;
} MyObject;

/* Ϣ */

typedef struct _relation {
    int         db_key;
    char        flag;   /* Τflag? */
} Relation;

typedef struct {
    int         elem_num;
    Relation    *elem;
} RelationArray;

/* identifier̾ѥϥåơ֥롥 */

typedef struct _global_table {
    SpdId       id;
    char        *name;
    struct _global_table *next;
} *GlobalTable;

/* ֥Ȥκѡ */

typedef struct {
  MyObject      *obj;
  MyObject      *alt_obj;
  Relation      *rel;
} DeleteField;

typedef struct {
    MyObject    *obj;
    Relation    *rel;
    SpdId       file_id;
    int         offset;
    int         length;
} DeclField;

typedef struct {
    int         type;
    Relation    *rel;
} RealTypeField;

typedef struct {
    SpdId       new_id;
    SpdId       old_id;
} ModifyField;

typedef struct {
    int         db_key_num;     /* elemѤ */
    SpdClassId  class_id;
    int         *db_key;
} CheckTable;

#ifdef SPD_MERGE_SDB
extern SpdSet           class_object_set[SAPID_I_P_CLASS_NUM];
extern RelationArray    relation_array[SAPID_I_P_REL_NUM];
#else
extern SpdSet           class_object_set[SPD_CID_NUM_I_MODEL];
extern RelationArray    relation_array[SPD_RID_NUM_I_MODEL];
#endif

/* ץȥ */

/* linklib.c */

MyObject *search_object_by_rel(MyObject *, int, int);

void push_modification_table(SpdSet, SpdId, SpdId);
void modify_id(int, SpdSet);
SpdBoolean is_identical_db_obj(SpdDBObj *db_obj1, SpdDBObj *db_obj2,
                               SpdClassId classId);

CheckTable *create_obj_check_table(int alloc_size, SpdClassId class_id);
void add_to_obj_check_table(CheckTable *table, int db_key);
SpdId lookup_obj_check_table(CheckTable *table, int db_key);
void free_obj_check_table(CheckTable *table);

void add_link_node(Link **head, MyObject *obj, Relation *rel);
void add_new_link(LinkArray *link_array, MyObject *obj, Relation *rel);
void make_relations(void);
void clear_relations(void);

/* sdbopt.c */
void optimize_file(void);
void optimizer(int optimize_level);

/* link4b_ident.c */

void push_global_table(SpdId id, char *name);
SpdId get_equal_global(char *name);
void free_global_table(void);

/* link4b_write.c */

void write_db_sdblink(void);

/* link4b_read.c */

int class_index(SpdClassId class_id);
int rel_index(SpdClassId class_id);
int offset_index(int file_no, SpdClassId class_id);
void init_id_offset(int file_num);
void free_id_offset(void);
int get_id_offset(int file_no, SpdClassId class_id);
void set_id_offset(int file_no, SpdClassId class_id, SpdObjId object_id);
void add_new_object(SpdClassId classId, SpdDBObj db_obj);
void add_new_relation(SpdClassId classId, SpdDBRel db_rel);
SpdId get_equal_db_obj(SpdDBObj *db_obj, SpdClassId class_id);

/******************************************************************************/

#ifdef SPD_MERGE_SDB

/* Added by H.Ohhashi 1999/06/30. */
extern char         *script_name;
extern char         *gl_sdb_dir;

/* src_table.c */

void put_src_file_id(int rep_no, SpdId file_id, int offset);
void init_src_array(int rep_no, char *path, int file_num);
void init_rep_array(int rep_no);
void initPfileFileTbl(int pfile_num);
void addPfileFileTbl(SpdId old_pfileId, int incAmount);
void modify_src_table(SpdSet modify_set);
void write_cp_src(void);
void free_rep_table(void);

#endif /* SPD_MERGE_SDB */

#endif /* _SDBLINK_H */
