/* 
 * Program:             $RCSfile: link_PSDB.c,v $  $Revision: 50.2 $
 * 
 * Purpose:             
 * 
 * Author:              N.Fukuyasu  2003/08/16
 *                      S.Yamamoto  2003/08/16
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: link_PSDB.c,v 50.2 2003/08/16 10:33:23 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "linklib.h"
#include "psdblink.h"

/* 
 * This function counts the number of objects and relations of all PSDBs.
 */

void count_PSDB_objs(int file_num, SpdObjId file_id[])
{
    SpdInfo *info;
    int     i;

    for (i = 0; i < file_num; i++) {
        spdSetSDBinfoFile(PSDB_DIR, spdObjIdStr(file_id[i]));
        info = spdOpenSDBinfo();
        count_objects_num(info, SPD_MID_P_MODEL);
        spdCloseInfo(info);
    }
}

/*
 * This function loads objects and relations from a PSDB.
 * A PSDB has no object which is identical with the object of other PSDB,
 * so "MergePair", the 4th argument of the function read_objects, is
 * unnecessary.
 */

void load_PSDB(SpdObjId file_id)
{
    static SpdBoolean   is_first_file = TRUE;
    static SpdObjId     max_id[SPD_LAST_CID_P_MODEL + 1];

    /*
     * The number used as a subscript of id_offset is only the range of
     * [SPD_FIRST_CID_P_MODEL, SPD_LAST_CID_P_MODEL], or SPD_CID_FILE.
     * But it is useful to the simplification of algorithm that Class ID
     * is used as subscript of array.
     * The subscript of max_id is also the same reason.
     */

    int id_offset[SPD_LAST_CID_P_MODEL + 1];
    SpdInfo *info;
    SpdConstString pmodel_path;
    char psdb_path[FILENAME_MAX + 1];

    if (is_first_file) {
        initialize_max_id(max_id, SPD_MID_P_MODEL);
        is_first_file = FALSE;
    }

    /* Offsets, generated from max_ids, for merging objects. */
    initialize_id_offset(id_offset, max_id, SPD_MID_P_MODEL);
    /* The references to the "file" appear in the "cfile_ifile". */
    id_offset[SPD_CID_FILE] = 0;

    /*
     * Read a PSDB.
     * The model directory and the information file name are temporally
     * changed. After reading, these are reset.
     */

    pmodel_path = spdGetModelDir(SPD_MID_P_MODEL);
    sprintf(psdb_path, "%s/%s", PSDB_DIR, spdObjIdStr(file_id));
    spdSetModelDir(SPD_MID_P_MODEL, psdb_path);

    spdSetSDBinfoFile(PSDB_DIR, spdObjIdStr(file_id));
    info = spdOpenSDBinfo();
    set_read_sdb_format(info);

    read_objects(NULL, info, SPD_MID_P_MODEL, NULL, id_offset, max_id);
    read_relations(NULL, info, SPD_MID_P_MODEL, NULL, id_offset);

    spdCloseInfo(info);
    spdResetSDBinfoFile();

    spdSetModelDir(SPD_MID_P_MODEL, pmodel_path);
}

/*
 * This function writes a P-model as "SDB/P-model".
 */

void write_Pmodel(void)
{
    SpdInfo     *info = spdOpenSDBinfo();

    if (SPD_STREQ(spdGetInfoString(info, "BINARY_FORMAT_SDB"), "TRUE")) {
        spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
    }
    write_objects(info, SPD_MID_P_MODEL);
    write_relations(info, SPD_MID_P_MODEL);
    spdCloseInfo(info);
}
