/* 
 * Program:             $RCSfile: object_db.c,v $  $Revision: 40.3 $
 * 
 * Purpose:             
 * 
 * Author:              N.Fukuyasu  2003/04/01
 *                      S.Yamamoto  2003/04/01        
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: object_db.c,v 40.3 2003/07/12 12:36:25 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/write_sdb_sub.h>
#include "linklib.h"

static void check_id_space(void);
static SpdDBRel *swap_relation(SpdDBRel *rel);

/*
 * Object, Relationɤ߹Ѥΰ补(optimize.cľܻȤ롥)
 */

SpdDBObj *obj_db[SPD_LAST_RID + 1];
int obj_db_n[SPD_LAST_RID + 1] = { 0 };
SpdDBRel *rel_db[SPD_LAST_RID + 1];
int rel_db_n[SPD_LAST_RID + 1] = { 0 };

/**********************************************************************/

/*
 * class_obj_num[]:
 *   ƥ饹ȤΥ֥ȤθĿɽ
 *   obj_dbrel_dbΰݤѤ롥
 */

static int class_obj_num[SPD_LAST_RID + 1] = { 0 };

void add_class_obj_num(SpdClassId cid, int num)
{
  if (spdIsValidClassOrRelationId(cid)) {
    class_obj_num[cid] += num;
  }
}

/**********************************************************************/

/*
 * search_identical_object[]:
 *   ƥ饹Ȥʥ֥Ȥ򸡺ؿؤΥݥ󥿡
 *   ɬפ˱ƺload_class()ƤӽФϿ롥
 *   ͤNULLξˤϸϹԤʤ
 */

static SearchObjFunc search_identical_object[SPD_LAST_RID + 1];
static SearchRelFunc search_merged_relation[SPD_LAST_RID + 1];

/*
 * initialize_search_functions():
 *   search_identical_object[]롥
 */

void initialize_search_functions(void)
{
  int cid;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    search_identical_object[cid] = NULL;
    search_merged_relation[cid] = NULL;
  }
}

/*
 * set_search_obj_function():
 *   cidΥ饹ɤ߹ݤˡؿfuncѤʥ֥Ȥ
 *   褦ꤹ롥
 */

void set_search_obj_function(SpdClassId cid, SearchObjFunc func)
{
  if (spdIsValidClassOrRelationId(cid)) {
    search_identical_object[cid] = func;
  }
}

/*
 * set_search_rel_function():
 *   cidΥ饹ɤ߹ݤˡؿfuncѤʴϢ
 *   褦ꤹ롥
 */

void set_search_rel_function(SpdClassId cid, SearchRelFunc func)
{
  if (spdIsValidClassOrRelationId(cid)) {
    search_merged_relation[cid] = func;
  }
}

/**********************************************************************/

/*
 * check_id_space():
 *   ɤ߹⤦ȤƤ륪֥/ϢοID_SPACEۤƤʤ
 *   롥
 */

static void check_id_space(void)
{
  SpdClassId cid;
  SpdConstString class_name;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    if (class_obj_num[cid] > SPD_ID_SPACE) {
      class_name = spdGetClassNameById(cid);
      spdErrorExit("Class \"%s\", num %d: Too many objects.\n",
           class_name, class_obj_num[cid]);
    }
  }
}

/*
 * allocate_db():
 *   ֥/Ϣɤ߹िɬפΰݤ롥
 *   obj_db, rel_dbϥ饹Ȥ̡˴롥
 *   ɤ߹߻̾󤻤Ԥ륯饹ˤĤƤϡ;ʬ˳ݤ뤳Ȥ
 *   ʤ뤬Τ鸫ФϤ鷺Ǥ롥
 *   Ϣ饹ϡѰդ٤ɤ߹ޤ롥
 */

void allocate_db(void)
{
  SpdClassId cid;

  check_id_space();

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    obj_db[cid] = NULL;
    rel_db[cid] = NULL;
    if (spdIsValidClassId(cid)) {
      if (class_obj_num[cid] > 0) {
    obj_db[cid] = spdMalloc(sizeof(SpdDBObj) * class_obj_num[cid]);
      }
    } else {
      if (class_obj_num[cid] > 0) {
    rel_db[cid] = spdMalloc(sizeof(SpdDBRel) * class_obj_num[cid]);
      }
    }
  }
}

/*
 * release_db():
 *   obj_db, rel_db롥
 */

void release_db(void)
{
  SpdClassId cid;
  int i;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    if (obj_db[cid] != NULL && cid == 0) {
      for (i = 0; i < obj_db_n[cid]; i++) {
    spdFreeStrsInObj(obj_db[cid][i]);
      }
      spdFree(obj_db[cid]);
      obj_db[cid] = NULL;
      obj_db_n[cid] = 0;
    }
    if (rel_db[cid] != NULL) {
      spdFree(rel_db[cid]);
      rel_db[cid] = NULL;
      rel_db_n[cid] = 0;
    }
  }
}

/*
 * add_new_object():
 *   ֥Ȥobj_dbɲä롥
 */

void add_new_object(SpdClassId cid, SpdDBObj obj)
{
  obj_db[cid][obj_db_n[cid]++] = obj;
}

/*
 * add_new_relation():
 *   Ϣrel_dbɲä롥
 */

void add_new_relation(SpdClassId cid, SpdDBRel rel)
{
  rel_db[cid][rel_db_n[cid]++] = rel;
}

/*
 * delete_object():
 *   MergePair˴Ťơ֥Ȥ롥
 */

void delete_object(SpdClassId cid, MergePair merge_pair)
{
  int i;

  for (i = 0; i < obj_db_n[cid]; i++) {
    if (search_merge_pair(merge_pair, obj_db[cid][i].Obj_id) != SPD_NON_ID) {
      obj_db[cid][i].Obj_id = SPD_NON_ID;
    }
  }
}

/*
 * delete_all_class_object():
 *   ٤ƤΥ饹ˤĤdelete_object()Ԥ
 */

void delete_all_class_object(MergePair merge_pair)
{
  SpdClassId cid;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    delete_object(cid, merge_pair);
  }
}

/*
 * delete_relation():
 *   MergePair˴ŤơϢ롥
 */

void delete_relation(SpdClassId cid, MergePair merge_pair)
{
  int i;

  for (i = 0; i < rel_db_n[cid]; i++) {
    if (search_merge_pair(merge_pair, rel_db[cid][i].Rel_id) != SPD_NON_ID) {
      rel_db[cid][i].Rel_id = SPD_NON_ID;
    }
  }
}

/*
 * delete_all_class_relation():
 *   ٤ƤδϢ饹ˤĤdelete_relation()Ԥ
 */

void delete_all_class_relation(MergePair merge_pair)
{
  SpdClassId cid;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    delete_relation(cid, merge_pair);
  }
}

/*
 * modify_relation():
 *   MergePair˴ŤơϢĤΥ֥IdѴ롥
 */

void modify_relation(SpdClassId cid, MergePair merge_pair)
{
  SpdObjId new_id;
  int i;

  for (i = 0; i < rel_db_n[cid]; i++) {
    new_id = search_merge_pair_recursive(merge_pair, rel_db[cid][i].Rel_p_id);
    if (new_id != SPD_NON_ID) {
      rel_db[cid][i].Rel_p_id = new_id;
    }
    new_id = search_merge_pair_recursive(merge_pair, rel_db[cid][i].Rel_c_id);
    if (new_id != SPD_NON_ID) {
      rel_db[cid][i].Rel_c_id = new_id;
    }
  }
}

/*
 * modify_all_relation():
 *   ٤ƤδϢ饹ˤĤmodify_relation()Ԥ
 */

void modify_all_relation(MergePair merge_pair)
{
  SpdClassId cid;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    modify_relation(cid, merge_pair);
  }
}

/*
 * swap_relation():
 *   ϢοƻҴطȿž롥
 */

static SpdDBRel *swap_relation(SpdDBRel *rel)
{
  SpdObjId obj_id;
  if (rel != NULL) {
    obj_id = rel->Rel_p_id;
    rel->Rel_p_id = rel->Rel_c_id;
    rel->Rel_c_id = obj_id;
  }
  return rel;
}

/**********************************************************************/

/*
 * load_class():
 *   fpcidΥ饹Υ֥Ȥ򤹤٤obj_db[cid]ɤ߹ࡥ
 *   ɤ߹֥ȤIdˤϡid_offset­ޤϿ롥
 *   search_identical_object[cid]NULLǤʤ̾󤻤Ԥ
 */

SpdObjId load_class(SpdClassId cid, FILE *fp, SpdInfo *info, MergePair merge_pair, int id_offset)
{
  SpdDBObj       obj;
  SpdObjId       o_id, max_id = SPD_NON_ID, old_id;
  SearchObjFunc  search = NULL;
  int            num;

  if (merge_pair != NULL) {
    search = search_identical_object[cid];
  }

  num = get_info_num_of_objs_by_cid(info, cid);

  while (num--) {
    obj = spdReadObj(fp, cid);
    old_id = obj.Obj_id;
    obj.Obj_id += id_offset;
    if (max_id < obj.Obj_id) {
      max_id = obj.Obj_id;
    }
    if (search != NULL && (o_id = search(obj, old_id)) != SPD_NON_ID) {
      add_merge_pair(merge_pair, obj.Obj_id, o_id);
    } else {
      add_new_object(cid, obj);
    }
  }

  return max_id;
}

/*
 * load_relation():
 *   fpcidδϢ饹δϢ򤹤٤rel_db[cid]ɤ߹ࡥ
 *   load_class()Ʊ֥ȤIdˤid_offset­ޤ롥
 *   load_class()̾󤻤Ԥ줿ˤmerge_pairˤäơ
 *   ֥ȤIdŬڤѴ롥
 */

void load_relation(SpdClassId cid, FILE *fp, SpdInfo *info, MergePair merge_pair, int id_offset[])
{
  SpdDBRel rel;
  static SpdRelId r_id[SPD_LAST_RID + 1] = { 0 };
  SpdClassId p_cid, c_cid;
  SearchRelFunc search = NULL;
  int num;

  if (merge_pair != NULL) {
    search = search_merged_relation[cid];
  }

  if (r_id[cid] == 0) {
    r_id[cid] = SPD_OBJ_ID_OFFSET(cid);
  }
  p_cid = getParentClassId(cid);
  c_cid = getChildClassId(cid);

  num = get_info_num_of_objs_by_cid(info, cid);

  while (num--) {
    /* 
     * relationˤĤƤidĤƤɬפϤʤ(mergeˤ)
     * optimizeѤ뤿id꿶äƤ롥
     * (optimize r_id == SPD_NON_ID ʤк줿Ϣ)
     */
    rel = spdReadRel(fp, cid, r_id[cid]++);
    rel.Rel_p_id += id_offset[p_cid];
    rel.Rel_c_id += id_offset[c_cid];
    switch (cid) {
    /*
     * δϢƻҴطȿž롥
     * ϻˤϸ᤹(write_relations)
     */
    case SPD_RID_TYPE_IS:
      swap_relation(&rel);
      break;
    default:
      break;
    }
    if (search == NULL || !search(rel, merge_pair)) {
      add_new_relation(cid, rel);
    }
  }
  if (merge_pair != NULL) {
    modify_relation(cid, merge_pair);
  }
}

/**********************************************************************/

/*
 * count_objects_num():
 *   ǥñ̤ǡǼƤ륪֥Ȥο򥯥饹Ȥinfo
 *   롥
 */

void count_objects_num(SpdInfo *info, SpdModelId mid)
{
  SpdClassId cid;

  for (cid = spdGetFirstCid(mid); cid <= spdGetLastRid(mid); cid++) {
    add_class_obj_num(cid, get_info_num_of_objs_by_cid(info, cid));
  }
}

/*
 * initialize_max_id():
 *   ֥ɤ߹߻Ѥmax_id롥
 */

void initialize_max_id(SpdObjId max_id[], SpdModelId mid)
{
  SpdClassId cid;

  for (cid = spdGetFirstCid(mid); cid <= spdGetLastCid(mid); cid++) {
    max_id[cid] = SPD_NON_ID;
  }
}

/*
 * initialize_id_offset():
 *   ֥ɤ߹߻ѤIDƳΤΥեåȤ
 *    max_idˤ롥
 */

void initialize_id_offset(int id_offset[], SpdObjId max_id[], SpdModelId mid)
{
  SpdClassId cid;

  for (cid = spdGetFirstCid(mid); cid <= spdGetLastCid(mid); cid++) {
    id_offset[cid] = (max_id[cid] == SPD_NON_ID ?
              0 : max_id[cid] - SPD_OBJ_ID_OFFSET(cid) + 1);
  }
}

/* Ѵؿ */

/*
 * read_objects():
 *   midǻꤵǥΥ֥Ȥɤ߹ࡥ
 *   fpNULLξˤϡǥѤSDB򥪡ץ󤷡NULLǤϤʤ
 *   ˤϡꤵ줿fpɤ߹ࡥ(ISDBɤ߹ߤ)
 */

void read_objects(FILE *fp, SpdInfo *info, SpdModelId mid, MergePair merge_pair, int id_offset[], SpdObjId max_id[])
{
  SpdClassId cid;
  SpdObjId max_id_tmp;
  SpdBoolean exec_close = FALSE;

  for (cid = spdGetFirstCid(mid); cid <= spdGetLastCid(mid); cid++) {
    if (fp == NULL) {
      fp = spdOpenSDBfile(spdGetModelDir(mid), spdGetClassNameById(cid),
              SAPID_O_READ);
      exec_close = TRUE;
    }

    max_id_tmp = load_class(cid, fp, info, merge_pair, id_offset[cid]);
    if (max_id != NULL && max_id_tmp != SPD_NON_ID) {
      max_id[cid] = max_id_tmp;
    }

    if (exec_close) {
      spdFclose(fp);
      fp = NULL;
    }
  }
}

/*
 * read_relations():
 *   midǻꤵǥδϢɤ߹ࡥ
 *   fpNULLξˤϡǥѤSDB򥪡ץ󤷡NULLǤϤʤ
 *   ˤϡꤵ줿fpɤ߹ࡥ(ISDBɤ߹ߤ)
 */

void read_relations(FILE *fp, SpdInfo *info, SpdModelId mid, MergePair merge_pair, SpdObjId id_offset[])
{
  SpdClassId cid;
  SpdBoolean exec_close = FALSE;

  for (cid = spdGetFirstRid(mid); cid <= spdGetLastRid(mid); cid++) {
    if (fp == NULL) {
      fp = spdOpenSDBfile(spdGetModelDir(mid), spdGetClassNameById(cid),
              SAPID_O_READ);
      exec_close = TRUE;
    }

    load_relation(cid, fp, info, merge_pair, id_offset);

    if (exec_close) {
      spdFclose(fp);
      fp = NULL;
    }
  }
}

/* Ѵؿ */

/*
 * write_objects():
 *   ꤵ줿饹Τ٤ƤΥ֥ȤSDB˽Ϥ롥
 */

void write_objects(SpdInfo *info, SpdModelId mid)
{
  FILE *fp;
  int i = 0;
  int obj_num;
  SpdClassId cid;
  SpdConstString class_name, model_dir;

  model_dir = spdGetModelDir(mid);
  for (cid = spdGetFirstCid(mid); cid <= spdGetLastCid(mid); cid++) {
    class_name = spdGetClassNameById(cid);
    fp = spdOpenSDBfile(model_dir, class_name, SAPID_O_WRITE);
    obj_num = 0;
    for (i = 0; i < obj_db_n[cid]; i++) {
      if (obj_db[cid][i].Obj_id != SPD_NON_ID) {
    spdWriteDBObj(fp, &obj_db[cid][i]);
    obj_num++;
      }
    }
    spdFclose(fp);
    spdSetInfoInt(info, class_name, obj_num);
  }
}

/*
 * write_relations():
 *   ꤵ줿饹Τ٤ƤΥ֥ȤSDB˽Ϥ롥
 */

void write_relations(SpdInfo *info, SpdModelId mid)
{
  FILE *fp;
  int i = 0;
  int rel_num;
  SpdClassId cid;
  SpdConstString class_name, model_dir;

  model_dir = spdGetModelDir(mid);
  for (cid = spdGetFirstRid(mid); cid <= spdGetLastRid(mid); cid++) {
    class_name = spdGetClassNameById(cid);
    fp = spdOpenSDBfile(model_dir, class_name, SAPID_O_WRITE);
    rel_num = 0;
    for (i = 0; i < rel_db_n[cid]; i++) {
      if (rel_db[cid][i].Rel_id != SPD_NON_ID) {
    switch (cid) {
      /*
       * ȿžƤƻҴط򸵤᤹(load_relation())
       */
    case SPD_RID_TYPE_IS:
      swap_relation(&rel_db[cid][i]);
      break;
    default:
      break;
    }
    spdWriteDBRel(fp, &rel_db[cid][i]);
    rel_num++;
      }
    }
    spdFclose(fp);
    spdSetInfoInt(info, class_name, rel_num);
  }
}
