/* 
 * Program:             $RCSfile: optimize.c,v $  $Revision: 40.6 $
 * 
 * Purpose:             
 * 
 * Author:              N.Fukuyasu  2003/04/01
 *                      S.Yamamoto  2003/04/01        
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: optimize.c,v 40.6 2003/09/01 07:01:31 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "linklib.h"
#include "sdblink5.h"

typedef SpdSet ExtObjPair;

typedef struct _ext_object_pair {
  ExtDBObj *ext_obj1;
  ExtDBObj *ext_obj2;
} ExtObjectPair;

static int compare_ext_db_obj(const ExtDBObj *a, const ExtDBObj *b);
static void initialize_bsearch(void);
static ExtDBObj *find_object(SpdObjId obj_id);
static void search_children(void);

static void initialize_optimization_field(void);
static void reduce_class_optimization_field(SpdClassId cid);
static void reduce_optimization_field(void);
static void release_optimization_field(void);

static ExtObjPair create_ext_obj_pair(void);
static void release_ext_obj_pair(ExtObjPair pair);
static void add_ext_obj_pair(ExtObjPair pair, ExtDBObj *src_obj, ExtDBObj *dst_obj);

static SpdBoolean same_tree(ExtDBObj *obj1, ExtDBObj *obj2);

static void modify_offset(ExtDBObj *obj, int offset);
static void make_relative_offset(void);

static void delete_tree_decl_type(MergePair tree_pair);
static void delete_tree_ident_type(MergePair tree_pair);
static void find_declaration_tree(ExtDBObj *src_obj, ExtDBObj *dst_obj, MergePair tree_pair);
static ExtDBObj *find_decl_decl_child(ExtDBObj *decl);
static ExtDBObj *find_decl_ident_child(ExtDBObj *decr);
static void make_ext_decl_table(void);
static void release_ext_decl_table(void);
static SpdBoolean different_ident(ExtDBObj *ident1, ExtDBObj *ident2);
static void optimize_declaration(void);

static void optimize_type(void);

static SpdObjId search_identical_operator(SpdDBObj obj, SpdObjId old_id);
static SpdObjId search_identical_identifier(SpdDBObj obj, SpdObjId old_id);
static SpdBoolean search_merged_ident_type(SpdDBRel rel, MergePair merge_pair);

/**********************************************************************/

static int optimize_level = 0;

void set_optimize_level(int level)
{
  optimize_level = level;
}

int get_optimize_level(void)
{
  return optimize_level;
}

/**********************************************************************/

/* object_db.cΥ֥Ȥ򤽤ΤޤѤ롥 */

extern SpdDBObj *obj_db[];
extern int obj_db_n[];
extern SpdDBRel *rel_db[];
extern int rel_db_n[];

/*
 * ext_obj_db:
 *   ץƥޥκݤɬפʾɲäobj_db
 *   obj_dbƱ饹Ȥ̡˴롥
 */

static ExtDBObj *ext_obj_db[SPD_LAST_RID + 1];
static int ext_obj_db_n[SPD_LAST_RID + 1] = { 0 };

/*
 * compare_ext_db_obj():
 *   ExtDBObjObj_idӤ롥
 *   ext_obj_dbbsearch()뤿sortݤѤ롥
 */

static int compare_ext_db_obj(const ExtDBObj *a, const ExtDBObj *b)
{
  return (a->obj->Obj_id - b->obj->Obj_id);
}

/*
 * initialize_bsearch():
 *   ext_obj_dbbsearch()뤿sortԤ
 */

static void initialize_bsearch(void)
{
  SpdClassId cid;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    if (ext_obj_db_n[cid] > 0) {
      SPD_QSORT(ext_obj_db[cid], ext_obj_db_n[cid],
          sizeof(ExtDBObj), compare_ext_db_obj);
    }
  }
}

/*
 * find_object():
 *   bsearch()Ѥext_obj_dbobj_idĥ֥Ȥ򸡺롥
 *   ext_obj_db[cid]obj->Obj_id˥ȤƤɬפ롥
 */

static ExtDBObj *find_object(SpdObjId obj_id)
{
  SpdClassId cid;
  SpdDBObj key_obj;
  /* ExtDBObj *res, key = { &key_obj, NULL, 0, 0 }; */
  ExtDBObj *res, key = { NULL /* Just dummy. */, NULL, 0, 0 };

  key.obj = &key_obj;
  cid = spdGetClassId(obj_id);
  key_obj.Obj_id = obj_id;
  res = bsearch(&key, ext_obj_db[cid], ext_obj_db_n[cid], sizeof(ExtDBObj),
                (int (*)(const void *, const void *))compare_ext_db_obj);
  return res;
}

/*
 * search_children():
 *   ٤ƤΥ֥ȤˤĤƹطλҥ֥ȤƤ
 *   modify_offset()ι®ΤTYP_COMPʳδϢˤĤƤ
 *   relationƤ
 */

static void search_children(void)
{
  SpdClassId cid;
  SpdObjId p_obj_id;
  ExtDBObj *parent;
  int i;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    if (spdIsValidRelationId(cid)) {
      for (i = 0; i < rel_db_n[cid]; i++) {
    p_obj_id = rel_db[cid][i].Rel_p_id;
    parent = find_object(p_obj_id);
    if (isCompRel(cid)) {
      add_child_relation(parent, &rel_db[cid][i]);
    } else {
      add_ext_relation(parent, &rel_db[cid][i]);
    }
      }
    }
  }
}

/*
 * initialize_optimization_field():
 *   ץƥޥΤˡext_obj_dbѰդ롥
 *   bsearch()Ǹ뤿ᡤext_obj_dbObj_idsortƤ
 *   ɬפ롥
 *   ץƥޥľ˸ƤӽФ롥
 */

static void initialize_optimization_field(void)
{
  SpdClassId cid;
  int i;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    ext_obj_db[cid] = NULL;
    if (obj_db_n[cid] > 0) {
      ext_obj_db[cid] = spdMalloc(sizeof(ExtDBObj) * obj_db_n[cid]);
      for (i = 0; i < obj_db_n[cid]; i++) {
    ext_obj_db[cid][i].obj = &obj_db[cid][i];
    ext_obj_db[cid][i].child = NULL;
    ext_obj_db[cid][i].child_num = 0;
    ext_obj_db[cid][i].child_size = 0;
    ext_obj_db[cid][i].rel = NULL;
    ext_obj_db[cid][i].rel_num = 0;
    ext_obj_db[cid][i].rel_size = 0;
      }
      ext_obj_db_n[cid] = obj_db_n[cid];
    }
  }

  initialize_bsearch();
}

/*
 * reduce_class_optimization_field():
 *   ץƥޥѤext_obj_dbIdݤäޤ޺줿ʬͤ롥
 *   ֥Ȥ(obj_dbIdSPD_NON_ID)ǸƤӽФ롥
 */

static void reduce_class_optimization_field(SpdClassId cid)
{
  int i, j = 0;

  for (i = 0; i < ext_obj_db_n[cid]; i++) {
    if (ext_obj_db[cid][i].obj->Obj_id != SPD_NON_ID) {
      if (i != j) {
    ext_obj_db[cid][j] = ext_obj_db[cid][i];
      }
      j++;
    } else {
      free_ext_db_obj(ext_obj_db[cid][i]);
    }
  }
  ext_obj_db_n[cid] = j;
}

/*
 * reduce_optimization_field():
 *   ٤ƤΥ饹ˤĤơreduce_class_optimization_field()Ԥ
 */

static void reduce_optimization_field(void)
{
  SpdClassId cid;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    reduce_class_optimization_field(cid);
  }
}

/*
 * release_optimization_field():
 *   ץƥޥѤ˳ݤ줿ext_db_obj롥
 *   ٤ƤΥץƥޥäǸƤӽФ롥
 */

static void release_optimization_field(void)
{
  SpdClassId cid;
  int i;

  for (cid = SPD_FIRST_CID; cid <= SPD_LAST_RID; cid++) {
    if (spdIsValidClassId(cid)) {
      for (i = 0; i < ext_obj_db_n[cid]; i++) {
    free_ext_db_obj(ext_obj_db[cid][i]);
      }
      spdFree(ext_obj_db[cid]);
    }
  }
}

/**********************************************************************/

static ExtObjPair create_ext_obj_pair(void)
{
  SpdSet pair;

  pair = spdCreateSet(sizeof(ExtObjectPair));

  return (ExtObjPair)pair;
}

static void release_ext_obj_pair(ExtObjPair pair)
{
  if (pair != NULL) {
    spdFreeSet(pair);
  }
}

static void add_ext_obj_pair(ExtObjPair pair, ExtDBObj *src_obj, ExtDBObj *dst_obj)
{
  ExtObjectPair obj_pair;

  obj_pair.ext_obj1 = src_obj;
  obj_pair.ext_obj2 = dst_obj;
  spdAddElem((SpdSet)pair, src_obj->obj->Obj_id, &obj_pair);
}

/**********************************************************************/

/*
 * same_tree():
 *   obj1롼ȤȤĥ꡼obj2롼ȤȤĥ꡼
 *   ɤ򸡺롥
 *
 * 르ꥺ
 *   (1). obj1 == obj2 ʤTRUE
 *   (2). ٤Ƥ°ƱǤʤ(줫°ۤʤʤ)FALSE
 *   (3). childοۤʤFALSE
 *   (4). obj1Τ٤ƤchildˤĤơobj2Τ٤Ƥchild椫same_tree()
 *        ȤʤΤĤʤFALSE
 *   (5). (2), (3), (4)ʳʤTRUE
 */

static SpdBoolean same_tree(ExtDBObj *obj1, ExtDBObj *obj2)
{
  SpdClassId obj1_rel_cid, obj2_rel_cid;
  ExtDBObj *c_obj1, *c_obj2;
  int i, j;

  /*
   * obj1obj2ϤɤNULLǤϤʤƱ饹Υ֥ȤǤ뤳Ȥ
   * ꤷƤ롥
   *
   * if (obj1 == NULL || obj2 == NULL) return FALSE;
   * if (spdGetClassId(obj1->obj->Obj_id) !=
   *     spdGetClassId(obj2->obj->Obj_id)) {
   *   return FALSE;
   * }
   */

  if (obj1 == obj2) return TRUE;
  if (!is_identical_object(obj1->obj, obj2->obj)) return FALSE;
  if (obj1->child_num != obj2->child_num) return FALSE;

  if (obj1->child_num != 0) {
    for (i = 0; i < obj1->child_num; i++) {
      obj1_rel_cid = spdGetClassId(obj1->child[i]->Rel_id);
      c_obj1 = find_object(obj1->child[i]->Rel_c_id);
      /*
       * childΥꥹȤobj1obj2Ʊ¤ǤݾڤʤΤǡ
       * ۤĴ٤Ƥ롥
       */
      for (j = i; j < obj2->child_num; j++) {
    obj2_rel_cid = spdGetClassId(obj2->child[j]->Rel_id);
    if (obj1_rel_cid != obj2_rel_cid) continue;
    c_obj2 = find_object(obj2->child[j]->Rel_c_id);
    if (same_tree(c_obj1, c_obj2)) {
      if (i != j) {
        SpdDBRel *tmp = obj2->child[i];
        obj2->child[i] = obj2->child[j];
        obj2->child[j] = tmp;
      }
      break;
    }
      }
      if (j == obj2->child_num) return FALSE;
    }
  }

  return TRUE;
}

/**********************************************************************/

/*
 * modify_offset():
 *   objƤȤ빽طoffsetХեåȤѴ
 *   objλҤˤĤƤƵŪ˼¹Ԥ롥
 *   decl_typeϹطǤϤʤoffsetĤΤХեåȲ롥
 */

static void modify_offset(ExtDBObj *obj, int offset)
{
  ExtDBObj *child;
  int i;

  if (obj == NULL) return;

  for (i = 0; i < obj->child_num; i++) {
    /* ٤Ƥιط(C_TYP_CMP)offset? */
    obj->child[i]->Rel_offset -= offset;
    child = find_object(obj->child[i]->Rel_c_id);
    modify_offset(child, offset);
  }

  if (spdGetClassId(obj->obj->Obj_id) == SPD_CID_DECLARATION) {
    for (i = 0; i < obj->rel_num; i++) {
      if (spdGetClassId(obj->rel[i]->Rel_id) == SPD_RID_DECL_TYPE &&
          obj->rel[i]->Rel_offset != SPD_NON_OFFSET) {
        obj->rel[i]->Rel_offset -= offset;
      }
    }
  }
}

/*
 * make_relative_offset():
 *   file_declǴϢդdeclarationʲΤ٤ƤιطˤĤ
 *   file_decloffsetХեåȤѴ롥
 */

static void make_relative_offset(void)
{
  ExtDBObj *child;
  int i;

  for (i = 0; i < rel_db_n[SPD_RID_FILE_DECL]; i++) {
    child = find_object(rel_db[SPD_RID_FILE_DECL][i].Rel_c_id);
    modify_offset(child, rel_db[SPD_RID_FILE_DECL][i].Rel_offset);
  }
}

/**********************************************************************/

/*
 * delete_tree_decl_type(), delete_tree_ident_type():
 *   declarationΥĥ꡼¾Υ֥ȤؤλȤ롥
 *   Ϣdecl_type, ident_typeǤ롥
 */

static void delete_tree_decl_type(MergePair tree_pair)
{
  SpdObjId decl_id;
  int i;

  for (i = 0; i < rel_db_n[SPD_RID_DECL_TYPE]; i++) {
    if (rel_db[SPD_RID_DECL_TYPE][i].Rel_id != SPD_NON_ID) {
      decl_id = rel_db[SPD_RID_DECL_TYPE][i].Rel_p_id;
      if (search_merge_pair(tree_pair, decl_id) != SPD_NON_ID) {
    rel_db[SPD_RID_DECL_TYPE][i].Rel_id = SPD_NON_ID;
      }
    }
  }
}

static void delete_tree_ident_type(MergePair tree_pair)
{
  SpdObjId ident_id;
  int i;

  for (i = 0; i < rel_db_n[SPD_RID_IDENT_TYPE]; i++) {
    if (rel_db[SPD_RID_IDENT_TYPE][i].Rel_id != SPD_NON_ID) {
      ident_id = rel_db[SPD_RID_IDENT_TYPE][i].Rel_p_id;
      if (search_merge_pair(tree_pair, ident_id) != SPD_NON_ID) {
    rel_db[SPD_RID_IDENT_TYPE][i].Rel_id = SPD_NON_ID;
      }
    }
  }
}

/*
 * find_declaration_tree():
 *   src_obj롼ȤȤĥ꡼ιϢʤ顤
 *   2ĤΥĥ꡼б륪֥Ȥ򸡺롥
 */

static void find_declaration_tree(ExtDBObj *src_obj, ExtDBObj *dst_obj, MergePair tree_pair)
{
  int i;
  ExtDBObj *c_src_obj, *c_dst_obj;

  /* if (src_obj == NULL || dst_obj == NULL) return; */

  if (src_obj == dst_obj) return;

  for (i = 0; i < src_obj->child_num; i++) {
    c_src_obj = find_object(src_obj->child[i]->Rel_c_id);
    c_dst_obj = find_object(dst_obj->child[i]->Rel_c_id);
    find_declaration_tree(c_src_obj, c_dst_obj, tree_pair);
    src_obj->child[i]->Rel_id = SPD_NON_ID;
  }
  add_merge_pair(tree_pair, src_obj->obj->Obj_id, dst_obj->obj->Obj_id);
}

/**********************************************************************/

/*
 * DeclProperty:
 *   declarationħդǡ
 *   ߤϡdeclarationidentifierΤߡ
 */

typedef struct _decl_property {
  ExtDBObj *decl;
  ExtDBObj *ident;
} DeclProperty;

/*
 * ext_decl:
 *   external_declarationΰ
 */

static DeclProperty *ext_decl = NULL;
static int ext_decl_num = 0;

/*
 * find_decl_decl_child():
 *   decl_declˤdeclarationdeclarator롥
 */

static ExtDBObj *find_decl_decl_child(ExtDBObj *decl)
{
  int i;

  if (decl == NULL || decl->child_num == 0) {
    /*
     * ISDBϤƤСξＰϿˤʤ뤳ȤϤʤ
     */
    spdWarning("Invalid ISDB.\n");
    return NULL;
  }
  for (i = 0; i < decl->child_num; i++) {
    if (spdGetClassId(decl->child[i]->Rel_id) == SPD_RID_DECL_DECL) {
      return find_object(decl->child[i]->Rel_c_id);
    }
  }

  return NULL;
}

/*
 * find_decl_ident_child():
 *   decl_identˤdeclaratoridentifier롥
 */

static ExtDBObj *find_decl_ident_child(ExtDBObj *decr)
{
  int i;

  if (decr == NULL) {
    /*
     * ISDBϤƤСξＰϿˤʤ뤳ȤϤʤ
     */
    spdWarning("Invalid ISDB.\n");
    return NULL;
  }
  for (i = 0; i < decr->child_num; i++) {
    if (spdGetClassId(decr->child[i]->Rel_id) == SPD_RID_DECL_IDENT) {
      return find_object(decr->child[i]->Rel_c_id);
    }
  }

  return NULL;
}

/*
 * make_ext_decl_table():
 *   file_declǴϢդ뤹٤Ƥdeclarationext_declϿ롥
 */

static void make_ext_decl_table(void)
{
  ExtDBObj *decl;
  ExtDBObj *decr, *ident;
  int i;

  ext_decl_num = rel_db_n[SPD_RID_FILE_DECL];
  ext_decl = spdMalloc(sizeof(DeclProperty) * ext_decl_num);

  for (i = 0; i < ext_decl_num; i++) {
    decl = find_object(rel_db[SPD_RID_FILE_DECL][i].Rel_c_id);
    ext_decl[i].decl = decl;
    decr = find_decl_decl_child(decl);
    ident = find_decl_ident_child(decr);
    ext_decl[i].ident = ident;
  }
}

/*
 * release_ext_decl_table():
 *   ext_decl롥
 */

static void release_ext_decl_table(void)
{
  if (ext_decl != NULL) {
    spdFree(ext_decl);
  }
  ext_decl = NULL;
  ext_decl_num = 0;
}

/*
 * different_ident():
 *   2Ĥidentifier(ident1ident2)ۤʤ뤫ɤȽꤹ롥
 *   TRUEξˤɬۤʤ뤬FALSEξˤƱȤϸ¤ʤ
 */

static SpdBoolean different_ident(ExtDBObj *ident1, ExtDBObj *ident2)
{
  int name1_len, name2_len;
  SpdString name1, name2;

  if (ident1 != ident2) {
    if (ident1 == NULL || ident2 == NULL) return TRUE;
    name1_len = ident1->obj->ivalue[SPD_AID_IDENTIFIER_SORT];
    name2_len = ident2->obj->ivalue[SPD_AID_IDENTIFIER_SORT];
    if (name1_len != name2_len) return TRUE;
    name1 = ident1->obj->Ident_name;
    name2 = ident2->obj->Ident_name;
    if (!SPD_STREQ(name1, name2)) return TRUE;
  }

  return FALSE;
}

/*
 * optimize_declaration():
 *   Ʊإåեͳ褹Ȼפdeclarationޡ롥
 *   declarationΥĥ꡼¤ƱɤȽꤹ롥
 *   ʤPIDBѤdeclarationν°եĴ٤٤Ǥ롥
 *   ٤Ƥexternal declarationǸ롥
 *   Ѥߤdeclarationext_decl[n].declNULLˤ롥
 *
 *   1.ޡǽdeclarationpair롥
 *   2.file_decl롥
 *   3.declaration˴ޤޤ륪֥ȤΤ¾Υĥ꡼
 *     ȤǽΤ륪֥ȤˤĤơޡ륪֥
 *     ȤȤpair롥
 *   4.declarationĥ꡼(طޤ)롥
 *   5.decl_type, ident_type롥
 *   6.줿֥ȤΤ¾Υĥ꡼黲ȤƤΤˤĤ
 *     ơϢ롥
 */

static void optimize_declaration(void)
{
  ExtObjPair decl_pair;
  int i, j;
  MergePair tree_pair;

  decl_pair = create_ext_obj_pair();

  make_ext_decl_table();

  for (i = 0; i < ext_decl_num - 1; i++) {
    if (ext_decl[i].decl == NULL) continue;
    for (j = i + 1; j < ext_decl_num; j++) {
      if (ext_decl[j].decl == NULL) continue;
      if (different_ident(ext_decl[i].ident, ext_decl[j].ident)) continue;
      if (same_tree(ext_decl[i].decl, ext_decl[j].decl)) {
    add_ext_obj_pair(decl_pair, ext_decl[j].decl, ext_decl[i].decl);
    ext_decl[j].decl = NULL;
      }
    }
  }
  release_ext_decl_table();

  tree_pair = create_merge_pair();

  for (i = 0; i < decl_pair->elemNum; i++) {
    ExtObjectPair *obj_pair;
    obj_pair = spdGetAElem(decl_pair, NULL);
    find_declaration_tree(obj_pair->ext_obj1, obj_pair->ext_obj2, tree_pair);
  }

  release_ext_obj_pair(decl_pair);

  delete_all_class_object(tree_pair);
  delete_tree_decl_type(tree_pair);
  delete_tree_ident_type(tree_pair);

  modify_all_relation(tree_pair);
  
  release_merge_pair(tree_pair);

  reduce_optimization_field();
}

/**********************************************************************/

static void optimize_type(void)
{
  MergePair merge_pair;
  int i, j;

  merge_pair = create_merge_pair();

  for (i = 0; i < ext_obj_db_n[SPD_CID_TYPE] - 1; i++) {
    if (ext_obj_db[SPD_CID_TYPE][i].obj->Obj_id == SPD_NON_ID) continue;
    if (ext_obj_db[SPD_CID_TYPE][i].obj->ivalue[SPD_AID_TYPE_SORT]
    != TYPE_DECLARATION) continue;
    for (j = i+1; j < ext_obj_db_n[SPD_CID_TYPE]; j++) {
      if (ext_obj_db[SPD_CID_TYPE][j].obj->Obj_id == SPD_NON_ID) continue;
      if (ext_obj_db[SPD_CID_TYPE][j].obj->ivalue[SPD_AID_TYPE_SORT]
      != TYPE_DECLARATION) continue;
      if (is_identical_object(ext_obj_db[SPD_CID_TYPE][i].obj,
                  ext_obj_db[SPD_CID_TYPE][j].obj)) {
    add_merge_pair(merge_pair, ext_obj_db[SPD_CID_TYPE][j].obj->Obj_id,
               ext_obj_db[SPD_CID_TYPE][i].obj->Obj_id);
    ext_obj_db[SPD_CID_TYPE][j].obj->Obj_id = SPD_NON_ID;
      }
    }
  }

  modify_all_relation(merge_pair);
  release_merge_pair(merge_pair);

  reduce_class_optimization_field(SPD_CID_TYPE);
}

/**********************************************************************/

/*
 * ISDB_optimize():
 *   ISDB->SDBoptimizer
 *   level1,3(̤Υ٥ϲ̤Υ٥ޤ)
 *     1: type̾
 *     3: Ʊإåեͳ褹external declarationΥޡ &
 *        Relative offset
 */

void ISDB_optimize(void)
{
  initialize_optimization_field();

  spdMessage("Optimizing.\n");

  if (optimize_level >= 3) {
    search_children();
    spdMessage("Making relative offset.\n");
    make_relative_offset();
    spdMessage("Optimize declaration.\n");
    optimize_declaration();
  }

  optimize_type();

  release_optimization_field();
}

/**********************************************************************/

/*
 * search_identical_operator():
 *   objФơޤǤɤ߹operator֥ȤƱ
 *   ֥Ȥ򸡺Υ֥Id֤
 *   ĤʤäˤSPD_NON_ID֤
 *   sizeof(type), castʳoperatorˤĤƤsortƱʤƱ
 *   operatorȤ롥
 */

static SpdObjId search_identical_operator(SpdDBObj obj, SpdObjId old_id)
{
  int sort;
  int i;

  sort = obj.ivalue[SPD_AID_OPERATOR_SORT];
  if (sort == OP_CAST || sort == OP_SIZEOF_TYPE) {
    return SPD_NON_ID;
  }

  /* operator tableаʲθϰϤϽ̾ǽ */
  for (i = 0; i < obj_db_n[SPD_CID_OPERATOR]; i++) {
    if (sort == obj_db[SPD_CID_OPERATOR][i].ivalue[SPD_AID_OPERATOR_SORT]) {
      return obj_db[SPD_CID_OPERATOR][i].Obj_id;
    }
  }
  return SPD_NON_ID;
}

/**********************************************************************/

#define is_global(obj) \
        ((obj).ivalue[SPD_AID_IDENTIFIER_SCOPE] == ID_SCOPE_GLOBAL)

static SpdObjId search_identical_identifier(SpdDBObj obj, SpdObjId old_id)
{
  int i;

  if (is_global(obj)) {
    for (i = 0; i < obj_db_n[SPD_CID_IDENTIFIER]; i++) {
      if (SPD_STREQ(obj.Ident_name, obj_db[SPD_CID_IDENTIFIER][i].Ident_name) &&
      is_identical_object(&obj, &obj_db[SPD_CID_IDENTIFIER][i])) {
    return obj_db[SPD_CID_IDENTIFIER][i].Obj_id;
      }
    }
  }

  return SPD_NON_ID;
}

/**********************************************************************/

static SpdBoolean search_merged_ident_type(SpdDBRel rel, MergePair merge_pair)
{
  return (search_merge_pair(merge_pair, rel.Rel_p_id) != SPD_NON_ID);
}

/**********************************************************************/

/*
 * initialize_search_functions_ISDB():
 *   ̾Ѥsearch_identical_object롥
 *   ̾󤻤ɬפʥ饹search_identical_objectƱ֥Ȥ
 *   뤿δؿϿ롥
 */

void initialize_search_functions_ISDB(void)
{
  initialize_search_functions();
  set_search_obj_function(SPD_CID_OPERATOR, search_identical_operator);
  set_search_obj_function(SPD_CID_IDENTIFIER, search_identical_identifier);
  set_search_rel_function(SPD_RID_IDENT_TYPE, search_merged_ident_type);
}
