/* 
 * Program:             $RCSfile: psdblink.c,v $  $Revision: 50.2 $
 * 
 * Purpose:             
 * 
 * Author:              N.Fukuyasu  2003/08/16
 *                      S.Yamamoto  2003/08/16
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: psdblink.c,v 50.2 2003/08/16 10:33:23 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "linklib.h"
#include "psdblink.h"

static void print_usage(void);
static void parse_option(int argc, char *argv[]);

static SpdConstString program_name;   /* This program name. */
static SpdString target_name = SPD_DEFAULT_TARGET_PROGRAM_NAME;

static void print_usage(void)
{
    static const char *message =
        "usage: %s [-d SDB_dir][-o target_name] files...\n";

    spdErrorExit(message, program_name);
}

static void parse_option(int argc, char *argv[])
{
    extern char *optarg;
    int ch;

    while ((ch = getopt(argc, argv, "o:d:")) != -1) {
        switch (ch) {
          case 'o':
            target_name = optarg;
            break;
          case 'd':
            spdSetSDBdir(optarg);
            break;
          default:
            print_usage();
        }
    }
}

int main(int argc, char *argv[])
{
    /* extern int optind; */
    SpdObjId    prog_id, *file_id;
    int         file_num, i;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parse_option(argc, argv);
    argc -= optind;
    argv += optind;
    if (argc < 1) {
        print_usage();
    }

    prog_id = spdSearchGlobalObj("program", target_name);
    if (prog_id == SPD_NON_ID) {
        spdErrorExit("\"%s\" is not found.\n", target_name);
    }

    file_num = argc;
    file_id = spdMalloc(sizeof(SpdObjId) * file_num);
    for (i = 0; i < file_num; i++) {
        file_id[i] = spdSearchGlobalObjEx("file", argv[i]);
        if (file_id[i] == SPD_NON_ID) {
            spdErrorExit("\"%s\" is not found.\n", argv[i]);
        }
    }

    spdMessage("Program name \"%s\".\n", target_name);

    spdUseModel(&P_model);

    count_PSDB_objs(file_num, file_id);
    allocate_db();
    initialize_search_functions();
    for (i = 0; i < file_num; i++) {
        load_PSDB(file_id[i]);
    }
    spdFree(file_id);

    write_Pmodel();

    release_db();

    return EXIT_SUCCESS;
}
