/* 
 * Program:             $RCSfile: sdblink5.c,v $  $Revision: 40.3 $
 * 
 * Purpose:             
 * 
 * Author:              N.Fukuyasu  2003/04/01
 *                      S.Yamamoto  2003/04/01        
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: sdblink5.c,v 40.3 2003/05/29 04:19:49 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "linklib.h"
#include "sdblink5.h"

static void print_usage(void);
static void parse_option(int argc, char *argv[]);

/**********************************************************************/

static SpdConstString   program_name;   /* This program name. */
static SpdString        target_name = SPD_DEFAULT_TARGET_PROGRAM_NAME;

static void print_usage(void)
{
  static const char *message =
    "usage: %s [-d SDB_dir][-o target_name][-O level][-b] files...\n";

  spdErrorExit(message, program_name);
}

static void parse_option(int argc, char *argv[])
{
  extern char *optarg;
  int ch;
  int level;

  while ((ch = getopt(argc, argv, "o:d:O:b")) != -1) {
    switch (ch) {
    case 'o':
      target_name = optarg;
      break;
    case 'd':
      spdSetSDBdir(optarg);
      break;
    case 'O':
      level = atoi(optarg);
      if (level < 1 || 3 < level) {
    print_usage();
      } else if (level == 2) {
    spdWarning("Obsolete \"-O2\" option is specified. Level 3 is selected.\n");
    level = 3;
      }
      set_optimize_level(level);
      break;
    case 'b':
      spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
      break;
    default:
      print_usage();
    }
  }
}

int main(int argc, char *argv[])
{
  extern int optind;
  SpdObjId *file_id, prog_id;
  int file_num;
  int i;

  program_name = spdBasename(argv[0]);
  spdInitialize(program_name);

  parse_option(argc, argv);
  argc -= optind;
  argv += optind;
  if (argc < 1) {
    print_usage();
  }

  file_num = argc;
  file_id = spdMalloc(sizeof(SpdObjId) * file_num);
  for (i = 0; i < file_num; i++) {
    if ((file_id[i] = spdSearchGlobalObjEx("file", argv[i])) == SPD_NON_ID) {
      spdErrorExit("\"%s\" is not found.\n", argv[i]);
    }
  }

  spdUseModel(&I_model);

  count_ISDB_objs(file_num, file_id);
  allocate_db();

  prog_id = make_a_program_object(target_name);
  read_file_objects_Imodel();
  initialize_search_functions_ISDB();
  for (i = 0; i < file_num; i++) {
    make_prog_file(prog_id, file_id[i]);
    load_ISDB(file_id[i]);
  }
  spdFree(file_id);

  if (get_optimize_level() > 0) {
    ISDB_optimize();
  }

  write_Imodel();

  /* Ǥ˲Ƥΰfree褦ȤƤޤ */
  /* release_db(); */

  return EXIT_SUCCESS;
}

/**********************************************************************/
