/* 
 * Program:             $RCSfile: sdbmerge.c,v $  $Revision: 40.5 $
 * 
 * Purpose:             
 * 
 * Author:              N.Fukuyasu  2003/04/01
 *                      S.Yamamoto  2003/04/01        
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: sdbmerge.c,v 40.5 2003/07/13 09:15:59 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "linklib.h"
#include "sdbmerge.h"

static void write_cfile_ifile(void);
static void add_cfile_ifile(SpdObjId cfile_id, SpdObjId ifile_id);
static void release_cfile_ifile(void);

static void write_pfile_file(void);
static void add_pfile_file(SpdObjId pfile_id, SpdObjId file_id);
static void release_pfile_file(void);

static void add_system_file_table(SpdObjId id, SpdConstString name);
static SpdObjId search_system_file_table(SpdString name);
static void release_system_file_table(void);
static SpdObjId search_identical_file(SpdDBObj obj, SpdObjId old_id);

static void count_SDB_objs(int SDB_num, char *SDB_dirs[]);
static void load_SDB(char *SDB_dir);

static void write_SDB_info(SpdInfo *info);
static void write_SDB(void);

static void print_usage(void);
static void parse_option(int argc, char *argv[]);

/**********************************************************************/

static RelTable cfile_ifile = RELATION_TABLE_EMPTY;

/*
 * write_cfile_ifile():
 *   analyzer/PIDB/PIDB.c pidbWriteCtoITable()νϷˤäơ
 *   PIDB/cfile_ifileϤ롥
 */

/*
 * ".c"ե".i"եбɽ"./SDB/PIDB/cfile_ifile"˽Ϥ롥
 * 1Ԥη
 * 
 *     <".c"եμֹ>\t<".i"եμֹ>
 * 
 * Ǥ롥
 */

static void write_cfile_ifile(void)
{
  FILE *fp;
  int i;

  fp = spdOpenSDBfile(PIDB_DIR, PIDB_CFILE_IFILE, SAPID_O_WRITE);
  for (i = 0; i < get_relation_table_num(cfile_ifile); i++) {
    fprintf(fp, "%d\t%d\n",
        get_relation_table_pobj_id(cfile_ifile, i),
        get_relation_table_cobj_id(cfile_ifile, i));
  }
  spdFclose(fp);
}

static void add_cfile_ifile(SpdObjId cfile_id, SpdObjId ifile_id)
{
  cfile_ifile = add_relation_element(cfile_ifile, cfile_id, ifile_id);
}

static void release_cfile_ifile(void)
{
  cfile_ifile = release_relation_table(cfile_ifile);
}

/**********************************************************************/

static RelTable pfile_file = RELATION_TABLE_EMPTY;

/*
 * write_pfile_file():
 *   analyzer/mkPfileFile/mkPfileFile.c writePfileFile()νϷ
 *   äơSRC/pfile_fileϤ롥
 */

static void write_pfile_file(void)
{
  FILE *fp;
  int i;

  fp = spdOpenSDBfile(SAPID_SRC_DIR, SRC_PFILE_FILE, SAPID_O_WRITE);
  for (i = 0; i < get_relation_table_num(pfile_file); i++) {
    fprintf(fp, "%d %d\n",
        get_relation_table_pobj_id(pfile_file, i),
        get_relation_table_cobj_id(pfile_file, i));
  }
  spdFclose(fp);
}

static void add_pfile_file(SpdObjId pfile_id, SpdObjId file_id)
{
  pfile_file = add_relation_element(pfile_file, pfile_id, file_id);
}

static void release_pfile_file(void)
{
  pfile_file = release_relation_table(pfile_file);
}

/**********************************************************************/

typedef struct _system_file {
  SpdObjId file_id;
  SpdConstString file_name;
} SystemFile;

static SystemFile *system_file_table = NULL;
static int system_file_table_size = 0;
static int system_file_num = 0;
#define SYSTEM_FILE_TABLE_BASE    32

static void add_system_file_table(SpdObjId id, SpdConstString name)
{
  if (system_file_table_size <= system_file_num) {
    system_file_table_size += SYSTEM_FILE_TABLE_BASE;
    system_file_table = spdRealloc(system_file_table,
                sizeof(SystemFile) * system_file_table_size);
  }
  system_file_table[system_file_num].file_id = id;
  system_file_table[system_file_num].file_name = spdStrdup(name);
  system_file_num++;
}

static SpdObjId search_system_file_table(SpdString name)
{
  int i;

  for (i = 0; i < system_file_num; i++) {
    if (SPD_STREQ(name, system_file_table[i].file_name)) {
      return system_file_table[i].file_id;
    }
  }
  return SPD_NON_ID;
}

static void release_system_file_table(void)
{
  int i;

  if (system_file_table != NULL) {
    for (i = 0; i < system_file_num; i++) {
      spdFree(system_file_table[i].file_name);
    }
    spdFree(system_file_table);
    system_file_table = NULL;
    system_file_table_size = 0;
    system_file_num = 0;
  }
}

static FILE *script_fp = NULL;
static char *target_SDB = SAPID_GL_SDB_DIR;

static SpdObjId search_identical_file(SpdDBObj obj, SpdObjId old_id)
{
  SpdObjId identical_file_id = SPD_NON_ID;
  SpdString file_name = obj.svalue[SPD_AID_FILE_NAME - SpdObjIValN];

  /* ̾󤻤ϥƥե(ХѥΥե)Τߡ */
  if (smuIsSystemFileName(file_name)) {
    identical_file_id = search_system_file_table(file_name);
    if (identical_file_id == SPD_NON_ID) {
      add_system_file_table(obj.Obj_id, file_name);
    }
  }

  if (identical_file_id == SPD_NON_ID) {
    /*
     * SRCPIDB򥳥ԡ뤿ΥץȤϤ롥
     */
    if (script_fp != NULL) {
      fprintf(script_fp, "cp %s/%s/%d %s/%s/%d\n",
          spdGetSDBdir(), SAPID_SRC_DIR, old_id,
          target_SDB, SAPID_SRC_DIR, obj.Obj_id);
      if (spdConvFileIdItoC(old_id) != SPD_NON_ID) {
    fprintf(script_fp, "cp %s/%s/%d%s %s/%s/%d%s\n",
        spdGetSDBdir(), PIDB_DIR, old_id, SPD_INFO_SUFFIX,
        target_SDB, PIDB_DIR, obj.Obj_id, SPD_INFO_SUFFIX);
    fprintf(script_fp, "cp %s/%s/%d%s %s/%s/%d%s\n",
        spdGetSDBdir(), PIDB_DIR, old_id, PIDB_EXPAND_SUFFIX,
        target_SDB, PIDB_DIR, obj.Obj_id, PIDB_EXPAND_SUFFIX);
    fprintf(script_fp, "cp %s/%s/%d%s %s/%s/%d%s\n",
        spdGetSDBdir(), PIDB_DIR, old_id, PIDB_DEFINE_SUFFIX,
        target_SDB, PIDB_DIR, obj.Obj_id, PIDB_DEFINE_SUFFIX);
    fprintf(script_fp, "cp %s/%s/%d%s %s/%s/%d%s\n",
        spdGetSDBdir(), PIDB_DIR, old_id, PIDB_IFEXPR_SUFFIX,
        target_SDB, PIDB_DIR, obj.Obj_id, PIDB_IFEXPR_SUFFIX);
      }
    }
  }

  return identical_file_id;
}

/**********************************************************************/

static void count_SDB_objs(int SDB_num, char *SDB_dirs[])
{
  SpdInfo *info;
  int i;

  for (i = 0; i < SDB_num; i++) {
    spdSetSDBdir(SDB_dirs[i]);
    info = spdOpenSDBinfo();
    count_objects_num(info, SPD_MID_I_MODEL);
    count_objects_num(info, SPD_MID_P_MODEL);
    spdCloseInfo(info);
  }
}

static void load_SDB(char *SDB_dir)
{
  SpdInfo *info;
  MergePair merge_pair;
  SpdObjId file_id, cfile_id, pfile_id, file_id_tmp;
  int id_offset[SPD_LAST_RID + 1];
  static SpdObjId max_id[SPD_LAST_RID + 1];
  static SpdBoolean is_first_SDB = TRUE;

  if (is_first_SDB) {
    initialize_max_id(max_id, SPD_MID_I_MODEL);
    initialize_max_id(max_id, SPD_MID_P_MODEL);
    is_first_SDB = FALSE;
  }

  initialize_id_offset(id_offset, max_id, SPD_MID_I_MODEL);
  initialize_id_offset(id_offset, max_id, SPD_MID_P_MODEL);

  spdSetSDBdir(SDB_dir);

  info = spdOpenSDBinfo();

  set_read_sdb_format(info);

  merge_pair = create_merge_pair();

  read_objects(NULL, info, SPD_MID_I_MODEL, merge_pair, id_offset, max_id);
  read_objects(NULL, info, SPD_MID_P_MODEL, merge_pair, id_offset, max_id);
  read_relations(NULL, info, SPD_MID_I_MODEL, merge_pair, id_offset);
  read_relations(NULL, info, SPD_MID_P_MODEL, merge_pair, id_offset);

  /*
   * PIDB/cfile_ifile ɤ߹ߡ
   */

  for (file_id = SPD_OBJ_ID_OFFSET(SPD_CID_FILE) + id_offset[SPD_CID_FILE];
       file_id <= max_id[SPD_CID_FILE]; file_id++) {
    cfile_id = spdConvFileIdItoC(file_id - id_offset[SPD_CID_FILE]);
    if (cfile_id != SPD_NON_ID) {
      cfile_id = cfile_id + id_offset[SPD_CID_FILE];
      file_id_tmp = search_merge_pair_recursive(merge_pair, cfile_id);
      if (file_id_tmp != SPD_NON_ID) {
    cfile_id = file_id_tmp;
      }
      add_cfile_ifile(cfile_id, file_id);
    }
  }

  /*
   * SRC/pfile_file ɤ߹ߡ
   */

  for (pfile_id = SPD_OBJ_ID_OFFSET(SPD_CID_PFILE) + id_offset[SPD_CID_PFILE];
       pfile_id <= max_id[SPD_CID_PFILE]; pfile_id++) {
    file_id = spdConvIdPfileToFile(pfile_id - id_offset[SPD_CID_PFILE]);
    if (file_id != SPD_NON_ID) {
      file_id = file_id + id_offset[SPD_CID_FILE];
      file_id_tmp = search_merge_pair_recursive(merge_pair, file_id);
      if (file_id_tmp != SPD_NON_ID) {
    file_id = file_id_tmp;
      }
      add_pfile_file(pfile_id, file_id);
    }
  }

  release_merge_pair(merge_pair);

  spdCloseInfo(info);
}

/*
 * write_SDB_info():
 *   SDBϤˡinfoեϤ롥
 */

static void write_SDB_info(SpdInfo *info)
{
  spdSetInfoInt(info, SPD_MKSTR(SapidVersion), SapidVersion);
  spdSetInfoInt(info, SPD_MKSTR(SapidRevision), SapidRevision);
  spdSetInfoInt(info, SPD_MKSTR(SapidPatchLevel), SapidPatchLevel);
  spdSetInfoInt(info, SPD_MKSTR(SapidSDBVersion), SapidSDBVersion);
  spdSetInfoInt(info, SPD_MKSTR(SapidPIDBVersion), SapidPIDBVersion);
  spdSetInfoString(info, "SDB4", "TRUE");
  spdSetInfoString(info, "BINARY_FORMAT_SDB",
           (spdGetWriteSDBformat() == SAPID_BINARY_FORMAT_SDB ?
            "TRUE" : "FALSE"));
  spdSetInfoString(info, "RELATIVE_OFFSET", "TRUE");
}

static void write_SDB(void)
{
  SpdInfo *info;

  info = spdOpenSDBinfo();

  write_SDB_info(info);

  write_objects(info, SPD_MID_I_MODEL);
  write_objects(info, SPD_MID_P_MODEL);
  write_relations(info, SPD_MID_I_MODEL);
  write_relations(info, SPD_MID_P_MODEL);

  spdCloseInfo(info);
}

/**********************************************************************/

#define CPSRC_SCRIPT_FILE "cp_src.sh"

static SpdConstString program_name;    /* This program name. */
static char *script_name = CPSRC_SCRIPT_FILE;

static void print_usage(void)
{
  static const char *message =
    "usage: %s [-b][-d SDB_dir][-s script_name] SDBs...\n";

  spdErrorExit(message, program_name);
}

static void parse_option(int argc, char *argv[])
{
  extern char *optarg;
  int ch;

  while ((ch = getopt(argc, argv, "bd:s:")) != -1) {
    switch (ch) {
    case 'b':
      spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
      break;
    case 'd':
      target_SDB = optarg;
      break;
    case 's':
      script_name = optarg;
      break;
    default:
      print_usage();
    }
  }
}

int main(int argc, char *argv[])
{
  extern int optind;
  int i;

  program_name = spdBasename(argv[0]);
  spdInitialize(program_name);

  parse_option(argc, argv);
  argc -= optind;
  argv += optind;
  if (argc < 1) {
    print_usage();
  }

  spdUseModel(&I_model);
  spdUseModel(&P_model);

  count_SDB_objs(argc, argv);
  allocate_db();

  script_fp = spdFopen(script_name, SPD_FOPEN_WT_MODE);

  initialize_search_functions();
  set_search_obj_function(SPD_CID_FILE, search_identical_file);
  for (i = 0; i < argc; i++) {
    load_SDB(argv[i]);
  }

  release_system_file_table();

  spdSetSDBdir(target_SDB);
  write_SDB();

  write_cfile_ifile();
  release_cfile_ifile();
  write_pfile_file();
  release_pfile_file();

  release_db();

  return EXIT_SUCCESS;
}
