/* 
 * Program:             $RCSfile: spdCpSrc.c,v $  $Revision: 30.26 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohhashi  1999/03/13
 *                      S.Yamamoto  1999/03/13
 * 
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  1999 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdCpSrc.c,v 30.26 2003/05/29 04:19:51 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if !defined(SPD_COPY_DOCUMENT)
#define TARGET_CLASS_NAME   "file"
#else
#define TARGET_CLASS_NAME   "docfile"
#endif

static void printUsage(SpdConstString cmdName);

/*
 * Print names and ids of all files to standard output stream.
 */

static void outputFileIds(void)
{
    int             i;
    SpdObjIdArray   files;

    files = spdGetObjIdArray(TARGET_CLASS_NAME, NULL, NULL);
    for (i = 0; i < files.size; i++) {
        printf("%s\n%d\n", spdGetName(files.id[i]), files.id[i]);
    }
    spdFreeObjIdArray(files);
}

/*
 * Analyzing options.
 */

static void parseOption(int argc, char *argv[], SpdConstString cmdName)
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB */
            spdSetSDBdir(optarg);
            break;
          default:
            spdWarning("'%c': Invalid option.\n", ch);
            printUsage(cmdName);
            break;
        }
    }
}

int main(int argc, char *argv[])
{
    SpdConstString program_name;
    SpdModelId  modelId;
    SpdDBId     dbId;

    /* Initialization. */
    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parseOption(argc, argv, program_name);

#if !defined(SPD_COPY_DOCUMENT)
    modelId = spdUseModel(&I_model);
#else
    modelId = spdUseModel(&D_model);
#endif

    dbId = spdOpenSDB();
    spdReadDB(modelId, dbId);
    outputFileIds();
    return (EXIT_SUCCESS);
}

/*
 * Printing usage.
 */

static void printUsage(SpdConstString cmdName)
{
    SpdConstString usageMessage =
        "\n"
        "Name\n"
        "   %s\n"
        "SYNOPSIS\n"
        "   %s [options] programName\n"
        "DESCRIPTION\n"
        "   -d SDBdirectory: \n";

    spdErrorExit(usageMessage, cmdName, cmdName);
}
