/* 
 * Program:             $RCSfile: spdCpSrc3.c,v $  $Revision: 40.3 $
 * 
 * Purpose:
 * 
 * Author:              S.Yamamoto  2003/04/03
 * 
 * (C) Copyright:       S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdCpSrc3.c,v 40.3 2003/06/04 10:22:22 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#if !defined(SPD_COPY_DOCUMENT)
#define TARGET_CLASS_NAME   "file"
#else
#define TARGET_CLASS_NAME   "docfile"
#endif

static void printUsage(SpdConstString cmdName);

static void copyFiles(void)
{
    int             i;
    SpdObjIdArray   files;

    files = spdGetObjIdArray(TARGET_CLASS_NAME, NULL, NULL);
    for (i = 0; i < files.size; i++) {
        char        src[FILENAME_MAX + 1];      
        FILE        *in, *out;
        int         c;

        /* spdMessage("\"%s\" %d.\n", spdGetName(files.id[i]), files.id[i]); */
        if (spdIsFileExtension(spdGetName(files.id[i]), ".c") ||
            spdIsFileExtension(spdGetName(files.id[i]), ".i")) {
            sprintf(src, "%s%s%s%s%d",
                    spdGetSDBdir(), SPD_PATH_SEP_STR,
                    SPD_TMP_SRC_DIR, SPD_PATH_SEP_STR,
                    files.id[i]);
            in = spdFopen(src, SPD_FOPEN_RB_MODE);
        } else {
            in = spdFopen(spdGetName(files.id[i]), SPD_FOPEN_RB_MODE);
        }
        out = spdOpenSourceById(files.id[i], SAPID_O_WRITE);
        while ((c = fgetc(in)) != EOF) {
            fputc(c, out);
        }
        spdFclose(out);
        spdFclose(in);
    }
    spdFreeObjIdArray(files);
}

/*
 * Analyzing options.
 */

static void parseOption(int argc, char *argv[], SpdConstString cmdName)
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB */
            spdSetSDBdir(optarg);
            break;
          default:
            spdWarning("'%c': Invalid option.\n", ch);
            printUsage(cmdName);
            break;
        }
    }
}

int main(int argc, char *argv[])
{
    SpdConstString program_name;
    SpdModelId  modelId;
    SpdDBId     dbId;

    /* Initialization. */
    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parseOption(argc, argv, program_name);

#if !defined(SPD_COPY_DOCUMENT)
    modelId = spdUseModel(&I_model);
#else
    modelId = spdUseModel(&D_model);
#endif

    dbId = spdOpenSDB();
    spdReadDB(modelId, dbId);
    copyFiles();
    return (EXIT_SUCCESS);
}

/*
 * Printing usage.
 */

static void printUsage(SpdConstString cmdName)
{
    const SpdString usageMessage =
        "\n"
        "Name\n"
        "   %s\n"
        "SYNOPSIS\n"
        "   %s [options] programName\n"
        "DESCRIPTION\n"
        "   -d SDBdirectory: \n";

    spdErrorExit(usageMessage, cmdName, cmdName);
}
