/* 
 * Program:             $RCSfile: spdGetObjId.c,v $  $Revision: 32.4 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu  1999/08/29
 *                      H.Ohhashi  1999/08/29
 *                      S.Yamamoto  1999/08/29
 * 
 * (C) Copyright:       N.Fukuyasu, H.Ohhashi and S.Yamamoto  1999 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdGetObjId.c,v 32.4 2003/06/28 15:32:23 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdConstString program_name;

/*
 * Default options.
 */

static SpdBoolean   flag_opt_a = FALSE;     /* Not add object. */ 
static SpdBoolean   flag_opt_b = FALSE;     /* Text mode. */
static SpdBoolean   flag_opt_s = FALSE;     /* Output object id. */

static SpdString    class_name = NULL;      /* Class name. */
#define DEFAULT_CLASS_NAME    "file"

static char     *usage = 
    "USAGE\n"
    "    %s [-abs][-c className][-d SDB-dir] name\n"
    "DESCRIPTION\n"
    "    Outputs object id of specified name if it is found in SDB.\n"
    "    Outputs SPD_NON_ID otherwise.\n"
    "WARNINGS\n"
    "    This program doesn't output warning for invalid arguments.\n"; 

static void print_usage(void)
{
    spdErrorExit(usage, program_name);
}

static void free_class_name(void)
{
    if (class_name != NULL) {
        spdFree(class_name);
        class_name = NULL;
    }
}

static void set_class_name(SpdString name)
{
    static SpdBoolean    is_first_call = TRUE;

    if (is_first_call) {
        atexit(free_class_name);
        is_first_call = FALSE;
    }
    free_class_name();
    if (name != NULL && !SPD_STREQ(name, "file") &&
        !SPD_STREQ(name, "program") && !SPD_STREQ(name, "library")) {
        spdWarning("Invalid class \"%s\".\n", name);
        name = NULL;
    }
    class_name = spdStrdup(name == NULL ? DEFAULT_CLASS_NAME : name);
}

static SpdString get_class_name(void)
{
    if (class_name == NULL) {
        set_class_name(NULL);
    }
    return (class_name);
}

int main(int argc, char *argv[])
{
    int         c;
    SpdObjId    obj_id;
    int         ret = -1;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    while ((c = getopt(argc, argv, "abc:d:s")) != EOF) {
        switch (c) {
          case 'a':                         /* Add object. */
            flag_opt_a = TRUE;
            break;
          case 'b':                         /* Binary format. */
            flag_opt_b = TRUE;
            break;
          case 'c':                         /* Class name. */
            set_class_name(optarg);
            break;
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 's':                         /* Silent mode. */
            flag_opt_s = TRUE;
            break;
          default:
            print_usage();
            break;
        }
    }
    argc -= optind;
    argv += optind;

    if (argc != 1) {
        print_usage();
    }

    if ((obj_id = spdSearchGlobalObj(get_class_name(), argv[0]))
        != SPD_NON_ID) {
        ret = obj_id;
    } else if (flag_opt_a) {
        /* The object isn't found. */ 
        ret = spdCreateGlobalObj(get_class_name(), argv[0]);
        if (flag_opt_b) {
            spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
        }
        spdWriteGlobalClass(get_class_name());
    }
    if (!flag_opt_s) {
        spdOutputForShell("%d", ret);
    }
    return (EXIT_SUCCESS);
}
