/*
 * Program:             $RCSfile: src_table.c,v $  $Revision: 30.16 $
 *
 * Purpose:
 *
 * Author:              H.Ohhashi  1999/06/30
 *                      S.Yamamoto  1997/06/30
 *
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  1999 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: src_table.c,v 30.16 2003/06/28 15:32:24 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/write_sdb_sub.h>
#include "ISDB.h"
#include "sdblink4c.h"

#define  SEARCH_BY_L_FILE   0
#define  SEARCH_BY_R_FILE   1

/*
 * ե򥳥ԡ뤿Υǡ¤ʤɡ
 * Added by H.Ohhashi 1999/06/30. 
 */

typedef struct {
    int           old_id;
    int           new_id; 
    int           is_ifile; /* .iե뤫ɤ */
} SrcItem;

typedef struct {
    int           size;
    SrcItem       *elem;
} SrcArray;

typedef struct {
    char          old_path[FILENAME_MAX + 1];
    SrcArray      src_array;
} RepItem;

typedef struct {
    int          size;
    RepItem      *elem;
} RepArray;

/* cfile_ifile, pfile_file򥳥ԡ뤿Υơ֥ */
typedef struct {
    SpdId    l_fileId;
    SpdId    r_fileId;
} SrcIdTblElem;

typedef struct {
    int         size;
    SrcIdTblElem    *elem;
} SrcIdTbl;

static RepArray         rep_array;
static SrcIdTbl c_i_tbl = {0, NULL};
static SrcIdTbl pf_f_tbl = {0, NULL};

/*
 * fileIdϿƤcfile_ifileź֤
 * modeˤꡤcfileõΤifileõΤꤹ롥
 */

static int lookUpSrcIdTbl(SpdId fileId, int mode, SrcIdTbl *tbl)
{
    int i;
    
    /* SPD_NON_IDξ -1֤ */
    if (fileId == SPD_NON_ID) return (-1);

    if (mode == SEARCH_BY_R_FILE) {
        for (i = 0; i < tbl->size; i++) {
            if (fileId == tbl->elem[i].r_fileId) return i;
        }
    } else if (mode == SEARCH_BY_L_FILE) {
        for (i = 0; i < tbl->size; i++) {
            if (fileId == tbl->elem[i].l_fileId) return i;
        }
    }
    return (-1);
}

/*
 * cfile_ifile뤿Υơ֥˿IDϿ롥
 */

static void addSrcIdTbl(SpdId ifileId, SpdId cfileId)
{
    int tbl_num;

    if ((tbl_num = lookUpSrcIdTbl(ifileId, SEARCH_BY_R_FILE,
                                  &c_i_tbl)) != -1) {
        c_i_tbl.elem[tbl_num].l_fileId = cfileId;
    } else if ((tbl_num = lookUpSrcIdTbl(cfileId, SEARCH_BY_L_FILE,
                                         &c_i_tbl)) != -1) {
        c_i_tbl.elem[tbl_num].r_fileId = ifileId;
    } else {
        c_i_tbl.elem = spdRealloc(c_i_tbl.elem,
                                  sizeof(SrcIdTblElem) *
                                  (c_i_tbl.size + 1));
        c_i_tbl.elem[c_i_tbl.size].r_fileId = ifileId;
        c_i_tbl.elem[c_i_tbl.size].l_fileId = cfileId;
        c_i_tbl.size++;
    }
}

/*
 * old_idбnew_idRepArrayõ
 */

static SpdId searchNewId(SpdId old_id)
{
    int i, j;
    SpdString   sdb_dir = spdGetSDBdir();

    for (i = 0; i < rep_array.size; i++) {
        if (!SPD_STREQ(sdb_dir, rep_array.elem[i].old_path)) continue;

        for (j = 0; j < rep_array.elem[i].src_array.size; j++) {
            if (old_id == rep_array.elem[i].src_array.elem[j].old_id)
                return rep_array.elem[i].src_array.elem[j].new_id;
        }
    }

    return SPD_NON_ID;
}

/*
 * Put source fileId.
 */

void put_src_file_id(int rep_no, SpdId file_id, int offset)
{
    static int file_cnt = 0;
    static int rep_cnt = 0;
    int    convId = SPD_NON_ID;
    int    new_convId;

    if (rep_cnt != rep_no) {
        rep_cnt = rep_no;
        file_cnt = 0;
    }
    rep_array.elem[rep_no].src_array.elem[file_cnt].old_id = file_id;
    rep_array.elem[rep_no].src_array.elem[file_cnt].new_id = file_id + offset;

    /* .iΥե뤫Ƚ(PIDB򥳥ԡ뤿) */
    if ((convId = spdConvFileIdItoC(file_id)) != SPD_NON_ID) {
        /* file_id .iξ */
        new_convId = searchNewId(convId);
        addSrcIdTbl(rep_array.elem[rep_no].src_array.elem[file_cnt].new_id, new_convId);
        rep_array.elem[rep_no].src_array.elem[file_cnt].is_ifile = TRUE;
    } else if ((convId = spdConvFileIdCtoI(file_id)) != SPD_NON_ID) {
        /* file_id .cξ */
        new_convId = searchNewId(convId);
        addSrcIdTbl(new_convId, rep_array.elem[rep_no].src_array.elem[file_cnt].new_id);
        rep_array.elem[rep_no].src_array.elem[file_cnt].is_ifile = FALSE;
    } else {
        rep_array.elem[rep_no].src_array.elem[file_cnt].is_ifile = FALSE;
    }
    file_cnt++;
}

/*
 * Memory allocation & put old_path.
 */

void init_src_array(int rep_no, char *path, int file_num)
{
    int i;

    rep_array.elem[rep_no].src_array.elem = 
        spdMalloc(sizeof(SrcItem) * file_num);
    rep_array.elem[rep_no].src_array.size = file_num;

    /* init src_array. */
    for (i = 0; i < rep_array.elem[rep_no].src_array.size; i++) {
        rep_array.elem[rep_no].src_array.elem[i].old_id = SPD_NON_ID;
        rep_array.elem[rep_no].src_array.elem[i].new_id = SPD_NON_ID;
        rep_array.elem[rep_no].src_array.elem[i].is_ifile = FALSE;
    }

    strcpy(rep_array.elem[rep_no].old_path, path);

}

/*
 * Memory allocation RepArray.
 */

void init_rep_array(int rep_no)
{
    rep_array.elem = spdMalloc(sizeof(RepItem) * rep_no);
    rep_array.size = rep_no;
}

/*
 * Memory allocation pfile_file table.
 */

void initPfileFileTbl(int pfile_num)
{
    pf_f_tbl.elem = spdMalloc(sizeof(SrcIdTblElem) * pfile_num);
    pf_f_tbl.size = pfile_num;
}

/*
 * Add pfile_file table.
 */

void addPfileFileTbl(SpdId old_pfileId, int incAmount)
{
    static int  tbl_num = 0;
    SpdId       old_fileId = spdConvIdPfileToFile(old_pfileId);

    /*
     * pfileΥ֥ȿϥޡѲʤ,
     * ҤȤĤϿƤФ褤
     */
    pf_f_tbl.elem[tbl_num].l_fileId = old_pfileId + incAmount;
    pf_f_tbl.elem[tbl_num].r_fileId = searchNewId(old_fileId);
    tbl_num++;
}

/*
 * Modify new file_id by optimizing.
 */

void modify_src_table(SpdSet modify_set)
{
    int i, j, k, dummy;
    ModifyField *elem;
    int tbl_num;

    for (i = 0; i < modify_set->elemNum; i++) {
        elem = spdGetAElem(modify_set, &dummy);

        /* Simple search. */
        for (j = 0; j < rep_array.size; j++) {
            for (k = 0; k < rep_array.elem[j].src_array.size; k++) {
                if (elem->old_id == rep_array.elem[j].src_array.elem[k].new_id) {
                    rep_array.elem[j].src_array.elem[k].new_id = elem->new_id;
                    break;
                }
            }
        }
    
        /* Modify cfile_ifile table */
        if ((tbl_num = lookUpSrcIdTbl(elem->old_id, 
                                      SEARCH_BY_R_FILE,
                                      &c_i_tbl)) != -1) {
            c_i_tbl.elem[tbl_num].r_fileId = elem->new_id;
        } else if ((tbl_num = lookUpSrcIdTbl(elem->old_id,
                                             SEARCH_BY_L_FILE,
                                             &c_i_tbl)) != -1) {
            c_i_tbl.elem[tbl_num].l_fileId = elem->new_id;
        }

        /* Modify pfile_file table */
        for (tbl_num = 0; tbl_num < pf_f_tbl.size; tbl_num++) {
            if (pf_f_tbl.elem[tbl_num].r_fileId == elem->old_id) {
                pf_f_tbl.elem[tbl_num].r_fileId = elem->new_id;
            }
        }
    }
}

/* 
 * ơ֥ե(cfile_ifile, pfile_file)񤭹ࡥ
 */

static void writeSrcIdTbl(SpdString dir, SpdString tblFile,
                          SrcIdTbl *tbl)
{
    FILE *fp;
    int i;

    fp = spdOpenSDBfile(dir, tblFile, SAPID_O_WRITE);

    for (i = 0; i < tbl->size; i++) {
        fprintf(fp, "%d %d\n", tbl->elem[i].l_fileId,
                tbl->elem[i].r_fileId);
    }
    spdFclose(fp);
}

/*
 * Write copy source file data.
 */

void write_cp_src(void)
{
    FILE    *fp;
    int     i, j;

    fp = spdFopen(script_name, SPD_FOPEN_WT_MODE);
  
    /* 륹ץȤϤ롥 */
    fprintf(fp, "#! /bin/sh -\n");
    fprintf(fp, "if [ ! -d %s/%s ]; then\n", gl_sdb_dir, SAPID_SRC_DIR);
    fprintf(fp, "    mkdir %s/%s\n", gl_sdb_dir, SAPID_SRC_DIR);
    fprintf(fp, "fi\n");

    fprintf(fp, "if [ ! -d %s/%s ]; then\n", gl_sdb_dir, PIDB_DIR);
    fprintf(fp, "    mkdir %s/%s\n", gl_sdb_dir, PIDB_DIR);
    fprintf(fp, "fi\n");

    for (i = 0; i < rep_array.size; i++) {
        for (j = 0; j < rep_array.elem[i].src_array.size; j++) {
            fprintf(fp, "cp %s/%s/%d %s/%s/%d\n", 
                    rep_array.elem[i].old_path, SAPID_SRC_DIR,
                    rep_array.elem[i].src_array.elem[j].old_id,
                    gl_sdb_dir, SAPID_SRC_DIR,
                    rep_array.elem[i].src_array.elem[j].new_id);
            /* Copy PIDB files ("*.{info,define,expand,ifexpr}"). */
            if (rep_array.elem[i].src_array.elem[j].is_ifile) {
                /* Added by S.Yamamoto 2002/05/05. */
                fprintf(fp, "cp %s/%s/%d%s %s/%s/%d%s\n", 
                        rep_array.elem[i].old_path, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].old_id,
                        SPD_INFO_SUFFIX,
                        gl_sdb_dir, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].new_id,
                        SPD_INFO_SUFFIX);
                fprintf(fp, "cp %s/%s/%d%s %s/%s/%d%s\n", 
                        rep_array.elem[i].old_path, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].old_id,
                        PIDB_EXPAND_SUFFIX,
                        gl_sdb_dir, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].new_id,
                        PIDB_EXPAND_SUFFIX);
                fprintf(fp, "cp %s/%s/%d%s %s/%s/%d%s\n", 
                        rep_array.elem[i].old_path, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].old_id,
                        PIDB_DEFINE_SUFFIX,
                        gl_sdb_dir, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].new_id,
                        PIDB_DEFINE_SUFFIX);
                /* Added by S.Yamamoto 2000/03/10. */
                fprintf(fp, "cp %s/%s/%d%s %s/%s/%d%s\n", 
                        rep_array.elem[i].old_path, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].old_id,
                        PIDB_IFEXPR_SUFFIX,
                        gl_sdb_dir, PIDB_DIR,
                        rep_array.elem[i].src_array.elem[j].new_id,
                        PIDB_IFEXPR_SUFFIX);
            }
        }
    }
    /* !! Copy ".isValidSDB" (Ū). */
    /* fprintf(fp, "touch %s/.isValidSDB-1\n", gl_sdb_dir); */
    /* fprintf(fp, "touch %s/.isValidSDB-2\n", gl_sdb_dir); */
  
    spdFclose(fp);

    writeSrcIdTbl(PIDB_DIR, PIDB_CFILE_IFILE, &c_i_tbl);
    writeSrcIdTbl(SAPID_SRC_DIR, SRC_PFILE_FILE, &pf_f_tbl);
}

/*
 * Free RepArray memory.
 */

void free_rep_table(void)
{
    int i;

    for (i = 0; i < rep_array.size; i++) {
        spdFree(rep_array.elem[i].src_array.elem);
    }

    if (rep_array.size)
        spdFree(rep_array.elem);

    if (c_i_tbl.size)
        spdFree(c_i_tbl.elem);

    if (pf_f_tbl.size)
        spdFree(pf_f_tbl.elem);
}
