/* 
 * Program:             $RCSfile: spdMkPgsql.c,v $  $Revision: 40.3 $
 * 
 * Purpose:
 * 
 * Author:              S.Ibuki  2002/10/01
 *                      S.Yamamoto  2002/10/01
 * 
 * (C) Copyright:       S.Ibuki and S.Yamamoto  2002 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
"$Id: spdMkPgsql.c,v 40.3 2003/05/18 08:12:01 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <libpq-fe.h>

static SpdBoolean   flag_s = OFF;
static char         *program_name;              /* This program name. */
static char         *model_name = NULL;

static void parse_option(int argc, char *argv[]);
static void print_usage(char *command_name);

static void dump_model(SpdModelId modelId);
static void dump_table(const char *class_name);

int escapquat(char *);

int main(int argc, char *argv[])
{
  int         i;
  SpdDBId     dbId;
  SpdModelId  mId;
  SpdModelId  ImodelId;
  SpdModelId  PmodelId;
  SpdModelId  AreaModelId;
  SpdModelId  BmodelId;
  SpdModelId  DmodelId;
  SpdModelId  SDAmodelId;
  SpdModelId  StmtModelId;
  SpdModelId  SareaModelId;
  SpdModelId  TypeViewId;
  SpdModelId  CommentViewId;
  SpdModelId  FlowViewId;
  SpdModelId  FCDViewId;
  SpdModelId  JmodelId;

  program_name = spdBasename(argv[0]);
  spdInitialize(program_name);
  parse_option(argc, argv);
  argc -= optind;
  argv += optind;

  ImodelId = spdUseModel(&I_model);
  PmodelId = spdUseModel(&P_model);
  AreaModelId = spdUseModel(&Area_model);
  BmodelId = spdUseModel(&B_model);
  DmodelId = spdUseModel(&D_model);
  SDAmodelId = spdUseModel(&SDA_model);
  StmtModelId = spdUseModel(&Stmt_model);
  SareaModelId = spdUseModel(&SDAarea_model);
  TypeViewId = spdUseModel(&TypeView_model);
  CommentViewId = spdUseModel(&Comment_view);
  FlowViewId = spdUseModel(&Flow_view);
  FCDViewId = spdUseModel(&FCD_view);
  JmodelId = spdUseModel(&J_model);

  dbId = spdOpenSDB();
  /* spdReadDB(SPD_NON_MID, dbId); */
  spdReadDB(ImodelId, dbId);
  spdReadDB(PmodelId, dbId);
  spdReadDB(AreaModelId, dbId);
  spdReadDB(BmodelId, dbId);
  spdReadDB(DmodelId, dbId);
  spdReadDB(SDAmodelId, dbId);
  spdReadDB(StmtModelId, dbId);
  spdReadDB(SareaModelId, dbId);
  spdReadDB(TypeViewId, dbId);
  spdReadDB(CommentViewId, dbId);
  spdReadDB(FlowViewId, dbId);
  spdReadDB(FCDViewId, dbId);
  spdReadDB(JmodelId, dbId);

  /* systemؿȤäƥǡ١ */
  // system("dropdb SDB");
  // system("createdb SDB");

  if (model_name != NULL) {
    dump_model(spdGetModelIdByName(model_name));
  } else if (argc == 0) {
    for (mId = SPD_FIRST_MID; mId <= SPD_LAST_MID; mId++) {
      dump_model(mId);
    }
  } else {
    for (i = 0; i < argc; i++) {
      dump_table(argv[i]);
    }
  }

  // spdCloseSDB(dbId);

  return (EXIT_SUCCESS);
}

static void dump_model(SpdModelId modelId)
{
  int         i;

  /* !! Temporarily added by S.Yamamoto 2002/07/28. */
  if (modelId == SPD_MID_J_MODEL) {
    spdWarning("dump_model(): J-model skipped.\n");
  }

  if (modelId < SPD_FIRST_MID || SPD_LAST_MID < modelId) {
    spdErrorExit("dump_model(): %d: Invalid model id.\n", modelId);
  }
  for (i = spdGetFirstCid(modelId); i <= spdGetLastCid(modelId); i++) {
    dump_table(spdGetClassNameById(i));
  }
  for (i = spdGetFirstRid(modelId); i <= spdGetLastRid(modelId); i++) {
    dump_table(spdGetClassNameById(i));
  }
}

static void dump_table(const char *class_name)
{
  int             attrNum;
  SpdDBType       attr_type;
  SpdObjId        obj_id;
  SpdCursor       *obj_csr;
  SpdClassId      classId;
  SpdAttrId       attrId;
  char            query_string[1024] = "", a[1024] = "",
    b[1024] = "", c[1024] = "", d[1024] = "";

  PGconn        *conn;      /* ǡ١ؤ³ݻ롥 */
  PGresult      *res;       /* 䤤碌η̤ݻ롥 */
  int           i;
  
  static char   text[1024];
   
  if ((classId = spdGetClassIdByName(class_name)) == SPD_NON_ID) {
    spdErrorExit("Invalid class name: \"%s\".\n", class_name);
  }

  conn = PQconnectdb("dbname=SDB");       /* ǡ١³롥 */
  if (PQstatus(conn) == CONNECTION_BAD) {
    spdError("Connection to database failed.\n");
    spdErrorExit(PQerrorMessage(conn));
  }

  if (strcmp(spdGetModelNameById(spdGetModelId(classId)), "J-model")) {
    sprintf(a, "CREATE TABLE \"%s\" (id integer primary key", class_name);
      
    attrNum = spdGetNumOfAttrs(classId);
    for (attrId = 0; attrId < attrNum; attrId++) {
      attr_type = spdGetAttrType(classId, attrId);
      switch (attr_type){
      case SPD_DBT_INT:
      case SPD_DBT_OID:
      case SPD_DBT_ENUM:
        strcpy(text, spdGetAttrNameByNo(classId, attrId));
        escapquat(text);
        sprintf(b,",\"%s\" integer",text);
        break;
      case SPD_DBT_STR:
        sprintf(b,",%s text", spdGetAttrNameByNo(classId, attrId));
        break;
      }
      strcat(c,b);
    }
    strcat(a,c);
    printf("%s);\n", a);
    sprintf(query_string, "%s);\n", a);
     
    res = PQexec(conn, query_string);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
      PQclear(res);
      PQfinish(conn);
      spdErrorExit("CREATE \"%s\" query failed.\n", class_name);
    }
    PQclear(res);
     
    strcpy(a, d);
    strcpy(b, d);
    strcpy(c, d);
     
    sprintf(query_string, "GRANT ALL ON \"%s\" TO PUBLIC;", class_name);

    res = PQexec(conn, query_string);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
        PQclear(res);
        PQfinish(conn);
        spdErrorExit("GRANT \"%s\" query failed.\n", class_name);
      }
    PQclear(res);
  }

  /*
   * Output attribute values.
   */
  obj_csr = spdGetObjIdInit(class_name, NULL);
   
  while ((obj_id = spdGetObjId(obj_csr)) != SPD_NON_ID) {
     
    sprintf(a,"INSERT INTO \"%s\" VALUES (%d", class_name, obj_id);
    for (attrId = 0; attrId < attrNum; attrId++) {
      attr_type = spdGetAttrType(classId, attrId);
      switch (attr_type) {
      case SPD_DBT_INT:
      case SPD_DBT_OID:
        sprintf(b,",%d", spdGetAttrIntVal(obj_id, attrId));
        break;
      case SPD_DBT_ENUM:
        if (flag_s) {
          int             val = spdGetAttrIntVal(obj_id, attrId);
          SpdConstString  str = spdGetEnumStrVal(classId,
                                                 attrId, val);
           
          sprintf(b,"%d (%s)\t", val, (str == NULL) ? "*** UNKNOWN ***" : str);
        } else {
          sprintf(b,",%d", spdGetAttrIntVal(obj_id, attrId));
        }
        break;
      case SPD_DBT_STR:
    escapquat(spdGetAttrStrVal(obj_id, attrId));
        sprintf(b,",'%s'", spdGetAttrStrVal(obj_id, attrId));
        break;
      case SPD_DBT_NON:
        /* Is it right? */
        break;
      }
      strcat(c, b);
    }
    strcat(a, c);
    sprintf(query_string, "%s);\n", a);
     
    res = PQexec(conn, query_string);
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
        PQclear(res);
        PQfinish(conn);
        spdErrorExit("%s query failed.\n", query_string);
        exit(1);
      }
    PQclear(res);
    
    strcpy(a, d);
    strcpy(b, d);
    strcpy(c, d);
  }
  
  spdFreeCursor(obj_csr);
  
  PQfinish(conn);   /* ǡ١Ȥ³롥 */ 
}

static void parse_option(int argc, char *argv[])
{
  int     ch;

  while ((ch = getopt(argc, argv, "d:m:s")) != -1) {
    switch (ch) {
    case 'd':                         /* Directory for SDB. */
      spdSetSDBdir(optarg);
      break;
    case 'm':                         /* Model name. */
      model_name = optarg;
      break;
    case 's':                         /* Symbolic representation. */
      flag_s = ON;
      break;
    default:
      spdError("'%c': Invalid option.\n", ch);
      print_usage(program_name);
      break;
    }
  }
}

/*
 * Printing usage.
 */

static void print_usage(char *command_name)
{
  const char  *usageMessage =
    "\n"
    "NAME\n"
    "    %s\n"
    "SYNOPSIS\n"
    "    %s [options] [className...]\n"
    "DESCRIPTION\n"
    "    -d SDB-directory: Directory for SDB (\"SDB\").\n"
    "    -m modelName: Specify model name.\n"
    "    -s: Symbolic representation.\n"
    "VERSION\n"
    "    %d.%d.%d\n";

  spdErrorExit(usageMessage,
               command_name, command_name,
               SapidVersion, SapidRevision, SapidPatchLevel);
}

int escapquat(char *text)
{
  int   i,n;
  char  buf[1024], *p, a[1] = "'", b[1] = "\\";

  n = strlen(text);
  p = text;
  while (*p!='\0') {
    if (*p == a[0]) {
      *p = '\0';
      strcpy(buf, p + 1);
      strcat(text, "''");
      strcat(text, buf);
      p++;
    } else if (*p == b[0]) {
      *p = '\0';
      strcpy(buf, p + 1);
      strcat(text, "\\\\");
      strcat(text, buf);
      p++;
    }
    p++;
  }
}
