/*
 * Program:             $RCSfile: PIDB_stream.c,v $  $Revision: 30.58 $
 *
 * Purpose:             եط
 *
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01 
 *
 * (C) Copyright:       T.Mimura and S.Yamamoto  1994 - 2003
 *                      This file is a product of the project Sapid. 
 */

#ifndef lint
static char rcsid[] =
    "$Id: PIDB_stream.c,v 30.58 2003/07/13 09:14:59 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DEBUG_INITIALIZE_PIDB */
/* #define DEBUG_READ_IFEXPR_INFO */
    
#if defined(lint)
#define DW_INITIALIZE_PIDB  printf      /* Just for cheking. */
#elif defined(DEBUG_INITIALIZE_PIDB)
#define DW_INITIALIZE_PIDB  spdMessage
#else
#define DW_INITIALIZE_PIDB  (void)
#endif

/* Temporarily added by S.Yamamoto 2003/05/21. */
#define SPD_READ_INT_ATTR(fp)           spdReadIntAttr((fp))

static SpdString        CPPinfoSuffix[] = {
    PIDB_DEFINE_SUFFIX,
    PIDB_EXPAND_SUFFIX,
    PIDB_IFEXPR_SUFFIX,
    NULL
};

static FILE     *CPPinfoFile[CPPinfoN];

static void     openPidb(SpdObjId ifileId);
static void     closePidb(void);

static void     releasePidb(void);
static void     releasePidbDefine(void);
static void     releasePidbExpand(void);
static void     releasePidbIfexpr(void);

static void     pidbReadLineSeparator(FILE *fp);
static void     pidbReadTupleSeparator(FILE *fp);

/*
 * Open PIDB files.
 * ifileId: ե(".i")objectId
 */

static void openPidb(SpdObjId ifileId)
{
    char        fname[FILENAME_MAX + 1];
    int         i;

    DW_INITIALIZE_PIDB("openPidb(): \"%d\".\n", ifileId);
    assert(defineInfo == NULL);

    for (i = 0; i < CPPinfoN; i++) {
        sprintf(fname, "%d%s", ifileId, CPPinfoSuffix[i]);
        CPPinfoFile[i] = spdOpenSDBfile(PIDB_DIR, fname, SAPID_O_READ);
    }
}

/*
 * Close PIDB files.
 */

static void closePidb(void)
{
    int     i;
    
    for (i = 0; i < CPPinfoN; i++) {
        if (CPPinfoFile[i] != NULL) {
            spdFclose(CPPinfoFile[i]);
            CPPinfoFile[i] = NULL;
        }
    }
}

/******************************************************************************/

static void pidbReadLineSeparator(FILE *fp)
{
    int     c;

    c = getc(fp);
    if (!IS_PIDB_LINE_SEPARATOR(c)) {
        spdErrorExit("pidbReadLineSeparator(): '%c': ftell %d.\n",
                     c, ftell(fp));
    }
}

static void pidbReadTupleSeparator(FILE *fp)
{
    int     c;

    c = getc(fp);
    if (!IS_PIDB_TUPLE_SEPARATOR(c)) {
        spdErrorExit("pidbReadTupleSeparator(): '%c': ftell %d.\n",
                     c, ftell(fp));
    }
    pidbReadLineSeparator(fp);
}

void pidbWriteLineSeparator(FILE *fp)
{
    putc(PIDB_LINE_SEPARATOR, fp);
}

void pidbWriteTupleSeparator(FILE *fp)
{
    putc(PIDB_TUPLE_SEPARATOR, fp);
    pidbWriteLineSeparator(fp);
}

/******************************************************************************/

/*
 * ֥ޥפɤ߹ࡥ
 */

static int pidbReadDefineInfo(void)
{
    int             i;
    FILE            *fp;
    static char     *INVALID_DEFINE_FILE = "PIDB: Invalid DEFINEinfo file.\n";
    
    assert(defineInfo == NULL);
    if (defineInfoN == 0) {
        return (0);
    }
    defineInfo = spdMalloc(sizeof(DEFINEinfo) * defineInfoN);
    
    fp = CPPinfoFile[CPPinfoDEFINE];
    for (i = 0; i < defineInfoN; i++) {
        int             id, j;
        DEFINEinfo     *def;
        
        /* ޥμֹ桥 */
        id = spdReadIntAttr(fp);
        pidbReadLineSeparator(fp);
        id--;                               /* id0롥 */
        def = &(defineInfo[id]);
        def->ID = id + 1;                   /* ID1顥 */

        /* ֡ */
        def->pos.type = spdReadIntAttr(fp);
        switch (def->pos.type) {
          case DEFINE_INITIALIZATION:
            break;
          case DEFINE_D_OPTION:
            break;
          case DEFINE_FILE:
            def->pos.fname = spdStrdup(spdReadStrAttr(fp));
            def->pos.start = SPD_READ_INT_ATTR(fp);
            def->pos.end = SPD_READ_INT_ATTR(fp);
            break;
          default:
            spdErrorExit(INVALID_DEFINE_FILE);
        }
        pidbReadLineSeparator(fp);
        
        /* ޥ̾ */
        def->name = spdStrdup(spdReadStrAttr(fp));
        pidbReadLineSeparator(fp);
        
        /* θĿ */
        def->narg = spdReadIntAttr(fp);
        pidbReadLineSeparator(fp);

        /* ִʸ */
        def->expansion = spdStrdup(spdReadStrAttr(fp));
        pidbReadLineSeparator(fp);

        if (def->narg > 0) {
            /* Ÿκݤ˰Ȥ */
           def->nref = spdReadIntAttr(fp);
           pidbReadLineSeparator(fp);
            
            if (def->nref > 0) {
                def->ref = spdMalloc(sizeof(ArgReference) * def->nref);
                for (j = 0; j < def->nref; j++) {
                    ArgReference   *r;
                    
                    r = &((def->ref)[j]);

                    /* ֡ */
                    r->nchar = spdReadIntAttr(fp);
                    r->argn = spdReadIntAttr(fp);
                    /* "#"ĤƤ뤫 */
                    r->stringify = spdReadIntAttr(fp);
                    /* "##"ˤ뤫 */
                    r->raw_before = spdReadIntAttr(fp);
                    /* "##"ˤ뤫 */
                    r->raw_after = spdReadIntAttr(fp);
                    pidbReadLineSeparator(fp);
                }
            }
            def->arg = spdMalloc(sizeof(char *) * def->narg);
            for (j = 0; j < def->narg; j++) {
                /* ̾ */
                (def->arg)[j] = spdStrdup(spdReadStrAttr(fp));
                pidbReadLineSeparator(fp);
            }
        } else {
            def->nref = 0;
        }

        /* ޥˤϥեΰ֡ */
        def->output = SPD_READ_INT_ATTR(fp);
        pidbReadLineSeparator(fp);

        pidbReadTupleSeparator(fp);        
    }
    /* Just checking file length. */
    {
        int     c;

        if ((c = getc(fp)) != EOF) {
            spdErrorExit("'%c': Invalid defineInfo length.\n", c);
        }
    }
    return (defineInfoN);
}

/*
 * ֥ޥŸפɤ߹ࡥ
 * աɬ˥ޥɤ߹Ǥȡ 
 */

static int pidbReadExpandInfo(void)
{
    int             i;
    FILE           *fp;
    
    if (expandInfoN == 0) {
        return (0);
    }
    expandInfo = spdMalloc(sizeof(EXPANDinfo) * expandInfoN);

    fp = CPPinfoFile[CPPinfoEXPAND];
    for (i = 0; i < expandInfoN; i++) {
        int             id, mid, j;
        DEFINEinfo      *mac;
        EXPANDinfo      *exp;
        
        /* Id. of macro expansion. */    
        id = spdReadIntAttr(fp);
        pidbReadLineSeparator(fp);
        id--;
        exp = &(expandInfo[id]);
        exp->ID = id + 1;
        
        /* Id. of macro. */
        mid = spdReadIntAttr(fp);
        pidbReadLineSeparator(fp);
        mid--;
        exp->macID = mid;
        mac = &(defineInfo[mid]);
        exp->mac = mac;
        
        /* ïΥޥƤΤ */
        exp->input.type = spdReadIntAttr(fp);
        switch (exp->input.type) {
          case CALL_FILE:
            /* 0: ե롥 */
            exp->input.fname = spdStrdup(spdReadStrAttr(fp));

#ifdef PIDB_USE_FILENAME_INDEX
            exp->input.fname_idx = pidbConvFileId(exp->input.fname);
#endif

            break;
          case CALL_MACROEXPANSION:
            /* 1: ޥŸ */
            exp->input.macexpID = spdReadIntAttr(fp) - 1;
            exp->input.macexp = &(expandInfo[exp->input.macexpID]);
            break;
          case CALL_MACROARG:
            /* 2: ޥΰ */
            exp->input.macexpID = spdReadIntAttr(fp) - 1;
            exp->input.macexp = &(expandInfo[exp->input.macexpID]);
            exp->input.argn = spdReadIntAttr(fp);
            break;
        }
        pidbReadLineSeparator(fp);
        
        /* ޥƤӽФΰ(ϰ) */
        exp->input.start = SPD_READ_INT_ATTR(fp);
        exp->input.end = SPD_READ_INT_ATTR(fp);
        pidbReadLineSeparator(fp);
        
        /* ˴ؤǡ */
        if (0 < exp->mac->narg) {
            exp->arg = spdMalloc(sizeof(MacroArgInfo) * exp->mac->narg);
            for (j = 0; j < exp->mac->narg; j++) {
                MacroArgInfo   *a = &((exp->arg)[j]);
                
                a->start = SPD_READ_INT_ATTR(fp);
                a->end = SPD_READ_INT_ATTR(fp);
                a->raw_length = spdReadIntAttr(fp);
                a->expanded_length = spdReadIntAttr(fp);
                a->stringified_length = spdReadIntAttr(fp);
                a->trailing_space = spdReadIntAttr(fp);
                a->following_space = spdReadIntAttr(fp);
                pidbReadLineSeparator(fp);
            }
        }
        
        /* ޥŸ̤ΰ֡ */
        exp->output.type = spdReadIntAttr(fp);
        switch (exp->output.type) {
          case EXPAND_FILE:                 /* ץץåϤ */
            exp->output.fname = spdStrdup(spdReadStrAttr(fp));

#ifdef PIDB_USE_FILENAME_INDEX
            exp->output.fname_idx = pidbConvFileId(exp->output.fname);
#endif

            break;
          case EXPAND_MACROARG:
            exp->output.macexpID = spdReadIntAttr(fp) - 1;
            exp->output.macexp = &(expandInfo[exp->output.macexpID]);
            exp->output.argn = spdReadIntAttr(fp);
            break;
        }
        pidbReadLineSeparator(fp);
        
        /* ޥŸ̤ΰ(ϰ) */
        exp->output.start = spdReadIntAttr(fp);
        exp->output.end = SPD_READ_INT_ATTR(fp);
        pidbReadLineSeparator(fp);
        
        pidbReadTupleSeparator(fp);
    }
    /* Just checking file length. */
    {
        int     c;

        if ((c = getc(fp)) != EOF) {
            spdErrorExit("'%c': Invalid expandInfo length.\n", c);
        }
    }
    return (expandInfoN);
}

static int pidbReadIfexprInfo(void)
{
    FILE *fp;
    int             i;

    if (ifexprInfoN == 0) {
        return (0);
    }
    ifexprInfo = spdMalloc(sizeof(IFEXPRinfo) * ifexprInfoN);
    
    fp = CPPinfoFile[CPPinfoIFEXPR];
    for (i = 0; i < ifexprInfoN; i++) {
        ifexprInfo[i].cfilename = spdStrdup(spdReadStrAttr(fp));
        ifexprInfo[i].clineno = spdReadIntAttr(fp);
        ifexprInfo[i].ifexpr = spdReadIntAttr(fp);
        ifexprInfo[i].ifilename = spdStrdup(spdReadStrAttr(fp));
        ifexprInfo[i].ilineno = spdReadIntAttr(fp);
        pidbReadLineSeparator(fp);

        pidbReadTupleSeparator(fp);

#ifdef DEBUG_READ_IFEXPR_INFO
        spdMessage("pidbReadIfexprInfo(): %d: %s %d %d %s %d.\n",
                   i, 
                   ifexprInfo[i].cfilename,
                   ifexprInfo[i].clineno,
                   ifexprInfo[i].ifexpr,
                   ifexprInfo[i].ifilename,
                   ifexprInfo[i].ilineno);
#endif /* DEBUG_READ_IFEXPR_INFO */
    }

    /* Added by S.Yamamoto 2000/03/24. */
    return (ifexprInfoN);
}

/******************************************************************************/

static void releasePidb(void)
{
    DW_INITIALIZE_PIDB("releasePidb(): defineInfo %p, defineInfoN %d, expandInfo %p, expandInfoN %d.\n", defineInfo, defineInfoN, expandInfo, expandInfoN);

    /* Release expandInfo[] before defineInfo[] released. */
    releasePidbExpand();
    releasePidbDefine();
    releasePidbIfexpr();
}

static void releasePidbDefine(void)
{
    int     i, j;

    for (i = 0; i < defineInfoN; i++) {
        DEFINEinfo     *def;

        def = &(defineInfo[i]);
        if (def->pos.type == DEFINE_FILE) {
            spdSafetyFree(def->pos.fname);
        }
        spdSafetyFree(def->name);
        spdSafetyFree(def->expansion);
        if (0 < def->narg) {
            if (0 < def->nref) {
                spdSafetyFree(def->ref);
            }
            /* ̾ */
            for (j = 0; j < def->narg; j++) {
                spdSafetyFree((def->arg)[j]);
            }
            spdSafetyFree(def->arg);
        }
    }
    if (defineInfo != NULL) {
        spdFree(defineInfo);
        defineInfo = NULL;
    }
    defineInfoN = 0;
}

static void releasePidbExpand(void)
{
    int     i;

    for (i = 0; i < expandInfoN; i++) {
        EXPANDinfo      *exp;

        exp = &(expandInfo[i]);
        if (exp->input.type == CALL_FILE) {
            spdSafetyFree(exp->input.fname);
        }

        if (0 < exp->mac->narg) {
            spdSafetyFree(exp->arg);
        }
        if (exp->output.type == EXPAND_FILE) {
            spdSafetyFree(exp->output.fname);
        }
    }
    if (expandInfo != NULL) {
        spdFree(expandInfo);
        expandInfo = NULL;
    }
    expandInfoN = 0;
}

static void releasePidbIfexpr(void)
{
    int     i;

    for (i = 0; i < ifexprInfoN; i++) {
        spdSafetyFree(ifexprInfo[i].cfilename);
        spdSafetyFree(ifexprInfo[i].ifilename);
    }
    if (ifexprInfo != NULL) {
        spdFree(ifexprInfo);
        ifexprInfo = NULL;
    }
    ifexprInfoN = 0;
}

/******************************************************************************/

#define HISTORY_MAX         16

/* 
 * Variables history[] and no are initialized in spdInitializePidbCacheSub().
 * 
 */

static struct {
    SpdObjId    ifileId;
    DEFINEinfo  *defineInfo;
    int         defineInfoN;
    EXPANDinfo  *expandInfo;
    int         expandInfoN;
    IFEXPRinfo  *ifexprInfo;
    int         ifexprInfoN;
} history[HISTORY_MAX];

static int          no;                 /* Index of history[]. */

static void spdInitializePidbCacheSub(void);

int spdInitializePidbCache(SpdObjId ifileId)
{
    int                 i;
    static SpdBoolean   is_first_call = TRUE;
    SpdInfo             *info;

    if (is_first_call) {
        spdInitializePidbCacheSub();
        is_first_call = FALSE;
    }

    DW_INITIALIZE_PIDB("spdInitializePidbCache(): no %d, ifileId %d.\n",
                       no, ifileId);

    if (ifileId == SPD_NON_ID) {
        spdErrorExit("spdInitializePidbCache(): Invalid ifileId %d.\n",
                     ifileId);
    }

    for (i = no; i < HISTORY_MAX; i++) {
        if (history[i].ifileId == ifileId) {
            defineInfo = history[i].defineInfo;
            defineInfoN = history[i].defineInfoN;
            expandInfo = history[i].expandInfo;
            expandInfoN = history[i].expandInfoN;
            ifexprInfo = history[i].ifexprInfo;
            ifexprInfoN = history[i].ifexprInfoN;
            DW_INITIALIZE_PIDB("    Hit: %d.\n", i);
            return (0);
        }
    }
    for (i = 0; i < no; i++) {
        if (history[i].ifileId == ifileId) {
            defineInfo = history[i].defineInfo;
            defineInfoN = history[i].defineInfoN;
            expandInfo = history[i].expandInfo;
            expandInfoN = history[i].expandInfoN;
            ifexprInfo = history[i].ifexprInfo;
            ifexprInfoN = history[i].ifexprInfoN;
            DW_INITIALIZE_PIDB("    Hit: %d.\n", i);
            return (0);
        }
    }

    /* no = (no - 1) % HISTORY_MAX; */  /* Variable no should be unsigned. */
    no = (no == 0) ? HISTORY_MAX - 1 : no - 1;
    DW_INITIALIZE_PIDB("    New no %d.\n", no);

    defineInfo = history[no].defineInfo;
    defineInfoN = history[no].defineInfoN;
    expandInfo = history[no].expandInfo;
    expandInfoN = history[no].expandInfoN;
    ifexprInfo = history[no].ifexprInfo;
    ifexprInfoN = history[no].ifexprInfoN;
    releasePidb();

    /* Read number of records. */
    info = spdOpenPIDBinfo(ifileId);
    defineInfoN = spdGetInfoInt(info, "define", -1);
    expandInfoN = spdGetInfoInt(info, "expand", -1);
    ifexprInfoN = spdGetInfoInt(info, "ifexpr", -1);
    spdCloseInfo(info);
    if (defineInfoN < 0 || expandInfoN < 0 || ifexprInfoN < 0) {
        spdError("spdInitializePidbCache(): Invalid PIDB info.\n");
        spdErrorExit("    defineInfoN %d, expandInfoN %d, ifexprInfoN %d.\n",
                     defineInfoN, expandInfoN, ifexprInfoN);
    }
    /* Read all records. */
    openPidb(ifileId);
    pidbReadDefineInfo();
    pidbReadExpandInfo();
    pidbReadIfexprInfo();
    closePidb();

    history[no].ifileId = ifileId;
    history[no].defineInfo = defineInfo;
    history[no].defineInfoN = defineInfoN;
    history[no].expandInfo = expandInfo;
    history[no].expandInfoN = expandInfoN;
    history[no].ifexprInfo = ifexprInfo;
    history[no].ifexprInfoN = ifexprInfoN;

    return (0);
}

static void spdInitializePidbCacheSub(void)
{
    int         i;

    for (i = 0; i < HISTORY_MAX; i++) {
        history[i].ifileId = SPD_NON_ID;
        history[i].defineInfo = NULL;
        history[i].defineInfoN = 0;
        history[i].expandInfo = NULL;
        history[i].expandInfoN = 0;
        history[i].ifexprInfo = NULL;
        history[i].ifexprInfoN = 0;
    }
    no = 0;
}

void spdReleasePidbCache(void)
{
    int         i;

    for (i = 0; i < HISTORY_MAX; i++) {
        defineInfo = history[i].defineInfo;
        defineInfoN = history[i].defineInfoN;
        expandInfo = history[i].expandInfo;
        expandInfoN = history[i].expandInfoN;
        ifexprInfo = history[i].ifexprInfo;
        ifexprInfoN = history[i].ifexprInfoN;
        releasePidb();
    }
    spdInitializePidbCacheSub();
}
