/* 
 * Program:             $RCSfile: lineTbl.c,v $  $Revision: 31.7 $
 * 
 * Purpose:             
 * 
 * Author:              M.Hakamata  1997/05/21
 *                      S.Yamamoto  1995/05/24
 * 
 * (C) Copyright:       M.Hakamata and S.Yamamoto  1997 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: lineTbl.c,v 31.7 2003/06/28 15:31:14 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
/* #include "lineTbl.h" */

/* bsearchӴؿ */

static int comparOffset(const int *a, const int *b)
{
    return ((*a < *b) ? -1 : ((*(b + 1) <= *a) ? 1 : 0));
}

#ifdef SAPID_AR3_KERNEL

/*
 * եγƹԤƬʸ֤˴ؤơ֥롥
 * : Ƭʸ֥ơ֥롥
 */

SpdLineTbl spdCreateLineTbl(char *p, int buflen)
{
    SpdLineTbl  ret;
    char        *head;                  /* First character in a line. */
    char        *start;                 /* First character in input buffer. */
    int         count = 0;
    
    head = start = p;
    ret.offset = NULL;
    while (*p != '\0') {
        if (*p++ == '\n') {                 /* UNIXξ礷ͤƤʤ */
            count++;
            ret.offset = (int *)spdRealloc(ret.offset, 
                                           sizeof(int) * count);
            ret.offset[count - 1] = head - start;
            head = p;
        }
    }
    
    if (p != head) {                        /* ԥɤʤԤξ硥 */
        count++;
        ret.offset = (int *)spdRealloc(ret.offset, sizeof(int) * count);
        ret.offset[count - 1] = head - start;
    }
    
    ret.offset = (int *)spdRealloc(ret.offset, sizeof(int) * (count + 1));
    ret.offset[count] = buflen;             /* Sentinel. */
    ret.linenum = count;                    /* Ԥ1Ϥޤ롥 */
    return (ret);
}

void spdFreeLineTbl(SpdLineTbl tbl)
{
    if (tbl.offset != NULL) {
        spdFree(tbl.offset);
    }
}

#endif /* SAPID_AR3_KERNEL */

/******************************************************************************/

#define HISTORY_MAX     32

static struct {
    SpdObjId    fileId;
    SpdLineTbl  tbl;
} history[HISTORY_MAX];

static int          no;                 /* Index of history[]. */

static void spdInitializeLineTbl(void);

SpdLineTbl spdGetLineTbl(SpdObjId fileId)
{
    char    *buf;
    int     buflen, i;

    static SpdBoolean   is_first_call = TRUE;

    if (is_first_call) {
        spdInitializeLineTbl();
        is_first_call = FALSE;
    }

    if (fileId == SPD_NON_ID) {
        spdErrorExit("spdGetLineTbl(): Invalid fileId %d.\n", fileId);
    }

    for (i = no; i < HISTORY_MAX; i++) {
        if (history[i].fileId == fileId) {
            return (history[i].tbl);
        }
    }
    for (i = 0; i < no; i++) {
        if (history[i].fileId == fileId) {
            return (history[i].tbl);
        }
    }

    /* no = (no - 1) % HISTORY_MAX; */  /* Variable no should be unsigned. */
    no = (no == 0) ? HISTORY_MAX - 1 : no - 1;
    spdFreeLineTbl(history[no].tbl);

    history[no].fileId = fileId;
    buflen = spdReadFileInBuf(fileId, &buf);
    history[no].tbl = spdCreateLineTbl(buf, buflen);
    spdFree(buf);

    return (history[no].tbl);
}

static void spdInitializeLineTbl(void)
{
    int         i;

    for (i = 0; i < HISTORY_MAX; i++) {
        history[i].fileId = SPD_NON_ID;
        history[i].tbl.offset = NULL;
        history[i].tbl.linenum = SPD_NON_LINENO;
    }
    no = 0;
}

void spdReleaseLineTbl(void)
{
    int         i;

    for (i = 0; i < HISTORY_MAX; i++) {
        spdFreeLineTbl(history[i].tbl);
    }
    spdInitializeLineTbl();
}

/******************************************************************************/

/*
 * ԤȥбʸΥեƬoffset (0Ϥޤ)
 * Ѵ롥ʸʤSPD_NON_OFFSET֤
 */

int spdCnvLineColumnToOffset(SpdObjId fileId, int line, int column)
{
    return (spdCnvLineColumnToOffsetSub(spdGetLineTbl(fileId),
                                        line, column));
}

int spdCnvLineColumnToOffsetSub(SpdLineTbl tbl, int line, int column)
{
    /* Invalid line number. */
    if (line < 1 || tbl.linenum < line) {
        return (SPD_NON_OFFSET);
    }

    /* Invalid column. */
    if (tbl.offset[line - 1] + (column - 1) >= tbl.offset[line]) {
        return (SPD_NON_OFFSET);
    }

    /* 1ϤޤΤǡ1Τ˺ʤ */
    return (tbl.offset[line - 1] + column - 1);
}

/*
 * եƬoffset"ܲ"Ѵ롥
 * Return value:
 *      0: Found.
 *     -1: Otherwise.
 */

int spdCnvOffsetToLineColumn(SpdObjId fileId, int offset,
                             int *line, int *column)
{
    return (spdCnvOffsetToLineColumnSub(spdGetLineTbl(fileId), offset,
                                        line, column));
}

int spdCnvOffsetToLineColumnSub(SpdLineTbl tbl, int offset,
                                int *line, int *column)
{
    int     *hit;

    hit = (int *)SPD_BSEARCH(&offset, tbl.offset, tbl.linenum, sizeof(int), 
                               comparOffset);
    if (hit == NULL) {
        return (-1);
    }

    /* hit꤬礦ɲԥɤξ硥 */
    /* 
     * if (*(hit+1) - 1 == offset && *hit != tbl.offset[tbl.linenum]) {
     *     return (-1);
     * }
     */
    *line = hit - tbl.offset + 1;
    *column = offset - *hit + 1;

    return (0);
}
