/* 
 * Program:             $RCSfile: printCPPinfo.c,v $  $Revision: 30.2 $
 * 
 * Purpose:             PIDB: printing messages.
 * 
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01
 * 
 * (C) Copyright:       T.Mimura and S.Yamamoto  1993 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: printCPPinfo.c,v 30.2 2001/05/27 12:05:13 yamamoto Exp yamamoto $";
#endif

#include <stdarg.h>
#include <stdio.h>
#include <Sapid/Sapid.h>

/*
 * Message output in verbose mode.
 */

void CPPinfoMessage(const char *format, ...)
{
    va_list     args;
    
    if (CPPinfoVerbose == TRUE) {
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
    }
}

/*
 * ޥɽ
 */

void    printDEFINEinfo(FILE *fp, int n)
{
    DEFINEinfo  *def;
    
    if (n < 0 || defineInfoN <= n) {
        return;
    }
    
    def = &(defineInfo[ n ]);
    fprintf(fp, "DEFINEinfo  ID %5d\n", n+1);
    /* fprintf(fp, "  ID: %5d\n", def->ID); */
    switch (def->pos.type) {
      case DEFINE_INITIALIZATION:
        fprintf(fp, "  define in Initialization.\n");
        break;
      case DEFINE_D_OPTION:
        fprintf(fp, "  define by -D option.\n");
        break;
      case DEFINE_FILE:
        fprintf(fp, "  define in '%s', from line %d to %d.\n",
                def->pos.fname, def->pos.start, def->pos.end);
        break;
    }
    fprintf(fp, "  name: '%s'\n", def->name);
    fprintf(fp, "  narg: %d\n", def->narg);
    fprintf(fp, "  expansion: '%s'\n", def->expansion);
    if (def->narg > 0) {
        int     i;

        fprintf(fp, "  nref: %d\n", def->nref);
        if (def->nref > 0) {
            fprintf(fp, "  reference:\n");
            for (i = 0; i < def->nref; i++) {
                ArgReference    *r;
                r = &((def->ref)[i]);
                fprintf(fp,"    %d chars, arg %d.\n", r->nchar, r->argn);
                if (r->raw_before == TRUE) {
                    fprintf(fp, "      ## operator before this reference.\n");
                }
                if (r->raw_after == TRUE) {
                    fprintf(fp, "      ## operator after  this reference.\n");
                }
                if (r->stringify == TRUE) {
                    fprintf(fp, "      #  operator before this reference.\n");
                }
            }
        }
        for (i = 0; i < def->narg; i++) {
            fprintf(fp, "  arg%2d: '%s'\n", i, (def->arg)[i]);
        }
        fprintf(fp, "  output: %d\n", def->output);
    }
}

void    printEXPANDinfo(FILE *fp, int n)
{
    EXPANDinfo  *exp;
    
    if (n < 0 || expandInfoN <= n) {
        return;
    }
    
    exp = &(expandInfo[ n ]);
    fprintf(fp, "EXPANDinfo  ID %5d\n", n+1);
    fprintf(fp, "  ID: %5d\n", exp->ID);
    fprintf(fp, "  called macro: ID %5d\n", exp->macID);
    switch (exp->input.type) {
      case CALL_FILE:
        fprintf(fp, "  input: '%s' %d-%d\n",
                exp->input.fname, exp->input.start, exp->input.end);
        break;
      case CALL_MACROEXPANSION:
        fprintf(fp, "  input: macroexpand(ID %d) %d-%d\n",
                exp->input.macexpID, exp->input.start, exp->input.end);
        break;
      case CALL_MACROARG:
        fprintf(fp, "  input: macroarg(ID %d  arg %d) %d-%d\n",
                exp->input.macexpID, exp->input.argn,
                exp->input.start, exp->input.end);
        break;
    }
    if (exp->mac->narg >0) {
        int     i;

        for (i = 0; i < exp->mac->narg; i++) {
            MacroArgInfo    *a;
            a = &((exp->arg)[i]);
            fprintf(fp, "  arg %2d: %d-%d\n", i, a->start, a->end);
            fprintf(fp, "    raw length        : %d\n", a->raw_length);
            fprintf(fp, "    expanded length   : %d\n", a->expanded_length);
            fprintf(fp, "    stringified length: %d\n", a->stringified_length);
            fprintf(fp, "    trailing space    : %d\n", a->trailing_space);
            fprintf(fp, "    following space   : %d\n", a->following_space);
        }
    }
    switch (exp->output.type) {
      case EXPAND_FILE:
        fprintf(fp, "  output: '%s' %d-%d\n",
                exp->output.fname, exp->output.start, exp->output.end);
        break;
      case EXPAND_MACROARG:
        fprintf(fp, "  output: macroarg(ID %d  arg %d) %d-%d\n",
                exp->output.macexpID, exp->output.argn,
                exp->output.start, exp->output.end);
        break;
    }
}
