/* 
 * Program:             $RCSfile: exprType.c,v $  $Revision: 30.20 $
 * 
 * Purpose:
 * 
 * Author:              T.Hidaka  1997/06/27
 *                      S.Yamamoto  1997/08/04
 * 
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: exprType.c,v 30.20 2003/07/08 06:19:41 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>

/*
 * η
 */

static SpdObjId smuGetConstantExprType(SpdObjId constantid)
{
    SpdString str;
    int strl, i;
    int size, sign, type;
    SpdBoolean strf;
    SpdObjId res;
    static SpdBoolean use_wchar_t = TRUE;

    str = spdGetAttrValString(constantid, "value");
    strl = strlen(str);

    strf = FALSE;
    size = sign = TYPE_UNKNOWN;
    
    switch (spdGetAttrValInt(constantid, "type")) {
      case CONST_INT:
        type = TYPE_INT;
        break;
      case CONST_FLOAT:
        type = TYPE_DOUBLE;
        break;
      case CONST_CHAR:
        type = TYPE_INT;
        break;
      case CONST_STRING:
        type = TYPE_CHAR;
        strf = TRUE;
        break;
      case CONST_WCHAR:
        if (use_wchar_t) {
            if ((res = smuFindTypedef("wchar_t")) != SPD_NON_ID) {
                return res;
            }
            spdWarning("wchar_t is not defined. Try to use 'unsigned long int'.\n");
            use_wchar_t = FALSE;
        }
        type = TYPE_INT;
        size = TYPE_LONG;
        sign = TYPE_UNSIGNED;
        break;
      default:
        type = TYPE_UNKNOWN;
        break;
    }

    if (strf == FALSE) {
        for (i = strl - 1; i >= 0 && isalpha(str[i]); i--) {
            switch (toupper(str[i])) {
              case 'L':
                size = TYPE_LONG;
                break;
              case 'F':
                if (type == TYPE_DOUBLE) {
                    type = TYPE_FLOAT;
                }
                break;
              case 'U':
                sign = TYPE_UNSIGNED;
                break;
            }
        }
    }
        
    res = smuFindType(TYPE_UNKNOWN, sign, size, type);
    if (res == SPD_NON_ID) {
        if (type == TYPE_INT) {
            /* Try implicit int. */
            res = smuFindType(TYPE_UNKNOWN, sign, size, TYPE_UNKNOWN);
        }
    } else if (strf) {                      /* String is "char *". */
        res = smuFindPointerTo(res);
    }

    return res;
}

/*
 * typedef̾﷿ؤѴ
 * !! const typedef-nameʤɤqualifierդξ̤б
 * !! ߤI-modelǤϺ
 */

static SpdObjId smuExtractTypedef(SpdObjId id)
{
    SpdObjId    ident, type;
    int         sort;
    
    sort = spdGetAttrValInt(id, "sort");
    if (sort == TYPE_TYPE_DEF) {
        ident = spdGetARelObj(id, "type_is", "type_id");
        if (ident != SPD_NON_ID) {
            type = spdGetARelObj(ident, "ident_type", "identifier_id");
            if (type != SPD_NON_ID) {
                id = smuExtractTypedef(type);
            }
        }
    }
    return (id);
}

/*
 * ʡ
 */

static SpdObjId smuTypePromotion(SpdObjId id)
{
    SpdObjId    typeId;
    int         sort, size, sign;

    sort = spdGetAttrValInt(id, "sort");
    size = spdGetAttrValInt(id, "size");
    sign = spdGetAttrValInt(id, "sign");

    /* typedef̾ηء */
    if (sort == TYPE_TYPE_DEF) {
        return smuExtractTypedef(id);
    }
    
    /* arraypointerء */
    if (sort == TYPE_ARRAY) {
        typeId = spdGetARelObj(id, "type_st", "p-id");
        return smuFindPointerTo(typeId);
    }
 
    if (smuObjIdIsTypeIntegral(id)) {
        /* char, shortintء */
        /* 
         * !! unsigned shortΰꤷʤФʤʤ
         * !! ǤintˤƤ롥
         */
        if (size != TYPE_LONG) {
            if (size == TYPE_SHORT) {
                return smuFindType(TYPE_UNKNOWN, sign, TYPE_UNKNOWN, TYPE_INT);
            }
        }
    }
    
    return id;
}

/*
 * 졥
 */

static SpdObjId smuTypeBalancing(SpdObjId type1, SpdObjId type2)
{
    type1 = smuTypePromotion(type1);
    type2 = smuTypePromotion(type2);
   
    /* Ȥ˥ݥ󥿤ξ -- ݥ󥿤ɤΰ = ptrdiff_t*/
    if (smuObjIdIsTypePointer(type1) &&
        smuObjIdIsTypePointer(type2)) {
        return smuFindType(TYPE_UNKNOWN, TYPE_UNKNOWN,
                           TYPE_UNKNOWN, TYPE_INT);
    }

    /* 줫ݥ󥿤ξ -- ݥ +  = ݥ */
    if (smuObjIdIsTypePointer(type1) && smuObjIdIsTypeIntegral(type2)) {
        return type1;
    }
    if (smuObjIdIsTypePointer(type2) && smuObjIdIsTypeIntegral(type1)) {
        return type2;
    }

    /* 줫 long double ξ --  */
    if (smuObjIdIsTypeDouble(type1) && smuObjIdIsTypeLong(type1)) {
        return type1;
    }
    if (smuObjIdIsTypeDouble(type2) && smuObjIdIsTypeLong(type2)) {
        return type2;
    }

    /* 줫 double ξ --  */
    if (smuObjIdIsTypeDouble(type1)) {
        return type1;
    }
    if (smuObjIdIsTypeDouble(type2)) {
        return type2;
    }

    /* 줫 unsigned long ξ --  */
    if (smuObjIdIsTypeLong(type1) && smuObjIdIsTypeUnsigned(type1)) {
        return type1;
    }
    if (smuObjIdIsTypeLong(type2) && smuObjIdIsTypeUnsigned(type2)) {
        return type2;
    }

    /* 줫 long ξ --  */
    if (smuObjIdIsTypeLong(type1)) {
        return type1;
    }
    if (smuObjIdIsTypeLong(type2)) {
        return type2;
    }

    /* 줫 unsigned int ξ --  */
    if (smuObjIdIsTypeUnsigned(type1)) {
        return type1;
    }
    if (smuObjIdIsTypeUnsigned(type2)) {
        return type2;
    }

    /*  int ξ */
    return type1;
}

/*
 * Ѽη
 */

static SpdObjId smuGetArithExprType(SpdObjId exprid)
{
    SpdObjId res = SPD_NON_ID;
    SpdObjId op = SPD_NON_ID;
    SpdObjId ident = SPD_NON_ID;
    SpdObjId constant = SPD_NON_ID;
    SpdRelArray exprs;
    SpdObjIdArray types;

    int nterms, nexprs;
    int opcat;
    int i;
    
    op = spdGetARelObj(exprid, "expr_op", "expression_id");
    exprs = spdGetRelSortedArray(exprid, "expr_expr", "p-id", NULL,
                                 spdCompareRelOffset);
    nexprs = exprs.size;
    if (0 < nexprs) {
        types.id = spdMalloc(sizeof(SpdRel) * nexprs);
        for (i = 0; i < nexprs; i++) {
            types.id[i] = smuGetExprType(exprs.rel[i].objectId);
        }
    } else {
        types.id = NULL;
    }

    if (op != SPD_NON_ID) {
        nterms = smuGetOperatorTerms(op);
        opcat = smuGetOperatorCategory(op);
    } else {
        nterms = 0;
        opcat = 0;
    }
    
    /* ޤοʬह */
    switch (nterms) {
      case 0:
        /* OP_SIZEOF_TYPE ξ(δؿǤ)ꤨʤ */
        /* äơѿ or λȤǤϤ */
        ident = spdGetARelObj(exprid, "ident_ref", "expression_id");
        if (ident != SPD_NON_ID) {
            res = spdGetARelObj(ident, "ident_type", "identifier_id");
        } else {
            constant = spdGetARelObj(exprid, "const_ref", "expression_id");
            if (constant != SPD_NON_ID) {
                res = smuGetConstantExprType(constant);
            }
        }
        break;
      case 1:
        /* 󥯥ȡǥ, ñ+/-, ä ʤɤ */
        /* ̤Ϥ٤ʬη */
        res = types.id[0];
        break;
      case 2:
        if (opcat & OPCATEGORY_TRASHRTYPE) {
            /* /եȱ黻ʤɤϺͤη */
            res = types.id[0];
        } else {
            /* ʳϷ */
            res = smuTypeBalancing(types.id[0], types.id[1]);
        }
        break;
      case 3:
        /* ?: (δؿǤ)ꤨʤ */
        break;
      default:
        break;
    }

    /* Added by S.Yamamoto 2000/09/19. */
    if (types.id != NULL) {
        spdFree(types.id);
    }
    spdFreeRelArray(exprs);

    return (res);
}

/******************************************************************************/

#if defined(SAPID_AR3_KERNEL)

/*
 * η롥
 * 
 * :
 * 2000/09/20 S.Yamamoto 
 * δؿϡmkExprType2ˤƤΤ߻Ѥ롥
 * ʤ̤ѼԤˤϸʤ
 * 
 * \subsection{ηؿ}
 * 
 * ߤsdbin4ϼȷδ֤δϢ``expr_type''ʤᡤ
 * ʲδؿѤƼη롥
 * 
 * \begin{SapidFuncDef}{SpdObjId}{smuGetExprType}
 * \SapidFuncArgItem{SpdObjId}{exprId}{}
 * \SapidDescriptionItem{
 * \SapidVarName{exprId}ɽη롥ͤ
 * \SapidClassName{type}饹Υ󥹥󥹤IDǤ롥
 * 
 * \SapidVarName{exprId}ɽ\SapidCode{return},
 * \SapidCode{break}ʤɤΤ褦ʤ˷¸ߤʤξ䡤
 * ᤿ǡ١¸ߤʤˤϡ
 * \SapidValue{SAPID\_NON\_ID}֤
 * }
 * \end{SapidFuncDef}
 */ 

SpdObjId smuGetExprType(SpdObjId exprid)
{
    SpdObjId res = SPD_NON_ID;
    SpdObjId op;
    SpdObjId type = SPD_NON_ID;
    SpdObjId cexpr = SPD_NON_ID;
    SpdObjId ident = SPD_NON_ID;
    SpdRelArray exprs;
    
    /* exprid  expression  ID Ǥʤнʤ */
    if (spdGetClassId(exprid) != SPD_CID_EXPRESSION) {
        return SPD_NON_ID;
    }
    
    op = spdGetARelObj(exprid, "expr_op", "expression_id");
    exprs = spdGetRelSortedArray(exprid, "expr_expr", "p-id", NULL,
                                 spdCompareRelOffset);
    if (exprs.size > 0) {
        cexpr = exprs.rel[0].objectId;
    }
    
    switch (smuGetExprSort(exprid)) {
      case EXPR_BREAK:
      case EXPR_CONTINUE:
      case EXPR_RETURN:
      case EXPR_GOTO:
        /* ANSI/ISO C ǤμǤʤнʤ */
        res = SPD_NON_ID;
        break;
      case EXPR_EMPTY:
        /* !! μvoid ? */
        res = smuFindType(TYPE_UNKNOWN, TYPE_UNKNOWN, TYPE_UNKNOWN, TYPE_VOID);
        break;
      case EXPR_SORT_ARITH:
      case EXPR_SORT_BIT:
      case EXPR_SORT_SHIFT:
        res = smuGetArithExprType(exprid);
        break;
      case EXPR_SORT_LOGICAL:
      case EXPR_SORT_COMPARE:
        /* int */
        res = smuFindType(TYPE_UNKNOWN, TYPE_UNKNOWN, TYPE_UNKNOWN, TYPE_INT);
        break;
      case EXPR_SORT_ARRAY:
        /* η */
        if (cexpr != SPD_NON_ID) {
            type = smuGetExprType(cexpr);
            type = smuExtractTypedef(type);
            if (smuObjIdIsTypePointer(type)) {
                res = spdGetARelObj(type, "type_st", "p-id");
            } else {
                res = SPD_NON_ID;
            }
        }
        break;
      case EXPR_SORT_POINTER_MEMBER:
      case EXPR_SORT_MEMBER:
        /* Фη */
        if (cexpr != SPD_NON_ID) {
            res = smuGetExprType(exprs.rel[exprs.size - 1].objectId);
        }
        break;
      case EXPR_SORT_POINTER:
        /* ݥ󥿤λؤ */
        if (cexpr != SPD_NON_ID) {
            type = smuGetExprType(cexpr);
            type = smuExtractTypedef(type);
            if (smuObjIdIsTypePointer(type)) {
                res = spdGetARelObj(type, "type_st", "p-id");
            } else {
                res = SPD_NON_ID;
            }
        }
        break;
      case EXPR_SORT_ADDRESS:
        /* 뷿ؤΥݥ */
        if (cexpr != SPD_NON_ID) {
            type = smuGetExprType(cexpr);
            res = smuFindPointerTo(type);
        }
        break;
      case EXPR_SORT_SIZEOF:
        /* "size_t" is defined as "unsigned int" in "stddef.h". */
        res = smuFindType(TYPE_UNKNOWN, TYPE_UNSIGNED, TYPE_UNKNOWN, TYPE_INT);
        break;
      case EXPR_SORT_CAST:
        /* 㥹ȸη */
        res = spdGetARelObj(op, "op_type", "operator_id");
        break;
      case EXPR_SORT_TRI:
        /* !!  (ΤȤη֤...) */
      case EXPR_SORT_COMMA:
      case EXPR_SORT_INITLIST:
        /* Type of the last sub-expression. */
        if (0 < exprs.size) {
            res = smuGetExprType(exprs.rel[exprs.size - 1].objectId);
        } else {
            /* When no sub-expression exists, the expression is empty. */
            res = smuFindType(TYPE_UNKNOWN, TYPE_UNKNOWN, TYPE_UNKNOWN,
                              TYPE_VOID); 
        }
        break;
      case EXPR_SORT_FUNCCALL:
        /* Type of function's return value. */
        if (cexpr != SPD_NON_ID) {
            ident = spdGetARelObj(cexpr, "ident_ref", "expression_id");
            if (ident != SPD_NON_ID) {
                /* ؿ */
                res = spdGetARelObj(ident, "ident_type", "identifier_id");
                if (res != SPD_NON_ID) {
                    /* ͤη */
                    res = spdGetARelObj(res, "type_st", "p-id");
                }
            }
        }
        break;
      case EXPR_SORT_BUILTIN:
        /* Բǽ */
        res = SPD_NON_ID;
        break;
    }

    spdFreeRelArray(exprs);
    return (res);
}

#endif /* defined(SAPID_AR3_KERNEL) */
