/* 
 * Program:             $RCSfile: getObjIdArray.c,v $  $Revision: 30.13 $
 * 
 * Purpose:
 * 
 * Author:              A.Watanabe  1996/05/15
 *                      S.Yamamoto  1996/06/05
 * 
 * (C) Copyright:       A.Watanabe and S.Yamamoto  1996 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
"$Id: getObjIdArray.c,v 30.13 2003/07/08 05:46:16 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdBoolean spdTrueById(SpdId dummy)
{
    return (TRUE);
}

static SpdBoolean spdTrueByRel(SpdRel dummy)
{
    return (TRUE);
}

static SpdBoolean spdTrueByOcc(SpdOcc dummy)
{
    return (TRUE);
}

void spdFreeObjIdArray(SpdObjIdArray array)
{
    spdSafetyFree(array.id);
}

void spdFreeRelIdArray(SpdRelIdArray array)
{
    spdSafetyFree(array.id);
}

void spdFreeRelArray(SpdRelArray array)
{
    spdSafetyFree(array.rel);
}

void spdFreeOccArray(SpdOccArray array)
{
    spdSafetyFree(array.occ);
}

SpdObjIdArray spdGetObjIdArray(SpdConstString className,
                               SpdConstString name,
                               SpdCondObjIdFunc cond)
{
    SpdCursor       *csr;
    SpdObjId        id;
    SpdObjIdArray   array = SPD_EMPTY_ARRAY;
    
    if (cond == NULL) {
        cond = spdTrueById;
    }
    
    csr = spdGetObjIdInit(className, name);
    if (csr != NULL) {
        while ((id = spdGetObjId(csr)) != SPD_NON_ID) {
            if ((*cond)(id)) {
                array.id =
                    spdRealloc(array.id, sizeof(SpdObjId) * (array.size + 1));
                array.id[array.size] = id;
                array.size++;
            }
        }
        spdFreeCursor(csr);
    }
    return (array);
}

SpdObjIdArray spdGetObjIdSortedArray(SpdConstString className,
                                     SpdConstString name,
                                     SpdCondObjIdFunc cond,
                                     SpdCompareObjIdFunc compar)
{
    SpdObjIdArray   array;
    
    array = spdGetObjIdArray(className, name, cond);
    SPD_QSORT(array.id, array.size, sizeof(SpdObjId), compar);
    return (array);
}

SpdRelArray spdGetRelArray(SpdObjId objectId,
                           SpdConstString relationName,
                           SpdConstString attrName,
                           SpdCondRelFunc cond)
{
    SpdCursor       *csr;
    SpdRel          rel;
    SpdRelArray     array = SPD_EMPTY_ARRAY;
    
    if (cond == NULL) {
        cond = spdTrueByRel;
    }
    
    csr = spdGetRelInit(objectId, relationName, attrName);
    while ((rel = spdGetRel(csr)).relationId != SPD_NON_ID) {
        if ((*cond)(rel)) {
            array.rel =
                spdRealloc(array.rel, sizeof(SpdRel) * (array.size + 1));
            array.rel[array.size] = rel;
            array.size++;
        }
    }
    spdFreeCursor(csr);
    return (array);
}

SpdRelArray spdGetRelSortedArray(SpdObjId objectId,
                                 SpdConstString relationName,
                                 SpdConstString attrName,
                                 SpdCondRelFunc cond,
                                 SpdCompareRelFunc compar)
{
    SpdRelArray     array;
    
    array = spdGetRelArray(objectId, relationName, attrName, cond);
    SPD_QSORT(array.rel, array.size, sizeof(SpdRel), compar);
    return (array);
}

SpdObjIdArray spdGetRelObjArray(SpdObjId objectId,
                                SpdConstString relationName,
                                SpdConstString attrName,
                                SpdCondObjIdFunc cond)
{
    SpdCursor       *csr;
    SpdObjId        id;
    SpdObjIdArray   array = SPD_EMPTY_ARRAY;
    
    if (cond == NULL) {
        cond = spdTrueById;
    }
    
    csr = spdGetRelObjInit(objectId, relationName, attrName);
    if (csr != NULL) {
        while ((id = spdGetRelObj(csr)) != SPD_NON_ID) {
            if ((*cond)(id)) {
                array.id =
                    spdRealloc(array.id, sizeof(SpdObjId) * (array.size + 1));
                array.id[array.size] = id;
                array.size++;
            }
        }
        spdFreeCursor(csr);
    }
    return (array);
}

SpdObjIdArray spdGetRelObjSortedArray(SpdObjId objectId,
                                      SpdConstString relationName,
                                      SpdConstString attrName,
                                      SpdCondObjIdFunc cond,
                                      SpdCompareObjIdFunc compar)
{
    SpdObjIdArray   array;
    
    array = spdGetRelObjArray(objectId, relationName, attrName, cond);
    SPD_QSORT(array.id, array.size, sizeof(SpdObjId), compar);
    return (array);
}

SpdRelIdArray spdGetRelIdArray(SpdObjId objectId1,
                               SpdObjId objectId2,
                               SpdConstString relationName,
                               SpdCondRelIdFunc cond)
{
    SpdCursor       *csr;
    SpdRelId        id;
    SpdRelIdArray   array = SPD_EMPTY_ARRAY;
    
    if (cond == NULL) {
        cond = spdTrueById;
    }
    
    csr = spdGetRelIdInit(objectId1, objectId2, relationName);
    if (csr != NULL) {
        while ((id = spdGetRelId(csr)) != -1) {
            if ((*cond)(id)) {
                array.id =
                    spdRealloc(array.id, sizeof(SpdRelId) * (array.size + 1));
                array.id[array.size] = id;
                array.size++;
            }
        }
        spdFreeCursor(csr);
    }
    return (array);
}

SpdRelIdArray spdGetRelIdSortedArray(SpdObjId objectId1,
                                     SpdObjId objectId2,
                                     SpdConstString relationName,
                                     SpdCondRelIdFunc cond,
                                     SpdCompareRelIdFunc compar)
{
    SpdRelIdArray   array;
    
    array = spdGetRelIdArray(objectId1, objectId2, relationName, cond);
    SPD_QSORT(array.id, array.size, sizeof(SpdRelId), compar);
    return (array);
}

SpdOccArray spdGetOffsetOfObjArray(SpdObjId objectId,
                                   SpdCondOccFunc cond)
{
    SpdCursor       *csr;
    SpdOcc          occ;
    SpdOccArray     array = SPD_EMPTY_ARRAY;
    
    if (cond == NULL) {
        cond = spdTrueByOcc;
    }
    
    csr = spdGetOffsetOfObjInit(objectId);
    if (csr != NULL) {
        while ((occ = spdGetOffsetOfObj(csr)).relationId != SPD_NON_ID) {
            if ((*cond)(occ)) {
                array.occ =
                    (SpdOcc *)spdRealloc(array.occ,
                                         sizeof(SpdOcc) * (array.size + 1));
                array.occ[array.size] = occ;
                array.size++;
            }
        }
        spdFreeCursor(csr);
    }
    return (array);
}

SpdOccArray spdGetOffsetOfObjSortedArray(SpdObjId objectId,
                                         SpdCondOccFunc cond,
                                         SpdCompareOccFunc compar)
{
    SpdOccArray     array;
    
    array = spdGetOffsetOfObjArray(objectId, cond);
    SPD_QSORT(array.occ, array.size, sizeof(SpdOcc), compar);
    return (array);
}

SpdOccArray spdGetIncludedOccArray(SpdObjId objectId,
                                   SpdConstString objectType,
                                   SpdObjId fileId,
                                   SpdCondOccFunc cond)
{
    SpdCursor       *csr;
    SpdOcc          occ;
    SpdOccArray     array = SPD_EMPTY_ARRAY;
    
    if (cond == NULL) {
        cond = spdTrueByOcc;
    }
    
    /* Modified by H.Ohhashi 1998/05/09. */
    /* csr = spdGetIncludedOccInit(objectId, objectType); */
    csr = spdGetIncludedOccInit(objectId, objectType, fileId);
    while ((occ = spdGetIncludedOcc(csr)).relationId != SPD_NON_ID) {
        if ((*cond)(occ)) {
            array.occ =
                spdRealloc(array.occ, sizeof(SpdOcc) * (array.size + 1));
            array.occ[array.size] = occ;
            array.size++;
        }
    }
    spdFreeCursor(csr);
    return (array);
}

SpdOccArray spdGetIncludedOccSortedArray(SpdObjId objectId,
                                         SpdConstString objectType,
                                         SpdObjId fileId,
                                         SpdCondOccFunc cond,
                                         SpdCompareOccFunc compar)
{
    SpdOccArray     array;
    
    array = spdGetIncludedOccArray(objectId, objectType, fileId, cond);
    SPD_QSORT(array.occ, array.size, sizeof(SpdOcc), compar);
    return (array);
}

/* 
 * Sample compare() functions.
 */

int spdCompareObjIdName(const SpdObjId *a, const SpdObjId *b)
{
    return (spdSafetyStrcmp(spdGetName(*a), spdGetName(*b)));
}

int spdCompareRelOffset(const SpdRel *a, const SpdRel *b)
{
    return (spdGetOffset(a->relationId) - spdGetOffset(b->relationId));
}

int spdCompareRelIdOffset(const SpdRelId *a, const SpdRelId *b)
{
    return (spdGetOffset(*a) - spdGetOffset(*b));
}

int spdCompareOccOffset(const SpdOcc *a, const SpdOcc *b)
{
    return (a->offset - b->offset);
}

int spdCompareOccName(const SpdOcc *a, const SpdOcc *b)
{
    return (spdSafetyStrcmp(spdGetName(a->objectId), spdGetName(b->objectId)));
}

/* 
 * Sample cond() functions.
 */

SpdBoolean spdObjIdIsDeclNormal(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == DECL_NORMAL);
}

SpdBoolean spdObjIdIsDeclFuncdef(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == DECL_FUNCDEF);
}

SpdBoolean spdObjIdIsDeclEnumerator(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == DECL_ENUMERATOR);
}

SpdBoolean spdObjIdIsDeclParameter(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == DECL_PARAMETER);
}

SpdBoolean spdObjIdIsDeclMember(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == DECL_MEMBER);
}

SpdBoolean spdObjIdIsDeclType(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == DECL_TYPE);
}

/*
 * Identifier.
 */

SpdBoolean spdObjIdIsIdVariable(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == ID_VARIABLE);
}

SpdBoolean spdObjIdIsIdFunction(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == ID_FUNCTION);
}

SpdBoolean spdObjIdIsIdMember(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == ID_MEMBER);
}

SpdBoolean spdObjIdIsIdEnumerator(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == ID_ENUMERATOR);
}

SpdBoolean spdObjIdIsIdTypename(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == ID_TYPENAME);
}

SpdBoolean spdObjIdIsIdTag(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == ID_TAG);
}
