/* 
 * Program:             $RCSfile: smuFileUtil.c,v $  $Revision: 31.23 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohhashi  1999/03/08
 *                      S.Yamamoto  1999/03/08
 * 
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  1999 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: smuFileUtil.c,v 31.23 2003/07/08 06:19:40 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#define PATH_NUM    16

static int spdReadFileInBufSub(SpdObjId file_id, char **buf);

struct SystemPath {
    char    *path[PATH_NUM];
    size_t  size;
};

static struct SystemPath    system_path;
static char                 *g_path_list = NULL;

/*
 * ֹfileIdǻꤵ줿եɤ߹ߤΤ˥ץ󤹤롥
 */

FILE *smuOpenSrcFile(SpdObjId fileId)
{
    FILE        *fp;
    char        fileName[FILENAME_MAX + 1];

#ifdef SPD_MY_FSEEK
    static int  is_first_call = TRUE;

    if (is_first_call) {
        int         i;

        is_first_call = FALSE;
        for (i = 0; i < FOPEN_MAX; i++) { 
            spdOpendFile[i].isOpened = FALSE;
        }
    }
#endif /* SPD_MY_FSEEK */

    switch (spdGetClassId(fileId)) {
      case SPD_CID_PFILE:
        fileId = spdConvIdPfileToFile(fileId);
        break;
      case SPD_CID_FILE:
        break;
      default:
        spdErrorExit("smuOpenSrcFile(): %d: Invalid file Id.\n", fileId);
        break;
    }

    sprintf(fileName, "%s/%s/%d", spdGetSDBdir(), SAPID_SRC_DIR, fileId);
    fp = spdFopen(fileName, SPD_FOPEN_RT_MODE);

#ifdef SPD_MY_FSEEK
    spdOpendFile[fileno(fp)].isOpened = TRUE;
    spdOpendFile[fileno(fp)].fileId = fileId;
    DW_MY_FSEEK("SPD_MY_FSEEK: Open: %d: smuOpenSrcFile(%d).\n",
                fileno(fp), fileId);
#endif /* SPD_MY_FSEEK */

    return (fp);
}

static int countPathNum(void)
{
    char    *path;
    int     count = 1;

    path = g_path_list;
    while (*path != '\0') {
        if (*path == ',') {
            count++;
        }
        path++;
    }
    return (count);
}

static SpdBoolean divPathList(void)
{
    int     i;

    system_path.size = countPathNum();
    /* printf("%d\n", system_path.size); */
    if (PATH_NUM < system_path.size) {
        return (FALSE);
    }
    for (i = 0; i < system_path.size - 1; i++) {
        system_path.path[i] = spdStrdupUntilChr(g_path_list, ',');
        g_path_list += strlen(system_path.path[i]) + 1;
    }
    system_path.path[i] = g_path_list;
    return (TRUE);
}

void smuSetSystemPath(char *path_list)
{
    if (path_list == NULL) {
        spdWarning("smuSetSystemPath(): Empty path list.\n");
        return;
    }
    g_path_list = path_list;
    if (!divPathList()) {
        spdErrorExit("smuSetSystemPath(): \"%s\": Too many pathes.\n",
                     path_list);
    }
}

/*    
 * ƥե뤫ݤȽꤹ롥
 * ѥ̾Ƭ'/'ξ˥ƥեȤߤʤ
 * Windowsξ'\\'θ롥
 */

SpdBoolean smuIsSystemFileName(SpdString file)
{
    int i;

    if (file == NULL) {
        return (FALSE);
    }

    if (g_path_list == NULL) {

#ifdef SPD_ON_WINDOWS_FS
        file = spdRemoveDriveName(file);
        return (file[0] == '/' || file[0] == '\\');
#else
        return (file[0] == '/');
#endif

    }

    for (i = 0; i < system_path.size; i++) {
        if (SPD_STREQ_N(file, system_path.path[i],
                        strlen(system_path.path[i]))) {
            return (TRUE);
        }
    }
    return (FALSE);
}

/*
 * ƥե뤫ݤȽꤹ롥
 */

SpdBoolean smuIsSystemFile(SpdObjId pfileId)
{
    return (smuIsSystemFileName(spdGetName(pfileId)));
}

/*
 * ƥ줿֥ȤݤȽꤹ롥
 */

SpdBoolean smuIsSystemObject(SpdObjId objectId)
{
    SpdOcc      occ;
    SpdObjId    declId, blockId, pfileId;

    switch (spdGetClassId(objectId)) {
      case SPD_CID_PFILE:
      case SPD_CID_FILE:
        /* .i Υե⥷ƥΥ֥ȤȤ롥 */
        return (smuIsSystemFileName(spdGetName(objectId)) &&
                spdIsFileExtension(spdGetName(objectId), ".i"));
      case SPD_CID_PROGRAM:
        /* programϥƥΥ֥ȤǤϤʤ */ 
        return (FALSE);
      case SPD_CID_TYPE:
        /* int, doubleȤ߹ߤηϥƥΥ֥ȤȤ롥 */
        if (spdGetAttrValInt(objectId, "type") != TYPE_UNKNOWN) {
            return (TRUE);
        }
        /* declarationĴ٤롥 */ 
        declId = spdGetARelObj(objectId, "decl_type", "c-id");
        return (declId == SPD_NON_ID ? TRUE : smuIsSystemObject(declId));
      default:
        occ = spdGetAnOffsetOfObj(objectId);
        /* pblockdefine_optionΥޥϡƥΥ֥ȤȤ롥 */
        if (occ.objectId == SPD_NON_ID || occ.fileId == SPD_NON_ID) {
            return (TRUE);
        }
        blockId = spdGetABlockId(occ.fileId, occ.offset);
        if (blockId != SPD_NON_ID) {
            /* I-modelΥ饹 */
            pfileId = spdGetARelObj(blockId, "file_block", "c-id");
        } else {
            pfileId = occ.fileId;
        }
        return (smuIsSystemFile(pfileId));
    }
}

/******************************************************************************/

/* 
 * եĹ֤
 * UNIXξϡspdGetFileSize()֤ͤƱ
 * WindowsϡԶڤ"\r\n""\n"ѴĹ
 *
 * !! ʤС饹file°֤٤
 */

int spdGetFileLength(SpdObjId fileId)
{
    /* Temporarily. */
    return (spdGetFileSize(fileId));
}

/* 
 * ե礭(stat()֤st_sizeƱ)֤
 * !! ʤС饹file°֤٤
 */

int spdGetFileSize(SpdObjId fileId)
{
    char            filePath[FILENAME_MAX + 1];

    switch (spdGetClassId(fileId)) {
      case SPD_CID_PFILE:
        fileId = spdConvIdPfileToFile(fileId);
        break;
      case SPD_CID_FILE:
        break;
      default:
        spdErrorExit("spdGetFileSize(): %d: Invalid file Id.\n", fileId);
        break;
    }
    sprintf(filePath, "%s/%s/%d", spdGetSDBdir(), SAPID_SRC_DIR, fileId);
    return (spdGetFileLengthByName(filePath));
}

int spdGetFileLengthByName(SpdConstString fileName)
{
    struct stat     st;

    spdStat(fileName, &st);
    return ((int)(st.st_size));
}

/******************************************************************************/

/*
 * եΥʬΥХåեݤƥեɤ߹ࡥ
 * ͤϡɤ߹ʸΥ('\0'ߥ͡ޤޤʤ)
 * 
 * fileIdˤС󡥺ϤȤ٤Ǥ롥fileId
 * ".c" (file饹뤤pfile饹)μֹǤ.
 */

int spdReadFileInBuf(SpdObjId fileId, char **buf)
{
    return (spdReadFileInBufSub(spdConvIdPfileToFile(fileId), buf));
}

/*
 * եΥʬΥХåեݤƥեɤ߹ࡥ
 * ͤɤ߹ʸΥ('\0'ߥ͡ޤޤʤ)
 */

int spdReadFileInBufByName(SpdConstString fname, char **buf)
{
    return (spdReadFileInBufSub(smuGetFileIdByName(fname), buf));
}

static int spdReadFileInBufSub(SpdObjId fileId, char **buf)
{
    int             s;
    FILE            *fp;

    s = spdGetFileLength(fileId);
    *buf = spdMalloc(s + 1);
    fp = smuOpenSrcFile(fileId);

#ifdef SPD_ON_WINDOWS_FS
    /* Note, fread() works in text mode in Win32Architecture,
       so we don't check return value of fread(). */
    s = spdFreadChar(*buf, s, fp);
#else   
    if (spdFreadChar(*buf, s, fp) != s) {
        spdErrorExit("spdReadFileInBufSub(): %d: Can't read file.\n", fileId);
    }
#endif

    spdFclose(fp);
    return (s);
}

/******************************************************************************/

int spdReadJavaInBufByName(SpdConstString fileName, char **buf)
{
    int             s;
    FILE            *fp;

    s = spdGetFileLengthByName(fileName);
    *buf = spdMalloc(s + 1);
    fp = spdOpenJavaSource(fileName, SAPID_O_READ);

#ifdef SPD_ON_WINDOWS_FS
    /* Note, fread() works in text mode in Win32Architecture,
       so we don't check return value of fread(). */
    s = spdFreadChar(*buf, s, fp);
#else   
    if (spdFreadChar(*buf, s, fp) != s) {
        spdErrorExit("spdReadJavaInBufByName(): %d: Can't read file.\n",
                     fileName);
    }
#endif

    spdFclose(fp);
    return (s);
}
