/* 
 * Program:             $RCSfile: smuGetAttr.c,v $  $Revision: 30.13 $
 * 
 * Purpose:             SDB Access Tool.
 * 
 * Author:              M.Terazawa  1995/12/28
 *                      S.Yamamoto  1996/01/03
 * 
 * (C) Copyright:       M.Terazawa and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: smuGetAttr.c,v 30.13 2003/07/30 16:26:42 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#define NOT_SPECIFIED       "not specified"

/* Program, not implemented yet. */

/* File, not implemented yet. */ 

/* Declaration. */
static SpdValName attrDeclarationSort[] = {
    { SPD_VAL_NAME3(DECL_NORMAL, "normal") },
    { SPD_VAL_NAME3(DECL_FUNCDEF, "funcdef") },
    { SPD_VAL_NAME3(DECL_MEMBER, "member") },
    { SPD_VAL_NAME3(DECL_ENUMERATOR, "enumerator") },
    { SPD_VAL_NAME3(DECL_PARAMETER, "parameter") },

    /* 
     * Added by S.Yamamoto 1996/05/24.
     * Commented by S.Yammoto 1997/06/24.
     * These macro are used in only declarator's field.
     */
    /* 
     * { SPD_VAL_NAME3(DECL_PROTOTYPE, "prototype") },
     * { SPD_VAL_NAME3(DECL_TAG, "tag") },
     * { SPD_VAL_NAME3(DECL_VARIABLE, "varaiable") },
     */

    { SPD_VAL_NAME3(DECL_TYPE, "type") },

    /*
     * Added by S.Yamamoto 1996/05/24.
     * Commented by S.Yamamoto 2000/04/03.
     */
    /* { SPD_VAL_NAME3(DECL_MACRO, "macro") }, */

    { -1, NULL, NULL }
};

static SpdValName attrDeclarationStorage[] = {
    { SPD_VAL_NAME3(STORAGE_NONE, NOT_SPECIFIED) },
    { SPD_VAL_NAME3(STORAGE_AUTO, "auto") },
    { SPD_VAL_NAME3(STORAGE_EXTERN, "extern") },
    { SPD_VAL_NAME3(STORAGE_TYPEDEF, "typedef") },
    { SPD_VAL_NAME3(STORAGE_STATIC, "static") },
    { SPD_VAL_NAME3(STORAGE_REGISTER, "register") },
    { -1, NULL, NULL }
};

/* Declarator. */
static SpdValName attrDeclaratorSort[] = {
    { SPD_VAL_NAME3(DECL_FUNCDEF, "funcdef") },
    { SPD_VAL_NAME3(DECL_MEMBER, "member") },
    { SPD_VAL_NAME3(DECL_ENUMERATOR, "enumrator") },
    { SPD_VAL_NAME3(DECL_PROTOTYPE, "prototype") },
    { SPD_VAL_NAME3(DECL_TAG, "tag") },
    { SPD_VAL_NAME3(DECL_VARIABLE, "variable") },
    { SPD_VAL_NAME3(DECL_TYPE, "type") },

    /* Added by S.Yamamoto 1997/06/24. */
    { SPD_VAL_NAME3(DECL_PARMTYPE, "parmType") },

    { -1, NULL, NULL }
};

/* Identifier. */
static SpdValName attrIdentifierSort[] = {
    { SPD_VAL_NAME3(ID_VARIABLE, "variable") },
    { SPD_VAL_NAME3(ID_FUNCTION, "function") },
    { SPD_VAL_NAME3(ID_MEMBER, "member") },
    { SPD_VAL_NAME3(ID_ENUMERATOR, "enumerator") },
    { SPD_VAL_NAME3(ID_TYPENAME, "type name") },
    { SPD_VAL_NAME3(ID_TAG, "tag") },
    { -1, NULL, NULL }
};

/* Optionaldecl. */
static SpdValName attrOptionaldeclSort[] = {
    { SPD_VAL_NAME3(OPT_STRUCT, "struct") },
    { SPD_VAL_NAME3(OPT_FUNCDEF, "funcdef") },
    { SPD_VAL_NAME3(OPT_INITIALIZER, "initalizer") },
    { SPD_VAL_NAME3(OPT_PROTOTYPE, "prototype") },

    /* Added by S.Yamamoto 1997/06/24. */
    { SPD_VAL_NAME3(OPT_FUNCPTR, "funcPtr") },

    { -1, NULL, NULL }
};

/* Block. */
static SpdValName attrBlockSort[] = {
    { SPD_VAL_NAME3(BLOCK_BASIC, "basic") },
    { SPD_VAL_NAME3(BLOCK_HIERACHICAL, "hierachical") },
    { SPD_VAL_NAME3(BLOCK_COMP, "compound") },
    { SPD_VAL_NAME3(BLOCK_BRANCH, "branch") },
    { SPD_VAL_NAME3(BLOCK_IF, "if") },
    { SPD_VAL_NAME3(BLOCK_SWITCH, "switch") },
    { SPD_VAL_NAME3(BLOCK_FOR, "for") },
    { SPD_VAL_NAME3(BLOCK_WHILE, "while") },
    { SPD_VAL_NAME3(BLOCK_DO, "do") },
    { -1, NULL, NULL }
};

/* Expression. */
static SpdValName attrExpressionSort[] = {
    { SPD_VAL_NAME3(EXPR_EMPTY, "empty") },
    { SPD_VAL_NAME3(EXPR_BREAK, "break") },
    { SPD_VAL_NAME3(EXPR_CONTINUE, "continue") },
    { SPD_VAL_NAME3(EXPR_RETURN, "return") },
    { SPD_VAL_NAME3(EXPR_GOTO, "goto") },
    { SPD_VAL_NAME3(EXPR_SORT_ARITH, "arithmetic") },
    { SPD_VAL_NAME3(EXPR_SORT_LOGICAL, "logical") },
    { SPD_VAL_NAME3(EXPR_SORT_ARRAY, "array") },
    { SPD_VAL_NAME3(EXPR_SORT_MEMBER, "member") },
    { SPD_VAL_NAME3(EXPR_SORT_POINTER, "pointer") },
    { SPD_VAL_NAME3(EXPR_SORT_ADDRESS, "address") },
    { SPD_VAL_NAME3(EXPR_SORT_TRI, "?:") },
    { SPD_VAL_NAME3(EXPR_SORT_SIZEOF, "sizeof") },
    { SPD_VAL_NAME3(EXPR_SORT_CAST, "cast") },
    { SPD_VAL_NAME3(EXPR_SORT_COMMA, "camma") },
    { SPD_VAL_NAME3(EXPR_SORT_FUNCCALL, "function call") },
    { SPD_VAL_NAME3(EXPR_SORT_INITLIST, "init-list") },
    { SPD_VAL_NAME3(EXPR_SORT_BUILTIN, "built-in function") },
    { SPD_VAL_NAME3(EXPR_SORT_BIT, "bit") },
    { SPD_VAL_NAME3(EXPR_SORT_SHIFT, "shift") },
    { SPD_VAL_NAME3(EXPR_SORT_COMPARE, "compare") },
    { SPD_VAL_NAME3(EXPR_SORT_POINTER_MEMBER, "pointer_member") },
    { -1, NULL, NULL }
};

static SpdValName attrExpressionLrtype[] = {
    { SPD_VAL_NAME3(EXPR_LR_UNKNOWN, "unknown") },

    { SPD_VAL_NAME3(EXPR_L_VALUE, "left value") },
    { SPD_VAL_NAME3(EXPR_R_VALUE, "right value") },
    { SPD_VAL_NAME3(EXPR_PL_VALUE, "left value with pointer") },
    { SPD_VAL_NAME3(EXPR_PR_VALUE, "right value with pointer") },
    { SPD_VAL_NAME3(EXPR_P_REF, "reference") },
    { SPD_VAL_NAME3(EXPR_LR_VALUE, "left and right value") },
    { SPD_VAL_NAME3(EXPR_PLR_VALUE, "left and right value with pointer") },
    { -1, NULL, NULL }
};

/* Label. */
static SpdValName attrLabelSort[] = {
    { SPD_VAL_NAME3(LABEL_NAMED, "named") },
    { SPD_VAL_NAME3(LABEL_CASE, "case") },
    { SPD_VAL_NAME3(LABEL_DEFAULT, "default") },
    { -1, NULL, NULL }
};

/* Constant. */
/* Note, attribute name is not "sort" but "type". */
static SpdValName attrConstantType[] = {
    { SPD_VAL_NAME3(CONST_INT, "integer") },
    { SPD_VAL_NAME3(CONST_FLOAT, "float") },
    { SPD_VAL_NAME3(CONST_CHAR, "char") },
    { SPD_VAL_NAME3(CONST_STRING, "string") },
    { SPD_VAL_NAME3(CONST_ENUM, "enum") },
    { SPD_VAL_NAME3(CONST_WCHAR, "wide character") },
    { -1, NULL, NULL }
};

/* Type. */
static SpdValName attrTypeSort[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, NOT_SPECIFIED) },
    { SPD_VAL_NAME3(TYPE_TYPE_DEF, "typedef") },
    { SPD_VAL_NAME3(TYPE_STRUCT_TAG, "struct") },
    { SPD_VAL_NAME3(TYPE_DECLARATION, "standard") },
    { SPD_VAL_NAME3(TYPE_POINTER, "pointer") },
    { SPD_VAL_NAME3(TYPE_ARRAY, "array") },
    { SPD_VAL_NAME3(TYPE_FUNC, "function") },
    { SPD_VAL_NAME3(TYPE_ENUM_SPECIFIER, "enumerate") },
    { SPD_VAL_NAME3(TYPE_UNION_TAG, "union") },
    { SPD_VAL_NAME3(TYPE_BUILTIN, "builtin") },
    { -1, NULL, NULL }
};

static SpdValName attrTypeQualifier[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, NOT_SPECIFIED) },
    { SPD_VAL_NAME3(TYPE_CONST, "const") },
    { SPD_VAL_NAME3(TYPE_VOLATILE, "volatile") },
    { -1, NULL, NULL }
};

static SpdValName attrTypeType[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, NOT_SPECIFIED) },
    { SPD_VAL_NAME3(TYPE_VOID, "void") },
    { SPD_VAL_NAME3(TYPE_INT, "int") },
    { SPD_VAL_NAME3(TYPE_CHAR, "char") },
    { SPD_VAL_NAME3(TYPE_FLOAT, "float") },
    { SPD_VAL_NAME3(TYPE_DOUBLE, "double") },
    { SPD_VAL_NAME3(TYPE_BUILTIN_TYPE, "builtin type") },
    { -1, NULL, NULL }
};

static SpdValName attrTypeSize[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, NOT_SPECIFIED) },
    { SPD_VAL_NAME3(TYPE_SHORT, "short") },
    { SPD_VAL_NAME3(TYPE_LONG, "long") },
    { SPD_VAL_NAME3(TYPE_LONG_LONG, "long long") },
    { -1, NULL, NULL }
};

static SpdValName attrTypeSign[] = {
    { SPD_VAL_NAME3(TYPE_UNKNOWN, NOT_SPECIFIED) },
    { SPD_VAL_NAME3(TYPE_SIGNED, "signed") },
    { SPD_VAL_NAME3(TYPE_UNSIGNED, "unsigned") },
    { -1, NULL, NULL }
};

/* Operator. */
static SpdValName attrOperatorSort[] = {
    { SPD_VAL_NAME3(OP_L_PAR, "(") },
    { SPD_VAL_NAME3(OP_R_PAR, ")") },
    { SPD_VAL_NAME3(OP_L_ARRAY, "[") },
    { SPD_VAL_NAME3(OP_R_ARRAY, "]") },
    { SPD_VAL_NAME3(OP_ARROW, "->") },
    { SPD_VAL_NAME3(OP_DOT, ".") },
    { SPD_VAL_NAME3(OP_NOT, "!") },
    { SPD_VAL_NAME3(OP_B_NOT, "~") },
    { SPD_VAL_NAME3(OP_INC, "++") },
    { SPD_VAL_NAME3(OP_DEC, "--") },
    { SPD_VAL_NAME3(OP_PLUS, "+") },
    { SPD_VAL_NAME3(OP_MINUS, "-") },
    { SPD_VAL_NAME3(OP_POINTER, "*") },
    { SPD_VAL_NAME3(OP_ADDRESS, "&") },
    { SPD_VAL_NAME3(OP_MUL, "*") },
    { SPD_VAL_NAME3(OP_DIV, "/") },
    { SPD_VAL_NAME3(OP_MOD, "%") },
    { SPD_VAL_NAME3(OP_ADD, "+") },
    { SPD_VAL_NAME3(OP_SUB, "-") },
    { SPD_VAL_NAME3(OP_SHL, "<<") },
    { SPD_VAL_NAME3(OP_SHR, ">>") },
    { SPD_VAL_NAME3(OP_LT, "<") },
    { SPD_VAL_NAME3(OP_LE, "<=") },
    { SPD_VAL_NAME3(OP_GT, ">") },
    { SPD_VAL_NAME3(OP_GE, ">=") },
    { SPD_VAL_NAME3(OP_EQ, "==") },
    { SPD_VAL_NAME3(OP_NE, "!=") },
    { SPD_VAL_NAME3(OP_B_AND, "&") },
    { SPD_VAL_NAME3(OP_B_XOR, "^") },
    { SPD_VAL_NAME3(OP_B_OR, "|") },
    { SPD_VAL_NAME3(OP_L_AND, "&&") },
    { SPD_VAL_NAME3(OP_L_OR, "||") },
    { SPD_VAL_NAME3(OP_ASSIGN, "=") },
    { SPD_VAL_NAME3(OP_A_ADD, "+=") },
    { SPD_VAL_NAME3(OP_A_SUB, "-=") },
    { SPD_VAL_NAME3(OP_A_MUL, "*=") },
    { SPD_VAL_NAME3(OP_A_DIV, "/=") },
    { SPD_VAL_NAME3(OP_A_MOD, "%=") },
    { SPD_VAL_NAME3(OP_A_AND, "&=") },
    { SPD_VAL_NAME3(OP_A_XOR, "^=") },
    { SPD_VAL_NAME3(OP_A_OR, "|=") },
    { SPD_VAL_NAME3(OP_A_SHL, "<<=") },
    { SPD_VAL_NAME3(OP_A_SHR, ">>=") },
    { SPD_VAL_NAME3(OP_COMMA, ",") },
    { SPD_VAL_NAME3(OP_SIZEOF, "sizeof") },
    { SPD_VAL_NAME3(OP_SIZEOF_TYPE, "sizeof(type)") },
    { SPD_VAL_NAME3(OP_CAST, "(type)") },

    /* Added by S.Yamamoto 1997/06/24. */
    { SPD_VAL_NAME3(OP_QUESTION, "?") },
    { SPD_VAL_NAME3(OP_COLON, ":") },
    { SPD_VAL_NAME3(OP_BUILTIN, "__builtin") },
    { SPD_VAL_NAME3(OP_BUILTIN_TYPE, "__builtin(type)") },

    { -1, NULL, NULL }
};

/* P-model. *******************************************************************/

/* Pblock. */
static SpdValName attrPblockSort[] = {
    { SPD_VAL_NAME3(PBLOCK_COMMENT, "comment") },
    { SPD_VAL_NAME3(PBLOCK_MACROCALL, "macrocall") },
    { SPD_VAL_NAME3(PBLOCK_NORMAL, "normal") },
    { SPD_VAL_NAME3(PBLOCK_INCLUDE, "include") },
    { SPD_VAL_NAME3(PBLOCK_DEFINE_FILE, "define_file") },
    { SPD_VAL_NAME3(PBLOCK_DEFINE_OPTION, "define_option") },
    { SPD_VAL_NAME3(PBLOCK_DEFINE_INIT, "define_init") },
    { SPD_VAL_NAME3(PBLOCK_UNDEF, "undef") },
    { SPD_VAL_NAME3(PBLOCK_IF, "if") },
    { SPD_VAL_NAME3(PBLOCK_ELSE, "else") },
    { SPD_VAL_NAME3(PBLOCK_ELIF, "elif") },
    { SPD_VAL_NAME3(PBLOCK_ENDIF, "endif") },
    { SPD_VAL_NAME3(PBLOCK_IFDEF, "ifdef") },
    { SPD_VAL_NAME3(PBLOCK_IFNDEF, "ifndef") },
    { SPD_VAL_NAME3(PBLOCK_PRAGMA, "pragma") },
    { SPD_VAL_NAME3(PBLOCK_LINE, "line") },
    { SPD_VAL_NAME3(PBLOCK_ERROR, "error") },
    { -1, NULL, NULL }
};

/* Hfile. */
static SpdValName attrHfileSort[] = {
    { SPD_VAL_NAME3(HFILE_ANGLE, "<...>") },
    { SPD_VAL_NAME3(HFILE_DQUOT, "\"...\"") },
    { -1, NULL, NULL }
};

/******************************************************************************/

typedef struct {
    int         classId;        /* Class id. */
    SpdString   attrName;       /* Attribute name. */
    SpdValName  *attrArray;     /* Pointer to attrXxxYyy[]. */ 
} AttributeTable;

static AttributeTable attributeTable[] = {
    /* I-model. */
    {SPD_CID_DECLARATION, "sort", attrDeclarationSort},
    {SPD_CID_DECLARATION, "storage", attrDeclarationStorage},
    {SPD_CID_DECLARATOR, "sort", attrDeclaratorSort},
    {SPD_CID_IDENTIFIER, "sort", attrIdentifierSort},
    {SPD_CID_OPTIONALDECL, "sort", attrOptionaldeclSort},
    {SPD_CID_BLOCK, "sort", attrBlockSort},
    {SPD_CID_EXPRESSION, "sort", attrExpressionSort},
    {SPD_CID_EXPRESSION, "lrtype", attrExpressionLrtype},
    {SPD_CID_LABEL, "sort", attrLabelSort},
    {SPD_CID_CONSTANT, "type", attrConstantType},
    {SPD_CID_TYPE, "sort", attrTypeSort},
    {SPD_CID_TYPE, "qualifier", attrTypeQualifier},
    {SPD_CID_TYPE, "type", attrTypeType},
    {SPD_CID_TYPE, "size", attrTypeSize},
    {SPD_CID_TYPE, "sign", attrTypeSign},
    {SPD_CID_OPERATOR, "sort", attrOperatorSort},

    /* P-model. */
    {SPD_CID_PBLOCK, "sort", attrPblockSort},
    {SPD_CID_HFILE, "sort", attrHfileSort},
};

/******************************************************************************/

SpdBoolean smuGetObjAttr(SpdObjId objId, SpdString attrName,
                         SpdString *retAttr, SpdString *retReadable)
{
    int         classId = spdGetClassId(objId);
    int         attrVal = spdGetAttrValInt(objId, attrName);

    return (smuGetClassAttr(classId, attrName, attrVal, retAttr, retReadable));
}

SpdBoolean smuGetClassAttr(int classId, SpdString attrName, int attrVal,
                           SpdString *retAttr, SpdString *retReadable)
{
    int         i, j;
    SpdValName  *attr;

    for (i = 0; i < SPD_ARRAY_SIZE(attributeTable); i++) {
        if (classId == attributeTable[i].classId &&
            SPD_STREQ(attrName, attributeTable[i].attrName)) {
            attr = attributeTable[i].attrArray;
            for (j = 0; attr[j].name != NULL; j++) {
                if (attr[j].value == attrVal) {
                    /* Found. */
                    if (retAttr != NULL) {
                        /* Discards "const". */
                        *retAttr = (SpdString)attr[j].name;
                    }
                    if (retReadable != NULL) {
                        /* Discards "const". */
                        *retReadable = (SpdString)attr[j].str;
                    }
                    return (TRUE);
                }
            }
            /* Invalid attribute value. */
            /* *retAttr = NULL; */
            if (retAttr != NULL) {
                *retAttr = "Invalid attribute value";
            }
            if (retReadable != NULL) {
                *retReadable = NULL;
            }
            return (FALSE);
        }
    }
    /* Invalid attribute name. */
    /* *retAttr = NULL; */
    if (retAttr != NULL) {
        *retAttr = "Invalid attribute name";
    }
    if (retReadable != NULL) {
        *retReadable = NULL;
    }
    return (FALSE);
}

/******************************************************************************/

static SpdString RelationName[] = {
    "blk_blk", "blk_decl", "blk_expr", "blk_label", "case_cexpr", 
    "const_ref", "decl_decl", "decl_ident", "decl_opt", "decl_type", 
    "expr_expr", "expr_op", "expr_type", "file_decl", "func_body", 
    "func_parm", "goto_labl", "ident_ref", "ident_type", "init_expr", 
    "op_type", "prog_file", "prot_type", "struct_decl", "type_cexpr", 
    "type_is", "type_st",

    /* Added by S.Yamamoto 1997/06/24. */
    "prot_parm", "funcptr_parm", "decr_type",

    NULL
};

static SpdString RelationPid[] = {
    "parent_block_id", "parent_block_id", "parent_block_id", 
    "block_id", "label_id", "expression_id", "declaration_id", 
    "declarator_id", "declaration_id", "declaration_id", 
    "outer_expression_id", "expression_id", "expression_id", "file_id", 
    "function_id", "optionaldecl_id", "expression_id", "expression_id", 
    "identifier_id", "optionaldecl_id", "operator_id", "program_id", 
    "optionaldecl_id", "optionaldecl_id", "type_id", "type_id", 
    "type_pointer_id",

    /* Added by S.Yamamoto 1997/06/24. */
    "optionaldecl_id", "optionaldecl_id", "declarator_id",

    NULL
};

static SpdString RelationCid[] = {
    "children_block_id", "children_declaration_id", 
    "children_expression_id", "label_id", "expression_id", "constant_id", 
    "declarator_id", "identifier_id", "optionaldecl_id", "type_id", 
    "inner_expression_id", "operator_id", "type_id", "declaration_id", 
    "block_id", "declaration_id", "lebel_id", "identifier_id", "type_id", 
    "expression_id", "type_id", "file_id", "type_id", "declaration_id", 
    "expression_id", "identifier_id", "type_pointed_id",

    /* Added by S.Yamamoto 1997/06/24. */
    "declaration_id", "declaration_id", "type_id",

    NULL
};

SpdString smuGetRelParentAttrName(SpdRelId relId)
{
    int     i;
    
    for (i = 0; RelationName[i] != NULL; i++) {
        if (SPD_STREQ(RelationName[i], spdGetClassName(relId))) {
            return (RelationPid[i]);
        }
    }
    return (NULL);
}

SpdString smuGetRelChildAttrName(SpdRelId relId)
{
    int     i;
    
    for (i = 0; RelationName[i] != NULL; i++) {
        if (SPD_STREQ(RelationName[i], spdGetClassName(relId))) {
            return (RelationCid[i]);
        }
    }
    return (NULL);
}
