/* 
 * Program:             $RCSfile: spdConvFileId.c,v $  $Revision: 31.12 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu  1999/08/29
 *                      H.Ohhashi  1999/08/29
 *                      S.Yamamoto  1999/08/29
 * 
 * (C) Copyright:       N.Fukuyasu, H.Ohhashi and S.Yamamoto  1999 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdConvFileId.c,v 31.12 2003/06/28 15:31:24 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

typedef struct _SpdSrcFileTblElem {
    SpdObjId    l_file_id;
    SpdObjId    r_file_id;
} SpdSrcFileTblElem;

typedef struct _SpdSrcFileTbl {
    size_t              size;
    SpdSrcFileTblElem   *elem;
} SpdSrcFileTbl;

static SpdSrcFileTbl    cfile_ifile;
static SpdSrcFileTbl    pfile_file;
static SpdString    current_sdb_dir = NULL;

static void freeSrcFileTbl(SpdSrcFileTbl *tbl)
{
    if (tbl->size) {
        spdFree(tbl->elem);
    }
    tbl->size = 0;
    tbl->elem = NULL;
}

static void freeAllSrcFileTbls(void)
{
    freeSrcFileTbl(&cfile_ifile);
    freeSrcFileTbl(&pfile_file);

    if (current_sdb_dir) {
        spdFree(current_sdb_dir);
        current_sdb_dir = NULL;
    }
}

/*
 * "SDB/PIDB/cfile_ifile"ɤ߹ࡥ
 */

static void readSrcFileTbl(SpdString dir, SpdString tblFile,
                           SpdSrcFileTbl *tbl)
{
    FILE        *fp;
    char        buf[SPD_LINE_LEN + 1];
    SpdObjId    l_file_id, r_file_id;

    if (!current_sdb_dir) current_sdb_dir = spdStrdup(spdGetSDBdir());
 
    /* SDBΥǥ쥯ȥ꤬ѹ줿Ƚꡥ*/
    if (!SPD_STREQ(current_sdb_dir, spdGetSDBdir())) {
        freeAllSrcFileTbls();
        current_sdb_dir = spdStrdup(spdGetSDBdir());
    }

    if (tbl->size != 0) return;
    
    fp = spdOpenSDBfile(dir, tblFile, SAPID_O_READ);
    while (fgets(buf, SPD_LINE_LEN, fp) != NULL) {
        sscanf(buf, "%d %d", &l_file_id, &r_file_id);
        tbl->elem = (SpdSrcFileTblElem *)
            spdRealloc(tbl->elem, sizeof(SpdSrcFileTblElem) *
                       (tbl->size + 1));

        tbl->elem[tbl->size].l_file_id = l_file_id;
        tbl->elem[tbl->size].r_file_id = r_file_id;
        tbl->size++;
    }
    spdFclose(fp);
    atexit(freeAllSrcFileTbls);
}

/*
 * "SDB/PIDB/cfile_ifile"˴Ťơ".i"б".c"fileId롥
 * "cfile_ifile"ϰԤ"cfileId ifileId"ҤƤ롥
 */

SpdObjId spdConvFileIdItoC(SpdObjId ifileId)
{
    int     i;

    readSrcFileTbl(PIDB_DIR, PIDB_CFILE_IFILE, &cfile_ifile);

    for (i = 0; i < cfile_ifile.size; i++) {
        if (cfile_ifile.elem[i].r_file_id == ifileId) {
            return (cfile_ifile.elem[i].l_file_id);
        }
    }
    return (SPD_NON_ID);
}

/*
 * "SDB/PIDB/cfile_ifile"˴Ťơ".c"б".i"fileId롥
 * "cfile_ifile"ϰԤ"cfileId ifileId"ҤƤ롥
 */

SpdObjId spdConvFileIdCtoI(SpdObjId cfileId)
{
    int     i;

    readSrcFileTbl(PIDB_DIR, PIDB_CFILE_IFILE, &cfile_ifile);

    for (i = 0; i < cfile_ifile.size; i++) {
        if (cfile_ifile.elem[i].l_file_id == cfileId) {
            return (cfile_ifile.elem[i].r_file_id);
        }
    }
    return (SPD_NON_ID);
}

/*
 * "SDB/SRC/pfile_file"˴ŤpfileIdбfileId롥
 * "pfile_file"ϰԤ"pfileId fileId"ҤƤ롥
 */

SpdObjId spdConvIdPfileToFile(SpdObjId pfileId)
{
    int     i;

    switch (spdGetClassId(pfileId)) {
      case SPD_CID_FILE:
        /* file饹IdξϤΤޤ֤*/
        return pfileId;
        break;
      case SPD_CID_PFILE:
        readSrcFileTbl(SAPID_SRC_DIR, SRC_PFILE_FILE, &pfile_file);
        for (i = 0; i < pfile_file.size; i++) {
            if (pfile_file.elem[i].l_file_id == pfileId) {
                return (pfile_file.elem[i].r_file_id);
            }
        }
        break;
      default:
        break;
    }
    return (SPD_NON_ID);
}

/*
 * ե̾fileId롥
 * Ʊ̾Υե̾ϰĤ¸ߤʤȲꤷƤ롥
 */

SpdObjId smuGetFileIdByName(SpdConstString fileName)
{
    return (spdGetAnObjId("file", fileName));
}
