/* 
 * Program:             $RCSfile: spdTraverse.c,v $  $Revision: 30.16 $
 * 
 * Purpose:             spdTraverse.
 * 
 * Author:              H.Suzuki  1998/02/04
 *                      S.Yamamoto  1998/02/08
 * Modified:            N.Iwamoto  1998/06/25
 * 
 * (C) Copyright:       H.Suzuki, N.Iwamoto and S.Yamamoto  1998 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdTraverse.c,v 30.16 2003/07/06 10:09:03 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/***** SpdStack *****/

static SpdStack *spdCreateStack(size_t elemSize)
{
    SpdStack *stk = (SpdStack *)spdMalloc(sizeof(SpdStack));

    stk->size = 0;
    stk->elemSize = elemSize;
    stk->elem = NULL;

    return stk;
}

static void spdFreeStack(SpdStack *stk)
{
    int i;

    for (i = 0; i < stk->size; i++) spdFree(stk->elem[i]);
    spdFree(stk->elem);
    spdFree(stk);
}

static void spdPushStack(void *elem, SpdStack *stk)
{
    stk->size++;
    stk->elem = (void **)spdRealloc(stk->elem, stk->elemSize * stk->size);
    stk->elem[stk->size - 1] = elem;
}

static void *spdPopStack(SpdStack *stk)
{
    void *retElem;

    if (stk->size == 0) return NULL;

    retElem = stk->elem[stk->size - 1];
    stk->size--;

    return retElem;
}

static void *spdGetTopOfStack(SpdStack *stk)
{
    if (stk->size < 1) return NULL;
    else return (stk->elem[stk->size - 1]);
}

/***** SpdTrvStackElem *****/

static SpdTrvStackElem *createTrvStackElem(SpdObjId objId, int depth, int num)
{
    SpdTrvStackElem *elem;

    elem = (SpdTrvStackElem *)spdMalloc(sizeof(SpdTrvStackElem));
    elem->objectId = objId;
    elem->type = SAPID_TRV_ACT_BEFORE;
    elem->actLv = SAPID_TRV_CALLED_NOTHING;
    elem->depth = depth;
    elem->num = num;

    return elem;
}

SpdStack *spdTrvCreateStack(void)
{
    return (spdCreateStack(sizeof(SpdTrvStackElem *)));
}

/***** ɲѴؿ *****/

void
spdTrvAddRel(SpdTrvActInfo *info, SpdClassId id)
{
    spdAddElem(info->trvRelSet, id, NULL);
}

void
spdTrvAddSortFunc(SpdTrvActInfo *info, SpdClassId id, SpdCompareRelFunc func)
{
    spdAddElem(info->sortFuncSet, id, (Contents)&func);
}

SpdTrvActRelFunc
spdTrvAddActFuncOfRelClass(SpdTrvActInfo *info, SpdClassId id,
                           SpdTrvActType atype, SpdTrvActRelFunc func)
{
    SpdSet              set;
    SpdTrvActRelFunc    old_func;

    switch (atype) {
      case SAPID_TRV_ACT_BEFORE:
        set = info->classBeforeRelFuncSet;
        break;
      case SAPID_TRV_ACT_AFTER:
        set = info->classAfterRelFuncSet;
        break;
      default:
        return NULL;
    }
    old_func = (SpdTrvActRelFunc)spdSearchElem(set, id);
    if (old_func) {
        spdDelElem(set, id);
    }
    spdAddElem(set, id, (Contents)&func);
    return (old_func);
}

SpdTrvActFunc
spdTrvAddActFuncOfClass(SpdTrvActInfo *info, SpdClassId id, 
                        SpdTrvActType atype, SpdTrvActFunc func)
{
    SpdSet          set;
    SpdTrvActFunc   old_func;

    switch (atype) {
      case SAPID_TRV_ACT_BEFORE:
        set = info->classBeforeFuncSet;
        break;
      case SAPID_TRV_ACT_AFTER:
        set = info->classAfterFuncSet;
        break;
      default:
        return NULL;
    }
    old_func = (SpdTrvActFunc)spdSearchElem(set, id);
    if (old_func) {
        spdDelElem(set, id);
    }
    spdAddElem(set, id, (Contents)&func);
    return (old_func);
}

SpdTrvActMidFunc
spdTrvAddActFuncMidOfClass(SpdTrvActInfo *info, SpdClassId id,
                           SpdTrvActMidFunc func)
{
    SpdSet              set;
    SpdTrvActMidFunc    old_func;

    set = info->classMiddleFuncSet;
    old_func = (SpdTrvActMidFunc)spdSearchElem(set, id);
    if (old_func) {
        spdDelElem(set, id);
    }
    spdAddElem(set, id, (Contents)&func);
    return (old_func);
}

SpdTrvActFunc
spdTrvAddActFuncOfObj(SpdTrvActInfo *info, SpdObjId id, 
                      SpdTrvActType atype, SpdTrvActFunc func)
{
    SpdSet          set;
    SpdTrvActFunc   old_func;

    switch (atype) {
      case SAPID_TRV_ACT_BEFORE:
        set = info->objBeforeFuncSet;
        break;
      case SAPID_TRV_ACT_AFTER:
        set = info->objAfterFuncSet;
        break;
      default:
        return NULL;
    }
    old_func = (SpdTrvActFunc)spdSearchElem(set, id);
    if (old_func) {
        spdDelElem(set, id);
    }
    spdAddElem(set, id, (Contents)&func);
    return (old_func);
}

/***** Ѵؿ *****/

SpdTrvActInfo *spdTrvInitActInfo(void)
{
    SpdTrvActInfo *info = (SpdTrvActInfo *)spdMalloc(sizeof(SpdTrvActInfo));
    
    info->classBeforeRelFuncSet = spdCreateSet(sizeof(SpdTrvActRelFunc));
    info->classAfterRelFuncSet = spdCreateSet(sizeof(SpdTrvActRelFunc));
    info->classBeforeFuncSet = spdCreateSet(sizeof(SpdTrvActFunc));
    info->classMiddleFuncSet = spdCreateSet(sizeof(SpdTrvActMidFunc));
    info->classAfterFuncSet = spdCreateSet(sizeof(SpdTrvActFunc));
    info->objBeforeFuncSet = spdCreateSet(sizeof(SpdTrvActFunc));
    info->objAfterFuncSet = spdCreateSet(sizeof(SpdTrvActFunc));
    info->sortFuncSet = spdCreateSet(sizeof(SpdTrvActFunc));
    info->trvRelSet = spdCreateSet(DEFAULT_INT_SET);
    info->trvStack = spdTrvCreateStack();
    
    return info;
}

void spdTrvFreeActInfo(SpdTrvActInfo *info)
{
    spdFreeSet(info->classBeforeRelFuncSet);
    spdFreeSet(info->classAfterRelFuncSet);
    spdFreeSet(info->classBeforeFuncSet);
    spdFreeSet(info->classMiddleFuncSet);
    spdFreeSet(info->classAfterFuncSet);
    spdFreeSet(info->objBeforeFuncSet);
    spdFreeSet(info->objAfterFuncSet);
    spdFreeSet(info->sortFuncSet);
    spdFreeSet(info->trvRelSet);
    spdFreeStack(info->trvStack);
    spdFree(info);
}

/***** Ѵؿ *****/

SpdTrvActRelFunc
spdTrvGetActFuncOfRelClass(SpdTrvActInfo *info, SpdClassId id,
                           SpdTrvActType atype)
{
    SpdTrvActRelFunc *f;
    SpdSet set;

    switch (atype) {
      case SAPID_TRV_ACT_BEFORE:
        set = info->classBeforeRelFuncSet;
        break;
      case SAPID_TRV_ACT_AFTER:
        set = info->classAfterRelFuncSet;
        break;
      default:
        return NULL;
    }

    if (info == NULL) return NULL;
    f = (SpdTrvActRelFunc *)spdSearchElem(set, id);

    if (f)
        return *f;
    else
        return NULL;
}

SpdTrvActFunc
spdTrvGetActFuncOfClass(SpdTrvActInfo *info, SpdClassId id,
                        SpdTrvActType atype)
{
    SpdTrvActFunc *f;
    SpdSet set;

    switch (atype) {
      case SAPID_TRV_ACT_BEFORE:
        set = info->classBeforeFuncSet;
        break;
      case SAPID_TRV_ACT_AFTER:
        set = info->classAfterFuncSet;
        break;
      default:
        return NULL;
    }

    if (info == NULL) return NULL;
    f = (SpdTrvActFunc *)spdSearchElem(set, id);

    if (f)
        return *f;
    else
        return NULL;
}

SpdTrvActMidFunc
spdTrvGetActFuncMidOfClass(SpdTrvActInfo *info, SpdClassId id)
{
    SpdTrvActMidFunc *f;
    SpdSet set;

    set = info->classMiddleFuncSet;

    if (info == NULL) return NULL;
    f = (SpdTrvActMidFunc *)spdSearchElem(set, id);

    if (f)
        return *f;
    else
        return NULL;
}

SpdTrvActFunc
spdTrvGetActFuncOfObj(SpdTrvActInfo *info, SpdObjId id, SpdTrvActType atype)
{
    SpdTrvActFunc *f;

    if (info == NULL) return NULL;
    f = (SpdTrvActFunc *)spdSearchElem((atype == SAPID_TRV_ACT_BEFORE) ? 
                                       info->objBeforeFuncSet : 
                                       info->objAfterFuncSet, id);

    if (f)
        return *f;
    else
        return NULL;
}

/***** ȥСѴؿ *****/

static void popAndFreeElem(SpdTrvActInfo *info)
{
    SpdTrvStackElem *elem;

    elem = (SpdTrvStackElem *)spdPopStack(info->trvStack);
    spdFree(elem);
}

static void popSameDepthElem(int depth, SpdTrvActInfo *info)
{
    SpdTrvStackElem *elem;

    elem = (SpdTrvStackElem *)spdGetTopOfStack(info->trvStack);
    while (elem->depth == depth) {
        elem = (SpdTrvStackElem *)spdPopStack(info->trvStack);
        spdFree(elem);
        elem = (SpdTrvStackElem *)spdGetTopOfStack(info->trvStack);
    }
}

static SpdBoolean isLastElemInSameDepth(SpdStack *trvStack)
{
    SpdTrvStackElem *elem1, *elem2;

    if (trvStack->size > 1) {
        elem1 = (SpdTrvStackElem *)trvStack->elem[trvStack->size - 1];
        elem2 = (SpdTrvStackElem *)trvStack->elem[trvStack->size - 2];
        if (elem1->depth != elem2->depth) return TRUE;
        else return FALSE;
    } else return TRUE;
}

static void traverseStack(SpdTrvActInfo *info, void *vp)
{
    SpdTrvActFunc actFunc;
    SpdTrvActRelFunc actFuncOfRelClass;
    SpdTrvActMidFunc actFuncMidOfClass;
    SpdCompareRelFunc *psortfunc, sortfunc;
    SpdRelArray relArr;
    SpdRel rel;
    SpdObjId objId, classId, relClassId, pClassId;
    SpdRelId relId;
    SpdTrvStackElem *elem, *cElem;
    int i, res, depth = 0, num = 0;

    while (info->trvStack->size > 0) {
        elem = (SpdTrvStackElem *)spdGetTopOfStack(info->trvStack);
        objId = elem->objectId;
        classId = spdGetClassId(objId);
        /* depthѤ numꤹ */
        if (elem->depth != depth) {
            depth = elem->depth;
            num = 0;
        } else if (elem->type == SAPID_TRV_ACT_BEFORE) num++;
        if (elem->num == -1) elem->num = num;
        else num = elem->num;

        /*** ϢéΥХåؿμ¹ ***/
        if (elem->actLv < SAPID_TRV_CALLED_BEFORE_REL
            && elem->type == SAPID_TRV_ACT_BEFORE
            && elem->pObjId != SPD_NON_ID) {
            relId = elem->rel.relationId;
            relClassId = spdGetClassId(relId);
            actFuncOfRelClass = spdTrvGetActFuncOfRelClass(info, relClassId, 
                                                           SAPID_TRV_ACT_BEFORE);
            if (actFuncOfRelClass != NULL) {
                elem->actLv = SAPID_TRV_CALLED_BEFORE_REL;
                res = actFuncOfRelClass(elem->pObjId, elem->rel, vp);
                if (res == SAPID_TRV_STOP) {
                    /* éʤ => Ʊdepthelemݥå */
                    if (elem->pObjId == SPD_NON_ID) {
                        popSameDepthElem(depth, info);
                        break;
                    } else {
                        popSameDepthElem(depth, info);
                        continue;
                    }
                } else if (res == SAPID_TRV_SKIP) {
                    if (elem->pObjId == SPD_NON_ID) {
                        popAndFreeElem(info);
                        break;
                    } else {
                        popAndFreeElem(info);
                        continue;
                    }
                } else if (res == SAPID_TRV_PAUSE) {
                    /*  */
                    return;
                }
            }
        }

        /*** BEFORE Хåؿμ¹ ***/
        if (elem->type == SAPID_TRV_ACT_BEFORE
            && elem->actLv < SAPID_TRV_CALLED_BEFORE_OBJ) {
            actFunc = spdTrvGetActFuncOfObj(info, objId, SAPID_TRV_ACT_BEFORE);
            if (actFunc == NULL)
                actFunc = spdTrvGetActFuncOfClass(info, classId,
                                                  SAPID_TRV_ACT_BEFORE);
            if (actFunc != NULL) {
                elem->actLv = SAPID_TRV_CALLED_BEFORE_OBJ;
                res = actFunc(objId, vp);
                if (res == SAPID_TRV_STOP) {
                    if (elem->pObjId == SPD_NON_ID) {
                        popAndFreeElem(info);
                        break;
                    } else {
                        popAndFreeElem(info);
                        continue;
                    }
                } else if (res == SAPID_TRV_SKIP) {
                    /* ʬȤμηpop */
                    if (elem->pObjId == SPD_NON_ID) {
                        popAndFreeElem(info);
                        break;
                    } else {
                        popAndFreeElem(info);
                        elem = (SpdTrvStackElem *)
                            spdGetTopOfStack(info->trvStack);
                        if (elem->depth == depth) popAndFreeElem(info);
                        continue;
                    }
                } else if (res == SAPID_TRV_PAUSE) {
                    /*  */
                    return;
                }
            }
        }

        /*** AFTER Хåؿμ¹ ***/
        if (elem->type == SAPID_TRV_ACT_AFTER
            && elem->actLv < SAPID_TRV_CALLED_AFTER_OBJ) {
            actFunc = spdTrvGetActFuncOfObj(info, objId, SAPID_TRV_ACT_AFTER);
            if (actFunc == NULL)
                actFunc = spdTrvGetActFuncOfClass(info, classId,
                                                  SAPID_TRV_ACT_AFTER);
            if (actFunc != NULL) {
                elem->actLv = SAPID_TRV_CALLED_AFTER_OBJ;
                res = actFunc(objId, vp);
                if (res == SAPID_TRV_STOP) {
                    if (elem->pObjId == SPD_NON_ID) {
                        popAndFreeElem(info);
                        break;
                    } else {
                        popAndFreeElem(info);
                        continue;
                    }
                } else if (res == SAPID_TRV_SKIP) {
                    /* ʬȤμηpop */
                    if (elem->pObjId == SPD_NON_ID) {
                        popAndFreeElem(info);
                        break;
                    } else {
                        popAndFreeElem(info);
                        elem = (SpdTrvStackElem *)
                            spdGetTopOfStack(info->trvStack);
                        if (elem->depth == depth) popAndFreeElem(info);
                        continue;
                    }
                } else if (res == SAPID_TRV_PAUSE) {
                    /*  */
                    return;
                }
            }
        }
        
        /*** ϢéäΥХåؿμ¹ ***/
        if (elem->actLv < SAPID_TRV_CALLED_AFTER_REL
            && elem->type == SAPID_TRV_ACT_AFTER
            && elem->pObjId != SPD_NON_ID) {
            relId = elem->rel.relationId;
            relClassId = spdGetClassId(relId);
            actFuncOfRelClass = spdTrvGetActFuncOfRelClass(info, relClassId,
                                                           SAPID_TRV_ACT_AFTER);
            if (actFuncOfRelClass != NULL) {
                actFuncOfRelClass(elem->pObjId, elem->rel, vp);
                res = elem->actLv = SAPID_TRV_CALLED_AFTER_REL;
                if (res == SAPID_TRV_PAUSE) {
                    /*  */
                    return;
                }
            }
        }

        /*** ϢȴϢδ֤ΥХåؿμ¹ ***/
        if (elem->actLv < SAPID_TRV_CALLED_MID_REL
            && elem->type == SAPID_TRV_ACT_AFTER
            && elem->pObjId != SPD_NON_ID
            && !isLastElemInSameDepth(info->trvStack)) {
            /* ƤclassIdбեåؿƤ */
            pClassId = spdGetClassId(elem->pObjId);
            actFuncMidOfClass = spdTrvGetActFuncMidOfClass(info, pClassId);
            if (actFuncMidOfClass != NULL) {
                elem->actLv = SAPID_TRV_CALLED_MID_REL;
                res = actFuncMidOfClass(elem->pObjId, elem->num, vp);
                if (res == SAPID_TRV_STOP) {
                    /* éʤ => Ʊdepthelemݥå */
                    if (elem->pObjId == SPD_NON_ID) {
                        popSameDepthElem(depth, info);
                        break;
                    } else {
                        popSameDepthElem(depth, info);
                        continue;
                    }
                } else if (res == SAPID_TRV_PAUSE) {
                    /*  */
                    return;
                }
            }
        }

        /*** åŸ ***/
        if (elem->type == SAPID_TRV_ACT_BEFORE) {
            /* type ѹ */
            elem->type = SAPID_TRV_ACT_AFTER;

            /* sort ؿλ (ꤵƤʤ offset  sort) */
            psortfunc = spdSearchElem(info->sortFuncSet, classId);
            if (psortfunc == NULL) sortfunc = spdCompareRelOffset;
            else sortfunc = *psortfunc;

            /* objId ȴϢΤҤ򤹤٤Ƽ */
            relArr = spdGetRelSortedArray(objId, "any", "p-id",
                                          NULL, sortfunc);
            /* éս˻Ҥ򥹥åѤ */
            for (i = relArr.size - 1; i >= 0; i--) {
                rel = relArr.rel[i];
                /* ꤷϢʳϥå */
                if (spdSearchElem(info->trvRelSet,
                                  spdGetClassId(rel.relationId)) == NULL)
                    continue;
                /* åפʬ뤿ᤳǤnumǤʤ
                   numϤϤ-1ǽ, ꤵ */
                cElem = createTrvStackElem(rel.objectId, depth + 1, -1);
                cElem->pObjId = elem->objectId;
                cElem->rel = rel;
                spdPushStack(cElem, info->trvStack);
            }
            spdFreeRelArray(relArr);
        } else {
            /* é꽪äΤǤ򥹥åݥåפƲ */
            if (elem->pObjId == SPD_NON_ID) {
                popAndFreeElem(info);
                break;
            } else
                popAndFreeElem(info);
        }
    }
}


/***** ȥСؿ *****/
/* id Υ֥Ȥ info ˽äƻҤΥ֥Ȥé */

void spdTraverse(SpdObjId id, SpdTrvActInfo *info, void *vp)
{
    SpdTrvStackElem *elem;

    if (info != NULL) {
        /* depth = 1ΥåѤ */
        elem = createTrvStackElem(id, 1, 0);
        elem->pObjId = SPD_NON_ID;
        spdPushStack(elem, info->trvStack);

        traverseStack(info, vp);
    }
}

/* info  trvStack ˽äƻҤΥ֥Ȥé */
/* SAPID_TRV_PAUSE Ʊ info ǸƤӽФ */

void spdTraverseContinue(SpdTrvActInfo *info, void *vp)
{
    if (info != NULL) {
        traverseStack(info, vp);
    }
}
