/* 
 * Program:             $RCSfile: xrefTable.c,v $  $Revision: 30.6 $
 * 
 * Purpose:
 * 
 * Author:              A.Watanabe  1996/12/21
 *                      S.Yamamoto  1997/01/26
 * 
 * (C) Copyright:       A.Watanabe and S.Yamamoto  1996 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: xrefTable.c,v 30.6 2003/06/28 15:31:24 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DEBUG */

#if !defined(SMU_XREF_USE_TRAVERSE)

static SmuXrefTable makeXrefTableSub(SmuXrefTable xref_table,
                                     SpdObjId caller_identId,
                                     SpdOccArray callee_ident_buf,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId));

static int spdCompareDeclOccName(const SpdOcc *a, const SpdOcc *b);

SpdBoolean spdOccIsDeclFuncdef(SpdOcc occ);
SpdBoolean spdOccIsIdFunction(SpdOcc occ);
SpdBoolean spdOccIsDeclProto(SpdOcc occ);

SmuXrefTables  smuMakeXrefTables(SpdObjId prog_id, SmuXrefFlag flag)
{
    int i, j;
    SpdOccArray decln_buf, callee_ident_buf;
    SpdObjId caller_declrId, caller_identId, caller_opt;
    SmuXrefTables xreftables = {{0,NULL},{0,NULL},{0,NULL},{0,NULL},{0,NULL}};
    /* int mode; */

    /* mode = spdGetSearchFlag(); */
    /* spdSetSearchFlag(SPD_SEARCH_ANY); */
    decln_buf = spdGetIncludedOccSortedArray(prog_id, "declaration",
                                             SPD_NON_ID,
                                             spdOccIsDeclFuncdef,
                                             spdCompareDeclOccName);

    for (i = 0; i < decln_buf.size; i++) {
        caller_declrId = spdGetARelObj(decln_buf.occ[i].objectId,
                                       "decl_decl", "any");
        caller_identId = spdGetARelObj(caller_declrId, "decl_ident", "any");
        
        caller_opt = spdGetARelObj(caller_declrId, "decl_opt", "any");
        
        callee_ident_buf = spdGetIncludedOccSortedArray(caller_opt,
                                                        "identifier",
                                                        SPD_NON_ID,
                                                        NULL,
                                                        spdCompareOccName);
#ifdef DEBUG
        fprintf(stderr, "+--funcdef%d: %s(%d)\n",
                i, spdGetName(caller_identId), caller_identId);
#endif
        
        xreftables.funcCallTable =
            makeXrefTableSub(xreftables.funcCallTable,
                             caller_identId, callee_ident_buf, flag,
                             spdObjIdIsIdFunction);
        xreftables.variableTable =
            makeXrefTableSub(xreftables.variableTable,
                             caller_identId, callee_ident_buf, flag,
                             spdObjIdIsIdVariable);
        xreftables.globalVariableTable =
            makeXrefTableSub(xreftables.globalVariableTable,
                             caller_identId, callee_ident_buf, flag,
                             smuIsGlobalVariable);
        xreftables.localVariableTable =
            makeXrefTableSub(xreftables.localVariableTable,
                             caller_identId, callee_ident_buf, flag,
                             smuIsLocalVariable);

        spdFreeOccArray(callee_ident_buf);

        /* ޥ̽ */
        callee_ident_buf = spdGetIncludedOccArray(caller_opt,
                                                  "macrocall",
                                                  SPD_NON_ID,
                                                  NULL);

        for (j = 0; j < callee_ident_buf.size; j++) {
            callee_ident_buf.occ[j].objectId =
                spdGetARelObj(callee_ident_buf.occ[j].objectId,
                              "macro_call", "any");
        }
        SPD_QSORT(callee_ident_buf.occ, callee_ident_buf.size, 
                    sizeof(SpdOcc), spdCompareOccName);

        xreftables.macroCallTable =
            makeXrefTableSub(xreftables.macroCallTable,
                             caller_identId, callee_ident_buf, flag,
                             NULL);
    
        spdFreeOccArray(callee_ident_buf);
    }
    spdFreeOccArray(decln_buf);

    /* spdSetSearchFlag(mode); */
    return xreftables;
}

static SmuXrefTable makeXrefTableSub(SmuXrefTable xref_table,
                                     SpdObjId caller_identId,
                                     SpdOccArray callee_ident_buf,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId))
{
    SpdSet  callee_set = SAPID_EMPTY_SET;
    int     i;

    for (i = 0; i < callee_ident_buf.size; i++) {
        if ((cond != NULL) &&
            !cond(callee_ident_buf.occ[i].objectId)) continue;

        if(flag) {
            if (spdSetContains(callee_set, callee_ident_buf.occ[i].objectId)) {
                continue;
            } else {
                spdAddElem(callee_set, callee_ident_buf.occ[i].objectId, NULL);
            }
        }

        xref_table.elem = spdRealloc(xref_table.elem,
                                     sizeof(SmuXrefTableElem) *
                                     (xref_table.size + 1));
        
#ifdef DEBUG
        fprintf(stderr, "|  +-- %d : %s(%d) -> %s(%d)\n", i,
                spdGetName(caller_identId), caller_identId,
                spdGetName(callee_ident_buf.occ[i].objectId),
                callee_ident_buf.occ[i].objectId);
#endif
    
        xref_table.elem[xref_table.size].callerId
            = caller_identId;
        xref_table.elem[xref_table.size].calleeId
            = callee_ident_buf.occ[i].objectId;
        xref_table.elem[xref_table.size].relationId
            = callee_ident_buf.occ[i].relationId;
        xref_table.size++;
    }
    spdFreeSet(callee_set);

    return xref_table;
}

void smuFreeXrefTables(SmuXrefTables tables)
{
    spdFree(tables.funcCallTable.elem);
    spdFree(tables.variableTable.elem);
    spdFree(tables.globalVariableTable.elem);
    spdFree(tables.localVariableTable.elem);
    spdFree(tables.macroCallTable.elem);
}

static int spdCompareDeclOccName(const SpdOcc *a, const SpdOcc *b)
{
    SpdObjId declr_a, declr_b, ident_a, ident_b;

    declr_a = spdGetARelObj(a->objectId, "decl_decl", "any");
    ident_a = spdGetARelObj(declr_a, "decl_ident", "any");

    declr_b = spdGetARelObj(b->objectId, "decl_decl", "any");
    ident_b = spdGetARelObj(declr_b, "decl_ident", "any");

    return (spdSafetyStrcmp(spdGetName(ident_a), spdGetName(ident_b)));
}

#endif /* !defined(SMU_XREF_USE_TRAVERSE) */

/******************************************************************************/

/*
 * Common functions for "xrefTable.c" and "xrefTableTraverse.c".
 */

SpdBoolean spdOccIsIdFunction(SpdOcc occ)
{
    return (spdGetAttrValInt(occ.objectId, "sort") == ID_FUNCTION);
}

SpdBoolean spdOccIsDeclFuncdef(SpdOcc occ)
{
    return (spdGetAttrValInt(occ.objectId, "sort") == DECL_FUNCDEF);
}

SpdBoolean spdOccIsDeclProto(SpdOcc occ)
{
    return (spdGetAttrValInt(occ.objectId, "sort") == DECL_PROTOTYPE);
}
