/* 
 * Program:             $RCSfile: xrefTableTraverse.c,v $  $Revision: 30.11 $
 * 
 * Purpose:
 * 
 * Author:              A.Watanabe  1996/12/21
 *                      S.Yamamoto  1997/01/26
 * 
 * (C) Copyright:       A.Watanabe and S.Yamamoto  1996 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: xrefTableTraverse.c,v 30.11 2003/06/28 15:31:24 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DEBUG */

#if defined(SMU_XREF_USE_TRAVERSE)

typedef struct _CallerBufElem {
    SpdObjId declnId;
    SpdObjId declrId;
    SpdObjId identId;
    SpdObjId optId;
    SpdObjId fileId;
    int offset;
    int length;       
} CallerBufElem;

typedef struct _CallerBuf {
    SpdObjId fileId;    
    int size;
    CallerBufElem *elem;
} CallerBuf;

typedef struct _CalleeBufElem {
    SpdObjId fileId;    
    SpdObjId callerId;
    SpdObjId calleeId;
    SpdRelId relationId;
} CalleeBufElem;

typedef struct _CalleeBuf {
    SpdObjId        fileId;    
    SpdObjId        callerId;
    SpdRelId        relationId;
    int             size;
    CalleeBufElem   *elem;
} CalleeBuf;

static SpdTrvActFunc beforeFile(SpdObjId id, CallerBuf *caller);

static SmuXrefTable makeXrefTableSub(SmuXrefTable xref_table,
                                     CalleeBuf callee,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId));

static SmuXrefTable makeXrefTableMacro(SmuXrefTable xref_table,
                                     SpdObjId caller_identId,
                                     SpdOccArray callee_ident_buf,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId));
    
SpdBoolean spdOccIsDeclFuncdef(SpdOcc occ);
SpdBoolean spdOccIsIdFunction(SpdOcc occ);
SpdBoolean spdOccIsDeclProto(SpdOcc occ);

static SpdTrvActFunc beforeFuncdef(SpdObjId id, CallerBuf *vp);

static int cmpCallerName(const CallerBufElem *a, const CallerBufElem *b);
/* 
 * static int cmpCalleeBuf(const CalleeBufElem *a, const CalleeBufElem *b);
 */

static SpdTrvActFunc beforeIdentifier(SpdObjId id, CalleeBuf *callee);
static SpdTrvActRelFunc beforeIdentRef(SpdObjId id, SpdRel rel,
                                      CalleeBuf *callee);

static void freeCallerBuf(CallerBuf caller);
static void freeCalleeBuf(CalleeBuf callee);

/*
 * ե󥹥ơ֥κ(ȥС)
 */

SmuXrefTables  smuMakeXrefTables(SpdObjId prog_id, SmuXrefFlag flag)
{
    SpdOccArray     callee_ident_buf;
    SpdTrvActInfo   *info1, *info2;
    SmuXrefTables   xreftables = {{0, NULL}, {0, NULL}, {0, NULL}, {0, NULL},
                                  {0, NULL}};
    CallerBuf       caller = {0, 0, NULL};
    CalleeBuf       callee = {0, 0, 0, 0, NULL};
    int             i, j;

    /*
     * part 1. 
     * ؿΰ롥
     */
    info1 = spdTrvInitActInfo();

    /* éϢ */    
    spdTrvAddRel(info1, SPD_RID_PROG_FILE);
    spdTrvAddRel(info1, SPD_RID_FILE_DECL);
    /* ãν */
    spdTrvAddActFuncOfClass(info1, SPD_CID_FILE, SAPID_TRV_ACT_BEFORE,
                            (SpdTrvActFunc)beforeFile);
    spdTrvAddActFuncOfClass(info1, SPD_CID_DECLARATION, SAPID_TRV_ACT_BEFORE,
                            (SpdTrvActFunc)beforeFuncdef);
    
    spdTraverse(prog_id, info1, &caller);
    SPD_QSORT(caller.elem, caller.size, sizeof(CallerBufElem), cmpCallerName);
    spdTrvFreeActInfo(info1);

    /*
     * part 2.
     * ؿμ̻Ҥ롥
     */
    for (i=0; i < caller.size; i++) {
        info2 = spdTrvInitActInfo();

        /* ʬ identifier ˤɤ夯δϢ */
        spdTrvAddRel(info2, SPD_RID_FUNC_BODY);
        spdTrvAddRel(info2, SPD_RID_BLK_BLK);
        spdTrvAddRel(info2, SPD_RID_BLK_EXPR);
        spdTrvAddRel(info2, SPD_RID_EXPR_EXPR);
        spdTrvAddRel(info2, SPD_RID_IDENT_REF);
        /*  identifier ˤɤ夯δϢ */
        spdTrvAddRel(info2, SPD_RID_BLK_DECL);
        spdTrvAddRel(info2, SPD_RID_DECL_DECL);
        spdTrvAddRel(info2, SPD_RID_DECL_OPT);
        spdTrvAddRel(info2, SPD_RID_INIT_EXPR);
        spdTrvAddRel(info2, SPD_RID_STRUCT_DECL);

        /* ident_ref  relationId Ͽ */
        spdTrvAddActFuncOfRelClass(info2, SPD_RID_IDENT_REF,
                                   SAPID_TRV_ACT_BEFORE,
                                   (SpdTrvActRelFunc)beforeIdentRef);

        spdTrvAddActFuncOfClass(info2, SPD_CID_IDENTIFIER,
                                SAPID_TRV_ACT_BEFORE,
                                (SpdTrvActFunc)beforeIdentifier);
    
        callee.fileId = caller.elem[i].fileId;
        callee.callerId = caller.elem[i].identId;
        spdTraverse(caller.elem[i].optId, info2, &callee);
        spdTrvFreeActInfo(info2);
    }
    /* 
     * SPD_QSORT(callee.elem, callee.size, sizeof(CalleeBufElem),
     *             cmpCalleeBuf);
     */
    
    /*
     * part 3.
     * ̻Ҥ(sort)̤˥ơ֥Ͽ롥
     */
    xreftables.funcCallTable =
        makeXrefTableSub(xreftables.funcCallTable, callee, flag,
                         spdObjIdIsIdFunction);
    xreftables.variableTable =
        makeXrefTableSub(xreftables.variableTable, callee, flag,
                         spdObjIdIsIdVariable);
    xreftables.globalVariableTable =
        makeXrefTableSub(xreftables.globalVariableTable, callee, flag,
                         smuIsGlobalVariable);
    xreftables.localVariableTable =
        makeXrefTableSub(xreftables.localVariableTable, callee, flag,
                         smuIsLocalVariable);
    
    freeCalleeBuf(callee);
    
    /*
     * part 4.
     * ޥν
     */
    for (i=0; i < caller.size; i++) {
        callee_ident_buf = spdGetIncludedOccArray(caller.elem[i].optId,
                                                  "macrocall",
                                                  SPD_NON_ID,
                                                  NULL);

        for (j = 0; j < callee_ident_buf.size; j++) {
            callee_ident_buf.occ[j].objectId =
                spdGetARelObj(callee_ident_buf.occ[j].objectId,
                              "macro_call", "any");
        }
        SPD_QSORT(callee_ident_buf.occ, callee_ident_buf.size, 
                    sizeof(SpdOcc), spdCompareOccName);

        xreftables.macroCallTable =
            makeXrefTableMacro(xreftables.macroCallTable,
                               caller.elem[i].identId, callee_ident_buf, flag,
                               NULL);
    
        spdFreeOccArray(callee_ident_buf);
    }
    freeCallerBuf(caller);
    
    return xreftables;
}

/*
 * եˬν
 */

static SpdTrvActFunc beforeFile(SpdObjId id, CallerBuf *caller)
{
    caller->fileId = id;

#ifdef DEBUG
    fprintf(stderr,"#file\t%s(%d)\n", spdGetName(id),id);
#endif

    return SAPID_TRV_GO;
}

/*
 * ؿˬν
 */

static SpdTrvActFunc beforeFuncdef(SpdObjId caller_declnId, CallerBuf *caller)
{
    SpdRel  rel;
    
    if (spdGetAttrValInt(caller_declnId, "sort") != DECL_FUNCDEF) {
        return (SpdTrvActFunc)SAPID_TRV_STOP;
    }
    
    caller->elem = spdRealloc(caller->elem, sizeof(CallerBufElem) * (caller->size + 1));

    caller->elem[caller->size].declnId = caller_declnId;
    caller->elem[caller->size].declrId = spdGetARelObj(caller_declnId, "decl_decl", "p-id");
    caller->elem[caller->size].identId = spdGetARelObj(caller->elem[caller->size].declrId, "decl_ident", "p-id");
    caller->elem[caller->size].optId   = spdGetARelObj(caller->elem[caller->size].declrId, "decl_opt", "p-id");
    
    rel = spdGetARel(caller->elem[caller->size].optId, "func_body", "p-id");
    caller->elem[caller->size].offset   = spdGetOffset(rel.relationId);
    caller->elem[caller->size].length   = spdGetLengthOfObj(rel.objectId);
    
#ifdef DEBUG
    fprintf(stderr, "%s(%d) --> %s(%d)\n",
            spdGetName(caller->fileId), caller->fileId,
            spdGetName(caller->elem[caller->size].identId),caller->elem[caller->size].identId);
#endif

    caller->size++;
    
    return SAPID_TRV_GO;
}

/*
 * ̻ˬν
 */

static SpdTrvActFunc beforeIdentifier(SpdObjId id, CalleeBuf *callee)
{

    callee->elem = spdRealloc(callee->elem,
                              sizeof(CalleeBufElem) * (callee->size + 1));

    callee->elem[callee->size].callerId = callee->callerId;
    callee->elem[callee->size].relationId = callee->relationId;
    callee->elem[callee->size].calleeId = id;

    callee->size++;

    return SAPID_TRV_GO;
}

/*
 * Ϣ(ident_ref)ˬν
 */

static SpdTrvActRelFunc beforeIdentRef(SpdObjId id, SpdRel rel,
                                       CalleeBuf *callee)
{
    callee->relationId = rel.relationId;
    return (SAPID_TRV_GO);
}

/*
 * ե󥹥ơ֥κ
 */

static SmuXrefTable makeXrefTableSub(SmuXrefTable xref_table,
                                     CalleeBuf callee,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId))
{
    SpdSet  callee_set = SAPID_EMPTY_SET;
    int     i;

    for (i = 0; i < callee.size; i++) {
        if ((cond != NULL) &&
            !cond(callee.elem[i].calleeId)) continue;

        if(flag) {
            if (spdSetContains(callee_set, callee.elem[i].calleeId)) {
                continue;
            } else {
                spdAddElem(callee_set, callee.elem[i].calleeId, NULL);
            }
        }

        xref_table.elem = spdRealloc(xref_table.elem, sizeof(SmuXrefTableElem) * (xref_table.size + 1));
        
#ifdef DEBUG
        fprintf(stderr, "refer %2d : %s(%d) --> %s(%d)\n", i,
                spdGetName(callee.elem[i].callerId),
                callee.elem[i].callerId,
                spdGetName(callee.elem[i].calleeId),
                callee.elem[i].calleeId);
#endif
    
        xref_table.elem[xref_table.size].callerId = callee.elem[i].callerId;
        xref_table.elem[xref_table.size].calleeId = callee.elem[i].calleeId;
        xref_table.elem[xref_table.size].relationId = callee.elem[i].relationId;
        xref_table.size++;
    }
    spdFreeSet(callee_set);

    return xref_table;
}

static SmuXrefTable makeXrefTableMacro(SmuXrefTable xref_table,
                                       SpdObjId caller_identId,
                                       SpdOccArray callee_ident_buf,
                                       SmuXrefFlag flag,
                                       SpdBoolean (*cond)(SpdObjId))
{
    SpdSet  callee_set = SAPID_EMPTY_SET;
    int     i;

    for (i = 0; i < callee_ident_buf.size; i++) {
        if ((cond != NULL) &&
            !cond(callee_ident_buf.occ[i].objectId)) continue;

        if(flag) {
            if (spdSetContains(callee_set, callee_ident_buf.occ[i].objectId)) {
                continue;
            } else {
                spdAddElem(callee_set, callee_ident_buf.occ[i].objectId, NULL);
            }
        }

        xref_table.elem = spdRealloc(xref_table.elem,
                                     sizeof(SmuXrefTableElem) *
                                     (xref_table.size + 1));
        
#ifdef DEBUG
        fprintf(stderr, "|  +-- %d : %s(%d) -> %s(%d)\n", i,
                spdGetName(caller_identId), caller_identId,
                spdGetName(callee_ident_buf.occ[i].objectId),
                callee_ident_buf.occ[i].objectId);
#endif
    
        xref_table.elem[xref_table.size].callerId
            = caller_identId;
        xref_table.elem[xref_table.size].calleeId
            = callee_ident_buf.occ[i].objectId;
        xref_table.elem[xref_table.size].relationId
            = callee_ident_buf.occ[i].relationId;
        xref_table.size++;
    }
    spdFreeSet(callee_set);

    return xref_table;
}

void smuFreeXrefTables(SmuXrefTables tables)
{
    spdFree(tables.funcCallTable.elem);
    spdFree(tables.variableTable.elem);
    spdFree(tables.globalVariableTable.elem);
    spdFree(tables.localVariableTable.elem);
    spdFree(tables.macroCallTable.elem);
}

static int cmpCallerName(const CallerBufElem *a, const CallerBufElem *b)
{
    if (a->identId != b->identId) {
        return (spdSafetyStrcmp(spdGetName(a->identId), spdGetName(b->identId)));
    }
    return 0;
}

/* 
 * static int cmpCalleeBuf(const CalleeBufElem *a, const CalleeBufElem *b)
 * {
 *     if (a->callerId != b->callerId) {
 *         return (spdSafetyStrcmp(spdGetName(a->callerId), spdGetName(b->callerId)));
 *     }
 *     if (a->calleeId != b->calleeId) {
 *         return (spdSafetyStrcmp(spdGetName(a->calleeId), spdGetName(b->calleeId)));
 *     }
 *     return spdGetOffset(a->relationId) - spdGetOffset(b->relationId);
 * }
 */

static void freeCallerBuf(CallerBuf caller)
{
    spdFree(caller.elem);
}

static void freeCalleeBuf(CalleeBuf callee)
{
    spdFree(callee.elem);
}

#endif /* defined(SMU_XREF_USE_TRAVERSE) */
