/* 
 * Program:             $RCSfile: headerUtil.c,v $  $Revision: 30.13 $
 * 
 * Purpose:
 * 
 * Author:              N.Ito, N.Fukuyasu and S.Yamamoto  1997/06/27
 * 
 * (C) Copyright:       N.Ito, N.Fukuyasu and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] = 
    "$Id: headerUtil.c,v 30.13 2003/02/15 06:34:55 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h> 
#include "SpdStrSub.h"

/*
 * Use fast isInHeader() which decide whether the object exists in a
 * header or not based on P-model.
 * Added by T.Hidaka and S.Yamamoto 1998/08/05.
 */

#define USE_FAST_isInHeader

static SpdBoolean allOffsetGetFileNameCfromI(char *iname, int offset, 
                                             int length, char *cname);

/*
 * SMUѤθȽؿ
 */

/* P-modelblockincludeʤ鿿֤Ҹ졥 */

SpdBoolean spdObjIdIsInclude(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == PBLOCK_INCLUDE);
}

/* P-modelblockdefineʤ鿿֤Ҹ졥 */

SpdBoolean spdObjIdIsDefine(SpdObjId id)
{
    return (spdGetAttrValInt(id, "sort") == PBLOCK_DEFINE_FILE);
}

/* 
 * SMUѤӴؿSpdObjId򥪥եåȤӤ롥  
 * IdoccäƤ뤳ȤȤƤ롥
 */

static int spdCompareIdOffset(const SpdObjId *a, const SpdObjId *b)
{
    return (spdGetAnOffsetOfObj(*a).offset - spdGetAnOffsetOfObj(*b).offset);
}

static int getFileNameCfromI_forNotInMacro(char *iname, int offset, char *cname)
{
    int         iline, icolumn, fileLen;
    SpdString   fileBuf = NULL;
    SpdLineTbl  tbl;
 
    tbl = spdGetLineTbl(smuGetFileIdByName(iname));
    if (spdCnvOffsetToLineColumnSub(tbl, offset, &iline, &icolumn) < 0) {
        return (-1);
    }
    fileLen = spdReadFileInBufByName(iname, &fileBuf);
    (void)lineConvertItoCwithTbl(cname, iline, fileBuf, fileLen, tbl);
    spdFree(fileBuf);
    return (TRUE);
}

static int getFileNameCfromI(char *iname, int offset, char *cname)
{
    MacroExpandRegion   mereg;
    MacroCallRegion mcreg;
    int         r;
     
    mereg.type = EXPAND_FILE;
    mereg.start = mereg.end = offset;
    mereg.fname = iname;

#ifdef PIDB_USE_FILENAME_INDEX
    mereg.fname_idx = pidbConvFileId(iname);
#endif

    r = reverseParseAnOutputText(&mereg, &mcreg);
    if (r < 0) {
        /* -1 SAPID_CPP_NOT_SUPPORTED */
        return r;
    }
    if (r == 0) {                           /* бʸ */
        strcpy(cname, mcreg.fname);
        return mcreg.start;
    }
    /* ޥŸ˴ޤޤʤ */
    r = getFileNameCfromI_forNotInMacro(iname, offset, cname);
    return r;
}

/******************************************************************************/

#if defined(USE_FAST_isInHeader)

int isInHeader(SpdObjId objectId)
{
    SpdOcc occ;
    SpdRegion ireg, creg;

    occ = spdGetAnOffsetOfObj(objectId);
    ireg = spdConvOccToRegion(occ);
    creg = spdConvRegionItoC(ireg);
    return (spdGetARelObj(creg.fileId, "hfile_file", "pfile_id") !=
            SPD_NON_ID);
}

#else /* !defined(USE_FAST_isInHeader) */

int isInHeader(SpdObjId objectId)
{
    int     cnameLen;
    SpdOcc  occ;
    char    cname[FILENAME_MAX + 1];
  
    occ = spdGetAnOffsetOfObj(objectId);
    cname[0] = '\0';

#ifdef BUGGY_PIDB_0109
    if(threeOffsetGetFileNameCfromI(spdGetName(occ.fileId), occ.offset, 
                                    spdGetLength(objectId), cname) == FALSE){
        return (FALSE);
    }
#else  /* not BUGGY_PIDB_0109 */
    switch(getFileNameCfromI(spdGetName(occ.fileId), occ.offset, cname)){
      case SAPID_CPP_NOT_SUPPORTED:
        spdErrorExit("getFileNameCfromI() returns SAPID_CPP_NOT_SUPPORTED.\n");
        break;
      case SPD_NON_ID:
        return (FALSE);
        break;
      default:
        break;
    }
#endif /* not BUGGY_PIDB_0109 */

    cnameLen = strlen(cname);
    /* 
     * if ((0 < cnameLen) && cname[cnameLen - 1] == 'h') {
     *     return (TRUE);
     * } else {
     *     return (FALSE);
     * }
     */
    return (0 < cnameLen && cname[cnameLen - 1] == 'h');
}

#endif /* !defined(USE_FAST_isInHeader) */

/******************************************************************************/

#ifdef BUGGY_PIDB_0109

/*
 * Ƭ,,ƤSPD_NON_ID֤ʤ
 * ΰƤoffsetФgetFileNameCfromI()Ƥ֡
 * ƤˤSPD_NON_ID֤äˡϤFALSE֤
 */

SpdBoolean threeOffsetGetFileNameCfromI(char *iname, int offset, 
                                        int length, char *cname)
{
    
    int headReturnValue, centerReturnValue, tailReturnValue;
    
    headReturnValue = getFileNameCfromI(iname, offset, cname);
    centerReturnValue = getFileNameCfromI(iname, offset + length / 2, cname);
    tailReturnValue = getFileNameCfromI(iname, offset + length , cname);
    
    if(headReturnValue == SAPID_CPP_NOT_SUPPORTED ||
       centerReturnValue == SAPID_CPP_NOT_SUPPORTED ||
       tailReturnValue == SAPID_CPP_NOT_SUPPORTED){
        spdErrorExit("getFileNameCfromI() returns SAPID_CPP_NOT_SUPPORTED.\n");
    }
    
    if(headReturnValue == SPD_NON_ID &&
       centerReturnValue == SPD_NON_ID &&
       tailReturnValue == SPD_NON_ID){
        
        if(allOffsetGetFileNameCfromI(iname, offset, length, cname) == FALSE){
            return (FALSE);
        } else {
            return (TRUE);
        }
    } else {
        return (TRUE);
    }
}

static SpdBoolean allOffsetGetFileNameCfromI(char *iname, int offset, 
                                             int length, char *cname)
{
    int         i;
    SpdBoolean  flag = FALSE;
    
    for (i = offset; i <= length; i++) {
        flag += (getFileNameCfromI(iname, i, cname) != SPD_NON_ID);
    }
    
    if (flag) {
        return (TRUE);
    } else {
        return (FALSE);
    }
    /* return (flag != FALSE); */
}

#endif /* BUGGY_PIDB_0109 */

/******************************************************************************/

/* P-modelheaderincludeʸϤ롥 */

SpdString allCatHeaderInfoAR4(SpdObjId ifileId, int catFlag)
{
    SpdObjId pfileId, cfileId, hfileId;
    SpdString res = NULL;
    SpdObjIdArray block_buf;
    int i;
    static SpdString save = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }
    
    if (catFlag == FALSE) {
        return res;
    }

    /* ȥåץ٥"*.c"ե̾롥 */
    pfileId = spdGetPfileIdIncludingIoffset(ifileId, 0);
    cfileId = spdGetCfileIdByPfileId(pfileId);

    /* "*.c"ե˽иblockΤߤ롥 */
    block_buf = spdGetRelObjSortedArray(cfileId, "file_block", "any", 
                                        spdObjIdIsInclude,
                                        spdCompareIdOffset);

    /* includeʸϤ롥 */
    for (i = 0; i < block_buf.size; i++){
        hfileId = spdGetARelObj(block_buf.id[i], "incld_hfile", "any");
        if (spdStrHfile(hfileId) != NULL) { 
            res = spdStrCat(res, "#include ");
            res = spdStrCat(res, spdStrHfile(hfileId));
            res = spdStrCat(res, "\n");
        }
    }
    
    spdFreeObjIdArray(block_buf);
    
    save = res;
    return res;
}
