/* 
 * Program:             $RCSfile: macroStr.c,v $  $Revision: 30.25 $
 * 
 * Purpose:
 * 
 * Author:              N.Ito, N.Fukuyasu and S.Yamamoto  1997/06/27
 * 
 * (C) Copyright:       N.Ito, N.Fukuyasu and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] = 
    "$Id: macroStr.c,v 30.25 2003/06/28 15:31:16 yamamoto Exp yamamoto $";
#endif

/* 
 * File name "spdStr_macro.c" is better.
 */

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "SpdStrSub.h"

/* Use cache facility in trimRegionSpaces(). */
#define USE_CACHE

#define IS_IN_REGION(x, end, start)     (x <= (end) - (start))
/* #define IS_IN_REGION(x, end, start)     (x < (end) - (start)) */

/* #define DEBUG_trimRegionSpaces */

static SpdString macroCreate(char *filename, int offset, int length);

static void trimRegionSpaces(int *offsetStart, int *offsetEnd, SpdConstString filename);
static int macroExists(char *filename, int offset, int length);

void macroInit(SpdObjId fileId)
{
    SpdString   srcname;

    if (spdGetOption("mode") == SAPID_I_MODE) {
        return;
    }
    
    if ((srcname = spdStrdup(spdGetName(fileId))) == NULL) {
        spdErrorExit("SpdStr: macroInit(): %d: Invalid file Id.\n", fileId);
    }

    if (spdIsFileExtension(srcname, ".i")) {
        if (spdInitializePidbCache(fileId) != 0) {
            spdErrorExit("Can't initialize: \"%s\".", srcname);
        }
    }

    spdFree(srcname);
}

SpdString spdStrMacro(SpdObjId objectId)
{
    static SpdString save = NULL;

    SpdString   res = NULL;
    char        *filename;
    int         length;
    SpdOcc      occ;
    int         dummy_offset = SPD_NON_OFFSET;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    if (objectId == SPD_NON_ID) {
        /* Added by S.Yamamoto 1998/08/09. */
        macroCreate(NULL, SPD_NON_OFFSET, SPD_NON_LENGTH);
        /* Added by S.Yamamoto 1998/08/25. */
        trimRegionSpaces(&dummy_offset, &dummy_offset, NULL);

        return (res);
    }
    if (spdGetOption("mode") == SAPID_I_MODE) {
        return (NULL);
    }

    occ = spdGetAnOffsetOfObj(objectId);
    /* Return if we can't get fileId. */
    if (occ.fileId == SPD_NON_ID) {
        return (NULL);
    }
    filename = spdGetName(occ.fileId);

    if ((length = spdGetLength(objectId)) == SPD_NON_LENGTH) {
        return (NULL);
    }

    if (macroExists(filename, occ.offset, length)) {
        res = spdStrCat(res, macroCreate(filename, occ.offset, length));
    }

    save = res;
    return (res);
}

/******************************************************************************/

/*
 * Moved from "macroAppearance.c" by S.Yamamoto 1998/08/09.
 *
 * macroCreate(): Create macro name.
 * macroArgCreateI(): Create macro arguments.
 */

static SpdString macroCreate(char *filename, int offset, int length)
{
    int         start, end;
    int         expStart, expEnd;
    SpdString   res = NULL;
    int         i;
    
    static SpdString    save = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }
    
    if (offset == SPD_NON_OFFSET && length == SPD_NON_LENGTH) {
        return (res);
    }

    start = offset;
    end = offset + length;

    for (i = 0; i < expandInfoN; i++) {
        if (expandInfo[i].output.type != EXPAND_FILE) {
            continue;
        }
        expStart = expandInfo[i].output.start;
        expEnd = expandInfo[i].output.end;
        /* Added by S.Yamamoto 1998/08/25. */
        if (start < expStart || expEnd < end) {
            continue;
        }

        {
            int     expStart1, expEnd1;

            expStart1 = expandInfo[i].output.start;
            expEnd1 = expandInfo[i].output.end;
            trimRegionSpaces(&expStart1, &expEnd1,
                             expandInfo[i].output.fname);
            expStart = expStart1;
            expEnd = expEnd1;
        }

        /*
         * Note by S.Yamamoto 1998/08/24.
         * Variable "expEnd" holds a position of last character of the macro
         * in "*.i" file. Variable "end" holds a postion of the next character
         * of target object because it is declared as (start + length).
         * So conditional expression (expEnd == end) for checking equality
         * is incorrect. Use correct expression (expEnd + 1 == end).
         */
        if (expStart == start && expEnd + 1 == end) {
            res = spdStrCat(res, expandInfo[i].mac->name);
            if (expandInfo[i].mac->narg == 0) {
                res = spdStrCat(res, "()");
            /* } else if(expandInfo[i].mac->narg != SPD_NON_ID) { */
            } else if(expandInfo[i].mac->narg != -1) {

#ifdef BUGGY_SPDSTR_0102
                if (macroArgCreateI(i, expandInfo[i].output.start, filename) ==
                   NULL) {
                    spdFree(res);
                    res = NULL;
                    res = spdStrCat(res, SPDSTR_ERROR_MACRO_IN_MACRO);
                    break;
                }
#endif /* BUGGY_SPDSTR_0102 */

                res = spdStrCat(res,
                                macroArgCreateI(i, expandInfo[i].output.start,
                                                filename));
            }
            break;
        }
    }

    save = res;
    return (res);
}

/******************************************************************************/

static void trimRegionSpaces(int *offsetStart, int *offsetEnd,
                             SpdConstString filename)
{
    int     start = *offsetStart, end = *offsetEnd;
    char    *cp;

    static char     *savedFilename = NULL, *cbuf = NULL;
    
    /* Release allocated resources. */
    if (filename == NULL) {
        if (savedFilename != NULL) {
            spdFree(savedFilename);
            savedFilename = NULL;
        }
        if (cbuf != NULL) {
            spdFree(cbuf);
            cbuf = NULL;
        }
        return;
    }

    /* Added by S.Yamamoto 1998/08/24. For bug compatibility? */
    if (start == end) {
        *offsetStart = start;
        *offsetEnd = end;
        return;
    }

#ifdef USE_CACHE
    if (savedFilename == NULL) {
        savedFilename = spdStrdup(filename);
        spdReadFileInBufByName(filename, &cbuf);
    } else if (!SPD_STREQ(savedFilename, filename)) {
        spdFree(savedFilename);
        savedFilename = spdStrdup(filename);
        spdFree(cbuf);
        spdReadFileInBufByName(filename, &cbuf);
    }
#else
    spdReadFileInBufByName(filename, &cbuf);
#endif  

    cp = cbuf + *offsetStart;
    while (!isgraph(*cp++) && start < end) {
        start++;
    }

    cp = cbuf + *offsetEnd;
    while (!isgraph(*cp--) && start < end) { 
        end--;
    }

#ifdef DEBUG_trimRegionSpaces
    {
        char    *cp1;
        int     i;

        spdMessage("trimRegionSpaces(): \"");
        cp1 = cbuf + *offsetStart;
        for (i = 0; IS_IN_REGION(i, *offsetEnd, *offsetStart); i++) {
            fputc(*cp1++, stderr);
        }
        fprintf(stderr, "\".\n");

        spdMessage("trimRegionSpaces(): \"");
        cp1 = cbuf + start;
        for (i = 0; IS_IN_REGION(i, end, start); i++) {
            fputc(*cp1++, stderr);
        }
        fprintf(stderr, "\".\n");

        fprintf(stderr, "\n");
    }
#endif /* DEBUG_trimRegionSpaces */

#ifndef USE_CACHE
    spdFree(cbuf);
    cbuf = NULL;
#endif  

    *offsetStart = start;
    *offsetEnd = end;
}

/******************************************************************************/

static int macroExists(char *filename, int offset, int length)
{
    MacroExpandRegion   target;

    target.type = EXPAND_FILE;
    target.start = offset;
    target.end = offset + length - 1;
    target.fname = filename;

#ifdef PIDB_USE_FILENAME_INDEX
    target.fname_idx = pidbConvFileId(filename);
#endif

    return (0 <= isInMacroExpand(&target, MINUS_VALUE, MINUS_VALUE));
}
