/* 
 * Program:             $RCSfile: spdStr_Pmodel.c,v $  $Revision: 30.5 $
 * 
 * Purpose:             Enhanced bsearch().
 * 
 * Author:              Y.Momiyama  1995/03/05
 *                      S.Yamamoto  1995/03/05
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 1998
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr_Pmodel.c,v 30.5 2002/10/21 07:25:57 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "SpdStrSub.h"

SpdString spdStrHfile(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 hfile ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_HFILE) {
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
      case HFILE_ANGLE:
        res = spdStrCat(res, "<");
        res = spdStrCat(res, spdGetName(objectId));
        res = spdStrCat(res, ">");
        break;
      case HFILE_DQUOT:
        res = spdStrCat(res, "\"");
        res = spdStrCat(res, spdGetName(objectId));
        res = spdStrCat(res, "\"");
        break;
      default:
        break;
    }

    save = res;
    return res;
}

SpdString spdStrMacroname(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 macroName ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_MACRONAME) {
        return res;
    }

    res = spdStrCat(res, spdGetName(objectId));

    save = res;
    return res;
}

SpdString spdStrMacrodeclarg(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 macroDeclArg ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_MACRODECLARG) {
        return res;
    }

    res = spdStrCat(res, spdGetName(objectId));

    save = res;
    return res;
}

SpdString spdStrMacrooptional(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;
    SpdRelArray macrodeclargArray;
    int i;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 macroOptional ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_MACROOPTIONAL) {
        return res;
    }

    res = spdStrCat(res, "(");
    macrodeclargArray = spdGetRelSortedArray(objectId, "mopt_mdecla",
                                             "p-id", NULL, spdCompareRelOffset);
    if (macrodeclargArray.size > 0) {
        res = spdStrCat(res,
                        spdStrMacrodeclarg(macrodeclargArray.rel[0].objectId));
    }
    for (i = 1; i < macrodeclargArray.size; i++) {
        res = spdStrCat(res, ", ");
        res = spdStrCat(res,
                        spdStrMacrodeclarg(macrodeclargArray.rel[i].objectId));
    }
    spdFreeRelArray(macrodeclargArray);

    res = spdStrCat(res, ")");

    save = res;
    return res;
}

SpdString spdStrMacropattern(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;
    SpdObjId macroname_id, macrooptional_id;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 macroPattern ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_MACROPATTERN) {
        return res;
    }

    macroname_id = spdGetARelObj(objectId, "mpatt_mname", "p-id");
    res = spdStrCat(res, spdStrMacroname(macroname_id));
    macrooptional_id = spdGetARelObj(objectId, "mpatt_mopt", "p-id");
    res = spdStrCat(res, spdStrMacrooptional(macrooptional_id));

    save = res;
    return res;
}

static int spdCompareRelOrder(const SpdRel *a, const SpdRel *b)
{
    return (spdGetAttrValInt(a->relationId, "order") -
            spdGetAttrValInt(b->relationId, "order"));
}

SpdString spdStrMacrobody(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;
    SpdRelArray mbody_mdeclaArray;
    int i;
    char *normstr = NULL, *pointer;
    int skip, sharp;
    char *append = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 macroBody ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_MACROBODY) {
        return res;
    }

    normstr = spdStrdup(spdGetAttrValString(objectId, "name"));
    pointer = normstr;
    mbody_mdeclaArray = spdGetRelSortedArray(objectId, "mbody_mdecla", "p-id",
                                             NULL, spdCompareRelOrder);
    for (i = 0; i < mbody_mdeclaArray.size; i++) {
        /* skipʸʬ normstr 򥳥ԡ */
        skip = spdGetAttrValInt(mbody_mdeclaArray.rel[i].relationId, "skip");
        append = spdMalloc(skip + 1);
        strncpy(append, pointer, skip);
        append[skip] = '\0';
        res = spdStrCat(res, append);
        spdFree(append);
        append = NULL;
    
        pointer += skip;                    /* normstr Υݥ󥿤ʤޤ */

        sharp = spdGetAttrValInt(mbody_mdeclaArray.rel[i].relationId, "sharp");
        if (sharp == 1) {
            res = spdStrCat(res, " #");
        }
        res = spdStrCat(res, spdStrMacrodeclarg(mbody_mdeclaArray.rel[i].objectId));
    }
    res = spdStrCat(res, pointer);
    spdFree(normstr);
    spdFreeRelArray(mbody_mdeclaArray);

    save = res;
    return res;
}

/******************************************************************************/

SpdString spdStrPblock(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 pblock ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_PBLOCK) {
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
        SpdObjId hfile_id;
        SpdObjId macropattern_id, macrobody_id;

      case PBLOCK_INCLUDE:
        res = spdStrCat(res, "#include ");
        hfile_id = spdGetARelObj(objectId, "incld_hfile", "p-id");
        res = spdStrCat(res, spdStrHfile(hfile_id));
        break;
      case PBLOCK_DEFINE_FILE:
        res = spdStrCat(res, "#define ");
        macropattern_id = spdGetARelObj(objectId, "def_mpatt", "p-id");
        res = spdStrCat(res, spdStrMacropattern(macropattern_id));
        macrobody_id = spdGetARelObj(objectId, "def_mbody", "p-id");
        if (macrobody_id != SPD_NON_ID) {
            res = spdStrCat(res, " ");
            res = spdStrCat(res, spdStrMacrobody(macrobody_id));
        }
        break;
      default:
        break;
    }

    save = res;
    return res;
}

/*
 * spdStrPblock Ǥ
 *
 * #include
 * #define
 *
 * Τߤʸƹ뤬,ʲδؿϤ
 * ƹʤ
 */
SpdString spdStrPblockCutNull(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 pblock ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_PBLOCK) {
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
        SpdObjId hfile_id;
        SpdObjId macropattern_id, macrobody_id;

      case PBLOCK_INCLUDE:
        hfile_id = spdGetARelObj(objectId, "incld_hfile", "p-id");
        if(spdStrHfile(hfile_id) != NULL){
            res = spdStrCat(res, "#include ");
            res = spdStrCat(res, spdStrHfile(hfile_id));
        }
        break;

      case PBLOCK_DEFINE_FILE:
        macropattern_id = spdGetARelObj(objectId, "def_mpatt", "p-id");
        if(spdStrMacropattern(macropattern_id) != NULL){
            res = spdStrCat(res, "#define ");
            res = spdStrCat(res, spdStrMacropattern(macropattern_id));
            macrobody_id = spdGetARelObj(objectId, "def_mbody", "p-id");
            if (macrobody_id != SPD_NON_ID) {
                res = spdStrCat(res, " ");
                res = spdStrCat(res, spdStrMacrobody(macrobody_id));
            }
        }
        break;

      default:
        break;
    }

    save = res;
    return res;
}
