/* 
 * Program:             $RCSfile: spdStr_block.c,v $  $Revision: 30.3 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu, S.Kawai and Y.Momiyama  1995/12/07
 * Modified:            S.Yamamoto  1995/12/08
 * 
 * (C) Copyright:       N.Fukuyasu, S.Kawai, Y.Momiyama
 *                      and S.Yamamoto  1995 - 1998
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr_block.c,v 30.3 2003/06/28 15:31:16 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "SpdStrSub.h"

extern int indent;                       /* ǥȿ */

SpdString spdStrBlock(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

#ifdef BUGGY_SPDSTR_0102
    SpdBoolean buggyFlag = FALSE;
    int modeOption;
#endif /* BUGGY_SPDSTR_0102 */

    DW_FUNC_IN("block", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return res;
    }

    res = spdStrCat(res, spdStrMacro(objectId));

#ifdef BUGGY_SPDSTR_0102
    if(res != NULL){
        if(SPD_STREQ(res, SPDSTR_ERROR_MACRO_IN_MACRO)){
            modeOption = spdGetOption("mode");
            spdSetOption("mode", SAPID_I_MODE);
            buggyFlag = TRUE;
            spdFree(res);
            res = NULL;
        }
    }
#endif /* BUGGY_SPDSTR_0102 */

    if(res != NULL){
        save = res;
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
      case BLOCK_BASIC:
        res = spdStrCat(res, spdStrBasic(objectId));
        break;
      case BLOCK_COMP:
        res = spdStrCat(res, spdStrCompound(objectId));
        break;
      case BLOCK_IF:                        /* Branch ֥å */
      case BLOCK_SWITCH:
      case BLOCK_FOR:
      case BLOCK_WHILE:
      case BLOCK_DO:
        res = spdStrCat(res, spdStrBranch(objectId));
        break;
    }

    DW_FUNC_OUT("block", res);

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }

    save = res;

#ifdef BUGGY_SPDSTR_0102
    if(buggyFlag)
        spdSetOption("mode", modeOption);
#endif /* BUGGY_SPDSTR_0102 */

    return res;
}

SpdString spdStrBasic(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     buf_n = 0, i;

    DW_FUNC_IN("basic", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return res;
    }
    if (spdGetAttrValInt(objectId, "sort") != BLOCK_BASIC) return res;

    csr = spdGetRelInit(objectId, "blk_expr", "any");
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);

    SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);

    for (i = 0; i < buf_n; i++) {
        if(i != 0) Indent();
        res = spdStrCat(res, spdStrExpression(buf[i].objectId));

        if(i < buf_n-1) res = spdStrCat(res, ";\n");
        else res = spdStrCat(res, ";");
    }
    if (buf != NULL) spdFree(buf);

    DW_FUNC_OUT("basic", res);

    save = res;
    return res;
}

SpdString spdStrBranch(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdObjId    expr_id, blk_id;
    SpdString   res = NULL;
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     buf_n = 0, i;

    DW_FUNC_IN("branch", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return res;
    }

    switch(spdGetAttrValInt(objectId, "sort")){
      case BLOCK_IF:
        res = spdStrCat(res, "if (");
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        res = spdStrCat(res, spdStrExpression(expr_id));
        res = spdStrCat(res, ") ");
        csr = spdGetRelInit(objectId, "blk_blk", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
        res = spdStrCat(res, spdStrBlock(buf[0].objectId));
        if (buf_n > 1) {
            res = spdStrCat(res, " else ");
            res = spdStrCat(res, spdStrBlock(buf[1].objectId));
        }
        if (buf != NULL) spdFree(buf);
        break;

      case BLOCK_SWITCH:
        res = spdStrCat(res, "switch (");
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        res = spdStrCat(res, spdStrExpression(expr_id));
        res = spdStrCat(res, ") ");
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        res = spdStrCat(res, spdStrBlock(blk_id));
        break;

      case BLOCK_FOR:
        res = spdStrCat(res, "for (");
        csr = spdGetRelInit(objectId, "blk_expr", "any");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);

        for (i = 0; i < buf_n; i++) {
            if(i != 0) res = spdStrCat(res, "; ");
            res = spdStrCat(res ,spdStrExpression(buf[i].objectId));
        }
        if (buf != NULL) spdFree(buf);

        res = spdStrCat(res, ") ");
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        res = spdStrCat(res, spdStrBlock(blk_id));
        break;

      case BLOCK_WHILE:
        res = spdStrCat(res, "while (");
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        res = spdStrCat(res, spdStrExpression(expr_id));
        res = spdStrCat(res, ") ");
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        res = spdStrCat(res, spdStrBlock(blk_id));
        break;

      case BLOCK_DO:
        res = spdStrCat(res, "do ");
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        res = spdStrCat(res, spdStrBlock(blk_id));
        res = spdStrCat(res, " while (");
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        res = spdStrCat(res, spdStrExpression(expr_id));
        res = spdStrCat(res, ");");
        break;
    }

    DW_FUNC_OUT("branch", res);

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }

    save = res;
    return res;
}

SpdString spdStrCompound(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     buf_n = 0, class_id, i;

    DW_FUNC_IN("compound", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return res;
    }

    if (spdGetAttrValInt(objectId, "sort") != BLOCK_COMP) return res;

    csr = spdGetRelInit(objectId, "blk_decl", "any"); /* blk_decl μ */
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);
    csr = spdGetRelInit(objectId, "blk_blk", "p-id"); /* blk_blk μ */
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);
    csr = spdGetRelInit(objectId, "blk_label", "any"); /* blk_labl μ */
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);
    SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);

    res = spdStrCat(res, "{");
    indent++;
    for (i = 0; i < buf_n; i++) {
        class_id = spdGetClassId(buf[i].objectId);
        if(class_id == SPD_CID_DECLARATION) {
            res = spdStrCat(res, "\n");
            Indent();
            res = spdStrCat(res, spdStrDeclaration(buf[i].objectId, FALSE));
        }
        else if(class_id == SPD_CID_BLOCK) {
            res = spdStrCat(res, "\n");
            Indent();
            res = spdStrCat(res, spdStrBlock(buf[i].objectId));
        }
        else if(class_id == SPD_CID_LABEL) {
            res = spdStrCat(res, "\n");
            indent--;
            Indent();
            indent++;
            res = spdStrCat(res, spdStrLabel(buf[i].objectId));
            res = spdStrCat(res, ":");
        }
    }
    if (buf != NULL) spdFree(buf);
    res = spdStrCat(res, "\n");
    indent--;
    Indent();
    res = spdStrCat(res, "}");

    DW_FUNC_OUT("compound", res);

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }

    save = res;
    return res;
}

SpdString spdStrLabel (SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;
    SpdObjId    expr_id;

#ifdef BUGGY_SPDSTR_0102
    SpdBoolean buggyFlag = FALSE;
    int modeOption;
#endif /* BUGGY_SPDSTR_0102 */

    DW_FUNC_IN("label", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 label ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_LABEL) {
        return res;
    }

    res = spdStrCat(res, spdStrMacro(objectId));

#ifdef BUGGY_SPDSTR_0102
    if(res != NULL){
        if(SPD_STREQ(res, SPDSTR_ERROR_MACRO_IN_MACRO)){
            modeOption = spdGetOption("mode");
            spdSetOption("mode", SAPID_I_MODE);
            buggyFlag = TRUE;
            spdFree(res);
            res = NULL;
        }
    }
#endif /* BUGGY_SPDSTR_0102 */

    if(res != NULL){
        save = res;
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
      case LABEL_CASE:
        res = spdStrCat(res, "case ");
        expr_id = spdGetARelObj(objectId, "case_cexpr", "any");
        res = spdStrCat(res, spdStrExpression(expr_id));
        break;
      case LABEL_DEFAULT:
        res = spdStrCat(res, "default");
        break;
      case LABEL_NAMED:
        res = spdStrCat(res, spdGetName(objectId));
        break;
      default:
        break;
    }

    DW_FUNC_OUT("label", res);

    save = res;

#ifdef BUGGY_SPDSTR_0102
    if(buggyFlag)
        spdSetOption("mode", modeOption);
#endif /* BUGGY_SPDSTR_0102 */

    return res;
}
