/* 
 * Program:             $RCSfile: spdStr_expr.c,v $  $Revision: 30.7 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu, S.Kawai and Y.Momiyama  1995/12/07
 * Modified:            S.Yamamoto  1995/12/08
 * 
 * (C) Copyright:       N.Fukuyasu, S.Kawai, Y.Momiyama
 *                      and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr_expr.c,v 30.7 2003/06/28 15:31:15 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "SpdStrSub.h"

#if defined(lint)
#define DW_EXPR_960830      printf      /* Just for cheking. */
#elif defined(DEBUG_EXPR_960830)
#define DW_EXPR_960830      spdMessage
#else
#define DW_EXPR_960830      (void)
#endif

SpdString spdStrExpression(SpdObjId objectId)
{
    SpdObjId    expr_id, label_id, op_id;
    SpdString   res = NULL;
    SpdCursor   *csr;
    SpdRel      rel, *buf = NULL;
    int         buf_n = 0, class_id, expr_sort, i;

    static SpdString    save = NULL;

#ifdef BUGGY_SPDSTR_0102
    SpdBoolean  buggyFlag = FALSE;
    int         modeOption;
#endif

    DW_FUNC_IN("expression", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }
    if ((spdGetClassId(objectId)) != SPD_CID_EXPRESSION) {
        return (res);
    }

    res = spdStrCat(res, spdStrMacro(objectId));

#ifdef BUGGY_SPDSTR_0102
    if (res != NULL && SPD_STREQ(res, SPDSTR_ERROR_MACRO_IN_MACRO)) {
        modeOption = spdGetOption("mode");
        spdSetOption("mode", SAPID_I_MODE);
        buggyFlag = TRUE;
        spdFree(res);
        res = NULL;
    }
#endif

    if (res != NULL) {
        save = res;
        return res;
    }

    expr_sort = smuGetExprSort(objectId);
    switch(expr_sort){
      case EXPR_EMPTY:
        break;
      case EXPR_BREAK:
        res = spdStrCat(res, "break");
        break;
      case EXPR_CONTINUE:
        res = spdStrCat(res, "continue");
        break;
      case EXPR_RETURN:
        res = spdStrCat(res, "return");
        if((expr_id = 
            spdGetARelObj(objectId, "expr_expr", "p-id")) != -1){
            res = spdStrCat(res, " ");
            res = spdStrCat(res, spdStrExpression(expr_id));
        }
        break;
      case EXPR_GOTO:
        res = spdStrCat(res, "goto ");
        label_id = spdGetARelObj(objectId,"goto_label","any");
        res = spdStrCat(res, spdStrLabel(label_id));
        break;
      case EXPR_SORT_SIZEOF: 
        expr_id = spdGetARelObj(objectId, "expr_expr", "p-id");
        op_id = spdGetARelObj(objectId, "expr_op", "p-id");
        if (expr_id == -1) {                /* sizeof(type) */
            res = spdStrCat(res, spdStrOperator(op_id));
        } else {                            /* sizeof(expression) */
            res = spdStrCat(res, spdStrOperator(op_id));
            /* res = spdStrCat(res, "("); */
            res = spdStrCat(res, spdStrExpression(expr_id));
            /* res = spdStrCat(res, ")"); */
        }
        break;
      case EXPR_SORT_FUNCCALL:              /* ؿƤӽФ */
        DW_EXPR_960830("FUNCCALL: %d", objectId);
        csr = spdGetRelInit(objectId, "expr_expr", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
            DW_EXPR_960830("\n --- expr_expr --- %d: %s ",
                           rel.objectId, spdStrExpression(rel.objectId));
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
      
        res = spdStrCat(res, spdStrExpression(buf[0].objectId));
        res = spdStrCat(res, "(");
        for(i=1; i<buf_n; i++) {            /* ؿƤӽФΰ */
            if(i != 1) res = spdStrCat(res, ", ");
            res = spdStrCat(res, spdStrExpression(buf[i].objectId));
        }
        res = spdStrCat(res, ")");
        spdFree(buf);
        DW_EXPR_960830("\n : %s\n", res);
        break;
      case EXPR_SORT_INITLIST:              /*  */
        res = spdStrCat(res, "{ ");
        csr = spdGetRelInit(objectId, "expr_expr", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
        for (i = 0; i < buf_n; i++) {
            if(i != 0) res = spdStrCat(res, ", ");
            res = spdStrCat(res, spdStrExpression(buf[i].objectId));
        }
        spdFree(buf);
        res = spdStrCat(res, " }");
        break;
      default:
        csr = spdGetRelInit(objectId, "expr_expr", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        csr = spdGetRelInit(objectId, "expr_op", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        csr = spdGetRelInit(objectId, "const_ref", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        csr = spdGetRelInit(objectId, "ident_ref", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
        for (i=0; i<buf_n; i++) {
            class_id = spdGetClassId(buf[i].objectId);
            if (class_id == SPD_CID_EXPRESSION) {
                res = spdStrCat(res, spdStrExpression(buf[i].objectId));
            } else if(class_id == SPD_CID_OPERATOR) {
                int op_sort;
                op_sort = spdGetAttrValInt(buf[i].objectId, "sort");
                switch (op_sort) {
                  case OP_MUL:
                  case OP_DIV:
                  case OP_MOD:
                  case OP_ADD:
                  case OP_SUB:
                  case OP_SHL:
                  case OP_SHR:
                  case OP_LT:
                  case OP_LE:
                  case OP_GT:
                  case OP_GE:
                  case OP_EQ:
                  case OP_NE:
                  case OP_B_AND:
                  case OP_B_XOR:
                  case OP_B_OR:
                  case OP_L_AND:
                  case OP_L_OR:
                  case OP_QUESTION:
                  case OP_COLON:
                  case OP_ASSIGN:
                  case OP_A_ADD:
                  case OP_A_SUB:
                  case OP_A_MUL:
                  case OP_A_DIV:
                  case OP_A_MOD:
                  case OP_A_AND:
                  case OP_A_XOR:
                  case OP_A_OR:
                  case OP_A_SHL:
                  case OP_A_SHR:
                    res = spdStrCat(res, " ");
                    break;
                  default:
                    break;
                }
                res = spdStrCat(res, spdStrOperator(buf[i].objectId));
                switch (op_sort) {
                  case OP_PLUS:
                  case OP_MINUS:
                  case OP_MUL:
                  case OP_DIV:
                  case OP_MOD:
                  case OP_ADD:
                  case OP_SUB:
                  case OP_SHL:
                  case OP_SHR:
                  case OP_LT:
                  case OP_LE:
                  case OP_GT:
                  case OP_GE:
                  case OP_EQ:
                  case OP_NE:
                  case OP_B_AND:
                  case OP_B_XOR:
                  case OP_B_OR:
                  case OP_L_AND:
                  case OP_L_OR:
                  case OP_QUESTION:
                  case OP_COLON:
                  case OP_ASSIGN:
                  case OP_A_ADD:
                  case OP_A_SUB:
                  case OP_A_MUL:
                  case OP_A_DIV:
                  case OP_A_MOD:
                  case OP_A_AND:
                  case OP_A_XOR:
                  case OP_A_OR:
                  case OP_A_SHL:
                  case OP_A_SHR:
                  case OP_COMMA:
                  case OP_SIZEOF:
                    res = spdStrCat(res, " ");
                    break;
                  default:
                    break;
                }
            } else if(class_id == SPD_CID_CONSTANT) {
                res = spdStrCat(res, spdStrConstant(buf[i].objectId));
            } else if(class_id == SPD_CID_IDENTIFIER) {
                res = spdStrCat(res, spdStrIdentifier(buf[i].objectId));
            }
        }
        spdFree(buf);
        break;
    }
    
    DW_FUNC_OUT("expression", res);
    
    if (save != NULL){
        spdFree(save);
        save = NULL;
    }

    save = res;

#ifdef BUGGY_SPDSTR_0102
    if (buggyFlag) {
        spdSetOption("mode", modeOption);
    }
#endif

    return (res);
}

SpdString spdStrConstant(SpdObjId objectId)
{
    SpdString           res = NULL;

    static SpdString    save = NULL;

    DW_FUNC_IN("constant", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 constant ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_CONSTANT) {
        return res;
    }

    res = spdStrCat(res, spdGetAttrValString(objectId, "value"));
    DW_FUNC_OUT("constant", res);
    save = res;
    return res;
}

SpdString spdStrOperator(SpdObjId objectId)
{
    SpdString   res = NULL;
    int         type_id_sizeof;
    int         type_id_cast;

    static SpdString    save = NULL;

#ifdef BUGGY_SPDSTR_0102
    SpdBoolean  buggyFlag = FALSE;
    int         modeOption;
#endif

    DW_FUNC_IN("operator", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_OPERATOR) {
        return res;
    }

    res = spdStrCat(res, spdStrMacro(objectId));

#ifdef BUGGY_SPDSTR_0102
    if (res != NULL && SPD_STREQ(res, SPDSTR_ERROR_MACRO_IN_MACRO)) {
        modeOption = spdGetOption("mode");
        spdSetOption("mode", SAPID_I_MODE);
        buggyFlag = TRUE;
        spdFree(res);
        res = NULL;
    }
#endif

    if(res != NULL){
        save = res;
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
      case OP_L_PAR:   res = spdStrCat(res, "("); break;
      case OP_R_PAR:   res = spdStrCat(res, ")"); break;
      case OP_L_ARRAY: res = spdStrCat(res, "["); break;
      case OP_R_ARRAY: res = spdStrCat(res, "]"); break;
      case OP_ARROW:   res = spdStrCat(res, "->"); break;
      case OP_DOT:     res = spdStrCat(res, "."); break;
      case OP_NOT:     res = spdStrCat(res, "!"); break;
      case OP_B_NOT:   res = spdStrCat(res, "~"); break;
      case OP_INC:     res = spdStrCat(res, "++"); break;
      case OP_DEC:     res = spdStrCat(res, "--"); break;
      case OP_PLUS:    res = spdStrCat(res, "+"); break;
      case OP_MINUS:   res = spdStrCat(res, "-"); break;
      case OP_POINTER: res = spdStrCat(res, "*"); break;
      case OP_ADDRESS: res = spdStrCat(res, "&"); break;
      case OP_MUL:     res = spdStrCat(res, "*"); break;
      case OP_DIV:     res = spdStrCat(res, "/"); break;
      case OP_MOD:     res = spdStrCat(res, "%"); break;
      case OP_ADD:     res = spdStrCat(res, "+"); break;
      case OP_SUB:     res = spdStrCat(res, "-"); break;
      case OP_SHL:     res = spdStrCat(res, "<<"); break;
      case OP_SHR:     res = spdStrCat(res, ">>"); break;
      case OP_LT:      res = spdStrCat(res, "<"); break;
      case OP_LE:      res = spdStrCat(res, "<="); break;
      case OP_GT:      res = spdStrCat(res, ">"); break;
      case OP_GE:      res = spdStrCat(res, ">="); break;
      case OP_EQ:      res = spdStrCat(res, "=="); break;
      case OP_NE:      res = spdStrCat(res, "!="); break;
      case OP_B_AND:   res = spdStrCat(res, "&"); break;
      case OP_B_XOR:   res = spdStrCat(res, "^"); break;
      case OP_B_OR:    res = spdStrCat(res, "|"); break;
      case OP_L_AND:   res = spdStrCat(res, "&&"); break;
      case OP_L_OR:    res = spdStrCat(res, "||"); break;
      case OP_QUESTION:res = spdStrCat(res, "?"); break;
      case OP_COLON:   res = spdStrCat(res, ":"); break;
      case OP_ASSIGN:  res = spdStrCat(res, "="); break;
      case OP_A_ADD:   res = spdStrCat(res, "+="); break;
      case OP_A_SUB:   res = spdStrCat(res, "-="); break;
      case OP_A_MUL:   res = spdStrCat(res, "*="); break;
      case OP_A_DIV:   res = spdStrCat(res, "/="); break;
      case OP_A_MOD:   res = spdStrCat(res, "%="); break;
      case OP_A_AND:   res = spdStrCat(res, "&="); break;
      case OP_A_XOR:   res = spdStrCat(res, "^="); break;
      case OP_A_OR:    res = spdStrCat(res, "|="); break;
      case OP_A_SHL:   res = spdStrCat(res, "<<="); break;
      case OP_A_SHR:   res = spdStrCat(res, ">>="); break;
      case OP_COMMA:   res = spdStrCat(res, ","); break;
      case OP_SIZEOF:  res = spdStrCat(res, "sizeof "); break;
      case OP_SIZEOF_TYPE:
        res = spdStrCat(res, "sizeof(");
        type_id_sizeof = spdGetARelObj(objectId, "op_type", "any");
        res = spdStrCat(res, spdStrTypeSpecifier(type_id_sizeof));

#ifdef BUGGY_SDBIN4_0110
        res = spdStrCat(res, spdStrTypeQualifier(type_id_sizeof, 
                                                 NULL, 0, TRUE));
#else  /* BUGGY_SDBIN4_0110 */
        res = spdStrCat(res, spdStrTypeQualifier(type_id_sizeof, NULL, 0));
#endif /* BUGGY_SDBIN4_0110 */

        res = spdStrCat(res, ")");
        break;
      case OP_CAST:
        res = spdStrCat(res, "(");
        type_id_cast = spdGetARelObj(objectId, "op_type", "any");
        res = spdStrCat(res, spdStrTypeSpecifier(type_id_cast));

#ifdef BUGGY_SDBIN4_0110
        res = spdStrCat(res, spdStrTypeQualifier(type_id_cast, 
                                                 NULL, 0, TRUE));
#else  /* BUGGY_SDBIN4_0110 */
        res = spdStrCat(res, spdStrTypeQualifier(type_id_cast, NULL, 0));
#endif /* BUGGY_SDBIN4_0110 */

        res = spdStrCat(res, ")");
        break;
      default:
        break;
    }

    DW_FUNC_OUT("operator", res);

    save = res;

#ifdef BUGGY_SPDSTR_0102
    if (buggyFlag) {
        spdSetOption("mode", modeOption);
    }
#endif

    return (res);
}
