/* 
 * Program:             $RCSfile: spdStr_file.c,v $  $Revision: 30.20 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu, S.Kawai and Y.Momiyama  1995/12/07
 * Modified:            S.Yamamoto  1995/12/08
 * 
 * (C) Copyright:       N.Fukuyasu, S.Kawai, Y.Momiyama
 *                      and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr_file.c,v 30.20 2003/06/28 15:31:15 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "SpdStrSub.h"

static SpdBoolean spdStrIsRemakeFile(int mode);
static stringInfomation *spdStrFilePushStrInfo(stringInfomation *strInfo, 
                                               int *strInfoIndex, char *str, 
                                               int offset, SpdObjId objectId);
static macroNameInfomation *pushMNameInfo(macroNameInfomation *mNameInfo, 
                                          int *mNameIndex, char *name,
                                          int offset,
                                          SpdObjId pblockId,
                                          headerMacroInfomation *hMacroInfo,
                                          int hMacroIndex);
static headerMacroInfomation *makeHMacroInfo(headerMacroInfomation *hMacroInfo,
                                             int *hMacroIndex);
static int spdStrFileCompOffset(const stringInfomation *strInfo1,
                                const stringInfomation *strInfo2);
static void spdStrFileStrInfoFree(stringInfomation *strInfo, int strInfoIndex);
static void mNameInfoFree(macroNameInfomation *mNameInfo, int mNameIndex);
static void hMacroInfoFree(headerMacroInfomation *hMacroInfo, int hMacroIndex);

static SpdBoolean isExpandedInclude(SpdObjId pblockId)
{
    SpdObjId hfileId, pfileId;

    hfileId = spdGetARelObj(pblockId, "incld_hfile", "p-id");
    pfileId = spdGetARelObj(hfileId, "hfile_file", "p-id");
    return (pfileId != SPD_NON_ID);
}

static SpdBoolean spdStrIsRemakeFile(int mode)
{
    static SpdBoolean flag = FALSE;

    if (mode == SPDSTR_IS_REMAKE_FILE_QUERY) {
        return (flag);
    } else if (mode == SPDSTR_IS_REMAKE_FILE_SET) {
        flag = TRUE;
        return (TRUE);
    } else {
        spdErrorExit("Unknown mode.\n");
        SPD_DUMMY_RETURN(TRUE);
    }
}

static stringInfomation *spdStrFilePushStrInfo(stringInfomation *strInfo, 
                                               int *strInfoIndex, char *str, 
                                               int offset, SpdObjId objectId)
{
    if(str == NULL) return strInfo;  

    if(*strInfoIndex == 0){
        strInfo = spdMalloc(sizeof(stringInfomation));
    } else {
        strInfo = spdRealloc(strInfo, 
                             sizeof(stringInfomation) * (*strInfoIndex + 1));
    }
    strInfo[*strInfoIndex].str = NULL;
    strInfo[*strInfoIndex].str = spdStrCat(strInfo[*strInfoIndex].str, str);
    strInfo[*strInfoIndex].offset = offset;
    strInfo[*strInfoIndex].objectId = objectId;
    (*strInfoIndex)++;
    return (strInfo);
}

static macroNameInfomation *pushMNameInfo(macroNameInfomation *mNameInfo, 
                                          int *mNameIndex, char *name,
                                          int offset,
                                          SpdObjId pblockId,
                                          headerMacroInfomation *hMacroInfo,
                                          int hMacroIndex)
{
    int i;
    
    if(name == NULL) return mNameInfo;  

    if(*mNameIndex == 0){
        mNameInfo = spdMalloc(sizeof(macroNameInfomation));
    } else {
        mNameInfo = spdRealloc(mNameInfo,
                               sizeof(macroNameInfomation)
                               * (*mNameIndex + 1));
    }
    mNameInfo[*mNameIndex].name = NULL;

    mNameInfo[*mNameIndex].name 
        = spdStrCat(mNameInfo[*mNameIndex].name, name);
    mNameInfo[*mNameIndex].offset = offset;
    mNameInfo[*mNameIndex].pblockId = pblockId;
    mNameInfo[*mNameIndex].withUndef = FALSE;

    /* ʣ withUndef ե饰ΩƤ롥*/
    for(i = 0; i < *mNameIndex; i++){
        if(SPD_STREQ(mNameInfo[*mNameIndex].name, mNameInfo[i].name)){
            if(mNameInfo[*mNameIndex].offset > mNameInfo[i].offset){
                mNameInfo[*mNameIndex].withUndef = TRUE;
            } else {
                mNameInfo[i].withUndef = TRUE;
            }
        }
    }

    /* إå define ȽʣƤ뤫 */
    for(i = 0; i < hMacroIndex; i++){
        if(SPD_STREQ(mNameInfo[*mNameIndex].name, hMacroInfo[i].name)){
            /* if(hMacroInfo[i].offset < mNameInfo[*mNameIndex].offset){ */
            mNameInfo[*mNameIndex].withUndef = TRUE;
            /* } */
        }
    }

    (*mNameIndex)++;

    return mNameInfo;
}

static headerMacroInfomation *makeHMacroInfo(headerMacroInfomation *hMacroInfo,
                                             int *hMacroIndex)
{
    int     i, j;

    for(i = 0; i < defineInfoN - 1; i++){
        for(j = i + 1; j < defineInfoN; j++){
            if(defineInfo[i].pos.type == DEFINE_FILE &&
               defineInfo[j].pos.type == DEFINE_FILE){
                if(SPD_STREQ(defineInfo[i].name, defineInfo[j].name) &&
                   (!spdIsFileExtension(defineInfo[i].pos.fname, ".c") ||
                    !spdIsFileExtension(defineInfo[j].pos.fname, ".c"))) {
                    if(*hMacroIndex == 0){
                        hMacroInfo = spdMalloc(sizeof(headerMacroInfomation));
                    } else {
                        hMacroInfo = spdRealloc(hMacroInfo,
                                                sizeof(headerMacroInfomation)
                                                * (*hMacroIndex + 1));
                    }
                    hMacroInfo[*hMacroIndex].name = NULL;
                    
                    hMacroInfo[*hMacroIndex].name 
                        = spdStrCat(hMacroInfo[*hMacroIndex].name, 
                                    defineInfo[i].name);

                    /* إå *.c ե
                       offset ʪʤΤǤǤϡ */
                    /*
                       hMacroInfo[*hMacroIndex].offset 
                       = SPD_MIN(defineInfo[i].pos.start, 
                       defineInfo[j].pos.start);
                       */
                    
                    (*hMacroIndex)++;
                }
            }
        }
    }
   
    return hMacroInfo;
}

/* Ӵؿ */

static int spdStrFileCompOffset(const stringInfomation *strInfo1,
                                const stringInfomation *strInfo2)
{
    return (strInfo1->offset - strInfo2->offset);
}

static void spdStrFileStrInfoFree(stringInfomation *strInfo, int strInfoIndex)
{
    int     i;

    for (i = 0; i < strInfoIndex; i++) {
        spdFree(strInfo[i].str);
    }
    spdFree(strInfo);
}

static void mNameInfoFree(macroNameInfomation *mNameInfo, int mNameIndex)
{
    int  i;

    for (i = 0; i < mNameIndex; i++) {
        spdFree(mNameInfo[i].name);
    }
    spdFree(mNameInfo);
}

static void hMacroInfoFree(headerMacroInfomation *hMacroInfo, int hMacroIndex)
{
    int     i;

    for (i = 0; i < hMacroIndex; i++) {
        spdFree(hMacroInfo[i].name);
    }
    spdFree(hMacroInfo);
}

/*
 * pblockdeclaration˴ޤޤƤ뤫ɤȽꤹؿޤޤƤ
 * TRUEޤޤƤʤFALSE֤
 * (declarationoffset) < (pblockoffset)Ȥ롥
 */

static SpdBoolean pblockIsIncluded(stringInfomation pblock,
                                   SpdObjId declId, SpdObjId fileId)
{
    SpdRegion decl_region, decl_C_region;
    SpdRelId rel_id;
    static int end_offset = 0;                  /* declarationκǸΰ֡ */
    static SpdObjId prev_declId = SPD_NON_ID; /* Ĵ٤declId. */

    if (prev_declId != declId) {
        prev_declId = declId;
        /* declId".i"ˤoffsetlength롥 */
        rel_id = spdGetARelId(fileId, declId, "file_decl");
        decl_region.fileId = fileId;                    /* .i  fileId */
        decl_region.offset = spdGetOffset(rel_id);      /* .i  offset */
        decl_region.length = spdGetLength(declId);      /* .i  length */

        /* ".c"ˤoffsetlengthѴ롥 */
        decl_C_region = spdConvRegionItoC(decl_region);
        /* ".c"ˤdeclarationκǸΰ֤롥 */
        end_offset = decl_C_region.offset + decl_C_region.length;
    }

    return (pblock.offset < end_offset);
}

SpdString spdStrFile (SpdObjId objectId)
{
    static SpdString    save = NULL;

    SpdString       res = NULL;
    SpdRel          decl_rel, *decl_buf = NULL; 
    int             i, j, decl_buf_n = 0;
    SpdCursor       *csr;
    SpdOcc          occ;
    SpdObjId        pfileId, cfileId, macroPatternId, macroNameId;
    SpdObjIdArray   block_buf;
    SpdRegion       region;
    int             tmpOffset;

    stringInfomation        *strInfo = NULL;
    int                     strInfoIndex = 0;

    macroNameInfomation     *mNameInfo = NULL;
    int                     mNameIndex = 0; 

    headerMacroInfomation   *hMacroInfo = NULL;
    int                     hMacroIndex = 0;

#ifdef DEBUG_INCLUDE_960830
    SpdObjId        type_id;
#endif

    DW_FUNC_IN("file", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 file ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_FILE) {
        return res;
    }

    spdStrIsRemakeFile(SPDSTR_IS_REMAKE_FILE_SET);

    if (spdGetOption("mode") == SAPID_C_MODE) {
        hMacroInfo = makeHMacroInfo(hMacroInfo, &hMacroIndex);
    }

    csr = spdGetRelInit(objectId, "file_decl", "p-id");
    while ((decl_rel = spdGetRel(csr)).relationId != -1) {
        decl_buf = spdRealloc(decl_buf, sizeof(SpdRel) * (decl_buf_n + 1));
        decl_buf[decl_buf_n++] = decl_rel;
    }
    spdFreeCursor(csr);

    for (i = 0; i < decl_buf_n; i++) {
        if (spdGetOption("mode") == SAPID_C_MODE) {
            region.fileId = objectId;
            region.offset = spdGetOffset(decl_buf[i].relationId);
            region.length = 1;
            region = spdConvRegionItoC(region);

            tmpOffset = region.offset;
        } else {
            tmpOffset = spdGetOffset(decl_buf[i].relationId);
        }
        if (tmpOffset < 0) {
            spdMessage("Illegal offset of declaration : objId: %d relId: %d\n",
                       decl_buf[i].objectId, decl_buf[i].relationId);
        }

        strInfo = spdStrFilePushStrInfo(strInfo, &strInfoIndex, 
                                        spdStrDeclaration(decl_buf[i].objectId,
                                                          FALSE),
                                        tmpOffset, decl_buf[i].objectId);
    }
    if (decl_buf != NULL) spdFree(decl_buf);

    DW_FUNC_OUT("file", res);

    if (spdGetOption("mode") == SAPID_C_MODE) {
        /* Commented by S.Yamamoto 2000/01/20. */
        /* SpdString tmp = NULL; */

        pfileId = spdGetARelObj(objectId, "cfile_ifile", "file_id");
        cfileId = spdGetCfileIdByPfileId(pfileId);

        /* include ʸ */
        block_buf = spdGetRelObjArray(cfileId, "file_block", "any", 
                                      spdObjIdIsInclude);
        for(i = 0; i < block_buf.size; i++){
            if (!isExpandedInclude(block_buf.id[i])) {
                continue;
            }
            occ = spdGetAnOffsetOfObj(block_buf.id[i]);
            strInfo 
                = spdStrFilePushStrInfo(strInfo, &strInfoIndex, 
                                        spdStrPblockCutNull(block_buf.id[i]),
                                        occ.offset, block_buf.id[i]);
        }
        spdFreeObjIdArray(block_buf);

        /* define ʸ */
        block_buf = spdGetRelObjArray(cfileId, "file_block", "any", 
                                      spdObjIdIsDefine);
        for (i = 0; i < block_buf.size; i++) {
            occ = spdGetAnOffsetOfObj(block_buf.id[i]);
            macroPatternId 
                = spdGetARelObj(block_buf.id[i], "def_mpatt", "p-id");
            macroNameId
                = spdGetARelObj(macroPatternId, "mpatt_mname", "p-id");
            
            mNameInfo 
                = pushMNameInfo(mNameInfo, &mNameIndex, 
                                spdStrMacroname(macroNameId), occ.offset, 
                                block_buf.id[i], hMacroInfo, hMacroIndex);
        }
        
        for (i = 0; i < mNameIndex; i++) {
            SpdString tmp = NULL;

            if (mNameInfo[i].withUndef) {
                tmp = spdStrCat(tmp, "#undef ");
                tmp = spdStrCat(tmp, mNameInfo[i].name);
                tmp = spdStrCat(tmp, "\n");
            }
            
            tmp = spdStrCat(tmp, spdStrPblockCutNull(mNameInfo[i].pblockId));
            strInfo = spdStrFilePushStrInfo(strInfo, &strInfoIndex, 
                                            tmp, mNameInfo[i].offset,
                                            block_buf.id[i]);
            spdFree(tmp);
        }
        
        spdFreeObjIdArray(block_buf);
    }

    SPD_QSORT(strInfo, strInfoIndex, sizeof(stringInfomation), 
                spdStrFileCompOffset); 
    
    /* 
     * Ȥ줿declaration, include, define
     * ϤˡǤdeclarationinclude, define
     * declarationθˤʤꡤѥǤʤʤ礬롥
     * 
     * Ūʽ֤ȤơΤ褦include, define
     * declaration˽Ф褦ˤ
     */
    for(i = 0; i < strInfoIndex; i++){
        if((spdGetClassId(strInfo[i].objectId) 
            == SPD_CID_DECLARATION)){
            for(j = i + 1; j < strInfoIndex; j++){
                if ((spdGetClassId(strInfo[j].objectId)
                     != SPD_CID_PBLOCK)
                    || !pblockIsIncluded(strInfo[j], strInfo[i].objectId,
                                         objectId)) break;
                res = spdStrCat(res, strInfo[j].str);
                res = spdStrCat(res, "\n\n");
            }
            res = spdStrCat(res, strInfo[i].str);
            res = spdStrCat(res, "\n\n");
            i = j - 1;
        } else {
            res = spdStrCat(res, strInfo[i].str);
            res = spdStrCat(res, "\n\n");
        }
    }

    spdStrFileStrInfoFree(strInfo, strInfoIndex);
    mNameInfoFree(mNameInfo, mNameIndex);
    hMacroInfoFree(hMacroInfo, hMacroIndex);

    save = res;
    return res;
}                   

SpdString spdStrAppDeclaration(SpdObjId fileId, SpdObjId declId)
{
    static SpdString save = NULL;
    SpdString   res = NULL;
    static SpdObjId *decl_table = NULL;
    static int table_num = 0;
    static SpdObjId saved_fileId = 0;
    SpdObjId type_id, ident_id, decl_id, file_id;
    SpdRel decr_rel;
    SpdCursor *decr_csr;
    SpdOccArray decl_occ;
    int i, type_sort;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* fileIdθƤӽФȰۤʤȤdecl_table롥 */
    if (fileId != saved_fileId) {
        saved_fileId = fileId;
        if (decl_table != NULL) {
            spdFree(decl_table);
            decl_table = NULL;
            table_num = 0;
        }
    }

    /* fileId Υ饹 file ǤʤС뤤ϡ
       declId Υ饹 declaration Ǥʤ NULL ֤ */
    if ((spdGetClassId(fileId)) != SPD_CID_FILE ||
        (spdGetClassId(declId)) != SPD_CID_DECLARATION) {
        return res;
    }

    type_id = spdGetARelObj(declId, "decl_type", "declaration_id");
    /* type  sort Υå
       TYPE_TYPE_DEF or TYPE_STRUCT_TAG or TYPE_ENUM_SPECIFIER ʤ */
    switch (type_sort = spdGetAttrValInt(type_id, "sort")) {
        /* if(decl -> type -> identifier -> declaration -> file != fileId)
           decl_table ϿƤ뤫ɤγǧ
           ϿƤʤϿ
           res = spdStrCat(res, spdStrDeclaration(declaration)); */
        /* TYPE_TYPE_DEF or TYPE_STRUCT_TAG ʤ */
        /* ˴ޤޤƤ declaration Ф
           spdStrAppDeclaration(fileId, declaration); */
      case TYPE_TYPE_DEF:
      case TYPE_STRUCT_TAG:
      case TYPE_ENUM_SPECIFIER:
        if ((ident_id = spdGetARelObj(type_id, "type_is",
                                      "type_id")) == SPD_NON_ID) break;
        decr_csr = spdGetRelInit(ident_id, "decl_ident", "identifier_id");
        while ((decr_rel = spdGetRel(decr_csr)).objectId
               != SPD_NON_ID) {
            if ((decl_id = spdGetARelObj(decr_rel.objectId, "decl_decl",
                                         "declarator_id"))
                != SPD_NON_ID) break;
        }
        spdFreeCursor(decr_csr);
#ifdef DEBUG_INCLUDE_960830
        fprintf(stderr, "%d: %s\n", decr_rel.objectId,
                spdStrDeclarator(decr_rel.objectId, FALSE));
#endif /* DEBUG_INCLUDE_960830 */
        file_id = spdGetARelObj(decl_id, "file_decl", "declaration_id");
#ifdef DEBUG_INCLUDE_960830_1
        fprintf(stderr,
                "type %d -> ident %d -> decr %d -> decl %d -> file %d\n",
                type_id, ident_id, decr_rel.objectId,
                decl_id, file_id);
#endif /* DEBUG_INCLUDE_960830_1 */
        if (file_id != SPD_NON_ID && file_id != fileId) {
            i = 0;
            while (i < table_num && decl_id != decl_table[i]) i++;
#ifdef DEBUG_INCLUDE_960830_1
            fprintf(stderr, "i = %d, table_num = %d\n", i, table_num);
#endif /* DEBUG_INCLUDE_960830_1 */
            if (i == table_num) {
                decl_table = spdRealloc(decl_table,
                                        sizeof(SpdObjId) * (table_num+1));
                decl_table[table_num++] = decl_id;
                if (type_sort != TYPE_ENUM_SPECIFIER) {
                    decl_occ = spdGetIncludedOccArray(decl_id,
                                                      "declaration", 
                                                      fileId, NULL);
                    for (i = 0; i < decl_occ.size; i++) {
                        SpdString tmp = NULL;
                        tmp = spdStrAppDeclaration(fileId,
                                                   decl_occ.occ[i].objectId);
                        if (tmp != NULL) {
                            res = spdStrCat(res, tmp);
                            res = spdStrCat(res, "\n\n");
                        }
                    }
                    spdFreeOccArray(decl_occ);
                }
                res = spdStrCat(res, spdStrDeclaration(decl_id, FALSE));
#ifdef DEBUG_INCLUDE_960830
                fprintf(stderr, "%d: %s\n",
                        decl_id, spdStrDeclaration(decl_id, FALSE));
#endif /* DEBUG_INCLUDE_960830 */
            }
        }
        break;
      default:
        break;
    }

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }
    save = res;
    return res;
}

SpdString spdStrAppPrototype(SpdObjId fileId, SpdObjId exprId)
{
    static SpdString save = NULL;
    SpdString   res = NULL;
    static SpdObjId *ident_table = NULL;
    static int table_num = 0;
    static SpdObjId saved_fileId = 0;
    SpdRelArray expr_rel, decr_rel;
    SpdObjId ident_id, decl_id, file_id, decr_id;
    SpdOccArray decl_occ;
    int i, j;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* fileId ƤӽФ줿Ȱۤʤä ident_table  free */
    if (fileId != saved_fileId) {
        saved_fileId = fileId;
        if (ident_table != NULL) {
            spdFree(ident_table);
            ident_table = NULL;
            table_num = 0;
        }
    }

    /* fileId Υ饹 file ǤʤС뤤ϡ
       exprId Υ饹 expression Ǥʤ NULL ֤ */
    if ((spdGetClassId(fileId)) != SPD_CID_FILE ||
        (spdGetClassId(exprId)) != SPD_CID_EXPRESSION) {
        return res;
    }

    expr_rel = spdGetRelSortedArray(exprId, "expr_expr",
                                    "outer_expression_id",
                                    NULL, spdCompareRelOffset);
    ident_id = spdGetARelObj(expr_rel.rel[0].objectId, "ident_ref",
                             "expression_id");
    spdFreeRelArray(expr_rel);

    /* prototype ̵ͭγǧ */
    decr_rel = spdGetRelArray(ident_id, "decl_ident", "identifier_id", NULL);
    for (j = 0; j < decr_rel.size; j++) {
        decr_id = decr_rel.rel[j].objectId;
        if (spdGetAttrValInt(decr_id, "sort") == DECL_PROTOTYPE) {
            /* declaration νи file ǧۤʤäƤ
               ident_table Ĵ٤ʤɲä prototype 
                */
            decl_id = spdGetARelObj(decr_id, "decl_decl", "declarator_id");
            file_id = spdGetARelObj(decl_id, "file_decl", "declaration_id");
            /* file_id  ʬȤ˲Ϥ줿եǤȤ
               ʤ궯ID Ĵ٤뤷ʤ*/
            if (file_id != SPD_NON_ID && file_id < fileId) {
                i = 0;
                while (i < table_num && ident_id != ident_table[i]) i++;
                if (i == table_num) {
                    ident_table = spdRealloc(ident_table,
                                             sizeof(SpdObjId) * (table_num+1));
                    ident_table[table_num++] = ident_id;

#ifdef DEBUG_INCLUDE_960914_TABLE
                    fprintf(stderr,
                            "AppProto: fileId: %d ident_table: %d file_id: %d ident_id: %d %s\n",
                            fileId, table_num, file_id, ident_id,
                            spdStrIdentifier(ident_id));
#endif /* DEBUG_INCLUDE_960914_TABLE */

                    {
                        SpdString tmp = NULL;

                        tmp = spdStrAppDeclaration(fileId, decl_id);
                        if (tmp != NULL) {
                            res = spdStrCat(res, tmp);
                            res = spdStrCat(res, "\n\n");
                        }

#ifdef DEBUG_INCLUDE_960914
                        fprintf(stderr, "AppProto: AppDecl: %d %s\n",
                                decl_id, tmp);
#endif /* DEBUG_INCLUDE_960914 */

                    }
                    decl_occ = spdGetIncludedOccArray(decl_id,
                                                      "declaration", 
                                                      fileId, NULL);
                    for (i = 0; i < decl_occ.size; i++) {
                        SpdString tmp = NULL;
                        tmp = spdStrAppDeclaration(fileId,
                                                   decl_occ.occ[i].objectId);
                        if (tmp != NULL) {
                            res = spdStrCat(res, tmp);
                            res = spdStrCat(res, "\n\n");
                        }
                    }
                    spdFreeOccArray(decl_occ);
                    res = spdStrCat(res, spdStrDeclaration(decl_id, FALSE));
#ifdef DEBUG_INCLUDE_960914
                    fprintf(stderr,
                            "AppProto: decl_id: %d %s\n",
                            decl_id, spdStrDeclaration(decl_id, FALSE));
#endif /* DEBUG_INCLUDE_960914 */

                }
            }
            break;
        }
    }
    spdFreeRelArray(decr_rel);

    save = res;
    return res;
}

SpdString spdStrDeclaration(SpdObjId objectId, SpdBoolean asProtoParm)
{
    static SpdString save = NULL;

    SpdString   res = NULL;
    SpdObjId    obj_id;
    SpdCursor   *csr;
    SpdRel      decl_rel, *decl_buf;
    SpdRelArray decl_s_buf;
    int         decl_buf_n, i, j;
    int         opt_proto;

#ifdef BUGGY_SDBIN3_0057
    SpdObjIdArray   decl_array;
#endif /* BUGGY_SDBIN3_0057 */

#ifdef BUGGY_SPDSTR_0102
    SpdBoolean buggyFlag = FALSE;
    int modeOption;
#endif /* BUGGY_SPDSTR_0102 */


    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 declaration ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_DECLARATION) {
        return res;
    }

    if (spdGetOption("mode") == SAPID_C_MODE 
        && spdStrIsRemakeFile(SPDSTR_IS_REMAKE_FILE_QUERY)) {
        if (isInHeader(objectId)) {
            return NULL;
        }
    }

    res = spdStrCat(res, spdStrMacro(objectId));

#ifdef BUGGY_SPDSTR_0102
    if(res != NULL){
        if(SPD_STREQ(res, SPDSTR_ERROR_MACRO_IN_MACRO)){
            modeOption = spdGetOption("mode");
            spdSetOption("mode", SAPID_I_MODE);
            buggyFlag = TRUE;
            spdFree(res);
            res = NULL;
        }
    }
#endif /* BUGGY_SPDSTR_0102 */

    if (res != NULL) {
        save = res;
        return res;
    }

#ifdef BUGGY_SDBIN3_0057
    /* ץȥǤ褦 declarator 
       Τ褦 declarator ¸ߤ text ֤ */
    decl_array = spdGetRelObjArray(objectId, "decl_decl", "p-id",
                                   spdObjIsDeclPrototype);
    if (decl_array.size != 0) {
        res = spdStrCat(res, spdGetObjText(objectId));
        save = res;
        return res;
    }
    spdFreeObjIdArray(decl_array);
#endif /* BUGGY_SDBIN3_0057 */

    /* 饹(storage) */
    switch (spdGetAttrValInt(objectId, "storage")) {
      case STORAGE_EXTERN:
        res = spdStrCat(res, "extern ");
        break;
      case STORAGE_STATIC:
        res = spdStrCat(res, "static ");
        break;
      case STORAGE_AUTO:
        res = spdStrCat(res, "auto ");
        break;
      case STORAGE_REGISTER:
        res = spdStrCat(res, "register ");
        break;
      case STORAGE_TYPEDEF:
#ifdef BUGGY_SDBIN3_0058
        /*
         * typedef int INT(); ȤޤƤʤ
         *  declaration ˳̤ޤޤƤ餳ηǤȤ
         * declaration  text 롥
         *  typedef int (INT); Ȥ褦ʤΤ
         * 줬櫓ǤϤʤ
         */
        {
            SpdString   obj = NULL;
            int i;

            obj = spdGetObjText(objectId);

            for (i = 0; i < strlen(obj); i++)
                if (obj[i] == '(') {
                    res = spdStrCat(res, obj);
#ifdef DEBUG_0058
                    fprintf(stderr, "typedef prototype: %d\n%s\n", objectId, res);
#endif /* DEBUG_0058 */
                    save = res;
                    return res;
                }
        }
#endif /* BUGGY_SDBIN3_0058 */
        res = spdStrCat(res, "typedef ");
        break;
      default:
        break;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
      case DECL_NORMAL:                     /* ѿΤʤؿ */
      case DECL_TYPE:                       /*  */
      case DECL_MEMBER:                     /* ¤ΡΤΥФ */
        obj_id = spdGetARelObj(objectId, "decl_type", "p-id");
        /* 
         * i = 0;
         * while (rel_class[i] != NULL &&
         *        (offset = spdGetOffset(spdGetARel(objectId, rel_class[i++],
         *                                          "c-id").relationId)) == -1);
         */
        res = spdStrCat(res, spdStrTypeSpecifier(obj_id));
        /* 
         * decl_type  declaration η򼨤ƤʤȤ
         * Х
         * > ϡΤ褦ˤ롥
         * > res = spdStrCat(res, spdStrType(obj_id));
         * >                         ^^^^^^^^^^
         * struct ηɽˡoffset ˤȽǤäΤǡ
         * spdStrType() Ѥʤ褦˻פ롥
         */
        decl_buf = NULL;
        decl_buf_n = 0;
        csr = spdGetRelInit(objectId, "decl_decl", "p-id");
        while ((decl_rel = spdGetRel(csr)).relationId != -1) {
            decl_buf = spdRealloc(decl_buf, sizeof(SpdRel) * (decl_buf_n + 1));
            decl_buf[decl_buf_n++] = decl_rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(decl_buf, decl_buf_n, sizeof(SpdRel), cmp_offset);
        if (spdGetAttrValInt(decl_buf[0].objectId, "sort") == DECL_TAG) {
            j = 1;
        } else {
            j = 0;
        }
        /* 
         * struct(union,enum)򼨤declaratorΤƤ롥
         * ϡtypeȤƹǤ롥
         * structdeclaratortypeǹǤʤϡ
         * for (i = 0; i < decl_buf_n; i++) {
         */
        if (j == 1) {
            SpdObjId opt_id;
            if ((opt_id = spdGetARelObj(decl_buf[0].objectId, "decl_opt",
                                        "p-id")) != SPD_NON_ID) {
                res = spdStrCat(res, " ");
                res = spdStrCat(res, spdStrOptionaldecl(opt_id));
            }
        }
        for (i = j; i < decl_buf_n; i++) {
            if (i > j) {
                res = spdStrCat(res, ", ");
            } else {
                res = spdStrCat(res, " ");
            }
            res = spdStrCat(res,
                            spdStrDeclarator(decl_buf[i].objectId, FALSE));
        }
        spdFree(decl_buf);
        res = spdStrCat(res, ";");
        break;
      case DECL_FUNCDEF:                    /* ؿ */
        if ((obj_id = spdGetARelObj(objectId, "decl_type", "p-id")) !=
            SPD_NON_ID) {
            /* i = 0;
               while (rel_class[i] != NULL &&
               (offset = spdGetOffset(spdGetARel(objectId, rel_class[i++],
               "c-id").relationId)) == -1);
               */
            res = spdStrCat(res, spdStrTypeSpecifier(obj_id));
        }
        /* parameter 
           SDB Ǥ struct  TAG  declarator ȤưäƤ뤿
           decl_decl ʣ¸ߤ롥
           äƼˡǤϤޤʤ
           > obj_id = spdGetARelObj(objectId, "decl_decl", "p-id");
           > res = spdStrCat(res, spdStrDeclarator(obj_id));
            offset κΤΤѤ롥
           */
        /* decl_buf = NULL;
           decl_buf_n = 0;
           csr = spdGetRelInit(objectId, "decl_decl", "p-id");
           while ((decl_rel = spdGetRel(csr)).relationId != -1) {
           decl_buf = spdRealloc(decl_buf, sizeof(SpdRel) * (decl_buf_n + 1));
           decl_buf[decl_buf_n++] = decl_rel;
           }
           spdFreeCursor(csr);
           SPD_QSORT(decl_buf, decl_buf_n, sizeof(SpdRel), cmp_offset);
           */
        decl_s_buf = spdGetRelSortedArray(objectId, "decl_decl",
                                          "declaration_id", NULL,
                                          spdCompareRelOffset);
        if (decl_s_buf.size > 0) {
            SpdObjId decl_id;
            SpdString tmp = NULL;
            decl_id = decl_s_buf.rel[decl_s_buf.size-1].objectId;
            tmp = spdStrDeclarator(decl_id, FALSE);
            if (obj_id != SPD_NON_ID && tmp != NULL) {
                /* type specifier ¸ߤ declarator  NULL Ǥʤ */
                res = spdStrCat(res, " ");
                res = spdStrCat(res, tmp);
            }
        }
        spdFreeRelArray(decl_s_buf);
        break;
      case DECL_PARAMETER:                  /* ؿˤ롤 parameter */
        if (asProtoParm == FALSE && spdObjIsFuncParameter(objectId)) {
            if ((obj_id = spdGetARelObj(objectId, "decl_type", "p-id")) !=
                SPD_NON_ID) {
                /* i = 0;
                   while (rel_class[i] != NULL &&
                   (offset = spdGetOffset(spdGetARel(objectId, rel_class[i++],
                   "c-id").relationId)) == -1);
                   */
                res = spdStrCat(res, spdStrTypeSpecifier(obj_id));
            }
            /* parameter 
               SDB Ǥ struct  TAG  declarator ȤưäƤ뤿
               decl_decl ʣ¸ߤ롥
               äƼˡǤϤޤʤ
               > obj_id = spdGetARelObj(objectId, "decl_decl", "p-id");
               > res = spdStrCat(res, spdStrDeclarator(obj_id));
                offset κΤΤѤ롥
               */
            /* decl_buf = NULL;
               decl_buf_n = 0;
               csr = spdGetRelInit(objectId, "decl_decl", "p-id");
               while ((decl_rel = spdGetRel(csr)).relationId != -1) {
               decl_buf = spdRealloc(decl_buf, sizeof(SpdRel) * (decl_buf_n + 1));
               decl_buf[decl_buf_n++] = decl_rel;
               }
               spdFreeCursor(csr);
               SPD_QSORT(decl_buf, decl_buf_n, sizeof(SpdRel), cmp_offset);
               */
            decl_s_buf = spdGetRelSortedArray(objectId, "decl_decl",
                                              "declaration_id", NULL,
                                              spdCompareRelOffset);
            if (decl_s_buf.size > 0) {
                SpdObjId decl_id;
                SpdString tmp = NULL;
                decl_id = decl_s_buf.rel[decl_s_buf.size-1].objectId;
                tmp = spdStrDeclarator(decl_id, FALSE);
                if (obj_id != SPD_NON_ID && tmp != NULL) {
                    /* type specifier ¸ߤ declarator  NULL Ǥʤ */
                    res = spdStrCat(res, " ");
                }
                res = spdStrCat(res, tmp);
            }
            spdFreeRelArray(decl_s_buf);
        } else {                            /* asProtoParm == TRUE || !isFuncParmeter */
            SpdString parm_str = NULL;
            opt_proto = spdGetOption("prototype");
            switch (opt_proto) {
              case SPDSTR_PROTOTYPE_NON:
                if ((obj_id = spdGetARelObj(objectId, "decl_type", "p-id")) !=
                    SPD_NON_ID) {
                    parm_str = spdStrCat(parm_str,
                                         spdStrTypeSpecifier(obj_id));
                }
                break;
              case SPDSTR_PROTOTYPE_TYPE:
              case SPDSTR_PROTOTYPE_ALL:
                if ((obj_id = spdGetARelObj(objectId, "decl_type", "p-id")) !=
                    SPD_NON_ID) {
                    parm_str = spdStrCat(parm_str,
                                         spdStrTypeSpecifier(obj_id));
                }
                break;
              case SPDSTR_PROTOTYPE_IDENT:
                break;
            }
            switch (opt_proto) {
              case SPDSTR_PROTOTYPE_NON:
                decl_s_buf = spdGetRelSortedArray(objectId, "decl_decl",
                                                  "declaration_id", NULL,
                                                  spdCompareRelOffset);
                if (decl_s_buf.size > 0) {
                    SpdObjId decl_id;
                    SpdString tmp = NULL;
                    decl_id = decl_s_buf.rel[decl_s_buf.size-1].objectId;
                    tmp = spdStrDeclarator(decl_id, asProtoParm);
                    if (parm_str != NULL && tmp != NULL) {
                        /* type specifier ¸ߤ declarator 
                           NULL Ǥʤ */
                        parm_str = spdStrCat(parm_str, " ");
                        parm_str = spdStrCat(parm_str, tmp);
                    }
                }
                spdFreeRelArray(decl_s_buf);
                break;
              case SPDSTR_PROTOTYPE_TYPE:
              case SPDSTR_PROTOTYPE_IDENT:
              case SPDSTR_PROTOTYPE_ALL:
                decl_s_buf = spdGetRelSortedArray(objectId, "decl_decl",
                                                  "declaration_id", NULL,
                                                  spdCompareRelOffset);
                if (decl_s_buf.size > 0) {
                    SpdObjId decl_id;
                    SpdString tmp = NULL;
                    decl_id = decl_s_buf.rel[decl_s_buf.size-1].objectId;
                    tmp = spdStrDeclarator(decl_id, asProtoParm);
                    if (parm_str != NULL && tmp != NULL) {
                        /* type specifier ¸ߤ declarator 
                           NULL Ǥʤ */
                        parm_str = spdStrCat(parm_str, " ");
                    }
                    parm_str = spdStrCat(parm_str, tmp);
                }
                spdFreeRelArray(decl_s_buf);
            }
            res = spdStrCat(res, parm_str);
            if (parm_str != NULL) spdFree(parm_str);
        }
        break;
      case DECL_ENUMERATOR:                 /* θġ */
        obj_id = spdGetARelObj(objectId, "decl_decl", "p-id");
        res = spdStrCat(res, spdStrDeclarator(obj_id, FALSE));
        break;
      default:
        break;
    }

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }
    save = res;

#ifdef BUGGY_SPDSTR_0102
    if(buggyFlag) {
        spdSetOption("mode", modeOption);
    }
#endif /* BUGGY_SPDSTR_0102 */

    return (res);
}
