/* 
 * Program:             $RCSfile: spdStr_type.c,v $  $Revision: 30.10 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu, S.Kawai and Y.Momiyama  1995/12/07
 * Modified:            S.Yamamoto  1995/12/08
 * 
 * (C) Copyright:       N.Fukuyasu, S.Kawai, Y.Momiyama
 *                      and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr_type.c,v 30.10 2003/06/28 15:31:16 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "SpdStrSub.h"

SpdString spdStrType(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }

    res = spdStrCat(res, spdStrTypeSpecifier(objectId));
    res = spdStrCat(res, " ");
#ifdef BUGGY_SDBIN4_0110    
    res = spdStrCat(res, spdStrTypeQualifier(objectId, NULL, 0, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
    res = spdStrCat(res, spdStrTypeQualifier(objectId, NULL, 0));
#endif /* BUGGY_SDBIN4_0110 */

    save = res;
    return res;
}

SpdString spdStrTypeDeclaration(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    DW_FUNC_IN("type", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sign")) {
      case TYPE_UNKNOWN: break;
      case TYPE_SIGNED: res = spdStrCat(res, "signed "); break;
      case TYPE_UNSIGNED: res = spdStrCat(res, "unsigned "); break;
      default: break;
    }

    switch (spdGetAttrValInt(objectId, "size")) {
      case TYPE_UNKNOWN: break;
      case TYPE_SHORT: res = spdStrCat(res, "short "); break;
      case TYPE_LONG: res = spdStrCat(res, "long "); break;

#ifdef ACCEPT_TYPE_LONG_LONG
      case TYPE_LONG_LONG: res = spdStrCat(res, "long long "); break;
#endif

      default: break;
    }

    switch (spdGetAttrValInt(objectId, "type")) {
      case TYPE_UNKNOWN: break;
      case TYPE_VOID: res = spdStrCat(res, "void"); break;
      case TYPE_INT: res = spdStrCat(res, "int"); break;
      case TYPE_CHAR: res = spdStrCat(res, "char"); break;
      case TYPE_FLOAT: res = spdStrCat(res, "float"); break;
      case TYPE_DOUBLE: res = spdStrCat(res, "double"); break;
      case TYPE_DOTS: res = spdStrCat(res, "..."); break;
      default: break;
    }

    DW_FUNC_OUT("type", res);

    save = res;
    return res;
}

SpdString spdStrTypeSpecifier (SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString   res = NULL;

    int         /* i, */ type_sort;
    SpdCursor   *csr;
    SpdObjId    obj_id, type_id;

#ifndef BUGGY_SDBIN3_0086
    SpdObjId    ident_id;
#endif

    DW_FUNC_IN("type_spe", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }
    
    /* Added by N.Iwamoto 1998/07/06 */
    /* type_qualifier  ON äϤ */
    if (spdGetOption("typeQualifier") == ON) {
        switch (spdGetAttrValInt(objectId, "qualifier")) {
          case TYPE_UNKNOWN: break;
          case TYPE_CONST: res = spdStrCat(res, "const "); break;
          case TYPE_VOLATILE: res = spdStrCat(res, "volatile "); break;
          default: break;
        }
    }
    
    switch (type_sort = spdGetAttrValInt(objectId, "sort")) {
      case TYPE_TYPE_DEF:
        
#if defined(BUGGY_SDBIN3_0086)
        res = spdStrCat(res, spdGetName(objectId));
#else
        ident_id = spdGetARelObj(objectId, "type_is", "p-id");
        res = spdStrCat(res, spdStrIdentifier(ident_id));
#endif
        
        break;
      case TYPE_STRUCT_TAG:
      case TYPE_UNION_TAG:
        if (type_sort == TYPE_STRUCT_TAG) {
            res = spdStrCat(res, "struct");
        } else if (type_sort == TYPE_UNION_TAG) {
            res = spdStrCat(res, "union");
        }
        
#if defined(BUGGY_SDBIN3_0086)
        {
            SpdString   tag_name;
            
            tag_name = spdGetName(objectId);
            obj_id = spdGetARelObj(objectId, "type_is", "p-id");
            if (!SPD_STREQ(tag_name, "")) { /* structդ */
                res = spdStrCat(res, " ");
                res = spdStrCat(res, tag_name);
            }
        }
#else
        if ((obj_id = spdGetARelObj(objectId, "type_is", "p-id")) !=
            SPD_NON_ID) {                 /* structդ */
            res = spdStrCat(res, " ");
            res = spdStrCat(res, spdGetName(obj_id));
        }
#endif /* BUGGY_SDBIN3_0086 */

        break;
      case TYPE_ENUM_SPECIFIER:
        res = spdStrCat(res, "enum");
        res = spdStrCat(res, " ");
        res = spdStrCat(res, spdGetName(objectId));
        break;
      case TYPE_FUNC:
        type_id = spdGetARelObj(objectId, "type_st", "p-id");
        res = spdStrCat(res, spdStrTypeSpecifier(type_id));
        break;
      case TYPE_ARRAY:
      case TYPE_POINTER:
        csr = spdGetRelInit(objectId, "type_st", "p-id");
        /* طʤʤ뤫ʬǤϤʤʤȤ */
        while (((obj_id = spdGetRelObj(csr)) != -1) &&
               (obj_id == objectId));
        spdFreeCursor(csr);

        res = spdStrCat(res, spdStrTypeSpecifier(obj_id));
        break;
      case TYPE_DECLARATION:
        res = spdStrCat(res, spdStrTypeDeclaration(objectId));
        break;
      default:
        break;
    }

    DW_FUNC_OUT("type_spe", res);

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }

    save = res;
    return res;
}

#ifdef BUGGY_SDBIN4_0110
SpdString spdStrTypeQualifier(SpdObjId objectId, SpdString identName,
                              int declSort, SpdBoolean isOperator)
#else /* BUGGY_SDBIN4_0110 */
SpdString spdStrTypeQualifier(SpdObjId objectId, SpdString identName,
                              int declSort)
#endif /* BUGGY_SDBIN4_0110 */
{
    static SpdString    save = NULL;
    SpdString   res = NULL, tmp = NULL;

    int     Sort;
    SpdObjId    objectId2, expr_id, type_id;

    DW_FUNC_IN("type_qua", save);

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }

    objectId2 = objectId;
    Sort = spdGetAttrValInt(objectId2, "sort");
    switch (Sort) {
      case TYPE_FUNC:
        if (declSort == DECL_VARIABLE || declSort == DECL_PARMTYPE
            || declSort == DECL_MEMBER || declSort == DECL_TYPE) {
            tmp = spdStrCat(tmp, "(");
        }
        tmp = spdStrCat(tmp, identName);
        if (declSort == DECL_VARIABLE || declSort == DECL_PARMTYPE
            || declSort == DECL_MEMBER || declSort == DECL_TYPE) {
            tmp = spdStrCat(tmp, ")");
        }
        if ((type_id = spdGetARelObj(objectId, "type_st", "p-id")) !=
            SPD_NON_ID)
#ifdef BUGGY_SDBIN4_0110
            res = spdStrCat(res, spdStrTypeQualifier(type_id, tmp,
                                                     declSort, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
        res = spdStrCat(res, spdStrTypeQualifier(type_id, tmp,
                                                 declSort));
#endif /* BUGGY_SDBIN4_0110 */
#ifdef DEBUG_TYPE_960830
        fprintf(stderr, "%d: %s\n", objectId, res);
#endif
        break;
      case TYPE_ARRAY:
        tmp = spdStrCat(tmp, identName);
        tmp = spdStrCat(tmp, "[");
        expr_id = spdGetARelObj(objectId, "type_cexpr", "p-id");
        tmp = spdStrCat(tmp, spdStrExpression(expr_id));
        tmp = spdStrCat(tmp, "]");
        if ((type_id = spdGetARelObj(objectId, "type_st", "p-id"))
            != SPD_NON_ID)
#ifdef BUGGY_SDBIN4_0110 
            res = spdStrCat(res, spdStrTypeQualifier(type_id, tmp,
                                                     declSort, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
        res = spdStrCat(res, spdStrTypeQualifier(type_id, tmp,
                                                 declSort));
#endif /* BUGGY_SDBIN4_0110 */
        break;
      case TYPE_POINTER:
        tmp = spdStrCat(tmp, "*");

        /* type_qualifier  ON äϤ */
        if (spdGetOption("typeQualifier") == ON) {
            switch (spdGetAttrValInt(objectId, "qualifier")) {
            case TYPE_UNKNOWN: break;
            case TYPE_CONST: tmp = spdStrCat(tmp, " const "); break;
            case TYPE_VOLATILE: tmp = spdStrCat(tmp, " volatile "); break;
            default: break;
            }
        }

        tmp = spdStrCat(tmp, identName);
        if ((type_id = spdGetARelObj(objectId, "type_st", "p-id"))
            != SPD_NON_ID) {
            SpdString tmp_2 = NULL;
            if (spdGetAttrValInt(type_id, "sort") == TYPE_ARRAY) {
                tmp_2 = spdStrCat(tmp_2, "(");
                tmp_2 = spdStrCat(tmp_2, tmp);
                tmp_2 = spdStrCat(tmp_2, ")");
#ifdef BUGGY_SDBIN4_0110
            } else if (isOperator && 
                       spdGetAttrValInt(type_id, "sort") == TYPE_FUNC) {
                tmp_2 = spdStrCat(tmp_2, "(");
                tmp_2 = spdStrCat(tmp_2, tmp);
                tmp_2 = spdStrCat(tmp_2, ")()");
#endif /* BUGGY_SDBIN4_0110 */
            } else {
                tmp_2 = spdStrCat(tmp_2, tmp);
            }
#ifdef BUGGY_SDBIN4_0110
            res = spdStrCat(res, spdStrTypeQualifier(type_id, tmp_2,
                                                     declSort, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
            res = spdStrCat(res, spdStrTypeQualifier(type_id, tmp_2,
                                                     declSort));
#endif /* BUGGY_SDBIN4_0110 */
        
            if (tmp_2 != NULL) spdFree(tmp_2);
        }
        break;
      default:
        res = spdStrCat(res, identName);
        break;
    }

    DW_FUNC_OUT("type_qua", res);

    if (tmp != NULL){ 
        spdFree(tmp);
        tmp = NULL;
    }

    if (save != NULL){ 
        spdFree(save);
        save = NULL;
    }

    save = res;
    return res;
}
