/* 
 * Program:             $RCSfile: spdStr3_block.c,v $  $Revision: 31.8 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto and S.Yamamoto  1998/06/28
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1998 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr3_block.c,v 31.8 2003/07/08 06:19:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr3.h>
#include "SpdStr3Sub.h"

int spdStr3BeforeBlockDecl(SpdObjId id, SpdRel rel, void *p)
{
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;

    spdStr3Indent(info);

    return SAPID_TRV_GO;
}

int spdStr3BeforeBlockBlock(SpdObjId id, SpdRel rel, void *p)
{
    int p_sort, c_sort;
    char **res;
    SpdStr3Info *info;
    
    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    p_sort = spdGetAttrValInt(id, "sort");
    c_sort = spdGetAttrValInt(rel.objectId, "sort");

    switch(p_sort){
      case BLOCK_FOR:
        *res = spdStrCat(*res, ") ");
        break;
    }

    /* else Ĥ˲Ԥ뤫ɤ 
       ̤ä饹ڡ else Ϥ롥 
       ̤ʤȤ ԡǥȤ else Ϥ롥 */
    if (info->else_flag) {
        if (info->basic_flag)
            spdStr3Indent(info);
        else 
            *res = spdStrCat(*res, " ");
        *res = spdStrCat(*res, "else ");
    }
        
    if (c_sort != BLOCK_COMP && p_sort != BLOCK_COMP) {
        if (info->else_flag && c_sort != BLOCK_BASIC) {
            spdQueuePushBack(&(info->indent_stack), rel.objectId);
        } else {
            info->indent ++;
        }
    }

    /* 
     * Modified for suppress message "suggest parentheses around && within ||"
     * from gcc by S.Yamamoto 2000/03/19.
     * 
     * if (!info->else_flag && c_sort != BLOCK_COMP && c_sort != BLOCK_BASIC || 
     *     c_sort == BLOCK_COMP && p_sort == BLOCK_COMP) {
     *     spdStr3Indent(info);
     * }
     */
    if ((!info->else_flag && c_sort != BLOCK_COMP && c_sort != BLOCK_BASIC) || 
        (c_sort == BLOCK_COMP && p_sort == BLOCK_COMP)) {
        spdStr3Indent(info);
    }

    info->else_flag = FALSE;
    info->basic_flag = FALSE;

    return SAPID_TRV_GO;
}

int spdStr3AfterBlockBlock(SpdObjId id, SpdRel rel, void *p)
{
    int p_sort, c_sort;
    char **res;
    SpdStr3Info *info;
    SpdObjId else_block_id;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    p_sort = spdGetAttrValInt(id, "sort");
    c_sort = spdGetAttrValInt(rel.objectId, "sort");

    if (c_sort != BLOCK_COMP && p_sort != BLOCK_COMP) {
        else_block_id = spdQueuePopBack(&(info->indent_stack));
        if (else_block_id != rel.objectId) {
            info->indent --;
            spdQueuePushBack(&(info->indent_stack), else_block_id);
        }
    }

    switch(p_sort){
      case BLOCK_DO:
        if (c_sort != BLOCK_COMP)
            spdStr3Indent(info);
        else
            *res = spdStrCat(*res, " ");
        *res = spdStrCat(*res, "while (");
        break;
    }

    return SAPID_TRV_GO;
}

int spdStr3BeforeBlockLabel(SpdObjId id, SpdRel rel, void *p)
{
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;

    info->indent --;
    spdStr3Indent(info);
    info->indent ++;

    return SAPID_TRV_GO;
}    

int spdStr3AfterBlockLabel(SpdObjId id, SpdRel rel, void *p)
{
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    *res = spdStrCat(*res, ":");

    return SAPID_TRV_GO;
}

int spdStr3BeforeBlockExpr(SpdObjId id, SpdRel rel, void *p)
{
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;

    if (spdGetAttrValInt(id, "sort") == BLOCK_BASIC)
        spdStr3Indent(info);

    return SAPID_TRV_GO;
}    

int spdStr3AfterBlockExpr(SpdObjId id, SpdRel rel, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    sort = spdGetAttrValInt(id, "sort");
    switch(sort){
      case BLOCK_BASIC:
        *res = spdStrCat(*res, ";");
        break;
      case BLOCK_IF:
      case BLOCK_SWITCH:
      case BLOCK_WHILE:
        *res = spdStrCat(*res, ") ");
        break;
      case BLOCK_DO:
        *res = spdStrCat(*res, ");");
        break;
    }

    return SAPID_TRV_GO;
}

int spdStr3BeforeBlock(SpdObjId id, void *p)
{
    int sort;
    char **res;
    char *str = NULL;
    SpdStr3Info *info;
    SpdObjId pblock_id;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    /* spdStr3 -c 1998/10/13 */
    if (spdGetOption("mode") == SAPID_C_MODE) {
        if ((pblock_id = isMacroCallBlock(id)) != SPD_NON_ID) {
            str = spdStr3(pblock_id);
            if (str != NULL) {
                *res = spdStrCat(*res, str);
                spdFree(str);
                str = NULL;
            }
            return SAPID_TRV_STOP;
        }
    }

    sort = spdGetAttrValInt(id, "sort");
    
    switch(sort){
      case BLOCK_COMP:
        *res = spdStrCat(*res, "{");
        info->indent ++;
        break;
      case BLOCK_IF:
        *res = spdStrCat(*res, "if (");
        break;
      case BLOCK_SWITCH:
        *res = spdStrCat(*res, "switch (");
        break;
      case BLOCK_FOR:
        *res = spdStrCat(*res, "for (");
        break;
      case BLOCK_WHILE:
        *res = spdStrCat(*res, "while (");
        break;
      case BLOCK_DO:
        *res = spdStrCat(*res, "do ");
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3MiddleBlock(SpdObjId id, int rel_count, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;
    SpdRelArray block_array;

    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    sort = spdGetAttrValInt(id, "sort");
    
    switch(sort){
      case BLOCK_IF:
        /* !! 1ܤϽϤ2ܤ else (;_;) */
        if (rel_count > 0) {
            /* *res = spdStrCat(*res, " else "); */
            info->else_flag = ON;
            /* else  block  BASIC ä(̤ʤ)
               indent ĤΤ flag  ON ˤ롥 */
            block_array = spdGetRelSortedArray(id, "blk_blk", "p-id", 
                                               NULL, spdCompareRelOffset);
            if (spdGetAttrValInt(block_array.rel[rel_count-1].objectId, "sort")
                == BLOCK_BASIC)
                info->basic_flag = ON;
            spdFreeRelArray(block_array);
        }
        break;
      case BLOCK_FOR: 
        /* !! 1, 2ܤ ";" 3ܤϽϤʤ */
        if (rel_count < 2) {
            *res = spdStrCat(*res, "; ");
        }
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3AfterBlock(SpdObjId id, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    sort = spdGetAttrValInt(id, "sort");
    
    switch(sort){
      case BLOCK_COMP:
        info->indent--;
        spdStr3Indent(info);
        *res = spdStrCat(*res, "}");
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3BeforeLabel(SpdObjId id, void *p)
{
    int sort;
    char **res;
    char *str = NULL;
    SpdStr3Info *info;
    SpdObjId pblock_id;

    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    /* spdStr3 -c 1998/10/13 */
    if (spdGetOption("mode") == SAPID_C_MODE) {
        if ((pblock_id = isMacroCallBlock(id)) != SPD_NON_ID) {       
            str = spdStr3(pblock_id);
            if (str != NULL) {
                *res = spdStrCat(*res, str);
                spdFree(str);
                str = NULL;
            }
            return SAPID_TRV_STOP;
        }
    }

    sort = spdGetAttrValInt(id, "sort");
    
    switch(sort){
      case LABEL_CASE:
        *res = spdStrCat(*res, "case ");
        break;
      case LABEL_DEFAULT:
        *res = spdStrCat(*res, "default");
        break;
      case LABEL_NAMED:
        *res = spdStrCat(*res, spdGetName(id));
        break;
      default:
        break;
    }
    return SAPID_TRV_GO;
}    
