/* 
 * Program:             $RCSfile: spdStr3_expr.c,v $  $Revision: 31.5 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto and S.Yamamoto  1998/06/28
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr3_expr.c,v 31.5 2003/07/08 06:19:42 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr3.h>
#include "SpdStr3Sub.h"

int spdStr3BeforeExprExpr(SpdObjId id, SpdRel rel, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    sort = spdGetAttrValInt(id, "sort");
    switch(sort){
      case EXPR_RETURN:
        *res = spdStrCat(*res, " ");
        break;
    }

    return SAPID_TRV_GO;
}

int spdStr3BeforeExprOp(SpdObjId id, SpdRel rel, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    sort = spdGetAttrValInt(rel.objectId, "sort");
    switch(sort){
      case OP_MUL:
      case OP_DIV:
      case OP_MOD:
      case OP_ADD:
      case OP_SUB:
      case OP_SHL:
      case OP_SHR:
      case OP_LT:
      case OP_LE:
      case OP_GT:
      case OP_GE:
      case OP_EQ:
      case OP_NE:
      case OP_B_AND:
      case OP_B_XOR:
      case OP_B_OR:
      case OP_L_AND:
      case OP_L_OR:
      case OP_QUESTION:
      case OP_COLON:
      case OP_ASSIGN:
      case OP_A_ADD:
      case OP_A_SUB:
      case OP_A_MUL:
      case OP_A_DIV:
      case OP_A_MOD:
      case OP_A_AND:
      case OP_A_XOR:
      case OP_A_OR:
      case OP_A_SHL:
      case OP_A_SHR:
        *res = spdStrCat(*res, " ");
        break;
      default:
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3AfterExprOp(SpdObjId id, SpdRel rel, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    sort = spdGetAttrValInt(rel.objectId, "sort");
    switch(sort){
      case OP_PLUS:
      case OP_MINUS:
      case OP_MUL:
      case OP_DIV:
      case OP_MOD:
      case OP_ADD:
      case OP_SUB:
      case OP_SHL:
      case OP_SHR:
      case OP_LT:
      case OP_LE:
      case OP_GT:
      case OP_GE:
      case OP_EQ:
      case OP_NE:
      case OP_B_AND:
      case OP_B_XOR:
      case OP_B_OR:
      case OP_L_AND:
      case OP_L_OR:
      case OP_QUESTION:
      case OP_COLON:
      case OP_ASSIGN:
      case OP_A_ADD:
      case OP_A_SUB:
      case OP_A_MUL:
      case OP_A_DIV:
      case OP_A_MOD:
      case OP_A_AND:
      case OP_A_XOR:
      case OP_A_OR:
      case OP_A_SHL:
      case OP_A_SHR:
      case OP_COMMA:
      case OP_SIZEOF:
        *res = spdStrCat(*res, " ");
        break;
      default:
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3BeforeExpression(SpdObjId id, void *p)
{
    int sort;
    char **res;
    char *str = NULL;
    SpdStr3Info *info;
    SpdObjId pblock_id;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    /* spdStr3 -c 1998/10/13 */
    if (spdGetOption("mode") == SAPID_C_MODE) {
        if ((pblock_id = isMacroCallBlock(id)) != SPD_NON_ID) {
            str = spdStr3(pblock_id);
            if (str != NULL) {
                *res = spdStrCat(*res, str);
                spdFree(str);
            }
            return SAPID_TRV_STOP;
        }
    }

    sort = smuGetExprSort(id);
    switch(sort){
      case EXPR_EMPTY:
        break;
      case EXPR_BREAK:
        *res = spdStrCat(*res, "break");
        break;
      case EXPR_CONTINUE:
        *res = spdStrCat(*res, "continue");
        break;
      case EXPR_RETURN:
        *res = spdStrCat(*res, "return");
        break;
      case EXPR_GOTO:
        *res = spdStrCat(*res, "goto ");
        break;
      case EXPR_SORT_INITLIST:              
        *res = spdStrCat(*res, "{ ");
        break;
      default:
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3MiddleExpression(SpdObjId id, int rel_count, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    sort = smuGetExprSort(id);
    switch(sort){
      case EXPR_SORT_FUNCCALL:
        /* 1ܤ "(" ǡʳ "," */
        if (rel_count == 0) {
            /* ʤ "(" Фޤ
               ΤȤ spdStr3AfterExpression() ǽФ褦ˤƤޤ */
            *res = spdStrCat(*res, "(");
        } else {
            *res = spdStrCat(*res, ", ");
        }
        break;
      case EXPR_SORT_INITLIST:
        *res = spdStrCat(*res, ", ");
        break;
    }    
    return SAPID_TRV_GO;
}

int spdStr3AfterExpression(SpdObjId id, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;
    SpdObjIdArray expr_array;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    sort = smuGetExprSort(id);
    switch(sort) {
      case EXPR_SORT_FUNCCALL:
        /*
         * ؿƤӽФΰ̵"("Ϥ롥
         * !! ʤΤǤȤǹͤ뤳ȡ
         */
        expr_array = spdGetRelObjArray(id, "expr_expr", "outer_expression_id", 
                                       NULL);
        if (expr_array.size == 1) {
            *res = spdStrCat(*res, "(");
        }
        *res = spdStrCat(*res, ")");
        spdFreeObjIdArray(expr_array);
        break;
      case EXPR_SORT_INITLIST:
        *res = spdStrCat(*res, " }");
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3BeforeOperator(SpdObjId id, void *p)
{
    int sort;
    char **res;
    /* char *str = NULL; */
    SpdStr3Info *info;
    /* SpdObjId pblock_id; */

    info = (SpdStr3Info *)p;
    res = &(info->str);
    
    /* spdStr3 -c 1998/10/13 */
    /* operator ̾󤻤ƤΤǤˡǤϤᡥ*/
    /* expr_op δϢ offset ʤФʤ */
    /* 
     * if (spdGetOption("mode") == SAPID_C_MODE) {
     *     if ((pblock_id = isMacroCallBlock(id)) != SPD_NON_ID) {
     *         str = spdStr3(pblock_id);
     *         if (str != NULL) {
     *             *res = spdStrCat(*res, str);
     *             spdFree(str);
     *             str = NULL;
     *         }
     *         return SAPID_TRV_STOP;
     *     }
     * }
     */

    sort = spdGetAttrValInt(id, "sort");
    switch(sort){
      case OP_L_PAR:   *res = spdStrCat(*res, "("); break;
      case OP_R_PAR:   *res = spdStrCat(*res, ")"); break;
      case OP_L_ARRAY: *res = spdStrCat(*res, "["); break;
      case OP_R_ARRAY: *res = spdStrCat(*res, "]"); break;
      case OP_ARROW:   *res = spdStrCat(*res, "->"); break;
      case OP_DOT:     *res = spdStrCat(*res, "."); break;
      case OP_NOT:     *res = spdStrCat(*res, "!"); break;
      case OP_B_NOT:   *res = spdStrCat(*res, "~"); break;
      case OP_INC:     *res = spdStrCat(*res, "++"); break;
      case OP_DEC:     *res = spdStrCat(*res, "--"); break;
      case OP_PLUS:    *res = spdStrCat(*res, "+"); break;
      case OP_MINUS:   *res = spdStrCat(*res, "-"); break;
      case OP_POINTER: *res = spdStrCat(*res, "*"); break;
      case OP_ADDRESS: *res = spdStrCat(*res, "&"); break;
      case OP_MUL:     *res = spdStrCat(*res, "*"); break;
      case OP_DIV:     *res = spdStrCat(*res, "/"); break;
      case OP_MOD:     *res = spdStrCat(*res, "%"); break;
      case OP_ADD:     *res = spdStrCat(*res, "+"); break;
      case OP_SUB:     *res = spdStrCat(*res, "-"); break;
      case OP_SHL:     *res = spdStrCat(*res, "<<"); break;
      case OP_SHR:     *res = spdStrCat(*res, ">>"); break;
      case OP_LT:      *res = spdStrCat(*res, "<"); break;
      case OP_LE:      *res = spdStrCat(*res, "<="); break;
      case OP_GT:      *res = spdStrCat(*res, ">"); break;
      case OP_GE:      *res = spdStrCat(*res, ">="); break;
      case OP_EQ:      *res = spdStrCat(*res, "=="); break;
      case OP_NE:      *res = spdStrCat(*res, "!="); break;
      case OP_B_AND:   *res = spdStrCat(*res, "&"); break;
      case OP_B_XOR:   *res = spdStrCat(*res, "^"); break;
      case OP_B_OR:    *res = spdStrCat(*res, "|"); break;
      case OP_L_AND:   *res = spdStrCat(*res, "&&"); break;
      case OP_L_OR:    *res = spdStrCat(*res, "||"); break;
      case OP_QUESTION:*res = spdStrCat(*res, "?"); break;
      case OP_COLON:   *res = spdStrCat(*res, ":"); break;
      case OP_ASSIGN:  *res = spdStrCat(*res, "="); break;
      case OP_A_ADD:   *res = spdStrCat(*res, "+="); break;
      case OP_A_SUB:   *res = spdStrCat(*res, "-="); break;
      case OP_A_MUL:   *res = spdStrCat(*res, "*="); break;
      case OP_A_DIV:   *res = spdStrCat(*res, "/="); break;
      case OP_A_MOD:   *res = spdStrCat(*res, "%="); break;
      case OP_A_AND:   *res = spdStrCat(*res, "&="); break;
      case OP_A_XOR:   *res = spdStrCat(*res, "^="); break;
      case OP_A_OR:    *res = spdStrCat(*res, "|="); break;
      case OP_A_SHL:   *res = spdStrCat(*res, "<<="); break;
      case OP_A_SHR:   *res = spdStrCat(*res, ">>="); break;
      case OP_COMMA:   *res = spdStrCat(*res, ","); break;
      case OP_SIZEOF:  *res = spdStrCat(*res, "sizeof "); break;
      case OP_SIZEOF_TYPE:
        *res = spdStrCat(*res, "sizeof("); 
        break;
      case OP_CAST:
        *res = spdStrCat(*res, "("); 
        break;
      default:
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3AfterOperator(SpdObjId id, void *p)
{
    int sort;
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    sort = spdGetAttrValInt(id, "sort");
    switch(sort){
      case OP_SIZEOF_TYPE:
      case OP_CAST:
        *res = spdStrCat(*res, ")"); 
        break;
    }
    return SAPID_TRV_GO;
}

int spdStr3BeforeConstant(SpdObjId id, void *p)
{
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    *res = spdStrCat(*res, spdGetAttrValString(id, "value"));
    return SAPID_TRV_GO;
}

int spdStr3BeforeIdentifier(SpdObjId id, void *p)
{
    char **res;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    *res = spdStrCat(*res, spdGetName(id));
    return SAPID_TRV_GO;
}
