/* 
 * Program:             $RCSfile: spdStr3_type.c,v $  $Revision: 31.11 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto and S.Yamamoto  1998/06/28
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStr3_type.c,v 31.11 2003/07/08 06:19:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr3.h>
#include "SpdStr3Sub.h"

static SpdString spdStr3TypeDeclaration(SpdObjId objectId);
static SpdString spdStr3TypeSpecifier (SpdObjId objectId);

int spdStr3AfterType(SpdObjId id, void *p)
{
    char **res, *str;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    str = spdStr3TypeSpecifier(id);
    *res = spdStrCat(*res, str);
 
#ifdef BUGGY_SDBIN4_0110    
    str = spdStr3TypeQualifier(id, NULL, 0, FALSE);
#else  /* BUGGY_SDBIN4_0110 */
    str = spdStr3TypeQualifier(id, NULL, 0);
#endif /* BUGGY_SDBIN4_0110 */

    if (str != NULL) {
        *res = spdStrCat(*res, " ");
        *res = spdStrCat(*res, str);
    }
    return SAPID_TRV_GO;
}

int spdStr3BeforeOpType(SpdObjId id, SpdRel rel, void *p)
{
    char **res, *str;
    SpdStr3Info *info;

    info = (SpdStr3Info *)p;
    res = &(info->str);

    str = spdStr3TypeSpecifier(rel.objectId);
    *res = spdStrCat(*res, str);
 
#ifdef BUGGY_SDBIN4_0110    
    str = spdStr3TypeQualifier(rel.objectId, NULL, 0, TRUE);
#else  /* BUGGY_SDBIN4_0110 */
    str = spdStr3TypeQualifier(rel.objectId, NULL, 0);
#endif /* BUGGY_SDBIN4_0110 */
    if (str != NULL) {
        *res = spdStrCat(*res, " ");
        *res = spdStrCat(*res, str);
    }
    return SAPID_TRV_SKIP;
}

static SpdString spdStr3TypeDeclaration(SpdObjId objectId)
{
    static SpdString    save = NULL;
    SpdString           res = NULL;
    int                 size, type;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }

    switch (spdGetAttrValInt(objectId, "sign")) {
      case TYPE_UNKNOWN: break;
      case TYPE_SIGNED: res = spdStrCat(res, "signed "); break;
      case TYPE_UNSIGNED: res = spdStrCat(res, "unsigned "); break;
      default: break;
    }

    switch (size = spdGetAttrValInt(objectId, "size")) {
      case TYPE_UNKNOWN: break;
      case TYPE_SHORT: res = spdStrCat(res, "short"); break;
      case TYPE_LONG: res = spdStrCat(res, "long"); break;

#ifdef ACCEPT_TYPE_LONG_LONG
      case TYPE_LONG_LONG: res = spdStrCat(res, "long long"); break;
#endif

      default: break;
    }

    type = spdGetAttrValInt(objectId, "type");
    if (type != TYPE_UNKNOWN && size != TYPE_UNKNOWN) {
        res = spdStrCat(res, " ");
    }

    switch (type) {
      case TYPE_UNKNOWN: break;
      case TYPE_VOID: res = spdStrCat(res, "void"); break;
      case TYPE_INT: res = spdStrCat(res, "int"); break;
      case TYPE_CHAR: res = spdStrCat(res, "char"); break;
      case TYPE_FLOAT: res = spdStrCat(res, "float"); break;
      case TYPE_DOUBLE: res = spdStrCat(res, "double"); break;
      case TYPE_DOTS: res = spdStrCat(res, "..."); break;
      default: break;
    }

    save = res;
    return res;
}

static SpdString spdStr3TypeSpecifier (SpdObjId objectId)
{
    static SpdString    save = NULL;

    SpdString           res = NULL;
    int                 type_sort;
    SpdCursor           *csr;
    SpdObjId            obj_id, type_id;

#ifndef BUGGY_SDBIN3_0086
    SpdObjId    ident_id;
#endif

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }

    /* Added by N.Iwamoto 1998/07/06 */
    /* type_qualifier  ON äϤ */
    if (spdGetOption("typeQualifier") == ON) {
        switch (spdGetAttrValInt(objectId, "qualifier")) {
          case TYPE_UNKNOWN: break;
          case TYPE_CONST: res = spdStrCat(res, "const "); break;
          case TYPE_VOLATILE: res = spdStrCat(res, "volatile "); break;
          default: break;
        }
    }

    switch (type_sort = spdGetAttrValInt(objectId, "sort")) {
      case TYPE_TYPE_DEF:
      
#if defined(BUGGY_SDBIN3_0086)
        res = spdStrCat(res, spdGetName(objectId));
#else
        ident_id = spdGetARelObj(objectId, "type_is", "p-id");
        res = spdStrCat(res, spdStr3Identifier(ident_id));
#endif

        break;
      case TYPE_STRUCT_TAG:
      case TYPE_UNION_TAG:
      case TYPE_ENUM_SPECIFIER:
        if (type_sort == TYPE_STRUCT_TAG) {
            res = spdStrCat(res, "struct");
        } else if (type_sort == TYPE_UNION_TAG) {
            res = spdStrCat(res, "union");
        } else if (type_sort == TYPE_ENUM_SPECIFIER) {
            res = spdStrCat(res, "enum");
        }

        if ((obj_id = spdGetARelObj(objectId, "type_is", "p-id")) !=
            SPD_NON_ID) {                 /* structդ */
            res = spdStrCat(res, " ");
            res = spdStrCat(res, spdGetName(obj_id));
        }

        break;
      case TYPE_FUNC:
        type_id = spdGetARelObj(objectId, "type_st", "p-id");
        res = spdStrCat(res, spdStr3TypeSpecifier(type_id));
        break;
      case TYPE_ARRAY:
      case TYPE_POINTER:
        csr = spdGetRelInit(objectId, "type_st", "p-id");
        /* طʤʤ뤫ʬǤϤʤʤȤ */
        while (((obj_id = spdGetRelObj(csr)) != -1) &&
               (obj_id == objectId));
        spdFreeCursor(csr);

        res = spdStrCat(res, spdStr3TypeSpecifier(obj_id));
        break;
      case TYPE_DECLARATION:
        res = spdStrCat(res, spdStr3TypeDeclaration(objectId));
        break;
      default:
        break;
    }

    if (save != NULL){
        spdFree(save);
        save = NULL;
    }

    save = res;
    return res;
}

#ifdef BUGGY_SDBIN4_0110
SpdString spdStr3TypeQualifier(SpdObjId objectId, SpdString identName,
                               int declSort, SpdBoolean isOperator)
#else /* BUGGY_SDBIN4_0110 */
SpdString spdStr3TypeQualifier(SpdObjId objectId, SpdString identName,
                               int declSort)
#endif /* BUGGY_SDBIN4_0110 */
{
    static SpdString    save = NULL;
    SpdString   res = NULL, tmp = NULL, tmp2 = NULL;

    int     Sort;
    SpdObjId    objectId2, expr_id, type_id;

    if (save != NULL) {
        spdFree(save);
        save = NULL;
    }

    /* objectId Υ饹 type ǤʤСNULL ֤ */
    if ((spdGetClassId(objectId)) != SPD_CID_TYPE) {
        return res;
    }

    objectId2 = objectId;
    Sort = spdGetAttrValInt(objectId2, "sort");
    switch (Sort) {
      case TYPE_FUNC:
        if (declSort == DECL_VARIABLE || declSort == DECL_PARMTYPE || 
            declSort == DECL_MEMBER || declSort == DECL_TYPE) {
            tmp = spdStrCat(tmp, "(");
        }

        tmp = spdStrCat(tmp, identName);

        if (declSort == DECL_VARIABLE || declSort == DECL_PARMTYPE || 
            declSort == DECL_MEMBER || declSort == DECL_TYPE) {
            tmp = spdStrCat(tmp, ")");
        }

        if ((type_id = spdGetARelObj(objectId, "type_st", "p-id")) !=
            SPD_NON_ID)
#ifdef BUGGY_SDBIN4_0110
            res = spdStrCat(res, spdStr3TypeQualifier(type_id, tmp,
                                                      declSort, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
        res = spdStrCat(res, spdStr3TypeQualifier(type_id, tmp,
                                                  declSort));
#endif /* BUGGY_SDBIN4_0110 */
        break;
      case TYPE_ARRAY:
        tmp = spdStrCat(tmp, identName);
        tmp = spdStrCat(tmp, "[");
        expr_id = spdGetARelObj(objectId, "type_cexpr", "p-id");
        tmp2 = spdStr3(expr_id);
        tmp = spdStrCat(tmp, tmp2);
        if (tmp2 != NULL) 
            spdFree(tmp2);
        tmp = spdStrCat(tmp, "]");
        if ((type_id = spdGetARelObj(objectId, "type_st", "p-id"))
            != SPD_NON_ID)
#ifdef BUGGY_SDBIN4_0110 
            res = spdStrCat(res, spdStr3TypeQualifier(type_id, tmp,
                                                      declSort, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
        res = spdStrCat(res, spdStr3TypeQualifier(type_id, tmp,
                                                  declSort));
#endif /* BUGGY_SDBIN4_0110 */
        break;
      case TYPE_POINTER:
        tmp = spdStrCat(tmp, "*");

        /* type_qualifier  ON äϤ */
        if (spdGetOption("typeQualifier") == ON) {
            switch (spdGetAttrValInt(objectId, "qualifier")) {
              case TYPE_UNKNOWN: break;
              case TYPE_CONST: tmp = spdStrCat(tmp, " const "); break;
              case TYPE_VOLATILE: tmp = spdStrCat(tmp, " volatile "); break;
              default: break;
            }
        }

        tmp = spdStrCat(tmp, identName);
        if ((type_id = spdGetARelObj(objectId, "type_st", "p-id"))
            != SPD_NON_ID) {
            SpdString tmp_2 = NULL;
            if (spdGetAttrValInt(type_id, "sort") == TYPE_ARRAY) {
                tmp_2 = spdStrCat(tmp_2, "(");
                tmp_2 = spdStrCat(tmp_2, tmp);
                tmp_2 = spdStrCat(tmp_2, ")");
#ifdef BUGGY_SDBIN4_0110
            } else if (isOperator && 
                       spdGetAttrValInt(type_id, "sort") == TYPE_FUNC) {
                tmp_2 = spdStrCat(tmp_2, "(");
                tmp_2 = spdStrCat(tmp_2, tmp);
                tmp_2 = spdStrCat(tmp_2, ")()");
#endif /* BUGGY_SDBIN4_0110 */
            } else {
                tmp_2 = spdStrCat(tmp_2, tmp);
            }
#ifdef BUGGY_SDBIN4_0110
            res = spdStrCat(res, spdStr3TypeQualifier(type_id, tmp_2,
                                                      declSort, FALSE));
#else  /* BUGGY_SDBIN4_0110 */
            res = spdStrCat(res, spdStr3TypeQualifier(type_id, tmp_2,
                                                      declSort));
#endif /* BUGGY_SDBIN4_0110 */
        
            if (tmp_2 != NULL) spdFree(tmp_2);
        }
        break;
      default:
        res = spdStrCat(res, identName);
        break;
    }

    if (tmp != NULL){ 
        spdFree(tmp);
        tmp = NULL;
    }

    if (save != NULL){ 
        spdFree(save);
        save = NULL;
    }

    save = res;
    return res;
}
