/* 
 * Program:             $RCSfile: alloc.c,v $  $Revision: 30.8 $
 * 
 * Purpose:             Memory management functions.
 * 
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01
 * 
 * (C) Copyright:       T.Mimura and S.Yamamoto  1993 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: alloc.c,v 30.8 2003/07/06 10:08:58 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* Memory trace. */
#define DW(f, l, ptr, size, msg) \
        spdMessage("MT: %8x %6x: %s: \"%s\" %d.\n", \
                   (ptr), (size), (msg), (f), (l))

/*
 * 顼դmalloc().
 *
 * Note, in some environment, malloc() and calloc() return a non-NULL
 * pointer if size is 0.
 */

void *spdMallocSub(size_t size)
{
    void    *ptr;
    
    if (size == 0) {
        return (NULL);
    } else if ((ptr = (void *)malloc(size)) == NULL) {
        spdErrorExit("spdMalloc(): Can't allocate memory.\n");
    }
    return (ptr);
}

void *spdMallocDbg(char *file, int line, size_t size)
{
    void    *tmp;

    tmp = spdMallocSub(size);
    DW(file, line, tmp, size, "malloc");
    return (tmp);
}

/*
 * 顼դcalloc().
 */

void *spdCallocSub(size_t nelem, size_t elsize)
{
    void    *ptr;
    
    if (nelem == 0 || elsize == 0) {
        return (NULL);
    } else if ((ptr = (void *)calloc(nelem, elsize)) == NULL) {
        spdErrorExit("spdCalloc(): Can't allocate memory.\n");
    }
    return (ptr);
}

void *spdCallocDbg(char *file, int line, size_t nelem, size_t elsize)
{
    void    *tmp;

    tmp = spdCallocSub(nelem, elsize);
    DW(file, line, tmp, nelem * elsize, "calloc");
    return (tmp);
}

/*
 * 顼դrealloc().
 */

void *spdReallocSub(void *ptr, size_t size)
{
    if (ptr == NULL) {
#ifdef SPD_DEBUG_ALLOC
        return (spdMallocSub(size));
#else
        return (spdMalloc(size));
#endif
    } else if (size == 0) {
        spdFree(ptr);
        return (NULL);
    } else if ((ptr = (void *)realloc(ptr, size)) == NULL) {
        spdErrorExit("spdRealloc(): Can't re-allocate memory.\n");
    }
    return (ptr);
}

void *spdReallocDbg(char *file, int line, void *ptr, size_t size)
{
    void    *tmp, *saved;

    saved = ptr;
    tmp = spdReallocSub(ptr, size);
    if (saved == tmp) {
        DW(file, line, ptr, size, "realloc-add");
    } else {
        DW(file, line, saved, 0, "realloc-free");
        DW(file, line, tmp, size, "realloc-new");
    }
    return (tmp);
}

/*
 * Sapidfree().
 */

void spdFreeDbg(char *file, int line, void *ptr)
{
    DW(file, line, ptr, 0, "free");
    free(ptr);
}

/*
 * 򸡺free().
 */

void *spdSafetyFreeSub(const void *ptr)
{
    if (ptr != NULL) {
#ifdef SPD_DEBUG_ALLOC
        free(ptr);
#else
        spdFree(ptr);
#endif
    }
    return (NULL);
}

void *spdSafetyFreeDbg(char *file, int line, void *ptr)
{
    DW(file, line, ptr, 0, "free");
    return (spdSafetyFreeSub(ptr));
}

/*
 * realloc()wrapperؿ.
 */

void spdIncRegionIntSub(void **ptr, int *nelem, int inc, size_t elsize)
{
    *nelem += inc;
    *ptr = spdRealloc(*ptr, (size_t)(*nelem) * elsize);
}

void spdIncRegionIntDbg(char *file, int line,
                        void **ptr, int *nelem, int inc, size_t elsize)
{
    void    *saved;

    saved = *ptr;
    spdIncRegionIntSub(ptr, nelem, inc, elsize);
    if (saved == *ptr) {
        DW(file, line, *ptr, *nelem * elsize, "incRegion-add");
    } else {
        DW(file, line, saved, 0, "incRegion-free");
        DW(file, line, *ptr, *nelem * elsize, "incRegion-new");
    }
}
