/* 
 * Program:             $RCSfile: message.c,v $  $Revision: 30.17 $
 * 
 * Purpose:             åϴؿ.
 * 
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01
 * 
 * (C) Copyright:       T.Mimura and S.Yamamoto  1993 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: message.c,v 30.17 2003/06/16 15:05:49 yamamoto Exp yamamoto $";
#endif

#include <stdarg.h>
#include <stdio.h>
#include <Sapid/Sapid.h>

#if defined(HAVE_SYSCONF) && defined(_SC_CLK_TCK)
#define SPD_CLK_TCK     (sysconf(_SC_CLK_TCK))
#else
#define SPD_CLK_TCK     CLK_TCK
#endif  

static int          _spd_debug = ON;
static const char   *this_program_name = NULL;

static void spdCleanUp(void);

/* 
 * Time profie.
 */

#if defined(HAVE_SYS_TIME_H) && defined(HAVE_SYS_TIMES_H)

static struct timeval   time1, time2;
static struct tms       times1, times2;

static void spdPrintTimeProf(void)
{
    int     seconds, milliseconds;
    int     diff;                           /* In milliseconds. */
    char    buf[32];

    /* Elapsed time. */
    gettimeofday(&time2, NULL);
    diff = (time2.tv_sec - time1.tv_sec) * 1000 +
        (time2.tv_usec - time1.tv_usec) / 1000;
    seconds = diff / 1000;
    milliseconds = diff % 1000;
    if (3600 < seconds) {
        sprintf(buf, "%d:%02d:%02d.%02d",
                seconds / 3600, (seconds % 3600) / 60, seconds % 60,
                milliseconds / 10);
    } else {
        sprintf(buf, "%d:%02d.%02d",
                seconds / 60, seconds % 60,
                milliseconds / 10);
    }

    /* User and system times. */
    times(&times2);
    spdMessage("Time profile: \"%.2fu %.2fs %s\".\n",
               (double)(times2.tms_utime - times1.tms_utime) / (double)SPD_CLK_TCK,
               (double)(times2.tms_stime - times1.tms_stime) / (double)SPD_CLK_TCK,
               buf);
}

#endif /* defined(HAVE_SYS_TIME_H) && defined(HAVE_SYS_TIMES_H) */

/*
 * General initialization.
 */

void spdInitialize(const char *this_prog)
{
    static SpdBoolean   is_first_call = TRUE;

    /*
     * Note:  
     * The functions will be called in the reverse order of their registration.
     */

    if (is_first_call) {
        is_first_call = FALSE;
        atexit(spdCleanUp);                 /* Called in the last. */
        if (getenv("SAPID_TIME_PROFILE") != NULL) {

#if defined(HAVE_SYS_TIME_H) && defined(HAVE_SYS_TIMES_H)
            gettimeofday(&time1, NULL);
            times(&times1);
            atexit(spdPrintTimeProf);
#endif
        }
        if (getenv("SAPID_MEMORY_PROFILE") != NULL) {
            _spdSafetyInitMemProf();
            atexit(spdPrintMemProf);
        }
    }
    spdCleanUp();                           /* Initialize memory. */

    /* assert(this_program_name == NULL); */
    this_program_name = spdStrdup(this_prog);
}

static void spdCleanUp(void)
{
    if (this_program_name != NULL) {
        spdFree(this_program_name);
        this_program_name = NULL;
    }
}

/*
 * Message.
 */

void spdMessage(const char *format, ...)
{
    va_list     args;
    
    fflush(stdout);
    fflush(stderr);

    fputs("Sapid: ", stderr);
    if (this_program_name != NULL) {
        fprintf(stderr, "%s: ", this_program_name);
    }
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fflush(stderr);
}

/*
 * Warning.
 */

void spdWarning(const char *format, ...)
{
    va_list     args;
    
    fflush(stdout);
    fflush(stderr);

    fputs("Warning: ", stderr);
    if (this_program_name != NULL) {
        fprintf(stderr, "%s: ", this_program_name);
    }
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fflush(stderr);
}

/*
 * Bug.
 */

void spdBugExit(const char *format, ...)
{
    va_list     args;
    
    fflush(stdout);
    fflush(stderr);

    fputs("Bug: ", stderr);
    if (this_program_name != NULL) {
        fprintf(stderr, "%s: ", this_program_name);
    }
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fflush(stderr);
    exit(EXIT_FAILURE);
}

/*
 * Error.
 */

void spdError(const char *format, ...)
{
    va_list     args;
    
    fflush(stdout);
    fflush(stderr);

    fputs("Error: ", stderr);
    if (this_program_name != NULL) {
        fprintf(stderr, "%s: ", this_program_name);
    }
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    
    fflush(stderr);
}

/*
 * Error exit.
 */

#if defined(__GNUC__) && defined(USE_VARARG_MACRO)

/* 
 * Macro spdErrorExit is defined as below:
 * #define spdErrorExit(fmt...)       _spdErrorExit(__FILE__, __LINE__, ## fmt)
 */
#undef spdErrorExit

#endif 

void spdErrorExit(const char *format, ...)
{
    va_list     args;
    
    fflush(stdout);
    fflush(stderr);

    fputs("Error: ", stderr);
    if (this_program_name != NULL) {
        fprintf(stderr, "%s: ", this_program_name);
    }
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    
    fflush(stderr);
    exit(EXIT_FAILURE);
}

void _spdErrorExit(const char *file, int line, const char *format, ...)
{
    va_list     args;

    fflush(stdout);
    fflush(stderr);

    fputs("Error: ", stderr);
    if (this_program_name != NULL) {
        fprintf(stderr, "%s: ", this_program_name);
    }
    if (_spd_debug == ON) {
        fprintf(stderr, "\"%s\", line %d: ", file, line);
    }

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    
    fflush(stderr);
    exit(EXIT_FAILURE);
}

/*
 * Used after spdMessage(), spdWarning(), spdError().
 *
 * For example:
 *     spdMessage("Long long long ...");
 *     spdMessageCont("long long ...");
 *     spdMessageCont("long message.\n");
 */

void spdMessageCont(const char *format, ...)
{
    va_list     args;
    
    fflush(stdout);
    fflush(stderr);

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    fflush(stderr);
}
