/* 
 * Program:             $RCSfile: path.c,v $  $Revision: 30.41 $
 * 
 * Purpose:             Miscllaneous functions.
 * 
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01
 * 
 * (C) Copyright:       T.Mimura and S.Yamamoto  1993 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: path.c,v 30.41 2003/05/29 04:18:51 yamamoto Exp yamamoto $";
#endif

#include <errno.h>
#include <stdio.h>
#include <Sapid/Sapid.h>

/*
 * ǥ쥯ȥ̾ȥե̾Ĥʤ롥
 */

char *spdCatPath(const char *str1, const char *str2)
{
    int     len;
    char    *namep;
    
    len = strlen(str1) + strlen(str2);
    namep = spdMalloc(len + 2);
    sprintf(namep, "%s/%s", str1, str2);
    return (namep);
}

/* WindowsΥѥ̾ɥ饤̾롥 */

char *spdRemoveDriveName(char *orgPath)
{
    char    *newPath;

    /* ХѥȽꤹؤŪˡ */
    if ((newPath = strchr(orgPath, ':')) != NULL) {
        return (newPath + 1);
    } else {
        return (orgPath);
    }
}

/*
 * ǥ쥯ȥ꤬¸ߤ뤫Ĵ٤롥dnameǥ쥯ȥǤϤʤ硤
 * åФƽλ롥
 *     Rerurn value:
 *         TRUE: ¸ߤ롥
 *         ʳ: ¸ߤʤ
 * 
 * : ̾νҸƱͤˡ¸ߤȤTRUE֤褦ѹ(1996/03/16)
 */

SpdBoolean spdIsDirectory(const char *dname)
{
    struct stat buf;
    
    if (stat(dname, &buf) == -1) {
        return (FALSE);
    }
    if ((buf.st_mode & S_IFMT) != S_IFDIR) {
        spdErrorExit("spdIsDirectory(): \"%s\": Not directory.\n", dname);
    }
    return (TRUE);
}

/*
 * ե뤬¸ߤ뤫Ĵ٤롥
 */

SpdBoolean spdIsFile(const char *path)
{
    struct stat buf;
    int s;
    
    if (stat(path, &buf) == -1) {
        return (FALSE);
    }
    s = buf.st_mode & S_IFMT;

    /* Modified by S.Yamamoto 1999/02/11. */
    /* return (s == S_IFREG || s == S_IFLNK); */
#ifdef WIN32
    return (s == S_IFREG);
#else
    return (s == S_IFREG || s == S_IFLNK);
#endif
}

/*
 * ǥ쥯ȥ롥
 */

void spdMkdir(SpdConstString path)
{
    if (mkdir(path, 0755) == -1) {
        spdErrorExit("spdMkdir(): \"%s\": Can't make directory.\n", path);
    }
}

/*
 * ǥ쥯ȥ꤬¸ߤʤкƵŪ˺롥
 * "./XXX/YYY/ZZZ/./../ZZZ/."Τ褦ŪʾθƤ롥
 */

SpdBoolean spdMkdirR(SpdConstString path)
{
    char    *newpath, *p;
    
    /* spdMessage("spdMkdirR(): \"%s\".\n", path); */
    if (!spdIsDirectory(path)) {
        /* Ǹ"/"ʲѥ롥 */
        newpath = spdStrdup(path);
        if ((p = strrchr(newpath, '/')) != NULL) {
            *p = 0;
            spdMkdirR(newpath);
        }
        spdFree(newpath);

        if (!spdIsDirectory(path)) {
            /* ꤵ줿ѥΤ롥 */
            spdMessage("spdMkdirR(): \"%s\": Making directory.\n", path);
            spdMkdir(path);
        }

        /* ǰΤᡤǥ쥯ȥ¸ߤǧ롥 */
        return (spdIsDirectory(path));
    } else {
        return (TRUE);
    }
}

/*
 * ե롥
 */

void spdUnlink(SpdConstString path)
{
    if (unlink(path) == -1) {
        spdErrorExit("spdUnlink(): \"%s\": Can't remove directory entry.\n",
                     path);
    }
}

#if !defined(WIN32)

/*
 * name1symbolic linkȤname2롥
 * C.f. "spdSymlink.cc" in Windows environment.
 */

void spdSymlink(SpdConstString name1, SpdConstString name2)
{
    spdMessage("spdSymlink(): \"%s\" <- \"%s\".\n", name1, name2);

    if (symlink(name1, name2) == -1) {
        spdErrorExit("spdSymlink(): \"%s\", \"%s\": Can't make a link.\n",
                     name1, name2);
    }
}

#endif /* !defined(WIN32) */

/******************************************************************************/

void spdStat(const char *path, struct stat *buf)
{
    if (stat(path, buf) != 0) {
        spdErrorExit("spdStat(): \"%s\": errno %d: stat() failed.\n",
                     path, errno);
    }
}

void spdFstat(int filedes, struct stat *buf)
{
    if (fstat(filedes, buf) != 0) {
        spdErrorExit("spdFstat(): %d: errno %d: fstat() failed.\n",
                     filedes, errno);
    }
}

#if !defined(WIN32)

void spdLstat(const char *path, struct stat *buf)
{
    if (lstat(path, buf) != 0) {
        spdErrorExit("spdLstat(): \"%s\": errno %d: lstat() failed.\n",
                     path, errno);
    }
}

#endif /* !defined(WIN32) */

/******************************************************************************/

/*
 * basename().
 *
 * basename("/usr/"):       "/"
 * spdBasename("/usr/"):    Error
 */

SpdConstString spdBasename(SpdConstString path)
{
    SpdConstString  cp;

    if (path == NULL) {
        spdErrorExit("spdBasename(): \"%s\": Invalid path.\n", path);
    }

    if ((cp = strrchr(path, '/')) != NULL) {
        cp++;
        if (*cp == '\0') {
            spdErrorExit("spdBasename(): \"%s\": Invalid trailing '/'.\n",
                         path);
        }
    } else {
        cp = path;
    }

    /* !! Temporarily. */
    return ((char *)cp);
}

/*
 * fileγĥҤextͿ줿ĥҤȽꤹ롥
 */

SpdBoolean spdIsFileExtension(SpdConstString file, SpdConstString ext)
{
    int     i, j;

    if (file == NULL || ext == NULL) {
        return (FALSE);
    }
    i = strlen(file) - 1;
    j = strlen(ext) - 1;
    if (i < j) {
        /* ĥ̾ե꤬̾ĹȤFALSE֤ */
        return (FALSE);
    }

    /*
     * ʸӤʸǤäFALSE֤
     */
    while (0 <= j) {
        if (file[i] != ext[j]) {
            return (FALSE);
        }
        i--;
        j--;
    }
    return (TRUE);
}

/* 
 * ե̾γĥҤoldExtnewExtѹ롥ΰ褬Ƥ롥
 * 
 * Return value:
 *     NULL:     Ϥ꤬롥
 *     ʳ: ĥҤѹ줿ե̾ؤݥ󥿡
 *
 * Note: ƤΰϸƤӽФ¦ǲ뤳ȡ
 */

SpdConstString spdCnvFileExtension(SpdConstString file,
                                   SpdConstString oldExt,
                                   SpdConstString newExt)
{
    int         f, o, n;
    char        *buf;
    char        *to;

    if (file == NULL || oldExt == NULL || newExt == NULL ||
        !spdIsFileExtension(file, oldExt)) {
        /* ϡ뤤fileoldExtʤNULL֤ */
        return (NULL);
    }

    f = strlen(file);
    o = strlen(oldExt);
    n = strlen(newExt);
    buf = to = spdMalloc(f - o + n + 1);

    /* Copy file without oldExt. */
    f = f - o;
    while (0 < f--) {
        *to++ = *file++;
    }

    /* Concatenate newExt. */
    while (*newExt != '\0') {
        *to++ = *newExt++;
    }

    /* Set terminator. */
    *to = '\0';

    return (buf);
}

/* 
 * ե̾γĥҤoldExtnewExtѹ롥˲Ūѹ롥
 * newExtoldExtĹˤϡƤӽФ¦ǽʬΰݤ뤳ȡ
 * 
 * Return value:
 *     NULL:     Ϥ꤬롥
 *     ʳ: ĥҤѹ줿ե̾ؤݥ󥿡
 */

SpdString spdReplaceFileExtension(SpdString file,
                                  SpdConstString oldExt,
                                  SpdConstString newExt)
{
    int             f, o, n;
    char            *cp;

    if (file == NULL || oldExt == NULL || newExt == NULL ||
        !spdIsFileExtension(file, oldExt)) {
        /* ϡ뤤fileoldExtʤNULL֤ */
        return (NULL);
    }

    f = strlen(file);
    o = strlen(oldExt);
    n = strlen(newExt);
    cp = file;

    /* Skip file without oldExt. */
    f = f - o;
    while (0 < f--) {
        cp++;
    }

    /* Concatenate newExt. */
    while (*newExt != '\0') {
        *cp++ = *newExt++;
    }

    /* Set terminator. */
    *cp = '\0';

    /* spdMessage("spdReplaceFileExtension(): \"%s\".\n", file); */
    return (file);
}
