/* 
 * Program:             $RCSfile: simplifyPath.c,v $  $Revision: 40.2 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohkubo  2001/08/17
 *                      S.Yamamoto  2001/08/17
 * 
 * (C) Copyright:       H.Ohkubo and S.Yamamoto  2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: simplifyPath.c,v 40.2 2001/09/09 04:57:38 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DEBUG */

#if defined(lint)
#define DW              printf      /* Just for cheking. */
#elif defined(DEBUG)
#define DW              spdMessage
#else
#define DW              (void)
#endif

/*
 * PREFIX is sentine:
 *     "/b":   For absolute path.
 *     "/../": For relative path.
 *
 * PREFIX_LEN is MAX(strlen("/b"), strlen("/../")).
 */

#define PREFIX_LEN      4

#define IS_SEP_OR_ZERO(x)   ((x) == SPD_PATH_SEP || (x) == '\0')

#define IS_DOT(p)           (*(p) == '.' && IS_SEP_OR_ZERO(*(p + 1)))
#define IS_DOT_DOT(p)       (*(p) == '.' && *((p) + 1) == '.' && \
                             IS_SEP_OR_ZERO(*(p + 2)))

/*
 * Variable r[] is used to return value of simplify_{absolute,relative}().
 */

static char r[FILENAME_MAX + PREFIX_LEN + 1];

static void simplify_absolute(SpdConstString p);
static void simplify_relative(SpdConstString p);

/* 
 * ǽʸ¤"."".."ơͿ줿path󲽤롥
 * ʤȤNULL֤
 * 
 * Note:
 * (1) relative path󲽤ˡCWD (Current Working Directory)
 *     ˴ؤѤʤ
 * (2) directoryfile¸ߤݾڤʤ
 *
 * Hisotry:
 * (1) "/.."򥨥顼ǤϤʤ"/"Ȥư褦ѹ
 * 
 * ToDo:
 * (1) WindowsǤϡɥ饤̾θ뤳ȡ
 * (2) WindowsǤϡ'/''\\'ξpath separatorȤưɬפ?
 * (3) "//"ΰϹθƤʤޤΤpathʤΤ?
 *     ޤɬפϤ뤫?
 */

SpdString spdSimplifyPath(SpdConstString path)
{
    if (path == NULL || path[0] == '\0') {
        return (NULL);
    } else if (path[0] == SPD_PATH_SEP) {
        simplify_absolute(path) ;
        DW("spdSimplifyPath(): simplify_absolute(): \"%s\".\n", path);
        /* Return "/" if the result is "". For example "/foo/..". */
        if (r[PREFIX_LEN] == '\0') {
            r[PREFIX_LEN] = SPD_PATH_SEP;
            r[PREFIX_LEN + 1] = '\0';
        }
    } else {
        simplify_relative(path) ;
        DW("spdSimplifyPath(): simplify_relative(): \"%s\".\n", path);
        /* Return "." if the result is "". For example "foo/..". */
        if (r[PREFIX_LEN] == '\0') {
            r[PREFIX_LEN] = '.';
            r[PREFIX_LEN + 1] = '\0';
        }
    }
    return (spdStrdup(&(r[PREFIX_LEN])));
}

/* 
 * absolute path󲽤롥
 * 
 * ToDo:
 * (1) (p[i] == SPD_PATH_SEP && p[i + 1] == '\0')˥å뤳ȡ
 */

static void simplify_absolute(SpdConstString p)
{
    int         i = 0;
    int         j = PREFIX_LEN;

    /* Set sentine. Don't mind r[0] and r[1]. */
    r[2] = SPD_PATH_SEP;
    r[3] = 'b';

    while (p[i] != '\0') {
        DW("while-beg:  i = %d '%c', j = %d.\n", i, p[i], j); 
        if (p[i] == SPD_PATH_SEP) {
            r[j++] = p[i++];
        } else if (IS_DOT_DOT(&(p[i]))) {
            /* 
             * component"..(/|\0)"ʤΤǡϤ'/'ޤ᤹
             * sentinelޤä顤᤮ʤΤǡ"/..""/"Ȥư
             */
            i += 2;
            j -= 2;
            while (r[j] != SPD_PATH_SEP) {
                j--;
            }
            if (j < PREFIX_LEN) {
                /* Treat "/.." as "/". */
                j = PREFIX_LEN;
            }
        } else if (IS_DOT(&(p[i]))) {
            /* component".(/|\0)"ʤΤǡåפ롥 */
            i++;
            j--;
        } else {
            /* componentϤ˥ԡ롥 */
            while (!IS_SEP_OR_ZERO(p[i])) {
                r[j++] = p[i++];
            }
        }
    }
    DW("while-exit: i = %d, j = %d.\n", i, j); 

    /* Set a terminator. */ 
    r[j++] = '\0';
    DW("          : i = %d, j = %d. (\"%s\")\n", i, j, &(r[PREFIX_LEN])); 

    /* Delete the last "/" if the result ends with a "/". */
    if (PREFIX_LEN < j - 2 && r[j - 2] == SPD_PATH_SEP) {
        r[j - 2] = '\0';
    }
}

/* 
 * relative path󲽤롥
 * 
 * absolute pahtȤ:
 * (1) sentinelޤ".."äȤΤޤ޽񤭹ࡥ
 *     CWDξѤʤΤǡ줬ʽǤ롥
 * (2) component".."ǡĴά̤κǸcomponent".."ʤ顤
 *     Τޤ޽񤭹ࡥ
 * (3) sentinel"/../"Ȥ롥(2)ɽƤ롥
 */

static void simplify_relative(SpdConstString p)
{
    int         i = 0;
    int         j = PREFIX_LEN;

    /* Set sentinel. */
    r[0] = SPD_PATH_SEP;
    r[1] = '.';
    r[2] = '.';
    r[3] = SPD_PATH_SEP;

    while (p[i] != '\0') {
        DW("while-beg:  i = %d '%c', j = %d.\n", i, p[i], j); 
        if (p[i] == SPD_PATH_SEP) {
            r[j++] = p[i++];
        } else if (IS_DOT_DOT(&(p[i]))) {
            i += 2;
            if (r[j - 4] == SPD_PATH_SEP && r[j - 3] == '.' &&
                r[j - 2] == '.' && r[j - 1] == SPD_PATH_SEP) {
                /* ľcomponent".."ʤΤǡʤ */
                r[j++] = '.';
                r[j++] = '.';
            } else {
                j -= 2;
                while (r[j] != SPD_PATH_SEP) {
                    j--;
                }
            }
        } else if (IS_DOT(&(p[i]))) {
            i++;
            j--;
        } else {
            while (!IS_SEP_OR_ZERO(p[i])) {
                r[j++] = p[i++];
            }
        }
    }
    DW("while-exit: i = %d, j = %d.\n", i, j); 

    if (j < PREFIX_LEN) {
        /* The last component is ".." without "/". */
        j = PREFIX_LEN;
    }

    /* Set a terminator. */ 
    r[j++] = '\0';
    DW("          : i = %d, j = %d. (\"%s\")\n", i, j, &(r[PREFIX_LEN])); 

    if (PREFIX_LEN < j - 2 && r[j - 2] == SPD_PATH_SEP) {
        r[j - 2] = '\0';
    }
}
