/* 
 * Program:             $RCSfile: spdIO.c,v $  $Revision: 30.11 $
 * 
 * Purpose:
 * 
 * Author:              S.Yamamoto  1998/12/05
 * 
 * (C) Copyright:       S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdIO.c,v 30.11 2003/05/21 00:13:22 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#ifdef SAPID_AR3_KERNEL

static char     cache[SPD_AR_CACHE_SIZE][SPD_LINE_LEN + 1];

int spdReadIntAttr(FILE *fp)
{
    return (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB ?
            spdReadIntAttrText(fp) :
            spdReadIntAttrBinary(fp));
}

int spdReadIntAttrText(FILE *fp)
{
    int             c, i = 0;
    struct stat     buf;
    static char     buffer[SPD_LINE_LEN + 1];
  
    while ((c = fgetc(fp)) != SDB_ATTR_SEPARATOR && c != EOF) {
        buffer[i++] = c;
    }
    buffer[i] = '\0';

    if (c == EOF) {
        spdFstat(fileno(fp), &buf);
        spdError("spdReadIntAttrText(): st_ino %d.\n", buf.st_ino);
        spdErrorExit("spdReadIntAttrText(): %d.\n", ftell(fp));
    }

    /* !! Should do range check. */
    return (atoi(buffer));
}
 
int spdReadIntAttrBinary(FILE *fp)
{
    int         value;

    if (fread(&value, sizeof(int), 1, fp) != 1) {
        spdErrorExit("spdReadIntAttrBinary(): %d.\n", ftell(fp));
    }
    return (SPD_NTOH_INT(value));
}

/******************************************************************************/

long spdReadLongAttr(FILE *fp)
{
    return (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB ?
            spdReadLongAttrText(fp) :
            spdReadLongAttrBinary(fp));
}

long spdReadLongAttrText(FILE *fp)
{
    int             c, i = 0;
    struct stat     buf;
    static char     buffer[SPD_LINE_LEN + 1];
  
    while ((c = fgetc(fp)) != SDB_ATTR_SEPARATOR && c != EOF) {
        buffer[i++] = c;
    }
    buffer[i] = '\0';

    if (c == EOF) {
        spdFstat(fileno(fp), &buf);
        spdError("spdReadLongAttrText(): st_ino %d.\n", buf.st_ino);
        spdErrorExit("spdReadLongAttrText(): %d.\n", ftell(fp));
    }

    /* !! Should do range check. */
    return (strtol(buffer, (char **)NULL, 10));
}
 
long spdReadLongAttrBinary(FILE *fp)
{
    long        value;

    if (fread(&value, sizeof(long), 1, fp) != 1) {
        spdErrorExit("spdReadLongAttrBinary(): %d.\n", ftell(fp));
    }
    return (SPD_NTOH_LONG(value));
}

/******************************************************************************/

const char *spdReadStrAttr(FILE *fp)
{
    return (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB ?
            spdReadStrAttrText(fp) :
            spdReadStrAttrBinary(fp));
}

char *spdReadStrAttrText(FILE *fp)
{
    int             c, i = 0, saved_n;
    struct stat     buf;
    static int      n = 0;

    while ((c = fgetc(fp)) != SDB_ATTR_SEPARATOR && c != EOF) {
        cache[n][i++] = c;
    }
    cache[n][i] = '\0';

    if (c == EOF) {
        spdFstat(fileno(fp), &buf);
        spdError("spdReadStrAttrText(): st_ino %d.\n", buf.st_ino);
        spdErrorExit("spdReadStrAttrText(): %d.", ftell(fp));
    }

    saved_n = n;
    n = (n + 1) % SPD_AR_CACHE_SIZE;
    return (cache[saved_n]);
}

char *spdReadStrAttrBinary(FILE *fp)
{
    int             c, i = 0, saved_n;
    struct stat     buf;
    static int      n = 0;

    while ((c = fgetc(fp)) != '\0' && c != EOF) {
        cache[n][i++] = c;
    }
    cache[n][i] = '\0';

    if (c == EOF) {
        spdFstat(fileno(fp), &buf);
        spdError("spdReadStrAttrBinary(): st_ino %d.\n", buf.st_ino);
        spdErrorExit("spdReadStrAttrBinary(): %d.\n", ftell(fp));
    }

    saved_n = n;
    n = (n + 1) % SPD_AR_CACHE_SIZE;
    return (cache[saved_n]);
}

/******************************************************************************/

void spdWriteIntAttr(FILE *fp, int value)
{
    if (spdGetWriteSDBformat() == SAPID_TEXT_FORMAT_SDB) {
        spdWriteIntAttrText(fp, value);
    } else {
        spdWriteIntAttrBinary(fp, value);
    }
}

void spdWriteIntAttrText(FILE *fp, int value)
{
    fprintf(fp, "%d%c", value, SDB_ATTR_SEPARATOR);
}

void spdWriteIntAttrBinary(FILE *fp, int value)
{
    value = SPD_HTON_INT(value);
    fwrite(&value, sizeof(int), 1, fp);
}

/******************************************************************************/

void spdWriteLongAttr(FILE *fp, long value)
{
    if (spdGetWriteSDBformat() == SAPID_TEXT_FORMAT_SDB) {
        spdWriteLongAttrText(fp, value);
    } else {
        spdWriteLongAttrBinary(fp, value);
    }
}

void spdWriteLongAttrText(FILE *fp, long value)
{
    fprintf(fp, "%ld%c", value, SDB_ATTR_SEPARATOR);
}

void spdWriteLongAttrBinary(FILE *fp, long value)
{
    value = SPD_HTON_LONG(value);
    fwrite(&value, sizeof(long), 1, fp);
}

/******************************************************************************/

void spdWriteStrAttr(FILE *fp, const char *value)
{
    if (spdGetWriteSDBformat() == SAPID_TEXT_FORMAT_SDB) {
        spdWriteStrAttrText(fp, value);
    } else {
        spdWriteStrAttrBinary(fp, value);
    }
}

void spdWriteStrAttrText(FILE *fp, const char *value)
{
    if (value != NULL) {
        fputs(value, fp);
    }
    fputc(SDB_ATTR_SEPARATOR, fp);
}

void spdWriteStrAttrBinary(FILE *fp, const char *value)
{
    if (value != NULL) {
        fputs(value, fp);
    }
    fputc('\0', fp);
}

#endif /* SAPID_AR3_KERNEL */
