/* 
 * Program:             $RCSfile: spdInfo.c,v $  $Revision: 30.33 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu  1996/11/07
 *                      S.Yamamoto  1996/11/07
 * Modified:            R.Miura  1998/04/28
 * 
 * (C) Copyright:       N.Fukuyasu, R.Miura and S.Yamamoto  1991 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdInfo.c,v 30.33 2003/08/16 10:32:27 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#define NOT_EXIST               (-1)

#define MAX_ATTR_STR_LEN        128
#define MAX_VAL_STR_LEN         128

#define SPD_MSG_CANT_OPEN_INFO  "\"%s\": Can't open info file.\n"

typedef struct {
    char    attr[MAX_ATTR_STR_LEN];
    char    val[MAX_VAL_STR_LEN];
} SpdInfoElem;

typedef struct _SpdInfoData {
    size_t      size;
    SpdInfoElem *elem;
} SpdInfoData;

static SpdInfo *spdOpenInfo(SpdConstString dir, SpdConstString baseName,
                            SpdConstString suffix, int infoType);
static SpdInfo *spdOpenInfoSub(SpdString infoFile, int infoType);

#ifdef DAPID

/* read_all_infoɤ߹ߥס*/
#define INFOTYPE_SETANDGET 0
#define INFOTYPE_ADDANDGET 1

static int get_attr_num(SpdInfo *info, const char *attr, int start_num);

#endif /* DAPID */

/******************************************************************************/

/*
 * The name of the database information file is
 * "SDB_DIR/<info_dir>/<info_base_name>SPD_INFO_SUFFIX."
 * Added by N.Fukuyasu 2003/08/11.
 */

#define SPD_DEFAULT_INFO_DIR    "."
#define SPD_DEFAULT_INFO_BASE   ""

static char info_dir[FILENAME_MAX + 1] = SPD_DEFAULT_INFO_DIR;
static char info_base_name[FILENAME_MAX + 1] = SPD_DEFAULT_INFO_BASE;

void spdSetSDBinfoFile(SpdConstString dir, SpdConstString baseName)
{
    strcpy(info_dir, dir);
    strcpy(info_base_name, baseName);
}

void spdResetSDBinfoFile(void)
{
    spdSetSDBinfoFile(SPD_DEFAULT_INFO_DIR, SPD_DEFAULT_INFO_BASE);
}

/*
 * Open SDB info.
 */

SpdInfo *spdOpenSDBinfo(void)
{
    return (spdOpenInfo((info_dir != NULL ? info_dir : "."),
                        (info_base_name != NULL ? info_base_name : ""),
                        SPD_INFO_SUFFIX, INFOTYPE_SETANDGET));
}

/*
 * Open ISDB info.
 */

SpdInfo *spdOpenISDBinfo(SpdObjId fileId)
{
    return (spdOpenInfo(ISDB_DIR, spdObjIdStr(fileId), SPD_INFO_SUFFIX,
            INFOTYPE_SETANDGET));
}

/*
 * Open PIDB info.
 */

SpdInfo *spdOpenPIDBinfo(SpdObjId fileId)
{
    return (spdOpenInfo(PIDB_DIR, spdObjIdStr(fileId), SPD_INFO_SUFFIX,
            INFOTYPE_SETANDGET));
}

/* 
 * Open Freja info.
 */

SpdInfo *spdOpenJSDBinfo(SpdConstString frejaName)
{
    return (spdOpenInfo(frejaName, "", SPD_INFO_SUFFIX, INFOTYPE_SETANDGET));
}

#ifdef DAPID

SpdInfo *spdOpenDSDBinfo(SpdConstString fileName)
{
    return (spdOpenInfo(DSDB_DIR, fileName, SPD_INFO_SUFFIX,
            INFOTYPE_SETANDGET));
}

SpdInfo *spdOpenDSDBsrcfraginfo(SpdConstString fileName)
{
    return (spdOpenInfo(DSDB_DIR, fileName, DSDB_SRCFRAG_INFO_SUFFIX,
            INFOTYPE_ADDANDGET));
}

/* 
 * Ʊ°̾attrnameФʣͤ¸뤿δؿ
 * srcfrag infoǻѤ롥spdAddInfoString()ͤɲä
 * spdGetInfoStringFirst()spdGetInfoStringNext()ͤ򼡡˼롥
 */

void spdAddInfoInt(SpdInfo *info, SpdString attrName, int attrValue)
{
    spdAddInfoString(info, attrName, spdItoa(attrValue));
}

void spdAddInfoString(SpdInfo *info, SpdString attrName, SpdConstString attrValue)
{
    SpdInfoData     *data;

    if (attrName == NULL) {
        return;
    }

    data = info->data;

    /* Add value. */
    data->elem = spdRealloc(data->elem,
                            sizeof(SpdInfoElem) * (data->size + 1));
    strcpy(data->elem[data->size].attr, attrName);
    strcpy(data->elem[data->size].val, attrValue);
    data->size++;

    info->modified = TRUE;
}

SpdInfoCursor spdGetInfoStringFirst(SpdInfo *info, SpdString attrname)
{
    static SpdInfoCursor    cur;

    cur.info = info;
    cur.start_num = 0;
    strcpy(cur.attrname, attrname);
    return (cur);
}

SpdString spdGetInfoStringNext(SpdInfoCursor *cur)
{
    int     i;

    i = get_attr_num(cur->info, cur->attrname, cur->start_num);
    if (i == NOT_EXIST) {
        return (NULL);
    } else {
        cur->start_num = i + 1;
        return (cur->info->data->elem[i].val);
    }
}

#endif /* DAPID */

/******************************************************************************/

static int get_attr_num(SpdInfo *info, const char *attr, int start_num)
{
    int         i;
    SpdInfoData *data;

    if (attr == NULL) {
        return NOT_EXIST;
    }
    data = info->data;

    for (i = start_num; i < data->size; i++) {
        if (SPD_STREQ(data->elem[i].attr, attr)) {
            return i;
        }
    }
    return NOT_EXIST;
}

/* 
 * type: 
 *     INFOTYPE_SETANDGET: spdSetInfospdGetInfoΤɹߡ
 *     INFOTYPE_ADDANDGET: spAddInfospdGetXxxFirst, spdGetXxxNextΤ
 *                         ɹߡ
 */

static void read_all_info(SpdInfo *info, int type)
{
    int         c, i;
    FILE        *fp;
    SpdInfoElem elem;
    SpdInfoData *data;

    data = info->data;
    if ((fp = spdSimpleFopen(info->filename, SPD_FOPEN_RT_MODE)) == NULL) {
        return;
    }

    while ((c = fgetc(fp), ungetc(c, fp), c != EOF)) {
        /* Get attribute name. */
        i = 0;
        while ((c = fgetc(fp)) != EOF) {
            if (c == '=') {
                break;
            } else if (c == '\n') {
                spdErrorExit("Invalid Info format.\n");
            }
            if (i < MAX_ATTR_STR_LEN - 1) {
                elem.attr[i++] = c;
            }
        }
        elem.attr[i] = '\0';

        /* Get value string. */
        i = 0;
        while ((c = fgetc(fp)) != EOF) {    
            if (c == '\n') {
                break;
            }
            if (i < MAX_VAL_STR_LEN - 1) {
                elem.val[i++] = c;
            }
        }
        elem.val[i] = '\0';
        if (c == EOF) {
            spdErrorExit("Invalid Info format.\n");
        }

        /* Read into "info->data". */
        if ((i = get_attr_num(info, elem.attr, 0)) == NOT_EXIST ||
            type == INFOTYPE_ADDANDGET) {
            data->elem = spdRealloc(data->elem,
                                    sizeof(SpdInfoElem) * (data->size + 1));
            strcpy(data->elem[data->size].attr, elem.attr);
            strcpy(data->elem[data->size].val, elem.val);
            data->size++;
        } else {
            strcpy(data->elem[i].val, elem.val);
        }
    }
    spdFclose(fp);
}

static void write_all_info(SpdInfo *info)
{
    int         i;
    FILE        *fp;
    SpdInfoData *data;    

    data = info->data;
    fp = spdFopen(info->filename, SPD_FOPEN_WT_MODE);
    for (i = 0; i < data->size; i++) {
        fprintf(fp, "%s=%s\n", data->elem[i].attr, data->elem[i].val);
    }
    spdFclose(fp);
}

static SpdInfo *spdOpenInfo(SpdConstString dir, SpdConstString baseName,
                            SpdConstString suffix, int infoType)
{
    char    filename[FILENAME_MAX + 1];
    SpdInfo *info;

    sprintf(filename, "%s/%s/%s%s", spdGetSDBdir(), dir, baseName, suffix);
    if ((info = spdOpenInfoSub(filename, infoType)) == NULL) {
        spdErrorExit(SPD_MSG_CANT_OPEN_INFO, filename);
    }
    return (info);
}

/*
 * infoTypeINFOTYPE_ADDANDGETξϽʣ
 */

static SpdInfo *spdOpenInfoSub(SpdString infoFile, int infoType)
{
    SpdInfo     *info;

    info = spdCalloc(1, sizeof(SpdInfo));
    strcpy(info->filename, infoFile);
    info->modified = FALSE;
    info->data = spdCalloc(1, sizeof(SpdInfoData));
    info->data->size = 0;
    read_all_info(info, infoType);
    return (info);
}

SpdString spdGetInfoString(SpdInfo *info, SpdConstString attrName)
{
    int     i;

    if (attrName == NULL) {
        return NULL;
    }

    if ((i = get_attr_num(info, attrName, 0)) == NOT_EXIST) {
        return NULL;
    } else {
        return info->data->elem[i].val;
    }
}

int spdGetInfoInt(SpdInfo *info, SpdConstString attrName, int defaultValue)
{
    SpdString   info_str;

    info_str = spdGetInfoString(info, attrName);
    return (info_str == NULL ? defaultValue : atoi(info_str));
}

void spdSetInfoString(SpdInfo *info,
                      SpdConstString attrName, SpdConstString attrValue)
{
    int         i;
    SpdInfoData *data;

    if (attrName == NULL) {
        return;
    }

    data = info->data;

    /* Add or modify value. */
    if ((i = get_attr_num(info, attrName, 0)) == NOT_EXIST) {
        data->elem = spdRealloc(data->elem,
                                sizeof(SpdInfoElem) * (data->size + 1));
        strcpy(data->elem[data->size].attr, attrName);
        strcpy(data->elem[data->size].val, attrValue);
        data->size++;
    } else {
        strcpy(data->elem[i].val, attrValue);
    }

    info->modified = TRUE;
}

void spdSetInfoInt(SpdInfo *info, SpdConstString attrName, int attrValue)
{
    spdSetInfoString(info, attrName, spdItoa(attrValue));
}

void spdCloseInfo(SpdInfo *info)
{
    if (info->modified) {
        write_all_info(info);
    }
    spdFree(info->data->elem);
    spdFree(info->data);
    spdFree(info);
}

/******************************************************************************/

/*
 * Read number of objects from SDB Info.
 * Added by S.Yamamoto 2001/09/09.
 */

int spdGetInfoNumOfObjs(SpdInfo *info, SpdConstString className)
{
    int         nums;

    nums = spdGetInfoInt(info, className, 0);
    if (SPD_ID_SPACE < nums) {
        spdErrorExit("spdGetInfoNumOfObjs(): Class \"%s\", num %d: Too many objects.\n",
                     className, nums);
    }  
    return (nums);
}

void spdSetInfoNumOfObjs(SpdInfo *info, SpdConstString className, int nums)
{
    if (SPD_ID_SPACE < nums) {
        spdErrorExit("spdSetInfoNumOfObjs(): Class \"%s\", num %d: Too many objects.\n",
                     className, nums);
    }  
    spdSetInfoInt(info, className, nums);
}
