/* 
 * Program:             $RCSfile: spdQueue.c,v $  $Revision: 30.12 $
 * 
 * Purpose:             
 * 
 * Author:              T.Suzuki  1997/06/25
 *                      S.Yamamoto  1997/06/27
 * 
 * (C) Copyright:       T.Suzuki and S.Yamamoto  1997 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdQueue.c,v 30.12 2003/06/28 15:31:19 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* For debug. */

#ifdef DEBUG

int enqueue_c = 0;
int dequeue_c = 0;
int clean_c   = 0;
int queues    = 0;      /* 줿queueο */
int reallocs  = 0;      /* realloc() */
int max       = 0;      /* maxκ͡ */

void
spdQueue_print_info(void)
{
    spdMessage("spdQueue: enqueue(): %d, dequeue(): %d, clean_queue(): %d.\n",
               enqueue_c, dequeue_c, clean_c);
    spdMessage("    queues: %d, realloc(): %d, max size:%d.\n",
               queues, reallocs, max);
}

#endif /* DEBUG */

/******************************************************************************/

SpdQueueElemType    _spd_queue_dummy;

/******************************************************************************/

/* Initialize the Queue */

SpdQueue spdQueueInit(void)
{
    SpdQueue queue;
    queue.buff = NULL;
    queue.head = queue.tail = queue.elem = queue.max = 0;
    return queue;
}

/* Clear contents of the Queue */

void spdQueueClean(SpdQueue *queue)
{
#ifdef DEBUG
    clean_c++;
#endif

    if (queue->buff != NULL) {
        spdFree(queue->buff);
    }

    queue->head = queue->tail = queue->elem = queue->max = 0;
    queue->buff = NULL;
}

/* Initialize Buffer */

static void init_buffer(SpdQueue *queue)
{
#ifdef DEBUG
    queues++;
#endif
    queue->max = QUEUE_ALLOC_SIZE;
    queue->buff = spdMalloc(sizeof(SpdQueueElemType) * queue->max);
    queue->elem = queue->head = 0;
    queue->tail = 0;
    return;
}

/* Expand buffer at the point */

static void expand_buffer(SpdQueue *queue)
{
    int     src, dist, size;

    queue->max += QUEUE_ALLOC_SIZE;
    queue->buff = spdRealloc(queue->buff,
                             sizeof(SpdQueueElemType) * queue->max);

#ifdef DEBUG
    reallocs++;
    if (max < queue->max) {
        max = queue->max;
    }
#endif

    /* ʤ򥳥ԡơüˤ褻롥 */
    if (queue->tail < queue->elem / 2 && queue->tail <  QUEUE_ALLOC_SIZE) {
        src  = 0;
        dist = queue->elem;
        size = queue->tail;
        queue->head = queue->elem + size;
    } else {
        src  = queue->tail;
        dist = queue->max - (queue->elem - queue->tail);
        size = queue->elem - queue->tail;
        queue->tail = queue->max - (queue->elem - queue->tail);
    }
    memmove(&(queue->buff[dist]), &(queue->buff[src]), size*sizeof(int));

#ifdef DEBUG
    spdMessage("spdQueue: src:%d dist:%d size:%d head:%d tail:%d\n",
               src, dist, size, queue->head, queue->tail);
    if (dist+size >= queue->max)
        spdMessage("expand_buffer(): Inner error. Illegal copy range.\n");
#endif

    return;
}

void _spdQueuePushBack(SpdQueue *queue, SpdQueueElemType value)
{
#ifdef DEBUG
    enqueue_c++;
#endif

    if (queue->buff == NULL) {              /*  */
        init_buffer(queue);
    } else if (queue->elem == queue->max) { /* ΰγĥ */
        expand_buffer(queue);
    }

    if (queue->max <= queue->head) {
        queue->head = 0;
    }

    queue->buff[queue->head++] = value;
    queue->elem++;
}

void _spdQueuePushFront(SpdQueue *queue, SpdQueueElemType value)
{
#ifdef DEBUG
    enqueue_c++;
#endif

    if (queue->buff == NULL) {
        init_buffer(queue);
    } else if (queue->elem == queue->max) {
        expand_buffer(queue);
    }

    if (queue->tail == 0)
        queue->tail = queue->max;

    queue->buff[--queue->tail] = value;
    queue->elem++;
}

SpdQueueElemType _spdQueuePopBack(SpdQueue *queue)
{
    SpdQueueElemType value = SPD_QUEUE_NIL_VALUE;

    if (queue->elem <= 0) {
        return value;
    }

    --queue->head;
    if (queue->head < 0)
        queue->head = queue->max-1;
    
    value = queue->buff[queue->head];
    queue->elem--;

    if (queue->elem == 0) {
        spdQueueClean(queue);
    }

    return value;
}

SpdQueueElemType _spdQueuePopFront(SpdQueue *queue)
{
    SpdQueueElemType value = SPD_QUEUE_NIL_VALUE;

#ifdef DEBUG
    dequeue_c++;
#endif
    
    if (queue->elem <= 0) {                 /* ǤʤȤ */
        return value;
    }

    /* ХåեۤƤʤindexƬˤ */
    if (queue->max <= queue->tail) {
        queue->tail = 0;
    }

    value = queue->buff[queue->tail++];
    queue->elem--;
    
    /* Is it necessary? 1997/06/29. */ 
    if (queue->elem == 0) {                 /* Ǥʤʤäΰ */
        spdQueueClean(queue);
    }

    return value;
}

/* 衤̤ **NOCHECK** */

/*
 * Ϣ³ΰ˰ưƬΥɥ쥹֤
 */

SpdQueueElemType* spdQueueToArray(SpdQueue *queue)
{

    if (queue->tail == queue->max) queue->tail = 0;

    if (queue->tail > queue->head) {
        queue->buff = spdRealloc(queue->buff, 
                                 sizeof(int)*(queue->max+queue->head));
        memmove(&(queue->buff[queue->max]),
                &(queue->buff[0]), queue->head * sizeof(int));
        
        queue->max  += queue->head;
        queue->head = queue->tail+queue->elem;
    }
    return &queue->buff[queue->tail];
}

void spdQueueSplice(SpdQueue *queue, int pos, int len, int* dt, int dtlen)
{
    spdError("spdQueueSplice(): Not implemented.\n");
}

void spdQueueDelete(SpdQueue *queue, int s_pos, int len)
{
    int e1, i;

    SpdQueueElemType *po = spdQueueToArray(queue);
    
    if (s_pos < 0 || s_pos > queue->elem) return;
    
    if (s_pos+len > queue->elem) {
        queue->head = queue->tail + s_pos;
        queue->elem = s_pos;
    }

    e1 = queue->elem - len;
 
    for (i=s_pos; i<e1; i++) {
        po[i] = po[i+len];
    }
    queue->head -= len;
    queue->elem -= len;
}

static int compar_int(const SpdQueueElemType *a, const SpdQueueElemType *b)
{
    return *(int*)a - *(int*)b;
}

void spdQueueSort(SpdQueue *queue,
                  int (*compar)(const SpdQueueElemType *, const SpdQueueElemType *))
{
    SpdQueueElemType *po = spdQueueToArray(queue);
    if (compar == NULL) compar = compar_int;
    
    SPD_QSORT(po, queue->elem, sizeof(int), compar);
}

/* 
 * description:
 * ߤΥ塼Ƭ顤ꤵ줿֤Ǥͤ롥
 * ǼƤǿ¿֤ͤ򸫤褦Ȥ硤SPD_NON_ID
 * äơ֤ͤͤˤꡤưλ뤳ȤϰŪǤϤʤ
 */

SpdQueueElemType _spdQueuePeek(SpdQueue *queue, int pos)
{
    if (pos < 0 || pos >= queue->elem) {
        /* ϰϥС, ȤꤢSPAID_NON_ID֤ */
        static SpdQueueElemType ret_value = SPD_QUEUE_NIL_VALUE;
        return ret_value;
    }

    pos += queue->tail;
    if (pos >= queue->max)
        pos -= queue->max;

    return queue->buff[pos];
}

/* 
 * description:
 * 塼Ƭ顤ꤵ줿֤Ǥͤѹ롥
 * ǼƤǿ¿֤ͤѹ뤳ȤϤǤʤ
 * Ԥˤ SPD_NON_ID֤
 * ʳϽ񤭹ΤƱ֤ͤ
 */

SpdQueueElemType _spdQueuePoke(SpdQueue *queue, int pos, SpdQueueElemType val)
{
    if (pos < 0 || pos >= queue->elem) {
        static SpdQueueElemType ret_value = SPD_QUEUE_NIL_VALUE;
        return ret_value;
    }
    
    pos += queue->tail;
    if (pos >= queue->max)
        pos -= queue->max;
    
    return queue->buff[pos] = val;
}

/*
 * Get Value of Front or Back
 */

SpdQueueElemType _spdQueueGetBack(SpdQueue *queue)
{
    return _spdQueuePeek(queue, queue->elem-1);
}

SpdQueueElemType _spdQueueGetFront(SpdQueue *queue)
{
    return _spdQueuePeek(queue, 0);
}

/*
 * Linear Search through Queue
 * output:
 *   Index of the first element which value is `val'.
 *   If No element is found, return -1.
 *
 * Currently, only `int' is encoded.
 */

int _spdQueueSearchIdx(SpdQueue *queue, SpdQueueElemType val)
{
    int i, e = (queue->tail < queue->head) ? queue->head : queue->max;


    for (i = queue->tail; i < e; i++) {

#ifdef SPD_QUEUE_UNION_TYPE_ELEM
        if (queue->buff[i].i == val.i) return i-queue->tail;
#else
        if (queue->buff[i] == val) return i-queue->tail;
#endif

    }
    if (queue->tail < queue->head) {
        return -1;
    }
    
    for (i = 0; i < queue->head; i++) {

#ifdef SPD_QUEUE_UNION_TYPE_ELEM
        if (queue->buff[i].i == val.i) return i;
#else
        if (queue->buff[i] == val) return i;
#endif

    }
    return -1;
}
