/* 
 * Program:             $RCSfile: spdSet2.c,v $  $Revision: 30.13 $
 * 
 * Purpose:             
 * 
 * Author:              M.Hakamata  1998/07/05
 *                      A.Yoshida  2003/07/07
 *                      S.Yamamoto  1998/07/06
 * 
 * (C) Copyright:       M.Hakamata, A.Yoshida and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] = 
"$Id: spdSet2.c,v 30.13 2003/07/08 06:19:37 yamamoto Exp yamamoto $";
#endif

#include <limits.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/spdSet2.h>

#define SPDSET_ELEM_NUM_AVERAGE    (3UL)
#define SPDSET_INITIAL_BUCKET_NUM  (23U)
#define SPDSET_POOL_INITIAL_CAPA   (255U)
#define SPDSET_POOL_CAPA_INCREMENT (255U)

static unsigned int hashI(SpdSet2 set, SpdElemValue key);
static unsigned int hashS(SpdSet2 set, SpdElemValue key);
static unsigned int hashL(SpdSet2 set, SpdElemValue key);
static unsigned int hashO(SpdSet2 set, SpdElemValue key);

static int elemEqualI(SpdElemValue e1, SpdElemValue e2);
static int elemEqualS(SpdElemValue e1, SpdElemValue e2);
static int elemEqualL(SpdElemValue e1, SpdElemValue e2);
static int elemEqualO(SpdElemValue e1, SpdElemValue e2);

static void rebuildSet(SpdSet2 set);
static SpdElem *allocSpdElem(SpdSet2 set);
static void freeSpdElem(SpdSet2 set, SpdElem *elem);

/* #define SPDSET2_DEBUG */
#ifdef SPDSET2_DEBUG
static void spdSet2_debug(SpdString msg, SpdSet2 set);
#endif

/*************************** Creation of SpdSet ***************************/

/* 
 * 
 * hash ١Ȥ뽸Ѥΰݡ롥
 */
#ifdef SPDSET2_MEMORY_DEBUG
#define DW(f, l, ptr, msg) \
        spdMessage("SpdSet2-MT: %8x : %s: \"%s\" %d.\n", \
                   (ptr), (msg), (f), (l))

SpdSet2 spdCreateSet2Dbg(int keyType, int valueType, char *file, int line)
#else
SpdSet2 spdCreateSet2(int keyType, int valueType)
#endif
{
    SpdSet2 set;

    set = spdMalloc(sizeof(SpdSet2_));

    set->keyType   = keyType;
    set->keyFree = NULL;
    set->keyCopy = NULL;

    switch (keyType) {
    case SPDSET_INT_KEY:
        set->keyHash = hashI;
        set->keyEqual = elemEqualI;
        break;
    case SPDSET_LONG_KEY:
        set->keyHash = hashL;
        set->keyEqual = elemEqualL;
        break;
    case SPDSET_STRING_KEY:
        set->keyHash = hashS;
        set->keyEqual = elemEqualS;
        set->keyCopy = spdElemCopyS;
        break;
    case SPDSET_OBJID_KEY:
        set->keyHash = hashO;
        set->keyEqual = elemEqualO;
        break;
    case SPDSET_COPY_STRING_KEY:
        set->keyHash = hashS;
        set->keyEqual = elemEqualS;
        set->keyFree = spdElemFreeS;
        set->keyCopy = spdElemCopyS;
        break;
    default:
        spdErrorExit("createHashSet(): Unknown keyType: %d\n", keyType);
    }

    set->valueType = valueType;
    set->valueFree = NULL;
    set->valueCopy = NULL;

    switch (valueType) {
    case SPDSET_INT_VALUE:
    case SPDSET_LONG_VALUE:
    case SPDSET_POINTER_VALUE:
    case SPDSET_OBJID_VALUE:
        break;
    case SPDSET_STRING_VALUE:
        set->valueCopy = spdElemCopyS;
        break;
    case SPDSET_COPY_STRING_VALUE:
        set->valueFree = spdElemFreeS;
        set->valueCopy = spdElemCopyS;
        break;
    case SPDSET_KEY_IS_VALUE:
        break;
    default:
        spdErrorExit("createHashSet(): Unknown keyType: %d\n", keyType);
    }

    /* ХåȤν */
    set->initialBktNum = SPDSET_INITIAL_BUCKET_NUM;
    set->bktNum        = 1U;
    set->bktArray      = spdMalloc(sizeof(SpdElem *));
    set->bktArray[0]   = NULL;
    set->rebuildNum    = SPDSET_ELEM_NUM_AVERAGE * (set->bktNum);

    /* ǿν */
    set->elemNum    = 0UL;

    /* SpdElem pool */
    set->freeList   = NULL;
    set->pool       = NULL;
    set->poolSize   = 0UL;
    set->poolUsed   = 0UL;
    set->poolInitialCapacity   = SPDSET_POOL_INITIAL_CAPA;
    set->poolCapacityIncrement = SPDSET_POOL_CAPA_INCREMENT;

#ifdef SPDSET2_DEBUG
    spdSet2_debug("called createHashSet", set);
#endif
#ifdef SPDSET2_MEMORY_DEBUG
    DW(file, line, set, "spdCreateSet2");
#endif
    return set;
}

/*
 * ϥåɽƹؿ
 * ʤХåȿϽ礬¸ߤ֡뤳ȤϤʤ
 */

static void rebuildSet(SpdSet2 set)
{
    unsigned int     newBktNum;
    unsigned int     oldBktNum;
    SpdElem          **newBktAry;
    SpdElem          **oldBktAry;
    register SpdElem **newBkt;
    register SpdElem *cur;
    register SpdElem *next;
    unsigned int     i;
  
    if (set->bktNum == UINT_MAX) { 
        /* ХåȤθĿϡunsigned intκͤޤǤ */
        set->rebuildNum = ULONG_MAX;
        return;
    }

    /* resize bucket number  */
    newBktNum = set->bktNum;
    if (newBktNum == 1U) {
        if (set->initialBktNum == 1U) {
            newBktNum = 23U;
        } else {
            newBktNum = set->initialBktNum;
        }
    } else if (newBktNum < 10000U) {
        newBktNum =  newBktNum * 2U + 1U;
    } else {
        if ((unsigned long)newBktNum + 10000UL > (unsigned long)UINT_MAX) {
            /* newBktNum κͤ UINT_MAX  */
            newBktNum = UINT_MAX;
        } else {
            newBktNum += 10000U;
        }
    }
    
    /* newBktNum20ʲʤ褦ˤ */
    /* 20 κ bit̺֥ե빽¤p.334 */
    if (newBktNum % 2U == 0U) {
        newBktNum++;
    }

    while(newBktNum % 3U == 0U || newBktNum % 5U == 0U
          || newBktNum % 7U == 0U || newBktNum % 11U == 0U
          || newBktNum % 13U == 0U || newBktNum % 17U == 0U
          || newBktNum % 19U == 0U) {
        newBktNum += 2U;
    }
  
    /* Хåݤ */
    newBktAry = spdMalloc(sizeof(SpdElem *) * newBktNum);
    if (newBktAry == NULL) { 
        /* ΰݤ˼ԤƤ⡤ȤꤢǽλˤϤʤ */
        /* Ǥ⡤ޤ̣ʤ(Ǥ˥̤þƤ) */
        /* פˡƹФˤƤ롥*/
        set->rebuildNum += 1000UL;
        return;
    }
    
    /*  newBktNum, newBktAry  set  */
    /* ͤ¸                */
    oldBktNum     = set->bktNum;
    oldBktAry     = set->bktArray;
    set->bktNum   = newBktNum;
    set->bktArray = newBktAry;
    
    for (i = 0; i < newBktNum; i++) 
        newBktAry[i] = NULL;

    /* ǤˤĤơϥå夷ľ(linkp  next ΤĤʤľ) */
    for (i = 0; i < oldBktNum; i++)
        for (cur = oldBktAry[i]; cur != NULL; cur = next) { 
            next = cur->next;
            
            /* ꥹȤǤդä롥 */
            newBkt = &(newBktAry[cur->rawHash % newBktNum]);
            cur->next = *newBkt;
            *newBkt = cur;
        }
    set->rebuildNum = (unsigned long)set->bktNum * SPDSET_ELEM_NUM_AVERAGE;
    spdFree(oldBktAry);
}

/* SpdElem롥 */

static SpdElem *allocSpdElem(SpdSet2 set)
{
    SpdElemBlk *curBlk;
    SpdElemBlk *newBlk;

    set->poolUsed++;

    if (set->poolSize < set->poolUsed) {
        newBlk = (SpdElemBlk *)spdMalloc(sizeof(SpdElemBlk));
        newBlk->usedNum = 1UL;
        if (set->poolSize == 0UL)
            newBlk->blkSize = set->poolInitialCapacity;
        else
            newBlk->blkSize = set->poolCapacityIncrement;

        newBlk->array = spdMalloc(sizeof(SpdElem) * newBlk->blkSize);
        if (newBlk->array == NULL) {
            spdErrorExit("SpdSet: allocSpdElem can't alloc %d element(s).\n",
                         newBlk->blkSize);
        }
        newBlk->prev = set->pool;
        
        set->pool = newBlk;
        set->poolSize += newBlk->blkSize;
        return newBlk->array;
    }

    curBlk = set->pool;

    /* Υ֥å̤ SpdElem ФȤ */
    if (curBlk->usedNum < curBlk->blkSize) {
        return curBlk->array + (curBlk->usedNum++);
    }

    /* ⤷ե꡼ꥹȤ˶, SpdElem ļäƤ */
    if (set->freeList != NULL) {
        SpdElem *elem;
        
        elem = set->freeList;
        set->freeList = elem->next;
        return elem;
    }
    spdWarning("allocSpdElem(): invalid structure of pool.\n");
    return NULL;
}   

/*************************** Free of SpdSet ***************************/

static void freeSpdElem(SpdSet2 set, SpdElem *elem)
{
    assert(set != NULL);
#ifdef SPDSET2_DEBUG
    spdSet2_debug("called freeSpdElem", set);
#endif
    if (set->keyFree != NULL) {
        set->keyFree(elem->key);
    }
    if (set->valueFree != NULL) {
        set->valueFree(elem->value);
    }

    /* ե꡼ꥹȤɲ */
    elem->next = set->freeList;
    set->freeList = elem;
    set->poolUsed--;
}

/*
 * Τ˳ݤΰ롥
 */

void spdFreeSet2(SpdSet2 set)
{
    SpdElem      *elem;
    unsigned int i;
    SpdElemBlk   *cur;
    SpdElemBlk   *prev;
    
#ifdef SPDSET2_DEBUG
    spdSet2_debug("called freeFreeSet2", set);
#endif
    if (set == NULL) return;
    
    /* SpdElem ˳ݤ value ΰ*/
    if (set->valueFree != NULL) {
        for (i = 0; i < set->bktNum; i++) {
            for (elem = set->bktArray[i]; elem != NULL; elem = elem->next) {
                set->valueFree(elem->value);
            }
        }
    }

    /* key  string ΤȤ, key ϳ˳ݤ */
    if (set->keyFree != NULL) {
        for (i = 0; i < set->bktNum; i++) {
            for (elem = set->bktArray[i]; elem != NULL; elem = elem->next) {
                set->keyFree(elem->key);
            }
        }
    }

    /* SpdElemBlk Ȥ˲ */
    for (cur = set->pool; cur != NULL; cur = prev) {
        prev = cur->prev;
        spdFree(cur->array);
        spdFree(cur);
    }

    spdFree(set->bktArray);
    spdFree(set);
}

/********************** Hash functions for elements **********************/

/*
 * ʸ󤫤顤 hash ͤ
 */

static unsigned int hashS(SpdSet2 set, SpdElemValue key)
{
    unsigned int hash = 0;
    SpdConstString s = (SpdConstString)key.ptr;
    
    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);

    if (s == NULL) return 0;
    
    /* ʸλ hash ͤ 0 */
    for (; *s != '\0'; s++) {
        hash += (hash << 3) + (*s);
    }
    return hash;
}

/*
 * int ͤ顤 hashͤ
 */

static unsigned int hashI(SpdSet2 set, SpdElemValue key)
{
    assert(set->keyType == SPDSET_INT_KEY);
    return (long)key.word * 1103515245;
}

/*
 * long ͤ顤 hashͤ
 */

static unsigned int hashL(SpdSet2 set, SpdElemValue key)
{
    assert(set->keyType == SPDSET_LONG_KEY);
    return key.long_word * 1103515245;
}

/*
 * SpdObjId ͤ顤 hashͤ
 */

static unsigned int hashO(SpdSet2 set, SpdElemValue key)
{
    assert(set->keyType == SPDSET_OBJID_KEY);
    return (long)key.objId * 1103515245;
}

/********************** Equality of elements **********************/

static int elemEqualI(SpdElemValue e1, SpdElemValue e2)
{
    return e1.word == e2.word;
}

static int elemEqualS(SpdElemValue e1, SpdElemValue e2)
{
    return e1.ptr == e2.ptr || SPD_STREQ(e1.ptr, e2.ptr);
}

static int elemEqualL(SpdElemValue e1, SpdElemValue e2)
{
    return e1.long_word == e2.long_word;
}

static int elemEqualO(SpdElemValue e1, SpdElemValue e2)
{
    return e1.objId == e2.objId;
}

/*************************** Free of elements **************************/

void spdElemFreeS(SpdElemValue elem)
{
    if (elem.ptr != NULL) {
        spdFree(elem.ptr);
    }
}

void spdElemFreeP(SpdElemValue elem)
{
    if (elem.ptr != NULL) {
        spdFree(elem.ptr);
    }
}

/*************************** Copy of elements **************************/

void spdElemCopyS(SpdElemValue *dst, SpdElemValue *src)
{
    dst->ptr = spdStrdup(src->ptr);
}

void spdElemMoveS(SpdElemValue *dst, SpdElemValue *src)
{
    dst->ptr = src->ptr;
    src->ptr = NULL;
}

/********************** Basic Operations of SpdSet **********************/

/*
 * ؤǤɲá
 */

SpdElem *spdAddElem2(SpdSet2 set, SpdElemValue key, SpdElemValue val)
{
    SpdElem *elem = NULL;
    unsigned int rawHash;
    
    assert(set != NULL);
#ifdef SPDSET2_DEBUG
    spdSet2_debug("called spdAddElem2", set);
#endif

#if 0
    if (spdContains2(set, key)) { /* already added. */
        return NULL;
    }

    if (set->elemNum >= set->rebuildNum) {
        rebuildSet(set);
    }

    elem = allocSpdElem(set);

    /* ݥ󥿤ΤĤʤ */
    rawHash = set->keyHash(set, key);
    elem->next = set->bktArray[rawHash % set->bktNum];
    set->bktArray[rawHash % set->bktNum] = elem;
#else
    rawHash = set->keyHash(set, key);
    for(elem = set->bktArray[rawHash % set->bktNum]; 
        elem != NULL; elem = elem->next) {
        if (elem->rawHash == rawHash && set->keyEqual(key, elem->key)) {
            return NULL; /* already added */
        }
    }

    if (set->elemNum >= set->rebuildNum) {
        rebuildSet(set);
    }

    elem = allocSpdElem(set);
    elem->next = set->bktArray[rawHash % set->bktNum];
    set->bktArray[rawHash % set->bktNum] = elem;
#endif
    set->elemNum++;
    
    /*  hash ͤ¸ */
    elem->rawHash = rawHash;

    /* key ʣ */
    if (set->keyType == SPDSET_COPY_STRING_KEY &&  key.ptr != NULL) {
        elem->key.ptr = spdStrdup(key.ptr);
    } else {
        elem->key = key;
    }

    /* value ʣ */
    switch (set->valueType) {
    case SPDSET_INT_VALUE:
    case SPDSET_LONG_VALUE:
    case SPDSET_POINTER_VALUE:
    case SPDSET_OBJID_VALUE:
    case SPDSET_STRING_VALUE:
        elem->value = val;
        break;
    case SPDSET_COPY_STRING_VALUE:
        if (val.ptr != NULL) {
            elem->value.ptr = spdStrdup(val.ptr);
        } else {
            elem->value = val; /* i.e. NULL */
        }
        break;
    case SPDSET_KEY_IS_VALUE:
        elem->value = elem->key;
        break;
    default:
        spdErrorExit("spdAddElem2(): unknwon value type %d.\n", 
                     set->valueType);
        break;
    }

    return elem;
}

/*
 *  key Ǥ뤫å,   
 * ФǤб SpdElem ؤΥݥ󥿤֤ 
 */

SpdElem *spdSearchElem2(SpdSet2 set, SpdElemValue key)
{
    SpdElem      *elem;     
    unsigned int rawHash;

    assert(set != NULL);

    rawHash = set->keyHash(set, key);
    elem = set->bktArray[rawHash % set->bktNum];

    for(; elem != NULL; elem = elem->next) {
        if (elem->rawHash == rawHash && set->keyEqual(key, elem->key)) {
            return elem;
        }
    }
    return NULL;
}

/*
 * key ǻꤷǤ򡤽礫õ롥
 * õ1֤Ԥ0֤
 */

SpdBoolean spdDelElem2(SpdSet2 set, SpdElemValue key)
{
    SpdElem *elem = NULL;
    SpdElem *ptr;
    int rawhash;
    int hash;

    assert(set != NULL);
#ifdef SPDSET2_DEBUG
    spdSet2_debug("called spdDelElem2", set);
#endif

    rawhash = set->keyHash(set, key);
    hash = rawhash % set->bktNum;
    ptr = set->bktArray[hash];
    if (ptr == NULL) { /* key Ŭ礹Ǥ¸ߤʤ */
        return FALSE;
    }
    if (ptr->rawHash == rawhash && set->keyEqual(ptr->key, key)) {
        elem = ptr;
        set->bktArray[hash] = ptr->next;
    } else {
        for (/* */; ptr->next != NULL; ptr = ptr->next) {
            if (ptr->next->rawHash == rawhash && 
                set->keyEqual(ptr->next->key, key)) {
                elem = ptr->next;
                ptr->next = ptr->next->next;
                break;
            }
        }
        if (elem == NULL) {  /* key Ŭ礹ǤĤʤä */
            return FALSE;
        }
    }

    freeSpdElem(set, elem);
    set->elemNum--;

    return TRUE;
}

/*
 *   key Ǥ¸ߤ뤫å 
 * ȯ, TRUE 򤫤, Ԥ FALSE ֤   
 */

SpdBoolean spdContains2(SpdSet2 set, SpdElemValue key)
{
    return spdSearchElem2(set, key) != NULL ? TRUE : FALSE;
}

/********************** Cursors for SpdSet **********************/

/*
 * Curosr 
 * ˤϡΤɤǤؤƤʤɬ spdSetCursorNext() 
 * ȤäơǽǤؤ褦ˤɬפ롣
 */

SetCursor spdSetCursorCreate(SpdSet2 set)
{
    SetCursor csr;

#ifdef SPDSET2_DEBUG
    spdSet2_debug("called spdSetCursorCreate", set);
#endif

    csr.set = set;
    csr.curBkt = 0;
    csr.curElem = NULL;

    return csr;
}

/*
 * Cursor ʤ롣ʤ̡Ǥ TRUE 򡢤ʤ(Ĥޤ
 * ٤ƤǤéä֤Ǥ) FALSE ֤
 */

SpdBoolean spdSetCursorNext(SetCursor *csr)
{
    unsigned int i;

    assert(csr != NULL);
#ifdef SPDSET2_DEBUG
    spdSet2_debug("called spdSetCursorNext", csr->set);
#endif
    if (csr->set == NULL) return FALSE;

    if (csr->curElem != NULL) {
        csr->curElem = csr->curElem->next;
        if (csr->curElem != NULL) return TRUE;
    }

    for (i = csr->curBkt; i < csr->set->bktNum; i++) {
        if (csr->set->bktArray[i] != NULL) {
            csr->curElem = csr->set->bktArray[i];
            csr->curBkt = i + 1;
            break;
        }
    }
    return csr->curElem != NULL;
}

/********************** Sets operations for SpdSet **********************/

/*
 * ꤷ set ʣ롥ޤΨ褯ʤ
 */

SpdSet2 spdDupSet2(SpdSet2 src)
{
    SpdSet2 dst;
    SetCursor csr;

#ifdef SPDSET2_DEBUG
    spdSet2_debug("called spdDupSet2", src);
#endif
    dst = spdCreateSet2(src->keyType, src->valueType);
    dst->keyFree = src->keyFree;
    dst->keyCopy = src->keyCopy;
    dst->valueFree = src->valueFree;
    dst->valueCopy = src->valueCopy;

    if (src->elemNum > 0) {
        dst->initialBktNum = src->elemNum / SPDSET_ELEM_NUM_AVERAGE + 1;
        dst->poolInitialCapacity = src->elemNum;
        dst->poolCapacityIncrement = src->poolCapacityIncrement;
    }

    if (src->keyCopy && src->valueCopy) {
        for (csr = spdSetCursorCreate(src); spdSetCursorNext(&csr);) {
            SpdElemValue key, value;
            src->keyCopy(&key, &csr.curElem->key);
            src->valueCopy(&value, &csr.curElem->value);
            spdAddElem2(dst, key, value);
        }
    } else if (src->keyCopy) {
        for (csr = spdSetCursorCreate(src); spdSetCursorNext(&csr);) {
            SpdElemValue key;
            src->keyCopy(&key, &csr.curElem->key);
            spdAddElem2(dst, key, csr.curElem->value);
        }
    } else if (src->valueCopy) {
        for (csr = spdSetCursorCreate(src); spdSetCursorNext(&csr);) {
            SpdElemValue value;
            src->valueCopy(&value, &csr.curElem->value);
            spdAddElem2(dst, csr.curElem->key, value);
        }
    } else {
        for (csr = spdSetCursorCreate(src); spdSetCursorNext(&csr);) {
            spdAddElem2(dst, csr.curElem->key, csr.curElem->value);
        }
    }
    return dst;
}

SpdSet2 spdUnionSet2(SpdSet2 s1, SpdSet2 s2)
{
    SpdSet2 res;
    SetCursor csr;

    assert(s1->keyType == s2->keyType
           && s1->valueType == s2->valueType
           && s1->keyCopy == s2->keyCopy
           && s1->valueCopy == s2->valueCopy);

    if (s1->elemNum < s2->elemNum) {
        SpdSet2 tmp;
        tmp = s1; s1 = s2; s2 = tmp;
    }
    res = spdDupSet2(s1);
    if (s2->keyCopy && s2->valueCopy) {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            if (!spdContains2(res, csr.curElem->key)) {
                SpdElemValue key, value;
                s2->keyCopy(&key, &csr.curElem->key);
                s2->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(res, key, value);
            }
        }
    } else if (s2->keyCopy) {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            if (!spdContains2(res, csr.curElem->key)) {
                SpdElemValue key;
                s2->keyCopy(&key, &csr.curElem->key);
                spdAddElem2(res, key, csr.curElem->value);
            }
        }
    } else if (s2->valueCopy) {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            if (!spdContains2(res, csr.curElem->key)) {
                SpdElemValue value;
                s2->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(res, csr.curElem->key, value);
            }
        }
    } else {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            spdAddElem2(res, csr.curElem->key, csr.curElem->value);
        }
    }
    return res;
}

SpdSet2 spdAddSet2(SpdSet2 s1, SpdSet2 s2)
{
    SetCursor csr;

    assert(s1->keyType == s2->keyType
           && s1->valueType == s2->valueType
           && s1->keyCopy == s2->keyCopy
           && s1->valueCopy == s2->valueCopy);

    if (s2->keyCopy && s2->valueCopy) {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            if (!spdContains2(s1, csr.curElem->key)) {
                SpdElemValue key, value;
                s2->keyCopy(&key, &csr.curElem->key);
                s2->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(s1, key, value);
            }
        }
    } else if (s2->keyCopy) {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            if (!spdContains2(s1, csr.curElem->key)) {
                SpdElemValue key;
                s2->keyCopy(&key, &csr.curElem->key);
                spdAddElem2(s1, key, csr.curElem->value);
            }
        }
    } else if (s2->valueCopy) {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            if (!spdContains2(s1, csr.curElem->key)) {
                SpdElemValue value;
                s2->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(s1, csr.curElem->key, value);
            }
        }
    } else {
        for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
            spdAddElem2(s1, csr.curElem->key, csr.curElem->value);
        }
    }
    return s1;
}

SpdSet2 spdIntersectionSet2(SpdSet2 s1, SpdSet2 s2)
{
    SpdSet2 res;
    SetCursor csr;

    assert(s1->keyType == s2->keyType
           && s1->valueType == s2->valueType
           && s1->keyCopy == s2->keyCopy
           && s1->valueCopy == s2->valueCopy);

    if (s1->elemNum > s2->elemNum) {
        SpdSet2 tmp;
        tmp = s1; s1 = s2; s2 = tmp;
    }

    res = spdCreateSet2(s1->keyType, s1->valueType);

    if (s1->keyCopy && s1->valueCopy) {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (spdContains2(s2, csr.curElem->key)) {
                SpdElemValue key, value;
                s1->keyCopy(&key, &csr.curElem->key);
                s1->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(res, key, value);
            }
        }
    } else if (s1->keyCopy) {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (spdContains2(s2, csr.curElem->key)) {
                SpdElemValue key;
                s1->keyCopy(&key, &csr.curElem->key);
                spdAddElem2(res, key, csr.curElem->value);
            }
        }
    } else if (s1->valueCopy) {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (spdContains2(s2, csr.curElem->key)) {
                SpdElemValue value;
                s1->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(res, csr.curElem->key, value);
            }
        }
    } else {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (spdContains2(s2, csr.curElem->key)) {
                spdAddElem2(res, csr.curElem->key, csr.curElem->value);
            }
        }
    }

    return res;
}

SpdSet2 spdDifferenceSet2(SpdSet2 s1, SpdSet2 s2)
{
    SpdSet2 res;
    SetCursor csr;

    assert(s1->keyType == s2->keyType
           && s1->valueType == s2->valueType
           && s1->keyCopy == s2->keyCopy
           && s1->valueCopy == s2->valueCopy);

    res = spdCreateSet2(s1->keyType, s1->valueType);

    if (s1->keyCopy && s1->valueCopy) {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (!spdContains2(s2, csr.curElem->key)) {
                SpdElemValue key, value;
                s1->keyCopy(&key, &csr.curElem->key);
                s1->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(res, key, value);
            }
        }
    } else if (s1->keyCopy) {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (!spdContains2(s2, csr.curElem->key)) {
                SpdElemValue key;
                s1->keyCopy(&key, &csr.curElem->key);
                spdAddElem2(res, key, csr.curElem->value);
            }
        }
    } else if (s1->valueCopy) {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (!spdContains2(s2, csr.curElem->key)) {
                SpdElemValue value;
                s1->valueCopy(&value, &csr.curElem->value);
                spdAddElem2(res, csr.curElem->key, value);
            }
        }
    } else {
        for (csr = spdSetCursorCreate(s1); spdSetCursorNext(&csr);) {
            if (!spdContains2(s2, csr.curElem->key)) {
                spdAddElem2(res, csr.curElem->key, csr.curElem->value);
            }
        }
    }

    return res;
}

SpdSet2 spdSubtractSet2(SpdSet2 s1, SpdSet2 s2)
{
    SetCursor csr;

    assert(s1->keyType == s2->keyType
           && s1->valueType == s2->valueType);

    for (csr = spdSetCursorCreate(s2); spdSetCursorNext(&csr);) {
        spdDelElem2(s1,  csr.curElem->key);
    }
    return s1;
}

/********************** helper function for debugging **********************/

#ifdef SPDSET2_DEBUG

static void spdSet2_debug(SpdString msg, SpdSet2 set)
{
    FILE *log;

    if ((log = fopen("/tmp/spdset2.log", "a")) == NULL) {
        spdErrorExit("Can't open log file for SpdSet2.\n");
    }

    fprintf(log, "SpdSet2: %s\n", msg);
    if (set == NULL) {
        fprintf(log, "  set is NULL\n");
        return;
    }
    fprintf(log, "  keyType = %d\n", set->keyType);
    fprintf(log, "  valueType = %d\n", set->valueType);

    fclose(log);
}

#endif
