/* 
 * Program:             $RCSfile: spdSet3.cc,v $  $Revision: 41.6 $
 * 
 * Purpose:
 * 
 * Author:              S.Yamamoto  2003/01/25
 * 
 * (C) Copyright:       S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdSet3.cc,v 41.6 2003/07/31 12:29:30 yamamoto Exp yamamoto $";
#endif

extern "C" {
#include <stdio.h>
#include <Sapid/Sapid.h>
}

#include <iostream>
#include <map>

using namespace std;

SpdSet3 spdCreateSet3(size_t elemSiz)
{
  SpdSet3       newSet = (SpdSet3)spdMalloc(sizeof(_SpdSet3));
  if (elemSiz <= sizeof(Contents3)) {
    newSet->map3 = new map<int, Contents3>;
    newSet->elemNum = 0;
    newSet->elemSiz = elemSiz;
    newSet->isFirstCall = TRUE;
    return (newSet);
  } else {
    spdErrorExit("spdCreateSet3(): %d: Too large element size.\n", elemSiz);
    SPD_DUMMY_RETURN(NULL);
  }
}

Contents spdAddElem3(SpdSet3 set, int key, Contents contents)
{
  Contents3             c;
  memcpy(&c, contents, SPD_SS3_ELEM_MAX_BYTE);

  map<int, Contents3>   *mp = (map<int, Contents3> *)(set->map3);
  mp->insert(pair<int, Contents3>(key, c));

  set->elemNum++;
  assert(set->elemNum == mp->size());

  map<int, Contents3>::iterator p;
  p = mp->find(key);
  return (p->second.x);
}

Contents spdSearchElem3(SpdSet3 set, int key)
{
  map<int, Contents3>   *mp = (map<int, Contents3> *)(set->map3);

  map<int, Contents3>::iterator p;
  p = mp->find(key);

  // fprintf(stderr, "spdSearchElem3(%d): %d.\n", key, p != mp->end());
  if (p == mp->end()) {
    return (NULL);
  } else { 
    return (p->second.x);
  }
}

SpdBoolean spdSetContains3(SpdSet3 set, int key)
{
  return (spdSearchElem3(set, key) != NULL);
}

SpdSet3 spdDupSet3(SpdSet3 set)
{
  SpdSet3       newSet = (SpdSet3)spdMalloc(sizeof(_SpdSet3));

  newSet->elemNum = set->elemNum;
  newSet->elemSiz = set->elemSiz;
  newSet->isFirstCall = TRUE;
  newSet->map3 = new map<int, Contents3>(*(map<int, Contents3> *)(set->map3));
  return (newSet);
}

int spdDelElem3(SpdSet3 set, int key)
{
  map<int, Contents3>   *mp = (map<int, Contents3> *)(set->map3);
  mp->erase(key);

  set->elemNum--;
  assert(set->elemNum == mp->size());
  return (1);
}

void spdFreeSet3(SpdSet3 set)
{
  map<int, Contents3>   *mp = (map<int, Contents3> *)(set->map3);

  delete mp;
  spdFree(set);
}

Contents spdGetArrayOfElem3(SpdSet3 set)
{
  Contents c = spdMalloc(set->elemNum * set->elemSiz);

  map<int, Contents3>   *mp = (map<int, Contents3> *)(set->map3);
  int                   i = 0;
  map<int, Contents3>::iterator p = mp->begin();

  while (p != mp->end()) {
    memcpy(((char *)c + i * set->elemSiz), p->second.x, set->elemSiz);
    p++;
    i++;
  }

  return (c);
}

Contents spdGetAElem3(SpdSet3 set, int *key)
{
  map<int, Contents3>           *mp = (map<int, Contents3> *)(set->map3);
  static map<int, Contents3>::iterator  p;

  if (set->isFirstCall) {
    set->isFirstCall = FALSE;
    p = mp->begin();
  }

  if (p != mp->end()) {
    *key = p->first;
    return (p++->second.x);
  } else {
    return (NULL);
  }
}

SpdSet3 spdUnionSet3(SpdSet3 set1, SpdSet3 set2)
{
  if (set1->elemSiz != set2->elemSiz) {
    return NULL;
  }

  // map<int, Contents3>           *mp1 = (map<int, Contents3> *)(set1->map3);
  // map<int, Contents3>::iterator p1 = mp1->begin();

  map<int, Contents3>           *mp2 = (map<int, Contents3> *)(set2->map3);
  map<int, Contents3>::iterator p2 = mp2->begin();

  SpdSet3                       set3 = spdDupSet3(set1);
  map<int, Contents3>           *mp3 = (map<int, Contents3> *)(set3->map3);
  // map<int, Contents3>::iterator p3 = mp3->begin();
  
  while (p2 != mp2->end()) {
    mp3->insert(pair<int, Contents3>(p2->first, p2->second));
    p2++;
  } 
  set3->elemNum = mp3->size();

  return (set3);
}

SpdSet3 spdIntersectionSet3(SpdSet3 set1, SpdSet3 set2)
{
  // spdMessage("spdIntersectionSet3(): Start.\n");

  if (set1->elemSiz != set2->elemSiz) {
    return NULL;
  }

  // map<int, Contents3>           *mp1 = (map<int, Contents3> *)(set1->map3);
  // map<int, Contents3>::iterator p1 = mp1->begin();

  // map<int, Contents3>           *mp2 = (map<int, Contents3> *)(set2->map3);
  // map<int, Contents3>::iterator p2 = mp2->begin();

  SpdSet3                       set3 = spdCreateSet3(set1->elemSiz);

  spdClearSetEnumeration3(set1);
  // spdMessage("spdIntersectionSet3(): set1->elemNum %d.\n", set1->elemNum);
  for (unsigned int i = 0; i < set1->elemNum; i++) {
    int         key;
    int         *contents;

    // spdMessage("spdIntersectionSet3(): i %d.\n", i);
    contents = (int *)spdGetAElem3(set1, &key);
    // spdMessage("spdIntersectionSet3(): a: key %d, value %d.\n",
    //            key, *contents);
    // spdMessage("spdIntersectionSet3(): i %d, spdSetContains3(%d) %d.\n",
    //             i, key, spdSetContains3(set2, key));
    if (spdSetContains3(set2, key)) {
      spdAddElem3(set3, key, contents);
      // spdMessage("spdIntersectionSet3(): b: key %d, value %d.\n",
      //             key, *contents);
    }
    // spdMessage("spdIntersectionSet3(): i %d, set3->elemNum %d.\n",
    //            i, set3->elemNum);
  }

  return (set3);
}

SpdSet3 spdDifferenceSet3(SpdSet3 set1, SpdSet3 set2)
{
  // spdMessage("spdDifferenceSet3(): Start.\n");

  if (set1->elemSiz != set2->elemSiz) {
    return NULL;
  }

  // map<int, Contents3>           *mp1 = (map<int, Contents3> *)(set1->map3);
  // map<int, Contents3>::iterator p1 = mp1->begin();

  // map<int, Contents3>           *mp2 = (map<int, Contents3> *)(set2->map3);
  // map<int, Contents3>::iterator p2 = mp2->begin();

  SpdSet3                       set3 = spdCreateSet3(set1->elemSiz);

  spdClearSetEnumeration3(set1);
  for (unsigned int i = 0; i < set1->elemNum; i++) {
    int         key;
    int         *contents;

    contents = (int *)spdGetAElem3(set1, &key);
    if (!spdSetContains3(set2, key)) {
      spdAddElem3(set3, key, contents);
    }
  }

  return (set3);
}

void spdClearSetEnumeration3(SpdSet3 set)
{
  set->isFirstCall = TRUE;
}
