/* 
 * Program:             $RCSfile: spdSetWrap.c,v $  $Revision: 50.3 $
 * 
 * Purpose:             
 * 
 * Author:              A.Yoshida   2003/06/09
 * 
 * (C) Copyright:       A.Yoshida and S.Yamamoto  2003 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] = 
"$Id: spdSetWrap.c,v 50.3 2003/07/08 05:46:13 yamamoto Exp yamamoto $";
#endif

#include <limits.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/spdSet2.h>

/**************  Wrapper funcions for spdAddElem2() ********************/

/*-------------- wrapper funciontions with string key -----------------*/

/*
 * key ʸvalue  key ƱǤǤɲ
 */

SpdElem *spdAddElemSK(SpdSet2 set, SpdConstString key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);
    assert(set->valueType == SPDSET_KEY_IS_VALUE);

    key_elem.ptr = (SpdSetAnyPointer)key;
    return spdAddElem2(set, key_elem, key_elem); 
    /* spdAddElem2() ignores the second key_elem. */
}

/*
 * key ʸ, value  int ͤǤǤɲ
 */

SpdElem *spdAddElemSI(SpdSet2 set, SpdConstString key, int val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);
    assert(set->valueType == SPDSET_INT_VALUE);

    key_elem.ptr = (SpdSetAnyPointer)key;
    val_elem.word = val;
    return spdAddElem2(set, key_elem, val_elem);
}       

/*
 * key, value ȤʸǤǤɲ
 * value  strdup Ѥ SpdElem ʣƤ롥
 * value  NULL ΤȤ NULL (!= "") ͤˤʤ롥
 */

SpdElem *spdAddElemSS(SpdSet2 set, SpdConstString key, SpdConstString val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);
    assert(set->valueType == SPDSET_STRING_VALUE
        || set->valueType == SPDSET_COPY_STRING_VALUE);

    key_elem.ptr = (SpdSetAnyPointer)key;
    val_elem.ptr = (SpdSetAnyPointer)val;
    return spdAddElem2(set, key_elem, val_elem);
}       

/*
 * key ʸ, value  ݥͤǤǤɲá
 * ݥͤ SpdElem Υѿ˥ԡ롥
 * ʤ, value λؤ object 뤫ɤϴΤ
 * ؤƤ data object ʣʤɤʤ
 */

SpdElem *spdAddElemSP(SpdSet2 set, SpdConstString key, SpdSetAnyPointer val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);
    assert(set->valueType == SPDSET_POINTER_VALUE);

    key_elem.ptr = (SpdSetAnyPointer)key;
    val_elem.ptr = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key ʸ, value  long ͤǤǤɲá
 */

SpdElem *spdAddElemSL(SpdSet2 set, SpdConstString key, long val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);
    assert(set->valueType == SPDSET_LONG_VALUE);

    key_elem.ptr = (SpdSetAnyPointer)key;
    val_elem.long_word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key ʸ, value  SpdObjId ͤǤǤɲá
 */

SpdElem *spdAddElemSO(SpdSet2 set, SpdConstString key, SpdObjId val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);
    assert(set->valueType == SPDSET_OBJID_VALUE);

    key_elem.ptr = (SpdSetAnyPointer)key;
    val_elem.objId = val;
    return spdAddElem2(set, key_elem, val_elem);
}


/*-------------- wrapper funciontions with integer key -----------------*/

/*
 * key  int , value  key ƱǤɲ
 * value  key Ʊͤ SpdElem Υѿ˳Ǽ
 */

SpdElem *spdAddElemIK(SpdSet2 set, int key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_INT_KEY);
    assert(set->valueType == SPDSET_KEY_IS_VALUE);

    key_elem.word = key;
    return spdAddElem2(set, key_elem, key_elem);
    /* spdAddElem2() ignores the second key_elem. */
}

/*
 * key, value Ȥ int ͤǤǤɲ
 * value  SpdElem Υѿ˳Ǽ
 */

SpdElem *spdAddElemII(SpdSet2 set, int key, int val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_INT_KEY);
    assert(set->valueType == SPDSET_INT_VALUE);

    key_elem.word = key;
    val_elem.word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  int , value ʸǤǤɲ
 * value  strdup Ѥ,SpdElem ΰʣ,
 * ؤΥݥ󥿤 SpdElem Υѿ˳Ǽ롥
 * val == NULL ξ ñ NULL (!= "") 롥
 */

SpdElem *spdAddElemIS(SpdSet2 set, int key, SpdConstString val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_INT_KEY);
    assert(set->valueType == SPDSET_STRING_VALUE
        || set->valueType == SPDSET_COPY_STRING_VALUE);

    key_elem.word = key;
    val_elem.ptr = (SpdSetAnyPointer)val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  int , value (data objectؤ)ݥ󥿤ǤǤɲ
 * value  SpdElem Υѿ˳Ǽ
 * ݥ󥿤 object 뤫ɤʤɤϴΤʤ
 */

SpdElem *spdAddElemIP(SpdSet2 set, int key, SpdSetAnyPointer val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_INT_KEY);
    assert(set->valueType == SPDSET_POINTER_VALUE);

    key_elem.word = key;
    val_elem.ptr = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  int , value  long ǤǤɲ
 */

SpdElem *spdAddElemIL(SpdSet2 set, int key, long val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_INT_KEY);
    assert(set->valueType == SPDSET_LONG_VALUE);

    key_elem.word = key;
    val_elem.long_word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  int , value  long ǤǤɲ
 */

SpdElem *spdAddElemIO(SpdSet2 set, int key, SpdObjId val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_INT_KEY);
    assert(set->valueType == SPDSET_OBJID_VALUE);

    key_elem.word = key;
    val_elem.objId = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*------------ wrapper funciontions with long integer key ---------------*/

/*
 * key  long , value  key ƱǤɲ.
 * value  key Ʊͤ SpdElem Υѿ˳Ǽ
 */

SpdElem *spdAddElemLK(SpdSet2 set, long key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_LONG_KEY);
    assert(set->valueType == SPDSET_KEY_IS_VALUE);

    key_elem.long_word = key;
    return spdAddElem2(set, key_elem, key_elem);
    /* spdAddElem2() ignores the second key_elem. */
}

/*
 * key  long , value  int ͤǤǤɲ
 * value  SpdElem Υѿ˳Ǽ
 */

SpdElem *spdAddElemLI(SpdSet2 set, long key, int val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_LONG_KEY);
    assert(set->valueType == SPDSET_INT_VALUE);

    key_elem.long_word = key;
    val_elem.word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  long , value ʸǤǤɲ
 * value  strdup Ѥ, SpdElem ΰʣ,
 * ؤΥݥ󥿤 SpdElemΥѿ˳Ǽ롥
 * val == NULL ξ ñ NULL (!= "") 롥
 */

SpdElem *spdAddElemLS(SpdSet2 set, long key, SpdConstString val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_LONG_KEY);
    assert(set->valueType == SPDSET_STRING_VALUE
        || set->valueType == SPDSET_COPY_STRING_VALUE);

    key_elem.long_word = key;
    val_elem.ptr = (SpdSetAnyPointer)val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  long, value (data objectؤ)ݥ󥿤ǤǤɲ
 * value  SpdElem Υѿ˳Ǽ
 * ݥ󥿤 object 뤫ɤʤɤϴΤʤ
 */

SpdElem *spdAddElemLP(SpdSet2 set, long key, SpdSetAnyPointer val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_LONG_KEY);
    assert(set->valueType == SPDSET_POINTER_VALUE);

    key_elem.long_word = key;
    val_elem.ptr = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  value  long ǤǤɲ
 */

SpdElem *spdAddElemLL(SpdSet2 set, long key, long val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_LONG_KEY);
    assert(set->valueType == SPDSET_LONG_VALUE);

    key_elem.long_word = key;
    val_elem.long_word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  long , value  SpdObjId ǤǤɲ
 */

SpdElem *spdAddElemLO(SpdSet2 set, long key, SpdObjId val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_LONG_KEY);
    assert(set->valueType == SPDSET_OBJID_VALUE);

    key_elem.long_word = key;
    val_elem.objId = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*------------- wrapper funciontions with SpdObjId key ---------------*/

/*
 * key  SpdObjId , value  key ƱǤɲ.
 * value  key Ʊͤ SpdElem Υѿ˳Ǽ
 */

SpdElem *spdAddElemOK(SpdSet2 set, SpdObjId key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);
    assert(set->valueType == SPDSET_KEY_IS_VALUE);

    key_elem.objId = key;
    return spdAddElem2(set, key_elem, key_elem);
    /* spdAddElem2() ignores the second key_elem. */
}

/*
 * key  SpdObjId , value  intͤǤǤɲ
 * value  SpdElem Υѿ˳Ǽ
 */

SpdElem *spdAddElemOI(SpdSet2 set, SpdObjId key, int val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);
    assert(set->valueType == SPDSET_INT_VALUE);

    key_elem.objId = key;
    val_elem.word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  SpdObjId , value ʸǤǤɲ
 * value  strdup Ѥ,SpdElem ΰʣ,
 * ؤΥݥ󥿤 SpdElem Υѿ˳Ǽ롥
 * val == NULL ξ ñ NULL (!= "") 롥
 */

SpdElem *spdAddElemOS(SpdSet2 set, SpdObjId key, SpdConstString val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);
    assert(set->valueType == SPDSET_STRING_VALUE
        || set->valueType == SPDSET_COPY_STRING_VALUE);

    key_elem.objId = key;
    val_elem.ptr = (SpdSetAnyPointer)val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  SpdObjId , value (data objectؤ)ݥ󥿤ǤǤɲ
 * value  SpdElem Υѿ˳Ǽ
 * ݥ󥿤 object 뤫ɤʤɤϴΤʤ
 */

SpdElem *spdAddElemOP(SpdSet2 set, SpdObjId key, SpdSetAnyPointer val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);
    assert(set->valueType == SPDSET_POINTER_VALUE);

    key_elem.objId = key;
    val_elem.ptr = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  SpdObj , value  long ǤǤɲ
 */

SpdElem *spdAddElemOL(SpdSet2 set, SpdObjId key, long val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);
    assert(set->valueType == SPDSET_LONG_VALUE);

    key_elem.objId = key;
    val_elem.long_word = val;
    return spdAddElem2(set, key_elem, val_elem);
}

/*
 * key  value  SpdObjId ǤǤɲ
 */

SpdElem *spdAddElemOO(SpdSet2 set, SpdObjId key, SpdObjId val)
{
    SpdElemValue key_elem;
    SpdElemValue val_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);
    assert(set->valueType == SPDSET_OBJID_VALUE);

    key_elem.objId = key;
    val_elem.objId = val;
    return spdAddElem2(set, key_elem, val_elem);
}


/**************  Wrapper funcions for spdSearchElem2() ********************/

/* 
 *  key Ǥ뤫å,   
 * ФǤб SpdElem ؤΥݥ󥿤֤
 */

SpdElem *spdSearchElemS(SpdSet2 set, SpdConstString key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);

    key_elem.ptr = (SpdSetAnyPointer)key;
    return spdSearchElem2(set, key_elem);
}

SpdElem *spdSearchElemI(SpdSet2 set, int key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_INT_KEY);

    key_elem.word = key;
    return spdSearchElem2(set, key_elem);
}

SpdElem *spdSearchElemL(SpdSet2 set, long key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_LONG_KEY);

    key_elem.long_word = key;
    return spdSearchElem2(set, key_elem);
}

SpdElem *spdSearchElemO(SpdSet2 set, SpdObjId key)
{
    SpdElemValue key_elem;

    assert(set->keyType == SPDSET_OBJID_KEY);

    key_elem.objId = key;
    return spdSearchElem2(set, key_elem);
}

/**************  Wrapper funcions for spdDelElem2() ********************/

/*
 * ǤκԤ
 */

SpdBoolean spdDelElemS(SpdSet2 set, SpdConstString key)
{
    SpdElemValue key_elem;

    assert(set != NULL);
    assert(set->keyType == SPDSET_STRING_KEY
        || set->keyType == SPDSET_COPY_STRING_KEY);

    key_elem.ptr = (SpdSetAnyPointer)key;
    return spdDelElem2(set, key_elem);
}

SpdBoolean spdDelElemI(SpdSet2 set, int key)
{
    SpdElemValue key_elem;

    assert(set != NULL);
    assert(set->keyType == SPDSET_INT_KEY);

    key_elem.word = key;
    return spdDelElem2(set, key_elem);
}

SpdBoolean spdDelElemL(SpdSet2 set, long key)
{
    SpdElemValue key_elem;

    assert(set != NULL);
    assert(set->keyType == SPDSET_LONG_KEY);

    key_elem.long_word = key;
    return spdDelElem2(set, key_elem);
}

SpdBoolean spdDelElemO(SpdSet2 set, SpdObjId key)
{
    SpdElemValue key_elem;

    assert(set != NULL);
    assert(set->keyType == SPDSET_OBJID_KEY);

    key_elem.objId = key;
    return spdDelElem2(set, key_elem);
}

/**************  Wrapper funcions for spdContains2() ********************/

SpdBoolean spdContainsS(SpdSet2 set, SpdConstString key)
{
    return spdSearchElemS(set, key) != NULL ? TRUE : FALSE;
}

SpdBoolean spdContainsI(SpdSet2 set, int key)
{
    return spdSearchElemI(set, key) != NULL ? TRUE : FALSE;
}

SpdBoolean spdContainsL(SpdSet2 set, long key)
{
    return spdSearchElemL(set, key) != NULL ? TRUE : FALSE;
}

SpdBoolean spdContainsO(SpdSet2 set, SpdObjId key)
{
    return spdSearchElemO(set, key) != NULL ? TRUE : FALSE;
}

/**************  Access funcions for spdElemValue ********************/

int spdSetGetValueI(SpdElemValue elem)
{
    return elem.word;
}

long spdSetGetValueL(SpdElemValue elem)
{
    return elem.long_word;
}

SpdConstString spdSetGetValueS(SpdElemValue elem)
{
    return (SpdConstString)elem.ptr;
}

SpdObjId spdSetGetValueO(SpdElemValue elem)
{
    return elem.objId;
}

SpdSetAnyPointer spdSetGetValueP(SpdElemValue elem)
{
    return elem.ptr;
}

/*--------------------------------------------------*/

void spdSetSetValueI(SpdElemValue *elem, int i)
{
    assert(elem != NULL);
    elem->word = i;
}

void spdSetSetValueL(SpdElemValue *elem, long i)
{
    assert(elem != NULL);
    elem->long_word = i;
}

void spdSetSetValueS(SpdElemValue *elem, SpdConstString s)
{
    assert(elem != NULL);   
    elem->ptr = (SpdSetAnyPointer)s;
}

void spdSetSetValueO(SpdElemValue *elem, SpdObjId id)
{
    assert(elem != NULL);
    elem->objId = id;
}

void spdSetSetValueP(SpdElemValue *elem, SpdSetAnyPointer ptr)
{
    assert(elem != NULL);
    elem->ptr = ptr;
}

/**************  Free funcions for SpdSet2 ********************/

void spdElemFreeSpdRel(SpdElemValue elem)
{
    spdFree(spdSetGetValueP(elem));
}

void spdElemFreeSpdOcc(SpdElemValue elem)
{
    spdFree(spdSetGetValueP(elem));
}

/**************  Copy funcions for SpdSet2 ********************/

void spdElemCopySpdRel(SpdElemValue *dst, SpdElemValue *src)
{
    dst->ptr = (SpdRel *)spdMalloc(sizeof(SpdRel));
    *(SpdRel *)dst->ptr = *(SpdRel *)src->ptr;
}

void spdElemMoveSpdRel(SpdElemValue *dst, SpdElemValue *src)
{
    dst->ptr = src->ptr;
    src->ptr = NULL;
}

void spdElemCopySpdOcc(SpdElemValue *dst, SpdElemValue *src)
{
    dst->ptr = (SpdOcc *)spdMalloc(sizeof(SpdOcc));
    *(SpdOcc *)dst->ptr = *(SpdOcc *)src->ptr;
}

void spdElemMoveSpdOcc(SpdElemValue *dst, SpdElemValue *src)
{
    dst->ptr = src->ptr;
    src->ptr = NULL;
}

