/* 
 * Program:             $RCSfile: spdStringPool.c,v $  $Revision: 30.5 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu  1998/09/09
 *                      S.Yamamoto  1998/09/09
 * 
 * (C) Copyright:       N.Fukuyasu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStringPool.c,v 30.5 2003/02/18 12:36:25 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DEBUG_STRING_POOL */

/*
 * Default size of string pool.
 */

/* #define SAPID_STRING_POOL_SIZE  0x1000 */
#define SAPID_STRING_POOL_SIZE  (1024 * 1024)

/******************************************************************************/

#ifdef DEBUG_STRING_POOL

static void print_current_state(SpdStringPool *pool)
{
    int                 i = 0;
    SpdStringPoolList   *pool_list_tmp;

    if (pool == NULL) {
        return;
    }

    pool_list_tmp = pool->pool_list;
    fprintf(stderr, "\n----- string pool (%x) -----\n", pool);
    while (pool_list_tmp) {
        i++;
        fprintf(stderr, "%2d: %x (%dbytes)\n", i, pool_list_tmp->str_pool,
                pool->pool_size);
        pool_list_tmp = pool_list_tmp->next;
    }
    fprintf(stderr, "space: %d pos: %x\n\n",
            pool->space_size, pool->current_pos);
}

#endif /* DEBUG_STRING_POOL */

/******************************************************************************/

/* Generate string pool. */

static void append_string_pool(SpdStringPool *pool)
{
    SpdStringPoolList *list_tmp;

    /*
     * copy_string_pool()Τ߸ƤФ졤poolNULLɤȽѤߡ
     *
     * if (pool == NULL) return;
     */

    /* 礭 pool->pool_size  string pool  str_pool ɲ */
    list_tmp = (SpdStringPoolList *)spdMalloc(sizeof(SpdStringPoolList));
    list_tmp->str_pool = (SpdString)spdMalloc(pool->pool_size);
    list_tmp->next = pool->pool_list;

    pool->pool_list = list_tmp;
    pool->space_size = pool->pool_size;
    pool->current_pos = pool->pool_list->str_pool;
}

/* Remove string pool. */

static void delete_string_pool(SpdStringPool *pool)
{
    SpdStringPoolList *list_tmp;

    /*
     * spdStrPoolFree()Τ߸ƤФ졤poolNULLɤȽѤߡ
     *
     * if (pool == NULL) return;
     */

    while (pool->pool_list) {
        list_tmp = pool->pool_list;
        pool->pool_list = pool->pool_list->next;
        spdFree(list_tmp->str_pool);
        spdFree(list_tmp);
    }
}

/* strstring poolΤΤɤȽꤹ롥 */

static SpdBoolean is_string_pool(SpdStringPool *pool, SpdString str)
{
    SpdStringPoolList *list_tmp;
    size_t size;

    /*
     * spdStrPoolCopy()Τ߸ƤФ졤poolNULLɤȽѤߡ
     *
     * if (pool == NULL) return FALSE;
     */

    size = pool->pool_size;

    for (list_tmp = pool->pool_list; list_tmp; list_tmp = list_tmp->next) {
        /* pool <= str < pool + POOL_SIZE */
        if (list_tmp->str_pool <= str &&
            str < list_tmp->str_pool + size) return TRUE;
    }

    return FALSE;
}

/*
 * string poolstr򥳥ԡ롥
 * string poolζ̤­ʤstring poolɲä롥
 */

static SpdString copy_string_pool(SpdStringPool *pool, SpdString str)
{
    SpdString str_tmp;
    size_t length;

    /*
     * spdStrPoolCopy(), spdStrPoolFastCopy()ƤФ졤
     * poolstrNULLɤȽѤߡ
     *
     * if (pool == NULL || str == NULL) return NULL;
     */

    length = strlen(str) + 1;   /* ʸstrĹ('\0'ޤ) */

    /* pool礭­ʤwarningϤstrcpy롥 */
    if (pool->pool_size < length) {
        /* spdWarning("SpdStringPool: Specified too long string.\n"); */
        return (spdStrdup(str));
    }

    /* ̤­ʤstring poolɲä롥 */
    if (pool->space_size < length) append_string_pool(pool);

    /* strstring pool˥ԡ롥 */
    str_tmp = pool->current_pos;
    strcpy(str_tmp, str);

    /* string poolζ̤ӶƬϤ롥 */
    pool->space_size -= length;
    pool->current_pos += length;

#ifdef DEBUG_STRING_POOL
    print_current_state(pool);
#endif

    return str_tmp;
}

/******************************************************************************/

/*
 * string poolstring poolؤΥݥ󥿤֤
 * sizeϡ٤˥󤹤礭ɽ
 */

SpdStringPool *spdStrPoolInit(size_t size)
{
    SpdStringPool *pool;

    pool = (SpdStringPool *)spdMalloc(sizeof(SpdStringPool));
    pool->pool_list = NULL;
    pool->pool_size = (0 < size) ? size : SAPID_STRING_POOL_SIZE;
    pool->space_size = 0;
    pool->current_pos = NULL;

    return pool;
}

/*
 * strstring pool˥ԡơΥݥ󥿤֤
 * strstring poolΤΤǤstr֤
 */

SpdString spdStrPoolCopy(SpdStringPool *pool, SpdString str)
{
    SpdString str_tmp;

    if (pool == NULL) {         /* pool  NULL ʤ NULL ֤ */
        spdWarning("spdStrPoolCopy(): Specified uninitialized string pool.\n");
        return NULL;
    }
    if (str == NULL) {          /* str  NULL ʤ NULL ֤ */
        return NULL;
    }

#ifdef DEBUG_STRING_POOL
    fprintf(stderr, "spdStrPoolCopy(): str = %s\n", str);
#endif

    /* str  pool ΤΤǤ str ֤ */
    if (is_string_pool(pool, str)) return str;

    str_tmp = copy_string_pool(pool, str);
    return str_tmp;
}

/*
 * strstring pool˥ԡơΥݥ󥿤֤
 * strstring poolΤ°Ƥ뤫ݤθϹԤʤ
 * ѼԤstrpool°ƤʤȤݾڤǤˤΤѤ뤳ȡ
 */

SpdString spdStrPoolFastCopy(SpdStringPool *pool, SpdString str)
{
    SpdString str_tmp;

    if (pool == NULL) {         /* pool  NULL ʤ NULL ֤ */
        spdWarning("spdStrPoolFastCopy(): Uninitialized string pool.\n");
        return NULL;
    }
    if (str == NULL) {          /* str  NULL ʤ NULL ֤ */
        return NULL;
    }

#ifdef DEBUG_STRING_POOL
    fprintf(stderr, "spdStrPoolFastCopy(): str = %s\n", str);
#endif

    str_tmp = copy_string_pool(pool, str);
    return str_tmp;
}

/*
 * string poolfree롥
 */

void spdStrPoolFree(SpdStringPool *pool)
{
    if (pool == NULL) return;

#ifdef DEBUG_STRING_POOL
    print_current_state(pool);
#endif

    delete_string_pool(pool);
    spdFree(pool);
}
