/* 
 * Program:             $RCSfile: string.c,v $  $Revision: 30.13 $
 * 
 * Purpose:             String manipulation functions.
 * 
 * Author:              T.Mimura  1993/12/01
 *                      S.Yamamoto  1994/09/01
 * 
 * (C) Copyright:       T.Mimura and S.Yamamoto  1993 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: string.c,v 30.13 2003/11/20 07:35:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/*
 * ʸʣԤ
 */

char *spdStrdup(const char *str)
{
    return (strcpy((char *)spdMalloc(strlen(str) + 1), str));
}

char *spdSafetyStrdup(const char *str)
{
    return ((str != NULL) ? spdStrdup(str) : NULL);
}

/* 
 * ʸfromƬ顤ʸcκǽνиľޤǤ򡤥ɥ쥹to
 * ԡ롥ʸcиʤϡʸfromΤ򥳥ԡ롥
 */

char *spdStrcpyUntilChr(char *to, const char *from, int c)
{
    while (*from != '\0' && *from != c) {
        *to++ = *from++;
    }
    *to = '\0';
    return (to);
}   

/* 
 * ʸstrƬ顤ʸcκǽνиľޤǤ
 * ʸȤʣơΥɥ쥹֤
 * ʸcиʤϡʸstrΤʣ롥
 */

char *spdStrdupUntilChr(const char *str, int c)
{
    char    *cp, *endp;
    int     len;

    endp = strchr(str, c);
    if (endp == NULL) {
        /* Chracter c is not found in string str. Copy whole string. */
        return (spdStrdup(str));
    } else {
        len = endp - str;
        cp = spdMalloc(len + 1);
        strncpy(cp, str, len);
        *(cp + len) = '\0';
        return (cp);
    }
}   

/*
 * strcat().
 */

/* SpdString spdStrcat(SpdString a, SpdString b) */

char *spdStrcat(char *a, char *b)
{
    /* SpdString res = NULL; */
    char    *res = NULL;

    if (b != NULL) {
        if (a == NULL) {
            res = spdStrdup(b);
        } else {
            res = spdRealloc(a, strlen(a) + strlen(b) + 1);
            strcat(res, b);
        }
    } else {
        res = a;
    }

    return (res);
}

/*
 * strcat() and strdup().
 */

char *spdStrcatDup(const char *str1, const char *str2)
{
    int     len;
    char    *cp;
    
    len = strlen(str1) + strlen(str2);
    cp = spdMalloc(len + 1);
    strcpy(cp, str1);
    strcat(cp, str2);
    return (cp);
}

int spdSafetyStrcmp(const char *a, const char *b)
{
    if (a == NULL && b == NULL) {
        return (0);
    } else if (a == NULL) {
        return (-1);
    } else if (b == NULL) {
        return (1);
    }
    return (strcmp(a, b));
}

SpdBoolean spdSafetyStreq(const char *a, const char *b)
{
    return ((a == NULL || b == NULL) ? FALSE : SPD_STREQ(a, b));
}

/*
 * convert an integer to a string.
 */

SpdConstString spdItoa(int x)
{
    static char     buf[SPD_MAX_INT_LEN];

    buf[0] = '\0';
    sprintf(buf, "%d", x);
    return (buf);
}

/*
 * Convert an object id to a string.
 * Added by N.Fukuyasu 2003/08/12.
 * 
 * We assumes that after buf_index's overflow, its value becomes 0.
 */

#define BUF_NUM     32        /* It should be 2 ^ x. */

SpdConstString spdObjIdStr(SpdObjId id)
{
    static char     buf[BUF_NUM][SPD_MAX_INT_LEN];
    static unsigned int buf_index = 0;

    sprintf(buf[buf_index % BUF_NUM], "%d", id);
    return buf[buf_index++ % BUF_NUM];
}
