/* 
 * Program:             $RCSfile: ddbin.c,v $  $Revision: 30.12 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: ddbin.c,v 30.12 2003/06/04 17:01:00 yamamoto Exp yamamoto $";
#endif

#define DDBINMAIN

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "ddbin.h"

#include "attribute.h"
#include "docfile.h"
#include "docfrag.h"
#include "docident.h"
#include "document.h"
#include "make_source.h"
#include "parse.h"
#include "textstring.h"
#include "write_dsdb.h"

static char *target_document_name = SPD_DEFAULT_TARGET_DOCUMENT_NAME;
static SpdConstString program_name;              /* This program name. */

static void parse_option(int argc, char *argv[]);
static void printUsage(SpdConstString command_name);
static void write_sapid_version(void);

void initObjLists(void)
{
    docobj_list = initLinklist();
    docfileobj_list = initLinklist();
    docfragobj_list = initLinklist();
    textstringobj_list = initLinklist();
    attributeobj_list = initLinklist();
    docidentobj_list = initLinklist();
    docidentinfo_list = initLinklist();
    initDocIdentSet();
}

static void freeObjectLists(void)
{
    /* Free Object Lists */
    freeDocumentObjList();
    freeDocFileObjList();
    freeDocFragObjList();
    freeTextstringObjList();
    freeAttributeObjList();
    freeDocIdentObjList();
    freeDocIdentInfoList();

    /* free Docident Set */
    freeDocIdentSet();
}

int main(int argc, char **argv)
{
    SpdModelId DmodelId;
    int i;

    DocObj *docobj;
    int ret_code;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);

    parse_option(argc, argv);
    argc -= optind;
    argv += optind;
    if (argc < 1) {
        spdErrorExit("No input file.\n");
    }

    DmodelId = spdUseModel(&D_model);

    for(i=0;i<argc;i++){
        initObjLists();
        /* create Document Object */
        docobj = addDocumentObj(target_document_name);
        spdMessage("input file: %s.\n",argv[i]);
        if((ret_code = parseDocfile(docobj,argv[i]))==EXIT_SUCCESS){
            write_dsdb(argv[i]);
            write_sapid_version();
            makeSourcecode(docobj);
        }

#ifdef DOC_DEBUG
        printDocumentList();
        printDocFileList();
        printDocFragList();
        printTextstringList();
        printAttributeList();
        printDocIdentList();
#endif /* ifdef DEBUG_DOC */
        freeObjectLists();
    }
    return ret_code;                        /* EXIT_FAILURE or EXIT_SUCCESS. */
}

static void parse_option(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "bd:o:v")) != -1) {
        switch (ch) {
          case 'b':                         /* Binary mode. */
            spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
            break;
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            spdMessage("SDB directory \"%s\".\n", spdGetSDBdir());
            break;
          case 'o':                         /* Document name. */
            target_document_name = optarg;
            spdMessage("Document name \"%s\".\n", target_document_name);
            break;
          case 'v':
            verbose_mode = 1;
            break;
          default:
            printUsage(program_name);
            break;
        }
    }
}

/*
 * Output version information.
 */

static SpdString make_val_str(int value)
{
    static char     val_str[SPD_LINE_LEN + 1];

    sprintf(val_str, "%d", value);
    return (val_str);
}

static void write_sapid_version(void)
{
    SpdInfo     *info;

    info = spdOpenSDBinfo();
    spdSetInfoString(info, SPD_MKSTR(SapidVersion),
                     make_val_str(SapidVersion));
    spdSetInfoString(info, SPD_MKSTR(SapidRevision),
                     make_val_str(SapidRevision));
    spdSetInfoString(info, SPD_MKSTR(SapidPatchLevel),
                     make_val_str(SapidPatchLevel));
    spdSetInfoString(info, SPD_MKSTR(SapidSDBVersion),
                     make_val_str(SapidSDBVersion));
    spdSetInfoString(info, SPD_MKSTR(SapidPIDBVersion),
                     make_val_str(SapidPIDBVersion));
    spdSetInfoString(info, "DDBIN", "TRUE");
    spdSetInfoString(info, "USE_DSDB", "TRUE");
    spdSetInfoString(info, "BINARY_FORMAT_SDB",
                     (spdGetWriteSDBformat() == SAPID_BINARY_FORMAT_SDB ?
                      "TRUE" : "FALSE"));
    spdCloseInfo(info);
}

/*
 * Usage.
 */

static void printUsage(SpdConstString command_name)
{
    const char  *usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [options] inputFiles\n"
        "DESCRIPTION\n"
        "    -d SDBdirectory.\n"
        "    -o DocumentName\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
