/* 
 * Program:             $RCSfile: docfrag.c,v $  $Revision: 30.7 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: docfrag.c,v 30.7 2003/06/04 17:01:00 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdUtil.h>
#include <Sapid/Dmodel.h>
#include "ddbin.h"
#include "docfrag.h"

DocFragObj *addDocFragObj(const int sort, const char *name, int offset)
{
    DocFragObj  *new;
    static int  docfragobj_num = SPD_OBJ_ID_OFFSET(SPD_CID_DOCFRAG);

    new = (DocFragObj *)spdCalloc(1, sizeof(DocFragObj));
    new->id = docfragobj_num++;
    new->sort = sort;
    new->offset = offset;
    new->name = spdStrdup(name);
    new->docfrag_list = initLinklist();
    new->attribute_list = initLinklist();
    new->docident_list = initLinklist();
    new->srctext = NULL;
    insertLinklistFirst(docfragobj_list, new);
    return new;
}

void freeDocFragObjList(void)
{
    Linklist lst = docfragobj_list;
    Node node = getFirstNode(lst);

    while(node){
        DocFragObj *p = (DocFragObj *)getNodeData(node);

        spdFree(p->name);
        freeLinklist(p->docfrag_list);
        freeLinklist(p->attribute_list);
        freeLinklist(p->docident_list);
        /*freeLinklist(p->textstring_list);*/
        if(p->srctext)
            spdFree(p->srctext);

        spdFree(p);
        node=getNextNode(node);
    }
    freeLinklist(lst);
}

#ifdef DOC_DEBUG

void printDocFragList(void)
{
    Node node = getFirstNode(docfragobj_list);
    Node node2;

    while(node){
        DocFragObj *p = (DocFragObj *)getNodeData(node);
        printf("DocFragObj:\n");
        printf("  id:%d\n",p->id);
        switch(p->sort){
          case DAPID_DOCFRAG_TEXTFRAG:
            printf("  type:TEXTFRAG\n");
            break;
          case DAPID_DOCFRAG_BASICELEM:
            printf("  type:BASICELEM\n");
            break;
          case DAPID_DOCFRAG_DEFINEELEM:
            printf("  type:DEFINEELEM\n");
            break;
          case DAPID_DOCFRAG_IDENTREF:
            printf("  type:IDENTREF\n");
            break;
          case DAPID_DOCFRAG_SRCCODE:
            printf("  type:SRCCODE\n");
            break;
        }
        printf("  name:%s\n",p->name);
        node=getNextNode(node);

        printf("  docfragid:");
        node2 = getFirstNode(p->docfrag_list);
        while(node2){
            DocFragObj *p2 = (DocFragObj *)getNodeData(node2);
            printf("%d ",p2->id);
            node2=getNextNode(node2);
        }
        printf("\n");
        printf("  textstringid:");
        node2 = getFirstNode(p->textstring_list);
        while(node2){
            TextstringObj *p2 = (TextstringObj *)getNodeData(node2);
            printf("%d ",p2->id);
            node2=getNextNode(node2);
        }
        printf("\n");
    }
}

#endif /* DOC_DEBUG */
