/* 
 * Program:             $RCSfile: docident.c,v $  $Revision: 30.12 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: docident.c,v 30.12 2003/07/06 10:11:30 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/spdSet2.h>
#include "ddbin.h"
#include "docident.h"

SpdSet2 docident_set;

void initDocIdentSet(void)
{
    docident_set = spdCreateSet2(SPDSET_COPY_STRING_KEY, SPDSET_POINTER_VALUE);
}

void freeDocIdentSet(void)
{
    spdFreeSet2(docident_set);
}

DocIdentInfo *searchDocIdent(char * name)
{
    SpdElem *elem = spdSearchElemS(docident_set, name);

    return elem ? (DocIdentInfo *)spdSetGetValueP(elem->value) : NULL;
}

DocIdentInfo *addDocIdentInfo(const int sort, const char *name)
{
    DocIdentInfo    *new;
    static int      docidentobj_num = SPD_OBJ_ID_OFFSET(SPD_CID_DOCIDENT);

    new = (DocIdentInfo *)spdCalloc(1, sizeof(DocIdentInfo));
    new->id = docidentobj_num++;
    new->sort = sort;
    new->name = spdStrdup(name);
    insertLinklistFirst(docidentinfo_list, new);

    spdAddElemSP(docident_set, name, new);

    return new;
}

void freeDocIdentInfoList(void)
{
    Linklist lst = docidentinfo_list;
    Node node = getFirstNode(lst);

    while(node){
        DocIdentInfo *p = (DocIdentInfo *)getNodeData(node);
        spdFree(p->name);
        spdFree(p);
        node=getNextNode(node);
    }
    freeLinklist(lst);
}

DocIdentObj *addDocIdentObj(DocIdentInfo *obj, int offset)
{
    DocIdentObj *new;

    new = (DocIdentObj *)spdCalloc(1,sizeof(DocIdentObj));
    new->docident = obj;
    new->offset = offset;
    insertLinklistFirst(docidentobj_list, new);

    return new;
}

void freeDocIdentObjList(void)
{
    Linklist lst = docidentobj_list;
    Node node = getFirstNode(lst);

    while(node){
        DocIdentObj *p = (DocIdentObj *)getNodeData(node);
        spdFree(p);
        node=getNextNode(node);
    }
    freeLinklist(lst);
}

/*
 * ؿμ̻Ҥγ̤
 */

char *removeParentheses(char *text)
{
    static char tmp[SPD_STRING_LEN + 1];
    int         i;

    for (i = 0; i < strlen(text); i++) {
        if (text[i] == '(') {
            break;
        }
        tmp[i] = text[i];
    }
    tmp[i] = '\0';
    return (tmp);
}

#ifdef DOC_DEBUG

void printDocIdentList(void)
{
    Node node = getFirstNode(docidentobj_list);
    Node node2;

    while(node){
        DocIdentObj *p = (DocIdentObj *)getNodeData(node);
        printf("DocIdentObj:\n");
        printf("  id:%d\n",p->id);
        switch(p->sort){
          case DID_VARIABLE:
            printf("  sort:VARIABLE\n");
            break;
          case DID_FUNCTION:
            printf("  sort:FUNCTION\n");
            break;
          case DID_MEMBER:
            printf("  sort:MEMBER\n");
            break;
          case DID_ENUMERATOR:
            printf("  sort:ENUMERATOR\n");
            break;
          case DID_TYPENAME:
            printf("  sort:TYPENAME\n");
            break;
          case DID_TAG:
            printf("  sort:TAG\n");
            break;
          default:
            printf("  sort:UNKNOWN\n");
            break;
        }
        printf("  name:%s\n",p->name);
        node=getNextNode(node);
        printf("\n");
    }
}

#endif /* DOC_DEBUG */
