/* 
 * Program:             $RCSfile: parse.c,v $  $Revision: 30.11 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: parse.c,v 30.11 2003/06/22 08:42:48 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include "ddbin.h"
#include "parse.h"
#include "docfile.h"
#include "docfrag.h"
#include "linklist.h"
#include "textstring.h"
#include "attribute.h"
#include "docident.h"

#include <Sapid/Sapid.h>
#include <Sapid/SpdUtil.h>
#include <Sapid/Dmodel.h>

#include "../xmlparser/xmlparser.h"
#include "../xmlparser/version.h"

DocFileObj *cur_docfileobj;
Linklist elem_stack;

static DocFragObj *getElemStackTop(void)
{
    return (DocFragObj *)getNodeData(getFirstNode(elem_stack));
}

static void pushElemStack(DocFragObj *elem_obj)
{
    insertLinklistFirst(elem_stack, elem_obj);
}

static DocFragObj *popElemStack(void)
{
    return (DocFragObj *)removeLinklistFirst(elem_stack);
}

static void processStartElement(XBit bit)
{
    char *name = (char *)bit->element_definition->name;
    char *defname;
    int  defsort = DID_UNKNOWN, offset;
    DocFragObj *docfragobj;
    Attribute attr;
    AttributeObj *attrobj;
    TextstringObj *textstringobj;

    if(SPD_STREQ(name,DEFINE_TAG)){
        DocIdentInfo *docidentinfo;
        DocIdentObj *docidentobj;
        /* create Define element Object */
        docfragobj = addDocFragObj(DOCFRAG_DEFELEM, name,
                                   bit->byte_offset);
        attr = bit->attributes;
        while(attr){
            if(SPD_STREQ(attr->definition->name,DEFELEM_ATTRNAME)){
                if(SPD_STREQ(attr->value,"")){
                    spdErrorExit("%d: defelem has name attr, but has no value.\n");
                }
                if((docidentinfo = searchDocIdent(attr->value))!=NULL){
                    spdWarning("%d: %s is previously defined.\n"
                               ,bit->byte_offset
                               ,attr->name_offset
                               );
                }else{
                    defname = attr->value;
                    offset = attr->value_offset;
                }
            }else if(SPD_STREQ(attr->definition->name,DEFELEM_ATTRSORT)){
                if(SPD_STREQ(attr->value,"")){
                    spdErrorExit("%d: defelem has sort attr, but has no value.\n");
                }
                if((docidentinfo = searchDocIdent(attr->value))!=NULL){
                    spdWarning("%d: %s is previously defined.\n"
                               ,bit->byte_offset
                               ,attr->name_offset
                               );
                }else{
                    /* Ȥꤢؿޥե */
                    if(SPD_STREQ(attr->value, "function")){
                        defsort = DID_FUNCTION;
                    }else if(SPD_STREQ(attr->value, "typedef")){
                        defsort = DID_TYPENAME;
                    }else if(SPD_STREQ(attr->value, "macro")){
                        defsort = DID_MACRONAME;
                    }else if(SPD_STREQ(attr->value, "file")){
                        defsort = DID_FILE;
                    }
                }
            }
            attr=attr->next;
        }
        if(strcmp(defname, "")){
            docidentinfo = addDocIdentInfo(defsort,defname);
            docidentobj = addDocIdentObj(docidentinfo, offset);
            insertLinklistFirst(docfragobj->docident_list, docidentobj);
        }
    }else if(SPD_STREQ(name,SRCCODE_TAG)){
        /* create Srccode element Object */
        docfragobj = addDocFragObj(DOCFRAG_SRCELEM, name,
                                   bit->byte_offset);
    }else if(SPD_STREQ(name,DOCIDENT_TAG)){
        /* create docident element Object */
        docfragobj = addDocFragObj(DOCFRAG_IDENTELEM, name,
                                   bit->byte_offset);
    }else if(SPD_STREQ(name,MACRODEF_TAG)){
        docfragobj = addDocFragObj(DOCFRAG_MACROELEM, name,
                                   bit->byte_offset);
    }else{
        /* create Basic element Object */
        docfragobj = addDocFragObj(DOCFRAG_BASICELEM, name,
                                   bit->byte_offset);
        /* Attribute */
        attr = bit->attributes;
        while(attr){
            /*attrobj = addAttributeObj(attr->definition->name, bit->byte_offset);*/
            attrobj = addAttributeObj(attr->definition->name, attr->name_offset);
            textstringobj = addTextstringObj(attr->value,attr->value_offset);
            attrobj->textstringObj = textstringobj;
            insertLinklistFirst(docfragobj->attribute_list, attrobj);
            attr = attr->next;
        }
    }

    if(isEmptyLinklist(elem_stack)){        /*ǽElementξ*/
        DocFileObj *parent = cur_docfileobj;
        insertLinklistFirst(parent->docfrag_list,docfragobj);
    }else{
        DocFragObj *parent = getElemStackTop();
        insertLinklistFirst(parent->docfrag_list,docfragobj);
    }
    pushElemStack(docfragobj);

    /* !! ToDo: LINK ELEMENTβϤ̤ݡȡ */
}

static void processEndElement(XBit bit)
{
    DocFragObj *docfragobj;
    docfragobj = popElemStack();
    docfragobj->length = bit->end_byte_offset - docfragobj->offset;
}

static void processDocFile(DocObj *parent, char *filename)
{
    DocFileObj *docfileobj = addDocFileObj(filename);
    insertLinklistFirst(parent->docfile_list, docfileobj);
    cur_docfileobj = docfileobj;
}

/* 
 * !!
 * 1999/07/04AR¤顤ҤȤĤʸSPD_LINE_LENޤǤǤ뤳Ȥ顤
 * ƥȤҤȤޤfragmentȤưΤϸǤϸ
 * 
 * SDBǲԤ쥳ɤζڤ˻ȤƤ뤳ȤȤ顤ʸ˲Ԥ
 * 뤳ȤȤʤ롥Ԥ򥨥פ?
 * ǤϡԤΤơfragmentʬ뤳Ȥˤ롥
 * Ԥ̵ͭξϼ롥
 * 
 * 2000/02/28:
 * Хʥ⡼ɤǤϡԥɤ⤽Τޤޥǡ١롥
 * ޤAR5ʤ顤SPD_LINE_LEN̵
 * Ȥꤢϡñ̤textstringȤ롥
 */

static void processTextdata(char *text, int offset,int length)
{
    char *p,*sp;
    DocFragObj *parent;

    parent = getElemStackTop();

    if (parent->sort & DOCFRAG_SRCELEM){
        /* parent is SRCELEM */
        /* preserve srctext for makeSourcecode() */
        parent->srctext = spdStrdup(text);
        parent->srcoffset = offset;
    }else if(parent->sort & DOCFRAG_IDENTELEM){
        /* parent is DOCIDENT */
        DocIdentInfo    *docidentinfo;
        DocIdentObj     *docidentobj;
        char            *tmp;

        /*
         * ؿˤĤƤϡ"<docident>printf()</docident>"
         * Ƚ񤯤Ȥ¿ᡤ"()"ʬȤǡ
         * Ȥμͳ٤򹭤᤿(2001/01/30) 
         */
        tmp = removeParentheses(text);
        if ((docidentinfo = searchDocIdent(tmp)) == NULL) {
            spdWarning("%d: %s is defined IMPLICITLY.\n", offset, text);
            docidentinfo = addDocIdentInfo(DID_UNKNOWN, text);
        }
        docidentobj = addDocIdentObj(docidentinfo, offset);
        insertLinklistFirst(parent->docfrag_list, docidentobj);

        /* !! ToDo: docidentμ(function, type, ...)ǡ١롥 */
    }else{
        /* parent is DOCELEM */
        TextstringObj *textstringobj;
        DocFragObj *docfragobj;

        /* make TEXTFRAG object */
        if (parent->sort & DOCFRAG_DOCELEM){
            docfragobj = addDocFragObj(DOCFRAG_TEXTFRAG, "", offset);
            docfragobj->length = strlen(text);
            insertLinklistFirst(parent->docfrag_list,docfragobj);
            parent = docfragobj;
        }

        /* make TEXTSTRING object */
        p = sp = text;
        while(*p != '\0'){
            if(*p=='\n'){
                char temp = *++p;
                *p = '\0';
                if(spdGetWriteSDBformat() == SAPID_TEXT_FORMAT_SDB){
                    /*ƥȥեޥåȤΤȤϲԤä()*/
                    *(p-1) = '\0';
                }
                textstringobj = addTextstringObj(sp,offset);
                insertLinklistFirst(parent->docfrag_list,textstringobj);
                *p = temp;
                offset += (p-sp);
                sp = p;
            }
            p++;
        }
        if(*sp != '\0'){
            textstringobj = addTextstringObj(sp,offset);
            insertLinklistFirst(parent->docfrag_list,textstringobj);
        }

        /*쥢르ꥺ*/
        /*
           p = sp = text;
           while(sp){
           if(*p=='\0'){
           textstringobj = addTextstringObj(sp,offset);
           insertLinklistFirst(parent->docfrag_list,textstringobj);
           sp=NULL;
           }else if(*p=='\n'){
           *p='\0';
           textstringobj = addTextstringObj(sp,offset);
           insertLinklistFirst(parent->docfrag_list,textstringobj);
           offset += (p-sp);
           sp=p+1;
           }
           p++;
           }
           */
    }
}

/* return: error 0 */

int parseDocfile(DocObj *parent, char *filename)
{
    Parser p;
    Entity ent = 0;
    InputSource source = 0;
    CharacterEncoding encoding = CE_unknown;

    if(verbose_mode)
        fprintf(stderr, "%s\n", rxp_version_string);

    elem_stack = initLinklist();

    ent = NewExternalEntity(0, 0, filename, 0, 0);
    if(ent)
        source = EntityOpen(ent);
    if(!source)
        return 0;

    /* create xmlparser */
    p = NewParser();
    ParserSetFlag(p, MergePCData, 1);

    if(ParserPush(p, source) == -1){
        ParserPerror(p, &p->xbit);
        return 0;
    }

  #ifdef DEBUG
    SetFileEncoding(Stdout, source->entity->encoding);
  #endif

    if(verbose_mode)
        fprintf(stderr, "Input encoding %s, output encoding %s\n",
                CharacterEncodingNameAndByteOrder[source->entity->encoding],
                CharacterEncodingNameAndByteOrder[encoding]);

    /* create Document File Object */
    processDocFile(parent, filename);

    while(1){
        XBit bit;
        bit = ReadXBit(p);
        switch(bit->type){
          case XBIT_eof:
            FreeXBit(bit);
            if(ent)
                FreeEntity(ent);
            return EXIT_SUCCESS;
          case XBIT_start:
            processStartElement(bit);
            break;
          case XBIT_empty:
            /* nob support empty tag */
            break;
          case XBIT_end:
            processEndElement(bit);
            break;
          case XBIT_pcdata:
            /* create Text Object */
            /*text node element βǤʤ
              parserxml-parserǤ뤿ᡥxmlǤϥɥȤľˤ
              ҤȤĤΥդ줿ǤƤʤ*/
            processTextdata((char *)bit->pcdata_chars,bit->byte_offset, bit->end_byte_offset - bit->byte_offset);
            break;
          case XBIT_cdsect:
            /* create Text Object */
            processTextdata((char *)bit->cdsect_chars,bit->byte_offset, bit->end_byte_offset - bit->byte_offset);
            break;
          case XBIT_comment:
            /* not support comment */
          case XBIT_pi:
            /* not support PI */
          case XBIT_dtd:
            /* not support DTD */
            continue;
            break;
          case XBIT_warning:
            /* never returned this value. but the bit should passed to
               the warning callback function.*/
          case XBIT_error:
            ParserPerror(p, bit);
          default:
            FreeXBit(bit);
            freeLinklist(elem_stack);
            return EXIT_FAILURE;
        }
    }
}
