/* 
 * Program:             $RCSfile: write_dsdb.c,v $  $Revision: 30.10 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: write_dsdb.c,v 30.10 2002/02/03 19:00:46 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "ddbin.h"
#include "write_dsdb.h"

static int write_dsdb_document(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node = getFirstNode(docobj_list);

    while(node){
        DocObj *docobj = (DocObj *)getNodeData(node);
        count++;

        obj = spdCreateTempObj(SPD_CID_DOCUMENT, docobj->id);
        spdSetTempObjAttrStrValByName(&obj,"name",docobj->name);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_docfile(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node = getFirstNode(docfileobj_list);

    while(node){
        DocFileObj *docfileobj = (DocFileObj *)getNodeData(node);
        count++;

        obj = spdCreateTempObj(SPD_CID_DOCFILE, docfileobj->id);
        spdSetTempObjAttrStrValByName(&obj,"name",docfileobj->name);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_docfrag(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node = getFirstNode(docfragobj_list);

    while(node){
        DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
        count++;

        obj = spdCreateTempObj(SPD_CID_DOCFRAG, docfragobj->id);
        spdSetTempObjAttrIntValByName(&obj,"sort",docfragobj->sort);
        spdSetTempObjAttrIntValByName(&obj,"length",docfragobj->length);
        spdSetTempObjAttrStrValByName(&obj,"name",docfragobj->name);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_textstring(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node = getFirstNode(textstringobj_list);

    while(node){
        TextstringObj *textstringobj = (TextstringObj *)getNodeData(node);
        count++;

        obj = spdCreateTempObj(SPD_CID_TEXTSTRING, textstringobj->id);
        spdSetTempObjAttrIntValByName(&obj,"length",strlen(textstringobj->text));
        spdSetTempObjAttrStrValByName(&obj,"text",textstringobj->text);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_attribute(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node = getFirstNode(attributeobj_list);

    while(node){
        AttributeObj *attributeobj = (AttributeObj *)getNodeData(node);
        count++;
        obj = spdCreateTempObj(SPD_CID_ATTRIBUTE, attributeobj->id);
        spdSetTempObjAttrIntValByName(&obj,"length",strlen(attributeobj->name));
        spdSetTempObjAttrStrValByName(&obj,"name",attributeobj->name);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_docident(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node = getFirstNode(docidentinfo_list);

    while(node){
        DocIdentInfo *docidentinfo = (DocIdentInfo *)getNodeData(node);
        count++;

        obj = spdCreateTempObj(SPD_CID_DOCIDENT, docidentinfo->id);
        spdSetTempObjAttrIntValByName(&obj,"sort",docidentinfo->sort);
        spdSetTempObjAttrIntValByName(&obj,"length",strlen(docidentinfo->name));
        spdSetTempObjAttrStrValByName(&obj,"name",docidentinfo->name);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_doc_docfile(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;
    DocObj *docobj;
    DocFileObj *docfileobj;

    node = getFirstNode(docobj_list);
    while(node){
        docobj = (DocObj *)getNodeData(node);
        node2 = getFirstNode(docobj->docfile_list);
        while(node2){
            count++;
            docfileobj = (DocFileObj *)getNodeData(node2);

            obj = spdCreateTempObj(SPD_RID_DOC_DOCFILE, 0);
            spdSetTempObjAttrIntValByName(&obj,"document_id",docobj->id);
            spdSetTempObjAttrIntValByName(&obj,"docfile_id",docfileobj->id);
            spdWriteTempObj(fp,&obj);
            spdDeleteTempObj(&obj);
            node2 = getNextNode(node2);
        }
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_docfile_docfrag(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;
    DocFileObj *docfileobj;
    DocFragObj *docfragobj;

    node = getFirstNode(docfileobj_list);
    while(node){
        docfileobj = (DocFileObj *)getNodeData(node);
        node2 = getFirstNode(docfileobj->docfrag_list);
        while(node2){
            count++;
            docfragobj = (DocFragObj *)getNodeData(node2);

            obj = spdCreateTempObj(SPD_RID_DOCFILE_DOCFRAG, 0);
            spdSetTempObjAttrIntValByName(&obj,"docfile_id",docfileobj->id);
            spdSetTempObjAttrIntValByName(&obj,"docfrag_id",docfragobj->id);
            spdSetTempObjAttrIntValByName(&obj,"offset",docfragobj->offset);
            spdWriteTempObj(fp,&obj);
            spdDeleteTempObj(&obj);
            node2 = getNextNode(node2);
        }
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_docelem_docfrag(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;
    DocFragObj *childfragobj;

    node = getFirstNode(docfragobj_list);
    while(node){
        DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
        if(docfragobj->sort & DOCFRAG_DOCELEM){
            node2 = getFirstNode(docfragobj->docfrag_list);
            while(node2){
                childfragobj = (DocFragObj *)getNodeData(node2);
                count++;

                obj = spdCreateTempObj(SPD_RID_DOCELEM_DOCFRAG, 0);
                spdSetTempObjAttrIntValByName(&obj,"docelem_id",docfragobj->id);
                spdSetTempObjAttrIntValByName(&obj,"docfrag_id",childfragobj->id);
                spdSetTempObjAttrIntValByName(&obj,"offset",childfragobj->offset);
                spdWriteTempObj(fp,&obj);
                spdDeleteTempObj(&obj);
                node2 = getNextNode(node2);
            }
        }
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_docelem_attribute(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;

    node = getFirstNode(docfragobj_list);
    while(node){
        DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
        node2 = getFirstNode(docfragobj->attribute_list);
        while(node2){
            AttributeObj *attributeobj;
            attributeobj = (AttributeObj *)getNodeData(node2);
            count++;

            obj = spdCreateTempObj(SPD_RID_DOCELEM_ATTRIBUTE, 0);
            spdSetTempObjAttrIntValByName(&obj,"docelem_id",docfragobj->id);
            spdSetTempObjAttrIntValByName(&obj,"attribute_id",attributeobj->id);
            spdSetTempObjAttrIntValByName(&obj,"offset",attributeobj->offset);
            spdWriteTempObj(fp,&obj);
            spdDeleteTempObj(&obj);
            node2 = getNextNode(node2);
        }
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_attribute_value(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node;

    node = getFirstNode(attributeobj_list);
    while(node){
        AttributeObj *attributeobj = (AttributeObj *)getNodeData(node);
        count++;

        obj = spdCreateTempObj(SPD_RID_ATTRIBUTE_VALUE, 0);
        spdSetTempObjAttrIntValByName(&obj,"attribute_id",attributeobj->id);
        spdSetTempObjAttrIntValByName(&obj,"value_id",attributeobj->textstringObj->id);
        spdSetTempObjAttrIntValByName(&obj,"offset",attributeobj->textstringObj->offset);
        spdWriteTempObj(fp,&obj);
        spdDeleteTempObj(&obj);
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_textfrag_textstring(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;
    TextstringObj *textstringobj;

    node = getFirstNode(docfragobj_list);
    while(node){
        DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
        /* process TEXTFRAG objects. */
        if(docfragobj->sort & DOCFRAG_TEXTFRAG){
            node2 = getFirstNode(docfragobj->docfrag_list); /* get the list of
                                                               textstring object*/
            while(node2){
                textstringobj = (TextstringObj *)getNodeData(node2);
                count++;

                obj = spdCreateTempObj(SPD_RID_TEXTFRAG_TEXTSTRING, 0);
                spdSetTempObjAttrIntValByName(&obj,"textfrag_id",docfragobj->id);
                spdSetTempObjAttrIntValByName(&obj,"textstring_id",textstringobj->id);
                spdSetTempObjAttrIntValByName(&obj,"offset",textstringobj->offset);
                spdWriteTempObj(fp,&obj);
                spdDeleteTempObj(&obj);
                node2 = getNextNode(node2);
            }
        }
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_identelem_docident(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;
    DocIdentObj *docidentobj;

    node = getFirstNode(docfragobj_list);
    while(node){
        DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
        /* process DOCIDENT objects. */
        if(docfragobj->sort & DOCFRAG_IDENTELEM){
            node2 = getFirstNode(docfragobj->docfrag_list);
            if(node2){
                docidentobj = (DocIdentObj *)getNodeData(node2);
                count++;

                obj = spdCreateTempObj(SPD_RID_IDENTELEM_DOCIDENT, 0);
                spdSetTempObjAttrIntValByName(&obj,"identelem_id",docfragobj->id);
                spdSetTempObjAttrIntValByName(&obj,"docident_id",docidentobj->docident->id);
                spdSetTempObjAttrIntValByName(&obj,"offset",docidentobj->offset);
                spdWriteTempObj(fp,&obj);
                spdDeleteTempObj(&obj);
                /* get A docident object.
                   this is a list but contains only one docident object. */
                /*node2 = getNextNode(node2);*/
            }
        }
        node = getNextNode(node);
    }
    return count;
}

static int write_dsdb_defelem_docident(FILE *fp)
{
    SpdDBObj obj;
    int count = 0;
    Node node,node2;
    DocIdentObj *docidentobj;

    node = getFirstNode(docfragobj_list);
    while(node){
        DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
        /* process DEFELEM objects. */
        if(docfragobj->sort & DOCFRAG_DEFELEM){
            node2 = getFirstNode(docfragobj->docident_list);
            if(node2){
                docidentobj = (DocIdentObj *)getNodeData(node2);
                count++;

                obj = spdCreateTempObj(SPD_RID_DEFELEM_DOCIDENT, 0);
                spdSetTempObjAttrIntValByName(&obj,"defelem_id",docfragobj->id);
                spdSetTempObjAttrIntValByName(&obj,"docident_id",docidentobj->docident->id);
                spdSetTempObjAttrIntValByName(&obj,"offset",docidentobj->offset);
                spdWriteTempObj(fp,&obj);
                spdDeleteTempObj(&obj);
                node2 = getNextNode(node2);
            }
        }
        node = getNextNode(node);
    }
    return count;
}

/* 
 * static int write_dsdb_srcelem_textstring(FILE *fp)
 * {
 *     int count = 0;
 *     Node node,node2;
 *     TextstringObj *textstringobj;
 *     node = getFirstNode(docfragobj_list);
 *     while(node){
 *         DocFragObj *docfragobj = (DocFragObj *)getNodeData(node);
 *         if(docfragobj->sort & DAPID_DOCFRAG_SRCCODE){
 *             // DOCFRAG_SRCCODE and DOCFRAG_ELEMENT are exclusive. 
 *             node2 = getFirstNode(docfragobj->textstring_list);
 *             while(node2){
 *                 textstringobj = (TextstringObj *)getNodeData(node2);
 *                 write_srcelem_textstring_sub(fp,                   
 *                                              docfragobj->id,       
 *                                              textstringobj->id,
 *                                              textstringobj->offset);    
 *                 count++;
 *                 node2 = getNextNode(node2);
 *             }
 *         }
 *         node = getNextNode(node);
 *     }
 *     return count;
 * }
 */

void write_dsdb(char *filename)
{
    int count;
    FILE *fp_db = spdOpenDSDB(filename,SAPID_O_WRITE);
    SpdInfo *fp_info = spdOpenDSDBinfo(filename);

    /** classes **/
    count = write_dsdb_document(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_CID_DOCUMENT), count);

    count = write_dsdb_docfile(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_CID_DOCFILE), count);

    count = write_dsdb_docfrag(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_CID_DOCFRAG), count);

    count = write_dsdb_textstring(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_CID_TEXTSTRING), count);

    count = write_dsdb_attribute(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_CID_ATTRIBUTE), count);

    count = write_dsdb_docident(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_CID_DOCIDENT), count);

    /** relations **/
    count = write_dsdb_doc_docfile(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_DOC_DOCFILE), count);

    count = write_dsdb_docfile_docfrag(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_DOCFILE_DOCFRAG), count);

    count = write_dsdb_docelem_docfrag(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_DOCELEM_DOCFRAG), count);

    count = write_dsdb_docelem_attribute(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_DOCELEM_ATTRIBUTE), count);

    count = write_dsdb_attribute_value(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_ATTRIBUTE_VALUE), count);

    count = write_dsdb_textfrag_textstring(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_TEXTFRAG_TEXTSTRING), count);

    count = write_dsdb_defelem_docident(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_DEFELEM_DOCIDENT), count);

    count = write_dsdb_identelem_docident(fp_db);
    spdSetInfoInt(fp_info, (SpdString)spdGetClassNameById(SPD_RID_IDENTELEM_DOCIDENT), count);

    /* 
     * count = write_dsdb_srcelem_textstring(fp_db);
     * spdSetInfoInt(fp_info, spdGetClassNameById(DAPID_REL_SRCELEM_TEXTSTRING),
     *               count);
     */

    if (spdGetWriteSDBformat() == SAPID_BINARY_FORMAT_SDB) {
        spdSetInfoString(fp_info, "BINARY_FORMAT_SDB", "TRUE");
    } else {
        spdSetInfoString(fp_info, "BINARY_FORMAT_SDB", "FALSE");
    }
    spdCloseInfo(fp_info);
}
