/* 
 * Program:             $RCSfile: dsdb.c,v $  $Revision: 30.14 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: dsdb.c,v 30.14 2003/07/08 05:48:46 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelAR.h>
#include "dsdb.h"

/* #define SPD_DUMP_READ */

void push_modification_table(SpdSet set, SpdId old_id, SpdId new_id)
{
    ModifyField     field;

    field.old_id = old_id;
    field.new_id = new_id;

    /* 
     * old_id򥭡Ȥơfield򽸹ɲä롥
     * spdAddElem()filedΥԡ򽸹ɲä롥
     */
    spdAddElem(set, old_id, &field);
}

SpdObjId get_modified_id(SpdSet set, SpdId old_id)
{
    ModifyField *element;

    if ((element=(ModifyField *)spdSearchElem(set, old_id)) != NULL) {
        return element->new_id;
    }else{
        return SPD_NON_ID;
    }
}

static int spdLoadDSDBsub(char *filename, SpdDBId dbId,
                          SpdObjId (*classfunc)(const char *dfilename,
                                                SpdObjId obj_id,
                                                SpdObjId old_id),
                          void (*relfunc)(const char *dfilename,
                                          SpdClassId class_id,
                                          SpdObjId old_p_id, SpdObjId old_c_id,
                                          SpdObjId new_p_id, SpdObjId new_c_id),
                          SpdClassId first_class_id, SpdClassId last_class_id,
                          SpdClassId first_rel_id, SpdClassId last_rel_id)
{
    FILE *fp;
    SpdInfo *info;
    SpdDBObj *db_obj;
    SpdObjId old_id, new_id, obj_id;
    SpdObjId p_id,c_id,new_p_id, new_c_id;
    SpdSet modify_set;
    SpdClassId class_id;
    int obj_num;

#ifdef SPD_DUMP_READ
    spdMessage("spdLoadDSDBsub(): filename \"%s\".\n", filename);
#endif

    modify_set = spdCreateSet(sizeof(ModifyField));

    fp = spdOpenDSDB_AR5(filename, SAPID_O_READ);
    info = spdOpenDSDBinfo(filename);
    spdSetReadSDBformat(SPD_STREQ(spdGetInfoString(info, "BINARY_FORMAT_SDB"),
                                    "TRUE") ?
                        SAPID_BINARY_FORMAT_SDB :
                        SAPID_TEXT_FORMAT_SDB);

    /* Τν */
    for (class_id = first_class_id;
         class_id <= last_class_id; class_id++) {
        obj_num = spdGetInfoInt(info, spdGetClassNameById(class_id), 0);
        if (obj_num == 0) {
            continue;
        }

        while (obj_num--) {
            db_obj = (SpdDBObj *)spdMalloc(sizeof(SpdDBObj));
            *db_obj = spddb_CreateDBObj(class_id, SPD_NON_ID);
            spddb_ReadTuple(fp, class_id, db_obj);
            if (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB &&
                fgetc(fp) != '\n') {
                spdFclose(fp);
                spdErrorExit("Invalid SDB: \"%s\".\n", filename);
            }
            /* Regist */
            old_id = db_obj->id;
            db_obj->id = SPD_NON_ID;
            obj_id = spddb_RegisterDBObj(db_obj, dbId);

            new_id = (*classfunc)(filename, obj_id, old_id);
            if(new_id != SPD_NON_ID){
                push_modification_table(modify_set, old_id, new_id);
            }
        }
    }

    /* Ϣν*/
    for (class_id = first_rel_id;
         class_id <= last_rel_id; class_id++) {
        obj_num = spdGetInfoInt(info, spdGetClassNameById(class_id), 0);
        if (obj_num == 0) {
            continue;
        }
        while (obj_num--) {
            SpdObjId relId;

            relId = spdCreateObj(class_id, SPD_NON_ID);
            spddb_ReadTuple(fp, class_id, spddb_GetDBObjById(relId));
            if (spdGetReadSDBformat() == SAPID_TEXT_FORMAT_SDB &&
                fgetc(fp) != '\n') {
                spdFclose(fp);
                spdErrorExit("Invalid SDB: \"%s\".\n", filename);
            }

            p_id = spdGetAttrIntVal(relId, SAPIDDB_ATTR_P_ID);
            c_id = spdGetAttrIntVal(relId, SAPIDDB_ATTR_C_ID);
            new_p_id = get_modified_id(modify_set, p_id);
            new_c_id = get_modified_id(modify_set, c_id);

            /* Ϣλؤ褬̵Ͽʤ*/
            if (new_c_id == SPD_NON_ID || new_p_id == SPD_NON_ID){
                /* Delete Obj */
                spdDeleteObj(relId);
                if(relfunc)
                    (*relfunc)(filename, spdGetClassId(relId),
                               p_id, c_id, new_p_id, new_c_id); /* CALLBACK */
            }else{
                /* Regist */
                spdSetAttrIntVal(relId, SAPIDDB_ATTR_P_ID, new_p_id);
                spdSetAttrIntVal(relId, SAPIDDB_ATTR_C_ID, new_c_id);
                if(relfunc)
                    (*relfunc)(filename, spdGetClassId(relId),
                               p_id, c_id, new_p_id, new_c_id); /* CALLBACK */
            }
        }
    }
    spdCloseInfo(info);
    spdFclose(fp);

    spdFreeSet(modify_set);

    return 0;
}

/* DmodelAR.c٤? */

SpdObjId spdLoadDSDB(char *filename, SpdDBId dbId,
                     SpdObjId (*classfunc)(const char *dfilename,
                                           SpdObjId obj_id,
                                           SpdObjId old_id),
                     void (*relfunc)(const char *dfilename,
                                     SpdClassId class_id,
                                     SpdObjId old_p_id, SpdObjId old_c_id,
                                     SpdObjId new_p_id, SpdObjId new_c_id))
{
    return spdLoadDSDBsub(filename, dbId, classfunc, relfunc,
                          SPD_FIRST_CID_D_MODEL, SPD_LAST_CID_D_MODEL,
                          SPD_FIRST_RID_D_MODEL, SPD_LAST_RID_D_MODEL);
}

/* ImodelAR.c٤? */

SpdObjId spdLoadISDB(char *filename, SpdDBId dbId,
                     SpdObjId (*classfunc)(const char *dfilename,
                                           SpdObjId obj_id,
                                           SpdObjId old_id),
                     void (*relfunc)(const char *dfilename,
                                     SpdClassId class_id,
                                     SpdObjId old_p_id, SpdObjId old_c_id,
                                     SpdObjId new_p_id, SpdObjId new_c_id))
{
    return spdLoadDSDBsub(filename, dbId, classfunc, relfunc,
                          SPD_FIRST_CID_I_MODEL, SPD_LAST_CID_I_MODEL,
                          SPD_FIRST_RID_I_MODEL, SPD_LAST_RID_I_MODEL);
}
