/* 
 * Program:             $RCSfile: mkDmodel.c,v $  $Revision: 30.20 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkDmodel.c,v 30.20 2003/06/03 15:28:01 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelAR.h>
#include <Sapid/ImodelARcompat.h>
#include "docident.h"
#include "dsdb.h"

int verbose = 0;

static char *target_software_name = SPD_DEFAULT_TARGET_SOFTWARE_NAME;
static char *target_document_name = SPD_DEFAULT_TARGET_DOCUMENT_NAME;
static SpdConstString program_name; /* This program name. */

static void parse_option(int argc, char *argv[]);
static void printUsage(SpdConstString command_name);
static SpdSet srcelem_set;

static SpdObjId softobj_id;
static SpdObjId docobj_id;

void loadISDBrel(const char *ifilename, SpdClassId class_id,
                 SpdObjId old_p_id, SpdObjId old_c_id,
                 SpdObjId new_p_id, SpdObjId new_c_id)
{
    SpdObjId new_srcelem_id;

    switch(class_id){
      case SPD_RID_FILE_DECL:
        new_srcelem_id = get_modified_id(srcelem_set,
                                         getSrcElemObjId(ifilename));
        if(new_srcelem_id == SPD_NON_ID){
            spdErrorExit("Can't find srcelem: \"%s\".\n", ifilename);
        }
        if(new_c_id == SPD_NON_ID){
            spdErrorExit("Invalid srcfrag id.\n");
        }
        /* Ϣ srcelem_srcfrag */
        spdCreateRel(SPD_RID_SRCELEM_SRCFRAG, SPD_NON_ID,
                     new_srcelem_id, new_c_id);
        break;
      default:
        break;
    }
}

SpdObjId loadISDBclass(const char *ifilename, SpdObjId obj_id, SpdObjId old_id)
{
    SpdObjId new_id;
    SpdClassId class_id;

    class_id = spdGetClassId(obj_id);

    switch(class_id){
      case SPD_CID_PROGRAM:                 /* FALLTHROUGH */
      case SPD_CID_FILE:
        spddb_UnRegisterDBObj(obj_id, spdGetDefaultDB());
        new_id = SPD_NON_ID;
        break;
      default:
        /* Regist obj */
        new_id = obj_id;
    }
    return new_id;
}

SpdObjId loadDSDBclass(const char *dfilename, SpdObjId obj_id, SpdObjId old_id)
{
    SpdObjId new_id;
    SpdClassId class_id;
    SpdConstString name;

    class_id = spdGetClassId(obj_id);
    switch(class_id){

      case SPD_CID_DOCUMENT:
        spddb_UnRegisterDBObj(obj_id, spdGetDefaultDB());
        new_id = SPD_NON_ID;
        break;
      case SPD_CID_DOCFILE:
        /* Ϣ doc_docfile  */
        spdCreateRel(SPD_RID_DOC_DOCFILE, SPD_NON_ID,
                     docobj_id, obj_id);
        new_id = obj_id;
        break;
      case SPD_CID_DOCFRAG:
        if(spdGetAttrIntValByName(obj_id, "sort") & DOCFRAG_SRCELEM){
            push_modification_table(srcelem_set, old_id, obj_id);
        }
        new_id = obj_id;
        break;
      case SPD_CID_DOCIDENT:
        name = spdGetName(obj_id);
        if((new_id=get_equal_docident(name)) == SPD_NON_ID){
            push_docident_table(obj_id, name);
            new_id = obj_id;
        }else{
            /* ̾ */
            spddb_UnRegisterDBObj(obj_id, spdGetDefaultDB());
        }
        break;
      default:
        new_id = obj_id;
    }

    return new_id;
}

void make_docident_identifier(void)
{
    SpdObjId identifier_id, docident_id;
    SpdCursor *csr;
    int i_sort, di_sort;

    csr = spdGetObjIdInit("identifier", NULL);
    while( (identifier_id = spdGetObjId(csr))!=SPD_NON_ID ){
        int scope;
        scope = spdGetAttrIntValByName(identifier_id, "scope");
        /* Ȥꤢextern  identifier */

        if (scope == ID_SCOPE_GLOBAL || scope == ID_SCOPE_NOT_GLOBAL) {
            const char *name;
            name = spdGetName(identifier_id);
            docident_id = get_equal_docident(name);

            if(docident_id != SPD_NON_ID){
                spdCreateRel(SPD_RID_DOCIDENT_IDENTIFIER, SPD_NON_ID,
                             docident_id, identifier_id);
                i_sort = spdGetAttrIntValByName(identifier_id, "sort");
                di_sort = spdGetAttrIntValByName(docident_id, "sort");
                /* docidentsortidentifierˤ碌ꤹ */

                if (di_sort != i_sort){
                    if(di_sort == DID_UNKNOWN){
                        spdSetAttrIntValByName(docident_id, "sort", i_sort);
                    }else{
                        spdWarning("Sort of document identifier don't match. >%s\n",
                                   name);
                    }
                }

            }
        }

    }
}

int main(int argc, char **argv)
{
    SpdModelId ImodelId, DmodelId;
    SpdDBId dbId;
    int i;
    SpdObjId docobj_id;
    SpdInfo *srcfrag_info;
    SpdInfoCursor cur_i,cur_o;
    SpdString ifilename;
    int srcfrag_offset;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parse_option(argc,argv);
    argc-=optind;
    argv+=optind;

    ImodelId = spdUseModel(&I_model);
    DmodelId = spdUseModel(&D_model);

    dbId = spdOpenSDB();
    spdReadDB(ImodelId, dbId);
    spdCreateDB(DmodelId, dbId);

    /* SOFTWARE object */
    softobj_id = spdCreateObj(SPD_CID_SOFTWARE, SPD_NON_ID);
    spdSetAttrStrValByName(softobj_id, "name", target_software_name);
    /* DOCUMENT object */
    docobj_id = spdCreateObj(SPD_CID_DOCUMENT, SPD_NON_ID);
    spdSetAttrStrValByName(docobj_id, "name", target_document_name);
    /* relation soft_doc */
    spdCreateRel(SPD_RID_SOFT_DOC, SPD_NON_ID, softobj_id, docobj_id);

    /* LINK */
    /****  D-modelʬν****/
    /* DSDBɤ߹ࡥ */
    for(i=0;i<argc;i++){
        srcelem_set = spdCreateSet(sizeof(ModifyField));

        spdLoadDSDB(argv[i], dbId, loadDSDBclass, NULL);

        /****  I-modelʬν****/
        srcfrag_info = spdOpenDSDBsrcfraginfo(argv[i]);
        cur_i = spdGetInfoStringFirst(srcfrag_info, SRCFRAGINFO_ATTR_IFILENAME);
        cur_o = spdGetInfoStringFirst(srcfrag_info, SRCFRAGINFO_ATTR_OFFSET);
        /* ڤФifile̾ϡDSDBinfo¸ƤΤǡ
           ɤ߹ߡifileISDB(SDB/DSDBˤե̾.i.DSDB)Ф
           󥯽 */
        while((ifilename = spdGetInfoStringNext(&cur_i))!=NULL){
            srcfrag_offset = atoi(spdGetInfoStringNext(&cur_o));
            spdLoadISDB(ifilename, dbId, loadISDBclass, loadISDBrel);
        }
        spdFreeSet(srcelem_set);
    }

    make_docident_identifier();

    /* WRITE */
    spdMessage("Writing D-model database.\n");
    spdCommitDB(ImodelId, dbId);
    spdCommitDB(DmodelId, dbId);

    spdCloseSDB(dbId);
    free_docident_table();

    return (EXIT_SUCCESS);
}

static void parse_option(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "bd:o:s:")) != -1) {
        switch (ch) {
          case 'b':                         /* Binary format. */
            spdSetWriteSDBformat(SAPID_BINARY_FORMAT_SDB);
            break;
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'o':                         /* Document name. */
            target_document_name = optarg;
            break;
          case 's':                         /* Software name. */
            target_software_name = optarg;
            break;
          default:
            printUsage(program_name);
            break;
        }
    }
}

static void printUsage(SpdConstString command_name)
{
    const char  *usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [options] \n"
        "DESCRIPTION\n"
        "    -d SDBdirectory.\n"
        "    -o DocumentName\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
