/* 
 * Program:             $RCSfile: SrcMacro.c,v $  $Revision: 41.17 $
 * 
 * Purpose:             
 * 
 * Author:              K.Toita  2001/01/10
 *                      S.Yamamoto  2001/08/23
 * 
 * (C) Copyright:       K.Toita and S.Yamamoto  2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: SrcMacro.c,v 41.17 2003/06/28 15:33:41 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <stdarg.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelARcompat.h>
#include <Sapid/SpdStr.h>
#include "chkDoc.h"
#include "printDoc.h"
#include "funcinfo.h"

SpdObjIdArray MacroDocidentIdArray = SPD_EMPTY_ARRAY;
SpdObjIdArray PBlockIdArray = SPD_EMPTY_ARRAY;

MacroInfoArray ValidMacroArray = EMPTY_MI_ARRAY;
MacroInfoArray ExistMacroArray = EMPTY_MI_ARRAY;
MacroInfoArray NoDocMacroArray = EMPTY_MI_ARRAY;
MacroInfoArray InvalidMacroArgNumArray = EMPTY_MI_ARRAY;
MacroInfoArray InvalidMacroArgNameArray = EMPTY_MI_ARRAY;
MacroInfoArray NotExistMacroArray = EMPTY_MI_ARRAY;

void setMacroInfointoArraySub(MacroInfoArray *miArray, MacroInfo mi)
{
    miArray->mi = spdRealloc(miArray->mi, sizeof(MacroInfo) * (miArray->size + 1));
    miArray->mi[miArray->size] = mi;
    miArray->size++;
}

void setMacroInfointoArray(MacroInfo mi)
{
    if(mi.status == STATUS_NOMACRO){
        setMacroInfointoArraySub(&NoDocMacroArray, mi);
    } else if (!flag_m.detail && (mi.status == STATUS_EXISTMACRO)) {
        setMacroInfointoArraySub(&ExistMacroArray, mi);
    } else if(mi.status & STATUS_INVALID_MACRO_ARGNUM){
        setMacroInfointoArraySub(&InvalidMacroArgNumArray, mi);
    } else if(mi.status & STATUS_INVALID_MACRO_ARGNAME){
        setMacroInfointoArraySub(&InvalidMacroArgNameArray, mi);
    } else{
        setMacroInfointoArraySub(&ValidMacroArray, mi);
    }
}

/* Υޥϴ˥åƤ뤫 */
SpdBoolean isCheckedMacro(SpdObjId pblockId)
{
    SpdObjId patternId, nameId;
    SpdString mname;
    int i;

    patternId = spdGetARelObj(pblockId, "def_mpatt", "pblock_id");
    nameId = spdGetARelObj(patternId, "mpatt_mname", "macroPattern_id");
    mname = spdStrdup(spdGetName(nameId));
    if (flag_f.detail) {
        for (i = 0; i < ValidMacroArray.size; i++) {
            if (SPD_STREQ(mname, ValidMacroArray.mi[i].src.name)) {
                spdFree(mname);
                return TRUE;
            }
        }
        for (i = 0; i < InvalidMacroArgNumArray.size; i++) {
            if (SPD_STREQ(mname, InvalidMacroArgNumArray.mi[i].src.name)) {
                spdFree(mname);
                return TRUE;
            }
        }
        for (i = 0; i < InvalidMacroArgNameArray.size; i++) {
            if (SPD_STREQ(mname, InvalidMacroArgNameArray.mi[i].src.name)) {
                spdFree(mname);
                return TRUE;
            }
        }
    } else {
        for (i = 0; i < ExistMacroArray.size; i++) {
            if (SPD_STREQ(mname, ExistMacroArray.mi[i].src.name)) {
                spdFree(mname);
                return TRUE;
            }
        }
    }
    for (i = 0; i < NoDocMacroArray.size; i++) {
        if (SPD_STREQ(mname, NoDocMacroArray.mi[i].src.name)) {
            spdFree(mname);
            return TRUE;
        }
    }
    spdFree(mname);
    return FALSE;
}

/*  docidentId Ͽ */
void setMacroDocidentIdintoArraySub(SpdObjId docidentId)
{
    MacroDocidentIdArray.id = spdRealloc(MacroDocidentIdArray.id,
                                         sizeof(SpdObjId) * (MacroDocidentIdArray.size +1));
    MacroDocidentIdArray.id[MacroDocidentIdArray.size] = docidentId;
    MacroDocidentIdArray.size++;
}

/* sort  macro  docidentId Ͽ */
void setMacroDocidentIdintoArray(void)
{
    SpdObjId docidentId;
    SpdCursor *docidentCsr;

    docidentCsr = spdGetObjIdInit("docident", NULL);
    while ((docidentId = spdGetObjId(docidentCsr)) != SPD_NON_ID) {
        if (spdGetAttrValInt(docidentId, "sort") == DID_MACRONAME) {
            setMacroDocidentIdintoArraySub(docidentId);
        }
    }
    spdFreeCursor(docidentCsr);
}

char *getMacroBody(SpdObjId pblockId)
{
    SpdObjId bodyId;
    bodyId = spdGetARelObj(pblockId, "def_mbody", "pblock_id");
    if(bodyId == SPD_NON_ID) {
        return NULL;
    } else {
        return spdStrdup(spdStr(bodyId));
    }
}

void printMacroSet(MacroSet mset)
{
    int i;
    printf("#define %s",mset.name);
    if(mset.arg_num != 0) {
        printf("(");
    }
    for (i = 0; i < mset.arg_num; i++) {
        printf("%s",mset.argname[i]);
        if(i != mset.arg_num - 1) {
            printf(",");
        }
    }
    if(mset.arg_num != 0) {
        printf(")");
    }
    printf(" %s\n",mset.body);
}

MacroSet getSrcMacroSet(SpdObjId pblockId)
{
    SpdObjId patternId, nameId, optId;
    SpdRelArray margArray;
    MacroSet mset = EMPTY_MS;
    int i;
    
    patternId = spdGetARelObj(pblockId, "def_mpatt", "pblock_id");
    nameId = spdGetARelObj(patternId, "mpatt_mname", "macroPattern_id");
    mset.name = spdStrdup(spdGetName(nameId));

    optId = spdGetARelObj(patternId, "mpatt_mopt", "macroPattern_id");
    if(optId != SPD_NON_ID) {
        margArray = spdGetRelSortedArray(optId, "mopt_mdecla",
                                         "p-id", NULL, spdCompareRelOffset);
        mset.arg_num = margArray.size;
        mset.argname = (char **)spdMalloc(sizeof(char *)*(mset.arg_num));

        for (i = 0; i < margArray.size; i++) {
            mset.argname[i] = trim((SpdString)spdGetName(margArray.rel[i].objectId));
        }
        spdFreeRelArray(margArray);
    } else {
        mset.arg_num = 0;
    }
    mset.body = getMacroBody(pblockId);

    return mset;
}

MacroSet getDocMacroSetSub(SpdObjId textfragId)
{
    SpdRelArray textArray;
    SpdString str, ptr;
    MacroSet mset = EMPTY_MS;    
    int i,j,k,n;
    
    textArray = spdGetRelSortedArray(textfragId, "textfrag_textstring",
                                     "p-id", NULL, spdCompareRelOffset);
    for (i = 0; i < textArray.size; i++) {
        str = spdStrdup((SpdString)spdGetAttrValString(textArray.rel[i].objectId,"text"));
        if ((ptr = strstr(str, MACRO_START)) != NULL) {
            /* #defineȤФ */
            for (j = 0; j < strlen(MACRO_START); j++) {
                ptr++;
            }

            /* #define  ޥ̾δ֤ζȤФ */
            while(*ptr == ' ') {
                ptr++;
            }

            /* ޥ̾ʸ */
            k = 0;
            while(*ptr != ' ' && *ptr !='(' && *ptr != '\0' && *ptr != '\t') {
                ptr++;
                k++;
            }

            /* ʸܣ*/
            mset.name = spdMalloc(sizeof(SpdString)*(k+1));

            /* ޥ̾򥳥ԡ */
            j = 0;
            ptr = ptr - k;
            while(k > 0) {
                mset.name[j++] = *ptr++;
                k--;
            }
            mset.name[j] = '\0';

            if (*ptr == '(') {              /* ĥޥ */
                SpdString   buf, argname;
                int         arglen;
                
                ptr++;                      /* '('ȤФ */
                arglen = 0;
                mset.arg_num = 1;

                /* οĴ٤ */
                while (*ptr != ')') {
                    if (*ptr == ',') {
                        mset.arg_num++;
                    }
                    ptr++;
                    arglen++;
                }
                mset.argname = (char **)spdMalloc(sizeof(char *)*(mset.arg_num));
                buf = spdMalloc(sizeof(SpdString)*(arglen+1));

                ptr = ptr - arglen;

                j = 0;
                n = 0;

                /* ̾򥳥ԡ */
                while (*ptr != ')') {
                    ptr++;
                    j++;
                    if (*ptr == ',' || *ptr == ')') {

                        argname = spdMalloc(sizeof(char *)*(j+1));
                        ptr = ptr - j;
                        k = 0;
                        while(j > 0) {
                            argname[k++] = *ptr++;
                            j--;
                        }
                        argname[k] = '\0';
                        mset.argname[n++] = spdStrdup((SpdString)trim(argname));
                        spdFree(argname);
                        if (*ptr == ',') {
                            ptr++;
                        }
                    }
                }                
                ptr++;                      /* ')'ȤФ */
            }

            /* body ޤǤζȤФ */
            while (*ptr == ' ' || *ptr == '\t') {
                ptr++;
            }

            if (*ptr == '\0') {             /* ΤߤΥޥ */
                mset.value = 1;             /* ΤߤΥޥͤ򣱤Ȥ */
            } else {

                /* body ĹĴ٤ */
                j = 0;
                while (*ptr != '\0') {
                    ptr++;
                    j++;
                }

                mset.body = spdMalloc(sizeof(SpdString)*(j+1));
                ptr = ptr - j;
                k = 0;
                while (j > 0) {
                    mset.body[k++] = *ptr++;
                    j--;
                }
                mset.body[k] = '\0';
            }
        }
        spdFree(str);
    }
    spdFreeRelArray(textArray);
    return mset;
}

MacroSet getDocMacroSet(SpdObjId docfragId)
{
    SpdRelArray docfragArray;
    SpdObjId childDocfragId;
    MacroSet mset = EMPTY_MS;
    int i;
    
    docfragArray =spdGetRelSortedArray(docfragId, "docfrag_docelem",
                                       "p-id", NULL, spdCompareRelOffset);
    for (i = 0; i < docfragArray.size; i++) {
        if (spdGetAttrValInt(docfragArray.rel[i].objectId, "sort") == DOCFRAG_MACROELEM) {
            childDocfragId = spdGetARelObj(docfragArray.rel[i].objectId,
                                           "docfrag_docelem", "p-id");
            spdFreeRelArray(docfragArray);
            return (getDocMacroSetSub(childDocfragId));
        }
    }
    spdFreeRelArray(docfragArray);
    return mset;
}

MacroInfo getMacroInfo(SpdObjId pblockId)
{
    SpdObjId patternId;
    MacroInfo mi = EMPTY_MI;

    mi.src = getSrcMacroSet(pblockId);
    patternId = spdGetARelObj(pblockId, "def_mpatt", "pblock_id");
    mi.name_id = spdGetARelObj(patternId, "mpatt_mname", "macroPattern_id");
    
    return mi;
}

void checkMacroInDoc(MacroInfo *mi)
{
    int invalid = 0;
    int i,j;
    
    flag_w = FALSE;                         /* warning flag  */
    for (i = 0; i < MacroDocidentIdArray.size; i++) {
        if (SPD_STREQ(spdGetName(MacroDocidentIdArray.id[i]),mi->src.name)) {
            /* ˤޥɥȤ¸ߤ */
            mi->docfrag_id = spdGetARelObj(MacroDocidentIdArray.id[i],
                                           "defelem_docident", "docident_id");
            mi->doc = getDocMacroSet(mi->docfrag_id);
            mi->status = STATUS_EXISTMACRO;

            if (!flag_m.detail) {
                putMessage(&(mi->message), "Exist.\n");
                exist_macro++;
                return;
            }
            if (mi->src.arg_num != mi->doc.arg_num) { /* ο㤦 */
                if (!flag_w) {
                    flag_w = TRUE;
                    putMessage(&(mi->message), "Warning! <BR />\n");
                }
                mi->status |= STATUS_INVALID_MACRO_ARGNUM;
                putMessage(&(mi->message), "Invalid ArgNum. %d <--> %d\n<BR />",
                           mi->src.arg_num, mi->doc.arg_num);
                invalid_macro_arg_num++;
                invalid++;
            } else {
                for (j = 0; j < mi->src.arg_num; j++) {
                    if(!SPD_STREQ(mi->src.argname[j],mi->doc.argname[j])) {
                        /* ̾㤦 */
                        if (!flag_w) {
                            flag_w = TRUE;
                            putMessage(&(mi->message), "Warning! <BR />\n");
                        }
                        mi->status |= STATUS_INVALID_MACRO_ARGNAME;
                        putMessage(&(mi->message), "Invalid Arg-%d Name. \"%s\" <--> \"%s\"<BR />\n",
                                   j+1, mi->src.argname[j], mi->doc.argname[j]);
                        invalid_macro_arg_name++;
                        invalid++;
                    }
                }
            }
            if (!flag_w) {                  /* ٹ1Ĥʤ OK ɽ */
                putMessage(&(mi->message), "OK\n");
                valid_macro++;
            }
            if (invalid) {
                invalid_macro++;
            }
            return;
        }
    }
    
    /* ˤޥɥȤ¸ߤʤ */
    mi->status = STATUS_NOMACRO;
    putMessage(&(mi->message), "No Document.\n");
    macro_no_doc++;
}

/* ˤޥɥȤˤ뤫Ĵ٤ */
void checkSrcProgramMacroSub(SpdObjId pfileId)
{
    SpdRelArray pblockArray;
    int i;
    MacroInfo mi = EMPTY_MI;
    
    pblockArray = spdGetRelSortedArray(pfileId, "file_block",
                                       "pfile_id", NULL, spdCompareRelOffset);
    for (i = 0; i < pblockArray.size; i++) {
        if (spdGetAttrValInt(pblockArray.rel[i].objectId, "sort") == PBLOCK_DEFINE_FILE) {
            if (!isCheckedMacro(pblockArray.rel[i].objectId)) {
                mi = getMacroInfo(pblockArray.rel[i].objectId);
                mi.pfile_id = smuGetFileIdByName((SpdString)spdGetName(pfileId));
                checkMacroInDoc(&mi);
                /* printMacroSet(mi.src); */
                /* printMacroSet(mi.doc); */
                setMacroInfointoArray(mi);
                if (flag_m.src || (!flag_m.src && (mi.docfrag_id != SPD_NON_ID))) { 
                    macro_num++;
                }
            }
        }
    }
}

SpdBoolean checkMacroInSrcSub(SpdObjId docidentId)
{
    int i;
    SpdObjId docfragId;

    docfragId = spdGetARelObj(docidentId, "defelem_docident", "docident_id");
    if (flag_m.detail) {
        for (i = 0; i < ValidMacroArray.size; i++) {
            if (docfragId == ValidMacroArray.mi[i].docfrag_id) {
                return TRUE;
            }
        }
        for (i = 0; i < InvalidMacroArgNumArray.size; i++) {
            if (docfragId == InvalidMacroArgNumArray.mi[i].docfrag_id) {
                return TRUE;
            }
        }
        for (i = 0; i < InvalidMacroArgNameArray.size; i++) {
            if (docfragId == InvalidMacroArgNameArray.mi[i].docfrag_id) {
                return TRUE;
            }
        }
    } else {
        for (i = 0; i < ExistMacroArray.size; i++) {
            if (docfragId == ExistMacroArray.mi[i].docfrag_id) {
                return TRUE;
            }
        }
    }
    return FALSE;
}

/* ɥȤΥޥˤ뤫Ĵ٤ */
void checkMacroInSrc(void)
{
    int i;
    MacroInfo mi = EMPTY_MI;

    for (i = 0; i < MacroDocidentIdArray.size; i++) {
        if (!checkMacroInSrcSub(MacroDocidentIdArray.id[i])) {
            mi.docfrag_id = spdGetARelObj(MacroDocidentIdArray.id[i],
                                          "defelem_docident", "docident_id");
            mi.doc = getDocMacroSet(mi.docfrag_id);
            mi.status = STATUS_NOSRC;
            putMessage(&(mi.message), "Not Exist.\n");
            setMacroInfointoArraySub(&NotExistMacroArray, mi);
            no_macro++;
        }
    }
}

void checkSrcProgramMacro(void)
{
    SpdObjId pfileId;
    SpdCursor *pfileCsr;

    /* docident_id Τ sort  macro ΤΤϿƤ */
    setMacroDocidentIdintoArray();
    
    /* ΥƥեʳΥޥɥȤˤ뤫Ĵ٤ */
    pfileCsr = spdGetObjIdInit("pfile", NULL);
    while ((pfileId = spdGetObjId(pfileCsr)) != SPD_NON_ID) {
        if (!smuIsSystemFile(pfileId)) {
            checkSrcProgramMacroSub(pfileId);
        }
    }

    /* ɥȤΥޥˤ뤫Ĵ٤ */
    if (flag_m.doc) {
        checkMacroInSrc();
    }
}
