/* 
 * Program:             $RCSfile: chkEnum.c,v $  $Revision: 40.6 $
 * 
 * Purpose:             
 * 
 * Author:              K.Toita  2002/01/05
 *                      S.Yamamoto  2002/01/05
 * 
 * (C) Copyright:       K.Toita - S.Yamamoto  2002 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: chkEnum.c,v 40.6 2003/07/13 09:17:29 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <stdarg.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelARcompat.h>
#include <Sapid/PmodelAR.h>
#include <Sapid/SpdStr.h>
#include "chkDoc.h"
#include "SrcElem.h"
#include "SrcType.h"
#include "funcinfo.h"

EnumInfoArray ValidEnumArray = EMPTY_EI_ARRAY;
EnumInfoArray ExistEnumArray = EMPTY_EI_ARRAY;
EnumInfoArray NoDocEnumArray = EMPTY_EI_ARRAY;
EnumInfoArray InvalidEnumNumArray = EMPTY_EI_ARRAY;
EnumInfoArray InvalidEnumNameArray = EMPTY_EI_ARRAY;
EnumInfoArray InvalidEnumValueArray = EMPTY_EI_ARRAY;
EnumInfoArray NotExistEnumArray = EMPTY_EI_ARRAY;

void setEnumInfointoArraySub(EnumInfoArray *eiArray, EnumInfo ei)
{
    eiArray->ei = spdRealloc(eiArray->ei,
                             sizeof(EnumInfo) * (eiArray->size + 1));
    eiArray->ei[eiArray->size] = ei;
    eiArray->size++;
}

void setEnumInfointoArray(EnumInfo ei)
{
    if (ei.status == STATUS_NOENUM) {
        setEnumInfointoArraySub(&NoDocEnumArray, ei);
    } else if (!flag_e.detail && (ei.status == STATUS_EXISTENUM)) {
        setEnumInfointoArraySub(&ExistEnumArray, ei);
    } else if (ei.status & STATUS_INVALID_ENUM_NUM) {
        setEnumInfointoArraySub(&InvalidEnumNumArray, ei);
    } else if (ei.status & STATUS_INVALID_ENUM_NAME) {
        setEnumInfointoArraySub(&InvalidEnumNameArray, ei);
    } else if (ei.status & STATUS_INVALID_ENUM_VALUE) {
        setEnumInfointoArraySub(&InvalidEnumValueArray, ei);
    } else {
        setEnumInfointoArraySub(&ValidEnumArray, ei);
    }
}

int getEnumValue(SpdObjId declId, int beforeValue)
{
    SpdObjId decrId, optId, exprId, constId;
    SpdString valstr;
    int val;
    
    decrId = spdGetARelObj(declId, "decl_decl", "declaration_id");
    optId = spdGetARelObj(decrId, "decl_opt", "declarator_id");
    if (optId == SPD_NON_ID) {
        /* äͤͿʤ+1֤ */
        return (beforeValue + 1);
    }
    exprId = spdGetARelObj(optId, "init_expr", "optionaldecl_id");
    constId = spdGetARelObj(exprId, "const_ref", "expression_id");
    valstr = spdStrdup(spdGetAttrStrValByName(constId, "value"));
    if (valstr != NULL) {
        val = atoi(valstr);
        spdFree(valstr);
        return val;
    } else {
        spdFree(valstr);
        return (beforeValue + 1);
    }
}

EnumSet getEnumSet(SpdObjId decrId)
{
    SpdObjId identId, optId;
    EnumSet eset = EMPTY_ES;
    int i;

    eset.decr_id = decrId;
    identId = spdGetARelObj(decrId, "decl_ident", "declarator_id");
    if (identId != SPD_NON_ID) {
        eset.name = spdStrdup(spdGetName(identId));
    }
    
    optId = spdGetARelObj(decrId, "decl_opt", "declarator_id");
    eset.enumrels = spdGetRelSortedArray(optId,
                                         "struct_decl", "optionaldecl_id",
                                         NULL, spdCompareRelOffset);
    eset.enum_num = eset.enumrels.size;
    eset.enum_name = (char **)spdMalloc(sizeof(char *)*(eset.enum_num));
    eset.enum_value = spdMalloc(sizeof(int)*(eset.enum_num));
    for (i = 0; i < eset.enumrels.size; i++) {
        eset.enum_name[i] = spdStrdup(getName(eset.enumrels.rel[i].objectId));
        /* getEnumValueäͤꤵʤ硤+1֤ */
        if (i == 0) {
            /* ǽͤϲʤȤȤʤΤǡ-1Ȥ */
            eset.enum_value[i] = getEnumValue(eset.enumrels.rel[i].objectId,
                                              -1);
        } else {
            eset.enum_value[i] = getEnumValue(eset.enumrels.rel[i].objectId,
                                              eset.enum_value[i-1]);
        }
    }
    return eset;
}

void printEnumSet(EnumSet eset)
{
    int i;

    printf("name:%s\n",eset.name);
    printf("decl:%d decr:%d\n",eset.decl_id, eset.decr_id);
    for (i = 0; i < eset.enum_num; i++) {
        printf("name:%s value:%d\n",eset.enum_name[i], eset.enum_value[i]);
    }
}

EnumInfo getEnumInfo(SpdObjId decrId)
{
    EnumInfo ei = EMPTY_EI;

    ei.src = getEnumSet(decrId);
    /* printEnumSet(ei.src); */
    return ei;
}

void checkEnumInDoc(EnumInfo *ei)
{
    SpdObjId id, identId, docidentId, defelemId, decrId;

    identId = spdGetARelObj(ei->src.decr_id, "decl_ident", "declarator_id");
    docidentId = spdGetARelObj(identId, "docident_identifier", "identifier_id");
    defelemId = spdGetARelObj(docidentId, "defelem_docident", "docident_id");

    ei->docfrag_id = defelemId;

    if (flag_e.src  || (!flag_e.src && ei->docfrag_id != SPD_NON_ID)) {
        enum_num++;
    }

    if (defelemId == SPD_NON_ID) {
        ei->status = STATUS_NOENUM;
        enum_no_doc++;
        putMessage(&(ei->message), "No Document.\n");
    } else {
        id = getSrcElemId(defelemId);
        id = spdGetARelObj(id, "srcelem_srcfrag", "srcelem_id");

        if (id == SPD_NON_ID) {
            ei->status = STATUS_NOENUM;
            enum_no_doc++;
            putMessage(&(ei->message), "No Document.\n");
        } else {
            ei->status = STATUS_EXISTENUM;

            decrId = spdGetARelObj(id, "decl_decl", "declaration_id");
            ei->doc = getEnumSet(decrId);
            ei->doc.decl_id = id;
            
            if (!flag_e.detail) {
                putMessage(&(ei->message), "Exist.\n");
                exist_enum++;
            }
        }
    }
}

void checkEnumDetail(EnumInfo *ei)
{
    int invalid = 0;

    /* enumοΥå */
    if (ei->doc.enum_num != ei->src.enum_num) {
        if(!flag_w){                        /* ƤΥ顼λ Warning! ɽ */
            flag_w = TRUE;
            putMessage(&(ei->message), "Warning! <BR />\n");
        }
        ei->status |= STATUS_INVALID_ENUM_NUM;
        putMessage(&(ei->message),
                   "Invalid Enum Num. \"%d\" <--> \"%d\"<BR />\n",
                   ei->src.enum_num, ei->doc.enum_num);
        invalid_enum_num++;                 /* count */
        invalid++;
    } else {
        /* enum ̾ͥå */
        int i;
        for (i = 0; i < ei->src.enum_num; i++) {
            if (!SPD_STREQ(ei->doc.enum_name[i], ei->src.enum_name[i])) {
                /* ̾㤦Ȥ */
                if (!flag_w) {
                    /* ƤΥ顼λ Warning! ɽ */
                    flag_w = TRUE;
                    putMessage(&(ei->message), "Warning! <BR />\n");
                }
                ei->status |= STATUS_INVALID_ENUM_NAME;
                putMessage(&(ei->message),
                           "Invalid Enum-%d Name. \"%s\" <--> \"%s\"<BR />\n",
                           i+1, ei->src.enum_name[i], ei->doc.enum_name[i]);
                invalid_enum_name++;
                invalid++;
            }
            if (ei->doc.enum_value[i] != ei->src.enum_value[i]) {
                /* ͤ㤦Ȥ */
                if (!flag_w) {
                    /* ƤΥ顼λ Warning! ɽ */
                    flag_w = TRUE;
                    putMessage(&(ei->message), "Warning! <BR />\n");
                }
                ei->status |= STATUS_INVALID_ENUM_VALUE;
                putMessage(&(ei->message),
                           "Invalid Enum-%d Value. \"%d\" <--> \"%d\"<BR />\n",
                           i+1, ei->src.enum_value[i], ei->doc.enum_value[i]);
                invalid_enum_value++;
                invalid++;
            }
        }
    }
    if (!flag_w) {
        /* ٹ1Ĥʤ OK ɽ */
        putMessage(&(ei->message), "OK\n");
        valid_enum++;
    }
    if (invalid) {
        invalid_enum++;
    }
}

SpdBoolean spdDeclIsDeclEnum(const SpdObjId declId)
{
    if (spdGetAttrValInt(declId, "sort") == DECL_ENUMERATOR) return TRUE;
    return FALSE;
}

void checkEnum(SpdObjId structdeclId)
{
    SpdObjId    optId, decrId, declId, identId;
    SpdCursor   *csr;
    SpdOcc      occ;
    EnumInfo    ei;
    
    optId = spdGetARelObj(structdeclId, "struct_decl", "declaration_id");
    decrId = spdGetARelObj(optId, "decl_opt", "optionaldecl_id");
    declId = spdGetARelObj(decrId, "decl_decl", "declarator_id");
    identId = spdGetARelObj(decrId, "decl_ident", "declarator_id");
    
    csr = spdGetOffsetOfObjInit(declId);
    occ = spdGetOffsetOfObj(csr);

    if (!is_system(occ)) {
        /* ץҤEnumȡ̾ΤʤEnum */
        if ((occ.fileId != SPD_NON_ID) && (identId != SPD_NON_ID) &&
            !is_declIdinArray(declId)) {
            spdDeclarationIdintoArray(declId);
            ei = getEnumInfo(decrId);
            ei.src.decl_id = declId;
            ei.pfile_id = getpfileId(declId);
            checkEnumInDoc(&ei);
            /* printEnumSet(ei.src); */
            /* printEnumSet(ei.doc); */

            if (flag_e.detail && (ei.status == STATUS_EXISTENUM)) {
                checkEnumDetail(&ei);
            }

            setEnumInfointoArray(ei);

            flag_w = FALSE;                 /* flag_w ν */
        }
    }
}

void checkSrcProgramEnum(void)
{
    SpdObjIdArray   declArray;
    int             i;
    
    declArray = spdGetObjIdArray("declaration", NULL, spdDeclIsDeclEnum);
    for (i = 0; i < declArray.size; i++) {
        checkEnum(declArray.id[i]);
    }
    spdFreeObjIdArray(declArray);
}
