/* 
 * Program:             $RCSfile: funcinfo.c,v $  $Revision: 40.8 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: funcinfo.c,v 40.8 2002/12/16 12:11:38 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <stdarg.h>
#include <Sapid/Sapid.h>
#include <Sapid/ImodelARcompat.h>
#include <Sapid/PmodelAR.h>
#include <Sapid/SpdStr.h>
#include "chkDoc.h"
#include "funcinfo.h"

/* Message. */

void vputMessage(char **message, char *format, va_list args)
{
    char buf[SPD_STRING_LEN + 1];

    vsprintf(buf, format, args);
    *message = spdStrcat(*message, buf);
}

void putMessage(char **message, char *format, ...)
{
    va_list args;

    va_start(args, format);
    vputMessage(message, format, args);
    va_end(args);
}

/* Commented by S.Yamamoto 2001/09/21. */
/* 
 * void clearMessage(FuncInfo *fi)
 * {
 *     if (fi->message) {
 *         spdFree(fi->message);
 *         fi->message = NULL;
 *     }
 * }
 */

void fprintfMessage(FILE *fp, char **message)
{
    if (message) {
        fprintf(fp, "%s", *message);
    }
}

/* declaration(prototype, arg)identifier name롥 */

SpdString getName(SpdObjId id)
{
    id = spdGetARelObj(id, "decl_decl", "declaration_id");
    id = spdGetARelObj(id, "decl_ident", "declarator_id");
    return (id == SPD_NON_ID ? "" : (SpdString)spdGetName(id));
}

char *getArgTypename(SpdObjId decl_id)
{
    SpdString str;
    static char name[BUF_MAX];
    SpdObjId id;

    id=spdGetARelObj(decl_id, "decl_decl", "declaration_id");
    id=spdGetARelObj(id, "decr_type", "declarator_id");  
    spdSetOption("prototype", SPDSTR_PROTOTYPE_TYPE);
    str=spdStr(id);
    if (str) {
        strcpy(name, str);
    }
    spdStrFree();
    return name;
}

char *getFuncTypename(SpdObjId func_id)
{
    SpdString str;
    static char name[BUF_MAX];
    SpdObjId typeid;
    SpdObjId id;

    id=spdGetARelObj(func_id, "decl_decl", "declaration_id");
    typeid=spdGetARelObj(id, "decr_type", "declarator_id");
    str=spdStr(typeid);
    if (str) {
        strcpy(name, str);
    } else {

#ifdef COMMENTED
        /* decr_type̵(funcid˴ؤƤϡʤȤϤ̵֤)
           decl_type롥*/
        typeid=spdGetARelObj(decl_id, "decl_type", "declaration_id");
        str=spdStr(typeid);
        if (str) {
            strcpy(name, str);
        }
#endif

    }
    spdStrFree();
    return name;
}

void freeFuncSet(FuncSet fset)
{
    int i;

    spdFree(fset.funcname);
    spdFree(fset.functype);

    for(i = 0; i < fset.arg_num; i++){
        spdFree(fset.argtype[i]);
        spdFree(fset.argname[i]);
    }
    spdFree(fset.argtype);
    spdFree(fset.argname);
    spdFreeRelArray(fset.argrels);
}

void freeFuncInfo(FuncInfo fi)
{
    freeFuncSet(fi.src);
    freeFuncSet(fi.doc);

    if(fi.message){
        spdFree(fi.message);
    }
}

#ifdef COMMENTED

void printFuncInfo(FuncInfo fi)
{
    int i;

    printf("%d:",fi.funcdecl_id);
    printf("%s %s", fi.functype, fi.funcname);
    printf("(");
    for (i=0;i<fi.arg_num;i++) {
        printf("%s %s", fi.argtype[i], fi.argname[i]);
        if (i < fi.arg_num-1) {
            printf(", ");
        }
    }
    printf(");\n");
    printf("%d:",fi.status);
    if (fi.status == STATUS_NOFUNC) {
        printf("Nofunc.\n");
    }
    if (fi.status & STATUS_EXISTFUNC) {
        printf("------------------------------Func is exist.\n");
    }
    if (fi.status & STATUS_INVALID_FUNCTYPE) {
        printf("------------------------------Func has invalid type.\n");
    }
    if (fi.status & STATUS_INVALID_ARGTYPE) {
        printf("------------------------------Func has invalid arg type.\n");
    }
    if (fi.status & STATUS_INVALID_ARGNAME) {
        printf("------------------------------Func has invalid arg name.\n");
    }
}

#endif /* COMMENTED */

/* ζ */
/* a, bϤδؿ,ζ򤬼Τա */

SpdString trim(SpdString a)
{
    SpdString str1;
    SpdString str2;
    int i,j,k;
    
    if (strlen(a) == 0) {
        return a;
    }

    str1 = spdStrdup(a);
    i = strlen(str1);
    
    /* ζ롥 */
    while (str1[--i] == ' ') {
        /* Nothing. */
    }
    
    str1[++i] = '\0';

    /* ζ */
    j = 0;
    while (str1[j] == ' ') {
        j++;
    }
    str2 = spdMalloc(sizeof(SpdString)*(i-j+1));
    k = 0;
    while (j <= i) {
        str2[k++] = str1[j++];
    }

    spdFree(str1);

    return str2;
}

FuncSet getFuncSet(SpdObjId declId)
{
    FuncSet fset = EMPTY_FS;
    int i;
    SpdObjId id;

    fset.decl_id = declId;
    fset.funcname = trim(getName(declId));
    fset.functype = trim(getFuncTypename(declId));

    id = spdGetARelObj(declId, "decl_decl", "declaration_id");
    id = spdGetARelObj(id, "decl_opt", "declarator_id");
    if (spdGetAttrValInt(id, "sort") == DECL_FUNCDEF) {
        fset.argrels = spdGetRelSortedArray(id, "func_parm", "optionaldecl_id",
                                            NULL, spdCompareRelOffset);
    } else {
        fset.argrels = spdGetRelSortedArray(id, "prot_parm", "optionaldecl_id",
                                            NULL, spdCompareRelOffset);
    }
    
    fset.arg_num = fset.argrels.size;
    fset.argtype = spdMalloc(sizeof(char *) * fset.arg_num);
    fset.argname = spdMalloc(sizeof(char *) * fset.arg_num);
    for (i = 0; i < fset.argrels.size; i++) {
        fset.argtype[i] = trim(getArgTypename(fset.argrels.rel[i].objectId));
        fset.argname[i] = trim(getName(fset.argrels.rel[i].objectId));
    }
    return fset;
}

FuncInfo getFuncInfo(SpdObjId declId)
{
    FuncInfo fi = EMPTY_FI;

    if (declId == SPD_NON_ID) {
        return fi;
    }

#if defined(SPD_CHK_DOC)
    fi.funcdecl_id = declId;
#elif defined(SPD_CHK_DOC2)
    fi.pfile_id = getpfileId(declId);
#endif /* defined(SPD_CHK_DOC2) */

    fi.src = getFuncSet(declId);
    
    fi.status = STATUS_NOFUNC;
    return fi;
}

/******************************************************************************/

#ifdef SPD_CHK_DOC2

SpdObjId getpfileId(SpdObjId declId)
{
    SpdOcc      occ;
    SpdObjId    block_id, pfile_id;
    
    occ = spdGetAnOffsetOfObj(declId);
    block_id = spdGetABlockId(occ.fileId, occ.offset);
    pfile_id = spdGetARelObj(block_id, "file_block", "c-id");
    pfile_id = smuGetFileIdByName((SpdString)spdGetName(pfile_id));
    return pfile_id;
}

#endif /* SPD_CHK_DOC2 */
