/* 
 * Program:             $RCSfile: getInfo.c,v $  $Revision: 40.10 $
 * 
 * Purpose:             
 * 
 * Author:              K.Toita  2001/12/01
 *                      S.Yamamoto  2001/12/01
 * 
 * (C) Copyright:       K.Toita and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: getInfo.c,v 40.10 2003/06/28 15:33:40 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "chkDoc.h"
#include "getInfo.h"

static int get_attr_num(SpdInfo *info, const char *attr, int start_num)
{
    int         i;
    SpdInfoData *data;

    if (attr == NULL) {
        return NOT_EXIST;
    }
    data = info->data;

    for (i = start_num; i < data->size; i++) {
        if (SPD_STREQ(data->elem[i].attr, attr)) {
            return i;
        }
    }
    return NOT_EXIST;
}

static void read_all_info(SpdInfo *info)
{
    int         c, i;
    FILE        *fp;
    SpdInfoElem elem;
    SpdInfoData *data;

    data = info->data;
    if ((fp = spdSimpleFopen(info->filename, SPD_FOPEN_RT_MODE)) == NULL) {
        return;
    }

    while ((c = fgetc(fp), ungetc(c, fp), c != EOF)) {
        /* Skip empty line and comment line. */
        if (c == '#') {
            c = fgetc(fp);
            while (c != '\n' && c != EOF) {
                c = fgetc(fp);
            }
            continue;
        } else if (c == '\n') {
            c = fgetc(fp);
            continue;
        }

        /* Get attribute name. */
        i = 0;
        while ((c = fgetc(fp)) != EOF) {
            if (c == '=') {
                break;
            } else if (c == '\n') {
                spdErrorExit("Invalid Info format.\n");
            }
            if (i < MAX_ATTR_STR_LEN - 1) {
                elem.attr[i++] = c;
            }
        }
        elem.attr[i] = '\0';

        /* Get value string. */
        i = 0;
        while ((c = fgetc(fp)) != EOF) {    
            if (c == '\n') {
                break;
            }
            if (i < MAX_VAL_STR_LEN - 1) {
                elem.val[i++] = c;
            }
        }
        elem.val[i] = '\0';
        if (c == EOF) {
            spdErrorExit("Invalid Info format.\n");
        }

        /* Read into "info->data". */
        if ((i = get_attr_num(info, elem.attr, 0)) == NOT_EXIST) {
            data->elem = spdRealloc(data->elem,
                                    sizeof(SpdInfoElem) * (data->size + 1));
            strcpy(data->elem[data->size].attr, elem.attr);
            strcpy(data->elem[data->size].val, elem.val);
            data->size++;
        } else {
            strcpy(data->elem[i].val, elem.val);
        }
    }
    spdFclose(fp);
}

/* check ե뤫SpdInfo */

SpdInfo *spdChkDocOpenInfo(SpdString checkFile)
{
    SpdInfo *info;
    SpdInfoData *data;

    info = spdCalloc(1, sizeof(SpdInfo));
    data = spdCalloc(1, sizeof(SpdInfoData));
    strcpy(info->filename, checkFile);
    info->data = data;
    info->data->size = 0;
    info->modified = FALSE;
    read_all_info(info);
    return (info);
}

static SpdBoolean getFlagInfoBool(SpdInfo *info, SpdString attr_name)
{
    SpdString val;

    val = spdGetInfoString(info, attr_name);
    if (val == NULL) {
        return FALSE;
    } else if (SPD_STREQ(val, ON_VAL)) {
        return TRUE;
    } else {
        return FALSE;
    }
}

void setCheckFlagSub(SpdInfo *info, Flag *flag, SpdString attr_src,
                     SpdString attr_doc, SpdString attr_detail)
{
    flag->src = getFlagInfoBool(info, attr_src);
    flag->doc = getFlagInfoBool(info, attr_doc);

    if (attr_detail != NULL) {              /* fileλNULL */
        flag->detail = getFlagInfoBool(info, attr_detail);
    } else {
        flag->detail = FALSE;
    }

    /* FALSEλϸʤ */
    if (!flag->src && !flag->doc && !flag->detail) {
        flag->main = FALSE;
    }
}


void printFlag(Flag flag)
{
    printf("flag main:%d src:%d doc:%d detail:%d\n",flag.main,flag.src,flag.doc,flag.detail);
}

/* SpdInfoflagꤹ */

void setCheckFlag(SpdInfo *info)
{
    /* optionͥ褹 */
    if (flag_f.main) {
        setCheckFlagSub(info, &flag_f, CHECK_FUNCSRC, CHECK_FUNCDOC, CHECK_FUNCDETAIL);
    }
    if (flag_t.main) {
        setCheckFlagSub(info, &flag_t, CHECK_TYPESRC, CHECK_TYPEDOC, CHECK_TYPEDETAIL);
    }
    if (flag_u.main) {
        setCheckFlagSub(info, &flag_u, CHECK_UNIONSRC, CHECK_UNIONDOC, CHECK_UNIONDETAIL);
    }
    if (flag_m.main) {
        setCheckFlagSub(info, &flag_m, CHECK_MACROSRC, CHECK_MACRODOC, CHECK_MACRODETAIL);
    }
    if (flag_e.main) {
        setCheckFlagSub(info, &flag_e, CHECK_ENUMSRC, CHECK_ENUMDOC, CHECK_ENUMDETAIL);
    }
    if (flag_file.main) {
        setCheckFlagSub(info, &flag_file, CHECK_FILESRC, CHECK_FILEDOC, NULL);
    }
}

/* SpdInfomodify flagꤹ */

void setModifyFlag(SpdInfo *info)
{
    SpdString val;
    /* Τ */

    if (modflag_f.main != NON_MODIFY) {
        val = spdGetInfoString(info, MODIFY_FUNC);
        if (val == NULL) {
            modflag_f.main = NON_MODIFY;
        } else if (SPD_STREQ(val,SRC)) {
            modflag_f.main = SRC_MODIFY;
        } else if (SPD_STREQ(val,DOC)) {
            modflag_f.main = DOC_MODIFY;
        } else {
            modflag_f.main = NON_MODIFY;
        }
        if (modflag_f.main != NON_MODIFY) {
            modflag_f.functype = getFlagInfoBool(info, MODIFY_FUNC_TYPE);
            modflag_f.argnum = getFlagInfoBool(info, MODIFY_FUNC_ARGNUM);
            modflag_f.argtype = getFlagInfoBool(info, MODIFY_FUNC_ARGTYPE);
            modflag_f.argname = getFlagInfoBool(info, MODIFY_FUNC_ARGNAME);
        }
        if (!modflag_f.functype && !modflag_f.argnum && !modflag_f.argtype && !modflag_f.argname) {
            modflag_f.main = NON_MODIFY;
        }
    }
    /* printf("main:%d functype:%d argnum:%d argtype:%d argname:%d\n",
     *      modflag_f.main, modflag_f.functype, modflag_f.argnum,
     *      modflag_f.argtype, modflag_f.argname);
     */
    if (modflag_t.main != NON_MODIFY) {
        val = spdGetInfoString(info, MODIFY_TYPE);
        if (val == NULL) {
            modflag_t.main = NON_MODIFY;
        } else if (SPD_STREQ(val,SRC)) {
            modflag_t.main = SRC_MODIFY;
        } else if (SPD_STREQ(val,DOC)) {
            modflag_t.main = DOC_MODIFY;
        } else {
            modflag_t.main = NON_MODIFY;
        }
        if (modflag_t.main != NON_MODIFY) {
            modflag_t.deftype = getFlagInfoBool(info, MODIFY_DEFTYPE);
            modflag_t.membernum = getFlagInfoBool(info, MODIFY_STRUCT_MEMBERNUM);
            modflag_t.membertype = getFlagInfoBool(info, MODIFY_STRUCT_MEMBERTYPE);
            modflag_t.membername = getFlagInfoBool(info, MODIFY_STRUCT_MEMBERNAME);
        }
        if (!modflag_t.deftype && !modflag_t.membernum &&
            !modflag_t.membertype && !modflag_t.membername) {
            modflag_t.main = NON_MODIFY;
        }
    }
    /* printf("main:%d deftype:%d membernum:%d membertype:%d membername:%d\n",
     *       modflag_t.main, modflag_t.deftype, modflag_t.membernum,
     *       modflag_t.membertype, modflag_t.membername);
     */
    if (modflag_u.main != NON_MODIFY) {
        val = spdGetInfoString(info, MODIFY_UNION);
        if (val == NULL) {
            modflag_u.main = NON_MODIFY;
        } else if (SPD_STREQ(val,SRC)) {
            modflag_u.main = SRC_MODIFY;
        } else if (SPD_STREQ(val,DOC)) {
            modflag_u.main = DOC_MODIFY;
        } else {
            modflag_u.main = NON_MODIFY;
        }
        if (modflag_u.main != NON_MODIFY) {
            modflag_u.membernum = getFlagInfoBool(info, MODIFY_UNION_MEMBERNUM);
            modflag_u.membertype = getFlagInfoBool(info, MODIFY_UNION_MEMBERTYPE);
            modflag_u.membername = getFlagInfoBool(info, MODIFY_UNION_MEMBERNAME);
        }
        if (!modflag_u.membernum && !modflag_u.membertype && !modflag_u.membername) {
            modflag_u.main = NON_MODIFY;
        }
    }
    /* printf("main:%d membernum:%d membertype:%d membername:%d\n",
     *        modflag_u.main, modflag_u.membernum,
     *        modflag_u.membertype, modflag_u.membername);
     */
    if (modflag_m.main != NON_MODIFY) {
        val = spdGetInfoString(info, MODIFY_MACRO);
        if (val == NULL) {
            modflag_m.main = NON_MODIFY;
        } else if (SPD_STREQ(val,SRC)) {
            modflag_m.main = SRC_MODIFY;
        } else if (SPD_STREQ(val,DOC)) {
            modflag_m.main = DOC_MODIFY;
        } else {
            modflag_m.main = NON_MODIFY;
        }
        if (modflag_m.main != NON_MODIFY) {
            modflag_m.argnum = getFlagInfoBool(info, MODIFY_MACRO_ARGNUM);
            modflag_m.argname = getFlagInfoBool(info, MODIFY_MACRO_ARGNAME);
        }
        if (!modflag_m.argnum && !modflag_m.argname) {
            modflag_m.main = NON_MODIFY;
        }

    }
    /* printf("main:%d argnum:%d argname:%d\n",
     *       modflag_m.main, modflag_m.argnum,modflag_m.argname);
     */
}
