/* 
 * Program:             $RCSfile: printDoc.c,v $  $Revision: 41.7 $
 * 
 * Purpose:             
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: printDoc.c,v 41.7 2002/10/21 07:28:22 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>
#include "chkDoc.h"
#include "printDoc.h"

SpdString convTextToHtml(SpdString str)
{
    int             i, j;
    static char     ret[BUF_MAX];

    j = 0;
    for (i = 0; i < strlen(str); i++) {
        switch (str[i]) {
          case '"':
            ret[j++] = '&';
            ret[j++] = 'q';
            ret[j++] = 'u';
            ret[j++] = 'o';
            ret[j++] = 't';
            ret[j++] = ';';
            break;
          case '<':
            ret[j++] = '&';
            ret[j++] = 'l';
            ret[j++] = 't';
            ret[j++] = ';';
            break;
          case '>':
            ret[j++] = '&';
            ret[j++] = 'g';
            ret[j++] = 't';
            ret[j++] = ';';
            break;
          case '&':
            ret[j++] = '&';
            ret[j++] = 'a';
            ret[j++] = 'm';
            ret[j++] = 'p';
            ret[j++] = ';';
            break;
          case ' ':
            ret[j++] = '&';
            ret[j++] = 'n';
            ret[j++] = 'b';
            ret[j++] = 's';
            ret[j++] = 'p';
            ret[j++] = ';';
            break;
          default:
            ret[j++] = str[i];
            break;
        }
    }
    if (BUF_MAX <= j) {
        spdErrorExit("Too long string.\n");
    }
    ret[j] = '\0';

    return (ret);
}

void fprintfTextfrag(FILE *fp, SpdObjId textfrag_id)
{
    int i;
    SpdRelArray text_array;
    
    text_array = spdGetRelSortedArray(textfrag_id, "textfrag_textstring",
                                      "textfrag_id", NULL, spdCompareRelOffset);
    for (i = 0; i < text_array.size; i++) {
        fprintf(fp, "%s", convTextToHtml((SpdString)spdGetAttrValString(text_array.rel[i].objectId,"text")));

        /* ǽȺǸʳϲԤ */
        /* 
         * if (i != 0 && i != text_array.size) { 
         *     fprintf(fp,"<br>\n");
         * } 
         */
    }
    spdFreeRelArray(text_array);
}

void fprintfIdentelem(FILE *fp, SpdObjId identelem_id)
{
    SpdObjId docident_id, defelem_id;

    docident_id = spdGetARelObj(identelem_id,"identelem_docident",
                                "identelem_id");
    defelem_id = spdGetARelObj(docident_id, "defelem_docident", "docident_id");
    
    if(defelem_id == SPD_NON_ID) {
        fprintf(fp, "%s",spdGetName(docident_id));
    } else {
        int sort = spdGetAttrValInt(docident_id, "sort");
        switch (sort) {
          case DID_FUNCTION:
            fprintf(fp, "<A HREF=\"func-detail.html#%d\" CLASS=\"Function\">%s()</A>",
                    docident_id,spdGetName(docident_id));
            break;
          case DID_TYPENAME:
            fprintf(fp, "<A HREF=\"type-detail.html#%d\" CLASS=\"Type\">%s</A>",
                    docident_id,spdGetName(docident_id));
            break;
          case DID_MACRONAME:
            fprintf(fp, "<A HREF=\"macro-detail.html#%d\" CLASS=\"Macro\">%s</A>",
                    docident_id,spdGetName(docident_id));
            break;
          case DID_FILE:
            fprintf(fp, "<A HREF=\"file-detail.html#%d\" CLASS=\"File\">%s</A>",
                    docident_id,spdGetName(docident_id));
            break;
          default:
            fprintf(fp, "<B>%s</B>",spdGetName(docident_id));
            break;
        }
    }
}

void fprintfSrcelem(FILE *fp, SpdObjId srcelem_id)
{
    SpdObjId decl_id;
    fprintf(fp, "<PRE>");
    decl_id = spdGetARelObj(srcelem_id,"srcelem_srcfrag",
                            "srcelem_id");
    spdSetOption("prototype", SPDSTR_PROTOTYPE_NON);
    spdSetOption("typeQualifier", ON);
    fprintf(fp, "%s", spdStr(decl_id));
    fprintf(fp, "</PRE>");
    spdStrFree();
}

void fprintfMacroelem(FILE *fp, SpdObjId docfrag_id)
{
    SpdObjId textfrag_id;
    
    textfrag_id = spdGetARelObj(docfrag_id, "docfrag_docelem", "p-id");
    fprintf(fp, "<PRE>\n");
    fprintfTextfrag(fp, textfrag_id);
    fprintf(fp, "</PRE>\n");

}

void fprintfDocfrag(FILE *fp, SpdObjId docfrag_id)
{
    SpdRelArray frag_array;
    int i;

    int sort = spdGetAttrValInt(docfrag_id, "sort");

    if(sort & DOCFRAG_TEXTFRAG){
        fprintfTextfrag(fp, docfrag_id);
    }else if(sort & DOCFRAG_IDENTELEM){
        fprintfIdentelem(fp, docfrag_id);
    }else if(sort & DOCFRAG_SRCELEM){
        fprintfSrcelem(fp, docfrag_id);
    }else if(sort & DOCFRAG_MACROELEM){
        fprintfMacroelem(fp, docfrag_id);
    }else if(sort & DOCFRAG_DOCELEM){
        frag_array = spdGetRelSortedArray(docfrag_id, "docelem_docfrag",
                                          "docelem_id", NULL, spdCompareRelOffset);
        for(i=0;i<frag_array.size;++i){
            fprintfDocfrag(fp, frag_array.rel[i].objectId);
        }
        spdFreeRelArray(frag_array);
    }
}
