/* 
 * Program:             $RCSfile: SIP2.h,v $  $Revision: 30.21 $
 * 
 * Purpose:
 * 
 * Author:              H.Suzuki  1998/11/06
 *                      S.Yamamoto  1998/11/06
 * 
 * (C) Copyright:       H.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/AreaModel.h>
#include <Sapid/AreaModelAR.h>

#define SIP_MAX_ARGC            16

/* 
 * SIP_MAX_PRINTF_ARGS should be less than NL_ARGMAX. 
 */
#define SIP_MAX_PRINTF_ARGS     7

typedef struct SipAreaInfo_t {
    SpdSet      defs;
    SpdSet      refs;
} SipAreaInfo;

typedef struct SipCallStackElem_t {
    SpdObjId    funcId;
    SpdObjId    prgPtr;
    SpdStack    *trvStack;
    SipAreaInfo areaInfo;
} SipCallStackElem;

typedef struct SipEvalStackElem_t {
    enum SipESType {
        SipES_Area = 0, SipES_Value, SipES_FuncSep, SipES_Fix
    } type;
    union {
        SpdObjId        areaId;
        SpdObjId        funcId;
        SpdValueData    value;
        SpdBoolean      isPrefix;
    } obj;
} SipEvalStackElem;

typedef struct SipStack_t {
    size_t      size;
    size_t      elemSize;
    void        **elem;
} SipStack;

typedef SpdSet SipFuncSet;
typedef int (* SipOpeFunc)(void);

/* extern SipFuncSet opeFuncs; */

/*** prototype ***/
/* eval.c */
SpdValueData sipEvalExpression(SpdObjId exprId);
SpdObjId sipGetExpressionArea(SpdObjId exprId);
void sipCallFunction(SpdObjId funcId);
void sipSetDebugMode(int onoff);
void sipSetProgramArgument(SpdObjId funcId, int argc, char *argv[]);

/* callstack.c */
void sipSetProgramPtr(SpdObjId prgPtr);
SpdObjId sipGetProgramPtr(void);
void sipPushCallStack(SpdObjId funcId);
SpdObjId sipGetCurrentFunctionId(void);
int sipGetDepthOfCallStack(void);
SpdObjId sipGetDecrFuncDef(SpdObjId funcId);
SipAreaInfo *sipGetCurrentAreaInfo(void);

/* step.c */
void sipGoStep(void);

/* system.c */
void sipSetOpeFunc(int opeSort, SipOpeFunc func);
void sipBegin(void);
void sipEnd(void);

/* libaccess.c */
SpdObjId sipGetFuncIdentId(SpdConstString funcname);

/******************************************************************************/

#define SIP_DEBUG_FLAG

#if defined(SIP_DEBUG_FLAG) || defined(lint)
#define SIP_DW_CONST        sipDwConst
#define SIP_DW_OPE          sipDwOpe
#else
#define SIP_DW_CONST        (void)
#define SIP_DW_OPE          (void)
#endif
