/* 
 * Program:             $RCSfile: callstack.c,v $  $Revision: 30.18 $
 * 
 * Purpose:
 * 
 * Author:              H.Suzuki  1998/11/06
 *                      S.Yamamoto  1998/11/06
 * 
 * (C) Copyright:       H.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: callstack.c,v 30.18 2003/07/08 05:48:29 yamamoto Exp yamamoto $";
#endif

#include "SIP2.h"
#include "SIP2-proto.h"

static SipStack *callStk;

/***** Areaؤ/Ⱦ *****/
SipAreaInfo *sipGetCurrentAreaInfo(void)
{
    SipCallStackElem *csElem = getTopOfCallStack();

    if (csElem == NULL) return NULL;
    else return (&csElem->areaInfo);
}

/* 륹å */
void sipCreateCallStack(void)
{
    callStk = sipCreateStack(sizeof(SipCallStackElem *));
}

/* 륹åβ */
void sipFreeCallStack(void)
{
    sipFreeStack(callStk);
}

/* 륹åǤκ */
static SipCallStackElem *sipCreateCallStackElem(SpdObjId funcId)
{
    SipCallStackElem *csElem;

    csElem = (SipCallStackElem *)spdMalloc(sizeof(SipCallStackElem));
    csElem->funcId = funcId;
    csElem->prgPtr = SPD_NON_ID;
    csElem->trvStack = spdTrvCreateStack();
    csElem->areaInfo.defs = spdCreateSet(DEFAULT_INT_SET);
    csElem->areaInfo.refs = spdCreateSet(DEFAULT_INT_SET);

    return csElem;
}

void printStorageStack(void)
{
    SpdObjId strgId, pId;

    strgId = spdGetCurrentStorage();
    printf("@c@");
    while (strgId != SPD_NON_ID) {
        switch (spdGetAttrIntValByName(strgId, "sort")) {
          case AREA_STORAGE_GLOBAL:
            pId = spdGetAParentObjId(strgId, SPD_RID_PROG_STORAGE);
            break;
          case AREA_STORAGE_FUNC:
            pId = spdGetAParentObjId(strgId, SPD_RID_FUNC_STORAGE);
          case AREA_STORAGE_LOCAL:
          case AREA_STORAGE_HEAP:
          default:
            pId = spdGetAParentObjId(strgId, SPD_RID_BLOCK_STORAGE);
            break;
        }
        printf("<%d>", pId);
        strgId = spdGetAParentObjId(strgId, SPD_RID_STORAGE_STORAGE);
    }
    printf("\n");
}

/* ѿΤ֤storage򤹤٤Ƴ */
static void createLocalStorages(SpdObjId blockId)
{
    SpdSet2 childBlocks;
    SpdObjId pblockId;
    SetCursor csr;

    if (spdGetAttrIntValByName(blockId, "sort") == BLOCK_COMP) {
        spdEnableStorage(spdCreateStorageArea(blockId));
        if (sipGetDebugMode()) {
            printf("@@CreateStorage");printStorageStack();
        }
    }

    childBlocks = spdGetChildObjIdSet(blockId, SPD_RID_BLK_BLK);
    for (csr = spdSetCursorCreate(childBlocks); spdSetCursorNext(&csr);) {
        pblockId = spdSetGetValueO(csr.curElem->value);
        createLocalStorages(pblockId);
    }
    spdFreeSet2(childBlocks);
}

/* 륹å˴ؿץå夹 */
void sipPushCallStack(SpdObjId funcId)
{
    SpdObjId optDeclId, bodyId;

    sipPushStack(sipCreateCallStackElem(funcId), callStk);

    /* Τ֤storage */
    optDeclId = spdGetAChildObjId(sipGetDecrFuncDef(funcId),
                                  SPD_RID_DECL_OPT);
    spdEnableStorage(spdCreateStorageArea(optDeclId));

    /* ѿΤ֤storage򤹤٤Ƴ */
    bodyId = spdGetAChildObjId(optDeclId, SPD_RID_FUNC_BODY);
    createLocalStorages(bodyId);

    sipSetTrvStackOfCallStack();

    if (sipGetDebugMode()) {
        printf("@@Push");printStorageStack();
    }
}

/* 륹åݥåפ */
SipCallStackElem *sipPopCallStack(void)
{
    while(spdGetCurrentStorage() != SPD_NON_ID)
        if (spdGetAttrIntValByName(spdDisableStorage(), "sort")
            == AREA_STORAGE_FUNC)
            break;

    if (sipGetDebugMode()) {
        printf("@@Pop ");printStorageStack();
    }

    return ((SipCallStackElem *)sipPopStack(callStk));
}

/* 륹åƬǤ(ݥåפʤ) */
SipCallStackElem *getTopOfCallStack(void)
{
    return ((SipCallStackElem *)sipGetTopOfStack(callStk));
}

void sipFreePoppedCallStack(void)
{
    SipCallStackElem *csElem = sipPopCallStack();
    sipFreeStack((SipStack *)csElem->trvStack);
    spdFree(csElem);
}

/* 륹åƬǤ Function_id  */
SpdObjId sipGetCurrentFunctionId(void)
{
    SipCallStackElem *csElem = getTopOfCallStack();

    if (csElem == NULL) return SPD_NON_ID;
    else return (csElem->funcId);
}

/* 륹åǿ() */
int sipGetDepthOfCallStack(void)
{
    return (callStk->size);
}

/* ؿ̾ؿ declarator  */
SpdObjId sipGetDecrFuncDef(SpdObjId funcId)
{
    SpdIdSet decrs;
    SpdObjId decrId;
    SetCursor csr;

    if (spdGetAttrIntValByName(funcId, "sort") == ID_FUNCTION) {
        decrs = spdGetParentObjIdSet(funcId, SPD_RID_DECL_IDENT);
        for (csr = spdSetCursorCreate(decrs); spdSetCursorNext(&csr);) {
            decrId = spdSetGetValueO(csr.curElem->value);
            if (spdGetAttrIntValByName(decrId, "sort") == DECL_FUNCDEF)
                return decrId;
        }
        spdFreeSet2(decrs);
    }

    return SPD_NON_ID;
}

/* ؿ̾¹Ԥ٤Ƭμ */
static SpdObjId sipGetFuncTopExpression(SpdObjId funcId)
{
    SpdObjId decrId;

    if ((decrId = sipGetDecrFuncDef(funcId)) != SPD_NON_ID)
        return (sipGetExpressionFromFuncDecr(decrId));
    else return SPD_NON_ID;
}

/* ͡ objId б exprId  */
static SpdObjId sipGetExpression(SpdObjId id)
{
    switch (spdGetClassId(id)) {
      case SPD_CID_IDENTIFIER:
        if (spdGetAttrIntValByName(id, "sort") == ID_FUNCTION)
            return (sipGetFuncTopExpression(id));
        else return SPD_NON_ID;
      default:
        return id;
    }
}

/* Set program pointer. */

void sipSetProgramPtr(SpdObjId prgPtr)
{
    SipCallStackElem    *top;

    /* objIdбexprId롥 */
    prgPtr = sipGetExpression(prgPtr);

    top = (SipCallStackElem *)sipGetTopOfStack(callStk);
    if (top != NULL) {
        top->prgPtr = prgPtr;
    }
}

SpdObjId sipGetProgramPtr(void)
{
    SipCallStackElem    *top;

    top = (SipCallStackElem *)sipGetTopOfStack(callStk);
    /* 
     * if (top == NULL) return SPD_NON_ID;
     * else return (top->prgPtr);
     */
    return ((top == NULL) ? SPD_NON_ID : top->prgPtr);
}
