/* 
 * Program:             $RCSfile: eval.c,v $  $Revision: 30.33 $
 * 
 * Purpose:
 * 
 * Author:              H.Suzuki  1998/11/06
 *                      S.Yamamoto  1998/11/06
 * 
 * (C) Copyright:       H.Suzuki and S.Yamamoto  1998 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: eval.c,v 30.33 2003/07/08 05:48:29 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <stdarg.h>
#include "SIP2.h"
#include "SIP2-proto.h"

SpdValueData spdCreateValueDataFromConstant(SpdObjId constId);

static void printEvalStack(void);
static void sipDwConst(const char *format, ...);
static void sipDwFuncCall(const char *format, ...);
static void sipDwFuncReturn(const char *format, ...);
static void sipDwIdent(const char *format, ...);
static void sipDwLibCall(const char *format, ...);
static void sipDwLibReturn(const char *format, ...);

static SipStack *evalStk = NULL;
static SpdTrvActInfo *info = NULL;

static int debugMode = OFF;
static int debugLevel = 0;

/* Evaluation stack. **********************************************************/

SipStack *sipCreateEvalStack(void)
{
    if (evalStk == NULL) {
        evalStk = sipCreateStack(sizeof(SipEvalStackElem *));
    }
    return (evalStk);
}

void sipFreeEvalStack(void)
{
    sipFreeStack(evalStk);
}

void sipPushEvalStack(SipEvalStackElem *esElem)
{
    if (esElem != NULL) {
        sipPushStack(esElem, evalStk);
    }
}

SipEvalStackElem *sipPopEvalStack(void)
{
    return ((SipEvalStackElem *)sipPopStack(evalStk));
}

SipEvalStackElem *sipGetTopOfEvalStack(void)
{
    return ((SipEvalStackElem *)sipGetTopOfStack(evalStk));
}

SipEvalStackElem *sipDupEvalStackElem(const SipEvalStackElem *elem)
{
    SipEvalStackElem    *dupElem;

    if (elem == NULL) {
        return NULL;
    }
    dupElem = sipCreateEvalStackElem();
    dupElem->type = elem->type;
    switch (elem->type) {
      case SipES_Area:
        dupElem->obj.areaId = elem->obj.areaId;
        break;
      case SipES_Value:
        dupElem->obj.value = elem->obj.value;
        break;
      case SipES_FuncSep:
        dupElem->obj.funcId = elem->obj.funcId;
        break;
      case SipES_Fix:
        dupElem->obj.isPrefix = elem->obj.isPrefix;
        break;
      default:
        spdErrorExit("sipDupEvalStackElem(): Invalid type %d.\n", elem->type);
        break;
    }
    return (dupElem);
}

/* spdGetRelSortedArrayѸȽؿ */

static SpdBoolean spdRelIsDeclParameter(const SpdRel rel)
{
    return (spdObjIdIsDeclParameter(rel.objectId));
}

/* ɾå˴ؿƽФѤ롥 */

static void sipCallFunc(SpdObjId funcId)
{
    SpdObjId            decrId, optDeclId, prmDecrId, identId, areaId;
    SpdRelArray         relArr;
    SipEvalStackElem    *esElem;
    SpdValueData        val;
    int                 i;

    /* 桼ؿݤȽꤹ롥 */
    if ((decrId = sipGetDecrFuncDef(funcId)) == SPD_NON_ID) {
        return;
    }

    /* 륹åѤࡥ */
    sipPushCallStack(funcId);

    optDeclId = spdGetAChildObjId(decrId, SPD_RID_DECL_OPT);
    relArr = spdGetRelSortedArray(optDeclId, "func_parm", "optionaldecl_id",
                                  spdRelIsDeclParameter,
                                  spdCompareRelOffset);

    sipDwFuncCall("[%s]\n", spdGetName(funcId));    
    debugLevel++;

    /* οݥåפ *//* !! Ĺΰб뤳ȡ */
    for (i = 0;
         i < relArr.size && sipGetTopOfEvalStack()->type != SipES_FuncSep;
         i++) {
        /* ѿ롥 */
        prmDecrId = spdGetAChildObjId(relArr.rel[i].objectId,
                                      SPD_RID_DECL_DECL);
        identId = spdGetAChildObjId(prmDecrId, SPD_RID_DECL_IDENT);
        areaId = spdGetVariableArea(identId);

        esElem = sipPopEvalStack();
        val = sipGetValueDataFromESElem(esElem);
        spdFree(esElem);
        spdSetValueData(areaId, val);
    }
    /* !! Ĥäͤݥåפ(ʤȤϤꤨʤ) */
    while (sipGetTopOfEvalStack()->type != SipES_FuncSep) {
        spdFree(sipPopEvalStack());
    }

    esElem = sipGetTopOfEvalStack();
    if (esElem->type == SipES_FuncSep) {
        esElem->obj.funcId = funcId;
    }
    spdFreeRelArray(relArr);

    sipSetProgramPtr(sipGetExpressionFromFuncDecr(decrId));
}

/* ؿƽФ(󥿡եؿ) */

void sipCallFunction(SpdObjId funcId)
{
    sipCallFunc(funcId);
}

/* main()ƤνԤʤ */

void sipSetProgramArgument(SpdObjId funcId, int argc, char *argv[])
{
    SipEvalStackElem *funcElem, *argcElem, *argvElem;
    SpdObjId declId, optId, parmDeclId, parmDecrId, parmId;
    SpdObjId pAreaId, cAreaId, stringId;
    SpdObjId argvId = SPD_NON_ID, typeId;
    SpdIdSet parmDecls;
    SpdValueData val;
    int i;
    SetCursor csr;

    /* SipES_FuncSepѤࡥ */
    funcElem = sipCreateEvalStackElem();
    funcElem->type = SipES_FuncSep;
    funcElem->obj.funcId = funcId;
    sipPushEvalStack(funcElem);

    declId = sipGetDecrFuncDef(funcId);
    optId = spdGetAChildObjId(declId, SPD_RID_DECL_OPT);
    parmDecls = spdGetChildObjIdSet(optId, SPD_RID_FUNC_PARM);
    for (csr = spdSetCursorCreate(parmDecls); spdSetCursorNext(&csr);) {
        parmDeclId = spdSetGetValueO(csr.curElem->value);
        parmDecrId = spdGetAChildObjId(parmDeclId, SPD_RID_DECL_DECL);
        parmId = spdGetAChildObjId(parmDecrId, SPD_RID_DECL_IDENT);
        if (SPD_STREQ(spdGetName(parmId), "argv")) {
            argvId = parmId;
            break;
        }
    }
    spdFreeSet2(parmDecls);

    if (argvId != SPD_NON_ID) {
        /* argc */
        argcElem = sipCreateEvalStackElem();
        argcElem->type = SipES_Value;
        argcElem->obj.value.type = SpdValue_Integer;
        argcElem->obj.value.v.intval = argc;

        /* char *argv[]Areaargcο */
        typeId = spdGetAChildObjId(spdGetAChildObjId(argvId,
                                                     SPD_RID_IDENT_TYPE),
                                   SPD_RID_TYPE_ST);
        pAreaId = spdCreateArrayArea(typeId, argc);

        /* οAreaidargvͤȤ롥 */
        argvElem = sipCreateEvalStackElem();
        argvElem->type = SipES_Value;
        argvElem->obj.value.type = SpdValue_Pointer;
        argvElem->obj.value.v.ptr.areaId = pAreaId;
        argvElem->obj.value.v.ptr.offset = 0; /* !! 0Ǥ? */

        /* 
         * argv[0], ... String Area,
         * areaIdǤγareavalueȤƥåȤ롥
         */
        for (i = 0; i < argc; i++) {
            /* !! 2SPD_NON_IDǤ褤ʬʤ */
            stringId = spdCreateStringArea(argv[i], SPD_NON_ID);
            val.type = SpdValue_Pointer;
            val.v.ptr.areaId = stringId;
            val.v.ptr.offset = 0;           /* !! 0Ǥ? */

            cAreaId = spdGetArrayElementArea(pAreaId, i);
            spdSetValueData(cAreaId, val);
        }

        /* Ѥࡥ */
        sipPushEvalStack(argvElem);
        sipPushEvalStack(argcElem);
    }
}

/* ؿƽФФƤ빽Ϣexpr_exprȽǤԤʤ */

static SpdBoolean isFromFuncCallExpr(const SpdRel *rel)
{
    SpdObjId    pexprId;

    if (rel == NULL || spdGetOffset(rel->relationId) < 0) {
        /* def_use ʤɹطʳΤΤéäƤޤȡpexprId 
           ؿƤӽФˤʤ뤳Ȥ뤿ᡢط˸ꤷȽꤹ롣 */
        return FALSE;
    }

    pexprId = spdGetAParentObjId(rel->objectId, SPD_RID_EXPR_EXPR);
    if (pexprId == SPD_NON_ID) {
        return (FALSE);
    } else if (spdGetAttrIntValByName(pexprId, "sort") == EXPR_SORT_FUNCCALL) {
        return (TRUE);
    } else {
        return (FALSE);
    }
}

static SpdBoolean isNormalExprOpeRel(const SpdRel *rel)
{
    int     opeSort;

    if (spdGetClassId(rel->relationId) == SPD_RID_EXPR_OP) {
        opeSort = spdGetAttrIntValByName(rel->objectId, "sort");
        if (opeSort != OP_L_AND && opeSort != OP_L_OR) {
            return (TRUE);
        }
    }
    return (FALSE);
}

/* ؿƽФexpressionƬݤȽǤ롥 */

static SpdBoolean isLeftOfFuncCallExprs(SpdObjId cExprId, SpdObjId pExprId)
{
    SpdRelArray exprs;
    SpdBoolean ret;

    exprs = spdGetRelSortedArray(pExprId, "expr_expr", "p-id",
                                 NULL, spdCompareRelOffset);
    if (cExprId == exprs.rel[0].objectId) {
        ret = TRUE;
    } else {
        ret = FALSE;
    }
    spdFreeRelArray(exprs);
    return (ret);
}

/* Area  externƤ뤫Ƚ */

static SpdBoolean isExternDeclArea(SpdObjId areaId)
{
    SpdObjId identId, decrId, declId;
    SpdIdSet decrs;
    SpdBoolean res = TRUE;
    SetCursor csr;

    identId = spdGetAParentObjId(areaId, SPD_RID_VARIABLE_AREA);
    decrs = spdGetParentObjIdSet(identId, SPD_RID_DECL_IDENT);
    for (csr = spdSetCursorCreate(decrs); spdSetCursorNext(&csr);) {
        decrId = spdSetGetValueO(csr.curElem->value);
        declId = spdGetAParentObjId(decrId, SPD_RID_DECL_DECL);
        if (spdGetAttrIntValByName(declId, "storage") != STORAGE_EXTERN) {
            res = FALSE;
            break;
        }
    }
    spdFreeSet2(decrs);

    if (res && sipGetDebugMode()) {
        SpdConstString nm;

        if ((nm = spdGetName(identId)) != NULL)
            printf("variable:%s is extern.\n", nm);
        else
            printf("variable:??? is extern.\n");
    }
    
    return res;
}

/* spdTraverse()SpdRelؿ */
/* 
 * expression ϢéȤ黻(expr_op)éꡤ
 * ؿƽФϡեեåȽé褦ˤ롥
 * 黻(&&, ||)ϥեåȽé롥
 */

static int sipCompareRel(const SpdRel *a, const SpdRel *b)
{
    if (isNormalExprOpeRel(a)) {
        return (1);
    } else if (isNormalExprOpeRel(b)) {
        return (-1);
    } else if (isFromFuncCallExpr(a) || isFromFuncCallExpr(b)) {
        return (-spdCompareRelOffset(a, b));
    } else {
        return (spdCompareRelOffset(a, b));
    }
}

/* spdTraverse()Actؿ ***************************************************/

static int sipBeforeExpr(SpdObjId exprId, void *vp)
{
    SipEvalStackElem    *esElem;
    SpdRelArray         cexprs;
    SpdValueData        val;

    switch (smuGetExprSort(exprId)) {
      case EXPR_SORT_FUNCCALL:
        /* SipES_FuncSepEvalStackѤࡥ */
        esElem = sipCreateEvalStackElem();
        esElem->type = SipES_FuncSep;
        esElem->obj.funcId = SPD_NON_ID;
        sipPushEvalStack(esElem);
        if (debugMode == ON) printEvalStack();
        break;
      case EXPR_SORT_TRI:
        if (debugMode == ON) {
            fprintf(stderr, "sipBeforeExpr(): Tripple Expression.\n");
        }
        cexprs = spdGetRelSortedArray(exprId, "expr_expr", "p-id",
                                      NULL, spdCompareRelOffset);

        /* ǽΥڥɤé롥 */
        spdTraverse(cexprs.rel[0].objectId, info, NULL);
        esElem = sipPopEvalStack();
        val = sipGetValueDataFromESElem(esElem);
        spdFree(esElem);

        /* ǽΥڥɤͤˤäéʬ */
        if (spdCastValueDataInt(val) != 0) {
            spdTraverse(cexprs.rel[1].objectId, info, NULL);
        } else {
            spdTraverse(cexprs.rel[2].objectId, info, NULL);
        }

        spdFreeRelArray(cexprs);

        /* Ｐʲéʤ */
        return (SAPID_TRV_STOP);
    }

    return (SAPID_TRV_GO);
}

static int sipAfterExprExpr(SpdObjId exprId, SpdRel rel, void *vp)
{
    SipEvalStackElem    *elem;
    SpdObjId            funcId;
    SipOpeFunc          libAcsFunc;
    int                 ret;

    if (spdGetAttrIntValByName(exprId, "sort") == EXPR_SORT_FUNCCALL &&
        isLeftOfFuncCallExprs(rel.objectId, exprId)) {
        /* ؿƽФƬexpression. */
        elem = sipPopEvalStack();
        funcId = elem->obj.value.v.funcId;
        /* 桼ؿݤȽꤹ롥 */
        if (sipGetDecrFuncDef(funcId) != SPD_NON_ID) {
            /* 
             * sipDwFuncCall("[%s]\n", spdGetName(funcId));    
             * debugLevel++;
             */
            sipCallFunc(funcId);
            ret = SAPID_TRV_PAUSE;
        } else {
            sipDwLibCall("[%s]\n", spdGetName(funcId));    
            libAcsFunc = sipGetlibAcsFunc(funcId);
            if (libAcsFunc) {
                ret = libAcsFunc();
            } else {
                spdErrorExit("\"%s\": Undefined library function.\n",
                             spdGetName(funcId));
            }
            sipDwLibReturn("[%s]\n", spdGetName(funcId));    
        }
    } else {
        ret = SAPID_TRV_GO;
    }
    return (ret);
}

static int sipBeforeExprOpe(SpdObjId exprId, SpdRel opeRel, void *vp)
{
    SpdRel              exprRel;
    SipEvalStackElem    *esElem;
    int                 sort;

    sort = spdGetAttrIntValByName(opeRel.objectId, "sort");
    if (sort == OP_INC || sort == OP_DEC) {
        /* 
         * 󥯥(ǥ)黻Ҥéˡ
         * or֤ξ򥹥åѤࡥ
         */
        esElem = sipCreateEvalStackElem();
        esElem->type = SipES_Fix;
        exprRel = spdGetAChildRel(exprId, SPD_RID_EXPR_EXPR);
        /* 
         * if (spdGetOffset(exprRel.relationId) > spdGetOffset(opeRel.relationId))
         *     esElem->obj.isPrefix = TRUE;
         * else
         *     esElem->obj.isPrefix = FALSE;
         */
        esElem->obj.isPrefix = 
            (spdGetOffset(exprRel.relationId) > spdGetOffset(opeRel.relationId));
        sipPushEvalStack(esElem);
        if (debugMode == ON) printEvalStack();
    }

    return (SAPID_TRV_GO);
}

static int sipBeforeOpe(SpdObjId opeId, void *vp)
{
    SpdString   astr, str;
    SipOpeFunc  opeFunc;
    int         ret;

    smuGetObjAttr(opeId, "sort", &astr, &str);
    /* printf("eval:operator  : %s\n", str); */
    opeFunc = sipGetOpeFunc(spdGetAttrIntValByName(opeId, "sort"));
    if (opeFunc) {
        ret = opeFunc();
    } else {
        spdErrorExit("Undefined callback function for operation \"%s\".\n",
                     str);
    }
    if (debugMode == ON) printEvalStack();
    return (ret);
}

static int sipAfterExpr(SpdObjId exprId, void *vp)
{
    SipEvalStackElem    *elem, *retElem;

    switch (spdGetAttrIntValByName(exprId, "sort")) {
      case EXPR_SORT_FUNCCALL:
        /* 
         * if (debugMode == ON) {
         *     spdDwIndent();
         *     fprintf(stderr, "sipAfterExpr(): Function exit.\n");
         * }
         */
        /* SipES_FuncSepEvalStackФ */
        elem = sipPopEvalStack();
        if (elem->type == SipES_FuncSep) {
            spdFree(elem);
        } else {
            /* ͤ롥 */
            retElem = elem;
            elem = sipPopEvalStack();
            if (elem != NULL && elem->type == SipES_FuncSep) {
                spdFree(elem);
            }
            sipPushEvalStack(retElem);
        }
        break;
      case EXPR_RETURN:
        sipDwFuncReturn("\n");
        if (sipGetTopOfEvalStack()->type != SipES_FuncSep) {
            elem = sipPopEvalStack();
            retElem = sipCreateEvalStackElem();
            retElem->type = SipES_Value;
            retElem->obj.value = sipGetValueDataFromESElem(elem);
            spdFree(elem);
            sipPushEvalStack(retElem);
            if (debugMode == ON) printEvalStack();
        }
        debugLevel--;
        break;
        /* Temporarily added by S.Yamamoto 2001/03/15. */
      default:
        {
            SpdObjId        opeId = spdGetARelObj(exprId, "expr_op", "p-id");
            SpdValueData    val;

            if (spdGetAttrValInt(opeId, "sort") == OP_L_AND ||
                spdGetAttrValInt(opeId, "sort") == OP_L_OR) {
                elem = sipPopEvalStack();
                val = sipGetValueDataFromESElem(elem);
                spdFree(elem);
                retElem = sipCreateEvalStackElem();
                retElem->type = SipES_Value;
                retElem->obj.value.type = SpdValue_Integer;
                retElem->obj.value.v.intval =
                    (spdCastValueDataInt(val) == 0) ? 0 : 1;
                sipPushEvalStack(retElem);
                if (debugMode == ON && spdCastValueDataInt(val) != 0 &&
                    spdCastValueDataInt(val) != 1) {
                    /* fprintf(stderr, "sipAfterExpr(): Fix value.\n"); */
                    SIP_DW_OPE("[&&]: sipAfterExpr(): Fix value: %d\n",
                               spdCastValueDataInt(val));
                }
            }
        }
        break;
    }
    return (SAPID_TRV_GO);
}

/******************************************************************************/

static int sipBeforeIdent(SpdObjId identId, void *vp)
{
    SipEvalStackElem    *esElem;
    SpdObjId            areaId;
    SpdValueData        value;

    switch (spdGetAttrIntValByName(identId, "sort")) {
      case ID_FUNCTION:
        value = spdGetFuncNameValueData(identId);
        if (value.type == SpdValue_FuncPointer) {
            esElem = sipCreateEvalStackElem();
            esElem->type = SipES_Value;
            esElem->obj.value = value;
            sipPushEvalStack(esElem);
        }
        break;
      case ID_VARIABLE:
        areaId = spdGetVariableArea(identId);
        if (isExternDeclArea(areaId)) {
            /* !! ѿɽ褦ˤ롥 */
            spdSetAttrIntValByName(areaId, "indirect", AREA_DIRECT);
            spdSetAttrIntValByName(areaId, "address", (int)stdin);
        }
        esElem = sipCreateEvalStackElem();
        esElem->type = SipES_Area;
        esElem->obj.areaId = areaId;
        sipPushStack(esElem, evalStk);
        break;
      default:
        break;
    }
    sipDwIdent("[%s]\n", spdGetName(identId));
    return (SAPID_TRV_GO);
}

static int sipBeforeConst(SpdObjId constId, void *vp)
{
    SipEvalStackElem    *esElem = sipCreateEvalStackElem();

    esElem->type = SipES_Value;
    esElem->obj.value = spdCreateValueDataFromConstant(constId);
    sipPushEvalStack(esElem);
    SIP_DW_CONST("[%s]\n", spdGetAttrValString(constId, "value"));
    return (SAPID_TRV_GO);
}

/* Function for SpdTrvInfo. ***************************************************/

void initSpdTrvInfo(void)
{
    info = spdTrvInitActInfo();

    /* SpdRelӴؿϿ롥 */
    spdTrvAddSortFunc(info, SPD_CID_EXPRESSION, sipCompareRel);

    /* Add relationships which are traversed. */
    spdTrvAddRel(info, SPD_RID_EXPR_EXPR);
    spdTrvAddRel(info, SPD_RID_EXPR_OP);
    spdTrvAddRel(info, SPD_RID_CONST_REF);
    spdTrvAddRel(info, SPD_RID_IDENT_REF);

    /* Add callback functions for relationships. */
    /* Action for some kind of functin call. */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_EXPR_EXPR,
                               SAPID_TRV_ACT_AFTER, sipAfterExprExpr);
    /* Action for "++" and "--". */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_EXPR_OP,
                               SAPID_TRV_ACT_BEFORE, sipBeforeExprOpe);

    /* Add callback functions for classes. */
    /* Action for some kind of functin call and "?:". */
    spdTrvAddActFuncOfClass(info, SPD_CID_EXPRESSION,
                            SAPID_TRV_ACT_BEFORE, sipBeforeExpr);
    /* Action for some kind of functin call and "return". */
    spdTrvAddActFuncOfClass(info, SPD_CID_EXPRESSION,
                            SAPID_TRV_ACT_AFTER, sipAfterExpr);
    /* Apply the operator to its operands. */
    spdTrvAddActFuncOfClass(info, SPD_CID_OPERATOR,
                            SAPID_TRV_ACT_BEFORE, sipBeforeOpe);
    /* Get its value. */
    spdTrvAddActFuncOfClass(info, SPD_CID_IDENTIFIER,
                            SAPID_TRV_ACT_BEFORE, sipBeforeIdent);
    /* Get its value. */
    spdTrvAddActFuncOfClass(info, SPD_CID_CONSTANT,
                            SAPID_TRV_ACT_BEFORE, sipBeforeConst);

    /* trvStackʤ褦ˤ롥 */
    /* !! ñ˥ץ¹ԤȤfree줺˻ĤäƤ롥 */
    /* 
     * sipFreeStack((SipStack)info->trvStack);
     * info->trvStack = NULL;
     */
}

void freeSpdTrvInfo(void)
{
    spdTrvFreeActInfo(info);
}

void sipSetTrvStackOfCallStack(void)
{
    SipCallStackElem *csElem = getTopOfCallStack();

    if (csElem != NULL) {
        info->trvStack = csElem->trvStack;
    } else {
        /* !! ⲽäݤΤǡtrvStackδ򤷤äꤹ */
        info->trvStack = spdTrvCreateStack();
    }
}

/* Expression evalution. ******************************************************/

static void evalExpression(SpdObjId exprId)
{
    SipCallStackElem *csElem;

    if (info->trvStack->size > 0) {
        spdTraverseContinue(info, NULL);
    } else {
        csElem = getTopOfCallStack();
        spdFreeSet(csElem->areaInfo.defs);
        spdFreeSet(csElem->areaInfo.refs);
        csElem->areaInfo.defs = spdCreateSet(DEFAULT_INT_SET);
        csElem->areaInfo.refs = spdCreateSet(DEFAULT_INT_SET);
        spdTraverse(exprId, info, NULL);
    }
}

/* expressionɾͤ롥 */
/* ֤ͤʤ(break, continue, return, ...)ϡNon Value֤ */

SpdValueData sipEvalExpression(SpdObjId exprId)
{
    SipEvalStackElem *elem;
    SpdValueData val;

    evalExpression(exprId);
    if (spdGetAttrIntValByName(exprId, "sort") == EXPR_RETURN) {
        /* returnʸΤȤpopʤ */
        return (sipGetValueDataFromESElem(NULL));
    } else {
        elem = (SipEvalStackElem *)sipGetTopOfStack(evalStk);
        if (elem != NULL) {
            if (elem->type != SipES_FuncSep) {
                elem = sipPopEvalStack();
                val = sipGetValueDataFromESElem(elem);
                spdFree(elem);
                return (val);
            } else {
                return (sipGetValueDataFromESElem(NULL));
            }
        } else {
            return (sipGetValueDataFromESElem(NULL));
        }
    }
}

/* expressionɾΰ롥 */

SpdObjId sipGetExpressionArea(SpdObjId exprId)
{
    SipEvalStackElem    *esElem;

    if (spdGetClassId(exprId) != SPD_CID_EXPRESSION) {
        return SPD_NON_ID;
    }
    evalExpression(exprId);
    esElem = sipPopEvalStack();
    if (esElem->type == SipES_Area) {
        return (esElem->obj.areaId);
    } else {
        return (SPD_NON_ID);
    }
}

/* StringAreaidʸ롥 */
/* !! ¤ƤAreaΰݤƤΥɥ쥹֤褦ˤ */

SpdString sipGetString(SpdObjId pAreaId)
{
    SpdObjId areaId;
    SpdIdSet areas;
    SpdValueData val;
    SpdString str;
    int i;

    areas = spdGetChildObjIdSet(pAreaId, SPD_RID_AREA_AREA);
    str = (SpdString)spdMalloc(sizeof(char) * areas->elemNum);
    for (i = 0; i < areas->elemNum; i++) {
        areaId = spdGetArrayElementArea(pAreaId, i);
        /* !! 2ϤȤꤢ0ˤƤ */
        val = spdGetValueData(areaId, 0);
        str[i] = val.v.intval;
    }
    spdFreeSet2(areas);

    return str;
}

SpdValueData sipGetValueDataFromESElem(SipEvalStackElem *elem)
{
    SpdValueData val;
    SpdObjId areaId;
    SipCallStackElem *csElem;

    if (elem != NULL) {
        switch (elem->type) {
          case SipES_Area:
            areaId = elem->obj.areaId;

            /* Area Info */
            csElem = getTopOfCallStack();
            spdAddElem(csElem->areaInfo.refs, (int)areaId, (Contents)&areaId);

            /* !! TYPE x[] ݥ󥿤ȤƼޤǤΰŪʽ
             * if (spdGetAttrIntValByName(areaId, "sort") != AREA_BASIC) {
             */
            if (spdGetAttrIntValByName(areaId, "sort") != AREA_BASIC
                && spdGetChildObjIdSet(areaId, SPD_RID_AREA_AREA)
                ->elemNum > 0) {
                val.type = SpdValue_Pointer;
                val.v.ptr.areaId = areaId;
                val.v.ptr.offset = 0;       /* !! Ȥꤢ0 */
                return (val);
            } else {
                /* !! 2ϤȤꤢ0ˤƤ */
                return (spdGetValueData(areaId, 0));
            }
          case SipES_Value:
            return (elem->obj.value);
          default:
            break;
        }
    }

    /* ʳ֤ */
    val.type = SpdValue_NonValue;
    val.v.intval = -1;
    return (val);
}

/* For AreaModel. *************************************************************/

SpdValueData spdCreateValueDataFromConstant(SpdObjId constId)
{
    SpdObjId        pAreaId;
    SpdValueData    retVal;
    SpdConstString  constVal, p;
    int             charVal, c, d, cnt;

    constVal = spdGetAttrValString(constId, "value");
    switch (spdGetAttrIntValByName(constId, "type")) {
      case CONST_INT:
        retVal.type = SpdValue_Integer;
        if (constVal[0] == '0') {
            if (constVal[1] == 'x' || constVal[1] == 'X')
                /* 16 */
                retVal.v.intval = strtol(constVal, NULL, 16);
            else
                /* 8 */
                retVal.v.intval = strtol(constVal, NULL, 8);
        } else {
            retVal.v.intval = strtol(constVal, NULL, 10);
        }
        break;
      case CONST_CHAR:
        p = constVal;
        /* !! ʣʸʤʸ̤ݡ */
        /* Ƭ ' ɤФ */
        while(*p == '\'') p++;
        charVal = *p;
        /* ץ󥹤Υå */
        if (charVal == '\\') {
            charVal = *(++p);
            if (isalpha(charVal)) {
                /* ʸץ */
                switch (charVal) {
                  case 'a': charVal = '\a'; break;
                  case 'b': charVal = '\b'; break;
                  case 'f': charVal = '\f'; break;
                  case 'n': charVal = '\n'; break;
                  case 'r': charVal = '\r'; break;
                  case 't': charVal = '\t'; break;
                  case 'v': charVal = '\v'; break;
                  case 'x':
                    /* 16ʥץ */
                    charVal = 0;
                    if (isxdigit(*(p + 1))) { /* \xμʸå */
                        c = 0;
                        /* 3ʾϺǸΣͤȤ */
                        while (isxdigit(d = *(++p))) {
                            charVal = c * 16;
                            if (isdigit(d)) c = d - '0';
                            else c = tolower(d) - 'a' + 10;
                            charVal += c;
                        }
                    } else spdErrorExit("\\xθ16ʿʤ\n");
                    break;
                  default:
                    /* ¾(\\ , \' , ...) \ μʸͤȤ */
                    break;
                }
            } else if (isdigit(charVal)) {
                /* 8ʥץ */
                charVal -= '0';
                cnt = 1;
                while (isdigit(d = *(++p))) {
                    if (++cnt > 3) {
                        spdWarning("ʣʸʸ̤ݡ:%s\n",
                                   constVal);
                        break;
                    }
                    charVal = charVal * 8 + (d - '0');
                }
            }
        }
        retVal.type = SpdValue_Integer;
        retVal.v.intval = charVal;
        break;
      case CONST_STRING:
        pAreaId = spdGetAChildObjId(constId, SPD_RID_STRING_AREA);
        retVal.type = SpdValue_RealPointer;
        retVal.v.pointerval = (void *)sipGetString(pAreaId);
        break;
      case CONST_FLOAT:
        /* !! double ݡȤ줿 double ˤ */
        retVal.type = SpdValue_Integer;
        retVal.v.intval = (int)strtod(constVal, NULL);
        break;
      default:
        retVal.type = SpdValue_NonValue;
        retVal.v.intval = -1;
        break;
    }

    return retVal;
}

/* Debug. *********************************************************************/

void sipSetDebugMode(int onoff)
{
    debugMode = onoff;
}

int sipGetDebugMode(void)
{
    return (debugMode);
}

static void printEvalStack(void)
{
    SipEvalStackElem    *esElem;
    SpdObjId            pareaId;
    int                 i;

    fflush(stdout);
    fflush(stderr);
    spdDwIndent();
    /* fprintf(stderr, " ES "); */
    fprintf(stderr, " ");
    for (i = 0; i < evalStk->size; i++) {
        esElem = (SipEvalStackElem *)evalStk->elem[i];
        switch (esElem->type) {
          case SipES_Area:
            if ((pareaId = spdGetAParentObjId(esElem->obj.areaId,
                                              SPD_RID_VARIABLE_AREA))
                != SPD_NON_ID) {
                fprintf(stderr, "[A:%s]", spdGetName(pareaId));
            } else {
                fprintf(stderr, "[A]");
            }
            break;
          case SipES_Value:
            fprintf(stderr, "[V:%d]", sipCastValueDataInt(esElem->obj.value));
            break;
          case SipES_FuncSep:
            if (esElem->obj.funcId != SPD_NON_ID) {
                fprintf(stderr, "[F:%s]", spdGetName(esElem->obj.funcId));
            } else {
                fprintf(stderr, "[F]");
            }
            break;
          case SipES_Fix:
            fprintf(stderr, "[Fix:%s]", esElem->obj.isPrefix ? "Pre" : "Post");
            break;
        }
    }
    fprintf(stderr, "\n");
    fflush(stderr);
}

/*
 * Debug write.
 */

void spdDwIndent(void)
{
    int     i;

    fflush(stdout);
    fflush(stderr);
    for (i = 0; i < debugLevel; i++) {
        fputs("    ", stderr);
    }
    fflush(stderr);
}

static void sipDwConst(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "Const ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        printEvalStack();
        fflush(stderr);
    }
}

static void sipDwFuncCall(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "Func ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        printEvalStack();
        fflush(stderr);
    }
}

static void sipDwFuncReturn(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "FuncReturn ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        fflush(stderr);
    }
}

static void sipDwIdent(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "Id ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        printEvalStack();
        fflush(stderr);
    }
}

static void sipDwLibCall(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "Lib ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        fflush(stderr);
    }
}

static void sipDwLibReturn(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "LibReturn ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        printEvalStack();
        fflush(stderr);
    }
}

void sipDwOpe(const char *format, ...)
{
    va_list     args;
    
    if (sipGetDebugMode()) {
        fflush(stdout);
        fflush(stderr);
        spdDwIndent();
        fprintf(stderr, "Op ");
        va_start(args, format);
        vfprintf(stderr, format, args);
        va_end(args);
        printEvalStack();
        fflush(stderr);
    }
}
