/* 
 * Program:             $RCSfile: libaccess.c,v $  $Revision: 30.25 $
 * 
 * Purpose:
 * 
 * Author:              H.Suzuki  1998/11/06
 *                      S.Yamamoto  1998/11/06
 * 
 * (C) Copyright:       H.Suzuki and S.Yamamoto  1998 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: libaccess.c,v 30.25 2003/04/13 03:19:28 yamamoto Exp yamamoto $";
#endif

#include "SIP2.h"
#include "SIP2-proto.h"
#include <ctype.h>

static SpdSet libAcsTbl;

/* int fclose(FILE *stream). */

static int libAcsFclose(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    rslt->obj.value.v.intval = fclose((FILE *)val.v.pointerval);

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* int fflush(FILE *stream). */

static int libAcsFflush(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    rslt->obj.value.v.intval = fflush((FILE *)val.v.pointerval);

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* FILE *fopen(const char *filename, const char *mode). */

static int libAcsFopen(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val[2];
    int                 i;

    for (i = 0; i < 2; i++) {
        elem = sipPopEvalStack();
        val[i] = sipGetValueDataFromESElem(elem);
        spdFree(elem);
    }

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_RealPointer;
    rslt->obj.value.v.pointerval
        = (void *)fopen((char *)sipCastValueDataInt(val[0]),
                        (char *)sipCastValueDataInt(val[1]));

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* int getc(FILE *stream). */

static int libAcsGetc(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    /* rslt->obj.value.v.intval = getc((FILE *)sipCastValueDataInt(val)); */
    rslt->obj.value.v.intval = getc((FILE *)val.v.pointerval);

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* int getchar(void). */

static int libAcsGetchar(void)
{
    SipEvalStackElem    *rslt;

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    rslt->obj.value.v.intval = getchar();

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* int isdigit(int c). */

static int libAcsIsdigit(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    rslt->obj.value.v.intval = isdigit(spdCastValueDataInt(val));

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* void perror(const char *s); */

static int libAcsPerror(void)
{
    SipEvalStackElem    *elem;
    SpdValueData        val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    perror((const char *)val.v.pointerval);

    return SAPID_TRV_GO;
}

/* int putc(int c, FILE *stream). */

static int libAcsPutc(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val[2];
    int                 i;

    for (i = 0; i < 2; i++) {
        elem = sipPopEvalStack();
        val[i] = sipGetValueDataFromESElem(elem);
        spdFree(elem);
    }

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    /* rslt->obj.value.v.intval = putc(sipCastValueDataInt(val[0]),          */
    /*                                 (FILE *)sipCastValueDataInt(val[1])); */
    rslt->obj.value.v.intval = putc(sipCastValueDataInt(val[0]),
                                    (FILE *)val[1].v.pointerval);

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* int putchar(int c); */

static int libAcsPutchar(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    rslt->obj.value.v.intval = putchar(spdCastValueDataInt(val));

    sipPushEvalStack(rslt);
    return SAPID_TRV_GO;
}

/* int setvbuf(FILE *stream, char *buf, int type, size_t size). */

static int libAcsSetvbuf(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val[4];
    int                 i;

    for (i = 0; i < 4; i++) {
        elem = sipPopEvalStack();
        val[i] = sipGetValueDataFromESElem(elem);
        spdFree(elem);
    }

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;

    /* Modified by S.Yamamoto 2000/12/13. */
    /* 
     * rslt->obj.value.v.intval = -1;
     */
    rslt->obj.value.v.intval = setvbuf((FILE *)val[0].v.pointerval,
                                       (char *)val[1].v.pointerval,
                                       sipCastValueDataInt(val[2]),
                                       (size_t)sipCastValueDataInt(val[3]));

    sipPushEvalStack(rslt);

    return SAPID_TRV_GO;
}

/******************************************************************************/

/*
 * int printf(const char *format,...).
 * formatȤΰθĿSIP_MAX_PRINTF_ARGSʲǤ뤳ȡ
 */

static int libAcsPrintf(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val[SIP_MAX_PRINTF_ARGS];
    int                 i;

    i = 0;
    while (sipGetTopOfEvalStack()->type != SipES_FuncSep) {
        if (SIP_MAX_PRINTF_ARGS <= i) {
            spdErrorExit("libAcsPrintf(): Too many arguments.\n");
        }
        elem = sipPopEvalStack();
        val[i++] = sipGetValueDataFromESElem(elem);
        spdFree(elem);
    }
    while (i < SIP_MAX_PRINTF_ARGS) {
        val[i].type = VALUE_INTEGER;
        val[i++].v.intval = 0;
    }

    /* 
     * fflush(stdout);
     * fflush(stderr);
     */

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    /* !! 1"char *", ʹߤintȷǤƤ롥 */
    assert((SIP_MAX_PRINTF_ARGS - 1) == 6);
    rslt->obj.value.v.intval = printf((char *)val[0].v.pointerval,
                                      sipCastValueDataInt(val[1]),
                                      sipCastValueDataInt(val[2]),
                                      sipCastValueDataInt(val[3]),
                                      sipCastValueDataInt(val[4]),
                                      sipCastValueDataInt(val[5]),
                                      sipCastValueDataInt(val[6]));
    /* 
     * fflush(stdout);
     */
    /* !! ʸΥݥ󥿤(ȽǤɬ?) */
    spdFree(val[0].v.pointerval);

    sipPushEvalStack(rslt);
    return (SAPID_TRV_GO);
}

/*
 * int fprintf(FILE *strm, const char *format,...).
 * formatȤΰθĿSIP_MAX_PRINTF_ARGSʲǤ뤳ȡ
 */

static int libAcsFprintf(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        fval, val[SIP_MAX_PRINTF_ARGS];
    int                 i;

    /* FILEݥ󥿤ͤ롥 */
    elem = sipPopEvalStack();
    fval = sipGetValueDataFromESElem(elem);

    i = 0;
    while (sipGetTopOfEvalStack()->type != SipES_FuncSep) {
        if (SIP_MAX_PRINTF_ARGS <= i) {
            spdErrorExit("libAcsFprintf(): Too many arguments.\n");
        }
        elem = sipPopEvalStack();
        val[i++] = sipGetValueDataFromESElem(elem);
        spdFree(elem);
    }
    while (i < SIP_MAX_PRINTF_ARGS) {
        val[i].type = VALUE_INTEGER;
        val[i++].v.intval = 0;
    }

    /* 
     * fflush(stdout);
     * fflush(stderr);
     * fflush((FILE *)fval.v.pointerval);
     */

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    /* !! 2"char *", ʹߤintȷǤƤ롥 */
    assert ((SIP_MAX_PRINTF_ARGS - 1) == 6);
    rslt->obj.value.v.intval = fprintf((FILE *)fval.v.pointerval,
                                       (char *)val[0].v.pointerval,
                                       sipCastValueDataInt(val[1]),
                                       sipCastValueDataInt(val[2]),
                                       sipCastValueDataInt(val[3]),
                                       sipCastValueDataInt(val[4]),
                                       sipCastValueDataInt(val[5]),
                                       sipCastValueDataInt(val[6]));
    /* 
     * fflush((FILE *)fval.v.pointerval);
     */
    /* !! ʸΥݥ󥿤(ȽǤɬ?) */
    spdFree(val[0].v.pointerval);

    sipPushEvalStack(rslt);
    return (SAPID_TRV_GO);
}

/* Table. *********************************************************************/

SpdObjId sipGetFuncIdentId(SpdConstString funcname)
{
    SpdObjIdArray   idents;
    SpdObjId        identId = SPD_NON_ID;
    int             i;

    idents = spdGetObjIdArray("identifier", funcname, NULL);
    for (i = 0; i < idents.size; i++) {
        if (spdGetAttrIntValByName(idents.id[i], "sort") == ID_FUNCTION) {
            identId = idents.id[i];
            break;
        }
    }
    spdFreeObjIdArray(idents);

    return identId;
}

void sipSetLibAccessFunc(SpdString funcname, SipOpeFunc func)
{
    SpdObjId        id;

    if ((id = sipGetFuncIdentId(funcname)) != SPD_NON_ID) {
        spdAddElem(libAcsTbl, id, (Contents)&func);
    }
}

void sipInitLibAccessTable(void)
{
    libAcsTbl = spdCreateSet(sizeof(void *));

    sipSetLibAccessFunc("fclose", libAcsFclose);
    sipSetLibAccessFunc("fflush", libAcsFflush);
    sipSetLibAccessFunc("fopen", libAcsFopen);
    sipSetLibAccessFunc("fprintf", libAcsFprintf);
    sipSetLibAccessFunc("getc", libAcsGetc);
    sipSetLibAccessFunc("getchar", libAcsGetchar);
    sipSetLibAccessFunc("isdigit", libAcsIsdigit);
    sipSetLibAccessFunc("perror", libAcsPerror);
    sipSetLibAccessFunc("printf", libAcsPrintf);
    sipSetLibAccessFunc("putc", libAcsPutc);
    sipSetLibAccessFunc("putchar", libAcsPutchar);
    sipSetLibAccessFunc("setvbuf", libAcsSetvbuf);
}

void sipFreeLibAccessTable(void)
{
    spdFreeSet(libAcsTbl);
}

/* !! SipOpeFuncäƤΤϤäȰ㤦롥 */

SipOpeFunc sipGetlibAcsFunc(SpdObjId funcId)
{
    SipOpeFunc      *retFunc;

    retFunc = (SipOpeFunc *)spdSearchElem(libAcsTbl, funcId);
    return (retFunc != NULL ? *retFunc : NULL);
}
