/* 
 * Program:             $RCSfile: operate.c,v $  $Revision: 30.38 $
 * 
 * Purpose:
 * 
 * Author:              H.Suzuki  1998/11/06
 *                      S.Yamamoto  1998/11/06
 * Modified:            H.Ohkubo  2001/04/08
 * 
 * (C) Copyright:       H.Suzuki, H.Ohkubo and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
"$Id: operate.c,v 30.38 2003/06/28 15:33:20 yamamoto Exp yamamoto $";
#endif

#include <stdarg.h>
#include "SIP2.h"
#include "SIP2-proto.h"

static SpdObjId getArrayAreaWithIndex(SpdObjId areaId, int index)
{
    SpdRel      pRel;
    SpdObjId    retAreaId, pAreaId;
    int         order, newIndex;

    pRel = spdGetAParentRel(areaId, SPD_RID_AREA_AREA);
    if (pRel.objectId == SPD_NON_ID) {
        /* Root area of array. */
        pAreaId = areaId;
        order = 0;
    } else {
        pAreaId = pRel.objectId;
        order = spdGetAttrIntValByName(pRel.relationId, "order");
    }

    newIndex = order + index;
    retAreaId = spdGetArrayElementArea(pAreaId, newIndex);
    /* 
     * if (retAreaId == SPD_NON_ID) {
     *     spdErrorExit("illegal access. area[%d], index[%d]\n",
     *                  pAreaId, newIndex);
     * } else {
     *     return retAreaId;
     * } 
     */
    return (retAreaId);
}

/* Arithmetic operators. ******************************************************/

/* ( ) : ̱黻 */

static int sipOpeNoAction(void)
{
    /* Do nothing. */
    return SAPID_TRV_GO;
}

/* Sub function of addition and subtraction. */

static int sipOpeAddAndSub(int sign)
{
    SipEvalStackElem    *elem1, *elem2, *rslt;
    SpdValueData        val1, val2;
    SpdObjId            areaId;
    int                 index;
    void                *pval;

    elem2 = sipPopEvalStack();
    elem1 = sipPopEvalStack();
    val1 = sipGetValueDataFromESElem(elem1);
    val2 = sipGetValueDataFromESElem(elem2);
    spdFree(elem1);
    spdFree(elem2);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    if ((val1.type == SpdValue_Pointer && val2.type != SpdValue_Pointer) || 
        (val1.type != SpdValue_Pointer && val2.type == SpdValue_Pointer)) {
        /* ݥͤβû */
        /* !! ΤȤؤΥݥ󥿤бƤ롥 */
        /* !! ¥ݥͤλϡcastƤ黻ʤФʤʤ */
        if (val1.type == SpdValue_Pointer) {
            areaId = val1.v.ptr.areaId;
            index = sign * val2.v.intval;
        } else {
            index = val1.v.intval;
            areaId = sign * val2.v.ptr.areaId;
        }
        rslt->obj.value.type = SpdValue_Pointer;
        rslt->obj.value.v.ptr.areaId = getArrayAreaWithIndex(areaId, index);
        SIP_DW_OPE("[+]: %d:areaId + %d = %d:areaId\n",
                   areaId, index, rslt->obj.value.v.ptr.areaId);
    } else if ((val1.type == SpdValue_RealPointer &&
                val2.type != SpdValue_RealPointer) ||
               (val1.type != SpdValue_RealPointer &&
                val2.type == SpdValue_RealPointer)) {
        if (val1.type == SpdValue_RealPointer) {
            pval = val1.v.pointerval;
            index = sign * val2.v.intval;
        } else {
            index = val1.v.intval;
            pval = val2.v.pointerval;       /* !! signʤ */
        }
        rslt->obj.value.type = SpdValue_RealPointer;
        rslt->obj.value.v.pointerval = (FILE *)pval + index;
        SIP_DW_OPE("[+]: %#x + %d = %#x\n",
                   pval, index, rslt->obj.value.v.pointerval);
    } else {
        rslt->obj.value.type = SpdValue_Integer;
        rslt->obj.value.v.intval = val1.v.intval + sign * val2.v.intval;
        SIP_DW_OPE("[+]: %d + %d = %d\n",
                   val1.v.intval, val2.v.intval, rslt->obj.value.v.intval);
    }
    sipPushEvalStack(rslt);

    return (SAPID_TRV_GO);
}

/* Sub function of binary operation. */

static int sipBinaryOpe(int opeSort)
{
    SipEvalStackElem    *elem1, *elem2, *rslt;
    SpdValueData        val1, val2;
    char                *sipDwPrefix;

    elem2 = sipPopEvalStack();
    elem1 = sipPopEvalStack();
    val1 = sipGetValueDataFromESElem(elem1);
    val2 = sipGetValueDataFromESElem(elem2);
    spdFree(elem1);
    spdFree(elem2);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    switch (opeSort) {
      case OP_MUL :
        rslt->obj.value.v.intval = val1.v.intval * val2.v.intval;
        sipDwPrefix = "[*]: %d * %d = %d\n";
        break;
      case OP_DIV :
        rslt->obj.value.v.intval = val1.v.intval / val2.v.intval;
        sipDwPrefix = "[/]: %d / %d = %d\n";
        break;
      case OP_MOD :
        rslt->obj.value.v.intval = val1.v.intval % val2.v.intval;
        sipDwPrefix = "[%%]: %d %% %d = %d\n";
        break;
      case OP_B_AND :
        rslt->obj.value.v.intval = val1.v.intval & val2.v.intval;
        sipDwPrefix = "[&]: %d & %d = %d\n";
        break;
      case OP_B_OR :
        rslt->obj.value.v.intval = val1.v.intval | val2.v.intval;
        sipDwPrefix = "[|]: %d | %d = %d\n";
        break;
      case OP_B_XOR :
        rslt->obj.value.v.intval = val1.v.intval ^ val2.v.intval;
        sipDwPrefix = "[^]: %d ^ %d = %d\n";
        break;
      case OP_SHL :
        rslt->obj.value.v.intval = val1.v.intval << val2.v.intval;
        sipDwPrefix = "[<<]: %d << %d = %d\n";
        break;
      case OP_SHR :
        rslt->obj.value.v.intval = val1.v.intval >> val2.v.intval;
        sipDwPrefix = "[>>]: %d >> %d = %d\n";
        break;
      case OP_EQ :
        rslt->obj.value.v.intval = val1.v.intval == val2.v.intval;
        sipDwPrefix = "[==]: %d == %d = %d\n";
        break;
      case OP_LT :
        rslt->obj.value.v.intval = val1.v.intval < val2.v.intval;
        sipDwPrefix = "[<]: %d < %d = %d\n";
        break;
      case OP_LE :
        rslt->obj.value.v.intval = val1.v.intval <= val2.v.intval;
        sipDwPrefix = "[<=]: %d <= %d = %d\n";
        break;
      case OP_GT :
        rslt->obj.value.v.intval = val1.v.intval > val2.v.intval;
        sipDwPrefix = "[>]: %d > %d = %d\n";
        break;
      case OP_GE :
        rslt->obj.value.v.intval = val1.v.intval >= val2.v.intval;
        sipDwPrefix = "[>=]: %d >= %d = %d\n";
        break;
      case OP_NE :
        rslt->obj.value.v.intval = val1.v.intval != val2.v.intval;
        sipDwPrefix = "[!=]: %d != %d = %d\n";
        break;
      default:
        spdErrorExit("sipBinaryOpe(): %d: Invalid operator.\n", opeSort);
        break;
    }
    sipPushEvalStack(rslt);
    SIP_DW_OPE(sipDwPrefix,
               val1.v.intval, val2.v.intval, rslt->obj.value.v.intval);

    return SAPID_TRV_GO;
}

/* + : û */

static int sipOpeAdd(void)
{
    sipOpeAddAndSub(1);
    SPD_DUMMY_RETURN(0);
}

/* - :  */

static int sipOpeSub(void)
{
    sipOpeAddAndSub(-1);
    SPD_DUMMY_RETURN(0);
}

/* * : 軻 */

static int sipOpeMul(void)
{
    sipBinaryOpe(OP_MUL);
    SPD_DUMMY_RETURN(0);
}

/* / :  */

static int sipOpeDiv(void)
{
    sipBinaryOpe(OP_DIV);
    SPD_DUMMY_RETURN(0);
}

/* % : ; */

static int sipOpeMod(void)
{
    sipBinaryOpe(OP_MOD);
    SPD_DUMMY_RETURN(0);
}

/* Unary operators. ***********************************************************/

/* Sub function of unary operation. */

static int sipUnaryOpe(int opeSort)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;
    char                *sipDwPrefix;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);
    spdFree(elem);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value.type = SpdValue_Integer;
    switch (opeSort) {
      case OP_PLUS :
        rslt->obj.value.v.intval = val.v.intval;
        sipDwPrefix = "[+]: + %d = %d\n";
        break;
      case OP_MINUS :
        rslt->obj.value.v.intval = -(val.v.intval);
        sipDwPrefix = "[-]: - %d = %d\n";
        break;
      case OP_B_NOT :
        rslt->obj.value.v.intval = ~(val.v.intval);
        sipDwPrefix = "[~]: ~ %d = %d\n";
        break;
      case OP_NOT :
        rslt->obj.value.v.intval = !(val.v.intval);
        sipDwPrefix = "[!]: ! %d = %d\n";
        break;
      default:
        spdErrorExit("sipUnaryOpe(): %d: Invalid operator.\n", opeSort);
        break;
    }
    sipPushEvalStack(rslt);
    SIP_DW_OPE(sipDwPrefix, val.v.intval, rslt->obj.value.v.intval);

    return SAPID_TRV_GO;
}

/* + : ñ+ */

static int sipOpePlus(void)
{
    sipUnaryOpe(OP_PLUS);
    SPD_DUMMY_RETURN(0);
}

/* - : ñ- */

static int sipOpeMinus(void)
{
    sipUnaryOpe(OP_MINUS);
    SPD_DUMMY_RETURN(0);
}

/* Bit operator. **************************************************************/

/* ΤߤоݤȤfloat, doubleФƤϥӥåȱ黻ϤǤʤ */

/* ~ :  */

static int sipOpeBitNot(void)
{
    sipUnaryOpe(OP_B_NOT);
    SPD_DUMMY_RETURN(0);
}

/* & : And */

static int sipOpeBitAnd(void)
{
    sipBinaryOpe(OP_B_AND);
    SPD_DUMMY_RETURN(0);
}

/* | : Or */

static int sipOpeBitOr(void)
{
    sipBinaryOpe(OP_B_OR);
    SPD_DUMMY_RETURN(0);
}

/* ^ : Xor */

static int sipOpeBitXor(void)
{
    sipBinaryOpe(OP_B_XOR);
    SPD_DUMMY_RETURN(0);
}

/* << : ե */

static int sipOpeShiftL(void)
{
    sipBinaryOpe(OP_SHL);
    SPD_DUMMY_RETURN(0);
}

/* >> : ե */

static int sipOpeShiftR(void)
{
    sipBinaryOpe(OP_SHR);
    SPD_DUMMY_RETURN(0);
}

/* Pointer. *******************************************************************/

/* * : ݥ */

static int sipOpePointer(void)
{
    SipEvalStackElem    *elem, *rslt;
    SpdValueData        val;
    SpdObjId            areaId;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);

    rslt = sipCreateEvalStackElem();
    switch (val.type) {
      case SpdValue_Pointer:
        areaId = val.v.ptr.areaId;
        while (spdGetAttrIntValByName(areaId, "sort") == AREA_ARRAY) {
            areaId = spdGetArrayElementArea(areaId, 0);
        }
        rslt->type = SipES_Area;
        rslt->obj.areaId = areaId;

        if (sipGetDebugMode()) {
            SpdConstString  nm;
            
            nm = spdGetName(spdGetAParentObjId(val.v.ptr.areaId,
                                               SPD_RID_VARIABLE_AREA));
            if (nm != NULL) {
                SIP_DW_OPE("[*]: *%s -> Area:%d\n", nm, rslt->obj.areaId);
            } else {
                SIP_DW_OPE("[*]: *(Area:%d) -> Area:%d\n",
                           val.v.ptr.areaId, rslt->obj.areaId);
            }
        }
        break;
      case SpdValue_FuncPointer:
        /* "funcp()""(* funcp)()"Ʊ뤹Τǲ⤷ʤ */
        rslt->type = SipES_Value;
        rslt->obj.value = val;
        SIP_DW_OPE("[*]: Func Pointer *(%d) -> %d\n",
                   val.v.funcId, rslt->obj.value.v.funcId);
        break;
      default:
        rslt->type = SipES_Value;
        rslt->obj.value.type = SpdValue_Integer;
        rslt->obj.value.v.intval = *(int *)val.v.pointerval;
        SIP_DW_OPE("[*]: *(%d) -> %d\n",
                   val.v.pointerval, rslt->obj.value.v.intval);
        break;
    }
    sipPushEvalStack(rslt);

    spdFree(elem);

    return SAPID_TRV_GO;
}

/* & : ɥ쥹 */

static int sipOpeAddress(void)
{
    SipEvalStackElem *elem, *rslt;
    SpdObjId areaId;

    elem = sipPopEvalStack();

    if (elem->type != SipES_Area) {
        /* AreaʳؤΥɥ쥹黻 */ 
        spdErrorExit("Invalid & operation.\n");
    }

    areaId = elem->obj.areaId;
    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    if (spdGetAttrIntValByName(areaId, "indirect") == AREA_DIRECT) {
        rslt->obj.value.type = SpdValue_RealPointer;
        rslt->obj.value.v.pointerval =
            (void *)spdGetAttrIntValByName(areaId, "address");
    } else {
        rslt->obj.value.type = SpdValue_Pointer;
        rslt->obj.value.v.ptr.areaId = elem->obj.areaId;
    }
    sipPushEvalStack(rslt);

    if (sipGetDebugMode()) {
        SpdConstString  nm;

        nm = spdGetName(spdGetAParentObjId(areaId, SPD_RID_VARIABLE_AREA));
        if (nm != NULL) {
            SIP_DW_OPE("[&]: &%s -> Area:%d\n", nm, rslt->obj.value.v.intval);
        } else {
            SIP_DW_OPE("[&]: &(Area:%d) -> Area:%d\n",
                       areaId, rslt->obj.value.v.intval);
        }
    }

    spdFree(elem);

    return SAPID_TRV_GO;
}

/* лȱ黻 *************************************************************/

/* [ ] : 󻲾 */

static int sipOpeArray(void)
{
    SipEvalStackElem *baseElem, *indxElem, *rslt;
    SpdObjId areaId, cAreaId, typeId;
    SpdRelId relId;
    SpdValueData val;
    int indx, indirect;
    void *adrs, *cAdrs;

    indxElem = sipPopEvalStack();
    baseElem = sipPopEvalStack();

    if (baseElem->type == SipES_Area) {
        switch (spdGetAttrIntValByName(baseElem->obj.areaId, "sort")) {
          case AREA_ARRAY:
            areaId = baseElem->obj.areaId;
            indirect = spdGetAttrIntValByName(areaId, "indirect");
            if (0 < spdGetChildObjIdSet(areaId, SPD_RID_AREA_AREA)->elemNum ||
                indirect == AREA_DIRECT) {
                val = sipGetValueDataFromESElem(indxElem);
                indx = spdCastValueDataInt(val);
                spdFree(baseElem);
                spdFree(indxElem);

                cAreaId = spdGetArrayElementArea(areaId, indx);
                if (indirect == AREA_DIRECT) {
                    if (cAreaId == SPD_NON_ID) {
                        typeId = spdGetAChildObjId(areaId, SPD_RID_AREA_TYPE);
                        typeId = spdGetAChildObjId(typeId, SPD_RID_TYPE_ST);

                        cAreaId = spdCreateArea(typeId);
                        relId = spdCreateRel(SPD_RID_AREA_AREA,
                                             SPD_NON_ID, areaId, cAreaId);
                        spdSetAttrIntValByName(relId, "order", indx);
                    }
                    if (spdGetAttrIntValByName(cAreaId, "indirect")
                        != AREA_DIRECT) {
                        adrs = (void *)spdGetAttrIntValByName(areaId,
                                                              "address");
                        /* !! adrstypeId˹碌Cast٤ */
                        cAdrs = (FILE *)adrs + indx;
                        spdSetAttrIntValByName(cAreaId, "address", (int)cAdrs);
                        spdSetAttrIntValByName(cAreaId, "indirect",
                                               AREA_DIRECT);
                    }
                }

                rslt = sipCreateEvalStackElem();
                rslt->type = SipES_Area;
                rslt->obj.areaId = cAreaId;

                if (sipGetDebugMode()) {
                    SpdConstString  nm;
                    
                    nm = spdGetName(spdGetAParentObjId(areaId,
                                                       SPD_RID_VARIABLE_AREA));
                    if (nm != NULL) {
                        SIP_DW_OPE("[[]]: %s[%d] -> Area:%d\n", nm, indx, cAreaId);
                    } else {
                        SIP_DW_OPE("[[]]: (Area:%d)[%d] -> Area:%d\n",
                                   areaId, indx, cAreaId);
                    }
                }

                sipPushEvalStack(rslt);
            } else {
                /* !! TYPE x[]Υݥ󥿤ȤƤμԤ */
                sipPushEvalStack(baseElem);
                sipPushEvalStack(indxElem);
                sipOpeAdd();
                sipOpePointer();
            }
            break;
          case AREA_BASIC:
          default:
            sipPushEvalStack(baseElem);
            sipPushEvalStack(indxElem);
            sipOpeAdd();
            sipOpePointer();
            break;
        }
    }

    return SAPID_TRV_GO;
}

/* -> : 黻 */

static int sipOpeArrow(void)
{
    return SAPID_TRV_GO;
}

/* . : ɥåȱ黻 */

static int sipOpeDot(void)
{
    return SAPID_TRV_GO;
}

/* 黻 *****************************************************************/

/* = : ñ */

static int sipOpeAssign(void)
{
    SipEvalStackElem    *lelem, *relem, *rslt; /* , ,  */
    SpdValueData        rval;               /*  */
    SpdObjId            valueId;
    SipCallStackElem    *csElem;

    relem = sipPopEvalStack();
    lelem = sipPopEvalStack();

    /* դΥå */
    if (lelem->type != SipES_Area) {
        spdFree(relem);
        spdFree(lelem);
        spdErrorExit("Areaʳؤ\n");
    }
    /* area info  */
    csElem = getTopOfCallStack();
    spdAddElem(csElem->areaInfo.defs, (int)lelem->obj.areaId,
               (Contents)&lelem->obj.areaId);

    /* ͤμ */
    rval = sipGetValueDataFromESElem(relem);
    spdFree(relem);

    /* area_obj  value_obj ˴ϢϤ(ŤvalueϾä) */
    valueId = spdGetAChildObjId(lelem->obj.areaId, SPD_RID_AREA_VALUE);
    if (valueId != SPD_NON_ID) {
        spdDeleteObj(valueId);
    }
    spdSetValueData(lelem->obj.areaId, rval);

    rslt = sipCreateEvalStackElem();
    rslt->type = SipES_Value;
    rslt->obj.value = rval;
    sipPushEvalStack(rslt);

    if (sipGetDebugMode()) {
        SpdObjId    pareaId;

        if ((pareaId = spdGetAParentObjId(lelem->obj.areaId,
                                          SPD_RID_VARIABLE_AREA)) != SPD_NON_ID) {
            SIP_DW_OPE("[=]: %s = %d\n", spdGetName(pareaId),
                       sipCastValueDataInt(sipGetValueDataFromESElem(lelem)));
        } else {
            SIP_DW_OPE("[=]: Area(%d) = %d\n", lelem->obj.areaId,
                       sipCastValueDataInt(sipGetValueDataFromESElem(lelem)));
        }
    }

    spdFree(lelem);

    return SAPID_TRV_GO;
}

/* ʣ黻 ***************************************************************/

static int sipOpeAndAssign(int opeSort)
{
    SipEvalStackElem    *lElem, *rElem, *dupElem;

    rElem = sipPopEvalStack();
    lElem = sipPopEvalStack();
    dupElem = sipDupEvalStackElem(lElem);
    sipPushEvalStack(dupElem);
    sipPushEvalStack(lElem);
    sipPushEvalStack(rElem);

    switch (opeSort) {
      case OP_A_ADD: sipOpeAdd(); break;
      case OP_A_SUB: sipOpeSub(); break;
      case OP_A_MUL: sipOpeMul(); break;
      case OP_A_DIV: sipOpeDiv(); break;
      case OP_A_MOD: sipOpeMod(); break;
      case OP_A_AND: sipOpeBitAnd(); break;
      case OP_A_OR : sipOpeBitOr(); break;
      case OP_A_XOR: sipOpeBitXor(); break;
      case OP_A_SHL: sipOpeShiftL(); break;
      case OP_A_SHR: sipOpeShiftR(); break;
      default:
        spdErrorExit("sipOpeAndAssign(): %d: Invalid operator.\n", opeSort);
        break;
    }
    sipOpeAssign();

    return SAPID_TRV_GO;
}

/* += : û */

static int sipOpeAssignAdd(void)
{
    sipOpeAndAssign(OP_A_ADD);
    SPD_DUMMY_RETURN(0);
}

/* -= :  */

static int sipOpeAssignSub(void)
{
    sipOpeAndAssign(OP_A_SUB);
    SPD_DUMMY_RETURN(0);
}

/* *= : 軻 */

static int sipOpeAssignMul(void)
{
    sipOpeAndAssign(OP_A_MUL);
    SPD_DUMMY_RETURN(0);
}

/* /= :  */

static int sipOpeAssignDiv(void)
{
    sipOpeAndAssign(OP_A_DIV);
    SPD_DUMMY_RETURN(0);
}

/* %= : ; */

static int sipOpeAssignMod(void)
{
    sipOpeAndAssign(OP_A_MOD);
    SPD_DUMMY_RETURN(0);
}

/* &= : And */

static int sipOpeAssignBitAnd(void)
{
    sipOpeAndAssign(OP_A_AND);
    SPD_DUMMY_RETURN(0);
}

/* |= : Or */

static int sipOpeAssignBitOr(void)
{
    sipOpeAndAssign(OP_A_OR);
    SPD_DUMMY_RETURN(0);
}

/* ^= : Xor */

static int sipOpeAssignBitXor(void)
{
    sipOpeAndAssign(OP_A_XOR);
    SPD_DUMMY_RETURN(0);
}

/* <<= : ե */

static int sipOpeAssignShiftL(void)
{
    sipOpeAndAssign(OP_A_SHL);
    SPD_DUMMY_RETURN(0);
}

/* >>= : ե */

static int sipOpeAssignShiftR(void)
{
    sipOpeAndAssign(OP_A_SHR);
    SPD_DUMMY_RETURN(0);
}

/* Logical operators. *********************************************************/

static int sipOpeNot(void)
{
    sipUnaryOpe(OP_NOT);
    SPD_DUMMY_RETURN(0);
}

static int sipOpeLogicAnd(void)
{
    SipEvalStackElem *elem;
    SpdValueData val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);

    SIP_DW_OPE("[&&]: Left value: %d\n", spdCastValueDataInt(val));
    /* ʤɾ롥 */
    if (spdCastValueDataInt(val) == 0) {
        sipPushEvalStack(elem);
        return SAPID_TRV_SKIP;
    } else {
        spdFree(elem);
        return SAPID_TRV_GO;
    }
}

static int sipOpeLogicOr(void)
{
    SipEvalStackElem *elem;
    SpdValueData val;

    elem = sipPopEvalStack();
    val = sipGetValueDataFromESElem(elem);

    SIP_DW_OPE("[||]: Left value: %d\n", spdCastValueDataInt(val));
    /* ʤɾ롥 */
    if (spdCastValueDataInt(val) != 0) {
        sipPushEvalStack(elem);
        return SAPID_TRV_SKIP;
    } else {
        spdFree(elem);
        return SAPID_TRV_GO;
    }
}

/* Comma operator. ************************************************************/

static int sipOpeComma(void)
{
    SipEvalStackElem    *lval, *rval;

    rval = sipPopEvalStack();
    lval = sipPopEvalStack();

    sipPushEvalStack(rval);
    spdFree(lval);

    return SAPID_TRV_GO;
}

/* ط黻 *******************************************************************/

/* == : Ƚ */

static int sipOpeEqual(void)
{
    sipBinaryOpe(OP_EQ);
    SPD_DUMMY_RETURN(0);
}

/* < : ʤ黻 */

static int sipOpeLT(void)
{
    sipBinaryOpe(OP_LT);
    SPD_DUMMY_RETURN(0);
}

/* <= : ʲ黻 */

static int sipOpeLE(void)
{
    sipBinaryOpe(OP_LE);
    SPD_DUMMY_RETURN(0);
}

/* > : ʤ黻 */

static int sipOpeGT(void)
{
    sipBinaryOpe(OP_GT);
    SPD_DUMMY_RETURN(0);
}

/* >= : ʾ黻 */

static int sipOpeGE(void)
{
    sipBinaryOpe(OP_GE);
    SPD_DUMMY_RETURN(0);
}

/* != : ͱ黻 */

static int sipOpeNotEqual(void)
{
    sipBinaryOpe(OP_NE);
    SPD_DUMMY_RETURN(0);
}

/* Increment and decrement operator. ******************************************/

/* Increment operator. */

static int sipOpeInc(void)
{
    SipEvalStackElem *oneElem, *dupElem, *fixElem;
    SpdValueData val;

    /* 黻Ҥΰ־ */
    fixElem = sipPopEvalStack();

    /* Ѥͤݻ */
    val = sipGetValueDataFromESElem(sipGetTopOfEvalStack());
    dupElem = sipCreateEvalStackElem();
    dupElem->type = SipES_Value;
    dupElem->obj.value = val;

    /* 1פѤ */
    oneElem = sipCreateEvalStackElem();
    oneElem->type = SipES_Value;
    oneElem->obj.value.type = SpdValue_Integer;
    oneElem->obj.value.v.intval = 1;
    sipPushEvalStack(oneElem);

    /* x += 1 黻 */
    sipOpeAssignAdd();

    if (fixElem->obj.isPrefix) {
        /* (Ѥݻͤ) */
        spdFree(dupElem);
    } else {
        /*  */
        spdFree(sipPopEvalStack());
        sipPushEvalStack(dupElem);
    }
    spdFree(fixElem);

    return SAPID_TRV_GO;
}

/* Decrement operator. */

static int sipOpeDec(void)
{
    SipEvalStackElem *oneElem, *dupElem, *fixElem;
    SpdValueData val;

    /* 黻Ҥΰ־ */
    fixElem = sipPopEvalStack();

    /* Ѥͤݻ */
    val = sipGetValueDataFromESElem(sipGetTopOfEvalStack());
    dupElem = sipCreateEvalStackElem();
    dupElem->type = SipES_Value;
    dupElem->obj.value = val;

    /* 1פѤ */
    oneElem = sipCreateEvalStackElem();
    oneElem->type = SipES_Value;
    oneElem->obj.value.type = SpdValue_Integer;
    oneElem->obj.value.v.intval = 1;
    sipPushEvalStack(oneElem);

    /* x -= 1 黻 */
    sipOpeAssignSub();

    if (fixElem->obj.isPrefix) {
        /* (Ѥݻͤ) */
        spdFree(dupElem);
    } else {
        /*  */
        spdFree(sipPopEvalStack());
        sipPushEvalStack(dupElem);
    }
    spdFree(fixElem);

    return SAPID_TRV_GO;
}

/* Initialization of operator libraries. **************************************/

void sipInitOpeFunc(void)
{
    sipSetOpeFunc(OP_L_PAR, sipOpeNoAction);
    sipSetOpeFunc(OP_R_PAR, sipOpeNoAction);

    sipSetOpeFunc(OP_L_ARRAY, sipOpeArray);
    sipSetOpeFunc(OP_R_ARRAY, sipOpeNoAction);
    sipSetOpeFunc(OP_ARROW, sipOpeArrow);
    sipSetOpeFunc(OP_DOT, sipOpeDot);

    sipSetOpeFunc(OP_POINTER, sipOpePointer);
    sipSetOpeFunc(OP_ADDRESS, sipOpeAddress);

    sipSetOpeFunc(OP_ADD, sipOpeAdd);
    sipSetOpeFunc(OP_SUB, sipOpeSub);
    sipSetOpeFunc(OP_MUL, sipOpeMul);
    sipSetOpeFunc(OP_DIV, sipOpeDiv);
    sipSetOpeFunc(OP_MOD, sipOpeMod);

    sipSetOpeFunc(OP_COMMA, sipOpeComma);

    sipSetOpeFunc(OP_B_NOT, sipOpeBitNot);
    sipSetOpeFunc(OP_B_AND, sipOpeBitAnd);
    sipSetOpeFunc(OP_B_OR, sipOpeBitOr);
    sipSetOpeFunc(OP_B_XOR, sipOpeBitXor);
    sipSetOpeFunc(OP_SHL, sipOpeShiftL);
    sipSetOpeFunc(OP_SHR, sipOpeShiftR);

    sipSetOpeFunc(OP_EQ, sipOpeEqual);
    sipSetOpeFunc(OP_NE, sipOpeNotEqual);
    sipSetOpeFunc(OP_LT, sipOpeLT);
    sipSetOpeFunc(OP_LE, sipOpeLE);
    sipSetOpeFunc(OP_GT, sipOpeGT);
    sipSetOpeFunc(OP_GE, sipOpeGE);

    sipSetOpeFunc(OP_PLUS, sipOpePlus);
    sipSetOpeFunc(OP_MINUS, sipOpeMinus);

    sipSetOpeFunc(OP_NOT, sipOpeNot);
    sipSetOpeFunc(OP_L_AND, sipOpeLogicAnd);
    sipSetOpeFunc(OP_L_OR, sipOpeLogicOr);

    sipSetOpeFunc(OP_ASSIGN, sipOpeAssign);
    sipSetOpeFunc(OP_A_ADD, sipOpeAssignAdd);
    sipSetOpeFunc(OP_A_SUB, sipOpeAssignSub);
    sipSetOpeFunc(OP_A_MUL, sipOpeAssignMul);
    sipSetOpeFunc(OP_A_DIV, sipOpeAssignDiv);
    sipSetOpeFunc(OP_A_MOD, sipOpeAssignMod);
    sipSetOpeFunc(OP_A_AND, sipOpeAssignBitAnd);
    sipSetOpeFunc(OP_A_OR, sipOpeAssignBitOr);
    sipSetOpeFunc(OP_A_XOR, sipOpeAssignBitXor);
    sipSetOpeFunc(OP_A_SHL, sipOpeAssignShiftL);
    sipSetOpeFunc(OP_A_SHR, sipOpeAssignShiftR);

    /* 
     * sipSetOpeFunc(OP_SIZEOF, sipOpeSizeof);
     * sipSetOpeFunc(OP_SIZEOF_TYPE, sipOpeSizeofType);
     * sipSetOpeFunc(OP_CAST, sipOpeCast);
     */

    /* !!  */
    sipSetOpeFunc(OP_CAST, sipOpeNoAction);

    sipSetOpeFunc(OP_INC, sipOpeInc);
    sipSetOpeFunc(OP_DEC, sipOpeDec);
}

/******************************************************************************/

/* Sizeof operators. */

/* 
 * static int sipOpeSizeof(void)
 * {
 * 
 *     return SAPID_TRV_GO;
 * }
 * 
 * static int sipOpeSizeofType(void)
 * {
 * 
 *     return SAPID_TRV_GO;
 * }
 */

/* Cast operator. */

/* 
 * static int sipOpeCast(void)
 * {
 *     return SAPID_TRV_GO;
 * }
 */
