/* 
 * Program:             $RCSfile: sint3-step.c,v $  $Revision: 40.5 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/12/11
 *                      S.Yamamoto  2001/12/11
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001 - 2002
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: sint3-step.c,v 40.5 2002/12/22 13:31:14 yamamoto Exp yamamoto $";
#endif

#include <limits.h>
#include "SIP2.h"
#include "SIP2-proto.h"

static SpdObjId sipSkipFvNode(SpdObjId curNodeId);

static SpdBoolean isTrue(SpdValueData val)
{
    /* 
     * if (spdCastValueDataInt(val) != 0) return TRUE;
     * else return FALSE;
     */
    return (spdCastValueDataInt(val) != 0);
}

/* 2ĤSpdValueDataӤ == (0), > (1), < (-1)֤ */

static int compareValueDataInt(SpdValueData val1, SpdValueData val2)
{
    if (val1.type == SpdValue_NonValue || val2.type == SpdValue_NonValue)
        return -1;
    else
        return (sipCastValueDataInt(val1) - sipCastValueDataInt(val2));
}

/* ؿƽлdeclaratorǽ˼¹Ԥ٤expression򸡺֤ */

SpdObjId sipGetExpressionFromFuncDecr(SpdObjId funcDecrId)
{
    SpdCursor *sdanode_cursor;
    SpdObjId declId, sdanodeId = SPD_NON_ID, blkEntryId = SPD_NON_ID;
    SpdObjId exprId = SPD_NON_ID;

    if (spdGetClassId(funcDecrId) == SPD_CID_DECLARATOR) {
        declId = spdGetARelObj(funcDecrId, "decl_decl", "declarator_id");
        sdanode_cursor = spdGetRelObjInit(declId, "blk_sdanode", "blkId");
        while ((sdanodeId = spdGetRelObj(sdanode_cursor)) != SPD_NON_ID) {
            if (spdGetClassId(sdanodeId) == SPD_CID_BLK_ENTRY) {
                blkEntryId = sdanodeId;
                break;
            }
        }
        spdFreeCursor(sdanode_cursor);

        if (blkEntryId == SPD_NON_ID) {
            spdErrorExit("Can't find blkEntry of function.\n");
        }

        /* ǽexprޤǤʬϤʤΤǰƻǤɤ롣 */
        exprId = sipSkipFvNode(blkEntryId);
        if (exprId == SPD_NON_ID) {
            sipFreePoppedCallStack();
            sipSetTrvStackOfCallStack();
            return (sipGetProgramPtr());
        }
    }
    return exprId;
}

/* exprIdexpressionＰɤĴ٤롣 */

SpdBoolean isBranchCondExpr(SpdObjId exprId)
{
    int edgeSort;
    SpdRel rel;

    rel = spdGetARel(exprId, "next_node", "curId");
    edgeSort = spdGetAttrValInt(rel.relationId, "sort");

    if (edgeSort == SPD_FV_E_BRANCH_TRUE || 
        edgeSort == SPD_FV_E_BRANCH_FALSE ) {
        return TRUE;
    } else {
        return FALSE;
    }
}

/* exprIdexpressionswitchʸξＰɤĴ٤롣 */

SpdBoolean isSwitchCondExpr(SpdObjId exprId)
{
    SpdObjId preNodeId;

    preNodeId = spdGetARelObj(exprId, "next_node", "nextId");
    if (spdGetClassId(preNodeId) == SPD_CID_BLK_ENTRY &&
        spdGetAttrValInt(preNodeId, "sort") == SPD_FV_B_SWITCH) {
        return TRUE;
    } else {
        return FALSE;
    }
}

SpdObjId getBranchNext(SpdObjId exprId, SpdValueData val)
{
    SpdBoolean bool;
    SpdCursor *nextNode_cursor;
    SpdRel rel;
    SpdObjId nextNodeId = SPD_NON_ID;

    bool = isTrue(val);
    nextNode_cursor = spdGetRelInit(exprId, "next_node", "curId");
    do {
        int edgeSort = -1;

        rel = spdGetRel(nextNode_cursor);
        edgeSort = spdGetAttrValInt(rel.relationId, "sort");
        if (edgeSort == SPD_FV_E_BRANCH_TRUE) {
            if (bool == TRUE) {
                nextNodeId = rel.objectId;
                break;
            }
        } else if (edgeSort == SPD_FV_E_BRANCH_FALSE) {
            if (bool == FALSE) {
                nextNodeId =  rel.objectId;
                break;
            }
        }
    } while (rel.objectId != SPD_NON_ID);
    spdFreeCursor(nextNode_cursor);

    return nextNodeId;
}

SpdObjId getSwitchNext(SpdObjId exprId, SpdValueData val)
{
    SpdCursor *nextNode_cursor;
    SpdRel rel;
    SpdObjId curNodeId, cexprId, nextNodeId = SPD_NON_ID;
    SpdValueData caseVal;

    /* switchBranch򤵤 */
    curNodeId = spdGetARelObj(exprId, "next_node", "curId");
    /* switch(a); Τ褦ʾ */
    if (spdGetClassId(curNodeId) == SPD_CID_BLK_EXIT &&
        spdGetAttrValInt(curNodeId, "sort") == SPD_FV_B_SWITCH) {
        return curNodeId;
    }
    /* switchBranchid */
    curNodeId = spdGetARelObj(curNodeId, "next_node", "curId");
    
    
    /* case٥ǰפΤ򤵤*/
    nextNode_cursor = spdGetRelInit(curNodeId, "next_node", "curId");
    do {
        rel = spdGetRel(nextNode_cursor);

        if (spdGetAttrValInt(rel.relationId, "sort") == SPD_FV_E_SWITCH_CASE) {
            cexprId = spdGetARelObj(rel.objectId, "case_cexpr", "any");
            caseVal = sipEvalExpression(cexprId);
            if (compareValueDataInt(val, caseVal) == 0) {
                nextNodeId = rel.objectId;
                break;
            }
        }
    } while (rel.objectId != SPD_NON_ID);
    spdFreeCursor(nextNode_cursor);

    if (nextNodeId != SPD_NON_ID) {
        return nextNodeId;
    }

    /* default٥뤫NOMATCHδϢ򤵤 */
    nextNode_cursor = spdGetRelInit(curNodeId, "next_node", "curId");
    do {
        rel = spdGetRel(nextNode_cursor);

        if (spdGetAttrValInt(rel.relationId, "sort") ==
            SPD_FV_E_SWITCH_DEFAULT) {
            nextNodeId = rel.objectId;
            break;
        } else if (spdGetAttrValInt(rel.relationId, "sort") ==
                   SPD_FV_E_SWITCH_NOMATCH) {
            nextNodeId = rel.objectId;
            break;
        }
    } while (rel.objectId != SPD_NON_ID);
    spdFreeCursor(nextNode_cursor);

    return nextNodeId;
}

/* ƻǤĤ */

static SpdObjId sipSkipFvNode(SpdObjId curNodeId)
{
    SpdObjId exprId = SPD_NON_ID;

    while((curNodeId = spdGetARelObj(curNodeId, "next_node", "curId"))
          != SPD_NON_ID) {
        if (spdGetClassId(curNodeId) == SPD_CID_EXPRESSION) {
            exprId = curNodeId;
            break;
        }
        
        /* ؿν꤬break롣 */
        if (spdGetClassId(curNodeId) == SPD_CID_BLK_EXIT &&
            spdGetAttrValInt(curNodeId, "sort") == SPD_FV_B_FUNC) {
            break;
        }
    }
    return exprId;
}

/* exprIdμ˼¹Ԥ뼰롥 */

static SpdObjId sipGetNextExpression(SpdObjId exprId, SpdValueData val)
{
    SpdObjId curNodeId, nextExprId = SPD_NON_ID;

    /* ޤʬʬ򤹤ޤ롣 */
    if (isBranchCondExpr(exprId) == TRUE) {
        curNodeId = getBranchNext(exprId, val);
    } else if (isSwitchCondExpr(exprId) == TRUE) {
        curNodeId = getSwitchNext(exprId, val);
    } else {
        curNodeId = exprId;
    }

    /* ȤϰƻʤΤexprФޤǤ롣 */
    if (curNodeId != exprId &&
        spdGetClassId(curNodeId) == SPD_CID_EXPRESSION) {
        nextExprId = curNodeId;
    } else {
        nextExprId = sipSkipFvNode(curNodeId);
    }
    
    if (nextExprId == SPD_NON_ID) {
        sipFreePoppedCallStack();
        sipSetTrvStackOfCallStack();
        return (sipGetProgramPtr());
    }

    return nextExprId;
}

/* Step execution. */

void sipGoStep(void)
{
    SpdObjId        prgPtr;
    SpdValueData    val;

    prgPtr = sipGetProgramPtr();
    val = sipEvalExpression(prgPtr);

    /* stepư(ؿƽФɾƤɾprgPtrۤʤ) */
    if (sipGetProgramPtr() == prgPtr) {
        sipSetProgramPtr(sipGetNextExpression(prgPtr, val));
    }
}
