/* 
 * Program:             $RCSfile: sint3.c,v $  $Revision: 40.5 $
 * 
 * Purpose:
 *
 * Author:              T.Saso  2001/12/11
 *                      S.Yamamoto  2001/12/11
 *
 * (C) Copyright:       T.Saso and S.Yamamoto  2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: sint3.c,v 40.5 2003/06/28 15:33:23 yamamoto Exp yamamoto $";
#endif

#include "SIP2.h"
#include "SIP2-proto.h"
#include <Sapid/ImodelARcompat.h>
#include <Sapid/PmodelAR.h>

static void usage(SpdConstString program_name)
{
    const char *usageMsg =
        "NAME\n"
        "    %s - Sapid INTerpreter\n"
        "SYNOPSIS\n"
        "    %s [%s options] [-- target program options]\n"
        "    -d dir      : Directory for SDB (\"SDB\").\n"
        "    -f function : Specify target function (\"main\").\n"
        "    -p program  : Specify target program name.\n"
        "    -v          : Print verbose message (OFF).\n";

    fprintf(stderr, usageMsg, program_name, program_name, program_name);
}

int main(int argc, char *argv[])
{
    SpdConstString  program_name;               /* This program name. */
    SpdObjId        progId, id;
    SpdDBId         dbId;
    SpdModelId      ImodelId, PmodelId, AreaModelId, FlowViewId;
    SpdConstString  targetProgName = NULL, targetFuncName = "main";
    int             c, i, sintArgc;
    char            *sintArgv[SIP_MAX_ARGC];

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);

    while ((c = getopt(argc, argv, "d:f:p:v")) != EOF) {
        switch (c) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'f':
            targetFuncName = optarg;
            break;
          case 'p':                         /* Target program name. */
            targetProgName = optarg;
            break;
          case 'v':                         /* Verbose mode. */
            sipSetDebugMode(ON);
            break;
          default:
            usage(program_name);
            return (EXIT_FAILURE);
            break;
        }
    }
    if (optind != argc && !SPD_STREQ(argv[optind - 1], "--")) {
        spdErrorExit("\"%s\": Invalid argument.\n", argv[optind]);
    }
    argc -= optind;
    argv += optind;
    optind = 1;

    ImodelId = spdUseModel(&I_model);
    PmodelId = spdUseModel(&P_model);
    AreaModelId = spdUseModel(&Area_model);
    FlowViewId = spdUseModel(&Flow_view);

    dbId = spdOpenSDB();
    spdReadDB(ImodelId, dbId);
    spdReadDB(PmodelId, dbId);
    spdCreateDB(AreaModelId, dbId);
    spdReadDB(FlowViewId, dbId);

    /* Set target program name. */
    if (targetProgName == NULL) { 
        if ((targetProgName = spdGetName(spdGetAnObjId("program", NULL))) ==
            NULL) {
            spdErrorExit("No program in SDB.\n");
        }
    }

    /* Set target program's arguments. */
    sintArgc = argc + 1;
    sintArgv[0] = (char *)targetProgName;
    for (i = 0; i < argc; i++) {
        sintArgv[i + 1] = argv[i];
    }
    if (sipGetDebugMode()) {
        spdMessage("init: sintArgc: %d.\n", sintArgc);
        for (i = 0; i < sintArgc; i++) {
            spdMessage("init: sintArgv[%d]: \"%s\".\n", i, sintArgv[i]);
        }
    }

    /* Set target program id. */
    if ((progId = spdGetAnObjId("program", targetProgName)) == SPD_NON_ID) {
        /* Program may be specified in full or relative path form. */
        if ((progId = spdGetAnObjId("program", spdBasename(targetProgName))) ==
            SPD_NON_ID) {
            spdErrorExit("\"%s\": No such program.\n", targetProgName);
        }

    }

    /* Set target function id. */
    if ((id = sipGetFuncIdentId(targetFuncName)) == SPD_NON_ID) {
        spdErrorExit("\"%s\": No such function.\n", targetFuncName);
    }
    sipBegin();
    spdSetGlobalStorage(spdCreateGlobalStorage(progId));
    sipSetProgramArgument(id, sintArgc, sintArgv);
    if (sipGetDebugMode()) {
        spdDwIndent();
        spdMessage("trac: sipCallFunction(): %s\n", spdGetName(id));
    }
    sipCallFunction(id);
    while ((id = sipGetProgramPtr()) != SPD_NON_ID) {
        if (sipGetDebugMode()) {
            spdDwIndent();
            spdMessage("trac: sipGoStep(): id %d.\n", id);
            {
                SpdOcc      occ;        /* Occurence in "*.i" file. */
                SpdRegion   region;     /* Region in "*.c" file. */
                int         line, column;

                /* idͿ줿鼰Фơ"*.i"νи롥 */
                occ = spdGetAnOffsetOfObj(id);
                if (occ.objectId != SPD_NON_ID) {
                    /* 
                     * spdDwIndent();
                     * spdMessage("trac: sipGoStep(): \"%s\", offset %d.\n",
                     *            spdGetName(occ.fileId), occ.offset);
                     */
                    /* "*.i"νи".c"regionѴ롥 */
                    region = spdConvRegionItoCbyPIDB(spdConvOccToRegion(occ));
                    /* 
                     * spdDwIndent();
                     * spdMessage("trac: sipGoStep(): \"%s\", offset %d.\n",
                     *            spdGetName(region.fileId), region.offset);
                     */
                    /* ".c"regionlinecolumnѴ롥 */
                    spdCnvOffsetToLineColumn(region.fileId, region.offset,
                                             &line, &column);
                    spdDwIndent();
                    spdMessage("trac: sipGoStep(): \"%s\", line %d.\n",
                               spdGetName(region.fileId), line);
                }
            }
        }
        sipGoStep();
    }
    if (sipGetDebugMode()) {
        spdMessage("trac: sipEnd():\n");
    }
    sipEnd();

    spdCloseSDB(dbId);
    return (EXIT_SUCCESS);
}
