/* 
 * Program:             $RCSfile: SaQuLa.h,v $  $Revision: 30.4 $
 * 
 * Purpose:             "Imakefile" of Sapid/tool/SIP/sip.
 *
 * Author:              N.Kurauchi  1997/02/08
 *                      S.Yamamoto  1996/02/08
 *
 * (C) Copyright:       N.Kurauchi and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#include <Sapid/Sapid.h>
#include <Sapid/sbdd.h>

#define INPUT_FILE_MAX_SIZE 1024
#define SYMTABLE_MAX        50
#define BDDTABLE_MAX        100
#define SUFTABLE_MAX        100
#define UNCOMPARISON        200
#define SBDDNODE_MAX        10000
#define SUF_SPD_EMPTY_ARRAY {0, NULL, NULL}
#define BUF_MAX             1024

/* define for symtable.type */
#define UNDEFINED -1
/* #define NIL 0 */
#define STRINGTABLE 1
#define INTTABLE 2
#define OBJIDTABLE 3
#define RELIDTABLE 4
#define OCCTABLE 5
#define SYMTABLE 6
#define BDDTABLE 7

/* define for symtable.meta_type */
#define ATOM 0
#define SET 1
#define BAG 2
#define LIST 3
#define ARRAY 4
#define STRUCTURE 5

/* define for reservetable.funcnum */
#define GET_STRING_ATTRIBUTE 10
#define GET_INT_ATTRIBUTE 11
#define GET_BY_REL 12

/* define for operator_name */
#define EQUAL 101
#define NOTEQUAL 102
#define NOTOP 200
#define ANDOP 201
#define OROP 202
#define XOROP 203
#define NANDOP 204
#define NOROP 205
#define XNOROP 206
#define GETOCC 100

/* define for masternum */
#define UNSUBORDINARY -2

typedef int ValueType;
typedef int Ope;

typedef struct _ObjIdValue {
    int masternum;
    SpdObjId id;
    int offset; /* for have_no_offset_attr class */
    SpdRelId rel_id; /* for getting offset */
} ObjIdValue;

typedef struct _StringValue {
    int masternum;
    SpdString value;
} StringValue;

typedef struct _IntValue {
    int masternum;
    int value;
} IntValue;

typedef struct _SaqulaObjIdArray {
    size_t size;
    ObjIdValue *oiv;
} SaqulaObjIdArray;

typedef struct _StringArray {
    size_t size;
    StringValue *sv;
} StringArray;

typedef struct _IntArray {
    size_t size;
    IntValue *iv;
} IntArray;

typedef struct _SpdBooleanArray {
    size_t size;
    SpdBoolean *value;
} SpdBooleanArray;

typedef struct _SpdBooleanArrayArray {
    size_t size;
    SpdBooleanArray *array;
} SpdBooleanArrayArray;

typedef struct _SpdIdStruct {
    size_t size;
    SpdObjId *master;
    SpdObjIdArray *objid_table;
} SpdIdStruct;

typedef struct _SpdOccStruct {
    size_t size;
    SpdObjId *master;
    SpdOccArray *occ_table;
} SpdOccStruct;

typedef struct _SpdStringStruct {
    size_t size;
    SpdObjId *master;
    StringArray *string_table;
} SpdStringStruct;

typedef union _Value {
    SaqulaObjIdArray objid_table;
    SpdRelIdArray relid_table;
    SpdOccArray occ_table;
    StringArray string_table;
    IntArray int_table;
    SpdBooleanArray bool_table;
    /* structure specification need to be changed */
    SpdIdStruct id_struct;
    SpdOccStruct occ_struct;
    SpdStringStruct string_struct;
} Value;

typedef struct _SymTable {
    SpdString name;
    Value value;
    ValueType type;
    ValueType meta_type;
    SpdBoolean isvariable;
    void *msp;
} SymTable;

typedef struct _SymTableArray {
    size_t size;
    SymTable **sp;
} SymTableArray;

typedef union _AttrValue {
    int integer;
    SpdString string;
} AttrValue;

typedef struct _CondAttrTable {
    size_t size;
    SpdString *attr_name;
    SpdBoolean *is_int;
    AttrValue *attrvalue;
} CondAttrTable;

typedef int FuncNum;

typedef struct _ArgList {
    SpdString stringarg[2];
    SymTable *symtablearg;
    SymTable *resultsp;
} MyArgList;

typedef struct _ReserveTable {
    size_t size;
    FuncNum *funcnum;
    MyArgList *arglist;
} ReserveTable;

typedef struct _SaqulaResult {
    Value value;
    ValueType type;
    ValueType meta_type;
} SaqulaResult;

typedef struct _BddTable {
    int val_num;
    bddp bddp;
    void *rs, *ls;
    Ope ope;
} BddTable;

typedef struct _SufNumArray {
    size_t size;
    int *num;
} SufNumArray;

typedef struct _SufTable {
    size_t size;
    SymTable **sp;
    SufNumArray *num;
} SufTable;

SaqulaResult *saqula(SpdString , SpdString);
void saqulaInit(void);
void saqulaQuit(void);

SymTable *get_by_condattr(SymTable *);
SymTable *get_by_rel(SymTable *, SpdString, SpdString);
SymTable *get_extent(SpdString);
/* SymTable *get_relation_extent(SpdString );*/
SymTable *get_int_attribute(SymTable *, SpdString);
SymTable *get_string_attribute(SymTable *, SpdString);

SymTable *getOcc(SymTable *, SpdString);
SymTable *convItoC(SymTable *);

SymTable *get_int_constant(int);
SymTable *get_string_constant(SpdString);
void print_condattrtableint(void);
void print_condattrtablestr(void);
void print_int_attribute(SymTable *);
void print_string_attribute(SymTable *);
void print_symtable(SymTable *);
void set_variable(SymTable *, SymTable *);
SymTable *symlook(SpdString);

void set_condattrtablestr(SpdString, SpdString);
void set_condattrtableint(SpdString, int);

void do_by_reservetable(void);

void setValue(SymTable *);

BddTable *bddlook(Ope, void *, void *);
SpdBooleanArrayArray getSufResult(BddTable *);
void checkBddTable(SpdBooleanArrayArray);
BddTable *getBddTable(int);
void checkCom(Ope, SymTable *, SymTable *);
void putSufSp(SymTable *, int);
void putSufSpToPost(SymTable *, int);
SymTable *getSufQuery(void);
void setSufResult(void);
void setValueToSp(SymTable *, SymTable *);
