%{

/* 
 * Program:             $RCSfile: lex.l,v $  $Revision: 30.3 $
 * 
 * Purpose:             "Imakefile" of Sapid/tool/SIP/sip.
 *
 * Author:              N.Kurauchi  1997/02/08
 *                      S.Yamamoto  1996/02/08
 *
 * (C) Copyright:       N.Kurauchi and S.Yamamoto  1997 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: lex.l,v 30.3 2002/10/21 07:28:30 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "SaQuLa.h"
#include "y.tab.h"

#undef YY_INPUT
#define YY_INPUT(b, r, ms)  ((r) = my_yyinput((b), (ms)))

char    *buf;
int     lineno = 1;

extern SpdBoolean   first_appear;

#if !defined(yywrap)
#define yywrap()            1               /* Treat EOF as an EOF. */
#endif

int my_yyinput(char *buf, int max_size);

%}

ws [ \t]+
comment #.*
nl \n

object (program|file|declaration|declarator|identifier|optionaldecl|block|expression|label|constant|type|operator)

intattribute (id|sort|storage|length|lrtype|type|qualifier|size|sign|offset)
stringattribute (name|document|value)

relation (blk_blk|blk_decl|blk_expr|blk_label|case_cexpr|const_ref|decl_decl|decl_ident|decl_opt|decl_type|expr_expr|expr_op|expr_type|file_decl|func_body|func_parm|goto_labl|ident_ref|ident_type|init_expr|op_type|prog_file|prot_type|struct_decl|type_cexpr|type_is|type_st)

operator (GetOcc|ConvItoC)

sortby (offsets|names)
    
rolename [a-z]+_[a-z]+(_[a-z]+)*

variable [A-Za-z][A-Za-z0-9]*
intconstant [1-9][0-9]*
stringconstant \"[A-Za-z_]*\"

dot (\.|->)
eq =
noteq !=

%%

{ws} ;
{nl} { lineno++; }
{comment} ;

DEFINE      { return DEFINE; }
AS          { return AS; }
SELECT      { return SELECT; }
FROM        { return FROM; }
IN          { return IN; }
WHERE       { return WHERE; }
SORT        { return SORT; }
BY          { return BY; }
NOT         { return NOT; }
AND         { return AND; }
OR          { return OR; }
XOR         { return XOR; }
NAND        { return NAND; }
NOR         { return NOR; }
XNOR        { return XNOR; }

{dot}       { return DOT; }
{eq}        { return EQ; }
{noteq}     { return NOTEQ; }

{object}s {
    yylval.name = spdStrdup(yytext);
    return OBJECT_EXTENT;
}

{object} {
    yylval.name = spdStrdup(yytext);
    return OBJECT;
}

{relation}s {
    yylval.name = spdStrdup(yytext);
    return RELATION_EXTENT;
}

{operator} {
    yylval.name = spdStrdup(yytext);
    return OPERATOR_NAME;
}

{sortby} {
    yylval.name = spdStrdup(yytext);
    return SORTBY;
}

{intattribute} {
    yylval.name = spdStrdup(yytext);
    return INTATTRIBUTE;
}

{stringattribute} {
    yylval.name = spdStrdup(yytext);
    return STRINGATTRIBUTE;
}

{relation} {
    yylval.name = spdStrdup(yytext);
    return RELATION;
}

{rolename} {
    yylval.name = spdStrdup(yytext);
    return ROLENAME;
}

{intconstant} {
    yylval.intval = atoi(yytext);
    return INTCONSTANT;
}

SPD_NON_ID {
    yylval.intval = SPD_NON_ID;
    return INTCONSTANT;
}

DECL_NORMAL {
    yylval.intval = DECL_NORMAL;
    return INTCONSTANT;
}

DECL_FUNCDEF {
    yylval.intval = DECL_FUNCDEF;
    return INTCONSTANT;
}

DECL_ENUMERATOR {
    yylval.intval = DECL_ENUMERATOR;
    return INTCONSTANT;
}

DECL_PARAMETER {
    yylval.intval = DECL_PARAMETER;
    return INTCONSTANT;
}

DECL_MEMBER {
    yylval.intval = DECL_MEMBER;
    return INTCONSTANT;
}

DECL_TYPE {
    yylval.intval = DECL_TYPE;
    return INTCONSTANT;
}

STORAGE_NONE {
    yylval.intval = STORAGE_NONE;
    return INTCONSTANT;
}

STORAGE_AUTO {
    yylval.intval =STORAGE_AUTO;
    return INTCONSTANT;
}

STORAGE_EXTERN {
    yylval.intval = STORAGE_EXTERN;
    return INTCONSTANT;
}

STORAGE_STATIC {
    yylval.intval = STORAGE_STATIC;
    return INTCONSTANT;
}

STORAGE_REGISTER {
    yylval.intval = STORAGE_REGISTER;
    return INTCONSTANT;
}

STORAGE_TYPEDEF {
    yylval.intval = STORAGE_TYPEDEF;
    return INTCONSTANT;
}

DECL_VARIABLE {
    yylval.intval = DECL_VARIABLE;
    return INTCONSTANT;
}

DECL_PROTOTYPE {
    yylval.intval = DECL_PROTOTYPE;
    return INTCONSTANT;
}

DECL_TAG {
    yylval.intval = DECL_TAG;
    return INTCONSTANT;
}

ID_VARIABLE {
    yylval.intval = ID_VARIABLE;
    return INTCONSTANT;
}

ID_FUNCTION {
    yylval.intval = ID_FUNCTION;
    return INTCONSTANT;
}

ID_MEMBER {
    yylval.intval = ID_MEMBER;
    return INTCONSTANT;
}

ID_ENUMERATOR {
    yylval.intval = ID_ENUMERATOR;
    return INTCONSTANT;
}

ID_TYPENAME {
    yylval.intval = ID_TYPENAME;
    return INTCONSTANT;
}

ID_TAG {
    yylval.intval = ID_TAG;
    return INTCONSTANT;
}

OPT_PROTOTYPE {
    yylval.intval = OPT_PROTOTYPE;
    return INTCONSTANT;
}

OPT_FUNCDEF {
    yylval.intval = OPT_FUNCDEF;
    return INTCONSTANT;
}

OPT_INITIALIZER {
    yylval.intval = OPT_INITIALIZER;
    return INTCONSTANT;
}

OPT_STRUCT {
    yylval.intval = OPT_STRUCT;
    return INTCONSTANT;
}

BLOCK_BASIC {
    yylval.intval = BLOCK_BASIC;
    return INTCONSTANT;
}

BLOCK_COMP {
    yylval.intval = BLOCK_COMP;
    return INTCONSTANT;
}

BLOCK_IF {
    yylval.intval = BLOCK_IF;
    return INTCONSTANT;
}

BLOCK_SWITCH {
    yylval.intval = BLOCK_SWITCH;
    return INTCONSTANT;
}

BLOCK_FOR {
    yylval.intval = BLOCK_FOR;
    return INTCONSTANT;
}

BLOCK_WHILE {
    yylval.intval = BLOCK_WHILE;
    return INTCONSTANT;
}

BLOCK_DO {
    yylval.intval = BLOCK_DO;
    return INTCONSTANT;
}

BLOCK_BRANCH {
    yylval.intval = BLOCK_BRANCH;
    return INTCONSTANT;
}

BLOCK_HIERACHICAL {
    yylval.intval = BLOCK_HIERACHICAL;
    return INTCONSTANT;
}

EXPR_EMPTY {
    yylval.intval = EXPR_EMPTY;
    return INTCONSTANT;
}

EXPR_BREAK {
    yylval.intval = EXPR_BREAK;
    return INTCONSTANT;
}

EXPR_CONTINUE {
    yylval.intval = EXPR_CONTINUE;
    return INTCONSTANT;
}

EXPR_RETURN {
    yylval.intval = EXPR_RETURN;
    return INTCONSTANT;
}

EXPR_GOTO {
    yylval.intval = EXPR_GOTO;
    return INTCONSTANT;
}

EXPR_SORT_ARITH {
    yylval.intval = EXPR_SORT_ARITH;
    return INTCONSTANT;
}

EXPR_SORT_LOGICAL {
    yylval.intval = EXPR_SORT_LOGICAL;
    return INTCONSTANT;
}

EXPR_SORT_ARRAY {
    yylval.intval = EXPR_SORT_ARRAY;
    return INTCONSTANT;
}

EXPR_SORT_MEMBER {
    yylval.intval = EXPR_SORT_MEMBER;
    return INTCONSTANT;
}

EXPR_SORT_POINTER {
    yylval.intval = EXPR_SORT_POINTER;
    return INTCONSTANT;
}

EXPR_SORT_ADDRESS {
    yylval.intval = EXPR_SORT_ADDRESS;
    return INTCONSTANT;
}

EXPR_SORT_TRI {
    yylval.intval = EXPR_SORT_TRI;
    return INTCONSTANT;
}

EXPR_SORT_SIZEOF {
    yylval.intval = EXPR_SORT_SIZEOF;
    return INTCONSTANT;
}

EXPR_SORT_CAST {
    yylval.intval = EXPR_SORT_CAST;
    return INTCONSTANT;
}

EXPR_SORT_COMMA {
    yylval.intval = EXPR_SORT_COMMA;
    return INTCONSTANT;
}

EXPR_SORT_FUNCCALL {
    yylval.intval = EXPR_SORT_FUNCCALL;
    return INTCONSTANT;
}

EXPR_SORT_INITLIST {
    yylval.intval = EXPR_SORT_INITLIST;
    return INTCONSTANT;
}

EXPR_SORT_BUILTIN {
    yylval.intval = EXPR_SORT_BUILTIN;
    return INTCONSTANT;
}

EXPR_SORT_BIT {
    yylval.intval = EXPR_SORT_BIT;
    return INTCONSTANT;
}

EXPR_SORT_SHIFT {
    yylval.intval = EXPR_SORT_SHIFT;
    return INTCONSTANT;
}

EXPR_SORT_COMPARE {
    yylval.intval = EXPR_SORT_COMPARE;
    return INTCONSTANT;
}

EXPR_L_VALUE {
    yylval.intval = EXPR_L_VALUE;
    return INTCONSTANT;
}

EXPR_R_VALUE {
    yylval.intval = EXPR_R_VALUE;
    return INTCONSTANT;
}

EXPR_PL_VALUE {
    yylval.intval = EXPR_PL_VALUE;
    return INTCONSTANT;
}

EXPR_PR_VALUE {
    yylval.intval = EXPR_PR_VALUE;
    return INTCONSTANT;
}

EXPR_P_REF {
    yylval.intval = EXPR_P_REF;
    return INTCONSTANT;
}

EXPR_LR_VALUE {
    yylval.intval = EXPR_LR_VALUE;
    return INTCONSTANT;
}

EXPR_PLR_VALUE {
    yylval.intval = EXPR_PLR_VALUE;
    return INTCONSTANT;
}

LABEL_NAMED {
    yylval.intval = LABEL_NAMED;
    return INTCONSTANT;
}

LABEL_CASE {
    yylval.intval = LABEL_CASE;
    return INTCONSTANT;
}

LABEL_DEFAULT {
    yylval.intval = LABEL_DEFAULT;
    return INTCONSTANT;
}

CONST_INT {
    yylval.intval = CONST_INT;
    return INTCONSTANT;
}

CONST_FLOAT {
    yylval.intval = CONST_FLOAT;
    return INTCONSTANT;
}

CONST_CHAR {
    yylval.intval = CONST_CHAR;
    return INTCONSTANT;
}

CONST_STRING {
    yylval.intval = CONST_STRING;
    return INTCONSTANT;
}

TYPE_UNKNOWN {
    yylval.intval = TYPE_UNKNOWN;
    return INTCONSTANT;
}

TYPE_TYPE_DEF {
    yylval.intval = TYPE_TYPE_DEF;
    return INTCONSTANT;
}

TYPE_STRUCT_TAG {
    yylval.intval = TYPE_STRUCT_TAG;
    return INTCONSTANT;
}

TYPE_UNION_TAG {
    yylval.intval = TYPE_UNION_TAG;
    return INTCONSTANT;
}

TYPE_DECLARATION {
    yylval.intval = TYPE_DECLARATION;
    return INTCONSTANT;
}

TYPE_POINTER {
    yylval.intval = TYPE_POINTER;
    return INTCONSTANT;
}

TYPE_ARRAY {
    yylval.intval = TYPE_ARRAY;
    return INTCONSTANT;
}

TYPE_FUNC {
    yylval.intval = TYPE_FUNC;
    return INTCONSTANT;
}

TYPE_ENUM_SPECIFIER {
    yylval.intval = TYPE_ENUM_SPECIFIER;
    return INTCONSTANT;
}

TYPE_CONST {
    yylval.intval = TYPE_CONST;
    return INTCONSTANT;
}

TYPE_VOLATILE {
    yylval.intval = TYPE_VOLATILE;
    return INTCONSTANT;
}

TYPE_VOID {
    yylval.intval = TYPE_VOID;
    return INTCONSTANT;
}

TYPE_INT {
    yylval.intval = TYPE_INT;
    return INTCONSTANT;
}

TYPE_CHAR {
    yylval.intval = TYPE_CHAR;
    return INTCONSTANT;
}

TYPE_FLOAT {
    yylval.intval = TYPE_FLOAT;
    return INTCONSTANT;
}

TYPE_DOUBLE {
    yylval.intval = TYPE_DOUBLE;
    return INTCONSTANT;
}

TYPE_SHORT {
    yylval.intval = TYPE_SHORT;
    return INTCONSTANT;
}

TYPE_LONG {
    yylval.intval = TYPE_LONG;
    return INTCONSTANT;
}

TYPE_SIGNED {
    yylval.intval = TYPE_SIGNED;
    return INTCONSTANT;
}

TYPE_UNSIGNED {
    yylval.intval = TYPE_UNSIGNED;
    return INTCONSTANT;
}

OP_L_PAR {
    yylval.intval = OP_L_PAR;
    return INTCONSTANT;
}

OP_R_PAR {
    yylval.intval = OP_R_PAR;
    return INTCONSTANT;
}

OP_L_ARRAY {
    yylval.intval = OP_L_ARRAY;
    return INTCONSTANT;
}

OP_R_ARRAY {
    yylval.intval = OP_R_ARRAY;
    return INTCONSTANT;
}

OP_ARROW {
    yylval.intval = OP_ARROW;
    return INTCONSTANT;
}

OP_DOT {
    yylval.intval = OP_DOT;
    return INTCONSTANT;
}

OP_NOT {
    yylval.intval = OP_NOT;
    return INTCONSTANT;
}

OP_B_NOT {
    yylval.intval = OP_B_NOT;
    return INTCONSTANT;
}

OP_INC {
    yylval.intval = OP_INC;
    return INTCONSTANT;
}

OP_DEC {
    yylval.intval = OP_DEC;
    return INTCONSTANT;
}

OP_PLUS {
    yylval.intval = OP_PLUS;
    return INTCONSTANT;
}

OP_MINUS {
    yylval.intval = OP_MINUS;
    return INTCONSTANT;
}

OP_POINTER {
    yylval.intval = OP_POINTER;
    return INTCONSTANT;
}

OP_ADDRESS {
    yylval.intval = OP_ADDRESS;
    return INTCONSTANT;
}

OP_MUL {
    yylval.intval = OP_MUL;
    return INTCONSTANT;
}

OP_DIV {
    yylval.intval = OP_DIV;
    return INTCONSTANT;
}

OP_MOD {
    yylval.intval = OP_MOD;
    return INTCONSTANT;
}

OP_ADD {
    yylval.intval = OP_ADD;
    return INTCONSTANT;
}

OP_SUB {
    yylval.intval = OP_SUB;
    return INTCONSTANT;
}

OP_SHL {
    yylval.intval = OP_SHL;
    return INTCONSTANT;
}

OP_SHR {
    yylval.intval = OP_SHR;
    return INTCONSTANT;
}

OP_LT {
    yylval.intval = OP_LT;
    return INTCONSTANT;
}

OP_LE {
    yylval.intval = OP_LE;
    return INTCONSTANT;
}

OP_GT {
    yylval.intval = OP_GT;
    return INTCONSTANT;
}

OP_GE {
    yylval.intval = OP_GE;
    return INTCONSTANT;
}

OP_EQ {
    yylval.intval = OP_EQ;
    return INTCONSTANT;
}

OP_NE {
    yylval.intval = OP_NE;
    return INTCONSTANT;
}

OP_B_AND {
    yylval.intval = OP_B_AND;
    return INTCONSTANT;
}

OP_B_XOR {
    yylval.intval = OP_B_XOR;
    return INTCONSTANT;
}

OP_B_OR {
    yylval.intval = OP_B_OR;
    return INTCONSTANT;
}

OP_L_AND {
    yylval.intval = OP_L_AND;
    return INTCONSTANT;
}

OP_L_OR {
    yylval.intval = OP_L_OR;
    return INTCONSTANT;
}

OP_QUESTION {
    yylval.intval = OP_QUESTION;
    return INTCONSTANT;
}

OP_COLON {
    yylval.intval = OP_COLON;
    return INTCONSTANT;
}

OP_ASSIGN {
    yylval.intval = OP_ASSIGN;
    return INTCONSTANT;
}

OP_A_ADD {
    yylval.intval = OP_A_ADD;
    return INTCONSTANT;
}

OP_A_SUB {
    yylval.intval = OP_A_SUB;
    return INTCONSTANT;
}

OP_A_MUL {
    yylval.intval = OP_A_MUL;
    return INTCONSTANT;
}

OP_A_DIV {
    yylval.intval = OP_A_DIV;
    return INTCONSTANT;
}

OP_A_MOD {
    yylval.intval = OP_A_MOD;
    return INTCONSTANT;
}

OP_A_AND {
    yylval.intval = OP_A_AND;
    return INTCONSTANT;
}

OP_A_XOR {
    yylval.intval = OP_A_XOR;
    return INTCONSTANT;
}

OP_A_OR {
    yylval.intval = OP_A_OR;
    return INTCONSTANT;
}

OP_A_SHL {
    yylval.intval = OP_A_SHL;
    return INTCONSTANT;
}

OP_A_SHR {
    yylval.intval = OP_A_SHR;
    return INTCONSTANT;
}

OP_COMMA {
    yylval.intval =OP_COMMA;
    return INTCONSTANT;
}

OP_SIZEOF {
    yylval.intval = OP_SIZEOF;
    return INTCONSTANT;
}

OP_SIZEOF_TYPE {
    yylval.intval = OP_SIZEOF_TYPE;
    return INTCONSTANT;
}

OP_CAST {
    yylval.intval = OP_CAST;
    return INTCONSTANT;
}

{stringconstant} {
    buf = spdMalloc(strlen(yytext));
    strcpy(buf, &yytext[1]);
    buf[strlen(buf)-1] = '\0';
    yylval.name = spdStrdup(buf);
    spdFree(buf);
    return STRINGCONSTANT;
}

{variable} {
    yylval.symtable = symlook(yytext);
    if(first_appear == TRUE){
        yylval.symtable->isvariable = TRUE;
        yylval.symtable->msp = NULL;
    }
    return VARIABLE;
}

. { return yytext[0]; }

%%

extern char myinput[];
extern char *myinputptr;
extern char *myinputlim;

extern int lineno;

int my_yyinput(char *buf, int max_size)
{
    int n;

    if (max_size < (myinputlim-myinputptr)) {
        n = max_size;
    } else {
        n = (int)(myinputlim - myinputptr);
    }

    if(n > 0){
        memcpy(buf, myinputptr, n);
        myinputptr += n;
    }
    return n;
}

void yyerror(const char *msg)
{
    printf("%d: %s at '%s'\n", lineno, msg, yytext);
}
