/* 
 * Program:             $RCSfile: saqula.c,v $  $Revision: 30.5 $
 * 
 * Purpose:             "Imakefile" of Sapid/tool/SIP/sip.
 *
 * Author:              N.Kurauchi  1997/02/08
 *                      S.Yamamoto  1997/02/08
 *
 * (C) Copyright:       N.Kurauchi and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: saqula.c,v 30.5 2003/07/20 11:47:50 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "SaQuLa.h"

char            *myinput;
char            *myinputptr;
int             myinputlim;
SpdBoolean      first_appear;
SpdObjId        prog_id;
SymTable        symtable[SYMTABLE_MAX];
CondAttrTable   condattrtable;
ReserveTable    reservetable;
SaqulaResult    result;
BddTable        bddtable[BDDTABLE_MAX];
int             val_num = 1;

SpdBoolean      p_option = 0;

int yyparse(void);
void yyerror(char *msg);

SaqulaResult *saqula(SpdString progname, SpdString input)
{

    if(p_option == 0){
        if((prog_id = spdTarget("I-model", progname)) == SPD_NON_ID){
            spdErrorExit("program \"%s\" not found.",progname);
        }
    } else {
        if((prog_id = spdTarget("P-model",progname)) == SPD_NON_ID){
            spdErrorExit("program \"%s\" not found.",progname);
        }
    }

    condattrtable.size = 0;
    reservetable.size = 0;

    /* initialize for flex */
    myinput = (char *)spdMalloc(sizeof(char)*1024);
    strcpy(myinput,input);
    myinputptr = myinput;
    myinputlim = (int)&myinput[strlen(myinput)];

    /* initialize for sbdd */
    bddinit(SBDDNODE_MAX);
    
    yyparse();
    return &result;
}

void saqulaInit(void)
{
}

void saqulaQuit(void)
{
}

void setValue(SymTable *arg1)
{
    if(arg1){
        result.type = arg1->type;
        result.meta_type = arg1->meta_type;
        result.value = arg1->value;
    }
}

SymTable *get_extent(SpdString buf)
{
    SpdString extent_name = NULL;
    SpdCursor *csr;
    SpdObjId objid;
    SymTable *sp;
    
    sp = symlook(buf);
    if(first_appear == FALSE) return(sp);
    
    extent_name = spdMalloc(strlen(buf));
    strncpy(extent_name, buf, strlen(buf)-1);
    extent_name[strlen(buf)-1] = '\0';

    sp->type = OBJIDTABLE;
    sp->meta_type = SET;
    sp->msp = NULL;
    sp->value.objid_table.oiv = NULL;
    sp->value.objid_table.size = 0;
    
    if(SPD_STREQ(buf, "program")){
        sp->value.objid_table.oiv
            = spdRealloc(sp->value.objid_table.oiv,
                         sizeof(ObjIdValue));
        sp->value.objid_table.oiv[0].masternum = UNSUBORDINARY;
        sp->value.objid_table.oiv[0].id = prog_id;
        sp->value.objid_table.size = 1;
        return(sp);
    }

    csr = spdGetObjIdInit(extent_name, NULL);
    while ((objid = spdGetObjId(csr)) != SPD_NON_ID){
        sp->value.objid_table.oiv
            = spdRealloc(sp->value.objid_table.oiv,
                         sizeof(ObjIdValue)*(sp->value.objid_table.size+1));
        sp->value.objid_table.oiv[sp->value.objid_table.size].masternum
            = UNSUBORDINARY;
        sp->value.objid_table.oiv[sp->value.objid_table.size].id
            = objid;
        sp->value.objid_table.oiv[sp->value.objid_table.size].offset
            = SPD_NON_OFFSET;
        sp->value.objid_table.oiv[sp->value.objid_table.size].rel_id
            = SPD_NON_ID;
        sp->value.objid_table.size++;
    }
    spdFreeCursor(csr);
    return(sp);
}

/* 
 * SymTable *get_relation_extent(buf)
 * char *buf;
 * {
 *     SymTable *sp;
 *     char *name = NULL;
 * 
 *     name = spdMalloc(strlen(buf));
 *     strncpy(name, buf, strlen(buf)-1);
 *     name[strlen(buf)-1] = '\0';
 * 
 *     sp = symlook(buf);
 *     if(first_appear == FALSE) return(sp);
 * 
 *     sp->type = RELIDTABLE;
 * 
 *     return(sp);
 * }
 */

SymTable *get_int_attribute(SymTable *arg1, SpdString arg2)
{
    SymTable *sp;
    char *table_name = NULL;
    int i;

    table_name = spdMalloc(strlen(arg1->name)+strlen(arg2)+2);
    sprintf(table_name, "%s.%s", arg1->name, arg2);
    sp = symlook(table_name);
    if((first_appear == FALSE)&&(sp->type != UNDEFINED)) return(sp);

    if(arg1->msp == NULL) sp->msp = arg1;
    else sp->msp = arg1->msp;

    if(arg1->type == UNDEFINED){
        i = reservetable.size;
        reservetable.funcnum = (FuncNum *)spdRealloc(reservetable.funcnum,
                                                     sizeof(FuncNum)*(i+1));
        reservetable.funcnum[i] = GET_INT_ATTRIBUTE;
        reservetable.arglist
            = (MyArgList *)spdRealloc(reservetable.arglist,
                                    sizeof(MyArgList)*(i+1));
        reservetable.arglist[i].stringarg[0]
            = (SpdString)spdMalloc(strlen(arg2)+1);
        strcpy(reservetable.arglist[i].stringarg[0], arg2);
        reservetable.arglist[i].symtablearg = arg1;
        reservetable.arglist[i].resultsp = sp;
        reservetable.size++;
        return sp;
    }

    sp->type = INTTABLE;
    sp->meta_type = arg1->meta_type;
    sp->value.int_table.iv = NULL;

    for(i=0;i<arg1->value.objid_table.size;i++){
        sp->value.int_table.iv
            = (IntValue *)spdRealloc(sp->value.int_table.iv,
                                     sizeof(IntValue)*(i+1));
        if(arg1->value.objid_table.oiv[i].masternum
           == UNSUBORDINARY)
            sp->value.int_table.iv[i].masternum = i;
        else sp->value.int_table.iv[i].masternum
            = arg1->value.objid_table.oiv[i].masternum;
        if(SPD_STREQ(arg2, "offset")){
            sp->value.int_table.iv[i].value
                = arg1->value.objid_table.oiv[i].offset;
        }
        else{
            sp->value.int_table.iv[i].value
                = spdGetAttrValInt(arg1->value.objid_table.oiv[i].id,
                                   arg2);
        }
    }
    sp->value.int_table.size = i;
    return(sp);
}

SymTable *get_string_attribute(SymTable *arg1, SpdString arg2)
{
    SymTable *sp;
    char *table_name = NULL, buf[BUF_MAX];
    int i;
    
    table_name = spdMalloc(strlen(arg1->name)+strlen(arg2)+2);
    sprintf(table_name, "%s.%s", arg1->name, arg2);
    sp = symlook(table_name);
    if((first_appear == FALSE)&&(sp->type != UNDEFINED)) return(sp);
    
    if(arg1->msp == NULL) sp->msp = arg1;
    else sp->msp = arg1->msp;

    /* set reservetable if arg1 == UNDEFINED */
    if(arg1->type == UNDEFINED){
        i = reservetable.size;
        reservetable.funcnum = spdRealloc(reservetable.funcnum,
                                          sizeof(FuncNum)*(i+1));
        reservetable.funcnum[i] = GET_STRING_ATTRIBUTE;
        reservetable.arglist = spdRealloc(reservetable.arglist,
                                          sizeof(MyArgList)*(i+1));
        reservetable.arglist[i].stringarg[0]
            = spdMalloc(strlen(arg2)+1);
        strcpy(reservetable.arglist[i].stringarg[0], arg2);
        reservetable.arglist[i].symtablearg = arg1;
        reservetable.arglist[i].resultsp = sp;
        reservetable.size++;
        return sp;
    }
    
    sp->type = STRINGTABLE;
    sp->meta_type = arg1->meta_type;
    sp->value.string_table.sv = NULL;

    for(i=0;i<arg1->value.objid_table.size;i++){
        sp->value.string_table.sv
            = (StringValue *)spdRealloc(sp->value.string_table.sv,
                                        sizeof(StringValue)*(i+1));
        if(arg1->value.objid_table.oiv[i].masternum
           == UNSUBORDINARY)
            sp->value.string_table.sv[i].masternum = i;
        else sp->value.string_table.sv[i].masternum
            = arg1->value.objid_table.oiv[i].masternum;
        strcpy(buf, spdGetAttrValString(arg1->value.objid_table.oiv[i].id,
                                        arg2));
        sp->value.string_table.sv[i].value
            = (SpdString)spdMalloc(strlen(buf)+1);
        strcpy(sp->value.string_table.sv[i].value, buf);
    }
    sp->value.string_table.size = i;
    return(sp);
}

SymTable *get_by_condattr(SymTable *arg1)
{
    SymTable *sp;
    int i, j, k;
    char buf[BUF_MAX];
    SpdBoolean satisfy;
    
    if(condattrtable.size == 0) return(arg1);
    
    strcpy(buf, arg1->name);
    for(i=0;i<condattrtable.size;i++){
        strcpy(&buf[strlen(buf)], condattrtable.attr_name[i]);
        if(condattrtable.is_int[i] == TRUE){
            sprintf(&buf[strlen(buf)], "%d",
                    condattrtable.attrvalue[i].integer);
        } else{
            strcpy(&buf[strlen(buf)], condattrtable.attrvalue[i].string);
        }
    }

    sp = symlook(buf);
    if((first_appear == FALSE)&&(sp->type != UNDEFINED)) return(sp);

    sp->type = OBJIDTABLE;

    k=0;
    sp->value.objid_table.size = 0;
    sp->value.objid_table.oiv = NULL;
    for(j=0;j<arg1->value.objid_table.size;j++){
    
        satisfy = TRUE;
        for(i=0;i<condattrtable.size;i++){
            if(condattrtable.is_int[i] == TRUE){
                if(condattrtable.attrvalue[i].integer
                   != spdGetAttrValInt(arg1->value.objid_table.oiv[j].id,
                                       condattrtable.attr_name[i]))
                    satisfy = FALSE;
            }
            else{
                if(strcmp(condattrtable.attrvalue[i].string,
                          spdGetAttrValString(arg1->value.objid_table.oiv[j].id,
                                              condattrtable.attr_name[i])))
                    satisfy = FALSE;
            }
        }
        if(satisfy == TRUE){
            sp->value.objid_table.oiv
                = spdRealloc(sp->value.objid_table.oiv,
                             sizeof(ObjIdValue)*(k+1));
            sp->value.objid_table.oiv[k].id
                = arg1->value.objid_table.oiv[j].id;
            k++;
            sp->value.objid_table.size = k;
        }
    }
    
    condattrtable.size = 0;
    return(sp);
}

SymTable *get_by_rel(SymTable *arg1, SpdString arg2, SpdString arg3)
{
    SymTable *sp;
    char buf[BUF_MAX];
    int i;
    SpdCursor *csr;
    SpdObjId id;
    SpdRelId rel_id;                        /* for offset */

    strcpy(buf, arg1->name);
    strcpy(&buf[strlen(buf)], arg2);
    strcpy(&buf[strlen(buf)], arg3);

    sp = symlook(buf);
    if((first_appear == FALSE)&&(sp->type != UNDEFINED)) return(sp);
    
    if(arg1->msp == NULL) sp->msp = arg1;
    else sp->msp = arg1->msp;

    if(arg1->type == UNDEFINED){
        i = reservetable.size;
        reservetable.funcnum = (FuncNum *)spdRealloc(reservetable.funcnum,
                                                     sizeof(FuncNum)*(i+1));
        reservetable.funcnum[i] = GET_BY_REL;
        reservetable.arglist
            = (MyArgList *)spdRealloc(reservetable.arglist,
                                    sizeof(MyArgList)*(i+1));
        reservetable.arglist[i].stringarg[0]
            = (SpdString)spdMalloc(strlen(arg2)+1);
        strcpy(reservetable.arglist[i].stringarg[0], arg2);
        reservetable.arglist[i].stringarg[1]
            = (SpdString)spdMalloc(strlen(arg3)+1);
        strcpy(reservetable.arglist[i].stringarg[1], arg3);
        reservetable.arglist[i].symtablearg = arg1;
        reservetable.arglist[i].resultsp = sp;
        reservetable.size++;
        return sp;
    }

    sp->type = OBJIDTABLE;
    /*    sp->meta_type = STRUCTURE;*/
    sp->value.objid_table.oiv = NULL;
    sp->value.objid_table.size = 0;
    /*    sp->value.id_struct.master = NULL;
          sp->value.id_struct.objid_table = NULL;*/

    for(i=0;i<arg1->value.objid_table.size;i++){
        /*        sp->value.id_struct.master
                  = spdRealloc(sp->value.id_struct.master,
                  sizeof(SpdObjId)*(i+1));
                  sp->value.id_struct.master[i] = arg1->value.objid_table.id[i];
                  sp->value.id_struct.objid_table
                  = spdRealloc(sp->value.id_struct.objid_table,
                  sizeof(SpdObjIdArray)*(i+1));
                  sp->value.id_struct.objid_table[i]
                  = spdGetRelObjArray(arg1->value.objid_table.id[i],
                  arg2, arg3, NULL);*/
        csr = spdGetRelObjInit(arg1->value.objid_table.oiv[i].id,
                               arg2, arg3);
        while((id = spdGetRelObj(csr)) != SPD_NON_ID){
            sp->value.objid_table.oiv
                = spdRealloc(sp->value.objid_table.oiv,
                             sizeof(ObjIdValue)*(sp->value.objid_table.size+1));
            sp->value.objid_table.oiv[sp->value.objid_table.size].id
                = id;
            if(arg1->value.objid_table.oiv[i].masternum == UNSUBORDINARY)
                sp->value.objid_table.oiv[sp->value.objid_table.size].masternum
                    = i;
            else
                sp->value.objid_table.oiv[sp->value.objid_table.size].masternum
                    = arg1->value.objid_table.oiv[i].masternum;

            /* getting offset */
            if((rel_id = spdGetARelId(arg1->value.objid_table.oiv[i].id,
                                      id, arg2)) == SPD_NON_ID){
                spdErrorExit("No relation %d < %s > %d",
                             arg1->value.objid_table.oiv[i].id,
                             arg2, id);
            }
            sp->value.objid_table.oiv[sp->value.objid_table.size].offset
                = spdGetOffset(rel_id);
            sp->value.objid_table.oiv[sp->value.objid_table.size].rel_id
                = rel_id;
        
            sp->value.objid_table.size++;
        }
    }
    /* sp->value.id_struct.size = i;*/
    return(sp);
}

void set_condattrtablestr(SpdString arg1, SpdString arg2)
{
    int i;

    i = condattrtable.size;
    
    condattrtable.attr_name = spdRealloc(condattrtable.attr_name,
                                         sizeof(SpdString)*(i+1));
    condattrtable.attr_name[i] = spdMalloc(strlen(arg1)+1);
    strcpy(condattrtable.attr_name[i], arg1);

    condattrtable.is_int = spdRealloc(condattrtable.is_int,
                                      sizeof(SpdBoolean)*(i+1));
    condattrtable.is_int[i] = FALSE;
    
    condattrtable.attrvalue = spdRealloc(condattrtable.attrvalue,
                                         sizeof(AttrValue)*(i+1));
    condattrtable.attrvalue[i].string = spdMalloc(strlen(arg2)+1);
    strcpy(condattrtable.attrvalue[i].string, arg2);
    condattrtable.size++;
}

void set_condattrtableint(SpdString arg1, int arg2)
{
    int i;

    i = condattrtable.size;
    
    condattrtable.attr_name = spdRealloc(condattrtable.attr_name,
                                         sizeof(SpdString)*(i+1));
    condattrtable.attr_name[i] = spdMalloc(strlen(arg1)+1);
    strcpy(condattrtable.attr_name[i], arg1);

    condattrtable.is_int = spdRealloc(condattrtable.is_int,
                                      sizeof(SpdBoolean)*(i+1));
    condattrtable.is_int[i] = TRUE;
    
    condattrtable.attrvalue = spdRealloc(condattrtable.attrvalue,
                                         sizeof(AttrValue)*(i+1));
    condattrtable.attrvalue[i].integer = arg2;
    condattrtable.size++;
}

void set_variable(SymTable *arg1, SymTable *arg2)
{
    arg1->type = arg2->type;
    arg1->meta_type = arg2->meta_type;
    arg1->value = arg2->value;
    if(arg2->msp != NULL) arg1->msp = arg2->msp;
}

void do_by_reservetable(void)
{
    int i;

    for(i=0;i<reservetable.size;i++){
        switch(reservetable.funcnum[i])
            {
              case GET_STRING_ATTRIBUTE:
                get_string_attribute(reservetable.arglist[i].symtablearg,
                                     reservetable.arglist[i].stringarg[0]);
                break;
        
              case GET_INT_ATTRIBUTE:
                get_int_attribute(reservetable.arglist[i].symtablearg,
                                  reservetable.arglist[i].stringarg[0]);
                break;

              case GET_BY_REL:
                get_by_rel(reservetable.arglist[i].symtablearg,
                           reservetable.arglist[i].stringarg[0],
                           reservetable.arglist[i].stringarg[1]);
                break;
            }
    }
}

SymTable *get_int_constant(int arg)
{
    SpdString buf;
    SymTable *sp;
    int i = 1, j = 10;

    while((arg/j) != 0){
        j *= 10;
        i++;
    }
    
    buf = spdMalloc(i+1);
    sprintf(buf, "%d", arg);

    sp = symlook(buf);
    sp->type = INTTABLE;
    sp->meta_type = ATOM;

    sp->value.int_table.iv
        = (IntValue *)spdRealloc(sp->value.int_table.iv,
                                 sizeof(IntValue));
    sp->value.int_table.iv[0].masternum = UNSUBORDINARY;
    sp->value.int_table.iv[0].value = arg;
    sp->value.int_table.size = 1;

    return sp;
}

SymTable *get_string_constant(SpdString arg)
{
    SpdString buf;
    SymTable *sp;

    buf = spdMalloc(strlen(arg)+3);
    sprintf(buf,"\"%s\"", arg);

    sp = symlook(buf);
    sp->type = STRINGTABLE;
    sp->meta_type = ATOM;

    sp->value.string_table.sv
        = (StringValue *)spdRealloc(sp->value.string_table.sv,
                                    sizeof(StringValue));
    sp->value.string_table.sv[0].masternum = UNSUBORDINARY;
    sp->value.string_table.sv[0].value
        = (SpdString)spdMalloc(strlen(arg)+1);
    strcpy(sp->value.string_table.sv[0].value, arg);
    sp->value.string_table.size = 1;

    return sp;
}

SymTable *symlook(SpdString arg1)
{
    SymTable *sp;

    for(sp = symtable; sp < &symtable[SYMTABLE_MAX]; sp++){
        if(sp->name && SPD_STREQ(sp->name, arg1)){
            first_appear = FALSE;
            return sp;
        }
        if(!sp->name){
            sp->name = spdStrdup(arg1);
            sp->type = UNDEFINED;
            /* sp->isvariable may be TRUE, but if so sp->isvariable become
               TRUE in lex.l */
            sp->isvariable = FALSE;
            first_appear = TRUE;
            return sp;
        }
    }
    yyerror("Too many variables\n");
    exit(1);
}
