/************************************************ 
 * BDD Manipulator - header file (Version 6.04) * 
 * Copyright (C) 1992 Kyoto University          *
 * Author : Shin-ichi MINATO    (1992/5/6)      *
 ************************************************/

/* 
 * Modified by S.Yamamoto 2000/03/18.
 * 
 * $Id: sbdd.c,v 30.2 2000/03/19 08:57:09 yamamoto Exp yamamoto $
 */

#ifndef lint
static char rcsid[] =
    "$Id: sbdd.c,v 30.2 2000/03/19 08:57:09 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include "sbdd.h"

#include <Sapid/Sapid.h>

#define SHORT	16
#define SMAX	((unsigned short)~0)

#define OPNIL	0
#define OPAND	1
#define OPXOR	3
#define OPR	2
#define OPVAL	4
#define OPCOF1	5
#define OPCOF2	6

#define BDDF	0
#define BDDT	NOT(BDDF)

#define LPOS	21
#define LMAX	(EXP(WORD-LPOS-1)-1)
#define LMASK	(LMAX<<LPOS)
#define NMAX	(EXP(LPOS-1)-1)
#define NMASK	(EXP(LPOS)-EXP(1))

#define EXP(x)		((unsigned) 1<<(x))
#define	NEG(x)		((x)< 0)
#define	NOT(x)		(~(x))
#define	ABS(x)		(NEG(x)? NOT(x): (x))

#define	NODE(x)		(NEG(x)? (((x)==BDDT)? NIL:(NOT(x)& NMASK)>>1)\
				:(((x)==BDDF)? NIL:((x)& NMASK)>>1))
#define ATOFF(f,k,f0,f1)\
	{ if(NEG(f))\
	    if(NOT(f) & 1)\
	    {  f0=NOT(node[k].ix1); f1=NOT(node[k].ix0); }\
	    else\
	    {  f0=NOT(node[k].ix0); f1=NOT(node[k].ix1); }\
	  else\
	    if((f) & 1)\
	    {  f0=node[k].ix1; f1=node[k].ix0; }\
	    else\
	    {  f0=node[k].ix0; f1=node[k].ix1; } }

#define LCONV(lev,f)\
	if(NEG(f))\
	{  if(f != BDDT)\
	     f = ((f)| LMASK)\
			& NOT((lev)-(NOT(f)& LMASK));}\
	else\
	{  if(f != BDDF) f = ((f) & ~LMASK)\
				|((lev)-((f) & LMASK));}

#define	INCREF(k) \
	((k==0)? 0: \
	  (++node[k].refc==0)? \
	      (--node[k].refc, bddgc(), \
	          (++node[k].refc==0)? \
		      (err("INCREF: refc overflow"),ERR):0 \
	      ) \
	      :0 \
	)

#define	DECREF(k) \
	((k==0)? 0: \
	  (node[k].refc--==0)? (err("DECREF: refc < 0"),ERR):0 \
	)

#define	NDTHASH(ix0,ix1) \
	((((ix0)^((ix0)>>10))+((ix1)^((ix1)>>15))*1117)&(ndtsize-1))

#define	OPTHASH(f,g) \
	((((f)+((f)>>16))+((g)+((g)>>18))*115)&(optsize-1))

struct nodetag
{
	int		ix0;
	int		ix1;
	int		next;
	unsigned short	refc;
};
struct opttag
{
	int		f;
	int		g;
	int		h;
};

static	int		nodemax;
static	struct nodetag	*node = NULL;
static	int		avx;
static	int		nodenum;

static	int		ndtsize;
static	int		*ndt = NULL;

static	int		optsize;
static	struct opttag	*opt = NULL;

static	int		errmode = 0;

static	void	gc1();
static	bddp	getnode();
static	int	andfalse();
static	int	nodect();
static	void	resetct();
static	void	err();
static	int	prob();

static	int	vfalse = 0;
static	int	vtrue = 1000000000;
static	int	(*rule)() = prob;

int	bddinit(max)
int	max;
{
	int	i;

	if(node != NULL)
	{
		spdFree(opt);
		spdFree(ndt);
		spdFree(node);
	}
	if((nodemax=max)<= 0)
	{
		err("bddinit: nodemax <= 0");
		return(ERR);
	}
	if(nodemax > NMAX+1)
	{
		err("bddinit: nodemax > 1048576");
		return(ERR);
	}
	for(ndtsize=1; ndtsize<nodemax; ndtsize<<=1);
	optsize = ndtsize>>2;
	node = (struct nodetag *)spdMalloc(sizeof(struct nodetag)*nodemax);
	if(node == NULL)
	{
		err("bddinit: not enough memory (node)");
		return(ERR);
	}
	ndt = (int *)spdMalloc(sizeof(int)*ndtsize);
	if(ndt == NULL)
	{
		err("bddinit: not enough memory (ndt)");
		spdFree(node);
		return(ERR);
	}
	opt = (struct opttag *)spdMalloc(sizeof(struct opttag)*optsize);
	if(opt == NULL)
	{
		err("bddinit: not enough memory (opt)");
		spdFree(ndt);
		spdFree(node);
		return(ERR);
	}
	avx = 0;
	for(i=1;i<nodemax;i++)
	{
		node[i-1].next = i;
		node[i-1].ix0 = BDDNIL;
		node[i-1].refc = 0;
	}
	node[nodemax-1].next = NIL;
	node[nodemax-1].ix0 = BDDNIL;
	node[nodemax-1].refc = 0;
	nodenum = 0;
	for(i=0;i<ndtsize;i++) ndt[i] = NIL;
	for(i=0;i<optsize;i++) opt[i].f = BDDNIL;

	bddprime(1);
	return(0);
}

bddp	bddcopy(f)
bddp	f;
{
	int	k;

	if((k=NODE(f))== NIL) return(f);
	if(k >= nodemax || node[k].ix0 == BDDNIL)
	{
		err("bddcopy: invalid argument");
		return(BDDNIL);
	}
	if(INCREF(k)== ERR) return(BDDNIL);
	return(f);
}

int	bddfree(f)
bddp	f;
{
	int	k;

	if((k=NODE(f))== NIL) return(0);
	if(k >= nodemax || node[k].ix0 == BDDNIL)
	{
		err("bddfree: invalid argument");
		return(ERR);
	}
	return(DECREF(k));
}

int	bddlevel(f)
bddp	f;
{
	int	k;

	if((k=NODE(f))== NIL) return(0);
	if(k >= nodemax || node[k].ix0 == BDDNIL)
	{
		err("bddlevel: invalid argument");
		return(ERR);
	}
	return(ABS(f)>>LPOS);
}

bddp	bddnot(f)
bddp	f;
{
	if(bddcopy(f)== BDDNIL) return(BDDNIL);
	return(NOT(f));
}

bddp	bddfalse()
{
	return(BDDF);
}

bddp	bddtrue()
{
	return(BDDT);
}

bddp	bddprime(level)
unsigned short	level;
{
	if(level <= 0)
	{
		err("bddprime: level <= 0");
		return(BDDNIL);
	}
	if(level >= LMAX)
	{
		err("bddprime: level overflow");
		return(BDDNIL);
	}
	return(getnode(level<<LPOS,BDDF,BDDT));
}

bddp	bddshift(f,level)
bddp	f;
int	level;
{
	if(f == BDDT || f == BDDF) return(f);
	if(level < 0)
	{
		err("bddshift: level < 0");
		return(BDDNIL);
	}
	if((ABS(f)>>LPOS)+level >= LMAX)
	{
		err("bddshift: level overflow");
		return(BDDNIL);
	}
	if(bddcopy(f)== BDDNIL) return(BDDNIL);
	if(NEG(f)) f = NOT(NOT(f)+(level<<LPOS));
	else	f += level<<LPOS;
	return(f);
}

bddp	bddshift2(f,level)
bddp	f;
int	level;
{
	if(f == BDDT || f == BDDF) return(f);
	if((ABS(f)>>LPOS)+level >= LMAX)
	{
		err("bddshift2: level overflow");
		return(BDDNIL);
	}
	if(bddcopy(f)== BDDNIL) return(BDDNIL);
	if(NEG(f)) f = NOT(NOT(f)+(level<<LPOS));
	else	f += level<<LPOS;
	return(f);
}

bddp	bddrstr(f,level,value)
bddp	f;
int	level;
char	value;
{
	int	k;
	int	flev;

	if(level <= 0)
	{
		err("bddrstr: level <= 0");
		return(BDDNIL);
	}
	if(level >= LMAX)
	{
		err("bddrstr: level overflow");
		return(BDDNIL);
	}
	if((k=NODE(f))== NIL) return(f);
	if(k >= nodemax || node[k].ix0 == BDDNIL)
	{
		err("bddrstr: invalid argument");
		return(BDDNIL);
	}
	flev = ABS(f)& LMASK;
	if((level<<LPOS) > flev)
		if(INCREF(k)== ERR) return(BDDNIL);
		else	return(f);
	else
	{
		bddp	f0,f1,h,h0,h1;
		int	hash,optf,optg,opth,optlev;

		ATOFF(f,k,f0,f1);
		LCONV(flev,f0);
		LCONV(flev,f1);
		if((level<<LPOS) == flev)
		{
			h = (value==0)? f0:f1;
			if((k=NODE(h))!= NIL) INCREF(k);
			return(h);
		}
		if(node[k].refc <= 1) hash = -1;
		else
		{
			optf = ABS(f);
			optf= (optf&~LMASK)|(OPR<<LPOS);
			optg= level<<LPOS;
			if(value==0) optg = NOT(optg);
			optlev = flev+EXP(LPOS);
			LCONV(optlev,optg);
			hash = OPTHASH(optf, optg);
			if( optf == opt[hash].f &&
	    	    	optg == opt[hash].g )
			{
				h = opt[hash].h;
				LCONV(optlev,h);
				if((k=NODE(h))!= NIL) INCREF(k);
				return(NEG(f)? NOT(h):h);
			}
		}
		if((h0=bddrstr(f0,level,value))== BDDNIL)
			return(BDDNIL);
		if((h1=bddrstr(f1,level,value))== BDDNIL)
		{
			bddfree(h0);
			return(BDDNIL);
		}
		if((h=getnode(flev,h0,h1))== BDDNIL)
		{
			bddfree(h0);
			bddfree(h1);
			return(BDDNIL);
		}
		if(hash != -1)
		{
			opth = NEG(f)? NOT(h):h;
			LCONV(optlev,opth);
			opt[hash].f = optf;
			opt[hash].g = optg;
			opt[hash].h = opth;
		}
		return(h);
	}
}

bddp	bddrstr0(f,level)
bddp	f;
short	level;
{
	return(bddrstr(f,level,0));
}

bddp	bddrstr1(f,level)
bddp	f;
short	level;
{
	return(bddrstr(f,level,1));
}

bddp	bddand(f,g)
bddp	f,g;
{
	int	fx,gx,hx;
	bddp	f0,f1,g0,g1;
	bddp	h,h0,h1;
	int	flev,glev;
	int	hash,optf,optg,opth,optlev;

	if((fx=NODE(f))== NIL)
	{
		if(f == BDDF) return(BDDF);
		if((gx=NODE(g))!= NIL && INCREF(gx)== ERR)
			return(BDDNIL);
		return(g);
	}
	if((gx=NODE(g))== NIL)
	{
		if(g == BDDF) return(BDDF);
		if(INCREF(NODE(f))== ERR) return(BDDNIL);
		return(f);
	}
	if(f == g)
		if(INCREF(NODE(f))== ERR)
			return(BDDNIL);
		else	return(f);
	else if(f == NOT(g)) return(BDDF);
	if(fx >= nodemax||node[fx].ix0 == BDDNIL)
	{
		err("bddand: invalid argument");
		return(BDDNIL);
	}
	if(gx >= nodemax||node[gx].ix0 == BDDNIL)
	{
		err("bddand: invalid argument");
		return(BDDNIL);
	}
	if((flev=ABS(f)& LMASK) < (glev=ABS(g)& LMASK)
	    || (flev == glev && f < g))
	{
		h=f; f=g; g=h;
		hx=flev; flev=glev; glev=hx;
		hx=fx; fx=gx; gx=hx;
	}
	if(node[fx].refc==1 && node[gx].refc==1) hash = -1;
	else
	{
		if(NEG(f)) optf = (f|LMASK)& NOT(OPAND<<LPOS);
		else	   optf = (f &~LMASK)|(OPAND<<LPOS);
		optg = g;
		optlev = flev+EXP(LPOS);
		LCONV(optlev,optg);
		hash = OPTHASH(optf, optg);
		if( optf == opt[hash].f &&
	    	    optg == opt[hash].g )
		{
			h = opt[hash].h;
			LCONV(optlev,h);
			if((hx=NODE(h))!= NIL) INCREF(hx);
			return(h);
		}
	}
	ATOFF(f,fx,f0,f1);
	LCONV(flev,f0);
	LCONV(flev,f1);
	if(flev != glev)
	{
		if((h0=bddand(f0,g))== BDDNIL) return(BDDNIL);
		if((h1=bddand(f1,g))== BDDNIL)
		{
			bddfree(h0);
			return(BDDNIL);
		}
	}
	else
	{
		ATOFF(g,gx,g0,g1);
		LCONV(glev,g0);
		LCONV(glev,g1);
		if((h0=bddand(f0,g0))== BDDNIL) return(BDDNIL);
		if((h1=bddand(f1,g1))== BDDNIL)
		{
			bddfree(h0);
			return(BDDNIL);
		}
	}
	if((h=getnode(flev,h0,h1))== BDDNIL)
	{
		bddfree(h0);
		bddfree(h1);
		return(BDDNIL);
	}
	if(hash != -1)
	{
		opth = h;
		LCONV(optlev,opth);
		opt[hash].f = optf;
		opt[hash].g = optg;
		opt[hash].h = opth;
	}
	return(h);
}

bddp	bddxor(f,g)
bddp	f,g;
{
	int	fx,gx,hx;
	bddp	f0,f1,g0,g1;
	bddp	h,h0,h1;
	int	flev,glev;
	int	hash,optf,optg,opth,optlev;

	if((fx=NODE(f))== NIL)
	{
		if((gx=NODE(g))!= NIL && INCREF(gx)== ERR)
			return(BDDNIL);
		if(f == BDDF) return(g);
		return(NOT(g));
	}
	if((gx=NODE(g))== NIL)
	{
		if(INCREF(NODE(f))== ERR) return(BDDNIL);
		if(g == BDDF) return(f);
		return(NOT(f));
	}
	if(f == g) return(BDDF);
	else if(f == NOT(g)) return(BDDT);
	if(fx >= nodemax||node[fx].ix0 == BDDNIL)
	{
		err("bddxor: invalid argument");
		return(BDDNIL);
	}
	if((gx=NODE(g))>= nodemax||node[gx].ix0 == BDDNIL)
	{
		err("bddxor: invalid argument");
		return(BDDNIL);
	}
	if((flev=ABS(f)& LMASK) < (glev=ABS(g)& LMASK)
	    || (flev == glev && f < g))
	{
		h=f; f=g; g=h;
		hx=flev; flev=glev; glev=hx;
		hx=fx; fx=gx; gx=hx;
	}
	if(node[fx].refc==1 && node[gx].refc==1) hash = -1;
	else
	{
		optf = ABS(f);
		optf = (optf &~LMASK)|(OPXOR<<LPOS);
		optg = ABS(g);
		optlev = flev+EXP(LPOS);
		LCONV(optlev,optg);
		hash = OPTHASH(optf, optg);
		if( optf == opt[hash].f &&
	    	    optg == opt[hash].g )
		{
			h = opt[hash].h;
			LCONV(optlev,h);
			if((hx=NODE(h))!= NIL) INCREF(hx);
			return(NEG(f^g)? NOT(h):h);
		}
	}
	ATOFF(f,fx,f0,f1);
	LCONV(flev,f0);
	LCONV(flev,f1);
	if(flev != glev)
	{
		if((h0=bddxor(f0,g))== BDDNIL) return(BDDNIL);
		if((h1=bddxor(f1,g))== BDDNIL)
		{
			bddfree(h0);
			return(BDDNIL);
		}
	}
	else
	{
		ATOFF(g,gx,g0,g1);
		LCONV(glev,g0);
		LCONV(glev,g1);
		if((h0=bddxor(f0,g0))== BDDNIL) return(BDDNIL);
		if((h1=bddxor(f1,g1))== BDDNIL)
		{
			bddfree(h0);
			return(BDDNIL);
		}
	}
	if((h=getnode(flev,h0,h1))== BDDNIL)
	{
		bddfree(h0);
		bddfree(h1);
		return(BDDNIL);
	}
	if(hash != -1)
	{
		opth = NEG(f^g)? NOT(h):h;;
		LCONV(optlev,opth);
		opt[hash].f = optf;
		opt[hash].g = optg;
		opt[hash].h = opth;
	}
	return(h);
}

bddp	bddor(f,g)
bddp	f,g;
{
	bddp	h;

	if((h=bddand(NOT(f),NOT(g))) == BDDNIL) return(BDDNIL);
	return(NOT(h));
}

bddp	bddnand(f,g)
bddp	f,g;
{
	bddp	h;

	if((h=bddand(f,g)) == BDDNIL) return(BDDNIL);
	return(NOT(h));
}

bddp	bddnor(f,g)
bddp	f,g;
{
	return(bddand(NOT(f),NOT(g)));
}

bddp	bddxnor(f,g)
bddp	f,g;
{
	int	h;

	if((h=bddxor(f,g)) == BDDNIL) return(BDDNIL);
	return(NOT(h));
}

int	bddimply(f,g)
bddp	f,g;
{
	return(andfalse(f,NOT(g)));
}

void	defval(vf,vt,r)
int	vf,vt;
int	(*r)();
{
	int	i;

	vfalse = vf;
	vtrue = vt;
	rule = r;
	for(i=0;i<optsize;i++) opt[i].f = BDDNIL;
}

int	bddval(f)
bddp	f;
{
	int	fx,flev,v;
	bddp	f0,f1;
	int	hash;

	if((fx=NODE(f))== NIL)
	{
		if(f == BDDF) return(vfalse);
		return(vtrue);
	}
	if(fx >= nodemax||node[fx].ix0 == BDDNIL)
	{
		err("bddval: invalid argument");
		return(ERR);
	}
	if(node[fx].refc == 1) hash = -1;
	else
	{
		hash = OPTHASH(OPVAL<<LPOS, f);
		if((OPVAL<<LPOS)== opt[hash].f && f == opt[hash].g )
			return(opt[hash].h);
	}
	flev = ABS(f)& LMASK;
	ATOFF(f,fx,f0,f1);
	LCONV(flev,f0);
	LCONV(flev,f1);
	v = (*rule)(flev>>LPOS,bddval(f0),bddval(f1));
	if(hash != -1)
	{
		opt[hash].f = OPVAL<<LPOS;
		opt[hash].g = f;
		opt[hash].h = v;
	}
	return(v);
}

int	bddused()
{
	return(nodenum);
}

int	bddsize(f)
bddp	f;
{
	int	num;

	num = nodect(f);
	resetct(f);
	return(num);
}

void	bddgc()
{
	int	i;

	for(i=0;i<nodemax;i++)
		if(node[i].ix0 != BDDNIL && node[i].refc == 0)
		{
			node[i].refc = 1;
			gc1(i);
		}
	for(i=0;i<optsize;i++) opt[i].f = BDDNIL;
}

void	emode(e)
char	e;
{
	errmode = e;
}

static	int	andfalse(f,g)
bddp	f,g;
{
	int	fx,gx;
	bddp	f0,f1,g0,g1;
	int	flev,glev;
	int	hash,optf,optg,optlev;

	if(f == NOT(g)) return(TRUE);
	if(f == BDDF || g == BDDF) return(TRUE);
	if(f == g) return(FALSE);
	if(f == BDDT || g == BDDT) return(FALSE);
	if((fx=NODE(f)) >= nodemax||node[fx].ix0 == BDDNIL)
	{
		err("andfalse: invalid argument");
		return(ERR);
	}
	if((gx=NODE(g)) >= nodemax||node[gx].ix0 == BDDNIL)
	{
		err("andfalse: invalid argument");
		return(ERR);
	}
	if((flev=ABS(f)& LMASK) < (glev=ABS(g)& LMASK)
	    || (flev == glev && f < g))
	{
		f0=f; f=g; g=f0;
		optf=flev; flev=glev; glev=optf;
		optf=fx; fx=gx; gx=optf;
	}
	if(node[fx].refc==1 && node[gx].refc==1) hash = -1;
	else
	{
		if(NEG(f)) optf = (f | LMASK)& NOT(OPAND<<LPOS);
		else	   optf = (f &~LMASK)| (OPAND<<LPOS);
		optg = g;
		optlev = flev+EXP(LPOS);
		LCONV(optlev,optg);
		hash = OPTHASH(optf, optg);
		if( optf == opt[hash].f &&
	    	    optg == opt[hash].g )
			return(opt[hash].h == BDDF);
	}
	ATOFF(f,fx,f0,f1);
	LCONV(flev,f0);
	LCONV(flev,f1);
	if(flev != glev)
	{
		if(andfalse(f0,g)== FALSE) return(FALSE);
		if(andfalse(f1,g)== FALSE) return(FALSE);
	}
	else
	{
		ATOFF(g,gx,g0,g1);
		LCONV(glev,g0);
		LCONV(glev,g1);
		if(andfalse(f0,g0)== FALSE) return(FALSE);
		if(andfalse(f1,g1)== FALSE) return(FALSE);
	}
	if(hash != -1)
	{
		opt[hash].f = optf;
		opt[hash].g = optg;
		opt[hash].h = BDDF;
	}
	return(TRUE);
}

static	void	gc1(k)
int	k;
{
	int	k0;

	DECREF(k);
	if(node[k].refc > 0) return;
	{
		int	hash,nx,nx2;

		hash = NDTHASH(node[k].ix0, node[k].ix1);
		if((nx=ndt[hash])== k) ndt[hash] = node[k].next;
		else
		{
			while(nx != k) nx = node[nx2=nx].next;
			node[nx2].next = node[k].next;
		}
		node[k].next = avx;
		avx = k;
		nodenum--;
		if((k0=NODE(node[k].ix0))!= NIL) gc1(k0);
		if((k0=NODE(node[k].ix1))!= NIL) gc1(k0);
		node[k].ix0 = BDDNIL;
		return;
	}
}

static	bddp	getnode(lev,f0,f1)
int	lev;
bddp	f0,f1;
{
	int	k,attr,nx,hash;
	bddp	f;

	if(f0 == f1)
	{
		if((k=NODE(f0))!= NIL) DECREF(k);
		return(f0);
	}
	if(ABS(f0)>ABS(f1))
	{
		if(NEG(f1))
		{
			f=f0; f0 = NOT(f1); f1 = NOT(f);
			attr=NOT(lev|1);
		}
		else
		{
			f=f0; f0=f1; f1=f; attr= lev|1;
		}
	}
	else
	{
		if(NEG(f0))
		{	
			f0=NOT(f0); f1=NOT(f1);
			attr = NOT(lev);
		}
		else	attr = lev;
	}
	LCONV(lev,f0);
	LCONV(lev,f1);
	hash = NDTHASH(f0,f1);
	nx=ndt[hash];
	while(nx != NIL)
	{
		if(f0 == node[nx].ix0 &&
		   f1 == node[nx].ix1 )
		{
			if((k=NODE(f0))!= NIL) DECREF(k);
			if((k=NODE(f1))!= NIL) DECREF(k);
			if(INCREF(nx)== ERR) return(BDDNIL);
			return(attr^(nx<<1));
		}
		nx = node[nx].next;
	}
	if((nx = avx) == NIL)
	{
		bddgc();
		if((nx = avx) == NIL)
		{
			err("getnode: node overflow");
			return(BDDNIL);
		}
	}
	avx = node[nx].next;
	node[nx].next = ndt[hash];
	ndt[hash] = nx;
	node[nx].ix0 = f0;
	node[nx].ix1 = f1;
	node[nx].refc = 1;
	nodenum++;
	return(attr^(nx<<1));
}

static	void	err(msg)
char	*msg;
{
	if(errmode == 2) return;
	fprintf(stderr,"*** SBDD Error %s ***\n",msg);
	if(errmode == 1) return;
	fprintf(stderr," nodenum: %d\t",nodenum);
	fprintf(stderr," nodemax: %d\n",nodemax);
	fprintf(stderr," ndtsize: %d\toptsize: %d\n",ndtsize,optsize);
	exit(1);
}

static	int	nodect(f)
bddp	f;
{
	int	k;

	if((k=NODE(f))== NIL) return(0);
	if(NEG(node[k].next)) return(0);
	node[k].next = NOT(node[k].next);
	return(1 + nodect(node[k].ix0) + nodect(node[k].ix1));
}

static	void	resetct(f)
bddp	f;
{
	int	k;

	if((k=NODE(f))==NIL) return;
	if( !NEG(node[k].next)) return;
	node[k].next = NOT(node[k].next);
	resetct(node[k].ix0);
	resetct(node[k].ix1);
}

static	int	prob(lev,v0,v1)
short	lev;
int	v0,v1;
{
	return((v0>>1)+(v1>>1));
}

bddp	bddcofact(f,g)
bddp	f,g;
{
	int	fx,gx,hx;
	bddp	f0,f1,g0,g1;
	bddp	h,h0,h1;
	int	flev,glev;
	int	hash,optf,optg,opth,optlev;

	if((gx=NODE(g))== NIL)
	{
		if(g == BDDF) return(BDDF);
		if(INCREF(NODE(f))== ERR) return(BDDNIL);
		return(f);
	}
	if((fx=NODE(f))== NIL) return(f);
	if(f == g) return(BDDT);
	if(f == NOT(g)) return(BDDF);
	if(fx >= nodemax||node[fx].ix0 == BDDNIL)
	{
		err("bddcofact: invalid argument");
		return(BDDNIL);
	}
	if(gx >= nodemax||node[gx].ix0 == BDDNIL)
	{
		err("bddcofact: invalid argument");
		return(BDDNIL);
	}
	flev=ABS(f)& LMASK;
	glev=ABS(g)& LMASK;
	if(node[fx].refc==1 && node[gx].refc==1) hash = -1;
	else
	{
		if(flev >= glev)
		{
			if(NEG(f)) optf = (f|LMASK)& NOT(OPCOF1<<LPOS);
			else	   optf = (f &~LMASK)|(OPCOF1<<LPOS);
			optg = g;
			optlev = flev+EXP(LPOS);
		}
		else
		{
			if(NEG(g)) optf = (g|LMASK)& NOT(OPCOF2<<LPOS);
			else	   optf = (g &~LMASK)|(OPCOF2<<LPOS);
			optg = f;
			optlev = glev+EXP(LPOS);
		}
		LCONV(optlev,optg);
		hash = OPTHASH(optf, optg);
		if( optf == opt[hash].f &&
	    	    optg == opt[hash].g )
		{
			h = opt[hash].h;
			LCONV(optlev,h);
			if((hx=NODE(h))!= NIL) INCREF(hx);
			return(h);
		}
	}
	if(flev > glev)
	{
		ATOFF(f,fx,f0,f1);
		LCONV(flev,f0);
		LCONV(flev,f1);
		if((h0=bddcofact(f0,g))== BDDNIL) return(BDDNIL);
		if((h1=bddcofact(f1,g))== BDDNIL)
		{
			bddfree(h0);
			return(BDDNIL);
		}
		if((h=getnode(flev,h0,h1))== BDDNIL)
		{
			bddfree(h0);
			bddfree(h1);
			return(BDDNIL);
		}
	}
	else if(flev < glev)
	{
		ATOFF(g,gx,g0,g1);
		LCONV(glev,g0);
		LCONV(glev,g1);
		if(g0 == BDDF)
		{
			if((h=bddcofact(f,g1))== BDDNIL)
				return(BDDNIL);
		}
		else if(g1 == BDDF)
		{
			if((h=bddcofact(f,g0))== BDDNIL)
				return(BDDNIL);
		}
		else
		{
			if((h0=bddcofact(f,g0))== BDDNIL)
				return(BDDNIL);
			if((h1=bddcofact(f,g1))== BDDNIL)
			{
				bddfree(h0);
				return(BDDNIL);
			}
			if((h=getnode(glev,h0,h1))== BDDNIL)
			{
				bddfree(h0);
				bddfree(h1);
				return(BDDNIL);
			}
		}
	}
	else
	{
		ATOFF(f,fx,f0,f1);
		LCONV(flev,f0);
		LCONV(flev,f1);
		ATOFF(g,gx,g0,g1);
		LCONV(glev,g0);
		LCONV(glev,g1);
		if(g0 == BDDF)
		{
			if((h=bddcofact(f1,g1))== BDDNIL)
				return(BDDNIL);
		}
		else if(g1 == BDDF)
		{
			if((h=bddcofact(f0,g0))== BDDNIL)
				return(BDDNIL);
		}
		else
		{
			if((h0=bddcofact(f0,g0))== BDDNIL) return(BDDNIL);
			if((h1=bddcofact(f1,g1))== BDDNIL)
			{
				bddfree(h0);
				return(BDDNIL);
			}
			if((h=getnode(flev,h0,h1))== BDDNIL)
			{
				bddfree(h0);
				bddfree(h1);
				return(BDDNIL);
			}
		}
	}
	if(hash != -1)
	{
		opth = h;
		LCONV(optlev,opth);
		opt[hash].f = optf;
		opt[hash].g = optg;
		opt[hash].h = opth;
	}
	return(h);
}

