%{

/* 
 * Program:             $RCSfile: yacc.y,v $  $Revision: 30.2 $
 * 
 * Purpose:             "Imakefile" of Sapid/tool/SIP/sip.
 *
 * Author:              N.Kurauchi  1997/02/08
 *                      S.Yamamoto  1996/02/08
 *
 * (C) Copyright:       N.Kurauchi and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: yacc.y,v 30.2 2003/06/24 08:12:56 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "SaQuLa.h"

extern SaqulaResult *result;
extern int val_num;

void yyerror(char *msg);
int yylex(void);

%}

%union{
    char        *name;
    int         intval;
    SymTable    *symtable;
    BddTable    *bddtable;
    Ope ope;
}

%token DEFINE
%token AS
%token SELECT
%token FROM
%token IN
%token WHERE
%token SORT
%token BY
%token DOT
%token EQ
%token NOTEQ
%token NOT
%token AND
%token OR
%token XOR
%token NAND
%token NOR
%token XNOR

%token <name> OBJECT_EXTENT
%token <name> RELATION_EXTENT
%token <name> OBJECT
%token <name> INTATTRIBUTE
%token <name> STRINGATTRIBUTE
%token <name> RELATION
%token <intval> INTCONSTANT
%token <name> STRINGCONSTANT
%token <symtable> VARIABLE
%token <name> ROLENAME
%token <name> SORTBY
%token <name> OPERATOR_NAME

%token <symtable> nil
%token <bddtable> true
%token <bddtable> false

%type <symtable> query
%type <bddtable> boolean

%nonassoc LOWER_WHERE
%nonassoc WHERE
%right IN
%left ','
%nonassoc LOWER_BOOLEAN
%right EQ NOTEQ
%nonassoc LOWER_RELATION
%left RELATION
%left DOT '('
%left OR NOR
%left XOR XNOR
%left AND NAND
%nonassoc NOT

%%

query_program
    : define_query_list query {
        setValue($2);
    }

    | query {
        setValue($1);
    }
    ;

define_query_list
    : define_query_list define_query

    | define_query
    ;

define_query
    : DEFINE VARIABLE AS query ';' {
        set_variable($2, $4);
    }
    ;

/* basic query */
query
    : nil

    | VARIABLE

    | '(' query ')' {
        $$ = $2;
    }

    | INTCONSTANT {
        $$ = get_int_constant($1);
    }

    | STRINGCONSTANT {
        $$ = get_string_constant($1);
    }

    /* | boolean %prec LOWER_BOOLEAN */
    ;

boolean
    : true

    | false

    | '(' boolean ')' {
        $$ = $2;
    }
    ;

/* comparison query */
boolean
    : query EQ query {
        $$ = bddlook(EQUAL, (void *)$1, (void *)$3);
    }

    | query NOTEQ query {
        $$ = bddlook(NOTEQUAL, (void *)$1, (void *)$3);
    }
    ;

/* boolean query */
boolean
    : NOT boolean {
        $$ = bddlook(NOTOP, (void *)$2, NULL);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddnot($2->bddp);
    }

    | boolean AND boolean {
        $$ = bddlook(ANDOP, (void *)$1, (void *)$3);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddand($1->bddp, $3->bddp);
    }

    | boolean OR boolean {
        $$ = bddlook(OROP, (void *)$1, (void *)$3);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddor($1->bddp, $3->bddp);
    }

    | boolean XOR boolean {
        $$ = bddlook(XOROP, (void *)$1, (void *)$3);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddor($1->bddp, $3->bddp);
    }

    | boolean NAND boolean {
        $$ = bddlook(NANDOP, (void *)$1, (void *)$3);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddor($1->bddp, $3->bddp);
    }

    | boolean NOR boolean {
        $$ = bddlook(NOROP, (void *)$1, (void *)$3);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddor($1->bddp, $3->bddp); }
    | boolean XNOR boolean {
        $$ = bddlook(XNOROP, (void *)$1, (void *)$3);
        $$->val_num = UNCOMPARISON;
        $$->bddp = bddor($1->bddp, $3->bddp);
    }
    ;

/* access query */
query
    :  query DOT INTATTRIBUTE {
        $$ = get_int_attribute($1, $3);
    }

    | query DOT STRINGATTRIBUTE {
        $$ = get_string_attribute($1, $3);
    }

    | query DOT RELATION %prec LOWER_RELATION {
        $$ = get_by_rel($1, $3, "any");
    } 

    | query DOT RELATION '[' ROLENAME ']' {
        $$ = get_by_rel($1, $3, $5);
    }

    /* | query DOT OPERATOR_NAME '(' query_list ')' */
    ;

/*
 * query_list
 *     : query
 *     | query_list ',' query
 *     ;
 */

/* collection query */
query
    : SELECT query FROM variable_list %prec LOWER_WHERE {
        do_by_reservetable();
        $$ = $2;
    }

    | SELECT query FROM variable_list WHERE boolean {
        checkBddTable(getSufResult($6));
        $$ = getSufQuery(); }

    | SORT '(' query ')' BY SORTBY {
    }
    ;

variable_list
    : variable_list ',' variable_in

    | variable_in
    ;

variable_in
    : VARIABLE IN query {
        set_variable($1, $3);
    }
    ;

/* others */
query
    : OBJECT_EXTENT {
        $$ = get_extent($1);
    }

    | query DOT OPERATOR_NAME '(' OBJECT ')' {
        if (SPD_STREQ($3, "GetOcc")) {
            $$ = getOcc($1, $5);
        } else {
            spdErrorExit("\"%s\": No method.\n", $3);
        }
    }

    | query DOT OPERATOR_NAME '(' ')' {
        if (SPD_STREQ($3, "ConvItoC")) {
            $$ = convItoC($1);
        } else {
            spdErrorExit("\"%s\": No method.\n", $3);
        }
    }

    /* | RELATION_EXTENT { $$ = get_relation_extent($1); }*/

    | query '(' cond_list ')' {
        $$ = get_by_condattr($1);
    }
    ;

cond_list
    : cond_list ',' equation

    | equation
    ;

equation
    : INTATTRIBUTE EQ INTCONSTANT {
        set_condattrtableint($1, $3);
    }

    | STRINGATTRIBUTE EQ STRINGCONSTANT {
        set_condattrtablestr($1, $3);
    }
    ;

%%
