/* 
 * Program:             $RCSfile: AR4region.c,v $  $Revision: 30.14 $
 * 
 * Purpose:             SpdRegion˴ؤؿ
 *
 * Author:              T.Hidaka  1997/06/27
 * Modified:            S.Yamamoto  1997/09/05
 *
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1997 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: AR4region.c,v 30.14 2002/10/21 07:28:07 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <tcl.h>
#include "tclAR.h"
#include "tclARSub.h"
#include "tclfilemap.h"

static SpdRegion list_to_region(Tcl_Interp *interp, char *str)
{
    SpdRegion   reg = SPD_NON_REGION;
    int         listargc;
    char        **listargv;
    SpdObjId    fileId;
    int         offset, length;
    
    if (Tcl_SplitList(interp, str, &listargc, &listargv) == TCL_OK) {
        if (listargc == 3) {
            if (tclAR_GetInt(interp, listargv[0], &fileId) != TCL_OK) {
                return reg;
            }
            if (tclAR_GetInt(interp, listargv[1], &offset) != TCL_OK) {
                return reg;
            }
            if (tclAR_GetInt(interp, listargv[2], &length) != TCL_OK) {
                return reg;
            }
        }
    }
    spdFree(listargv);
    
    reg.fileId = fileId;
    reg.offset = offset;
    reg.length = length;

    return reg;
}

static SpdOcc list_to_occ(Tcl_Interp *interp, char *str)
{
    SpdOcc      occ = SPD_NON_OCC;
    int         listargc;
    char        **listargv;
    SpdObjId    objectId, fileId, relationId;
    int         offset;
    
    if (Tcl_SplitList(interp, str, &listargc, &listargv) == TCL_OK) {
        if (listargc == 4) {
            if (tclAR_GetInt(interp, listargv[0], &objectId) != TCL_OK) {
                return occ;
            }
            if (tclAR_GetInt(interp, listargv[1], &fileId) != TCL_OK) {
                return occ;
            }
            if (tclAR_GetInt(interp, listargv[2], &offset) != TCL_OK) {
                return occ;
            }
            if (tclAR_GetInt(interp, listargv[3], &relationId) != TCL_OK) {
                return occ;
            }
        }
    }
    spdFree(listargv);
    
    occ.objectId = objectId;
    occ.fileId = fileId;
    occ.offset = offset;
    occ.relationId = relationId;

    return occ;
}

static char *region_to_list(Tcl_Interp *interp, SpdRegion reg)
{
    static char     buf[100];

    sprintf(buf, "%d %d %d", reg.fileId, reg.offset, reg.length);
    return buf;
}

static char *occ_to_list(Tcl_Interp *interp, SpdOcc occ)
{
    static char buf[100];

    sprintf(buf, "%d %d %d %d", occ.objectId, occ.fileId,
            occ.offset, occ.relationId);
    return buf;
}

int tcl_spdConvRegionItoC(ClientData clientData, Tcl_Interp *interp,
                          int argc, char *argv[])
{
    SpdRegion   ireg, creg;

    /*  : i-reg */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    ireg = list_to_region(interp, argv[1]);
    creg = spdConvRegionItoC(ireg);
    Tcl_SetResultString(region_to_list(interp, creg));
    
    return TCL_OK;
}

int tcl_spdConvRegionCtoI(ClientData clientData, Tcl_Interp *interp,
                          int argc, char *argv[])
{
    SpdRegion   ireg, creg;

    /*  : c-reg */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    creg = list_to_region(interp, argv[1]);
    ireg = spdConvRegionCtoI(creg);
    Tcl_SetResultString(region_to_list(interp, ireg));
    
    return TCL_OK;
}

/* 
 * spdGetRegionText(): SpdRegion ʬե뤫롥
 */

int tcl_spdGetRegionText(ClientData clientData, Tcl_Interp *interp,
                         int argc, char *argv[])
{
    FILE        *fp;
    char        *buf;
    SpdRegion   reg;
    
    /*  : reg */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    reg = list_to_region(interp, argv[1]);
    if (reg.fileId == SPD_NON_ID) {
        spdErrorExit("Region not found.\n");
    }
    buf = spdMalloc(reg.length + 1);

    /* Modified by H.Ohhashi 1999/04/05. */
    /* reg.fileId = spdSearchGlobalObj("file", spdGetName(reg.fileId)); */
    fp = smuOpenSrcFile(reg.fileId);
    SPD_FSEEK(fp, reg.offset, SEEK_SET);
    spdFreadChar(buf, reg.length, fp);
    spdFclose(fp);

    Tcl_SetResult(interp, buf, (Tcl_FreeProc *)free);
    
    return TCL_OK;
}

/* 
 * spdConvOccToRegion(): SpdOccSpdRegionѴ롥
 */

int tcl_spdConvOccToRegion(ClientData clientData, Tcl_Interp *interp,
                           int argc, char *argv[])
{
    SpdOcc      occ;
    SpdRegion   reg;
    
    /*  : occ */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    occ = list_to_occ(interp, argv[1]);
    reg = spdConvOccToRegion(occ);
    Tcl_SetResultString(region_to_list(interp, reg));
    
    return TCL_OK;
}

int tcl_spdConvRegionToOcc(ClientData clientData, Tcl_Interp *interp,
                           int argc, char *argv[])
{
    SpdOcc      occ;
    SpdRegion   reg;
    
    /*  : reg */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    reg = list_to_region(interp, argv[1]);
    occ = spdConvRegionToOcc(reg);
    Tcl_SetResultString(occ_to_list(interp, occ));
    
    return TCL_OK;
}

int tcl_spdConvRegionTclToSpd(ClientData clientData, Tcl_Interp *interp,
                              int argc, char *argv[])
{
    struct offsetmap    *map;
    SpdRegion           reg;
    struct region       from, to;

    /*  : region */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    reg = list_to_region(interp, argv[1]);
    
    map = makeFile2TkMap(interp, reg.fileId);

    from.offset = reg.offset;
    from.length = reg.length;

    to = convRegionRev(map, from);

    reg.offset = to.offset;
    reg.length = to.length;

    Tcl_SetResultString(region_to_list(interp, reg));

    return TCL_OK;
}

int tcl_spdConvRegionSpdToTcl(ClientData clientData, Tcl_Interp *interp,
                              int argc, char *argv[])
{
    struct offsetmap    *map;
    SpdRegion           reg;
    struct region       from, to;

    /*  : region */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    reg = list_to_region(interp, argv[1]);
    
    map = makeFile2TkMap(interp, reg.fileId);
    from.offset = reg.offset;
    from.length = reg.length;

    to = convRegion(map, from);

    reg.offset = to.offset;
    reg.length = to.length;

    Tcl_SetResultString(region_to_list(interp, reg));

    return TCL_OK;
}
