/* 
 * Program:             $RCSfile: SMU.c,v $  $Revision: 30.13 $
 * 
 * Purpose:
 *
 * Author:              T.Hidaka  1997/06/27
 * Modified:            S.Yamamoto  1997/09/05
 *
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1997 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: SMU.c,v 30.13 2003/06/28 15:33:26 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <tcl.h>
#include "tclAR.h"
#include "tclARSub.h"

/* ݡȤ٤ؿ ********************************************************

SpdObjIdArray spdGetObjIdSortedArray(SpdString className,
                                     SpdString name,
                                     SpdBoolean (*cond)(const SpdObjId),
                                     int (*compar)(const SpdObjId *,
                                                   const SpdObjId *));
SpdObjIdArray spdGetObjIdOfModelSortedArray(SpdString className,
                                            SpdString name,
                                            SpdString model,
                                            SpdBoolean (*cond)(const SpdObjId),
                                            int (*compar)(const SpdObjId *,
                                                          const SpdObjId *));
SpdRelArray spdGetRelSortedArray(SpdObjId objectId,
                                 SpdString relationName,
                                 SpdString attrName,
                                 SpdBoolean (*cond)(const SpdRel),
                                 int (*compar)(const SpdRel *,
                                               const SpdRel *));
SpdObjIdArray spdGetRelObjSortedArray(SpdObjId objectId,
                                      SpdString relationName,
                                      SpdString attrName,
                                      SpdBoolean (*cond)(const SpdObjId),
                                      int (*compar)(const SpdObjId *,
                                                    const SpdObjId *));
SpdRelIdArray spdGetRelIdSortedArray(SpdObjId objectId1,
                                     SpdObjId objectId2,
                                     SpdString relationName,
                                     SpdBoolean (*cond)(const SpdRelId),
                                     int (*compar)(const SpdRelId *,
                                                   const SpdRelId *));
SpdOccArray spdGetOffsetOfObjSortedArray(SpdObjId objectId,
                                         SpdBoolean (*cond)(const SpdOcc),
                                         int (*compar)(const SpdOcc *,
                                                       const SpdOcc *));
SpdOccArray spdGetIncludedOccSortedArray(SpdObjId objectId,
                                         SpdString objectType,
                                         SpdBoolean (*cond)(const SpdOcc),
                                         int (*compar)(const SpdOcc *,
                                                       const SpdOcc *));

int spdCompareObjIdName(const SpdObjId *a, const SpdObjId *b);
int spdCompareRelOffset(const SpdRel *a, const SpdRel *b);
int spdCompareRelIdOffset(const SpdRelId *a, const SpdRelId *b);
int spdCompareOccOffset(const SpdOcc *a, const SpdOcc *b);
int spdCompareOccName(const SpdOcc *a, const SpdOcc *b);

SpdBoolean spdObjIdIsDeclNormal(const SpdObjId id);
SpdBoolean spdObjIdIsDeclFuncdef(const SpdObjId id);
SpdBoolean spdObjIdIsDeclEnumerator(const SpdObjId id);
SpdBoolean spdObjIdIsDeclParameter(const SpdObjId id);
SpdBoolean spdObjIdIsDeclMember(const SpdObjId id);
SpdBoolean spdObjIdIsDeclType(const SpdObjId id);

SpdBoolean spdObjIdIsIdVariable(const SpdObjId id);
SpdBoolean spdObjIdIsIdFunction(const SpdObjId id);
SpdBoolean spdObjIdIsIdMember(const SpdObjId id);
SpdBoolean spdObjIdIsIdEnumerator(const SpdObjId id);
SpdBoolean spdObjIdIsIdTypename(const SpdObjId id);
SpdBoolean spdObjIdIsIdTag(const SpdObjId id);

SpdBoolean  smuIsArgument(const SpdObjId ident_id);
SpdBoolean  smuIsGlobalVariable(const SpdObjId ident_id);
SpdBoolean  smuIsLocalVariable(const SpdObjId ident_id);
SpdBoolean  smuIsStaticId(const SpdObjId ident_id);

SpdString smuGetRelParentAttrName(SpdRelId relId);
SpdString smuGetRelChildAttrName(SpdRelId relId);

SpdObjId spdGetExprType(SpdObjId exprid);

SpdBoolean smuObjIdIsTypePointer(SpdObjId id);
SpdBoolean smuObjIdIsTypeIntegral(SpdObjId id);
SpdBoolean smuObjIdIsTypeDouble(SpdObjId id);
SpdBoolean smuObjIdIsTypeLong(SpdObjId id);
SpdBoolean smuObjIdIsTypeUnsigned(SpdObjId id);

SpdObjId smuFindType(int qualifier, int sign, int size, int type);
SpdObjId smuFindPointerOf(SpdObjId type);

int smuGetOperatorTerms(int sort);
int smuGetOperatorCategory(int sort);
int smuInitOperatorCategory(void);

*******************************************************************************/

int tcl_smuGetAProgramName(ClientData clientData, Tcl_Interp *interp,
                           int argc, char *argv[])
{
    /* ʤ */
    if (argc != 1) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    Tcl_SetResultString(smuGetAProgramName());

    return TCL_OK;
}

int tcl_smuGetObjAttr(ClientData clientData, Tcl_Interp *interp,
                      int argc, char *argv[])
{
    SpdString attr, readable;
    int res;
    SpdId id;
    int intRes;
    
    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    if (tclAR_GetInt(interp, argv[1], &id) != TCL_OK) {
        return TCL_ERROR;
    }

    res = smuGetObjAttr(id, argv[2], &attr, &readable);

    if (res) {
        Tcl_AppendElement(interp, attr);
        Tcl_AppendElement(interp, readable);
    } else {
        intRes = spdGetAttrValInt(id, argv[2]);
        if (intRes != SPD_NON_ID) {
            Tcl_SetResultInt(intRes);
        } else {
            Tcl_SetResultString(spdGetAttrValString(id, argv[2]));
        }
    }
    
    return TCL_OK;
}

int tcl_smuGetClassAttr(ClientData clientData, Tcl_Interp *interp,
                        int argc, char *argv[])
{
    SpdString attr, readable;
    int res;
    SpdClassId id;
    int value;
    
    if (argc != 4) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    if (tclAR_GetInt(interp, argv[1], &id) != TCL_OK) {
        return TCL_ERROR;
    }
    if (tclAR_GetInt(interp, argv[3], &value) != TCL_OK) {
        return TCL_ERROR;
    }

    res = smuGetClassAttr(id, argv[2], value, &attr, &readable);

    Tcl_AppendElement(interp, attr);
    Tcl_AppendElement(interp, readable);

    if (res) {
        return TCL_OK;
    } else {
        return TCL_ERROR;
    }
}

static Tcl_DString convSmuXrefTable(SmuXrefTable table)
{
    Tcl_DString str;
    char buf[40];
    int i;

    Tcl_DStringInit(&str);
    for (i = 0; i < table.size; i++) {
        sprintf(buf, "%d %d %d", table.elem[i].callerId,
                table.elem[i].calleeId, table.elem[i].relationId);
        Tcl_DStringAppendElement(&str, buf);
    }
    
    return str;
}

int tcl_smuMakeXrefTables(ClientData clientData, Tcl_Interp *interp,
                          int argc, char *argv[])
{
    SmuXrefTables xrefs;
    Tcl_DString res, str;
    SpdObjId prog_id;
    int flag;

    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    if (tclAR_GetInt(interp, argv[1], &prog_id) != TCL_OK) {
        return TCL_ERROR;
    }

    if (SPD_STREQ(argv[2], "smuXrefEachObj")) {
        flag = smuXrefEachObj;
    } else if (SPD_STREQ(argv[2], "smuXrefEachOcc")) {
        flag = smuXrefEachOcc;
    } else {
        interp->result = "wrong parameter.";
        return TCL_ERROR;
    }

    xrefs = smuMakeXrefTables(prog_id, flag);

    Tcl_DStringInit(&res);

    str = convSmuXrefTable(xrefs.funcCallTable);
    Tcl_DStringAppendElement(&res, Tcl_DStringValue(&str));
    Tcl_DStringFree(&str);

    str = convSmuXrefTable(xrefs.variableTable);
    Tcl_DStringAppendElement(&res, Tcl_DStringValue(&str));
    Tcl_DStringFree(&str);

    str = convSmuXrefTable(xrefs.globalVariableTable);
    Tcl_DStringAppendElement(&res, Tcl_DStringValue(&str));
    Tcl_DStringFree(&str);

    str = convSmuXrefTable(xrefs.localVariableTable);
    Tcl_DStringAppendElement(&res, Tcl_DStringValue(&str));
    Tcl_DStringFree(&str);

    str = convSmuXrefTable(xrefs.macroCallTable);
    Tcl_DStringAppendElement(&res, Tcl_DStringValue(&str));
    Tcl_DStringFree(&str);

    Tcl_DStringResult(interp, &res);

    smuFreeXrefTables(xrefs);

    return TCL_OK;
}

/* Moved and renamed from SMU by H.Ohhashi 1999/11/24. */
  
int tcl_spdConvIdPfileToFile(ClientData clientData, Tcl_Interp *interp,
                             int argc, char *argv[])
{
    SpdObjId pfile_id;

    /* 1ġ */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    if (tclAR_GetInt(interp, argv[1], &pfile_id) != TCL_OK) {
        return TCL_ERROR;
    }

    Tcl_SetResultInt(spdConvIdPfileToFile(pfile_id));

    return TCL_OK;
}

/*
 * ƥե뤫ݤȽꤹ롥
 */

int tcl_smuIsSystemFile(ClientData clientData, Tcl_Interp *interp,
             int argc, char *argv[])
{
    SpdObjId pfileId;
    char buf[40];
    /*  : pfileId */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    if (tclAR_GetInt(interp, argv[1], &pfileId) != TCL_OK) {
        return TCL_ERROR;
    }

    /* ƥեIDʤ0,Ǥʤ0򤤤 */
    sprintf(buf, "%d", smuIsSystemFile(pfileId));
    Tcl_AppendElement(interp, buf);
    
    return TCL_OK;
}

/*
 * ƥ줿֥ȤݤȽꤹ롥
 */

int tcl_smuIsSystemObject(ClientData clientData, Tcl_Interp *interp,
             int argc, char *argv[])
{
    SpdObjId objectId;
    char buf[40];
    /*  : objectId */
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    if (tclAR_GetInt(interp, argv[1], &objectId) != TCL_OK) {
        return TCL_ERROR;
    }
    
    /* ƥobjectIdʤ0, Ǥʤ0򤤤 */
    sprintf(buf, "%d", smuIsSystemObject(objectId));
    Tcl_AppendElement(interp, buf);
    
    return TCL_OK;
}
