/* 
 * Program:             $RCSfile: tclfilemap.c,v $  $Revision: 30.15 $
 * 
 * Purpose:
 *
 * Author:              T.Hidaka  1997/06/27
 * Modified:            S.Yamamoto  1997/09/05
 *
 * (C) Copyright:       T.Hidaka and S.Yamamoto  1997 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: tclfilemap.c,v 30.15 2003/06/28 15:33:26 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <sys/types.h>      /* For stat(). */
#include <sys/stat.h>       /* For stat(). */
#include <Sapid/Sapid.h>

#ifndef KANJI
#define KANJI       /* For tcl.h */
#endif

#include <tcl.h>
#include "tclfilemap.h"

#define IS_SJIS(c)  (((c) >= 0x81 && (c) <= 0x9f) || \
                     ((c) >= 0xe0 && (c) <= 0xfc))

/* #define DEBUG */

#if defined(lint)
#define DW                  printf      /* Just for cheking. */
#elif defined(DEBUG)
#define DW                  spdMessage
#else
#define DW                  (void)
#endif

struct offsetmap *createMap(void)
{
    struct offsetmap    *res;

    res = spdMalloc(sizeof(struct offsetmap));
    if (res) {
        res->mapsize = res->mapallocsize = 0;
        res->mapnode = NULL;
    }
    return res;
}

int deleteMap(struct offsetmap *map)
{
    if (map) {
        map->mapsize = map->mapallocsize = 0;
        if (map->mapnode)
            spdFree(map->mapnode);
        map->mapnode = NULL;
        spdFree(map);
    }
    return 0;
}

int clearMap(struct offsetmap *map)
{
    if (map) {
        map->mapsize = 0;
    }
    return 0;
}

/* offsetmap˥Ρɤɲä롥 */

int addMap(struct offsetmap *map, struct region from, struct region to)
{
    struct offsetmap_node   n;

    if (to.offset < 0) {
        /* to.offsetϼưŪ˵뤳ȤǤ롥 */
    }
    DW("%d, %d -> %d, %d.\n", from.offset, from.length, to.offset, to.length);
    
    n.from = from;
    n.to = to;

    if (map) {
        if (map->mapallocsize <= map->mapsize) {
            map->mapallocsize += MAP_ALLOC_STEP_SIZE;
            if (map->mapnode) {
                map->mapnode = spdRealloc(map->mapnode,
                                          map->mapallocsize *
                                          sizeof(struct offsetmap_node));
            } else {
                map->mapnode = spdMalloc(map->mapallocsize *
                                         sizeof(struct offsetmap_node));
            }
        }
        if (map->mapnode == NULL) {         /* allocation error */
            return -1;
        }
        map->mapnode[(map->mapsize)++] = n;
    }
    return 0;
}

/* offsetmapǡѴоݤȤʤäƤΰꤹȡ
   offsetѴоݤƬoffset + length֤ */

struct region convRegion(struct offsetmap *map, struct region from)
{
    int             i, diff;
    struct region   to;

    if (map == NULL) {                      /* Ѵʤ */
        return from;
    }

    /* եåȤ롥 */
    diff = 0;
    for (i = 0; i < map->mapsize; i++) {
        if (map->mapnode[i].from.offset > from.offset) {
            break;
        }

        diff += (map->mapnode[i].from.length - map->mapnode[i].to.length);
    }

    to.offset = from.offset - diff;

    /* length뤿offset + lengthΰ֤ΥեåȤ롥 */
    diff = 0;
    for (i = 0; i < map->mapsize; i++) {
        if (map->mapnode[i].from.offset > from.offset + from.length) {
            break;
        }

        diff += (map->mapnode[i].from.length - map->mapnode[i].to.length);
    }

    to.length = (from.offset + from.length - diff) - to.offset;

    return to;
}

struct region convRegionRev(struct offsetmap *map, struct region to)
{
    int             i, diff;
    struct region   from;

    if (map == NULL) {                      /* Ѵʤ */
        return to;
    }

    /* եåȤ롥 */
    diff = 0;
    for (i = 0; i < map->mapsize; i++) {
        if (map->mapnode[i].to.offset > to.offset) {
            break;
        }

        diff += (map->mapnode[i].from.length - map->mapnode[i].to.length);
    }

    from.offset = to.offset + diff;

    /* length뤿offset + lengthΰ֤ΥեåȤ롥 */
    diff = 0;
    for (i = 0; i < map->mapsize; i++) {
        if (map->mapnode[i].to.offset > to.offset + to.length) {
            break;
        }

        diff += (map->mapnode[i].from.length - map->mapnode[i].to.length);
    }

    from.length = (to.offset + to.length + diff) - from.offset;

    return from;
}

/******************************************************************************/

static void makeMap_JIS(struct offsetmap *map, const unsigned char *str)
{
    int             c, c1;
    struct region   from, to;
    enum {T_ASCII, T_KANJI, T_KANA }    kanji = T_ASCII;

    from.offset = from.length = 0;
    to.offset = to.length = 0;
    
    while ((c = *str++) != '\0') {
        if (c == '\033') {
            if (!strncmp((const char *)str, "$B", 2) ||
                !strncmp((const char *)str, "$@", 2)) {
                kanji = T_KANJI;
                str += 2;
                from.length += 3;
            } else if (!strncmp((const char *)str, "(J", 2) ||
                       !strncmp((const char *)str, "(B", 2)) {
                kanji = T_ASCII;
                str += 2;
                from.length += 3;
                if (from.length || to.length) {
                    /* mapnode Ͽ */
                    addMap(map, from, to);

                    from.offset += from.length;
                    to.offset += to.length;
                    from.length = to.length = 0;
                }
            } else if (!strncmp((const char *)str, "(I", 2)) {
                kanji = T_KANA;
                str += 2;
                from.length += 3;
            } else {
                from.offset++;
                to.offset++;
            }
        } else if (kanji == T_KANJI) {
            c1 = *str++;
            if (c1 == '\0') break;
            from.length += 2;
            to.length++;
        } else if (kanji == T_KANA) {
            from.length++;
            to.length++;
        } else {
            to.offset++;
            from.offset++;
        }
    }
}

static void makeMap_EUC(struct offsetmap *map, const unsigned char *str)
{
    int             c;
    struct region   from, to;

    from.offset = from.length = 0;
    to.offset = to.length = 0;
    
    while ((c = *str++) != '\0') {
        if (c == 0x8e) {                    /* SS2. */
            str++;
            from.length += 2;
            to.length ++;
        } else if (c == 0x8f) {             /* SS3. */
            str+=2;
            from.length += 3;
            to.length ++;
        } else if (c & 0x80) {
            str++;
            from.length += 2;
            to.length ++;
        } else {
            if (from.length || to.length) {
                /* mapnodeϿ */
                addMap(map, from, to);
        
                from.offset += from.length;
                to.offset += to.length;
                from.length = to.length = 0;
            }
            from.offset++;
            to.offset++;
        }
    }

    if (from.length || to.length) {
        /* mapnodeϿ */
        addMap(map, from, to);
    
        from.offset += from.length;
        to.offset += to.length;
        from.length = to.length = 0;
    }
}

static void makeMap_SJIS(struct offsetmap *map, const unsigned char *str)
{
    int             c;
    struct region   from, to;
    
    from.offset = from.length = 0;
    to.offset = to.length = 0;

    while ((c = *str++) != '\0') {
        if (IS_SJIS(c)) {
            str++;
            from.length+=2;
            to.length++;
        } else {
            if (from.length || to.length) {
                /* mapnode Ͽ */
                addMap(map, from, to);
        
                from.offset += from.length;
                to.offset += to.length;
                from.length = to.length = 0;
            }

            from.offset++;
            to.offset++;
        }
    }

    if (from.length || to.length) {
        /* mapnode Ͽ */
        addMap(map, from, to);
    
        from.offset += from.length;
        to.offset += to.length;
        from.length = to.length = 0;
    }
}

static unsigned char    fname[FILENAME_MAX] = ""; /* bufferơ */
static struct offsetmap *file2tkmap = NULL;

struct offsetmap *makeFile2TkMap(Tcl_Interp *interp, SpdObjId fileId)
{
    unsigned char   *fn, *str, *p;

    fn = spdGetName(fileId);
    if (SPD_STREQ((const char *)fn, (const char *)fname) && file2tkmap) {
        return file2tkmap;                  /* mapϺѤǤ롥 */
    }

    /* mapΰݤ롥 */
    if (file2tkmap == NULL) {
        file2tkmap = createMap();
    }
    if (file2tkmap == NULL) {
        return NULL;
    }

    /* map롥 */
    clearMap(file2tkmap);

    /* mapκ
       Cf. Tcl_Encode in "tcl7.6jp/generic/tclKanjiUtil.c". */

    /* եɤ߹ࡥ */
    /* docfileϼѰդ롥!! SMU줿褤? */
    if (SPD_STREQ(spdGetClassName(fileId), "docfile")) {
        char        filePath[FILENAME_MAX + 1];
        struct stat st;
        FILE        *fp;
         
        sprintf(filePath, "%s/%s/%d", spdGetSDBdir(), SAPID_SRC_DIR, fileId);
        if (stat(filePath, &st) != 0) {
            spdErrorExit("spdGetFileSize(): %d: Can't stat().\n", fileId);
        }
        str = spdMalloc(st.st_size + 1);
        fp = spdFopen(filePath, SPD_FOPEN_RT_MODE);
 
        if (spdFreadChar(str, st.st_size, fp) != st.st_size) {
            spdErrorExit("spdReadFileInBufSub(): %d: Can't read file.\n", fileId);
        }
        spdFclose(fp);
    } else {
        spdReadFileInBufByName((char *)fn, (char **)&str);
    }
    
    for (p = str; *p; p++) {
        /* JISɤȽ('\033'Ƚ)Τ߼ǹԤ */
        /* !! Tclkanji codeޥɤǤϤޤȽǤʤ礬롥 */
        if (*p == '\033') {
            break;
        }

        /* '{', '}'ȥꥹȤȤߤʤƤޤΤǡ'*'˽񤭴롥 */
        /* ʸȽ꡿mapˤϱƶϤʤ */
        if (*p == '{' || *p == '}') {
            *p = '*';
        }
    }
    if (!*p) {
        /* Check kanji code. */
        if (Tcl_VarEval(interp, "kanji code {", str, "}", NULL) != TCL_OK) {
            DW("makeFile2TkMap(): Not kanji.\n");
            return NULL;
        }
    }
    DW("makeFile2TkMap(): %s: Kanji.\n", interp->result);

    if (*p || SPD_STREQ(interp->result, "JIS")) {
        makeMap_JIS(file2tkmap, str);
    } else if (SPD_STREQ(interp->result, "EUC")) {
        makeMap_EUC(file2tkmap, str);
    } else if (SPD_STREQ(interp->result, "SJIS")) {
        makeMap_SJIS(file2tkmap, str);
    }
    
    strcpy((char *)fname, (const char *)fn);
    spdFree(str);
    return file2tkmap;
}

int deleteFile2TkMap(void)
{
    deleteMap(file2tkmap);
    return 0;
}
