/* 
 * tclAppInit.c --
 *
 *  Provides a default version of the main program and Tcl_AppInit
 *  procedure for Tcl applications (without Tk).
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclAppInit.c 1.17 96/03/26 12:45:29
 */

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <tcl.h>
#include <tk.h>

/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */

/* CPP directive "#ifndef WIN32" is added by S.Yamamoto 1997/10/17. */

#ifndef WIN32
extern int matherr(void);
int *tclDummyMathPtr = (int *)matherr;
#endif

#ifdef TCL_TEST
EXTERN int      Tcltest_Init _ANSI_ARGS_((Tcl_Interp *interp));
#endif /* TCL_TEST */

/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *  This is the main program for the application.
 *
 * Results:
 *  None: Tcl_Main never returns here, so this procedure never
 *  returns either.
 *
 * Side effects:
 *  Whatever the application does.
 *
 *----------------------------------------------------------------------
 */

int
main(argc, argv)
    int argc;           /* Number of command-line arguments. */
    char **argv;        /* Values of command-line arguments. */
{
    Tk_Main(argc, argv, Tcl_AppInit);
    return (EXIT_SUCCESS);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_AppInit --
 *
 *  This procedure performs application-specific initialization.
 *  Most applications, especially those that incorporate additional
 *  packages, will have their own version of this procedure.
 *
 * Results:
 *  Returns a standard Tcl completion code, and leaves an error
 *  message in interp->result if an error occurs.
 *
 * Side effects:
 *  Depends on the startup script.
 *
 *----------------------------------------------------------------------
 */

#include "tclAR.h"
#include "tclARSub.h"

int
Tcl_AppInit(interp)
    Tcl_Interp *interp;     /* Interpreter for application. */
{
    if (Tcl_Init(interp) == TCL_ERROR) {
        return (TCL_ERROR);
    }
    if (Tk_Init(interp) == TCL_ERROR) {
        return (TCL_ERROR);
    }

#if (7 <= TCL_MAJOR_VERSION) && (5 <= TCL_MINOR_VERSION)
    Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *)NULL);
#endif

#ifdef TK_TEST
    if (Tktest_Init(interp) == TCL_ERROR) {
        return (TCL_ERROR);
    }
    Tcl_StaticPackage(interp, "Tktest", Tktest_Init,
                      (Tcl_PackageInitProc *)NULL);
#endif /* TK_TEST */

    Sapidtclar_Init(interp);

    /*
     * Call the init procedures for included packages.  Each call should
     * look like this:
     *
     * if (Mod_Init(interp) == TCL_ERROR) {
     *     return (TCL_ERROR);
     * }
     *
     * where "Mod" is the name of the module.
     */

    /*
     * Call Tcl_CreateCommand for application-specific commands, if
     * they weren't already created by the init procedures called above.
     */

    /*
     * Specify a user-specific startup file to invoke if the application
     * is run interactively.  Typically the startup file is "~/.apprc"
     * where "app" is the name of the application.  If this line is deleted
     * then no user-specific startup file will be run under any conditions.
     */

    Tcl_SetVar(interp, "tcl_rcFileName", "~/.tclshrc", TCL_GLOBAL_ONLY);
    return (TCL_OK);
}
