/* 
 * Program:             $RCSfile: testSpdSet.c,v $  $Revision: 40.19 $
 * 
 * Purpose:
 * 
 * Author:              S.Yamamoto  2003/01/25
 * 
 * (C) Copyright:       S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: testSpdSet.c,v 40.19 2003/06/28 15:32:44 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

/* #define DW_MEM_PROF */
#define TEST8_SPDGETARRAYOFELEM

#define MAX_TEST_ELEM           16
/* #define MAX_TEST_ELEM           1024 */
/* #define MAX_TEST_ELEM           (2 * 1024) */
/* #define MAX_TEST_ELEM           (128 * 1024) */

typedef struct {
    int     x;
    int     y;
} Point;

static int compare_int(int *a, int *b);

int main(int argc, char *argv[])
{
    int             i, j, *ip;
    SpdSet          set, set2, set3;
    SpdConstString  program_name;           /* This program name. */

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    spdInitMemProf();

    /* 
     * Test 1 (Create and Add).
     */

    spdMessage("Test 1.\n");
    set = spdCreateSet(sizeof(int));

    for (i = 0; i < MAX_TEST_ELEM; i++) {
        j = MAX_TEST_ELEM - i;
        spdAddElem(set, i, &j);
    }
    j = MAX_TEST_ELEM - i;
    spdAddElem(set, i, &j);

    /* 
     * Test 2 (Search).
     */

    spdMessage("Test 2.\n");
    for (i = 0; i < MAX_TEST_ELEM; i++) {
        ip = spdSearchElem(set, i);
        fprintf(stderr, "spdSearchElem(set, %d): %d.\n", i, *ip);
        assert(spdSetContains(set, i));
    }
    ip = spdSearchElem(set, i);
    fprintf(stderr, "spdSearchElem(set, %d): %d.\n", i, *ip);

    /* 
     * Test 3 (Del and Free).
     */

    spdMessage("Test 3.\n");
    set2 = spdDupSet(set);          /* Duplicate for test 4. */
    for (i = 0; i < MAX_TEST_ELEM; i++) {
        spdDelElem(set, i);
    }
    spdDelElem(set, i);
    assert(set->elemNum == 0);
    spdFreeSet(set);

    /* 
     * Test 4 (Duplicate).
     */

    spdMessage("Test 4.\n");
    for (i = 0; i < MAX_TEST_ELEM; i++) {
        ip = spdSearchElem(set2, i);
        fprintf(stderr, "spdSearchElem(set2, %d): %d.\n", i, *ip);
    }
    ip = spdSearchElem(set2, i);
    fprintf(stderr, "spdSearchElem(set2, %d): %d.\n", i, *ip);

    /* 
     * Test 5 (spdGetArrayOfElem).
     */

    {
        int     *ip, *jp;

        spdMessage("Test 5.\n");
        ip = jp = (int *)spdGetArrayOfElem(set2);
        SPD_QSORT(jp, set2->elemNum, sizeof(int), compare_int);
        for (i = 0; i < set2->elemNum; i++) {
            fprintf(stderr, "spdGetArrayOfElem(): %d: %d.\n", i, *ip++);
        }
        spdFree(jp);
    }

    /* 
     * Test 6 (spdGetAElem).
     */

    {
        int     *ip, j;

        spdMessage("Test 6.\n");
        for (i = 0; i < set2->elemNum; i++) {
            ip = (int *)spdGetAElem(set2, &j);
            fprintf(stderr, "spdGetAElem(): key %d, val %d.\n", j, *ip);
            assert(*ip + j == MAX_TEST_ELEM);
        }
        /* !! ǸǤϲ٤ǤǤ롥ƤϤʤ */
        /* fprintf(stderr, "spdGetAElem(): key %d, val %d.\n", j, *ip); */
        /* assert(spdGetAElem(set2, &j) == NULL); */
    }

    /* 
     * Test 7 (spdUnionSet).
     */
    
    i = -1;
    j = MAX_TEST_ELEM - i;
    fprintf(stderr, "Test 7: set2->elemNum %d.\n", set2->elemNum);
    spdAddElem(set2, i, &j);
    fprintf(stderr, "Test 7: set2->elemNum %d.\n", set2->elemNum);
    spdDelElem(set2, 11);
    spdDelElem(set2, 12);
    spdDelElem(set2, 13);
    fprintf(stderr, "Test 7: set2->elemNum %d.\n", set2->elemNum);

    set3 = spdCreateSet(sizeof(int));
    for (i = 0; i < MAX_TEST_ELEM * 2; i++) {
      j = MAX_TEST_ELEM - i;
      spdAddElem(set3, i, &j);
    }

    {
      SpdSet      set4;

      spdMessage("Test 7.\n");
      set4 = spdUnionSet(set2, set3);
      for (i = 0; i < set4->elemNum; i++) {
        ip = (int *)spdGetAElem(set4, &j);
        fprintf(stderr, "spdGetAElem(): key %d, val %d.\n", j, *ip);
        assert(*ip + j == MAX_TEST_ELEM);
      }
      spdFreeSet(set4);
    }

    /* 
     * Test 8 (spdIntersectionSet).
     */
    
    {
      SpdSet    set4;
      int   *ip, *jp;

      spdMessage("Test 8.\n");

#ifdef CHECK_INPUT
      fprintf(stderr, "Test 8: set2: elemNum %d.\n", set2->elemNum);
      spdClearSetEnumeration(set2);
      for (i = 0; i < set2->elemNum; i++) {
        ip = (int *)spdGetAElem(set2, &j);
        fprintf(stderr, "Test 8: set2: key %d, val %d.\n", j, *ip);
      }
      fprintf(stderr, "Test 8: set3: elemNum %d.\n", set3->elemNum);
      spdClearSetEnumeration(set3);
      for (i = 0; i < set3->elemNum; i++) {
        ip = (int *)spdGetAElem(set3, &j);
        fprintf(stderr, "Test 8: set3: key %d, val %d.\n", j, *ip);
      }
#endif

      set4 = spdIntersectionSet(set2, set3);
      fprintf(stderr, "Test 8: set4->elemNum %d.\n", set4->elemNum);

#ifdef TEST8_SPDGETARRAYOFELEM

      ip = jp = (int *)spdGetArrayOfElem(set4);
      SPD_QSORT(jp, set4->elemNum, sizeof(int), compare_int);
      for (i = 0; i < set4->elemNum; i++) {
        fprintf(stderr, "spdGetArrayOfElem(): %d: %d.\n", i, *ip++);
      }
      spdFree(jp);

#else /* !TEST8_SPDGETARRAYOFELEM */

      spdClearSetEnumeration(set4);
      for (i = 0; i < set4->elemNum; i++) {
        ip = (int *)spdGetAElem(set4, &j);
        fprintf(stderr, "spdGetAElem(): key %d, val %d.\n", j, *ip);
      }

#endif /* !TEST8_SPDGETARRAYOFELEM */

      spdFreeSet(set4);
    }

    /* 
     * Test 9 (spdDifferenceSet).
     */
    
    {
      SpdSet    set4;
      int   *ip, *jp;

      spdMessage("Test 9.\n");
      set4 = spdDifferenceSet(set3, set2);
      fprintf(stderr, "Test 8: set4->elemNum %d.\n", set4->elemNum);

#ifdef TEST8_SPDGETARRAYOFELEM

      ip = jp = (int *)spdGetArrayOfElem(set4);
      SPD_QSORT(jp, set4->elemNum, sizeof(int), compare_int);
      for (i = 0; i < set4->elemNum; i++) {
        fprintf(stderr, "spdGetArrayOfElem(): %d: %d.\n", i, *ip++);
      }
      spdFree(jp);

#else /* !TEST8_SPDGETARRAYOFELEM */

      spdClearSetEnumeration(set4);
      for (i = 0; i < set4->elemNum; i++) {
        ip = (int *)spdGetAElem(set4, &j);
        fprintf(stderr, "spdGetAElem(): key %d, val %d.\n", j, *ip);
      }

#endif /* !TEST8_SPDGETARRAYOFELEM */

      spdFreeSet(set4);
    }

    spdFreeSet(set2);
    spdFreeSet(set3);
    return (EXIT_SUCCESS);
}

static int compare_int(int *a, int *b)
{
    return (*a - *b);
}
