/* 
 * Program:             $RCSfile: comment.c,v $  $Revision: 51.6 $
 * 
 * Purpose:             
 *
 * Author:              N.Atsumi    2003/07/31
 *                      S.Yamamoto  2003/07/31
 *
 * (C) Copyright:       N.Atsumi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: comment.c,v 51.6 2003/11/04 03:39:43 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spie.h"
#include "style.h"

#define MAX_COMMENT_LEN     40

void putComment(FILE *fp, SpdString comment, int column)
{
    char    style_start[TEMP_MAX], style_end[TEMP_MAX];
    int     column_num = 0;

    if (comment == NULL) return;
    if (flag_c) {
        sprintf(style_start, "<span class=\"%s\">",
                color_style[COMMENT_STYLE].style);
        sprintf(style_end, "</span>\n");
    } else {
        sprintf(style_start, "<font color=\"%s\">",
                color_style[COMMENT_STYLE].color);
        sprintf(style_end, "</font>\n");
    }
    fputs(style_start, fp);
    while (column_num != column && *comment != '\0') {
        fputc(*comment, fp);
        comment++;
        column_num++;
    }
    fputs(style_end, fp);
}

SpdObjIdArray getCommentIds(SpdObjId ident)
{
    SpdObjId        declr;
    SpdRelArray     declr_buf, lcomment_buf, pcomment_buf;
    SpdObjIdArray   comment_buf = SPD_EMPTY_ARRAY;
    SpdString       comment_str;
    int             i, j, k, strage;

    declr_buf = spdGetRelSortedArray(ident, "decl_ident", "c-id", NULL,
                                     spdCompareRelOffset);
    for (i = 0; i < declr_buf.size; i++) {
        declr = spdGetARelObj(declr_buf.rel[i].objectId, "decl_decl", "c-id");
        strage = spdGetAttrValIntOfObj(declr, "storage");
        if (strage == STORAGE_EXTERN) continue;
        lcomment_buf = spdGetRelSortedArray(declr, "lcomment_declaration",
                                            "c-id", NULL, spdCompareRelOffset);
        for (j = 0; j < lcomment_buf.size; j++) {
            if (lcomment_buf.rel[j].objectId == SPD_NON_ID) continue;
            pcomment_buf = spdGetRelSortedArray(lcomment_buf.rel[j].objectId,
                                                "pcomment_lcomment",
                                                "c-id", NULL,
                                                spdCompareRelOffset);
            for (k = 0; k < pcomment_buf.size; k++) {
                comment_buf.id = spdRealloc(comment_buf.id, sizeof(SpdObjId) *
                                            (comment_buf.size + 1));
                comment_buf.id[comment_buf.size] =
                    pcomment_buf.rel[k].objectId;
                comment_buf.size++;
            }
            spdFreeRelArray(pcomment_buf);
        }
        spdFreeRelArray(lcomment_buf);
    }
    spdFreeRelArray(declr_buf);
    return comment_buf;
}

SpdObjIdArray getMacroComment(SpdObjId mname_id)
{
    SpdObjId        tmp;
    SpdRelArray     lcomment_buf, pcomment_buf;
    SpdObjIdArray   comment_buf = SPD_EMPTY_ARRAY;
    int             i, j;

    tmp = spdGetARelObj(mname_id, "mpatt_mname", "macroName_id");
    tmp = spdGetARelObj(tmp, "def_mppat", "macroPattern_id");
    lcomment_buf = spdGetRelSortedArray(tmp, "lcomment_pblock", "pblock_id",
                                       NULL, spdCompareRelOffset);
    for (i = 0; i < lcomment_buf.size; i++) {
        if (lcomment_buf.rel[i].objectId == SPD_NON_ID) continue;
        pcomment_buf = spdGetRelSortedArray(lcomment_buf.rel[i].objectId,
                                            "pcomment_lcomment",
                                            "c-id", NULL,
                                            spdCompareRelOffset);
        for (j = 0; j < pcomment_buf.size; j++) {
            comment_buf.id = spdRealloc(comment_buf.id, sizeof(SpdObjId) *
                                        (comment_buf.size + 1));
            comment_buf.id[comment_buf.size] = pcomment_buf.rel[j].objectId;
            comment_buf.size++;
        }
        spdFreeRelArray(pcomment_buf);
    }
    spdFreeRelArray(lcomment_buf);
    return comment_buf;
}

void putOneLineComment(FILE *fp, SpdObjId ident)
{
    SpdObjIdArray   comments;
    SpdString       comment_str;
    char            *ret;

    comments = getCommentIds(ident);
    if (comments.size) {
        comment_str = spdGetAttrValString(comments.id[0], "str");
        ret = strchr(comment_str, '\n');
        if (ret != NULL)
            *ret = '\0';
        putComment(fp, comment_str, MAX_COMMENT_LEN);
        fputs(genTagStr("br", NULL, NULL), fp);
    }
    spdFreeObjIdArray(comments);
}

void putOneLineCommentForMacro(FILE *fp, SpdObjId mname_id)
{
    SpdObjIdArray   comments;
    SpdString       comment_str;
    char            *ret;

    comments = getMacroComment(mname_id);
    if (comments.size) {
        comment_str = spdGetAttrValString(comments.id[0], "str");
        if (comment_str != NULL) {
            /* Ԥڤ */
            ret = strchr(comment_str, '\n');
            if (ret != NULL)
                *ret = '\0';
            putComment(fp, comment_str, MAX_COMMENT_LEN);
        }
    }
    spdFreeObjIdArray(comments);
}

void putSpecCommentForMacro(FILE *fp, SpdObjId mname_id)
{
    SpdObjIdArray   comments;
    SpdString       comment_str;
    int             i;

    fprintf(fp, "<td colspan=\"3\" valign=\"top\"><b>%s</b>\n", COMMENT_TITLE);
    comments = getMacroComment(mname_id);
    for (i = 0; i < comments.size; i++) {
        comment_str = spdGetAttrValString(comments.id[i], "str");
        putComment(fp, comment_str, -1);
    }
    spdFreeObjIdArray(comments);
    fprintf(fp, "</td>\n");
}

void putSpecComment(FILE *fp, SpdObjId ident)
{
    SpdObjIdArray   comments;
    SpdString       comment_str;
    int             i;

    fprintf(fp, "<td colspan=\"3\" valign=\"top\"><b>%s</b>\n", 
            COMMENT_TITLE);
    comments = getCommentIds(ident);
    for (i = 0; i < comments.size; i++) {
        comment_str = spdGetAttrValString(comments.id[i], "str");
        if (comment_str == NULL || *comment_str == '\0') continue;
        fputs("<pre>", fp);
        putComment(fp, comment_str, -1);
        fputs("</pre>", fp);
    }
    spdFreeObjIdArray(comments);
    fprintf(fp, "</td>\n");
}
