/* 
 * Program:             $RCSfile: config.c,v $  $Revision: 30.19 $
 * 
 * Purpose:             
 *
 * Author:              H.Ohhashi  1997/07/01
 *                      S.Yamamoto  1997/10/17
 *
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  1997 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: config.c,v 30.19 2003/06/28 15:34:13 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spie.h"

#define BG_COLOR "bg_color"
#define TEXT_COLOR "text_color"
#define LINK_COLOR "link_color"
#define ALINK_COLOR "alink_color"
#define VLINK_COLOR "vlink_color"

#define DEFAULT_BG "#FFF8DC"
#define DEFAULT_TEXT "#000000"
#define DEFAULT_LINK "#0000FF"
#define DEFAULT_ALINK "#FF0000"
#define DEFAULT_VLINK "#00008B"

ColorStyle color_style[MAX_STYLE_NUM] = {
    {"comment_color",   "#690000", "Comment"},  /* COMMENT */
    {"define_color",    "#003700", "Define"},   /* DEFINE */
    {"pp_color",        "#003700", "PP"},       /* PP */
    {"no_expand_color", "#999999", "NoExpand"}, /* NO_EXPAND */
    {"file_color",      "#0000FF", "File"},     /* FILE */
    {"func_color",      "#E10000", "Func"},     /* FUNC */
    {"var_color",       "#A50052", "Var"},      /* VAR */
    {"l_var_color",     "#750052", "L_Var"},    /* L_VAR */
    {"arg_color",       "#C50052", "Arg"},      /* ARG */
    {"macro_color",     "#009700", "Macro"},    /* MACRO */
    {"member_color",    "#3F00BF", "Member"},   /* MEMBER */
    {"enum_color",      "#FF8040", "Enum"},     /* ENUM */
    {"type_color",      "#004387", "Type"},     /* TYPE */
    {"const_color",     "#505050", "Const"},    /* CONST */
    {"tag_color",       "#FE00FE", "Tag"},      /* TAG */
    {"doc_color",       "#2693FF", "Doc"},      /* DOC */
    {"line_num_color",  "#000000", "LineNum"},  /* LINE */
    {"memo_color",      "#000000", "Memo"}      /* MEMO */
};

/* HTMLեο꤬Ҥ줿".spie"ɤߤؿ */
/* ĶѿHOMEѤ".spie"桼Υۡǥ쥯ȥ֤ */

void getConfigure(void)
{
    FILE *fp;
    char line[TEMP_MAX];
    char *colon_point;                      /* ':'Υݥ */
    char temp[TEMP_MAX];
    SpdString config_var;
    SpdString value;
    SpdString homeDir;
    SpdString configFile;
    int length, i;

    if ((homeDir = getenv("HOME")) == NULL) {
        spdErrorExit("\"HOME\" not found.\n");
    }
    configFile = spdCatPath(homeDir, CONFIG_FILE);

    /* ǥեȤ */
    bg_color = spdStrdup(DEFAULT_BG);
    text_color = spdStrdup(DEFAULT_TEXT);
    link_color = spdStrdup(DEFAULT_LINK);
    alink_color = spdStrdup(DEFAULT_ALINK);
    vlink_color = spdStrdup(DEFAULT_VLINK);

    if ((fp = spdSimpleFopen(configFile, SPD_FOPEN_RT_MODE)) == NULL) {
        /* ".spie"¸ߤʤ硤ǥեȤοѤ롥 */
        spdMessage("\"%s\" not found. Use default colors.\n", configFile);
        /* return˲롥 */
        spdFree(configFile);
        return;
    }

    /* spdMessage("Reading \"%s\".\n", configFile); */
    while ((fgets(line, TEMP_MAX, fp)) != NULL) {
        /* ԤƬ'#'ξȲԤΤߤξɤФ */
        if (line[0] == '#') continue;

        /* Ԥ˥󤬤ʤɤФ */
        if ((colon_point = strchr(line, ':')) == NULL) continue;

        /* ':'ʸФ */
        length = colon_point - line;
        strncpy(temp, line, length);
        temp[length] = '\0';

        /* ѿ(SPACE,TAB) */
        config_var = removeBlankString(temp);

        /* ݥ󥿤1Ŀʤ롥 */
        colon_point++;

        /* ͤ */
        value = removeBlankString(colon_point);
        if (SPD_STREQ(config_var, BG_COLOR)) {
            spdFree(bg_color);
            bg_color = spdStrdup(value);
        } else if (SPD_STREQ(config_var, TEXT_COLOR)) {
            spdFree(text_color);
            text_color = spdStrdup(value);
        } else if (SPD_STREQ(config_var, LINK_COLOR)) {
            spdFree(link_color);
            link_color = spdStrdup(value);
        } else if (SPD_STREQ(config_var, ALINK_COLOR)) {
            spdFree(alink_color);
            alink_color = spdStrdup(value);
        } else if (SPD_STREQ(config_var, VLINK_COLOR)) {
            spdFree(vlink_color);
            vlink_color = spdStrdup(value);
        } else {
            for (i = 0; i < MAX_STYLE_NUM; i++) {
                if (SPD_STREQ(config_var, color_style[i].name)) {
                    strncpy(color_style[i].color, value, COLOR_LENGTH);
                    break;
                }
            }
            if (i == MAX_STYLE_NUM) {
                spdWarning("Invalid configuration variable %s in %s.\n", 
                           config_var, configFile);
            }
        }
    }

    spdFclose(fp);
    spdFree(configFile);
}
