/* 
 * Program:             $RCSfile: file_stat.c,v $  $Revision: 30.5 $
 * 
 * Purpose:             
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/02/24
 * 
 * (C) Copyright:       K.Hayashi and S.Yamamoto  1995 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: file_stat.c,v 30.5 2001/08/26 08:05:38 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <sys/types.h>      /* For stat(). */
#include <sys/stat.h>       /* For stat(). */
#include <Sapid/Sapid.h>

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#include "spie.h"

char *get_m_time(char *fname)
{
    struct stat     buf;
    char            *date;
    
    if (stat(fname, &buf) != 0) {
        return ("");
    }
    date = ctime(&(buf.st_mtime));

    /* Note, date has its own newline character. */
    *(strchr(date, '\n')) = '\0';

    return (date);
}

/******************************************************************************/

#ifndef WIN32

char *get_owner(char *fname)
{
    struct stat     buf;
    struct passwd   *pwd;
    static char     name_buf[64];
    int             i;
    
    if (stat(fname, &buf) != 0) {
        return ("");
    }

    if ((pwd = getpwuid(buf.st_uid)) == NULL) {
        sprintf(name_buf, "%ld", (long)buf.st_uid);
    } else if (0 < strlen(pwd->pw_gecos)) {
        for (i = 0; pwd->pw_gecos[i] != '\0' && pwd->pw_gecos[i] != ','; i++) {
            name_buf[i] = pwd->pw_gecos[i];
        }
        name_buf[i] = '\0';
    } else {
        strcpy(name_buf, pwd->pw_name);
    }
    return (name_buf);
}

#else /* WIN32 */

#include <windows.h>
#include <fstream.h>

static SpdBoolean   isWindows95(void);

char *get_owner(char *fname)
{
    static char     name_buf[64];

    DWORD       bufsize, accountsize, refdomainsize;
    char        info[1023], owner[255];
    char        account[255];
    BOOL        success;
    char        refdomain[255];
    BOOL        owndef;
    PSID        psid = owner;
    SID_NAME_USE    snu;

    if (isWindows95()) {
        return ("");
    }

    /* 
     * GetFileSecurity()ؿϥեǥ쥯ȥ˴ؤꤵ줿
     * ޤǽʾϸƤӽФ¦Υøˤä
     * ¤ޤ
     */

    success = GetFileSecurity(
        fname, /* ե̾ʸΥɥ쥹 */
        OWNER_SECURITY_INFORMATION, /* ׵᤹ */
        (PSECURITY_DESCRIPTOR)info, /* ƥһҤΥɥ쥹 */ 
        sizeof(info), /* ƥһҤΥХåե */
        &bufsize); /* ɬפʥХåե Υɥ쥹 */
    if (!success) {
        /* 
         * cout << "GetFileSecurity(): Error " << GetLastError() << endl;
         * if (bufsize) {
         *     cout << "GetFileSecurity(): buffer is less\n";
         *     cout << "\tRequire " << bufsize << "byte.\n" << endl;
         * }
         * exit(1);
         */
        spdErrorExit("GetFileSecurity(): Error.\n");
    }

    /* 
     * GetSecurityDescriptorOwner()ؿϥƥһҤ
     * ʡޤ
     */
    
    success = GetSecurityDescriptorOwner(
        (PSECURITY_DESCRIPTOR)info, /* ƥһҤΥɥ쥹 */ 
        (PSID *)&psid, /* ʡSIDؤݥ󥿤Υɥ쥹 */ 
        &owndef); /* ǥեȤɤ򼨤ե饰Υɥ쥹 */
    if (!success) {
        /* 
         * cout << "GetSecurityDescriptorOwner(): Error.\n" << endl;
         * exit(1);
         */
        spdErrorExit("GetSecurityDescriptorOwner(): Error.\n");
    }

    accountsize = sizeof(account);
    refdomainsize = sizeof(refdomain);

    /* 
     * LookupAccountSid()ؿϥƥ̻(SID)ä
     * SIDΥ̾ȤSIDǽ˸Ĥäɥᥤ
     * ̾ޤ
     */

    success = LookupAccountSid(
        NULL, /* ƥ̾ʸΥɥ쥹 */ 
        psid, /* ƥ̻ҤΥɥ쥹 */ 
        account, /* ̾ʸΥɥ쥹 */ 
        &accountsize, /* ʸΥΥɥ쥹 */ 
        refdomain, /* ȤɥᥤʸΥɥ쥹 */ 
        &refdomainsize, /* ɥᥤʸΥΥɥ쥹 */ 
        &snu); /* SIDμι¤򼨤ɥ쥹 */ 
    if (!success) {
        /* 
         * cout << "LookupAccountSid(): Error.\n";
         * cout << "Require " << accountsize << " byte for account.\n";
         * cout << "Require " << refdomainsize;
         * cout << " byte for referenced domain.\n" << endl;
         * exit(1);
         */
        spdErrorExit("LookupAccountSid(): Error.\n");
    }

    strcpy(name_buf, account);
    return (name_buf);
}

static SpdBoolean isWindows95(void)
{
    OSVERSIONINFO osinfo;

    osinfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&osinfo);

    /* Win32s, Win95 or WinNT. */
    switch (osinfo.dwPlatformId) {
      case VER_PLATFORM_WIN32_WINDOWS:
        /* Windows 95. */
        return TRUE;
        break;
      case VER_PLATFORM_WIN32_NT:
        /* Windows NT. */
        return FALSE;
        break;
      default:
        /* Win32s. */
        return TRUE;
        break;
    }
}

#endif /* WIN32 */
