/*
 * Program:             $RCSfile: sort.c,v $  $Revision: 30.19 $
 *
 * Purpose:             
 *
 * Author:              H.Ohhashi
 * Modified:            S.Yamamoto  1995/03/13
 *
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: sort.c,v 30.19 2003/06/28 15:34:15 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spie.h"

/* Added by A.Watanabe and S.Yamamoto 2001/09/05. */
#define SORTED_BY_ID

#ifdef SPIE_MORE_SORT

static int compareXrefElem(SmuXrefTableElem *a, SmuXrefTableElem *b)
{
    int     c;

    if ((c = spdSafetyStrcmp(spdGetName(a->calleeId),
                         spdGetName(b->calleeId))) != 0) {
        return (c);
    } else if ((c = spdSafetyStrcmp(spdGetName(a->callerId),
                                spdGetName(b->callerId))) != 0) {
        return (c);
    } else {
        return (0);
    }
}

#endif /* SPIE_MORE_SORT */

#ifdef SPIE_MORE_SORT

/*
 * Ӵؿ()
 * ӤΤӤspdGetName()Ƥ֤ΤϸΨ?
 */

static int compareDefineName(DefElem *a, DefElem *b)
{
    int     c;

    if ((c = spdSafetyStrcmp(spdGetName(a->ident_id),
                         spdGetName(b->ident_id))) != 0) {
        return (c);
    } else if ((c = a->file_id - b->file_id) != 0) {
        return (c);
    } else if ((c = a->offset - b->offset) != 0) {
        return (c);
    } else {
#ifdef SORTED_BY_ID
        if ((c = a->ident_id - b->ident_id) != 0) {
            return (c);
        }
#endif
        return (0);
    }
}

#else /* not SPIE_MORE_SORT */

static int compareDefineName(DefElem *a, DefElem *b)
{
    int     c;

    c = spdSafetyStrcmp(spdGetName(a->ident_id), spdGetName(b->ident_id));
    if (c > 0) return(1);
    if (c < 0) return(-1);
    return(0);
}

#endif  /* not SPIE_MORE_SORT */

#ifdef SPIE_MORE_SORT

/*
 * Ӵؿ(и)
 */

static int compareDefineFileOffset(DefElem *a, DefElem *b)
{
    int     c;

    if ((c = a->file_id - b->file_id) != 0) {
        return (c);
    } else if ((c = a->ident_id - b->ident_id) != 0) {
        return (c);
    } else {
        return (a->offset - b->offset);
    }

}

static int compareVarTable(VarTableElem *a, VarTableElem *b)
{
    int     c;

    if ((c = a->file_id - b->file_id) != 0) {
        return (c);
    } else if ((c = a->var_id - b->var_id) != 0) {
        return (c);
    } else {
        return (a->offset - b->offset);
    }
}

#endif /* SPIE_MORE_SORT */

#ifndef BUGGY_SDBLINK4_0107

#ifdef SPIE_MORE_SORT

/* Ӵؿ() */

static int compareConstValue(ConstOccElem *a, ConstOccElem *b)
{
    int c;
    
    if ((c = (spdSafetyStrcmp(a->org_value, b->org_value))) != 0) {
        return (c);
    } else if ((c = a->file_id - b->file_id) != 0) {
        return (c);
    } else if ((c = a->offset - b->offset) != 0) {
        return (c);
    } else if ((c = a->const_id - b->const_id) != 0) {
        return (c);
    } else {
        return (0);
    }
}

#else /* not SPIE_MORE_SORT */

static int compareConstValue(ConstOccElem *a, ConstOccElem *b)
{
    return (spdSafetyStrcmp(a->value, b->value)); 
}

#endif /* not SPIE_MORE_SORT */

#endif /* BUGGY_SDBLINK4_0107 */

/* 
 * Added by K.Maeda 1999/11/24.
 * Ӵؿ(إåե)
 */

static int compareHeaderFileName(HeaderFileElem *a, HeaderFileElem *b)
{
    return(spdSafetyStrcmp(a->name , b->name));
}

/*
 * ɥȤǤoffsetˤӴؿ
 */

static int compareDocDefElemOffset(DocDefElem *a, DocDefElem *b)
{
    int c;
    
    if ((c = a->docfile_id - b->docfile_id) != 0) {
        return (c);
    } else {
        return (a->offset - b->offset);
    }
}

/*
 * ѿ˳ǼǤ򥽡Ȥ.
 * -bץHTML⡼ɤξ硤̵Ȥʤ.
 */

void sortElement(void)
{
    if (flag_b != ON || language_type != LATEX_MODE) {
        SPD_QSORT(func_def.elem, func_def.size, sizeof(DefElem), 
                    compareDefineName);
    }

    /* ޥ̾˥Ȥ롥 */
    SPD_QSORT(macro_def.elem, macro_def.size, sizeof(DefElem), 
                compareDefineName);
    if (language_type != LATEX_MODE) {
        /* ϥפHTML or XHTMLξ硥 */
        SPD_QSORT(var_def.elem, var_def.size, sizeof(DefElem), 
                    compareDefineName);
        SPD_QSORT(l_var_def.elem, l_var_def.size, sizeof(DefElem), 
                    compareDefineName);
        SPD_QSORT(arg_def.elem, arg_def.size, sizeof(DefElem), 
                    compareDefineName);
        SPD_QSORT(enum_def.elem, enum_def.size, sizeof(DefElem), 
                    compareDefineName); 
        SPD_QSORT(tag_def.elem, tag_def.size, sizeof(DefElem), 
                    compareDefineName); 
        SPD_QSORT(member_def.elem, member_def.size, sizeof(DefElem), 
                    compareDefineName); 
        SPD_QSORT(type_def.elem, type_def.size, sizeof(DefElem), 
                    compareDefineName);
    
        /* value򥭡ˤƥȤ롥 */
#ifndef BUGGY_SDBLINK4_0107
        SPD_QSORT(constant.elem, constant.size, sizeof(ConstOccElem), 
                    compareConstValue);
#endif

        /* Added by K.Maeda 1999/11/24. */
        SPD_QSORT(header.elem, header.size, sizeof(HeaderFileElem), 
                    compareHeaderFileName);

        SPD_QSORT(doc_def.elem, doc_def.size, sizeof(DocDefElem),
                    compareDocDefElemOffset);

#ifdef SPIE_MORE_SORT
        SPD_QSORT(l_var_occ.elem, l_var_occ.size, sizeof(DefElem), 
                    compareDefineFileOffset);
        SPD_QSORT(arg_occ.elem, arg_occ.size, sizeof(DefElem), 
                    compareDefineFileOffset);
        SPD_QSORT(member_occ.elem, member_occ.size, sizeof(DefElem), 
                    compareDefineFileOffset);
        SPD_QSORT(enum_occ.elem, enum_occ.size, sizeof(DefElem), 
                    compareDefineFileOffset);
        SPD_QSORT(type_occ.elem, type_occ.size, sizeof(DefElem), 
                    compareDefineFileOffset);
        SPD_QSORT(tag_occ.elem, tag_occ.size, sizeof(DefElem), 
                    compareDefineFileOffset);
        SPD_QSORT(var_table.elem, var_table.size, sizeof(VarTableElem), 
                    compareVarTable);
#endif
    }
}

#ifdef SPIE_MORE_SORT

void sortXrefTables(SmuXrefTables xreftables)
{
    SPD_QSORT(xreftables.funcCallTable.elem, xreftables.funcCallTable.size,
                sizeof(SmuXrefTableElem), 
                compareXrefElem);
    SPD_QSORT(xreftables.variableTable.elem, xreftables.variableTable.size,
                sizeof(SmuXrefTableElem), 
                compareXrefElem);
    SPD_QSORT(xreftables.globalVariableTable.elem,
                xreftables.globalVariableTable.size, sizeof(SmuXrefTableElem), 
                compareXrefElem);
    SPD_QSORT(xreftables.localVariableTable.elem,
                xreftables.localVariableTable.size, sizeof(SmuXrefTableElem), 
                compareXrefElem);
}

#endif /* SPIE_MORE_SORT */
