/*
 * Program:             $RCSfile: spieSub.c,v $  $Revision: 30.65 $
 *
 * Purpose:             Main file of spieSub.
 *
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/13
 *                      A.Watanabe  1997/03/26
 *
 * (C) Copyright:       K.Hayashi, A.Watanabe and S.Yamamoto  1995 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spieSub.c,v 30.65 2003/08/11 10:24:48 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <sys/types.h>      /* For stat(). */
#include <sys/stat.h>       /* For stat(). */
#include <Sapid/Sapid.h>

#define GLOBAL_DEFINE
#include "spie.h"

#include "JavaScript.h"

extern char *optarg;
extern int optind, opterr;

static void spieSetTipsHtmlDir(void);

/* ˡɽλؿ */

static void printUsage(void)
{
    spdErrorExit("Usage: %s [options][program]\n", program_name);
}

/*
 * HTMLեǼѤΥǥ쥯ȥ̾Υҡΰ롥
 */

static void spieFreeHtmlDir(void)
{
    if (directory != NULL) {
        spdFree(directory);
        directory = NULL;
    }
    if (flag_T) {
        if (tips_dir != NULL) {
            spdFree(tips_dir);
            tips_dir = NULL;
        }
    }
}

/* HTMLեǼѤΥǥ쥯ȥ롥 */

static void spieSetHtmlDir(SpdString topDir, SpdObjId prog_id)
{
    char        temp[SPD_STRING_LEN + 1];
    struct stat stat_buf;
    
    sprintf(temp, "%s/%d", topDir, prog_id);
    directory = spdStrdup(temp);
    /* HTMLեǼǥ쥯ȥ롥 */
    if (stat(directory, &stat_buf) != 0) {
        /* ǥ쥯ȥե¸ߤʤ硤롥 */
        if (flag_v) {
            spdMessage("Making directory \"%s\".\n", directory);
        }
        spdMkdirR(directory);
    } else if ((stat_buf.st_mode & S_IFMT) == S_IFDIR) { 
        /*
         * Ʊ̾Υǥ쥯ȥ꤬¸ߤƤ硤ʤˤ⤷ʤ
         * ŪˤwarningϤɬפ뤫⤷ʤ 
         * spdWarning("Directory \"%s\" exists.\n", directory); 
         */
    } else {
        /* ǥ쥯ȥ̾Ʊ̾Υե뤬¸ߤ硥 */
        spdErrorExit("File \"%s\" exists.\n", directory);
    }
}

/* TipsHTMLեǼѤΥǥ쥯ȥ롥 */

static void spieSetTipsHtmlDir(void)
{
    char        temp[SPD_STRING_LEN + 1];
    struct stat stat_buf;
    
    strcpy(temp, directory);
    tips_dir = spdCatPath(temp, tipsHtmlDir);
    /* HTMLեǼǥ쥯ȥ롥 */
    if (stat(tips_dir, &stat_buf) != 0) {
        /* ǥ쥯ȥե¸ߤʤ硤롥 */
        if (flag_v) {
            spdMessage("Making directory \"%s\".\n", tips_dir);
        }
        spdMkdirR(tips_dir);
    } else if ((stat_buf.st_mode & S_IFMT) == S_IFDIR) { 
        /*
         * Ʊ̾Υǥ쥯ȥ꤬¸ߤƤ硤ʤˤ⤷ʤ
         * ŪˤwarningϤɬפ뤫⤷ʤ 
         * spdWarning("Directory \"%s\" exists.\n", tips); 
         */
    } else {
        /* ǥ쥯ȥ̾Ʊ̾Υե뤬¸ߤ硥 */
        spdErrorExit("File \"%s\" exists.\n", tips_dir);
    }
}

/*--------------------------------------------------------------*/
/*              оݥץ̾                          */
/*--------------------------------------------------------------*/

static void spieGetTargetProgName(int argc, char *argv[])
{
    /* ϤоݤȤʤץ̾롥 */
    if (optind == argc) {
        target_program_name = smuGetAProgramName();
    } else if (optind == (argc - 1)) {
        target_program_name = argv[optind];
    } else {
        printUsage();
    }
}

/*--------------------------------------------------------------*/
/*              ץ                                  */
/*--------------------------------------------------------------*/

static void spieAnalyzeOption(int argc, char *argv[])
{
    int c;
    int exit_code = EXIT_SUCCESS;

    while ((c = getopt(argc, argv, "a:bcd:f:ilmnpq:st:vwxCD:S:T")) != EOF) {
        switch (c) {
          case 'a':
            author = optarg;
            break;
          case 'b':
            flag_b = ON;
            break;
          case 'c':
            flag_c = ON;
            break;
          case 'd':
            spdSetSDBdir(optarg);
            break;
          case 'i':
            flag_i = ON;
            break;
          case 'l':
            flag_l = ON;
            break;
          case 'm':
            flag_m = ON;
            break;
          case 'n':
            flag_n = ON;
            break;
          case 'p':
            flag_p = ON;
            break;
          case 'q':
            flag_q = ON;
            lib_directory = optarg;
            break;
          case 's':
            flag_s = ON;
            break;
          case 't':
            if (SPD_STREQ(optarg, OUTPUT_TYPE_HTML)) {
                language_type = HTML_MODE;
            } else if (SPD_STREQ(optarg, OUTPUT_TYPE_LATEX)) {
                language_type = LATEX_MODE;
            } else if (SPD_STREQ(optarg, OUTPUT_TYPE_XHTML)) {
                language_type = XHTML_MODE;
            } else {
                spdErrorExit("Invalid output type \"%s\".\n", optarg);
            }
            break;
          case 'v':
            flag_v = ON;
            break;
          case 'w':
            flag_w = ON;
            break;
          case 'x':
            flag_x = ON;
            break;
          case 'C':
            flag_C = ON;
            break;
          case 'D':
            htmlTopDir = optarg;
            break;
          case 'S':
            smuSetSystemPath(optarg);
            break;
          case 'T':
            flag_T = ON;
            break;
          default:
            exit_code = EXIT_FAILURE;
            break;
        }
    }

    if (exit_code != EXIT_SUCCESS) {
        printUsage();
    }

    spieGetTargetProgName(argc, argv);
}

/*
 * compare function. objectId򾺽˥Ȥ뤿ᡥ
 */

static int spdCompareObjId(const SpdObjId *a, const SpdObjId *b)
{
    if (*a > *b) return 1;
    if (*a < *b) return -1;
    return 0;
}

/*--------------------------------------------------------------*/
/*              оݥץID                    */
/*--------------------------------------------------------------*/

static SpdObjIdArray spieGetProgramIds(void)
{
    SpdObjIdArray prog_ids;

#ifdef SAPID_MEMORY_PROFILE
    spdMessage("\"%s\": %d.\n", __FILE__, __LINE__);
    spdPrintMemProf();
#endif

    /*
     * program ID򾺽˥ȤƼ롥
     * IDŤСΤΤȲꡥ
     */
    prog_ids = spdGetObjIdSortedArray("program", NULL, NULL,
                                      spdCompareObjId);
    return prog_ids;
}

/*--------------------------------------------------------------*/
/*              եID                      */
/*                      ե꡼ɬפ                    */
/*--------------------------------------------------------------*/

static SpdObjIdArray spieGetFileIds(SpdObjId prog_id)
{
    SpdObjIdArray file_buf;
    SpdObjId pfile_id;
    int i;

    file_buf = spdGetRelObjSortedArray(prog_id, "prog_file", "p-id",
                                       NULL, spdCompareObjIdName);

    if (language_type != LATEX_MODE) {
        for (i = 0; i < file_buf.size; ++i) {
            /*
             * getPBlockؿϥإåեƵƤӽФǽ뤿ᡤ
             * pfile_idˤƤ롥Τᡤcfile_ifilepfile_id
             * ɬפ. 
             */
            pfile_id = spdGetARelObj(file_buf.id[i], "cfile_ifile", "file_id");
            getPBlock(pfile_id);
            getOcc(file_buf.id[i], &l_var_occ, spdOccIsIdVariable);
            getOcc(file_buf.id[i], &arg_occ, spdOccIsIdArgument);
            getOcc(file_buf.id[i], &member_occ, spdOccIsIdMember);
            getOcc(file_buf.id[i], &enum_occ, spdOccIsIdEnum);
            getTypeOccInExpression(file_buf.id[i]);

#ifndef BUGGY_SDBLINK4_0107
            getConstant(file_buf.id[i]);
#endif
            /*
             * getPBlockθgetDeclarationƤ֤褦ˤ.
             * (macro_callΥѿfunc_idƤޤ)
             */
            getDeclaration(file_buf.id[i]);
        }       
    } else {
        for (i = 0; i < file_buf.size; ++i) {
            getDeclaration(file_buf.id[i]);
        }
    }

    return(file_buf);
}

/*--------------------------------------------------------------*/
/*              ϴؿ                                        */
/*--------------------------------------------------------------*/

static void spieOutput(SpdObjId prog_id, SpdObjIdArray file_buf)
{
    if (language_type != LATEX_MODE) {
        /* ϤHTMLեοԤ */
        if (!flag_c)
            getConfigure();
        /* HTMLǼǥ쥯ȥ򥻥åȡ롥 */
        spieSetHtmlDir(htmlTopDir, prog_id);
        if (flag_p) {
            putJSTitleFunc();
        }
        if (flag_q)
            putQuickJumpHtml();
        /* TipsHTMLǼǥ쥯ȥ򥻥åȡ롥 */
        if (flag_T) {
            putJSTipsFunc();
            spieSetTipsHtmlDir();
        }
        /* ƼHTMLե롥 */
        putHtml(file_buf.id, file_buf.size);
        /* HTML뤿spdSed script file롥 */
        putScript(file_buf.id, file_buf.size);

        spieFreeHtmlDir();
    } else {
        putSpecLatex();
    }
}

/****************************************************************/
/*              ᥤؿ                                      */
/****************************************************************/

#ifdef SPIE_MORE_SORT
extern void sortXrefTables(SmuXrefTables xreftables);
#endif

int main(int argc, char *argv[])
{
    SpdObjIdArray   prog_buf;
    SpdObjIdArray   file_buf;
    SpdDBId dbId;
    SpdModelId ImodelId, PmodelId, DmodelId, CviewId;
    int             i;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);

    /* SpdStrqualifierϤ뤿ꡥ */
    spdSetOption("typeQualifier", ON);

    spieAnalyzeOption(argc, argv);

    ImodelId = spdUseModel(&I_model);
    PmodelId = spdUseModel(&P_model);
    if (flag_x) {
        DmodelId = spdUseModel(&D_model);
    }
    if (flag_C) {
        CviewId = spdUseModel(&Comment_view);
    }

    dbId = spdOpenSDB();

    spdReadDB(ImodelId, dbId);
    spdReadDB(PmodelId, dbId);
    if (flag_x) {
        spdReadDB(DmodelId, dbId);
    }
    if (flag_C) {
        spdReadDB(CviewId, dbId);
    }
    
    prog_buf = spieGetProgramIds();

    /* programο֤ */
    for (i = 0; i < prog_buf.size; i++) {
        /* ƽнդ뤳ȡ */
#ifdef SPIE_TRAVERSE
        traverse = spieTraverse(prog_buf.id[i]);
        xreftables = spieMakeXrefTables(prog_buf.id[i], traverse,
                                        smuXrefEachOcc);
#else
        xreftables = smuMakeXrefTables(prog_buf.id[i], smuXrefEachOcc);
#endif

#ifdef SPIE_MORE_SORT
        sortXrefTables(xreftables);
#endif  

        makeProtoTable();

        if (flag_x) {
            /* sort  macro  docidentId Ͽ */
            setMacroDocidentIdintoArray();
        }

        file_buf = spieGetFileIds(prog_buf.id[i]);

        if (flag_x) {
            getDocDefine();
        }

        /* Ǥ򥽡Ȥ롥 */
        sortElement();

#ifdef DEBUG
        elementDump();
#endif
        spieOutput(prog_buf.id[i], file_buf);

        spdFreeObjIdArray(file_buf);
        smuFreeXrefTables(xreftables);
#if defined(SPIE_TRAVERSE)
        spieFreeTraverse(traverse);
#endif  
        freeProtoTable();
        elementFree();
    }
    spdFreeOption();
    spdFreeObjIdArray(prog_buf);

    spdCloseSDB(dbId);
    
#ifdef SAPID_MEMORY_PROFILE
    spdMessage("\"%s\": %d.\n", __FILE__, __LINE__);
    spdPrintMemProf();
#endif

    return (EXIT_SUCCESS);
}
