/* 
 * Program:             $RCSfile: type.c,v $  $Revision: 30.5 $
 * 
 * Purpose:             
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            A.Watanabe  1995/02/24
 *                      S.Yamamoto  1995/02/24
 * 
 * (C) Copyright:       K.Hayashi, A.Watanabe and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: type.c,v 30.5 2002/12/16 12:12:05 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spie.h"

static SpdString getTypeName_raw(SpdObjId type_id);

static SpdString getTypeName_raw(SpdObjId type_id)
{
    SpdString   text;
    SpdObjId    expr_id;
    static char buf[SPD_LINE_LEN + 1];
    
    strcpy(buf, "");
    switch (spdGetAttrValInt(type_id, "qualifier")) {
      case TYPE_CONST:
        strcat(buf, "const ");
        break;
      case TYPE_VOLATILE:
        strcat(buf, "volatile ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "sign")) {
      case TYPE_SIGNED:
        strcat(buf, "signed ");
        break;
      case TYPE_UNSIGNED:
        strcat(buf, "unsigned ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "size")) {
      case TYPE_SHORT:
        strcat(buf, "short ");
        break;
      case TYPE_LONG:
        strcat(buf, "long ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "type")) {
      case TYPE_VOID:
        strcat(buf, "void ");
        break;
      case TYPE_INT:
        strcat(buf, "int ");
        break;
      case TYPE_CHAR:
        strcat(buf, "char ");
        break;
      case TYPE_FLOAT:
        strcat(buf, "float ");
        break;
      case TYPE_DOUBLE:
        strcat(buf, "double ");
        break;
      case TYPE_DOTS:
        strcat(buf, "... ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "sort")) {
      case TYPE_TYPE_DEF:
        strcat(buf, spdGetName(type_id));
        strcat(buf, " ");
        break;
      case TYPE_STRUCT_TAG:
        strcat(buf, "struct ");
        strcat(buf, spdGetName(type_id));
        strcat(buf, " ");
        break;
      case TYPE_DECLARATION:
        break;
      case TYPE_POINTER:
        strcat(buf, "*");
        break;
      case TYPE_ARRAY:
        strcat(buf, "[");
        if ((expr_id = spdGetARelObj(type_id, "type_cexpr", "any")) != -1)
            if ((text = spdGetObjText(expr_id)) != NULL)
                strcat(buf, text);
        strcat(buf, "]");
        break;
      case TYPE_FUNC:
        strcat(buf, "()");
        break;
      case TYPE_ENUM_SPECIFIER:
        strcat(buf, "enum ");
        strcat(buf, spdGetName(type_id));
        strcat(buf, " ");
        break;
    }
    return buf;
}

SpdString getTypeName(SpdObjId type_id)
{
    char        buf[1024];
    int         inner_type_id;
    static char res[1024];
    
    strcpy(buf, "");

    if ((inner_type_id = spdGetARelObj(type_id, "type_st", "type_pointer_id"))
        != -1) {
        strcat(buf, getTypeName(inner_type_id));
    }
    strcat(buf, getTypeName_raw(type_id));

    strcpy(res, buf);
    return res;
}

SpdString getDeclStr(SpdObjId type_id, char *decl)
{
    char        buf[1024];
    SpdObjId    expr_id, inner_type_id;
    char        *text;
    static char res[1024];
    
    strcpy(buf, "");
    switch (spdGetAttrValInt(type_id, "qualifier")) {
      case TYPE_CONST:
        strcat(buf, "const ");
        break;
      case TYPE_VOLATILE:
        strcat(buf, "volatile ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "sign")) {
      case TYPE_SIGNED:
        strcat(buf, "signed ");
        break;
      case TYPE_UNSIGNED:
        strcat(buf, "unsigned ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "size")) {
      case TYPE_SHORT:
        strcat(buf, "short ");
        break;
      case TYPE_LONG:
        strcat(buf, "long ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "type")) {
      case TYPE_VOID:
        strcat(buf, "void ");
        break;
      case TYPE_INT:
        strcat(buf, "int ");
        break;
      case TYPE_CHAR:
        strcat(buf, "char ");
        break;
      case TYPE_FLOAT:
        strcat(buf, "float ");
        break;
      case TYPE_DOUBLE:
        strcat(buf, "double ");
        break;
      case TYPE_DOTS:
        strcat(buf, "... ");
        break;
    }
    switch (spdGetAttrValInt(type_id, "sort")) {
      case TYPE_TYPE_DEF:
        strcat(buf, spdGetName(type_id));
        strcat(buf, " ");
        break;
      case TYPE_STRUCT_TAG:
        strcat(buf, "struct ");
        strcat(buf, spdGetName(type_id));
        strcat(buf, " ");
        break;
      case TYPE_DECLARATION:
        break;
      case TYPE_POINTER:
        strcat(buf, "*");
        break;
      case TYPE_ENUM_SPECIFIER:
        strcat(buf, "enum ");
        strcat(buf, spdGetName(type_id));
        strcat(buf, " ");
        break;
    }
    strcat(buf, decl);
    switch (spdGetAttrValInt(type_id, "sort")) {
      case TYPE_ARRAY:
        strcat(buf, "[");
        if ((expr_id = spdGetARelObj(type_id, "type_cexpr", "any")) != -1)
            if ((text = spdGetObjText(expr_id)) != NULL)
                strcat(buf, text);
        strcat(buf, "]");
        break;
      case TYPE_FUNC:
        strcat(buf, "()");
        break;
    }
    
    if ((inner_type_id = spdGetARelObj(type_id, "type_st", "type_pointer_id"))
        != -1)
        return getDeclStr(inner_type_id, buf);
    strcpy(res, buf);
    return res;
}
