/* 
 * Program:             $RCSfile: traverse.c,v $  $Revision: 30.5 $
 * 
 * Purpose:
 * 
 * Author:              A.Watanabe  2001/01/29
 *                      S.Yamamoto  2001/01/29
 * 
 * (C) Copyright:       A.Watanabe and S.Yamamoto  2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: traverse.c,v 30.5 2002/10/21 07:28:55 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spie.h"

#if defined(SPIE_TRAVERSE)

static SpdTrvActFunc beforeFile(SpdObjId id, SpieTraverse *trav);
static SpdTrvActFunc afterFile(SpdObjId id, SpieTraverse *trav);
static SpdTrvActRelFunc beforeRelDecl(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav);
static SpdTrvActRelFunc beforeRelDecr(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav);
static SpdTrvActRelFunc afterRelDecr(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav);
static SpdTrvActRelFunc beforeRelIdent(SpdObjId id, SpdRel rel,
                                       SpieTraverse *trav);
static SpdTrvActRelFunc beforeRelExpr(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav);
static SpdTrvActRelFunc beforeRelConst(SpdObjId id, SpdRel rel,
                                       SpieTraverse *trav);

static SmuXrefTable makeXrefTableSub(SmuXrefTable xref_table,
                                     CalleeBuf callee,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId));

SpieTraverse spieTraverse(SpdObjId prog_id)
{
    SpdTrvActInfo   *info;
    SpieTraverse trav;

    trav.fileId = SPD_NON_ID;
    trav.funcId = SPD_NON_ID;

    trav.decln_buf.size = 0;
    trav.decln_buf.occ = NULL;
    trav.ident_buf.size = 0;
    trav.ident_buf.occ = NULL;
    trav.expr_buf.size = 0;
    trav.expr_buf.occ = NULL;
    trav.const_buf.size = 0;
    trav.const_buf.occ = NULL;
    trav.callee.size = 0;
    trav.callee.elem = NULL;
    
    info = spdTrvInitActInfo();

    /* file. */
    spdTrvAddRel(info, SPD_RID_PROG_FILE);
    /*
     * declaration.
     */    
    /* file -> declaration. */    
    spdTrvAddRel(info, SPD_RID_FILE_DECL);
    /* block -> declaration. */
    spdTrvAddRel(info, SPD_RID_BLK_DECL);
    /* optionaldecl  -> declaration. */
    spdTrvAddRel(info, SPD_RID_STRUCT_DECL);
    spdTrvAddRel(info, SPD_RID_PROT_PARM);
    spdTrvAddRel(info, SPD_RID_FUNC_PARM);
    spdTrvAddRel(info, SPD_RID_FUNCPTR_PARM);
    
    /* declarator. */    
    spdTrvAddRel(info, SPD_RID_DECL_DECL);
    /* optionaldecl */
    spdTrvAddRel(info, SPD_RID_DECL_OPT);

    /* block. */
    spdTrvAddRel(info, SPD_RID_FUNC_BODY);
    spdTrvAddRel(info, SPD_RID_BLK_BLK);

    /*
     * expression.
     */
    spdTrvAddRel(info, SPD_RID_BLK_EXPR);
    
    spdTrvAddRel(info, SPD_RID_BLK_LABEL);
    spdTrvAddRel(info, SPD_RID_CASE_CEXPR);
    
    spdTrvAddRel(info, SPD_RID_EXPR_EXPR);
    
    spdTrvAddRel(info, SPD_RID_INIT_EXPR);

    /* identifier. */
    spdTrvAddRel(info, SPD_RID_IDENT_REF);
    /* constant. */
    spdTrvAddRel(info, SPD_RID_CONST_REF);
    /* 󥵥 */
    spdTrvAddRel(info, SPD_RID_DECR_TYPE);
    spdTrvAddRel(info, SPD_RID_TYPE_CEXPR);
    spdTrvAddRel(info, SPD_RID_TYPE_ST);
    
    /* ãν */
    /* file. */
    spdTrvAddActFuncOfClass(info, SPD_CID_FILE, SAPID_TRV_ACT_BEFORE,
                            (SpdTrvActFunc)beforeFile);
    spdTrvAddActFuncOfClass(info, SPD_CID_FILE, SAPID_TRV_ACT_AFTER,
                            (SpdTrvActFunc)afterFile);
    /* declaration. */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_FILE_DECL, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecl);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_BLK_DECL, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecl);

    spdTrvAddActFuncOfRelClass(info, SPD_RID_STRUCT_DECL, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecl);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_PROT_PARM, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecl);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_FUNC_PARM, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecl);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_FUNCPTR_PARM, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecl);
    /* declarator. */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_DECL_DECL, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelDecr);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_DECL_DECL, SAPID_TRV_ACT_AFTER,
                               (SpdTrvActRelFunc)afterRelDecr);
    /* expression. */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_BLK_EXPR, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelExpr);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_CASE_CEXPR, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelExpr);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_EXPR_EXPR, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelExpr);
    spdTrvAddActFuncOfRelClass(info, SPD_RID_INIT_EXPR, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelExpr);
    /* identifier. */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_IDENT_REF, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelIdent);
    /* constant. */
    spdTrvAddActFuncOfRelClass(info, SPD_RID_CONST_REF, SAPID_TRV_ACT_BEFORE,
                               (SpdTrvActRelFunc)beforeRelConst);

    spdTraverse(prog_id, info, &trav);
    spdTrvFreeActInfo(info);

    return trav;
}

/* file. */

static SpdTrvActFunc beforeFile(SpdObjId id, SpieTraverse *trav)
{
    trav->fileId = id;
    trav->funcId = SPD_NON_ID;
    return SAPID_TRV_GO;
}

static SpdTrvActFunc afterFile(SpdObjId id, SpieTraverse *trav)
{
    trav->fileId = SPD_NON_ID;
    trav->funcId = SPD_NON_ID;
    return SAPID_TRV_GO;
}

/* declaration. */

static SpdTrvActRelFunc beforeRelDecl(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav)
{
    SpdOcc occ;
    SpdOccArray *array = &(trav->decln_buf);

    occ.fileId = trav->fileId;
    occ.relationId = rel.relationId;
    occ.objectId = rel.objectId;
    occ.offset = spdGetAbsoluteOffset(occ.relationId, occ.fileId);

    array->occ = spdRealloc(array->occ, sizeof(SpdOcc) * (array->size + 1));
    
    array->occ[array->size] = occ;
    array->size++;
    
    return SAPID_TRV_GO;
}

/* declarator. */

static SpdTrvActRelFunc beforeRelDecr(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav)
{
    if (spdGetAttrValInt(id, "sort") == DECL_FUNCDEF) {
        trav->funcId = spdGetARelObj(rel.objectId, "decl_ident", "p-id");
    }
    return SAPID_TRV_GO;
}

static SpdTrvActRelFunc afterRelDecr(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav)
{
    if (spdGetAttrValInt(id, "sort") == DECL_FUNCDEF) {
        trav->funcId = SPD_NON_ID;
    }
    return SAPID_TRV_GO;
}

/* identifier. */

static SpdTrvActRelFunc beforeRelIdent(SpdObjId id, SpdRel rel,
                                       SpieTraverse *trav)
{
    SpdOcc occ;
    SpdOccArray *array = &(trav->ident_buf);
    CalleeBufElem callee;
    CalleeBuf *array2 = &(trav->callee);

    occ.fileId = trav->fileId;
    occ.relationId = rel.relationId;
    occ.objectId = rel.objectId;
    occ.offset = spdGetAbsoluteOffset(occ.relationId, occ.fileId);

    callee.fileId = trav->fileId;
    callee.callerId = trav->funcId;
    callee.relationId = rel.relationId;
    callee.calleeId = rel.objectId;
    
    array->occ = spdRealloc(array->occ, sizeof(SpdOcc) * (array->size + 1));
    array->occ[array->size] = occ;
    array->size++;

    array2->elem = spdRealloc(array2->elem, sizeof(CalleeBufElem) * (array2->size + 1));
    array2->elem[array2->size] = callee;
    array2->size++;

    return SAPID_TRV_GO;
}

/* expression. */

static SpdTrvActRelFunc beforeRelExpr(SpdObjId id, SpdRel rel,
                                      SpieTraverse *trav)
{
    SpdOcc occ;
    SpdOccArray *array = &(trav->expr_buf);


    occ.fileId = trav->fileId;
    occ.relationId = rel.relationId;
    occ.objectId = rel.objectId;
    occ.offset = spdGetAbsoluteOffset(occ.relationId, occ.fileId);

    array->occ = spdRealloc(array->occ, sizeof(SpdOcc) * (array->size + 1));
    array->occ[array->size] = occ;
    array->size++;
    
    return SAPID_TRV_GO;
}

/* constant. */

static SpdTrvActRelFunc beforeRelConst(SpdObjId id, SpdRel rel,
                                       SpieTraverse *trav)
{
    SpdOcc occ;
    SpdOccArray *array = &(trav->const_buf);


    occ.fileId = trav->fileId;
    occ.relationId = rel.relationId;
    occ.objectId = rel.objectId;
    occ.offset = spdGetAbsoluteOffset(occ.relationId, occ.fileId);

    array->occ = spdRealloc(array->occ, sizeof(SpdOcc) * (array->size + 1));
    array->occ[array->size] = occ;
    array->size++;
    
    return SAPID_TRV_GO;
}

void spieFreeTraverse(SpieTraverse trav)
{
    spdFreeOccArray(trav.decln_buf);
    spdFreeOccArray(trav.ident_buf);
    spdFreeOccArray(trav.expr_buf);
    spdFreeOccArray(trav.const_buf);
    spdFree(trav.callee.elem);
}

/*
 * ե󥹥ơ֥κ(ȥС)
 */

SmuXrefTables  spieMakeXrefTables(SpdObjId prog_id, SpieTraverse trav,
                                  SmuXrefFlag flag)
{
    SmuXrefTables   xreftables = {{0, NULL}, {0, NULL}, {0, NULL}, {0, NULL},
                                      {0, NULL}};
    
    xreftables.funcCallTable =
        makeXrefTableSub(xreftables.funcCallTable, trav.callee, flag,
                         spdObjIdIsIdFunction);
    xreftables.variableTable =
        makeXrefTableSub(xreftables.variableTable, trav.callee, flag,
                         spdObjIdIsIdVariable);
    xreftables.globalVariableTable =
        makeXrefTableSub(xreftables.globalVariableTable, trav.callee, flag,
                         smuIsGlobalVariable);
    xreftables.localVariableTable =
        makeXrefTableSub(xreftables.localVariableTable, trav.callee, flag,
                         smuIsLocalVariable);
    
    return xreftables;
}

/*
 * ե󥹥ơ֥κ
 */

static SmuXrefTable makeXrefTableSub(SmuXrefTable xref_table,
                                     CalleeBuf callee,
                                     SmuXrefFlag flag,
                                     SpdBoolean (*cond)(SpdObjId))
{
    SpdSet  callee_set = SAPID_EMPTY_SET;
    int     i;

    for (i = 0; i < callee.size; i++) {
        if (cond != NULL && !cond(callee.elem[i].calleeId)) {
            continue;
        }

        if (flag) {
            if (spdSetContains(callee_set, callee.elem[i].calleeId)) {
                continue;
            } else {
                spdAddElem(callee_set, callee.elem[i].calleeId, NULL);
            }
        }

        xref_table.elem = spdRealloc(xref_table.elem,
                                     sizeof(SmuXrefTableElem) *
                                     (xref_table.size + 1));
        
        xref_table.elem[xref_table.size].callerId = callee.elem[i].callerId;
        xref_table.elem[xref_table.size].calleeId = callee.elem[i].calleeId;
        xref_table.elem[xref_table.size].relationId = callee.elem[i].relationId;
        xref_table.size++;
    }
    spdFreeSet(callee_set);

    return xref_table;
}

#endif /* defined(SPIE_TRAVERSE) */
