/* 
 * Program:             $RCSfile: lib_version.c,v $  $Revision: 30.8 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohhashi  2000/01/21
 *                      S.Yamamoto  2000/01/21
 * 
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: lib_version.c,v 30.8 2003/06/28 15:33:56 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spdCmpVersion.h"

/* 
 * ؿ
 */

SpdBoolean spdDeclrIsDeclFuncDef(const SpdObjId declrId)
{
    return (spdGetAttrValInt(declrId, "sort") == DECL_FUNCDEF);
}

SpdBoolean spdDeclnIsDeclParm(SpdRel rel)
{
    return (spdGetAttrValInt(rel.objectId, "sort") == DECL_PARAMETER);
}


SpdBoolean spdRelIsDeclFuncDef(SpdRel rel)
{
    return (spdGetAttrValInt(rel.objectId, "sort") == DECL_FUNCDEF); 
}

SpdBoolean spdRelIsDeclVariable(SpdRel rel)
{
    return (spdGetAttrValInt(rel.objectId, "sort") == DECL_VARIABLE); 
}

SpdBoolean spdRelIsDeclType(SpdRel rel)
{
    return (spdGetAttrValInt(rel.objectId, "sort") == DECL_TYPE); 
}

SpdBoolean spdRelIsDeclTag(SpdRel rel)
{
    return (spdGetAttrValInt(rel.objectId, "sort") == DECL_TAG); 
}

/******************************************************************************/

/*
 * Ϣ".c"region뤿short-cutؿ
 */

SpdRegion spdGetRegionC(SpdRel rel)
{
    SpdRegion region_i, region_c;

    region_i.fileId = spdGetFileId(rel.relationId);
    region_i.offset = spdGetAbsoluteOffset(rel.relationId, 
                                           region_i.fileId);
    region_i.length = spdGetLength(rel.objectId);

    region_c = spdConvRegionItoC(region_i);
    return region_c;
}

/*
 * identifierƤdeclaratorȤδϢ롥
 */

SpdRel spdGetDefRel(SpdObjId identId)
{
    SpdRelArray rel_buf;
    SpdRel ret = {SPD_NON_ID, SPD_NON_ID};

    switch (spdGetAttrValInt(identId, "sort")) {
      case ID_FUNCTION:
        rel_buf = spdGetRelArray(identId, "decl_ident", "identifier_id",
                                 spdRelIsDeclFuncDef);
        break;
      case ID_VARIABLE:
        rel_buf = spdGetRelArray(identId, "decl_ident", "identifier_id",
                                 spdRelIsDeclVariable);
        break;
      case ID_TYPENAME:
        rel_buf = spdGetRelArray(identId, "decl_ident", "identifier_id",
                                 spdRelIsDeclType);
        break;
      case ID_TAG:
        rel_buf = spdGetRelArray(identId, "decl_ident", "identifier_id",
                                 spdRelIsDeclTag);
        break;
      default:
        break;
    }

    if (rel_buf.size == 0) return ret;

    /* ret.objectId = rel_buf.rel[0].objectId; */
    ret = spdGetARel(rel_buf.rel[0].objectId, "decl_ident",
                     "declarator_id");
    spdFreeRelArray(rel_buf);
    return ret;
}

/*
 * identIdսregion롥
 */

SpdRegion spdGetDefRegion(SpdObjId identId)
{
    SpdRegion regionC;
    SpdRel decl_ident_rel;

    decl_ident_rel = spdGetDefRel(identId);
    regionC = spdGetRegionC(decl_ident_rel);
    return regionC;
}

/*************************************************************/

/*
 * ident_idGlobalǤ뤫ɤȽꡥ
 */

SpdBoolean spdIsGlobalIdent(SpdObjId ident_id)
{
    SpdObjId    decl_id, decr_id;
    SpdCursor   *csr;
    SpdBoolean  isGlobal = FALSE;
    
    if (spdGetClassId(ident_id) != SPD_CID_IDENTIFIER)
        return FALSE;

    csr = spdGetRelObjInit(ident_id, "decl_ident", "identifier_id");
    while ((decr_id = spdGetRelObj(csr)) != SPD_NON_ID) {
        decl_id = spdGetARelObj(decr_id, "decl_decl", "declarator_id");
        if (spdGetARelObj(decl_id, "file_decl", "declaration_id") 
            != SPD_NON_ID) {
            isGlobal = TRUE;
            break;
        }
    }
    spdFreeCursor(csr);
    return isGlobal;
}

/*
 * relationId°programID롥
 */

SpdObjId spdGetProgramIdByRelId(SpdRelId relationId)
{
    SpdObjId file_id = spdGetFileId(relationId);

    return (spdGetARelObj(file_id, "prog_file", "file_id"));
}

/*
 * program_idprogram椫ident_idбidentifierID롥
 */

SpdObjId spdGetIdentIdWithVersion(SpdObjId ident_id,
                                  SpdObjId program_id)
{
    int i;
    SpdObjIdArray ident_array;
    SpdRel rel;
    SpdObjId ret_ident_id = SPD_NON_ID;
    int attr_sort = spdGetAttrValInt(ident_id, "sort");

    /*
     * Υåident_ididentifierΥ饹Ǥꡤ
     * globalǤ뤳Ȥγǧ
     */
    if (spdGetClassId(ident_id) != SPD_CID_IDENTIFIER)
        return SPD_NON_ID;
    if (spdIsGlobalIdent(ident_id) == FALSE) return SPD_NON_ID;
  
    /* ident_idƱ̾identifierID򤹤٤롥*/
    ident_array = spdGetObjIdArray("identifier", 
                                   spdGetName(ident_id), NULL);

    for (i = 0; i < ident_array.size; i++) {
        /* globalǤ뤫ɤ*/
        if (spdIsGlobalIdent(ident_array.id[i]) == FALSE) continue;

        /* sort°ƱɤӤ롥*/
        if (attr_sort != 
            spdGetAttrValInt(ident_array.id[i], "sort")) continue;

        /* identifier°programID롥*/
        rel = spdGetARel(ident_array.id[i], "decl_ident",
                         "identifier_id");
        /* programIdӡ*/
        if (program_id == spdGetProgramIdByRelId(rel.relationId)) {
            ret_ident_id = ident_array.id[i];
            break;
        }
    }
    spdFreeObjIdArray(ident_array);

    return ret_ident_id;
}  

/******************************************************************************/

/*
 * identId餽identifier줿declaration
 * (ץȥפexternǤϤʤ)ID롥
 */

SpdObjId spdGetDefDeclnIdByIdentId(SpdObjId identId)
{
    int i;
    SpdObjIdArray declr_buf = {0, NULL};
    SpdObjId decln_id;
    SpdObjId ret_id = SPD_NON_ID;

    /* identifierμˤäƽۤʤ롥*/
    switch (spdGetAttrValInt(identId, "sort")) {
      case ID_VARIABLE:
        /* declr(ʣ)->decln(storageexternǤʤ)*/
        declr_buf = spdGetRelObjArray(identId, "decl_ident", "identifier_id",
                                      NULL);
        for (i = 0; i < declr_buf.size; i++) {
            decln_id = spdGetARelObj(declr_buf.id[i], "decl_decl",
                                     "declarator_id");
            if (spdGetAttrValInt(decln_id, "storage") == STORAGE_EXTERN) continue;

            ret_id = decln_id;
        }
        break;
      case ID_FUNCTION:
        /* declr(ʣ¸ߤ뤦sortfuncdefǤ)*/
        declr_buf = spdGetRelObjArray(identId, "decl_ident", "identifier_id",
                                      spdDeclrIsDeclFuncDef);
        /* DECL_FUNCDEF¸ߤʤ硥*/
        if (declr_buf.size == 0) break;
        ret_id = spdGetARelObj(declr_buf.id[0], "decl_decl",
                               "declarator_id");
        break;
      case ID_MEMBER:
        /* !! global¸ߤʤα*/
        break;
      case ID_ENUMERATOR:
        /* !! global¸ߤʤα*/
        break;
      case ID_TYPENAME:
        /* declr->decln(̾󤻤Ƥ?)*/
        ret_id = spdGetARelObj(spdGetARelObj(identId, "decl_ident", 
                                             "identifier_id"),
                               "decl_decl", "declarator_id");
        break;
      case ID_TAG:
        /* declr(ʣ¸ߤ뤦Ϣ decl_opt )*/
        declr_buf = spdGetRelObjArray(identId, "decl_ident", "identifier_id",
                                      NULL);
        for (i = 0; i < declr_buf.size; i++) {
            if (spdGetARelObj(declr_buf.id[i], "decl_opt", "declarator_id") !=
                SPD_NON_ID) {
                ret_id = spdGetARelObj(declr_buf.id[i], "decl_decl",
                                       "declarator_id");
                break;
            }
        }
        break;
      default:
        break;
    }
    if (declr_buf.size != 0) {
        spdFreeObjIdArray(declr_buf);
    }
    return ret_id;
}

/*************************************************************/

/*
 * FuncInfo ν
 */

static void initialFuncInfo(FuncInfo *info)
{
    info->rtnTypeId = SPD_NON_ID;
    info->argNum = 0;
    info->argInfo = NULL;
}

/*
 * FuncInfoǻѤҡΰβ
 */

static void freeFuncInfo(FuncInfo *info)
{
    if (info->argNum > 0) {
        spdFree(info->argInfo);
    }
    info->rtnTypeId = SPD_NON_ID;
    info->argNum = 0;
    info->argInfo = NULL;
}

/*
 * ArgInfoκ
 */

static void mkFuncArgInfo(FuncInfo *info, SpdObjId declnId)
{
    int i;
    SpdRelArray func_parms = {0, NULL};
    SpdObjId declr_id, opt_id;

    declr_id = spdGetARelObj(declnId, "decl_decl", "declaration_id");
    opt_id = spdGetARelObj(declr_id, "decl_opt", "declarator_id");

    func_parms = spdGetRelSortedArray(opt_id, "func_parm", "optionaldecl_id",
                                      spdDeclnIsDeclParm, spdCompareRelOffset);

    if (func_parms.size == 0) return;

    info->argNum = func_parms.size;
    info->argInfo = spdMalloc(sizeof(ArgInfo)*func_parms.size);

    for (i = 0; i < func_parms.size; i++) {
        declr_id = spdGetARelObj(func_parms.rel[i].objectId, "decl_decl",
                                 "declaration_id");
        info->argInfo[i].typeId = spdGetARelObj(declr_id, "decr_type", 
                                                "declarator_id");
        info->argInfo[i].identId = spdGetARelObj(declr_id, "decl_ident", 
                                                 "declarator_id");
    }
    spdFreeRelArray(func_parms);
}

/*
 * FuncInfoκ
 */

static void mkFuncInfo(FuncInfo *info, SpdObjId declnId)
{
    info->rtnTypeId = spdGetARelObj(declnId, "decl_type", "declaration_id");
    mkFuncArgInfo(info, declnId);
}

/*************************************************************/

/*
 * declarationʸӡ
 */

SpdBoolean spdIsSameDecln(SpdObjId declr_a, SpdObjId declr_b)
{
    SpdString str_a;
    SpdString str_b;
    SpdBoolean ret = FALSE;

    if (declr_a == SPD_NON_ID || declr_b == SPD_NON_ID) return FALSE;

    /* !! segmentation faultȯ... */
    /*
     * spdStrˤꡤ.iʸƹ롥
     * segmentation faultɤŪ֡
     */
    spdSetOption("mode", SAPID_I_MODE);
    str_a = spdStrdup(spdStr(declr_a));
    str_b = spdStrdup(spdStr(declr_b));
    spdSetOption("mode", SAPID_C_MODE);

    if (SPD_STREQ(str_a, str_b)) ret = TRUE;
    spdFree(str_a);
    spdFree(str_b);
    return ret;
}

/*
 * ؿѹμĴ, μɽ֤ͤ
 * ϴؿǤ declaration ID
 */

int spdGetFuncModifySort(SpdObjId pre_decln_id, 
                         SpdObjId cur_decln_id)
{
    int i;
    FuncInfo pre_info, cur_info;
    int modify_sort = MODIFY_BODY;
    SpdString pre_typeStr, cur_typeStr;

    initialFuncInfo(&pre_info);
    initialFuncInfo(&cur_info);

    mkFuncInfo(&pre_info, pre_decln_id);
    mkFuncInfo(&cur_info, cur_decln_id);

    /* ͤηΥå*/
    pre_typeStr = spdStrdup(spdStr(pre_info.rtnTypeId));
    cur_typeStr = spdStrdup(spdStr(cur_info.rtnTypeId));
    if (!SPD_STREQ(pre_typeStr, cur_typeStr)) {
        modify_sort = MODIFY_INTERFACE;
    }
    spdFree(pre_typeStr);
    spdFree(cur_typeStr);

    /* οΥå*/
    if (pre_info.argNum != cur_info.argNum) {
        modify_sort = MODIFY_INTERFACE;
    }

    /* ġΰΥå*/
    for (i = 0; i < pre_info.argNum; i++) {
        if (!SPD_STREQ(spdGetName(pre_info.argInfo[i].identId),
                         spdGetName(cur_info.argInfo[i].identId))) {
            modify_sort = MODIFY_INTERFACE;
            break;
        }

        pre_typeStr = spdStrdup(spdStr(pre_info.argInfo[i].typeId));
        cur_typeStr = spdStrdup(spdStr(cur_info.argInfo[i].typeId));
        if (!SPD_STREQ(pre_typeStr, cur_typeStr)) {
            modify_sort = MODIFY_INTERFACE;
        }
        spdFree(pre_typeStr);
        spdFree(cur_typeStr);
    }

    freeFuncInfo(&pre_info);
    freeFuncInfo(&cur_info);
    return modify_sort;
}
