/* 
 * Program:             $RCSfile: mkVersionTbl.c,v $  $Revision: 30.5 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohhashi  2000/01/21
 *                      S.Yamamoto  2000/01/21
 * 
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: mkVersionTbl.c,v 30.5 2003/07/08 06:22:20 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include "spdCmpVersion.h"

/*************************************************************/
  
/*
 * ơ֥Ʊ̾Υ֥Ȥ¸ߤTRUE,ʤFALSE֤
 */

static SpdBoolean isContainElemByName(IdentTbl *tbl, SpdString name)
{
    int i;

    /* 
     * !! ñ㥵ȤƤʤBSEARCHϻȤʤ
     * !! ϥåȤ
     */
    for (i = 0; i < tbl->size; i++) {
        if (SPD_STREQ(name, tbl->elem[i].name)) return TRUE;
    }
    return FALSE;
}

/*
 * ơ֥˼̻ҤϿ롥
 */

static void addElemInVerTable(IdentTbl *tbl, SpdRel rel, 
                              SpdObjIdArray *prog_list, int idx)
{
    int i;
    SpdObjId next_ident_id;

    tbl->elem = spdRealloc(tbl->elem, sizeof(IdentTblElem)*(tbl->size+1));
    tbl->elem[tbl->size].name = spdStrdup(spdGetName(rel.objectId));
    tbl->elem[tbl->size].arraySize = 0;

    /* СʬݤSPD_NON_IDˤ*/
    tbl->elem[tbl->size].idArray = spdMalloc(sizeof(VerIdent)*
                                             prog_list->size);
    tbl->elem[tbl->size].arraySize = prog_list->size;
    for (i = 0; i < prog_list->size; i++) {
        tbl->elem[tbl->size].idArray[i].identId = SPD_NON_ID;
        tbl->elem[tbl->size].idArray[i].programId = SPD_NON_ID;
    }

    tbl->elem[tbl->size].idArray[idx].identId = rel.objectId;
    tbl->elem[tbl->size].idArray[idx].programId = prog_list->id[idx];

    /* ѥСˤƱ켱̻ҤϿ롥*/
    for (i = idx+1; i < prog_list->size; i++) {
        next_ident_id = spdGetIdentIdWithVersion(rel.objectId, prog_list->id[i]);

        if (next_ident_id != SPD_NON_ID) {
            tbl->elem[tbl->size].idArray[i].identId = 
                next_ident_id;
            tbl->elem[tbl->size].idArray[i].programId = 
                prog_list->id[i];
        }
    }
    tbl->size++;
}

/*
 * Сơ֥롥
 */

void mkVersionTable(IdentTbl *tbl, SpdRel rel, 
                    SpdObjIdArray *prog_list, int idx)
{
    /* Τߤ¸ߤʤ̻ҤϽ롥*/
    if (spdGetDefDeclnIdByIdentId(rel.objectId) == SPD_NON_ID) return;

    if (isContainElemByName(tbl, spdGetName(rel.objectId))) return;

    addElemInVerTable(tbl, rel, prog_list, idx);
}

/*************************************************************/

/*
 * ơ֥Ѥΰ
 */

static void freeIdentTable(IdentTbl *tbl)
{
    int i;

    for (i = 0; i < tbl->size; i++) {
        spdFree(tbl->elem[i].name);
        spdFree(tbl->elem[i].idArray);
        tbl->elem[i].name = NULL;
        tbl->elem[i].idArray = NULL;
        tbl->elem[i].idArray = 0;
    }
    spdFree(tbl->elem);
    tbl->elem = NULL;
    tbl->size = 0;
}

/*
 * ٤ƤΥơ֥ΰ롥
 */

void freeAllTables(void)
{
    freeIdentTable(&funcTbl);
    freeIdentTable(&variableTbl);
    freeIdentTable(&typenameTbl);
    freeIdentTable(&tagTbl);
}

/*************************************************************/

#ifdef DEBUG

/*
 * ơ֥ΥפϤ롥
 */

static void dumpTable(IdentTbl *tbl, SpdString title)
{
    int i, j;

    fprintf(stderr, "### %s ###\n\n", title);

    for (i = 0; i < tbl->size; i++) {
        fprintf(stderr, "%s\n", tbl->elem[i].name);

        for (j = 0; j < tbl->elem[i].arraySize; j++) {
            fprintf(stderr, " - %d %d\n", tbl->elem[i].idArray[j].programId,
                    tbl->elem[i].idArray[j].identId);
        }
        fprintf(stderr, "\n");
    }
}

/*
 * ơ֥ǧѡ
 */

void dumpAllTables(void)
{
    dumpTable(&funcTbl, "Function");
    dumpTable(&variableTbl, "Variable");
    dumpTable(&typenameTbl, "Typename");
    dumpTable(&tagTbl, "Tag");
}

#endif

/*************************************************************/

/*
 * Ӵؿ
 */

static int spdCompareIdentName(IdentTblElem *a, IdentTblElem *b)
{
    int     c;

    c = strcmp(a->name, b->name);
    if (c > 0) return(1);
    if (c < 0) return(-1);
    return(0);
}

/*
 * ơ֥򥽡Ȥ롥
 */

void sortAllTables(void)
{
    SPD_QSORT(funcTbl.elem, funcTbl.size, sizeof(IdentTblElem),
                spdCompareIdentName);
    SPD_QSORT(variableTbl.elem, variableTbl.size, sizeof(IdentTblElem),
                spdCompareIdentName);
    SPD_QSORT(typenameTbl.elem, typenameTbl.size, sizeof(IdentTblElem),
                spdCompareIdentName);
    SPD_QSORT(tagTbl.elem, tagTbl.size, sizeof(IdentTblElem),
                spdCompareIdentName);
}
