/* 
 * Program:             $RCSfile: spdCmpVersion.c,v $  $Revision: 30.5 $
 * 
 * Purpose:
 * 
 * Author:              H.Ohhashi  2000/01/21
 *                      S.Yamamoto  2000/01/21
 * 
 * (C) Copyright:       H.Ohhashi and S.Yamamoto  2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdCmpVersion.c,v 30.5 2002/10/21 07:28:35 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#define  GLOBAL_DEFINE      /* Define global variables. */
#include "spdCmpVersion.h"

/*
 * compare function. objectId򾺽˥Ȥ뤿ᡥ
 */

static int spdCompareObjId(const SpdObjId *a, const SpdObjId *b)
{
    if (*a > *b) return 1;
    if (*a < *b) return -1;
    return 0;
}

/*
 * program ID 롥
 */

static SpdObjIdArray spdGetProgList(void)
{
    SpdObjIdArray prog_ids;

    if ((spdTarget("P-model", NULL)) == SPD_NON_ID) {
        spdErrorExit("Program not found.\n");
    }

    /*
     * program ID򾺽˥ȤƼ롥
     * IDŤСΤΤȲꡥ
     */
    prog_ids = spdGetObjIdSortedArray("program", NULL, NULL,
                                      spdCompareObjId);
    return prog_ids;
}

/*
 * Initialization.
 */

static void initialize(char *argv[])
{
    spdInitialize(spdBasename(argv[0]));
}

/*
 * analyze option.
 */

static void analyzeOption(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:D:")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            spdMessage("SDB directory \"%s\".\n", spdGetSDBdir());
            break;
          case 'D':                         /* Directory for HTML. */
            outputDir = optarg;
            break;
          default:
            spdErrorExit("invalid option.\n");
            break;
        }
    }
}

/*
 * globalidentifier롥
 */

static void spdGetGlobalIdent(SpdObjIdArray *prog_ids)
{
    int i, j, k;
    SpdObjIdArray file_ids, decln_ids;
    SpdRel ident_rel;
    SpdRegion region_c;

    for (i = 0; i < prog_ids->size; i++) {
        file_ids = spdGetRelObjArray(prog_ids->id[i], "prog_file", 
                                     "program_id", NULL);
    
        for (j = 0; j < file_ids.size; j++) {
            /* global򤹤٤Ƽ롥*/
            decln_ids = spdGetRelObjArray(file_ids.id[j], "file_decl",
                                          "file_id", NULL);
            for (k = 0; k < decln_ids.size; k++) {
                ident_rel = spdGetARel(spdGetARelObj(decln_ids.id[k], 
                                                     "decl_decl", "declaration_id"),
                                       "decl_ident", "declarator_id");

                region_c = spdGetRegionC(ident_rel);
                /* 
                 * if (smuIsSystemFile(spdGetName(region_c.fileId))) continue;
                 */
                if (smuIsSystemFile(region_c.fileId)) {
                    continue;
                }
                switch (spdGetAttrValInt(ident_rel.objectId, "sort")) {
                  case ID_VARIABLE:
                    mkVersionTable(&variableTbl, ident_rel, prog_ids, i);
                    break;
                  case ID_FUNCTION:
                    mkVersionTable(&funcTbl, ident_rel, prog_ids, i);
                    break;
                  case ID_MEMBER:
                    /*### memberglobalʤΤ¸ߤʤΤǤ ?*/
                    break;
                  case ID_ENUMERATOR:
                    /*### enumeratorglobalʤΤ¸ߤʤΤǤ ?*/
                    break;
                  case ID_TYPENAME:
                    mkVersionTable(&typenameTbl, ident_rel, prog_ids, i);
                    break;
                  case ID_TAG:
                    mkVersionTable(&tagTbl, ident_rel, prog_ids, i);
                    break;
                  default:
                    break;
                }
            }
            spdFreeObjIdArray(decln_ids);
        }
        spdFreeObjIdArray(file_ids);
    }
}

/*************************************************************/

/*
 * Main function.
 */

int main(int argc, char *argv[])
{
    SpdObjIdArray prog_ids;

    initialize(argv);
    analyzeOption(argc, argv);

    prog_ids = spdGetProgList();
    spdGetGlobalIdent(&prog_ids);
    sortAllTables();
    writeAllTables();

#ifdef DEBUG
    dumpAllTables();
#endif

    spdFreeObjIdArray(prog_ids);
    freeAllTables();
    return EXIT_SUCCESS;
}
