/* 
 * Program:             $RCSfile: macro.c,v $  $Revision: 30.10 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto  1996/12/15
 *                      S.Yamamoto  1997/01/16
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1996 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: macro.c,v 30.10 2002/12/16 12:11:51 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h> 
#include <Sapid/SpdStr.h>
#include "common.h"

SpdStringArray include_array = SPD_EMPTY_ARRAY;

SpdRegion makeSpdRegion(SpdObjId file_id, int offset, int length)
{
    SpdRegion region;
    
    region.fileId = file_id;
    region.offset = offset;
    region.length = length;
    return (region);
}

SpdString strCatHeader(SpdObjId hfile_id)
{
    SpdString res = NULL;

    if (spdGetAttrValInt(hfile_id, "sort") == HFILE_ANGLE){
        /* #include <...> */
        res = strCat(res, "#include <");
        res = strCat(res, spdGetName(hfile_id));
        res = strCat(res, ">");
    } else if (spdGetAttrValInt(hfile_id, "sort") == HFILE_DQUOT){
        res = strCat(res, "#include \"");
        res = strCat(res, spdGetName(hfile_id));
        res = strCat(res, "\"");
    }
    return (res);
}

void makeUsrHeader(SpdString file_name)
{
    FILE    *fpin, *fpout;
    char    s[SPD_LINE_LEN + 1];

    if ((fpin = smuOpenSrcFile(smuGetFileIdByName(file_name))) == NULL) {
        spdErrorExit("\"%s\": Can't read file.\n", file_name);
    }
    
    chdir(cwd);
    fpout = spdFopen(file_name, SPD_FOPEN_AT_MODE);

    while (fgets(s, SPD_LINE_LEN, fpin)){
        fputs(s, fpout);
    }
    spdFclose(fpin);

    spdFclose(fpout);
    chdir(dir); 
}

SpdObjId getNeededInclude(SpdObjId file_id)
{
    SpdObjId hfile_id, cfile_id, hfile_id2;
    int i, flag = ON;
    
    hfile_id = spdGetARelObj(file_id, "hfile_file", "pfile_id");
    if (hfile_id == SPD_NON_ID){
        return (SPD_NON_ID);
    } else {
        cfile_id = spdGetCfileIdByPfileId(file_id);
        hfile_id2 = spdGetARelObj(cfile_id, "hfile_file", "pfile_id");
        if (flag_u){
            if (hfile_id2 != SPD_NON_ID){
                return (SPD_NON_ID);
            } else {
                flag = OFF;
                for (i = 0; i < include_array.size; i++){
                    if (strstr(include_array.str[i], spdGetName(hfile_id)) != NULL){
                        flag = ON;
                        break;
                    }
                }
            }
        } else if (hfile_id2 != SPD_NON_ID && 
                   spdGetAttrValInt(hfile_id2, "sort") != HFILE_DQUOT){
            return (SPD_NON_ID);
        } else {
            /* ɸإåեʳ˥󥯥롼 */
            flag = OFF;
            for (i = 0; i < include_array.size; i++){
                if (strstr(include_array.str[i], spdGetName(hfile_id)) != NULL){
                    flag = ON;
                    break;
                }
            }
        }
        if (flag == OFF){
            if (flag_u && spdGetAttrValInt(hfile_id, "sort") == HFILE_DQUOT){
                makeUsrHeader(spdGetName(hfile_id));
            }
            include_array = addStrToArray(include_array, strCatHeader(hfile_id));
        }
    }   
    
    return (file_id);
}

SpdStringArray getNeededMacro(SpdSet *decln_set)
{
    SpdOccArray call_array;
    SpdRelArray decln_array = SPD_EMPTY_ARRAY;
    SpdObjIdArray pblock_array;
    SpdStringArray array = SPD_EMPTY_ARRAY;
    SpdRel file_rel;
    SpdObjId name_id, mpatt_id, def_id, pfile_id, pblock_id, ifile_id;
    SpdRegion decln_reg, macro_reg;
    SpdSet macro_set = spdCreateSet(DEFAULT_INT_SET);
    int i, j, k, l, flag;
    
    /* include_array = SPD_EMPTY_ARRAY; */
    
    decln_array = getSortedRelArray(decln_set);
    
    for(i = 0; i < decln_array.size; i++){
        ifile_id = spdGetARelObj(decln_array.rel[i].objectId, "file_decl", 
                                 "declaration_id");
        pblock_id = spdGetABlockId(ifile_id, 
                                   spdGetOffset(decln_array.rel[i].relationId));
        pfile_id = spdGetARelObj(pblock_id, "file_block", "pblock_id");
        if (getNeededInclude(pfile_id) != SPD_NON_ID){
            spdDelElem(*decln_set, decln_array.rel[i].objectId);
        } else {
            decln_reg = 
                makeSpdRegion(ifile_id, spdGetOffset(decln_array.rel[i].relationId), 
                              spdGetLength(decln_array.rel[i].objectId));
            pblock_array = spdGetBlockIdArrayOfMcall(decln_reg);
            for (j = 0; j < pblock_array.size; j++){
                call_array = spdGetIncludedOccArray(pblock_array.id[j], 
                                                    "macroCall", 
                                                    SPD_NON_ID, NULL);
                for (k = 0; k < call_array.size; k++){
                    name_id = spdGetARelObj(call_array.occ[k].objectId, 
                                            "mcall_mname", "macroCall_id");
                    if (spdAddElem(macro_set, name_id, NULL) != NULL){
                        mpatt_id = spdGetARelObj(name_id, "mpatt_mname", 
                                                 "macroName_id");
                        def_id = spdGetARelObj(mpatt_id, "def_mpatt", 
                                               "macroPattern_id");
                        file_rel = spdGetARel(def_id, "file_block", "pblock_id");
                        macro_reg = makeSpdRegion(file_rel.objectId, 
                                                  spdGetOffset(file_rel.relationId), 
                                                  spdGetLength(def_id));
                        if (getNeededInclude(file_rel.objectId) == SPD_NON_ID){
                            flag = OFF;
                            for (l = 0; l < array.size; l++){
                                if (!strcmp(array.str[l], 
                                            spdGetTextByRegion(macro_reg))){
                                    flag = ON;
                                    break;
                                }
                            }
                            if (flag == OFF){
                                array = 
                                    addStrToArray(array, 
                                                  spdStrdup(spdGetTextByRegion(macro_reg)));
                            }
                        }
                    }
                }
            }
        }   
    }
    
    spdFreeRelArray(decln_array);
    spdFreeSet(macro_set);
    
    return (array);
}
