/* 
 * Program:             $RCSfile: slicer_file.c,v $  $Revision: 30.7 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto  1996/12/15
 *                      S.Yamamoto  1997/01/16
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1996 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: slicer_file.c,v 30.7 2002/10/27 11:41:50 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h> 
#include <Sapid/SpdStr.h>
#include "common.h"
#include "spdSlicer4.h"

SpdStringArray func_array = SPD_EMPTY_ARRAY;
SpdStringArray init_array = SPD_EMPTY_ARRAY;
SpdStringArray decl_array = SPD_EMPTY_ARRAY;

void classifyDeclForSlicer(SpdRelArray decln_array)
{
    SpdObjIdArray declr_array;
    SpdObjId target_id, opt_id, decln_id;
    SpdString type, decl, opt, declr;
    SpdSet set = spdCreateSet(DEFAULT_INT_SET);
    int i, j;
    char *p;
    
    spdSetOption("mode", SAPID_C_MODE);
    
    for (i = 0; i < decln_array.size; i++){
        target_id = decln_array.rel[i].objectId;
        if (spdGetAttrValInt(target_id, "sort") == DECL_FUNCDEF){
            func_array = addStrToArray(func_array, spdStrdup(getText(target_id))); 
        }else if (checkStor(target_id, STORAGE_STATIC)){
            /* static ξ */
            init_array = addStrToArray(init_array, spdStrdup(getText(target_id))); 
        }else{
            declr_array = spdGetRelObjArray(target_id, "decl_decl", 
                                            "declaration_id", NULL);
            if (isOptionalDecl(declr_array)){
                for (j = 0; j < declr_array.size; j++){
                    if ((opt_id = 
                         spdGetARelObj(declr_array.id[j], "decl_opt", 
                                       "declarator_id")) != SPD_NON_ID && 
                        spdGetAttrValInt(opt_id, "sort") == OPT_INITIALIZER){
                        /* Ƥ */
                        if ((decln_id = getIdOfStatic(declr_array.id[j]))
                            != SPD_NON_ID){
                            /* static ξ */
                            if (spdAddElem(set, target_id, NULL) != NULL){
                                init_array = 
                                    addStrToArray(init_array, 
                                                  spdStrdup(getText(target_id))); 
                            }
                            if (spdAddElem(set, decln_id, NULL) != NULL){
                                init_array = 
                                    addStrToArray(init_array, 
                                                  spdStrdup(getText(decln_id))); 
                            }
                        } else if ((decln_id = getIdOfExtern(declr_array.id[j])) 
                                   !=  SPD_NON_ID){
                            /* externξ */
                            if (spdAddElem(set, decln_id, NULL) != NULL){
                                decl_array = 
                                    addStrToArray(decl_array, 
                                                  spdStrdup(getText(decln_id))); 
                            }
                            if (spdAddElem(set, target_id, NULL) != NULL){
                                init_array = 
                                    addStrToArray(init_array, 
                                                  spdStrdup(getText(target_id))); 
                            }
                        } else {
                            if (spdAddElem(set, target_id, NULL) != NULL){
                                init_array = 
                                    addStrToArray(init_array, 
                                                  spdStrdup(getText(target_id))); 
                            }
                            type = 
                                spdStrdup(spdStr(spdGetARelObj(target_id, 
                                                               "decl_type", 
                                                               "declaration_id")));
                            opt = spdStrdup(getText(opt_id));
                            declr = spdStrdup(getText(declr_array.id[j]));
                
                            p = strstr(declr, opt);
                            if (*(p-1) == ' '){
                                *(p-1) = ';';
                                *p = '\0';
                            }else {
                                *p = ';';
                                *(p+1) = '\0';
                            }
                            decl = NULL;
                            decl = strCat(decl, "extern ");
                            decl = strCat(decl, type);
                            decl = strCat(decl, declr);
                            decl_array = addStrToArray(decl_array, 
                                                       spdStrdup(decl)); 
                            spdFree(type);
                            spdFree(opt);
                            spdFree(declr);
                            spdFree(decl);
                        }
                    } else if (spdAddElem(set, target_id, NULL) != NULL){
                        decl_array = 
                            addStrToArray(decl_array, 
                                          spdStrdup(getText(target_id))); 
                    }
                }
            }else if (spdAddElem(set, target_id, NULL) != NULL){
                decl_array = addStrToArray(decl_array, 
                                           spdStrdup(getText(target_id))); 
            }
            spdFreeObjIdArray(declr_array);
        }   
    }
    spdFreeSet(set);
}

void makeSlicedSourceFile(SpdString file, SpdString include_file, 
                          SpdStringArray init_array, SpdStringArray func_array)
{
    FILE *fp, *fp_inc;
    int i;

    if (flag_d){
        chdir(cwd);
    }

    fp = spdFopen(file, SPD_FOPEN_WT_MODE);
    if ((fp_inc = smuOpenSrcFile(smuGetFileIdByName(include_file))) != NULL) {
        fprintf(fp, "#include \"%s\"\n", include_file);
        spdFclose(fp_inc);
    }
    for (i = 0; i < init_array.size; i++){
        fprintf(fp, "%s\n", init_array.str[i]);
    }
    for (i = 0; i < func_array.size; i++){
        fprintf(fp, "%s\n", func_array.str[i]);
    }
    spdFclose(fp);

    if (flag_d){
        chdir(dir);
    } 
    return;
}

void makeSlicedFile(SpdRelArray decln_array, SpdStringArray macro_array, 
                    SpdStringArray include_array, SpdString func_name)
{
    SpdString include_file = NULL, 
    source_file = NULL;
    
    classifyDeclForSlicer(decln_array);
    
    include_file = strCat(include_file, func_name);
    include_file = strCat(include_file, ".h");
    
    source_file = strCat(source_file, func_name);
    source_file = strCat(source_file, ".c");
    
    makeIncludeFile(include_file, include_array, macro_array, decl_array);
    makeSlicedSourceFile(source_file, include_file, init_array, func_array); 
   
    spdFree(include_file);
    spdFree(source_file);
    freeStringArray(include_array);
    freeStringArray(macro_array);
    freeStringArray(decl_array);
    freeStringArray(func_array);
    freeStringArray(init_array);
}
