/* 
 * Program:             $RCSfile: util.c,v $  $Revision: 30.6 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto  1996/12/15
 *                      S.Yamamoto  1997/01/16
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1996 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: util.c,v 30.6 2003/06/28 15:33:58 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h> 
#include "common.h"

SpdObjId getIdOfStatic(SpdObjId declr_id)
{
    SpdObjIdArray declr_array;
    SpdObjId ident_id, decln_id;
    int i;

    ident_id = spdGetARelObj(declr_id, "decl_ident", "declarator_id");
    declr_array = spdGetRelObjArray(ident_id, "decl_ident", "identifier_id", NULL);
    for (i = 0; i < declr_array.size; i++) {
        decln_id = spdGetARelObj(declr_array.id[i], "decl_decl", "declarator_id");
        if (checkStor(decln_id, STORAGE_STATIC)) {
            return decln_id;
        }
    }
    spdFreeObjIdArray(declr_array);
    return SPD_NON_ID;
}       

/* program  objectId  file  objectId ν */
void getFileSet(SpdObjId prog_id, SpdSet *file_set)
{
    SpdObjIdArray file_buf;
    int i;

    file_buf = spdGetRelObjArray(prog_id, "prog_file", "program_id", NULL);
    for (i = 0; i < file_buf.size; i++) {
        spdAddElem(*file_set, file_buf.id[i], NULL);
    }
    spdFreeObjIdArray(file_buf);
    return;
}

#if 0
/* sort  ID_FUNCTION Ǥ뤫ɤĴ٤ */
SpdBoolean spdOccIsIdFunction(SpdOcc occ)
{
    return (spdGetAttrValInt(occ.objectId, "sort") == ID_FUNCTION);
}
#endif

/* ؿ̾ؿ objectId  */
void getFuncIdSet(SpdSet *file_set, SpdStringArray func_array, SpdSet *decln_set)
{
    SpdOccArray ident_array;
    SpdObjIdArray declr_array;
    SpdObjId file_id, decln_id, ident_id;
    SpdSet set = spdCreateSet(DEFAULT_INT_SET);
    int i, j, k, l, flag = OFF;
    
    for(i = 0; i < (*file_set)->elemNum; i++) {
        spdGetAElem(*file_set, &file_id);
        ident_array = spdGetIncludedOccArray(file_id, "identifier", 
                                             file_id, spdOccIsIdFunction);
        for (j = 0; j < ident_array.size; j++) {
            for (k = 0; k < func_array.size; k++) {
                ident_id = ident_array.occ[j].objectId;
                if (SPD_STREQ(spdGetName(ident_id), func_array.str[k])) {
                    spdAddElem(set, k, NULL);
                    declr_array = spdGetRelObjArray(ident_id, "decl_ident", 
                                                    "identifier_id", NULL);
                    for (l = 0; l < declr_array.size; l++) {
                        decln_id = spdGetARelObj(declr_array.id[l], "decl_decl", 
                                                 "declarator_id");
                        spdAddElem(*decln_set, decln_id, NULL);
                    }
                    spdFreeObjIdArray(declr_array);
                }
            }
        }
        spdFreeOccArray(ident_array);
    }
    
    for (i = 0; i < func_array.size; i++) {
        if (!spdSetContains(set, i)) {
            spdError("Function \"%s\" not found.\n", func_array.str[i]);
            flag = ON;
        }
    }
    
    if (flag) {
        exit(1);
    }
    
    spdFreeSet(set);
    return;
}

/* ʸ˳Ǽ */
SpdStringArray addStrToArray(SpdStringArray array, SpdString string)
{
    array.str = spdRealloc(array.str, 
                           (sizeof(SpdString) * (array.size + 1)));
    array.str[array.size++] = string;
    
    return (array);
}

/* ʸΰ */
void freeStringArray(SpdStringArray array)
{
    int i;
    
    for (i = 0; i < array.size; i++) {
        spdFree(array.str[i]);
    }
    spdFree(array.str);
}

/* ꤷid°storageꤷstorgeɤĴ٤ */
SpdBoolean checkStor(SpdObjId id, int storage)
{
    if (spdGetAttrValInt(id, "storage") == storage) {
        return TRUE;
    }
    return FALSE;
}

/* spdStr.c  */
SpdString strCat(SpdString a, SpdString b)
{
    SpdString res = NULL;

    if (b != NULL) {
        if (a == NULL) {
            res = spdMalloc(strlen(b)+1);
            strcpy(res, b);
        } else {
            res = spdRealloc(a, strlen(a)+strlen(b)+1);
            strcat(res, b);
        }
    } else {
        res = a;
    }

    return res;
}

SpdRelArray getSortedRelArray(SpdSet *decln_set)
{
    SpdRelArray decln_array = SPD_EMPTY_ARRAY;
    SpdRel rel;
    SpdObjId decln_id;
    int i;

    for (i = 0; i < (*decln_set)->elemNum; i++) {
        spdGetAElem(*decln_set, &decln_id);
        rel = spdGetARel(decln_id, "file_decl", "declaration_id");
        decln_array.rel = spdRealloc(decln_array.rel, 
                                     sizeof(SpdRel) * (decln_array.size + 1));
        decln_array.rel[decln_array.size].objectId = decln_id;
        decln_array.rel[decln_array.size].relationId = rel.relationId;  
        decln_array.size++;
    }

    SPD_QSORT(decln_array.rel, decln_array.size, 
                sizeof(SpdRel), spdCompareRelOffset) ;
    
    return (decln_array);
}
