/* 
 * Program:             $RCSfile: spdSlicerView4.c,v $  $Revision: 30.2 $
 * 
 * Purpose:
 * 
 * Author:              N.Iwamoto  1996/12/15
 *                      S.Yamamoto  1997/01/16
 * 
 * (C) Copyright:       N.Iwamoto and S.Yamamoto  1996 - 1997
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdSlicerView4.c,v 30.2 2002/10/27 11:41:54 yamamoto Exp yamamoto $";
#endif

#include <stdio.h> 
#include <tcl.h>
#include <tk.h>
#include <Sapid/Sapid.h>
#include <Sapid/tclAR.h>
#include "common.h" 
#include "spdSlicer4.h" 

char *dir;
char *cwd;

int flag_t = OFF;
int flag_u = OFF;
int flag_d = OFF;

/* void globalArrayInit(void); */

void globalArrayInit(void)
{
    include_array.str = NULL;
    include_array.size = 0;
    decl_array.str = NULL;
    decl_array.size = 0;
    func_array.str = NULL;
    func_array.size = 0;
    init_array.str = NULL;
    init_array.size = 0;

    return;
}    

int tclSlicer(ClientData clientdata, Tcl_Interp *interp, int argc, 
              char *argv[])
{
    SpdObjId prog_id;
    SpdStringArray funcname_array = SPD_EMPTY_ARRAY;
    SpdSet file_set=spdCreateSet(DEFAULT_INT_SET), 
    decln_set=spdCreateSet(DEFAULT_INT_SET);
    SpdStringArray macro_array = SPD_EMPTY_ARRAY;
    SpdRelArray decln_array = SPD_EMPTY_ARRAY;
    int i;


    globalArrayInit();

    prog_id = atoi(argv[1]);
    funcname_array = addStrToArray(funcname_array, argv[2]);
    getFileSet(prog_id, &file_set);

    getFuncIdSet(&file_set, funcname_array, &decln_set);

    getNeededDeclSet(&decln_set);

    macro_array = getNeededMacro(&decln_set); 

    decln_array = getSortedRelArray(&decln_set);

    spdFreeSet(decln_set);
    spdFreeSet(file_set);
    
    classifyDeclForSlicer(decln_array);

    for (i = 0; i < include_array.size; i++){
        if (Tcl_SetVar(interp, "include", include_array.str[i],
                       TCL_LIST_ELEMENT | TCL_APPEND_VALUE | 
                       TCL_LEAVE_ERR_MSG) == NULL){
            return TCL_ERROR;
        }
    }
    
    for (i = 0; i < macro_array.size; i++){
        if (Tcl_SetVar(interp, "macro", macro_array.str[i],
                       TCL_LIST_ELEMENT | TCL_APPEND_VALUE | 
                       TCL_LEAVE_ERR_MSG) == NULL){
            return TCL_ERROR;
        }
    }

    for (i = 0; i < decl_array.size; i++){
        if (Tcl_SetVar(interp, "decl", decl_array.str[i],
                       TCL_LIST_ELEMENT | TCL_APPEND_VALUE | 
                       TCL_LEAVE_ERR_MSG) == NULL){
            return TCL_ERROR;
        }
    }

    for (i = 0; i < init_array.size; i++){
        if (Tcl_SetVar(interp, "init", init_array.str[i],
                       TCL_LIST_ELEMENT | TCL_APPEND_VALUE | 
                       TCL_LEAVE_ERR_MSG) == NULL){
            return TCL_ERROR;
        }
    }

    for (i = 0; i < func_array.size; i++){
        if (Tcl_SetVar(interp, "func", func_array.str[i],
                       TCL_LIST_ELEMENT | TCL_APPEND_VALUE | 
                       TCL_LEAVE_ERR_MSG) == NULL){
            return TCL_ERROR;
        }
    }

    freeStringArray(include_array);
    freeStringArray(macro_array);
    freeStringArray(decl_array);
    freeStringArray(func_array);
    freeStringArray(init_array);

    return TCL_OK;
}

int main(int argc, char *argv[])
{
    Tcl_Interp  *interp;
    static char tclsrc[FILENAME_MAX + 1] = "/usr/local/Sapid/lib";

#if (TK_MAJOR_VERSION < 4) || (TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION < 2)
    Tk_Window   mainWin;
#endif

    /* Tcl ե̾δ */
    {
        char *libpath = getenv("SAPID_LIBDIR");
        if (libpath != NULL) {
            strncpy(tclsrc, libpath, sizeof(tclsrc));
        }
        
        strncat(tclsrc, "/spdSlicerView4.tcl", sizeof(tclsrc));
    }

    /* Tcl Interpreter  */
    interp = Tcl_CreateInterp();

#if (TK_MAJOR_VERSION < 4) || (TK_MAJOR_VERSION == 4) && (TK_MINOR_VERSION < 2)
    mainWin = Tk_CreateMainWindow(interp, NULL, tclfile, tclfile);
    if (mainWin == NULL) {
        spdErrorExit("%s\n", interp->result);
    }
#endif

    if (Tcl_AppInit(interp) == TCL_ERROR)
        spdErrorExit("%s\n", interp->result);

    /* TclƤФ륳ޥɤ */
    Tcl_CreateCommand(interp, "spdSlicer", tclSlicer,
                      (ClientData)"New", (Tcl_CmdDeleteProc *)NULL);
    
    /* ᥤ */
    if (Tcl_EvalFile(interp, tclsrc) == TCL_ERROR) {
        spdErrorExit("%s\n", interp->result);
    }

    Tk_MainLoop();

    /* λ */
    Tcl_Eval(interp, "exit");
    return (EXIT_SUCCESS);
}

int Tcl_AppInit(Tcl_Interp *interp)
{
    if (Tcl_Init(interp) == TCL_ERROR) {
        fprintf(stderr, "%s\n", interp->result);
        return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR) {
        fprintf(stderr, "%s\n", interp->result);
        return TCL_ERROR;
    }

    Sapidtclar_Init(interp);
    return (TCL_OK);
}
