/* 
 * Program:             $RCSfile: spdDump_get.c,v $  $Revision: 30.8 $
 * 
 * Purpose:
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * Modified:            R.Miura  1997/06/22
 * 
 * (C) Copyright:       K.Hayashi, R.Miura and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdDump_get.c,v 30.8 2002/10/13 09:32:41 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>

int spdCompareRelOffsetAndName(const SpdRel *a, const SpdRel *b)
{
    int         offset1, offset2;
    SpdObjId    objId1, objId2;
    
    offset1 = spdGetOffset((*a).relationId);
    objId1 = (*a).objectId;
    offset2 = spdGetOffset((*b).relationId);
    objId2 = (*b).objectId;

    if (offset1 == -1 && offset2 != -1) {
        return (1);
    } else if (offset1 != -1 && offset2 == -1) {
        return (-1);
    } else if (spdGetClassId(objId1) == SPD_CID_FILE) {
        /* 
         * if (spdCompareObjIdName(&objId1, &objId2) != 0) {
         *     return (spdCompareObjIdName(&objId1, &objId2));
         * } else {
         *     return (offset1 - offset2);
         * }
         */
        return (spdCompareObjIdName(&objId1, &objId2) != 0 ?
                spdCompareObjIdName(&objId1, &objId2) :
                offset1 - offset2);
    } else {
        return (offset1 - offset2); 
    }
}

int spdCompareObjOffsetAndName(const SpdObjId *a, const SpdObjId *b)
{
    SpdRelArray     relBuf;
    SpdRel          rel1, rel2;
 
    relBuf = spdGetRelSortedArray(*a, "any", "c-id", 
                                  NULL, spdCompareRelOffsetAndName);
    rel1 = relBuf.rel[0];
    spdFreeRelArray(relBuf);

    relBuf = spdGetRelSortedArray(*b, "any", "c-id", 
                                  NULL, spdCompareRelOffsetAndName);
    rel2 = relBuf.rel[0];    
    spdFreeRelArray(relBuf);

    return (spdCompareRelOffsetAndName(&rel1, &rel2));
}

int spdCompareObjAttr(const SpdObjId *a, const SpdObjId *b)
{
    int         i, argn;
    int         iattr1, iattr2;
    SpdString   sattr1, sattr2;
    SpdClassId  classId = spdGetClassId(*a);

    for (i = 1; (argn = class_table[classId].seq[i]) >= 0; i++) {
        if (SPD_IS_SVAL(argn)) {                   /* String */
            sattr1 = spdGetAttrValString(*a, class_table[classId].
                                         sattr[argn - IVAL_NUM]);
            sattr2 = spdGetAttrValString(*b, class_table[classId].
                                         sattr[argn - IVAL_NUM]);
            if (!(spdSafetyStrcmp(sattr1, sattr2) == 0))
                return (spdSafetyStrcmp(sattr1, sattr2));
        } else {                            /* Integer */
            iattr1 = spdGetAttrValInt(*a, class_table[classId].
                                      iattr[argn]);
            iattr2 = spdGetAttrValInt(*b, class_table[classId].
                                      iattr[argn]);
            if (iattr1 != iattr2) {
                return (iattr1 - iattr2);
            }
        }
    }
    return (0);
}

/* Get id of file. */

int getFileId(int relationId)
{
    SpdObjId        objectId, fileId;
    SpdRelArray     rel_buf;
    int             i;
    
    objectId = spdGetAttrValIntOfRel(relationId, "p-id");
    if (spdGetClassId(objectId) == SPD_CID_FILE) {
        return (spdGetAttrValIntOfRel(relationId, "p_file_id"));
        /* ʽʤƤ⤤褦ʵ뤬 */
    }

    /* spdSetSearchFlag(SPD_SEARCH_ANY); */

    rel_buf = spdGetRelSortedArray(objectId, "any", "c-id", 
                                   NULL, spdCompareRelOffsetAndName);
    for (i = 0; i < rel_buf.size; i++) {
#if 0
        /* äƤ뤫ʬʤid˼Ԥ뤳Ȥ롥 */
        /* 
         * if (class_table[spdGetClassId(rel_buf.rel[i].relationId)].typ 
         *     & C_TYP_CMP) { 
         *     fileId = getFileId(rel_buf.rel[i].relationId);
         *     if (fileId != -1) 
         *         break;
         * }
         */
#endif

        fileId = getFileId(rel_buf.rel[i].relationId);
        if (fileId != -1)  {
            break;
        }
    }
    spdFreeRelArray(rel_buf);
    
    return (fileId);
}
