/* 
 * Program:             $RCSfile: spdDump_read.c,v $  $Revision: 30.9 $
 * 
 * Purpose:
 * 
 * Author:              K.Hayashi  1995/02/20
 * Modified:            S.Yamamoto  1995/03/05
 * Modified:            R.Miura  1997/06/22
 * 
 * (C) Copyright:       K.Hayashi, R.Miura and S.Yamamoto  1995 - 2000
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdDump_read.c,v 30.9 2003/06/28 15:34:03 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>

extern int getFileId(int); 
extern int spdCompareRelOffsetAndName(const SpdRel *a, const SpdRel *b);
extern int spdCompareObjOffsetAndName(const SpdObjId *a, const SpdObjId *b);
extern int spdCompareObjAttr(const SpdObjId *a, const SpdObjId *b);

static int print_id = 0;
static int print_txt = 0;
static int print_relnum = 0;
static int no_trans = 0;
static int print_tree = 0;
static int print_sep = 0;
static int show_help = 0;
static int sort_type = 0;   /* 0: Attr  1: Offset ͥ */

static void printObjAttr(SpdClassId classId, SpdObjId objId)
{
    int i, argn, iattr;
    SpdString sattr;
    SpdString retAttr, retReadable; 

    for (i = 1; (argn = class_table[classId].seq[i]) >= 0; i++) {
        if (SPD_IS_SVAL(argn)) {            /* String */
            sattr = spdGetAttrValString(objId, class_table[classId].
                                        sattr[argn - IVAL_NUM]);

            if (SPD_STREQ(sattr, "")) {
                printf("\t(null)");
            } else {
                printf("\t%s", sattr);
            }

        } else {                            /* Integer */
            iattr = spdGetAttrValInt(objId, class_table[classId].iattr[argn]);
            if (no_trans) {
                printf("\t%d", iattr);
            } else {
                if (smuGetClassAttr(classId, 
                                    class_table[classId].iattr[argn], 
                                    iattr, &retAttr, &retReadable)) {
                    printf("\t%s", retAttr);
                } else {
                    printf("\t%d", iattr);
                }
            }
        }
    }
}

/* ƤؤδϢ򤿤ɤä Tree ɽ(Ƶ) */
static void printTree(SpdRel pRel, int level)
{
    int i, j;
    SpdRelArray relBuf;

    putchar('\t');

    for (i = 0; i < level; i++) {
        printf("  ");
    }
    
    printf("--%s (%s:%d)-->", 
           spdGetClassName(pRel.relationId),
           spdGetName(getFileId(pRel.relationId)),
           /* spdGetOffset(pRel.relationId)); offsetȼѹ */
           spdGetAbsoluteOffset(pRel.relationId, 
                                getFileId(pRel.relationId)));

    if (print_id) 
        printf(" %d", pRel.objectId);

    if (SPD_STREQ(spdGetClassName(pRel.objectId), "file")) {
        /* Ƶ */
        printf(" file(%s) \n", spdGetName(pRel.objectId));
        return;
    }
    putchar('\n');
    
    /* Ƶ */
    relBuf = spdGetRelSortedArray(pRel.objectId, "any", "c-id", NULL, 
                                  spdCompareRelOffsetAndName);
    
    for(j = 0; j < relBuf.size; j++) {
        printTree(relBuf.rel[j], level + 1);
    }
    spdFreeRelArray(relBuf);

    if (level == 0) {
        printf("\n");
    }
}

/* ֥Ȥɽ */
static void printObjClass(SpdClassId classId, SpdObjId objId)
{
    int j;
    char *string;
    SpdRelArray pRelBuf, cRelBuf;
    SpdRelId relId;
    SpdObjId fileId;

    pRelBuf = spdGetRelSortedArray(objId, "any", "c-id", 
                                   NULL, spdCompareRelOffsetAndName);

    cRelBuf = spdGetRelSortedArray(objId, "any", "p-id", 
                                   NULL, spdCompareRelOffset);

    /* ƤؤδϢҤȤĤʤ */
    if (pRelBuf.size == 0) {

        if (print_id) 
            printf("%d", objId);

        printf("\tNO_PARENT");

        if (print_txt) {
            printf("\t%.10s",
                   ((string = spdGetObjText(objId)) == NULL ? "(NULL)" : string));
        }

        printObjAttr(classId, objId); 

        if (print_relnum) 
            printf("\tP%d C%d", pRelBuf.size, cRelBuf.size);

        printf("\n");
        return;
    }

    /* ƤؤδϢ */
    for (j = 0; j < pRelBuf.size; j ++) {

        relId = pRelBuf.rel[j].relationId;

        if (print_id)
            printf("%d", objId);

        fileId = getFileId(relId);
        printf("\t[%s:%d] %s", 
               ((string = spdGetName(fileId)) == NULL ? "(NULL)" : string),
               /* spdGetOffset(relId), offsetȼѹ */
               spdGetAbsoluteOffset(relId, fileId),
               spdGetClassName(relId));

        if (print_txt) {
            printf("\t%.10s", spdGetObjText(objId));
        }

        printObjAttr(classId, objId); 

        if (print_relnum) 
            printf("\tP%d C%d", pRelBuf.size, cRelBuf.size);

        printf("\n");

        if (print_tree)
            printTree(pRelBuf.rel[j], 0);
    }
    spdFreeRelArray(pRelBuf);
    spdFreeRelArray(cRelBuf);
}

static int spdCompareObj(const SpdObjId *a, const SpdObjId *b)
{
    if (sort_type) {
        /* Offset ͥ */
        if (spdCompareObjOffsetAndName(a, b) != 0)
            return (spdCompareObjOffsetAndName(a, b));
        else 
            return (spdCompareObjAttr(a, b));
    } else {
        /* Attr ͥ */
        if (spdCompareObjAttr(a, b) != 0)
            return (spdCompareObjAttr(a, b));
        else 
            return (spdCompareObjOffsetAndName(a, b));
    }   
}

static void readObjClass(SpdClassId classId)
{
    SpdObjIdArray objIdBuf;
    int i;
    
    /* 
     * objIdBuf = spdGetObjIdOfModelSortedArray(spdGetClassNameById(classId),
     *                                          NULL, "I", NULL,
     *                                          spdCompareObj);
     */
    objIdBuf = spdGetObjIdSortedArray(spdGetClassNameById(classId), NULL, 
                                      NULL, spdCompareObj);

    for (i = 0; i < objIdBuf.size; i ++) {
        printObjClass(classId, objIdBuf.id[i]);
        if (print_sep)
            printf("\t...............................\n");
    }
    spdFreeObjIdArray(objIdBuf);
}

int main(int argc, char *argv[])
{
    char *target_program_name = NULL;
    char *classname = NULL;
    SpdClassId classId;
    SpdInfo *info;
    int c;

    spdInitialize(spdBasename(argv[0]));

    while ((c = getopt(argc, argv, "d:ixrntshO")) != EOF) {
        switch (c) {
          case 'd':             /* Directory for SDB. */
            spdSetSDBdir(optarg);
            spdMessage("SDB directory \"%s\".\n", spdGetSDBdir());
            break;
          case 'i':
            print_id = 1;
            break;
          case 'x':
            print_txt = 1;
            break;
          case 'n':
            no_trans = 1;
            break;
          case 't':
            print_tree = 1;
            break;
          case 's':
            print_sep = 1;  
            break;
          case 'r' :
            print_relnum = 1;
            break;
          case 'h' :
            show_help = 1;
            break;
          case 'O' :
            sort_type = 1;
            break;
          default:
            show_help = 1;
            break;
        }
    }

    if (argc != optind + 1) { 
        show_help = 1;
    }

    /* إפɽ */
    if (show_help) {
        fprintf(stderr, "Usage: spdDumpSDB [options] classname   \n");
        fprintf(stderr, "    -i : ֥ID ղ      \n");
        fprintf(stderr, "    -x : spdGetObjText η̤ɽ \n");
        fprintf(stderr, "    -n : ޥ̤Ѵ             \n");
        fprintf(stderr, "    -t : Tree ɽղ            \n");
        fprintf(stderr, "    -s : object ֤Υѥ졼\n");
        fprintf(stderr, "    -r : ƻҤؤδϢɽ       \n");
        fprintf(stderr, "    -h : help ɽ                  \n");
        fprintf(stderr, "    -O : Offset ͥ襽          \n");
        return (EXIT_SUCCESS);
    }

    classname = argv[optind]; 
    target_program_name = smuGetAProgramName();

    if ((classId = spdGetClassIdByName(classname)) == SPD_NON_ID) {
        /* classname  */
        spdErrorExit("Invalid class name \"%s\".\n", classname); 
    }

    if (spdTarget("P-model", target_program_name) == SPD_NON_ID) {
        spdErrorExit("Program \"%s\" not found.\n", target_program_name); 
    }   

    /* Print "SDB/.info". */
    info = spdOpenSDBinfo();
    printf("*** classname=%s ***\n", classname);
    printf("Sapid-%s", spdGetInfoString(info, "SapidVersion"));
    printf(".%s", spdGetInfoString(info, "SapidRevision"));
    printf(".%s\n", spdGetInfoString(info, "SapidPatchLevel")); 
    printf("SapidSDBVersion=%s\n", spdGetInfoString(info, "SapidSDBVersion"));
    printf("SapidPIDBVersion=%s\n", spdGetInfoString(info, "SapidPIDBVersion"));
    printf("SDB4=%s\n", spdGetInfoString(info, "SDB4"));
    printf("RELATIVE_OFFSET=%s\n", spdGetInfoString(info, "RELATIVE_OFFSET"));
    spdCloseInfo(info);

    if (sort_type) {
        printf("Offset priority sort\n");
    } else {
        printf("Attribute priority sort\n");
    }
    
    readObjClass(classId);

    return (EXIT_SUCCESS);
}
