/* 
 * Program:             $RCSfile: spdDumpTable.c,v $  $Revision: 30.18 $
 * 
 * Purpose:
 * 
 * Author:              S.Yamamoto  1996/12/28
 * 
 * (C) Copyright:       S.Yamamoto  1996 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdDumpTable.c,v 30.18 2003/05/29 04:21:31 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static void parse_option(int argc, char *argv[]);
static void dump_obj_table(void);
static void dump_rel_table(void);

int main(int argc, char *argv[])
{
    SpdConstString  program_name;              /* This program name. */
    SpdString       target_program_name;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parse_option(argc, argv);

    target_program_name = smuGetAProgramName();
    if (spdTarget("P-model", target_program_name) == SPD_NON_ID) {
        spdErrorExit("Program \"%s\" not found.\n", target_program_name);
    }
    dump_obj_table();
    dump_rel_table();
    return (EXIT_SUCCESS);
}

static void dump_obj_table(void)
{
    int         i, j;
    SpdClassId  classId = SPD_NON_CID;
    SpdAttrId   aid, maxAid;
    
    /* printf("# Object table:\n"); */
    for (i = 0; i < object_db_n; i++) {
        if (spdGetClassId(object_db[i].Obj_id) != classId) {
            classId = spdGetClassId(object_db[i].Obj_id);

            /* Output class name. */
            printf("# %s/%s\n", spdGetModelNameById(spdGetModelId(classId)),
                   spdGetClassName(object_db[i].Obj_id));

            /* Output attribute names. */
            printf("# ");
            maxAid = spdGetLastAttrIdInt(classId);
            for (aid = spdGetFirstAttrIdInt(classId); aid <= maxAid; aid++) {
                printf("%s\t", spdGetAttrName(classId, aid));
            }
            maxAid = spdGetLastAttrIdStr(classId);
            for (aid = spdGetFirstAttrIdStr(classId); aid <= maxAid; aid++) {
                printf("%s\t", spdGetAttrName(classId, aid));
            }
            printf("\n");
        }

        for (j = 0; j < SpdObjIValN; j++) { /* Integer attribute. */
            if (class_table[classId].iattr[j] == NULL) {
                break;
            }
            printf("%d\t", object_db[i].ivalue[j]);
        }
        for (j = 0; j < SpdObjSValN; j++) { /* String attribute. */
            if (class_table[classId].sattr[j] == NULL) {
                break;
            }
            printf("\"%s\"\t",
                   (object_db[i].svalue[j] ? object_db[i].svalue[j] : ""));
        }
        printf("\n");
    }
}

static void dump_rel_table(void)
{
    int         i, j;
    SpdClassId  relId = SPD_NON_CID;
    
    /* printf("# Relation table:\n"); */
    for (i = 0; i < relation_db_n; i++) {
        if (spdGetClassId(relation_db[i].Rel_id) != relId) {
            relId = spdGetClassId(relation_db[i].Rel_id);

            /* Output relationship name. */
            printf("# %s/%s\n", spdGetModelNameById(spdGetModelId(relId)),
                    spdGetClassName(relation_db[i].Rel_id));
        }

        for (j = 0; j < SpdRelIValN; j++) { /* Integer attribute. */
            if (class_table[relId].iattr[j] == NULL) {
                break;
            }
            printf("%d\t", relation_db[i].ivalue[j]);
        }
        printf("\n");
    }
}

static void parse_option(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            spdMessage("SDB directory \"%s\".\n", spdGetSDBdir());
            break;
          default:
            spdErrorExit("'%c': Invalid option.\n", ch);
            break;
        }
    }
}
