/* 
 * Program:             $RCSfile: spdDumpTable4.c,v $  $Revision: 30.16 $
 * 
 * Purpose:
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdDumpTable4.c,v 30.16 2003/05/29 04:21:47 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdConstString   program_name;           /* This program name. */
static SpdString        model_name = NULL;

static void parse_option(int argc, char *argv[]);
static void print_usage(SpdConstString command_name);

static void dump_model(SpdModelId modelId);
static void dump_obj_table(SpdClassId classId);
static void dump_rel_table(SpdClassId relId);

static void output_class_header(SpdAttrId firstAidInt, SpdAttrId lastAidInt,
                                SpdAttrId firstAidStr, SpdAttrId lastAidStr,
                                SpdClassId classId);

int main(int argc, char *argv[])
{
    int         i;
    SpdDBId     dbId;
    SpdModelId  mId;
    SpdModelId  ImodelId;
    SpdModelId  PmodelId;
    SpdModelId  AreaModelId;
    SpdModelId  BmodelId;
    SpdModelId  DmodelId;
    SpdModelId  SDAmodelId;
    SpdModelId  StmtModelId;
    SpdModelId  SareaModelId;
    SpdModelId  TypeViewId;
    SpdModelId  CommentViewId;
    SpdModelId  FlowViewId;
    SpdModelId  FCD_view_id;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parse_option(argc, argv);
    argc -= optind;
    argv += optind;

    ImodelId = spdUseModel(&I_model);
    PmodelId = spdUseModel(&P_model);
    AreaModelId = spdUseModel(&Area_model);
    BmodelId = spdUseModel(&B_model);
    DmodelId = spdUseModel(&D_model);
    SDAmodelId = spdUseModel(&SDA_model);
    StmtModelId = spdUseModel(&Stmt_model);
    SareaModelId = spdUseModel(&SDAarea_model);
    TypeViewId = spdUseModel(&TypeView_model);
    CommentViewId = spdUseModel(&Comment_view);
    FlowViewId = spdUseModel(&Flow_view);
    FCD_view_id = spdUseModel(&FCD_view);

    dbId = spdOpenSDB();
    spdReadDB(SPD_NON_MID, dbId);

    if (model_name != NULL) {
        dump_model(spdGetModelIdByName(model_name));
    } else if (argc == 0) {
        for (mId = SPD_FIRST_MID; mId <= SPD_LAST_MID; mId++) {
            dump_model(mId);
        }
    } else {
        for (i = 0; i < argc; i++) {
            SpdClassId  classId = spdGetClassIdByName(argv[i]);

            if (SPD_IS_CLS_ID(classId)) {
                dump_obj_table(classId);
            } else if (SPD_IS_REL_ID(classId)) {
                dump_rel_table(classId);
            }
        }
    }
    spdCloseSDB(dbId);

    return (EXIT_SUCCESS);
}

static void dump_model(SpdModelId modelId)
{
    int         i;

    if (modelId < SPD_FIRST_MID || SPD_LAST_MID < modelId) {
        spdErrorExit("dump_model(): %d: Invalid model id.\n", modelId);
    }

    for (i = spdGetFirstCid(modelId); i <= spdGetLastCid(modelId); i++) {
        dump_obj_table(i);
    }
    for (i = spdGetFirstRid(modelId); i <= spdGetLastRid(modelId); i++) {
        dump_rel_table(i);
    }
}

static void dump_obj_table(SpdClassId classId)
{
    int         i;
    SpdAttrId   aid, firstAidInt, lastAidInt, firstAidStr, lastAidStr;
    
    /* Class header. */
    firstAidInt = spdGetFirstAttrIdInt(classId);
    lastAidInt = spdGetLastAttrIdInt(classId);
    firstAidStr = spdGetFirstAttrIdStr(classId);
    lastAidStr = spdGetLastAttrIdStr(classId);
    output_class_header(firstAidInt, lastAidInt, firstAidStr, lastAidStr,
                        classId);
    
    for (i = 0; i < object_db_n; i++) {
        if (spdGetClassId(object_db[i].Obj_id) != classId) {
            continue;
        }
        /* Integer attribute. */
        for (aid = firstAidInt; aid <= lastAidInt; aid++) {
            printf("%d\t", object_db[i].ivalue[aid]);
        }
        /* String attribute. */
        for (aid = firstAidStr; aid <= lastAidStr; aid++) {
            printf("\"%s\"\t",
                   (object_db[i].svalue[aid - IVAL_NUM] ?
                    object_db[i].svalue[aid - IVAL_NUM] : ""));
        }
        printf("\n");
    }
}

static void dump_rel_table(SpdClassId relId)
{
    int         i;
    SpdAttrId   aid, firstAidInt, lastAidInt, firstAidStr, lastAidStr;

    firstAidInt = spdGetFirstAttrIdInt(relId);
    lastAidInt = spdGetLastAttrIdInt(relId);
    firstAidStr = spdGetFirstAttrIdStr(relId);
    lastAidStr = spdGetLastAttrIdStr(relId);
    output_class_header(firstAidInt, lastAidInt,
                        firstAidStr, lastAidStr,
                        relId);

    for (i = 0; i < relation_db_n; i++) {
        if (spdGetClassId(relation_db[i].Rel_id) != relId) {
            continue;
        }
        /* Integer attribute. */
        for (aid = firstAidInt; aid <= lastAidInt; aid++) {
            printf("%d\t", relation_db[i].ivalue[aid]);
        }
        printf("\n");
    }
}

static void output_class_header(SpdAttrId firstAidInt, SpdAttrId lastAidInt,
                                SpdAttrId firstAidStr, SpdAttrId lastAidStr,
                                SpdClassId cId)
{
    SpdAttrId   aId;

    /* Output class name. */
    printf("# %s/%s (%s, %d)\n",
           spdGetModelNameById(spdGetModelId(cId)),
           spdGetClassNameById(cId),
           SPD_IS_CLS_ID(cId) ? "Class" : "Relationship",
           cId);

    /* Output attribute names. */
    printf("# ");
    for (aId = firstAidInt; aId <= lastAidInt; aId++) {
        printf("%s\t", spdGetAttrName(cId, aId));
    }
    for (aId = firstAidStr; aId <= lastAidStr; aId++) {
        printf("%s\t", spdGetAttrName(cId, aId));
    }

    printf("\n");
}

static void parse_option(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:m:s")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'm':                         /* Model name. */
            model_name = optarg;
            break;
          default:
            spdError("'%c': Invalid option.\n", ch);
            print_usage(program_name);
            break;
        }
    }
}

/*
 * Printing usage.
 */

static void print_usage(SpdConstString command_name)
{
    const char  *usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [options] [className...]\n"
        "DESCRIPTION\n"
        "    -d SDB-directory: Directory for SDB (\"SDB\").\n"
        "    -m modelName: Specify model name.\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
