/* 
 * Program:             $RCSfile: spdDumpTable5.c,v $  $Revision: 30.29 $
 * 
 * Purpose:
 * 
 * Author:              N.Koyama  2000/03/16
 *                      S.Yamamoto  2000/03/16
 * 
 * (C) Copyright:       N.Koyama and S.Yamamoto  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdDumpTable5.c,v 30.29 2003/05/29 04:21:52 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static SpdBoolean   flag_s = OFF;
static SpdConstString program_name;              /* This program name. */
static char         *model_name = NULL;

static void parse_option(int argc, char *argv[]);
static void print_usage(SpdConstString command_name);

static void dump_model(SpdModelId modelId);
static void dump_table(const char *class_name);

int main(int argc, char *argv[])
{
    int         i;
    SpdDBId     dbId;
    SpdModelId  mId;
    SpdModelId  ImodelId;
    SpdModelId  PmodelId;
    SpdModelId  AreaModelId;
    SpdModelId  BmodelId;
    SpdModelId  DmodelId;
    SpdModelId  SDAmodelId;
    SpdModelId  StmtModelId;
    SpdModelId  SareaModelId;
    SpdModelId  TypeViewId;
    SpdModelId  CommentViewId;
    SpdModelId  FlowViewId;
    SpdModelId  FCDViewId;
    SpdModelId  JmodelId;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);
    parse_option(argc, argv);
    argc -= optind;
    argv += optind;

    ImodelId = spdUseModel(&I_model);
    PmodelId = spdUseModel(&P_model);
    AreaModelId = spdUseModel(&Area_model);
    BmodelId = spdUseModel(&B_model);
    DmodelId = spdUseModel(&D_model);
    SDAmodelId = spdUseModel(&SDA_model);
    StmtModelId = spdUseModel(&Stmt_model);
    SareaModelId = spdUseModel(&SDAarea_model);
    TypeViewId = spdUseModel(&TypeView_model);
    CommentViewId = spdUseModel(&Comment_view);
    FlowViewId = spdUseModel(&Flow_view);
    FCDViewId = spdUseModel(&FCD_view);
    JmodelId = spdUseModel(&J_model);

    dbId = spdOpenSDB();
    /* spdReadDB(SPD_NON_MID, dbId); */
    spdReadDB(ImodelId, dbId);
    spdReadDB(PmodelId, dbId);
    spdReadDB(AreaModelId, dbId);
    spdReadDB(BmodelId, dbId);
    spdReadDB(DmodelId, dbId);
    spdReadDB(SDAmodelId, dbId);
    spdReadDB(StmtModelId, dbId);
    spdReadDB(SareaModelId, dbId);
    spdReadDB(TypeViewId, dbId);
    spdReadDB(CommentViewId, dbId);
    spdReadDB(FlowViewId, dbId);
    spdReadDB(FCDViewId, dbId);
    spdReadDB(JmodelId, dbId);

    if (model_name != NULL) {
        dump_model(spdGetModelIdByName(model_name));
    } else if (argc == 0) {
        for (mId = SPD_FIRST_MID; mId <= SPD_LAST_MID; mId++) {
            dump_model(mId);
        }
    } else {
        for (i = 0; i < argc; i++) {
            dump_table(argv[i]);
        }
    }
    spdCloseSDB(dbId);

    return (EXIT_SUCCESS);
}

static void dump_model(SpdModelId modelId)
{
    int         i;

    /* !! Temporarily added by S.Yamamoto 2002/07/28. */
    if (modelId == SPD_MID_J_MODEL) {
        spdWarning("dump_model(): J-model skipped.\n");
    }

    if (modelId < SPD_FIRST_MID || SPD_LAST_MID < modelId) {
        spdErrorExit("dump_model(): %d: Invalid model id.\n", modelId);
    }
    for (i = spdGetFirstCid(modelId); i <= spdGetLastCid(modelId); i++) {
        dump_table(spdGetClassNameById(i));
    }
    for (i = spdGetFirstRid(modelId); i <= spdGetLastRid(modelId); i++) {
        dump_table(spdGetClassNameById(i));
    }
}

static void dump_table(const char *class_name)
{
    int             attrNum;
    SpdDBType       attr_type;
    SpdObjId        obj_id;
    SpdCursor       *obj_csr;
    SpdClassId      classId;
    SpdAttrId       attrId;

    if ((classId = spdGetClassIdByName(class_name)) == SPD_NON_ID) {
        spdErrorExit("Invalid class name: \"%s\".\n", class_name);
    }

    /* !! Temporarily added by S.Yamamoto 2002/07/28. */
    /* 
     * if (spdGetModelId(classId) == SPD_MID_J_MODEL) {
     *     spdWarning("dump_table(): \"%s\" skipped.\n", class_name);
     * }
     */
    
    /*
     * Output class name.
     */
    printf("# %s/%s (%s, %d)\n",
           spdGetModelNameById(spdGetModelId(classId)), class_name,
           (spdGetClassSort(classId) & SAPIDDB_SORT_OBJ) ?
           "Class" : "Relationship",
           classId);
  
    /*
     * Output attribute names.
     * AR5ǤϥեɤľܥǤʤ?
     */
    if (spdGetClassSort(classId) & SAPIDDB_SORT_OBJ) {
        printf("# id\t");
    } else {
        printf("# id\t");
    }        
    attrNum = spdGetNumOfAttrs(classId);
    for (attrId = 0; attrId < attrNum; attrId++) {
        printf("%s\t", spdGetAttrNameByNo(classId, attrId));
    }
    printf("\n");
  
    /*
     * Output attribute values.
     */
    obj_csr = spdGetObjIdInit(class_name, NULL);
    while ((obj_id = spdGetObjId(obj_csr)) != SPD_NON_ID) {
        printf("%d\t", obj_id);
        for (attrId = 0; attrId < attrNum; attrId++) {
            attr_type = spdGetAttrType(classId, attrId);
            switch (attr_type) {
              case SPD_DBT_INT:
              case SPD_DBT_OID:
                printf("%d\t", spdGetAttrIntVal(obj_id, attrId));
                break;
              case SPD_DBT_ENUM:
                if (flag_s) {
                    int             val = spdGetAttrIntVal(obj_id, attrId);
                    SpdConstString  str = spdGetEnumStrVal(classId,
                                                           attrId, val);
                
                    printf("%d (%s)\t", val,
                           (str == NULL) ? "*** UNKNOWN ***" : str);
                } else {
                    printf("%d\t", spdGetAttrIntVal(obj_id, attrId));
                }
                break;
              case SPD_DBT_STR:
                printf("\"%s\"\t", spdGetAttrStrVal(obj_id, attrId));
                break;
              case SPD_DBT_NON:
                /* Is it right? */
                break;
            }
        }
        printf("\n");
    }
    spdFreeCursor(obj_csr);
}

static void parse_option(int argc, char *argv[])
{
    int     ch;

    while ((ch = getopt(argc, argv, "d:m:s")) != -1) {
        switch (ch) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'm':                         /* Model name. */
            model_name = optarg;
            break;
          case 's':                         /* Symbolic representation. */
            flag_s = ON;
            break;
          default:
            spdError("'%c': Invalid option.\n", ch);
            print_usage(program_name);
            break;
        }
    }
}

/*
 * Printing usage.
 */

static void print_usage(SpdConstString command_name)
{
    const char  *usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [options] [className...]\n"
        "DESCRIPTION\n"
        "    -d SDB-directory: Directory for SDB (\"SDB\").\n"
        "    -m modelName: Specify model name.\n"
        "    -s: Symbolic representation.\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
