/* 
 * Program:             $RCSfile: SpdFlatten.h,v $  $Revision: 30.3 $
 * 
 * Purpose:
 * 
 * Author:              T.Watanabe  1998/04/02
 *                      S.Yamamoto  1998/04/04
 * 
 * (C) Copyright:       T.Watanabe and S.Yamamoto  1998
 *                      This file is a product of the project Sapid.
 */

#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/SpdStr.h>

#define EXPR                1
#define DECL                0

#define EXPR_ASSIGN         (EXPR_SORT_POINTER_MEMBER + 1)
#define EXPR_SINGLE_L_ARITH (EXPR_SORT_POINTER_MEMBER + 2)
#define EXPR_SINGLE_R_ARITH (EXPR_SORT_POINTER_MEMBER + 3)

#define spdStrCat(x, y)     spdStrcat((x), (y))

typedef struct _Node {
    int objectId;
    int sort;
    SpdString str;
    int child_n;
    struct _Node **child;
} Node;

typedef struct _Stack {
    Node **node;
    int num;
} Stack;

typedef struct {
    SpdString *str;
    int num;
} StatementArray;

typedef struct {
    StatementArray expr;
    StatementArray decl;
} Block;

/* block.c */
Block *initBlock(void);
void freeBlock(Block *);
void printBlock(Block *);
void addElementToBlock(Block *, SpdString, int);
Block *copyBlock(Block *);

/* spdSimplize.c */
Stack *initStack(void);

void freeNode(Node *);
void freeStack(Stack *);

Node  *mkNode(int);
Node *copyNode(Node *);
Block *mkSimpleExpr(SpdObjId);
void  spdEntryDeclarator(SpdString, SpdString, Block *);
Node *haveNoSibling(Stack *, Node *, Block *);
Node  *namedNode(Node *);
Node  *mkNewTree(Node *, Node *);
SpdString getTypeStr(SpdObjId);
Node *spdAnalTree(Stack *,Node *, Block *);
void spdExprTree(Node *);
void swap(Node **, Node **);
SpdBoolean isNeedPar(Node *, Node *);
int getExprPrimary(Node *);
SpdString namedLabel(SpdId);
SpdString spdMkStatement(StatementArray, SpdString);
SpdString spdMkStatementButOne(StatementArray, SpdString);

/* spdStr_*.c */
Block *spdStrExpression2(SpdObjId); 
Block *spdStrBlock2(SpdObjId);
Block *spdStrBasic2(SpdObjId);
Block *spdStrBranch2(SpdObjId);
Block *spdStrCompound2(SpdObjId);
SpdString spdStr2(SpdObjId); 
Block *spdStrDeclarator2(SpdObjId, SpdBoolean);
Block *spdStrOptionaldecl2(SpdObjId, SpdString);
Block *spdStrDeclaration2(SpdObjId, SpdBoolean);
SpdString spdStrFile2(SpdObjId); 

SpdString spdStrConstant(SpdObjId);
SpdString spdStrDeclaration(SpdObjId, SpdBoolean);
SpdString spdStrDeclarator(SpdObjId, SpdBoolean);
SpdString spdStrExpression(SpdObjId); 
SpdString spdStrIdentifier(SpdObjId);
SpdString spdStrLabel(SpdObjId);
SpdString spdStrOperator(SpdObjId);
SpdString spdStrOptionaldecl(SpdObjId);
SpdString spdStrType(SpdObjId);
SpdString spdStrTypeSpecifier(SpdObjId);
SpdString spdStrTypeQualifier(SpdObjId, SpdString, int);

/* Utility functions. */

/* SpdString spdStrCat(SpdString, SpdString); */
SpdString spdStrIndent(void);
SpdBoolean spdObjIsFuncParameter(const SpdObjId);
int cmp_offset(const SpdRel *, const SpdRel *);
