/* 
 * Program:             $RCSfile: spdFltBlock.c,v $  $Revision: 30.5 $
 * 
 * Purpose:
 * 
 * Author:              T.Watanabe  1998/04/02
 *                      S.Yamamoto  1998/04/04
 * 
 * (C) Copyright:       T.Watanabe and S.Yamamoto  1998
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdFltBlock.c,v 30.5 2003/06/28 15:34:07 yamamoto Exp yamamoto $";
#endif

#include "SpdFlatten.h"

extern int indent;                       /* ǥȿ */

Block *spdStrBlock2(SpdObjId objectId)
{
    Block *block = NULL;

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return NULL;
    }

    switch (spdGetAttrValInt(objectId, "sort")) {
      case BLOCK_BASIC:            
        block = spdStrBasic2(objectId);
        break;
      case BLOCK_COMP:
        block = spdStrCompound2(objectId);
        break;
      case BLOCK_IF:                        /* Branch ֥å */
      case BLOCK_SWITCH:
      case BLOCK_FOR:
      case BLOCK_WHILE:
      case BLOCK_DO:
        block = spdStrBranch2(objectId); 
        break;
    }
    return block;
}

Block *spdStrBasic2(SpdObjId objectId)
{
    SpdString   eres = NULL, dres = NULL;
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     buf_n = 0, i;
    Block *cblock = NULL, *pblock = NULL;

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return NULL;
    }
    if (spdGetAttrValInt(objectId, "sort") != BLOCK_BASIC) return NULL;

    csr = spdGetRelInit(objectId, "blk_expr", "any");
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);

    SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);

    for (i = 0; i < buf_n; i++) {
        if(cblock != NULL) spdFree(cblock);
        cblock = spdStrExpression2(buf[i].objectId);
        eres = spdMkStatement(cblock->expr, eres);
        dres = spdMkStatement(cblock->decl, dres);
    }
    if (buf != NULL) spdFree(buf);
    pblock = initBlock();
    if(dres != NULL) addElementToBlock(pblock, dres, DECL);
    if(eres != NULL) addElementToBlock(pblock, eres, EXPR);
    if(cblock != NULL) freeBlock(cblock);
    return pblock;
}

Block *spdStrBranch2(SpdObjId objectId)
{
    SpdObjId    expr_id, blk_id, pblk_id;
    SpdString   eres = NULL, dres = NULL;
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     buf_n = 0, j, is_elif;
    Block *cblock = NULL, *pblock = NULL;
    SpdString   tmp_res = NULL;
    SpdString second_for_exprs = NULL, third_for_expr = NULL;

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return NULL;
    }

    switch(spdGetAttrValInt(objectId, "sort")){
    case BLOCK_IF:
      pblk_id = spdGetARelObj(objectId, "blk_blk", "c-id"); 
      is_elif = (spdGetAttrValInt(pblk_id, "sort") == BLOCK_IF) ? 1 : 0;
      /* if-expr */
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        if(cblock != NULL) freeBlock(cblock);
        cblock = spdStrExpression2(expr_id);
        dres = spdMkStatement(cblock->decl, dres);
        eres = spdMkStatementButOne(cblock->expr, eres);
        if(is_elif == 0){                   /* else-ifʸǤʤ */
            tmp_res = spdStrCat(tmp_res, "\n");
            tmp_res = spdStrCat(tmp_res, spdStrIndent());
        }
        tmp_res = spdStrCat(tmp_res, "if (");   
        tmp_res = spdStrCat(tmp_res, cblock->expr.str[0]);
        tmp_res = spdStrCat(tmp_res, ") ");
        /* if-block */
        csr = spdGetRelInit(objectId, "blk_blk", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
        if(cblock != NULL) freeBlock(cblock);
        if(spdGetAttrValInt(buf[0].objectId, "sort") != BLOCK_COMP){
            tmp_res = spdStrCat(tmp_res, "{");
            indent++;
            cblock = spdStrBlock2(buf[0].objectId);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                tmp_res = spdStrCat(tmp_res, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j >= 0; j--){
                tmp_res = spdStrCat(tmp_res, cblock->expr.str[j]);
            }
            tmp_res = spdStrCat(tmp_res, "\n");
            indent--;
            tmp_res = spdStrCat(tmp_res, spdStrIndent());
            tmp_res = spdStrCat(tmp_res, "}");
        } else {
            cblock = spdStrBlock2(buf[0].objectId);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                tmp_res = spdStrCat(tmp_res, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j > 0; j--){
                tmp_res = spdStrCat(tmp_res, cblock->expr.str[j]);
            }
            tmp_res = spdStrCat(tmp_res, "\n");
            tmp_res = spdStrCat(tmp_res, spdStrIndent());
            tmp_res = spdStrCat(tmp_res, cblock->expr.str[0]);
        }
        if (buf_n > 1) {                    /* else֥å */
            int blk_sort = spdGetAttrValInt(buf[1].objectId, "sort"); 
            tmp_res = spdStrCat(tmp_res, " else ");
            if(cblock != NULL) freeBlock(cblock);
            /* comp or notǤȤäȴñ... */
            if(blk_sort == BLOCK_COMP){
                cblock = spdStrBlock2(buf[1].objectId);
                for(j = cblock->decl.num - 1; j >= 0; j--){
                    dres = spdStrCat(dres, cblock->decl.str[j]);
                }
                for(j = cblock->expr.num -1; j >= 0; j--){
                    if(j == 0){
                        tmp_res = spdStrCat(tmp_res, "\n");
                        tmp_res = spdStrCat(tmp_res, spdStrIndent());
                    }
                    tmp_res = spdStrCat(tmp_res, cblock->expr.str[j]);
                }
            } else if(blk_sort == BLOCK_IF){
                cblock = spdStrBlock2(buf[1].objectId);
                for(j = cblock->decl.num - 1; j >= 0; j--){
                    dres = spdStrCat(dres, cblock->decl.str[j]);
                }
                if(cblock->expr.num == 2){
                    eres = spdStrCat(eres, cblock->expr.str[1]);
                    tmp_res = spdStrCat(tmp_res, cblock->expr.str[0]);
                } else {
                    tmp_res = spdStrCat(tmp_res, cblock->expr.str[0]);
                }     
            } else {
                tmp_res = spdStrCat(tmp_res, "{");
                indent++;
                cblock = spdStrBlock2(buf[1].objectId);
                for(j = cblock->decl.num - 1; j >= 0; j--){
                    tmp_res = spdStrCat(tmp_res, cblock->decl.str[j]);
                }
                for(j = cblock->expr.num - 1; j >= 0; j--){
                    tmp_res = spdStrCat(tmp_res, cblock->expr.str[j]);
                }
                tmp_res = spdStrCat(tmp_res, "\n");
                indent--;
                tmp_res = spdStrCat(tmp_res, spdStrIndent());
                tmp_res = spdStrCat(tmp_res, "}");  
            } 
        }
        if (buf != NULL) spdFree(buf);
        pblock = initBlock();
        if(dres != NULL) addElementToBlock(pblock, dres, DECL);
        if(tmp_res != NULL) addElementToBlock(pblock, tmp_res, EXPR);
        if(eres != NULL) addElementToBlock(pblock, eres, EXPR);
        if(cblock != NULL) freeBlock(cblock);
        return pblock;
        break;
      case BLOCK_SWITCH:
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        if(cblock != NULL) freeBlock(cblock);
        cblock = spdStrExpression2(expr_id);
        /* switch expr part*/
        eres = spdMkStatementButOne(cblock->expr, eres);

        eres = spdStrCat(eres, "\n");
        eres = spdStrCat(eres, spdStrIndent());
        eres = spdStrCat(eres, "switch ("); 
        eres = spdStrCat(eres, cblock->expr.str[0]);
        eres = spdStrCat(eres, ") ");

        dres = spdMkStatement(cblock->decl, dres);

        /* switch block part */
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        if(cblock != NULL) freeBlock(cblock);
        cblock = spdStrBlock2(blk_id);
        for(j = cblock->decl.num - 1; j >= 0; j--){
            eres = spdStrCat(eres, cblock->decl.str[j]);
        }
        for(j = cblock->expr.num - 1; j > 0; j--){
            eres = spdStrCat(eres, cblock->expr.str[j]);
        }
        eres = spdStrCat(eres, "\n");
        eres = spdStrCat(eres, spdStrIndent());
        eres = spdStrCat(eres, cblock->expr.str[0]);
        break;
      case BLOCK_FOR:
        /* for expr part */
        csr = spdGetRelInit(objectId, "blk_expr", "any");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);

        /* tmp_resˤΥ֥åĤäƤ */
        tmp_res = spdStrCat(tmp_res, "\n");
        tmp_res = spdStrCat(tmp_res, spdStrIndent());
        tmp_res = spdStrCat(tmp_res, "for (");  
        /* 켰 */
        if(cblock != NULL) freeBlock(cblock);       
        cblock = spdStrExpression2(buf[0].objectId);        
        eres = spdMkStatement(cblock->expr, eres);
        /* 켰϶ʸ */
        tmp_res = spdStrCat(tmp_res, "; ");
        dres = spdMkStatement(cblock->decl, dres);
        /* (Ｐ) */
        if(cblock != NULL) freeBlock(cblock);       
        cblock = spdStrExpression2(buf[1].objectId);        
        eres = spdMkStatementButOne(cblock->expr, eres);
        indent++;
        second_for_exprs = spdMkStatementButOne(cblock->expr, second_for_exprs);
        indent--;
        tmp_res = spdStrCat(tmp_res, cblock->expr.str[0]);
        tmp_res = spdStrCat(tmp_res, ";");
        dres = spdMkStatement(cblock->decl, dres);
        /* 軰 */
        if(cblock != NULL) freeBlock(cblock);       
        cblock = spdStrExpression2(buf[2].objectId);        
        indent++;
        third_for_expr = spdMkStatement(cblock->expr, third_for_expr);
        indent--;
        dres = spdMkStatement(cblock->decl, dres);
        eres = spdStrCat(eres, tmp_res);
        eres = spdStrCat(eres, ") ");
        if (buf != NULL) spdFree(buf);
        /* for-block part*/
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        if(cblock != NULL) freeBlock(cblock);
        if(spdGetAttrValInt(blk_id, "sort") != BLOCK_COMP){
            eres = spdStrCat(eres, "{");
            indent++;
            cblock = spdStrBlock2(blk_id);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
            indent--;

            /* continue  */    
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres,  namedLabel(objectId));
            eres = spdStrCat(eres, ":");

            eres = spdStrCat(eres, third_for_expr);
            eres = spdStrCat(eres, second_for_exprs);
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres, spdStrIndent());
            eres = spdStrCat(eres, "}");
        } else {
            cblock = spdStrBlock2(blk_id);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j > 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
            /* continue  */    
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres,  namedLabel(objectId));
            eres = spdStrCat(eres, ":");

            eres = spdStrCat(eres, third_for_expr);
            eres = spdStrCat(eres, second_for_exprs);
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres, spdStrIndent());
            eres = spdStrCat(eres, cblock->expr.str[0]);
        }
        break;
      case BLOCK_WHILE:
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        if(cblock != NULL) spdFree(cblock);
        /* while expr part */
        cblock = spdStrExpression2(expr_id);
        eres = spdMkStatementButOne(cblock->expr, eres);
        indent++;
        tmp_res = spdMkStatementButOne(cblock->expr, tmp_res);
        indent--;
        eres = spdStrCat(eres, "\n");
        eres = spdStrCat(eres, spdStrIndent());
        eres = spdStrCat(eres, "while (");  
        eres = spdStrCat(eres, cblock->expr.str[0]);
        eres = spdStrCat(eres, ") ");

        dres = spdMkStatement(cblock->decl, dres);
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        /* while block part */
        if(cblock != NULL) spdFree(cblock);
        if(spdGetAttrValInt(blk_id, "sort") != BLOCK_COMP){
            indent++;
            cblock = spdStrBlock2(blk_id);
            eres = spdStrCat(eres, "{");
            for(j = cblock->decl.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }

            /* continue  */    
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres,  namedLabel(objectId));
            eres = spdStrCat(eres, ":");

            indent--;
            eres = spdStrCat(eres, tmp_res);
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres, spdStrIndent());
            eres = spdStrCat(eres, "}");
        } else {
            cblock = spdStrBlock2(blk_id);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j > 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
            /* continue  */    
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres,  namedLabel(objectId));
            eres = spdStrCat(eres, ":");

            eres = spdStrCat(eres, tmp_res);
            eres = spdStrCat(eres, "\n");
            eres = spdStrCat(eres, spdStrIndent());
            eres = spdStrCat(eres, cblock->expr.str[0]);
        }
        break;
      case BLOCK_DO:
        eres = spdStrCat(eres, "\n");
        eres = spdStrCat(eres, spdStrIndent());
        eres = spdStrCat(eres, "do ");
        blk_id = spdGetARelObj(objectId,"blk_blk","p-id");
        /* do block part */
        if(cblock != NULL) spdFree(cblock);
        if(spdGetAttrValInt(blk_id, "sort") != BLOCK_COMP){    
            eres = spdStrCat(eres, "{");
            indent++;
            cblock = spdStrBlock2(blk_id);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
            indent--;

            tmp_res = spdStrCat(tmp_res, "}");
        } else {
            cblock = spdStrBlock2(blk_id);
            for(j = cblock->decl.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->decl.str[j]);
            }
            for(j = cblock->expr.num - 1; j > 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
            tmp_res = spdStrCat(tmp_res, cblock->expr.str[0]);
        }
        /* do expr part */
        expr_id = spdGetARelObj(objectId,"blk_expr","any");
        if(cblock != NULL) freeBlock(cblock); 
        cblock = spdStrExpression2(expr_id);

        /* continue  */        
        eres = spdStrCat(eres, "\n");
        eres = spdStrCat(eres,  namedLabel(objectId));
        eres = spdStrCat(eres, ":");

        indent++;
        eres = spdMkStatementButOne(cblock->expr, eres);
        indent--;
        eres = spdStrCat(eres, "\n");
        eres = spdStrCat(eres, spdStrIndent());
        eres = spdStrCat(eres, tmp_res);    
        eres = spdStrCat(eres, "while (");  
        eres = spdStrCat(eres, cblock->expr.str[0]);
        eres = spdStrCat(eres, ")");
        eres = spdStrCat(eres, ";");

        dres = spdMkStatement(cblock->decl, dres);
        break;
    }

    pblock = initBlock();
    if(dres != NULL) addElementToBlock(pblock, dres, DECL);
    if(eres != NULL) addElementToBlock(pblock, eres, EXPR);
    if(cblock != NULL) freeBlock(cblock);
    return pblock;
}

Block *spdStrCompound2(SpdObjId objectId)
{
    SpdString   eres = NULL, dres = NULL;
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     buf_n = 0, class_id, i, j;
    Block *cblock = NULL;                   /* Ȥ¦֥å */
    Block *pblock = initBlock();            /* ޤȤ¦֥å */

    if ((spdGetClassId(objectId)) != SPD_CID_BLOCK) {
        return NULL;
    }

    if (spdGetAttrValInt(objectId, "sort") != BLOCK_COMP) return NULL;

    csr = spdGetRelInit(objectId, "blk_decl", "any"); /* blk_decl μ */
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);
    csr = spdGetRelInit(objectId, "blk_blk", "p-id"); /* blk_blk μ */
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);
    csr = spdGetRelInit(objectId, "blk_label", "any"); /* blk_labl μ */
    while ((rel = spdGetRel(csr)).relationId != -1) {
        buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
        buf[buf_n++] = rel;
    }
    spdFreeCursor(csr);
    SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);

    indent++;
    for (i = 0; i < buf_n; i++) {
        class_id = spdGetClassId(buf[i].objectId);
        if(class_id == SPD_CID_DECLARATION) {
            if(cblock != NULL) freeBlock(cblock); 
            cblock = spdStrDeclaration2(buf[i].objectId, FALSE);
            for(j=cblock->decl.num - 1; j >= 0; j--){
                dres = spdStrCat(dres, cblock->decl.str[j]);
            }
            for(j=cblock->expr.num - 1; j >= 0; j--){
                if(j == cblock->expr.num - 1){
                    eres = spdStrCat(eres, "\n"); 
                    eres = spdStrCat(eres, spdStrIndent()); 
                }
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
        }
        else if(class_id == SPD_CID_BLOCK) {
            if(cblock != NULL) freeBlock(cblock); 
            cblock = spdStrBlock2(buf[i].objectId);
            for(j=cblock->decl.num - 1; j >= 0; j--){
                dres = spdStrCat(dres, cblock->decl.str[j]);
            }
            for(j=cblock->expr.num - 1; j >= 0; j--){
                eres = spdStrCat(eres, cblock->expr.str[j]);
            }
        }
        else if(class_id == SPD_CID_LABEL) {
            eres = spdStrCat(eres, "\n"); 
            indent--;
            eres = spdStrCat(eres, spdStrIndent());
            indent++;
            eres = spdStrCat(eres, spdStrLabel(buf[i].objectId));
            eres = spdStrCat(eres, ":");
        }
    }
    if (buf != NULL) spdFree(buf);
    /*  eres = spdStrCat(eres, "\n");*/
    indent--;
    addElementToBlock(pblock,  "}", EXPR);
    if(eres != NULL) addElementToBlock(pblock, eres, EXPR);
    if(dres != NULL) addElementToBlock(pblock, dres, EXPR);
    addElementToBlock(pblock,  "{", EXPR);
    return pblock;
}

SpdString namedLabel(SpdId id)
{
    SpdString prefix = "FLT_CONTINUE";
    SpdString name;
    int figure;
    char tmp[10];

    figure = sprintf(tmp, "%d", id);
    name = spdMalloc(sizeof(char) * 13); 
    strncpy(name, prefix, 13);
    name = spdStrCat(name, tmp);
    return name;
}


SpdString spdMkStatement(StatementArray elem, SpdString str){
    int j;

    for(j = elem.num - 1; j >= 0; j--){
        str = spdStrCat(str, "\n");
        str = spdStrCat(str, spdStrIndent());
        str = spdStrCat(str, elem.str[j]);
        str = spdStrCat(str, ";");
    }    
    return str;
}

SpdString spdMkStatementButOne(StatementArray elem, SpdString str){
    int j;

    for(j = elem.num - 1; j > 0; j--){
        str = spdStrCat(str, "\n");
        str = spdStrCat(str, spdStrIndent());
        str = spdStrCat(str, elem.str[j]);
        str = spdStrCat(str, ";");
    }    
    return str;
}
