/* 
 * Program:             $RCSfile: spdSimplize.c,v $  $Revision: 30.9 $
 * 
 * Purpose:
 * 
 * Author:              T.Watanabe  1998/04/02
 *                      S.Yamamoto  1998/04/04
 * 
 * (C) Copyright:       T.Watanabe and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdSimplize.c,v 30.9 2003/06/28 15:34:06 yamamoto Exp yamamoto $";
#endif

#include "SpdFlatten.h"

Stack *initStack(void)
{
    Stack *stack = (Stack *)spdMalloc(sizeof(Stack));

    stack->node = (Node **)spdMalloc(sizeof(Node *));
    stack->num = 0;
    return stack;
}

void freeStack(Stack *stack)
{
    int i;

    if(stack == NULL) return;
    for(i=0; i < stack->num; i++){
        freeNode(stack->node[i]);
    }
    spdFree(stack->node);
    spdFree(stack); 
}

void freeNode(Node *node)
{
    int i;

    if(node == NULL) return;
    if(node->child_n == 0){
        if(node->str != NULL) spdFree(node->str);
        spdFree(node); 
        return;
    }
    for(i=0;i < node->child_n; i++){
        freeNode(node->child[i]);
    }
    if(node->child != NULL) spdFree(node->child); 
    if(node->str != NULL) spdFree(node->str);
    spdFree(node);
}

Node *mkNode(int objectId)
{
    Node *node = (Node *)spdMalloc(sizeof(Node));

    node->objectId = objectId;
    node->sort = -1;
    node->str = NULL;
    node->child_n = 0;
    node->child = NULL;
    return (node);
}

Node *copyNode(Node *org)
{
    int i;
    Node *node = mkNode(org->objectId);

    node->sort = org->sort;
    node->str = spdStrCat(node->str, org->str);
    node->child_n = org->child_n;
    if(node->child_n != 0){
        node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
        for(i=0; i < org->child_n; i++){
            node->child[i] = copyNode(org->child[i]);
        }
    }
    return node;
}

Node *namedNode(Node *node)
{
    static int surfix = 0;
    SpdString prefix = "newid_";
    SpdString name;
    int figure;
    char tmp[10];

    figure = sprintf(tmp, "%d", surfix);
    name = spdMalloc(sizeof(char) * (figure + 1)); 
    strncpy(name, tmp, (figure + 1));
    surfix++;
    node->str = spdStrCat(node->str, prefix);
    node->str = spdStrCat(node->str, name);
    return node;
}


/* node1node2ǤĤʤĥ꡼ */
Node *mkNewTree(Node *node1, Node *node2)
{
    Node *p_node, *l_node;

    p_node = mkNode(SPD_NON_ID);
    p_node->sort = EXPR_ASSIGN;
    p_node->child_n = 3;
    p_node->child = (Node **)spdMalloc(sizeof(Node *) * p_node->child_n);
    l_node = copyNode(node1);
    p_node->child[0] = l_node;
    p_node->child[2] = node2;
    p_node->child[1] = mkNode(SPD_NON_ID);
    p_node->child[1]->str = spdStrCat(p_node->child[1]->str, " = ");
    return p_node;
}

SpdString getTypeStr(SpdObjId id)
{
    if (id == SPD_NON_ID) {
        spdErrorExit("\" %d \" is not found.\n", id);
    }
    /* 
     * return (spdStrType(smuGetExprType(id)));
     */
    return (spdStrType(spdGetARelObj(id, "expr_type", "expression_id")));
}

void swap(Node **node1, Node **node2)
{
    Node *tmp_node;

    tmp_node = *node1;
    *node1 = *node2;
    *node2 = tmp_node;
}

Block *mkSimpleExpr(SpdObjId objectId)
{
    Node *root;
    Stack *store_tree;
    int i;
    Node *res_node = NULL;
    Block *block = NULL;

    store_tree = initStack();
    block = initBlock();
    root = mkNode(objectId); 
    spdExprTree(root);
    res_node = spdAnalTree(store_tree, root, block);
    addElementToBlock(block, res_node->str, EXPR);
    /* single ExpressionˤΤexpr */
    for(i = 0; i < store_tree->num; i ++){
        res_node = spdAnalTree(store_tree, store_tree->node[i], block);
        addElementToBlock(block, res_node->str, EXPR);
    }
    freeNode(root); 
    freeStack(store_tree);
    
    return block;
}

void spdEntryDeclarator(SpdString name, SpdString type, Block *block){
    block->decl.str[block->decl.num] = NULL;
    block->decl.str[block->decl.num] = 
        spdStrCat(block->decl.str[block->decl.num], type); 
    block->decl.str[block->decl.num] = 
        spdStrCat(block->decl.str[block->decl.num], name);
    block->decl.num++;
    block->decl.str = (SpdString *)spdRealloc(block->decl.str, 
                               sizeof(SpdString) * (block->decl.num + 1));
    return;
}

SpdBoolean isNeedPar(Node *par, Node *child)
{
    int pprimary, cprimary;
    pprimary = getExprPrimary(par);
    cprimary = getExprPrimary(child);

    if(pprimary > cprimary) return TRUE;
    return FALSE;
}

int getExprPrimary(Node *node){
    int sort, op_sort;
    int primary = -1;
    SpdObjId opId, exprId;

    exprId = node->objectId;
    if(spdGetClassId(exprId) != SPD_CID_EXPRESSION){
        return SPD_NON_ID;
    }
    sort = node->sort;
    switch(sort){
    case EXPR_SORT_ARRAY:       
    case EXPR_SORT_MEMBER:
    case EXPR_SORT_POINTER_MEMBER:
    case EXPR_SORT_FUNCCALL:
        primary = 17;
        break;
    case EXPR_SORT_POINTER:
    case EXPR_SORT_ADDRESS:
    case EXPR_SORT_SIZEOF:
        primary = 15;
        break;
    case EXPR_SORT_CAST:
        primary = 14;
        break;
    case EXPR_SORT_SHIFT:
        primary = 11;
        break;
    case EXPR_SORT_TRI:
        primary = 3;
        break;
    case EXPR_ASSIGN:
        primary = 2;
        break;
    case EXPR_SORT_COMMA:
        primary = 1;
        break;
    case EXPR_SINGLE_L_ARITH:
        primary = 15;
        break;
    case EXPR_SINGLE_R_ARITH:
        primary = 16;
        break;
    default:
        opId = spdGetARelObj(exprId, "expr_op", "expression_id");
        op_sort = spdGetAttrValInt(opId, "sort");
        switch(op_sort){
        case OP_NOT:
        case OP_B_NOT:
        case OP_PLUS:
        case OP_MINUS:
            primary = 15;
            break;
        case OP_MUL:
        case OP_DIV:
        case OP_MOD:
            primary = 13;
            break;
        case OP_ADD:
        case OP_SUB:
            primary = 12;
            break;
        case OP_SHL:
        case OP_SHR:
            primary = 11;
            break;
        case OP_LT:
        case OP_LE:
        case OP_GT:
        case OP_GE:
            primary = 10;
            break;
        case OP_EQ:
        case OP_NE:
            primary = 9;
            break;
        case OP_B_AND:
            primary = 8;
            break;
        case OP_B_XOR:
            primary = 7;
            break;
        case OP_B_OR:
            primary = 6;
            break;
        case OP_L_AND:
            primary = 5;
            break;
        case OP_L_OR:
            primary = 4;
            break;
        default:
            primary = 20;
            break;
        }
        break;
    }
    return primary;
}

/* ʬ˷郎зʤλҤˤ */
Node *haveNoSibling(Stack *stack, Node *node, Block *block)
{
    Node *newNode = NULL;
    SpdString type = NULL;

    if(node->child_n == 0) return node;
    if(node->sort == EXPR_ASSIGN || node->sort == EXPR_SORT_POINTER ||
       node->sort == EXPR_SORT_ARRAY || node->sort == EXPR_SORT_MEMBER ||
       node->sort == EXPR_SORT_ADDRESS || node->sort == EXPR_SORT_FUNCCALL ||
       node->sort == EXPR_SORT_POINTER_MEMBER){
        node = spdAnalTree(stack, node, block);
    } else {
        newNode = mkNode(SPD_NON_ID);
        newNode = namedNode(newNode);
        type = getTypeStr(node->objectId);
        spdEntryDeclarator(newNode->str, type, block);
        stack->node = spdRealloc(stack->node, 
                                 sizeof(Node *) * (stack->num + 1));
        stack->node[stack->num++] = mkNewTree(newNode, node); 
        swap(&(node), &(newNode));
    }
    return node;
}

Node *spdAnalTree(Stack *stack, Node *node, Block *block)
{
    int i, top_node = FALSE;
    static int nest_num = -1;
    Node *newNode = NULL;

    nest_num++;
    if(nest_num == 0) top_node = TRUE;

    switch(node->sort){
      case EXPR_SORT_MEMBER:
      case EXPR_SORT_POINTER_MEMBER:
        if(node->child_n == 0) break;
        if(isNeedPar(node, node->child[0])){
            node->str = spdStrCat(node->str, "(");
            node->child[0] = spdAnalTree(stack, node->child[0], block);
            node->str = spdStrCat(node->str, node->child[0]->str);
            node->str = spdStrCat(node->str, ")");
            for(i=1; i< node->child_n; i++) {    
                node->child[i] = spdAnalTree(stack, node->child[i], block);
                node->str = spdStrCat(node->str, node->child[i]->str);
            }
        } else {
            for(i=0; i< node->child_n; i++) {    
                node->child[i] = spdAnalTree(stack, node->child[i], block);
                node->str = spdStrCat(node->str, node->child[i]->str);
            }
        }
        break;
      case EXPR_SORT_ARRAY:
        if(node->child_n == 0) break;
        if(isNeedPar(node, node->child[0])){
            node->str = spdStrCat(node->str, "(");
            node->child[0] = spdAnalTree(stack, node->child[0], block);
            node->str = spdStrCat(node->str, node->child[0]->str);
            node->str = spdStrCat(node->str, ")");
        } else {
            node->child[0] = spdAnalTree(stack, node->child[0], block);
            node->str = spdStrCat(node->str, node->child[0]->str);
        }                
        node->str = spdStrCat(node->str, "[");
        node->child[1] = haveNoSibling(stack, node->child[1], block);
        node->str = spdStrCat(node->str, node->child[1]->str); 
        node->str = spdStrCat(node->str, "]");
        break;
      case EXPR_SORT_POINTER:
      case EXPR_SORT_ADDRESS:
        for(i=0; i< node->child_n; i++) {    
            node->child[i] =  haveNoSibling(stack, node->child[i], block);
            node->str = spdStrCat(node->str, node->child[i]->str);
        }
        break;
      case EXPR_ASSIGN:
        if(top_node == TRUE){     /*  ϤƤexpressionλ */
            for(i = 0; i < node->child_n; i++){
                node->child[i] = spdAnalTree(stack, node->child[i], block);
                node->str = spdStrCat(node->str, node->child[i]->str);
            }
        } else {                /* ϤƤи */
            if(node->child_n == 0) break;
            newNode = mkNode(SPD_NON_ID);
            stack->node = spdRealloc(stack->node, 
                                     sizeof(Node *) * (stack->num + 1)); 
            stack->node[stack->num++] = node;

                node->child[0] = spdAnalTree(stack, node->child[0], block);
                newNode->str = spdStrCat(newNode->str, node->child[0]->str);
                for(i=0; i < node->child[0]->child_n; i++){
                    freeNode(node->child[0]->child[i]);
                }
                node->child[0]->child_n = 0;
            swap(&(node), &(newNode));
        }       
        break;
    
      case EXPR_SORT_LOGICAL:
      case EXPR_SORT_COMPARE:
/*
        if(top_node == TRUE) node = haveNoSibling(stack, node, block);
        else {
            node->sort = EXPR_SORT_ARITH; 
            node = spdAnalTree(stack, node, block);
        }
*/
        node->str = spdStrCat(node->str, spdStrExpression(node->objectId));
        break;
      case EXPR_SORT_FUNCCALL:
        if(node->child_n == 0) break;
        node->child[0] = spdAnalTree(stack, node->child[0], block);
        if(isNeedPar(node, node->child[0])){
            node->str = spdStrCat(node->str, "(");
            node->str = spdStrCat(node->str, node->child[0]->str);
            node->str = spdStrCat(node->str, ")");
        } else {
            node->str = spdStrCat(node->str, node->child[0]->str);
        }
        node->str = spdStrCat(node->str, "(");
        for(i=1; i< node->child_n; i++) {          
            if(i != 1) node->str = spdStrCat(node->str, ", ");
            node->child[i] = haveNoSibling(stack, node->child[i], block);
            node->str = spdStrCat(node->str, node->child[i]->str);
        }
        node->str = spdStrCat(node->str, ")");
        break;
      case EXPR_SINGLE_L_ARITH:
        node->str = spdStrCat(node->str, node->child[0]->str);
        if(isNeedPar(node, node->child[1])){
            node->str = spdStrCat(node->str, "(");
            node->child[1] = haveNoSibling(stack, node->child[1], block);
            node->str = spdStrCat(node->str, node->child[1]->str);
            node->str = spdStrCat(node->str, ")");
        } else {
            node->child[1] = haveNoSibling(stack, node->child[1], block);
            node->str = spdStrCat(node->str, node->child[1]->str);
        }
        break;
      case EXPR_SINGLE_R_ARITH:
        if(isNeedPar(node, node->child[0])){
            node->str = spdStrCat(node->str, "(");
            node->child[0] = haveNoSibling(stack, node->child[0], block);
            node->str = spdStrCat(node->str, node->child[0]->str);
            node->str = spdStrCat(node->str, ")");
        } else {
            node->child[0] = haveNoSibling(stack, node->child[0], block);
            node->str = spdStrCat(node->str, node->child[0]->str);
        }
        node->str = spdStrCat(node->str, node->child[1]->str);
        break;
      /* Ūʲ(ŸˤΤޤɽƤ롣) */
      case EXPR_SORT_TRI: 
        node->str = spdStrCat(node->str, spdStrExpression(node->objectId));
        break;
      case EXPR_SORT_CAST: 
      default: 
        for(i=0; i < node->child_n; i++){
            if(node->child[i]->child_n == 0){
                node->str = spdStrCat(node->str, node->child[i]->str);      
            } else {
                node->child[i] = haveNoSibling(stack, node->child[i], block);
                node->str = spdStrCat(node->str, node->child[i]->str);
            }
        }
        break;
    }

    if(top_node == 1) nest_num = -1;
    return node;
}

void spdExprTree(Node *node)
{
    SpdCursor   *csr;
    SpdRel  rel, *buf = NULL;
    int     i, buf_n = 0, expr_par = 0, blk_sort = 0;
    SpdString res = NULL;
    SpdObjId class_id, expr_id, label_id, op_id, blk_id;

    node->sort = smuGetExprSort(node->objectId);
    switch(node->sort){
    case EXPR_EMPTY:
        break;
    case EXPR_BREAK:
        res = spdStrCat(res, "break");
        node->str = res; 
        break;
    case EXPR_CONTINUE:
        blk_id = spdGetARelObj(node->objectId, "blk_expr", "any");
        while(1){
            blk_sort = spdGetAttrValInt(blk_id, "sort");
            if(blk_sort == BLOCK_FOR || blk_sort == BLOCK_DO ||
               blk_sort == BLOCK_WHILE) break;
            blk_id = spdGetARelObj(blk_id, "blk_blk", "c-id");
        }
        res = spdStrCat(res, "goto ");
        res = spdStrCat(res,  namedLabel(blk_id));

        node->str = res; 
        break;
    case EXPR_RETURN:
        res = spdStrCat(res, "return");
        if((expr_id = 
            spdGetARelObj(node->objectId, "expr_expr", "p-id")) != -1){
            res = spdStrCat(res, " ");
            node->child_n = 1;
            node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
            node->child[0] = mkNode(expr_id);
            spdExprTree(node->child[0]);
        }
        node->str = res; 
        break;
    case EXPR_GOTO:
        res = spdStrCat(res, "goto ");
        label_id = spdGetARelObj(node->objectId,"goto_label","any");
        res = spdStrCat(res, spdStrLabel(label_id));
        node->str = res; 
        break;
    case EXPR_SORT_FUNCCALL:              /* ؿƤӽФ */
        csr = spdGetRelInit(node->objectId, "expr_expr", "p-id");
        while ((rel = spdGetRel(csr)).relationId != -1) {
            buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
            buf[buf_n++] = rel;
        }
        spdFreeCursor(csr);
        SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
    
        node->child_n = buf_n;
        node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
        for(i=0; i< node->child_n; i++){
            node->child[i] = mkNode(buf[i].objectId);
            spdExprTree(node->child[i]);
        }
        spdFree(buf);
        break;
    default:
        switch(node->sort){ 
        case EXPR_SORT_SIZEOF: 
            expr_id = spdGetARelObj(node->objectId, "expr_expr", "p-id");
            op_id = spdGetARelObj(node->objectId, "expr_op", "p-id");
            if(expr_id == SPD_NON_ID ) {     /* sizeof(type) */
                res = spdStrCat(res, spdStrOperator(op_id));
            }
            else {                              /* sizeof(expression) */
                res = spdStrCat(res, spdStrOperator(op_id));
                /*  res = spdStrCat(res, "("); */
                node->child_n = 1;
                node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
                node->child[0] = mkNode(expr_id);
                spdExprTree(node->child[0]);
                /*  res = spdStrCat(res, ")"); */
            }
            node->str = res;
            break;
        case EXPR_SORT_ARRAY:                 /* []ΡɤȤʤ */
            csr = spdGetRelInit(node->objectId, "expr_expr", "p-id");
            while ((rel = spdGetRel(csr)).relationId != -1) {
                buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
                buf[buf_n++] = rel;
            }
            spdFreeCursor(csr);
            SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
    
            node->child_n = buf_n;
            node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
            for(i=0; i< node->child_n; i++){
                node->child[i] = mkNode(buf[i].objectId);
                spdExprTree(node->child[i]);
            }
            spdFree(buf);
            break;
        case EXPR_SORT_CAST:
            expr_id = spdGetARelObj(node->objectId, "expr_expr", "p-id");
            op_id = spdGetARelObj(node->objectId, "expr_op", "p-id");
            res = spdStrCat(res, spdStrOperator(op_id));
            node->child_n = 1;
            node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
            node->child[0] = mkNode(expr_id);
            spdExprTree(node->child[0]);
            node->str = res;
            break;
        case EXPR_SORT_LOGICAL:
        case EXPR_SORT_COMPARE:
        default:
            csr = spdGetRelInit(node->objectId, "expr_expr", "p-id");
            while ((rel = spdGetRel(csr)).relationId != -1) {
                buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
                buf[buf_n++] = rel;             
            }       
            spdFreeCursor(csr);
            csr = spdGetRelInit(node->objectId, "expr_op", "p-id");
            while ((rel = spdGetRel(csr)).relationId != -1) {
                buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
                buf[buf_n++] = rel;
                if(spdGetAttrValInt(rel.objectId, "sort") == OP_L_PAR)
                node->sort =  expr_par;
            }
            spdFreeCursor(csr);
            csr = spdGetRelInit(node->objectId, "const_ref", "p-id");
            while ((rel = spdGetRel(csr)).relationId != -1) {
                buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
                buf[buf_n++] = rel;
            }
            spdFreeCursor(csr);
            csr = spdGetRelInit(node->objectId, "ident_ref", "p-id");
            while ((rel = spdGetRel(csr)).relationId != -1) {
                buf = spdRealloc(buf, sizeof(SpdRel) * (buf_n + 1));
                buf[buf_n++] = rel;
            }
            spdFreeCursor(csr);
            SPD_QSORT(buf, buf_n, sizeof(SpdRel), cmp_offset);
    
            if(node->sort == expr_par){
                node->objectId = buf[1].objectId;
                spdExprTree(node);
            } else {
                node->child_n = buf_n;
                node->child = (Node **)spdMalloc(sizeof(Node *) * node->child_n);
                for(i=0; i<node->child_n; i++){
                    class_id = spdGetClassId(buf[i].objectId);
                    if(class_id == SPD_CID_EXPRESSION) {
                        node->child[i] = mkNode(buf[i].objectId);
                        spdExprTree(node->child[i]);
                    } 
                    else if(class_id == SPD_CID_OPERATOR) {
                        int op_sort;
                        op_sort = spdGetAttrValInt(buf[i].objectId, "sort");
                        switch (op_sort) {
                        case OP_MUL:
                        case OP_DIV:
                        case OP_MOD:
                        case OP_ADD:
                        case OP_SUB:
                        case OP_SHL:
                        case OP_SHR:
                        case OP_LT:
                        case OP_LE:
                        case OP_GT:
                        case OP_GE:
                        case OP_EQ:
                        case OP_NE:
                        case OP_B_AND:
                        case OP_B_XOR:
                        case OP_B_OR:
                        case OP_L_AND:
                        case OP_L_OR:
                        case OP_QUESTION:
                        case OP_COLON:
                        case OP_ASSIGN:
                        case OP_A_ADD:
                        case OP_A_SUB:
                        case OP_A_MUL:
                        case OP_A_DIV:
                        case OP_A_MOD:
                        case OP_A_AND:
                        case OP_A_XOR:
                        case OP_A_OR:
                        case OP_A_SHL:
                        case OP_A_SHR:
                            res = spdStrCat(res, " ");
                            break;
                        default:
                            break;
                        }
                    res = spdStrCat(res, spdStrOperator(buf[i].objectId));
                        switch (op_sort) {
                      case OP_PLUS:
                      case OP_MINUS:
                      case OP_MUL:
                      case OP_DIV:
                      case OP_MOD:
                      case OP_ADD:
                      case OP_SUB:
                      case OP_SHL:
                      case OP_SHR:
                      case OP_LT:
                      case OP_LE:
                      case OP_GT:
                      case OP_GE:
                      case OP_EQ:
                      case OP_NE:
                      case OP_B_AND:
                      case OP_B_XOR:
                      case OP_B_OR:
                      case OP_L_AND:
                      case OP_L_OR:
                      case OP_QUESTION:
                      case OP_COLON:
                        case OP_ASSIGN:
                      case OP_A_ADD:
                      case OP_A_SUB:
                      case OP_A_MUL:
                      case OP_A_DIV:
                      case OP_A_MOD:
                      case OP_A_AND:
                      case OP_A_XOR:
                      case OP_A_OR:
                      case OP_A_SHL:
                      case OP_A_SHR:
                      case OP_COMMA:
                        case OP_SIZEOF:
                            res = spdStrCat(res, " ");
                            break;
                        default:
                            break;
                        }
                        if(op_sort == OP_ASSIGN || op_sort == OP_A_ADD ||
                           op_sort == OP_A_SUB || op_sort == OP_A_MUL ||
                           op_sort == OP_A_DIV || op_sort == OP_A_MOD ||
                           op_sort == OP_A_AND || op_sort == OP_A_XOR ||
                           op_sort == OP_A_OR  ||  op_sort == OP_A_SHL ||
                           op_sort == OP_A_SHR ){ 
                            node->sort = EXPR_ASSIGN;
                            node->child[i] = mkNode(buf[i].objectId);
                            node->child[i]->str = res;
                        }else if(op_sort == OP_INC || op_sort == OP_DEC){
                            if(i == 0) node->sort = EXPR_SINGLE_L_ARITH;
                            else node->sort = EXPR_SINGLE_R_ARITH;
                            node->child[i] = mkNode(buf[i].objectId);
                            node->child[i]->str = res;
                        } else {
                            node->child[i] = mkNode(buf[i].objectId);
                            node->child[i]->str = res;
                            res = NULL;
                        }
                    }
                    else if(class_id == SPD_CID_CONSTANT) {
                        res = spdStrCat(res, spdStrConstant(buf[i].objectId));
                        spdFree(node->child);
                        node->child_n = 0;
                        node->str = res; 
                    }
                    else if(class_id == SPD_CID_IDENTIFIER) {
                        res = spdStrCat(res, spdStrIdentifier(buf[i].objectId));
                        spdFree(node->child);
                        node->child_n = 0;
                        node->str = res; 
                    }
                }
            }  
            spdFree(buf);
            break;  
        }
        break;
    }
}
