/* 
 * Program:             $RCSfile: spdMkFCDGIdxSub.c,v $  $Revision: 41.5 $
 * 
 * Purpose:             
 *
 * Author:              N.Atsumi  2001/11/29
 *                      S.Yamamoto  2001/11/29 
 *
 * (C) Copyright:       N.Atsumi and S.Yamamoto  2001 - 2002 
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
"$Id: spdMkFCDGIdxSub.c,v 41.5 2003/06/03 15:28:48 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#define DEFAULT_FCDG_SIZE 6

typedef struct _FCDG_NODE {
    SpdObjId obj;
    SpdObjIdArray dep_array;
} FCDG_NODE;

typedef struct _FCDG {
    FCDG_NODE *node;
    int size;
} FCDG;

static void put_node(SpdObjId node, SpdSet output_buf, FCDG *fcdg);
static void put_dep_no(SpdObjId node, int no, FCDG *fcdg);

int verbose_flag = 0;  /* 1:verbose output */
int debug_flag = 0;    /* 1:debug output */
SpdString fcdg_file;
SpdString spec_dir;
SpdString write_mode;
FILE *fcdg_idx;

static void putObjName(SpdObjId node)
{
    SpdObjId lib_id;
    SpdString lib_name;

    lib_id = spdGetARelObj(node, "fcdg_lib", "p-id");
    lib_name = spdGetName(lib_id);
    if (lib_name == NULL) {
        lib_name = spdStrdup("(null)");
        if (debug_flag) {
            spdError("Can't get Name of library function[%d]!\n", lib_id);
        }
    }
    fprintf(fcdg_idx, "%s", lib_name);
}

static void putOpStr(SpdObjId op)
{
    int sort;

    sort = spdGetAttrValIntOfObj(op, "sort");
    switch (sort) {
      case OP_NOT:
        fprintf(fcdg_idx, "!");
        break;
      case OP_LT:
        fprintf(fcdg_idx, "<");
        break;
      case OP_LE:
        fprintf(fcdg_idx, "<=");
        break;
      case OP_GT:
        fprintf(fcdg_idx, ">");
        break;
      case OP_GE:
        fprintf(fcdg_idx, ">=");
        break;
      case OP_EQ:
        fprintf(fcdg_idx, "==");
        break;
      case OP_NE:
        fprintf(fcdg_idx, "!=");
        break;
      default:
        if (verbose_flag) {
            spdWarning("Operator [%d] is not supported!\n", op);
        }
        break;
    }
}

static void putValueStr(SpdObjId val)
{
    SpdObjId p_id, m_id, mn_id;
    SpdOcc occ;
    SpdString mname;

    occ = spdGetAnOffsetOfObj(val);
    p_id = spdGetABlockId(occ.fileId, occ.offset);
    m_id = spdGetARelObj(p_id, "call_mcall", "p-id");
    mn_id = spdGetARelObj(m_id, "mcall_mname", "p-id");
    if (mn_id != SPD_NON_ID)
        mname = spdGetName(mn_id);
    else
        mname = spdGetAttrValString(val, "value");
    if (mname == NULL) {
        mname = spdStrdup("(null)");
        if (debug_flag) {
            spdError("Can't get Name of Constant(Macro)[%d]!\n", val);
        }
    }
    fprintf(fcdg_idx, "%s", mname);
}

static void putCondStr(SpdObjId node)
{
    SpdObjId op, val;

    op = spdGetARelObj(node, "cond_op", "p-id");
    putOpStr(op);
    fprintf(fcdg_idx, "\t");
    val = spdGetARelObj(node, "cond_val", "p-id");
    putValueStr(val);
}

static FCDG *fcdg_init(void)
{
    int i;
    FCDG *fcdg;
    fcdg = spdMalloc(sizeof(FCDG));
    fcdg->node = spdMalloc(sizeof(FCDG_NODE) * DEFAULT_FCDG_SIZE);
    fcdg->size = 0;
    for (i = 0; i < DEFAULT_FCDG_SIZE; i++) {
        fcdg->node[i].obj = SPD_NON_ID;
    }
    return fcdg;
}

static int fcdgCompareOffset(const SpdObjId *a, const SpdObjId *b)
{
    SpdObjId expr_a, expr_b;
    SpdOcc occ_a, occ_b;

    expr_a = spdGetARelObj(*a, "fcdg_expr", "p-id");
    expr_b = spdGetARelObj(*b, "fcdg_expr", "p-id");
    occ_a = spdGetAnOffsetOfObj(expr_a);
    occ_b = spdGetAnOffsetOfObj(expr_b);
    return (occ_a.offset - occ_b.offset);
}

static void regist_fcdg(SpdObjId node, FCDG *fcdg)
{
    SpdObjIdArray dep_array;
    int i;

    fcdg->node[fcdg->size].obj = node;
    fcdg->node[fcdg->size].dep_array =
        spdGetRelObjSortedArray(node, "libdata_depend", "p-id", NULL,
                                fcdgCompareOffset);
    dep_array = fcdg->node[fcdg->size].dep_array;
    if (!dep_array.size) {
        spdFreeObjIdArray(dep_array);
    } else {
        fcdg->size++;
        for (i = 0; i < dep_array.size; i++) {
            regist_fcdg(dep_array.id[i], fcdg);
        }
        if (!(fcdg->size % DEFAULT_FCDG_SIZE)) {
            spdRealloc(fcdg->node, sizeof(FCDG_NODE) * DEFAULT_FCDG_SIZE *
                       (fcdg->size / DEFAULT_FCDG_SIZE + 1));
        }
    }
}

static void fcdg_free(FCDG *fcdg)
{
    int i;
    for (i = 0; i < fcdg->size; i++) {
        spdFreeObjIdArray(fcdg->node[i].dep_array);
    }
    spdFree(fcdg->node);
}

static void putCtrlNode(SpdObjId node, SpdSet output_buf, FCDG *fcdg)
{
    SpdObjIdArray ctrl_array;
    int i;

    ctrl_array = spdGetRelObjSortedArray(node, "libctrl_depend", "p-id",
                                         NULL, fcdgCompareOffset);
    for (i = 0; i < ctrl_array.size; i++) {
        if (!spdSetContains(output_buf, ctrl_array.id[i])) {
            put_dep_no(ctrl_array.id[i], 0, fcdg);
            put_node(ctrl_array.id[i], output_buf, fcdg);
        }
    }
    spdFreeObjIdArray(ctrl_array);
}

static void put_node(SpdObjId node, SpdSet output_buf, FCDG *fcdg)
{
    int sort, i;
    for (i = 0; i < fcdg->size; i++) {
        if (node == fcdg->node[i].obj) {
            fprintf(fcdg_idx, "$%d:", i);
        }
    }
    sort = spdGetAttrValIntOfObj(node, "sort");
    if (spdSetContains(output_buf, node)) return;
    spdAddElem(output_buf, node, &node);
    if (sort == FCDG_CONDITION) {
        fprintf(fcdg_idx, "[\t");
        putCondStr(node);
        fprintf(fcdg_idx, "\t]\t{\t");
        putCtrlNode(node, output_buf, fcdg);
        fprintf(fcdg_idx, "\t}\t");
    } else {
        putObjName(node);
        fprintf(fcdg_idx, "\t");
    }
}

static void put_dep_no(SpdObjId node, int no, FCDG *fcdg)
{
    int i, j;

    if (!no) no = fcdg->size;
    for (i = 0; i <= no; i++) {
        for (j = 0; j < fcdg->node[i].dep_array.size; j++) {
            if (node == fcdg->node[i].dep_array.id[j]) {
                fprintf(fcdg_idx, "$%d;", i);
                break;
            }
        }
    }
}

static void dump_dep(FCDG *fcdg, SpdObjId node, int no, SpdSet output_buf)
{
    int i, j;
    if (!spdSetContains(output_buf, node)) {
        put_dep_no(node, no, fcdg);
        put_node(node, output_buf, fcdg);
        for (i = 0; i < fcdg->size; i++) {
            if (fcdg->node[i].obj == node) {
                for (j = 0; j < fcdg->node[i].dep_array.size; j++) {
                    dump_dep(fcdg, fcdg->node[i].dep_array.id[j], i, output_buf);
                }
                break;
            }
        }
    }
}

static void dump_fcdg(FCDG *fcdg)
{
    int i, j;
    SpdSet output_buf;
    SpdObjId var_id, var_expr;
    SpdOcc var_occ;
    SpdRegion var_i_reg, var_p_reg;
    output_buf = spdCreateSet(sizeof(SpdObjId));
    for (i = 0; i < fcdg->size; i++) {
        if (!spdSetContains(output_buf, fcdg->node[i].obj)) {
            var_expr = spdGetARelObj(fcdg->node[i].obj, "def_var", "p-id");
            var_id = spdGetARelObj(var_expr, "ident_ref", "p-id");
            var_occ = spdGetAnOffsetOfObj(var_expr);
            var_i_reg = spdConvOccToRegion(var_occ);
            var_p_reg = spdConvRegionItoCbyPIDB(var_i_reg);
            fprintf(fcdg_idx, "NAME:%d_%d\n", var_id, var_p_reg.offset);
            if (debug_flag) {
                if ((var_id == SPD_NON_ID) ||
                    (var_p_reg.offset == SPD_NON_OFFSET)) {
                    spdWarning("Can't get Offset!\n"\
                               "DEF NODE  ID : %d\n"\
                               "DEF EXPR  ID : %d\n"\
                               "DEF IDENT ID : %d\n"\
                               "DEF  OFFSET  : %d\n", fcdg->node[i].obj,
                               var_expr, var_id, var_p_reg.offset);
                }
            }
            put_node(fcdg->node[i].obj, output_buf, fcdg);
        }
        for (j = 0; j < fcdg->node[i].dep_array.size; j++) {
            dump_dep(fcdg, fcdg->node[i].dep_array.id[j], i, output_buf);
        }
    }
    spdFreeSet(output_buf);
}

static void get_fcdg(SpdObjId top)
{
    FCDG *fcdg;
    fcdg = fcdg_init();
    regist_fcdg(top, fcdg);
    dump_fcdg(fcdg);
    fcdg_free(fcdg);
}

static void analyze_func(SpdObjId func)
{
    SpdCursor *fcdg_csr;
    SpdObjId top_id;
    SpdSet node_set;

    fcdg_csr = spdGetRelObjInit(func, "func_fcdg", "p-id");
    while((top_id = spdGetRelObj(fcdg_csr)) != SPD_NON_ID) {
        node_set = spdCreateSet(sizeof(SpdObjId));
        get_fcdg(top_id);
        fprintf(fcdg_idx, "\n");
        spdFreeSet(node_set);
    }
    spdFreeCursor(fcdg_csr);
}

static void getFCDG(SpdObjId prog)
{
    SpdCursor *file_csr;
    SpdObjId file_id;

    file_csr = spdGetRelObjInit(prog, "prog_file", "p-id");
    while ((file_id = spdGetRelObj(file_csr)) != SPD_NON_ID) {
        SpdCursor   *decl_csr;
        SpdObjId    decl_id;
        SpdSet      func_set;
        SpdObjId p_file, file;
        int         i;
        
        func_set = spdCreateSet(sizeof(SpdObjId));
        decl_csr = spdGetRelObjInit(file_id, "file_decl", "p-id");
        while ((decl_id = spdGetRelObj(decl_csr)) != SPD_NON_ID) {
            int     sort;

            sort = spdGetAttrValIntOfObj(decl_id, "sort");
            if (sort == DECL_FUNCDEF) {
                spdAddElem(func_set, decl_id, &decl_id);
            }
        }
        spdFreeCursor(decl_csr);
        if (func_set->elemNum) {
            p_file = spdConvFileIdItoC(file_id);
            file = spdConvIdPfileToFile(p_file);
            fprintf(fcdg_idx, "SPEC:%s/%d/%d\n", spec_dir, prog, file);
        }
        for (i = 0; i < func_set->elemNum; i++) {
            SpdObjId    func_id;

            func_id = *(SpdObjId *)spdGetAElem(func_set, NULL);
            analyze_func(func_id);
        }
        spdFreeSet(func_set);
    }
    spdFreeCursor(file_csr);
}

void opt_analyze(int argc, char *argv[])
{
    int     c;

    while ((c = getopt(argc, argv, "d:f:s:Dvw:")) != EOF) {
        switch (c) {
          case 'd':                         /* Directory for SDB. */
            spdSetSDBdir(optarg);
            break;
          case 'f':
            fcdg_file = spdStrdup(optarg);
            break;
          case 's':
            spec_dir = spdStrdup(optarg);
            break;
          case 'D':
            debug_flag ^= 1;
            break;
          case 'v':
            verbose_flag ^= 1;
            break;
          case 'w':
            write_mode = spdStrdup(optarg);
            break;
          default:
            spdErrorExit("'%s': Invalid option.\n", argv[optind-1]);
            break;
        }
    }
}

int main(int argc, char *argv[])
{
    SpdConstString prog_name;
    SpdModelId fcd_id, i_id, p_id;
    SpdDBId db_id;
    SpdCursor *prog_csr;
    SpdObjId prog_id;

    prog_name = spdBasename(argv[0]);
    spdInitialize(prog_name);
    opt_analyze(argc, argv);

    i_id = spdUseModel(&I_model);
    p_id = spdUseModel(&P_model);
    fcd_id = spdUseModel(&FCD_view);
    db_id = spdOpenSDB();
    spdReadDB(i_id, db_id);
    spdReadDB(p_id, db_id);
    spdReadDB(fcd_id, db_id);

    prog_csr = spdGetObjIdInit("program", NULL);

    fcdg_idx = spdFopen(fcdg_file, write_mode);
    spdFree(fcdg_file);
    spdFree(write_mode);
    while ((prog_id = spdGetObjId(prog_csr)) != SPD_NON_ID) {
        fprintf(fcdg_idx, "PROG:%s\n", spdGetName(prog_id));
        getFCDG(prog_id);
    }
    spdFreeCursor(prog_csr);

    spdFree(spec_dir);
    spdFclose(fcdg_idx);
    spdCloseSDB(db_id);
    return (EXIT_SUCCESS);
}
