#! PerlCmd
#
# Program:              $RCSfile: spdSed.pl,v $  $Revision: 30.10 $
# 
# Purpose:
# 
# Author:               T.Suzuki  1999/03/08
#                       S.Yamamoto  1999/03/08
# 
# (C) Copyright:        T.Suzuki and S.Yamamoto  1999 - 2002
#                       This file is a product of the project Sapid.
# 

# 
# $Id: spdSed.pl,v 30.10 2002/12/29 14:59:34 yamamoto Exp yamamoto $
# 

#
# spdSed -f scriptFile [-o outputFile] [inputFile] 
#

# Enhanced script format:
# (1) In each command line, <offset> --> <file>:<offset> available.
# (2) For filename table,
#     # <fileName> ==> <replacedFileName>

$tmpdir = "/tmp";

%strs = ();             # <FileName> => string
%freplace = ();         # <fileName> => <replacedFileName>

use Getopt::Std;

# getopts('Rf:o:v');    # Option "R" enable file replace mode.
# getopts('e:f:o:v');
getopts("f:o:v");
if (! -e $opt_f) {
    print "\"$opt_f\": Can't open script file.\n";
    exit;
}

open(SCR, "$opt_f") || die "\"$opt_f\": Can't open script file.";

$cfile = "";

while (<SCR>) {
    if ($_ =~ /^\# (\S+) ==> (\S+)/) {
        # Set filename replace table.
        if ($opt_v) { print "Debug: input (==>)    : $1, $2\n"; }
        $freplace{$1} = $2;
        next;
    } elsif ($_ =~ /^\#.*/) {
        # Skip comment.
        next;
    } elsif ($_ =~ /^[ADMR] (\d+|[^:]*:\d+) (.*)/) {
        # Command line (should check string length).
        $cfile = '';

        $offs = $&;
        $tail = $';
        if ($offs =~ s/\s([^:]+):/ /) {	# Input file name specified.
            $cfile = $1;
            $_ = $offs . $tail;
        }           
        if ($opt_v) { print "Debug: input (command): $1, $_"; }
    }

    $strs{$cfile} .= $_;
}

# print %strs;
# exit 0;

while (($fnm, $str) = each(%strs)) {
    # First, check replace table.
    # If not exist, try to create new file named "*.gen" in default.
    #
    # However -R flag (replace file) is available and specified,
    # try to rename old file to "*.orig" and replace same name file.

    if ($opt_v) { print "Debug: fnm \"$fnm\", str $str"; }
    
    if ($freplace{$fnm} ne '') {
        # exist in replace table
        $nfn = $freplace{$fnm};
        $ofn = $fnm;
    } elsif (!$opt_R) {
        # New File Generate mode (default).
        $nfn = "$fnm.gen";
        $ofn = $fnm;
    } else {
        # Replace mode.
        rename ($fnm, "$fnm.orig");
        $nfn = $fnm;
        $ofn = "$fnm.orig";
    }

    # Temporarily script file.
    $tmpfnm = "$tmpdir/spdSed-$fnm";

    open(TM, ">$tmpfnm") || die print "Can't open $tmpfnm.\n";
    print TM $str;
    close(TM);

    $ofn = '/dev/null' if (! -e $ofn);

    if ($fnm eq '') {
        # Original behavior:
        #     Input file is from cmd-line (no concern when not specfied),
        #     Output file is -o when specfied (else maybe stdout).
        $opts  = "-f $tmpfnm";
        $opts .= " -o $opt_o" if ($opt_o ne '');
        $opts .= join(" ", @ARGV);
        if ($opt_v) { print "spdSedSub $opts\n"; }
        `spdSedSub $opts`;
    } else {
        if ($opt_v) { print "spdSedSub -f $tmpfnm -o $nfn $ofn\n"; }
        `spdSedSub -f $tmpfnm -o $nfn $ofn`;
    }
    
    # Delete temporary script file.
    unlink($tmpfnm);
}

exit 0;
