/* 
 * Program:             $RCSfile: spdStrTester.c,v $  $Revision: 31.14 $
 * 
 * Purpose:
 * 
 * Author:              N.Fukuyasu, S.Kawai and Y.Momiyama  1995/12/07
 *                      N.Ito  1997/08/29
 *                      S.Yamamoto  1995/12/08
 * 
 * (C) Copyright:       N.Fukuyasu, S.Kawai, Y.Momiyama, N.Ito
 *                      and S.Yamamoto  1995 - 2001
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdStrTester.c,v 31.14 2003/06/28 15:34:08 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

#ifdef SAPID_USE_SPDSTR3
#include <Sapid/SpdStr3.h>
#else
#include <Sapid/SpdStr.h>
#endif

#define DEFAULT_MODE    SAPID_C_MODE
#define ERROR_MESSAGE \
        "Usage: %s [-ic][-d dir][-q] fileName\n              %s [-ic][-d dir][-q] programName objectId\n"

#ifdef BUGGY_SPDSTR_0102
#define OPTIONS     "icd:mq"
#else
#define OPTIONS     "icd:q"
#endif

static void do_spdStr(SpdObjId id)
{
    SpdString   result;

    if ((result = spdStr(id)) != NULL) {
        puts(result);
    }
}

int main(int argc, char *argv[])
{
    SpdObjId    prog_id;
    int         c;                          /* Option character. */
    int         modeOption = SAPID_NON_OPTION; /* Saved spdOption's mode. */
    SpdCursor   *file_csr;
    SpdBoolean  foundFlag = FALSE;
    
    spdInitialize(spdBasename(argv[0]));

    while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
        switch (c) {
          case 'i':
            modeOption = SAPID_I_MODE;
            break;
          case 'c':
            modeOption = SAPID_C_MODE;
            break;

          case 'd':                         /* SDB directory. */
            spdSetSDBdir(optarg);
            break;

#ifdef BUGGY_SPDSTR_0102
          case 'm':
            spdSetOption("buggy-SpdStr-0102", ON);
            break;
#endif /* BUGGY_SPDSTR_0102 */

          case 'q':                         /* typeQualifierϤ롥 */
            spdSetOption("typeQualifier", ON);
            break;
          default:
            spdErrorExit(ERROR_MESSAGE, argv[0], argv[0]);
            break; 
        }
    }
    
    if (argc == optind + 1) {
        /* "*.c"뤤"*.i"Τߤꤵ줿硥 */
        SpdString   progName = NULL;
        SpdObjId    file_id;

        progName = smuGetAProgramName();
        if (spdIsFileExtension(argv[optind], ".i")) {
            if (modeOption == SAPID_C_MODE) {
                if ((prog_id = spdTarget("P-model", progName)) ==
                    SPD_NON_ID) {
                    spdErrorExit("Program \"%s\" not found.\n", progName);
                }
            } else {
                modeOption = SAPID_I_MODE;
                if ((prog_id = spdTarget("I-model", progName)) == 
                    SPD_NON_ID) {
                    spdErrorExit("Program \"%s\" not found.\n", progName);
                }
            }       
            spdSetOption("mode", modeOption);

            file_csr = spdGetRelObjInit(prog_id, "prog_file", "any");
            while ((file_id = spdGetRelObj(file_csr)) != SPD_NON_ID) {
                if (SPD_STREQ(spdGetName(file_id), argv[optind])) {
                    do_spdStr(file_id);
                    spdStrFree();
                    foundFlag = TRUE;
                    break;
                }
            }
            if (foundFlag) {
                spdFreeCursor(file_csr);
            }
        } else if (spdIsFileExtension(argv[optind], ".c")) {
            SpdObjId    cfile_id, ifile_id;

            if (modeOption == SAPID_I_MODE) {
                if ((prog_id = spdTarget("I-model", progName)) ==
                    SPD_NON_ID) {
                    spdErrorExit("Program \"%s\" not found.\n", progName);
                }
            } else {
                modeOption = SAPID_C_MODE;
                if ((prog_id = spdTarget("P-model", progName)) ==
                    SPD_NON_ID) {
                    spdErrorExit("Program \"%s\" not found.\n", progName);
                }
            }       
            spdSetOption("mode", modeOption);

            /* Added by N.Fukuyasu 1999/08/28. */
            /* ".c" Υե̾ ".i" ΥեID 롥 */
            cfile_id = smuGetFileIdByName(argv[optind]);
            ifile_id = spdConvFileIdCtoI(cfile_id);

            file_csr = spdGetRelObjInit(prog_id, "prog_file", "any");
            while ((file_id = spdGetRelObj(file_csr)) != SPD_NON_ID) {
                /* Modified by N.Fukuyasu 1999/08/28. */
                /* if (SPD_STREQ(spdGetName(file_id), fileName)) { */
                if (file_id == ifile_id) {
                    do_spdStr(file_id);
                    spdStrFree();
                    foundFlag = TRUE;
                    break;
                }
            }
            if (foundFlag) {
                spdFreeCursor(file_csr);
            }
        } else {
            spdErrorExit("Invalid file name: \"%s\".\n", argv[optind]);
        }
    } else if (argc == optind + 2) {
        /* ץ̾ȥֹ֥椬ꤵ줿硥 */
        if (modeOption == SAPID_NON_OPTION) {
            modeOption = DEFAULT_MODE;
        }
        prog_id = spdTarget((modeOption == SAPID_I_MODE ?
                             "I-model" : "P-model"), argv[optind]);
        if (prog_id ==  SPD_NON_ID) {
            spdErrorExit("Program \"%s\" not found.\n", argv[optind]);
        }
        spdSetOption("mode", modeOption);

        do_spdStr(atoi(argv[optind + 1]));
        spdStrFree();
    } else {
        spdErrorExit(ERROR_MESSAGE, argv[0], argv[0]);
    }

    return (EXIT_SUCCESS);
}
